/*
 *M*****************************************************************************
 *M*  Product:  XEROX OSG/OSCG/EIDC S2X Board                                  *
 *M*****************************************************************************
 *M*                                                                           *
 *M*                    COPYRIGHT (C) 2001 XEROX CORPORATION                   *
 *M*                    ALL RIGHTS RESERVED                                    *
 *M*                                                                           *
 *M*****************************************************************************
 *M*
 *M*    MODULE NAME  :  encImage.c
 *M*
 *M*    PURPOSE      :  JBIG2 compression
 *M*
 *M*    DESCRIPTION  :  processes a line of input image and performs JBIG
 *M*							compression
 *M*
 *******************************************************************************
 **     AUTHOR IDENTITY                                                        *
 *******************************************************************************
 **
 **  Initials  Name
 **  --------  ------------------------------
 **  mdd       Marlin Diehl
 **   nl       Nathan Lund
 **   rn       Ramesh Nagarajan
 **   pt       Phil Turkovitz
 **
 *******************************************************************************
 **     SVCS MODULE REVISION HISTORY:                                          *
 *******************************************************************************
 **  Revision    Date    Author  Description
 **  --------  --------  ------  ----------------------
 **  0.1			02.01.21  ??      
 *******************************************************************************
 */

/*****************/
/* Include files */
/*****************/
#define TYPE_bool /* hack to avoid conflict for bool type */
#include <stdio.h>
#include <stdlib.h>
//#include <time.h>
#include "jbig2.h"

// SILX headers
#include "dgp_err_handle.h"
#include "writejbigCB.h"
#include "jbig2apiCB.h"

#include "segment.h"
#include "encgenreg.h"
#include "stdtypes.h"

/*********************/
/* Constant #defines */
/*********************/


#define	PRINTF(x)		//printf x

/*******************************/
/* Declaration specific to PC  */
/*******************************/
#define PCVERSION
#ifdef PCVERSION
	#define  inline
	#define  restrict
	#define  if_prob(a)
	typedef  unsigned int	n32;
#endif


/*********************/
/* Global Variables  */
/*********************/
#define JB2HEADERSIZE 8
static Dgp_UInt8 jb2header[] = { 0x97, 0x4a, 0x42, 0x32, 0x0d, 0x0a, 0x1a, 0x0a };
static void encBufferAsImage(
             Dgp_State *dgp_state,
             void* cookie,
             Dgp_UInt8     *buffer,
             Dgp_Int32     bufWidth,
             Dgp_Int32     bufHeight,
             Dgp_Int32     bpl,
             Dgp_Int32     bufStartY);

/*F*****************************************************************************
 *F*
 *F*  Function Name:  jbigEncode
 *F*
 *F*  Operation:      entry point to JBIG2 encoding algorithm
 *F*
 *F*  Function
 *F*  Returns:        Nothing
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  02.01.21		??
 *F****************************************************************************/
int jbigEncode (uint8 *restrict imgptr, JBIG2State *restrict state, int rows)
{ 

	Dgp_Int32 ret = cSuccessful;
    Dgp_Bool noStripes = DGP_FALSE;
    Dgp_Int32 resolution = 300;
    Jb2State *jb2state = NULL;

	switch (state->processingState) {
	case initialize:    // initilaizing and processing the firstBand
        {
            Dgp_State* dgp_state = NULL;
            int i;

            /* Init dgp state */    
            ret = Dgp_initDgpState(&dgp_state, 
                    NULL,
                    NULL,
                    stderr, 
                    Dgp_fileErrPrinter, 
                    DGP_TRUE, 
                    DGP_TRUE);

            ret += createJb2StateAPICB (NULL, dgp_state, 2, 1, DGP_FALSE, &jb2state, NULL);

            // write jbig2header to buffer
            for(i = 0; i< JB2HEADERSIZE; i++){
                *(state->outCurAddress)= jb2header[i];
                state->outCurAddress++;
            }

 // start with Page Info segment
  writeJb2PageInfo( 	 state->pixels,
  						 state->scanlines,
  						 resolution,
  						 resolution,
						 jb2state,
						 &noStripes);

            /* Write 0x00: random-access organisation, page count known */
            /* Write 0x01: sequential organisation, page count known */
            /* Write 0x02: random-access organisation, page count not known */
            /* Write 0x03: sequential organisation, page count not known */
            *(state->outCurAddress) = 0x03;
            state->outCurAddress++;
            if ( ret  != cSuccessful){
		        if(dgp_state)	
                    Dgp_cleanupDgpState((Dgp_State **)(&dgp_state));
                return 0;
            }
            state->dgp_state  = (void* )dgp_state;
            state->enc_cookie = (void* )jb2state;
            state->compressedSize = 0;
		    PRINTF (("<jbigEncode> Initialize\n"));

        }
	    break;

	case process:
        {
            int lastBufInPage;
            //encode one buffer at a time
            if(state->bandNumber >= state->totalBandCnt)
                lastBufInPage = 1;
            else 
                lastBufInPage = 0;
            // encode one buffer
            encBufferAsImage(
                     state->dgp_state, 
                     state->enc_cookie, 
                     imgptr,            // input buffer
                     state->pixels,     // buffer width (same for  page width)
                     rows,              // buffer Height
                     state->bytesperline,
                     state->compressedSize);  // Y coord of starting position

            state->compressedSize += rows;
            // finish and get the data
            if(lastBufInPage)
            {
               uint8* dictData = NULL, *pgData = NULL;
               uint32 dictDataSize = 0, pgDataSize = 0;

               jb2state = (Jb2State*) state->enc_cookie;

               ret += finishJb2PageCB (jb2state, NULL);
               ret += getJb2CompressedData (jb2state, &pgData, &pgDataSize);

               if(pgDataSize!=0 && pgData){
                   memcpy(state->outCurAddress, pgData, pgDataSize);
                   state->outCurAddress += pgDataSize;
                   free(pgData);
               }            
            }
		    PRINTF (("<jbigEncode> Process\n"));
		    //compressOneRow (imgptr, state, rows); skip this function
        }
		break;
	
	case finish:
        //clean up 
        closeJb2StateAPICB(state->enc_cookie, NULL);
        Dgp_cleanupDgpState((Dgp_State**)(&(state->dgp_state)));

		//Any completion work
		PRINTF (("<jbigEncode> Finish\n"));
		break;
	}

	return 1;

}


static void
encBufferAsImage(
             Dgp_State *dgp_state,
             void* cookie,
             Dgp_UInt8     *buffer,
             Dgp_Int32     bufWidth,
             Dgp_Int32     bufHeight,
             Dgp_Int32     bpl,
             Dgp_Int32     bufStartY)
{

    Jb2State *jb2state = (Jb2State*) cookie;
	jb2FileParams *params = jb2GetFileParams ((jb2File) jb2state->jb2f);

	jb2Segment regseg;
	DgpImage_Packed resid = NULL;
	Dgp_Int32 longsperline, i;


	regseg.segtype = JB2SEGTYPE_GENREG_IMM;
	regseg.pagenum = 1; 
	regseg.nrefsegs = 0;
	regseg.refsegs = (Dgp_Int32 *)NULL;
	regseg.refretain = (Dgp_Bool *)NULL;

    regseg.reginfo.regwidth = bufWidth;
	regseg.reginfo.regheight = bufHeight;
	regseg.reginfo.regxpos = 0;
    if (bufStartY == 0)
    	regseg.reginfo.regypos = 0;
    else
    	regseg.reginfo.regypos = bufStartY - 1;


	longsperline = bpl / 4;
    if (bpl % 4) fprintf (stderr, "buffer length problem !\n");
	DgpImage_wrap(dgp_state, (Dgp_UInt32 *) buffer, 
		bufWidth, bufHeight, longsperline, &resid); 

    regseg.reginfo.combop = COMBOP_OR;
	regseg.reginfo.rendered = DGP_TRUE;
	regseg.reginfo.im = resid;
	if (params->sdarith)
	{
		// arithmetic coding
	    regseg.data.genreg.mmr = DGP_FALSE;
	    regseg.data.genreg.abparams.rtemplate = params->sdabparams.rtemplate;
    	for (i = 0; i < 3; i++)
	    {
    	    regseg.data.genreg.abparams.atx[i] = params->sdabparams.atx[i];
    	    regseg.data.genreg.abparams.aty[i] = params->sdabparams.aty[i];
	    }
    	regseg.data.genreg.abparams.tpon = params->sdabparams.tpon;
    }
    else
        regseg.data.genreg.mmr = DGP_TRUE;

	regseg.data.genreg.data = (Dgp_UInt8 *)NULL;
	regseg.data.genreg.datalen = 0;

	jb2EncodeGenRegion((jb2File)jb2state->jb2f, 
		&regseg);

    return;
}
