/*
 *M*****************************************************************************
 *M*  Product:  XEROX OSG/OSCG/EIDC S2X Board                                  *
 *M*****************************************************************************
 *M*                                                                           *
 *M*                    COPYRIGHT (C) 2001 XEROX CORPORATION                   *
 *M*                    ALL RIGHTS RESERVED                                    *
 *M*                                                                           *
 *M*****************************************************************************
 *M*
 *M*    MODULE NAME  :  jbigtsk.c
 *M*
 *M*    PURPOSE      :  JBIG2 Task 
 *M*
 *M*    DESCRIPTION  :  Handles all the task related stuff for JBIG2 Encoding 
 *M*                    
 *M*
 *******************************************************************************
 **     AUTHOR IDENTITY                                                        *
 *******************************************************************************
 **
 **  Initials  Name
 **  --------  ------------------------------
 **  mdd       Marlin Diehl
 **   nl       Nathan Lund
 **   rn       Ramesh Nagarajan
 **   pt       Phil Turkovitz
 **
 *******************************************************************************
 **     SVCS MODULE REVISION HISTORY:                                          *
 *******************************************************************************
 **  Revision    Date    Author  Description
 **  --------  --------  ------  ----------------------
 **  0.1  		02.01.21  rn		Template Task for JBIG2 Encoding
 *******************************************************************************
 */

/*****************/
/* Include files */
/*****************/
//#include <eti/ds.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <mm.h>
#include "jbig2.h"
#include "stdtypes.h"    /* type definitions for this product */
#ifndef FILE_READ
	#include "videoio.h"
#endif

/*********************/
/* Constant #defines */
/*********************/
#define JBIG2_STATE_IDLE        0
#define JBIG2_STATE_OPEN        1
#define JBIG2_STATE_WORK        2
#define JBIG2_STATE_CLOSE       3
#define JBIG2_STATE_ABORT       4

#define JBIG2_BUSY						1
#define JBIG2_PIXEL_RANGE_ERROR		2
#define JBIG2_SCANLINE_RANGE_ERROR		3

#define VIDEOFORMAT     DIAGONAL        // LIVEVIDEO or DIAGONAL or COUNTER

//#define DIAG

 
/*********************/
/* Macro definitions */
/*********************/
#define PRINTF(x) //printf x
 
/************/
/* Typedefs */
/************/
typedef unsigned char bit;
 
/*************************/
/* Structure definitions */
/*************************/
 
/*****************************************/
/* External data accessed by this module */
/*****************************************/
extern FILE *ifile;
 
/*********************************************/
/* External function accessed by this module */
/*********************************************/
extern int jbigEncode (bit *, JBIG2State *, int);
 
/***********************************/
/* Data promoted to a global level */
/***********************************/
bit  outimage[(7200*5100*3)/8];

/***********************/
/* Local data (static) */
/***********************/
static int16		jbigState;
static JBIG2State	state;
static uint32		eTime, time656;
 
/****************************************/
/* Local function declarations (static) */
/****************************************/
static void jbigOpen(void);
static bool jbigDoBand(void);
static void jbigClose(void);

/***************************************************************************/

/*******************************/
/* Global function definitions */
/*******************************/    

/*F*****************************************************************************
 *F*
 *F*  Function Name:  jbigTaskInit
 *F*
 *F*  Operation:      Powerup initialization.
 *F*
 *F*  Function
 *F*  Returns:        Nothing
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  01.08.06  mdd     Creation
 *F****************************************************************************/
/*F*/ void jbigTaskInit(void) 
{
	PRINTF(("<jbigTaskInit>\n"));
	jbigState = JBIG2_STATE_IDLE;
}



/*F*****************************************************************************
 *F*
 *F*  Function Name:  g4ValidateRequest
 *F*
 *F*  Operation:      Confirm validity of parent's request and
 *F*                  confirm that task isn't already busy.
 *F*
 *F*  Function
 *F*  Returns:        TRUE= request accepted, FALSE= request denied
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  01.08.06  mdd     Create wrapper
 *F****************************************************************************/
/*F*/ bool jbigValidateRequest(
/*F*/         uint32 pixels,    /* input image size in fastscan direction */
/*F*/         uint32 scanlines, /* input image size in slowscan direction */
/*F*/         uint32 *errnum )  /* addr to receive code for valDenyType */
{
	PRINTF(("<jbigValidateRequest>\n"));

	/* if task is busy , return nonzero error code */
	if( jbigState != JBIG2_STATE_IDLE )
	{
		*errnum = JBIG2_BUSY;
		return( FALSE );
	}

	/* if any input parameter is out of range , return nonzero valDenyType */
	/* FIXME; add code to check min/max range of pixels & scanlines input */

	/* OK to start processing new image */
    if ((pixels % 8)) {
		*errnum = JBIG2_PIXEL_RANGE_ERROR;
		return (FALSE);
	}
	else {
		state.pixels    = pixels;
		state.scanlines = scanlines;
		jbigState = JBIG2_STATE_OPEN;
		*errnum = 0;
		return (TRUE);
	}
}

/*F*****************************************************************************
 *F*
 *F*  Function Name:  jbigTask
 *F*
 *F*  Operation:      An OS task to
 *F*                  get raster data from the scanner input queue and
 *F*                  apply G3 compression to the raster data and
 *F*                  put the compressed data to the host output queue.
 *F*                  .
 *F*
 *F*  Function
 *F*  Returns:        Nothing
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  01.08.02  mdd     Creation
 *F****************************************************************************/
/*F*/  bool jbigTask(void)
{
  bool contFlag = TRUE;

  if( jbigState == JBIG2_STATE_IDLE ) return (contFlag);

  switch (jbigState)
  {  
     case JBIG2_STATE_WORK:
          if( !jbigDoBand() )
          jbigState = JBIG2_STATE_CLOSE;
          break;

     case JBIG2_STATE_OPEN:
          jbigOpen();
          jbigState = JBIG2_STATE_WORK;
          break;

     case JBIG2_STATE_CLOSE:
          jbigClose();
		  contFlag = FALSE;
          //jbigState = JBIG2_STATE_IDLE;
          break;

     default:
       PRINTF(("illegal jbigState found"));
       break;
  }

  return (contFlag);
}


/***************************************************************************/

/***************************************/
/* Local function definitions (static) */
/***************************************/
static bool readPBMFile(uint8 *inBuffer, int32 bufHeight, int32 pixels, int32 bytesperline,FILE* ifile)
{
    int i;
    uint8 *p;
    uint32 *q;
    if(pixels % 32){
        uint8 tempBuffer[7200*JBIG2_STRIPSPERBAND/8];
	    fread (tempBuffer, bufHeight * pixels / 8, 1, ifile);
        for(i = 0; i < bufHeight; i++){
            uint8 *inptr = inBuffer + bytesperline * i;
            memcpy(inptr, tempBuffer + i * pixels / 8,  pixels / 8);
        }
    }else{
	    fread (inBuffer,  bufHeight * bytesperline, 1, ifile);
    }
    for (i = 0, p = inBuffer, q = (uint32 *)inBuffer;   i < bufHeight; i++) {
        int32 j;
        for (j = 0; j < bytesperline; j += sizeof(uint32)) {
            uint32 tmp = 0;
            int32 ix;
            for (ix = 0; ix < (int)sizeof(uint32); ix++) {
                tmp |= (*p++) << (32 - (ix + 1) * 8);
            }
            *q++ = tmp;
        }
    }
   return TRUE;
}

/*F*****************************************************************************
 *F*
 *F*  Function Name:  jbigOpen
 *F*
 *F*  Operation:      Init the task to process a  new image.
 *F*
 *F*  Function
 *F*  Returns:        Nothing
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  01.08.02  mdd     Create wrapper
 *F****************************************************************************/
/*F*/
 static void jbigOpen(void) 
{
#ifndef FILE_READ
	SCODE sc;
#endif

  PRINTF(("<jbigOpen>\n"));
#ifdef DIAG
  printf ("<g4ValidateRequest> Diagonal Line TestPattern\n");
  state.pixels    	= state.pixels*8;
#endif
  state.bytesperline    = (state.pixels + 31)/32 * 4;
  state.outCurAddress   = outimage;

  state.bufferId = 0;

  // Number of scanlines input may not be integer multiple of JBIG2_LINESPERSTRIP.
  // Two cases for scanlines in remainder of scanlines/JBIG2_LINESPERSTRIP division.
  // Rounding up requires that last strip be padded with scanlines of zerofill.
  // Truncation make the video path lose those last few lines of input image.
  // For roundup of trailing scanlines into additional strip:
  state.totalStripCnt = state.scanlines;
  state.lastStripCnt  = 0;

  // Roundup trailing strips into last band that may be only partially filled.
  state.totalBandCnt = (state.totalStripCnt)/JBIG2_STRIPSPERBAND + 1;
  state.lastBandCnt  = state.totalStripCnt % JBIG2_STRIPSPERBAND;
  if (state.lastBandCnt == 0) {		// Always end with the lastBand for completion
	  state.lastBandCnt = JBIG2_STRIPSPERBAND;
	  state.totalBandCnt--;
  }

  state.stripNumber = 0;
  state.bandNumber = 0;
  state.processingState = initialize;
  eTime = 0; time656 = 0;

  //---fixme  hmpv_write_fcnt(0);
  //---fixme move the read to jbigDoBand
#ifdef FILE_READ
//    fread (&state.inpBuffer[state.bufferId][0], JBIG2_STRIPSPERBAND*state.bytesperline, 1, ifile);
  readPBMFile(&state.inpBuffer[state.bufferId][0], JBIG2_STRIPSPERBAND, state.pixels, state.bytesperline, ifile);
#else
  //Getting Data from SPA FPGA, 656 port initialization stuff...
  //DsKick to start consumer and spaResetToggle to start producer

  //**** FIXME LATER.... currently input image is color, hence divide by 3
  //***** Faking color 24 bpp image as binary
  sc = VideoioInOpen (&state.inpBuffer[0][0], &state.inpBuffer[1][0],
                (state.bytesperline*JBIG2_STRIPSPERBAND),
                state.pixels/8, state.scanlines/3, VIDEOFORMAT);
   if (sc != 0) {
        printf ("<ptOpen> VideoioInOpen return sc = 0x%x\n", sc);
        exit (-1);
   }

  state.inpBuffer[0][0] = 0x00;
  state.inpBuffer[0][1] = 0xf0; 
#endif

//----fixme  time656 += hmpv_read_fcnt();
//----fixme  hmpv_write_fcnt(0);

  // Initialize and encode the first strip
  jbigEncode (&state.inpBuffer[state.bufferId][0], &state, JBIG2_STRIPSPERBAND);
}

/*F*****************************************************************************
 *F*
 *F*  Function Name:  jbigDoBand
 *F*
 *F*  Operation:      Do a band of image processing then return control
 *F*                  to scheduler.  The smallest unit of work is a STRIP.
 *F*                  The number of STRIPS processed in a  band is given by
 *F*                  JBIG2_STRIPSPERBAND.
 *F*
 *F*                  The constant JBIG2_STRIPSPERBAND can be tuned/adjusted 
 *F*                  up/down to tradeoff time alloted to this task against
 *F*                  time alloted to other tasks.
 *F*
 *F*                  The constant JBIG2_LINESPERSTRIP is fixed by the
 *F*                  implementation of the compression algorithm.
 *F*
 *F*  Function
 *F*  Returns:        TRUE= more bands follow, FALSE= last band found
 *F*                  The last band may contain less stips than STRIPSPERBAND.
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  01.08.02  mdd     Create wrapper
 *F****************************************************************************/
static bool jbigDoBand(void)
{
  state.processingState = process;

  if(( ++state.bandNumber ) != state.totalBandCnt ) {	
    // all bands but last band take this path

//fixme	eTime += hmpv_read_fcnt();	// don't time the fread
//fixme	hmpv_write_fcnt(0);
#ifdef FILE_READ
	if (state.bandNumber == state.totalBandCnt-1)
		//fread (&state.inpBuffer[(state.bufferId)^1][0], state.lastBandCnt*state.bytesperline, 1, ifile);
        readPBMFile(&state.inpBuffer[(state.bufferId)^1][0], state.lastBandCnt, 
                    state.pixels, state.bytesperline, ifile);
	else
	    //fread (&state.inpBuffer[(state.bufferId)^1][0], JBIG2_STRIPSPERBAND*state.bytesperline, 1, ifile);
        readPBMFile(&state.inpBuffer[(state.bufferId)^1][0], JBIG2_STRIPSPERBAND, 
                    state.pixels, state.bytesperline, ifile);
#else
        // Wait until the Nth captured frame is available
        VideoioInWaitScanDone (&state.inpBuffer[state.bufferId][0]);
 
        // DsContinue to re-start transfer & spaResetToggle to re-start producer
        VideoioInGetScan (0);
#endif
//fixme	time656 += hmpv_read_fcnt();
//fixme	hmpv_write_fcnt(0);
    
    // G4 Compression
    jbigEncode (&state.inpBuffer[state.bufferId][0], &state, JBIG2_STRIPSPERBAND);
    state.stripNumber += JBIG2_STRIPSPERBAND;
    state.bufferId = (state.bufferId)^1;

    PRINTF(("<jbigDoBand> band %u ends w/strip %u \n",
	   state.bandNumber, state.stripNumber));
    return( TRUE );	// signal more bands
  } else {	
    // only last band takes this path, it could be a full band also
    // FIXME: last strip of last band may need raster zerofill appended
    // FIXME: for (JBIG2_LINESPERSTRIP - lastStripCnt ) scanlines.

    jbigEncode (&state.inpBuffer[state.bufferId][0], &state, state.lastBandCnt);
    state.stripNumber += state.lastBandCnt;
      
    PRINTF(("<jbigDoBand> band %u ends w/strip %u \n",
	    state.bandNumber, state.stripNumber));

    return( FALSE );	// signal last band
  }
}



/*F*****************************************************************************
 *F*
 *F*  Function Name:  jbigClose
 *F*
 *F*  Operation:      Flush pipeline to complete processing of current image.
 *F*
 *F*  Function
 *F*  Returns:        Nothing
 *F*
 *F*  Function Revision History:
 *F*    Date    Author  Description
 *F*  --------  ------  ----------------------
 *F*  01.08.02  mdd     Create wrapper
 *F****************************************************************************/
/*F*/ static void jbigClose(void)
{
   PRINTF(("<jbigClose>\n"));

   state.processingState = finish;
   jbigEncode (&state.inpBuffer[0][0], &state, JBIG2_STRIPSPERBAND);

//----fixme   eTime += hmpv_read_fcnt();

   printf("<jbigtsk> JBIG2 Encoding cycles = %ld\n", eTime);
   printf("<jbigtsk> JBIG2 Encoding Time for this image = %3.3f\n", ((float)eTime * 0.00000000463));
   printf("<jbigtsk> 656 Xfer or File Read time this image = %3.3f\n", ((float)time656 * 0.00000000463));
   printf("<jbigtsk> Compressed Size = %d\n", state.outCurAddress - outimage);

#ifndef PCVERSION
   printf("<jbigtsk> JBIG2 Time for 8.5x11 doc. = %3.3f\n", ((float)eTime * 6.600 * 5.100 * 0.00463 )/(state.pixels*state.scanlines));
#endif

#ifdef FILE_READ
   fclose(ifile);
#endif
}


/* Return the outimage address location and the compressed size */
uint8* getAddressAndSize (uint32 *size)
{
    *size = state.outCurAddress - outimage;
	//*size = state.compressedSize;
	return (outimage);
}



