
#include "StdAfx.h"
#include "Compression.h"
#include "Debug.h"

ICompression::ICompression (void) :
     m_width (0),
     m_height (0),
     m_bpl (0),
     m_bpp (0),
     m_channels (0),
     m_padding (0),
     m_pCompressedData (NULL),
     m_compressedDataSize (0)
{};

ICompression::~ICompression (void)
{
	ClearOutputBuffer();
}

void
ICompression::ClearOutputBuffer (void)
{
	if (m_pCompressedData)
	{
		delete [] m_pCompressedData;
		m_pCompressedData = NULL;
	}
	m_compressedDataSize = 0;

	return;
}


BOOL
ICompression::SetSize (ULONG width, ULONG height, ULONG bytesPerLine,
		       ULONG bitsPerPixel, ULONG channels)
{

  if (width == 0 || height == 0 || bytesPerLine == 0 || bitsPerPixel == 0 ||
      channels == 0) return FALSE;

  m_width = width;
  m_height = height;
  m_bpl = bytesPerLine;
  m_bpp = bitsPerPixel;
  m_channels = channels;

  ULONG expected_bpl = (width * m_bpp + 7) / 8;
  m_padding = m_bpl - expected_bpl;

  DBG_WATCH( "expected_bpl", expected_bpl );
  DBG_WATCH( "bpl", m_bpl );
  DBG_WATCH( "padding", m_padding );

  return TRUE;
}

BOOL 
ICompression::AllocateOutputBuffer ( LONG expectedCompressionFactor )
{
	
	ULONG uncompressedSize = ((m_width + m_padding) * m_height * m_channels * m_bpp + 7) / 8;
	ULONG newSize;

	// the G4 "compression" sometimes expands the data size ....
	if (expectedCompressionFactor < 0)
	{
		newSize	= uncompressedSize * -expectedCompressionFactor;
	}
	else
	{
		newSize = uncompressedSize / expectedCompressionFactor;
	}

	// we already have a buffer that is large enough
	if (m_compressedDataSize >= newSize)
	{
		return TRUE;
	}

	// allocate a new buffer
	ClearOutputBuffer();	
	m_compressedDataSize = newSize;
	m_pCompressedData = new unsigned char [m_compressedDataSize];
	if (m_pCompressedData)
	{
		return TRUE;
	}

	return FALSE;
}

// default implementation: no compression, simple mem copy
BOOL
ICompression::Compress (unsigned char* imageData, ULONG dataSize,
			 unsigned char** compressedData, 
			 ULONG* compressedDataSize)
{
	
	AllocateOutputBuffer ( 1 );

	memcpy (static_cast<void*> (m_pCompressedData), static_cast<const void*> (imageData), m_compressedDataSize);

	*compressedData = m_pCompressedData;
	*compressedDataSize = m_compressedDataSize;
	
	return TRUE;
}
