#include "stdafx.h"

#include "OemClassFactory.h"

#include "Debug.h"

//////////////////////////////////////////////////////////////////////////
//
// DLL Interface
//
//////////////////////////////////////////////////////////////////////////

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved )
{

#if 0

    static Stopwatch tmrCal( "Calibration" );
    static Stopwatch tmrFake( "Fake" );

    tmrCal.start();

    for( size_t ii=0; ii < 1000000; ++ii)
    {
        tmrFake.start();
        tmrFake.stop();
    }

    tmrCal.stop();
    
    DbgStream() << tmrCal << std::endl;

#endif

    static Stopwatch tmr( "DllMain" );
    
    tmr.start();

    DBG_TRACE( "DllMain" );

    DBG_WATCH( "hModule", (void*)(hModule) );

    switch (ul_reason_for_call)
    {
        case DLL_PROCESS_ATTACH: DBG_MSG( "DLL_PROCESS_ATTACH" ); break;
        case DLL_PROCESS_DETACH: DBG_MSG( "DLL_PROCESS_DETACH" ); break;
        case DLL_THREAD_ATTACH : DBG_MSG( "DLL_THREAD_ATTACH" ); break;
        case DLL_THREAD_DETACH : DBG_MSG( "DLL_THREAD_DETACH" ); break;
        
        default: break;
    }
 
    DbgStream() << tmr << std::endl;

    return TRUE;
}

STDAPI DllGetClassObject( const CLSID & rclsid,
                          const   IID &   riid,
                          void **ppvObj )
{
    DBG_TRACE( "DllGetClassObject" );

    *ppvObj = NULL;

    // Can we create this component?  The only class that this DLL knows
    // how to make is CLSID_OEMRENDER.
    if( rclsid != CLSID_OEMRENDER )
    {
        DBG_ERROR( "CLASS_E_CLASSNOTAVAILABLE" );
        return CLASS_E_CLASSNOTAVAILABLE;
    }

    OemClassFactory * pCF = new OemClassFactory;

    if( !pCF )
    {
        DBG_ERROR( "E_OUTOFMEMORY" );
        return E_OUTOFMEMORY;
    }

    HRESULT hr = pCF->QueryInterface( riid, ppvObj );
    
    // We call release here because our reference count will be two here.
    // We get one from the construction and another from the call to
    // QueryInterface.
    pCF->Release();


    // Despite the fact that MSDN says you return S_OK on success, you must
    // really return the handle from QueryInterface.

    DBG_WATCH( "hr", (void*)(hr) );
    return hr;
}

STDAPI DllCanUnloadNow( void )
{
    DBG_TRACE( "DllCanUnloadNow" );
    return S_OK;
}
