
#include "StdAfx.h"
#include "ijl.h"
#include "jpeg.h"

BOOL IJpeg::Compress (unsigned char* imageData, ULONG dataSize,
			 unsigned char** compressedData, 
			 ULONG* compressedDataSize)
{
  BOOL bres, memErr;
  IJLERR jerr;
  // Allocate the IJL JPEG_CORE_PROPERTIES structure.
  JPEG_CORE_PROPERTIES jcprops;

  LONG compressionFactor = 10;

  do {
  AllocateOutputBuffer (compressionFactor);

  memErr = FALSE;
  bres = TRUE;

  __try
    {
      
      // Initialize the Intel(R) JPEG Library.
	jerr = ijlInit(&jcprops);
      if(IJL_OK != jerr)
	{
	  bres = FALSE;
	  __leave;
	}
      // Set up information to write from the pixel buffer.
      jcprops.DIBWidth = m_width;
      jcprops.DIBHeight = m_height; // Implies a bottom-up DIB.
      jcprops.DIBBytes = imageData;
      jcprops.DIBPadBytes = m_padding;
      jcprops.DIBChannels = m_channels;
      jcprops.DIBColor = IJL_RGB;
      jcprops.JPGWidth = m_width;
      jcprops.JPGHeight = m_height;
      jcprops.JPGFile = NULL;
      jcprops.JPGBytes = m_pCompressedData;
      jcprops.JPGSizeBytes = m_compressedDataSize;
      jcprops.JPGChannels = m_channels;
      jcprops.JPGColor = IJL_YCBCR;
      jcprops.JPGSubsampling = IJL_NONE; // No subsampling as long as we work on 8x8 ...
	  jcprops.jquality = m_quality; // Select "good" image quality
	  // Write the actual JPEG image from the pixel buffer.
	    jerr = ijlWrite(&jcprops,IJL_JBUFF_WRITEWHOLEIMAGE);
      if(IJL_OK != jerr)
	{
	  if (jerr == IJL_BUFFER_TOO_SMALL) memErr = TRUE;
	  bres = FALSE;
	  __leave;
	}
    } // __try
      __finally
	{
	  *compressedData = m_pCompressedData;
	  *compressedDataSize = jcprops.JPGSizeBytes;
	  // Clean up the Intel(R) JPEG Library.
	    ijlFree(&jcprops);
	}

  if (compressionFactor > 1) compressionFactor /= 2;
  else if (compressionFactor == 1) compressionFactor = -2;
  else compressionFactor *= 2;

  } while (memErr);

  return bres;
} // EncodeToJPEGBuffer()

