#include "stdafx.h"

#include <INITGUID.H> // This will force the implementation of the guids.
#include <PRCOMOEM.H>

#include "OemBase.h"
#include "OemClassFactory.h"
#include "Debug.h"
#include "misc.h"

static Stopwatch startToEndTmr( "Start-to-end" );

//JBTODO -- Multiplicity of IOemBase?
// I am not sure whether or not it is safe to use "this" as the private
// portion of the PDEV object.


// I am not at all sure that we need this method.  See the DDK documentation.
// for details.
BOOL APIENTRY DrvQueryDriverInfo( DWORD dwMode, PVOID pBuffer, DWORD cbBuf,
                                  PDWORD  pcbNeeded )
{   
    DBG_TRACE( "DrvQueryDriverInfo" );

    BOOL ok = FALSE;

    if( dwMode == DRVQUERY_USERMODE )
    {
        if( pcbNeeded )
        {
            *pcbNeeded = sizeof(DWORD);
        }

        if( pBuffer )
        {
            if( cbBuf >= sizeof(DWORD) )
            {
                ok = TRUE;
                *static_cast<DWORD *>(pBuffer) = TRUE;
            }
        }
        
        if( !pBuffer && pcbNeeded )
        {
            ok = TRUE;
        }
    }

    return ok;
}

// Here was are just initializing our brushes to be solid black and
// solid white.

BRUSHOBJ IOemBase::m_blackBr = { 0x00000000, NULL, 0 };
BRUSHOBJ IOemBase::m_whiteBr = { 0x00FFFFFF, NULL, 0 };


// This array maps the global hook functions needed for our hook function
// table to their identifying in constants.  It is not a member of IOemBase
// because you cannot make arrays of indefinite length member variables.
// (At least I cannot figure out how.)  This array's sole purpose in life
// is to be used by EnableDriver().

static const DRVFN IOemBase_OEMHookFuncs[] =
{
    { INDEX_DrvAlphaBlend,          (PFN) OEMAlphaBlend          },
    { INDEX_DrvBitBlt,              (PFN) OEMBitBlt              },
    { INDEX_DrvCopyBits,            (PFN) OEMCopyBits            },
    { INDEX_DrvDitherColor,         (PFN) OEMDitherColor         },
    { INDEX_DrvEndDoc,              (PFN) OEMEndDoc              },
    
    { INDEX_DrvEscape,              (PFN) OEMEscape              },
    { INDEX_DrvFillPath,            (PFN) OEMFillPath            },
    { INDEX_DrvFontManagement,      (PFN) OEMFontManagement      },
    { INDEX_DrvGetGlyphMode,        (PFN) OEMGetGlyphMode        },
    { INDEX_DrvGradientFill,        (PFN) OEMGradientFill        },

    { INDEX_DrvLineTo,              (PFN) OEMLineTo              },
    { INDEX_DrvNextBand,            (PFN) OEMNextBand            },
    { INDEX_DrvPlgBlt,              (PFN) OEMPlgBlt              },
    { INDEX_DrvQueryAdvanceWidths,  (PFN) OEMQueryAdvanceWidths  },
    { INDEX_DrvQueryFont,           (PFN) OEMQueryFont           },    

    { INDEX_DrvQueryFontData,       (PFN) OEMQueryFontData       },
    { INDEX_DrvQueryFontTree,       (PFN) OEMQueryFontTree       },
    { INDEX_DrvRealizeBrush,        (PFN) OEMRealizeBrush        },
    { INDEX_DrvSendPage,            (PFN) OEMSendPage            },
    { INDEX_DrvStartBanding,        (PFN) OEMStartBanding        },

    { INDEX_DrvStartDoc,            (PFN) OEMStartDoc            },
    { INDEX_DrvStartPage,           (PFN) OEMStartPage           },
    { INDEX_DrvStretchBlt,          (PFN) OEMStretchBlt          },
    { INDEX_DrvStretchBltROP,       (PFN) OEMStretchBltROP       },  
    { INDEX_DrvStrokeAndFillPath,   (PFN) OEMStrokeAndFillPath   },   

    { INDEX_DrvStrokePath,          (PFN) OEMStrokePath          },
    { INDEX_DrvTextOut,             (PFN) OEMTextOut             },
    { INDEX_DrvTransparentBlt,      (PFN) OEMTransparentBlt      }
};


// The number of elements in the above array.

static const size_t IOemBase_numOEMHookFuncs =
    sizeof(IOemBase_OEMHookFuncs) / sizeof(DRVFN);



// This table is similar to the one above except that it maps the
// various contants to EngXxxx functions that they can be punted
// to.  This array (and its size variable below) is used by the
// constructor to initialize the punt functions table.
static const DRVFN IOemBase_EngHookFuncs[] =
{
    { INDEX_DrvAlphaBlend,          (PFN) EngAlphaBlend          },
    { INDEX_DrvBitBlt,              (PFN) EngBitBlt              },
    { INDEX_DrvCopyBits,            (PFN) EngCopyBits            },
//  { INDEX_DrvDitherColor,         (PFN) EngDitherColor         },
    { INDEX_DrvFillPath,            (PFN) EngFillPath            },

    { INDEX_DrvGradientFill,        (PFN) EngGradientFill        },
    { INDEX_DrvLineTo,              (PFN) EngLineTo              },
    { INDEX_DrvPlgBlt,              (PFN) EngPlgBlt              },
    { INDEX_DrvStretchBlt,          (PFN) EngStretchBlt          },
    { INDEX_DrvStretchBltROP,       (PFN) EngStretchBltROP       },  
    
    { INDEX_DrvStrokeAndFillPath,   (PFN) EngStrokeAndFillPath   },   
    { INDEX_DrvStrokePath,          (PFN) EngStrokePath          },
    { INDEX_DrvTextOut,             (PFN) EngTextOut             },
    { INDEX_DrvTransparentBlt,      (PFN) EngTransparentBlt      },    
};

static const size_t IOemBase_numEngHookFuncs =
    sizeof(IOemBase_EngHookFuncs)/sizeof(DRVFN);

//////////////////////////////////////////////////////////////////////////
//
// IOemBase non-exported methods.
//
//////////////////////////////////////////////////////////////////////////

IOemBase::IOemBase( void )
:
 m_refCount( 1 ),  // Do not call AddRef() on construction!
 m_pOEMHelp( NULL ),
 m_hookPtr( NULL ),
 m_isBanded( false ),
 m_curPage( 0 ),
 m_jobId( 0 )
{ 
    m_bandOrigin.x = 0;
    m_bandOrigin.y = 0;
    
    DBG_TRACE_OBJ( "IOemBase::IOemBase" );

    // Our base class implements these three methods.
    implementMethod( NAME_GetImplementedMethod   );
    implementMethod( NAME_GetInfo                );
    implementMethod( NAME_PublishDriverInterface );
    implementMethod( NAME_EnablePDEV             );
    implementMethod( NAME_ResetPDEV              );
    implementMethod( NAME_DisablePDEV            );
    implementMethod( NAME_EnableDriver           );
    implementMethod( NAME_DisableDriver          );

    // Initialize the punt function table with the EngXxxx functions.
    // Because of the way that std::set objects work, any attempt to
    // insert methods on top of these will fail.  Thus, we will always
    // get the behaviour of calling EngXxxx functions when we punt when
    // those functions exist.  We try (as much as possible) to ignore the
    // functions given to us in EnableDriver().
    for( size_t ii = 0; ii < IOemBase_numEngHookFuncs; ++ii )
    {
        m_puntFunctions.insert( IOemBase_EngHookFuncs[ii] );
    }
}

IOemBase::~IOemBase( void )
{
    DBG_TRACE_OBJ( "IOemBase::~IOemBase" );

    if( m_pOEMHelp )
    {
        m_pOEMHelp->Release();
        m_pOEMHelp = NULL;
    }

    delete [] m_hookPtr; m_hookPtr = NULL;

    DbgStream() << startToEndTmr  << std::endl;
}

void IOemBase::implementMethod( const std::basic_string<CHAR> & method )
{
    DBG_TRACE_OBJ( "IOemBase::implementMethod" );

    m_methodNameTable.insert( method );
}

PFN IOemBase::getPuntFunc( ULONG iFunc )
{
    const DRVFN key = { iFunc, NULL };
    std::set<DRVFN>::const_iterator fItr = m_puntFunctions.find( key );

    PFN pfn = NULL;

    if( fItr != m_puntFunctions.end() )
    {
        pfn = (*fItr).pfn;
    }
    else
    {
        // JBTOD -- SetLastError();
    }

    return pfn;
}
 
// Convenience method to turn a unicode string represented as an array
// into a std::string<WCHAR>.  It is probably an error to call this
// method with a NULL pointer since you get "<NULL>" as your answer,
// but for debugging, it's fine.
std::basic_string<WCHAR> IOemBase::makeStr( const WCHAR * pwStr )
{
    std::basic_string<WCHAR> str;

    if( pwStr )
    {
        for( ; *pwStr; ++pwStr )
        {
            str += *pwStr;
        }
    }
    else
    {
        str += WCHAR('<');
        str += WCHAR('N');
        str += WCHAR('U');
        str += WCHAR('L');
        str += WCHAR('L');
        str += WCHAR('>');
    }

    return str;
}

const size_t IOemBase::bitsPixel( void ) const
{
    return m_gdiInfo.cBitsPixel * m_gdiInfo.cPlanes;
}
   
// Note that because we initialize m_bandOrigin in DrvStartPage()
// bandOverrun() will produce the correct answer if called during
// DrvSendPage() instead of DrvNextBand().  (That is, it will give
// 0 as the answer.)
long IOemBase::bandOverrun( SURFOBJ * pso ) const
{
    return (m_bandOrigin.y + pso->sizlBitmap.cy) - m_gdiInfo.ulVertRes;
}

// Note that because we initialize m_bandOrigin in DrvStartPage()
// bandRows() will produce the correct answer if called during
// DrvSendPage() instead of DrvNextBand().  (That is, it will give
// pso->sizlBitmap.cy as the answer, which will be equal to 
// m_gdi_info.ulVertRes.)
size_t IOemBase::bandRows( SURFOBJ * pso ) const
{
    return (bandOverrun(pso) < 0)? pso->sizlBitmap.cy  : 
                                   (m_gdiInfo.ulVertRes - m_bandOrigin.y);
}


std::string IOemBase::
getSetting( PDEVOBJ pdevobj, const std::string & setting )
{
    DWORD options  = 0;
    DWORD cbNeeded = 0;

    this->m_pOEMHelp->DrvGetDriverSetting( pdevobj, setting.c_str(), NULL, 0, &cbNeeded, &options );

    if( cbNeeded )
    {
        char * output = new char[cbNeeded];

        this->m_pOEMHelp->DrvGetDriverSetting( pdevobj, setting.c_str(), output, cbNeeded, &cbNeeded, &options );

        std::string ret( output );

        delete [] output;

        return ret;
    }

    return std::string();
}

//////////////////////////////////////////////////////////////////////////
//
// IOemBase IUnknown methods.
//
// These are just basic implementations of IUnknown.  (See SDK help.)
// These are thread-safe.  I cannot imagine that you will ever need
// to understand or use these, let alone change them.  You can think of
// them as part of the black magic that makes the DLL work. 
//
//////////////////////////////////////////////////////////////////////////

HRESULT __stdcall IOemBase::QueryInterface( const IID & iid, void ** ppv )
{    
    DBG_TRACE_OBJ( "IOemBase::QueryInterface" );

    if( (iid == IID_IUnknown    ) ||
        (iid == IID_IPrintOemUni)  )
    {
        *ppv = this;
        this->AddRef();
        return S_OK;
    }

    *ppv = NULL;

    return E_NOINTERFACE;
}

ULONG __stdcall IOemBase::AddRef( void )
{
    DBG_TRACE_OBJ( "IOemBase::AddRef" );
    
    return InterlockedIncrement( &m_refCount );
}

ULONG __stdcall IOemBase::Release( void ) 
{
    DBG_TRACE_OBJ( "IOemBase::Release" );

    if( InterlockedDecrement( &m_refCount ) == 0 )
    {
        delete this;
        return 0;
    }
    
    return m_refCount;
}

//////////////////////////////////////////////////////////////////////////
//
// IOemBase IOemPrintUni Methods
//
//////////////////////////////////////////////////////////////////////////

// By implementing GetImplementedMethod() this way, nobody else ever has
// to implement it again.  They just have to call implementMethod() in
// their constructor.

HRESULT __stdcall IOemBase::GetImplementedMethod( PSTR pMethodName )
{
    DBG_TRACE_OBJ( "IOemBase::GetImplementedMethod" );

    std::basic_string<CHAR> str( pMethodName );

    const bool ok = ( m_methodNameTable.find( str ) !=
                      m_methodNameTable.end() );

    return ok? S_OK : S_FALSE;

    // The documentation indicates that we should return E_FAIL if a method
    // is not found, but the sample code returns S_FALSE.
    //  return ok? S_OK : E_FAIL; JBTODO
}


// I don't think anybody should ever have to override GetInfo() in a
// derived class.
HRESULT __stdcall IOemBase::
GetInfo( DWORD dwMode, PVOID pBuffer, DWORD cbSize, PDWORD pcbNeeded )
{
    DBG_TRACE_OBJ( "IOemBase::GetInfo" );

    // pcbNeeded is required.
    if( !pcbNeeded )
    {
        SetLastError(ERROR_INVALID_PARAMETER);
        return E_FAIL;
    }
   
    // Make sure that dwMode is valid.  Note that there are two constants
    // defined in the header file but not in the documentation.  I have
    // included them below.
    switch( dwMode )
    {
        case OEMGI_GETSIGNATURE: break;
        case OEMGI_GETVERSION  : break;
        
        case OEMGI_GETPUBLISHERINFO   :
        {
            // This is undocumented, but done in the DDK example code.
            *pcbNeeded = sizeof(PUBLISHERINFO);
            return E_FAIL;
        }

        case OEMGI_GETINTERFACEVERSION: // Undocumented.
        default                       :
        {
            *pcbNeeded = 0;
            SetLastError(ERROR_INVALID_PARAMETER);
            return E_FAIL;
        }
    }

    // For now, all cases will need a DWORD.
    *pcbNeeded = sizeof( DWORD );

    // If pBuffer is NULL, that is ok.  We just have to tell our caller
    // how much memory they would have needed.  Otherwise, we need to
    // cough up some information.
    if( pBuffer )
    {
        if( cbSize < *pcbNeeded )
        {
            SetLastError(ERROR_INSUFFICIENT_BUFFER);
            return E_FAIL;
        }
    

        switch( dwMode )
        {
            case OEMGI_GETSIGNATURE: *(PDWORD)pBuffer = OEM_SIGNATURE; break;
            case OEMGI_GETVERSION  : *(PDWORD)pBuffer = OEM_VERSION;   break;
            
            default : assert( NULL ); break;
            
            // default should be impossible because we
        }
    }

    return S_OK;
}

// I don't think anybody should ever have to override
// PublishDriverInterface().  Note that this method is a way of giving
// the driver interface to the IPrintOemUni object, not a way of asking
// for it.  The pointer comes in as pIUnknown and gets stored as
// m_pOEMHelp.
//
// JBTODO -- I don't know when the object represented by the returned
// handle is supposed to be released.  Based on the sample code with the
// DDK, I am currently releasing it in DisableDriver() and in the destructor
// (if I still have one for some reason.)  This seems somewhat
// counterintuitive since it gets created here and not in EnableDriver().
HRESULT __stdcall IOemBase::PublishDriverInterface( IUnknown *pIUnknown )
{
    DBG_TRACE_OBJ( "IOemBase::PublishDriverInterface" );

    HRESULT hResult = S_OK;

    // If we already have a driver interface, we do not need to make
    // another one.  While this makes intuitive sense, it is interesting that
    // we don't care whether the IUnknown interface we get passed happens
    // to be the same instance as the one we have or not.
    // JBTODO:  Is this the correct behaviour?
    if( !m_pOEMHelp )
    {
        // This is effectively a down-cast.  Exactly why we do it this way is unclear to 
        // me, but this is apparently what we are supposed to do.  (I would have just
        // passed in a pointer to an IPrintOemDrvierUni to begin with.)
        HRESULT qryRes = pIUnknown->QueryInterface( IID_IPrintOemDriverUni,
                      reinterpret_cast<void**>(&m_pOEMHelp) );

        if( qryRes != S_OK )
        {
            m_pOEMHelp = NULL;
            SetLastError(E_NOINTERFACE);
            hResult    = E_FAIL;
        }
    }

    return hResult;
}

// I can't think of any reason you will have to override EnableDriver().
HRESULT __stdcall IOemBase::
EnableDriver( DWORD dwDriverVersion, DWORD cbSize, PDRVENABLEDATA pded )
{
    DBG_TRACE_OBJ( "IOemBase::EnableDriver" );

    // Store for later use by derived classes.
    m_driverVersion = dwDriverVersion;

    DBG_WATCH( "m_driverVersion", m_driverVersion );
    
    // We really have to be given pded here.  That's the entire point of the
    // function.  Thus, I claim that a null pded is falure.
    if( !pded )
    {
        DBG_ERROR( "pded == NULL" );
        return E_FAIL;
    }

    // Given that DRVENABLEDATA is a structure of known size, it seems
    // absurd that we would ever get too little data, but we get a value
    // so we check it.
    if( cbSize < sizeof(DRVENABLEDATA) )
    {
        DBG_ERROR( "cbSize < sizeof(DRVENABLEDATA)" );
        return E_FAIL;
    }
    
    // My understanding is that EnableDriver() should not be called twice for a 
    // given instance without calling DisableDriver() in between.  The reason I
    // demand that m_hookPtr be NULL as opposed to simply deleting it is that
    // we give m_hookPtr to our caller as it's pded->pdrvfn value.  The only
    // time it is safe to delete this will be in DisableDriver().  If you ever
    // get failure here, we will have to rethink how we pass back the hook table.
    if( m_hookPtr )
    {
        DBG_ERROR( "EnableDriver() called twice in a row." );
        SetLastError(ERROR_INTERNAL_ERROR);
        return E_FAIL;
    }

    m_hookPtr = new DRVFN[IOemBase_numOEMHookFuncs];

    if( !m_hookPtr )
    {
        DBG_ERROR( "Out of memory" );
        SetLastError(ERROR_NOT_ENOUGH_MEMORY);
        return E_FAIL;
    }

    // Make a special copy just for this caller.  We do this because
    // pded->pdrvfn is not const.
    memcpy( m_hookPtr, IOemBase_OEMHookFuncs, sizeof( IOemBase_OEMHookFuncs ) );
    
    pded->iDriverVersion = PRINTER_OEMINTF_VERSION;
    pded->c              = IOemBase_numOEMHookFuncs;
    pded->pdrvfn         = m_hookPtr;
        
    return S_OK;
}

// You should never have to override DisableDriver.
HRESULT __stdcall IOemBase::DisableDriver( void )
{
    DBG_TRACE_OBJ( "IOemBase::DisableDriver" );

    if( m_pOEMHelp )
    {
        m_pOEMHelp->Release();
        m_pOEMHelp = NULL;
    }
    
    delete [] m_hookPtr; m_hookPtr = NULL;

    return S_OK;
}

/*************************************************************************
  EnablePDEV and DisablePDEV are a pair.  If you implement one, you
  will probably have to implement the other.

  If you do implement either of these, your EnablePDEV() method should call
  IOemBase::EnablePDEV (or a lower base class version) as it's FIRST
  operation.  This is because several things happen in EnablePDEV that
  always need to happen.

  - m_printerName must be set.
  - m_gdiInfo must be set.
  - m_devInfo must be set.
  - The data from pded must be inserted into m_puntFunctions.

  Similarly if you implement DisablePDEV(), your version should call
  IOemBase::DisablePDEV() (or a lower base class version) as its LAST
  operation.  At present, DisablePDEV does not actually do anything,
  but it is much like a destructor call.  Be good and call it anyway.

**************************************************************************/

HRESULT __stdcall IOemBase::EnablePDEV(
    PDEVOBJ         pdevobj,
    PWSTR           pPrinterName,
    ULONG           cPatterns,
    HSURF          *phsurfPatterns,
    ULONG           cjGdiInfo,
    GDIINFO        *pGdiInfo,
    ULONG           cjDevInfo,
    DEVINFO        *pDevInfo,
    DRVENABLEDATA  *pded,
    OUT PDEVOEM    *pDevOem)
{
    DBG_TRACE_OBJ( "IOemBase::EnablePDEV" );

    //
    // Stash the printer name, GDI info, and Dev info for future
    // use by derived classes.
    //

    if( pPrinterName )
    {
        m_printerName = std::basic_string<WCHAR>(pPrinterName);
    }
    else
    {
        // I am not sure if this is appropriate behaviour.  Missing the
        // printer name hardly seems fatal though.
        m_printerName = std::basic_string<WCHAR>(L"<NULL>");
    }
    
    DBG_WATCH( "m_printerName", m_printerName );

    if( !pGdiInfo )
    {
        DBG_ERROR( "pGdiInfo == NULL" );
        SetLastError(ERROR_BAD_ARGUMENTS);
        return E_FAIL;
    }

    m_gdiInfo = *pGdiInfo;

    DBG_WATCH( "ulVertRes", m_gdiInfo.ulVertRes );
    DBG_WATCH( "ulHorzRes", m_gdiInfo.ulHorzRes );

    if( !pDevInfo )
    {
        DBG_ERROR( "pDevInfo == NULL" );
        SetLastError(ERROR_BAD_ARGUMENTS);
        return E_FAIL;
    }

    m_devInfo = *pDevInfo;

    if( !pded )
    {
        DBG_ERROR( "pded == NULL" );
        SetLastError(ERROR_BAD_ARGUMENTS);
        return E_FAIL;
    }

    const ULONG numPuntFuncs = pded->c;
    DRVFN *        puntFuncs = pded->pdrvfn;

    // I am not sure if this is really a failure.  I don't know if
    // a NULL pded->pdrvfn value simply indicates no punt functions.
    if( !puntFuncs )
    {
        DBG_ERROR( "puntFuncs == NULL" );
        SetLastError(ERROR_BAD_ARGUMENTS);
        return E_FAIL;
    }

    // Note that this is actually a secondary way of punting.  We already
    // inserted a bunch of EngXxxx functions into this set in the
    // constructor.  Any function which overlaps here will not be inserted.
    // (This is because that is how a set behaves.)  However, if there is
    // some function for which there is no EngXxxx version, we can get it
    // this way.
    for( size_t ii = 0; ii < numPuntFuncs; ++ii )
    {
        m_puntFunctions.insert( puntFuncs[ii] );
    }

    *pDevOem = this;

    return S_OK;
}

// 
/*

 I am not completely sure of the correct way to implement ResetPDEV().
 Try reading the documentation.  I hope you get more from it than
 I did.  Most of the information below was determined experimentally and
 may be wrong.  However, it is a good starting point.

 --DrvEnablePDEV() is always called for a new instance of an
   IPrintOemUni, regardless of whether it is going to be used in a
   ResetPDEV() call.
 --DrvEnablePDEV() is always called before ResetPDEV().  Thus, by the
   the time an new object is passed as pdevobjNew to ResetPDEV(), it
   already has some non-trivial state.
 --The state passed to its DrvEnablePDEV() is the correct state.  Thus,
   m_gdiInfo, m_devInfo, and m_printerName should not be copied from
   the old object.
 --EnableDriver() is always called before ResetPDEV().  Thus m_hookPtr
   m_puntFunctions and are not dealt with here.
 --The state passed to EnableDriver() is correct, thus m_driverVersion need
   not be worried about here.  (This one does not seem to matter, though,
   one way or the other as I have only ever seen it take on one value.
 --The old object will never be used again, so although you cannot leave
   it corrupted, you can leave pointers NULL, etc.  It is okay to transfer
   a pointer to a heap object instead of copying it.  (Be careful to NULL
   out the pointers in the old object, however, so that they do not get
   deleted prematurely.)
 --m_jobId and m_docName are invalid at this point ResetPDEV() is called,
   so we go ahead and copy them.  However, if we end up getting valid
   ones in DrvStartDoc(), then those take precedence.
 --ResetPDEV() is effectively a static function.  Note that there is no use
   of this-> anywhere in ResetPDEV().
 --ResetPDEV() can happen in a number of circumstances, including mid-document
   or when starting a new document.
   
 I have added an extra level of indirection (initFrom()) to
 limit the number of reinterpret_cast operations we have to do.  When you
 dervive a class, you must leave ResetPDEV() as it is and write your own
 version of initFrom().
*/

HRESULT IOemBase::ResetPDEV( PDEVOBJ pdevobjOld, PDEVOBJ pdevobjNew )
{
    DBG_TRACE_OBJ( "IOemBase::ResetPDEV" );

    IOemBase * oldBase = reinterpret_cast<IOemBase *>(pdevobjOld->pdevOEM);
    IOemBase * newBase = reinterpret_cast<IOemBase *>(pdevobjNew->pdevOEM);

    DBG_WATCH( "oldBase", (void*)(oldBase) );
    DBG_WATCH( "newBase", (void*)(newBase) );

    return newBase->initFrom( oldBase );
}

// Your initFrom should call your base class version of initFrom() as
// it's first operation.  This of it as being part of a sort of 
// construction.  See also the description of ResetPDEV().
HRESULT IOemBase::initFrom( IOemBase * old )
{
    DBG_TRACE_OBJ( "IOemBase::initFrom" );

    //
    // I believe the following section is correct.
    //

    this->m_isBanded = old->m_isBanded;
    this->m_curPage  = old->m_curPage;    
    this->m_jobId    = old->m_jobId;
    this->m_docName  = old->m_docName;

    this->m_bandOrigin = old->m_bandOrigin;

    //
    // We do not touch these here!  (See ResetPDEV())
    //
    // this->m_gdiInfo
    // this->m_devInfo
    // this->m_printerName
    // this->m_driverVersion
    // this->m_hookPtr
    // this->m_puntFunctions

    // This is does not get touched here either.
    //
    // this->m_refCount
    
    //
    // I have no idea if this is the correct way of handling m_pOEMHelp
    // or not.  It seems to work this way.
    //

    if( this->m_pOEMHelp )
    {
        this->m_pOEMHelp->Release();
    }
    
    this->m_pOEMHelp = old->m_pOEMHelp;
    this->m_pOEMHelp->AddRef();


    return S_OK;
}

// See comments for EnablePDEV().

HRESULT __stdcall IOemBase::DisablePDEV( PDEVOBJ )
{
    DBG_TRACE_OBJ( "IOemBase::DisablePDEV" );

    return S_OK;
}

//////////////////////////////////////////////////////////////////////////
//
// IOemBase DDI Hook Functions (Implemented)
//
//////////////////////////////////////////////////////////////////////////

    
/*
  Aside from it's role stated in the Win2K SDK, DrvStartDoc has an
  additional obligation when you inherit from IOemBase.  It must
  sometimes set m_curPage to 0, m_jobId, and m_docName.

  The method returns FALSE if either dwJobId or pwszDocName are NULL.  This
  allows you to write a derived class version of DrvStartDoc() that calls
  the base class to take care of its base-class requirements.  You know that
  both the name and id were valid if it returns true.  (If it returns FALSE,
  then it is up to you whether to continue or not.)

  Basically, I believe that FALSE is returned when DrvStartDoc() is called
  because of a call to ResetPDEV() and TRUE is returned when we are really
  starting a document.  (Known from experimental evidence only!)

  You should call the base class version of DrvStartDoc() as your very first
  operation in your DrvStartDoc().
*/

BOOL IOemBase::DrvStartDoc( SURFOBJ *pso, PWSTR pwszDocName, DWORD dwJobId )
{
    DBG_TRACE_OBJ( "IOemBase::DrvStartDoc" );

    startToEndTmr.start();

    DBG_WATCH( "pwszDocName", makeStr(pwszDocName) );
    DBG_WATCH( "dwJobId", dwJobId       );

    // dwJobId and pwszDocName may sometimes be NULL.  Experimentally, I
    // have determined that this occurs only when DrvStartDoc() is called
    // because our object was initialized with ResetPDEV().  Thus, the if()
    // statement.  When both are valid, we are really starting a new
    // document.  When (either or?) both are invalid, we are just
    // reinitializing from a call to ResetPDEV() and we do not want to
    // overwrite our previous values.  Nor do we want to reset the page
    // number.
    if( dwJobId && pwszDocName )
    {
        m_curPage  = 0;    
        m_jobId    = dwJobId;
        m_docName  = makeStr( pwszDocName );
        
        return TRUE;
    }

    return FALSE;
}

/*
 DrvEngDoc() may not have to be implemented by your derived class.
 Thus, it assumes success.  Despite the fact that it does not
 effect the state of IOemBase, you can call it from your
 implementation to be consistent.  You should call the base class version
 as your last operation.  (It is sort of destructor like.)

 JBTODO -- Should DrvEndDoc() set m_curPage = 0, m_jobId = 0, m_docName
 to an empty string, and m_isBanded to false?
*/
BOOL IOemBase::DrvEndDoc( SURFOBJ *pso, FLONG fl )
{
    DBG_TRACE_OBJ( "IOemBase::DrvEndDoc" );

    startToEndTmr.stop();

    return TRUE;
}

/*
  Aside from it's roll stated in the Win2K SDK, DrvStartPage() has an
  additional obligation when you inherit from IOemBase.  It must
  incriment m_curPage, set m_isBanded to false, and initialize 
  m_bandOrigin to (0,0).  You should always (probably?) call the base
  class version as your first operation in your DrvStartPage().
*/

BOOL IOemBase::DrvStartPage( SURFOBJ *pso )
{
    DBG_TRACE_OBJ( "IOemBase::DrvStartPage" );

    m_isBanded = false;
 
    ++m_curPage;

    // These used to be initialized in DrvStartBanding().  This made more
    // semantic sense.  However, it also meant that these values were
    // poorly defined when there was no banding.  Since it is someimes useful
    // to treat and unbanded page as a banded page with a single band, it
    // seems like initializing these guys here makes sense.
    m_bandOrigin.x = 0;
    m_bandOrigin.y = 0;

    return TRUE;
}

/*

  DrvSendPage() fails by default since I cannot imagine a working printer
  driver that does not override DrvSendPage().  I have an implimentation
  here so that derived classes can call IOemBase::DrvSendPage() as their
  first action.  (This just makes it consistent.  DrvSendPage() could be
  a pure virtual function.)
*/

BOOL IOemBase::DrvSendPage( SURFOBJ *pso ) 
{
    DBG_TRACE_OBJ( "IOemBase::DrvSendPage" );

    return FALSE;
}


/*
  Aside from it's roll stated in the Win2K SDK, DrvStartBanding()
  has an additional obligation when you inherit from IOemBase.
  It must set m_isBanded to true.

  You will note that the default version of DrvStartBanding() succeeds
  while DrvNextBand() fails.  This is because banding is inevitably
  a driver-specific algorithm.  DrvStartBanding(), however will
  almost always be implemented the same way.  Thus, we can save
  some work by making a working default version.  DrvNextBand(),
  however, will almost always have to be implemented, so we 
  make that fail.  Note that DrvNextBand() is not really abstract
  in the sense that one can imagine a "working" printer driver that
  does not support banding at all.

  You should call the base class version of DrvStartBanding() as your
  very first operation in your implementation of DrvStartBanding().
*/

BOOL IOemBase::DrvStartBanding( SURFOBJ * pso, POINTL * pptl )
{
    DBG_TRACE_OBJ( "IOemBase::DrvStartBanding" );

    m_isBanded = true;
    
    // m_bandOrigin is initialized in DrvStartPage() so that it
    // has a well-defined value even when we are not banded.

    *pptl = m_bandOrigin;

    return TRUE;
}

// See comments for DrvStartBanding().
//
// You should call the base class version of DrvNextBand() as your
// very first operation in your implementation of DrvNextBand().

BOOL IOemBase::DrvNextBand( SURFOBJ * pso, POINTL *pptl )
{
    DBG_TRACE_OBJ( "IOemBase::DrvNextBand" );

    // (-1,-1) indicates we are done banding.
    pptl->x = -1;
    pptl->y = -1;

    return FALSE;
}

// We clearly cannot handle any escapes by default.
ULONG IOemBase::DrvEscape( SURFOBJ *, ULONG iEsc, ULONG cjIn, PVOID pvIn, 
                           ULONG cjOut, PVOID pvOut )
{
    DBG_TRACE_OBJ( "IOemBase::DrvEscape" );
    
    return 0; // Indicates escape is "not supported."
}

//////////////////////////////////////////////////////////////////////////
//
// IOemBase DDI Hook Functions (Punted)
//
//////////////////////////////////////////////////////////////////////////

BOOL IOemBase::DrvBitBlt(
    SURFOBJ   *psoTrg,
    SURFOBJ   *psoSrc,
    SURFOBJ   *psoMask,
    CLIPOBJ   *pco,
    XLATEOBJ  *pxlo,
    RECTL     *prclTrg,
    POINTL    *pptlSrc,
    POINTL    *pptlMask,
    BRUSHOBJ  *pbo,
    POINTL    *pptlBrush,
    ROP4      rop4 )
{
    DBG_TRACE_OBJ( "IOemBase::DrvBitBlt" );
    PFN_DrvBitBlt pfn = PFN_DrvBitBlt( getPuntFunc( INDEX_DrvBitBlt ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoTrg, psoSrc, psoMask, pco, pxlo, prclTrg, pptlSrc, pptlMask, pbo, pptlBrush, rop4 );
}

BOOL IOemBase::DrvStretchBlt(
    SURFOBJ         *psoDest,
    SURFOBJ         *psoSrc,
    SURFOBJ         *psoMask,
    CLIPOBJ         *pco,
    XLATEOBJ        *pxlo,
    COLORADJUSTMENT *pca,
    POINTL          *pptlHTOrg,
    RECTL           *prclDest,
    RECTL           *prclSrc,
    POINTL          *pptlMask,
    ULONG            iMode )
{
    DBG_TRACE_OBJ( "IOemBase::DrvStretchBlt" );
    PFN_DrvStretchBlt pfn = PFN_DrvStretchBlt( getPuntFunc( INDEX_DrvStretchBlt ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoDest, psoSrc, psoMask, pco, pxlo, pca, pptlHTOrg,  prclDest, prclSrc, pptlMask, iMode );
}

BOOL IOemBase::DrvCopyBits(
    SURFOBJ        *psoDest,
    SURFOBJ        *psoSrc,
    CLIPOBJ        *pco,
    XLATEOBJ       *pxlo,
    RECTL          *prclDest,
    POINTL         *pptlSrc )
{
    DBG_TRACE_OBJ( "IOemBase::DrvCopyBits" );
    PFN_DrvCopyBits pfn = PFN_DrvCopyBits( getPuntFunc( INDEX_DrvCopyBits ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoDest, psoSrc, pco, pxlo, prclDest, pptlSrc );
}

BOOL IOemBase::DrvTextOut(
    SURFOBJ    *pso,
    STROBJ     *pstro,
    FONTOBJ    *pfo,
    CLIPOBJ    *pco,
    RECTL      *prclExtra,
    RECTL      *prclOpaque,
    BRUSHOBJ   *pboFore,
    BRUSHOBJ   *pboOpaque,
    POINTL     *pptlOrg,
    MIX         mix )
{
    DBG_TRACE_OBJ( "IOemBase::DrvTextOut" );
    PFN_DrvTextOut pfn = PFN_DrvTextOut( getPuntFunc( INDEX_DrvTextOut ) );
    if( !pfn ) { return FALSE; }
    return pfn( pso, pstro, pfo, pco, prclExtra, prclOpaque, pboFore, pboOpaque, pptlOrg, mix );
}

BOOL IOemBase::DrvStrokePath(
    SURFOBJ    *pso,
    PATHOBJ    *ppo,
    CLIPOBJ    *pco,
    XFORMOBJ   *pxo,
    BRUSHOBJ   *pbo,
    POINTL     *pptlBrushOrg,
    LINEATTRS  *plineattrs,
    MIX         mix )
{
    DBG_TRACE_OBJ( "IOemBase::DrvStrokePath" );
    PFN_DrvStrokePath pfn = PFN_DrvStrokePath( getPuntFunc( INDEX_DrvStrokePath ) );
    if( !pfn ) { return FALSE; }
    return pfn( pso, ppo, pco, pxo, pbo, pptlBrushOrg, plineattrs, mix );
}

BOOL IOemBase::DrvFillPath(
    SURFOBJ    *pso,
    PATHOBJ    *ppo,
    CLIPOBJ    *pco,
    BRUSHOBJ   *pbo,
    POINTL     *pptlBrushOrg,
    MIX         mix,
    FLONG       flOptions )
{    
    DBG_TRACE_OBJ( "IOemBase::DrvFillPath" );
    PFN_DrvFillPath pfn = PFN_DrvFillPath( getPuntFunc( INDEX_DrvFillPath ) );
    if( !pfn ) { return FALSE; }
    return pfn( pso, ppo, pco, pbo, pptlBrushOrg, mix, flOptions );
}

BOOL IOemBase::DrvStrokeAndFillPath(
    SURFOBJ    *pso,
    PATHOBJ    *ppo,
    CLIPOBJ    *pco,
    XFORMOBJ   *pxo,
    BRUSHOBJ   *pboStroke,
    LINEATTRS  *plineattrs,
    BRUSHOBJ   *pboFill,
    POINTL     *pptlBrushOrg,
    MIX         mixFill,
    FLONG       flOptions )
{
    DBG_TRACE_OBJ( "IOemBase::DrvStrokeAndFillPath" );
    PFN_DrvStrokeAndFillPath pfn = PFN_DrvStrokeAndFillPath( getPuntFunc( INDEX_DrvStrokeAndFillPath ) );
    if( !pfn ) { return FALSE; }
    return pfn( pso, ppo, pco, pxo, pboStroke, plineattrs, pboFill, pptlBrushOrg, mixFill, flOptions );
}


BOOL IOemBase::DrvRealizeBrush(
    BRUSHOBJ   *pbo,
    SURFOBJ    *psoTarget,
    SURFOBJ    *psoPattern,
    SURFOBJ    *psoMask,
    XLATEOBJ   *pxlo,
    ULONG       iHatch )
{
    DBG_TRACE_OBJ( "IOemBase::DrvRealizeBrush" );

    // Apparently, we just need to return true.  I am not sure why.  Also,
    // call apparently cannot be punted (or isn't meant to be.

    return TRUE;

    //PFN_DrvRealizeBrush pfn = PFN_DrvRealizeBrush( getPuntFunc( INDEX_DrvRealizeBrush ) );
    //if( !pfn ) { return FALSE; }
    //return pfn( pbo, psoTarget, psoPattern, psoMask, pxlo, iHatch );
}

PIFIMETRICS IOemBase::DrvQueryFont(
    DHPDEV      dhpdev,
    ULONG       iFile,
    ULONG       iFace,
    ULONG      *pid )
{
    DBG_TRACE_OBJ( "IOemBase::DrvQueryFont" );
    PFN_DrvQueryFont pfn = PFN_DrvQueryFont( getPuntFunc( INDEX_DrvQueryFont ) );
    if( !pfn ) { return NULL; }
    return pfn( dhpdev, iFile, iFace, pid );
}

PVOID IOemBase::DrvQueryFontTree(
    DHPDEV      dhpdev,
    ULONG       iFile,
    ULONG       iFace,
    ULONG       iMode,
    ULONG      *pid )
{
    DBG_TRACE_OBJ( "IOemBase::DrvQueryFontTree" );
    PFN_DrvQueryFontTree pfn = PFN_DrvQueryFontTree( getPuntFunc( INDEX_DrvQueryFontTree ) );
    if( !pfn ) { return NULL; }
    return pfn( dhpdev, iFile, iFace, iMode, pid );
}

LONG IOemBase::DrvQueryFontData(
    DHPDEV      dhpdev,
    FONTOBJ    *pfo,
    ULONG       iMode,
    HGLYPH      hg,
    GLYPHDATA  *pgd,
    PVOID       pv,
    ULONG       cjSize )
{
    DBG_TRACE_OBJ( "IOemBase::DrvQueryFontData" );
    PFN_DrvQueryFontData pfn = PFN_DrvQueryFontData( getPuntFunc( INDEX_DrvQueryFontData ) );
    if( !pfn ) { return FD_ERROR; }
    return pfn( dhpdev, pfo, iMode, hg, pgd, pv, cjSize );
}

BOOL IOemBase::DrvQueryAdvanceWidths(
    DHPDEV      dhpdev,
    FONTOBJ    *pfo,
    ULONG       iMode,
    HGLYPH     *phg,
    PVOID       pvWidths,
    ULONG       cGlyphs )
{
    DBG_TRACE_OBJ( "IOemBase::DrvQueryAdvancedWidths" );
    PFN_DrvQueryAdvanceWidths pfn = PFN_DrvQueryAdvanceWidths( getPuntFunc( INDEX_DrvQueryAdvanceWidths) );
    if( !pfn ) { return FALSE; }
    return pfn( dhpdev, pfo, iMode, phg, pvWidths, cGlyphs );
}

ULONG IOemBase::DrvFontManagement(
    SURFOBJ    *pso,
    FONTOBJ    *pfo,
    ULONG       iMode,
    ULONG       cjIn,
    PVOID       pvIn,
    ULONG       cjOut,
    PVOID       pvOut )
{
    DBG_TRACE_OBJ( "IOemBase::DrvFontMangement" );
    PFN_DrvFontManagement pfn = PFN_DrvFontManagement( getPuntFunc( INDEX_DrvFontManagement ) );
    if( !pfn ) { return ULONG(-1); } // -1 is one of many failure choices.
    return pfn( pso, pfo, iMode, cjIn, pvIn, cjOut, pvOut );
}

ULONG IOemBase::DrvGetGlyphMode( DHPDEV dhpdev, FONTOBJ *pfo )
{
    DBG_TRACE_OBJ( "IOemBase::DrvGetGlyphMode" );
    PFN_DrvGetGlyphMode pfn = PFN_DrvGetGlyphMode( getPuntFunc( INDEX_DrvGetGlyphMode ) );
    if( !pfn ) { return FO_GLYPHBITS; } // Appears to be the best choice. JBTODO
    return pfn( dhpdev, pfo );
}

ULONG IOemBase::DrvDitherColor(
    DHPDEV  dhpdev,
    ULONG   iMode,
    ULONG   rgbColor,
    ULONG  *pulDither )
{
    DBG_TRACE_OBJ( "IOemBase::DrvDitherColor" );
    PFN_DrvDitherColor pfn = PFN_DrvDitherColor( getPuntFunc( INDEX_DrvDitherColor ) );
    if( !pfn ) { return DCR_SOLID; } // Appears to be the best choice. JBTODO
    return pfn( dhpdev, iMode, rgbColor, pulDither );
}

BOOL IOemBase::DrvLineTo(
    SURFOBJ    *pso,
    CLIPOBJ    *pco,
    BRUSHOBJ   *pbo,
    LONG        x1,
    LONG        y1,
    LONG        x2,
    LONG        y2,
    RECTL      *prclBounds,
    MIX         mix )
{
    DBG_TRACE_OBJ( "IOemBase::DrvLineTo" );
    PFN_DrvLineTo pfn = PFN_DrvLineTo( getPuntFunc( INDEX_DrvLineTo ) );
    if( !pfn ) { return FALSE; }
    return pfn( pso, pco, pbo, x1, y1, x2, y2, prclBounds, mix );
}

BOOL IOemBase::DrvStretchBltROP(
    SURFOBJ         *psoDest,
    SURFOBJ         *psoSrc,
    SURFOBJ         *psoMask,
    CLIPOBJ         *pco,
    XLATEOBJ        *pxlo,
    COLORADJUSTMENT *pca,
    POINTL          *pptlHTOrg,
    RECTL           *prclDest,
    RECTL           *prclSrc,
    POINTL          *pptlMask,
    ULONG            iMode,
    BRUSHOBJ        *pbo,
    ROP4             rop4 )
{
    DBG_TRACE_OBJ( "IOemBase::DrvStretchBltROP" );
    PFN_DrvStretchBltROP pfn = PFN_DrvStretchBltROP( getPuntFunc( INDEX_DrvStretchBltROP ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoDest, psoSrc, psoMask, pco, pxlo, pca, pptlHTOrg, prclDest, prclSrc, pptlMask, iMode, pbo, rop4 );
}

BOOL IOemBase::DrvPlgBlt(
    SURFOBJ         *psoDst,
    SURFOBJ         *psoSrc,
    SURFOBJ         *psoMask,
    CLIPOBJ         *pco,
    XLATEOBJ        *pxlo,
    COLORADJUSTMENT *pca,
    POINTL          *pptlBrushOrg,
    POINTFIX        *pptfixDest,
    RECTL           *prclSrc,
    POINTL          *pptlMask,
    ULONG           iMode )
{
    DBG_TRACE_OBJ( "IOemBase::DrvPlgBlt" );
    PFN_DrvPlgBlt pfn = PFN_DrvPlgBlt( getPuntFunc( INDEX_DrvPlgBlt ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoDst, psoSrc, psoMask, pco, pxlo, pca, pptlBrushOrg, pptfixDest, prclSrc, pptlMask, iMode );
}

BOOL IOemBase::DrvAlphaBlend(
    SURFOBJ    *psoDest,
    SURFOBJ    *psoSrc,
    CLIPOBJ    *pco,
    XLATEOBJ   *pxlo,
    RECTL      *prclDest,
    RECTL      *prclSrc,
    BLENDOBJ   *pBlendObj )
{
    DBG_TRACE_OBJ( "IOemBase::DrvAlphaBlend" );
    PFN_DrvAlphaBlend pfn = PFN_DrvAlphaBlend( getPuntFunc( INDEX_DrvAlphaBlend ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoDest, psoSrc, pco, pxlo, prclDest, prclSrc, pBlendObj );
}

BOOL IOemBase::DrvGradientFill(
    SURFOBJ    *psoDest,
    CLIPOBJ    *pco,
    XLATEOBJ   *pxlo,
    TRIVERTEX  *pVertex,
    ULONG       nVertex,
    PVOID       pMesh,
    ULONG       nMesh,
    RECTL      *prclExtents,
    POINTL     *pptlDitherOrg,
    ULONG       ulMode )
{
    DBG_TRACE_OBJ( "IOemBase::DrvGradientFill" );
    PFN_DrvGradientFill pfn = PFN_DrvGradientFill( getPuntFunc( INDEX_DrvGradientFill ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoDest, pco, pxlo, pVertex, nVertex, pMesh, nMesh, prclExtents, pptlDitherOrg, ulMode );
}

BOOL IOemBase::DrvTransparentBlt(
    SURFOBJ    *psoDst,
    SURFOBJ    *psoSrc,
    CLIPOBJ    *pco,
    XLATEOBJ   *pxlo,
    RECTL      *prclDst,
    RECTL      *prclSrc,
    ULONG      iTransColor,
    ULONG      ulReserved )
{
    DBG_TRACE_OBJ( "IOemBase::DrvTransparentBlt" );
    PFN_DrvTransparentBlt pfn = PFN_DrvTransparentBlt( getPuntFunc( INDEX_DrvTransparentBlt ) );
    if( !pfn ) { return FALSE; }
    return pfn( psoDst, psoSrc, pco, pxlo, prclDst, prclSrc, iTransColor, ulReserved );
}


//////////////////////////////////////////////////////////////////////////
//
// CustomGdiSurface
//
//////////////////////////////////////////////////////////////////////////

void CustomGdiSurface::
Create( SURFOBJ * pSrc, const size_t hRes, const size_t vRes )
{
    DBG_TRACE_OBJ( "CustomGdiSurface::Create" );

    const SIZEL sizl = { hRes, vRes };

    Destroy();

    // lWidth takes into account DWORD aligning.
    LONG lWidth = LONG( ((sizl.cx + 31) & ~ULONG(31)) / 8 );

    ULONG iFormat = BMF_1BPP;

    FLONG fl = (BMF_NOZEROINIT | BMF_TOPDOWN);

    // JBTODO -- setting pvBits to NULL makes EngCreateBitmap allocate
    // memory for us.  What I don't know is how this memory gets deleted.
    const PVOID pvBits = NULL;

    m_hBmp = EngCreateBitmap( sizl, lWidth, iFormat, fl, pvBits );

    if( !m_hBmp )
    {
        DBG_ERROR( "Failed to create Bitmap." );
        return;
    }

    PDEVOBJ pdevobj = reinterpret_cast<PDEVOBJ>(pSrc->dhpdev);

    HSURF hsurf   = reinterpret_cast<HSURF>(m_hBmp);
    HDEV  hdev    = reinterpret_cast<HDEV>(pdevobj->hEngine);
    FLONG flHooks = 0; // JBTODO: Not sure what this does.

    if( !EngAssociateSurface( hsurf, hdev, flHooks ) )
    {
        DBG_ERROR( "Failed to associate surface." );
        return;
    }

    m_pSurf = EngLockSurface( hsurf );

    if( !m_pSurf )
    {
        DBG_ERROR( "Failed to make bitmap into surface." );
        EngDeleteSurface( hsurf );
        hsurf = NULL;
    }
}

void CustomGdiSurface::Destroy( void )
{
    DBG_TRACE_OBJ( "CustomGdiSurface::Destroy" );

    if( m_hBmp )
    {
        if( m_pSurf )
        {
            EngUnlockSurface( m_pSurf );
            m_pSurf = NULL;
        }

        EngDeleteSurface( reinterpret_cast<HSURF>(m_hBmp) );
        m_hBmp = NULL;
    }
}
