#include "stdafx.h"

#include "Debug.h"

#include "OemClassFactory.h"

#include "OemTiff.h"
#include "OemTurboSeg.h"

__stdcall OemClassFactory::OemClassFactory( void )
: m_refCount(1)
{
    DBG_TRACE_OBJ( "OemClassFactory::OemClassFactory" );
}

ULONG __stdcall OemClassFactory::AddRef( void )
{
    DBG_TRACE_OBJ( "OemClassFactory::AddRef" );
    return InterlockedIncrement( &m_refCount );
}

ULONG __stdcall OemClassFactory::Release( void ) 
{
    DBG_TRACE_OBJ( "OemClassFactory::Release" );

    if( InterlockedDecrement( &m_refCount ) == 0 )
    {
        delete this;
        
        return 0;
    }
    
    return m_refCount;
}

HRESULT OemClassFactory::QueryInterface( const IID & iid, void ** ppvObj )
{
    DBG_TRACE_OBJ( "OemClassFactory::QueryInterface" );

    if( (iid == IID_IUnknown     ) ||
        (iid == IID_IClassFactory)  )
    {
        this->AddRef();
        *ppvObj = this;
        return S_OK;
    }

    *ppvObj = NULL;
    
    return E_NOINTERFACE;
}

HRESULT OemClassFactory::
CreateInstance( IUnknown * pUnk, const IID & riid, void ** ppvObj )
{
    DBG_TRACE_OBJ( "OemClassFactory::CreateInstance" );

    // We do not support any kind of aggregation.
    if( pUnk != NULL )
    {
        return CLASS_E_NOAGGREGATION;
    }

    IOemBase * pOemBase = new IOemTurboSeg;
    //IOemBase * pOemBase = new IOemTiff;

    if( !pOemBase )
    {
        return E_OUTOFMEMORY;
    }
   
    // Get the requested interface.
    HRESULT hr = pOemBase->QueryInterface( riid, ppvObj );

    // Release the IUnknown pointer.  (Re
    // (If QueryInterface failed, component will delete itself.)
    pOemBase->Release();

    return hr;
}


HRESULT __stdcall  OemClassFactory::LockServer( BOOL fLock )
{
    DBG_TRACE_OBJ( "OemClassFactory::LockServer" );

    HRESULT hRes = CoLockObjectExternal( this, fLock, false );
    
    switch( hRes )
    {
        case S_OK         :
        case E_OUTOFMEMORY:
        case E_UNEXPECTED :  break;
        
        default: hRes = E_FAIL; break;
    }

    return hRes;
}
