#include "stdafx.h"
#include "OemPlane.h"
#include "debug.h"
#include "misc.h"


IOemPlane::IOemPlane( void )
{
    DBG_TRACE_OBJ( "IOemPlane::IOemPlane" );
}


/////////////////////////////////////////////////////////////////////////
// IOemBase methods
//////////////////////////////////////////////////////////////////////////

HRESULT IOemPlane::initFrom( IOemBase * old )
{
    DBG_TRACE( "IOemPlane::initFrom" );

    this->IOemBase::initFrom( old );

    return E_FAIL;
}


BOOL IOemPlane::DrvStartPage( SURFOBJ *pso )
{
    DBG_TRACE_OBJ( "IOemPlane::DrvStartPage" );
    DBG_WATCH( "Page number", m_curPage );

    IOemBase::DrvStartPage( pso );
    
    PDEVOBJ pdevobj = reinterpret_cast<PDEVOBJ>(pso->dhpdev);

    // Set the working bitmap to white by default.
    memset( pso->pvScan0, 0xff, pso->sizlBitmap.cy * pso->lDelta );

	return TRUE;
}

BOOL IOemPlane::DrvNextBand( SURFOBJ * pso, POINTL * pptl )
{
    DBG_TRACE_OBJ( "IOemPlane::DrvNextBand" );

    IOemBase::DrvNextBand( pso, pptl );

    if( !pso )
    {
        assert( pso );
        return FALSE;
    }

    if( !pptl )
    {
        assert( pptl );
        return FALSE;
    }


    PDEVOBJ pdevobj = reinterpret_cast<PDEVOBJ>(pso->dhpdev);

    const size_t rows = bandRows( pso );

    // All that the fancy math does here is get a byte aligned value for
    // the row size.  This may be greater than sizlBitmap.cx in some
    // cases.  It may also be less than lDelta.
    const size_t bytesInRow = (((bitsPixel() * pso->sizlBitmap.cx) + 7) & ~ULONG(7)) / 8;

    // Get a properly typed pointer to our bitmap.
    unsigned char * base = static_cast<unsigned char *>(pso->pvScan0);

    // Write out byte-aligned data one scan line at a time.
    for( size_t row = 0; row < rows; ++row )
    {
        writeData( pdevobj, base + (row*pso->lDelta), bytesInRow );
    }

    // If there is work left to be done, overrun will be negative.
    if( bandOverrun(pso) < 0 )
    {
        // In that case we must set m_bandOrigin to point to the top of
        // the undrawn portion of the page.
        m_bandOrigin.y += pso->sizlBitmap.cy;

        // Set the surface back to flat white for the next band.
        memset( pso->pvScan0, 0xff, pso->lDelta * pso->sizlBitmap.cy );
    }
    else
    {    
        // Setting both x and y to -1 is how we tell our caller there is
        // nothing more to be done.
        m_bandOrigin.x = -1;
        m_bandOrigin.y = -1;
    }

    *pptl = m_bandOrigin;

    return TRUE;
}

BOOL IOemPlane::DrvSendPage( SURFOBJ *pso )
{
    DBG_TRACE_OBJ( "IOemPlane::DrvSendPage" );

    IOemBase::DrvSendPage( pso );

    // The documentation seems to say that DrvSendPage() gets
    // called even when banding is turned on, however it does not
    // seem to get called in realitiy.  In any case, we check for
    // it just to make sure.
    if( m_isBanded )
    {
        return TRUE;
    }

    PDEVOBJ pdevobj = reinterpret_cast<PDEVOBJ>(pso->dhpdev);

    // All that the fancy math does here is get a byte aligned value for
    // the row size.  This may be greater than sizlBitmap.cx in some
    // cases.  It may also be less than lDelta.
    const size_t bytesInRow = (((bitsPixel() * pso->sizlBitmap.cx) + 7) & ~ULONG(7)) / 8;

    unsigned char * base = static_cast<unsigned char *>(pso->pvScan0);

    // Write out byte-aligned data one scan line at a time.

    for( size_t row = 0; row < pso->sizlBitmap.cy; ++row )
    {
        writeData( pdevobj, base + (row*pso->lDelta), bytesInRow );
    }

    return TRUE;
}
