#ifndef OemPlane_h_INCLUDED
#define OemPlane_h_INCLUDED

#include "OemBase.h"


/*

  IOemPlane renders everything on a GDI created Device Independent bitmap
  and then sends the data from said bitmap one scanline at a time to its own
  writeData() method.  This is a very nearly-working implementation of a
  simple printer driver.  All you must do is add appropriate header
  information and implement writeData() to call DrvWriteSpoolBuf() and  you
  would have a working bitmap or uncompressed TIFF printer driver.

*/

class IOemPlane : public IOemBase
{
public:

    IOemPlane( void );
       
    //
    // DDI hook methods.
    // 

    virtual BOOL DrvStartPage   ( SURFOBJ * ); 
    virtual BOOL DrvNextBand    ( SURFOBJ *, POINTL * );
    virtual BOOL DrvSendPage    ( SURFOBJ * );

protected:

    
    // You must send exactly one full byte-aligned scanline to each call
    // to writeData() (although it is unlikely you would ever call it.)
    //
    // When you derive a class, you override this method to get the data
    // as it is written out.  It takes a pointer to the data you want to write
    // and the size of the buffer as arguemts.  It returns the number of bytes
    // writen.  We implement a default version to help us build, but it always
    // fails.  (This method could be pure-virtual.)
    virtual size_t writeData( PDEVOBJ, unsigned char *, size_t )
    {
        return 0;
    }

    virtual HRESULT initFrom( IOemBase * );
    
private:

};

#endif
