#include "stdafx.h"

#include "OemTiff.h"
#include "debug.h"
#include "misc.h"

IOemTiff::IOemTiff( void )
:
 m_tiff( "c:\\spool.tiff" )
{
    DBG_TRACE_OBJ( "IOemTiff::IOemTiff" );
}

HRESULT IOemTiff::initFrom( IOemBase * old )
{
    DBG_TRACE( "IOemTiff::initFrom" );

    IOemTiff * oldThis = dynamic_cast<IOemTiff *>(old);

    if( !oldThis )
    {
        return E_FAIL;
    }

    this->IOemPlane::initFrom( old );

    this->m_tiff = oldThis->m_tiff;

    return S_OK;
}


size_t IOemTiff::writeData( PDEVOBJ pdevobj, unsigned char * buf, size_t count )
{
    m_tiff.writeSpoolBuf( buf, count );
    return count;
}

/////////////////////////////////////////////////////////////////////////
// IOemBase methods
//////////////////////////////////////////////////////////////////////////

BOOL IOemTiff::DrvStartDoc( SURFOBJ *pso, PWSTR pwszDocName, DWORD dwJobId )
{
    DBG_TRACE_OBJ( "IOemTiff::DrvStartDoc" );

    DBG_WATCH( "pwszDocName", makeStr( pwszDocName ) );
    DBG_WATCH( "dwJobId", dwJobId );

    // Fulfill our requirment to the base class
    
    if( IOemPlane::DrvStartDoc( pso, pwszDocName, dwJobId ) )
    {
        m_tiff.init( m_gdiInfo.ulHorzRes, m_gdiInfo.ulVertRes, bitsPixel() );
    }

    return TRUE;
}

BOOL IOemTiff::DrvStartPage( SURFOBJ *pso )
{
    DBG_TRACE_OBJ( "IOemTiff::DrvStartPage" );

    IOemPlane::DrvStartPage( pso );

    // If we are not on the first page, we send he directory for the last page. 
    // (This is because we need some informatio about the current page to
    // finish the IFD.)
    if( m_curPage != 1 )
    {
        m_tiff.sendLastPageIfd( m_gdiInfo.ulHorzRes, m_gdiInfo.ulVertRes, bitsPixel() );
    }

    return TRUE;
}

BOOL IOemTiff::DrvEndDoc( SURFOBJ *pso, FLONG fl )
{
    DBG_TRACE_OBJ( "IOemTiff::DrvEndDoc" );

    IOemPlane::DrvEndDoc( pso, fl );

    // Write out the last IFD.
    m_tiff.sendVeryLastPageIfd();
    
    return TRUE;

}