#include "StdAfx.h"
#include "TfxOutput.h"
#include "debug.h"

 ITfxOutput::ITfxOutput ( void ) : 
m_instanceID (0),
m_outIoDoc (0),
m_outIoPage (0),
m_outIoSelector (0),
m_outDocType (DOCTYPE_TIFFFX_M1),
m_sbh (NULL)
{
  IO_InitInstance(0, &m_instanceID); 
}

 ITfxOutput::~ITfxOutput ( void )
{
  IO_EndInstance(m_instanceID); 
}

BOOL 
 ITfxOutput::TfxStartDoc ()
{
  LONG status = 0;
  CHAR version = 0;
  ULONG options = 0;
  
  // santity check
  if (!m_instanceID) return FALSE;

  // create the spool buffer feeder
  m_sbh  = new TiffSpoolBufferFeeder;

	// only open new document if no current doucment is open
    if (m_outIoDoc == 0)
      {
        status = IO_CreateDocument(m_instanceID, m_outDocType, 
				   (LONG) m_sbh, version, 
				   options, mrcioRead, 
				   mrcioWrite, mrcioSeek, 
				   mrcioFileSize, mrcioTell, 
				   &m_outIoDoc);
      }
    else
      {
        status = 1;
      }
  
  if (status != 0)
    return FALSE;
  
  return TRUE;
}


BOOL 
 ITfxOutput::TfxStartPage ( PDEVOBJ pdevobj,
		IPrintOemDriverUni* pOEMHelp )
{
  
  LONG status = 0;
  
  // sanity check
    if (!m_outIoDoc) return FALSE;
  
  // set member variables
  m_pdevobj = pdevobj; 
  m_pOEMHelp = pOEMHelp;

  // open page only if no page is currently open
    if (m_outIoPage == 0)
      {
        status = IO_AddPageStart(m_outIoDoc,
				 -1 /* Append to the end */,
				 &m_outIoPage);
      }
    else
      {
        status = 1;
      }
  
  if (status != 0)
    return FALSE;
  
  return TRUE;
}

BOOL 
 ITfxOutput::TfxAddImage (ULONG layer, ULONG xpos, ULONG ypos, 
			  ULONG width, ULONG height, ULONG xres, ULONG yres, 
			  unsigned char* buffer, ULONG bufferSize, IO_COMPRESSION comp)  
{
  
  IO_COLORSPACE colorSpace = IO_COLORSPACE_DEFAULT;
  IO_COLORTYPE colorType;
  ULONG outIoImage;
  
  LONG status;
  ULONG quality = 75; //default JPEG quality
    
    DBG_WATCH( "Add Image :: layer", layer );
  DBG_WATCH( "Add Image :: xpos", xpos );
  DBG_WATCH( "Add Image :: ypos", ypos );
  DBG_WATCH( "Add Image :: width", width );
  DBG_WATCH( "Add Image :: height", height );
  DBG_WATCH( "Add Image :: bufferSize", bufferSize );
  DBG_WATCH( "Add Image :: comp", comp );
  
  // sanity check
    if (!m_outIoPage) return FALSE;
  
  if (layer % 2 == 0)
    {
      // any even layer image is the selector at 1 bpp
        colorType = IO_COLORTYPE_BINARY;
    }
  else
    {
      // any odd image layer is at 24 bpp
        colorType = IO_COLORTYPE_COLOR24;
		// our JPEG encoder spits out YCbCr at the moment ....
		colorSpace = IO_COLORSPACE_YCBCR;
    }
  
  status = IO_AddImage(m_outIoPage, layer, colorType, xpos, ypos,
		       xres, yres, width, height, height, NULL, &outIoImage);
  DBG_WATCH( "status of IO_AddImage", status );
  
  if (status == 0)
    {
      status = IO_SetCompression (outIoImage, comp);
    }
  
  IO_COMPRESSION io_comp = comp;
  if (status == 0)
    {
      status = IO_ImageWriteStart(outIoImage, 0, 
				  &io_comp, &quality, &colorSpace);
      DBG_WATCH( "status of IO_ImageWriteStart", status );
      DBG_WATCH( "Add Image :: io_comp", io_comp );
    }
  if (comp != io_comp)
    {
      status = 1;
    }
  if (status == 0)
    {
      ULONG bytesWritten = 0;
      status = IO_CompressedWriteStrip (outIoImage, 0, height,
					buffer, bufferSize, &bytesWritten);
      DBG_WATCH( "status of IO_CompressedWriteStrip", status );
      
      if (bytesWritten != bufferSize) status = 1;
    }
  if (status == 0)
    {
      status = IO_ImageWriteFinish (outIoImage);
      DBG_WATCH( "status of IO_ImageWriteFinish", status );
    }
  
  
  if (status != 0)
    return FALSE;
  
  return TRUE;
}


BOOL 
 ITfxOutput::TfxWriteFirstSelectorStrip (ULONG width, ULONG height, 
					 ULONG stripHeight, ULONG xres, ULONG yres, 
					 unsigned char* buffer, ULONG bufferSize, IO_COMPRESSION comp)  
{
  
  IO_COLORSPACE colorSpace = IO_COLORSPACE_DEFAULT;
  IO_COLORTYPE colorType = IO_COLORTYPE_BINARY;
  const ULONG layer = 2; // the main selector is always layer 2
  ULONG quality = 75; //default JPEG quality
  
  LONG status;
  
  DBG_WATCH( "FirstSelectorStrip :: width", width );
  DBG_WATCH( "FirstSelectorStrip :: height", height );
  DBG_WATCH( "FirstSelectorStrip :: stripHeight", stripHeight );
  DBG_WATCH( "FirstSelectorStrip :: bufferSize", bufferSize );
  DBG_WATCH( "FirstSelectorStrip :: comp", comp );
  
  // sanity check
  if (!m_outIoPage) return FALSE;
  
  // store usefull variables
  m_xRes = xres;
  m_yRes = yres;
  m_stripHeight = stripHeight;
  m_width = width;
  m_height = height;
  m_selComp = comp;
  m_stripCount = 0;
  
  // mark the last byte written previously. Those bytes 
  // are for the previous page
  m_sbh->markCurrentPosition ();

  status = IO_AddImage(m_outIoPage, layer, colorType, 0, 0,
		       xres, yres, width, height, 
		       stripHeight, NULL, &m_outIoSelector);
  DBG_WATCH( "status of IO_AddImage", status );
  
  if (status == 0)
    {
      status = IO_SetCompression (m_outIoSelector, m_selComp);
      
      IO_COMPRESSION io_comp = m_selComp;
      if (status == 0)
	{
	  status = IO_ImageWriteStart(m_outIoSelector, 0, 
				      &io_comp, &quality, &colorSpace);
	  DBG_WATCH( "status of IO_ImageWriteStart", status );
	  DBG_WATCH( "FirstSelectorStrip :: io_comp", io_comp );
	}
      if (m_selComp != io_comp)
	{
	  status = 1;
	}
      if (status == 0)
	{
	  // now the IFD of the previous page is updated
	  // write out the previous page
      BOOL wrote_spool = m_sbh->sendMarkedToSpoolBuf (m_pdevobj, m_pOEMHelp);
	  if (!wrote_spool) return FALSE;

	  return TfxWriteNextSelectorStrip (buffer, bufferSize);
	}
      else
	{
	  return FALSE;
	}
    }

  if (status != 0)
    return FALSE;
  
  return TRUE;
}

BOOL ITfxOutput::TfxWriteNextSelectorStrip (unsigned char* buffer,
					    ULONG bufferSize)
{
  
    LONG status;
  
  IO_COLORSPACE colorSpace = IO_COLORSPACE_DEFAULT;
  IO_COLORTYPE colorType = IO_COLORTYPE_BINARY;

  ULONG bytesWritten = 0;
  status = IO_CompressedWriteStrip (m_outIoSelector, 
                    m_stripCount * m_stripHeight, m_stripHeight,
					buffer, bufferSize, &bytesWritten);
  DBG_WATCH( "status of IO_CompressedWriteStrip", status );
      
  if (bytesWritten != bufferSize)
	{
	  DBG_WATCH( "problem writing Strip", bytesWritten );
	  DBG_WATCH( "problem writing Strip", bufferSize );
	}
  m_stripCount++;

  if (status == 0 && m_height <= m_stripCount * m_stripHeight)
    {
      status = IO_ImageWriteFinish (m_outIoSelector);
      DBG_WATCH( "status of IO_ImageWriteFinish", status );
      m_outIoSelector = 0;
    }
  
  
  if (status != 0)
    return FALSE;
  
  DBG_WATCH( "TfxWriteNextSelectorStrip returning TRUE ", status );
  return TRUE;
}

BOOL 
 ITfxOutput::TfxEndPage ()
{
  LONG status = 0;
  
  // sanity check
    if (!m_outIoPage) return FALSE;
  
  status = IO_AddPageFinish(m_outIoPage);
  m_outIoPage = 0;
  
  if (status != 0)
    return FALSE;
  
  return TRUE;
}


BOOL 
 ITfxOutput::TfxEndDoc ()
{
  
  ULONG status = 0;
  BOOL ret_val = FALSE;

  // sanity check
  if (!m_outIoDoc) return ret_val;
  
  status = IO_CloseDocument (m_outIoDoc);
  m_outIoDoc = 0;
  
  // send whatever is left to the spool buffer
  if (status == 0)
  {
	ret_val = m_sbh->sendToSpoolBuf (m_pdevobj, m_pOEMHelp);
  }

  delete m_sbh;
  m_sbh = NULL;
  
  return ret_val;
  
}

