/*************************************************************************
**
**  DigiPaper Codebase
**    
**  Pedro Felzenszwalb, January 1999
**  most of the code was stolen from tc2ps
**  
**  26.7.2000 / dhtesq.
**  Resurrection...
**
**	15.3.2001 / dhtesq.
**	Re-resurrection for tfx...
**
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/DgpPdf/DgpPdf.c,v 1.1 2001/03/20 01:36:44 dtracy Exp $
 *
 */

/*
 * This program reads in a tfx file and converts it to PDF.
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "dgp_state.h"
#include "dgp_err_printers.h"
#include "DgpPdf_util.h"
#include "dgp_cl.h"

#ifdef _WIN32
#include <string.h>
#endif

/*
#ifndef _WIN32
#include "undecl-stdio.h"
#endif
*/

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/DgpPdf/DgpPdf.c,v 1.1 2001/03/20 01:36:44 dtracy Exp $");


/*
 * Default values for command line arguments.
 */
#define DEF_OUTFILENAME "foo.tfx.pdf"
#define DEF_VERBOSE     DGP_FALSE
#define DEF_RES         DGP_PS_DPI
#define DEF_RES_STR     DGP_PS_DPI_STR
#define DEF_PPH         DGP_FALSE
#define DEF_PPW         DGP_FALSE
#define DEF_LFONTS      DGP_FALSE
#define DEF_ERRFILENAME "stderr"
#define DEF_NOE         DGP_FALSE

typedef struct {
  Dgp_Int32 dpi;
  Dgp_Bool verbose;
  //Dgp_Bool pph;
  //Dgp_Bool ppw;
  Dgp_Bool lfonts;
  Dgp_Int8 *out_filename;
  Dgp_Int8 *err_filename;
  Dgp_Bool noe;
} CLOpts;

static CLOpts cl_opts = {DEF_RES, DEF_VERBOSE, /*DEF_PPH, DEF_PPW,*/ DEF_LFONTS, DEF_OUTFILENAME, DEF_ERRFILENAME, DEF_NOE};

// Displayed when showing usage information.
static const Dgp_Int8 *cl_header = "Usage: DgpPdf [options] foo.tfx";

/*
 * The definition of the command line options.
 */
static DgpCl_Option cl_cmdln_options[] = {
  DGPCL_HELP_CMD,
  DGPCL_RES_CMD(cl_opts.dpi, DEF_RES_STR),
  DGPCL_VERBOSE_CMD(cl_opts.verbose),
  //DGPCL_NEW_FLAG_CMD("ppw", cl_opts.ppw, "Use most popular page width in document"),
  //DGPCL_NEW_FLAG_CMD("pph", cl_opts.pph, "Use most popular page height in document"),
  DGPCL_NEW_FLAG_CMD("lfonts", cl_opts.lfonts, "Use local fonts for per-page tokens"),
  DGPCL_OUTFNAME_CMD(cl_opts.out_filename, DEF_OUTFILENAME),
  DGPCL_NEW_STRING_CMD("e", cl_opts.err_filename, "Set error file name", DEF_ERRFILENAME),
  DGPCL_NEW_FLAG_CMD("noe", cl_opts.noe, "Stifle error output; overrides -e"),
  {NULL}
};

/*************************************************************************/
// An argument validator...
static ErrCode
validateArgs(Dgp_State *dgp_state,
             Dgp_Int32 unparsed_argc, 
             Dgp_Int8 **unparsed_argv,
             FILE *err_out,
             void *validate_cookie)
{
  DGP_PROLOGUE("validateArgs()", DGP_PERRSTATE);
  CLOpts *my_opts = (CLOpts *)validate_cookie;

  unparsed_argc = unparsed_argc;
  unparsed_argv = unparsed_argv;

  if(my_opts->dpi > 2400 || my_opts->dpi < 0) {
    DGPCL_PRINT_ERRF0(err_out, "Valid resolutions are 0 < res < 2400...");
    DGP_ERR(cErrInvalidCmdLine);
  }

  NORMAL_RETURN;
}
/*************************************************************************/

/*************************************************************************/
static void
initializeState(DgpPdfState *state)
{
  state->dpi = 0;
  state->outfile = NULL;

  state->use_local_fonts = DGP_FALSE;
  state->use_pop_pg_height = DGP_FALSE;
  state->use_pop_pg_width = DGP_FALSE;
  state->verbose = DGP_FALSE;

  state->tok_hash = NULL;
  state->toks_pg = NULL;
  state->residual_toks = NULL;
  state->num_residual_toks = 0;
  state->num_uniqids = 0;

  state->zephDoc = NULL;
  state->cache = NULL;
  state->dictionary = NULL;
  state->pos_block = NULL;
  state->entry = NULL;

  /*
  state->dictblock = NULL;
  state->extensions = NULL;
  state->num_extensions = NULL;
  state->num_toks = NULL;
  state->toks = NULL;
  */
  return;
}
/*************************************************************************/

/*************************************************************************/
Dgp_Int32
main(Dgp_Int32 argc, Dgp_Int8 **argv)
    {
  DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);

  Dgp_Int8      *in_filename = NULL, *errfname = NULL, *minus_path = NULL;
  Dgp_Bool      no_err_out = DGP_FALSE;
  DgpPdfState   state;
  Dgp_State     *dgp_state = NULL;
  uds           docUds;
#ifdef _WIN32
  Dgp_Int8              filesep = '\\';
#else
  Dgp_Int8              filesep = '/';
#endif

  if(argc == 1){
      DgpCl_printHelp(stderr, cl_header, cl_cmdln_options);
      MAIN_NORMAL_RETURN;
  }
  docUds.fp = NULL;

  DGP_ERR_PASS(Dgp_initDgpState(&(dgp_state), NULL, NULL, stderr, Dgp_fileErrPrinter, DGP_TRUE, DGP_FALSE));

  /* Init options struct. */
  initializeState(&state);

  DGP_ERR_CHECK_NO_MSG(DgpCl_parse(dgp_state, &argc, argv, cl_cmdln_options, stderr, cl_header, validateArgs, &cl_opts)){
      if(DGP_ERROR_IS(cErrInvalidCmdLine)){
          MAIN_NORMAL_RETURN;
      }
  }
  in_filename = argv[0];
  state.verbose = cl_opts.verbose;
  state.dpi = cl_opts.dpi;
  state.use_local_fonts = cl_opts.lfonts;
  //state.use_pop_pg_height = cl_opts.pph;
  //state.use_pop_pg_width = cl_opts.ppw;
  errfname = cl_opts.err_filename;
  no_err_out = cl_opts.noe;

  if((cl_opts.out_filename == NULL) || (cl_cmdln_options[4].found == DGP_FALSE)) {
      // The user has not specified the output filename...
      DGP_CHECK_ALLOC_CLEANUP(state.out_filename = (Dgp_Int8 *)DGP_MALLOC((strlen(in_filename) + 5) * sizeof(Dgp_Int8)));
      state.out_filename[0] = '\0';      
      sprintf(state.out_filename, "%s.pdf", in_filename);
  } else {
      // Copy the command line string, unless stdout has been specified.
      if(!strcmp(cl_opts.out_filename, "stdout")){
          state.out_filename = NULL;
      } else {
          // -o argument could be a path or a destination filename.
          // If it ends in a filesep, use it as a prepended path.
          if(cl_opts.out_filename[strlen(cl_opts.out_filename) - 1] == filesep){
              // Prepend path...
              minus_path = strrchr(in_filename, filesep);
              if(minus_path == NULL){
                  minus_path = in_filename;
              } else {
                  minus_path += 1;
              }
              DGP_CHECK_ALLOC_CLEANUP(state.out_filename = 
                  (Dgp_Int8 *)DGP_MALLOC((strlen(cl_opts.out_filename) + strlen(minus_path) + 6) 
                  * sizeof(Dgp_Int8)));
              strcpy(state.out_filename, cl_opts.out_filename);
              state.out_filename = strcat(state.out_filename, minus_path);
              state.out_filename = strcat(state.out_filename, ".pdf");
          } else {
              // -o argument is destination filename, no manipulation required.
              DGP_CHECK_ALLOC_CLEANUP(state.out_filename = (Dgp_Int8 *)DGP_MALLOC((strlen(cl_opts.out_filename) + 1) * sizeof(Dgp_Int8)));
              strcpy(state.out_filename, cl_opts.out_filename);
          }          
      }
  }

  docUds.fseek  = (fseek_func)fseek;
  docUds.ftell  = (ftell_func)ftell;
  docUds.fread  = (fread_func)fread;
  docUds.fwrite = (fwrite_func)fwrite;
  state.docUds = &docUds;

  /* Open all files needed. */
  docUds.fp = fopen(in_filename, "rb");
  DGP_ERR_GEN_CLEANUP((!docUds.fp), cErrFileNotFound);

  if(state.out_filename){
      state.outfile = fopen(state.out_filename, "wb");
      DGP_ERR_GEN_CLEANUP((!state.outfile), cErrFileNotFound);
  } else {
      state.outfile = stdout;
  }

  if(!no_err_out){
    if(strcmp(errfname, "stderr")) {
      DGP_ERR_GEN_CLEANUPF1((!(state.errfile = fopen(errfname, "w"))), cErrFileNotFound,
	      "I can't open error log file \'%s\' for writing.", errfname);
      dgp_state->err_state->userCookie = state.errfile;
    } else {
      state.errfile = stderr;
    }
  } else {
    dgp_state->err_state->userCookie = NULL;
  }

  /* Translate the tfx file... */
  DGP_ERR_PASS_CLEANUP(DgpPdf(&state, dgp_state));

  /* Close all files. */
  DGP_ERR_GEN_CLEANUPF0((fclose(state.outfile) < 0), cErrUndefined, "I can't close the output file...");
  DGP_ERR_GEN_CLEANUPF0((fclose(docUds.fp) < 0), cErrUndefined, "I can't close the output file...");

  if((state.errfile) && (state.errfile != stderr)) {
    DGP_ERR_GEN_CLEANUPF0((fclose(state.errfile) < 0), cErrUndefined, "I can't close the error file.");
  }

  if(state.out_filename){
      DGP_FREE(state.out_filename);
  }
  Dgp_cleanupDgpState(&dgp_state);
  
  MAIN_NORMAL_RETURN;
  
ERR_CLEANUP:  
  if(state.out_filename){
      DGP_FREE(state.out_filename);
  }
  if(state.outfile){
      fclose(state.outfile);
  }
  if(docUds.fp){
      fclose(docUds.fp);
  }
  if(state.errfile && (state.errfile != stderr)){
      fclose(state.errfile);
  }
  if(dgp_state){
      Dgp_cleanupDgpState(&dgp_state);
  }
  MAIN_ERR_RETURN;
}
/*************************************************************************/
