/*************************************************************************
**
**  DigiPaper Codebase
**    
**  Pedro Felzenszwalb, January 1999
**  most of the code was stolen from tc2ps
**  
**	15.3.2001 / dhtesq.
**	Re-resurrection for tfx...
**
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/DgpPdf/DgpPdf_util.c,v 1.1 2001/03/20 01:36:44 dtracy Exp $
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "dgp_state.h"
#include "dgp_err_control.h"

#include "list.h"
#include "DgpPdf_util.h"
#include "DgpPdf_ops.h"
#include "encodings.h"

#include "zlib.h"

#include "image_defs.h"
#include "zephpage.h"
#include "zephprivate.h"        // Need private structures for xif handles...

#include "resolution_t.h"       // libcoord...
#include "libsupercache.h"
#include "sutokposblock.h"

#include "insertbox.h"          // libtc_util...

//#include "xiff.h"

/*
#ifndef _WIN32
#include "undecl-stdio.h"
#endif
*/

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/DgpPdf/DgpPdf_util.c,v 1.1 2001/03/20 01:36:44 dtracy Exp $");


/* adds an object to the xref table */
ErrCode
addXref(Dgp_Int32       obj, 
        Dgp_Int32       offset, 
        DgpPdfState     *state, 
        PdfCookie       *pdfCookie,
        Dgp_State       *dgp_state) {
    DGP_PROLOGUE("addXref()", DGP_PERRSTATE);

    state = state;

    if (pdfCookie->xrefPos == pdfCookie->xrefSize) {
      pdfCookie->xrefSize *= 2;
      DGP_CHECK_ALLOC_CLEANUP(pdfCookie->xref = DGP_REALLOC(pdfCookie->xref, pdfCookie->xrefSize * sizeof(xrefElt)));
    }
    pdfCookie->xref[pdfCookie->xrefPos].obj = obj;
    pdfCookie->xref[pdfCookie->xrefPos].offset = offset;
    pdfCookie->xrefPos++;
    
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/* Init a pdf cookie. */
ErrCode
initPdfCookie(DgpPdfState *state, PdfCookie *pdfCookie, Dgp_State *dgp_state) {
    DGP_PROLOGUE("initPdfCookie()", DGP_PERRSTATE);

    state = state;

    pdfCookie->offset = 0;
    pdfCookie->objCount = 1;
    DGP_CHECK_ALLOC_CLEANUP(pdfCookie->xref = DGP_MALLOC(sizeof(xrefElt) * XREF_SIZE));
    pdfCookie->xrefPos = 0;
    pdfCookie->xrefSize = XREF_SIZE;

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

void
cleanupPdfCookie(DgpPdfState *state, PdfCookie *pdfCookie, Dgp_State *dgp_state) {	
  state = state;
  DGP_FREE(pdfCookie->xref);
}

// (I found these preprocessed out while resurrecting -- dhtesq.)
#if 0
static void
DB_save_pim(DgpImage_Packed pim)
{
  static Dgp_Int32 id = 0;
  Dgp_Int8 str[256];

  sprintf(str, "packed_im%03d.pgm", id);
  imSave(pim, str);
  id++;
}

static void
DB_save_bitmap(Dgp_UInt32 *bmp, Dgp_Int32 w, Dgp_Int32 h, Dgp_Int32 lpl) 
{
  DgpImage_Packed pim = NULL;

  pim = pimWrap(bmp, w, h, lpl);

  DB_save_pim(pim);

  pimUnwrap(pim);
}

static void
DB_print_ulong(Dgp_UInt32 l)
{
  Dgp_Int32 i;
  Dgp_Int32 num_bits = sizeof(Dgp_UInt32) * BITS_UCHAR;
  Dgp_UInt32 mask = (1 << (num_bits - 1));

  for(i = 0; i < num_bits; i++) {
    if((i%8) == 0) printf(" ");
    printf("%d", ((l & mask) ? 1:0));
    mask = mask >> 1;
  }
}

static void
DB_print_uchar(UCHAR c)
{
  Dgp_Int32 i;
  Dgp_Int32 num_bits = BITS_UCHAR;
  UCHAR mask = (1 << (num_bits - 1));

  for(i = 0; i < num_bits; i++) {
    printf("%d", ((c & mask) ? 1:0));
    mask = mask >> 1;
  }
  printf(" ");
}


static void
DB_print_bitmap(UCHAR *bmp, Dgp_Int32 w, Dgp_Int32 h, Dgp_Int32 lpl) 
{
  Dgp_Int32 i;
  Dgp_Int32 buf_len = lpl * sizeof(Dgp_UInt32) * h;

  w=w;

  for(i = 0; i < buf_len; i++) {
    DB_print_uchar(bmp[i]);
    if(((i+1) % 4) == 0) printf("\n");
  }
}
#endif
///////////////////////////////////////////////////////////////////////

/* For both of these functions:
    if *cookie is true, e1 is a tokpos, otherwise it is a FontToken. */

// 1.8.2000 / dhtesq.
// The notion of a toknum doesn't seem to exist anymore.
// I'll call two tokens equal if they have the same address,
// and two tokposes equal if they point to the same token.
//
// Hashing function will remain just modulo DGP_SIZE_OF_HT.

/* Comparator for ftoken hash table. */
static Dgp_Bool
isequal_ft(void *cookie, DgpAdt_ht_elt e1, DgpAdt_ht_elt e2)
{
  if(*((Dgp_Bool *)cookie)) {
    //return((Dgp_Bool)(((tokpos *)e1)->toknum == ((FontToken *)e2)->token->toknum));
    return((Dgp_Bool)(((tokpos *)e1)->tok == ((FontToken *)e2)->token));
  }
  //return((Dgp_Bool)(((FontToken *)e1)->token->toknum == ((FontToken *)e2)->token->toknum));
  return((Dgp_Bool)(((FontToken *)e1)->token == ((FontToken *)e2)->token));
}

/* Hash function for ftoken hash table. */
static Dgp_UInt16
do_hash_ft(void *cookie, DgpAdt_ht_elt e1)
{
  if(*((Dgp_Bool *)cookie)) {
    //return((Dgp_UInt16)(((tokpos *)e1)->toknum % DGP_SIZE_OF_HT));
    return((Dgp_UInt16)( (Dgp_UInt32)(((tokpos *)e1)->tok) % DGP_SIZE_OF_HT));
  }
  //return((Dgp_UInt16)(((FontToken *)e1)->token->toknum % DGP_SIZE_OF_HT));
  return((Dgp_UInt16)( (Dgp_UInt32)(((FontToken *)e1)->token) % DGP_SIZE_OF_HT));
}

/* Make the type casting for a 'const *' happy. */
typedef const FontToken *cFT_ptr;

/* Sort FontTokens by height (descending). */
#if 0
static int
sort_by_height(const void *a, const void *b)
{
  const FontToken *fa = (*((const cFT_ptr*)a));
  const FontToken *fb = (*((const cFT_ptr*)b));
  
  /* Sort based on height... */
  return(fb->height - fa->height);
}
#endif // 0

/* 
 * Sort tokens by font_num (ascending), and by char_code within each font
 * (ascending). 
 */
static int
sort_by_font(const void *a, const void *b)
{
  const FontToken *fa = *((const cFT_ptr*)a);
  const FontToken *fb = *((const cFT_ptr*)b);

  if(fa->font_num == fb->font_num) {
    return(fa->char_code - fb->char_code);
  }
  return(fa->font_num - fb->font_num);
}

/* Write a token as inline image. */
ErrCode
write_token(token       *tok, 
            DgpPdfState *state, 
            Dgp_Int32   *length,
            Dgp_State   *dgp_state) {
    DGP_PROLOGUE("write_token()", DGP_PERRSTATE);

	Dgp_Int32       width, height;
	ReusableBuffer  enc_rebuf = NULL;
	Dgp_Int8        *g4_buf;
	Dgp_Int32       buf_len;
	FILE            *outf = state->outfile;

	DGP_ERR_PASS(InitReusableBuffer(&enc_rebuf, dgp_state));

	width = tok->width;
	height = tok->height;

	/* Output inline image. */
	*length = fprintf(outf, "\
BI\n\
/W %d\n\
/H %d\n\
/BPC 1\n\
/IM true\n\
/F /CCF\n\
/DP <</EndOfBlock false /K -1 /Columns %d /Rows %d>>\n\
ID\n", (int)width, (int)height, (int)width, (int)height);
	/* Note that undecoded image width is padded to multiple of 8... */

	if ((tok->longsperline == 0) && (tok->width > 0)) {
		/* We have a compressed token. */
        if(fwrite(tok->bits, 1, tok->uses, outf) != (size_t)(tok->uses)){
            DGP_ERR_CLEANUP(cErrIO_WriteError);
        }
        *length += tok->uses;
	} else {
		/* g4 compress token... */
        DGP_ERR_PASS_CLEANUP(g4encode(  enc_rebuf, 
                                        (void *)tok->bits,
                                        width, 
                                        height, 
                                        tok->longsperline,
                                        state, 
                                        &g4_buf, 
                                        &buf_len,
                                        dgp_state));
		/* Write image data. */
        if (fwrite(g4_buf, 1, buf_len, outf) != (size_t)(buf_len)){
            DGP_ERR_CLEANUP(cErrIO_WriteError);
        }
		*length += buf_len;		
	}	
	
	/* End inline image. */
	*length += fprintf(outf, "EI\n");

	CleanupReusableBuffer(&enc_rebuf);
    NORMAL_RETURN;

ERR_CLEANUP:

	CleanupReusableBuffer(&enc_rebuf);
	ERR_RETURN;
}


/* Write out the font data and the font definitions. */
/* Either write height classes and char procs that blat pieces of height classes for each character
   or write one inline image per char proc */
static ErrCode
write_fonts(FontToken       **sort_ft, 
            Dgp_Int32       ft_num, 
            Dgp_Int32       num_fonts, 
            Dgp_Int32       first_font_num,
            DgpPdfState     *state, 
            PdfCookie       *pdfCookie,
            Dgp_State       *dgp_state) {
	
    DGP_PROLOGUE("write_fonts()", DGP_PERRSTATE);

	Dgp_Int32   chars, bbox_width, bbox_height, charProcObj, encodingObj;
	FILE        *outf = state->outfile;
	Dgp_Int32   firstFontObj;
	Dgp_Int32   i, j;

	/* Sort the tokens by font, and code. */
	qsort(sort_ft, ft_num, sizeof(*sort_ft), sort_by_font);
	
	/* Write encoding array. */
	encodingObj = pdfCookie->objCount++;
	DGP_ERR_PASS_CLEANUP(addXref(encodingObj, pdfCookie->offset, state, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<</Differences [0", (int)encodingObj);
    /* Write encoding table. */
    for (j = 0; j < DGP_CHARS_PER_FONT; j++) {
        pdfCookie->offset += fprintf(outf, " /c%d", (int)j);
    }
    pdfCookie->offset += fprintf(outf, "]>>\n\
endobj\n");

    firstFontObj = pdfCookie->objCount;
	for (i = 0; i < num_fonts; i++) {
		/* Chars in this font. */
		chars = MIN(DGP_CHARS_PER_FONT, ft_num);

		/* Maximum width & height in this font. */
		bbox_width = 0;
		bbox_height = 0;
		for (j = 0; j < chars; j++) {
			bbox_width = MAX(bbox_width, sort_ft[j]->width);
			bbox_height = MAX(bbox_height, sort_ft[j]->height);
		}

		/* Write font header to the file. */
		DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset, state, pdfCookie, dgp_state));
		pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /Font\n\
/Subtype /Type3\n\
/Name /F%d\n\
/FirstChar 0\n\
/LastChar %d\n\
/Widths [", (int)pdfCookie->objCount++, (int)(first_font_num + i), (int)(chars - 1));

		/* Write character widths. */
		for (j = 0; j < chars; j++) {
		  if (j != 0){
		    pdfCookie->offset += fprintf(outf, " ");
		  }
		  pdfCookie->offset += fprintf(outf, "%d", sort_ft[j]->width);
		}
		pdfCookie->offset += fprintf(outf, "\
]\n\
/CharProcs <<");

		/* Write character drawing refs. */
		charProcObj = pdfCookie->objCount;
		pdfCookie->objCount += chars;
		for (j = 0; j < chars; j++) {
		  pdfCookie->offset += fprintf(outf, " /c%d %d 0 R", (int)j, (int)(charProcObj + j));
		}
		pdfCookie->offset += fprintf(outf, ">>\n\
/FontBBox [0 0 %d %d]\n\
/FontMatrix [0.001 0 0 0.001 0 0]\n\
/Resources <</ProcSet [/PDF /ImageB]>>\n\
/Encoding %d 0 R\n\
>>\n\
endobj\n", (int)bbox_width, (int)bbox_height, (int)encodingObj);

		/* Write char procs to draw each character. */
		for (j = 0; j < chars; j++) {
			Dgp_Int32 length, tmp;
			/* Image width is padded to multiple of 8. */
			DGP_ERR_PASS_CLEANUP(addXref(charProcObj+j, pdfCookie->offset, state, pdfCookie, dgp_state));
			pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<</Length %d 0 R>>\n\
stream\n", (int)(charProcObj+j), (int)pdfCookie->objCount);
			length = fprintf(outf, "\
%d 0 0 0 %d %d d1\n\
%d 0 0 %d 0 0 cm\n", (int)sort_ft[j]->width, (int)sort_ft[j]->width-1, (int)sort_ft[j]->height-1, 
				(int)sort_ft[j]->width, (int)sort_ft[j]->height);

			/* Write inline image. */
			DGP_ERR_PASS_CLEANUP(write_token(sort_ft[j]->token, state, &tmp, dgp_state));
			length += tmp;
			pdfCookie->offset += length;
			pdfCookie->offset += fprintf(outf, "\
endstream\n\
endobj\n");

			/* Write length of stream. */
			DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset, state, pdfCookie, dgp_state));
			pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
%d\n\
endobj\n", (int)pdfCookie->objCount++, (int)length);
		}

		/* Ok, advance to next font... */
		sort_ft += chars;
		ft_num -= chars;
	}

	/* Write a font resource object. */
	pdfCookie->fontResourceObj = pdfCookie->objCount;
	pdfCookie->objCount++;
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->fontResourceObj, pdfCookie->offset, state, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(outf, "%d 0 obj\n<< ", (int)pdfCookie->fontResourceObj);
	for (i = 0; i < num_fonts; i++) {
	  pdfCookie->offset += fprintf(outf, "/F%d %d 0 R ", (int)(first_font_num + i), 
				       (int)(firstFontObj + (i * (2 * DGP_CHARS_PER_FONT + 1))));
	}
	pdfCookie->offset += fprintf(outf, ">>\nendobj\n");

    NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}

/* 
   Put array of FontTokens into fonts.
   Currently, we just stick the tokens into a font until it is full, then
   start a new one.
*/
static void
assign_fonts(FontToken      **sort_ft, 
             Dgp_Int32      ft_num, 
             Dgp_Int32      *num_new_fonts,
             Dgp_Int32      first_font_num)
{
  Dgp_Int32 i, cur_char_code = 0, cur_font_num = first_font_num;

  for(i = 0; i < ft_num; i++) {
    if(cur_char_code == DGP_CHARS_PER_FONT) {
        cur_char_code = 0;
        cur_font_num++;
    }
    sort_ft[i]->font_num = cur_font_num;
    sort_ft[i]->char_code = cur_char_code;
    cur_char_code++;
  }
  *num_new_fonts = (cur_font_num - first_font_num) + 1;
  return;
}


/*
 * Figure out an assignment of the global tokens to fonts.
 */
static ErrCode
figure_global_fonts(FontTokenInfo   *ftoken_info, 
                    DgpPdfState     *state, 
                    PdfCookie       *pdfCookie,
                    Dgp_State       *dgp_state)
{
  DGP_PROLOGUE("figure_global_fonts()", DGP_PERRSTATE);

  Dgp_Int32         i, j, k, glb_idx, loc_idx;
  DgpAdt_ht_elt     in_table = NULL;
  Dgp_Int32         ftoken_num = 0, global_ft_num = 0, num_global_fonts = 0;
  DgpAdt_HashTable  ft_hash = (DgpAdt_HashTable)NULL;
  FontToken         **sort_ft = NULL, *ft = NULL;
  token             *cur_tok;
  jb2SymDictSegment	*seg = &state->segment->data.symdict;
  //suTokenDictionary dict = state->dictionary;


  /* Sum up total number of tokens. */
  ftoken_num = 0;
  global_ft_num = 0;
  for(i = 0; i < state->num_uniqids; i++) {
    ftoken_num += dict->tokCount[i];
    
    /* If the dict ref count is > 1, then the dictionary is definitely global. */
    if(ftoken_info->dict_ref_count[ftoken_info->mono_uniqids[i]] > 1) {
      global_ft_num += dict->tokCount[i];
    }
  }

  /* Allocate our playspace. */
  // Remember, this is being done for all tokens in the document (not page-by-page).
  DGP_CHECK_ALLOC_CLEANUP(sort_ft = (FontToken **)DGP_MALLOC(global_ft_num * sizeof(*sort_ft)));
  DGP_CHECK_ALLOC_CLEANUP(ft = (FontToken *)DGP_MALLOC(ftoken_num * sizeof(*ft)));

  /* Create hash table so that we can find a specific ftoken from a given toknum. */
  DGP_ERR_PASS_CLEANUP(DgpAdt_htNew(&ft_hash,
                                    &(ftoken_info->e1_is_a_tokpos), 
                                    isequal_ft,
                                    do_hash_ft, 
                                    NULL, 
                                    DGP_SIZE_OF_HT,
                                    dgp_state));

  glb_idx = 0;
  loc_idx = global_ft_num;
  for(i = 0; i < state->num_uniqids; i++) {
    if(ftoken_info->dict_ref_count[ftoken_info->mono_uniqids[i]] > 1) {
        // This is a global dictionary...
        k = glb_idx;
        glb_idx += dict->tokCount[i];
    } else {
        // Local...
        k = loc_idx;
        loc_idx += dict->tokCount[i];
    }
    
    /* Get FontToken info from tokens, global tokens are in the beginning of
       the ft array, local tokens at the end. */
    // (...but global tokens have a higher index / dhtesq.)
    cur_tok = dict->toks[i];
    for(j = 0; j < dict->tokCount[i]; j++) {
        if (k < global_ft_num){
            // We have yet to reach the number of global tokens...
            sort_ft[k] = ft + k;
        }
        ft[k].token = cur_tok;
        ft[k].height = cur_tok->height;
        ft[k].width = cur_tok->width;

        /* Font number and code will be assigned later. */
        ft[k].font_num = DGP_UNASSIGNED;
        ft[k].char_code = DGP_UNASSIGNED;

        /* Add the ft to the hash table. */
        DGP_ERR_PASS_CLEANUP(DgpAdt_htAddToTable(ft_hash, (DgpAdt_ht_elt)(ft + k), &in_table));
        DGP_CHECK_NOT_NULL(in_table);

        cur_tok++;
        k++;
    }
  }

  if(global_ft_num > 0) {
    /* Now we fill in the global fonts.  The local tokens are put into
       fonts later, based on the page they are used on. */
      // 3.8.2000 / dhtesq.
      // Doesn't look like local tokens are put into fonts at all.  Note if the encoder
      // has done its job well, local tokens should be singletons.  These get
      // collectively g4'd and recorded as an image.
    assign_fonts(sort_ft, global_ft_num, &num_global_fonts, 0);
    
    /* Write the global fonts to the outfile. */
    DGP_ERR_PASS_CLEANUP(write_fonts(sort_ft, global_ft_num, num_global_fonts, 0, state, pdfCookie, dgp_state));
  } else {
      /* 3.8.2000 / dhtesq.
       * If there is no "global" dictionary (say we have a one-page document)
       * then we'll never make the above call to assign fonts.  Make sure we have
       * a valid fontResourceObj.
       */
    pdfCookie->fontResourceObj = 0;
  }


  ftoken_info->ftokens = ft;
  ftoken_info->sort_ftokens = sort_ft;
  ftoken_info->ftoken_num = ftoken_num;
  ftoken_info->global_ftoken_num = global_ft_num; 
  ftoken_info->num_global_fonts = num_global_fonts;
  ftoken_info->ftokens_hash = ft_hash;
  
  NORMAL_RETURN;

ERR_CLEANUP:
  if(ft_hash) DgpAdt_htFree(ft_hash);
  if(ft) DGP_FREE(ft);
  if(sort_ft) DGP_FREE(sort_ft);

  ERR_RETURN;
}


/* Create one large image out of a bunch of positions. */
static ErrCode
make_local_image(DgpAdt_List    local_fboxes, 
                 Dgp_Int32      pg_width, 
                 Dgp_Int32      pg_height, 
                 DgpPdfState    *state, 
                 DgpImage_Packed *_pim, 
                 Dgp_Int32      *_lx, 
                 Dgp_Int32      *_ly,
                 Dgp_State      *dgp_state)
{
    DGP_PROLOGUE("make_local_image()", DGP_PERRSTATE);

    DgpAdt_ListNode ln = DGP_NULLLISTNODE;
    FontPos *fbox = NULL;
    token *tok = NULL;
    Dgp_UInt16 lx = (Dgp_UInt16)pg_width, ly = (Dgp_UInt16)pg_height, ux = 0, uy = 0;
    DgpImage_Packed pim = NULL;
    tokpos box = {0, 0, 0};
    Dgp_Int32 width, height;

    state = state;
    *_pim = NULL;

    /* First figure out how big the image needs to be .*/
    DGPADT_FOREACH(ln, (local_fboxes)) {
      fbox = (FontPos *)(DGPADT_LIGET(ln));
      tok = fbox->ftoken->token;
      
      if(fbox->xpos < lx) lx = fbox->xpos;
      if(fbox->ypos < ly) ly = fbox->ypos;
      if((fbox->xpos + tok->width) > ux) ux = (fbox->xpos + tok->width);
      if((fbox->ypos + tok->height) > uy) uy = (fbox->ypos + tok->height);

      /* Indicate that the token should be ignored from now on. */
      fbox->ftoken->font_num = DGP_IGNORE;
    }
    
    /* Create the image for the local tokens. */
    width = (((ux - lx) + 31) / 32)*32;
    height = (uy - ly);
    DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_PACKED, width, height, (void *)&pim));
    
    /* Insert each box into the new image. */
    DGPADT_FOREACH(ln, (local_fboxes)) {
      fbox = (FontPos *)(DGPADT_LIGET(ln));
      tok = fbox->ftoken->token;
      
      /* Convert coords relative to image and UL corner of box, with
	     origin in UL corner. */
      box.xpos = fbox->xpos - lx;
      box.ypos = uy - fbox->ypos - (fbox->ftoken->token->height);
      
      if((tok->longsperline == 0) && (tok->width > 0)) {
      	/* We have a single-use token... */
	    DGP_ERR_PASS_CLEANUP(InsertCompressedBoxPacked( pim, 
                                                        box.xpos,
                                                        box.ypos, 
                                                        tok->width,
                                                        tok->height, 
                                                        tok->bits, 
                                                        tok->uses,
                                                        DGP_FALSE,                             
                                                        dgp_state));
      } else {
          InsertBoxPacked(pim, &box, (fbox->ftoken->token), DGP_FALSE);
      }
    }
    
	*_pim = pim;
	*_lx = lx;
	*_ly = ly;

	NORMAL_RETURN;

ERR_CLEANUP:

	if(pim) DgpImage_free(pim);

	ERR_RETURN;
}


/* 
   sort by font (ascending), baseline, xpos;
   any postions whose ftokens have the font_num DGP_IGNORE are moved to the
   beginning of the array, so they can be ignored.
 */
static int
sort_positions(const void *a, const void *b)
{
  const FontPos *fa = (const FontPos *)a;
  const FontPos *fb = (const FontPos *)b;

  /* first check for ignored tokens */
  if(fa->ftoken->font_num == DGP_IGNORE) {
    if(fb->ftoken->font_num == DGP_IGNORE) {
      return(DGP_EQUAL);
    }
    return(DGP_LESS_THAN);
  } else if(fb->ftoken->font_num == DGP_IGNORE) {
    return(DGP_GREATER_THAN);
  }

  if (fa->ftoken->font_num != fb->ftoken->font_num) {
	  return (fa->ftoken->font_num - fb->ftoken->font_num);
  } else {
	  if (fa->ypos != fb->ypos) {
		  return (fa->ypos - fb->ypos);
	  } else {
		  return (fa->xpos - fb->xpos);
	  }
  }
}

Dgp_Int32
print_char(Dgp_Int8 *outf, Dgp_Int32 char_code) {
	if (char_code == '\r') {
		return sprintf(outf, "\\r");
	} else if (char_code == '\n') {
		return sprintf(outf, "\\n");
	} else if (char_code == '\t') {
		return sprintf(outf, "\\t");
	} else if (char_code == '\b') {
		return sprintf(outf, "\\b");
	} else if (char_code == '\f') {
		return sprintf(outf, "\\f");
	} else if (char_code == '\\') {
		return sprintf(outf, "\\\\");
	} else if (char_code == '(') {
		return sprintf(outf, "\\(");
	} else if (char_code == ')') {
		return sprintf(outf, "\\)");
	} else {
		*(Dgp_UInt8 *)outf = (Dgp_UInt8)char_code;
		return 1;
	}
}

/* 
  Function which will write out each subifd image, leaving it compressed
    as it is.

  I converted the ipcore interface for reading compressed images to zephyr's,
  which presently has no strip support.  I write out the g4 masks, but I can't 
  get them to work yet (29.9.2000) on the pdf side. / dhtesq.
*/
static ErrCode
write_subifd_images(zephPageInfo    zephPage, 
                    zephImageInfo   *zephImage,
                    Dgp_Int32       *_num_images, 
                    residImage      **_images,
                    DgpPdfState     *state, 
                    PdfCookie       *pdfCookie,
                    Dgp_State       *dgp_state)
{
    DGP_PROLOGUE("write_subifd_images()", DGP_PERRSTATE);
    
    FILE *outf = state->outfile;
    Dgp_UInt32 i, j, layers = 0;


    Dgp_UInt8 *cmp_data = NULL;
    ReusableBuffer enc_rebuf = NULL;  
    residImage *images = NULL;
    Dgp_Int32 num_images = 0, mask_name = 0;

    Dgp_Int32 zephImageIndex;
    zephImageInfo zimage, zmask;
    zephReadState zrs;

    _zephPrivatePageInfo *_privPage;
    _privPage = (_zephPrivatePageInfo *)zephPage->privateInfo;

	DGP_ERR_PASS(InitReusableBuffer(&enc_rebuf, dgp_state));

	/* We really only want to check the background and selector (if there is one). */
    // Let's try to capture images on higher planes... / dhtesq.)
	//layers = MIN(1, zephPage->layerCount);
    layers = zephPage->layerCount;

	/* Write out each subifd image for this page. */
	for(i = 0, zephImageIndex = 0; i < layers; i++) {
		for(j = 0; j < zephPage->imageCounts[i]; j++) {
            zimage = zephImage[zephImageIndex++];

			/* Skip main IFD. */
			if(i == ZEPH_MAINIMAGE_LAYER) continue;
      
			/* Check for valid color type and compression type. */
            switch(zimage->compType){
            case XIF_G4:
                break;
            case XIF_TOKENIZATION:  case XIF_LLTOKENIZATION:
                continue;
                break;
            case XIF_JPEG:
                break;
            default:
				DGP_WARNF0("Unsupported image type in subifd, will be ignored.");
				continue;
                break;
            }
	
            /* Do we have a mask?... */
            if(zimage->maskInfo && state->verbose){
                DGP_WARNF0("Page contains masked image;  output is likely to be corrupt.");
            }

			/* Read in the compressed data and write it as a separate
			   image in the pdf/postscript file. */
            DGP_ERR_GEN_CLEANUPF0(zimage->height != zimage->stripHeight, cErrNoImpl, "Stripped images not supported.");
            /* Get a buffer. */
            GetReusableBufferPtr(cmp_data, (zimage->compSize * sizeof(*cmp_data)), enc_rebuf);
            DGP_ERR_PASS_CLEANUP(zephReadImageStart(  zimage,
                                                      (void *)cmp_data, 0,
                                                      NULL, 0,
                                                      0, 0,
                                                      ZEPHRT_COMPRESSED,
                                                      ORIENT_RD,      // Not really relevant here.
                                                      &zimage->res,
                                                      zimage->xPos, zimage->yPos,
                                                      zimage->width, zimage->height,
                                                      &zimage->compSize,
                                                      0, 0,
                                                      &zrs));
            // We count on this function to throw an error if the read
            // comes back with less than compSize bytes...
            DGP_ERR_PASS_CLEANUP(zephReadImageEnd(    zimage,
                                                      &zrs,
                                                      &zimage->compSize,
                                                      &zimage->res,
                                                      zimage->width, zimage->height,
                                                      zimage->width, zimage->height));

            /* Write image object. */
            DGP_CHECK_ALLOC_CLEANUP(images = DGP_REALLOC(images, sizeof(residImage)*(num_images+1)));

            images[num_images].width = zimage->width;
            images[num_images].height = zimage->stripHeight;
            images[num_images].xres = (Dgp_Int32)(DgpCoord_float(zimage->res.xRes));
            images[num_images].yres = (Dgp_Int32)(DgpCoord_float(zimage->res.yRes));
            images[num_images].lx = zimage->xPos;
            images[num_images].ly = zephPage->height - (zimage->yPos + zimage->stripHeight);
            images[num_images].obj = pdfCookie->objCount++;
            images[num_images].is_mask = DGP_FALSE;
            DGP_ERR_PASS_CLEANUP(addXref(images[num_images].obj, pdfCookie->offset, state, pdfCookie, dgp_state));

            switch(zimage->compType){
            case XIF_G4:
                pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Name /I%d\n\
/Width %d\n\
/Height %d\n\
/BitsPerComponent 1\n\
/ImageMask true\n\
/Filter /CCITTFaxDecode\n\
/DecodeParms << /EndOfBlock false /K -1 /Columns %d /Rows %d >>\n\
/Length %d\n\
>>\n\
stream\n",      (int)images[num_images].obj, 
                (int)num_images, 
                (int)zimage->width, (int)zimage->stripHeight,
                (int)zimage->width, (int)zimage->stripHeight, (int)zimage->compSize);
                break;
            case XIF_JPEG:
                // Write a reference to the mask object if we have one...
                if(zimage->maskInfo){
                    mask_name = images[num_images].obj + 1;
                    pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Name /I%d\n\
/Mask %d 0 R\n\
/Width %d\n\
/Height %d\n\
/BitsPerComponent 8\n\
/ColorSpace /DeviceRGB\n\
/Filter /DCTDecode\n\
/Length %d\n\
>>\n\
stream\n",          (int)images[num_images].obj, 
                    (int)num_images,
                    (int)mask_name,
                    (int)zimage->width, (int)zimage->stripHeight,
                    (int)zimage->compSize);
                } else {
                    pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Name /I%d\n\
/Width %d\n\
/Height %d\n\
/BitsPerComponent 8\n\
/ColorSpace /DeviceRGB\n\
/Filter /DCTDecode\n\
/Length %d\n\
>>\n\
stream\n",          (int)images[num_images].obj, 
                    (int)num_images,
                    (int)zimage->width, (int)zimage->stripHeight,
                    (int)zimage->compSize);
                }
                break;
            default:
                break;
            }

            /* Write image data. */
            if(fwrite(cmp_data, 1, zimage->compSize, outf) !=
                (size_t)(zimage->compSize)){
                DGP_ERRF0(cErrIO_WriteError, "Error writing image data.");
            }
            pdfCookie->offset += zimage->compSize;

            pdfCookie->offset += fprintf(outf, "\
endstream\n\
endobj\n");

            num_images++;

            if(zimage->maskInfo){
                zmask = zimage->maskInfo;
                /* Get a buffer for the mask. */
                GetReusableBufferPtr(cmp_data, (zmask->compSize * sizeof(*cmp_data)), enc_rebuf);

                /* Read the compressed mask strip. */
                DGP_ERR_PASS_CLEANUP(zephReadImageStart(zmask,
                                                        (void *)cmp_data, 0,
                                                        NULL, 0,
                                                        0, 0,
                                                        ZEPHRT_COMPRESSED,
                                                        ORIENT_RD,      // Not really relevant here.
                                                        &zmask->res,
                                                        zmask->xPos, zmask->yPos,
                                                        zmask->width, zmask->height,
                                                        &zmask->compSize,
                                                        0, 0,
                                                        &zrs));
                // We count on this function to throw an error if the read
                // comes back with less than compSize bytes...
                DGP_ERR_PASS_CLEANUP(zephReadImageEnd(  zmask,
                                                        &zrs,
                                                        &zmask->compSize,
                                                        &zmask->res,
                                                        zmask->width, zmask->height,
                                                        zmask->width, zmask->height));


                /* Write image object. */
                DGP_CHECK_ALLOC_CLEANUP(images = DGP_REALLOC(images, sizeof(residImage)*(num_images + 1)));
                              
                images[num_images].width = zmask->width;
                images[num_images].height = zmask->stripHeight;
                images[num_images].xres = (Dgp_Int32)(DgpCoord_float(zmask->res.xRes));
                images[num_images].yres = (Dgp_Int32)(DgpCoord_float(zmask->res.yRes));
                images[num_images].lx = zmask->xPos;
                images[num_images].ly = zephPage->height - (zmask->yPos + zmask->stripHeight);
                images[num_images].obj = pdfCookie->objCount++;
                images[num_images].is_mask = DGP_TRUE;
                DGP_ERR_PASS_CLEANUP(addXref(images[num_images].obj, pdfCookie->offset, state, pdfCookie, dgp_state));
                               
                switch(zmask->compType){
                case XIF_G4:
                    pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/ImageMask true\n\
/Decode [0 1]\n\
/Name /I%d\n\
/Width %d\n\
/Height %d\n\
/BitsPerComponent 1\n\
/Filter /CCITTFaxDecode\n\
/DecodeParms << /EndOfBlock false /K -1 /Columns %d /Rows %d >>\n\
/Length %d\n\
>>\n\
stream\n",          (int)images[num_images].obj, 
                    (int)num_images, 
                    (int)zmask->width, (int)zmask->stripHeight,
                    (int)zmask->width, (int)zmask->stripHeight, 
                    (int)zmask->compSize);
                    break;
                default:
                    DGP_WARNF0("No implementation for this mask type;  continuing...");
                    break;
                }                
                
                /* Write image data. */
                if(fwrite(cmp_data, 1, zmask->compSize, outf) != (size_t)(zmask->compSize)){
                    DGP_ERRF0(cErrIO_WriteError, "Error writing image data.");
                }
                pdfCookie->offset += zmask->compSize;

                pdfCookie->offset += fprintf(outf, "\
endstream\n\
endobj\n");

                //mask_name = images[num_images].obj;
                num_images++;
            }
        } // image loop
    } // layer loop

	*_num_images = num_images;
	*_images = images;
	CleanupReusableBuffer(&enc_rebuf);
	NORMAL_RETURN;

ERR_CLEANUP:
	CleanupReusableBuffer(&enc_rebuf);
    if (images){
		DGP_FREE(images);
    }
	ERR_RETURN;
}


/*
   This function actually outputs the positions for the page.  We sort the
   positions into fonts and baseline.
   We do an xshow for each baseline.
*/
static ErrCode
write_page_positions(FontPos        *fboxes, 
                     Dgp_Int32      num_boxes, 
                     Dgp_Int32      xres, 
                     Dgp_Int32      yres, 
                     DgpPdfState    *state, 
					 Dgp_Int8       *buf, 
                     Dgp_Int32      buf_size, 
                     Dgp_Int32      *_buf_used,
                     Dgp_State      *dgp_state) {
    DGP_PROLOGUE("write_page_positions()", DGP_PERRSTATE);

    Dgp_Int32 cur_x = 0, cur_y = 0, off_x, off_y;
    Dgp_Int32 tmp;
    Dgp_Int32 cur_font;
    Dgp_Int32 line_start = 0, line_end, line_y;
    Dgp_Int32 buf_remain = buf_size;

    /* Sort boxes into fonts baseline, xpos. */
    qsort(fboxes, num_boxes, sizeof(*fboxes), sort_positions);

    /* Begin text. */
    if (buf_size < 50)
      goto buffer_full;
    tmp = sprintf(buf, "%f 0 0 %f 0 0 cm\nBT\n", (Dgp_Float32)state->dpi/(Dgp_Float32)xres, (Dgp_Float32)state->dpi/(Dgp_Float32)yres);
    //tmp = sprintf(buf, "BT\n");
    buf_remain -= tmp;
    buf += tmp;

    /* Move past any DGP_IGNORE tokens (already written to a local image) 
       and DGP_UNASSIGNED tokens, which are not in any font. */
    while ((line_start < num_boxes) && 
	   ((fboxes[line_start].ftoken->font_num == DGP_IGNORE) || (fboxes[line_start].ftoken->font_num == DGP_UNASSIGNED)))
      line_start++;

    /* Blat all boxes. */
    while (line_start < num_boxes) {
      cur_font = fboxes[line_start].ftoken->font_num;
      
      /* Select font. */
      if (buf_remain < 50)
	goto buffer_full;
      tmp = sprintf(buf, "/F%d 1000 Tf\n", (int)cur_font);
      buf_remain -= tmp;
      buf += tmp;
		
      /* Blat all characters in this font. */
      while ((line_start < num_boxes) && (fboxes[line_start].ftoken->font_num == cur_font)) {			
	
	/* Blat all characters in this font in this line. */
	line_y = fboxes[line_start].ypos;
	line_end = line_start;
	while ((line_end < num_boxes) && (fboxes[line_end].ftoken->font_num == cur_font) &&
	       (fboxes[line_end].ypos == line_y))
	  line_end++;
	
	/* Position for first character. */
	off_x = fboxes[line_start].xpos-cur_x;
	off_y = fboxes[line_start].ypos-cur_y;
	cur_x += off_x;
	cur_y += off_y;

	if (buf_remain < 50)
	  goto buffer_full;
	tmp = sprintf(buf, "%d %d Td\n", (int)off_x, (int)off_y);
	buf_remain -= tmp;
	buf += tmp;
	
	if ((line_start+1) < line_end) {
	  /* Do an xshow. */
	  if (buf_remain < 50)
	    goto buffer_full;
	  tmp = sprintf(buf, "[(");
	  buf_remain -= tmp;
	  buf += tmp;

	  /* First character. */
	  tmp = print_char(buf, fboxes[line_start].ftoken->char_code);
	  buf_remain -= tmp;
	  buf += tmp;

	  line_start++;
	  /* Remaining characters. */
	  while (line_start < line_end) {
	    Dgp_Int32 offset;			
	    
	    if (buf_remain < 50)
	      goto buffer_full;

	    /* Position offset between last character and current character. */
	    offset = (fboxes[line_start-1].xpos+fboxes[line_start-1].ftoken->width)-fboxes[line_start].xpos;					
	    if (offset != 0) {						
	      tmp = sprintf(buf, ")%d(", (int)offset);
	      buf_remain -= tmp;
	      buf += tmp;
	    }
					
	    tmp = print_char(buf, fboxes[line_start].ftoken->char_code);
	    buf_remain -= tmp;
	    buf += tmp;
	    line_start++;
	  }
 	  /* End xshow. */
	  tmp = sprintf(buf, ")]TJ\n");
	  buf_remain -= tmp;
	  buf += tmp;
	} else {
	  /* Show a single char. */
	  if (buf_remain < 50)
	    goto buffer_full;
	  tmp = sprintf(buf, "(");
	  buf_remain -= tmp;
	  buf += tmp;
	  tmp = print_char(buf, fboxes[line_start].ftoken->char_code);					
	  buf_remain -= tmp;
	  buf += tmp;
	  tmp = sprintf(buf, ")Tj\n");
	  buf_remain -= tmp;
	  buf += tmp;
	  line_start++;
	}
      }
    }

    /* End text. */
    if (buf_remain < 3)
      goto buffer_full;
    tmp = sprintf(buf, "ET\n");
    buf_remain -= tmp;
    buf += tmp;

    *_buf_used = buf_size-buf_remain;

    NORMAL_RETURN;

buffer_full:
    *_buf_used = -1;
    DGP_ERR_NO_MSG(cErrIntBufOverflow);

//ERR_CLEANUP:    
    ERR_RETURN;
}
				
/* qsort comparator function to sort array of ints. */
#if 0
static int
sort_ints(const void *a, const void *b)
{
  return((*(const Dgp_Int32 *)a) - (*(const Dgp_Int32 *)b));
}
#endif // 0


/* Count the references from each page to each mono dictionary. */
static ErrCode
get_dict_info(FontTokenInfo         *ftoken_info,
              DgpPdfState           *state,
              Dgp_State             *dgp_state)
{
  DGP_PROLOGUE("get_dict_info()", DGP_PERRSTATE);
  
  Dgp_UInt32 i, j, k;

  zephPageInfo  page;
  zephImageInfo zephImage;

  /* Variables needed for the zephyr calls. */
  Dgp_UInt32 *uniqids, num_uniqids, uniqid, num_mono_uniqids = 0;
  Dgp_UInt32 *mono_uniqids = NULL;
  ReusableBuffer pos_block_rebuf = NULL;
  Dgp_Int32 *dict_ref_count = NULL;

  Dgp_UInt32 max_uniqid = ftoken_info->max_uniqid;
  
  DGP_ERR_PASS(InitReusableBuffer(&pos_block_rebuf, dgp_state));

  DGP_CHECK_ALLOC_CLEANUP(dict_ref_count = (Dgp_Int32 *)DGP_CALLOC((ftoken_info->max_uniqid + 1), 
                                           sizeof(*dict_ref_count)));
  DGP_CHECK_ALLOC_CLEANUP(mono_uniqids = (Dgp_UInt32 *)DGP_CALLOC((ftoken_info->max_uniqid + 1), 
                                         sizeof(*mono_uniqids)));

  for(i = 0; i < state->zephDoc->pageCount; i++) {
    DGP_ERR_PASS_CLEANUP(zephOpenPage(state->zephDoc,
                                      (Dgp_UInt16)i,
                                      &page));

    DGP_ERR_PASS_CLEANUP(zephOpenImage( page, 
                                        ZEPH_MAINIMAGE_LAYER,           // These macros select the first image 
                                        ZEPH_MAINIMAGE_NUMBER,          // on the selector layer.
                                        &zephImage));
    
    /* Count up regular dict references. */
    /*for(j = 0; (Dgp_UInt32)j < num_uniqids; j++) {
      ftoken_info->dict_ref_count[(Dgp_Int32)uniqids[j]]++;
    }*/
    num_uniqids = zephImage->libCount;
    uniqids = zephImage->libIds;

    for(j = 0; (Dgp_UInt32)j < num_uniqids; j++) {
      uniqid = uniqids[j];

      if(uniqid > max_uniqid) {
        DGP_CHECK_ALLOC_CLEANUP(dict_ref_count = (Dgp_Int32 *)DGP_REALLOC(dict_ref_count, 
                                                 (uniqid + 1) * sizeof(*dict_ref_count)));
        
        memset((dict_ref_count + max_uniqid + 1), 0, 
               ((uniqid - max_uniqid) * sizeof(*dict_ref_count)));

        max_uniqid = uniqid;
      }

      dict_ref_count[(Dgp_Int32)uniqid]++;

      /* mono dict */
      for(k = 0; k < num_mono_uniqids; k++) {
          if(mono_uniqids[k] == uniqid) break;
      }
      if(k == num_mono_uniqids) {
          mono_uniqids[num_mono_uniqids] = uniqid;
          num_mono_uniqids++;
      }      
    }
    zephCloseImage(&zephImage);
    zephClosePage(&page);
  }
  
  ftoken_info->num_mono_uniqids = num_mono_uniqids;

  ftoken_info->dict_ref_count = dict_ref_count;
  dict_ref_count = NULL;
  ftoken_info->mono_uniqids = mono_uniqids;
  mono_uniqids = NULL;

  /* Cleanup... */
  CleanupReusableBuffer(&pos_block_rebuf);
  
  NORMAL_RETURN;

ERR_CLEANUP:
  if(dict_ref_count) (DGP_FREE)(dict_ref_count);
  if(mono_uniqids) (DGP_FREE)(mono_uniqids);

  CleanupReusableBuffer(&pos_block_rebuf);

  ERR_RETURN;
}


/* Clean up all of the memory we've accumulated in the ftoken_info struct. */
static void
cleanup_ftoken_info(FontTokenInfo **ftoken_info, DgpPdfState *state, Dgp_State *dgp_state)
{
  state = state;
  if(*ftoken_info) {

    if((*ftoken_info)->dict_ref_count) DGP_FREE((*ftoken_info)->dict_ref_count);
    if((*ftoken_info)->ftokens_hash) DgpAdt_htFree((*ftoken_info)->ftokens_hash);
    if((*ftoken_info)->ftokens) DGP_FREE((*ftoken_info)->ftokens);
    if((*ftoken_info)->sort_ftokens) DGP_FREE((*ftoken_info)->sort_ftokens);
    if((*ftoken_info)->mono_uniqids) DGP_FREE((*ftoken_info)->mono_uniqids);

    DGP_FREE(*ftoken_info);
    *ftoken_info = NULL;
  }
}


/* Setup the ftoken_info struct. */
static ErrCode
init_ftoken_info(FontTokenInfo      **_ftoken_info, 
                 DgpPdfState        *state, 
                 Dgp_State          *dgp_state)
{
  DGP_PROLOGUE("init_ftoken_info()", DGP_PERRSTATE);

  FontTokenInfo *ftoken_info = NULL;
  Dgp_Int32 num_dicts;
  
  DGP_ERR_PASS(zephGetSymlibCount(state->zephDoc, &num_dicts));

  DGP_CHECK_ALLOC_CLEANUP(ftoken_info = (FontTokenInfo *)DGP_CALLOC(1, sizeof(*ftoken_info)));

  ftoken_info->max_uniqid = num_dicts - 1;

  *_ftoken_info = ftoken_info;

  NORMAL_RETURN;

ERR_CLEANUP:
  cleanup_ftoken_info(&ftoken_info, state, dgp_state);

  ERR_RETURN;
}

/* Write pdf prolog. */
void
write_pdf_prolog(DgpPdfState *state, PdfCookie *pdfCookie) {
	FILE *outf = state->outfile;
	
	/* Write header. */
	pdfCookie->offset += fprintf(outf, "\
%%PDF-1.3\n");

    return;
}

/* Write pdf page catalog. */
ErrCode
write_pdf_catalog(DgpPdfState   *state, 
                  Dgp_Int32     num_pages, 
                  PdfCookie     *pdfCookie, 
                  Dgp_State     *dgp_state) {
    DGP_PROLOGUE("write_pdf_catalog()", DGP_PERRSTATE);
	FILE *outf = state->outfile;
	Dgp_Int32 i;

	/* Catalog object, references the pages object. */
	DGP_ERR_PASS(addXref(pdfCookie->objCount, pdfCookie->offset, state, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /Catalog\n\
/Pages %d 0 R\n\
>>\n\
endobj\n", (int)pdfCookie->objCount, (int)(pdfCookie->objCount+1));
	pdfCookie->catalogObj = pdfCookie->objCount;
	pdfCookie->objCount++;

	/* Pages object, references each page. */
	DGP_ERR_PASS(addXref(pdfCookie->objCount, pdfCookie->offset, state, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /Pages\n\
/Count %d\n\
/Kids [", (int)pdfCookie->objCount, (int)num_pages);	
	pdfCookie->pagesObj = pdfCookie->objCount;
	pdfCookie->objCount++;
	pdfCookie->firstPageObj = pdfCookie->objCount;
	pdfCookie->objCount += num_pages;
	/* Write one reference for each object page. */
	for (i = 0; i < num_pages; i++) {
        if(i > 0){
			pdfCookie->offset += fprintf(outf, " ");
        }
		pdfCookie->offset += fprintf(outf, "%d 0 R", (int)(pdfCookie->firstPageObj + i));
	}
	pdfCookie->offset += fprintf(outf, "\
]\n\
>>\n\
endobj\n");

    NORMAL_RETURN;
}

ErrCode
write_local_image(DgpAdt_List   fboxes, 
                  Dgp_Int32     pg_width, 
                  Dgp_Int32     pg_height, 
                  Dgp_Int32     image_num, 
                  Dgp_Int32     *imageObj, 
                  Dgp_Int32     *lx, 
                  Dgp_Int32     *ly, 
                  Dgp_Int32     *width, 
                  Dgp_Int32     *height,
                  DgpPdfState   *state, 
                  PdfCookie     *pdfCookie,
                  Dgp_State     *dgp_state) {
    DGP_PROLOGUE("write_local_image()", DGP_PERRSTATE);
    
    DgpImage_Packed     pim = NULL;
    Dgp_Int32           buf_len;
    Dgp_Int8            *g4_buf;
    ReusableBuffer      enc_rebuf = NULL;
    FILE                *outf = state->outfile;

    DGP_ERR_PASS(InitReusableBuffer(&enc_rebuf, dgp_state));

    DGP_ERR_PASS_CLEANUP(make_local_image(fboxes, pg_width, pg_height, state, &pim, lx, ly, dgp_state));

    *width = DGPIMG_WIDTH(pim);
    *height = DGPIMG_HEIGHT(pim);
    /* g4 compress image... */				
    DGP_ERR_PASS_CLEANUP(g4encode(  enc_rebuf, 
                                    (void *)(DGPIMG_GETROW(pim, 0)),
                                    *width, *height, 
                                    (DGPIMG_GETROWWIDTH(pim) / sizeof(Dgp_UInt32)),
                                    state, 
                                    &g4_buf, 
                                    &buf_len,
                                    dgp_state));

    /* Output the image object... */
    *imageObj = pdfCookie->objCount++;
    DGP_ERR_PASS_CLEANUP(addXref(*imageObj, pdfCookie->offset, state, pdfCookie, dgp_state));
    pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Name /I%d\n\
/Width %d\n\
/Height %d\n\
/BitsPerComponent 1\n\
/ImageMask true\n\
/Filter /CCITTFaxDecode\n\
/DecodeParms << /EndOfBlock false /K -1 /Columns %d /Rows %d >>\n\
/Length %d\n\
>>\n\
stream\n", (int)(*imageObj), (int)image_num, (int)(*width), (int)(*height), (int)(*width), (int)(*height), (int)buf_len); 

	/* Write image data. */
    if(fwrite(g4_buf, 1, buf_len, outf) != (size_t)(buf_len)){
        DGP_ERR_CLEANUP(cErrIO_WriteError);
    }
	pdfCookie->offset += buf_len;

	pdfCookie->offset += fprintf(outf, "\
endstream\n\
endobj\n");

	/* Cleanup. */
	DgpImage_free(pim);
	CleanupReusableBuffer(&enc_rebuf);
	NORMAL_RETURN;

ERR_CLEANUP:
	if(pim) DgpImage_free(pim);
	CleanupReusableBuffer(&enc_rebuf);
	ERR_RETURN;
}

/* Write a bunch of pdf pages from a xif document. */
static ErrCode
write_pdf_pages(FontTokenInfo       *ftoken_info,
                DgpPdfState         *state, 
                PdfCookie           *pdfCookie,
                Dgp_State           *dgp_state) {
    DGP_PROLOGUE("write_pdf_pages()", DGP_PERRSTATE);

    FILE *outf = state->outfile;

    zephPageInfo    zephPage = NULL;
    zephImageInfo   *zephImage = NULL;
    Dgp_UInt32       zephImageIndex = 0;

    Dgp_UInt32      i, j, k, magic_tokenized_index = 0;
    Dgp_Int32       contentObj;
    Dgp_UInt32       num_boxes = 0;

    FontToken       *cur_ft = NULL;
    tokpos          *boxes = NULL;
    FontPos         *fboxes = NULL, *cur_fbox = NULL;

    DgpAdt_HashTable ft_hash = ftoken_info->ftokens_hash;
    DgpAdt_ht_elt   in_table = NULL;
    DgpAdt_List     local_fboxes = NULL;
    DgpAdt_ListNode dummy;

    ReusableBuffer  fboxes_rebuf = NULL;

    Dgp_Int32       num_chars = 0;
    Dgp_Int8        *content = NULL;
    Dgp_Int32       content_size, content_used;
    Dgp_Int8        *comp_content, *content_write_pos = NULL;
    Dgp_Int32       comp_content_size;
    Dgp_Int32       tmp;
    Dgp_UInt32       num_images = 0, images_on_page = 0;

    residImage      *images = NULL;
    z_stream        strm;
    Dgp_Bool        fetched = DGP_FALSE;

    /* zlib cookie to compress content descriptions, these should come from the allocator
       structure but I need to write wrappers. */
    strm.zalloc = NULL;
    strm.zfree = NULL;
    strm.opaque = NULL; 

    /* Set the hash table cookie, since we are now referencing with tokpositions. */
    ftoken_info->e1_is_a_tokpos = DGP_TRUE;    
    
    DGP_ERR_PASS_CLEANUP(InitReusableBuffer(&fboxes_rebuf, dgp_state));
 
    /* Write each page... */
    for (i = 0; i < state->zephDoc->pageCount; i++) { 		
      num_images = 0;
      zephImageIndex = 0;
      images_on_page = 0;
      images = NULL;
      
      DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&local_fboxes, dgp_state));

      // Get boxes and dictionaries for this page.
      DGP_ERR_PASS_CLEANUP(zephOpenPage(state->zephDoc, (Dgp_UInt16)i, &zephPage));
      // How many images are there on the page?
      for(j = 0; j < zephPage->layerCount; j++){
	images_on_page += zephPage->imageCounts[j];
      }
      
      DGP_CHECK_ALLOC_CLEANUP(zephImage = (zephImageInfo *)DGP_CALLOC(1, 
								      sizeof(*zephImage) * images_on_page));
      for(j = 0, zephImageIndex = 0; j < zephPage->layerCount; j++){
	for(k = 0; k < zephPage->imageCounts[j]; k++){
	  if((j == ZEPH_MAINIMAGE_LAYER) && (k == ZEPH_MAINIMAGE_NUMBER)){
	    magic_tokenized_index = zephImageIndex;
	  }
	  
	  DGP_ERR_PASS_CLEANUP(zephOpenImage(zephPage, j, k, &zephImage[zephImageIndex++]));
	}
      }
 
      /* Write subifd images. */
      DGP_ERR_PASS_CLEANUP(write_subifd_images(   zephPage, 
						  zephImage, 
						  &num_images, 
						  &images, 
						  state, 
						  pdfCookie, 
						  dgp_state));
      
      DGP_ERR_PASS_CLEANUP(suGetTokenPositionBlockH(state->cache,
						    i,
						    zephImage[magic_tokenized_index],
						    DGP_TRUE,
						    &(state->entry)));
      fetched = DGP_TRUE;
      DGP_ERR_PASS_CLEANUP(suExtractTokenPositionBlock(state->entry,
                                                         &state->pos_block));
      /* Get the tokpos info for the page. */
      boxes = state->pos_block->pos;
      num_boxes = state->pos_block->posCount;
      
      /* Are there tokens in this page? */
      if (num_boxes > 0) {
	
	/* Get fboxes buffer large enough for this page's boxes. */
	GetReusableBufferPtr(fboxes, (num_boxes * sizeof(*fboxes)), fboxes_rebuf);
      
	cur_fbox = fboxes;
	num_chars = 0;
	for(j = 0; j < num_boxes; j++) {
	  /* Find the ftoken for this position. */
	  DGP_ERR_PASS_CLEANUP(DgpAdt_htIsInTable(ft_hash, (boxes + j), &in_table));
	  DGP_CHECK_NULL_CLEANUP(in_table);

	  cur_ft = (FontToken *)in_table;
	  cur_fbox->ftoken = cur_ft;

	  /* x coord is fine, but we need to flip y coords because postscript
	     thinks the origin is in the lower left. */
	  cur_fbox->xpos = (Dgp_UInt16)(boxes[j].xpos);              
	  // Bug squashed.  Need to subtract one from token height here, else boundary tokens overrun... / dhtesq.
	  cur_fbox->ypos = (Dgp_UInt16)(zephPage->height - 1 - (boxes[j].ypos + boxes[j].tok->height - 1));

	  /* Needed for color, may be filled in later if needed. */
	  cur_fbox->parent = NULL;

	  /* Figure out if we need to do something with this ftoken. */
	  switch(cur_ft->font_num) {
	  case DGP_UNASSIGNED:
	    /* Token is not in a font... */
	    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(local_fboxes, cur_fbox, &dummy));
	    break;
	  default:
	    num_chars++;
	  }	
	  cur_fbox++;
	}

	/* Write local image with tokens that are not in fonts. */
	// Looks like any per-page singletons get blatted onto a packed image and then
	// g4'd. / dhtesq.
	if (DGPADT_LILEN(local_fboxes) > 0) {
	  /* Make space for image in the array that keeps track of local images. */
	  DGP_CHECK_ALLOC_CLEANUP(images = DGP_REALLOC(images, sizeof(residImage)*(num_images + 1)));
	  
	  DGP_ERR_PASS_CLEANUP(write_local_image( local_fboxes, 
						  zephPage->width, 
						  zephPage->height, 
						  num_images,
						  &images[num_images].obj, 
						  &images[num_images].lx, 
						  &images[num_images].ly, 
						  &images[num_images].width, 
						  &images[num_images].height, 
						  state, 
						  pdfCookie,
						  dgp_state));
			       
          /* Image resolution is the same as page resolution... */
	  images[num_images].xres = DgpCoord_floor(zephPage->res.xRes);
	  images[num_images].yres = DgpCoord_floor(zephPage->res.yRes);
	  
	  num_images++;
	}
      }

      /* Build page content description. */
      content_size = num_chars * 1000 + num_images * 1000;
      content_used = 0;
      DGP_CHECK_ALLOC_CLEANUP(content = (Dgp_Int8 *)DGP_MALLOC(content_size * sizeof(Dgp_Int8)));
      content_write_pos = content;
			
      /* Keep trying to build content description if we didn't allocate enough space. */
      do {
	/* Write references to local images. */
	for (j = 0; j < num_images; j++) {
	  // To keep up appearances, omit masked images we can't really mask.
	  //if(images[j].is_mask == DGP_FALSE){
	  content_used += sprintf(content + content_used, "q\n%f 0 0 %f %f %f cm\n/I%d Do\nQ\n", 
				  (Dgp_Float32)(images[j].width * state->dpi) / (Dgp_Float32)images[j].xres, 
				  (Dgp_Float32)(images[j].height * state->dpi) / (Dgp_Float32)images[j].yres, 
				  (Dgp_Float32)(images[j].lx * state->dpi) / (Dgp_Float32)images[j].xres, 
				  (Dgp_Float32)(images[j].ly * state->dpi) / (Dgp_Float32)images[j].yres, 
				  (int)j);
	  //}
	}
	tmp = 0;
	
	if (num_boxes > 0) {
	  /* Write out the tokens that are in fonts. */
	  DGP_ERR_CHECK_NO_MSG(write_page_positions( fboxes, 
						     num_boxes, 
						     DgpCoord_floor(zephPage->res.xRes), 
						     DgpCoord_floor(zephPage->res.yRes),
						     state, 
						     content + content_used, 
						     content_size - content_used, 
						     &tmp,
						     dgp_state)) {
	    if (DGP_ERROR_IS(cErrIntBufOverflow)){
	      /* Ran out of space in content buffer. */
	      DGP_FREE(content);
	      content_size *= 2;
	      content_used = 0;
	      DGP_CHECK_ALLOC_CLEANUP(content = DGP_MALLOC(content_size * sizeof(Dgp_Int8)));
	    } else {
	      goto ERR_CLEANUP;
	    }
	  } else {
	    content_used += tmp;
	  }
	}
      } while (tmp == -1);
      
      /* Compress content. */
      if (deflateInit(&strm, DGP_Z_COMPRESSION) != Z_OK){
	DGP_ERR_CLEANUP(cErrCompressionError);
      }
			
      /* This should always be enough memory for the compressed data. */
      comp_content_size = content_used * 2 + 12;
      DGP_CHECK_ALLOC_CLEANUP(comp_content = DGP_MALLOC(comp_content_size * sizeof(Dgp_Int8)));
      
      strm.next_in = content;
      strm.avail_in = content_used;
      strm.next_out = comp_content;
      strm.avail_out = comp_content_size;
      
      if (deflate(&strm, Z_FINISH) != Z_STREAM_END){
	DGP_ERR_CLEANUP(cErrCompressionError);
      }
      DGP_FREE(content);
      if (deflateEnd(&strm) != Z_OK){
	DGP_ERR_CLEANUP(cErrCompressionError);
      }
      
		/* Write compressed content. */
      contentObj = pdfCookie->objCount++;
      DGP_ERR_PASS_CLEANUP(addXref(contentObj, pdfCookie->offset, state, pdfCookie, dgp_state));
      pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<< /Length %d /Filter /FlateDecode >>\n\
stream\n", (int)contentObj, (int)(strm.total_out));
				
      if (fwrite(comp_content, 1, strm.total_out, outf) != (size_t)(strm.total_out)){
	DGP_ERR_CLEANUP(cErrIO_WriteError);
      }
      pdfCookie->offset += strm.total_out;
      DGP_FREE(comp_content);
      
      /* End content object. */
      pdfCookie->offset += fprintf(outf, "\
endstream\n\
endobj\n");		

      /* Write page object. */
      DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->firstPageObj+i, pdfCookie->offset, state, pdfCookie, dgp_state));
      pdfCookie->offset += fprintf(outf, "\
%d 0 obj\n\
<<\n\
/Type /Page\n\
/Parent %d 0 R\n\
/MediaBox [0 0 %d %d]\n\
/Contents %d 0 R\n", (int)(pdfCookie->firstPageObj + i), (int)pdfCookie->pagesObj, 
				   (int)((Dgp_Float32)(zephPage->width * state->dpi) / (DgpCoord_float(zephPage->res.xRes))),
				   (int)((Dgp_Float32)(zephPage->height * state->dpi) / (DgpCoord_float(zephPage->res.yRes))), (int)contentObj);

      /* Resources might reference some images... */
      if (num_images > 0) {

	pdfCookie->offset += fprintf(outf, "\
/Resources << /ProcSet [/Text /PDF /ImageB] /Font %d 0 R /XObject <<", (int)pdfCookie->fontResourceObj);

	/* reference each image */
	for (j = 0; j < num_images; j++) {
	  pdfCookie->offset += fprintf(outf, " /I%d %d 0 R", (int)j, (int)images[j].obj);
	}
	pdfCookie->offset += fprintf(outf, ">> >>\n>>\nendobj\n");
      } else {

	pdfCookie->offset += fprintf(outf, "\
/Resources << /ProcSet [/Text /PDF] /Font %d 0 R >>\n\
>>\n\
endobj\n", (int)pdfCookie->fontResourceObj);
	
      }

      DGP_REPORTF3(state->verbose, "%s:  page %d/%d written.", state->out_filename, (int)i, (int)state->zephDoc->pageCount);

      /* Cleanup page... */
      if(state->entry){
	DGP_ERR_PASS_CLEANUP(suReleaseReference(state->entry));
	state->entry = NULL;
      }
      // Add more releases here if dealing with color...
      for(j = 0; j < images_on_page; j++){
	DGP_ERR_PASS_CLEANUP(zephCloseImage(&zephImage[j]));
      }
      DgpAdt_liFree(local_fboxes);
      if(images) DGP_FREE(images);
      images = NULL;
      if(zephImage) DGP_FREE(zephImage);
      zephImage = NULL;
      
      DGP_ERR_PASS_CLEANUP(zephClosePage(&zephPage));
      zephPage = NULL;
	}

    /* Cleanup... */
    DGP_FREE(zephImage);
    zephImage = NULL;
    CleanupReusableBuffer(&fboxes_rebuf);
    NORMAL_RETURN;

ERR_CLEANUP:	
    CleanupReusableBuffer(&fboxes_rebuf);
    if (local_fboxes){
      DgpAdt_liFree(local_fboxes);
    }
    if(images){
        DGP_FREE(images);
    }
    if(zephImage){
        for(j = 0; j < zephImageIndex; j++){
	  DGP_ERR_PASS(zephCloseImage(&zephImage[j]));
        }
        DGP_FREE(zephImage);
    }
    if(fetched){
      DGP_ERR_PASS(suReleaseReference(state->entry));
    }
    if(zephPage){
      DGP_ERR_PASS(zephClosePage(&zephPage));
    }

    ERR_RETURN;
}

/* Sort objects in xref table. */
static int
sort_by_obj(const void *a, const void *b)
{
  const xrefElt *ea = a;
  const xrefElt *eb = b;

  return (ea->obj - eb->obj);
}

void
write_pdf_end(DgpPdfState *state, PdfCookie *pdfCookie) {
	FILE *outf = state->outfile;
	Dgp_Int32 xrefPos = pdfCookie->offset;
	Dgp_Int32 i;

	/* Write xref table & trailer. */
	pdfCookie->offset += fprintf(outf, "\
xref\n\
0 %d\n\
0000000000 65535 f \n", (int)(pdfCookie->xrefPos + 1));

	/* Sort xref objects. */
	qsort(pdfCookie->xref, pdfCookie->xrefPos, sizeof(xrefElt), sort_by_obj);

	for (i = 0; i < pdfCookie->xrefPos; i++) {
		pdfCookie->offset += fprintf(outf, "%010d 00000 n \n", (int)pdfCookie->xref[i].offset);
	}

	pdfCookie->offset += fprintf(outf, "\
trailer\n\
<<\n\
/Size %d\n\
/Root %d 0 R\n\
>>\n\
startxref\n\
%d\n\
%%%%EOF\n", (int)(pdfCookie->xrefPos + 1), (int)pdfCookie->catalogObj, (int)xrefPos);

    return;
}

ErrCode
DgpPdf(DgpPdfState *state, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("DgpPdf()", DGP_PERRSTATE);

  PdfCookie pdfCookie;
  Dgp_Int32 pos;
  suOptions options;

  FontTokenInfo *ftoken_info = NULL;
  
  DGP_ERR_PASS_CLEANUPF0(zephOpenDocument(dgp_state, state->docUds, &state->zephDoc),
                         "DgpPdf: Cannot open document\n");

  DGP_ERR_PASS_CLEANUP(suSetDefaults(&options, dgp_state));
  DGP_ERR_PASS_CLEANUP(suInit(&options,
			      state->zephDoc,
                              &state->cache));

  /* Initialize the ftoken_info structure. */
  DGP_ERR_PASS_CLEANUP(init_ftoken_info(&ftoken_info, state, dgp_state));

  /* Figure out which dictionaries (color (if any) and regular) are global
     and which are local (local means the dict is only referenced on
     one page). 
    
    Sort of.  This will only count dictionary references in the first image on the selector.
    Later we'll check these for 'globality', i.e., references > 1. / dhtesq.
  */
  DGP_ERR_PASS_CLEANUP(get_dict_info(ftoken_info, state, dgp_state));

  /* Init pdf cookie. */
  DGP_ERR_PASS_CLEANUP(initPdfCookie(state, &pdfCookie, dgp_state));

  /* Write prolog of pdf file. */
  write_pdf_prolog(state, &pdfCookie);

  /* Make sure there is actually something in the document to convert. */
  if(state->zephDoc->pageCount > 0){
    /* Read all of the regular (non-color) dictionaries, so we can write
       the global fonts at the beginning of the ps file. */
    DGP_ERR_PASS_CLEANUP(retrieveAllDictionaries( state, 
                                                  ftoken_info->num_mono_uniqids,
                                                  ftoken_info->mono_uniqids,
                                                  dgp_state));

    pos = pdfCookie.offset;   // (file position)
    DGP_ERR_PASS_CLEANUP(figure_global_fonts(ftoken_info, state, &pdfCookie, dgp_state));

    DGP_REPORTF1(state->verbose, "%d bytes used for global fonts...", pdfCookie.offset - pos);

    DGP_ERR_PASS_CLEANUP(write_pdf_catalog(state, state->zephDoc->pageCount, &pdfCookie, dgp_state));

    DGP_ERR_PASS_CLEANUP(suReleaseReference(state->entry));

    /* Write out the pages. */
	pos = pdfCookie.offset;
    DGP_ERR_PASS_CLEANUP(write_pdf_pages(ftoken_info, state, &pdfCookie, dgp_state));

    DGP_REPORTF1(state->verbose, "%d bytes used for per-page fonts...", pdfCookie.offset - pos);
  }

  /* Write end of pdf document. */
  write_pdf_end(state, &pdfCookie);

  /* Clean up. */
  /* Cleanup all of the dictionaries. */
  DGP_ERR_PASS_CLEANUP(suDestroy(state->cache));

  DGP_ERR_PASS_CLEANUP(zephCloseDocument(&state->zephDoc));
  state->zephDoc = NULL;
  /*
  if(state->extensions != NULL) DGP_FREE(state->extensions);
  if(state->num_extensions != NULL) DGP_FREE(state->num_extensions);
  if(state->toks != NULL) DGP_FREE(state->toks);
  if(state->num_toks != NULL) DGP_FREE(state->num_toks);
  */

  cleanup_ftoken_info(&ftoken_info, state, dgp_state);

  cleanupPdfCookie(state, &pdfCookie, dgp_state);

  NORMAL_RETURN;

ERR_CLEANUP:
  DGP_ERR_PASS_CLEANUP(suDestroy(state->cache));
  if(state->zephDoc){
      DGP_ERR_PASS(zephCloseDocument(&state->zephDoc));
  }
  cleanup_ftoken_info(&ftoken_info, state, dgp_state);
  ERR_RETURN;
}
