/*************************************************************************
**
**  DigiPaper Codebase
**    
**  Pedro Felzenszwalb, January 1999
** 
**  28.7.2000 / dhtesq.
**  Resurrection...  
**
**	15.3.2001 / dhtesq.
**	Re-resurrection for tfx...
**
**************************************************************************/

/*
 * RCS Header
 *
 * $Header: /project/icons/cvsroot/digipaper/src/DgpPdf/DgpPdf_util.h,v 1.1 2001/03/20 01:36:44 dtracy Exp $
 */

#ifndef __DGPPDF_UTIL_H
#define __DGPPDF_UTIL_H

#include "libzephyr.h"
#include "hash_table.h"
#include <stdio.h>
/*
#ifndef _WIN32
#include "undecl-stdio.h"
#endif
*/
#include "util.h"

/* IMPORTANT, DO NOT CHANGE.  no garbage collecting should be done or else
   the pointers we have to tokens will be invalid!!! */
//#define DGP_GC_AFTER          -1        // This is irrelevant with the new interface... / dhtesq.

#define DGP_SIZE_OF_HT  32764

#define DGP_PS_DPI      72
#define DGP_PS_DPI_STR  "72"
/* Typical maximum string length in postscript. */
//#define DGP_MAX_STRING_LEN 65535

/* Characters per font. */
#define DGP_CHARS_PER_FONT 256

/* Value for unassigned font code or font number. */
#define DGP_UNASSIGNED  -1
/* value for font num of token already written to an image */
#define DGP_IGNORE      -2
/* value for font num of token which is local */
#define DGP_LOCAL       -3

/* values for qsort */
#define DGP_LESS_THAN   -1
#define DGP_EQUAL        0
#define DGP_GREATER_THAN 1

/* compress knob for page content */
#define DGP_Z_COMPRESSION 9

/* default initial size of xref table */
#define XREF_SIZE 10

/* info about residual images */
typedef struct {
	Dgp_Int32 obj;
	Dgp_Int32 width, height;
	Dgp_Int32 xres, yres;
	Dgp_Int32 lx, ly;
    Dgp_Bool is_mask;
} residImage;

/* entry in pdf xref table */
typedef struct {
	Dgp_Int32 obj;
	Dgp_Int32 offset;
} xrefElt;

/* keeps track of pdf file offset, objects and etc */
typedef struct {
	Dgp_Int32 offset;
	Dgp_Int32 objCount;
	Dgp_Int32 catalogObj;
	Dgp_Int32 pagesObj;
	Dgp_Int32 firstPageObj;
	Dgp_Int32 fontResourceObj;
	xrefElt *xref;
	Dgp_Int32 xrefSize;
	Dgp_Int32 xrefPos;
} PdfCookie;

/* structure used to sort tokens into fonts in the ps file. */
typedef struct {
  Dgp_UInt16 width, height; /* token width, token height in pixels */
  Dgp_Int32 font_num, char_code; /* number of font, code in the font holding this token */
  token *token; /* original token */
} FontToken;

/* associates a tokpos with the font character holding the bitmap for the
   token. */
struct _FontPos {
  Dgp_UInt16 xpos, ypos;
  struct _FontPos *parent;
  FontToken *ftoken;
};
typedef struct _FontPos FontPos;

/* structure to pass around info about the fonts defined in the ps file. */
typedef struct {
  FontToken *ftokens;     /* array of all font tokens */
  FontToken **sort_ftokens;  /* array of sortable font token ptrs, which 
				point into 'ftokens' */
  Dgp_Int32 ftoken_num;         /* total # of font tokens */
  Dgp_Int32 global_ftoken_num;  /* # of global font tokens */
  Dgp_Int32 num_global_fonts;   /* # of global fonts */

  Dgp_Int32 *dict_ref_count;    /* reference count for each dictionary */
  Dgp_UInt32 *mono_uniqids;     /* uniqids of non-color dictionaries */
  Dgp_UInt32 num_mono_uniqids; /* number of mono dictionaries */
  /*Dgp_Int32 total_num_uniqids;*/ /* total number of dictionaries in the file */
  Dgp_UInt32 max_uniqid;       /* largest uniqid in the file */

  DgpAdt_HashTable ftokens_hash; /* hash table of all font tokens, 
            			     referenced by toknum */
  Dgp_Bool e1_is_a_tokpos; /* cookie for hashtable */
} FontTokenInfo;

/* Structure for passing around commandline options for DgpPdf. */
// ...and since it's going around anyway, we'll store some of the document
// and dictionary context lost when xiff_ops went extinct. / dhtesq.
typedef struct {
  Dgp_Bool use_local_fonts;         /* use local fonts instead of residuals */
  Dgp_Bool use_pop_pg_width;        /* use doc max page width for all widths */
  Dgp_Bool use_pop_pg_height;       /* use doc max page height for all heights */
  Dgp_Int32 dpi;                    /* dpi for all pages in document */
  Dgp_Bool verbose;
  Dgp_Int8 *out_filename;           /* NULL if stdout... */

  FILE *outfile, *errfile; /* files for input, output, and errors */

  zephDocumentInfo zephDoc;
  uds *docUds;
  suXiffCache cache;
  suCacheEntry entry;
  jb2Segment *segment;
  suTokenDictionary dictionary;
  suTokenPositionBlock pos_block;

  Dgp_Int32 num_uniqids;
  DgpAdt_HashTable tok_hash;
  token *toks_pg;
  token *residual_toks;
  Dgp_Int32 num_residual_toks;

  /*
  This information can be found in the supercache dictionary{}.
  */
  //void **dictblock;
  Dgp_Int32 *num_toks;
  token **toks;
  Dgp_Int32 *num_extensions;
  Extension **extensions;
} DgpPdfState;

/* Do it... */
ErrCode
DgpPdf(DgpPdfState *state, Dgp_State *dgp_state);

#endif 


