/*************************************************************************
**
**  DigiPaper Codebase
**    
**  Pedro Felzenszwalb, January 1999
**  most of the code was stolen from tc2ps
**  
**
**  26.7.2000 / dhtesq.
**  Resurrection;  liberr compliance etc.
**
**	15.3.2001 / dhtesq.
**	Re-resurrection for tfx...
**
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/DgpPdf/encodings.c,v 1.1 2001/03/20 01:36:44 dtracy Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "dgp_state.h"

#include "g4.h"
#include "DgpPdf_util.h"
#include "encodings.h"

/*
#ifndef _WIN32
#include "undecl-stdio.h"
#endif
*/

/* Try to g4 encode this many times, then give up. */
#define DGP_MAX_TRIES 5
/* Increase buffer by 50% on failure. */
#define DGP_BUFFER_INCREASE 1.5


#ifdef _LITTLE_ENDIAN
void
convert_ulong_2_uchar(void *buf, Dgp_Int32 buf_len)
{
  Dgp_Int32 i;
  Dgp_UInt32 tmp_ulong = 0, *lbuf = (Dgp_UInt32 *)buf;
  Dgp_UInt8 *cbuf = (Dgp_UInt8 *)buf;

  for(i = 0; i < (Dgp_Int32)buf_len; i+=sizeof(Dgp_UInt32)) {
    tmp_ulong = *(lbuf)++;

    if(BITS_Dgp_UInt32 == 64) {
      *(cbuf++) = (Dgp_UInt8)((tmp_ulong >> 56) & 0xFF);
      *(cbuf++) = (Dgp_UInt8)((tmp_ulong >> 48) & 0xFF);
      *(cbuf++) = (Dgp_UInt8)((tmp_ulong >> 40) & 0xFF);
      *(cbuf++) = (Dgp_UInt8)((tmp_ulong >> 32) & 0xFF);
    }
    *(cbuf++) = (Dgp_UInt8)((tmp_ulong >> 24) & 0xFF);
    *(cbuf++) = (Dgp_UInt8)((tmp_ulong >> 16) & 0xFF);
    *(cbuf++) = (Dgp_UInt8)((tmp_ulong >> 8) & 0xFF);
    *(cbuf++) = (Dgp_UInt8)(tmp_ulong & 0xFF);    
  }

  return;
}

/* convert a buffer made for char access to a buffer for long access */
void
convert_uchar_2_ulong(void *buf, Dgp_Int32 buf_len)
{
  Dgp_Int32 i;
  Dgp_UInt32 tmp_ulong = 0, *lbuf = (Dgp_UInt32 *)buf;
  Dgp_UInt8 *cbuf = (Dgp_UInt8 *)buf;

  for(i = 0; i < (Dgp_Int32)buf_len; i+= sizeof(Dgp_UInt32)) {
    tmp_ulong = 0;
    if(BITS_ULONG == 64) {
      tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF) << 56;
      tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF) << 48;
      tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF) << 40;
      tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF) << 32;
    }
    tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF) << 24;
    tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF) << 16;
    tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF) << 8;
    tmp_ulong |= ((Dgp_UInt32)(*(cbuf++)) & 0xFF);

    *(lbuf++) = tmp_ulong;
  }

  return;
}
#endif


/* allocate and initialize ReusableBuffer */
ErrCode
InitReusableBuffer(ReusableBuffer *_rebuf, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("InitReusableBuffer()", DGP_PERRSTATE);
  ReusableBuffer rebuf = NULL;

  if(!(*_rebuf)) {
      DGP_CHECK_ALLOC(rebuf = (ReusableBuffer)DGP_CALLOC(sizeof(*rebuf), 1));

      rebuf->a = dgp_state->a;
    
      *_rebuf = rebuf;
  }

  NORMAL_RETURN;
}

/* cleanup any memory held in the ReusableBuffer */
void
CleanupReusableBuffer(ReusableBuffer *_rebuf)
{
  ReusableBuffer rebuf = *_rebuf;
  Dgp_Allocators *a;

  if(rebuf) {
    a  = &rebuf->a;
    if(rebuf->buf) {
      (a->free)(rebuf->buf);
    }
    rebuf->buf_len = 0;
    (a->free)(rebuf);
    *_rebuf = NULL;
  }

  return;
}

/* g4 encode data and write it to the file, encoding the resulting binary data
   according to the command line parameters (hex, ascii85 or binary). */
ErrCode
g4encode(ReusableBuffer rebuf, 
         void *raw_buf, 
         Dgp_Int32 width,
         Dgp_Int32 height, 
         Dgp_Int32 longs_per_row, 
         DgpPdfState *state, 
         Dgp_Int8 **_buf, 
         Dgp_Int32 *_buf_len,
         Dgp_State *dgp_state)
{
  DGP_PROLOGUE("g4encode()", DGP_PERRSTATE);

  Dgp_UInt8 *g4_buf = NULL;
  Dgp_Int32 cur_buf_len = 0, ret_buf_len = 0;
  Dgp_Int32 num_tries = 0;
  Dgp_Bool try_again = DGP_FALSE;

  state = state;
  cur_buf_len = longs_per_row * sizeof(Dgp_UInt32) * height;

  /* Try to g4 encode the data, if it fails, make the buffer larger and 
     try again. */
  do {
    try_again = DGP_FALSE;

    /* Get encoding buffer. */
    // This being Pedro, GetReusableBufferPtr is a _macro_. / dhtesq.
    GetReusableBufferPtr(g4_buf, (cur_buf_len * sizeof(*g4_buf)), rebuf);
    DGP_CHECK_NULL_CLEANUP(g4_buf);
    
    ret_buf_len = cur_buf_len;
    
    /* Try g4 encoding the data. */
    DGP_ERR_CHECK(g4Encode((Dgp_UInt32 *)raw_buf, width, height, longs_per_row,
        g4_buf, (size_t *)&ret_buf_len, DGP_FALSE, dgp_state)){
        num_tries++;
        if(num_tries >= DGP_MAX_TRIES){
            DGP_ERR_GENF1(DGP_TRUE, cErrCompressionError, "g4 encoding failed after %d attempts.", num_tries);
        }
        try_again = DGP_TRUE;
        cur_buf_len = (Dgp_Int32)(cur_buf_len * DGP_BUFFER_INCREASE);
    }

  } while(try_again);

  *_buf = g4_buf;
  *_buf_len = ret_buf_len;

  NORMAL_RETURN;

ERR_CLEANUP:
  ERR_RETURN;
}
