/*
 *
 * verify-ops.c
 *
 * 5.10.2000 / dhtesq.
 * Support functions for dgpver.
 *
 *
 */

#include "verify.h"

void
DgpVer_reportFindings( Dgp_State *dgp_state,
                Dgp_Int8 **infiles,
                Dgp_Bool documentVerified,
                Dgp_Int32 refBoxCount,
                Dgp_Int32 scanBoxCount,
                Dgp_Int32 extraScanClassCount,
                DgpAdt_List stinkers,
                Dgp_Int32 x_reg,
                Dgp_Int32 y_reg,
                FILE *bb)
{
  //DGP_VOID_PROLOGUE("reportFindings()", DGP_PERRSTATE);

  dgp_state = dgp_state;

  if(documentVerified == DGP_TRUE){
      fprintf(stdout, "\nAuthenticity of \"%s\" verified.\n", infiles[1]);
      if(bb){
          DgpVer_writeStinkers(stinkers, bb, infiles, x_reg, y_reg, DGP_FALSE);
      }
      return;
  }

  // What kind of failure?...
  if(extraScanClassCount){
      fprintf(stdout, "\nVerification failure.  Parts of \"%s\" are unrecognizable.\n", infiles[1]);
      if(bb){
          // Stinkers are scan boxes...
          DgpVer_writeStinkers(stinkers, bb, infiles, x_reg, y_reg, DGP_FALSE);
      }
      return;
  }

  if(scanBoxCount > refBoxCount){
      fprintf(stdout, "\nVerification failure.  Document \"%s\" may contain additions.\n", infiles[1]);
      if(bb){
          // Stinkers are scan boxes...
          DgpVer_writeStinkers(stinkers, bb, infiles, x_reg, y_reg, DGP_FALSE);
      }
      return;
  }

  if(scanBoxCount < refBoxCount){
      fprintf(stdout, "\nVerification failure.  Document \"%s\" may contain erasures.\n", infiles[1]);
      if(bb){
          // Stinkers are ref boxes...
          DgpVer_writeStinkers(stinkers, bb, infiles, x_reg, y_reg, DGP_TRUE);
      }
      return;
  }

/*
Relocations should be detected as misregistrations, erasures, and/or additions...

  if(scanBoxCount == refBoxCount){
      fprintf(stdout, "\nVerification failure.  Document \"%s\" may contain relocations.\n", infiles[1]);
      VOID_RETURN;
  }
*/

  return;
}


/*
Searches xrefList's classes for a token (on pgnum 1) which matches ref_box (on pgnum 0);  
ref_box already exists somewhere in the xrefList's classes.

Returns DGP_TRUE if the match is not found within the specified distance (see below).

x_reg, y_reg, heightDiff are for registration of the matched box;  tolerance is radius of the circle
that defines "around here".  scan_box is used for its position information.

Not very smart.
*/
static INLINE Dgp_Bool
tokenHasNoMatchAroundHere(  dgplssy_cc          *ref_box,
                            dgplssy_cc          *scan_box,
                            Dgp_Int32   x_reg,
                            Dgp_Int32   y_reg,
                            Dgp_Int32   heightDiff,
                            Dgp_Int32   tolerance,
                            DgpAdt_List xrefList)
{
  DGP_VOID_PROLOGUE("tokenHasNoMatchAroundHere()", NULL);
  DgpAdt_ListNode ln0, ln1, ln2;
  dgplssy_cc *sbox, *rbox;
  Dgp_Int32 dx, dy, ypos;
  Dgp_Float64 d;
  ccXref *xref = NULL;
  tokCBoxType *cbox = NULL;

  scan_box = scan_box;

  DGPADT_FOREACH(ln0, xrefList){
      xref = (ccXref *)DGPADT_LIGET(ln0);
      DGP_VOID_ASSERT(xref != NULL);
      //scan_box = xref->box;

      // Might have a loner...
      if(!xref->classSiblings && (xref->box->pgnum >= 1)){
          continue;
      }


      DGPADT_FOREACH(ln1, xref->classSiblings){
          cbox = tokCMemberBox(ln1);
          DGP_VOID_ASSERT(cbox != NULL);
          rbox = (dgplssy_cc *)cbox->userdata;
          DGP_VOID_ASSERT(rbox != NULL);
          if(rbox == ref_box){
              // Check this class for scan matches in the region.
              DGPADT_FOREACH(ln2, xref->classSiblings){
                  cbox = tokCMemberBox(ln2);
                  DGP_VOID_ASSERT(cbox != NULL);
                  sbox = (dgplssy_cc *)cbox->userdata;
                  DGP_VOID_ASSERT(sbox != NULL);

                  if(sbox->pgnum == 0) continue;
                  // Try a bottom alignment if it's closer...
                  if(abs(rbox->y - y_reg - sbox->y) > heightDiff){
                      // Align with bottom...
                      ypos = sbox->y + (sbox->bitmap.height - rbox->bitmap.height);
                  } else {
                      // Align with top...
                      ypos = sbox->y;
                  }

                  dx = abs(rbox->x - sbox->x - x_reg);
                  dy = abs(rbox->y - ypos - y_reg);
                  d = sqrt(dx*dx + dy*dy);
                  if(d < (DGPVER_NEIGHBORHOOD_MATCH_RADIUS * (Dgp_Float64)tolerance)){
                      return DGP_FALSE;
                  }
              }
              return DGP_TRUE;
          }
      }
  }

  // Since our xref list is based on scanned tokens, it's possible that a given reference token
  // exists nowhere on the class sibling lists.  Reference tokens matched nowhere on the scan
  // trigger this condition.  That's the only way we can reach here.
  return DGP_TRUE;
}


// Just reports if any of the passed cc is still on the passed (binary) image.
// xoffset and yoffset will be added to the passed cc's coordinates.
static INLINE Dgp_Bool 
ccHasDisappeared(   Dgp_UInt32 *image,
                    Dgp_Int32 lpl,
                    dgplssy_cc *box,
                    Dgp_Int32 xoffset,
                    Dgp_Int32 yoffset)
{
  //DGP_VOID_PROLOGUE("ccHasDisappeared()", NULL);

  Dgp_Int32 i, j, xpos, longstouched;
  Dgp_UInt32 *runner, *rowptr;
  Dgp_UInt32 leftmask, rightmask, mask = 0;

  image += (lpl * (box->y + yoffset));                  // Starting row...
  image += (box->x + xoffset) / DGP_BITS_UINT32;        // Starting column...

  xpos = (box->x + xoffset) & (DGP_BITS_UINT32 - 1);
  longstouched = (xpos + box->bitmap.width - 1) / DGP_BITS_UINT32 + 1;
  leftmask = (~0ul) >> xpos;

  if (((xpos + box->bitmap.width) & (DGP_BITS_UINT32 - 1)) == 0) {
      rightmask = ~0ul;
  }
  else {
      rightmask = (~0ul) <<
          (DGP_BITS_UINT32 - 1 - ((xpos + box->bitmap.width - 1) & (DGP_BITS_UINT32 - 1)));
  }

  if(longstouched == 1) {
      leftmask &= rightmask;
      rightmask = leftmask;
  }

  for(i = 0, rowptr = image; i < box->bitmap.height; i++, rowptr += lpl){
      runner = rowptr;
      for(j = 0; j < longstouched; j++){          
          if(j == 0) mask = leftmask;
          if(j == (longstouched - 1)) mask = rightmask;
          if(*runner++ & mask){
              return DGP_FALSE;
          }
      }
  }
  return DGP_TRUE;
}


/*************************************************************************
 *
 * Here's the brains of the verification...
 *
 *************************************************************************/
ErrCode
DgpVer_processInfiles(
	Dgp_State *dgp_state, 
	TCLYDocCookie tdc, 
	DgpTdc_Cookie llTdc,
	Dgp_Int8 **infiles,
    tokOptions *tokopt,
    Dgp_Int32 driftTolerance,
    Dgp_Int8 *postMortem,
    FILE *bb,
	Dgp_Bool report) 
{
  DGP_PROLOGUE("DgpVer_processInfiles()", DGP_PERRSTATE);

  /* This is a variable that is passed into
   * WriteCompressed(). It is passed into writeDictionary()
   * for successive calls - and is only cleared
   * when it is time for a clear dictionary.
   */  
  TCLYSectionBlockPtr sbPtr = NULL;
  DgpTdc_SectionBlock *llSb = NULL;
  TCImagePtr tcImagePtr[2];
  Dgp_Int32 pageNum = 0;            /* pageNum is the global page count for all images */
  Dgp_Int32 i, j;                      /* j is the page number within an image */
  Dgp_Int32 boxCount[2];
  Dgp_Int32 newClassesFound[2], numMatches, lonerCount, lonerClassCount, lastLonerClassCount;
  DgpAdt_List boxList[2], matchList = NULL, lonerList = NULL, classList = NULL, xrefList = NULL, secondaryXrefList = NULL;
  DgpAdt_List stinkers = NULL, confidents = NULL, secondaryConfidents = NULL, refSuccessStories = NULL;
  DgpAdt_ListNode ln, aln, closest_node = NULL, node = NULL, retnode;
  dgplssy_cc *box = NULL, *ref_box = NULL, *scan_box = NULL;
  dgplssy_cc **ccArray;
  Dgp_Int32 ccArrayMemberCount;
  ccXref *xref = NULL;
  Dgp_Int32 ypos, ref_ypos, xd, yd;
  Dgp_Float64 d, d_min;
  Dgp_Int32 xRegGlobal, yRegGlobal, xRegLocal, yRegLocal, unusedInt;
  Dgp_Int32 passCount, extraBoxesAxed;
  //Dgp_Int32 dilationRadius = tokopt->dilationRadius;
  //Dgp_Int32 widthDiff = tokopt->widthDiff;
  Dgp_Int32 heightDiff = tokopt->heightDiff;
  Dgp_Int32 dx, dy;
  Dgp_Int32 scanTokensFound = 0;
  tokClass *c = NULL;
  tokOffset *offset = NULL;
  //tokCBoxType *cbox = NULL;
  Dgp_Bool documentVerified = DGP_FALSE;
  TIFF *tiff = NULL;

  tcImagePtr[0] = tcImagePtr[1] = NULL;

  // Set up the stinkers, confidents lists.
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&stinkers, dgp_state));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&confidents, dgp_state));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&refSuccessStories, dgp_state));

  /* Start a section. */
  DGP_ERR_PASS_CLEANUP(tcly_doc_startSection(	tdc, 
		                        				sbPtr,
												0,
                                                report));
  DGP_ERR_PASS_CLEANUP(DgpTdc_startSection(llTdc, llSb, report));

  // Initial pass.  The idea here is to
  // 1) generate a dictionary for the reference page, and
  // 2) generate a list of tokens on the scan that do *not* match the dictionary.
  for(j = 0; j < 2; j++){
      DGP_ERR_PASS_CLEANUP(tc_readTiffEx(	dgp_state, 
                                            infiles[j], 
                                            0,
                                            j,
                                            &tcImagePtr[j]));
      if(tcImagePtr[j] != NULL) {
          /* We got the page... */
          DGP_REPORTF1(report, "Analyzing \'%s\'...", infiles[j]);
      
          if((tcImagePtr[j]->xRes < 1) || (tcImagePtr[j]->yRes < 1)){
              tcImagePtr[j]->xRes = tcImagePtr[j]->yRes = DEF_OUTPUT_RES;
              DGP_WARNF1("No resolution given in input file;  assuming %d.", tcImagePtr[j]->xRes);
          }
      
          /* Check if we are dealing with binary image.  If not, bolt. */
          DGP_ERR_GENF0((tcImagePtr[j]->bpp > 1), cErrPanic,
              "I don't know what to do with more than 1 bit per pixel...");

          if(j == 0){
              DGP_ERR_PASS_CLEANUP(tcly_doc_addMonoPage(  tcImagePtr[j]->image,
                                                          tdc, 
														  llTdc,
                                                          tcImagePtr[j]->xSize, 
                                                          tcImagePtr[j]->ySize,
                                                          tcImagePtr[j]->bpl, 
                                                          tcImagePtr[j]->xRes, 
                                                          tcImagePtr[j]->yRes, 
                                                          pageNum++,
                                                          DGP_FALSE,
                                                          report));
              tcly_doc_retrieveBoxList(tdc, &boxList[j]);
              tcly_doc_retrieveBoxCount(tdc, &boxCount[j]);
          } else {
              // Run this multiple times, removing positional matches in between calls.
              // When finished, the box list will only contain unmatched tokens.
              //
              // Remember we're not adding anything to the dictionary here;  that comes entirely
              // from the fiducial page.
              DGP_ERR_PASS_CLEANUP(tcly_doc_addMonoPageVerify(  tcImagePtr[j]->image,
                                                                tdc, 
																llTdc,
                                                                tcImagePtr[j]->xSize, 
                                                                tcImagePtr[j]->ySize,
                                                                tcImagePtr[j]->bpl, 
                                                                tcImagePtr[j]->xRes, 
                                                                tcImagePtr[j]->yRes, 
                                                                pageNum++,
																MakeModelsFromRunsWithoutTellingTokenizer,
                                                                report,
                                                                &newClassesFound[j],
                                                                &boxCount[j],
                                                                &boxList[j],
                                                                &xrefList,
                                                                &numMatches,
                                                                &matchList));

              // Get some global registration data.  We're not interested in the registration numbers per se;  we want to know
			  // which matched scan tokens match a reference token in roughly the same place.
			  // Some of the matches on the scan
              // are "accidental" -- they correspond to reference tokens that are far away.  So we'll check here for a rough positional match;
              // if one exists, we'll remove the scan token from the scan.  A subsequent call to the cc finder will tag the remainder as loners.
              scanTokensFound = numMatches;
              DGP_ERR_PASS_CLEANUP(DgpVer_retrieveGlobalRegistration(   dgp_state,
                                                                        xrefList,
                                                                        confidents,
                                                                        driftTolerance,
                                                                        &xRegGlobal,
                                                                        &yRegGlobal,
                                                                        &scanTokensFound,
                                                                        report));

              // Now use the list of confidently registered tokens to wipe the scan.
              DGP_ERR_PASS_CLEANUP(DgpVer_removeCCsFromImage(   dgp_state,
                                                                confidents,
                                                                tcImagePtr[j]->image,
                                                                tcImagePtr[j]->xSize,
                                                                tcImagePtr[j]->ySize,
                                                                tcImagePtr[j]->bpl,
                                                                32));     // Pixr boundary...

			  // Now the scan contains mismatches and additions.
              // If any of the remaining tokens matches, it's an accident.  These are all loners.
              DGP_ERR_PASS_CLEANUP(tcly_doc_addMonoPageVerify(  tcImagePtr[j]->image,
                                                                tdc, 
																llTdc,
                                                                tcImagePtr[j]->xSize, 
                                                                tcImagePtr[j]->ySize,
                                                                tcImagePtr[j]->bpl, 
                                                                tcImagePtr[j]->xRes, 
                                                                tcImagePtr[j]->yRes, 
                                                                pageNum++,
																MakeModelsFromRunsWithoutTellingTokenizer,
                                                                report,
                                                                &lonerClassCount,
                                                                &lonerCount,
                                                                &lonerList,
                                                                NULL,
                                                                NULL,
                                                                NULL));
              // Shouldn't find anything new here...
              DGP_ASSERT(lonerClassCount == newClassesFound[1]);

              /* DGP_ERR_PASS_CLEANUP(writeMonoPage(dgp_state, tcImagePtr[j], "d:\\testbed\\tif\\foo.tif", &tiff, DGP_TRUE, DGP_TRUE)); */
          }
      }
  }

  // In a perfect world we'd be done by now, but symbol matching even with very lenient parameters
  // will account for only ~90% of cc's.
  DGP_REPORTF1(report, "I was able to match ~%.1f%% of the cc's by pure dilation...", 
      100 *((Dgp_Float32)(boxCount[0] - lonerCount)/(Dgp_Float32)boxCount[0]));

  // Subsequent token matching will take place against bitmaps that may have been ravaged by adjacent blattings etc.
  // Be a little more forgiving with the dilation now.
  tokopt->dilationRadius = 20;		// Non-tp preserving...

  // We have a rough (non-local) registration of the scan at this point.

  // Some tokens on the scan ("loners") haven't matched anything on the reference.
  // We try to account for them by blatting the probable corresponding reference bitmap
  // over their sorry selves.
  //
  // Our registration hasn't been perfect, so we'll (effectively) blat a dilated bitmap.
  // In practice this does a good job of wiping out the first part of the pair/triplet/whatever.
  //
  // When we run the cc finder again, this will have the benefit of having separated
  // letter combinations that are accidentally closed by the print-and-scan process (this happens
  // quite often).
  //
  // We'll repeat this until there are no new classes (we've matched everything), or
  // the class count stops changing (a bugger persists that won't be matched).

  lastLonerClassCount = -1;
  passCount = 0;
  while((lonerClassCount || lonerCount) && (lastLonerClassCount != lonerClassCount)){
      lastLonerClassCount = lonerClassCount;

      DGP_ERR_PASS_CLEANUP(DgpVer_sortListByPosition(dgp_state, lonerList, &ccArray, &ccArrayMemberCount));

      for(i = 0; i < ccArrayMemberCount; i++){
          box = ccArray[i];
          DGP_ASSERT(box);

          closest_node = NULL;
          d_min = (Dgp_Float64)INT_MAX;

		  // For overblatting, we need to be as close as possible.  Get some local registration data by looking
		  // at matched cc's within a certain radius of this box.
          DGP_ERR_PASS_CLEANUP(DgpVer_retrieveLocalRegistration(dgp_state, xrefList, box, driftTolerance, xRegGlobal, yRegGlobal, &xRegLocal, &yRegLocal));

		  // What is the closest reference box to this scan box?
          DGPADT_FOREACH(aln, boxList[0]){
              ref_box = (dgplssy_cc *)DGPADT_LIGET(aln);
              // Don't use the same reference box twice for an overblat.
              if(ref_box->tok == DGPVER_SIAMESETWIN){
                  continue;
              }

              // Try a bottom alignment if it's closer...
              if(abs(ref_box->y - yRegLocal - box->y) > heightDiff){
                  // Align with bottom...
                  ypos = box->y + (box->bitmap.height - ref_box->bitmap.height);
              } else {
                  // Align with top...
                  ypos = box->y;
              }

              xd = abs(ref_box->x - box->x - xRegLocal);
              yd = abs(ref_box->y - ypos - yRegLocal);
              if((xd > (DGPVER_OVERBLAT_RADIUS * driftTolerance)) || (yd > (DGPVER_OVERBLAT_RADIUS * driftTolerance))) continue;
              d = sqrt(xd*xd + yd*yd);
              if(d < d_min){
                  // The reference box we're using ought to have no scan match in the area...
                  if(tokenHasNoMatchAroundHere(ref_box, box, xRegLocal, yRegLocal, heightDiff, driftTolerance, xrefList)){
                      d_min = d;
                      closest_node = aln;
                  }
              }
          }
          if(!closest_node){
              DGP_REPORTF4(report, "No reference box found near %d x %d loner @ %d, %d;", box->bitmap.width, box->bitmap.height, box->x, box->y);
              // This could mean a cc has been added to the scan...
              continue;
          }
          ref_box = (dgplssy_cc *)DGPADT_LIGET(closest_node);
          closest_node = NULL;
          
          // Do we want to align the reference token with the top or the bottom of the Siamese twin?
          if(abs(ref_box->y - yRegLocal - box->y) > heightDiff){
              // Align with bottom...
              ypos = box->y + (box->bitmap.height - ref_box->bitmap.height);
              ref_ypos = ref_box->y + (box->bitmap.height - ref_box->bitmap.height);
          } else {
              // Align with top...
              ypos = box->y;
              ref_ypos = ref_box->y;
          }

          // Can't get the dilation working, but who needs it when blatting is this fast?
          // Blat our bitmap at its nominal position and at all n-pixel displacements, which
          // for these purposes is equivalent to dilating the bitmap and then blatting it.
          for(dx = -2; dx < 3; dx++){
              for(dy = -2; dy < 3; dy++){
                  DgpBlat_blitBitsOp((Dgp_UInt32 *)(tcImagePtr[1]->image),
                      tcImagePtr[1]->bpl / sizeof(Dgp_Int32),
                      0, 0,
                      tcImagePtr[1]->xSize + 32, tcImagePtr[1]->ySize + 32,
                      ref_box->bitmap.bits,
                      ref_box->bitmap.longsperline,
                      ref_box->bitmap.width,
                      ref_box->bitmap.height,
                      box->x + 32 + dx, 
                      ypos + 32 + dy,       // Boundary around pixr...
                      DGPBLAT_OP_COVER,		// (~reference & scan)
                      dgp_state);
              }
          }          

          // Reciprocate on the reference image.  If this overblatting process leaves cc's on the scan,
          // it means the reference token was too small;  if it leaves cc's on the reference, it means
          // the reference token was too big.  This might reveal an erasure on the scan.
          for(dx = -1; dx < 2; dx++){
              for(dy = -1; dy < 2; dy++){
                  DgpBlat_blitBitsOp((Dgp_UInt32 *)(tcImagePtr[0]->image),
                      tcImagePtr[0]->bpl / sizeof(Dgp_Int32),
                      0, 0,
                      tcImagePtr[0]->xSize + 32, tcImagePtr[0]->ySize + 32,
                      box->bitmap.bits,
                      box->bitmap.longsperline,
                      box->bitmap.width,
                      box->bitmap.height,
                      ref_box->x + 32 + dx, 
                      ref_ypos + 32 + dy,       // Boundary around pixr...
                      DGPBLAT_OP_COVER,
                      dgp_state);
              }
          }

          // This is suboptimal, but I'm going to use the cc's tok field for userdata -- don't need it for anything else.
          // Record the fact that we've used this cc to bash apart a Siamese twin, and so it won't have anything to positionally
          // match with later.
          ref_box->tok = DGPVER_SIAMESETWIN;

          // Check if contained tokens are still there...
          // (This happens with disconnected "e"'s, for example.)
          extraBoxesAxed = 0;
          DGPADT_FOREACH(node, lonerList){
              scan_box = (dgplssy_cc *)DGPADT_LIGET(node);
              if(scan_box == box) continue;
              if((scan_box->x > box->x - 2) && (scan_box->x + scan_box->bitmap.width < box->x + box->bitmap.width + 2)){
                  if((scan_box->y > box->y - 2) && (scan_box->y + scan_box->bitmap.height < box->y + box->bitmap.height + 2)){

                      if(ccHasDisappeared(  (Dgp_UInt32 *)tcImagePtr[1]->image,
                                            tcImagePtr[1]->bpl / sizeof(Dgp_Int32),
                                            scan_box,
                                            32, 32)){
                          extraBoxesAxed++;
                      }
                  }
              }
          }

          scanTokensFound++;
          //scanTokensFound -= extraBoxesAxed;
      } // end loner list traversal...

      //goto ENDGAME;
      DGP_FREE(ccArray);
      // Free loner list bitmaps...
      DGPADT_FOREACH(ln, lonerList){
          box = (dgplssy_cc *)DGPADT_LIGET(ln);
          DGP_FREE(box->bitmap.bits);
      }

      // Now try to match what's left on the scan against the existing dictionary.
      // Twice:  once to get tokens that can be removed, and again to get a valid list of remaining, unmatched tokens.
      //if(passCount == 1) goto ENDGAME;
      DGP_ERR_PASS_CLEANUP(tcly_doc_addMonoPageVerify(  tcImagePtr[1]->image,
                                                        tdc, 
														llTdc,
                                                        tcImagePtr[1]->xSize, 
                                                        tcImagePtr[1]->ySize,
                                                        tcImagePtr[1]->bpl, 
                                                        tcImagePtr[1]->xRes, 
                                                        tcImagePtr[1]->yRes, 
                                                        pageNum++,
														MakeModelsFromRunsWithoutTellingTokenizer,
                                                        report,
                                                        &lonerClassCount,
                                                        &lonerCount,
                                                        &lonerList,
                                                        &secondaryXrefList,
                                                        &numMatches,
                                                        NULL));
      scanTokensFound += numMatches;

      DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&secondaryConfidents, dgp_state));

      DGP_ERR_PASS_CLEANUP(DgpVer_retrieveGlobalRegistration(   dgp_state,
                                                                secondaryXrefList,
                                                                secondaryConfidents,
                                                                driftTolerance,
                                                                &unusedInt,
                                                                &unusedInt,
                                                                &scanTokensFound,
                                                                report));

      // Use the list of confidently registered tokens to wipe the scan.
      DGP_ERR_PASS_CLEANUP(DgpVer_removeCCsFromImage(   dgp_state,
                                                        secondaryConfidents,
                                                        tcImagePtr[1]->image,
                                                        tcImagePtr[1]->xSize,
                                                        tcImagePtr[1]->ySize,
                                                        tcImagePtr[1]->bpl,
                                                        32));     // Pixr boundary...

      // Cleanup...
      DGPADT_FOREACH(ln, lonerList){
          box = (dgplssy_cc *)DGPADT_LIGET(ln);
          DGP_FREE(box->bitmap.bits);
      }

      DGP_ERR_PASS_CLEANUP(tcly_doc_addMonoPageVerify(  tcImagePtr[1]->image,
                                                        tdc, 
														llTdc,
                                                        tcImagePtr[1]->xSize, 
                                                        tcImagePtr[1]->ySize,
                                                        tcImagePtr[1]->bpl, 
                                                        tcImagePtr[1]->xRes, 
                                                        tcImagePtr[1]->yRes, 
                                                        pageNum++,
														MakeModelsFromRunsWithoutTellingTokenizer,
                                                        report,
                                                        &lonerClassCount,
                                                        &lonerCount,
                                                        &lonerList,
                                                        NULL,
                                                        &numMatches,
                                                        NULL));

      if(secondaryXrefList){
          DGPADT_FOREACH(ln, secondaryXrefList){
              xref = (ccXref *)DGPADT_LIGET(ln);
              DGP_ASSERT(xref);
              DGP_FREE(xref);
          }
          DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(secondaryXrefList));
          secondaryXrefList = NULL;
      }
      DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(secondaryConfidents));
      secondaryConfidents = NULL;

      //if(passCount == 1) goto ENDGAME;
      DGP_REPORTF3(report, "Pass %d: %d cc's, %d new classes...", passCount++, lonerCount, lonerClassCount);
  } // end while

  // Cleanup...
  DGPADT_FOREACH(ln, lonerList){
      box = (dgplssy_cc *)DGPADT_LIGET(ln);
      DGP_FREE(box->bitmap.bits);
  }

  if(lonerClassCount){
      // There is a cc in the scan we have been unable to match with anything in the reference.
      // This constitutes a verification failure -- addition or modification.
      DGP_REPORTF1(report, "%d unmatchable tokens persist in the scan.", lonerClassCount);

      // Everything on the loner list is now a stinker.
      DGPADT_FOREACH(ln, lonerList){
          box = (dgplssy_cc *)DGPADT_LIGET(ln);
          DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(stinkers, (void *)box, &retnode));
      }
      goto ENDGAME;
  }

  scanTokensFound += lonerCount;
  if(scanTokensFound != boxCount[0]){
      // There is an erasure or addition on the scan, or for separated character extinguished by one overblat.
      DGP_REPORTF2(report, "All scanned tokens match, but box counts do not: reference %d, scan %d.", boxCount[0], scanTokensFound);
      // Don't bail yet -- if there's a mismatch, we'd like to know where it is.
      //goto ENDGAME;
  }

  // If we reach here, we have matched every cc in the scan above our size threshold with something in
  // the reference dictionary.  Additionally, the "number of tokens" in the scan and reference are
  // identical, meaning scan erasures are only possible if a token has been added somewhere else on the page.
  //
  // So the remaining fraud modes are:
  //
  // 1) cc's on the scan have been replaced with other tokens from the reference dictionary.
  // 2) cc's have been erased from the scan and relocated elsewhere on the page.
  // 3) cc's on the scan have been transposed.
  // (4) there was an erasure/addition and we're just trying to figure out where.)
  //

  // We're done with the tokenizer;  get the class list...

  DGP_ERR_PASS_CLEANUP(tcly_doc_endSection(	tdc,
                                            &sbPtr,
                                            DGP_TRUE, 
                                            report));
  DGP_ERR_PASS_CLEANUP(tcly_doc_retrieveClassList(tdc, sbPtr, &classList));

  DGPADT_FOREACH(ln, classList){
      c = (tokClass *)DGPADT_LIGET(ln);
      DGP_ASSERT(c != NULL);

      DGPADT_FOREACH(aln, c->members){          
          offset = (tokOffset *)DGPADT_LIGET(aln);
          DGP_ASSERT(offset != NULL);
          ref_box = (dgplssy_cc *)offset->userdata;
          DGP_ASSERT(ref_box != NULL);

          if((ref_box->pgnum >= 1) || (ref_box->tok == DGPVER_SIAMESETWIN)){
              continue;
          }

          // For each class member we find on the reference, we require that one exist on the scan
          // in the same place, within our registration, give or take some slop.
          closest_node = NULL;
          d_min = (Dgp_Float64)INT_MAX;
          DGPADT_FOREACH(node, c->members){
              offset = (tokOffset *)DGPADT_LIGET(node);
              DGP_ASSERT(offset != NULL);
              scan_box = (dgplssy_cc *)offset->userdata;
              DGP_ASSERT(scan_box != NULL);

              if(scan_box->pgnum == 0){
                  continue;
              }

              // I'd like to use local registration here, but the tokenizer scrambles the xref classSibling list after closing the section.
              //DGP_ERR_PASS_CLEANUP(DgpVer_retrieveLocalRegistration(dgp_state, xrefList, scan_box, driftTolerance, xRegGlobal, yRegGlobal, &xRegLocal, &yRegLocal));
              xd = abs(ref_box->x - scan_box->x - xRegGlobal);
              yd = abs(ref_box->y - scan_box->y - yRegGlobal);
              d = sqrt(xd*xd + yd*yd);
              if((d < d_min) && (d < (DGPVER_POSITIONAL_MATCH_RADIUS * driftTolerance))){
                  d_min = d;
                  closest_node = node;
              }
          }
          if(!closest_node){
              // Until we can get the tokenizer to behave, trust there's no erasure if the reference token was used for global registration.
              if(ref_box->tok == DGPVER_GLOBAL_REGISTRATOR){
                  continue;
              }
              // Possible erased (or relocated) token;  failure.
              DGP_REPORTF4(report, "Unable to find match for %d x %d reference token at %d, %d.", 
                  ref_box->bitmap.width, ref_box->bitmap.height, ref_box->x, ref_box->y);
              DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(stinkers, (void *)ref_box, &retnode));

              //goto ENDGAME;
          }

          DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refSuccessStories, (void *)ref_box, &retnode));
      }
  }

  // Get the success stories off the reference image.
  DGP_ERR_PASS_CLEANUP(DgpVer_removeCCsFromImage(   dgp_state,
                                                    refSuccessStories,
                                                    tcImagePtr[0]->image,
                                                    tcImagePtr[0]->xSize,
                                                    tcImagePtr[0]->ySize,
                                                    tcImagePtr[0]->bpl,
                                                    32));     // Pixr boundary...

  // Everything left on the reference image is an overblatting failure or an accidental symbol match.
  DGP_ERR_PASS_CLEANUP(tcly_doc_addMonoPageVerify(  tcImagePtr[0]->image,
                                                    tdc, 
													llTdc,
                                                    tcImagePtr[0]->xSize, 
                                                    tcImagePtr[0]->ySize,
                                                    tcImagePtr[0]->bpl, 
                                                    tcImagePtr[0]->xRes, 
                                                    tcImagePtr[0]->yRes, 
                                                    pageNum++,
													MakeModelsFromRunsWithoutTellingTokenizer,
                                                    report,
                                                    &lonerClassCount,
                                                    &lonerCount,
                                                    &lonerList,
                                                    NULL,
                                                    NULL,
                                                    NULL));

  DGP_ERR_PASS_CLEANUP(DgpTdc_endSection(llTdc, &((tdoc_state *)llTdc)->curBlock, DGP_TRUE, report));

  // Anything on the reference image at this point indicates an overblatting failure...probable erasure.
  DGP_REPORTF1(report && lonerCount, "Erasure on scan?  %d cc's persist on the reference.", lonerCount);

  // Add reference loners to stinkers...
  DGPADT_FOREACH(ln, lonerList){
      box = (dgplssy_cc *)DGPADT_LIGET(ln);
      DGP_FREE(box->bitmap.bits);
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(stinkers, (void *)box, &retnode));
  }

  if(DGPADT_LILEN(stinkers)){
      // We have erasures...
      scanTokensFound = -1;
      goto ENDGAME;
  }

  // Pass the forty.
  documentVerified = DGP_TRUE;

ENDGAME:
  if(postMortem){
      DGP_ERR_PASS_CLEANUP(writeMonoPage(dgp_state, tcImagePtr[0], postMortem, &tiff, DGP_TRUE, DGP_TRUE));
  }

  for(j = 0; j < 2; j++){
      /* We don't need the image any more. */
      tc_freeImage(dgp_state, tcImagePtr[j]);
      tcImagePtr[j] = NULL;
  }

  DGP_REPORTF2(report, "\"%s\": %d cc's", infiles[0], boxCount[0]);
  DGP_REPORTF2(report, "\"%s\": %d cc's", infiles[1], boxCount[1]);

  DGPADT_FOREACH(ln, boxList[0]){
      box = (dgplssy_cc *)DGPADT_LIGET(ln);
      DGP_FREE(box->bitmap.bits);
  }
  DGPADT_FOREACH(ln, boxList[1]){
      box = (dgplssy_cc *)DGPADT_LIGET(ln);
      DGP_FREE(box->bitmap.bits);
  }

  if(matchList) DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(matchList));
  matchList = NULL;
  if(lonerList){
      /*
      DGPADT_FOREACH(ln, lonerList){
          box = (dgplssy_cc *)DGPADT_LIGET(ln);
          DGP_FREE(box->bitmap.bits);
      }*/
      DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(lonerList));
      lonerList = NULL;
  }
  if(xrefList){
      DGPADT_FOREACH(ln, xrefList){
          xref = (ccXref *)DGPADT_LIGET(ln);
          DGP_ASSERT(xref);
          DGP_FREE(xref);
      }
      DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(xrefList));
      xrefList = NULL;
  }

  DgpVer_reportFindings(dgp_state, infiles, documentVerified, boxCount[0], scanTokensFound, lonerClassCount, stinkers, xRegGlobal, yRegGlobal, bb);
  if(stinkers) DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(stinkers));
  stinkers = NULL;
  if(confidents) DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(confidents));
  confidents = NULL;
  if(refSuccessStories) DGP_ERR_PASS_CLEANUP(DgpAdt_liFree(refSuccessStories));
  refSuccessStories = NULL;

  NORMAL_RETURN;
  
ERR_CLEANUP:
  if(matchList) DGP_ERR_PASS(DgpAdt_liFree(matchList));
  if(lonerList) DGP_ERR_PASS(DgpAdt_liFree(lonerList));
  if(stinkers) DGP_ERR_PASS(DgpAdt_liFree(stinkers));
  if(confidents) DGP_ERR_PASS(DgpAdt_liFree(confidents));
  if(refSuccessStories) DGP_ERR_PASS(DgpAdt_liFree(refSuccessStories));
  if(xrefList){
      DGPADT_FOREACH(ln, xrefList){
          xref = (ccXref *)DGPADT_LIGET(ln);
          DGP_ASSERT(xref);
          DGP_FREE(xref);
      }
      DGP_ERR_PASS(DgpAdt_liFree(xrefList));
      xrefList = NULL;
  }

  for(j = 0; j < 2; j++){
      tc_freeImage(dgp_state,tcImagePtr[j]);
      tcImagePtr[j] = NULL;
  }
  
  ERR_RETURN;
}
/*************************************************************************/

#if 0
/*************************************************************************/
ErrCode
DgpVer_dilateCC(    Dgp_State *dgp_state,
                    dgplssy_cc *box,
                    PIXR **dest)
{
    DGP_PROLOGUE("DgpVer_dilateCC", DGP_PERRSTATE);

    PIXR *source = NULL;
    STREL *strel = NULL;
    Dgp_Int32 xRes = DEF_OUTPUT_RES;
    Dgp_Int32 yRes = DEF_OUTPUT_RES;
    UInt8 *w_image = NULL;
    Dgp_Int32 y;

    DGP_ERR_PASS_CLEANUP(w_getStrel(    DGP_PIPWSTATE,
                                        strel2,
                                        &strel));
    DGP_ERR_PASS_CLEANUP(w_createPixrExtended(	DGP_PIPWSTATE,
                                        (Int32)box->bitmap.width,
                                        (Int32)box->bitmap.height,
                                        1, 
                                        Int32_TO_ScaledInt1616((Int32)xRes), 
                                        Int32_TO_ScaledInt1616((Int32)yRes), 
                                        32,      // Frame...
                                        1, 1, NULL,
                                        &source));

    //DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(DGP_PIPWSTATE, source, &w_image));
    
    for(y = 0; y < box->bitmap.height; y++){
          w_getLine(DGP_PIPWSTATE, source, y, (UInt8 **)&w_image);
          memcpy(w_image, box->bitmap.bits + (y * box->bitmap.longsperline * sizeof(Dgp_Int32)), box->bitmap.longsperline * sizeof(Dgp_Int32));
    }

    DGP_ERR_PASS_CLEANUP(w_dilate(  DGP_PIPWSTATE,
                            source,
                            strel,
                            dest));

    DGP_ERR_PASS_CLEANUP(w_destroyPixr( DGP_PIPWSTATE, &source));

      /*
      lonerDR = 1;

      //dilationLpl = ((ref_box->bitmap.width + (1 * lonerDR) + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32);
      //DGP_CHECK_ALLOC_CLEANUP(dilationBuffer = (Dgp_UInt32 *)DGP_CALLOC(1, dilationLpl *
        //                                        (ref_box->bitmap.height + (2 * lonerDR)) * sizeof(*dilationBuffer)));

      //memcpy(dilationBuffer, ref_box->bitmap.bits, ref_box->bitmap.height * dilationLpl * sizeof(*dilationBuffer));

      //DGP_ERR_PASS_CLEANUP(DgpVer_dilateCC(dgp_state, ref_box, &dilationPixr));
                                        
      //DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(DGP_PIPWSTATE, dilationPixr, (UInt8 **)&dilationBuffer));
      //DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(DGP_PIPWSTATE, dilationPixr, &dilationLpl));
      //dilationLpl /= sizeof(*dilationBuffer);
      */
                         
      /*
      DgpBlat_blitBitsOp((Dgp_UInt32 *)(tcImagePtr[1]->image),
                         tcImagePtr[1]->bpl / sizeof(Dgp_Int32),
                         0, 0,
                         tcImagePtr[1]->xSize, tcImagePtr[1]->ySize,
                         dilationBuffer,
                         dilationLpl,
                         ref_box->bitmap.width + (1 * lonerDR),
                         ref_box->bitmap.height + (2 * lonerDR),
                         box->x + 32, 
                         ypos + 32,       // Boundary around pixr...
                         DGPBLAT_OP_XOR,
                         dgp_state);

      //DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &dilationPixr));
      //dilationBuffer = NULL;
      */
    
    NORMAL_RETURN;

ERR_CLEANUP:
    if(source){
        DGP_ERR_PASS_CLEANUP(w_destroyPixr( DGP_PIPWSTATE, &source));
    }

    ERR_RETURN;
}
/*************************************************************************/
#endif

/*************************************************************************/
ErrCode 
writeMonoPage(  Dgp_State                   *dgp_state,
                TCImagePtr                  tcImagePtr,
                Dgp_Int8                    *outName,
                TIFF                        **tiff,
                Dgp_Bool                    firstPage,
                Dgp_Bool                    lastPage)
{
    DGP_PROLOGUE("writeMonoPage", DGP_PERRSTATE);

    DGP_ERR_PASS_CLEANUPF1(w_writeEx(DGP_PIPWSTATE, 
                                     outName,
                                     (PIXR *)tcImagePtr->readPixrPtr,
                                     TIFF_G4,
                                     tiff, 
                                     firstPage,
                                     lastPage),
                                     "I couldn't write the pixr for '%s'.\n", outName);
      
    NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}
/*************************************************************************/
