
#include <stdio.h>

#include "dgp_state.h"
#include "dgp_err_control.h"
#include "PsgFix.h"

#include "pdf_util.h"

/* adds an object to the xref table */
ErrCode
addXref(Dgp_Int32       obj, 
        Dgp_Int32       offset, 
        PdfCookie       *pdfCookie,
        Dgp_State       *dgp_state) {
    DGP_PROLOGUE("addXref()", DGP_PERRSTATE);

	if (pdfCookie->xrefPos == pdfCookie->xrefSize) {
		pdfCookie->xrefSize *= 2;
		DGP_CHECK_ALLOC_CLEANUP(pdfCookie->xref = DGP_REALLOC(pdfCookie->xref, pdfCookie->xrefSize * sizeof(xrefElt)));
	}
	pdfCookie->xref[pdfCookie->xrefPos].obj = obj;
	pdfCookie->xref[pdfCookie->xrefPos].offset = offset;
	pdfCookie->xrefPos++;

	NORMAL_RETURN;

ERR_CLEANUP:
	ERR_RETURN;
}

/* init a pdf cookie */
ErrCode
initPdfCookie(PdfCookie *pdfCookie, Dgp_State *dgp_state) {
    DGP_PROLOGUE("initPdfCookie()", DGP_PERRSTATE);
    Dgp_UInt32 XREF_SIZE = 20;

	pdfCookie->offset = 0;

    /* reserve object # 2 for the catalog, # 1 for pages, #3,4 for linearized files */
	pdfCookie->objCount = 3;
    pdfCookie->catalogObj = 2;
    pdfCookie->PagesObj = 1;
    
    DGP_CHECK_ALLOC_CLEANUP(pdfCookie->xref = DGP_MALLOC(sizeof(xrefElt) * XREF_SIZE));
	pdfCookie->xrefPos = 0;
    pdfCookie->DictObj = 0;
	pdfCookie->xrefSize = XREF_SIZE;
    pdfCookie->numPages = 0;
    pdfCookie->pageStart = 0;
    pdfCookie->ProcObj = 0;

    DGP_ERR_PASS (DgpAdt_liNew (&(pdfCookie->pageObjList), dgp_state));
    DGP_ERR_PASS (DgpAdt_liNew (&(pdfCookie->dictList), dgp_state));

    pdfCookie->hasMRC = DGP_FALSE;

    NORMAL_RETURN;

ERR_CLEANUP:
    DGP_FREE(pdfCookie->xref);
	ERR_RETURN;
}

void
cleanupPdfCookie(PdfCookie *pdfCookie, Dgp_State *dgp_state) 
{	
    DgpAdt_ListNode ln;
	if(pdfCookie->xref) DGP_FREE(pdfCookie->xref);
    pdfCookie->xref = NULL;

    DGPADT_FOREACH(ln, pdfCookie->pageObjList)
    {
        DGP_FREE (ln->userdata);
    }
    DgpAdt_liClear (pdfCookie->pageObjList);
    DgpAdt_liClear (pdfCookie->dictList);
    DgpAdt_liFree (pdfCookie->pageObjList);
    DgpAdt_liFree (pdfCookie->dictList);
}   


/* write pdf prolog */
void
write_pdf_prolog(uds *u, PdfCookie *pdfCookie) {
	
	/* write header */
	pdfCookie->offset += fprintf(u->fp, "\
%%PDF-1.4\n%%%c%c%c%c\n", 250, 251, 252, 253);

    return;
}

/* write pdf catalog and pages */
ErrCode
write_pdf_catalog(PdfCookie     *pdfCookie, 
                  uds           *u,
                  Dgp_State     *dgp_state) 
{
    DGP_PROLOGUE("write_pdf_pages()", DGP_PERRSTATE);
    
    DgpAdt_ListNode ln;
    pageInfo *info;

    /* Write page object. */	
    DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->PagesObj, pdfCookie->offset, pdfCookie, dgp_state));

    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Type /Pages\n\
/Count %ld\n\
/Kids [ ", pdfCookie->PagesObj, pdfCookie->pageObjList->nelt);

        /* Write one reference for each object page. */
	    DGPADT_FOREACH(ln, pdfCookie->pageObjList) 
        {
            info = (pageInfo*) ln->userdata;
		    pdfCookie->offset += fprintf(u->fp, "%ld 0 R ", info->pageObj);
	    }
	    pdfCookie->offset += fprintf(u->fp, "\
]\n\
>>\n\
endobj\n\n");

     /* Catalog object, references the pages object. */
     DGP_ERR_PASS(addXref(pdfCookie->catalogObj, pdfCookie->offset, pdfCookie,
dgp_state));
     pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Type /Catalog\n\
/Pages %ld 0 R\n\
>>\n\
endobj\n\n", pdfCookie->catalogObj, pdfCookie->PagesObj);

        NORMAL_RETURN;

ERR_CLEANUP:
        ERR_RETURN;
}


ErrCode
write_jbig2_page(PdfCookie     *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *data,
                   Dgp_UInt32    dataSize,
                   Dgp_UInt32    width,
                   Dgp_UInt32    height,
                   Dgp_UInt32    pageDPI)
{

    DGP_PROLOGUE("write_jbig2_object()", DGP_PERRSTATE);
    Dgp_Int8        iname[64], sizeStr[256];
    Dgp_UInt32      sizeStrLen;
    Dgp_UInt32      pdfWidth, pdfHeight;
    Dgp_UInt32      pdfDefaultDPI = 72;
    pageInfo*       info;

    pdfWidth = width * pdfDefaultDPI / pageDPI;
    pdfHeight = height * pdfDefaultDPI / pageDPI;

    /* Page description */
	
    /* name the image */
    sprintf (iname, "IID%ld", pdfCookie->objCount);

    /* save the number of the page object in a variable and in the page list */
    info = DGP_MALLOC (sizeof (*info));
    info->dictRef = pdfCookie->DictObj;
    info->pageObj = pdfCookie->objCount;
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pdfCookie->pageObjList, (void*) info, NULL));

	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj <<\n\
/Type /Page\n\
/Resources <<\n\
/XObject <<\n\
/%s %ld 0 R\n\
>>\n\
/ProcSet %ld 0 R\n\
>>\n\
/Parent %ld 0 R\n\
/Contents %ld 0 R\n\
/MediaBox [ 0 0 %ld %ld ]\n\
>>\n\
endobj\n\n", pdfCookie->objCount++, iname, info->pageObj + 2, pdfCookie->ProcObj,
    pdfCookie->PagesObj, 
    info->pageObj + 1, pdfWidth, pdfHeight);

    /* Object size */
    sizeStrLen = sprintf (sizeStr, "q %ld 0 0 %ld 0 0 cm /%s Do Q", pdfWidth, pdfHeight, iname);
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount , pdfCookie->offset, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Length %ld\n\
>>\n\
stream\n\
%s\n\
endstream\n\
endobj\n\n", pdfCookie->objCount++, sizeStrLen, sizeStr);

    
    /* JBIG2 Position Block as XObject/Image */
    DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount,
                        pdfCookie->offset, pdfCookie, dgp_state));

    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Width %ld\n\
/Height %ld\n\
/ColorSpace /DeviceGray\n\
/BitsPerComponent 1\n\
/Name /%s\n\
/Filter /JBIG2Decode\n", pdfCookie->objCount++, width, height, iname);
    if (pdfCookie->DictObj > 0) pdfCookie->offset += fprintf (u->fp, "\
/DecodeParms << /JBIG2Globals %ld 0 R >>\n", pdfCookie->DictObj);
    pdfCookie->offset += fprintf (u->fp, "\
/Length %ld\n\
>>\n\
stream\n", dataSize);
    u->fwrite (data, dataSize, 1, u->fp);
      pdfCookie->offset += dataSize;
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");
  
  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

ErrCode
write_jbig2_global(PdfCookie     *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *dict,
                   Dgp_UInt32    dictSize)
{

    DGP_PROLOGUE("write_jbig2_global()", DGP_PERRSTATE);
    
    if (dictSize > 0)
    {

    /* Dictionary */
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset,
        pdfCookie, dgp_state));
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pdfCookie->dictList, (void*) pdfCookie->objCount, NULL));

    pdfCookie->DictObj = pdfCookie->objCount;
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<< /Length %ld\n\
>>\n\
stream\n", pdfCookie->objCount++, dictSize); 
  u->fwrite (dict, dictSize, 1, u->fp);
    pdfCookie->offset += dictSize; 
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");
    }
    else
    {
        pdfCookie->DictObj = 0;
    }

    if (pdfCookie->ProcObj == 0)
    {
    /* Object Type */
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset,
        pdfCookie, dgp_state));
    /* add this object to the list of dicts, because it is a shared object */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pdfCookie->dictList, (void*) pdfCookie->objCount, NULL));

    pdfCookie->ProcObj = pdfCookie->objCount;
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
[\n\
/PDF /ImageB\n\
]\n\
endobj\n\n", pdfCookie->objCount++);
    }

  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/* Sort objects in xref table. */
static int 
sort_by_obj(const void *a, const void *b)
{
  const xrefElt *ea = a;
  const xrefElt *eb = b;

  return (ea->obj - eb->obj);
}


void
write_pdf_end(PdfCookie *pdfCookie, uds *u) 
{
	Dgp_UInt32 xrefPos = pdfCookie->offset;
	Dgp_UInt32 i, objgap, j;

	/* Sort xref objects. */ 
	qsort(pdfCookie->xref, pdfCookie->xrefPos, sizeof(xrefElt), sort_by_obj);
	/* Write xref table & trailer. */
	pdfCookie->offset += fprintf(u->fp, "\
xref\n\
0 %ld\n\
0000000000 65535 f \n", pdfCookie->xref[pdfCookie->xrefPos-1].obj + 1);


	for (i = 0; i < pdfCookie->xrefPos; i++) {
        if(i > 0){
            objgap = pdfCookie->xref[i].obj - pdfCookie->xref[i-1].obj;
            for(j = 1; j< objgap; j++){
                	pdfCookie->offset += fprintf(u->fp, "0000000000 00000 f \n");
            }

        }
		pdfCookie->offset += fprintf(u->fp, "%010ld 00000 n \n", pdfCookie->xref[i].offset);
        
	}

	pdfCookie->offset += fprintf(u->fp, "\
trailer\n\
<<\n\
/Size %ld\n\
/Root %ld 0 R\n\
>>\n\
startxref\n\
%ld\n\
%%%%EOF\n",  pdfCookie->xref[pdfCookie->xrefPos-1].obj + 1, pdfCookie->catalogObj, xrefPos);

    return;
}
