
#include <stdio.h>

#include "dgp_state.h"
#include "dgp_err_control.h"
#include "PsgFix.h"

#include "pdf_util.h"

static Dgp_UInt32 startPos (PdfCookie *pdfCookie, Dgp_UInt32 Obj);
static Dgp_UInt32 endPos (PdfCookie *pdfCookie, Dgp_UInt32 Obj);
static void copyFile (size_t start_pos, size_t end_pos,
               Dgp_Int8* buffer, Dgp_UInt32 buffer_size, Dgp_UInt32* objMap,
               FILE* in, FILE* out);
static Dgp_UInt32* remapObjects (PdfCookie* pdfCookie, Dgp_State *dgp_state);
static void createHintStream (Dgp_State* dgp_state, Dgp_UInt32* fPXref, Dgp_UInt32 fPXref_len, Dgp_UInt32* mXref, Dgp_UInt32 mXref_len, 
                              PdfCookie *pdfCookie, Dgp_UInt32* objMap, Dgp_Int8* buffer, Dgp_UInt32* buffer_size);
static Dgp_UInt32 log2 (Dgp_UInt32 val);
static Dgp_Bool PB (Dgp_UInt32 w, Dgp_Int8 n, Dgp_UInt32* outWord, Dgp_Int8* outMod, 
	 Dgp_Int8* out_stream);

ErrCode
linearize_pdf_file(PdfCookie *pdfCookie, FILE* in, FILE* out, Dgp_State *dgp_state) 
{
    DGP_PROLOGUE ("linearize_pdf_file", DGP_PERRSTATE);
    Dgp_Int8 *buffer;
    Dgp_UInt32 buf_size = 16384, ret_size;
    Dgp_UInt32 page;
    size_t next_pos, size, end_pos;
    size_t end_firstPage, mXref_fpos;
    Dgp_UInt32* objectMap;
    Dgp_UInt32 firstPageXref[10], *mainXref;
    Dgp_UInt32 fXref = 1, mXref = 1;
    Dgp_UInt32 i, mXref_pos, fXref_pos;
    Dgp_UInt32 hint_len, lin_size;
    pageInfo *info;

    objectMap = remapObjects (pdfCookie, dgp_state);
    DGP_ERR_GEN(!objectMap, 1);

    DGP_CHECK_ALLOC(mainXref = DGP_MALLOC (pdfCookie->objCount * sizeof (Dgp_UInt32)));
    DGP_CHECK_ALLOC(buffer = DGP_MALLOC (buf_size));

    fprintf(out, "%%PDF-1.4\n%%%c%c%c%c\n\n", 250, 251, 252, 253);
    
    /* get the file into a well defined state */
    fflush (in);
    end_pos = ftell (in);

    /* go to the beginning of the file */
    fseek (in, 0, SEEK_SET);

    /* generate space for linearized object */
    firstPageXref[fXref++] = ftell (out);
    info = (pageInfo*) pdfCookie->pageObjList->first->userdata;
    lin_size = fprintf (out, "%d 0 obj\n\
<< /Linearized 1.0\n\
/L %d \n\
/H [10000 500]\n\
/O %d\n\
/E 10000 \n\
/N %d\n\
/T %d  >>\n\
endobj\n  \n", 
        objectMap[pdfCookie->catalogObj] - 1,
        end_pos,
        objectMap[info->pageObj],
        pdfCookie->pageObjList->nelt,
        end_pos);

    /* generate space for 1st page XREF */
    fXref_pos = ftell (out);
    memset (buffer, 32, 256);
    buffer[255] = '\n';
    fwrite(buffer, 1, 256, out);

    /* copy catalog */
    next_pos = startPos (pdfCookie, pdfCookie->catalogObj);
    size = 1024;
    fseek (in, next_pos, SEEK_SET);
    size = fread (buffer, 1, size, in);
    firstPageXref[fXref++] = ftell (out);
    copyFile (next_pos, next_pos + size, buffer, buf_size, objectMap, in, out);

    /* copy 3 page objects for first page */
    if (pdfCookie->pageObjList->nelt >= 1)
    {
        info = (pageInfo*) pdfCookie->pageObjList->first->userdata;
        page = info->pageObj;
        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page++);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page++);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
    }

    /* copy dictionary and ProcSet */
    if (pdfCookie->dictList->nelt > 1)
    {
        page = (Dgp_UInt32) pdfCookie->dictList->first->userdata;
        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page++);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
    }

    /* estimate size of the hint stream */
    firstPageXref[fXref++] = ftell (out);
    fprintf (out, "%d 0 obj\nendobj\n\n", 
        pdfCookie->objCount + 1);
    hint_len = 130 + 10 * pdfCookie->pageObjList->nelt + 
        5 * pdfCookie->dictList->nelt;
    memset (buffer, 32, hint_len);
    buffer[hint_len - 1] = '\n';
    fwrite(buffer, 1, hint_len, out);
    end_firstPage = ftell (out);

    /* All other pages */

    /* copy 3 page objects for all other pages page */
    if ( pdfCookie->pageObjList->nelt >= 2)
    {
        DgpAdt_ListNode ln = pdfCookie->pageObjList->first->next;
        while (ln)
        {
            info = (pageInfo*) ln->userdata;
            page = info->pageObj;
            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page++);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page++);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
            
            ln = ln->next;
        }
    }
    
    /* copy other dictionaries */
    if ( pdfCookie->dictList->nelt >= 3)
    {
        DgpAdt_ListNode ln = pdfCookie->dictList->first->next->next;
        while (ln)
        {
            page = (Dgp_UInt32) ln->userdata;
            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
            
            ln = ln->next;
        }
    }

    /* copy page tree */
    next_pos = startPos (pdfCookie, pdfCookie->PagesObj);
    end_pos = endPos (pdfCookie, pdfCookie->PagesObj);
    mainXref[mXref++] = ftell (out);
    copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

    /* create main XRef */
    mXref_pos = ftell (out);
    fprintf (out, "xref\n0 %d\n", mXref);
    mXref_fpos = ftell (out);
    fprintf (out, "0000000000 65535 f \n", mXref);
    for (i = 1; i < mXref; i++)
        fprintf (out, "%010ld 00000 n \n", mainXref[i]);
    fprintf (out, "trailer\n<< /Size %d >>\nstartxref\n%d\n%%%%EOF\n", mXref, fXref_pos);

    /* create first page XRef */
    fflush (out);
    end_pos = ftell (out);
    fseek (out, fXref_pos, SEEK_SET);
    fprintf (out, "xref\n%d %d\n", mXref, fXref - 1);
    for (i = 1; i < fXref; i++)
        fprintf (out, "%010ld 00000 n \n", firstPageXref[i]);
    fprintf (out, "trailer\n<< /Size %d\n/Prev %d\n/Root %d 0 R >>\nstartxref\n0\n%%%%EOF\n", 
        mXref + fXref - 1, mXref_pos, objectMap[pdfCookie->catalogObj]);

    /* create the hint stream */
    ret_size = buf_size;
    createHintStream (dgp_state, firstPageXref, fXref, mainXref, mXref, pdfCookie, 
        objectMap, buffer, &ret_size);
    fseek (out, firstPageXref[fXref-1], SEEK_SET);
    if (ret_size <= hint_len) 
        fwrite (buffer, ret_size, 1, out);
    else
        DGP_ERR_PRINTERF1("not enough space for hint tables -- %d bytes missing\n",
        ret_size - hint_len);

    /* update the linearized object */
    fseek (out, firstPageXref[1], SEEK_SET);
    info = (pageInfo*) pdfCookie->pageObjList->first->userdata;
    next_pos = fprintf (out, "%d 0 obj\n\
<< /Linearized 1.0\n\
/L %d\n\
/H [%d %d]\n\
/O %d\n\
/E %d\n\
/N %d\n\
/T %d >>\n\
endobj\n\n", 
        objectMap[pdfCookie->catalogObj] - 1,
        end_pos, firstPageXref[fXref - 1], hint_len,
        objectMap[info->pageObj],
        end_firstPage,
        pdfCookie->pageObjList->nelt,
        mXref_fpos);
    while (lin_size > next_pos + 1)
    {
        fputc (' ', out);
        next_pos++;
    }

    fclose (in);
    DGP_FREE (objectMap);
    DGP_FREE (mainXref);
    DGP_FREE (buffer);
    
    NORMAL_RETURN;
}

static Dgp_UInt32 startPos (PdfCookie *pdfCookie, Dgp_UInt32 Obj)
{
    Dgp_UInt32 i;
    size_t first_pos = 0;

    for (i = 0; i < pdfCookie->xrefPos; i++)
    {
        if (pdfCookie->xref[i].obj == Obj)
        {
            first_pos = pdfCookie->xref[i].offset;
            break;
        }
    }

    return first_pos;
}

static Dgp_UInt32 endPos (PdfCookie *pdfCookie, Dgp_UInt32 Obj)
{
    Dgp_UInt32 i;
    size_t size = 0;

    for (i = 0; i < pdfCookie->xrefPos; i++)
    {
        if (pdfCookie->xref[i].obj == Obj)
        {
            if (i+1 < pdfCookie->xrefPos) 
            {
                size = pdfCookie->xref[i+1].offset;
                break;
            }
        }
    }

    return size;
}

static void copyFile (size_t start_pos, size_t end_pos,
               Dgp_Int8* buffer, Dgp_UInt32 buffer_size, Dgp_UInt32* objMap,
               FILE* in, FILE* out)
{
    Dgp_Bool check = DGP_TRUE;
    size_t temp = 1, size = end_pos - start_pos, size_used = buffer_size;

    fseek (in, start_pos, SEEK_SET);

    while (size > 0 && temp > 0)
    {
        if (buffer_size > size) size_used = size;
        temp = fread (buffer, 1, size_used, in);
        if (check)
        {
            Dgp_Int8 *bufptr, *stream, *nextpos, *endpos;
            Dgp_UInt32 objnr = atoi (buffer);
    
            /* set the end marker */
            /* terminate the buffer to limit the search */
            if (temp < buffer_size) buffer[temp] = '\0';
            endpos = buffer + temp;

            /* renumber the object */
            if (objnr > 0 && objnr < 100000)
                    fprintf (out, "%d 0 obj", objMap[objnr]);
            bufptr = strstr (buffer, "obj");
            bufptr += 3;

            /* look for a stream, and terminate the buffer there */
            stream = strstr (bufptr, "stream");
            if (stream)
            {
                *stream = '\0';
                check = DGP_FALSE;
            }

            nextpos = strstr (bufptr, " 0 R");
            while (nextpos)
            {
                Dgp_Int8* pChar = nextpos;

                /* go to the beginning of this reference */
                while (*pChar == ' ' || *pChar == '\n' || *pChar == '\r')
                    pChar--;
                while (*pChar != ' ' && *pChar != '\n' && *pChar != '\r')
                    pChar--;
                pChar++;
                objnr = atoi (pChar);

                fwrite (bufptr, 1, pChar - bufptr, out);
                fprintf (out, "%d 0 R", objMap[objnr]);

                bufptr = nextpos + 4;
                nextpos = strstr (bufptr, " 0 R");
            }

            if (stream) *stream = 's';
            fwrite (bufptr, 1, endpos - bufptr, out);
        } 
        else
            fwrite (buffer, 1, temp, out);

        size -= temp;
    }
    return;
}

static Dgp_UInt32* remapObjects (PdfCookie* pdfCookie, Dgp_State *dgp_state)
{
    DgpAdt_ListNode ln;
    Dgp_UInt32 index = 1 ;
    Dgp_UInt32 first;
    Dgp_UInt32* map = NULL;
    pageInfo *info;

    if (pdfCookie->pageObjList->nelt == 0) return NULL;
    map = DGP_MALLOC (pdfCookie->objCount * sizeof (Dgp_UInt32));

    /* 3 objects per page */
    if (pdfCookie->pageObjList->nelt > 1)
    {
        ln = pdfCookie->pageObjList->first->next; 
        while (ln)
        {
            info = (pageInfo*) ln->userdata;
            first = info->pageObj;
            map[first++] = index++;
            map[first++] = index++;
            map[first] = index++;
            ln = ln->next;
        }
    }
    
    /* all dicts after the first and proc set */
    if (pdfCookie->dictList->nelt > 2)
    {
        ln = pdfCookie->dictList->first->next->next; 
        while (ln)
        {
            first = (Dgp_UInt32) ln->userdata;
            map[first] = index++;
            ln = ln->next;
        }
    }

    /* page tree object */
    map[pdfCookie->PagesObj] = index++;

    /**** first page objects */
    /* linerize object */
    index++;

    /* catalog */
    map[pdfCookie->catalogObj] = index++;

    /* 3 objects for the first page */
    ln = pdfCookie->pageObjList->first; 
    info = (pageInfo*) ln->userdata;
    first = info->pageObj;
    map[first++] = index++;
    map[first++] = index++;
    map[first] = index++;

    /* first page dict plus proc set */
    ln = pdfCookie->dictList->first;     
    first = (Dgp_UInt32) ln->userdata;
    map[first++] = index++;
    if (pdfCookie->dictList->nelt > 1)
        map[first++] = index++;

    return map;
}

static void createHintStream (Dgp_State *dgp_state, Dgp_UInt32* fPXref, Dgp_UInt32 fPXref_len, Dgp_UInt32* mXref, Dgp_UInt32 mXref_len, 
                              PdfCookie *pdfCookie, Dgp_UInt32* objMap, Dgp_Int8* buffer, Dgp_UInt32* buffer_size)
{
    Dgp_UInt32 tmp_long, tmp_buffer = 0;
    Dgp_UInt32 *dictSize, *dictNr, *pageSize, i, maxSize, minSize;
    DgpAdt_ListNode ln;
    Dgp_Int8 tmp_mod = 0, *pBuf = buffer;
    Dgp_Int8 *shared_buf, *main_buf, temp_char;
    Dgp_Int32 shared_size, main_size;
    DgpAdt_List dictList = pdfCookie->dictList;
    DgpAdt_List pageList = pdfCookie->pageObjList;
    Dgp_Int8 size3, size5, size10, size11, size12;

    /* get the sizes and object numbers of all dictionary objects */
    dictSize = DGP_MALLOC (4 * (dictList->nelt + 4));
    dictNr = DGP_MALLOC (4 * (fPXref_len + mXref_len));
    memset (dictNr, 0, 4 * (fPXref_len + mXref_len));

    /* linearized object */
    dictSize[0] = fPXref[2] - fPXref[1];
    /* catalog object */
    dictSize[1] = fPXref[3] - fPXref[2];
    /* first page objects */
    dictSize[2] = fPXref[6] - fPXref[3];
    /* first page dicts */
    /* symbol dict */
    ln = dictList->first;
    if (ln)
    {
        dictSize[3] = fPXref[7] - fPXref[6];
        dictNr[objMap[(Dgp_UInt32) ln->userdata]] = 3;

        /* proc set object */
        ln = ln->next;
        if (ln)
        {
            dictSize[4] = fPXref[8] - fPXref[7];
            dictNr[objMap[(Dgp_UInt32) ln->userdata]] = 4;
    
            /* hist stream */
            dictSize[5] = mXref[1] - fPXref[8];
        }
        else
            dictSize[4] = mXref[1] - fPXref[8];
    }
    else
        dictSize[3] = mXref[1] - fPXref[8];

        
    /* rest of the dictionaries */
    if (ln) ln = ln->next;
    for (i = 6; ln != NULL; ln = ln->next, i++)
    {
        tmp_long = objMap[(Dgp_UInt32) ln->userdata];
        dictSize[i] = mXref[tmp_long + 1] - mXref[tmp_long];
        dictNr[tmp_long] = i;
    }
    minSize = mXref[mXref_len - 1];
    maxSize = 0;
    for (i = 0; i < dictList->nelt + 4; i++)
    {
        if (dictSize[i] < minSize) minSize = dictSize[i];
        if (dictSize[i] > maxSize) maxSize = dictSize[i];
    }

    /* generate the sharedTable header */
    /* 1  objNr of first shared object */
    if (dictList->nelt > 2)
        tmp_long = objMap[(Dgp_UInt32) dictList->first->next->next->userdata];
    else
        tmp_long = 0;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 2  position of first shared object */
    tmp_long = mXref[tmp_long];
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 3  first page shared objects */
    tmp_long = 6;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 4 total number of shared objects */
    tmp_long = dictList->nelt + 4;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 5 unused */
    tmp_long = 0;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 6 least length of a shared object in bytes */
    tmp_long = minSize;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 7 number of bits needed to represent the difference between (6) and the largest object */
    tmp_long = log2(maxSize - minSize);
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* write the dictionary sizes */
    temp_char = (Dgp_Int8) tmp_long;
    for (i = 0; i < dictList->nelt + 4; i++)
    {
        if (PB ((Dgp_UInt32) dictSize[i] - minSize, temp_char, &tmp_buffer, &tmp_mod, pBuf)) 
            pBuf += 4;
    }
    /* write the 0 bit */
    for (i = 0; i < dictList->nelt + 4; i++)
    {
        if (PB ((Dgp_UInt32)0, 1, &tmp_buffer, &tmp_mod, pBuf)) 
            pBuf += 4;
    }
    /* flush any remaining bytes */
    shared_size = pBuf - buffer;
    if (tmp_mod)
    {
        Dgp_Int8 store = tmp_mod;
        PB ((Dgp_UInt32)0, (Dgp_Int8) (32 - tmp_mod), &tmp_buffer, &tmp_mod, pBuf);
        shared_size += (store + 7) / 8;
    }

    /* copy the shared buffer to a temp space */
    shared_buf = DGP_MALLOC (shared_size);
    memcpy (shared_buf, buffer, shared_size);
/*    for (i = 0; i < shared_size; i++, shared_buf++)
        printf ("%03d  %02x\n", i, (unsigned char) *shared_buf); */
    
     /* reset the temporary variables */
    pBuf = buffer;
    tmp_buffer = 0;
    tmp_mod = 0;

    /* get the page sizes */
    pageSize = DGP_MALLOC (4 * pageList->nelt);
    pageSize[0] = fPXref[6] - fPXref[3];
    for (i = 1; i < pageList->nelt; i++)
    {
        pageSize[i] = mXref[3*i + 1] - mXref[3*i - 2];
    }
    minSize = mXref[mXref_len - 1];
    maxSize = 0;
    for (i = 0; i < pageList->nelt; i++)
    {
        if (pageSize[i] < minSize) minSize = pageSize[i];
        if (pageSize[i] > maxSize) maxSize = pageSize[i];
    }

    /* generate the Page Offset Table header */
    /* 1  least number of objects in a page */
    tmp_long = 3;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 2  position first Page page object */
    tmp_long = fPXref[3];
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 3  bits needed to encode the difference between most and least objects per page */
    tmp_long = 0;
    size3 = (Dgp_Int8) tmp_long;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 4 least length of page in bytes */
    tmp_long = minSize;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 5 number of bits to represent difference between largest and smallest page */
    tmp_long = log2(maxSize - minSize);
    size5 = (Dgp_Int8) tmp_long;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 6, 7, 8 unused */
    tmp_long = 0;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;
    if (PB (tmp_long, 32, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 9 same as #5 */
    tmp_long = log2(maxSize - minSize);
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 10 bits needed to represent greatest number of shared obj references */
    tmp_long = 2;
    size10 = (Dgp_Int8) tmp_long;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 11 bits needed to represent the number of total shared objects */
    tmp_long = log2(dictList->nelt + 1);
    size11 = (Dgp_Int8) tmp_long;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 12 bits needed to represent the fractional position */
    tmp_long = 1;
    size12 = (Dgp_Int8) tmp_long;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* 13 denominator for the fractional position */
    tmp_long = 2;
    if (PB (tmp_long, 16, &tmp_buffer, &tmp_mod, pBuf)) pBuf += 4;

    /* write the number of objects per page */
    if (size3 > 0)
    {
        for (i = 0; i < pageList->nelt; i++)
        {
            if (PB ((Dgp_UInt32) 0, size3, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
        }
    }

    /* write the length of the page */
    if (size5 > 0)
    {
        for (i = 0; i < pageList->nelt; i++)
        {
            if (PB (pageSize[i], size5, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
        }
    }

    /* write the number of shared objects per page */
    if (size10 > 0)
    {
        if (PB ((Dgp_UInt32) 0, size10, &tmp_buffer, &tmp_mod, pBuf)) 
            pBuf += 4;
        for (i = 1; i < pageList->nelt; i++)
        {
            if (PB ((Dgp_UInt32) 2, size10, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
        }
    }

    /* write the shared object id */
    if (size11 > 0)
    {
        for (i = 1; i < pageList->nelt; i++)
        {
            if (PB (dictNr[objMap[pdfCookie->ProcObj]], size11, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
            if (PB (dictNr[i], size11, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
        }
    }

    /* write the shared object position */
    if (size12 > 0)
    {
        for (i = 1; i < pageList->nelt; i++)
        {
            if (PB ((Dgp_UInt32) 0, size12, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
            if (PB ((Dgp_UInt32) 0, size12, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
        }
    }

    /* write the length of the page again */
    if (size5 > 0)
    {
        for (i = 0; i < pageList->nelt; i++)
        {
            if (PB (pageSize[i], size5, &tmp_buffer, &tmp_mod, pBuf)) 
                pBuf += 4;
        }
    }
    /* flush any remaining bytes */
    main_size = pBuf - buffer;
    if (tmp_mod)
    {
        Dgp_Int8 store = tmp_mod;
        PB ((Dgp_UInt32)0, (Dgp_Int8) (32 - tmp_mod), &tmp_buffer, &tmp_mod, pBuf);
        tmp_long += (store + 7) / 8;
    }
    /* copy the shared buffer to a temp space */
    main_buf = DGP_MALLOC (main_size);
    memcpy (main_buf, buffer, main_size);

    /* create the object */
    *buffer_size = sprintf (buffer, "%d 0 obj\n<< /Length %d\n/S %d >>\nstream\n", 
        pdfCookie->objCount + 1, main_size + shared_size, main_size);
    pBuf = buffer + *buffer_size;
    memcpy (pBuf, main_buf, main_size);
    pBuf += main_size;
    memcpy (pBuf, shared_buf, shared_size);
    pBuf += shared_size;
    *buffer_size += main_size + shared_size;
    *buffer_size += sprintf (pBuf, "\nendstream\nendobj\n\n");

    DGP_FREE (dictSize);
    DGP_FREE (pageSize);
    DGP_FREE (dictNr);
    DGP_FREE (shared_buf);
    DGP_FREE (main_buf);

    return;
}

static Dgp_UInt32 log2 (Dgp_UInt32 val)
{
    Dgp_UInt32 mask = 0x80000000;
    Dgp_UInt32 count = 32;

    if (val == 0) return 0;

    while (mask > val)
    {
        count--;
        mask = mask >> 1;
    }

    return count;
}

/* Takes n bits, right-justified in w, and puts them into the output bit
   stream. outWord contains a unfinished word, and outMod the number of bit
   that are already stored in this word. These are state variables that the
   calling program must maintain */
static Dgp_Bool PB (Dgp_UInt32 w, Dgp_Int8 n, Dgp_UInt32* outWord, Dgp_Int8* outMod, 
	 Dgp_Int8* out_stream)
{
  const Dgp_Int8 bpw = 32;
  Dgp_Int8 newOutMod = (n + *outMod) % bpw;
  Dgp_Bool ret_val = DGP_FALSE;

  if (n == 0) return ret_val;
  if(newOutMod > *outMod) 
    {
      /* Just storing a partial field */					
      w = w << ((Dgp_Int8)(bpw - newOutMod));					
      *outWord |=  w;
    } 
  else 
    {
      /* Crossing a word boundary */
      Dgp_UInt32 temp;

      ret_val = DGP_TRUE;
      if(newOutMod == 0) 
      {
          /* Filling a word exactly */					
          temp = *outWord | w;						
          *outWord = 0;							
      } 
      else 
      {
          /* There are bits left over for the next word */			
          temp = *outWord | (w >> newOutMod);
          *outWord = w << ((Dgp_Int8)(bpw - newOutMod));
      }
      
      /* copy the new word to the output stream */
      *(out_stream++) = (Dgp_Int8)((temp >> 24) & 0xFF);
      *(out_stream++) = (Dgp_Int8)((temp >> 16) & 0xFF);
      *(out_stream++) = (Dgp_Int8)((temp >> 8) & 0xFF);
      *(out_stream++) = (Dgp_Int8)(temp & 0xFF);
  }									
  *outMod = newOutMod;
  return ret_val;
}
