/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1996-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef JPEGINT_H_INCLUDED
#define JPEGINT_H_INCLUDED

#ifdef _MSC_VER
#pragma pack(push, jpegint_h, 1)
#endif

/*
 * jpegint.h
 *
 * Copyright (C) 1991-1995, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file provides common declarations for the various JPEG modules.
 * These declarations are considered internal to the JPEG library; most
 * applications using the library shouldn't need to include this file.
 */

IP_RCSINFO(jpegint_h_RCSInfo, "$RCSfile: jpegint.h,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:20 $");

/* Declarations for both compression & decompression */

typedef enum {			/* Operating modes for buffer controllers */
	JBUF_PASS_THRU,		/* Plain stripwise operation */
	/* Remaining modes require a full-image buffer to have been created */
	JBUF_SAVE_SOURCE,	/* Run source subobject only, save output */
	JBUF_CRANK_DEST,	/* Run dest subobject only, using saved data */
	JBUF_SAVE_AND_PASS	/* Run both subobjects, save output */
} J_BUF_MODE;

/* Values of global_state field (jdapi.c has some dependencies on ordering!) */
#define CSTATE_START	100	/* after create_compress */
#define CSTATE_SCANNING	101	/* start_compress done, write_scanlines OK */
#define CSTATE_RAW_OK	102	/* start_compress done, write_raw_data OK */
#define CSTATE_WRCOEFS	103	/* jpeg_write_coefficients done */
#define DSTATE_START	200	/* after create_decompress */
#define DSTATE_INHEADER	201	/* reading header markers, no SOS yet */
#define DSTATE_READY	202	/* found SOS, ready for start_decompress */
#define DSTATE_PRELOAD	203	/* reading multiscan file in start_decompress*/
#define DSTATE_PRESCAN	204	/* performing dummy pass for 2-pass quant */
#define DSTATE_SCANNING	205	/* start_decompress done, read_scanlines OK */
#define DSTATE_RAW_OK	206	/* start_decompress done, read_raw_data OK */
#define DSTATE_BUFIMAGE	207	/* expecting jpeg_start_output */
#define DSTATE_BUFPOST	208	/* looking for SOS/EOI in jpeg_finish_output */
#define DSTATE_RDCOEFS	209	/* reading file in jpeg_read_coefficients */
#define DSTATE_STOPPING	210	/* looking for EOI in jpeg_finish_decompress */


/* Declarations for compression modules */

/* Master control module */
struct jpeg_comp_master {
  ErrCode (CDECL *prepare_for_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *pass_startup) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *finish_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);

  /* State variables made visible to other modules */
  boolean call_pass_startup;	/* True if pass_startup must be called */
  boolean is_last_pass;		/* True during last pass */
};

/* Main buffer control (downsampled-data buffer) */
struct jpeg_c_main_controller {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	J_BUF_MODE  pass_mode);
  ErrCode (CDECL *process_data) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	JSAMPARRAY  input_buf,
	JDIMENSION *in_row_ctr,
	JDIMENSION  in_rows_avail);
};

/* Compression preprocessing (downsampling input buffer control) */
struct jpeg_c_prep_controller {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	J_BUF_MODE  pass_mode);
  ErrCode (CDECL *pre_process_data) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	JSAMPARRAY  input_buf,
	JDIMENSION *in_row_ctr,
	JDIMENSION  in_rows_avail,
	JSAMPIMAGE  output_buf,
	JDIMENSION *out_row_group_ctr,
	JDIMENSION  out_row_groups_avail);
};

/* Coefficient buffer control */
struct jpeg_c_coef_controller {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	J_BUF_MODE  pass_mode);
  ErrCode (CDECL *compress_data) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	JSAMPIMAGE  input_buf,
	boolean	   *no_suspend);
};

/* Colorspace conversion */
struct jpeg_color_converter {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *color_convert) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	JSAMPARRAY  input_buf,
	JSAMPIMAGE  output_buf,
	JDIMENSION  output_row,
	int	    num_rows);
};

/* Downsampling */
struct jpeg_downsampler {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *downsample) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	JSAMPIMAGE  input_buf,
	JDIMENSION  in_row_index,
	JSAMPIMAGE  output_buf,
	JDIMENSION  out_row_group_index);

  boolean need_context_rows;	/* TRUE if need rows above & below */
};

/* Forward DCT (also controls coefficient quantization) */
struct jpeg_forward_dct {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  /* perhaps this should be an array??? */
  ErrCode (CDECL *forward_DCT) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	jpeg_component_info *compptr,
	JSAMPARRAY  sample_data,
	JBLOCKROW   coef_blocks,
	JDIMENSION  start_row,
	JDIMENSION  start_col,
	JDIMENSION  num_blocks);
};

/* Entropy encoding */
struct jpeg_entropy_encoder {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	boolean	    gather_statistics);
  ErrCode (CDECL *encode_mcu) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	JBLOCKROW  *MCU_data,
	boolean	   *no_suspend);
  ErrCode (CDECL *finish_pass) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
};

/* Marker writing */
struct jpeg_marker_writer {
  /* write_any_marker is exported for use by applications */
  /* Probably only COM and APPn markers should be written */
  ErrCode (CDECL *write_any_marker) (
	IPState	   *pIPState,
	j_compress_ptr cinfo,
	int	    marker,
	const JOCTET *dataptr,
	unsigned int datalen);
  ErrCode (CDECL *write_file_header) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *write_frame_header) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *write_scan_header) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *write_file_trailer) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
  ErrCode (CDECL *write_tables_only) (
	IPState	   *pIPState,
	j_compress_ptr cinfo);
};


/* Declarations for decompression modules */

/* Master control module */
struct jpeg_decomp_master {
  ErrCode (CDECL *prepare_for_output_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *finish_output_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);

  /* State variables made visible to other modules */
  boolean is_dummy_pass;	/* True during 1st pass for 2-pass quant */
};

/* Input control module */
struct jpeg_input_controller {
  ErrCode (CDECL *consume_input) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	int	   *retcode);
  ErrCode (CDECL *reset_input_controller) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *start_input_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *finish_input_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);

  /* State variables made visible to other modules */
  boolean has_multiple_scans;	/* True if file has multiple scans */
  boolean eoi_reached;		/* True when EOI has been consumed */
};

/* Main buffer control (downsampled-data buffer) */
struct jpeg_d_main_controller {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	J_BUF_MODE  pass_mode);
  ErrCode (CDECL *process_data) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	JSAMPARRAY  output_buf,
	JDIMENSION *out_row_ctr,
	JDIMENSION  out_rows_avail);
};

/* Coefficient buffer control */
struct jpeg_d_coef_controller {
  ErrCode (CDECL *start_input_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *consume_data) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	int	   *retcode);
  ErrCode (CDECL *start_output_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *decompress_data) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	JSAMPIMAGE  output_buf,
	int	   *retcode);
  /* Pointer to array of coefficient virtual arrays, or NULL if none */
  jvirt_barray_ptr *coef_arrays;
};

/* Decompression postprocessing (color quantization buffer control) */
struct jpeg_d_post_controller {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	J_BUF_MODE  pass_mode);
  ErrCode (CDECL *post_process_data) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	JSAMPIMAGE  input_buf,
	JDIMENSION *in_row_group_ctr,
	JDIMENSION  in_row_groups_avail,
	JSAMPARRAY  output_buf,
	JDIMENSION *out_row_ctr,
	JDIMENSION  out_rows_avail);
};

/* Marker reading & parsing */
struct jpeg_marker_reader {
  ErrCode (CDECL *reset_marker_reader) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  /* Read markers until SOS or EOI.
   * Returns same codes as are defined for jpeg_consume_input:
   * JPEG_SUSPENDED, JPEG_REACHED_SOS, or JPEG_REACHED_EOI.
   */
  ErrCode (CDECL *read_markers) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	int	   *recode);
  /* Read a restart marker --- exported for use by entropy decoder only */
  jpeg_marker_parser_method read_restart_marker;
  /* Application-overridable marker processing methods */
  jpeg_marker_parser_method process_COM;
  jpeg_marker_parser_method process_APPn[16];

  /* State of marker reader --- nominally internal, but applications
   * supplying COM or APPn handlers might like to know the state.
   */
  boolean saw_SOI;		/* found SOI? */
  boolean saw_SOF;		/* found SOF? */
  int next_restart_num;		/* next restart number expected (0-7) */
  unsigned int discarded_bytes;	/* # of bytes skipped looking for a marker */
};

/* Entropy decoding */
struct jpeg_entropy_decoder {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *decode_mcu) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	JBLOCKROW  *MCU_data,
	boolean	   *no_suspend);
};

/* Inverse DCT (also performs dequantization) */
typedef ErrCode (CDECL *inverse_DCT_method_ptr) (
    IPState    *pIPState,
    JSAMPLE    *range_limit,
    jpeg_component_info *compptr,
    JCOEFPTR	coef_block,
    JSAMPARRAY	output_buf,
    JDIMENSION	output_col);

struct jpeg_inverse_dct {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  /* It is useful to allow each component to have a separate IDCT method. */
  inverse_DCT_method_ptr inverse_DCT[MAX_COMPONENTS];
};

/* Upsampling (note that upsampler must also call color converter) */
struct jpeg_upsampler {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *upsample) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	JSAMPIMAGE  input_buf,
	JDIMENSION *in_row_group_ctr,
	JDIMENSION  in_row_groups_avail,
	JSAMPARRAY  output_buf,
	JDIMENSION *out_row_ctr,
	JDIMENSION  out_rows_avail);

  boolean need_context_rows;	/* TRUE if need rows above & below */
};

/* Colorspace conversion */
struct jpeg_color_deconverter {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *color_convert) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	JSAMPIMAGE  input_buf,
	JDIMENSION  input_row,
	JSAMPARRAY  output_buf,
	int	    num_rows);
};

/* Color quantization or color precision reduction */
struct jpeg_color_quantizer {
  ErrCode (CDECL *start_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	boolean	    is_pre_scan);
  ErrCode (CDECL *color_quantize) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo,
	JSAMPARRAY  input_buf,
	JSAMPARRAY  output_buf,
	int	    num_rows);
  ErrCode (CDECL *finish_pass) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
  ErrCode (CDECL *new_color_map) (
	IPState	   *pIPState,
	j_decompress_ptr cinfo);
};


/* Miscellaneous useful macros */

#undef MAX
#define MAX(a,b)	((a) > (b) ? (a) : (b))
#undef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))


/* We assume that right shift corresponds to signed division by 2 with
 * rounding towards minus infinity.  This is correct for typical "arithmetic
 * shift" instructions that shift in copies of the sign bit.  But some
 * C compilers implement >> with an unsigned shift.  For these machines you
 * must define RIGHT_SHIFT_IS_UNSIGNED.
 * RIGHT_SHIFT provides a proper signed right shift of an INT32 quantity.
 * It is only applied with constant shift counts.  SHIFT_TEMPS must be
 * included in the variables of any routine using RIGHT_SHIFT.
 */

#ifdef RIGHT_SHIFT_IS_UNSIGNED
#define SHIFT_TEMPS	INT32 shift_temp;
#define RIGHT_SHIFT(x,shft)  \
	((shift_temp = (x)) < 0 ? \
	 (shift_temp >> (shft)) | ((~((INT32) 0)) << (32-(shft))) : \
	 (shift_temp >> (shft)))
#else
#define SHIFT_TEMPS
#define RIGHT_SHIFT(x,shft)	((x) >> (shft))
#endif


/* Short forms of external names for systems with brain-damaged linkers. */

#ifdef NEED_SHORT_EXTERNAL_NAMES
#define jinit_compress_master	jICompress
#define jinit_c_master_control	jICMaster
#define jinit_c_main_controller	jICMainC
#define jinit_c_prep_controller	jICPrepC
#define jinit_c_coef_controller	jICCoefC
#define jinit_color_converter	jICColor
#define jinit_downsampler	jIDownsampler
#define jinit_forward_dct	jIFDCT
#define jinit_huff_encoder	jIHEncoder
#define jinit_phuff_encoder	jIPHEncoder
#define jinit_marker_writer	jIMWriter
#define jinit_master_decompress	jIDMaster
#define jinit_d_main_controller	jIDMainC
#define jinit_d_coef_controller	jIDCoefC
#define jinit_d_post_controller	jIDPostC
#define jinit_input_controller	jIInCtlr
#define jinit_marker_reader	jIMReader
#define jinit_huff_decoder	jIHDecoder
#define jinit_phuff_decoder	jIPHDecoder
#define jinit_inverse_dct	jIIDCT
#define jinit_upsampler		jIUpsampler
#define jinit_color_deconverter	jIDColor
#define jinit_1pass_quantizer	jI1Quant
#define jinit_2pass_quantizer	jI2Quant
#define jinit_merged_upsampler	jIMUpsampler
#define jinit_memory_mgr	jIMemMgr
#define jdiv_round_up		jDivRound
#define jround_up		jRound
#define jcopy_sample_rows	jCopySamples
#define jcopy_block_row		jCopyBlocks
#define jzero_far		jZeroFar
#define jpeg_zigzag_order	jZIGTable
#define jpeg_natural_order	jZAGTable
#endif /* NEED_SHORT_EXTERNAL_NAMES */


/* Compression module initialization routines */
EXTERN ErrCode CDECL 
jinit_compress_master (
    IPState    *pIPState,
    j_compress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_c_master_control (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean	transcode_only);
EXTERN ErrCode CDECL 
jinit_c_main_controller (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean	need_full_buffer);
EXTERN ErrCode CDECL 
jinit_c_prep_controller (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean	need_full_buffer);
EXTERN ErrCode CDECL 
jinit_c_coef_controller (
    IPState    *pIPState,
    j_compress_ptr cinfo,
    boolean	need_full_buffer);
EXTERN ErrCode CDECL 
jinit_color_converter (
    IPState    *pIPState,
    j_compress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_downsampler (
    IPState    *pIPState,
    j_compress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_forward_dct (
    IPState    *pIPState,
    j_compress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_huff_encoder (
    IPState    *pIPState,
    j_compress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_phuff_encoder (
    IPState    *pIPState,
    j_compress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_marker_writer (
    IPState    *pIPState,
    j_compress_ptr cinfo);
/* Decompression module initialization routines */
EXTERN ErrCode CDECL 
jinit_master_decompress (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_d_main_controller (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    boolean	need_full_buffer);
EXTERN ErrCode CDECL 
jinit_d_coef_controller (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    boolean	need_full_buffer);
EXTERN ErrCode CDECL 
jinit_d_post_controller (
    IPState    *pIPState,
    j_decompress_ptr cinfo,
    boolean	need_full_buffer);
EXTERN ErrCode CDECL 
jinit_input_controller (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_marker_reader (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_huff_decoder (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_phuff_decoder (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_inverse_dct (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_upsampler (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_color_deconverter (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_1pass_quantizer (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_2pass_quantizer (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
EXTERN ErrCode CDECL 
jinit_merged_upsampler (
    IPState    *pIPState,
    j_decompress_ptr cinfo);
/* Memory manager initialization */
EXTERN ErrCode CDECL 
jinit_memory_mgr (
    IPState    *pIPState,
    j_common_ptr cinfo);

/* Utility routines in jutils.c */
EXTERN long jdiv_round_up JPP((long a, long b));
EXTERN long jround_up JPP((long a, long b));
EXTERN ErrCode CDECL 
jcopy_sample_rows (
    IPState    *pIPState,
    JSAMPARRAY	input_array,
    int		source_row,
    JSAMPARRAY	output_array,
    int		dest_row,
    int		num_rows,
    JDIMENSION	num_cols);
EXTERN ErrCode CDECL 
jcopy_block_row (
    IPState    *pIPState,
    JBLOCKROW	input_row,
    JBLOCKROW	output_row,
    JDIMENSION	num_blocks);
EXTERN ErrCode CDECL 
jzero_far (
    IPState    *pIPState,
    void FAR   *target,
    size_t	bytestozero);
/* Constant tables in jutils.c */
extern const int jpeg_zigzag_order[]; /* natural coef order to zigzag order */
extern const int jpeg_natural_order[]; /* zigzag coef order to natural order */

/* Suppress undefined-structure complaints if necessary. */

#ifdef INCOMPLETE_TYPES_BROKEN
#ifndef AM_MEMORY_MANAGER	/* only jmemmgr.c defines these */
struct jvirt_sarray_control { long dummy; };
struct jvirt_barray_control { long dummy; };
#endif
#endif /* INCOMPLETE_TYPES_BROKEN */

#ifdef _MSC_VER
#pragma pack(pop, jpegint_h)
#endif

#endif /* JPEGINT_H_INCLUDED */

/*
$Log: jpegint.h,v $
Revision 1.1.1.1  1999/09/20 18:18:20  jlin
Imported sources

Revision 1.5  1998/10/19 23:51:15  aber
jpegint.h -> /product/ipcore/ipshared/include/RCS/jpegint.h,v

CommonCheck: Add pragma pack for structure alignment consistency between
client includes and pre-build libraries.

Revision 1.4  1998/01/13 23:45:45  ddavies
jpegint.h -> /product/ipcore/ipshared/include/RCS/jpegint.h,v

CommonCheck: Add IPState parameter to all public and almost all private procedures.

Revision 1.3  1997/07/11 01:44:02  ddavies
jpegint.h -> /product/ipcore/ipshared/include/RCS/jpegint.h,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.2  1996/11/06 00:30:38  xzhu
jpegint.h -> /product/ipcore/ipshared/include/RCS/jpegint.h,v

Changed some functions prototype to fix error return problem.

*/
