/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	grytable.c:
 *
 *          Lookup tables:
 *              Int32   ip_make4bppGrayRotationTable()
 *		Int32	ip_makeGrayRotationTable()
 *		Int32	i_getResconTableSize()
 *		Int32	i_make2To3ResconTable()
 *		Int32	i_make1To2ResconTable()
 *		Int32	ip_make8bppRescaleTable()
 *		Int32	ip_make8bppAreaMapTable()
 *
 */


#include <stddef.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"

    /*  prototypes */
#include "shrip.pub"	/* for c4bppGrayRotTableSize, c8bppGrayRotTableSize */
#include "shrscal.pub"
#include "shrscal.prv"
#include "memory.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: grytable.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/* PRIVATE */
/**************************************************************************
 *  ip_make4bppGrayRotationTable  ()
 *	Input:
 *	tableSize:		if *pp4bppGrayRotationTable != NULL,
 *				this is the size of the table
 *				passed in.  If *pp4bppGrayRotation
 *				table is NULL, this is ignored.
 *	p4bppGrayRotationTable:	ptr to ptr to table
 *
 *	On return:
 *		gray rotation table is allocated if necessary and
 *		filled with data.
 **************************************************************************/
Int32  CDECL 
ip_make4bppGrayRotationTable (
    IPState    *pIPState,
    Int32	tableSize,
    UInt16    **p4bppGrayRotationTable)
{
PROLOGUE;
 Int32		index;		/* index into table */
Int32		pix, x, y;	/* pixel, x and y offsets */
UInt16         *locTab;         /* local table pointer */

#define cPixelMask4bpp		0xF
#define cXFracMask4bpp	0x7
#define cYFracMask4bpp	0x7
#define cXFracPos4bpp		7
#define cYFracPos4bpp		4


	/* allocate the table if one doesn't already exist */
    locTab = *p4bppGrayRotationTable;
    if (locTab == NULL)
    {
	locTab = (UInt16 *)CALLOC(1, c4bppGrayRotTableSize);
	if (locTab == NULL)
	{
	    return ia_nomem;
	}
    }
    else	/* is the supplied table the right size? */
    {
	if (tableSize != c4bppGrayRotTableSize)
	    return ia_invalidParm;
    }

	/* Fill the table.  For each x:y:pixel value, we compute the fraction
	 * of the pixel value that goes to each of the four pixels in the
	 * long int.
	 */

    for (index = 0; index < 1024; index++)
    {
	pix = (index  & cPixelMask4bpp);
	pix = (pix << 4);
	x = (index >> cXFracPos4bpp) & cXFracMask4bpp;
	y = (index >> cYFracPos4bpp) & cYFracMask4bpp;

	    /* Only the portion of the pixel specified by the y coord lies
	     * above the boundary to the next lower line.  That fraction
	     * of the pixel will be used in this X location, so let's
	     * compute it now.  There's a table in the big comments at the
	     * top of the file that explains this computation.  Basically,
	     * we assume that the y coords 0, 1, 2, ..., 7 correspond to
	     * 15/16, 13/16, ..., 1/16 of the pixel lying above the
	     * boundary.
	     */
	pix = pix * (15 - 2*y)/16;

	locTab[index] =	(((pix * (15 - 2*x)/16) & 0xFF) << 8) |
		        (((pix * (15 - 2*(x ^ cXFracMask4bpp))/16) & 0xFF));
    }	/* end of loop that fills the table */

	/* things worked, so return the table pointer */
    *p4bppGrayRotationTable = locTab;
    NORMAL_RETURN;
}


/* PRIVATE */
/**************************************************************************
 *  ip_makeGrayRotationTable()
 *	Input:
 *	tableSize:		if *ppGrayRotationTable != NULL,
 *				this is the size of the table
 *				passed in.  If *ppGrayRotation
 *				table is NULL, this is ignored.
 *	pGrayRotationTable:	ptr to ptr to table
 *
 *	On return:
 *		gray rotation table is allocated if necessary and
 *		filled with data.
 **************************************************************************/
Int32  CDECL 
ip_makeGrayRotationTable (
    IPState    *pIPState,
    Int32	tableSize,
    UInt16    **pGrayRotationTable)
{
PROLOGUE;
 Int32		index;		/* index into table */
Int32		pix, x, y;	/* pixel, x and y offsets */
UInt16         *locTab;        /* local table pointer */

#define cPixMask	0xFF
#define cXFracMask	0x7
#define cYFracMask	0x7
#define cXFracPos	10
#define cYFracPos	7

	/* allocate the table if one doesn't already exist */
    locTab = *pGrayRotationTable;
    if (locTab == NULL)
    {
	locTab = (UInt16 *)CALLOC(1, c8bppGrayRotTableSize);
	if (locTab == NULL)
	{
	    return ia_nomem;
	}
    }
    else	/* is the supplied table the right size? */
    {
	if (tableSize != c8bppGrayRotTableSize)
	    return ia_invalidParm;
    }

	/* Fill the table.  For each x:y:pixel value, we compute the fraction
	 * of the pixel value that goes to each of the four pixels in the
	 * long int.
	 */

    for (index = 0; index < 8192; index++)
    {
	pix = (index << 1)  & cPixMask;
	x = (index >> cXFracPos) & cXFracMask;
	y = (index >> cYFracPos) & cYFracMask;

	    /* Only the portion of the pixel specified by the y coord lies
	     * above the boundary to the next lower line.  That fraction
	     * of the pixel will be used in this X location, so let's
	     * compute it now.  There's a table in the big comments at the
	     * top of the file that explains this computation.  Basically,
	     * we assume that the y coords 0, 1, 2, ..., 7 correspond to
	     * 15/16, 13/16, ..., 1/16 of the pixel lying above the
	     * boundary.
	     */
	pix = pix * (15 - 2*y)/16;

	locTab[index] =
		(((pix * (15 - 2*x)/16) & 0xFF) << 8) |
		(((pix * (15 - 2*(x ^ cXFracMask))/16) & 0xFF));
    }	/* end of loop that fills the table */

	/* things worked, so return the table pointer */
    *pGrayRotationTable = locTab;
    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *  i_getResconTableSize()
 *	Get the size of the resolution conversion tables.
 *
 *	Input:
 *		srcRes:		resolution of source
 *		destRes:	resolution of depth
 *		depth:		depth of src and dest images
 *		pTableSize:	ptr to returned size of table in bytes.
 *				The returned size may be 0 if no table is
 *				needed.
 *
 **************************************************************************/
Int32  CDECL 
i_getResconTableSize (
    IPState    *pIPState,
    UInt32	srcRes,
    UInt32	destRes,
    UInt32	depth,
    UInt32     *pTableSize)
{
	/* Just do it */
   PROLOGUE;
   if (2 * srcRes == destRes)	/* doing 1 to 2 resolution conversion */
   {
	if (depth == 8)
	    *pTableSize = 0;
	else
	    *pTableSize = 256*sizeof(UInt32);
    }
    else
    if (2 * destRes == 3 * srcRes)	/* doing 2 to 3 res conversion */
    {
	if (depth == 8)
	    *pTableSize = 0;
	else
	    *pTableSize = 256*sizeof(UInt32);
    }
    else	/* illegal resolution conversion */
	return ia_invalidParm;

    NORMAL_RETURN;
}


/* PUBLIC */
/**************************************************************************
 *  i_make1To2ResconTable()
 *	Make table used when double image resolution.
 *
 *	Input:
 *		ppResconTable:	ptr to variable which holds pointer to table
 *				If the *ppResconTable is not NULL, we assume
 *				that the table has already been allocated.
 *				If *ppResconTable == NULL, the table is
 *				allocated here and *ppResconTable is set
 *				to point to it.
 *		depth:		depth of image
 *
 *	On return:
 *		2x gray expansion table is generated
 **************************************************************************/
Int32  CDECL 
i_make1To2ResconTable (
    IPState    *pIPState,
    UInt32    **ppResconTable,
    UInt32	depth)
{
PROLOGUE;
 Int32		index;		/* index into table */
UInt32		pixelHigh, pixelLow;
UInt32		tableSize;
UInt32         *locTab;         /* local table pointer */

    if (depth == 8)	/* no table needed in the version of the code */
	NORMAL_RETURN;

	/* Allocate the table if necessary */
    if (*ppResconTable == NULL)
    {
	i_getResconTableSize (pIPState, 1, 2, depth, &tableSize);
	*ppResconTable = (UInt32 *)MALLOC(tableSize);
    }

    if (*ppResconTable == NULL)
	return ia_nomem;

    locTab = (UInt32 *) *ppResconTable;

	/* if this is 4 bit/pixel, turn pixel pairs AB into pixel triplets
	 * as in:
	 *	28    24    20     16    12        8     4     0
	 * |  0  |  0  |  0  |  0  |  A  | (A+B)/2 |  B  |  0  |
	 * We don't put the A/2 and B/2 terms on the end because that
	 * discards the LS bits.  We can't afford to do that, so the
	 * connecting terms are explicitly computed during the
	 * resolution conversion process.
	 */
    for (index = 0; index < 256; index++)
    {
	pixelHigh = index >> 4;
	pixelLow = index & 0x0F;
	    /* This table turns every pair of pixels into 5 pixels */
	locTab[index] = (pixelHigh << 12) |
				  (((pixelHigh + pixelLow + 1) / 2) << 8) |
				  (pixelLow << 4);
    }
    NORMAL_RETURN;
}



/* PUBLIC */
/**************************************************************************
 *  i_make2To3ResconTable()
 *	Make table used in 200x200x4 to 300x300x4 resolution conversion.
 *
 *	Input:
 *		ppResconTable:	ptr to variable which holds pointer to table
 *				If the *ppResconTable is not NULL, we assume
 *				that the table has already been allocated.
 *				If *ppResconTable == NULL, the table is
 *				allocated here and *ppResconTable is set
 *				to point to it.
 *		depth:		depth of image
 *
 *	On return:
 *		1.5x gray expansion table is generated
 **************************************************************************/
Int32  CDECL 
i_make2To3ResconTable (
    IPState    *pIPState,
    UInt32    **ppResconTable,
    UInt32	depth)
{
PROLOGUE;
 Int32		index;		/* index into table */
UInt32		pixelHigh, pixelLow;
UInt32		tableSize;
UInt32         *locTab;         /* local table pointer */

    if (depth == 8)	/* no table needed in the version of the code */
	NORMAL_RETURN;

	/* Allocate the table if necessary */
    if (*ppResconTable == NULL)
    {
	i_getResconTableSize (pIPState, 2, 3, depth, &tableSize);
	*ppResconTable = (UInt32 *)MALLOC(tableSize);
    }

    if (*ppResconTable == NULL)
	return ia_nomem;

    locTab = (UInt32 *) *ppResconTable;

	/* if this is 4 bit/pixel, turn pixel pairs into pixel triplets */
    for (index = 0; index < 256; index++)
    {
	pixelHigh = index >> 4;
	pixelLow = index & 0x0F;
	    /* This table turns every pair of pixels into three pixels */
	locTab[index] = (pixelHigh << 8) |
				  (((pixelHigh + pixelLow) / 2) << 4) |
				  pixelLow;
    }
    NORMAL_RETURN;
}




/*  This table is currently arrange as one big table indexed by:
 * | 14 |   13   12   11    |  10     9      8  | 7 6 5 4 3 2 1 | 0 |
 * | tab| 3b xIOCoord frac  | 3b yIOCoord frac  |  7b pixel val | 0 |
 *
 * The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
 * yIOCoord bits are its 3 MS fraction bits.  The 7 pixel bits are
 * the 7 MS pixel bits.
 * The 0 at the end makes this an index into a table of short ints.
 * The top bit, tab, determines if we are using the high half of the table;
 * see comments at the start of arbRescale8bppLine.
 * Some of the entries may be unused because it is impossible to reference
 * some entries with the given magnification factor.
 * This large table makes building the table relatively easy but
 * makes the inner loop of arbRescale8bppLine kinda messy mostly because
 * the table index is made of 4 different variables.
 * One way to make tables referencing easier is to make lots of smaller
 * tables.  We have not done this yet but it is described here so that
 * we won't loose the idea.  
 * We have one small table that may be used several times in succession
 * and then move on to the next table as we move in the x direction
 * across the image.  Let's say the magnification factor is 1.001.  Then
 * we the 3 MS bits of the xIOCoord are 000 for several destination pixels.
 * Why not package that as one table that is only indexed by pixel value.
 * Each new table would probably consist of a count of the numer of
 * times to use this table as we're moving in the x direction, a count
 * of the number of times we should use this entire series of tables in
 * the y direction, a pointer to the next table in the x direction, a
 * pointer to the next table in the y direction, and 2 pointers to tables
 * that are indexed by pixel value.  Why 2 pointers?  Well, for example, if 
 * we are doing the first line after sync'ing we only need one source
 * line to make the destination line and hence one table reference
 * gives us the entire pixel value (this is just as it is now.  sometimes
 * we need 2 source lines to make a destination line and so we reference
 * the table twice.)  Instead of keeping xIOCoord in the inner loop,
 * we keep track of it when we make the table.  All the inner loop has
 * to do is decrement the counter and go to the next table when it 
 * reaches 0 (ditto for the y direction).  This turns out to be a
 * big win if the 3 MS bits of the fraction are the same for several
 * pixels (this happens when the magnification factor is less than 1.125).
 * If the magnification factor is greater than 1.5 we basically break
 * even since we are spending more time duplicating pixels than 
 * referencing the table anyway.  If the magnification factor is between
 * 1.125 and 1.5, we win some but not as much (at least on a sparc)
 * because we are changing tables between each destination pixel.
 * 
 * How may of these stuctures are in the x direction (or y direction
 * for that matter)?  It may be more that 8 (since we are using 3 fraction
 * bits).  Say that we need to drop out an extra destination pixel
 * every 2.3 source pixels.  extraPixel (see arbRescale8bppLine) will
 * be 2.3, 1.3, 0.3, 2.6, 1.6, 0.6, 2.9, 1.9, 0.9, 3.2, etc.
 * Remember that we put out an extra destination pixel every time
 * extraPixel is 0.  The fractional part of xIOCoord is reset to 0 each
 * time that we put out the extra destination pixel.  This means that
 * we see 3 different xIOCoord values so that we only need 3 tables
 * that are referenced by pixel value.  However, we need more than 3 
 * structures since the count of referencing the table may change when
 * the integer part goes from 2 to 3 or we might need another table.
 * How may structures do we have to have?  Well, in the worst case I guess
 * we would have to do the entire line and so we could need many structures.
 * This could increase our overall memory usage somewhat.  Maybe 
 * there's some way to stop earlier and just claim that we're reusing
 * the structures that we've already calculated.  I'm not sure about
 * this part.....
 */

/* PRIVATE */
/**************************************************************************
 *  ip_make8bppRescaleTable()
 *	Make tables used in arbitrary rescaling by linear interpolation.
 *
 *	Input:
 *	ppRescaleTable		pointer to pointer to rescale table
 *	ppStateTable		pointer to pointer to state table
 *	wD			width of the destination image
 *	xOIIncrement		amount to get to next input pixel in output 
 *				pixel coordinate space.  split integer.
 *	Split integers are already in form to be used in index.
 *
 *	On return:
 *		rescale table is generated
 *		state table is generated
 *************************************************************************/
Int32  CDECL 
ip_make8bppRescaleTable (
    IPState    *pIPState,
    UInt16    **ppRescaleTable,
    UInt8     **ppStateTable,
    Int32	wD,
    UInt32	xOIIncrement)
{
PROLOGUE;
 Int32		 xIndex, yIndex;
Int32		 xOICoord1, xOICoord2, xOICoord3, xOICoord4;
Float32		 xFracLeft, xFracRight, yFracTop;
UInt32		 pixel;
UInt32		 pixelLeft, pixelLRem;
UInt32		 pixelLeftTop, pixelLeftBot, pixelLRemTop, pixelLRemBot;
UInt32		 pixelRight, pixelRRem;
UInt32		 pixelRightTop, pixelRightBot, pixelRRemTop, pixelRRemBot;
Int32		 x, y, withRound, withoutRound;
UInt8		*pRescaleTable, *pStateTable;
/* const Float32	 weights[8] = {16.0F/16.0F, 15.0F/16.0F, 11.0F/16.0F, 8.0F/16.0F, 8.0F/16.0F, 4.0F/16.0F, 1.0F/16.0F, 0.0F/16.0F}; */
const Float32	 xWeights[] = {8.0F/8.0F, 7.0F/8.0F, 6.0F/8.0F, 5.0F/8.0F, 4.0F/8.0F, 3.0F/8.0F, 2.0F/8.0F, 1.0F/8.0F};
const Float32	 yWeights[] = {8.0F/8.0F, 7.0F/8.0F, 6.0F/8.0F, 5.0F/8.0F, 4.0F/8.0F, 3.0F/8.0F, 2.0F/8.0F, 1.0F/8.0F};

#define cResXSig	0x8		/* number of x table positions */
#define cResYSig	0x8		/* number of y table positions */

/* cResXFPos, cResXFMask, cResXFAMask, cResXFRound, cResXLSMask,
 * cResXIInc, cResXIMask, cResXIPos,
 * cResYFPos, cResYFMask, cResYFAMask, cResYFInc, cResYIInc,
 * cResYIMask and cResYIPos are all defined in shrscal.prv.
 */

#define cResXSmooth	0x4000
#define cResYSmooth	0x8000

	/* states in the state machine.  see comments in srescale.c */
#define cRes_2_2_State		0x0
#define cRes_2_1_1_State	0x1
#define cRes_1_2_1_State	0x2
#define cRes_1_1_2_State	0x3
#define cRes_1_1_1_1_State	0x4

   *ppRescaleTable = (UInt16 *)CALLOC(32768, sizeof(UInt16));
    if (! *ppRescaleTable)
    {
	return ia_nomem;
    }
    pRescaleTable = (UInt8 *)*ppRescaleTable;

    *ppStateTable = (UInt8 *)MALLOC((wD/4)+1);
    if (! *ppStateTable)
    {
	FREE(pRescaleTable);
	return ia_nomem;
    }
    pStateTable = *ppStateTable;

    for (x = 0; x < cResXSig; x++)
    {
	    xIndex = x << cResXFPos;

	    xFracLeft = xWeights[x];
		/* Rounding causes a problem when
		 * the next 3MS bits should have been 111 but we rounded to
		 * 000.  For example, consider xIndex of
		 * 0x1800 and xOIIncrement of 2572.  After adding these 2
		 * together and rounding we index with 000 and get a 
		 * weight of 1.0.  If xOICoord really was 0x1800 this
		 * is the wrong thing to do since we wanted to index by
		 * 111 and get a weight of 7/8.  If xOICoord was
		 * large enough to round up, we won't use the 2nd pixel
		 * anyway (we'll be in a state that uses 1_1 instead of
		 * 2) so this won't be a problem.  
		 * For a reason, that I have yet to reconstruct rounding 
		 * seems to be necessary - e.g. with a magnification factor
		 * of ~8.0 we seem to get jaggies without the rounding.
		 */
	    withRound = ((xOIIncrement + xIndex + cResXFRound) & cResXFMask) >> cResXFPos;
	    withoutRound = ((xOIIncrement + xIndex) & cResXFMask) >> cResXFPos;
	    if ((withRound == 0) & (withoutRound == 0x7))
		xFracRight = xWeights[withoutRound];
	    else xFracRight = xWeights[withRound];

	    for (y = 0; y < cResYSig; y++)
	    {
		yIndex = (y << cResYFPos);
		yFracTop = yWeights[y];

		for (pixel = 0; pixel < 256; pixel += 2)
		{
                        /* Fill the table for the given pixel.  Part of
                         * this pixel will go into the left destination
                         * pixel (pixelLeft) and part will go into the
                         * right destination pixel (pixelRight).  Since
                         * we would like to fill up the table as quickly
                         * as possible and get pixel values to sum
                         * as closely as possible to the total value,
                         * we will also compute pixelLRem as the amount
                         * that should go into the smoothing portion of the
                         * table.  In addition, we must take into account
                         * how much of pixelLeft, etc. can be put in
                         * the destination pixel because we are only
                         * using a portionof the pixel in the y direction.
                         * Hence, Top and Bot.
			 * *Rem values used to go into places computed as
			 * OICoord ^ RES_*FMask.  Since these 
			 * values are in a special part of the table
			 * (indicated by RES_*SMOOTH) let's just index
			 * them by xFrac and save ourselves in the inner
			 * loop.
                         */		    
		    pixelLeft = (UInt32) (xFracLeft * pixel);
		    pixelLRem = pixel - pixelLeft;
		    pixelLeftTop = (UInt32) (pixelLeft * yFracTop);
		    pixelLeftBot = pixelLeft - pixelLeftTop;
		    pixelLRemTop = (UInt32) (pixelLRem * yFracTop);
		    pixelLRemBot = pixelLRem - pixelLRemTop;

		    pixelRight = (UInt32) (xFracRight * pixel);
		    pixelRRem = pixel - pixelRight;
		    pixelRightTop = (UInt32) (pixelRight * yFracTop);
		    pixelRightBot = pixelRight - pixelRightTop;
		    pixelRRemTop = (UInt32) (pixelRRem * yFracTop);
		    pixelRRemBot = pixelRRem - pixelRRemTop;

		    *((UInt16 *)&pRescaleTable[xIndex + yIndex + pixel]) = 
			(UInt16)((pixelLeftTop << 8) | pixelRightTop);
		    *((UInt16 *)&pRescaleTable[cResYSmooth + xIndex + yIndex + pixel]) = 
			(UInt16)((pixelLeftBot << 8) | pixelRightBot);
		    *((UInt16 *)&pRescaleTable[cResXSmooth + xIndex + yIndex + pixel]) = 
			(UInt16)((pixelLRemTop << 8) | pixelRRemTop);
		    *((UInt16 *)&pRescaleTable[cResYSmooth + cResXSmooth + xIndex + yIndex + pixel]) = 
			(UInt16)((pixelLRemBot << 8) | pixelRRemBot);
		}

	    }	/* end for y */
    }		/* end for x */

        /* prepare the state that we will encounter on the way across the
         * line.  Do it know because its easier to calculate it once
         * than in the inner loop of each line.
         */
    xOICoord1 = 0;
    for (x = 0; x < wD; x += 4)
    {
	xOICoord2 = (xOICoord1 & cResXFAMask) + xOIIncrement;
	xOICoord3 = (xOICoord2 & cResXFAMask) + xOIIncrement;
	xOICoord4 = (xOICoord3 & cResXFAMask) + xOIIncrement;

	    /* Do we cross a source pixel between 1st and 2nd dest? */
	if ((xOICoord2 >> cResXIPos) == 0)
	{
		/* Then its either cRes_2_2_State or cRes_2_1_1_State. 
		 * Do we cross a source pixel between 3rd and 4th dest?
		 */
	    if ((xOICoord4 >> cResXIPos) == 0)
		*pStateTable++ = cRes_2_2_State;
	    else 
		*pStateTable++ = cRes_2_1_1_State;
	}
	else
	{
		/* Then its either cRes_1_2_1_State or cRes_1_1_2_State
		 * or cRes_1_1_2_State. 
		 */
	    if ((xOICoord3 >> cResXIPos) == 0)
		*pStateTable++ = cRes_1_2_1_State;
	    else if ((xOICoord4 >> cResXIPos) == 0)
		*pStateTable++ = cRes_1_1_2_State;
	    else *pStateTable++ = cRes_1_1_1_1_State;
	}
	xOICoord1 = (xOICoord4 + xOIIncrement) & cResXFAMask;
    }
    NORMAL_RETURN;
}

/*  This table is currently arrange as one big table indexed by:
 * | 14 |   13   12   11    |  10     9      8  | 7 6 5 4 3 2 1 | 0 |
 * | tab| 3b xIOCoord frac  | 3b yIOCoord frac  |  7b pixel val | 0 |
 *
 * The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
 * yIOCoord bits are its 3 MS fraction bits.  The 7 pixel bits are
 * the 7 MS pixel bits.
 * The 0 at the end makes this an index into a table of short ints.
 * The top bit, tab, determines if we are using the high half of the table;
 * see comments at the start of arbRescale8bppLine.
 * Some of the entries may be unused because it is impossible to reference
 * some entries with the given magnification factor.
 * This large table makes building the table relatively easy but
 * makes the inner loop of arbRescale8bppLine kinda messy mostly because
 * the table index is made of 4 different variables.
 * One way to make tables referencing easier is to make lots of smaller
 * tables.  We have not done this yet but it is described here so that
 * we won't loose the idea.  
 * We have one small table that may be used several times in succession
 * and then move on to the next table as we move in the x direction
 * across the image.  Let's say the magnification factor is 1.001.  Then
 * we the 3 MS bits of the xIOCoord are 000 for several destination pixels.
 * Why not package that as one table that is only indexed by pixel value.
 * Each new table would probably consist of a count of the numer of
 * times to use this table as we're moving in the x direction, a count
 * of the number of times we should use this entire series of tables in
 * the y direction, a pointer to the next table in the x direction, a
 * pointer to the next table in the y direction, and 2 pointers to tables
 * that are indexed by pixel value.  Why 2 pointers?  Well, for example, if 
 * we are doing the first line after sync'ing we only need one source
 * line to make the destination line and hence one table reference
 * gives us the entire pixel value (this is just as it is now.  sometimes
 * we need 2 source lines to make a destination line and so we reference
 * the table twice.)  Instead of keeping xIOCoord in the inner loop,
 * we keep track of it when we make the table.  All the inner loop has
 * to do is decrement the counter and go to the next table when it 
 * reaches 0 (ditto for the y direction).  This turns out to be a
 * big win if the 3 MS bits of the fraction are the same for several
 * pixels (this happens when the magnification factor is less than 1.125).
 * If the magnification factor is greater than 1.5 we basically break
 * even since we are spending more time duplicating pixels than 
 * referencing the table anyway.  If the magnification factor is between
 * 1.125 and 1.5, we win some but not as much (at least on a sparc)
 * because we are changing tables between each destination pixel.
 * 
 * How may of these stuctures are in the x direction (or y direction
 * for that matter)?  It may be more that 8 (since we are using 3 fraction
 * bits).  Say that we need to drop out an extra destination pixel
 * every 2.3 source pixels.  extraPixel (see arbRescale8bppLine) will
 * be 2.3, 1.3, 0.3, 2.6, 1.6, 0.6, 2.9, 1.9, 0.9, 3.2, etc.
 * Remember that we put out an extra destination pixel every time
 * extraPixel is 0.  The fractional part of xIOCoord is reset to 0 each
 * time that we put out the extra destination pixel.  This means that
 * we see 3 different xIOCoord values so that we only need 3 tables
 * that are referenced by pixel value.  However, we need more than 3 
 * structures since the count of referencing the table may change when
 * the integer part goes from 2 to 3 or we might need another table.
 * How may structures do we have to have?  Well, in the worst case I guess
 * we would have to do the entire line and so we could need many structures.
 * This could increase our overall memory usage somewhat.  Maybe 
 * there's some way to stop earlier and just claim that we're reusing
 * the structures that we've already calculated.  I'm not sure about
 * this part.....
 */


/* PRIVATE */
/**************************************************************************
 *  ip_make8bppAreaMapTable()
 *	Make table used in rescaling using an area mapping algorithm.
 *
 *	Input:
 *	ppRescaleTableTop	pointer to pointer to rescale table
 *	wD			width of destination image 
 *	hD			height of destination image 
 *	xIOIncrement	amount to get to next input pixel in output 
 *				pixel coordinate space.  split integer.
 *	yIOIncrement	amount to get to next line in output pixel 
 *				coordinate space.  split integer.
 *
 *	On return:
 *		rescale table is generated
 **************************************************************************/
Int32  CDECL 
ip_make8bppAreaMapTable (
    IPState    *pIPState,
    UInt16    **ppRescaleTable,
    UInt8     **ppStateTable,
    Int32	wD,
    Int32	hD,
    Int32	xIOIncrement,
    Int32	yIOIncrement,
    Int32	xCount)
{
PROLOGUE;
 Int32		 xIndex, yIndex;
Int32		 xIOCoord, yIOCoord, yIOCoordTop, extraPixel;
Float32		 xFrac1, xFrac2, yFrac1;
UInt32		 pixel, pixelHi, pixelLo;
UInt32		 xDoneCount, yDoneCount;
UInt32		 xDone[8], yDone[8], yIOCValues[8];
Int32		 i, x, y, xChange;
UInt8		*pRescaleTable, *pStateTable;
UInt32		 state;

#define cResXCHGMask	0x4000

#define cExtraPixel1Start	0
#define cExtraPixel1StartExtra	1
#define cExtraPixel1Finish	2
#define cExtraPixel1FinishXCount2 3
#define cExtraPixel2Start	4
#define cExtraPixel2Finish	5
#define cExtraPixel3Start	6
#define cExtraPixel3Finish	7
#define cExtraPixel4Start	8
#define cExtraPixel4Finish	9
#define cExtraPixel5OrOverStart	10
#define cExtraPixel5OrOverCont	11

#define cRescaleStart		0
#define cRescaleStartExtra	1
#define cRescaleFinish		2
#define cRescaleExtra1Kludge	3

#define cExtraPixelDec1		0x4000
#define cExtraPixelDec2		0x8000
#define cExtraPixelDec3		0xC000
#define cExtraPixelDec4		0x10000


   *ppRescaleTable = (UInt16 *)CALLOC(8192, sizeof(UInt32));
    if (! *ppRescaleTable)
    {
	return ia_nomem;
    }
    pRescaleTable = (UInt8 *)*ppRescaleTable;
    *ppStateTable = (UInt8 *)MALLOC((wD/4)+1);
    if (! *ppStateTable)
    {
	FREE(pRescaleTable);
	return ia_nomem;
    }
    pStateTable = *ppStateTable;

    xIOCoord= yIOCoord = 0;
	/* must be the same as code that rescales a line */
    xIOIncrement = xIOIncrement >> 2;
    xCount = xCount >> 2;
    yIOIncrement = yIOIncrement >> 5;

    xDoneCount = yDoneCount = 0;
    for (i = 0; i < 8; i++)
    {
	xDone[i] = yDone[i] = yIOCValues[i] = 0;
    }
	/* first find the yIOCoord for each possible entry into the table.
	 * Note that although the index is only 3 bits we use all the
	 * significance that we can to calculate the pixel values.
	 * The table is indexed by the 3 frac bits of yIOCoord (for the 
	 * bottom part of the pixel) and 1 minus the 3 fract bits of 
	 * yIOCoord (for the top part of the pixel).  If those frac bits
	 * are 0, the table is only indexed once.
	 * Keep the values in yIOCValues.  yDone (and xDone) indicate
	 * if we have found the value for a particular index.
	 * yDone (and xDone) keep the total number of values that have
	 * been found.
	 */
    for (y = 0; y < hD; y++)
    {
	    /* have we found the values for this index? */
	if (yDone[(yIOCoord & cResYFMask) >> cResYFPos] == 0)
	{
	    yDoneCount++;
	    yDone[(yIOCoord & cResYFMask) >> cResYFPos] = 1;
	    yIOCValues[(yIOCoord & cResYFMask) >> cResYFPos] = yIOCoord;
	}

	    /* how about the 1-coord value? */
	yIOCoordTop = ((yIOCoord & cResYIMask) + cResYIInc) - yIOCoord;
	if (yDone[((cResYIInc - (yIOCoord & cResYFMask)) & cResYFMask) >> cResYFPos] == 0)
	{
	    yDoneCount++;
	    yDone[((cResYIInc - (yIOCoord & cResYFMask)) & cResYFMask) >> cResYFPos] = 1;
	    yIOCValues[((cResYIInc - (yIOCoord & cResYFMask)) & cResYFMask) >> cResYFPos] = yIOCoordTop;
	}

	if (yDoneCount == 8) 
	    break;
	yIOCoord += yIOIncrement;
    }

    for (x = 0; x < wD; x++)
    {
	if (xDone[(xIOCoord & cResXFMask) >> cResXFPos] != 1)
	{
	    xFrac1 = ((Float32)((xIOCoord >> cResXIPos) + 1)) - 
	      (((Float32)xIOCoord) / 16384);
	    xFrac2 = ((Float32)((xIOCoord + xIOIncrement) - 
	      ((xIOCoord & cResXIMask) + cResXIInc))) / 16384;
	    xIndex = xIOCoord & cResXFMask;

	    xDone[(xIOCoord & cResXFMask) >> cResXFPos] = 1;
	    xDoneCount++;

	    for (y = 0; y < cResYIInc; y += cResYFInc)
	    {
		    /* if we can didn't generate this value, approximate
		     * it with the 3 most significant bits.
		     */
		if (yDone[y >> cResYFPos] == 0) 
		    yIOCoord = y;
		else yIOCoord = yIOCValues[y >> cResYFPos];
		yFrac1 = ((Float32)((yIOCoord >> cResYIPos) + 1)) - 
	      	      (((Float32)yIOCoord) / 2048);
		yIndex = xIndex + y;

		for (pixel = 0; pixel < 256; pixel += 2)
		{
		    *((UInt16 *)&pRescaleTable[yIndex + pixel]) = 
	              (UInt16)(((((UInt32)(pixel * xFrac1 * yFrac1)) & 0xFF)
				<< 8) |
	                       (((UInt32)(pixel * xFrac2 * yFrac1)) & 0xFF));
		    *((UInt16 *)&pRescaleTable[yIndex + pixel + cResXCHGMask]) = 
	              (UInt16)(((((UInt32)(pixel * xFrac1 * yFrac1)) & 0xFF)
				<< 8) |
	                       (((UInt32)(pixel * xFrac2 * yFrac1)) & 0xFF));
		}

	    }	/* end for y */
	    if (xDoneCount == 8) break;
	}	/* end if xDone */

	    /* calculate the tables used when the least signficant
	     * bit of the index changes.
	     */
	if ((xIOIncrement & cResXFMask) == 0)
	    xChange = (((xIOCoord + xIOIncrement) ^ xIOCoord) & cResXLSMask);
	else xChange = 0;
	if (xChange)
	{
	    xFrac1 = ((Float32)((xIOCoord >> cResXIPos) + 1)) - 
	      (((Float32)xIOCoord) / 16384);
	    xFrac2 = ((Float32)((xIOCoord + xIOIncrement) - 
	      ((xIOCoord & cResXIMask) + cResXIInc))) / 16384;
	    xIndex = xIOCoord & cResXFMask;

	    for (y = 0; y < cResYIInc; y += cResYFInc)
	    {
		    /* if we can never generate this value just 
		     * continue on to the next.
		     */
		if (yDone[y >> cResYFPos] == 0) 
		    yIOCoord = y;
		else yIOCoord = yIOCValues[y >> cResYFPos];

		yFrac1 = ((Float32)((yIOCoord >> cResYIPos) + 1)) - 
	      	      (((Float32)yIOCoord) / 2048);
		yIndex = xIndex + y;

	        for (pixel = 0; pixel < 256; pixel += 2)
	        {
		    pixelHi = *((UInt16 *)&pRescaleTable[
		      yIndex + pixel + cResXCHGMask]) & 0xFF00;
		    pixelLo = (((UInt32)(pixel * xFrac2 * yFrac1)) & 0xFF);
	            *((UInt16 *)&pRescaleTable[
		      yIndex + pixel + cResXCHGMask]) =
			  (UInt16) (pixelHi | pixelLo);
	        }

	    }	/* end for y */

	}
        xIOCoord += xIOIncrement;
    }		/* end for x */


    extraPixel = xCount;
    state = cRescaleStart;
    for (x = 0; x < wD; x += 4)
    {
        switch (extraPixel >> cResXIPos)
        {
        case 1:
	    if ((state == cRescaleStart) || (state == cRescaleExtra1Kludge))
	    {
	        *pStateTable++ = cExtraPixel1Start;
		if (state == cRescaleExtra1Kludge)
			/* undo the kludge */
		    extraPixel += cExtraPixelDec1;

		    /* Occassionally we may not want to drop out an
		     * extra pixel.  This can happen once during a
		     * loop iteration but not more than once.
		     */
		if ((extraPixel >> cResXIPos) < 2)
		{
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		    if ((extraPixel >> cResXIPos) < 2)
		    {
			extraPixel = (extraPixel & cResXFAMask) + xCount;
			state = cRescaleStart;
		    }
		    else
		    {
			extraPixel -= cExtraPixelDec1;
			state = cRescaleStartExtra;
		    }
		}
		else
		{
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		    if ((extraPixel >> cResXIPos) < 2)
			state = cRescaleStartExtra;
		    else state = cRescaleStart;
		}
	        if ((extraPixel >> cResXIPos) >= 2)
		{
			/* we want to stay in case 1 with state equal to
			 * cRescaleStart or cRescaleStartExtra but this
			 * would put us in case 2. so decrement extraPixel
			 * and introduce a new state.
			 */
		    extraPixel -= cExtraPixelDec1;
		    state = cRescaleExtra1Kludge;
		}
	    }
	    else if (state == cRescaleStartExtra)
	    {
	        *pStateTable++ = cExtraPixel1StartExtra;
		extraPixel = (extraPixel & cResXFAMask) + xCount;
		if ((extraPixel >> cResXIPos) < 2)
		{
		    extraPixel = (extraPixel & cResXFAMask) + xCount;
		    if ((extraPixel >> cResXIPos) >= 2)
		    {
			extraPixel -= cExtraPixelDec1;
			state = cRescaleStart;
		    }
		    else state = cRescaleStartExtra;
		}
		else
		{
		    extraPixel -= cExtraPixelDec1;
		    state = cRescaleStart;
		}
	    }
	    else	/* must be cRescaleFinish */
	    {
	        if ((xCount >> cResXIPos) == 2)
	        {
	            *pStateTable++ = cExtraPixel1FinishXCount2;
	            extraPixel = (extraPixel & cResXFAMask) + xCount;
	            state = cRescaleStart;
	        }
	        else
	        {
	            *pStateTable++ = cExtraPixel1Finish;
	            extraPixel = (extraPixel & cResXFAMask) + xCount - cExtraPixelDec2;
	            state = cRescaleFinish;
	        }
	    }
	    break;
        case 2:
	    if (state == cRescaleStart)
	    {
	        *pStateTable++ = cExtraPixel2Start;
	        extraPixel = (extraPixel & cResXFAMask) + xCount;
	        state = cRescaleFinish;
	    }
	    else	/* must be cRescaleFinish */
	    {
	        *pStateTable++ = cExtraPixel2Finish;
	        extraPixel = (extraPixel & cResXFAMask) + xCount - cExtraPixelDec1;
	        state = cRescaleFinish;
	    }
	    break;
        case 3:
	    if (state == cRescaleStart)
	    {
	        *pStateTable++ = cExtraPixel3Start;
	        extraPixel = (extraPixel & cResXFAMask) + xCount;
	        state = cRescaleStart;
	    }
	    else	/* must be cRescaleFinish */
	    {
	        *pStateTable++ = cExtraPixel3Finish;
	        extraPixel = (extraPixel & cResXFAMask) + xCount;
	        state = cRescaleFinish;
	    }
	    break;
        case 4:
	    if (state == cRescaleStart)
	    {
	        *pStateTable++ = cExtraPixel4Start;
	        extraPixel -= cExtraPixelDec3;
	        state = cRescaleFinish;
	    }
	    else	/* must be cRescaleFinish */
	    {
	        *pStateTable++ = cExtraPixel4Finish;
	        extraPixel = (extraPixel & cResXFAMask) + xCount;
	        state = cRescaleStart;
	    }
	    break;
        default:
	    if (state == cRescaleStart)
	    {
	        *pStateTable++ = cExtraPixel5OrOverStart;
	        extraPixel -= cExtraPixelDec3;
	        state = cRescaleFinish;
	    }
	    else	/* must be cRescaleFinish */
	    {
	        *pStateTable++ = cExtraPixel5OrOverCont;
	        extraPixel -= cExtraPixelDec4;
	        state = cRescaleFinish;
	    }
	    break;
        }
    }

    NORMAL_RETURN;
}

/*

$Log: grytable.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.21  1998/01/13 23:46:49  ddavies
grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.20  1995/12/13 02:41:30  ejaquith
grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.19  1995/12/13  01:18:04  ddavies
 * ddavies on Tue Dec 12 17:15:34 PST 1995
 *
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Change ip_makeGrayRotationTable and ip_make4bppGrayRotationTable so that
 * they can take an existing table as an option.  They now take a pointer to a
 * pointer to a rotation table.  If the *pointer is NULL, a table is created,
 * otherwise we use the one passed in after checking its size with the new
 * tableSize parameters.
 *
 * Revision 1.18  1995/10/18  03:51:46  ddavies
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.17  1994/09/17  17:46:17  ddavies
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Change private procedures that make tables for the 200->300 and 100->200
 * resolution conversion procedures so they are public, not private.
 * Also change them so that if they are supplied with table space, they do not
 * allocate tables.  Also add a procedure, i_getResconTableSize that clients
 * use to decide what size tables to allocate.  The new resolution conversion
 * procedures are called i_make1To2ResconTable and i_make2To3ResconTable.
 *
 * Revision 1.16  1994/07/27  15:57:48  ejaquith
 * ejaquith on Wed Jul 27 08:55:04 PDT 1994
 *
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Made ...GrayRotationTable functions be called with table pointers instead
 * of globals.  Added local pointers in functions for speed.
 *
 * Revision 1.15  1994/02/04  20:12:40  dferrell
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.14  1993/12/02  19:21:54  ddavies
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.13  1993/11/14  19:05:16  ddavies
 * ddavies on Sun Nov 14 11:04:54 PST 1993
 *
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Replace stdio.h with stddef.h at Leigh's suggestion.
 *
 * Revision 1.12  1993/11/13  00:40:43  ddavies
 * ddavies on Fri Nov 12 16:40:31 PST 1993
 *
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Include <stdio.h> to get a definition of NULL.
 *
 * Revision 1.11  1993/11/05  01:26:08  ddavies
 * grytable.c -> /product/ipcore/ipshared/src/RCS/grytable.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

