/*
** Copyright (c) 1992, 1993  Xerox Corporation.  All Rights Reserved.
** Copyright   protection   claimed   includes   all   forms  and
** matters of copyrightable  material and information now allowed
** by statutory or judicial law or hereinafter granted, including
** without  limitation,  material  generated  from  the  software
** programs  which  are  displayed  on the  screen such as icons,
** screen display looks, etc.
*/

/*
 * i_aieutil.c
 *
 *	Contains AIE utility functions
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "memory.pub"
#include "except.prv"
#include "aiedefs.pub"
#include "aiedefs.prv"

/* Prototypes */
#include "shrip.pub"
#include "shrip.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: i_aieutl.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:10 $");

/*
** function:
**	ip_tableInvert ()
** parameters:
**	old			old table
**	new			new table
** purpose:
**	to invert the input table and store the result in new.
**
** return:
**	ia_successful, ia_notSameSize, or ia_nomem
*/

Int32 CDECL 
ip_tableInvert (
    IPState    *pIPState,
    aTable     *pOld,
    aTable     *pNew)
{
    PROLOGUE;
    Int32        i, tableLength;
    Int32	*tmp;

    tableLength = pOld->length;
    ExceptRaise (tableLength != pNew->length, ia_notSameLength);
    ExceptRaise (!(tmp = (Int32 *) CALLOC(tableLength, sizeof(Int32))), ia_nomem);

    for (i = 0; i < tableLength; i++)
	tmp[tableLength - 1 - i] = pOld->table[i];

    for (i = 0; i < tableLength; i++)
	pNew->table[i] = (tableLength - 1) - tmp[i];

    FREE (tmp);

    NORMAL_RETURN;
}


/*
** function:
**	ip_tableApplyGamma()
** parameters:
**	g			gamma value
**	pOld			old table
**	pNew			new table
**	invert			invert table
** purpose:
**	to apply the gamma value to the Table.  if the gamma
**	value is > 1.0, we need to invert the table.
** return:
**	ia_successful
*/

Int32 CDECL 
ip_tableApplyGamma (
    IPState    *pIPState,
    Float64	g,
    aTable     *pOld,
    aTable     *pNew,
    Bool	invert)
{
    PROLOGUE;
    Int32           i, tableLength;
    Float64         f, val;

    tableLength = pOld->length;
    ExceptRaise (tableLength != pNew->length, ia_notSameLength);

    f = (1.0 * (tableLength - 1)) / pow((tableLength - 1), g);
    for (i = 0; i < tableLength; i++)
    {
	val = pow((double) pOld->table[i], g);
	pNew->table[i] = (int) (val * f);
    }

    if (invert)	/* was: g > 1.0 */ 
	return ip_tableInvert (pIPState, pNew, pNew);

    NORMAL_RETURN;
}


/*
** function:
**	ip_tableApplyGain ()
** parameters:
**	g			gain value
**	pOld			old table
**	pNew			new table
** purpose:
**	to apply the gain value to the tables. The resulting values are clamped
**	to the length of the table, which is sort of silly, if you think
**	about it. Perhaps Tables need a range, but that complicates things.
**
** return:
**	ia_successful, ia_notSameLength
*/

Int32 CDECL 
ip_tableApplyGain (
    IPState    *pIPState,
    Float64	g,
    aTable     *pOld,
    aTable     *pNew)
{
    PROLOGUE;
    int             i;

    ExceptRaise (pOld->length != pNew->length, ia_notSameLength);

    for (i = 0; i < pOld->length; i++)
    {
	pNew->table[i] = (int) (MIN_1((pOld->length - 1), g * (pOld->table[i])));
    }

    NORMAL_RETURN;
}


/*
** function:
**	ip_tableFold ()
** parameters:
**	pFirst			first table
**	pSecond			second table
**	pNew			new table
**	invert			invert
** purpose:
**	to fold 2 tables into 1 by first indexing the
**	second table and using that value as the index
**	of the first table. Invert is to invert the
**	second table before applying???
**	note:
**	  first, second, and new may be same array.
**	  tables should all be the same size
** return:
**	ia_successful, ia_notSameLength
*/

Int32 CDECL 
ip_tableFold (
    IPState    *pIPState,
    aTable     *pFirst,
    aTable     *pSecond,
    aTable     *pNew)
{
    PROLOGUE;
    int             i, tableLength;
    Int32           *n;

    tableLength = pFirst->length;
    ExceptRaise (tableLength != pSecond->length || pNew->length != pSecond->length,
		 ia_notSameLength);

    ExceptRaise (!(n = (Int32 *) CALLOC(tableLength, sizeof(Int32))), ia_nomem);

    for (i = 0; i < tableLength; i++)
	n[i] = pFirst->table[pSecond->table[i]];

    for (i = 0; i < tableLength; i++)
	pNew->table[i] = n[i];

    FREE(n);
    NORMAL_RETURN;
}



/*
** function:
**	ip_tableApply
**
** parameters:
**	pH			histogram -- assumed to have 256 entries
**	pTrc			trc
**	pNew			new histogram
**
** purpose:
**	to apply the trc to the histogram,
**	note:
**	  this is different from other apply_XXX()
**	  because we allocate the new buffer here.
**	  [I'm not quite sure why]
**
** return:
**	ia_nomem, ia_successful
*/

Int32 CDECL 
ip_tableApply (
    IPState    *pIPState,
    aTable     *pH,
    aTable     *pTable,
    aTable    **ppNew)
{
    PROLOGUE;
    Int32           i, j;
    Int32          *t;
    Int32          *new_h;
    double          mult;

    ERR_PASS(ip_tableAllocate (pIPState, pTable->length, ppNew));

    t = (*ppNew)->table;

    ExceptRaise (!(new_h = (Int32 *) CALLOC(pTable->length, sizeof(Int32))),
		 ia_nomem);

    /* Scale the histogram to the length of the Table */
    mult = (double) pTable->length / 256.0;

    for (i = 0; i < 256; i++)
	new_h[(int) (i * mult)] = pH->table[i];

    for (i = 0; i < pTable->length; i++)
    {
	j = pTable->table[i];
	t[j] += new_h[i];
    }

    FREE(new_h);

    NORMAL_RETURN;
}


/*
** function:
**	ip_tableScale ()
**
** parameters:
**	mode			mode for scaling
**	pOld			old table
**	pNew			new table
**
** purpose:
**	to apply the appropriate scaling technique.
**	mode == 1:
**		simple scale down from oldsize to newsize
**	mode == 2:
**		sum the contents of the old table cells into
**		the new table.  ex:  if mult is 4, then each new
**		table cell will contain the sum of 4 old cells.
**	mode == 3:
**		simple scale up from oldsize to newsize
**		ex:  if mult is 4, then each new table cell
**		will repeat 4*the contents of the old cell 4 times.
** returns:
**	ip_successful, ...
*/

Int32 CDECL 
ip_tableScale (
    IPState    *pIPState,
    Int32	mode,
    aTable     *pOld,
    aTable     *pNew)
{
    PROLOGUE;
    Int32           i, j, cnt;
    Int32           val;
    Float64         mult;

    switch (mode)
    {
    case 1:
	ip_tableDecimate (pIPState, pOld, pNew);
	break;

    case 2:
	mult = (double) pOld->length / (pNew->length * 1.0);
	for (i = 0; i < pNew->length; i++)
	{
	    val = 0;
	    for (j = 0; j < (int) mult; j++)
		val += pOld->table[(int) (i * mult) + j];
	    pNew->table[i] = val;
	}
	break;

    case 3:
	mult = (double) pOld->length / (pNew->length * 1.0);
	for (i = 0, cnt = 0; i < pNew->length; i++)
	{
	    for (j = 0; j < (int) mult; j++)
		pNew->table[cnt++] = (long) (mult * pOld->table[i]);
	}
	break;

    default:
	return ia_invalidParm;
    }

    NORMAL_RETURN;
}

/*
** function:
**	ip_Interpolate()
** purpose:
**	to provide a library form of the sandpiper "trcint"
**	program which is used to interpolate a set of points.
**	What it does [Peter's explanation] is to take a set of points along
**	a curve, where in[i] is the index into the resulting array,
**	and out[i] is the content at the resulting array[in[i]].
**	The resulting array will be in[points-1] - in[0] + 1 in length.
**
** arguments:
**	points =    Number of points in "in" and "out" arrays
**	in =	    Array of "input" values, i.e. indices into the resulting table
**	out =	    Array of "output" values corresponding to each input value
**	res =	    Number of points that were created
**	result =    Resulting array with points interpolated.
**
*/

Int32           CDECL 
ip_Interpolate (
    IPState    *pIPState,
    Int32	points,
    Int32      *in,
    Int32      *out,
    Int32      *res,
    Int32     **result)
{
    PROLOGUE;
    Int32           x = 0, *knot_array, *new_array, newx;
    Int32           int_x, *int_y;
    Float64        *slope, *val;
    int             ii;
    Int32          *rptr;

    *res = in[points - 1] - in[0] + 1;
    *result = (Int32 *) CALLOC(*res, sizeof(Int32));
    rptr = *result;
    ExceptRaise (!rptr, ia_nomem);

    knot_array = (Int32 *) CALLOC(2, sizeof(Int32));
    new_array = (Int32 *) CALLOC(2, sizeof(Int32));
    int_y = (Int32 *) CALLOC(2, sizeof(Int32));
    slope = (Float64 *) CALLOC(2, sizeof(Float64));
    val = (Float64 *) CALLOC(2, sizeof(Float64));

    /* fill trc array by linear interpolation */
    for (ii = 0; ii < points;)
    {
	if (ii == 0)
	{
	    knot_array[0] = in[ii];
	    knot_array[1] = out[ii];
	    ii++;
	    x = *knot_array;
	    for (int_x = 0; int_x < x; int_x++)
		*rptr++ = knot_array[1];
	}
	new_array[0] = in[ii];
	new_array[1] = out[ii];
	ii++;
	newx = *new_array;

	/* if knots are at the same x location, then skip to the next knot */
	while (newx == x)
	{
	    if (ii == points)	/* this should not happen... */
	    {
		*rptr = newx;	/* it happened, added 9/27/93 */
		return ia_internal;
	    }
	    new_array[0] = in[ii];
	    new_array[1] = out[ii];
	    ii++;
	    newx = *new_array;
	}

	*rptr++ = knot_array[1];

	/* if knots aren't adjacent, interpolate intervening trc values */
	if (newx - x > 1)
	{
	    slope[1] = (Float64) (new_array[1] - knot_array[1]) / (Float64) (newx - x);
	    val[1] = (Float64) knot_array[1];

	    /* print out trc values */
	    for (int_x = x + 1; int_x < newx; int_x++)
	    {
		val[1] += slope[1];
		int_y[1] = (Int32) (val[1] + .5);
		*rptr++ = int_y[1];
	    }
	}
	/* assign old x & y values */
	x = newx;
	knot_array[1] = new_array[1];
    }

    *rptr++ = knot_array[1];

    FREE(knot_array);
    FREE(new_array);
    FREE(int_y);
    FREE(slope);
    FREE(val);

    NORMAL_RETURN;
}

/*
** function:
**	ip_tableInterpolate()
** purpose:
**	to provide a library form of the sandpiper "trcint"
**	program which is used to interpolate a set of points.
**	What it does [Peter's explanation] is to take a set of points along
**	a curve, where in[i] is the index into the resulting array,
**	and out[i] is the content at the resulting array[in[i]].
**	The resulting array will be in[points-1] - in[0] + 1 in length.
**	This differs from the above function in that it creates a Table.
**
** arguments:
**	points =	Number of points in "in" and "out" arrays
**	in =		Array of "input" values, i.e. indices into the resulting table
**	out =		Array of "output" values corresponding to each input value
**	ppTableResult =	Resulting Table with points interpolated.
**
*/

Int32           CDECL 
ip_tableInterpolate (
    IPState    *pIPState,
    Int32	points,
    Int32      *in,
    Int32      *out,
    aTable    **ppTableResult)
{
    PROLOGUE;
    Int32           x = 0, *knot_array, *new_array, newx;
    Int32           int_x, *int_y;
    Float64        *slope, *val;
    int             ii;
    Int32          *rptr, pointsOut;

    pointsOut = in[points - 1] - in[0] + 1;
    ERR_PASS( ip_tableAllocate (pIPState, pointsOut, ppTableResult));
    rptr = ip_tableData (pIPState, *ppTableResult);
 
    /*
     * If you can figure out why these are CALLOC'ed rather than just auto,
     * you win the prize, because I don't know.
     */
    knot_array = (Int32 *) CALLOC(2, sizeof(Int32));
    new_array = (Int32 *) CALLOC(2, sizeof(Int32));
    int_y = (Int32 *) CALLOC(2, sizeof(Int32));
    slope = (Float64 *) CALLOC(2, sizeof(Float64));
    val = (Float64 *) CALLOC(2, sizeof(Float64));

    /* fill trc array by linear interpolation */
    for (ii = 0; ii < points;)
    {
	if (ii == 0)
	{
	    knot_array[0] = in[ii];
	    knot_array[1] = out[ii];
	    ii++;
	    x = *knot_array;
	    for (int_x = 0; int_x < x; int_x++)
		*rptr++ = knot_array[1];
	}
	new_array[0] = in[ii];
	new_array[1] = out[ii];
	ii++;
	newx = *new_array;

	/* if knots are at the same x location, then skip to the next knot */
	while (newx == x)
	{
	    if (ii == points)	/* this should not happen... */
	    {
		*rptr = newx;	/* it happened, added 9/27/93 */
		return ia_internal;
	    }
	    new_array[0] = in[ii];
	    new_array[1] = out[ii];
	    ii++;
	    newx = *new_array;
	}

	*rptr++ = knot_array[1];

	/* if knots aren't adjacent, interpolate intervening trc values */
	if (newx - x > 1)
	{
	    slope[1] = (Float64) (new_array[1] - knot_array[1]) / (Float64) (newx - x);
	    val[1] = (Float64) knot_array[1];

	    /* print out trc values */
	    for (int_x = x + 1; int_x < newx; int_x++)
	    {
		val[1] += slope[1];
		int_y[1] = (Int32) (val[1] + .5);
		*rptr++ = int_y[1];
	    }
	}
	/* assign old x & y values */
	x = newx;
	knot_array[1] = new_array[1];
    }

    *rptr++ = knot_array[1];

    FREE(knot_array);
    FREE(new_array);
    FREE(int_y);
    FREE(slope);
    FREE(val);

    NORMAL_RETURN;
}

/*
** function:
**	ip_tableDecimate
**
** purpose:
**	Decimate a table. Note that "decimation" here also includes
**	the case where one has more output than input, which is really
**	"extrapolation" (did you know that the origin of the word "decimate"
**	is "kill every tenth man, determined by drawing straws?")
**
** arguments:
**	outPoints =	Number of points in the output table
**	pIn =		Input table
**	ppOut =		Output table
**
** returns:
**	ia_successful or ia_nomem
**/

Int32 CDECL 
ip_tableDecimate (
    IPState    *pIPState,
    aTable     *pIn,
    aTable     *pOut)
{
    PROLOGUE;
    Int32           i, j;
    Int32          *rptr;
    Int32           d, s;

    rptr = ip_tableData (pIPState, pOut);

    if (pIn->length >= pOut->length)
    {
	d = pIn->length / pOut->length;
	s = d >> 1;

	for (i = j = 0; i < pOut->length; i++)
	{
	    *rptr++ = (Int32) (pIn->table[j] >> s);
	    j += d;
	}
    }
    else if (pIn->length < pOut->length)
    {
	d = pOut->length / pIn->length;

	for (i = 0; i < pIn->length; i++)
	{
	    for (j = 0; j < d; j++)
		*rptr++ = (Int32) pIn->table[i];
	}
    }

    NORMAL_RETURN;
}

/*
** ip_tableAllocate
**
**	Allocate a Table
**
** arguments:
**	length = length of Table
**	ppTable = pointer to returned table
**
** action:
**	trc allocated, or NULL if no memory
**
** returns:
*	ia_successful or ia_nomem
*/

ErrCode CDECL 
ip_tableAllocate (
    IPState    *pIPState,
    Int32	length,
    aTable    **ppTable)
{
    PROLOGUE;
    aTable *tab;

    /* don't forget the stupid [1] */
    tab = (aTable *) CALLOC (1, sizeof(aTable) + (length - 1) * sizeof(Int32));
    if (tab != NULL)
    {
	tab->length = length;
	*ppTable = tab;
    }
    else
	__error__ = ia_nomem;

    NORMAL_RETURN;
}

/*
** ip_tableFree 
**
**	Free a Table
**/

Int32 CDECL 
ip_tableFree (
    IPState    *pIPState,
    aTable     *pTable)
{
    PROLOGUE;
    FREE (pTable);
    NORMAL_RETURN;
}


/*******************************************************************************
 * ip_tableData
 *
 *	Returns a pointer to the data for a table
 *
 *******************************************************************************/

Int32 * CDECL 
ip_tableData (
    IPState    *pIPState,
    aTable     *pTable)
{
    PROLOGUE;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    return &(pTable->table[0]);
}

/*******************************************************************************
 * ip_tableLength
 *
 *	Returns the length of a table 
 *
 *******************************************************************************/

Int32 CDECL 
ip_tableLength (
    IPState    *pIPState,
    aTable     *pTable)
{
    PROLOGUE;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    return pTable->length;
}

/*

$Log: i_aieutl.c,v $
Revision 1.1.1.1  1999/09/20 18:18:10  jlin
Imported sources

Revision 1.9  1998/01/13 23:47:00  ddavies
i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.8  1997/11/13 23:21:16  aber
i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v

CommonCheck: Updated the AIE algorithms to match (or closely match) the latest DITC
AIE used in Sandpiper. The updates include the following:
1. Exclude a frame (5%) around the exterior in almost all statistics.
2. Make sure the black-point and white-point are included in the
   computations.
3. ... this was improved upon to not use just the first region with a
   black-or-white-point, but to use the ones with the largest variance.
4. The statistics are accumulated in cells 4 across and 4 down. This
   is not done if the width and height are less than 512.
5. Black-point and white-point do not include 0 and 255 respectively.
   They are the darkest and whitest not quite at these saturation points.
6. If the number of colors or gray levels is too low, less aggressive
   correction is applied.
7. Lots of thresholds changed.
8. Luminance channel is computed with a different formula.

Revision 1.7  1995/12/13 02:41:57  ejaquith
i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.6  1994/07/12  23:55:27  ejaquith
 * i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v
 *
 * removed #include "iadef.prv".
 *
 * Revision 1.5  1994/06/28  01:57:01  macleod
 * macleod on Mon Jun 27 18:56:44 PDT 1994
 *
 * i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v
 *
 * A bunch of routines were calling calloc and free when they should
 * have been calling CALLOC and FREE.
 *
 * Revision 1.4  1994/02/04  20:13:53  dferrell
 * i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.3  1993/12/02  19:22:28  ddavies
 * i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.2  1993/11/05  01:27:06  ddavies
 * i_aieutl.c -> /product/ipcore/ipshared/src/RCS/i_aieutl.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/

