/*************************************************************
 *  Copyright (c) 1992, 1993 Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 * i_histo.c -- procedures that produce histograms
 *           Int32	i_histoRect()		wrapper for fast or slow version *
 *	     Int32	i_histoSubSampled8()	c-version
 *	     Int32	i_histoSubSampled4()	c-version
 *
 *	     Int32	i_histoRect4()		computes 4 bpp histo of rect
 *
 *	     Int32	i_histoMaskRect()	histogram only stuff in mask
 */


#include <stdio.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "frames.pub"
#include "iaerror.pub"
#include "imageref.pub"
#include "lmask.prv"
#include "rmask.prv"

    /* prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_histo.c,v $");

#define cNibbleMask     0x0f

/*-----------------------------------------------------------------------*
 *                         Grayscale Histogram                           *
 *-----------------------------------------------------------------------*/

/* PUBLIC */
/*************************************************************************
 *  i_histoRect()
 *  		Takes an 8-bit grayscale image, a rectangle and a pointer
 *              to the (empty) histogram. The rectangle is specified by 
 *              its (x, y) offset, width and height.
 *              When complete, histo(i) contains the number of pixels in
 *              the rectangle of the gray image having intensity i.
 *              The histogram is represented as an array of 256 Int32's.
 *
 *	Input:
 *      	pS = 	Pointer to upper left corner of source image's frame
 *      	Bpl =	bytes/line of source including frame.
 *     		x,y =	upper left corner of rectangle relative to upper 
 *               	left corner of *frame*, not image.
 *      	w =  	width of rectangle in pixels
 *     		h =  	height of rectangle in pixels
 *      	histo	pointer to array of 256 Int32's.  Array assumed to 
 *               	be initially filled with 0's
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32   CDECL 
i_histoRect (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	Bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32      *histo)
{

    PROLOGUE;
    pS += (y * Bpl) + x;		/* account for rectangle */

	/* Here's the loop that does all the real work */
    for (; h > 0; h--, pS += Bpl)
    {
	ip_calcHistogramLine (pIPState, pS, w, histo);
    }

    NORMAL_RETURN;
}


/* PUBLIC */
/*************************************************************************
 *  i_histoRect4()
 *  		Takes a 4-bit grayscale image, a rectangle and a pointer
 *              to the (empty) histogram. The rectangle is specified by 
 *              its (x, y) offset, width and height.
 *              When complete, histo(i) contains the number of pixels in
 *              the rectangle of the gray image having intensity i.
 *              The histogram is represented as an array of 16 Int32's.
 *
 *	Input:
 *      	pS = 	Pointer to upper left corner of source image's frame
 *      	Bpl =	bytes/line of source including frame.
 *     		x,y =	upper left corner of rectangle relative to upper 
 *               	left corner of *frame*, not image.
 *      	w =  	width of rectangle in pixels including frame
 *     		h =  	height of rectangle in pixels including frame
 *      	histo	pointer to array of 16 Int32's.  Array assumed to 
 *               	be initially filled with 0's
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32   CDECL 
i_histoRect4 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	Bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32      *histo)
{
PROLOGUE;
 Int32	 index, height;
UInt8	*pSLoop;
Int32	 byteHisto[256] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    pS += (y * Bpl) + x/2;		/* account for rectangle */

	/* the way this routine works is that it calculates a histogram
	 * of the image using bytes.  Each byte represents a pair of
	 * 4 bit pixels.  We then compute the 4-bit histogram by adding
	 * the approprate counts from the byte histogram together.  For
	 * example, say entry number 18 decimal, or 0x12, has a count
	 * of 10035.  This means there were 10035 places in the image
	 * where both nibble 1 and nibble 2 appeared together.  We add
	 * 10035 to the counts for both nibble 1 and nibble 2 in the
	 * output histogram.
	 */
	/* Here's the loop that does all the real work */
    for (pSLoop = pS, height = h; height > 0; pSLoop += Bpl, height--)
    {
	    /* if the left edge of the rectangle started on an odd boundary,
	     * remove the count due to the leftmost pixel.  We do this
	     * now to make the cache work right.
	     */
	if (x & 1)
	    histo[UCHAR_ACCESS(&pSLoop[0]) >> 4]--;

	ip_calcHistogramLine (pIPState, pSLoop, (w + 1)/2, &(byteHisto[0]));

	    /* if the right edge of the rectangle was on an even pixel
	     * boundary, we need to remove the count due to the extra
	     * nibble on the right end of the rightmost byte.
	     * Since we really want to see if (x+w-1) is even, we
	     * test to see if (x+w) is odd.
	     */
	if (((x+w) & 1) != 0)
	    histo[UCHAR_ACCESS(&pSLoop[(w-1)/2]) & 0xF]--;

    }

    for (index = 0; index < 256; index++)
    {
	histo[index >> 4] += byteHisto[index];
	histo[index & 0xF] += byteHisto[index];
    }

    NORMAL_RETURN;
}


/* This is a public routine */
/* PUBLIC */
/*************************************************************************
 *  i_histoSubSampled8(): 
 *              Takes an 8-bit grayscale image and a pointer to the 
 *              the (empty) histogram of intensity values, represented as
 *              an array of 256 ints.
 *              When complete, histo(i) contains the number of pixels in
 *		the gray image having intensity i.
 *              Computes histogram based on subsampled subset of original
 *              data.
 *
 *	pS =	 Pointer to upper left corner of source image's frame
 *	w =	 width of image in pixels including frame
 *	h =	 height of image in pixels including frame
 *	Bpl =	 bytes/line of source including frame.
 *      factor = subsampling factor, 1 = 1:1, 2=1:2, 3=1:4, etc.
 *	histo =	 pointer to array of 256 Int32's.  Array assumed to be initially
 *		 filled with 0's
 *
 *		Note that this version does not count the bytes in the
 *		last partial 32-bit word of a source line if it exists.
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32  CDECL 
i_histoSubSampled8 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    register Int32 factor,
    Int32      *histo)
{
PROLOGUE;
 UInt8	*pSByte, *pSEnd;
Int32	 hDist, scanLineVertDist, sampleVertDist, lineDropDist;
Int32	 hOffset;

    /* psm 11/9/93: don't forget to ignore the frame */
    pS += (Bpl << cLogFrameBits) + 8*cFrameBytes;	/* 8 bits/pixel */
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

	/* point to just below the last line.  this is used by the
	 * scan line loop.
	 */
    pSEnd = pS + h * Bpl;

	/* Originally, we just dropped down one line for each
	 * horizontal move.  This is too steep when we're supposed to
	 * look at a strip that's very wide, but not very tall and
	 * too steep for strips that are very tall and thin.  To
	 * get around these problems, we use separate factors for
	 * how far to move horizontally, how far to drop vertically
	 * after each sample and how far to drop vertically at the
	 * start of each scan lin.  hDist tells us how far to move
	 * horizontally.  scanLineVertDist tells
	 * us how far to move vertically in each horizontal step.
	 * sampleVertDist tells us how far to move vertically at
	 * the start of each scan line.
	 */
    hDist = MIN(factor, w/16);	/* deal with narrow strips */
    hDist = MAX(hDist, 1);

	/* if the image is flat, don't move down very far between lines */
    scanLineVertDist = MAX(1, MIN(factor, h/16)) * Bpl;

	/* If the image is flat and wide, don't move down on every vertical
	 * sample.
	 */
    sampleVertDist = MIN(1, (h/32)/(w/hDist));

	/* This is how far we move in bytes each time we cross the image.
	 */
    sampleVertDist = sampleVertDist * Bpl;
    lineDropDist = ((w/hDist) * sampleVertDist) + w;

	/* To save vars, we actually move across each scan path from
	 * the lower right to the upper left.
	 */
    for (hOffset = hDist; pS + lineDropDist < pSEnd; pS += scanLineVertDist)
    {
        for (pSByte = pS + lineDropDist - hOffset;
	     pSByte >= pS;
	     pSByte -= hDist + sampleVertDist)
        {
            histo[UCHAR_ACCESS(&pSByte[0])]++;
        }
	if (--hOffset == 0)
	    hOffset = hDist;
    }
    
    NORMAL_RETURN;
}

/* This is a public routine */
/* PUBLIC */
/*************************************************************************
 *  i_histoSubSampled4(): 
 *              Takes a 4-bit grayscale image and a pointer to the 
 *              the (empty) histogram of intensity values, represented as
 *              an array of 256 ints.
 *              When complete, histo(i) contains the number of pixels in
 *		the gray image having intensity i.
 *              Computes histogram based on subsampled subset of original
 *              data.
 *
 *	pS =	 Pointer to upper left corner of source image's frame
 *	w =	 width of image in pixels including frame
 *	h =	 height of image in pixels including frame
 *	Bpl =	 bytes/line of source including frame.
 *      factor = subsampling factor, 1 = 1:1, 2=1:2, 3=1:4, etc.
 *	histo =	 pointer to array of 16 Int32's.  Array assumed to be initially
 *		 filled with 0's
 *
 *		Note that this version does not count the bytes in the
 *		last partial 32-bit word of a source line if it exists.
 *
 *	Return code:
 *		ia_successful	good completion
 ************************************************************************/
Int32  CDECL 
i_histoSubSampled4 (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	w,
    Int32	h,
    Int32	Bpl,
    register Int32 factor,
    Int32      *histo)
{
PROLOGUE;
 UInt8	*pSByte, *pSEnd;
Int32	 hDist, scanLineVertDist, sampleVertDist, lineDropDist;
Int32	 hOffset;

    /* psm 11/9/93: don't forget to ignore the frame */
    pS += (Bpl << cLogFrameBits) + 4*cFrameBytes;	/* 4 bits/pixel */
    w -= 2 * cFrameBits;
    h -= 2 * cFrameBits;

	/* point to just below the last line.  this is used by the
	 * scan line loop.
	 */
    pSEnd = pS + h * Bpl;

	/* Originally, we just dropped down one line for each
	 * horizontal move.  This is too steep when we're supposed to
	 * look at a strip that's very wide, but not very tall and
	 * too steep for strips that are very tall and thin.  To
	 * get around these problems, we use separate factors for
	 * how far to move horizontally, how far to drop vertically
	 * after each sample and how far to drop vertically at the
	 * start of each scan lin.  hDist tells us how far to move
	 * horizontally.  scanLineVertDist tells
	 * us how far to move vertically in each horizontal step.
	 * sampleVertDist tells us how far to move vertically at
	 * the start of each scan line.
	 */
    hDist = MIN(factor/2, w/32);	/* deal with narrow strips */
    hDist = MAX(hDist, 1);

	/* if the image is flat, don't move down very far between lines */
    scanLineVertDist = MAX(1, MIN(factor, h/16)) * Bpl;

	/* If the image is flat and wide, don't move down on every vertical
	 * sample.
	 */
    sampleVertDist = MIN(1, (h/32)/(w/hDist));

	/* This is how far we move in bytes each time we cross the image.
	 */
    sampleVertDist = sampleVertDist * Bpl;
    lineDropDist = ((w/(2*hDist)) * sampleVertDist) + w/2;

	/* To save vars, we actually move across each scan path from
	 * the lower right to the upper left.
	 */
    for (hOffset = hDist; pS + lineDropDist < pSEnd; pS += scanLineVertDist)
    {
        for (pSByte = pS + lineDropDist - hOffset;
	     pSByte >= pS;
	     pSByte -= hDist + sampleVertDist)
        {
            histo[UCHAR_ACCESS(&pSByte[0]) & cNibbleMask]++;
        }
	if (--hOffset == 0)
	    hOffset = hDist;
    }

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compute histogram of intensity samples of image pixels covered by a mask.
 *
 * LongDescription:::
 *	Given a gray image and a mask, histogram those parts of the image
 *	which have a corresponding "1" in the mask. *note* that this
 *	function does not clear the histogram; that should be done by
 *	the caller (or not, if the caller is accumulating to the histogram,
 *	for example, when histogramming multiple regions)
 *	Histogram only the area in the given rectangle (and under the 
 *	mask).
 *
 * Input:::
 *  pS::		Source image, pointer to start of frame.
 *  x, y::		upper left corner of rectangle relative to
 *			upper left corner of source image's frame
 *  w::			rectangle width
 *  h::			rectangle height
 *  sBpl::		Bytes between lines in source image
 *  sDepth::		Source image depth, 4 or 8 supported
 *  pM::		Mask image, should be 1 bit, same size as pS,
 *			pointer to frame
 *  mBpl::		Bytes between lines in mask
 *
 * Output:::
 *  pHisto::		The histogram
 *
 * Returns:::
 *  ia_successful or error code
 *
 * FunctionCategory:::
 *  histogram
 *
 * Keywords:::
 *  histogram, gray
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  JCraig
 *
 * EndDocumentation:::
 **************************************************************************/

Int32 CDECL 
i_histoMaskRect (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	sBpl,
    Int32	sDepth,
    UInt8      *pM,
    Int32	mBpl,
    Int32      *pHisto)
{
PROLOGUE;
 Int32	 i, firstWordBits;
UInt32	 firstWordMask, lastWordMask;
Int32    byteHisto[256] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    if ((w < 0) || (h < 0))
	return ia_invalidParm;

    if ((w == 0) || (h == 0))
	NORMAL_RETURN;

    /* Point to the byte at the upper left corner of the rectangle */
    if (sDepth == 4)
        pS += (y * sBpl) + (x / 2);
    else 
	pS += (y * sBpl) + x;
    pM += (y * mBpl) + ((x / 32) * 4);

    firstWordMask = rightMask[(~x) & 0x1F];
    /* is the rectangle fully contained within one word?  if so, we
     * may need to mask some bits off the right as well.
     */
    firstWordBits = 32 - (x & 0x1F);
    if (w <= firstWordBits)
    {
	firstWordMask &= leftMask[(x + w - 1) & 0x1F];
	lastWordMask = 0;
	firstWordBits = w;
    }
    else
	/* does the rectangle end on a word boundary?  if so, the
	 * inner loop of ip_calcHistoMaskLine8 or ip_calcHistoMaskLine4
	 * will deal with this word and there is no last, partial
	 * word to do.  setting lastWordMask to 0, makes this happen.
	 */
	if ((w - firstWordBits) & 0x1F)
            lastWordMask = leftMask[(x + w - 1) & 0x1F];
	else
	    lastWordMask = 0;


    for (i = 0; i < h; i++, pS += sBpl, pM += mBpl)
    {
	if (sDepth == 8)
	    ip_calcHistoMaskLine8 (pIPState, pS,
				   (UInt32 *)pM, x & 0x1F, firstWordMask,
				   firstWordBits, lastWordMask, w, pHisto);
	else	/* sDepth == 4 */
	    ip_calcHistoMaskLine4 (pIPState, pS,
				   (UInt32 *)pM, x & 0x1F, firstWordMask,
				   firstWordBits, lastWordMask, w,
				   pHisto, byteHisto);
    }

    if (sDepth == 4)
	for (i = 0; i < 256; i++)
	{
	    pHisto[i >> 4] += byteHisto[i];
	    pHisto[i & 0xF] += byteHisto[i];
	}

    NORMAL_RETURN;

}
		    
/*
$Log: i_histo.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.7  1998/01/13 23:48:35  ddavies
i_histo.c -> /product/ipcore/ipshared/src/RCS/i_histo.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1997/07/03 21:56:42  jcraig
i_histo.c -> /product/ipcore/ipshared/src/RCS/i_histo.c,v

remove i_histo and i_histo4 because their functionality is
available in i_histoRect and i_histoRect4 respectively.
rename i_histoMask to i_histoMaskRect because it now takes a rectangle
Only those pixels under the mask in that rectangle will be histogrammed.

 * Revision 1.5  1995/12/13  02:44:42  ejaquith
 * i_histo.c -> /product/ipcore/ipshared/src/RCS/i_histo.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.4  1994/07/06  20:08:50  ddavies
 * ddavies on Wed Jul 6 13:08:27 PDT 1994
 *
 * i_histo.c -> /product/ipcore/ipshared/src/RCS/i_histo.c,v
 *
 * Turn 1 into 1UL to keep compiler from complaining about overflow.
 *
 * Revision 1.3  1994/06/16  22:45:53  macleod
 * macleod on Thu Jun 16 15:43:57 PDT 1994
 *
 * i_histo.c -> /product/ipcore/ipshared/src/RCS/i_histo.c,v
 *
 * Added i_histoMask, which takes a gray image (4 or 8 bit) and a binary
 * mask, and histograms only those pixels in the image which have a
 * corresponding bit set in the mask.
 * Note that the histogram is NOT cleared at the start, in order to
 * allow for accumulating to histograms based on multiple calls.
 *
 * Revision 1.2  1994/03/25  22:39:26  ddavies
 * ddavies on Fri Mar 25 14:31:01 PST 1994
 *
 * i_histo.c -> /product/ipcore/ipshared/src/RCS/i_histo.c,v
 *
 * Add implementations for i_histo4 and i_histoRect4.  These routines compute
 * histograms of full 4 bit images and rectangles inside 4 bit images respectively.
 * Also change calls to ip_calcFullHistogram to calls to ip_calcHistogramLine.
 * The new routine calculates the histogram of a single image line instead of
 * doing the entire image.  This is in keeping with the new ipshared philosophy.
 * Delete all the little loops that print out the histogram, we have lots of
 * other ways to do this.
 * Change the implementation of i_histoRect so that it does what the comments say
 * it does, that is x and y are specified relative to the OUTSIDE of the frame,
 * not the inside.  ART- you'll have to put in the "+ cFrameBits" phrases
 * that are currently commented out in your code.
 * Fixed i_histoRect so that it calculates the histogram of exactly the pixels
 * inside the rectangle instead of moving the rectangle over to the nearest
 * word boundary.
 * i_histoRect4 now works the same way.
 * Changed the way the subsampled histogram routines work.  When given a very
 * narrow or very flat image, these routines now do more horizontal samples
 * or more vertical samples respectively.  That is, they modify the subsampling
 * factor horizontally or vertically (or both) to get more samples in images
 * with extreme aspect ratios.  The old code just sampled along diagonals.  When
 * the diagonals didn't fit in the image, it just stopped.
 *
 * Revision 1.1  1994/01/19  19:41:04  macleod
 * Initial revision
 *
*/


