/*************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
 *************************************************************/

/*
 *  i_intops.c	misc. operations that extract integer information
 *
 * Public routines:
 *		Int32	i_getFirstOnPixel()
 *		Int32	i_getLastOnPixel()
 */

#include "types.pub"
#include "ipscore.prv"

#include "iaerror.pub"
#define	_USE_FIRST_BIT_TABLE_		/* want both tables from iopbyte.prv */
#define	_USE_LAST_BIT_TABLE_
#include "iopbyte.prv"

    /* Prototypes */
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_intops.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:11 $");

/* PUBLIC */
/**************************************************************************
 * i_getFirstOnBit()
 *		Given a source word find the first on bit starting
 *		from the left side of the word.  Bits are numbered
 *		from left to right.
 *      
 *      Input:
 *              source = source word in which pixel is to be found
 *		pIndex = pointer to variable which will contain index
 *			 of on pixel
 *
 *	Returns:
 *		ia_successful	if on pixel is found
 *		ia_noOnPixel	if no on pixel is found 
 *
 *      Current implementation assumes 1 bit/pixel.
 *************************************************************************/
Int32   CDECL 
i_getFirstOnBit (
    IPState    *pIPState,
    UInt32	source,
    Int32      *pIndex)
{
PROLOGUE;
 UInt8	currentByte;

    if (source == 0)
    {
	*pIndex = 0; 
	return ia_noOnPixel;
    }

    currentByte = (UInt8)(source >> 24);
    if (currentByte)
    {
	*pIndex = firstBitTable[currentByte] ;
	NORMAL_RETURN;
    }

    currentByte = (UInt8)((source >> 16) & 0xFF);
    if (currentByte)
    {
	*pIndex = firstBitTable[currentByte] + 8;
	NORMAL_RETURN;
    }

    currentByte = (UInt8)((source >> 8) & 0xFF);
    if (currentByte)
    {
	*pIndex = firstBitTable[currentByte] + 16;
	NORMAL_RETURN;
    }

    currentByte = (UInt8)(source & 0xFF);
    if (currentByte)
    {
	*pIndex = firstBitTable[currentByte] + 24;
	NORMAL_RETURN;
    }

	/* should never execute this statement */
    return ia_noOnPixel;
}

/* PUBLIC */
/**************************************************************************
 * i_getLastOnBit()
 *		Given a source word find the last on bit starting
 *		from the left side of the word (or the first on bit
 *		starting from the right side of the word).
 *		Bits are numbered from left to right.
 *      
 *      Input:
 *              source = source word in which pixel is to be found
 *		pIndex = pointer to variable which will contain index
 *			 of on pixel
 *
 *	Returns:
 *		ia_successful	if on pixel is found
 *		ia_noOnPixel	if no on pixel is found 
 *
 *      Current implementation assumes 1 bit/pixel.
 *************************************************************************/
Int32   CDECL 
i_getLastOnBit (
    IPState    *pIPState,
    UInt32	source,
    Int32      *pIndex)
{
PROLOGUE;
 UInt8	currentByte;

    if (source == 0)
    {
	*pIndex = 0; 
	return ia_noOnPixel;
    }

    currentByte = (UInt8)(source & 0xFF);
    if (currentByte)
    {
	*pIndex = lastBitTable[currentByte] + 24;
	NORMAL_RETURN;
    }

    currentByte = (UInt8)((source >> 8) & 0xFF);
    if (currentByte)
    {
	*pIndex = lastBitTable[currentByte] + 16;
	NORMAL_RETURN;
    }

    currentByte = (UInt8)((source >> 16) & 0xFF);
    if (currentByte)
    {
	*pIndex = lastBitTable[currentByte] + 8;
	NORMAL_RETURN;
    }

    currentByte = (UInt8)(source >> 24);
    if (currentByte)
    {
	*pIndex = lastBitTable[currentByte];
	NORMAL_RETURN;
    }

	/* should never execute this statement */
    return ia_noOnPixel;
}

/*

$Log: i_intops.c,v $
Revision 1.1.1.1  1999/09/20 18:18:11  jlin
Imported sources

Revision 1.4  1998/01/13 23:48:38  ddavies
i_intops.c -> /product/ipcore/ipshared/src/RCS/i_intops.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1995/12/13 02:44:45  ejaquith
i_intops.c -> /product/ipcore/ipshared/src/RCS/i_intops.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/13  23:50:52  dferrell
 * i_intops.c -> /product/ipcore/ipshared/src/RCS/i_intops.c,v
 *
 * Fixed type conversion warnings for MS Visual C++ compiler.
 *
 * Revision 1.1  1994/01/22  03:49:40  ddavies
 * Initial revision
 *
 * Revision 1.7  1993/12/02  19:27:15  ddavies
 * sintops.c -> /product/ipcore/ipshared/src/RCS/sintops.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.6  1993/11/05  02:22:44  ddavies
 * sintops.c -> /product/ipcore/ipshared/src/RCS/sintops.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


