/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* 
 *  i_props.c:  Upper level code for copying images.  This is the remnant
 *		of more complex operations on images (pr-ops) that want away.
 *
 *         Int32    i_copyImage()
 */


#include <stdio.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

    /* prototypes */
#include "shrpixr.pub"
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: i_props.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:12 $");

/* PUBLIC */
/*************************************************************
 *  i_copyImage():  returns with source image copied into destination
 *	Input:
 *		sPtr = source image (includes frame)
 *		w = width of the image (including frame)
 *		h = height of the image (including frame)
 *		d = the depth of a pixel
 *		sBpl = bytes/line of the source image
 *		dPtr = destination image (includes frame)
 *		dBpl = bytes/line of the dest image
 *
 *	On return:
 *		Destination image contains copy of source image
 *		The source and destination image bytes/line need not
 *		be multiples of 4.  This is one of the very few operations
 *		for which this is true.
 *
 *	Return code:
 *		ia_successful	successful completion
 *
 *	Calling routine MUST setBorder if appropriate.
 *************************************************************/
Int32  CDECL 
i_copyImage (
    IPState    *pIPState,
    UInt8      *sPtr,
    Int32	w,
    Int32	h,
    Int32	d,
    Int32	sBpl,
    UInt8      *dPtr,
    Int32	dBpl)
{
PROLOGUE;
 Int32		 frameInBytes, imageInBytes, line;
UInt8		*dPtrOrig;

	/* Use an assembly routine originally written to remove frames
         * from pixrects.  It uses double word copies - important on an SS2.
         */

    frameInBytes = (cFrameBytes * d);		/* a multiple of 4! */
    imageInBytes = (((w - 2*cFrameBits) * d) + 7) / 8;	/* not necessarily a
						 * multiple of 4 */
    sPtr += (sBpl << cLogFrameBits) + frameInBytes;
    dPtr += (dBpl << cLogFrameBits) + frameInBytes;
    dPtrOrig = dPtr;

	/* copy each line up into position starting with the topmost image line.
	 * The starting position for the source accounts for the number
	 * of upper pad lines (cFrameBits) and the width of the left frame
	 * (frameInBytes).  The assembly routine "ip_removeFrameFromLine" just
	 * copies the memory it's told to copy - it really has little to do
	 * with frames.  It does NOT assume that each line is a multiple of 4
	 * bytes in length or that the source and dest pointers point to
	 * word boundaries (unlike almost all other ipshared primitives).
	 */
    for (line = h - 2*cFrameBits;
	 line > 0;
	 sPtr += sBpl, dPtr += dBpl, line--)
    {
	ip_removeFrameFromLine (pIPState, dPtr, sPtr, imageInBytes);
    }

    ip_clearRightFrame (pIPState, dPtrOrig,
			w-(2*cFrameBits), h-(2*cFrameBits), dBpl, d);
    NORMAL_RETURN;

}

/*

$Log: i_props.c,v $
Revision 1.1.1.1  1999/09/20 18:18:12  jlin
Imported sources

Revision 1.3  1998/01/13 23:49:16  ddavies
i_props.c -> /product/ipcore/ipshared/src/RCS/i_props.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1995/12/13 02:45:20  ejaquith
i_props.c -> /product/ipcore/ipshared/src/RCS/i_props.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1994/01/21  21:13:22  ddavies
 * Initial revision
 *
 * Revision 1.11  1993/12/02  19:21:32  ddavies
 * fprops.c -> /product/ipcore/ipshared/src/RCS/fprops.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.10  1993/11/05  01:25:20  ddavies
 * fprops.c -> /product/ipcore/ipshared/src/RCS/fprops.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


