/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  i_utils.c:
 *      Includes
 *
 *           IPState init/accessor/mutator functions
 *               i_IPStateGetSize()
 *               i_IPStateInit ()
 *               i_IPStateSetAllocator ()
 *               i_IPStateSetErrorProc ()
 *               i_IPStateGetErrorProc ()
 *               i_IPStateSetStrelInfo ()
 *               i_IPStateGetStrelInfo ()
 *		 ip_IPStateSetLongAccessFlag ()
 *		 ip_IPStateGetLongAccessFlag ()
 *		 ip_IPStateSetPurifyRunning ()
 *		 ip_IPStateGetPurifyRunning ()
 *
 *		i_printErrorInfo()
 *		i_printf()
 *
 */

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdarg.h>
#include <stdio.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "types.pub"
#include "ipscore.prv"
#include "ipserrs.pub"
#include "iaerror.pub"
#include "defines.pub"

#include "shrutils.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: i_utils.c,v $; $Revision: 1.2 $; $Date: 2000/03/24 22:12:03 $");

/*-------------- Default allocators: just call malloc, etc. --------------*/

#ifdef DEBUG_HEAP
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Default memory allocation procedure, std lib malloc() call
 *
 * Input:::
 *  count::        	number of bytes of memory to allocate  
 *  hasPointers::    	will this memory be used to contain pointers?
 *  thing::          	client info pointer
 *  file::           	only used if DEBUG_HEAP is defined.  name of file
 *			that called routine
 *  line::           	only used if DEBUG_HEAP is defined.  line in file
 *			from which routine is called
 *
 * Returns:::
 *  pointer to allocated memory
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
static void *  CDECL
ip_defaultMalloc (
    size_t              count,
    Bool                hasPointers,
    void               *thing,
    char               *file,
    Int32               line)
{
    thing = thing; file = file; line = line; hasPointers = hasPointers;
    return malloc(count);
}		 
#else /*DEBUG_HEAP*/
static void * CDECL
ip_defaultMalloc (
    size_t              count,
    Bool                hasPointers,
    void               *thing)
{
    thing = thing;
    hasPointers = hasPointers;
    return malloc(count);
}		 
#endif /*DEBUG_HEAP*/


#ifdef DEBUG_HEAP
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Default memory allocation procedure, std lib calloc()
 *
 * Input:::
 *  nel::            	number of elements to allocate
 *  count::        	number of bytes of each element to allocate  
 *  hasPointers::    	will this memory be used to contain pointers?
 *  thing::          	client info pointer
 *  file::           	only used if DEBUG_HEAP is defined.  name of file
 *			that called routine
 *  line::           	only used if DEBUG_HEAP is defined.  line in file
 *			from which routine is called
 *
 * Returns:::
 *  pointer to memory
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
static void *  CDECL
ip_defaultCalloc (
    size_t              nel,
    size_t              count,
    Bool                hasPointers,
    void               *thing,
    char               *file,
    Int32               line)
{
    thing = thing; file = file; line = line; hasPointers = hasPointers;
    return calloc (nel, count);
}
#else /*DEBUG_HEAP*/
static void * CDECL
ip_defaultCalloc (
    size_t              nel,
    size_t              count,
    Bool                hasPointers,
    void               *thing)
{
    thing = thing;
    hasPointers = hasPointers;
    return calloc (nel, count);
}
#endif /*DEBUG_HEAP*/


#ifdef DEBUG_HEAP
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Default memory allocation procedure, std lib realloc()
 *
 * Input:::
 *  ptr::           	original pointer to memory
 *  size::           	new size (in bytes) of block
 *  hasPointers::    	will this memory be used to contain pointers?
 *  thing::          	client info pointer
 *  file::           	only used if DEBUG_HEAP is defined.  name of file
 *			that called routine
 *  line::           	only used if DEBUG_HEAP is defined.  line in file
 *			from which routine is called
 *
 * Returns:::
 * pointer to memory
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
static void *  CDECL
ip_defaultRealloc (
    void               *ptr,
    size_t              size,
    Bool                hasPointers,
    void               *thing,
    char               *file,
    Int32               line)
{
    thing = thing; file = file; line = line; hasPointers = hasPointers;
    return realloc (ptr, size);
}
#else /*DEBUG_HEAP*/
static void * CDECL
ip_defaultRealloc (
    void               *ptr,
    size_t              size,
    Bool                hasPointers,
    void               *thing)
{
    thing = thing; 
    hasPointers = hasPointers;
    return realloc (ptr, size);
}
#endif /*DEBUG_HEAP*/


#ifdef DEBUG_HEAP
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Default memory freeing procedure, std lib free()
 *
 * Input:::
 *  ptr::       	pointer to memory to be free'd     
 *  size::          	number of bytes pointed to by ptr
 *  thing::          	client info pointer
 *  file::           	only used if DEBUG_HEAP is defined.  name of file
 *			that called routine
 *  line::           	only used if DEBUG_HEAP is defined.  line in file
 *			from which routine is called
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
static void CDECL
ip_defaultFree (
    void               *ptr,
    size_t              size,
    void               *thing,
    char               *file,
    Int32               line)
{
    thing = thing; file = file; line = line; size = size;
    free(ptr, size);
}
#else /*DEBUG_HEAP*/
static void CDECL
ip_defaultFree (
    void               *ptr,
    size_t              size,
    void               *thing)
{
    thing = thing;
    size = size;
    free(ptr);
}
#endif /*DEBUG_HEAP*/




/*-------------------------IPState functions------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *	Return size of IPState structure.
 *
 * Output:::
 *  pIPStateSize::     	pointer to size of IPState state structure
 *
 * Returns:::
 *  cSuccessful		successfully returned size of IPState
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_getSizeOfIPState (
    Int32      *pIPStateSize)
{
    PROLOGUE;

	/* Make compiler happy */
    __error__ = __error__;

	/* The IPState structure has a fixed size */
    *pIPStateSize = sizeof (IPState);

    return ia_successful;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Initialize a pre-allocated IPState structure.
 *
 * Input:::
 *  pIPState::     	pointer to IPState structure to be filled in
 *  imageMalloc::    	proc called when allocating images.
 *                      If NULL, default to system call via ip_defaultMalloc().
 *  imageCalloc::    	proc called when allocating and clearing images
 *                      If NULL, default to system call via ip_defaultCalloc().
 *  imageRealloc::   	proc called when reallocating images
 *                      If NULL, default to system call via ip_defaultRealloc().
 *  imageFree::      	proc called when deallocating images
 *                      If NULL, default to system call via ip_defaultFree().
 *  imageInfo::      	client ptr given to memory routines when allocating
 *			images
 *  dataMalloc::     	proc called when allocating data
 *                      If NULL, default to system call via ip_defaultMalloc().
 *  dataCalloc::     	proc called when allocating and clearing data
 *                      If NULL, default to system call via ip_defaultCalloc().
 *  dataRealloc::    	proc called when reallocating data
 *                      If NULL, default to system call via ip_defaultRealloc()
 *  dataFree::       	proc called when deallocating for data
 *                      If NULL, default to system call via ip_defaultFree().
 *  dataInfo::       	client ptr given to memory routines when allocating
 *			data
 *  errorProc::    	client procedure called to print error messages.
 *			The type of this procedure is defined in
 *			ipshared/include/types.pub.
 *  errorInfo::		client ptr that's included in every call
 *			to errorProc.
 *  maxFilter::		A value in the IPState structure is decremented each
 *			time ERR_PASS or ERR_PASS_CLEANUP is executed.  When
 *			the value reaches 0, it is reset to maxFilter and
 *			yieldProc is called.  This allows the client to
 *			abort processes without knowing about the OS.
 *  yieldProc::		The procedure called to yield control to the client.
 *			This takes pointer to void as its parameter.  The
 *			pointer is set to yieldInfo (below)
 *			as a means of identifying the executing process.
 *			If yieldProc == NULL, it will never be called.
 *  yieldInfo::		This client-supplied data is passed back by the
 *			yieldProc.
 *  ipStateSize::	number of bytes in structure being passed in.
 *
 * Returns:::
 *  cSuccessful		successfully initialized IPState
 *
 * Notes:::
 * If any of the allocation/deallocation routines are null,
 * sets them to a hopefully reasonable default.
 * However,  errorStream or *Info the are left alone if null.
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, jcraig, ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateInit (
    IPState            *pIPState,
    MallocType          imageMalloc,
    CallocType          imageCalloc,
    ReallocType         imageRealloc,
    FreeType            imageFree,
    void               *imageInfo,
    MallocType          dataMalloc,
    CallocType          dataCalloc,
    ReallocType         dataRealloc,
    FreeType            dataFree,
    void               *dataInfo,
    ClientErrorProc     errorProc,
    void	       *errorInfo,
    UInt32		maxFilter,
    ClientYieldProc	yieldProc,
    void	       *yieldInfo,
    Int32		ipStateSize)
{
    PROLOGUE;
    MallocType	 mallocProc;
    CallocType	 callocProc;
    ReallocType	 reallocProc;
    FreeType	 freeProc;
    Int32	 realIPStateSize;

    /* make sure that the client really handed us a pointer. */
    ERR_GEN(pIPState == NULL, cErrNullIPState);

    /* Make sure it's big enough */
    ERR_PASS (i_getSizeOfIPState(&realIPStateSize));
    ERR_GEN (ipStateSize < realIPStateSize, cErrInvalidParm);

    /* first the image allocators */
    mallocProc = (imageMalloc == NULL) ? ip_defaultMalloc : imageMalloc;
    callocProc = (imageCalloc == NULL) ? ip_defaultCalloc : imageCalloc;
    freeProc = (imageFree == NULL) ? ip_defaultFree : imageFree;
    reallocProc = (imageRealloc == NULL) ? ip_defaultRealloc : imageRealloc;
    ERR_PASS( i_IPStateSetAllocator(pIPState, cAllocatorImage,
				    mallocProc, callocProc,
				    reallocProc, freeProc, imageInfo));

    /* now the data allocators */
    mallocProc = (dataMalloc == NULL) ? ip_defaultMalloc : dataMalloc;
    callocProc = (dataCalloc == NULL) ? ip_defaultCalloc : dataCalloc;
    freeProc = (dataFree == NULL) ? ip_defaultFree : dataFree;
    reallocProc = (dataRealloc == NULL) ? ip_defaultRealloc : dataRealloc;
    ERR_PASS( i_IPStateSetAllocator(pIPState, cAllocatorData, 
				    mallocProc, callocProc,
				    reallocProc, freeProc, dataInfo));

	/* We don't know what to do with the strel info thing at
	 * this level, so just clear it.
	 */
    pIPState->strelInfo = NULL;

    /* now the error stream.  note it can be null */
    ERR_PASS( i_IPStateSetErrorProc (pIPState, errorProc, errorInfo));

    /* now the error stream.  note it can be null */
    ERR_PASS( i_IPStateSetYieldProc (pIPState, maxFilter,
				     yieldProc, yieldInfo));

    /* Assume long accesses are allowed.  If not, the client will fix
     * it later.
     */
    ERR_PASS( ip_IPStateSetLongAccessFlag (pIPState, cTrue));

    /* Now the purifyRunning flag.  Assume it's not running.  Regression
     * test code can fix this later.  It's a secret hack.
     */
    ERR_PASS( ip_IPStateSetPurifyIsRunning (pIPState, cFalse));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets allocation procedures in pIPState
 *
 * LongDescription:::
 * Sets the malloc, calloc, realloc, and free functions for
 * an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *  allocator::      	determines whether these allocators for data or images
 *  pMalloc::        	proc called when allocating memory
 *  pCalloc::         	proc called when allocating and clearing memory
 *  pRealloc::       	proc called when reallocating memory
 *  pFree::          	proc called when deallocating memory
 *  allocInfo::      	client ptr given to memory routines
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateSetAllocator (
    IPState            *pIPState,
    Int32               allocator,
    MallocType          pMalloc,
    CallocType          pCalloc,
    ReallocType         pRealloc,
    FreeType            pFree,
    void               *allocInfo)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    ERR_GEN((allocator >= cAllocatorCount) || (allocator < 0),
	    cErrInvalidParm);

    pIPState->mallocVec[allocator] = pMalloc;
    pIPState->callocVec[allocator] = pCalloc;
    pIPState->reallocVec[allocator] = pRealloc;
    pIPState->freeVec[allocator] = pFree;
    pIPState->dataVec[allocator] = allocInfo;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the error stream in an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *  errorProc::		Pointer to procedure used to print error messages.
 *  errorInfo::		Pointer to client data included in every call
 *			to errorProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateSetErrorProc (
    IPState            *pIPState,
    ClientErrorProc     errorProc,
    void	       *errorInfo)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    pIPState->errorProc = errorProc;
    pIPState->errorInfo = errorInfo;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the error stream from an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *
 * Output:::
 *  pErrorProc::	Pointer to pointer to procedure used to print
 *			error messages.
 *  pErrorInfo::	Pointer to pointer to client data included in
 *			every call to errorProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateGetErrorProc (
    IPState            *pIPState,
    ClientErrorProc    *pErrorProc,
    void	      **pErrorInfo)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    ERR_GEN (pErrorProc == NULL || pErrorInfo == NULL,
	     cErrNullPtr);
    *pErrorProc = pIPState->errorProc;
    *pErrorInfo = pIPState->errorInfo;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the yield proc and filter in an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *  maxFilter::		Value specifying how often yieldProc is called.
 *			yieldProc is called once for every maxFilter
 *			times ERR_PASS or ERR_PASS_CLEANUP is executed.
 *  yieldProc::		Pointer to procedure used to yield control back to
 *			the client.
 *  yieldInfo::		pointer to client info that is passed back as an
 *			argument to yieldProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateSetYieldProc (
    IPState            *pIPState,
    Int32		maxFilter,
    ClientYieldProc	yieldProc,
    void	       *yieldInfo)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    pIPState->yieldProc = yieldProc;
    pIPState->yieldInfo = yieldInfo;
    pIPState->maxFilter = maxFilter;
    pIPState->curFilter = maxFilter;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the yield procedure and max filter value from an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *
 * Output:::
 *  pMaxFilter::	Pointer to returned maximum filter value.
 *			The yieldProc is called every maxFilter times
 *			ERR_PASS or ERR_PASS_CLEANUP is invoked.
 *  pYieldProc::	Pointer to pointer to yield procedure.
 *  pYieldInfo::	Pointer to client info that is passed back as
 *			an argument to yieldProc.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod, ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateGetYieldProc (
    IPState            *pIPState,
    Int32	       *pMaxFilter,
    ClientYieldProc    *pYieldProc,
    void	      **pYieldInfo)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    ERR_GEN (pYieldProc == NULL || pMaxFilter == NULL,
	     cErrNullPtr);
    *pYieldProc = pIPState->yieldProc;
    *pYieldInfo = pIPState->yieldInfo;
    *pMaxFilter = pIPState->maxFilter;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Set the StrelInfo structure for the IPState structure.
 *
 * LongDescription:::
 *  The IPShared level routines have no idea what this is, but we
 *  can set the pointer here for the sake of completeness.
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *  pStrelInfo::     	pointer to StrelInfo structure
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateSetStrelInfo (
    IPState            *pIPState,
    StrelInfoPtr        pStrelInfo)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    pIPState->strelInfo = pStrelInfo;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the StrelInfo structure for the IPState structure.
 *
 * LongDescription:::
 *  The IPShared level routines have no idea what this is, but we
 *  can return the pointer here for the sake of completeness.
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *
 * Output:::
 *  pStrelInfo::     	pointer to pointer to StrelInfo structure
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
i_IPStateGetStrelInfo (
    IPState            *pIPState,
    StrelInfoPtr       *pStrelInfo)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    *pStrelInfo = pIPState->strelInfo;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the enableLongAccesses flag in an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *  enableLongAccesses:: Boolean set to cTrue if long (>= 64 bit) accesses
 *			 should be enabled.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
ip_IPStateSetLongAccessFlag (
    IPState            *pIPState,
    Bool		enableLongAccesses)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    pIPState->enableLongAccesses = enableLongAccesses;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the enableLongAccesses flag from an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *
 * Output:::
 *  pEnableLongAccesses:: Pointer to returned Boolean that will hold the
 *			  contents of the enableLongAccesses flag.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
ip_IPStateGetLongAccessFlag (
    IPState            *pIPState,
    Bool	       *pEnableLongAccesses)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    ERR_GEN (pEnableLongAccesses == NULL,
	     cErrNullPtr);
    *pEnableLongAccesses = pIPState->enableLongAccesses;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Sets the purifyRunning flag in an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *  purifyIsRunning::	Boolean set to cTrue if Purify is running, cFalse
 *			otherwise.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
ip_IPStateSetPurifyIsRunning (
    IPState            *pIPState,
    Bool		purifyIsRunning)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    pIPState->purifyIsRunning = purifyIsRunning;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get the purifyIsRunning flag from an IPState structure
 *
 * Input:::
 *  pIPState::      	pointer to IPCore state structure
 *
 * Output:::
 *  pPurifyIsRunning::	Pointer to returned Boolean that will hold the
 *			contents of the purifyIsRunning flag.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
ip_IPStateGetPurifyIsRunning (
    IPState            *pIPState,
    Bool	       *pPurifyIsRunning)
{
    PROLOGUE;

    ERR_GEN (pIPState == NULL, cErrNullIPState);
    ERR_GEN (pPurifyIsRunning == NULL,
	     cErrNullPtr);
    *pPurifyIsRunning = pIPState->purifyIsRunning;

    NORMAL_RETURN;
}




/* ------------------------ Printing Utilities --------------------------*/

/* i_printErrorInfo()
 * Called by error macros when an error has occurred.  Currently does
 * nothing but does provide a convenient place to set breakspoints to
 * catch all error returns from a procedure.
 * 
 * Input parameters:
 *	filename	name of file in which error occurred
 *	line_num	line number in which error occurred
 *	error		error code
 *
 * Returns
 *	ia_successful
 */
Int32 CDECL 
i_printErrorInfo (
    IPState    *pIPState,
    const char *filename,
    Int32	line_num,
    ErrCode	error)
{
    
    PROLOGUE;
    filename = filename;
    line_num = line_num;
    error = error;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Print formatted error message using client's error printing procedure
 *
 * LongDescription:::
 * Print a formatted error string using a client's error printing procedure.
 * We should use this function instead of fprintf or printf.
 *
 * Input:::
 *  errorProc::      	Procedure used to print the error string.
 *			For example, this could be fprintf.
 *  errorInfo::		Pointer to private info used by errorProc.
 *			For example, this could be stderr.
 *  format::       	Pointer to the standard formatting string.
 *  ...::       	optional args required by formatting string.
 *
 * Notes:::
 *  Why does this routine exist?  Why not just print to stderr or stdio?
 *  On Windows and Win95 platforms, printing to stderr or stdio does
 *  nothing.  After a spirited debate between Win95 proponents and ANSI C
 *  purists, we bowed to the inevitable.  Those platforms
 *  do have special debugging functions that can display a string in
 *  special debugging windows.  Therefore, all this function does is
 *  format the string and call whatever string-displaying function
 *  the client provides.
 *
 * Returns:::
 *  cSuccessful		successfully printed error message
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * Citations:::
 *  See page 174 of "The C Programming Language, Second Edition" for an
 *  illustration of vsprintf.
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_printf (
    IPState    *pIPState,
    ClientErrorProc errorProc,
    void       *errorInfo,
    const char *format,
    		...)
{
    PROLOGUE;
    char	errorString[256];
    va_list	args;

	/* To be a good little C function, this routine should
	 * consume all its args before returning, whether or not
	 * errorProc is NULL. Therefore, we fill errorString whether
	 * or not errorProc is NULL.  If errorProc is not NULL,
	 * we also print the string.
	 * Use va_start to initialize the args pointer.  Then
	 * use vsprintf to form the string from the formatted
	 * string and the optional remaining args.
	 */
    va_start(args, format);
    vsprintf(&(errorString[0]), format, args);
    va_end(args);

    if (errorProc != NULL)
    {
	errorProc(errorInfo, &(errorString[0]));
    }

    NORMAL_RETURN;
}


/*
 $Log: i_utils.c,v $
 Revision 1.2  2000/03/24 22:12:03  dtracy
 VxWorks-friendly modifications.  New targets 'sparc_vxworks' and 'debugsparc_vxworks' added to src Makefile.

 Revision 1.1.1.1  1999/09/20 18:18:13  jlin
 Imported sources

 Revision 1.7  1998/01/21 21:45:13  ddavies
 i_utils.c -> /product/ipcore/ipshared/src/RCS/i_utils.c,v

 Add IPState initialization and allocation routines iGetSizeOfIPState,
 i_IPStateInit, i_IPStateSetAllocator, i_IPSetErrorProc, i_IPStateGetErrorProc,
 i_IPStateSetYieldProc, i_IPStateGetYieldProc, i_IPStateSetStrelInfo,
 i_IPStateGetStrelInfo, ip_IPStateSetLongAccessFlag, ip_IPStateGetLongAccessFlag,
 ip_IPStateSetPurifyIsRunning, ip_IPStateGetPurifyIsRunning.
 The Strel routines don't know anything about strels, they just access
 the pointer in the IPState structure.
 These were all added because IPState is now an IPShared thing.  Therefore,
 the accessors must be in IPShared.

 Revision 1.6  1998/01/13 23:50:27  ddavies
 i_utils.c -> /product/ipcore/ipshared/src/RCS/i_utils.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.5  1996/04/17 23:25:27  ejaquith
 i_utils.c -> /product/ipcore/ipshared/src/RCS/i_utils.c,v

 Fixed auto-doc stuff that is lingering there for demo of IPShared routines.

 Revision 1.4  1996/02/26 19:42:03  ddavies
 ddavies on Mon Feb 26 11:40:55 PST 1996

 i_utils.c -> /product/ipcore/ipshared/src/RCS/i_utils.c,v

 Fix bug in i_printf.  It now reads all the variable arguments and ends
 the vararg list with va_end whether or not errorProc is used.  This makes
 it a big tidier.

 Revision 1.3  1995/12/13 02:46:44  ejaquith
 i_utils.c -> /product/ipcore/ipshared/src/RCS/i_utils.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.2  1995/10/07  02:38:28  ddavies
 * ddavies on Fri Oct 6 19:35:56 PDT 1995
 *
 * i_utils.c -> /product/ipcore/ipshared/src/RCS/i_utils.c,v
 *
 * Add implementation for i_printf.  This routine should be used instead of
 * fprintf for printing ipwrap stuff and on the rare occasions when someone
 * is trying to debug something in ipshared with print statements.  It
 * uses vsprintf to produce a string and allows a client-supplied procedure
 * to display the string somewhere, somehow.  In a sane world, fprintf
 * would work, but Windows3.1, Win95 and NT require special care.
 *
 * Revision 1.1  1995/08/17  22:19:32  jcraig
 * Initial revision
 *
 */

