/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	wctrans.c:	
 *
 *  Public routine:
 *      i_wc_biwaveletTransform()  
 *      i_wc_biwaveletTrans_LUT()
 *      i_wc_biwaveletTrans_Lift()
 *      i_wc_init_filtertbl()
 *      i_wc_waveArithInit()
 *      i_wc_dataArithInit()
 *      i_wc_waveArithCoding()
 *      i_wc_dataArithTerm()
 *      i_wc_waveArithTerm() 
 *      i_wc_getArithCodelen()
 *      i_wc_waveHuffInit()
 *      i_wc_dc_huffCoding()
 *      i_wc_ac_huffCoding()
 *      i_wc_set_huffTable()
 *      i_wc_emit_huffcode()
 *      i_wc_image2buffer()
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

   /* prototypes */
#include "wavelib.pub"

#include "wavelib.prv"

IP_RCSINFO(i_wcapi_c_RCSInfo, "$RCSfile: i_wcapi.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set up filter look up table.
 *
 * LongDescription:::
 *  filter parameters have been scale 256*128. Max image pixel is 255,
 *  after one dimention transform, Max coef is about 255*1.5. We can
 *  build look up table to reduce multiplications.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *  f_lowpass::      Ptr to low pass filter
 * f_highpass::      Ptr to high pass filter
 *     lowlen::      low pass filter length
 *    highlen::      high pass filter length
 *     tbllen::      lookup table length
 *
 * Output:::
 *   lowTab::        Ptr to low pass look up table address
 *  highTab::        Ptr to high pass look up table address
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_init_filtertbl (
    IPState    *pIPState,
    Int32      *f_lowpass,
    Int32      *f_highpass,
    Int32	lowlen,
    Int32	highlen,
    Int32	tbllen,
    Int32     **lowTab,
    Int32     **highTab)
{
    PROLOGUE;

    ERR_PASS(ip_init_filtertbl (pIPState,  f_lowpass, f_highpass, lowlen,
				highlen, tbllen, lowTab, highTab ));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    n level Biorthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *  f_lowpass::     Ptr to lowpass filter
 * f_highpass::     Ptr to highpass filter
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_biwaveletTransform (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32      *f_lowpass,
    Int32      *f_highpass,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;

    ERR_PASS(ip_waveBifilterTrans (pIPState, coefw, coefh, bpl, 
				   f_lowpass, f_highpass, tempbuf, coefbuf)); 
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    n level Biorthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    options::     Select different filters 
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_biwaveletTrans_Lift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32       options,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;

    if( options & ORD4FILTER )
	ERR_PASS(ip_waveD4filterLift (pIPState, coefw, coefh, bpl, 
				      tempbuf, coefbuf)); 
    else if( options & BI44FILTER )
	ERR_PASS(ip_waveBi44filterLift (pIPState, coefw, coefh, bpl, 
				      tempbuf, coefbuf)); 
    else if( options & TT2AFILTER )
	ERR_PASS(ip_waveTTfilter (pIPState, coefw, coefh, bpl, 
				      tempbuf, coefbuf));
    else if( options & D16FILTER )
	ERR_PASS(ip_waveD16filter (pIPState, coefw, coefh, bpl, 
				      tempbuf, coefbuf)); 
    else	
	ERR_PASS(ip_waveBifilterLift (pIPState, coefw, coefh, bpl, 
				      tempbuf, coefbuf)); 
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    n level Biorthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *  f_lowpass::     Ptr to lowpass filter LUT
 * f_highpass::     Ptr to highpass filter LUT
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_biwaveletTrans_LUT (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32     **f_lowpass,
    Int32     **f_highpass,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;

    ERR_PASS(ip_waveTrans_LUT (pIPState, coefw, coefh, bpl, 
			       f_lowpass, f_highpass, tempbuf, coefbuf)); 
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding init.
 *
 * LongDescription:::
 *  Arithmetic init. must be called before coding.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *         pdata::   Ptr to data used in waveCallbackFunc
 *    outputFunc::   Ptr to call back function (write out compressed data)
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_waveArithInit (
    IPState    *pIPState,
    void       *pdata,
    waveCallbackFunc outputFunc)
{
    PROLOGUE;

    ERR_PASS(ip_waveArithInit (pIPState, pdata, outputFunc));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding data init.
 *
 * LongDescription:::
 *  Arithmetic coding can support multi coding status. For every data set
 *  you have to call data init to create coding infomation stucture.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *       infoIdx::     Coding data set index
 *  no_of_symbol::     Number of symbol of coding data set
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_dataArithInit (
    IPState    *pIPState,
    Int32	infoIdx,
    Int32	no_of_symbol)
{
    PROLOGUE;

    ERR_PASS(ip_dataArithInit (pIPState, infoIdx, no_of_symbol));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding.
 *
 * LongDescription:::
 *  Arithmetic coding.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *    index::     Current symbol's index
 *  bitCode::     Fix bits coding or ataptive arithmetic coding.
 *  infoIdx::     Coding data set index
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_waveArithCoding (
    IPState    *pIPState,
    Int16	index,
    Int16	bitCode,
    Int32	infoIdx) 
{
    PROLOGUE;

    ERR_PASS(ip_waveArithCoding (pIPState, index, bitCode, infoIdx));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding data term.
 *
 * LongDescription:::
 *  Term one data set's arothmetic coding info and free memory. 
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *   infoIdx::        Data set's index
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_dataArithTerm (
    IPState    *pIPState,
    Int32	infoIdx)
{
    PROLOGUE;

    ERR_PASS(ip_dataArithTerm (pIPState, infoIdx));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding term.
 *
 * LongDescription:::
 *  Arithmetic coding term.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *   flag::       data, may be used in later version
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_waveArithTerm (
    IPState    *pIPState,
    Int32	flag)
{
    PROLOGUE;

    ERR_PASS(ip_waveArithTerm (pIPState, flag));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get compressed code length.
 *
 * LongDescription:::
 *  Get arithmetic code length.
 *
 * Input:::
 *
 * Output:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *    codeLen::       Ptr to codeLen
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_getArithCodelen (
    IPState    *pIPState,
    Int32      *codeLen)
{
    PROLOGUE;

    ERR_PASS(ip_getArithCodelen (pIPState, codeLen));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding init.
 *
 * LongDescription:::
 *  Create huffman coding table for DC and AC ( copy from jpeg ) and init.
 * coding coef. 
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      pdata::      Ptr to data used in IO callback func
 * outputFunc::      Ptr to callback function
 *   huffInfo::      Ptr to huffman coding structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_waveHuffInit (
    IPState    *pIPState,
    void       *pdata,
    waveCallbackFunc outputFunc,
    WHUFFINFO  *huffInfo)
{
    PROLOGUE;

    ERR_PASS(ip_waveHuffInit (pIPState, pdata, outputFunc, huffInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Emit remain zero len or emit remain bits.
 *
 * LongDescription:::
 *   Emit remain zero len or emit remain bits.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *    flag::      1: emit zero len, 0: emit remain bits 
 *  phuffInfo::      Ptr to huffman coding structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_clearZerolen (
    IPState    *pIPState,
    Int32	flag,
    WHUFFINFO  *phuffInfo)
{
  PROLOGUE;

  ERR_PASS(ip_clearZerolen (pIPState, flag, phuffInfo));

  NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding for dc.
 *
 * LongDescription:::
 *  Do dc huffman coding.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *     symbol::      Data to be coded 
 *  phuffInfo::      Ptr to huffman coding structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_dc_huffCoding (
    IPState    *pIPState,
    Int16	symbol,
    WHUFFINFO  *phuffInfo) 
{
    PROLOGUE;

    ERR_PASS(ip_dc_huffCoding (pIPState, symbol, phuffInfo));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding for ac.
 *
 * LongDescription:::
 *  Do ac huffman coding.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *     symbol::      Data to be coded 
 *  phuffInfo::      Ptr to huffman coding structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_ac_huffCoding (
    IPState    *pIPState,
    Int16	symbol,
    WHUFFINFO  *phuffInfo) 
{
    PROLOGUE;

    ERR_PASS(ip_ac_huffCoding (pIPState, symbol, phuffInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set huffman table for current coding.
 *
 * LongDescription:::
 *  Huffman table for DC, AC , luminent and color are different. Before
 *  coding, we have to set the right table for current coding.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *       chan::      Channel number 
 *     dcFlag::      DC flag ( 1: dc huffman table )
 *  phuffInfo::      Ptr to huffman coding structure
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_set_huffTable (
    IPState    *pIPState,
    Int32	chan,
    Int32	dcFlag,
    WHUFFINFO  *phuffInfo)
{
    PROLOGUE;

    ERR_PASS(ip_set_huffTable (pIPState, chan, dcFlag, phuffInfo));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Copy 8bit/per pixel image to 16bit/per coef. coef-buffer.
 *
 * LongDescription:::
 *    Copy 8bit/per pixel image to 16bit/per coef. coef-buffer.
 *    Directly copy gray scale image. RGB color image will be converted 
 *    into YCrCb color image.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *    ncomps::      Number of image planets
 *       bpl::      Bytes per line of image
 *         x::      Begining x position
 *         y::      Begining line
 *         w::      Width of copying image
 *         h::      Height of copying image
 *  imagebuf::      Ptr to image buffer
 *
 * Output:::
 *  coefbuf::        Ptr to coef-buffer
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wc_image2buffer (
    IPState    *pIPState,
    Int32	ncomps,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8     **imagebuf,
    Int16     **coefbuf)
{
    PROLOGUE;

    ERR_PASS( ip_image2buffer (pIPState,  ncomps, bpl, x, y, w, h, imagebuf,
			       coefbuf) );
    NORMAL_RETURN;
}


/*
$Log: i_wcapi.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.5  1998/11/30 19:24:08  xzhu
xzhu on Mon Nov 30 11:23:19 PST 1998

i_wcapi.c -> /product/ipcore/ipshared/src/RCS/i_wcapi.c,v

Using options to select different methods for transform.

Revision 1.4  1998/10/20 19:32:26  xzhu
xzhu on Tue Oct 20 12:31:54 PDT 1998

i_wcapi.c -> /product/ipcore/ipshared/src/RCS/i_wcapi.c,v

Added new wrap function for new transform algorithm.

Revision 1.3  1998/01/13 23:50:31  ddavies
i_wcapi.c -> /product/ipcore/ipshared/src/RCS/i_wcapi.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1997/09/17 23:30:05  xzhu
xzhu on Wed Sep 17 16:30:04 PDT 1997

i_wcapi.c -> /product/ipcore/ipshared/src/RCS/i_wcapi.c,v

CommonCheck: Huff-info will be a passed in variable and not a global variable. So we can get huffman coding infomation directly from Huff-info.

Revision 1.1  1996/10/16 00:02:27  xzhu
Initial revision

*/
