/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	wctrans.c:	
 *
 *  Public routine:
 *      i_wd_waveletInvertTrans()  
 *      i_wd_waveletInvertTrans_LUT()
 *      i_wd_waveletInvertTrans_Lift()
 *      i_wd_init_filtertbl() 
 *      i_wd_arithDecodingInit()
 *      i_wd_dataDecodingInit()
 *      i_wd_arithDecoding()
 *      i_wd_dataDecodingTerm()
 *      i_wd_planetDecodingEnd()
 *      i_wd_waveDehuffInit()
 *      i_wd_dc_decoding()
 *      i_wd_ac_decoding()
 *      i_wd_set_huffTable()
 *      i_wd_buffer2image()
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

   /* prototypes */
#include "wavelib.pub"

#include "wavelib.prv"

IP_RCSINFO(i_wdapi_c_RCSInfo, "$RCSfile: i_wdapi.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set up invert filter look up table.
 *
 * LongDescription:::
 *  filter parameters have been scale 256*6. Max image pixel is 255,
 *  after one dimention transform, Max coef is about 255*1.5. We can
 *  build look up table to reduce multiplications.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *  r_lowpass::      Ptr to inverse low pass filter
 * r_highpass::      Ptr to inverse high pass filter
 *     lowlen::      inverse low pass filter length
 *    highlen::      inverse high pass filter length
 *     tbllen::      lookup table length
 *
 * Output:::
 *  lowTab::         low pass look up table
 *  highTab::        high pass look up table
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_init_invertTbl (
    IPState    *pIPState,
    Int32      *r_lowpass,
    Int32      *r_highpass,
    Int32	lowlen,
    Int32	highlen,
    Int32	tbllen,
    Int32     **lowTab,
    Int32     **highTab)
{
    PROLOGUE;

    ERR_PASS(ip_init_invertTbl (pIPState,  r_lowpass, r_highpass, lowlen,
				highlen, tbllen, lowTab, highTab ));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    n level Biorthogonal wavelet invert transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet invert transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *  r_lowpass::     Ptr to inverse lowpass filter
 * r_highpass::     Ptr to inverse highpass filter
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_waveletInvertTrans (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32      *r_lowpass,
    Int32      *r_highpass,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;

    ERR_PASS(ip_waveInvertTrans (pIPState, coefw, coefh, bpl, 
				 r_lowpass, r_highpass, tempbuf, coefbuf)); 
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    n level Biorthogonal wavelet invert transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet invert transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *    options::     Select different filters 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_waveletInvertTrans_Lift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32       options,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;

    if( options & BI44FILTER )
    {
	ERR_PASS(ip_wave44InvertLift (pIPState, coefw, coefh, bpl, 
				    tempbuf, coefbuf)); 
    }
    else if( options & ORD4FILTER )
    {
	ERR_PASS(ip_waveD4InvertLift (pIPState, coefw, coefh, bpl, 
				    tempbuf, coefbuf)); 
    }
    else if( options & TT2AFILTER )
    {
	ERR_PASS(ip_waveTTInvert (pIPState, coefw, coefh, bpl, 
				    tempbuf, coefbuf)); 
    }
    else
	ERR_PASS(ip_waveInvertLift (pIPState, coefw, coefh, bpl, 
				    tempbuf, coefbuf)); 
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    n level Biorthogonal wavelet invert transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet invert transform on nth level with LUT.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *     lowTab::     Ptr to inverse lowpass LUT
 *    highTab::     Ptr to inverse highpass LUT
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_waveletInvertTrans_LUT (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32     **lowTab,
    Int32     **highTab,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;

    ERR_PASS(ip_waveInvertTrans_LUT (pIPState, coefw, coefh, bpl, 
				     lowTab, highTab, tempbuf, coefbuf)); 
    
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmic decoding init.
 *
 * LongDescription:::
 *  Initiate arithmic decoding.
 *
 * Input::: 
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      pdata::     Ptr to user data for output.
 *  inputFunc::     Ptr to call back output function
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_arithDecodingInit (
    IPState    *pIPState,
    void       *pdata,
    waveCallbackFunc inputFunc)
{
    PROLOGUE;

    ERR_PASS( ip_arithDecodingInit (pIPState, pdata, inputFunc));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmic decoding data init.
 *
 * LongDescription:::
 *  Initiate arithmetic decoding for assigned data set.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *       infoIdx::     Decoding data set index ( same as coding)
 *  no_of_symbol::     Number of symbol of coding data set
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_dataDecodingInit (
    IPState    *pIPState,
    Int32	infoIdx,
    Int32	no_of_symbol)
{
    PROLOGUE;

    ERR_PASS( ip_dataDecodingInit (pIPState, infoIdx, no_of_symbol) );

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic decoding.
 *
 * LongDescription:::
 *  Arithmetic decoding.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *  bitCode::     Fix bits coding or ataptive arithmetic coding
 *  infoIdx::     Data set index of current symbol
 *
 * Output:::
 *   pIndex::     Ptr to current symbol index 
 *   symbol::     Ptr to decoded current symbol
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_arithDecoding (
    IPState    *pIPState,
    Int16      *pIndex,
    Int16      *symbol,
    Int16	bitCode,
    Int32	infoIdx) 
{
    PROLOGUE;

    ERR_PASS( ip_arithDecoding (pIPState, pIndex, symbol, bitCode, infoIdx) );

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic decoding data term.
 *
 * LongDescription:::
 *  Term one data set's arithmetic decoding info and free memory. 
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *   infoIdx::        Data set's index
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_dataDecodingTerm (
    IPState    *pIPState,
    Int32	infoIdx)
{
    PROLOGUE; 

    ERR_PASS( ip_dataDecodingTerm (pIPState, infoIdx) );

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic decoding ending alligment.
 *
 * LongDescription:::
 *  Seperated gray color coding. For every planet there is ending
 * alligment. 
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *   number::        Max extral bytes
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_planetDecodingEnd (
    IPState    *pIPState,
    Int32	number)
{
    PROLOGUE; 

    ERR_PASS( ip_planetDecodingEnd (pIPState, number) );

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman decoding init.
 *
 * LongDescription:::
 *  Create huffman decoding table for DC and AC ( copy from jpeg ) and init.
 * decoding coef. 
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      pdata::      Ptr to data used in IO callback func
 *  inputFunc::      Ptr to callback function
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_waveDehuffInit (
    IPState    *pIPState,
    void       *pdata,
    waveCallbackFunc inputFunc)
{
    PROLOGUE;

    ERR_PASS(ip_waveDehuffInit (pIPState, pdata, inputFunc));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    AC Huffman decoding.
 *
 * LongDescription:::
 *    AC Huffman decoding and will call ip_huffDecoding() to decode 
 *    huffman code.
 *
 * Input:::
 *
 * Output:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *    symbol::       Ptr to decoded symbol 
 *       len::       Ptr to bits length
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_ac_decoding (
    IPState    *pIPState,
    Int16      *symbol,
    Int16      *len)
{
    PROLOGUE;
    
    ERR_PASS(ip_ac_decoding (pIPState, symbol, len));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    DC Huffman decoding.
 *
 * LongDescription:::
 *    DC Huffman decoding and will call ip_huffDecoding() to decode 
 *    huffman code.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *    symbol::       Ptr to decoded symbol 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_dc_decoding (
    IPState    *pIPState,
    Int16      *symbol)
{
    PROLOGUE;

    ERR_PASS( ip_dc_decoding (pIPState, symbol) );

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set huffman table for current decoding.
 *
 * LongDescription:::
 *  Huffman table for DC, AC , luminent and color are different. Before
 * decoding, we have to set the right table for current decoding.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *       chan::      Channel number 
 *     dcFlag::      DC flag ( 1: dc huffman table )
 * 
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_set_huffTable (
    IPState    *pIPState,
    Int32	chan,
    Int32	dcFlag)
{
    PROLOGUE;

    ERR_PASS(ip_set_dehuffTable (pIPState, chan, dcFlag));

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Copy coef-buffer to image buffer.
 *
 * LongDescription:::
 *   After inverse wavelet transform, we copy decompressed image from
 *   coef-buffer to images buffer. If it is color image, we transform
 *   YCrCb to RGB.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *    ncomps::      Number of image planets
 *       bpl::      Bytes per line
 *         x::      Begining x position
 *         y::      Begining line
 *         w::      Width of coef-buffer (x+w) < bpl
 *         h::      Height of coef-buffer (y+h) < image height
 *   coefbuf::      Ptr to coef buffer
 *
 * Output:::
 *   imagebuf::     Ptr to image buffer point
 *        min::     Ptr to min of transformed pixel
 *        max::     Ptr max of transformed pixel
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
i_wd_buffer2image (
    IPState    *pIPState,
    Int32	ncomps,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8     **imagebuf,
    Int16     **coefbuf,
    Int16      *min,
    Int16      *max)
{
    PROLOGUE;

    ERR_PASS( ip_buffer2image (pIPState,  ncomps, bpl, x, y, w, h, imagebuf, 
			       coefbuf, min, max ));

    NORMAL_RETURN;
}

/*
$Log: i_wdapi.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.5  1998/11/30 19:24:49  xzhu
xzhu on Mon Nov 30 11:24:16 PST 1998

i_wdapi.c -> /product/ipcore/ipshared/src/RCS/i_wdapi.c,v

Using options to select different method for invert transform.

Revision 1.4  1998/10/20 19:37:19  xzhu
xzhu on Tue Oct 20 12:36:43 PDT 1998

i_wdapi.c -> /product/ipcore/ipshared/src/RCS/i_wdapi.c,v

Added wrap function for new invert transform algorithm.

Revision 1.3  1998/01/13 23:50:32  ddavies
i_wdapi.c -> /product/ipcore/ipshared/src/RCS/i_wdapi.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1997/09/17 23:22:36  xzhu
xzhu on Wed Sep 17 16:22:03 PDT 1997

i_wdapi.c -> /product/ipcore/ipshared/src/RCS/i_wdapi.c,v

Include wavelib.pub first, then include wavelib.prv.

Revision 1.1  1996/10/16 00:02:30  xzhu
Initial revision

*/
