/*<HTML><HEAD><TITLE>wshed.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  i_wshed.c:  Watershed "flooding" algorithm
 *
 *	i_watershedSegment ()
 *
 */

/* turns on counting statistics */
#define DEBUG_C 0
/* turns on timing progress printouts */
#define DEBUG_P 0
/* turns on early abort for preflood tag plane debugging */
#define DEBUG 0

#if DEBUG_C || DEBUG_P
#include "stdio.h"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
#endif

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

#include "shrcc.pub"
#include "shrip.pub"

#include "wshed.prv"
#include "shrapp.pub"


#if DEBUG_P
#include "shrpixr.pub"
#endif

#if DEBUG_C
Int32 xQcount = 0;
Int32 nQcount = 0;
Int32 dQcount = 0;
Int32 mQcount = 0;
Int32 excount = 0;
#endif

/* local prototype */
Int32 CDECL 
ip_WSCCCallback (
    IPState    *pIPState,
    void       *p,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	m,
    Int32	nRuns,
    TCScanLineRun *pRuns);



/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Computes the size and stride of the tag plane, given width and height
 *
 * LongDescription:::
 *  
 *
 * PrivateDescription:::
 *
 * Input:::
 *  w::			Width of the image to be segmented
 *  h::			Height of the image to be segmented
 *
 * Output:::
 *  bplT::		Bytes per line of the tag plane
 *  bytes::		Size in bytes of the tag plane
 *
 * Returns:::
 *  ia_successful
 *	
 * Notes:::
 *  Width and height are set to the framed width and height of the tag
 *  plane (the frames are not the same size as standard PIXR frames).
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
i_getWatershedTagPlaneSize (
    IPState    *pIPState,
    Int32      *w,
    Int32      *h,
    Int32      *bplT,
    Int32      *bytes)
{
    PROLOGUE;

    /* we want a 1 pixel "frame" around our tag plane.  The tag plane
     * needs to be 16 bits deep.
     */

    *w = (*w) + (2 * cWSTagPlaneFramePix);
    *h = (*h) + (2 * cWSTagPlaneFramePix);

    *bplT = (((2 * (*w)) + 3) / 4) * 4;
    *bytes = (*bplT) * (*h);

    NORMAL_RETURN;

}  /* i_getWatershedTagPlaneSize () */


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Callback function for connected component finder used to tag the minima
 *
 * LongDescription:::
 *  
 *
 * PrivateDescription:::
 *
 * Input:::
 *  p::			Pointer to user data packet
 *  x::			x ordinate of upper left BB of this CC
 *  y::			y ordinate of upper left BB of this CC
 *  w::			width of the BB of this CC
 *  h::			height of the BB of this CC
 *  m::			mass of this CC (number of pixels == 1)
 *  nRuns::		number of runs in this CC
 *  pRuns::		pointer to run list for this CC
 *
 * Output:::
 *
 * Returns:::
 *  ia_successful	if completed successfully
 *  ia_internal		if maximum number of tags exceeded
 *	
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
ip_WSCCCallback (
    IPState    *pIPState,
    void       *p,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    Int32	m,
    Int32	nRuns,
    TCScanLineRun *pRuns)
{
    PROLOGUE;
    ccWSPacket *packet;
    Int32 spl, cRow, l, r, ll;
    UInt16 cTag;
    UInt16 *imp, *rowp, *rp;
    TCScanLineRun *thisRun;

    /* keep the compiler quiet */
    m = m;
    nRuns = nRuns;

    packet = (ccWSPacket *)p;

    cTag = packet->currentTag;
    spl = (UInt16)(packet->tagSpl);
    imp = packet->tagImage;
    cRow = y + h - 1 - cFrameBits;
    rowp = imp + (spl * cRow);
    ll = x + w + 2;

    /* blit the tag value for this CC into the tag image */
    /* runs are in bottom-right to top-left order */
    thisRun = pRuns;
    while (thisRun != NULL)
    {
        l = (Int32)(thisRun->runLeft) - cFrameBits;
	r = (Int32)(thisRun->runRight) - cFrameBits;

        /* is this run on a new row? */
	if (r >= ll)
	{
	    cRow--;
	    rowp -= spl;
	}
	ll = l - 1;

	/* paint the run */
        for (rp = rowp + l; l <= r; l++, rp++)
	    USHORT_ACCESS (rp) = cTag;

        thisRun = thisRun->pNext;
    }

    /* increment the tag for the next CC */
    packet->currentTag = cTag + 1;

    /* check for max tag value */
    if (packet->currentTag == cWSMaxTag)
        return (ia_internal);

    NORMAL_RETURN;

}  /* ip_WSCCCallback () */


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Enqueues an entry for (row, col) in the FIFO pointed to by depthQueue
 *
 * LongDescription:::
 *  The enqueued entry appears at the tail
 *
 * PrivateDescription:::
 *
 * Input:::
 *  row::		row location of the pixel being queued
 *  col::		col location of the pixel being queued
 *  depthQueue::	pointer to the destination FIFO
 *
 * Output:::
 *
 * Returns:::
 *  ia_successful	if the entry was queued
 *  is_nomem		if no memory available to alloc or expand the queue
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
ip_WSEnqueue (
    IPState    *pIPState,
    UInt16	row,
    UInt16	col,
    WSDepthQueue *depthQueue)
{
    PROLOGUE;
    Int32 tail;

#if DEBUG_C
    nQcount++;
#endif

    if (depthQueue->entries == depthQueue->length)
    {
        if (depthQueue->length == 0)
	{
#if DEBUG_C
            mQcount++;
#endif

	    /* alloc the queue */
	    (depthQueue->queue) = (WSQueueEntry *)MALLOC (
	                    (sizeof (WSQueueEntry) * cWSQueueBlockSize));

	    if (depthQueue->queue == NULL)
		return (ia_nomem);

	    depthQueue->length = cWSQueueBlockSize;
	}
	else
	{
#if DEBUG_C
	    xQcount++;
#endif
	    /* expand the queue */
	    (depthQueue->queue) = (WSQueueEntry *)REALLOC ((depthQueue->queue),
	                      ((sizeof (WSQueueEntry)) * 
			      (depthQueue->length + cWSQueueBlockSize)));

	    if (depthQueue->queue == NULL)
		return (ia_nomem);
    
	    if (depthQueue->tail != 0)
	    {
	        if (depthQueue->tail < cWSQueueBlockSize)
		{
		    MEMCPY (
		       &(depthQueue->queue[depthQueue->length]),
		       depthQueue->queue,
		       depthQueue->tail * sizeof (WSQueueEntry));
		    depthQueue->tail += depthQueue->length;
		}
		else
		{
		    MEMCPY (
		    &(depthQueue->queue[depthQueue->length]),
		    &(depthQueue->queue[depthQueue->tail - cWSQueueBlockSize]),
		    cWSQueueBlockSize * sizeof (WSQueueEntry));
		    depthQueue->tail -= cWSQueueBlockSize;
		}
	    }
	    else
		depthQueue->tail += depthQueue->length;

	    depthQueue->length += cWSQueueBlockSize;
	}
    }

    tail = depthQueue->tail;

    /* we use a post-increment FIFO ring, so fill in the current tail
     * index and then increment it.
     */
    depthQueue->queue[tail].row = row;
    depthQueue->queue[tail].col = col;

    if (tail == depthQueue->length - 1)
        depthQueue->tail = 0;
    else
        (depthQueue->tail)++;

    (depthQueue->entries)++;

    NORMAL_RETURN;

}  /* ip_WSEnqueue () */


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Dequeues the entry at the head of the FIFO pointed to by depthQueue
 *
 * LongDescription:::
 *  
 *
 * PrivateDescription:::
 *
 * Input:::
 *  depthQueue::		pointer to the depthQueue being referenced
 *
 * Output:::
 *
 * Returns:::
 *  ia_successful
 *	
 * Notes:::
 *  Does not check for empty queue in production version
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
ip_WSDequeue (
    IPState    *pIPState,
    WSDepthQueue *depthQueue)
{
    PROLOGUE;
#if DEBUG_C
    dQcount++;
#endif

#if DEBUG
    /* can't dequeue from an empty queue */
    if ((depthQueue->entries) == 0)
        return (ia_internal);
#endif

    /* increment the head index */
    if (depthQueue->head == (depthQueue->length - 1))
        depthQueue->head = 0;
    else
        (depthQueue->head)++;

    (depthQueue->entries)--;

    NORMAL_RETURN;

}  /* ip_WSDequeue () */


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Four neighbor watershed flooding function
 *
 * LongDescription:::
 *  Checks the four neighbors of the head pixel in the currentDepth FIFO
 *  to determine if the pixel should receive a region tag or become a
 *  boundary pixel.  The pixel is then dequeued.  If the center pixel is
 *  found to be a boundary, its corresponding pixel in the binary
 *  watershed boundary image is set to 1.  Any neighbors which were 
 *  previously untagged and not enqueued are enqueued.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  depthQueueArray::	pointer to array of depth queues
 *  currentDepth::	depth index of center pixel
 *  iS::		pointer to 8 bit source image (gradient)
 *  bplS::		bytes per line of source image
 *  iT::		pointer to 16 bit tag plane
 *  bplT::		bytes per line of the tag plane
 *  iD::		pointer to binary watershed boundary image
 *  bplD::		bytes per line of the watershed boundary image
 *
 * Output:::
 *
 * Returns:::
 *  ia_successful	if completed successfully
 *  ia_nomem		if out of memory while enqueueing
 *	
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
ip_WSCheckNeighbors4 (
    IPState    *pIPState,
    WSDepthQueue *depthQueueArray,
    Int32	currentDepth,
    UInt8      *iS,
    Int32	bplS,
    UInt16     *iT,
    Int32	splT,
    UInt8      *iD,
    Int32	bplD)
{
    PROLOGUE;
    Int32 h, toQueueCount = 0;
    UInt16 row, col;
    UInt16 oTag = 0, cTag;
    Bool boundary = cFalse;
    UInt8  *ip;
    UInt16 *tp;
    WSToBeQueued  toQueue[4];
    WSDepthQueue *cQ;

    cQ = &(depthQueueArray[currentDepth]);
    h = cQ->head;
    row = cQ->queue[h].row;
    col = cQ->queue[h].col;

    ip = iS + (row * bplS) + col;
    tp = iT + (row * splT) + col;

    /* left */
    cTag = USHORT_ACCESS (tp - 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row;
	toQueue[toQueueCount].col = col - 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip - 1));
	toQueue[toQueueCount].pTag = tp - 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
	if (cTag >= cWSFirstTag)
	    oTag = cTag;
    }

    /* up */
    cTag = USHORT_ACCESS (tp - splT);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row - 1;
	toQueue[toQueueCount].col = col;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip - bplS));
	toQueue[toQueueCount].pTag = tp - splT;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* right */
    cTag = USHORT_ACCESS (tp + 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row;
	toQueue[toQueueCount].col = col + 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip + 1));
	toQueue[toQueueCount].pTag = tp + 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* down */
    cTag = USHORT_ACCESS (tp + splT);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row + 1;
	toQueue[toQueueCount].col = col;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip + bplS));
	toQueue[toQueueCount].pTag = tp + splT;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* if multiple tags in neighborhood, mark as boundary */
    if (boundary)
    {
        USHORT_ACCESS (tp) = cWSBoundary;
	SET_IMAGE_BIT (iD + (row * bplD), col);
    }
    /* otherwise, tag with local neighborhood tag */
    else
    {
	if (oTag != 0)
	    /* tag with local neighborhood tag */
	    USHORT_ACCESS (tp) = oTag;

	/* now enqueue and tag all of the previously Init neighbors */
	for (h = 0; h < toQueueCount; h++)
	{
	    __error__ = ip_WSEnqueue (
	             pIPState,
	             toQueue[h].row,
		     toQueue[h].col, 
		     (WSDepthQueue *)(&(depthQueueArray[toQueue[h].depth])));
	    if (__error__ != ia_successful)
		goto abort;

	    USHORT_ACCESS ((toQueue[h].pTag)) = cWSQueued;
	}
    }

    __error__ = ip_WSDequeue (pIPState, cQ);

abort:
    NORMAL_RETURN;

}  /* ip_WSCheckNeighbors4 () */


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Eight neighbor watershed flooding function
 *
 * LongDescription:::
 *  Checks the eight neighbors of the head pixel in the currentDepth FIFO
 *  to determine if the pixel should receive a region tag or become a
 *  boundary pixel.  The pixel is then dequeued.  If the center pixel is
 *  found to be a boundary, its corresponding pixel in the binary
 *  watershed boundary image is set to 1.  Any neighbors which were 
 *  previously untagged and not enqueued are enqueued.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  depthQueueArray::	pointer to array of depth queues
 *  currentDepth::	depth index of center pixel
 *  iS::		pointer to 8 bit source image (gradient)
 *  bplS::		bytes per line of source image
 *  iT::		pointer to 16 bit tag plane
 *  bplT::		bytes per line of the tag plane
 *  iD::		pointer to binary watershed boundary image
 *  bplD::		bytes per line of the watershed boundary image
 *
 * Output:::
 *
 * Returns:::
 *  ia_successful	if completed successfully
 *  ia_nomem		if out of memory while enqueueing
 *	
 * Notes:::
 *  This function has a different flow than the 4 neighbor version.
 *  If the same flow is used it is often the case that given an 
 *  8-connected gradient image, pixels in the queue may have no neighbors
 *  with true region tags (they were enqueued by being next to what
 *  turned out to be a boundary pixel).  This can cause infinite looping
 *  since these pixels may never be reached by flooding at the current
 *  depth.  So, we modify the flow to defer enqueueing of the neighbors
 *  until we know whether the center pixel is a boundary pixel or not.
 *  If it is a boundary pixel, its neighbors are not enqueued at this
 *  point, otherwise they are.
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
ip_WSCheckNeighbors8 (
    IPState    *pIPState,
    WSDepthQueue *depthQueueArray,
    Int32	currentDepth,
    UInt8      *iS,
    Int32	bplS,
    UInt16     *iT,
    Int32	splT,
    UInt8      *iD,
    Int32	bplD)
{
    PROLOGUE;
    Int32 h, toQueueCount = 0;
    UInt16 row, col;
    UInt16 oTag = 0, cTag;
    Bool boundary = cFalse;
    UInt8  *ip;
    UInt16 *tp;
    WSDepthQueue *cQ;
    WSToBeQueued  toQueue[8];

    cQ = &(depthQueueArray[currentDepth]);
    h = cQ->head;
    row = cQ->queue[h].row;
    col = cQ->queue[h].col;

    ip = iS + (row * bplS) + col;
    tp = iT + (row * splT) + col;

    /* Note that this function differs from the 4 neighbor one in
     * that it defers all enqueuing until the center pixel tag value
     * is known.  This is necessary when the gradient image contains
     * thin, 8-connected edges which can cause leakage of tags into
     * regions which will never be flooded at a particular level which
     * leads to infinite looping.  To avoid this, only previously
     * untagged neighbors which are adjacent to a true tag (not a
     * boundary tag) are enqueued.
     */

    /* up and left */
    cTag = USHORT_ACCESS (tp - splT - 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row - 1;
	toQueue[toQueueCount].col = col - 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip - bplS - 1));
	toQueue[toQueueCount].pTag = tp - splT - 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    oTag = cTag;
	}
    }

    /* up */
    cTag = USHORT_ACCESS (tp - splT);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row - 1;
	toQueue[toQueueCount].col = col;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip - bplS));
	toQueue[toQueueCount].pTag = tp - splT;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* up and right */
    cTag = USHORT_ACCESS (tp - splT + 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row - 1;
	toQueue[toQueueCount].col = col + 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip - bplS + 1));
	toQueue[toQueueCount].pTag = tp - splT + 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* right */
    cTag = USHORT_ACCESS (tp + 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row;
	toQueue[toQueueCount].col = col + 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip + 1));
	toQueue[toQueueCount].pTag = tp + 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
	if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}

    }

    /* down and right */
    cTag = USHORT_ACCESS (tp + splT + 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row + 1;
	toQueue[toQueueCount].col = col + 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip + bplS + 1));
	toQueue[toQueueCount].pTag = tp + splT + 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* down */
    cTag = USHORT_ACCESS (tp + splT);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row + 1;
	toQueue[toQueueCount].col = col;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip + bplS));
	toQueue[toQueueCount].pTag = tp + splT;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* down and left */
    cTag = USHORT_ACCESS (tp + splT - 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row + 1;
	toQueue[toQueueCount].col = col - 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip + bplS - 1));
	toQueue[toQueueCount].pTag = tp + splT - 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* left */
    cTag = USHORT_ACCESS (tp - 1);
    if (!cTag)  /* is this pixel in the init state? */
    {
	/* mark to be enqueued */
	toQueue[toQueueCount].row = row;
	toQueue[toQueueCount].col = col - 1;
	toQueue[toQueueCount].depth = MAX (currentDepth, 
	                              UCHAR_ACCESS (ip - 1));
	toQueue[toQueueCount].pTag = tp - 1;
	toQueueCount++;
    }
    else  /* pixel wasn't in init state */
    {
        /* does the pixel have a region tag? */
        if (cTag >= cWSFirstTag)
	{
	    /* is this the first region tag in the neighborhood? */
	    if ((oTag >= cWSFirstTag) && (cTag != oTag))
	        boundary = cTrue;
	    else
	        oTag = cTag;
	}
    }

    /* if multiple tags in neighborhood, mark as boundary */
    if (boundary)
    {
        USHORT_ACCESS (tp) = cWSBoundary;
	SET_IMAGE_BIT (iD + (row * bplD), col);
    }
    /* otherwise, tag with local neighborhood tag */
    else
    {
	if (oTag != 0)
	    /* tag with local neighborhood tag */
	    USHORT_ACCESS (tp) = oTag;

	/* now enqueue and tag all of the previously Init neighbors */
	for (h = 0; h < toQueueCount; h++)
	{
	    __error__ = ip_WSEnqueue (
	             pIPState,
	             toQueue[h].row,
		     toQueue[h].col, 
		     (WSDepthQueue *)(&(depthQueueArray[toQueue[h].depth])));
	    if (__error__ != ia_successful)
		goto abort;

	    USHORT_ACCESS ((toQueue[h].pTag)) = cWSQueued;
	}
    }

    /* dequeue the center pixel */
    __error__ = ip_WSDequeue (pIPState, cQ);

abort:
    NORMAL_RETURN;


}  /* ip_WSCheckNeighbors8 () */


/**************************************************************************
 * BeginDocumentation:::
 *
 *
 * OneLineDescription:::
 *  Watershed based segmentation of an image
 *
 * LongDescription:::
 *  The watershed algorithm is a segmentation scheme based on region
 *  growing using a flooding algorithm which detects ridge lines which
 *  form boundaries between multiple "watershed" regions.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  connectType::	0 = four neighbor flooding, 1 = eight neighbor
 *			flooding.  Note that the watershed boundaries
 *			will have inverse connectivity.
 *  iS::		pointer to 8-bit input image (gradient)
 *  wS::		width of 8-bit input image
 *  hS::		height of 8-bit input image
 *  bplS::		bytes per line of 8-bit input image
 *  iE::		pointer to binary edge image (boundaries of the seeds) 
 *  wE::		width of binary edge image
 *  hE::		height of binary edge image
 *  bplE::		bytes per line of binary edge image
 *  iM::		pointer to binary minima image ("region seeds")
 *  wM::		width of binary minima image
 *  hM::		height of binary minima image
 *  bplM::		bytes per line of binary minima image
 *  wD::		width of binary watershed boundary image
 *  hD::		height of binary watershed boundary image
 *  bplD::		bytes per line of binary watershed boundary image
 *
 *
 * Output:::
 *  iD::		pointer to binary destination image (watershed
 *			boundaries)
 *
 * Returns:::
 *  ia_successful	if completed successfully
 *  ia_nomem		if out of memory
 *	
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  gradient
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work_in_progress
 *
 * Authors/Modifications:::
 *  dferrell
 *
 * Citations:::
 *
 * EndDocumentation:::
 *
 **************************************************************************/
Int32 CDECL 
i_watershedSegment (
    IPState    *pIPState,
    Int32	connectType,
    UInt8      *iS,
    Int32	wS,
    Int32	hS,
    Int32	bplS,
    UInt8      *iE,
    Int32	wE,
    Int32	hE,
    Int32	bplE,
    UInt8      *iM,
    Int32	wM,
    Int32	hM,
    Int32	bplM,
    UInt16     *iT,
    Int32	wT,
    Int32	hT,
    Int32	bplT,
    UInt8      *iD,
    Int32	wD,
    Int32	hD,
    Int32	bplD)
{
    PROLOGUE;
    Int32 currentDepth,
	  depth,
	  row, col,
	  splT,
	  ret = ia_successful;
    UInt16 currentTag = cWSFirstTag,
	   largestTag;
    UInt8  *rp;
    UInt16 *tp;
    WSDepthQueue *depthQueueArray;
    ccWSPacket packet;


    iS += (bplS << cLogFrameBits) + (8 * cFrameBytes);
    wS -= 2 * cFrameBits;
    hS -= 2 * cFrameBits;
    iD += (bplD << cLogFrameBits) + cFrameBytes;
    wD -= 2 * cFrameBits;
    hD -= 2 * cFrameBits;
    iE += (bplE << cLogFrameBits) + cFrameBytes;
    wE -= 2 * cFrameBits;
    hE -= 2 * cFrameBits;
    wT -= 2 * cWSTagPlaneFramePix;
    hT -= 2 * cWSTagPlaneFramePix;

    /* Note that we don't adjust the minima image parameters since we'll
     * use them when we call the CC finder.
     */

    /* short words (pixels) per line in the tag image */
    splT = bplT / 2;

    /* init the depth queue array */
    depthQueueArray = (WSDepthQueue *)CALLOC (256, sizeof (WSDepthQueue));
    if (depthQueueArray == NULL)
    {
        __error__ = ia_nomem;
        goto cleanup;
    }

    /* would it pay off to run a histogram of the gradient image to help
     * set the queue lengths?
     */
    for (depth = 0; depth < 256; depth++)
    {
#if 0
        /* start with 512 length queues */
        depthQueueArray[depth].queue = (WSQueueEntry *) MALLOC (
	                    cWSQueueBlockSize * sizeof (WSQueueEntry));
        if (depthQueueArray[depth].queue == NULL)
	{
	    __error__ = ia_nomem;
	    goto cleanup;
	}
        depthQueueArray[depth].length = cWSQueueBlockSize;
        depthQueueArray[depth].entries = 0;
        depthQueueArray[depth].head = 0;
        depthQueueArray[depth].tail = 0;
#else
        /* start with no queues allocated */
        depthQueueArray[depth].queue = NULL;
        depthQueueArray[depth].length = 0;
        depthQueueArray[depth].entries = 0;
        depthQueueArray[depth].head = 0;
        depthQueueArray[depth].tail = 0;
#endif
    }

#if DEBUG_P
    fprintf (stdout, "Init Queues: %f\n", StopTime ());
#endif

    /* init the frame edges of the tag plane to be boundaries */
    tp = iT + (splT * (cWSTagPlaneFramePix / 2)) + (cWSTagPlaneFramePix / 2);
    for (col = 0; col < wT + 2; col++, tp++)
        USHORT_ACCESS (tp) = cWSBoundary;
    tp = iT + (splT * cWSTagPlaneFramePix) + (cWSTagPlaneFramePix / 2);
    for (row = 0; row < hT; row++, tp += splT)
    {
        USHORT_ACCESS (tp) = cWSBoundary;
	USHORT_ACCESS (tp + wT + (cWSTagPlaneFramePix / 2)) = cWSBoundary;
    }
    tp = iT + (splT * (cWSTagPlaneFramePix + hT)) + (cWSTagPlaneFramePix / 2);
    for (col = 0; col < wT + 2; col++, tp++)
        USHORT_ACCESS (tp) = cWSBoundary;


#if DEBUG_P
    fprintf (stdout, "Init Tag Frames: %f\n", StopTime ());
#endif

    /* move the tag image pointer inside its frame */
    iT += (splT * cWSTagPlaneFramePix) + cWSTagPlaneFramePix;

    /* Add the existing CC's in the minima image to the tag plane,
     * keeping track of the number of tags.  
     */
    packet.tagImage = iT;
    packet.tagSpl = splT;
    packet.currentTag = currentTag;

    __error__ = i_findConnectedComponentRuns (pIPState, iM, bplM, 0, 0, wM, hM,
                                            0, 0,
                                            connectType,
                                            (void *)(&packet), 
					    ip_WSCCCallback,
					    &ret);
    if (__error__ != ia_successful)
        goto cleanup;
    if (ret != ia_successful)
        goto cleanup;

    /* this assumes no new tags will be created later */
    largestTag = packet.currentTag;

#if DEBUG_P
    fprintf (stdout, "Init Tag with minima: %f\n", StopTime ());
#endif

    /* Add the boundary pixels of each minima region to the array of 
     * depth queues.
     */
    /* for now we are passed the border image.  This should be changed. */
    /* scan the border image, queueing all "on" pixels properly */
    for (row = 0, rp = iE; row < hE; row++, rp += bplE)
    {
/*
        Int32 wpl;
	UInt32 edgeW;

	wpl = (bplE + 3) / 4);
*/
        /* this should be done a word at a time */
        for (col = 0; col < wE; col++)
	{
	    if (GET_IMAGE_BIT (rp, col))
	    {
	        __error__ = ip_WSEnqueue (pIPState,
				       (UInt16) row,
				       (UInt16) col, 
				       (WSDepthQueue *)
					    (&(depthQueueArray[
						UCHAR_ACCESS (iS +
							      (row * bplS) +
							      col)])));

		if (__error__ != ia_successful)
		    goto cleanup;

		/* mark it as queued in the tag plane */
		USHORT_ACCESS (iT + (row * splT) + col) = cWSQueued;
	    }
	}
    }

#if DEBUG_P
    fprintf (stdout, "Enqueue minima boundaries: %f\n", StopTime ());
#endif

#if DEBUG
    NORMAL_RETURN;
#endif

    /* For each depth in the queue, process all pixel entries at that
     * depth.
     */
    currentDepth = 0;
    while (currentDepth < 256)
    {
        if (connectType == 0)
	{
	    while (depthQueueArray[currentDepth].entries != 0)
	    {
		__error__ = ip_WSCheckNeighbors4 (pIPState, 
						depthQueueArray, currentDepth, 
						iS, bplS, iT, splT, iD, bplD);

		if (__error__ != ia_successful)
		    goto cleanup;
	    }
	}
	else
	{
	    while (depthQueueArray[currentDepth].entries != 0)
	    {
		__error__ = ip_WSCheckNeighbors8 (pIPState,
						depthQueueArray, currentDepth, 
						iS, bplS, iT, splT, iD, bplD);

		if (__error__ != ia_successful)
		    goto cleanup;
	    }
	}

	/* free the now empty queue at this depth */
	if (depthQueueArray[currentDepth].queue != NULL)
	{
	    FREE (depthQueueArray[currentDepth].queue);
	    depthQueueArray[currentDepth].queue = NULL;
	}

	currentDepth++;
    }

#if DEBUG_P
    fprintf (stdout, "Flooding: %f\n", StopTime ());
#endif


#if DEBUG_C
    fprintf (stdout, "%ld enqueue calls and\n%ld dequeue calls\nwith %ld queue expansions and\n%ld queue allocations.\n%ld neighborhood exceptions\n",
             nQcount, dQcount, xQcount, mQcount, excount);
#endif

    /* free up the queue array */
    for (currentDepth = 0; currentDepth < 256; currentDepth++)
        if (depthQueueArray[currentDepth].queue != NULL)
	    FREE (depthQueueArray[currentDepth].queue);

    FREE (depthQueueArray);

    NORMAL_RETURN;

cleanup:
    /* free up the queue array */
    for (currentDepth = 0; currentDepth < 256; currentDepth++)
        if (depthQueueArray[currentDepth].queue != NULL)
	    FREE (depthQueueArray[currentDepth].queue);

    FREE (depthQueueArray);

    NORMAL_RETURN;

}  /* i_watershedSegment () */



/*  
 * $Log: i_wshed.c,v $
 * Revision 1.1.1.1  1999/09/20 18:18:13  jlin
 * Imported sources
 *
 * Revision 1.6  1998/01/13 23:50:34  ddavies
 * i_wshed.c -> /product/ipcore/ipshared/src/RCS/i_wshed.c,v
 *
 * CommonCheck: Add IPState parameter to almost all procedures.
 *
 * Revision 1.5  1996/10/30 21:55:33  dferrell
 * dferrell on Wed Oct 30 14:52:43 PPET 1996
 *
 * i_wshed.c -> /product/ipcore/ipshared/src/RCS/i_wshed.c,v
 *
 * Made 4-neighbor flooding function defer enqueues until the final state
 * of the center pixel is known.  This avoids trapping enqueued pixels
 * in local "wells" causing infinite looping (which can be a major
 * performance hit).
 *
 * Revision 1.4  1996/07/26 17:31:49  dferrell
 * dferrell on Fri Jul 26 10:31:04 PDT 1996
 *
 * i_wshed.c -> /product/ipcore/ipshared/src/RCS/i_wshed.c,v
 *
 * Changed all function declaration styles to the one-parameter-per-line
 * style that the nag scripts want.
 *
 * Revision 1.3  1996/07/24 18:20:36  dferrell
 * dferrell on Wed Jul 24 11:20:20 PDT 1996
 *
 * i_wshed.c -> /product/ipcore/ipshared/src/RCS/i_wshed.c,v
 *
 * Fixed compiler warnings from PC.
 *
 * Revision 1.2  1996/07/24 01:36:39  dferrell
 * dferrell on Tue Jul 23 18:36:27 PDT 1996
 *
 * i_wshed.c -> /product/ipcore/ipshared/src/RCS/i_wshed.c,v
 *
 * Fixed compiler warnings from NT build.
 *
 * Revision 1.1  1996/07/22 23:09:09  dferrell
 * Initial revision
 *
 *
 */

/*</PRE></BODY>*/
