/*****************************************************************
 *  Copyright (c) 1989, 1990, 1991, 1992.                        *
 *  Xerox Corporation.  All rights reserved.                     *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  iautils.c:
 *      Includes
 *
 *           Emil's timing utilities
 *               void    StartTime()
 *               Float32   StopTime()
 *
 */

#ifdef __SOLARIS2__
#include "time.h"   /* for clock () and clock_t */
#endif

#include "types.pub"
#include "ipscore.prv"

#ifdef _UNIX_
/* rusage is a BSDism which Sun climains is going away in a future
 * (post 5.3) world.
 */
#include <sys/time.h>
#include <sys/resource.h>
#ifndef RUSAGE_SELF	/* in case it wasn't defined by resource.h */
#include <sys/rusage.h>	/* try looking here in Solaris 2.3 anyway */
#endif
#endif /* _UNIX_ */

#if defined(__WATCOMC__) || defined (_WIN32) || defined(macintosh)
#include <time.h>
#endif


	/* prototypes */
#include "shrpixr.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
IP_RCSINFO(RCSInfo, "$RCSfile: iautils.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:13 $");


/*--------------------------------------------------------------------*
 *                          Timing Procs                              *
 *--------------------------------------------------------------------*/
#ifdef _UNIX_

/* rusage is a BSDism which Sun climains is going away in a future
 * (post 5.3) world.
 */

#ifdef __SOLARIS2__  /* use clock () if on Solaris 2.x */
static clock_t before;
static clock_t after;
#else
static  struct rusage before;
static  struct rusage after;
#endif

/* PUBLIC */
/*********************************************************************
 * StartTime()
 *		Starting timing the execution of code
 *
 *	Inputs:	none
 *
 *	StartTime cannot be nested i.e. the code sequence
 *	StartTime() ... StartTime() ... StopTime() ... StopTime()
 *	is illegal.
 *********************************************************************/
void	CDECL 
StartTime (void)
{
#ifdef __SOLARIS2__  /* use clock () if on Solaris 2.x */
    before = clock ();
#else
    getrusage(RUSAGE_SELF, &before);
#endif
}


/* PUBLIC */
/*********************************************************************
 * StopTime()
 *		Stop timing the execution of code and return
 *		elapsed time.
 *
 *	Inputs:	none
 *
 *********************************************************************/
Float32  CDECL 
StopTime (void)
{
    Int32       secs, usecs;
    Float32     timeInSecs;

#ifdef __SOLARIS2__
    /* keep compilers quiet */
    secs = usecs = 0;

    after = clock ();

    timeInSecs = ((Float32)after - (Float32)before) / (Float32)CLOCKS_PER_SEC;
#else
    getrusage(RUSAGE_SELF, &after);

/* This is a mess.  [Obsolete] Solaris rusage structures come
 * with all-new, all-singing, all-dancing nanosecond resolution.
 * Unfortunately, there aseems no a priori way to determine whether
 * we have microseconds or nanoseconds.
 * As a desperation measure, I'm (RM) guessing that if NANOSEC is
 * defined in sys/time.h, the time is in NANOSEConds.  At least
 * this words for SunOS4.1, Solaris2.2 and Solaris2.3.
 * Who knows about other Unix implementations?
 */
#ifdef NANOSEC
    usecs = after.ru_utime.tv_nsec - before.ru_utime.tv_nsec;
    secs = after.ru_utime.tv_sec - before.ru_utime.tv_sec;
    if (usecs < 0) {
        usecs = usecs + NANOSEC;
        secs = secs - 1;
    }
    timeInSecs = secs + ((Float32) usecs) / (Float32) NANOSEC;

#else	/* rusage is measured in microseconds, not nanoseconds */

    usecs = after.ru_utime.tv_usec - before.ru_utime.tv_usec;
    secs = after.ru_utime.tv_sec - before.ru_utime.tv_sec;
    if (usecs < 0) {
        usecs = usecs + 1000000;
        secs = secs - 1;
    }
    timeInSecs = secs + ((Float32) usecs) / 1000000.0;

#endif	 /* NANOSEC */
#endif   /* Solaris 2.x or not */

    return timeInSecs;
}

#endif  /* UNIX */

#if defined(__WATCOMC__) || defined(_WIN32) || defined(macintosh)

static clock_t prev_time;

void CDECL 
StartTime (void)
{
    prev_time = clock();
}

Float32 CDECL 
StopTime (void)
{
    return(( (Float32) ( clock() - prev_time ))/CLOCKS_PER_SEC );
}

#endif


/*

$Log: iautils.c,v $
Revision 1.1.1.1  1999/09/20 18:18:13  jlin
Imported sources

Revision 1.18  1998/01/15 01:46:48  ddavies
ddavies on Wed Jan 14 17:46:47 PST 1998

iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v

CommonCheck: Fix WATCOM complaints about unused procedure args.

Revision 1.17  1998/01/13 23:50:35  ddavies
iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.16  1996/02/06 19:21:17  dferrell
dferrell on Tue Feb 6 11:20:09 PST 1996

iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v

Removed Mac specific timing functions as CodeWarrior has correct
implementations of the ANSI time stuff.

 * Revision 1.15  1995/12/13  02:46:50  ejaquith
 * iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v
 *
 * CommonCheck: Added ";" to end of IP_RCSINFO() macro.
 *
 * Revision 1.14  1995/10/19  00:43:52  jcraig
 * jcraig on Wed Oct 18 17:43:03 PDT 1995
 *
 * iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v
 *
 * define StartTime and StopTime for _WIN32 systems
 *
 * Revision 1.13  1995/03/31  16:54:43  dferrell
 * dferrell on Fri Mar 31 08:53:52 PST 1995
 *
 * iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v
 *
 * Made use of all variables in StopTime () to keep gcc under Solaris happy
 * and quiet.
 *
 * Revision 1.12  1994/07/21  20:47:56  dferrell
 * iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v
 *
 * Added use of clock () function when we compile under solaris 2.x so that
 * we get timing info in our tests.
 *
 * Revision 1.11  1994/02/04  01:37:29  dferrell
 * dferrell on Thu Feb 3 17:36:52 PST 1994
 *
 * iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v
 *
 * Added Mac versions of StartTime () and StopTime ().  Resolution is 1/60 sec.
 *
 * Revision 1.10  1993/12/02  18:49:29  ddavies
 * iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v
 *
 * Put in Richard Mylarnik's fix to StopTime.  In Solaris 2.x, the time is
 * returned in nanoseconds instead of microseconds.  Richard guesses that the
 * way you can tell this is happening is that NANOSEC is defined in those
 * systems and not in the others.  This seems to work so far.
 * Also put RCS magic into a macro that doesn't cause compiler warnings.
 *
 * Revision 1.9  1993/11/05  01:45:02  ddavies
 * iautils.c -> /product/ipcore/ipshared/src/RCS/iautils.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


