/*
 * jcexpand.c
 *
 * Copyright (C) 1991, 1992, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains image edge-expansion routines.
 * These routines are invoked via the edge_expand method.
 */

#include "jpeg.h"
#include "jpeg.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: jcexpand.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:14 $");

/*
 * Expand an image so that it is a multiple of the MCU dimensions.
 * This is to be accomplished by duplicating the rightmost column
 * and/or bottommost row of pixels.  The image has not yet been
 * downsampled, so all components have the same dimensions.
 */

METHODDEF Int32 CDECL
edge_expand (compress_info_ptr cinfo,
	     Int32 input_cols, Int32 input_rows,
	     Int32 output_cols, Int32 output_rows,
	     JSAMPIMAGE image_data)
{
  /* Expand horizontally */
  if (input_cols < output_cols) {
    register JSAMPROW ptr;
    register JSAMPLE pixval;
    register Int32 count;
    register Int32 row;
    Int16 ci;
    Int32 numcols = output_cols - input_cols;

    for (ci = 0; ci < cinfo->num_components; ci++) {
      for (row = 0; row < input_rows; row++) {
	ptr = image_data[ci][row] + (input_cols-1);
	pixval = GETJSAMPLE(*ptr++);
	for (count = numcols; count > 0; count--)
	  *ptr++ = pixval;
      }
    }
  }

  /* Expand vertically */
  /* This happens only once at the bottom of the image, */
  /* so it needn't be super-efficient */
  if (input_rows < output_rows) {
    register Int32 row;
    Int16 ci;
    JSAMPARRAY this_component;

    for (ci = 0; ci < cinfo->num_components; ci++) {
      this_component = image_data[ci];
      for (row = input_rows; row < output_rows; row++) {
	jcopy_sample_rows(this_component, input_rows-1, this_component, row,
			  1, output_cols);
      }
    }
  }

  return ia_successful;
}


/*
 * The method selection routine for edge expansion.
 */

GLOBAL Int32 CDECL
jselexpand (compress_info_ptr cinfo)
{
  /* just one implementation for now */
  cinfo->cmethods->edge_expand = edge_expand;

  return ia_successful;
}
/*
$Log: jcexpand.c,v $
Revision 1.1.1.1  1999/09/20 18:18:14  jlin
Imported sources

Revision 1.5  1995/12/13 02:46:57  ejaquith
jcexpand.c -> /product/ipcore/ipshared/src/RCS/jcexpand.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.4  1994/10/21  00:49:11  lperry
 * jcexpand.c -> /product/ipcore/ipshared/src/RCS/jcexpand.c,v
 *
 * Get rid of UINT32, etc., so these files can work with Peter's interpreter.
 *
 * Revision 1.3  1994/07/21  00:15:41  lperry
 * jcexpand.c -> /product/ipcore/ipshared/src/RCS/jcexpand.c,v
 *
 * Change error handling from old setjmp/longjmp to function return codes.
 *
 * Revision 1.2  1994/06/27  16:44:34  lperry
 * jcexpand.c -> /product/ipcore/ipshared/src/RCS/jcexpand.c,v
 *
 * Moved static variables to the "cinfo" structure to achieve reentrancy.
 *
 * Revision 1.1  1994/02/27  02:20:21  lperry
 * Initial revision
 *
*/
