/***************************************************************
    Copyright (c) 1995, Xerox Corporation.  All rights reserved. 
    Copyright protection claimed includes all forms and matters
    of copyrightable material and information now allowed by
    statutory or judicial law or hereafter granted, including
    without limitation, material generated from the software
    programs which are displayed on the screen such as icons,
    screen display looks, etc.
***************************************************************/

#include <math.h>

#define JPEG_INTERNALS

#include "jinclude.h"
#include "jpeglib.h"
#include "ipscore.prv"
#include "jdct.h"		/* Private declarations for DCT subsystem */

IP_RCSINFO(RCSInfo, "$RCSfile: jdsdct.c,v $Revision: 1.1.1.1 $Date:");

#ifdef SDCT_SUPPORTED

/*
 * This module is specialized to the case DCTSIZE = 8.
 */

#if DCTSIZE != 8
  Sorry, this code only copes with 8x8 DCTs. /* deliberate syntax err */
#endif

#define   HORIZONTAL 0
#define   VERTICAL   1
#define   MODE0      0
#define   MODE1      1
#define   MODE2      2
#define   AVE_MODE   0
#define   SLOPE_MODE 1
#define   DCTSIZE_1   DCTSIZE-1

Int32     q_dv[4][3], q_dh[4][3];
ISLOW_MULT_TYPE   qtbl_sdct[3][DCTSIZE2];

/*   
 * Initialize for smart IDCT.
 */

GLOBAL ErrCode CDECL 
jpeg_smart_dct_init (
    IPState    *pIPState,
    j_decompress_ptr cinfo)
{
  PROLOGUE;
  jpeg_component_info *compptr;
  ISLOW_MULT_TYPE * quantptr, *pqv, *pqh;
  Int32            sv, sh;
  Int32            ctr, ci, idx;
  float            fweight[DCTSIZE] = {(float)0.0, (float)0.03806, 
				       (float)0.146446, (float)0.308658,
				       (float)0.49999, (float)0.691341, 
				       (float)0.853553, (float)0.961939};
  float           tmp1, tmp2;
  double          temp;

  for (ci = 0; ci < cinfo->comps_in_scan; ci++) 
  {
      
      compptr = cinfo->cur_comp_info[ci];
      quantptr = (ISLOW_MULT_TYPE *) compptr->dct_table;
      for(ctr=0; ctr<DCTSIZE2; ++ ctr)
	  qtbl_sdct[ci][ctr] = quantptr[jpeg_zigzag_order[ctr]];
      pqv = pqh = qtbl_sdct[ci];
      tmp1 = tmp2 = (float)0.0;
      for(ctr=0; ctr<DCTSIZE; ++ctr, pqh += DCTSIZE, pqv ++)
      {
	  sv = sh =0;
	  for(idx=0; idx<DCTSIZE; ++idx)
	  {
	      sv += pqv[idx*DCTSIZE]*pqv[idx*DCTSIZE];
	      sh += pqh[idx]*pqh[idx];
	  }
	  tmp1 += (fweight[ctr] * (float)sv);
	  tmp2 += (fweight[ctr] * (float)sh);
      }
      temp = sqrt((double)tmp1)/420.0 ;
      q_dv[0][ci] = (Int32)((float)temp * 20.0 + 0.5);
      q_dv[1][ci] = (Int32)((float)temp * 50.0 + 0.5);
      q_dv[2][ci] = (Int32)((float)temp * 7.0 + 0.5);
      q_dv[3][ci] = (Int32)((float)temp * 12.0 + 0.5);
      temp = sqrt((double)tmp2)/420.0;
      q_dh[0][ci] = (Int32)((float)temp * 20.0 + 0.5);
      q_dh[1][ci] = (Int32)((float)temp * 50.0 + 0.5);
      q_dh[2][ci] = (Int32)((float)temp * 7.0 + 0.5);
      q_dh[3][ci] = (Int32)((float)temp * 12.0 + 0.5);
  }

  NORMAL_RETURN;
}

LOCAL ErrCode CDECL
h_filter(
    IPState    *pIPState,
    UInt8      *inData,
    UInt8      *outData,
    UInt8      *edgemap,
    Int32	mode)
{
    PROLOGUE;
    Int32      row, col;
    UInt8      *pin, *pout, *pedge;
    Int16      sum, num;

    pin = inData;
    pout = outData;
    pedge = edgemap;

    MEMCOPY(outData, inData, DCTSIZE2); 
    if(mode == AVE_MODE)
    {   
	for(row=0; row<DCTSIZE; ++row)
	{
	    for(col=0; col<DCTSIZE; ++col, ++pin, ++pedge, ++pout)
	    {
		num = 1;
		sum = *pin;
		if(col >0 && !(*(pedge-1)))
		{
		    sum += *(pin-1);
		    ++num;
		}
		if(!(*pedge))
		{
		    sum += *(pin+1);
		    ++num;
		}
		if( num >1)
		    *pout =(sum+1)/num;
	    }
	}
    }
    else
    {
	for(row=0; row<DCTSIZE; ++row)
	{
	    for(col=0; col<DCTSIZE; ++col, ++pin, ++pedge, ++pout)
	    {
		num = 2;
		sum = (*pin)<<1;
		if(col >0 && !(*(pedge-1)))
		{
		    sum += *(pin-1);
		    ++num;
		}
		if(!(*pedge))
		{
		    sum += *(pin+1);
		    ++num;
		}
		if( num >2)
		    *pout = (sum+1)/num;
	    }
	}
    }

    NORMAL_RETURN;
}

LOCAL ErrCode CDECL
v_filter(
    IPState    *pIPState,
    UInt8      *inData,
    UInt8      *outData,
    UInt8      *edgemap,
    Int32       mode)
{
    PROLOGUE;
    Int32      row, col;
    UInt8      *pin, *pout, *pedge;
    Int16      sum, num;

    MEMCOPY(outData, inData, DCTSIZE2);
    if(mode == AVE_MODE)
    {   
	for(row=0; row<DCTSIZE; ++row)
	{
	    pin = inData +row;
	    pout = outData +row;
	    pedge = edgemap +row;

	    for(col=0; col<DCTSIZE; ++col, pin+=DCTSIZE, 
		pout+=DCTSIZE, pedge+=DCTSIZE)
	    {
		num = 1;
		sum = *pin;
		if(col >0 && !(*(pedge-DCTSIZE)))
		{
		    sum += *(pin-DCTSIZE);
		    ++num;
		}
		if(!(*pedge))
		{
		    sum += *(pin+DCTSIZE);
		    ++num;
		}
		if(num >1)
		    *pout = (sum+1)/num;
	    }
	}
    }
    else
    {
	for(row=0; row<DCTSIZE; ++row)
	{
	    pin = inData +row;
	    pout = outData +row;
	    pedge = edgemap +row;

	    for(col=0; col<DCTSIZE; ++col, pin+=DCTSIZE, 
		pout+=DCTSIZE, pedge+=DCTSIZE)
	    {
		num = 2;
		sum = *pin<<1;
		if(col >0 && !(*(pedge-DCTSIZE)))
		{
		    sum += *(pin-DCTSIZE);
		    ++num;
		}
   		if(!(*pedge))
		{
		    sum += *(pin+DCTSIZE);
		    ++num;
		}
		if(num > 2)
		    *pout = (sum+1)/num;
	    }
	}
    }
    
    NORMAL_RETURN;
}
	 
LOCAL ErrCode CDECL
findedge(
    IPState    *pIPState,
    UInt8      *inData,
    UInt8      *map,
    Int32	threshold,
    Int32	direction,
    Int32	mode)
{
    PROLOGUE;
    Int32      ctr, idx;
    UInt8     *ptemp, *pmap;
    Int32      step, step1, step2;
    Int32      dis;

    if(direction == HORIZONTAL)
    {
	step1 = DCTSIZE;
	step2 = 1;
    }
    else
    {
	step1 = 1;
	step2 = DCTSIZE;
    }
    if(mode != MODE1)
    {
	for(ctr=0, step=0; ctr<DCTSIZE; ++ctr, step+=step1)
	{
	    ptemp = inData+step;
	    pmap = map + step;
	    for(idx=0; idx<DCTSIZE_1; ++idx, ptemp+= step2, pmap+= step2)
	    {
		if(abs(*ptemp - *(ptemp+step2)) > threshold)
		    *pmap = 1;
	    }
	    *pmap =1;
	}
    }
    else
    {
	Int32   thre1, dis1, dis2;
	thre1 = threshold /3;
	if(thre1 > 32) thre1 = 32;
	for(ctr=0, step=0; ctr<DCTSIZE; ++ctr, step+=step1)
	{
	    ptemp = inData+step;
	    pmap = map + step;
	    for(idx=0; idx<DCTSIZE_1; ++idx, ptemp+= step2, pmap+= step2)
	    {
		dis = *ptemp - *(ptemp+step2);
		if(abs(dis) > threshold) *pmap = 1;
		else if(abs(dis) > thre1)
		{
		    dis1= (idx >0)? (*(ptemp-step2) - *(ptemp+step2)):dis;
		    dis2= (idx < (DCTSIZE_1-1))? 
			(*ptemp - *(ptemp+step2+step2)):dis;
		    if((dis*dis1 >0 && abs(dis1) > threshold) ||
		       (abs(dis2) > threshold && dis*dis2 >0))
			*pmap =1;
		}
	    }
	    *pmap =1;
	}
    }

    NORMAL_RETURN;
}

GLOBAL ErrCode CDECL 
jpeg_smart_dct (
    IPState    *pIPState,
    JSAMPLE    *range_limit,
    jpeg_component_info *compptr,
    JCOEFPTR	coef_block,
    JSAMPARRAY	output_buf,
    JDIMENSION	output_col)
{
  PROLOGUE;

  UInt8 inData[DCTSIZE2], *pinData, max, min;
  UInt8 outData[DCTSIZE2];
  UInt8 v_edge[DCTSIZE2], h_edge[DCTSIZE2];
  DCTELEM coefData[DCTSIZE2];
  JCOEF newCoef[DCTSIZE2];
  ISLOW_MULT_TYPE * quantptr, *pq;
  JCOEFPTR inptr, newptr;
  JSAMPROW outptr;
  int ctr, idx;
  Int32 ci;

  Int32    t0, t1, t2;   /* threshold delt */
  Int32    smooth, dis;

  quantptr = (ISLOW_MULT_TYPE *) compptr->dct_table;
  ci = compptr->component_index;

  smooth = ((Int32)3 * quantptr[0]) >>3;

  /* copy decompressed data to buffer */
  pinData = inData;
  max = 0;
  min = 255;
  for (ctr = 0; ctr < DCTSIZE; ctr++) 
  {
    outptr = output_buf[ctr] + output_col;
    MEMCOPY(pinData, outptr, DCTSIZE);
    for(idx=0; idx < DCTSIZE; ++idx)
    {
	if( pinData[idx] < min) min = pinData[idx];
	if( pinData[idx] > max) max = pinData[idx];
    }
    pinData += DCTSIZE;
  }
  t0 = (Int32) max;
  t1 = (Int32) min;
  dis = t0 - t1;

  if( dis > 0 )
  {
  /* delt type filter */
      t0 = ((dis<<2)+5)/10;
      if(dis <= smooth ) 
	  t1 = t2 = smooth;
      else 
      {
	  if( dis < q_dh[0][ci] ) t1 = (dis *6)/10;
	  else if( dis < q_dh[1][ci] ) t1 = (dis>>2) + q_dh[2][ci];
	  else t1 = ((dis * 3)/20) + q_dh[3][ci];
	  if( dis < q_dv[0][ci] ) t2 = (dis *6)/10;
	  else if( dis < q_dv[1][ci] ) t2 = (dis>>2) + q_dv[2][ci];
	  else t2 = ((dis * 3)/20) + q_dv[3][ci];
      }

      MEMZERO(v_edge, DCTSIZE2);
      ERR_PASS(findedge (pIPState, inData, v_edge, t0, VERTICAL, MODE0));
      ERR_PASS(v_filter (pIPState, inData, outData, v_edge, AVE_MODE));
      
      MEMZERO(h_edge, DCTSIZE2);
      ERR_PASS(findedge (pIPState, outData, h_edge, t1, HORIZONTAL, MODE1));
      ERR_PASS(h_filter (pIPState, inData, outData, h_edge, AVE_MODE));

      MEMZERO(v_edge, DCTSIZE2);
      ERR_PASS(findedge (pIPState, outData, v_edge, t2, VERTICAL, MODE1));
      for(ctr=0; ctr<3; ++ctr)
      {
	  ERR_PASS(h_filter (pIPState, inData, outData, h_edge, SLOPE_MODE)); 
	  ERR_PASS(v_filter (pIPState, outData, inData, v_edge, SLOPE_MODE));  
      }
  }
  else
  {
      NORMAL_RETURN;
  }

  /* forward DCT */
  ERR_PASS(jpeg_fdct_islow (pIPState, inData, coefData));

  /* new coef. modification */
  pq = qtbl_sdct[ci];
  inptr = coef_block;
  newptr = newCoef;
  for(ctr=0; ctr<DCTSIZE2; ++ctr )
  {
      *newptr = coefData[ctr] >> 3;
      t0 = *pq >>1;
      if( *newptr > (*inptr +t0))
	  *newptr = *inptr +t0;
      else if(*newptr < (*inptr -t0))
	  *newptr = *inptr -t0;
      ++pq;
      ++newptr;
      ++inptr;
  }

  /* inverse DCT */
  ERR_PASS(jpeg_idct_islow (pIPState, range_limit, compptr, newCoef,
			    output_buf, output_col));

  NORMAL_RETURN;
} 
 
#endif /* SMART_DCT_SUPPORTED */

/*
$Log: 

*/
