/*
 *	SPARC opcode lister
 *
 *	Convert SPARC Assembly code to something readable.
 *
 *	17-Jan-91 11:28:	MAC : Add new emulator opcodes for custom sparc
 *	30-Mar-90 16:29:06
 *	18-May-90 14:00:02	MAC : Adapt to multiple header format.
 *				  Also added -s option to print symbol table
 *				  and code to skip over other sections besides
 *				  just the text.  Handle un-stripped files.
 *	19-Mar-93		DD  : Converted to be a SPARC disassembler.
 *				  Now takes input from memory block instead
 *				  of a file.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "sparc.prv"	/* contains all info on SPARC ops */

    /* Prototypes */
#include "shrip.pub"
#include "shrip.prv"

#ifdef __GNUC__
#include "ansiprot.h"	/* for fprintf */
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: lister.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* The DAE typedefs have the "long" removed from the definition of Int32.
 * The fprintf's all insist that the numbers they are printing are
 * long.  Therefore, we need to explicitly cast the args in fprintf.
 * To do this, we'll make up an new type.
 */
typedef unsigned long int ULInt;

/* Character arrays used to disassemble opcodes */

static const char *op02[] = {
	"unimp0", "unimp1", "b", "unimp3", "sethi", "unimp5", "fb", "cb"
	};

static const char *icode[] = {
	"n", "e", "le", "l", "leu", "lu", "neg", "vs",
	"a", "ne", "g", "ge", "gu", "geu", "pos", "vc"
	};

/*	We don't decode coprocessor branches
static const char *cbcode[] = {
	"n", "123", "12", "13", "1", "23", "2", "3",
	"a", "0", "03", "02", "023", "01", "013", "012"
	};
*/

static const char *op23[] = {
	"add", "and", "or", "xor", "sub", "andn", "orn", "xnor",
	"addx", "unimp89", "unimp8A", "unimp8B", "subx", "unimp8D", "unimp8E", "unimp8F",
	"addcc", "andcc", "orcc", "xorcc", "subcc", "andncc", "orncc", "xnorncc",
	"addxcc", "unimp99", "unimp9A", "unimp9B", "subxcc", "unimp9D", "unimp9E", "unimp9F",
	"taddcc", "tsubcc", "taddcctv", "tsubcctv", "mulscc", "sll", "srl", "sra",
	"rdy", "rdpsr", "rdwim", "rdtbr", "unimpAC", "unimpAD", "unimpAE", "unimpAF",
	"wry", "wrpsr", "wrwim", "wrtbr", "fpop1", "fpop2", "cpop1", "cpop2",
	"jmpl", "rett", "t", "iflush", "save", "restore", "unimpBE", "unimpBF"
	};

static const char *op33[] = {
	"ld", "ldub", "ldu", "ldd", "st", "stb", "st", "std",
	"unimpC8", "ldsb", "ldsh", "unimpCB", "unimpCC", "ldstub", "unimpCE", "swap",
	"lda", "lduba", "lduha", "ldda", "sta", "stba", "stha", "stda",
	"unimpD8", "ldsba", "ldsha", "unimpDB", "unimpDC", "ldstuba", "unimpDE", "swapa",
	"ldf", "ldfsr", "unimpE2", "lddf", "stf", "stfsr", "stdfq", "stdf",
	"unimpE8", "unimpE9", "unimpEA", "unimpEB", "unimpEC", "unimpED", "unimpEE", "unimpEF",
	"ldc", "ldcsr", "unimpF2", "lddc", "stc", "stcsr", "stdcq", "stdc",
	"unimpF8", "unimpF9", "unimpFA", "unimpFB", "unimpFC", "unimpFD", "unimpFE", "unimpFF"
	};

static const char *reg[] = {
	"%g0", "%g1", "%g2", "%g3", "%g4", "%g5", "%g6", "%g7",
	"%o0", "%o1", "%o2", "%o3", "%o4", "%o5", "%o6", "%o7",
	"%l0", "%l1", "%l2", "%l3", "%l4", "%l5", "%l6", "%l7",
	"%i0", "%i1", "%i2", "%i3", "%i4", "%i5", "%i6", "%i7"
	};



/* This is a public procedure */
/* PUBLIC */
/*
 *  i_disassembleSPARCOps():  Interpret a block of memory as SPARC opcodes,
 *				Print the disassembled stuff to stdout.
 *
 * This routine interprets a block of memory as a group of SPARC opcodes.
 * It disassembles each one, printing the result to standard error.  It
 * does not attempt to convert branch offsets to addresses, it just
 * prints them.
 *
 *	pOp:		Pointer to UInt32's.  Each one is a SPARC op.
 *	opCount:	Number of SPARC ops to be disassembled.
 */
ErrCode CDECL 
i_disassembleSPARCOps (
    IPState    *pIPState,
    UInt32     *pOp,
    UInt32	opCount)
{
    PROLOGUE;
    for (; opCount > 0; opCount--)
    {
	ip_printSPARCOp (pIPState, (UInt32) pOp, *pOp);
	pOp++;
    }
    fprintf(stderr, "\n");
    NORMAL_RETURN;
}

/* PRIVATE */
/* This is a private procedure */
/*
 * ip_printSPARCOp disassembles and prints one opcode on stdout
 *
 *	addr:		address of opcode
 *	opcode:		the SPARC opcode
 */
void CDECL
ip_printSPARCOp(
    IPState    *pIPState,
    UInt32	addr,
    UInt32	opcode)
{
    PROLOGUE;
    UInt32	op, op3, op2;
    UInt32	rd, rs1, rs2;
    UInt32	annul, cond, disp22;
    UInt32	i, asi;
    UInt32	imm13, const22, disp30;
    UInt32	push, pop, storeOp, trap;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    push = 0;
    pop = 0;
    
    /* extract various fields */
    op = (opcode >> cOpPos) & 3;
    op2 = (opcode >> cOp2Pos) & 7;
    op3 = (opcode >> cOp3Pos) & 0x3f;
    rd = (opcode >> cRdPos) & 0x1f;
    rs1 = (opcode >> cRS1Pos) & 0x1f;
    rs2 = (opcode >> cRS2Pos) & 0x1f;
    annul = (opcode >> cAnnulPos) & 1;
    cond = (opcode >> cTestCPos) & 0xf;
    disp22 = opcode & 0x3fffff;
    const22 = opcode & 0x3fffff;
    disp30 = opcode & 0x3fffffff;
    i = (opcode >> cIsCPos) & 1;
    asi = (opcode >> cASIPos) & 0xff;
    imm13 = opcode & 0x1fff;

	/* store opcodes are regular in a weird sort of way */
    storeOp = ( (op == 3) && ((op3 & 0x0C) == 0x000100) );

    trap = (((op << 6) | op3) == 0xba);		/* non-zero means this is
						 * a trap opcode */

    /* print address and raw opcode */
    fprintf(stderr, "%08lx:  ", (ULInt) addr);			/* output address */
    fprintf(stderr, "%08lx	", (ULInt) opcode);

    /* output data */
    switch (op)
    {
	case 0:		/* branch or sethi */
	    switch (op2)
	    {
		case 2:	/* branch */
		case 6:	/* floating point branch */
		case 7:	/* coprocessor branch */
		    fprintf(stderr, "%s%s", op02[op2], icode[cond]);
		    if (annul) fprintf(stderr, ",a");
		    fprintf(stderr, "	0x%lx", (ULInt) disp22);
		    break;
		case 4:	/* sethi */
		    fprintf(stderr, "%s	0x%lx, %s",
				    op02[op2], (ULInt) const22, reg[rd]);
		    break;
		default:
		    fprintf(stderr, "%s", op02[op2]);
	    };
	    break;
	case 1:		/* call */
	    fprintf(stderr, "call	0x%lx", (ULInt) disp30);
	    break;
	case 2:		/* many regular opcodes */
	    if (i == 0)
	    {		/* two source registers */
		fprintf(stderr, "%s", op23[op3]);
		if (trap) fprintf(stderr, "%s", icode[cond]);
		fprintf(stderr, "	%s, %s", reg[rs1], reg[rs2]);
		if (!trap) fprintf(stderr, ", %s", reg[rd]);
	    }
	    else
	    {		/* 1 source register + constant */
		fprintf(stderr, "%s", op23[op3]);
		if (trap) fprintf(stderr, "%s", icode[cond]);
		fprintf(stderr, "	%s, ", reg[rs1]);
		fprintf(stderr, "0x%lx", (ULInt) imm13);
		if (!trap) fprintf(stderr, ", %s", reg[rd]);
	    }
	    break;
	case 3:		/* load/store opcodes */
	    if (i == 0)
	    {		/* two source registers */
		fprintf(stderr, "%s	", op33[op3]);
		if (!storeOp)
		{
		    fprintf(stderr, "[%s + %s]", reg[rs1], reg[rs2]);
		    if (asi) fprintf(stderr, "%ld", (ULInt) asi);
		    fprintf(stderr, ", %s", reg[rd]);
		}
		else
		{
		    fprintf(stderr, "%s, ", reg[rd]);
		    fprintf(stderr, "[%s + %s]", reg[rs1], reg[rs2]);
		    if (asi) fprintf(stderr, "%ld", (ULInt) asi);
		};
	    }
	    else
	    {		/* 1 source register + constant */
		fprintf(stderr, "%s	", op33[op3]);
		if (!storeOp)
		{
		    fprintf(stderr, "[%s + ", reg[rs1]);
		    fprintf(stderr, "0x%lx]", (ULInt) imm13);
		    fprintf(stderr, ", %s", reg[rd]);
		}
		else {
		    fprintf(stderr, "%s, ", reg[rd]);
		    fprintf(stderr, "[%s + ", reg[rs1]);
		    fprintf(stderr, "0x%lx]", (ULInt) imm13);
		};
	    }
    }
    fprintf(stderr, "\n");
}

/*

$Log: lister.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.14  1998/01/13 23:51:47  ddavies
lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.13  1996/02/23 00:58:50  ddavies
ddavies on Thu Feb 22 16:57:56 PST 1996

lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v

Add casts to long args in fprintf.  The DAE version of Int32 omits the long
so fprintf complains when it gets an int and is expecting a long.

Revision 1.12  1995/12/13 02:48:03  ejaquith
lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.11  1994/06/20  20:48:34  ddavies
 * ddavies on Mon Jun 20 13:47:56 PDT 1994
 *
 * lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v
 *
 * Add CDECL to declaration of i_disassembleSPARCOps.  This guarantees that
 * ALL routines have CDECLs, even if they don't use them.
 *
 * Revision 1.10  1994/05/03  00:54:28  jcraig
 * jcraig on Mon May 2 17:54:07 PDT 1994
 *
 * lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v
 *
 * add ifdef __GNUC__ around include of ansiprot.h
 *
 * Revision 1.9  1994/03/26  00:58:45  ddavies
 * lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v
 *
 * Change all the fprintf formats that print long ints or unsigned long ints.
 * They are all preceeded by a "l" to make gcc happy.
 *
 * Revision 1.8  1993/12/02  19:24:27  ddavies
 * lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.7  1993/11/05  01:49:07  ddavies
 * lister.c -> /product/ipcore/ipshared/src/RCS/lister.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


