/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miexpand.c: contains machine-independent primitives for expansion.
 *
 *         2x, 4x, 8x, 16x, 32x expansion procedures:
 *		Int32		ip_expand2Asm()
 *		Int32		ip_expand4Asm()
 *		Int32		ip_expand8Asm()
 *		Int32		ip_expand16Asm()
 *
 */


#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "iaerror.pub"
#include "imageref.pub"
#include "frames.pub"

    /*  prototypes */
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miexpand.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");
/* default: suppress all debugging */
#define  TIMING         0

/* Enable debugging info */
#ifdef  DEBUG_ON
#define  TIMING         1
#endif  /* DEBUG_ON */

/* The lookup tables used by the expansion code are generated in exptable.c
 */


/*********************************************************************
 ***************       2x Expansion Procedures       *****************
 ********************************************************************/

/* PRIVATE */
/*
 *  ip_expand2Asm()
 *		Puts expanded image data image region.  
 *              Each pixel in the source pr is mapped to a 2x2 square
 *              of 16 identical pixels in the dest.
 *    
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *		           (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *			   (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *              pTabExp2 = pointer to expansion table byte
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.  This is a C implementation which is
 *	normally an assembly routine - hence the name.
 */
Int32  CDECL 
ip_expand2Asm (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt16     *pTabExp2)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt	/* sun version */
{

    PROLOGUE;
    Int32       wD, hD;
    Int32       i; 			/* outer loop */
    Int32       j;			/* inner loop */
    UInt8      *pS,*psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt16     *pdWord;			/* inner loop */

    wD = w + w;
    hD = h + h;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the next
	 */
    for (i = 0; i < h; i++) {
        psByte = pS;
        pdWord = (UInt16 *)pD;
        for (j = 0; j < bpBorderlessLine; j++) {

                /* obtain 8 bits from the source */
                /* get the corresponding 16 bits from the table,
                 * and put it in the destination byte */
            *pdWord++ = pTabExp2[(Int32)(*psByte++)];
        }
        memcpy( (void *)(pD+dBpl), (void *)pD, dbpBorderlessLine );
	pS += sBpl;
	pD += (dBpl << 1);
    }

    NORMAL_RETURN;
}

#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt	/* PC VERSION */

{
    PROLOGUE;
    Int32	hS;      /* # of rows in source */
    Int32	j;       /* line counter */
    Int32	i;       /* source byte, dest word counter */
    UInt8      *sptr;    /* pointer into source buffer */
    Int32	prDWpl;  /* # of 16-bit words per dest line */
    UInt8      *pS; 	/* inner loop */
    UInt16      *pD;	/* inner loop */
    UInt8      *pDline;    /* destination pointer */
    Int32	destStep;
    Int32	dbpBorderlessLine;
 
    hS     = h;
    prDWpl = ( (w + w + 31) >> 5 ) << 1; /* # of short words */
    dbpBorderlessLine = prDWpl << 1;	/* number of visible dest bytes/line */

    destStep = dBpl + dBpl;

           
    pS     = ps;
    pDline = pd;
      
                /* for each line in the pixrects */ 
    for (j = 0; j < hS; j++) {
      

        sptr = pS;
	pD   = (UInt16 *)pDline;
              /*------------------------------------------------------------
               * Expand the lsb of the more significant pair of bytes first, 
               * then the msb of the more significant pair of bytes, 
               * then the lsb of the less significant pair of bytes, 
               * then the msb of the less significant pair of bytes. 
               * 
               * This is done so pixels will be in proper order on PC
               * and can be flipped into 'sun' format ok.
	       * BUG: we write 8 dest bytes each time through the loop,
	       * not 4 dest bytes.  Therefore, increment i by 8, not 4.
               *------------------------------------------------------------*/ 

        for (i = 0; i < dbpBorderlessLine; i += 8, sptr += 4) { 
              
            *pD++ = pTabExp2[(Int32) *(sptr + 2) ];
            *pD++ = pTabExp2[(Int32) *(sptr + 3) ]; 
            if ( i + 2 >= dbpBorderlessLine ) break;

            *pD++ = pTabExp2[(Int32) *sptr ]; 
            *pD++ = pTabExp2[(Int32) *(sptr + 1) ]; 
        }  
        memcpy( (void *)(pDline+dBpl), (void *)pDline, dbpBorderlessLine );
	pS     += sBpl;
	pDline += destStep;
    } 

    NORMAL_RETURN;
}

#endif
	    

/*********************************************************************
 ***************       4x Expansion Procedures       *****************
 ********************************************************************/

/* PRIVATE */
/*
 *  ip_expand4Asm()   
 *		Puts expanded image data in dest image region.
 *              Each pixel in the source image is mapped to a 4x4 square
 *              of 16 identical pixels in the dest.
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *			   (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *		           (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *              pTabExp4 = pointer to expansion table byte
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.  This is a C implementation which is
 *	normally an assembly routine - hence the name.
 */
Int32  CDECL 
ip_expand4Asm (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp4)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

{
    PROLOGUE;
    Int32       wD, hD;
    Int32       i; 			/* outer loop */
    Int32       j;			/* inner loop */
    UInt8      *pS, *psByte, *pD;
					     /* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;  /* inner loop */

    wD = w << 2;
    hD = h << 2;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following 3 destination rows.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0; j < bpBorderlessLine; j++)
	{

                /* obtain 8 bits from the source */
                /* get the corresponding 32 bits from the table,
                 * and put it in the destination byte */
            *pdWord++ = pTabExp4[(Int32)(*psByte++)];
        }
	/* copy this line to the next three lines.  Do this using one
	 * overlapped copy - i.e. copy from here to one line down, but
	 * do it for *Three* lines total.  This copies the first line to
	 * the second, then the (new) second to the third, then the (new)
	 * third to the fourth.
	 * BUG FIX: That trick doesn't work on Win32 in NT.  It is
	 * undefined behaviour in memcpy and Microsoft made it different,
	 * bless their hearts.
	 * So, we copy the first line to the second, third and fourth
	 * lines.  Originally, the code copied the first line to the
	 * second, then the first pair to the second pair.  While that
	 * was cuter, it does thrash the processor's cache more since
	 * it has to read the first two lines instead of just the
	 * first one.
	 */
	for (psByte = pD + dBpl; psByte < pD + 4*dBpl; psByte += dBpl)
	{
	    memcpy( (void *)psByte,
		    (void *)pD,
		     (dbpBorderlessLine + 3) & ~3);
	}
	pS += sBpl;
	pD += 4*dBpl;	/* skip four output lines */
    }

    NORMAL_RETURN;
}

#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

{
    PROLOGUE;
    Int32	wD, hD;
				    /* outer loop */
    Int32	i; 
    Int32	j;	/* inner loop */
    UInt8      *pS, *psByte, *pD;
					     /* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32      *pdWord;  /* inner loop */

    wD = w << 2;
    hD = h << 2;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following 3 destination rows.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0; j < bpBorderlessLine; j += 4, psByte += 4)
	{

                /* obtain 8 bits from the source,
                 * get the corresponding 32 bits from the table,
                 * and put them in the destination word.
		 * Since the PC naturally reads bytes from right to left
		 * within a 32 bit word, we must read index
		 * psByte to read them from left to right.  Reading
		 * from right to left keeps the image format the same
		 * as the SPARC's.
		 */
            *pdWord++ = pTabExp4[(Int32)(psByte[3])];
	    if (j+1 >= bpBorderlessLine) break;	/* quit if done */
            *pdWord++ = pTabExp4[(Int32)(psByte[2])];
	    if (j+2 >= bpBorderlessLine) break;	/* quit if done */
            *pdWord++ = pTabExp4[(Int32)(psByte[1])];
	    if (j+3 >= bpBorderlessLine) break;	/* quit if done */
            *pdWord++ = pTabExp4[(Int32)(psByte[0])];

        }
	/* copy this line to the next three lines.  Do this using one
	 * overlapped copy - i.e. copy from here to one line down, but
	 * do it for *Three* lines total.  This copies the first line to
	 * the second, then the (new) second to the third, then the (new)
	 * third to the fourth.
	 * BUG FIX: That trick doesn't work on Win32 in NT.  It is
	 * undefined behaviour in memcpy and Microsoft made it different,
	 * bless their hearts.
	 * So, we copy the first line to the second, third and fourth
	 * lines.  Originally, the code copied the first line to the
	 * second, then the first pair to the second pair.  While that
	 * was cuter, it does thrash the processor's cache more since
	 * it has to read the first two lines instead of just the
	 * first one.
	 */
	for (psByte = pD + dBpl; psByte < pD + 4*dBpl; psByte += dBpl)
	{
	    memcpy( (void *)psByte,
		    (void *)pD,
		    (dbpBorderlessLine + 3) & ~3);
	}

	pS += sBpl;
	pD += 4*dBpl;	/* skip four output lines */
    }

    NORMAL_RETURN;
}

#endif



/*********************************************************************
 ****************      8x  Expansion Procedures       ****************
 ********************************************************************/
/* PRIVATE */
/*
 *  ip_expand8Asm()
 *		Puts expanded image data image region.  
 *              Each pixel in the source pr is mapped to a 8x8 square
 *              of 64 identical pixels in the dest.
 *    
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *		           (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *		           (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *              pTabExp8 = pointer to expansion table byte
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.  This is a C implementation which is
 *	normally an assembly routine - hence the name.
 */
Int32  CDECL 
ip_expand8Asm (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp8)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
{
    PROLOGUE;
    Int32       wD, hD;
    Int32       i;			/* outer loop */
    Int32       j, dj;			/* inner loop */
    UInt8      *pS,*psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;			/* inner loop */
    Int32 	source;
    UInt32     *tabExp8Hi;		/* pts to upper half of expansion tab */

    wD = w << 3;
    hD = h << 3;

	/* point to upper half of expansion table */
    tabExp8Hi = pTabExp8 + 256;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j++, dj += 8)
	{

                /* obtain 8 bits from the source */
                /* get the corresponding 64 bits from the table,
                 * and put them in the destination words */
	    source = (Int32)(*psByte++);
            *pdWord++ = tabExp8Hi[source];
	    if (dj + 4 >= dbpBorderlessLine) break;	/* written last word? */
	    *pdWord++ = pTabExp8[source];
        }
	/* copy this line to the next seven lines.  Do this using one
	 * overlapped copy - i.e. copy from here to one line down, but
	 * do it for *Seven* lines total.  This copies the first line to
	 * the second, then the (new) second to the third, and so on.
	 * BUG FIX: That trick doesn't work on Win32 in NT.  It is
	 * undefined behaviour in memcpy and Microsoft made it different,
	 * bless their hearts.
	 * So, we copy the first line to the second, third and fourth
	 * lines.  Originally, the code copied the first line to the
	 * second, then the first pair to the second pair and so on.
	 * While that was cuter, it does thrash the processor's cache
	 * more since it has to read the first expansionFactor/2 lines
	 * instead of just the first one.
	 */
	for (psByte = pD + dBpl; psByte < pD + 8*dBpl; psByte += dBpl)
	{
	    memcpy( (void *)psByte,
		    (void *)pD,
		    (dbpBorderlessLine + 3) & ~3);
	}

	pS += sBpl;
	pD += 8*dBpl;	/* skip 8 dest lines */
    }

    NORMAL_RETURN;
}

	    
#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

{
    PROLOGUE;
    Int32	wD, hD;
    Int32	i;			/* outer loop */
    Int32	j, dj;			/* inner loop */
    UInt8      *pS, *psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;			/* inner loop */
    Int32	source;
    UInt32     *tabExp8Hi;		/* pts to upper half
					 * to expansion tab */
    wD = w << 3;
    hD = h << 3;

	/* point to upper half of expansion table */
    tabExp8Hi = pTabExp8 + 256;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j += 4, dj += 32, psByte += 4)
	{

                /* obtain 8 bits from the source */
                /* get the corresponding 64 bits from the table,
                 * and put them in the destination words.
		 * Since the PC goes through memory backwards, we have
		 * to  */
	    source = (Int32)(psByte[3]);
            *pdWord++ = tabExp8Hi[source];
	     if (dj + 4 >= dbpBorderlessLine) break;	/* written last word? */
	    *pdWord++ = pTabExp8[source];
	     if (dj + 8 >= dbpBorderlessLine) break;

	    source = (Int32)(psByte[2]);
            *pdWord++ = tabExp8Hi[source];
	     if (dj + 12 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp8[source];
	     if (dj + 16 >= dbpBorderlessLine) break;

	    source = (Int32)(psByte[1]);
            *pdWord++ = tabExp8Hi[source];
	     if (dj + 20 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp8[source];
	     if (dj + 24 >= dbpBorderlessLine) break;

	    source = (Int32)(psByte[0]);
            *pdWord++ = tabExp8Hi[source];
	     if (dj + 28 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp8[source];
        }
	/* copy this line to the next seven lines.  Do this using one
	 * overlapped copy - i.e. copy from here to one line down, but
	 * do it for *Seven* lines total.  This copies the first line to
	 * the second, then the (new) second to the third, and so on.
	 * BUG FIX: That trick doesn't work on Win32 in NT.  It is
	 * undefined behaviour in memcpy and Microsoft made it different,
	 * bless their hearts.
	 * So, we copy the first line to the second, third and fourth
	 * lines.  Originally, the code copied the first line to the
	 * second, then the first pair to the second pair and so on.
	 * While that was cuter, it does thrash the processor's cache
	 * more since it has to read the first expansionFactor/2 lines
	 * instead of just the first one.
	 */
	for (psByte = pD + dBpl; psByte < pD + 8*dBpl; psByte += dBpl)
	{
	    memcpy( (void *)psByte,
		    (void *)pD,
		    (dbpBorderlessLine + 3) & ~3);
	}

	pS += sBpl;
	pD += 8*dBpl;	/* skip 8 dest lines */
    }

    NORMAL_RETURN;
}


#endif /* ALAPCA_IMAGE_FORMAT
 */


/*********************************************************************
 ****************     16x  Expansion Procedures       ****************
 ********************************************************************/
/* PRIVATE */
/*
 *  ip_expand16Asm()
 *		Puts expanded image data image region.  
 *              Each pixel in the source pr is mapped to a 16x16 square
 *              of 256 identical pixels in the dest.
 *    
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *		           (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *		           (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *		pTabExp16= pointer to 16x expansion table
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.  This is a C implementation which is
 *	normally an assembly routine - hence the name.
 */
Int32  CDECL 
ip_expand16Asm (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp16)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

{
    PROLOGUE;
    Int32	wD, hD;
    Int32	i;			/* outer loop */ 
    Int32	j, dj;			/* inner loop */
    UInt8      *pS,*psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;			/* inner loop */
    Int32 	source;

    wD = w << 4;
    hD = h << 4;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j++, dj += 16)
	{

                /* obtain 8 bits from the source */
                /* get the corresponding 8*16 bits,
                 * and put them in the destination words */
	    source = (Int32)(*psByte++);
            *pdWord++ = pTabExp16[source + 3*256];
	    if (dj + 4 >= dbpBorderlessLine) break;	/* written last word? */
	    *pdWord++ = pTabExp16[source + 2*256];
	    if (dj + 8 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source + 256];
	    if (dj + 12 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source];
        }
	/* copy this line to the next fifteen lines.  Do this using one
	 * overlapped copy - i.e. copy from here to one line down, but
	 * do it for *Seven* lines total.  This copies the first line to
	 * the second, then the (new) second to the third, and so on.
	 * BUG FIX: That trick doesn't work on Win32 in NT.  It is
	 * undefined behaviour in memcpy and Microsoft made it different,
	 * bless their hearts.
	 * So, we copy the first line to the second, third and fourth
	 * lines.  Originally, the code copied the first line to the
	 * second, then the first pair to the second pair and so on.
	 * While that was cuter, it does thrash the processor's cache
	 * more since it has to read the first expansionFactor/2 lines
	 * instead of just the first one.
	 */
	for (psByte = pD + dBpl; psByte < pD + 16*dBpl; psByte += dBpl)
	{
	    memcpy( (void *)psByte,
		    (void *)pD,
		    (dbpBorderlessLine + 3) & ~3);
	}

	pS += sBpl;
	pD += 16*dBpl;	/* skip 16 dest lines */
    }

    NORMAL_RETURN;
}

	    
#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

{
    PROLOGUE;
    Int32	wD, hD;
    Int32	i;			/* outer loop */ 
    Int32	j, dj;			/* inner loop */
    UInt8      *pS,*psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;			/* inner loop */
    Int32	source;

    wD = w << 4;
    hD = h << 4;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j += 4, dj += 64, psByte += 4)
	{

                /* obtain 8 bits from the source */
                /* get the corresponding 8*16 bits,
                 * and put them in the destination words.
		 * Since the PC accesses bytes within the 32-bit
		 * word from right to left, we need to index psByte
		 * to get the byte we really want. */
	    source = (Int32)(psByte[3]);
            *pdWord++ = pTabExp16[source + 3*256];
	    if (dj + 4 >= dbpBorderlessLine) break;	/* written last word? */
	    *pdWord++ = pTabExp16[source + 2*256];
	    if (dj + 8 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source + 256];
	    if (dj + 12 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source];
	    if (dj + 16 >= dbpBorderlessLine) break;

	    source = (Int32)(psByte[2]);
            *pdWord++ = pTabExp16[source + 3*256];
	    if (dj + 20 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source + 2*256];
	    if (dj + 24 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source + 256];
	    if (dj + 28 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source];
	    if (dj + 32 >= dbpBorderlessLine) break;

	    source = (Int32)(psByte[1]);
            *pdWord++ = pTabExp16[source + 3*256];
	    if (dj + 36 >= dbpBorderlessLine) break;	/* written last word? */
	    *pdWord++ = pTabExp16[source + 2*256];
	    if (dj + 40 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source + 256];
	    if (dj + 44 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source];
	    if (dj + 48 >= dbpBorderlessLine) break;

	    source = (Int32)(psByte[0]);
            *pdWord++ = pTabExp16[source + 3*256];
	    if (dj + 52 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source + 2*256];
	    if (dj + 56 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source + 256];
	    if (dj + 60 >= dbpBorderlessLine) break;
	    *pdWord++ = pTabExp16[source];
        }
	/* copy this line to the next fifteen lines.  Do this using one
	 * overlapped copy - i.e. copy from here to one line down, but
	 * do it for *Seven* lines total.  This copies the first line to
	 * the second, then the (new) second to the third, and so on.
	 * BUG FIX: That trick doesn't work on Win32 in NT.  It is
	 * undefined behaviour in memcpy and Microsoft made it different,
	 * bless their hearts.
	 * So, we copy the first line to the second, third and fourth
	 * lines.  Originally, the code copied the first line to the
	 * second, then the first pair to the second pair and so on.
	 * While that was cuter, it does thrash the processor's cache
	 * more since it has to read the first expansionFactor/2 lines
	 * instead of just the first one.
	 */
	for (psByte = pD + dBpl; psByte < pD + 16*dBpl; psByte += dBpl)
	{
	    memcpy( (void *)psByte,
		    (void *)pD,
		    (dbpBorderlessLine + 3) & ~3);
	}

	pS += sBpl;
	pD += 16*dBpl;		/* skip 16 dest lines */
    }

    NORMAL_RETURN;
}

#endif /* _ALPACA_IMAGE_FMT_ */

/*

$Log: miexpand.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.6  1998/01/13 23:52:08  ddavies
miexpand.c -> /product/ipcore/ipshared/src/RCS/miexpand.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.5  1997/04/17 02:35:49  ddavies
ddavies on Wed Apr 16 19:34:02 PDT 1997

miexpand.c -> /product/ipcore/ipshared/src/RCS/miexpand.c,v

Eliminate unused prSBpl in PC version of ip_expand2Asm.  Also change
increment of i in PC version's inner loop - we process 8 dest bytes
per iteration, not 4.

Revision 1.4  1995/12/13 02:48:51  ejaquith
miexpand.c -> /product/ipcore/ipshared/src/RCS/miexpand.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.3  1994/07/27  15:51:31  ejaquith
 * miexpand.c -> /product/ipcore/ipshared/src/RCS/miexpand.c,v
 *
 * Reentrancy!
 * Removed globals and #include "iadef.prv"
 * Added pointers to table bytes to all functions in place of globals
 *
 * Revision 1.2  1994/05/18  21:36:41  ddavies
 * ddavies on Wed May 18 14:34:07 PDT 1994
 *
 * miexpand.c -> /product/ipcore/ipshared/src/RCS/miexpand.c,v
 *
 * ip_expand4Asm, ip_expand8Asm and ip_expand16Asm had been using an
 * undefined feature of memcpy to replicate image lines.  They would set
 * the source to line 1, the dest to line 2 and ask to copy 3 lines.  In
 * most implementations, this copies line 1 to line 2, then line 2 to line 3
 * and line 3 to line 4.  However, the Win32 implementation of memcpy in
 * NT does something else.  We now copy each line independently without
 * overlap.
 *
 * Revision 1.1  1994/01/06  03:29:04  ddavies
 * Initial revision
 *
 * Revision 1.24  1993/12/02  19:26:31  ddavies
 * sexpand.c -> /product/ipcore/ipshared/src/RCS/sexpand.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.23  1993/11/05  02:21:25  ddavies
 * ddavies on Thu Nov 4 18:21:12 PST 1993
 *
 * sexpand.c -> /product/ipcore/ipshared/src/RCS/sexpand.c,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.
 *

*/


