/*****************************************************************
 *  Copyright (c) 1995.                              		 *
 *  Xerox Corporation.  All rights reserved.                     *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miexprc.c: fast functions for binary column expansion
 *
 *	Private:
 *		Int32	ip_expand2Cols ()
 *		Int32	ip_expand4Cols ()
 *		Int32	ip_expand8Cols ()
 *		Int32	ip_expand16Cols ()
 *		Int32	ip_expand2Rows ()
 *		Int32	ip_expand4Rows ()
 *		Int32	ip_expand8Rows ()
 *		Int32	ip_expand16Rows ()
 *
 */


#include "types.pub"
#include "ipscore.prv"
#include "memory.pub"
#include "frames.pub"
#include "iaerror.pub"

#include "shrscal.pub"
#include "shrscal.prv"
#include "shrip.prv"


IP_RCSINFO(RCSInfo, "$RCSfile: miexprc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");


/* PRIVATE */
/*
 *  ip_expand2Cols
 *		Puts expanded image data in dest image region.  
 *              Each pixel in the source pr is mapped to 2
 *              identical pixels in the dest.
 *    
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *		           (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *			   (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *              pTabExp2 = pointer to expansion table byte
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.
 */
Int32  CDECL 
ip_expand2Cols (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt16     *pTabExp2)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt	/* sun version */
{

    PROLOGUE;
    Int32       wD, hD;
    Int32       i; 			/* outer loop */
    Int32       j;			/* inner loop */
    UInt8      *pS,*psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt16     *pdWord;			/* inner loop */

    wD = w + w;
    hD = h + h;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the next
	 */
    for (i = 0; i < h; i++) {
        psByte = pS;
        pdWord = (UInt16 *)pD;
        for (j = 0; j < bpBorderlessLine; j++) 
	{
	    /* obtain 8 bits from the source 
	     * get the corresponding 16 bits from the table,
	     * and put it in the destination byte */
            *pdWord = pTabExp2[(Int32)(*psByte)];
	    pdWord++;
	    psByte++;
        }
	pS += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;
}

#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt	/* PC VERSION */

{
    PROLOGUE;
    Int32	hS;      /* # of rows in source */
    Int32	j;       /* line counter */
    Int32	prSBpl;  /* bytes per source line */
    Int32	i;       /* source byte, dest word counter */
    UInt8      *sptr;    /* pointer into source buffer */
    Int32	prDWpl;  /* # of 16-bit words per dest line */
    UInt8      *pS; 	 /* inner loop */
    UInt16     *pD;	 /* inner loop */
    UInt8      *pDline;  /* destination pointer */
    Int32	dbpBorderlessLine;
 
    hS     = h;
    prSBpl = ( (w + 31) >> 5 ) << 2;	/* # of bytes = # of words * 4 */
    prDWpl = ( (w + w + 31) >> 5 ) << 1; /* # of short words */
    dbpBorderlessLine = prDWpl << 1;


    pS     = ps;
    pDline = pd;
      
                /* for each line in the pixrects */ 
    for (j = 0; j < hS; j++) {
      

        sptr = pS;
	pD   = (UInt16 *)pDline;
              /*------------------------------------------------------------
               * Expand the lsb of the more significant pair of bytes first, 
               * then the msb of the more significant pair of bytes, 
               * then the lsb of the less significant pair of bytes, 
               * then the msb of the less significant pair of bytes. 
               * 
               * This is done so pixels will be in proper order on PC
               * and can be flipped into 'sun' format ok. 
               *------------------------------------------------------------*/ 

        for (i = 0; i < dbpBorderlessLine; i += 8, sptr += 4) { 
              
            *pD++ = pTabExp2[(Int32) *(sptr + 2) ];
            *pD++ = pTabExp2[(Int32) *(sptr + 3) ]; 
            if ( i + 2 >= dbpBorderlessLine ) break;

            *pD++ = pTabExp2[(Int32) *sptr ]; 
            *pD++ = pTabExp2[(Int32) *(sptr + 1) ]; 
        }  
	pS     += sBpl;
	pDline += dBpl;
    } 

    NORMAL_RETURN; 
}

#endif
	    


/* PRIVATE */
/*
 *  ip_expand4Cols 
 *		Puts expanded image data in dest image region.
 *              Each pixel in the source image is mapped to 4
 *              identical pixels in the dest.
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *			   (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *		           (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *              pTabExp4 = pointer to expansion table byte
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.
 */
Int32  CDECL 
ip_expand4Cols (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp4)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

{
    PROLOGUE;
    Int32	wD, hD;
				    /* outer loop */
    Int32	i; 
    Int32	j;		    /* inner loop */
    UInt8      *pS, *psByte, *pD;
				    /* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;  /* inner loop */

    wD = w << 2;
    hD = h << 2;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following 3 destination rows.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0; j < bpBorderlessLine; j++)
	{
	    /* obtain 8 bits from the source
	     * get the corresponding 32 bits from the table,
	     * and put it in the destination byte */

            *pdWord = pTabExp4[(Int32)(*psByte)];
	    pdWord++;
	    psByte++;
        }
	pS += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;
}

#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

{
    PROLOGUE;
    Int32	wD, hD;
				    /* outer loop */
    Int32	i; 
    Int32	j;	/* inner loop */
    UInt8      *pS, *psByte, *pD;
					     /* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;  /* inner loop */

    wD = w << 2;
    hD = h << 2;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following 3 destination rows.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0; j < bpBorderlessLine; j += 4, psByte += 4)
	{
	    /* obtain 8 bits from the source,
	     * get the corresponding 32 bits from the table,
	     * and put them in the destination word.
	     * Since the PC naturally reads bytes from right to left
	     * within a 32 bit word, we must read index
	     * psByte to read them from left to right.  Reading
	     * from right to left keeps the image format the same
	     * as the SPARC's.
	     */
            *pdWord = pTabExp4[(Int32)(psByte[3])];
	    pdWord++;
	    if (j+1 >= bpBorderlessLine) 
	        break;	/* quit if done */
            *pdWord = pTabExp4[(Int32)(psByte[2])];
	    pdWord++;
	    if (j+2 >= bpBorderlessLine) 
	        break;	/* quit if done */
            *pdWord = pTabExp4[(Int32)(psByte[1])];
	    pdWord++;
	    if (j+3 >= bpBorderlessLine) 
	        break;	/* quit if done */
            *pdWord = pTabExp4[(Int32)(psByte[0])];
	    pdWord++;
        }

	pS += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;
}

#endif



/* PRIVATE */
/*
 *  ip_expand8Cols
 *		Puts expanded image data image region.  
 *              Each pixel in the source pr is mapped to 8
 *              identical pixels in the dest.
 *    
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *		           (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *		           (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *              pTabExp8 = pointer to expansion table byte
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.
 */
Int32  CDECL 
ip_expand8Cols (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp8)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
{
    PROLOGUE;
    Int32	wD, hD;
    Int32	i;			/* outer loop */
    Int32	j, dj;			/* inner loop */
    UInt8      *pS, *psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32      *pdWord;		/* inner loop */
    Int32	source;
    UInt32      *tabExp8Hi;		/* pts to upper half
					 * to expansion tab */

    wD = w << 3;
    hD = h << 3;

	/* point to upper half of expansion table */
    tabExp8Hi = pTabExp8 + 256;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j++, dj += 8)
	{
	    /* obtain 8 bits from the source
	     * get the corresponding 64 bits from the table,
	     * and put them in the destination words */
	    source = (Int32)(*psByte);
	    psByte++;
            *pdWord = tabExp8Hi[source];
	    pdWord++;
	    if (dj + 4 >= dbpBorderlessLine) 
	        break;	/* written last word? */
	    *pdWord = pTabExp8[source];
	    pdWord++;
        }

	pS += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;
}

	    
#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

{
    PROLOGUE;
    Int32	wD, hD;
				    
    Int32	i;			/* outer loop */
    Int32	j, dj;			/* inner loop */
    UInt8      *pS, *psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32      *pdWord;		/* inner loop */
    Int32	source;
    UInt32      *tabExp8Hi;		/* pts to upper half
					 * to expansion tab */
    wD = w << 3;
    hD = h << 3;

	/* point to upper half of expansion table */
    tabExp8Hi = pTabExp8 + 256;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j += 4, dj += 32, psByte += 4)
	{
	    /* obtain 8 bits from the source
	     * get the corresponding 64 bits from the table,
	     * and put them in the destination words.
	     * Since the PC goes through memory backwards, we have
	     * to.
	     */

	    source = (Int32)(psByte[3]);
            *pdWord = tabExp8Hi[source];
	    pdWord++;
	    if (dj + 4 >= dbpBorderlessLine) 
	        break;	/* written last word? */
	    *pdWord = pTabExp8[source];
	    pdWord++;
	    if (dj + 8 >= dbpBorderlessLine) 
	        break;

	    source = (Int32)(psByte[2]);
            *pdWord = tabExp8Hi[source];
            pdWord++;
	    if (dj + 12 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp8[source];
	    pdWord++;
	    if (dj + 16 >= dbpBorderlessLine) 
	        break;

	    source = (Int32)(psByte[1]);
            *pdWord = tabExp8Hi[source];
	    pdWord++;
	    if (dj + 20 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp8[source];
	    pdWord++;
	    if (dj + 24 >= dbpBorderlessLine) 
	        break;

	    source = (Int32)(psByte[0]);
            *pdWord = tabExp8Hi[source];
	    pdWord++;
	    if (dj + 28 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp8[source];
	    pdWord++;
        }

	pS += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;

}


#endif /* ALAPCA_IMAGE_FORMAT
 */


/* PRIVATE */
/*
 *  ip_expand16Cols
 *		Puts expanded image data image region.  
 *              Each pixel in the source pr is mapped to 16
 *              identical pixels in the dest.
 *    
 *	Input:  
 *		ps =       pointer to upper left corner of source binary image
 *		           (does not include frame)
 *		sBpl =     source bytes/line
 *		pd =       pointer to upper left corner of dest binary image
 *		           (does not include frame)
 *		dBpl =     destination bytes/line
 *		w =        width of source image ( does not include frame)
 *		h =        height of source image (does not include frame)
 *		pTabExp16= pointer to 16x expansion table
 *
 *	This low-level procedure requires that the correct values are input
 *	for pointers and dimensions.
 */
Int32  CDECL 
ip_expand16Cols (
    IPState    *pIPState,
    UInt8      *ps,
    Int32	sBpl,
    UInt8      *pd,
    Int32	dBpl,
    Int32	w,
    Int32	h,
    UInt32     *pTabExp16)

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

{
    PROLOGUE;
    Int32	wD, hD;
    Int32	i;			/* outer loop */ 
    Int32	j, dj;			/* inner loop */
    UInt8      *pS, *psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32      *pdWord;		/* inner loop */
    Int32	source;

    wD = w << 4;
    hD = h << 4;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j++, dj += 16)
	{
	    /* obtain 8 bits from the source
	     * get the corresponding 8*16 bits,
	     * and put them in the destination words 
	     */
	    source = (Int32)(*psByte);
	    psByte++;
            *pdWord = pTabExp16[source + 3*256];
	    pdWord++;
	    if (dj + 4 >= dbpBorderlessLine) 
	        break;	/* written last word? */
	    *pdWord = pTabExp16[source + 2*256];
	    pdWord++;
	    if (dj + 8 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source + 256];
	    pdWord++;
	    if (dj + 12 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source];
	    pdWord++;
        }

	pS += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;

}

	    
#elif _ALPACA_IMAGE_FMT_ == cAlpacaPCFmt

{
    PROLOGUE;
    Int32	wD, hD;
    Int32	i;			/* outer loop */ 
    Int32	j, dj;			/* inner loop */
    UInt8      *pS, *psByte, *pD;	/* inner loop */
    Int32	bpBorderlessLine;
    Int32	dbpBorderlessLine;
    UInt32     *pdWord;			/* inner loop */
    Int32	source;

    wD = w << 4;
    hD = h << 4;

    bpBorderlessLine = (w + 7) >> 3;
    dbpBorderlessLine = (wD + 7) >> 3;

    pS = ps;
    pD = pd;

	/* For each source row, we expand its columns into the
	 * next destination row; then we copy the destination
	 * row to the following destination row.  Finally we
	 * bump the source pointer to the start of the next
	 * row and the destination pointer to the start of the
	 * row after the last one touched.
	 */
    for (i = 0; i < h; i++)
    {
        psByte = pS;
        pdWord = (UInt32 *)pD;
        for (j = 0, dj = 0; j < bpBorderlessLine; j += 4, dj += 64, psByte += 4)
	{
	    /* obtain 8 bits from the source
	     * get the corresponding 8*16 bits,
	     * and put them in the destination words.
	     * Since the PC accesses bytes within the 32-bit
	     * word from right to left, we need to index psByte
	     * to get the byte we really want. 
	     */

	    source = (Int32)(psByte[3]);
            *pdWord = pTabExp16[source + 3*256];
	    pdWord++;
	    if (dj + 4 >= dbpBorderlessLine) 
	        break;	/* written last word? */
	    *pdWord = pTabExp16[source + 2*256];
	    pdWord++;
	    if (dj + 8 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source + 256];
	    pdWord++;
	    if (dj + 12 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source];
	    pdWord++;
	    if (dj + 16 >= dbpBorderlessLine) 
	        break;

	    source = (Int32)(psByte[2]);
            *pdWord = pTabExp16[source + 3*256];
	    pdWord++;
	    if (dj + 20 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source + 2*256];
	    pdWord++;
	    if (dj + 24 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source + 256];
	    pdWord++;
	    if (dj + 28 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source];
	    pdWord++;
	    if (dj + 32 >= dbpBorderlessLine) 
	        break;

	    source = (Int32)(psByte[1]);
            *pdWord = pTabExp16[source + 3*256];
	    pdWord++;
	    if (dj + 36 >= dbpBorderlessLine) 
	        break;	/* written last word? */
	    *pdWord = pTabExp16[source + 2*256];
	    pdWord++;
	    if (dj + 40 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source + 256];
	    pdWord++;
	    if (dj + 44 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source];
	    pdWord++;
	    if (dj + 48 >= dbpBorderlessLine) 
	        break;

	    source = (Int32)(psByte[0]);
            *pdWord = pTabExp16[source + 3*256];
	    pdWord++;
	    if (dj + 52 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source + 2*256];
	    pdWord++;
	    if (dj + 56 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source + 256];
	    pdWord++;
	    if (dj + 60 >= dbpBorderlessLine) 
	        break;
	    *pdWord = pTabExp16[source];
	    pdWord++;
        }

	pS += sBpl;
	pD += dBpl;
    }

    NORMAL_RETURN;

}

#endif /* _ALPACA_IMAGE_FMT_ */



/*
 $Log: miexprc.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.4  1998/01/13 23:52:11  ddavies
 miexprc.c -> /product/ipcore/ipshared/src/RCS/miexprc.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.3  1997/04/15 19:37:14  ddavies
 miexprc.c -> /product/ipcore/ipshared/src/RCS/miexprc.c,v

 Fix bug in ip_expand2Cols.  It was expanding twice as many pixels as
 necessary in each line, then overwriting the excess pixels in the next
 line.  This could produce trash in the frame.

 Revision 1.2  1995/12/13 02:49:02  ejaquith
 miexprc.c -> /product/ipcore/ipshared/src/RCS/miexprc.c,v

 CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.1  1995/02/28  00:00:36  dferrell
 * Initial revision
 *

 */

