/*****************************************************************
 *  Copyright (c) 1993, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* 
 *  miflush.c:  C implementation of lower level of cache flushing routines:
 *	   Int32    i_flushICache()
 */

#include <stddef.h>
#include <string.h>

#include "types.pub"
#include "ipscore.prv"

    /* prototypes */
#include "shrip.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: miflush.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* This file contains machine-independent versions of cache flushing
 * routines.  Since cache flushing is virtually always a machine-dependent
 * operation, these are dummy routines.  They're here for clients who
 * want to call them no matter what.
 */


/* PUBLIC */
/**************************************************************************
 *  i_flushICache()
 *		Flushes the instrution cache if one exists, otherwise does
 *		a bunch of nops, we hope.
 *
 *	Input: 
 *		pFlushAddr	Ptr to first location to be flushed
 *		bufLength	length of buffer to be flushed
 *
 *	Results:
 *		Attempts to flush the instruction cache.  We use the
 *		IFLUSH opcode on SPARC processors.  Some manufacturers
 *		may connect this so it causes an illegal_instruction trap.
 *		Bummer.
 *
 *	Returns:
 *		nothing
 **************************************************************************/
void  CDECL 
i_flushICache (
    IPState    *pIPState,
    void       *pFlushAddr,
    Int32	bufLength)
{
	/* This is a dummy version of the routine.
	 * It's only called directly from the DAE$.
	 */
    PROLOGUE;
    pFlushAddr = pFlushAddr;		/* keep compilers happy */
    bufLength = bufLength;
    pIPState = pIPState;
    __error__ = __error__;
}


/*

 $Log: miflush.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:52:16  ddavies
 miflush.c -> /product/ipcore/ipshared/src/RCS/miflush.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/05/29 02:11:34  ddavies
 Initial revision


 */

