/*****************************************************************
 *  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  migrybin.c: contains dummy primitives for combining gray and binary images
 *
 *         Combine Gray image with Binary image:
 *                  Int32     	ip_grayBinarySelLineLong()
 *                  Int32	ip_grayBinarySelLineLongNoFG()
 *                  Int32	ip_grayBinarySelLineLongNoBG()
 *                  Int32	ip_grayBinarySelLineLongNoFGNoBG()
 *                  Int32     	ip_gray4bppSelLineLong()
 *                  Int32	ip_gray4bppSelLineLongNoFG()
 *                  Int32	ip_gray4bppSelLineLongNoBG()
 *                  Int32	ip_gray4bppSelLineLongNoFGNoBG()
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"

/* prototypes */
#include "shrip.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: migrybn3.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");

/* Some routines are supplied only so the linker will be happy, but should
 * never be called.  When MMX code is present, the versions of the 
 * long routines implemented in assembly language will be called.  When
 * no assembly code is present, this file is linked in.  The calling
 * routine will see that there is no assembly code and will call the
 * ordinary reduction routines.  The only reason there routines exist
 * is to keep the linker happy.
 */

/* PRIVATE */
/*************************************************************************
 *  ip_grayBinarySelLineLong()
 *		The composite image has a foreground pixel wherever the
 *		selector has a 1 and a background pixel wherever the
 *		selector has a 0.
 *		This routine uses long (64-bit) operations.
 *
 *	Input:
 *		pBack =		pointer to gray background image (inside frame)
 *		pFore =		pointer to gray foreground image (inside frame)
 *		pDest =		pointer to destination image (inside frame)
 *		pSel =		pointer to binary selector image (inside frame)
 *		width =		width of all images in pixels
 *************************************************************************/
void  CDECL 
ip_grayBinarySelLineLong (
    IPState    *pIPState,
    UInt32     *pBack,
    UInt32     *pFore,
    UInt32     *pDest,
    UInt8      *pSel,
    Int32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pBack = pBack;
    pFore = pFore;
    pSel = pSel;
    pDest = pDest;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}


/* PRIVATE */
/*************************************************************************
 *  ip_grayBinarySelLineLongNoFG()
 *		The composite image has a virtual foreground pixel wherever the
 *		selector has a 1 and a background pixel wherever the
 *		selector has a 0.
 *		This routine uses long (64-bit) operations.
 *
 *	Input:
 *		pBack =		pointer to gray background image (inside frame)
 *		pDest =		pointer to destination image (inside frame)
 *		pSel =		pointer to binary selector image (inside frame)
 *		width =		width of all images in pixels
 *************************************************************************/
void  CDECL 
ip_grayBinarySelLineLongNoFG (
    IPState    *pIPState,
    UInt32     *pBack,
    UInt32     *pDest,
    UInt8      *pSel,
    Int32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pBack = pBack;
    pSel = pSel;
    pDest = pDest;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}


/* PRIVATE */
/*************************************************************************
 *  ip_grayBinarySelLineLongNoBG()
 *		The composite image has a foreground pixel wherever the
 *		selector has a 1 and a virtual background pixel wherever the
 *		selector has a 0.
 *		This routine uses long (64-bit) operations.
 *
 *	Input:
 *		pFore =		pointer to gray foreground image (inside frame)
 *		pDest =		pointer to destination image (inside frame)
 *		pSel =		pointer to binary selector image (inside frame)
 *		width =		width of all images in pixels
 *************************************************************************/
void  CDECL 
ip_grayBinarySelLineLongNoBG (
    IPState    *pIPState,
    UInt32     *pFore,
    UInt32     *pDest,
    UInt8      *pSel,
    Int32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pFore = pFore;
    pSel = pSel;
    pDest = pDest;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}



/* PRIVATE */
/*************************************************************************
 *  ip_grayBinarySelLineLongNoFGNoBG()
 *		The composite image has a virtual foreground pixel wherever the
 *		selector has a 1 and a virtual background pixel wherever the
 *		selector has a 0.
 *		This routine uses long (64-bit) operations.
 *
 *	Input:
 *		pDest =		pointer to destination image (inside frame)
 *		pSel =		pointer to binary selector image (inside frame)
 *		width =		width of all images in pixels
 *************************************************************************/
void  CDECL 
ip_grayBinarySelLineLongNoFGNoBG (
    IPState    *pIPState,
    UInt32     *pDest,
    UInt8      *pSel,
    Int32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pSel = pSel;
    pDest = pDest;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLineLong()
 *  		Combines a gray foreground and a gray background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose foreground pixel.  A 0xF in the selector
 *		=> choose background pixel.  Other values result in weighted
 *		combinations of foreground and background.
 *		This routine uses 64-bit accesses.
 *	Input:
 *		pBack =   	Pointer to left end of background source
 *				image line INSIDE frame
 *		pFore =   	Pointer to left end of foreground source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLineLong (
    IPState    *pIPState,
    UInt8      *pBack,
    UInt8      *pFore,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pBack = pBack;
    pFore = pFore;
    pSel = pSel;
    pDLong = pDLong;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLineLongNoFG()
 *  		Combines a virtual black foreground and a gray background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose foreground pixel.  A 0xF in the selector
 *		=> choose background pixel.  Other values result in weighted
 *		combinations of foreground and background.
 *		This routine uses 64-bit accesses.
 *	Input:
 *		pBack =   	Pointer to left end of background source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLineLongNoFG (
    IPState    *pIPState,
    UInt8      *pBack,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pBack = pBack;
    pSel = pSel;
    pDLong = pDLong;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLineLongNoBG()
 *  		Combines a foreground and a virtual white background image
 *		line together using a binary selector plane.  A 0 in the
 *		selector => choose foreground pixel.  A 0xF in the selector
 *		=> choose background pixel.  Other values result in weighted
 *		combinations of foreground and background.
 *		This routine uses 64-bit accesses.
 *	Input:
 *		pFore =   	Pointer to left end of foreground source
 *				image line INSIDE frame
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLineLongNoBG (
    IPState    *pIPState,
    UInt8      *pFore,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pFore = pFore;
    pSel = pSel;
    pDLong = pDLong;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}


/* PRIVATE */
/*************************************************************************** 
 *  ip_gray4bppSelLineLongNoFGNoBG()
 *  		Combines a virtual black foreground and a virtual white
 *		background image line together using a binary selector plane.
 *		A 0 in the selector => choose foreground pixel.  A 0xF in
 *		the selector => choose background pixel.  Other values
 *		result in weighted combinations of foreground and background.
 *		This routine uses 64-bit accesses.
 *	Input:
 *      	pSel =  	Pointer to left end of binary selector
 *				image line INSIDE frame.
 *		pDLong =	Pointer to left end of gray dest image line
 *				INSIDE frame.
 *      	width =		Width of dest image in pixels excluding frame
 *
 **************************************************************************/
void  CDECL 
ip_gray4bppSelLineLongNoFGNoBG (
    IPState    *pIPState,
    UInt8      *pSel,
    UInt32     *pDLong,
    UInt32	width)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pSel = pSel;
    pDLong = pDLong;
    width = width;
    pIPState = pIPState;
    __error__ = __error__;
}


/*

 $Log: migrybn3.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:15  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:52:30  ddavies
 migrybn3.c -> /product/ipcore/ipshared/src/RCS/migrybn3.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1997/06/06 04:20:04  ddavies
 Initial revision


 */

