/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *	migsrot1.c:
 *
 *          Low-level ops:
 *				ip_rotate4bppGrayLine();
 *
 */


#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "imageref.pub"
#include "memory.pub"

    /*  prototypes */
#include "shrip.pub"
#include "shrip.prv"
#include "shrpixr.pub"	/* contains types used by shrpixr.prv */
#include "shrpixr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: migsrot1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:15 $");
#define UNIBBLE_HIGH_ACCESS(uChar) ((((UInt8)(uChar)) >> 3) & 0x1E)
#define UNIBBLE_LOW_ACCESS(uChar) ((((UInt8)(uChar)) << 1) & 0x1E)
#define cMaxWhite4bpp 0x1E


/* PRIVATE */
/**************************************************************************
 *  ip_rotate4bppGrayLine()
 *		Generates one rotated line.
 *
 *	Input:
 *		pS =	pointer to beginning of next line of unrotated 
 *			gray data
 *		sBpl =	source bytes/line
 *		pD =	pointer to next line of rotated gray data
 *		w =	width of stipe in pixels
 *		vertSinRotAngle = used to adjust the X coord when moving
 *			vertically.
 *		horizSinRotAngle = used to adjust the Y coord when moving
 *			horizontally.
 *		xIOCoord = split integer: x position of current input
 *			pixel inoutput's frame of referecne
 *		yIOCoord = split integer: y position of current input
 *			pixel in output's frame of reference
 *		xCount = increment this count at each move towards the
 *			right.  While xCount < 0, we are off the image.
 *			distance to horizontal edge.
 *		yCount = increment this count at each move towards the
 *			bottom.  While yCount < 0, we are off the image.
 *			distance to vertical edge.
 *		nibble = which nibble in the byte should be used to start
 *              p4bppGrayRotationTable = ptr to gray rotation table
 *		hS =	height of source image.  This is down here because
 *			it's a late addition.
 **************************************************************************/
void  CDECL 
ip_rotate4bppGrayLine (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	sBpl,
    UInt8      *pD,
    Int32	w,
    Int32	vertSinRotAngle,
    Int32	horizSinRotAngle,
    Int32	xIOCoord,
    Int32	yIOCoord,
    UInt32	xCount,
    UInt32	yCount,
    Int32	nibble,
    UInt16     *p4bppGrayRotationTable,
    Int32	hS)
{
    PROLOGUE;
    UInt32      gray0, gray1;		/* do 8 pixels at a time */
    UInt8	inputTop, inputBottom;	/* input pixels */
    UInt8      *pTopLineTab, *pBotLineTab;	/* pointers into
					         * grayRotationTable. */
    Int32	col;			/* iteration variables */
    Int32	xMask;			/* used to select 3 MS fraction
					 * bits from shifted xIOCoord */
    Int32	xInc;			/* used to increment integer
					 * part of xIOCoord */
    UInt32	grayTemp;		/* holds value returned from
					 * gray rotate table */
    Int32	xVInc;			/* holds amount by which we
					 * change xIOCoord for vertical
					 * movement. */

#define cXFracMask4bpp		0x700	/* selects 3 MS X frac bits */
#define cXIOCoordInc4bpp	0x800	/* used to increment xIOCoord */
#define	cYFracMask4bpp		0xE0	/* selects 3 MS Y frac bits */
#define cYIOCoordInc4bpp	0x100	/* used to increment yIOCoord */

	/* It would be polite to check for all sorts of error conditions
	 * here, but the assembly routine doesn't check so we won't either.
	 * Nonetheless, keep the compiler happy.
	 */
    pIPState = pIPState;
    __error__ = __error__;

	/* Just because it's important in what's coming up, let's set the
	 * format of the table lookup index:
	 * |  10    9   8      |  7     6      5   | 4 3 2 1 | 0
	 * | 3b xIOCoord (fff) | 3b yIOCoord (fff) | pixel   |
	 *
	 * The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
	 * yIOCoord bits are its 3 MS fraction bits.
	 */

	/* We'd like to have the three MS fraction bits in xIOC aligned with
	 * their position in the table lookup.
	 */
    xIOCoord = xIOCoord >> 5;
    xVInc = vertSinRotAngle >> 5;
    xMask = cXFracMask4bpp;	/* used to select the 3 fraction bits */
    xInc = cXIOCoordInc4bpp;		/* used to increment integer part */

	/* Ditto for yIOCoord.  This leaves us with 8 bits of fraction. */
    yIOCoord = yIOCoord >> 8;

	/* The horizSinRotAngle quantity is added to yIOCoord in the inner loop.
	 * Align it with yIOCoord to make this easier. */
    horizSinRotAngle = horizSinRotAngle >> 8;

	/* we initialize the regs making up the gray box to 0 since no
	 * input pixels have been inserted yet.
	 */
    gray0 = gray1 = 0;

	/* Looks like we're going to have to start writing real inner loop
	 * code pretty soon.  Hard to put it off much longer.  But wait!
	 * We can still discuss how many flavors of loop we need.
	 * There are two types -  one checks xCount and yCount on
	 * every read from the input image because we're near a boundary, and
	 * one is used when we're totally inside the input image.
	 *
	 */
    for (col = 0; col < w; ) /* loop through dest cols */
    {

	/* which inner loop shall we use? */
	if (((yCount+2)  >= (UInt32) hS) ||
	    ((yCount-3)  >= (UInt32) hS) ||
	    ((xCount+10) >= (UInt32) w) ||
	    ((xCount-1)  >= (UInt32) w))

	    {	/* at least part of the input is outside the input image */

		/* There are two types of inner-inner loop.  One starts in
		 * col 0 and one starts in col -1.  Both types end their
		 * sweeps when they reach col 7.  If start from col -1, we
		 * remember to skip the first block of the loop.
		 */

		/* The Y portion of rTopLineTab should be inverted so
		 * it references the portion of the pixel that lies
		 * in the *next* line.  The ! Y portion of rBotLineTab 
		 * should not be inverted so it get the portion of the 
		 * input pixel that lies in the current line.  We 
		 * implement the inversion by adding the maximum Y coord 
		 * (7) to pTopLineTab and subtracting each Y coord
		 * from this pointer.
		 */

		/* get ptr to portion of table that will be used for
		 * for this xIOCoord.
		 */
	    pBotLineTab = (UInt8 *) p4bppGrayRotationTable +
			  (xIOCoord & xMask);
	    pTopLineTab = pBotLineTab + cYFracMask4bpp;
	    if (xIOCoord >= 0)
		{	/* backup one column */
		xCount--;
		nibble = nibble ^ 1;
		if (nibble == 1)
		    pS--;
		xIOCoord -= xInc;
		yIOCoord -= horizSinRotAngle;
		    /* did that take us above row -1 */
		if ((yIOCoord + cYIOCoordInc4bpp) < 0)
		    {	/* We went up too high, so move down one input line */
		    pS +=sBpl;
		    yCount++;
		    yIOCoord += cYIOCoordInc4bpp;
		    xIOCoord -= xVInc;
		    }
		}

	    if (nibble == 0)
		{
		    /*
		     * We're calculating pixel -1 relative to the output block.  
		     * Check to see if we're on the input image or not.
		     */
                if ((xCount >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 24;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[0]))])) << 24;
		    }
                if ((xCount >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {	/* lower pix is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 24;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl]))])) << 24;
		    }
	        if ( (yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 0 */
                if (((xCount+1) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[0]))])) << 16;
		    }
                if (((xCount+1) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+0]))])) << 16;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }
	    
		    /* column 1 */
                if (((xCount+2) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[1]))])) << 8;
		    }
                if (((xCount+2) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+1]))])) << 8;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 2 */
                if (((xCount+3) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[1]))]));
		    }
                if (((xCount+3) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+1]))]));
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 3 */
                if (((xCount+4) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
	            grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* top pixel is on the image */
	            grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[2]))]));
		    }
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
                if (((xCount+4) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {	/* lower pix is off the image so use max white value */
		    grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* lower pixel is on the image */
	            grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+2]))]));
		    }
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
		if ( (yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 4 */
                if (((xCount+5) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[2]))])) << 16;
		    }
                if (((xCount+5) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+2]))])) << 16;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 5 */
                if (((xCount+6) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[3]))])) << 8;
		    }
                if (((xCount+6) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+3]))])) << 8;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 6 */
                if (((xCount+7) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[3]))]));
		    }
                if (((xCount+7) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+3]))]));
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 7 */
                if (((xCount+8) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) >> 8;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[4]))])) >> 8;
		    }
                if (((xCount+8) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) >> 8;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+4]))])) >> 8;
		    }
		/* nibble = 1; */
		    /* Note that we didn't try to advance the pointers.
		     * That's because we're already in the right position
		     * to start the next sweep.
		     */

	        }
	    else	/* nibble = 1 */
		{
		    /*
		     * We're calculating pixel -1 relative to the output block.  
		     * Check to see if we're on the input image or not.
		     */
                if ((xCount >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 24;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[0]))])) << 24;
		    }
                if ((xCount >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {	/* lower pix is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 24;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl]))])) << 24;
		    }
	        if ( (yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 0 */
                if (((xCount+1) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[1]))])) << 16;
		    }
                if (((xCount+1) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+1]))])) << 16;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }
	    
		    /* column 1 */
                if (((xCount+2) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[1]))])) << 8;
		    }
                if (((xCount+2) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+1]))])) << 8;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 2 */
                if (((xCount+3) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* top pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[2]))]));
		    }
                if (((xCount+3) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray0 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+2]))]));
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 3 */
                if (((xCount+4) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
	            grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* top pixel is on the image */
	            grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[2]))]));
		    }
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
                if (((xCount+4) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {	/* lower pix is off the image so use max white value */
		    grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* lower pixel is on the image */
	            grayTemp = (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+2]))]));
		    }
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
		if ( (yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 4 */
                if (((xCount+5) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[3]))])) << 16;
		    }
                if (((xCount+5) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 16;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+3]))])) << 16;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 5 */
                if (((xCount+6) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[3]))])) << 8;
		    }
                if (((xCount+6) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) << 8;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+3]))])) << 8;
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 6 */
                if (((xCount+7) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[4]))]));
		    }
                if (((xCount+7) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp]));
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+4]))]));
		    }
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 7 */
                if (((xCount+8) >= (UInt32) w) || (yCount >= (UInt32) hS))
		    {	/* top pixel is off the image so use max white value */
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) >> 8;
		    }
	        else
		    {	/* top pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) + 
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[4]))])) >> 8;
		    }
                if (((xCount+8) >= (UInt32) w) || ((yCount+1) >= (UInt32) hS))
		    {
		    gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
		        cMaxWhite4bpp])) >> 8;
		    }
	        else
		    {	/* lower pixel is on the image */
	            gray1 += (UInt32) 
	              ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	                UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl+4]))])) >> 8;
		    }
		/* nibble = 0; */
		    /* Note that we didn't try to advance the pointers.
		     * That's because we're already in the right position
		     * to start the next sweep.
		     */

	        }
	    }
	else
	    {
		/* This is the normal inner loop */
		/* All the comments for the checking loop apply here.
		 * The only difference is that we don't test xCount and
		 * yCount on each input sample.  Note that we also
		 * can just freely increment or decrement yCount here
		 * since we're guaranteed to be not near the top or
		 * bottom of the image.
		 */

		/* set up current table pointer */
	    pBotLineTab = (UInt8 *) p4bppGrayRotationTable +
			      (xIOCoord & xMask);
	    pTopLineTab = pBotLineTab + cYFracMask4bpp;

		/* Here's where we jump to the correct block of the
		 * checking loop.  the block is based on the integer 
		 * part of the xIOCoord since each block calculated a 
		 * byte.
		 */
	    if (xIOCoord >= 0)
		{	/* we must be in column 0.  Back up 1 column */
		xCount--;
		nibble = nibble ^ 1;
		if (nibble == 1)
		    pS--;
		xIOCoord -= xInc;
		yIOCoord -= horizSinRotAngle;
		if ((yIOCoord + cYIOCoordInc4bpp)  < 0)
		    {	/* we are above row -1 */
		    pS += sBpl;
		    yCount++;
		    yIOCoord += cYIOCoordInc4bpp;
		    xIOCoord -= xVInc;
		    }
		}

	    if (nibble == 0)
		{
		        /* column -1 */
		inputTop = UCHAR_ACCESS(&pS[0]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl]);
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)])) << 24;
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)])) << 24;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[0]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl]);
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 0 */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)])) << 16;
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)])) << 16;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 1 */
		inputTop = UCHAR_ACCESS(&pS[1]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl+1]);
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)])) << 8;
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)])) << 8;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[1]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl+1]);
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 2 */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)]));
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)]));
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 3 */
		inputTop = UCHAR_ACCESS(&pS[2]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl+2]);
	        grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)]));
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
	        grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)]));
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[2]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl+2]);
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 4 */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)])) << 16;
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)])) << 16;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 5 */
		inputTop = UCHAR_ACCESS(&pS[3]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl+3]);
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)])) << 8;
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)])) << 8;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[3]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl+3]);
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 6 */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)]));
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)]));
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 7 */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[4]))])) >> 8;
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(UCHAR_ACCESS(&pS[sBpl+4]))])) >> 8;

		    /* Note that we dind't try to advance the pointers.
		     * They're already in the right position to start the
		     * next sweep.
		     */
		/* nibble = 1; */
		}
	    else
		{
		        /* column -1 */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[0]))])) << 24;
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(UCHAR_ACCESS(&pS[sBpl]))])) << 24;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 0 */
		inputTop = UCHAR_ACCESS(&pS[1]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl+1]);
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)])) << 16;
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)])) << 16;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[1]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl+1]);
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 1 */
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)])) << 8;
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)])) << 8;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
		    xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 2 */
		inputTop = UCHAR_ACCESS(&pS[2]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl+2]);
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)]));
	        gray0 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)]));
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[2]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl+2]);
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 3 */
	        grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)]));
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
	        grayTemp = (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)]));
	        gray0 += grayTemp >> 8;
	        gray1 += grayTemp << 24;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 4 */
		inputTop = UCHAR_ACCESS(&pS[3]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl+3]);
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)])) << 16;
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)])) << 16;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[3]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl+3]);
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 5 */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)])) << 8;
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)])) << 8;
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 6 */
		inputTop = UCHAR_ACCESS(&pS[4]);
		inputBottom = UCHAR_ACCESS(&pS[sBpl+4]);
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputTop)]));
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_HIGH_ACCESS(inputBottom)]));
	        if ((yIOCoord += horizSinRotAngle) >= 0)
		    {	/* entered row 0.  Move up one input row. */
		    pS -= sBpl;
		    inputTop = UCHAR_ACCESS(&pS[4]);
		    inputBottom = UCHAR_ACCESS(&pS[sBpl+4]);
		    yIOCoord -= cYIOCoordInc4bpp;
	            xIOCoord += xVInc;
		    yCount--;
		    }

		    /* column 7 */
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pTopLineTab[-(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputTop)])) >> 8;
	        gray1 += (UInt32) 
	          ( *(UInt16 *) &(pBotLineTab[(yIOCoord & cYFracMask4bpp) +
	            UNIBBLE_LOW_ACCESS(inputBottom)])) >> 8;

		    /* Note that we dind't try to advance the pointers.
		     * They're already in the right position to start the
		     * next sweep.
		     */
		/* nibble = 0; */
		}

	    }

	    /* We've completed the formation of this row.  Write it
	     * out.  Note that we always write gray0.  If the line 
	     * ends after that, don't write gray1.
	     */
	gray0 = (gray0 + 0x08080808) & 0xF0F0F0F0;
	/* gray0 &= 0xF0F0F0F0; */
	gray0 |= (gray0 << 4);
	gray1 = (gray1 + 0x08080808) & 0xF0F0F0F0;
	/* gray1 &= 0xF0F0F0F0; */
	gray1 = (gray1 >> 8) | (gray1 >> 4);
	grayTemp = (((gray0 & 0x0000FF00) << 8) | (gray0 & 0xFF000000)) |
		   (((gray1 & 0x00FF00FF) >> 8) | (gray1 & 0x000000FF));
	ULONG_ACCESS(&(pD[col>>1])) = grayTemp;
	col += 8;
	gray0 = gray1 = 0;
	pS += 4;
	xCount += 8;

    }	/* end of loop that processes all columns in the line */
}	/* end of ip_rotate4bppGrayLine */


/*

$Log: migsrot1.c,v $
Revision 1.1.1.1  1999/09/20 18:18:15  jlin
Imported sources

Revision 1.10  1998/01/13 23:52:33  ddavies
migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.9  1996/10/05 02:22:48  ddavies
migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v

Change the semantics of xCount and yCount.  Instead of trying to arrange
for them to be negative outside the image, we just to unsigned compares of
xCount >= width and yCount >= height.  If these are true, we're off the image.
Also get rid of the fixup of xCount at the left boundary.

Revision 1.8  1995/12/13 02:50:04  ejaquith
migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.7  1995/12/13  01:52:45  ddavies
 * migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v
 *
 * In keeping with the new, resolution-independent, method representing skew
 * angles, use two versions of sinRotAngle - horizSinRotAngle is used to adjust
 * y coords when moving horizontally and vertSinRotAngle is used to adjust
 * x coords when moving vertically.  Also fix a bug in which xCount was
 * being adjusted for use to detect the right edge of the image when it was
 * equal to 0.  In this case, it may still be necessary to read at pixel with
 * x coord = -1, but we wouldn't substitute a white pixel since xCount would
 * have been changed to be relative to the right edge of the image instead of
 * being relative to the left edge.
 *
 * Revision 1.6  1995/10/18  04:03:32  ddavies
 * migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
 * Revision 1.5  1994/11/07  22:06:12  ddavies
 * migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v
 *
 * Revision 1.4  1994/07/15  23:39:30  ejaquith
 * migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v
 *
 * Removed #include "iadef.prv"
 *
 * Revision 1.3  1994/02/25  02:33:43  dferrell
 * migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v
 *
 * Changed parameters for ip_rotateGrayLine4bpp.
 * xCount and yCount are now UInt32 rather than Int32.  This allowed me to
 * fix a problem on the Mac with overflow/wrap-around detection of signed
 * integers.
 * Modified checking for being close to the edge (Yes!) of the source image
 * by looking at xCount and yCount.  We now compare them to a constant instead
 * of checking for an overflow in a signed int.
 *
 * Revision 1.2  1994/02/04  20:18:20  dferrell
 * migsrot1.c -> /product/ipcore/ipshared/src/RCS/migsrot1.c,v
 *
 * Changed #include "memory.prv" to #include "memory.pub"
 *
 * Revision 1.1  1994/02/02  23:54:57  ddavies
 * Initial revision
 *
 * Revision 1.10  1993/12/02  19:20:28  ddavies
 * cgsrot.c -> /product/ipcore/ipshared/src/RCS/cgsrot.c,v
 *
 * Put RCS magic into macro that doesn't cause compiler warnings.
 *
 * Revision 1.9  1993/11/05  01:23:24  ddavies
 * cgsrot.c -> /product/ipcore/ipshared/src/RCS/cgsrot.c,v
 *

*/


