/*****************************************************************
 *  Copyright (c) 1996 Xerox Corporation.  All rights reserved.  *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  mimatvec.c: matrix / vector multiply routines
 *
 *		Bool	ip_longOpMathLibLinked()
 *
 *		Int32	ip_multMatrix8Vector32()
 *
 */


#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"
#include "iaerror.pub"

/* prototypes */
#include "fxpmath.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: mimatvec.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");


/* PRIVATE */
/*************************************************************************
 *  ip_longOpMathLibLinked()
 *		This routine returns cTrue if accelerated code that can
 *		do long (64 bit or greater) are linked..
 *
 *************************************************************************/
Bool  CDECL 
ip_longOpMathLibLinked (
    IPState    *pIPState)
{
	/* If we're here, there are no fancy assembly code libraries
	 * linked, so return cFalse.
	 */
    PROLOGUE;
    __error__ = __error__;	/* keep the compiler happy */
    pIPState = pIPState;

    return cFalse;
}


/* PRIVATE */
/*************************************************************************
 *  ip_multMatrix8Vector32()
 *		This routine does the real work of multiplying a vector
 *		by a matrix of weights, producing another vector.  It
 *		may be implemented in some enhanced assembly language.
 *		See i_matvec_mult_8BA.
 *
 *	Input:
 *		pMatrix =	pointer to upper left corner of weight matrix
 *		pInVector =	ptr to first element of input vector.
 *		pOutVector =	ptr to first element of output vector.
 *		numRows =	number of rows in weight matrix.  This
 *				is also the number of outputs to produce.
 *		numCols =	number of columns in weight matrix.  This is
 *				also the length of the input vector.
 *************************************************************************/
void  CDECL 
ip_multMatrix8Vector32 (
    IPState    *pIPState,
    Int8       *pMatrix,
    Int32      *pInVector,
    Int32      *pOutVector,
    Int32	numRows,
    Int32	numCols)
{
	/* this is a dummy routines.  It's only here to keep the
	 * linker happy.  It should never be called.  The caller
	 * is supposed to see if MMX code is available.  If so,
	 * call it.  If not, just do the op in the caller.
	 */
    PROLOGUE;
    pMatrix = pMatrix;
    pInVector = pInVector;
    pOutVector = pOutVector;
    numRows = numRows;
    numCols = numCols;
    pIPState = pIPState;
    __error__ = __error__;
}

/*

$Log: mimatvec.c,v $
Revision 1.1.1.1  1999/09/20 18:18:16  jlin
Imported sources

Revision 1.2  1998/01/13 23:52:43  ddavies
mimatvec.c -> /product/ipcore/ipshared/src/RCS/mimatvec.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.1  1996/10/30 05:53:29  ddavies
Initial revision


 */

