/*****************************************************************
 *  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  miredm.c: Dummy C implementations of column reduction primitives
 *
 *	     Dummy procedures (should never be called)
 *		void	ip_compressColsOrLineLong()
 *		void	ip_compressColsAndLineLong()
 *
 *		void	ip_compress4ColsOrLineLong()
 *		void	ip_compress4ColsAndLineLong()
 *
 *		void	ip_compress8ColsOrLineLong()
 *		void	ip_compress8ColsAndLineLong()
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "types.pub"
#include "ipscore.prv"
#include "defines.pub"

/* prototypes */
#include "shrscal.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miredcm.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:16 $");


/* Some routines are supplied only so the linker will be happy, but should
 * never be called.  When MMX code is present, the versions of the 
 * long routines implemented in assembly language will be called.  When
 * no assembly code is present, this file is linked in.  The calling
 * routine will see that there is no assembly code and will call the
 * ordinary reduction routines.  The only reason there routines exist
 * is to keep the linker happy.
 */

/* PRIVATE */
/*************************************************************************
 *  ip_compressColsOrLineLong()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the two corresponding pixels in the src 
 *		image are ON.  The two pixels are in adjacent columns.
 *		This is called by ip_compressColsOr when the processor
 *		supports long operations.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *************************************************************************/
void  CDECL 
ip_compressColsOrLineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
}


/* PRIVATE */
/*************************************************************************
 *  ip_compressColsAndLineLong()
 *		The reduced image line has an ON pixel whenever
 *		both of the corresponding pixels in the src 
 *		image are ON.  The two pixels are in adjacent columns.
 *		This is called by ip_compressColsAnd when the processor
 *		supports long operations.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *************************************************************************/
void  CDECL 
ip_compressColsAndLineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
}


/* PRIVATE */
/*************************************************************************
 *  ip_compress4ColsOrLineLong()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the four corresponding pixels in the src 
 *		image are ON.  The four pixels are in adjacent columns.
 *		This is called by ip_compressColsOr when the processor
 *		supports long operations.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *************************************************************************/
void  CDECL 
ip_compress4ColsOrLineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
}


/* PRIVATE */
/*************************************************************************
 *  ip_compress4ColsAndLineLong()
 *		The reduced image line has an ON pixel whenever
 *		all four of the corresponding pixels in the src 
 *		image are ON.  The four pixels are in adjacent columns.
 *		This is called by ip_compressColsAnd when the processor
 *		supports long operations.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *************************************************************************/
void  CDECL 
ip_compress4ColsAndLineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
}



/* PRIVATE */
/*************************************************************************
 *  ip_compress8ColsOrLineLong()
 *		The reduced image line has an ON pixel whenever
 *		at least 1 of the eight corresponding pixels in the src 
 *		image are ON.  The eight pixels are in adjacent columns.
 *		This is called by ip_compressColsOr when the processor
 *		supports long operations.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *************************************************************************/
void  CDECL 
ip_compress8ColsOrLineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
}


/* PRIVATE */
/*************************************************************************
 *  ip_compress8ColsAndLineLong()
 *		The reduced image line has an ON pixel whenever
 *		all eight of the corresponding pixels in the src 
 *		image are ON.  The eight pixels are in adjacent columns.
 *		This is called by ip_compressColsAnd when the processor
 *		supports long operations.
 *
 *	Input:
 *		pS =		pointer to source image (inside frame)
 *		srcWidth =	width of source in pixels
 *		pD =		pointer to destination image (inside frame)
 *      	dstOffset =	Offset in bytes to rightmost dest word
 *		srcOffset =	Offset in bytes to leftmost byte in rightmost
 *				group of source bytes that will form the
 *				rightmost dest word
 *		srcLimit =	Offset in bytes to rightmost source byte.
 *************************************************************************/
void  CDECL 
ip_compress8ColsAndLineLong (
    IPState    *pIPState,
    UInt8      *pS,
    Int32	srcWidth,
    UInt8      *pD,
    Int32	dstOffset,
    Int32	srcOffset,
    Int32	srcLimit)
{
	/* This is a dummy routine, it should never be called.
	 */
    PROLOGUE;
    pIPState = pIPState;
    __error__ = __error__;
    srcWidth = srcWidth;
    pS = pS;
    pD = pD;
    dstOffset = dstOffset;
    srcOffset = srcOffset;
    srcLimit = srcLimit;
}



/*

 $Log: miredcm.c,v $
 Revision 1.1.1.1  1999/09/20 18:18:16  jlin
 Imported sources

 Revision 1.2  1998/01/13 23:53:05  ddavies
 miredcm.c -> /product/ipcore/ipshared/src/RCS/miredcm.c,v

 CommonCheck: Add IPState parameter to almost all procedures.

 Revision 1.1  1996/11/04 21:47:56  ddavies
 Initial revision


 */
