/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * miscgarb:	Scale-to-gray, arbitrary downsampling, inner loops and
 *		machine-independent internal stuff
 *
 * Exports:
 *	Int32	    ip_scaleToGrayArb4bppLine ()
 *	Int32	    ip_makeScaleToGrayArbLineTab ()
 *	Int32	    ip_scaleToGrayArbLineTabBytes ()
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
	/* prototypes */
#include "shrscgr.pub"
#include "shrscgr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: miscgarb.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");
/***********************************************************************
 * ip_scaleToGrayArb4bppLine
 *
 *	Inner loop of scale-to-gray. See the comments in i_scgarb.c
 *	to find out what's going on.
 *
 * Arguments:
 *	destP	    -	    Destination image
 *	sourceP	    -	    Source image
 *	sWpl	    -	    32-bit word distance between lines
 *	filterHeight-	    Height of filter
 *	wD	    -	    Number of output pixels to generate
 *	pLineTab    -	    The line table
 *	filterRow   -	    Pointer to the current filter row
 *	filterNextRow-	    Longword offset for next filter row
 *
 * Returns:
 *	nothing
 ***********************************************************************/

void CDECL 
ip_scaleToGrayArb4bppLine (
    IPState    *pIPState,
    UInt32     *destP,
    UInt32     *sourceP,
    Int32	sWpl,
    Int32	filterWidth,
    Int32	filterHeight,
    Int32	wD,
    void       *lTab,
    Int32      *filterRow,
    Int32	filterNextRow)
{
    PROLOGUE;
    UInt32	outputWord = 0, mask;
    Int32	pixelSums, offset, shift0, filterIndex;
    Int32	rowCtr, *filterBase, shift1;
    Int32	col = 0, cWords;
    SGLineTab  *pLineTab = (SGLineTab *) lTab;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    mask = (1 << filterWidth) - 1;

    /* Unroll the first four cases for a bit O' speed gain */
    switch (filterHeight)
    {
    case 1:
	for (cWords = (wD >> 3); cWords > 0; cWords--)
	{
	    filterBase = filterRow /* + pLineTab->phaseOffset */;
	    for (col = 8; col > 0; col--, pLineTab++)
	    {
		pixelSums = 0;
		offset = pLineTab->offset;
		shift0 = pLineTab->shift.mi.shift0;
	    
		if (pLineTab->shift.mi.shift1 >= 0)
		{
		    shift1 = pLineTab->shift.mi.shift1;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];
		}
		else
		{
		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		}

		outputWord = (outputWord << 4) | (pixelSums >> 24);
	    }


	    *destP++ = outputWord;
	}
	break;

    case 2:
	for (cWords = (wD >> 3); cWords > 0; cWords--)
	{
	    for (col = 8; col > 0; col--, pLineTab++)
	    {
		pixelSums = 0;
		offset = pLineTab->offset;
		shift0 = pLineTab->shift.mi.shift0;
		filterBase = filterRow /* + pLineTab->phaseOffset */;
	    
		if (pLineTab->shift.mi.shift1 >= 0)
		{
		    shift1 = pLineTab->shift.mi.shift1;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];
		    offset += sWpl; 
		    filterBase += filterNextRow;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];
		}
		    
		else
		{
		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		    pixelSums += filterBase[((sourceP[offset+sWpl] >> shift0) & mask) + filterNextRow];
		}

		outputWord = (outputWord << 4) | (pixelSums >> 24);
	    }


	    *destP++ = outputWord;
	}
	break;

    case 3:
	for (cWords = (wD >> 3); cWords > 0; cWords--)
	{
	    for (col = 8; col > 0; col--, pLineTab++)
	    {
		pixelSums = 0;
		offset = pLineTab->offset;
		shift0 = pLineTab->shift.mi.shift0;
		filterBase = filterRow /* + pLineTab->phaseOffset */;
	    
		if (pLineTab->shift.mi.shift1 >= 0)
		{
		    shift1 = pLineTab->shift.mi.shift1;
		
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];

		    offset += sWpl; 
		    filterBase += filterNextRow;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];

		    offset += sWpl; 
		    filterBase += filterNextRow;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];
		}
		    
		else
		{
		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		    filterBase += filterNextRow;

		    pixelSums += filterBase[(sourceP[offset+sWpl] >> shift0) & mask];
		    filterBase += filterNextRow;

		    pixelSums += filterBase[(sourceP[offset+2*sWpl] >> shift0) & mask];
		}

		outputWord = (outputWord << 4) | (pixelSums >> 24);
	    }


	    *destP++ = outputWord;
	}
	break;
    
    case 4:
	for (cWords = (wD >> 3); cWords > 0; cWords--)
	{
	    for (col = 8; col > 0; col--, pLineTab++)
	    {
		pixelSums = 0;
		offset = pLineTab->offset;
		shift0 = pLineTab->shift.mi.shift0;
		filterBase = filterRow /* + pLineTab->phaseOffset */;
	    
		if (pLineTab->shift.mi.shift1 >= 0)
		{
		    shift1 = pLineTab->shift.mi.shift1;
		
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];

		    offset += sWpl; 
		    filterBase += filterNextRow;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];

		    offset += sWpl; 
		    filterBase += filterNextRow;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];

		    offset += sWpl; 
		    filterBase += filterNextRow;
		    filterIndex = (sourceP[offset] << shift0) |
				  (sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];
		}
		    
		else
		{
		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		    offset += sWpl;
		    filterBase += filterNextRow;

		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		    offset += sWpl;
		    filterBase += filterNextRow;

		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		    offset += sWpl;
		    filterBase += filterNextRow;

		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		}

		outputWord = (outputWord << 4) | (pixelSums >> 24);
	    }


	    *destP++ = outputWord;
	}
	break;
    

    default:

	for (cWords = (wD >> 3); cWords > 0; cWords--)
	{
	    for (col = 8; col > 0; col--, pLineTab++)
	    {
		pixelSums = 0;
		offset = pLineTab->offset;
		shift0 = pLineTab->shift.mi.shift0;
		filterBase = filterRow /* + pLineTab->phaseOffset */;
		
		if (pLineTab->shift.mi.shift1 >= 0)
		{
		    shift1 = pLineTab->shift.mi.shift1;
		    for (rowCtr = filterHeight; rowCtr > 0; rowCtr--)
		    {
			filterIndex = (sourceP[offset] << shift0) |
				    (sourceP[offset+1] >> shift1);
			pixelSums += filterBase[filterIndex & mask];
			offset += sWpl; 
			filterBase += filterNextRow;
		    }
		}
		else
		{
		    for (rowCtr = filterHeight; rowCtr > 0; rowCtr--)
		    {
			pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
			offset += sWpl;
			filterBase += filterNextRow;
		    }
		}

		outputWord = (outputWord << 4) | (pixelSums >> 24);
	    }

	    *destP++ = outputWord;
	}
	break;
    }

    /* Do the remaining word */
    if (wD & 7)
    {
	outputWord = 0;
	for (col = wD & 7; col > 0; col--, pLineTab++)
	{
	    pixelSums = 0;
	    offset = pLineTab->offset;
	    shift0 = pLineTab->shift.mi.shift0;
	    filterBase = filterRow /* + pLineTab->phaseOffset */;
	    
	    if (pLineTab->shift.mi.shift1 >= 0)
	    {
		UInt16 shift1;
		shift1 = pLineTab->shift.mi.shift1;
		for (rowCtr = filterHeight; rowCtr > 0; rowCtr--)
		{
		    filterIndex = (sourceP[offset] << shift0) |
				(sourceP[offset+1] >> shift1);
		    pixelSums += filterBase[filterIndex & mask];
		    offset += sWpl; 
		    filterBase += filterNextRow;
		}
	    }
	    else
	    {
		for (rowCtr = filterHeight; rowCtr > 0; rowCtr--)
		{
		    pixelSums += filterBase[(sourceP[offset] >> shift0) & mask];
		    offset += sWpl;
		    filterBase += filterNextRow;
		}
	    }

	    outputWord = (outputWord << 4) | (pixelSums >> 24);
	}

	*destP++ = outputWord << ((8 - (wD & 7)) * 4);
    }
}

/***********************************************************************
 *
 * ip_scaleToGrayArbLineTabBytes
 *
 *	Computes the size, in bytes, of block that must be allocated for
 *	i_makeScaleToGray4bppArbLineTAb. (It's done this way to support
 *	applications like the DAE which must do their own allocation)
 * 
 *	filter	-	The filter being used, created by above
 *      xD      -       X-offset into "virtual" image
 *	wD	-	Width of destination image
 *      wVirt   -       Width of "virtual" destination image
 *	wS	-	Width of source rectangle
 *	sBpl	-	inter-line distance, in bytes, in source image
 *
 * Returns:
 *	number of bytes, 0 = error 
 *
 ***********************************************************************/

Int32 CDECL 
ip_scaleToGrayArbLineTabBytes (
    IPState    *pIPState,
    void       *filter,
    Int32	xD,
    Int32	wD,
    Int32	wVirt,
    Int32	wS,
    Int32	sBpl)
{
    /* Shut up the compiler */
    PROLOGUE;

	/* Keep compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    xD = wVirt = wS = sBpl;
    filter = filter;
    return wD * (Int32)sizeof(SGLineTab);
}

/***********************************************************************
 * 
 * ip_makeScaleToGrayArbLineTab
 *
 *	Assuming we've already created the set of filters for a given
 *	reduction factor, we now need to make a "line table" for a given
 *	source offset, width, and destination offset. This table contains,
 *	for each output pixel, an offset for the starting source word,
 *	two shifts and masks, and an offset into the filter table
 *	for the given column phase. (The row phase is obviously the
 *	same for all pixels on the row, so we'll compute the start of
 *	the row of filters for that phase).
 * 
 * Arguments:
 *	filter  -	filter LUT returned by i_makeScaleToGrayArbFilter
 *      xD      -       X-offset into "virtual" image, without frame
 *	wD	-	Width of destination image
 *      wVirt   -       Width of "virtual" destination image, without frame
 *	wS	-	Width of source rectangle
 *	sBpl	-	Source image distance between lines, in bytes
 *	lineTab	-	Line table structure, allocated by caller,
 *			size returned by i_makeScaleToGrayArbLineTabBytes
 *	
 * Returns:
 *	ia_successful or error code
 *
 ***********************************************************************/

Int32 CDECL 
ip_makeScaleToGrayArbLineTab (
    IPState    *pIPState,
    void       *filter,
    Int32	xD,
    Int32	wD,
    Int32	wVirt,
    Int32	wS,
    Int32	sBpl,
    void       *lTab)
{
    PROLOGUE;
    Int32	 i, filterOffset, startPixel, startBit, startWord, shift;
    Float64	 sourcePixel, sourceInc;
    FilterDesc	*fDesc = (FilterDesc *) filter;
    SGLineTab	*lineTab = (SGLineTab *) lTab;

    /* filterOffset is the number of source pixels from the center
       to the left edge of the filter */
    filterOffset = fDesc->width / 2;

    /* Shut the compiler up */
    sBpl = sBpl;

    /* sourceInc is the scaling factor, i.e. sourceInc pixels per dest pixel */
    sourceInc = (Float64) wS / (Float64) wVirt;
    sourcePixel = xD * sourceInc + cFrameBits;

    for (i = 0; i < wD; i++)
    {
	startPixel = (Int32)((sourcePixel - filterOffset) + cEpsilon);
	startBit = 32 - (startPixel & 31);
	startWord = startPixel >> 5;
	lineTab[i].offset = startWord;

	/* Does it straddle a word? */
	if (startBit >= fDesc->width)
	{
	    /* If not, the second word shift is -1 */
	    shift = startBit - fDesc->width;
	    lineTab[i].shift.mi.shift0 = (UInt16)shift;
	    lineTab[i].shift.mi.shift1 = -1;
	}
	else
	{
	    /* Straddles a word, so make two shifts/masks. The first word
	       is shifted left, the second is shifted right*/
	    lineTab[i].shift.mi.shift0 = (UInt16)(fDesc->width - startBit);
	    shift = 32 - (fDesc->width - startBit);
	    lineTab[i].shift.mi.shift1 = (Int16)shift;
	}

	sourcePixel += sourceInc;
    }

    NORMAL_RETURN;
}

/*
$Log: miscgarb.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.7  1998/01/13 23:53:21  ddavies
miscgarb.c -> /product/ipcore/ipshared/src/RCS/miscgarb.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.6  1995/12/13 02:51:49  ejaquith
miscgarb.c -> /product/ipcore/ipshared/src/RCS/miscgarb.c,v

CommonCheck: Added ";" to end of IP_RCSINFO() macro.

 * Revision 1.5  1995/10/16  17:51:01  dferrell
 * miscgarb.c -> /product/ipcore/ipshared/src/RCS/miscgarb.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.4  1995/04/06  17:11:15  macleod
 * miscgarb.c -> /product/ipcore/ipshared/src/RCS/miscgarb.c,v
 *
 * Changed the interfaces to the line-table building functions to
 * take a source width, "virtual" (scaled) width, and an x-offset and
 * width relative to the "virtual" image.
 * Note that the x-offset does not include a frame.
 *
 * Revision 1.3  1995/01/13  02:02:39  macleod
 * miscgarb.c -> /product/ipcore/ipshared/src/RCS/miscgarb.c,v
 *
 * Took out "phases" stuff. Made the code somewhat faster by getting rid
 * of the "phases" stuff and recognizing (duh) that if the pixels were
 * masked after they were shifted, the mask was a constant.
 *
 * Revision 1.2  1995/01/04  18:45:50  macleod
 * miscgarb.c -> /product/ipcore/ipshared/src/RCS/miscgarb.c,v
 *
 * forgot the #ifdef __GNUC__ around the #include "ansiprot.h"
 *
 * Revision 1.1  1995/01/04  01:15:47  macleod
 * Initial revision
 *
*/
