/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	wcarith.c
 *
 *  Private routine:
 *        ip_waveArithInit()
 *        ip_dataArithInit()
 *        ip_bitplusfollow()
 *        ip_waveArithCoding()
 *        ip_dataArithTerm()
 *        ip_waveArithTerm()
 *        ip_getArithCodelen()
 *
 */

#include <stdio.h>
#include <math.h>

#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwcari_c_RCSInfo, "$RCSfile: miwcari.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

#define BUFFLEN  1000

static ARITHINFO mainInfo;

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding init.
 *
 * LongDescription:::
 *  Arithmetic init. It maust be called before coding.
 *
 * Input:::
 *       pdata::     Ptr to userData.
 *  outputFunc::     Ptr to callback function
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveArithInit (
    IPState    *pIPState,
    void       *pdata,
    Int32 (CDECL *outputFunc) (
		    IPState    *pIPState,
		    void       *userData,
		    UInt8      *buf,
		    UInt32	count))
{
    PROLOGUE;
    ARITHINFO *pInfo;

    pInfo = &mainInfo;
    pInfo->pdata = pdata;
    pInfo->IOPath = outputFunc; 
    pInfo->low =0;
    pInfo->high = TOP_VALUE;

    pInfo->buffer =0;
    pInfo->bits_to_go = 8;
    pInfo->bits_to_follow =0;
    pInfo->bufIdx =0;   /* save buffer length  */
    pInfo->codeLen =0;
    pInfo->outBuf= (UInt8 *) IP_CALLOC(sizeof(UInt8), BUFFLEN);

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding data init.
 *
 * LongDescription:::
 *  Arithmetic coding can support multi coding status. For every data set
 *  you have to call data init to create coding infomation stucture.
 *
 * Input:::
 *       infoIdx::     Coding data set index
 *  no_of_symbol::     Number of symbol of coding data set
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_dataArithInit (
    IPState    *pIPState,
    Int32	infoIdx,
    Int32	no_of_symbol)
{
    PROLOGUE;
    Int16      idx;
    Int16      *symbol_to_index, *index_to_symbol;
    Int32      *freq, *cum_freq;
    SINGLEARITHINFO  *pInfo;

    mainInfo.info[infoIdx] = (void *) IP_CALLOC(1, sizeof(SINGLEARITHINFO));
    pInfo = mainInfo.info[infoIdx];
    pInfo->no_of_symbol = (Int16)no_of_symbol;
    pInfo->no_of_index = (Int16)no_of_symbol+1;
    pInfo->eof_index = (Int16)no_of_symbol+1;
    pInfo->symbol_to_index = (Int16 *)IP_CALLOC(sizeof(Int16), 
		     pInfo->no_of_index*2);
    pInfo->index_to_symbol = pInfo->symbol_to_index + pInfo->no_of_index;
    pInfo->freq = (Int32 *)IP_CALLOC(sizeof(Int32), (pInfo->no_of_index+1)*2);
    pInfo->cum_freq = pInfo->freq + pInfo->no_of_index+1;

    symbol_to_index = pInfo->symbol_to_index;
    index_to_symbol = pInfo->index_to_symbol;
    for(idx=0; idx<pInfo->no_of_symbol; ++idx)
    {
	symbol_to_index[idx]= idx+1;
	index_to_symbol[idx+1] = idx;
    }
    freq = pInfo->freq;
    cum_freq = pInfo->cum_freq;
    for(idx=0; idx<=pInfo->no_of_index; ++idx)
    {
	freq[idx] =1;
	cum_freq[idx] = pInfo->no_of_index - idx;
    }
    freq[0] =0;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Emit bits.
 *
 * LongDescription:::
 *   Emit arithmetic coding output bits.
 *
 * Input:::
 *     bit::        Current bit status ( 1, or 0 )
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_bitplusfollow (
    IPState    *pIPState,
    Int16	bit)
{
    PROLOGUE;
    ARITHINFO *pInfo;

    pInfo = &mainInfo;    
    pInfo->buffer >>= 1;
    if(bit) pInfo->buffer |= 0x80;
    if( --pInfo->bits_to_go == 0)
    {
	if(pInfo->bufIdx >= BUFFLEN)
	{
	    pInfo->IOPath(pIPState, pInfo->pdata, pInfo->outBuf, pInfo->bufIdx);
	    pInfo->bufIdx =0;
	    pInfo->codeLen += BUFFLEN;
	}
	*(pInfo->outBuf+pInfo->bufIdx) = pInfo->buffer;
	++ pInfo->bufIdx;
	pInfo->bits_to_go = 8;
    }
    while( pInfo->bits_to_follow >0)
    {
	if( pInfo->bits_to_follow <= pInfo->bits_to_go)
	{
	    pInfo->buffer >>= pInfo->bits_to_follow;
	    if( !bit )
		pInfo->buffer |= (0xff)<<(8-pInfo->bits_to_follow);
	    pInfo->bits_to_go -= pInfo->bits_to_follow;
	    pInfo->bits_to_follow =0;
	}
	else
	{
	    pInfo->buffer >>= pInfo->bits_to_go;
	    if(!bit) pInfo->buffer |= (0xff)<<(8-pInfo->bits_to_go);
	    pInfo->bits_to_follow -= pInfo->bits_to_go;
	    pInfo->bits_to_go =0;
	}

	if( pInfo->bits_to_go == 0)
	{
	    if(pInfo->bufIdx >= BUFFLEN)
	    {
		pInfo->IOPath(pIPState, pInfo->pdata,
			      pInfo->outBuf, pInfo->bufIdx);
		pInfo->bufIdx =0;
		pInfo->codeLen += BUFFLEN;
	    }
	    *(pInfo->outBuf + pInfo->bufIdx) = pInfo->buffer;
	    ++ pInfo->bufIdx;
	    pInfo->bits_to_go = 8;
	}
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding.
 *
 * LongDescription:::
 *  Arithmetic coding.
 *
 * Input:::
 *   symbol::     Current symbol
 *  bitCode::     Fix bits coding or ataptive arithmetic coding.
 *  infoIdx::     Coding data set index
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveArithCoding (
    IPState    *pIPState,
    Int16	symbol,
    Int16	bitCode,
    Int32	infoIdx) 
{
    PROLOGUE;
    Int32      range, high, low;
    Int32      i;
    Int16      index;
    SINGLEARITHINFO  *pInfo;

    high = mainInfo.high;
    low = mainInfo.low;
    range = high - low+1;

    if(bitCode !=0)
    {
	pInfo = NULL;
	if(bitCode < 3)
	{
	    if (bitCode >1) 
	    {
		high = low -1 + range;  
		low += range >> 1; 
	    }
	    else
		high = low -1 + (range >> 1);
	}
	else
	{
 	    index = bitCode >> 8;
	    bitCode &= (1<<index) -1;
	    high = low -1 + ((range*(bitCode+1))>>index);
	    low += (range * bitCode)>>index;
	    bitCode += index<<8;
	}
	index =0;
    }
    else
    {
	pInfo = mainInfo.info[infoIdx];
	index = pInfo->symbol_to_index[symbol];
	high = mainInfo.low + (range * pInfo->cum_freq[index-1])/
	    pInfo->cum_freq[0] -1;
	low = mainInfo.low + (range * pInfo->cum_freq[index])/
	    pInfo->cum_freq[0];
    }


    for(;;)
    {
	if( high < HALF)
	    ip_bitplusfollow (pIPState, 0);
	else if(low >= HALF)
	{
	    ip_bitplusfollow (pIPState, 1);
	    low -= HALF;
	    high -= HALF;
	}
	else if(low >= FIRST_QTR && high < THIRD_QTR)
	{
	    ++ mainInfo.bits_to_follow;
	    low -= FIRST_QTR;
	    high -= FIRST_QTR;
	}
	else break;
	low = 2*low;
	high = 2*high +1;
    }

    mainInfo.low = low;
    mainInfo.high = high;

    if(bitCode) NORMAL_RETURN;

    if(index == pInfo->eof_index) NORMAL_RETURN;
    
    if(pInfo->cum_freq[0] >= MAX_FREQUENCY)
    {
	Int32 cum;
	cum =0;
	for(i=pInfo->no_of_index; i>=0; --i)
	{
	    pInfo->freq[i] = (pInfo->freq[i]+1)/2;
	    pInfo->cum_freq[i] = cum;
	    cum += pInfo->freq[i];
	}
    }
    for(i = index; pInfo->freq[i]== pInfo->freq[i-1]; --i);
    if( i< index )
    {
	Int16 c_1, c_2;
	c_1 = pInfo->index_to_symbol[i];
	c_2 = pInfo->index_to_symbol[index];
	pInfo->index_to_symbol[i] = c_2;
	pInfo->index_to_symbol[index] = c_1;
	pInfo->symbol_to_index[c_1] = index;
	pInfo->symbol_to_index[c_2] = (Int16)i;
    }
    ++ pInfo->freq[i];
    while(i>0)
    {
	-- i;
	++ pInfo->cum_freq[i];
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding data term.
 *
 * LongDescription:::
 *  Term one data set's arothmetic coding info and free memory. 
 *
 * Input:::
 *   infoIdx::        Data set's index
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_dataArithTerm (
    IPState    *pIPState,
    Int32	infoIdx)
{
    PROLOGUE;
    SINGLEARITHINFO  *pInfo;

    pInfo = mainInfo.info[infoIdx];
    IP_FREE(pInfo->symbol_to_index);
    IP_FREE(pInfo->freq);
    IP_FREE(pInfo);
    mainInfo.info[infoIdx] = NULL;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic coding term.
 *
 * LongDescription:::
 *  Arithmetic coding term.
 *
 * Input:::
 *   flag::         data, may be used in later version.
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveArithTerm (
    IPState    *pIPState,
    Int32	flag)
{
    PROLOGUE;
    ARITHINFO   *pInfo;
    Int32       idx;
 
    flag =0;
    pInfo = &mainInfo;
    ++ pInfo->bits_to_follow; 
    if(pInfo->low < FIRST_QTR)
	ip_bitplusfollow (pIPState, 0);
    else
	ip_bitplusfollow (pIPState, 1);

    pInfo->buffer >>= pInfo->bits_to_go;
    if(pInfo->bufIdx >= BUFFLEN)
    {
	pInfo->IOPath(pIPState, pInfo->pdata, pInfo->outBuf, pInfo->bufIdx);
	pInfo->bufIdx =0;
	pInfo->codeLen += BUFFLEN;
    }
    *(pInfo->outBuf+pInfo->bufIdx) = pInfo->buffer;
    ++ pInfo->bufIdx;
    pInfo->IOPath(pIPState, pInfo->pdata, pInfo->outBuf, pInfo->bufIdx);
    pInfo->codeLen += pInfo->bufIdx;
    pInfo->bufIdx =0;

    if(pInfo->buffer != 0)
    {
	pInfo->buffer =0;
	pInfo->IOPath(pIPState, pInfo->pdata, &pInfo->buffer, 1);
	pInfo->IOPath(pIPState, pInfo->pdata, &pInfo->buffer, 1);
	pInfo->codeLen += 2;
    }
    else
    {
	pInfo->buffer =0;
	pInfo->IOPath(pIPState, pInfo->pdata, &pInfo->buffer, 1);
	pInfo->codeLen += 1;
    }
    pInfo->bits_to_go =8;

    IP_FREE(pInfo->outBuf);
    pInfo->outBuf = NULL;

/* clearn up all data set info */
    for(idx=0; idx<NUM_DATA_SET; ++idx)
    {
	if(mainInfo.info[idx] == NULL) continue;
	ip_dataArithTerm (pIPState, idx);
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get compressed code length.
 *
 * LongDescription:::
 *  Get arithmetic code length.
 *
 * Input:::
 *
 * Output:::
 *    codeLen::       Ptr to codeLen
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_getArithCodelen (
    IPState    *pIPState,
    Int32      *codeLen)
{
    PROLOGUE;

    *codeLen = mainInfo.codeLen + mainInfo.bufIdx;

    NORMAL_RETURN;
}

/*
$Log: miwcari.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.4  1998/01/13 23:53:46  ddavies
miwcari.c -> /product/ipcore/ipshared/src/RCS/miwcari.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.3  1997/09/17 22:47:57  xzhu
xzhu on Wed Sep 17 15:47:56 PDT 1997

miwcari.c -> /product/ipcore/ipshared/src/RCS/miwcari.c,v

CommonCheck: Added #include wavelib.pub, since some changes made in huffman coding.

Revision 1.2  1997/07/11 01:51:20  ddavies
miwcari.c -> /product/ipcore/ipshared/src/RCS/miwcari.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.1  1996/10/16 00:02:33  xzhu
Initial revision

*/
