/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	wctrans.c:	
 *
 *  Private routine:
 *      ip_waveBifilterTrans()   
 *      ip_init_filetertbl()
 *      ip_waveTrans_LUT()
 *      ip_waveBifilterLift()
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwctra_c_RCSInfo, "$RCSfile: miwctra.c,v $; $Revision: 1.1.1.1 $; $Datd: $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    N level Biorthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *  f_lowpass::     Ptr to lowpass filter
 * f_highpass::     Ptr to highpass filter
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveBifilterTrans (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32      *f_lowpass,
    Int32      *f_highpass,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp;
    Int16       *coef, *buffer, *linebuf, *phigh, *plow, *pSrc, *ptm;
    Int16       t0,t1,t2,t3,t4;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* collumn */
    for(row=0; row<coefw; ++row)
    {
/* copy image */
	for(idx=0, col=row; idx<coefh; ++idx, col+=bpl)
	    *(linebuf+idx)=*(coef+col);
	for(idx=1, col=coefh-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}

/* convelution */
	plow = coef+ row;
	phigh = coef+ (halfh * bpl)+ row;
	ptm = buffer;
	for(idx=0; idx<coefh; ++idx, plow+= bpl, phigh+=bpl, ++ptm)
	{
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    t4 = *(ptm+8) + *(ptm);
	    temp = f_lowpass[0] * t0 +
		f_lowpass[1] * t1 +
		f_lowpass[2] * t2 +
		f_lowpass[3] * t3 +
		f_lowpass[4] * t4;
	    *plow = (Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefh) break;
	    ++ ptm;
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    temp =f_highpass[0] * t0 +
		f_highpass[1] * t1 +
		f_highpass[2] * t2 +
		f_highpass[3] * t3;
	    *phigh = (Int16)((temp+ONEHALF)>>SCALEBITS);
	}
    }

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+= bpl)
    {
/* copy image */
	MEMCPY(linebuf, pSrc, coefw*sizeof(Int16));
	for(idx=1, col=coefw-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	plow = pSrc;
	phigh = plow + halfw;
	ptm = buffer;
	for(idx=0; idx<coefw; ++idx, ++ptm)
	{
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    t4 = *(ptm+8) + *(ptm);
	    temp =  f_lowpass[0] * t0 +
		f_lowpass[1] * t1 +
		f_lowpass[2] * t2 +
		f_lowpass[3] * t3 +
		f_lowpass[4] * t4;
	    *plow++ = (Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefw) break;
	    ++ ptm;
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    temp = f_highpass[0] * t0 +
		f_highpass[1] * t1 +
		f_highpass[2] * t2 +
		f_highpass[3] * t3;
	    *phigh++ = (Int16)((temp+ONEHALF)>>SCALEBITS);
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    N level Biorthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveBifilterLift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp;
    Int16       *coef, *buffer, *linebuf, *phigh, *plow, *pSrc, *ptm, para;
    Int32       dp2, dp1, d0, dn1, sp1, s0, sn1, ddp1, dd0, ss0;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* collumn */
    for(row=0; row<coefw; ++row)
    {
/* copy image */
	for(idx=0, col=row; idx<coefh; ++idx, col+=bpl)
	    *(linebuf+idx)=*(coef+col);
	for(idx=1, col=coefh-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}

/* convelution */
	plow = coef+ row;
	phigh = coef+ (halfh * bpl)+ row;
	ptm = buffer;
	dp2 = ((Int32)(*(ptm+1))<< SCALEBITS) + ALPH0 * (*ptm + *(ptm+2));
	dp1 = ((Int32)(*(ptm+3))<< SCALEBITS) + ALPH0 * (*(ptm+2) + *(ptm+4));
	d0 = ((Int32)(*(ptm+5))<< SCALEBITS) + ALPH0 * (*(ptm+4) + *(ptm+6));
	para = (Int16)((dp2 +dp1 +ONEHALF)>>SCALEBITS);
	sp1 = ((Int32)(*(ptm+2))<< SCALEBITS) + ALPH1 * para;
	para = (Int16)((dp1 + d0 +ONEHALF)>>SCALEBITS);
	s0 = ((Int32)(*(ptm+4))<< SCALEBITS) + ALPH1 * para; 
	para = (Int16)((sp1 + s0 +ONEHALF)>>SCALEBITS);
	ddp1 = dp1 + ALPH2 * para;
	for(idx=0; idx<coefh; ++idx, plow+= bpl, phigh+=bpl, ptm+= 2)
	{
	    dn1 =((Int32)(*(ptm+7))<< SCALEBITS)+ALPH0 * (*(ptm+6) + *(ptm+8));
	    para = (Int16)((d0 + dn1 +ONEHALF)>>SCALEBITS);
	    sn1 =((Int32)(*(ptm+6))<< SCALEBITS) + ALPH1 * para;
	    para = (Int16) ((s0 + sn1+ONEHALF)>>SCALEBITS);
	    dd0 = d0 + ALPH2 * para;
	    para = (Int16) ((ddp1 + dd0+ONEHALF)>>SCALEBITS);
	    ss0 = s0 + ALPH3 * para;
	    temp = ALPH4 * ((ss0 +ONEHALF)/ONEHALF1);
	    *plow = (Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefh) break;
	    temp = dd0 / ALPH4;
	    *phigh = (Int16)(temp);
	    dp2 = dp1;
	    dp1 = d0;
	    d0 = dn1;
	    sp1 = s0;
	    s0 = sn1;
	    ddp1 = dd0;
	}
    }

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+= bpl)
    {
/* copy image */
	MEMCPY(linebuf, pSrc, coefw*sizeof(Int16));
	for(idx=1, col=coefw-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	plow = pSrc;
	phigh = plow + halfw;
	ptm = buffer;
	dp2 = ((Int32)(*(ptm+1))<< SCALEBITS) + ALPH0 * (*ptm + *(ptm+2));
	dp1 = ((Int32)(*(ptm+3))<< SCALEBITS) + ALPH0 * (*(ptm+2) + *(ptm+4));
	d0 = ((Int32)(*(ptm+5))<< SCALEBITS) + ALPH0 * (*(ptm+4) + *(ptm+6));
	para = (Int16)((dp2 +dp1 +ONEHALF)>>SCALEBITS);
	sp1 = ((Int32)(*(ptm+2))<< SCALEBITS) + ALPH1 * para;
	para = (Int16)((dp1 + d0 +ONEHALF)>>SCALEBITS);
	s0 = ((Int32)(*(ptm+4))<< SCALEBITS) + ALPH1 * para; 
	para = (Int16)((sp1 + s0 +ONEHALF)>>SCALEBITS);
	ddp1 = dp1 + ALPH2 * para;
	for(idx=0; idx<coefw; ++idx, ptm += 2)
	{
	    dn1 =((Int32)(*(ptm+7))<< SCALEBITS)+ALPH0 * (*(ptm+6) + *(ptm+8));
	    para = (Int16)((d0 + dn1 +ONEHALF)>>SCALEBITS);
	    sn1 =((Int32)(*(ptm+6))<< SCALEBITS) + ALPH1 * para;
	    para = (Int16) ((s0 + sn1+ONEHALF)>>SCALEBITS);
	    dd0 = d0 + ALPH2 * para;
	    para = (Int16) ((ddp1 + dd0+ONEHALF)>>SCALEBITS);
	    ss0 = s0 + ALPH3 * para;
	    temp = ALPH4 * ((ss0 +ONEHALF)/ONEHALF1);
	    *plow++ = (Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefw) break;
	    temp = dd0 / ALPH4;
	    *phigh++ = (Int16)(temp);
	    dp2 = dp1;
	    dp1 = d0;
	    d0 = dn1;
	    sp1 = s0;
	    s0 = sn1;
	    ddp1 = dd0;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    N level 4 tap orthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do 4 tap orthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveD4filterLift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp;
    Int16       *coef, *buffer, *linebuf, *phigh, *plow, *pSrc, *ptm;
    Float32     dp1, d0, dn1, sp1, s0;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* collumn */
    for(row=0; row<coefw; ++row)
    {
/* copy image */
	for(idx=0, col=row; idx<coefh; ++idx, col+=bpl)
	    *(linebuf+idx)=*(coef+col);
	for(idx=1, col=coefh-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}

/* convelution */
	plow = coef+ row;
	phigh = coef+ (halfh * bpl)+ row;
	ptm = buffer;
	dp1 = (Float32)(*(ptm+3)) - D4P0 * (Float32)(*(ptm+2));
	d0 = (Float32)(*(ptm+5)) -D4P0 * (Float32)(*(ptm+4));
	sp1 = (Float32)(*(ptm+2)) + (D4P1 * dp1) + (D4P2 * d0) ;
	for(idx=0; idx<coefh; ++idx, plow+= bpl, phigh+=bpl, ptm+= 2)
	{
	    dn1 = (Float32)(*(ptm+7)) - D4P0 * (Float32)(*(ptm+6));
	    s0  = (Float32)(*(ptm+4)) + (D4P1 * d0) + (D4P2 * dn1);
	    temp = (Int32)( D4P3 * s0) ;
	    *plow = (Int16)(temp);
	    if( ++idx >= coefh) break;
	    temp = (Int32)(D4P4 * (d0 + sp1));
	    *phigh = (Int16)(temp);
	    d0 = dn1;
	    sp1 = s0;
	}
    }

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+= bpl)
    {
/* copy image */
	MEMCPY(linebuf, pSrc, coefw*sizeof(Int16));
	for(idx=1, col=coefw-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	plow = pSrc;
	phigh = plow + halfw;
	ptm = buffer;
	dp1 = (Float32)(*(ptm+3)) - D4P0 * (Float32)(*(ptm+2));
	d0 = (Float32)(*(ptm+5)) -D4P0 * (Float32)(*(ptm+4));
	sp1 = (Float32)(*(ptm+2)) + (D4P1 *dp1) + (D4P2 * d0);
	for(idx=0; idx<coefw; ++idx, ptm += 2)
	{
	    dn1 = (Float32)(*(ptm+7)) - D4P0 * (Float32)(*(ptm+6));
	    s0  = (Float32)(*(ptm+4)) + (D4P1 * d0) + (D4P2 * dn1);
	    temp = (Int32)(D4P3 * s0 );
	    *plow++ = (Int16)(temp);
	    if( ++idx >= coefw) break;
	    temp = (Int32)(D4P4 * (d0 + sp1));
	    *phigh++ = (Int16)(temp);
	    d0 = dn1;
	    sp1 = s0;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    N level (4,2) biorthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do (4,2) biorthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveBi44filterLift (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh;
    Int16       *coef, *buffer, *linebuf, *phigh, *plow, *pSrc, *ptm;
    Int32       dp1, dp2, d0, dn1, s0, p1, p2;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer + 6;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* collumn */
    for(row=0; row<coefw; ++row)
    {
/* copy image */
	for(idx=0, col=row; idx<coefh; ++idx, col+=bpl)
	    *(linebuf+idx)=*(coef+col);
	for(idx=1, col=coefh-1; idx<7; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}

/* convelution */
	plow = coef+ row;
	phigh = coef+ (halfh * bpl)+ row;
	ptm = buffer;
	p1 = (Int32) (*(ptm+2)) + (Int32) (*(ptm+4));
	p2 = (Int32) (*ptm) + (Int32) (*(ptm+6));
	dp2 = (Int32)(*(ptm+3)) -((9*p1 - p2)/16);
	p1 = (Int32) (*(ptm+4)) + (Int32) (*(ptm+6));
	p2 = (Int32) (*(ptm+2)) + (Int32) (*(ptm+8));
	dp1 = (Int32)(*(ptm+5)) -((9*p1 - p2)/16);
	p1 = (Int32) (*(ptm+6)) + (Int32) (*(ptm+8));
	p2 = (Int32) (*(ptm+4)) + (Int32) (*(ptm+10));
	d0 = (Int32)(*(ptm+7)) - ((9*p1 -p2)/16);
	for(idx=0; idx<coefh; ++idx, plow+= bpl, phigh+=bpl, ptm+= 2)
	{
	    p1 = (Int32) (*(ptm+8)) + (Int32) (*(ptm+10));
	    p2 = (Int32) (*(ptm+6)) + (Int32) (*(ptm+12));
	    dn1 = (Int32)(*(ptm+9)) - ((9*p1 -p2)/16);
	    s0 = (Int32)(*(ptm+6)) + ((9*(d0+dp1) - (dp2+ dn1))/32);
	    *plow = (Int16) s0;
	    if( ++idx >= coefh) break;
	    *phigh = (Int16) d0;
	    dp2 = dp1;
	    dp1 = d0;
	    d0 = dn1;
	}
    }

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+= bpl)
    {
/* copy image */
	MEMCPY(linebuf, pSrc, coefw*sizeof(Int16));
	for(idx=1, col=coefw-1; idx<7; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	plow = pSrc;
	phigh = plow + halfw;
	ptm = buffer;
	p1 = (Int32) (*(ptm+2)) + (Int32) (*(ptm+4));
	p2 = (Int32) (*ptm) + (Int32) (*(ptm+6));
	dp2 = (Int32)(*(ptm+3)) -((9*p1 - p2)/16);
	p1 = (Int32) (*(ptm+4)) + (Int32) (*(ptm+6));
	p2 = (Int32) (*(ptm+2)) + (Int32) (*(ptm+8));
	dp1 = (Int32)(*(ptm+5)) -((9*p1 - p2)/16);
	p1 = (Int32) (*(ptm+6)) + (Int32) (*(ptm+8));
	p2 = (Int32) (*(ptm+4)) + (Int32) (*(ptm+10));
	d0 = (Int32)(*(ptm+7)) - ((9*p1 -p2)/16);
	for(idx=0; idx<coefw; ++idx, ptm += 2)
	{
	    p1 = (Int32) (*(ptm+8)) + (Int32) (*(ptm+10));
	    p2 = (Int32) (*(ptm+6)) + (Int32) (*(ptm+12));
	    dn1 = (Int32)(*(ptm+9)) - ((9*p1 -p2)/16);
	    s0 = (Int32)(*(ptm+6)) + ((9*(d0+dp1) - (dp2+ dn1))/32);
	    *plow++ = (Int16) s0;
	    if( ++idx >= coefw) break;
	    *phigh++ = (Int16) d0;
	    dp2 = dp1;
	    dp1 = d0;
	    d0 = dn1;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    N level TT wavelet transform.
 *
 * LongDescription:::
 *    Do TT wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveTTfilter (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh;
    Int16       *coef, *buffer, *linebuf, *phigh, *plow, *pSrc, *ptm;
    Int32       sp2, sp1, s0, sn1, sn2, d0;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer + 6;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* collumn */
    for(row=0; row<coefw; ++row)
    {
/* copy image */
	for(idx=0, col=row; idx<coefh; ++idx, col+=bpl)
	    *(linebuf+idx)=*(coef+col);
	for(idx=1, col=coefh-1; idx<7; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}

/* convelution */
	plow = coef+ row;
	phigh = coef+ (halfh * bpl)+ row;
	ptm = buffer;
	sp2 = ( *(ptm + 2) + *(ptm + 3) )/2;
	sp1 = ( *(ptm + 4) + *(ptm + 5) )/2;
	s0 = ( *(ptm + 6) + *(ptm + 7) )/2;
	sn1 = ( *(ptm + 8) + *(ptm + 9) )/2;
	for(idx=0; idx<coefh; ++idx, plow+= bpl, phigh+=bpl, ptm+= 2)
	{
	    *plow = (Int16) s0;
	    if( ++idx >= coefh) break;
	    sn2 = ( *(ptm + 10) + *(ptm+11) )/2;
	    d0 = *(ptm+6) - *(ptm+7) + (3 *(sp2 - sn2)- 22*(sp1 - sn1)+32)/64;
	    *phigh = (Int16) d0;
	    sp2 = sp1;
	    sp1 = s0;
	    s0 = sn1;
	    sn1 = sn2;
	}
    }

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+= bpl)
    {
/* copy image */
	MEMCPY(linebuf, pSrc, coefw*sizeof(Int16));
	for(idx=1, col=coefw-1; idx<7; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	plow = pSrc;
	phigh = plow + halfw;
	ptm = buffer;
	sp2 = ( *(ptm + 2) + *(ptm +3) )/2;
	sp1 = ( *(ptm + 4) + *(ptm +5) )/2;
	s0 = ( *(ptm + 6) + *(ptm +7) )/2;
	sn1 = ( *(ptm + 8) + *(ptm +9) )/2;
	for(idx=0; idx<coefw; ++idx, ptm += 2)
	{
	    *plow++ = (Int16) s0;
	    if( ++idx >= coefw) break;
	    sn2 = ( *(ptm + 10) + *(ptm +11) )/2;
	    d0 = *(ptm+6) - *(ptm+7) + (3 *(sp2 - sn2)- 22*(sp1 - sn1)+ 32)/64;
	    *phigh++ = (Int16) d0;
	    sp2 = sp1;
	    sp1 = s0;
	    s0 = sn1;
	    sn1 = sn2;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    N level 16 tap orthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do 16 tap orthogonal wavelet transform on nth level.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveD16filter (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh;
    Int16       *coef, *buffer, *linebuf, *phigh, *plow, *pSrc, *ptm;
    Int32       s0, d0;

    coef = *coefbuf;
    buffer = *tempbuf;

    linebuf = buffer + 15;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

/* collumn */
    for(row=0; row<coefw; ++row)
    {
/* copy image */
	for(idx=0, col=row; idx<coefh; ++idx, col+=bpl)
	    *(linebuf+idx)=*(coef+col);
	for(idx=1, col=coefh-1; idx<15; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	*(linebuf+col+idx) = *(linebuf+col-idx);

/* convelution */
	plow = coef+ row;
	phigh = coef+ (halfh * bpl)+ row;
	ptm = linebuf;
	for(idx=0; idx<coefh; ++idx, plow+= bpl, phigh+=bpl, ptm+= 2)
	{
	    s0 = (D16H0 * (*ptm)) + (D16H1 * (*ptm+1)) +
		 (D16H2 * (*ptm+2)) + (D16H3 * (*ptm+3)) +
		 (D16H4 * (*ptm+4)) + (D16H5 * (*ptm+5)) +
		 (D16H6 * (*ptm+6)) + (D16H7 * (*ptm+7)) +
		 (D16H8 * (*ptm+8)) + (D16H9 * (*ptm+9)) +
		 (D16HA * (*ptm+10)) + (D16HB * (*ptm+11)) +
		 (D16HC * (*ptm+12)) + (D16HD * (*ptm+13)) +
		 (D16HE * (*ptm+14)) + (D16HF * (*ptm+15));
	    *plow = (Int16)((s0 + ONEHALF)/ONEHALF1);
	    if( ++idx >= coefh) break;
	    d0 = (D16H0 * (*ptm+1)) - (D16H1 * (*ptm)) +
		 (D16H2 * (*ptm-1)) - (D16H3 * (*ptm-2)) +
		 (D16H4 * (*ptm-3)) - (D16H5 * (*ptm-4)) +
		 (D16H6 * (*ptm-5)) - (D16H7 * (*ptm-6)) +
		 (D16H8 * (*ptm-7)) - (D16H9 * (*ptm-8)) +
		 (D16HA * (*ptm-9)) - (D16HB * (*ptm-10)) +
		 (D16HC * (*ptm-11)) - (D16HD * (*ptm-12)) +
		 (D16HE * (*ptm-13)) - (D16HF * (*ptm-14));
	    *phigh = (Int16)((d0 + ONEHALF)/ONEHALF1);
	}
    }

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+= bpl)
    {
/* copy image */
	MEMCPY(linebuf, pSrc, coefw*sizeof(Int16));
	for(idx=1, col=coefw-1; idx<15; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	*(linebuf+col+idx) = *(linebuf+col-idx);
	plow = pSrc;
	phigh = plow + halfw;
	ptm = linebuf;
	for(idx=0; idx<coefw; ++idx, ptm += 2)
	{
	    s0 = (D16H0 * (*ptm)) + (D16H1 * (*ptm+1)) +
		 (D16H2 * (*ptm+2)) + (D16H3 * (*ptm+3)) +
		 (D16H4 * (*ptm+4)) + (D16H5 * (*ptm+5)) +
		 (D16H6 * (*ptm+6)) + (D16H7 * (*ptm+7)) +
		 (D16H8 * (*ptm+8)) + (D16H9 * (*ptm+9)) +
		 (D16HA * (*ptm+10)) + (D16HB * (*ptm+11)) +
		 (D16HC * (*ptm+12)) + (D16HD * (*ptm+13)) +
		 (D16HE * (*ptm+14)) + (D16HF * (*ptm+15));
	    *plow++ = (Int16)((s0 + ONEHALF)/ONEHALF1);
	    if( ++idx >= coefw) break;
	    d0 = (D16H0 * (*ptm+1)) - (D16H1 * (*ptm)) +
		 (D16H2 * (*ptm-1)) - (D16H3 * (*ptm-2)) +
		 (D16H4 * (*ptm-3)) - (D16H5 * (*ptm-4)) +
		 (D16H6 * (*ptm-5)) - (D16H7 * (*ptm-6)) +
		 (D16H8 * (*ptm-7)) - (D16H9 * (*ptm-8)) +
		 (D16HA * (*ptm-9)) - (D16HB * (*ptm-10)) +
		 (D16HC * (*ptm-11)) - (D16HD * (*ptm-12)) +
		 (D16HE * (*ptm-13)) - (D16HF * (*ptm-14));
	    *phigh++ = (Int16)((d0 + ONEHALF)/ONEHALF1);
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set up filter look up table.
 *
 * LongDescription:::
 *  filter parameters have been scale 256*128. Max image pixel is 255,
 *  after one dimention transform, Max coef is about 255*1.5. We can
 *  build look up table to reduce multiplications.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *  f_lowpass::      Ptr to low pass filter
 * f_highpass::      Ptr to high pass filter
 *     lowlen::      low pass filter length
 *    highlen::      high pass filter length
 *     tbllen::      lookup table length
 *
 * Output:::
 *   lowTab::         low pass look up table
 *  highTab::        high pass look up table
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_init_filtertbl (
    IPState    *pIPState,
    Int32      *f_lowpass,
    Int32      *f_highpass,
    Int32	lowlen,
    Int32	highlen,
    Int32	tbllen,
    Int32     **lowTab,
    Int32     **highTab)
{
    PROLOGUE;
    Int32  idx, col;
    
    for(col=0; col<tbllen; ++col)
    {
	for(idx =0; idx <lowlen; ++idx)
	    *(lowTab[idx]+col) = f_lowpass[idx]*col;
    }

    for(col=0; col<tbllen; ++col)
    {
	for(idx =0; idx <highlen; ++idx)
	    *(highTab[idx]+col) = f_highpass[idx]*col;
    }


    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    N level Biorthogonal wavelet transform.
 *
 * LongDescription:::
 *    Do biorthogonal wavelet transform on nth level by use of LUT.
 *
 * Input:::
 *   pIPState::     Ptr to IP struct for memory alloc/freeing routines
 *      coefw::     nth level width
 *      coefh::     nth level height 
 *        bpl::     bytes per line of coef-buffer
 *     lowTab::     Ptr to lowpass filter LUT
 *    highTab::     Ptr to highpass filter LUT
 *    tempbuf::     Ptr to temp buffer
 *
 * Output:::
 *    coefbuf::     Ptr to coef-buffer. 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_waveTrans_LUT (
    IPState    *pIPState,
    Int32	coefw,
    Int32	coefh,
    Int32	bpl,
    Int32     **lowTab,
    Int32     **highTab,
    Int16     **tempbuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    Int32       row, col, idx, halfw, halfh, temp, l_s;
    Int16       *coef, *buffer, *linebuf, *phigh, *plow, *pSrc, *ptm;
    Int16       t0,t1,t2,t3,t4;
    Int32       *low1, *low2, *low3, *low4, *low5;
    Int32       *high1, *high2, *high3, *high4;

    coef = *coefbuf;
    buffer = *tempbuf;
    low1 = lowTab[0];
    low2 = lowTab[1];
    low3 = lowTab[2];
    low4 = lowTab[3];
    low5 = lowTab[4];
    high1 = highTab[0];
    high2 = highTab[1];
    high3 = highTab[2];
    high4 = highTab[3];

    linebuf = buffer +4;
    halfh = (coefh+1)/2;
    halfw = (coefw+1)/2;

    l_s = coefw*sizeof(Int16);
/* collumn */
    for(row=0; row<coefw; ++row)
    {
/* copy image */
	for(idx=0, col=row; idx<coefh; ++idx, col+=bpl)
	    *(linebuf+idx)=*(coef+col);
	for(idx=1, col=coefh-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}

/* convelution */
	plow = coef+ row;
	phigh = coef+ (halfh * bpl)+ row;
	ptm = buffer;
	for(idx=0; idx<coefh; ++idx, plow+= bpl, phigh+=bpl, ++ptm)
	{
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    t4 = *(ptm+8) + *(ptm);
	    temp = (t0 > 0) ? low1[t0] : -low1[-t0];
	    temp += (t1 > 0) ? low2[t1] : -low2[-t1];
	    temp += (t2 > 0) ? low3[t2] : -low3[-t2];
	    temp += (t3 > 0) ? low4[t3] : -low4[-t3];
	    temp += (t4 > 0) ? low5[t4] : -low5[-t4];
	    *plow = (Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefh) break;
	    ++ptm;
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    temp = (t0 > 0) ? high1[t0] : -high1[-t0];
	    temp += (t1 > 0) ? high2[t1] : -high2[-t1];
	    temp += (t2 > 0) ? high3[t2] : -high3[-t2];
	    temp += (t3 > 0) ? high4[t3] : -high4[-t3];
	    *phigh = (Int16)((temp +ONEHALF)>>SCALEBITS);
	}
    }

/* row */
    for(row=0, pSrc=coef; row<coefh; ++row, pSrc+= bpl)
    {
/* copy image */
	MEMCPY(linebuf, pSrc, l_s);
	for(idx=1, col=coefw-1; idx<5; ++idx)
	{
	    *(linebuf-idx) = *(linebuf+idx);
	    *(linebuf+col+idx) = *(linebuf+col-idx);
	}
	plow = pSrc;
	phigh = plow + halfw;
	ptm = buffer;
	for(idx=0; idx<coefw; ++idx, ++ptm)
	{
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    t4 = *(ptm+8) + *(ptm);
	    temp = (t0 > 0) ? low1[t0] : -low1[-t0];
	    temp += (t1 > 0) ? low2[t1] : -low2[-t1];
	    temp += (t2 > 0) ? low3[t2] : -low3[-t2];
	    temp += (t3 > 0) ? low4[t3] : -low4[-t3];
	    temp += (t4 > 0) ? low5[t4] : -low5[-t4];
	    *plow++ = (Int16)((temp +ONEHALF)>>SCALEBITS);
	    if( ++idx >= coefw) break;
	    ++ptm;
	    t0 = *(ptm+4);
	    t1 = *(ptm+5) + *(ptm+3);
	    t2 = *(ptm+6) + *(ptm+2);
	    t3 = *(ptm+7) + *(ptm+1);
	    temp = (t0 > 0) ? high1[t0] : -high1[-t0];
	    temp += (t1 > 0) ? high2[t1] : -high2[-t1];
	    temp += (t2 > 0) ? high3[t2] : -high3[-t2];
	    temp += (t3 > 0) ? high4[t3] : -high4[-t3];
	    *phigh++ = (Int16)((temp+ONEHALF)>>SCALEBITS);
	}
    }

    NORMAL_RETURN;
}


/*
$Log: miwctra.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.5  1998/11/30 19:40:12  xzhu
xzhu on Mon Nov 30 11:40:02 PST 1998

miwctra.c -> /product/ipcore/ipshared/src/RCS/miwctra.c,v

Added new functions.

Revision 1.4  1998/10/20 17:32:52  xzhu
xzhu on Tue Oct 20 10:31:14 PDT 1998

miwctra.c -> /product/ipcore/ipshared/src/RCS/miwctra.c,v

Create new algorithm (lift) for transform.

Revision 1.3  1998/01/13 23:53:49  ddavies
miwctra.c -> /product/ipcore/ipshared/src/RCS/miwctra.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1997/09/17 22:47:59  xzhu
miwctra.c -> /product/ipcore/ipshared/src/RCS/miwctra.c,v

CommonCheck: Added #include wavelib.pub, since some changes made in huffman coding.

Revision 1.1  1996/10/16 00:02:49  xzhu
Initial revision

*/
