/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	wccolor.c: 
 *
 *  Private routine:
 *        ip_image2buffer ()
 *
 */


#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"
#include "imageref.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwcycr_c_RCSInfo, "$RCSfile: miwcycr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

#define R_Y_OFF		 0     		/* offset to R => Y section */
#define R_CB_OFF	(1* 257)
#define R_CR_OFF	(2* 257) 
		
#define G_Y_OFF		(3* 257)	/* offset to G => Y section */
#define G_CB_OFF	(4* 257)
#define G_CR_OFF	(5* 257)

#define B_Y_OFF		(6* 257)	/* etc. */
#define B_CB_OFF	(7* 257)
#define B_CR_OFF        (8* 257)

#define TABLE_SIZE	(9* 257)
#define CENTER_OFF      (CENTER<<SCALEBITS)

static Int32 redPara[3] = {9797, -5529, 16384};
static Int32 greenPara[3] = {19234, -10854, -13719};
static Int32 bluePara[3] = {3735, 16384, -2664};

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Copy 8bit/per pixel image to 16bit/per coef. coef-buffer.
 *
 * LongDescription:::
 *    Copy 8bit/per pixel image to 16bit/per coef. coef-buffer.
 *    Directly copy gray scale image. RGB color image will be converted 
 *    into YCrCb color image.
 *
 * Input:::
 *    ncomps::      Number of image planets
 *       bpl::      Bytes per line
 *         x::      Begining x position
 *         y::      Begining line
 *         w::      Width of copying image (x+w) < bpl
 *         h::      Height of copying image (y+h) < image height
 *  imagebuf::      Ptr to image buffer
 *
 * Output:::
 *  coefbuf::        Ptr to coef-buffer
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_image2buffer (
    IPState    *pIPState,
    Int32	ncomps,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8     **imagebuf,
    Int16     **coefbuf)
{
    PROLOGUE;
    UInt8              *image[3], *red, *green, *blue;
    Int16              *coef[3], cb, cr;
    Int32              halfw;
    Int32              row, col, idx, r, g, b, crow;
    Int32              cidx, width;
    Int32              *rgb_ycc;


/* set up look up table for RGB-YCrCb */
    rgb_ycc = (Int32 *)IP_CALLOC(sizeof(Int32), TABLE_SIZE);

    coef[0] = *coefbuf;
    image[0] = *imagebuf;
    if( ncomps > 1)
    {
	coef[1] = *(coefbuf+1);
	coef[2] = *(coefbuf+2);
	image[1] = *(imagebuf+1);
	image[2] = *(imagebuf+2);

	for(col=0; col<256; ++col)
	{
	    *(rgb_ycc + R_Y_OFF + col) = redPara[0]*col;
	    *(rgb_ycc + R_CB_OFF +col) = redPara[1]*col;
	    *(rgb_ycc + R_CR_OFF +col) = redPara[2]*col;
	    *(rgb_ycc + G_Y_OFF + col) = greenPara[0]*col;
	    *(rgb_ycc + G_CB_OFF +col) = greenPara[1]*col;
	    *(rgb_ycc + G_CR_OFF +col) = greenPara[2]*col;
	    *(rgb_ycc + B_Y_OFF + col) = ONEHALF + (bluePara[0]*col);
	    *(rgb_ycc + B_CB_OFF +col) = ONEHALF + 
		CENTER_OFF + (bluePara[1]*col);
	    *(rgb_ycc + B_CR_OFF +col) = ONEHALF + 
		CENTER_OFF + (bluePara[2]*col);
	}
    
	red = image[0]+ y*bpl;
	green = image[1]+ y*bpl;
	blue = image[2]+ y*bpl;
	halfw = (w+1)>>1;
	width = w-1;
	for(row=y, crow=0; crow<h; ++row, ++crow)
	{
	    for(col=x, idx=0, cidx=0; idx<w; ++col, ++idx)
	    {
/* RGB to YCbCr */
		r = UCHAR_ACCESS((red +col));
		g = UCHAR_ACCESS((green +col));
		b = UCHAR_ACCESS((blue +col));
		*(coef[0]+ idx)= (Int16)(( rgb_ycc[r + R_Y_OFF] + 
		    rgb_ycc[g + G_Y_OFF]+ rgb_ycc[ b + B_Y_OFF] )>> SCALEBITS);
		cb = (Int16)(( rgb_ycc[r + R_CB_OFF] + rgb_ycc[g + G_CB_OFF] + 
			 rgb_ycc[b + B_CB_OFF] )>> SCALEBITS);
		*(coef[1] + cidx) += cb;

		cr = (Int16)(( rgb_ycc[r + R_CR_OFF] + rgb_ycc[g + G_CR_OFF] + 
			 rgb_ycc[b + B_CR_OFF] )>> SCALEBITS);
		*(coef[2] + cidx) += cr;
		if( (idx&1) ) ++cidx;
		else if( idx == width )
		{
		    *(coef[1]+ cidx) += cb;
		    *(coef[2]+ cidx) += cr;
		}
	    }
	    red += bpl;
	    green += bpl;
	    blue += bpl;
	    coef[0] += w;
	    if( (crow&1) )
	    {
		for(cidx=0; cidx<halfw; ++cidx)
		{
		    *(coef[1]+cidx) >>= 2;
		    *(coef[2]+cidx) >>= 2;
		}
		coef[1] += halfw;
		coef[2] += halfw;
	    }
	    else if( crow == (h-1))
	    {
		for(cidx=0; cidx<halfw; ++cidx)
		{
		    *(coef[1]+cidx) >>= 1;
		    *(coef[2]+cidx) >>= 1;
		}
	    }
	}
    }
    else
    {
	image[0] += y*bpl;
	for(row=0; row<h; ++row)
	{
	    for(col=x, idx=0; idx<w; ++col, ++idx)
		*(coef[0]+idx) = UCHAR_ACCESS((image[0]+col));
	    image[0] += bpl;
	    coef[0] += w;
	}
    }

    IP_FREE(rgb_ycc);
    
    NORMAL_RETURN;
}


/*
$Log: miwcycr.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.3  1998/01/13 23:53:50  ddavies
miwcycr.c -> /product/ipcore/ipshared/src/RCS/miwcycr.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1997/09/17 22:48:01  xzhu
miwcycr.c -> /product/ipcore/ipshared/src/RCS/miwcycr.c,v

CommonCheck: Added #include wavelib.pub, since some changes made in huffman coding.

Revision 1.1  1996/10/16 00:02:38  xzhu
Initial revision

*/
