 /*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*	wdarith.c:
 *
 *  Private routine:
 *        ip_arithDecodingInit()
 *        ip_dataDecodingInit()
 *        ip_inputbits()
 *        ip_arithDecoding()
 *        ip_dataDecodingTerm()
 *        ip_planetDecodingEnd()
 *
 */

#include <stdio.h>
#include <math.h>

#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwdari_c_RCSInfo, "$RCSfile: miwdari.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

static ARITHINFO demainInfo;
 
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get current input data bit for decompressing.
 *
 * LongDescription:::
 *  Get current input data bit for decompressing.
 *
 * Input::: 
 *
 * Output:::
 *      bit::     Ptr to input bit.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_inputbits (
    IPState    *pIPState,
    Int32      *bit)
{
    PROLOGUE;
    ARITHINFO  *pInfo;

    pInfo = &demainInfo;
    if(pInfo->bits_to_go == 0)
    {
	pInfo->IOPath(pIPState, pInfo->pdata, &pInfo->buffer, sizeof(UInt8));
	pInfo->bytes[0] = pInfo->bytes[1];
	pInfo->bytes[1] = pInfo->buffer;
	pInfo->bits_to_go =8;
    }
    *bit = (Int32)(pInfo->buffer &1);
    pInfo->buffer >>= 1;
    -- pInfo->bits_to_go;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmic decoding init.
 *
 * LongDescription:::
 *  Arithmic decoding init.
 *
 * Input::: 
 *      pdata::     Ptr to user data for output.
 *  inputFunc::     Ptr to call back input function
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_arithDecodingInit (
    IPState    *pIPState,
    void       *pdata,
    Int32 (CDECL *inputFunc) (
		    IPState    *pIPState,
		    void       *userData,
		    UInt8      *buf,
		    UInt32	count))
{
    PROLOGUE;
    Int16      idx;
    Int32      value, bit;
    ARITHINFO  *pInfo;

    pInfo = &demainInfo;
    pInfo->pdata = pdata;
    pInfo->IOPath = inputFunc; 
    pInfo->low =0;
    pInfo->high = TOP_VALUE;

    pInfo->buffer =0;
    pInfo->bits_to_go = 0;
    pInfo->garbage_bits =0;

    pInfo->value = value =0;
    for(idx=0; idx<CODE_VALUE_BITS; ++idx)
    {
	ERR_PASS(ip_inputbits (pIPState, &bit));
	value = 2*value + bit;
    }
    pInfo->value = value;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic decoding data init.
 *
 * LongDescription:::
 *  Initiate rithmetic decoding for assigned data set.
 *
 * Input:::
 *       infoIdx::     Decoding data set index ( same as coding)
 *  no_of_symbol::     Number of symbol of coding data set
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_dataDecodingInit (
    IPState    *pIPState,
    Int32	infoIdx,
    Int32	no_of_symbol)
{
    PROLOGUE;
    Int16      idx;
    Int16      *symbol_to_index, *index_to_symbol;
    Int32      *freq, *cum_freq;
    SINGLEARITHINFO  *pInfo;

    demainInfo.info[infoIdx] = (void *) IP_CALLOC(1, sizeof(SINGLEARITHINFO));
    pInfo = demainInfo.info[infoIdx];
    pInfo->no_of_symbol = (Int16)no_of_symbol;
    pInfo->no_of_index = (Int16)no_of_symbol+1;
    pInfo->eof_index = (Int16)no_of_symbol+1;
    pInfo->symbol_to_index = (Int16 *)IP_CALLOC(sizeof(Int16), 
		     pInfo->no_of_index*2);
    pInfo->index_to_symbol = pInfo->symbol_to_index + pInfo->no_of_index;
    pInfo->freq = (Int32 *)IP_CALLOC(sizeof(Int32), 
		     (pInfo->no_of_index+1)*2);
    pInfo->cum_freq = pInfo->freq + pInfo->no_of_index+1;

    symbol_to_index = pInfo->symbol_to_index;
    index_to_symbol = pInfo->index_to_symbol;
    for(idx=0; idx<pInfo->no_of_symbol; ++idx)
    {
	symbol_to_index[idx]= idx+1;
	index_to_symbol[idx+1] = idx;
    }
    freq = pInfo->freq;
    cum_freq = pInfo->cum_freq;
    for(idx=0; idx<=pInfo->no_of_index; ++idx)
    {
	freq[idx] =1;
	cum_freq[idx] = pInfo->no_of_index - idx;
    }
    freq[0] =0;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic decoding.
 *
 * LongDescription:::
 *  Arithmetic decoding.
 *
 * Input:::
 *  bitCode::     Fix bits coding or ataptive arithmetic coding
 *  infoIdx::     Data set index of current symbol
 *
 * Output:::
 *   symbol::     Ptr to decoded current symbol
 *   pIndex::     Ptr to current symbol index 
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_arithDecoding (
    IPState    *pIPState,
    Int16      *pIndex,
    Int16      *symbol,
    Int16	bitCode,
    Int32	infoIdx) 
{
    PROLOGUE;
    Int32      range, high, low, cum, value, bit;
    Int16      index, i;
    SINGLEARITHINFO *pInfo;

    high = demainInfo.high;
    low = demainInfo.low;
    value = demainInfo.value;
    range = high - low +1; 

    if( bitCode )
    {
	index =0;
	pInfo = NULL;
	if(bitCode & 0xff00)
	{
	    bit = bitCode >>8;
	    *symbol = (Int16)((((value - low+1)<<bit)-1)/range);
	    high = low -1 + (((*symbol+1)*range)>>bit);
	    low += ((*symbol) * range)>>bit;
	}
	else
	{
	    *symbol = (Int16)((((value - low +1) << 1) - 1)/range);
	    if ( *symbol ) 
	    {
		high = low -1 + range;  
		low += range >> 1; 
	    }
	    else
		high = low-1 + (range >> 1);
	}
    }
    else
    {
	pInfo = demainInfo.info[infoIdx];
	cum = (((value-low)+1)*pInfo->cum_freq[0]-1)/range;
	for(index=1; pInfo->cum_freq[index]>cum; index++);

	high = low + (range*pInfo->cum_freq[index-1])/pInfo->cum_freq[0] -1;
	low = low + (range * pInfo->cum_freq[index])/pInfo->cum_freq[0];
    }

    for(;;)
    {
	if(high < HALF)
	{
	    /* nothing */
	}
	else if(low>= HALF)
	{
	    value -= HALF;
	    low -= HALF;
	    high -= HALF;
	}
	else if (low>= FIRST_QTR && high < THIRD_QTR)
	{
	    value -= FIRST_QTR;
	    low -= FIRST_QTR;
	    high -= FIRST_QTR;
	}
	else break;
	low = low *2;
	high = 2*high +1;
	ERR_PASS(ip_inputbits (pIPState, &bit));
	value = 2*value+bit;
    }
    
    demainInfo.low = low;
    demainInfo.high = high;
    demainInfo.value = value;

    if( bitCode )
    {
	NORMAL_RETURN;
    }
    *pIndex = index;
    if(index == pInfo->eof_index) NORMAL_RETURN;
    *symbol = pInfo->index_to_symbol[index];
    
    if(pInfo->cum_freq[0] == MAX_FREQUENCY)
    {
	cum =0;
	for(i=pInfo->no_of_index; i>=0; --i)
	{
	    pInfo->freq[i] = (pInfo->freq[i]+1)/2;
	    pInfo->cum_freq[i] = cum;
	    cum += pInfo->freq[i];
	}
    }
    for(i = index; pInfo->freq[i]== pInfo->freq[i-1]; --i);
    if( i< index )
    {
	Int16 c_1, c_2;
	c_1 = pInfo->index_to_symbol[i];
	c_2 = pInfo->index_to_symbol[index];
	pInfo->index_to_symbol[i] = c_2;
	pInfo->index_to_symbol[index] = c_1;
	pInfo->symbol_to_index[c_1] = index;
	pInfo->symbol_to_index[c_2] = i;
    }
    ++ pInfo->freq[i];
    while(i>0)
    {
	-- i;
	++ pInfo->cum_freq[i];
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic decoding data term.
 *
 * LongDescription:::
 *  Term one data set's arothmetic decoding info and free memory. 
 *
 * Input:::
 *   infoIdx::        Data set's index
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_dataDecodingTerm (
    IPState    *pIPState,
    Int32	infoIdx)
{
    PROLOGUE; 
    SINGLEARITHINFO  *pInfo;

    pInfo = demainInfo.info[infoIdx];
    IP_FREE(pInfo->symbol_to_index);
    IP_FREE(pInfo->freq);
    IP_FREE(pInfo);
    demainInfo.info[infoIdx] = NULL;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Arithmetic decoding ending alligment.
 *
 * LongDescription:::
 *  Seperated gray color coding. For every planet there is ending
 * alligment. 
 *
 * Input:::
 *   number::        Max extral bytes
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_planetDecodingEnd (
    IPState    *pIPState,
    Int32	number)
{
    PROLOGUE; 
    Int32       col;

    col =0;
    while( col < number &&( demainInfo.bytes[0] != 0 ||
			       demainInfo.bytes[1] != 0 ) )
    {
	demainInfo.bytes[0] = demainInfo.bytes[1];
	demainInfo.IOPath(pIPState, demainInfo.pdata,  
			  &demainInfo.bytes[1], sizeof(UInt8));
	++col;
    }

    NORMAL_RETURN;
}


/*
$Log: miwdari.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.5  1998/01/13 23:53:51  ddavies
miwdari.c -> /product/ipcore/ipshared/src/RCS/miwdari.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.4  1997/09/17 22:48:03  xzhu
miwdari.c -> /product/ipcore/ipshared/src/RCS/miwdari.c,v

CommonCheck: Added #include wavelib.pub, since some changes made in huffman coding.

Revision 1.3  1997/07/11 01:51:27  ddavies
miwdari.c -> /product/ipcore/ipshared/src/RCS/miwdari.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.2  1996/11/08 18:11:29  xzhu
xzhu on Fri Nov 8 10:11:03 PST 1996

miwdari.c -> /product/ipcore/ipshared/src/RCS/miwdari.c,v

Fixed bug in decoding term checking.

Revision 1.1  1996/10/16 00:02:36  xzhu
Initial revision


*/
