 /*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*	wdcolor.c:
 *
 *  Private routine:
 *      ip_buffer2image ()
 *
 */

#include <stdio.h>
#include <math.h>
#include "types.pub"
#include "ipscore.prv"
#include "errmacs.pub"
#include "memory.pub"
#include "ipserrs.pub"
#include "shrutils.pub"
#include "imageref.pub"

#include "wavelib.pub"
#include "wavelib.prv"

IP_RCSINFO(miwdycr_c_RCSInfo, "$RCSfile: miwdycr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:17 $");

#define MAXSAMPLE 255

static Int32 crPara[2] = {45940, -23400};
static Int32 cbPara[2] = {-11276, 58064};

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Copy coef-buffer to image buffer.
 *
 * LongDescription:::
 *   After inverse wavelet transform, we copy decompressed image from
 *   coef-buffer to images buffer. If it is color image, we transform
 *   YCrCb to RGB.
 *
 * Input:::
 *    ncomps::      Number of image planets
 *       bpl::      Bytes per line
 *         x::      Begining x position
 *         y::      Begining line
 *         w::      Width of coef-buffer (x+w) < bpl
 *         h::      Height of coef-buffer (y+h) < image height
 *   coefbuf::      Ptr to coef buffer
 *
 * Output:::
 *   imagebuf::     Ptr to image buffer point
 *        min::     Ptr to min value before limit
 *        max::     Ptr to max value before limit
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL 
ip_buffer2image (
    IPState    *pIPState,
    Int32	ncomps,
    Int32	bpl,
    Int32	x,
    Int32	y,
    Int32	w,
    Int32	h,
    UInt8     **imagebuf,
    Int16     **coefbuf,
    Int16      *min,
    Int16      *max)
{
    PROLOGUE;
    UInt8              *image[3];
    Int16              *coef[3], pixel, yy, cr, cb;
    Int32              row, crow, col, idx, cidx;
    Int32              halfw, halfh;
    Int32              *r_cr, *g_cb, *g_cr, *b_cb;

/* set up look up table for RGB-YCrCb */
    r_cr = (Int32 *)IP_CALLOC(sizeof(Int32), (4*(MAXSAMPLE+1)));

    *min = 0x7fff;
    *max = -0x7fff;
    coef[0] = *coefbuf;
    image[0] = *imagebuf + y*bpl;
    if( ncomps > 1)
    {
	image[1] = *(imagebuf+1) + y*bpl;
	image[2] = *(imagebuf+2) + y*bpl;
	coef[1] = *(coefbuf+1);
	coef[2] = *(coefbuf+2);
	g_cb = r_cr + MAXSAMPLE +1;
	g_cr = g_cb + MAXSAMPLE +1;
	b_cb = g_cr + MAXSAMPLE +1;
	for(col=0; col<=MAXSAMPLE; ++col)
	{
	    r_cr[col] = (ONEHALF + (crPara[0]*(col - CENTER)))>>SCALEBITS;
	    g_cr[col] = crPara[1]*(col - CENTER);
	    g_cb[col] = cbPara[0]*(col - CENTER);
	    b_cb[col] = (ONEHALF + (cbPara[1]*(col - CENTER)))>>SCALEBITS;
	}

	halfw = (w+1)>>1;
	halfh = (h+1)>>1;
	for(row=y, crow=0; crow<h; ++row, ++crow)
	{
	    cb = *coef[1];
	    cr = *coef[2];
	    for(col=x, idx=0, cidx=0; idx<w; ++col, ++idx)
	    {
		yy = *(coef[0]+idx);
		if( !(idx&1) )
		{    
		    cb = *(coef[1]+cidx);
		    cr = *(coef[2]+cidx);
		    if(cb <0) cb =0;
		    if(cr <0) cr =0; 
		    if(cb > MAXSAMPLE) cb = MAXSAMPLE;
		    if(cr > MAXSAMPLE) cr = MAXSAMPLE;
		}
		else ++cidx;
/*		if(yy < 0) yy=0;
	        if(yy > MAXSAMPLE) yy = MAXSAMPLE;  */
		pixel = yy + (Int16)r_cr[cr];
		if(pixel < *min) *min = pixel;
		if(pixel > *max) *max = pixel;
		if(pixel > 255) pixel = 255;
		if(pixel < 0) pixel =0;
		UCHAR_ACCESS((image[0]+ col)) = (UInt8) pixel;
		pixel= yy + (Int16)(( g_cb[cb]+ g_cr[cr]+ ONEHALF)>>SCALEBITS);
		if(pixel < *min) *min = pixel;
		if(pixel > *max) *max = pixel;
		if(pixel > 255) pixel = 255;
		if(pixel < 0) pixel =0;
		UCHAR_ACCESS((image[1]+ col)) = (UInt8) pixel;
		pixel = yy + (Int16) b_cb[cb];
		if(pixel < *min) *min = pixel;
		if(pixel > *max) *max = pixel;
		if(pixel > 255) pixel = 255;
		if(pixel < 0) pixel =0;
		UCHAR_ACCESS((image[2]+ col)) = (UInt8) pixel;
	    }
	    image[0] += bpl;
	    image[1] += bpl;
	    image[2] += bpl;
	    coef[0] += w;
	    if( (crow&1) )
	    {
		coef[1] += halfw;
		coef[2] += halfw;
	    }
	}
    }
    else
    {
	for(row=0; row<h; ++row)
	{
	    for(col=x, idx=0; idx<w; ++col, ++idx)
	    {
		pixel = *(coef[0]+idx);
		if( pixel < *min) *min = pixel;
		if( pixel > *max) *max = pixel;
		if( pixel < 0) UCHAR_ACCESS((image[0]+col)) =0;
		else if( pixel > 255) UCHAR_ACCESS((image[0]+col)) = 255;
		else UCHAR_ACCESS((image[0]+col)) = (UInt8) pixel;
	    }
	    image[0] += bpl;
	    coef[0] += w;
	}
    }

    IP_FREE(r_cr);
    
    NORMAL_RETURN;
}


/*
$Log: miwdycr.c,v $
Revision 1.1.1.1  1999/09/20 18:18:17  jlin
Imported sources

Revision 1.3  1998/01/13 23:53:56  ddavies
miwdycr.c -> /product/ipcore/ipshared/src/RCS/miwdycr.c,v

CommonCheck: Add IPState parameter to almost all procedures.

Revision 1.2  1997/09/17 22:48:09  xzhu
miwdycr.c -> /product/ipcore/ipshared/src/RCS/miwdycr.c,v

CommonCheck: Added #include wavelib.pub, since some changes made in huffman coding.

Revision 1.1  1996/10/16 00:02:41  xzhu
Initial revision


*/
