;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcbswap.asm:  template for assembly code.

	.486		; compile code for a 486 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

INCLUDE	pcproctp.prv

; There seem to be two popular file formats on PCs.  In one, the leftmost
; byte in an image is in the leftmost byte of a 32-bit word.  That's the
; format we use.  In the other, the leftmost image byte is in byte 0
; of a 32-bit word.  That's the same as our format on big endian machines,
; but different on little endian machines.  The first format has the
; advantage that left and right shifts within 32-bit words work correctly.
; The second format has the advantage that moving from left to right
; always means incrementing memory addresses.
;
; This file contains low level code to convert between these formats.
; There are three routines.  Two of them swap the bytes of each word in
; an image line.  One of these uses rotates and is intended for 386 platforms.
; The other uses the 486's bswap opcode.  The third procedure is called
; to return a pointer to the proper routine for the current platform.
;
; UPDATE: We have added 8 more routines.  One does 486 style byteswapping and
; takes a far source pointer.  One does 486 style byteswapping and takes
; a far dest pointer.  Two more do 386 style byteswapping and take
; far dest or source pointers.  Two more do 486 style byteswapping where the
; data is inverted before being written; one takes a far dest pointer and the
; other a far source pointer.  The last two do 386 style byteswapping and
; invert the data.
;
; How does ip_getByteSwapProc decide what platform we're on?  It uses the
; procedure specified (sort of) on pages 3-43 and 3-44 of the "Intel486
; Microprocessor Family Programmer's Reference Manual".  The idea is that
; 486 processors have a bit in the flags register that 386's don't have.
; This is the AC or Alignment Check mode bit.  When set (and when the AM
; bit in control register 0 is set), memory accesses that are not properly
; aligned will cause traps.  We don't want to cause any traps, but we do
; want to see if the bit exists.  We do this by reading the flags,
; flipping the bits, cycling the new value through the flags, and seeing
; if it stayed flipped.  If so, this is a 486.  If not, this is a 386.
; ip_getByteSwapProc treats the arg passed in as the address of a pointer to
; a procedures.  It writes the address of the start of either
; ip_swapLineBytes386 or ip_swapLineBytes486 into the function pointer
; using this address.
;
; The byte swapping routines are straightforward.  Both take image
; pointers and the number of words in the line as arguments.  Both
; proceed down the line, byteswapping as we go.  Both use ECX as a loop
; counter.  The routines that use far pointers use ES as the other segment
; register.

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; In the SPARC algorithms, we went to a great deal of trouble to use double
; word (64 bit) write operations.  Writes on an SS2 are slow, so it was
; beneficial.  The 386/486 has relatively fast 32 bit writes and no 64 bit
; writes, so we will discard that complication.

; Constant definitions for the "direction" argument.
; DO NOT CHANGE THESE WITHOUT CHANGING THE CORRESPONDING DEFINITIONS IN
; shrpixr.pub!!!!!
cDIBToPixr	EQU	0	; the source is a DIB, the dest is a Pixr
cPixrToDIB	EQU	1	; the source is a Pixr, the dest is a DIB
cPixrToPixr	EQU	2	; both source an dest are Pixrs

_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcbswap.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment (byte swap routines):
; EAX:	data word
; EBX:	dst image address 
; ECX:	word count
; EDX:	src image address
; EBP:	Addresses locals and parameters on stack
; ESI:	
; EDI:	
; ESP:	stack pointer

rData		EQU eax
rWCount		EQU ecx
rDstAddr	EQU ebx
rSrcAddr	EQU edx

; Local Variables:
; procType:	type of processor we're running on now
; srcAddr:	address of left end of current source line
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPDWORD TYPEDEF NEAR PTR DWORD
FPBYTE TYPEDEF FAR PTR BYTE
FPDWORD TYPEDEF FAR PTR DWORD

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF


; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Code used when we have far pointers (16 bit world)
; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

IFNDEF AsmForNT

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_getByteSwapProc.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers
; The arguments are declared in the order specified by cbswap.c

ip_getByteSwapProcStart::
ip_getByteSwapProc	PROC NEAR C PUBLIC USES EAX EBX ECX,\
	pIPState:DWORD, invertWords:DWORD,
	direction:DWORD, pSwapToFarProc:DWORD,
	pSwapFromFarProc:DWORD, pSwapNearProc:DWORD
	LOCAL procType:DWORD
; All registers used are also pushed on the stack automatically.

; make the assembler happy
	mov	eax, pIPState

; Figure out whether we're running on a 386 or a 486 (or above)
	INVOKE	ip_getProcessorType, pIPState, ADDR procType
	cmp	procType, cProcIs386
	je	gspIs386

; This is a 486
	cmp	invertWords, 0	; Should we invert the words?
	jne	gspInvertWords486	; yes, look at the direction

; Do not invert the data words, look at the direction to decide which input
; pointer is a far pointer.
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDest486
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrc486
; Neither the source nor the dest is far
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear486
	jmp	gspFin			; restore things are return

gspSetDIBSrc486:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesFromFar486
	jmp	gspFin

gspSetDIBDest486:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesToFar486
	jmp	gspFin

; The words should be inverted while being byteswapped.
gspInvertWords486:
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDestInverted486
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrcInverted486
; Neither the source nor the dest is far.
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear486
	jmp	gspFin

gspSetDIBSrcInverted486:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedFromFar486
	jmp	gspFin

gspSetDIBDestInverted486:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedToFar486
	jmp	gspFin



gspIs386:
; This is a 386
	cmp	invertWords, 0	; Should we invert the words?
	jne	gspInvertWords386	; yes, look at the direction

; Do not invert the data words, look at the direction to decide which input
; pointer is a far pointer.
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDest386
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrc386
; Neither the source nor the dest is far
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear386
	jmp	gspFin			; restore things are return

gspSetDIBSrc386:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesFromFar386
	jmp	gspFin

gspSetDIBDest386:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesToFar386
	jmp	gspFin

; The words should be inverted while being byteswapped.
gspInvertWords386:
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDestInverted386
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrcInverted386
; Neither the source nor the dest is far.
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear386
	jmp	gspFin

gspSetDIBSrcInverted386:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedFromFar386
	jmp	gspFin

gspSetDIBDestInverted386:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedToFar386

gspFin:
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_getByteSwapProc ENDP



; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Code used when we have NO far pointers (32 bit world)
; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ELSE	; we're assembling for NT and Win95
; Well, there really aren't any far pointers on these platforms, but
; code running above this may well expect exactly one of the three
; output function pointers to change anyway, so we'll be very careful
; to change the right one.  However, there are only four procedures
; in this case, ip_swapLineBytesNear486, ip_swapLineBytesInvertedNear486,
; ip_swapLineBytesNear386 and ip_swapLineBytesInvertedNear386.

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_getByteSwapProc.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers
; The arguments are declared in the order specified by cbswap.c

ip_getByteSwapProcStart::
ip_getByteSwapProc	PROC NEAR C PUBLIC USES EAX EBX ECX,\
	pIPState:DWORD, invertWords:DWORD,
	direction:DWORD, pSwapToFarProc:DWORD,
	pSwapFromFarProc:DWORD, pSwapNearProc:DWORD
	LOCAL procType:DWORD
; All registers used are also pushed on the stack automatically.

; make the assembler happy
	mov	eax, pIPState

; Use direction to eliminate warning
	mov	eax, direction

; Figure out whether we're running on a 386 or a 486 (or above)
	INVOKE	ip_getProcessorType, pIPState, ADDR procType
	cmp	procType, cProcIs386
	je	gspIs386

; This is a 486
	cmp	invertWords, 0		; Should we invert the words?
	jne	gspInvertWords486	; yes, look at the direction

; Do not invert the data words, look at the direction to decide which input
; pointer is a far pointer.
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDest486
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrc486
; Neither the source nor the dest is far
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear486
	jmp	gspFin			; restore things are return

gspSetDIBSrc486:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear486
	jmp	gspFin

gspSetDIBDest486:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear486
	jmp	gspFin

; The words should be inverted while being byteswapped.
gspInvertWords486:
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDestInverted486
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrcInverted486
; Neither the source nor the dest is far.
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear486
	jmp	gspFin

gspSetDIBSrcInverted486:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear486
	jmp	gspFin

gspSetDIBDestInverted486:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear486
	jmp	gspFin


gspIs386:
; This is a 386
	cmp	invertWords, 0	; Should we invert the words?
	jne	gspInvertWords386	; yes, look at the direction

; Do not invert the data words, look at the direction to decide which input
; pointer is a far pointer.
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDest386
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrc386
; Neither the source nor the dest is far
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear386
	jmp	gspFin			; restore things are return

gspSetDIBSrc386:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear386
	jmp	gspFin

gspSetDIBDest386:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesNear386
	jmp	gspFin

; The words should be inverted while being byteswapped.
gspInvertWords386:
	cmp	direction, cPixrToDIB	; is the dest a DIB?
	je	gspSetDIBDestInverted386
	cmp	direction, cDIBToPixr	; is the source a DIB?
	je	gspSetDIBSrcInverted386
; Neither the source nor the dest is far.
	mov	eax, pSwapNearProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear386
	jmp	gspFin

gspSetDIBSrcInverted386:
	mov	eax, pSwapFromFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear386
	jmp	gspFin

gspSetDIBDestInverted386:
	mov	eax, pSwapToFarProc
	mov	DWORD PTR [eax], OFFSET ip_swapLineBytesInvertedNear386

gspFin:
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_getByteSwapProc ENDP


ENDIF	; AsmForNT




; ----------------------------------------------------------------------
; The byte swap procedures are really simple.  They just loop through
; the image line byte-swapping each word.

; ----------------------------------------------------------------------
;	386 Procedures
;--------------------------------------------------------------------

; ----------------------------------------------------------------------
;	ip_swapLineBytesNear386
;--------------------------------------------------------------------

ip_swapLineBytesNear386Start::
ip_swapLineBytesNear386	PROC NEAR C PUBLIC USES EAX EBX ECX EDX, \
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

slbn386Loop:
	mov	al, [rSrcAddr+1]
	mov	ah, [rSrcAddr+0]
	rol	eax, 16			; swap these bytes into position
	mov	al, [rSrcAddr+3]
	mov	ah, [rSrcAddr+2]	; the word is fully formed
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbn386Loop

; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesNear386 ENDP


; ----------------------------------------------------------------------
;	ip_swapLineBytesInvertedNear386
;--------------------------------------------------------------------

ip_swapLineBytesInvertedNear386Start::
ip_swapLineBytesInvertedNear386	PROC NEAR C PUBLIC USES EAX EBX ECX EDX, \
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

slbin386Loop:
	mov	al, [rSrcAddr+1]
	mov	ah, [rSrcAddr+0]
	rol	eax, 16			; swap these bytes into position
	mov	al, [rSrcAddr+3]
	mov	ah, [rSrcAddr+2]	; the word is fully formed
	not	eax
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbin386Loop

; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesInvertedNear386 ENDP


; ----------------------------------------------------------------------
;	ip_swapLineBytesToFar386
;--------------------------------------------------------------------

ip_swapLineBytesToFar386Start::
ip_swapLineBytesToFar386	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:FPDWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	les	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

slbtf386Loop:
	mov	al, [rSrcAddr+1]
	mov	ah, [rSrcAddr+0]
	rol	eax, 16			; swap these bytes into position
	mov	al, [rSrcAddr+3]
	mov	ah, [rSrcAddr+2]	; the word is fully formed
	mov	ES:[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbtf386Loop

; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesToFar386 ENDP



; ----------------------------------------------------------------------
;	ip_swapLineBytesFromFar386
;--------------------------------------------------------------------

ip_swapLineBytesFromFar386Start::
ip_swapLineBytesFromFar386	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:FPDWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	les	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

slbff386Loop:
	mov	al, ES:[rSrcAddr+1]
	mov	ah, ES:[rSrcAddr+0]
	rol	eax, 16			; swap these bytes into position
	mov	al, ES:[rSrcAddr+3]
	mov	ah, ES:[rSrcAddr+2]	; the word is fully formed
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbff386Loop

; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesFromFar386 ENDP



; ----------------------------------------------------------------------
;	ip_swapLineBytesInvertedToFar386
;--------------------------------------------------------------------

ip_swapLineBytesInvertedToFar386Start::
ip_swapLineBytesInvertedToFar386 PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:FPDWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	les	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

slbitf386Loop:
	mov	al, [rSrcAddr+1]
	mov	ah, [rSrcAddr+0]
	rol	eax, 16			; swap these bytes into position
	mov	al, [rSrcAddr+3]
	mov	ah, [rSrcAddr+2]	; the word is fully formed
	not	eax			; invert the word
	mov	ES:[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbitf386Loop

; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesInvertedToFar386 ENDP



; ----------------------------------------------------------------------
;	ip_swapLineBytesFromFar386
;--------------------------------------------------------------------

ip_swapLineBytesInvertedFromFar386Start::
ip_swapLineBytesInvertedFromFar386 PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:FPDWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	les	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

slbiff386Loop:
	mov	al, ES:[rSrcAddr+1]
	mov	ah, ES:[rSrcAddr+0]
	rol	eax, 16			; swap these bytes into position
	mov	al, ES:[rSrcAddr+3]
	mov	ah, ES:[rSrcAddr+2]	; the word is fully formed
	not	eax			; invert the word
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbiff386Loop

; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesInvertedFromFar386 ENDP



;--------------------------------------------------------------------
;	486 Procedures
;--------------------------------------------------------------------

; ----------------------------------------------------------------------
;	ip_swapLineBytesNear486
;--------------------------------------------------------------------

ip_swapLineBytesNear486Start::
ip_swapLineBytesNear486	PROC NEAR C PUBLIC USES EAX EBX ECX EDX,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

	cmp	rWCount, 8
	jbe	slbn486LastLoop		; do them 1 at a time if the count
					; is less than or equal to 8

; There are at least 9 words left in the line, do a block of 8
slbn486Loop8x:
	mov	eax, [rSrcAddr]
	bswap	eax
	mov	[rDstAddr], eax
	mov	eax, [rSrcAddr+1*4]
	bswap	eax
	mov	[rDstAddr+1*4], eax
	mov	eax, [rSrcAddr+2*4]
	bswap	eax
	mov	[rDstAddr+2*4], eax
	mov	eax, [rSrcAddr+3*4]
	bswap	eax
	mov	[rDstAddr+3*4], eax
	mov	eax, [rSrcAddr+4*4]
	bswap	eax
	mov	[rDstAddr+4*4], eax
	mov	eax, [rSrcAddr+5*4]
	bswap	eax
	mov	[rDstAddr+5*4], eax
	mov	eax, [rSrcAddr+6*4]
	bswap	eax
	mov	[rDstAddr+6*4], eax
	mov	eax, [rSrcAddr+7*4]
	bswap	eax
	mov	[rDstAddr+7*4], eax
	add	rSrcAddr, 4*8
	add	rDstAddr, 4*8
	sub	rWCount, 8
	cmp	rWCount, 8
	ja	slbn486Loop8x		; do another block of 8 if there are
					; another 8 words ready

; There are 8 or fewer words left, do them 1 at a time
slbn486LastLoop:
	cmp	rWCount, 0
	je	slbn486Done		; nothing left to do
slbn486LastLoopCont:
	mov	eax, [rSrcAddr]
	bswap	eax
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbn486LastLoopCont

;
; We're done, so return.  The assembler turns this into a PILE of code!
slbn486Done:
	ret
ip_swapLineBytesNear486 ENDP


; ----------------------------------------------------------------------
;	ip_swapLineBytesInvertedNear486
;--------------------------------------------------------------------

ip_swapLineBytesInvertedNear486Start::
ip_swapLineBytesInvertedNear486	PROC NEAR C PUBLIC USES EAX EBX ECX EDX,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

	cmp	rWCount, 8
	jbe	slbin486LastLoop	; do them 1 at a time if the count
					; is less than or equal to 8

; There are at least 9 words left in the line, do a block of 8
slbin486Loop8x:
	mov	eax, [rSrcAddr]
	bswap	eax
	not	eax
	mov	[rDstAddr], eax
	mov	eax, [rSrcAddr+1*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+1*4], eax
	mov	eax, [rSrcAddr+2*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+2*4], eax
	mov	eax, [rSrcAddr+3*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+3*4], eax
	mov	eax, [rSrcAddr+4*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+4*4], eax
	mov	eax, [rSrcAddr+5*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+5*4], eax
	mov	eax, [rSrcAddr+6*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+6*4], eax
	mov	eax, [rSrcAddr+7*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+7*4], eax
	add	rSrcAddr, 4*8
	add	rDstAddr, 4*8
	sub	rWCount, 8
	cmp	rWCount, 8
	ja	slbin486Loop8x		; do another block of 8 if there are
					; another 8 words ready

; There are 8 or fewer words left, do them 1 at a time
slbin486LastLoop:
	mov	eax, [rSrcAddr]
	bswap	eax
	not	eax
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbin486LastLoop

;
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesInvertedNear486 ENDP


; ----------------------------------------------------------------------
;	ip_swapLineBytesToFar486
;--------------------------------------------------------------------

ip_swapLineBytesToFar486Start::
ip_swapLineBytesToFar486	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:FPDWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	les	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

	cmp	rWCount, 8
	jbe	slbtf486LastLoop	; do them 1 at a time if the count
					; is less than or equal to 8

; There are at least 9 words left in the line, do a block of 8
slbtf486Loop8x:
	mov	eax, [rSrcAddr]
	bswap	eax
	mov	ES:[rDstAddr], eax
	mov	eax, [rSrcAddr+1*4]
	bswap	eax
	mov	ES:[rDstAddr+1*4], eax
	mov	eax, [rSrcAddr+2*4]
	bswap	eax
	mov	ES:[rDstAddr+2*4], eax
	mov	eax, [rSrcAddr+3*4]
	bswap	eax
	mov	ES:[rDstAddr+3*4], eax
	mov	eax, [rSrcAddr+4*4]
	bswap	eax
	mov	ES:[rDstAddr+4*4], eax
	mov	eax, [rSrcAddr+5*4]
	bswap	eax
	mov	ES:[rDstAddr+5*4], eax
	mov	eax, [rSrcAddr+6*4]
	bswap	eax
	mov	ES:[rDstAddr+6*4], eax
	mov	eax, [rSrcAddr+7*4]
	bswap	eax
	mov	ES:[rDstAddr+7*4], eax
	add	rSrcAddr, 4*8
	add	rDstAddr, 4*8
	sub	rWCount, 8
	cmp	rWCount, 8
	ja	slbtf486Loop8x		; do another block of 8 if there are
					; another 8 words ready

; There are 8 or fewer words left, do them 1 at a time
slbtf486LastLoop:
	mov	eax, [rSrcAddr]
	bswap	eax
	mov	ES:[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbtf486LastLoop

;
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesToFar486 ENDP


; ----------------------------------------------------------------------
;	ip_swapLineBytesFromFar486
;--------------------------------------------------------------------

ip_swapLineBytesFromFar486Start::
ip_swapLineBytesFromFar486	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:FPDWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	les	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

	cmp	rWCount, 8
	jbe	slbff486LastLoop	; do them 1 at a time if the count
					; is less than or equal to 8

; There are at least 9 words left in the line, do a block of 8
slbff486Loop8x:
	mov	eax, ES:[rSrcAddr]
	bswap	eax
	mov	[rDstAddr], eax
	mov	eax, ES:[rSrcAddr+1*4]
	bswap	eax
	mov	[rDstAddr+1*4], eax
	mov	eax, ES:[rSrcAddr+2*4]
	bswap	eax
	mov	[rDstAddr+2*4], eax
	mov	eax, ES:[rSrcAddr+3*4]
	bswap	eax
	mov	[rDstAddr+3*4], eax
	mov	eax, ES:[rSrcAddr+4*4]
	bswap	eax
	mov	[rDstAddr+4*4], eax
	mov	eax, ES:[rSrcAddr+5*4]
	bswap	eax
	mov	[rDstAddr+5*4], eax
	mov	eax, ES:[rSrcAddr+6*4]
	bswap	eax
	mov	[rDstAddr+6*4], eax
	mov	eax, ES:[rSrcAddr+7*4]
	bswap	eax
	mov	[rDstAddr+7*4], eax
	add	rSrcAddr, 4*8
	add	rDstAddr, 4*8
	sub	rWCount, 8
	cmp	rWCount, 8
	ja	slbff486Loop8x		; do another block of 8 if there are
					; another 8 words ready

; There are 8 or fewer words left, do them 1 at a time
slbff486LastLoop:
	mov	eax, ES:[rSrcAddr]
	bswap	eax
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbff486LastLoop

;
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesFromFar486 ENDP


; ----------------------------------------------------------------------
;	ip_swapLineBytesInvertedToFar486
;--------------------------------------------------------------------

ip_swapLineBytesInvertedToFar486Start::
ip_swapLineBytesInvertedToFar486 PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:FPDWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	les	rDstAddr, dstAddr
	mov	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

	cmp	rWCount, 8
	jbe	slbitf486LastLoop	; do them 1 at a time if the count
					; is less than or equal to 8

; There are at least 9 words left in the line, do a block of 8
slbitf486Loop8x:
	mov	eax, [rSrcAddr]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr], eax
	mov	eax, [rSrcAddr+1*4]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr+1*4], eax
	mov	eax, [rSrcAddr+2*4]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr+2*4], eax
	mov	eax, [rSrcAddr+3*4]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr+3*4], eax
	mov	eax, [rSrcAddr+4*4]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr+4*4], eax
	mov	eax, [rSrcAddr+5*4]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr+5*4], eax
	mov	eax, [rSrcAddr+6*4]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr+6*4], eax
	mov	eax, [rSrcAddr+7*4]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr+7*4], eax
	add	rSrcAddr, 4*8
	add	rDstAddr, 4*8
	sub	rWCount, 8
	cmp	rWCount, 8
	ja	slbitf486Loop8x		; do another block of 8 if there are
					; another 8 words ready

; There are 8 or fewer words left, do them 1 at a time
slbitf486LastLoop:
	mov	eax, [rSrcAddr]
	bswap	eax
	not	eax
	mov	ES:[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbitf486LastLoop

;
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesInvertedToFar486 ENDP


; ----------------------------------------------------------------------
;	ip_swapLineBytesInvertedFromFar486
;--------------------------------------------------------------------

ip_swapLineBytesInvertedFromFar486Start::
ip_swapLineBytesInvertedFromFar486 PROC NEAR C PUBLIC USES EAX EBX ECX EDX ES,\
	pIPState:DWORD, srcAddr:FPDWORD, dstAddr:DWORD, srcPitch:DWORD

; make the assembler happy
	mov	eax, pIPState

; assume that srcAddr and dstAddr are 32-bit aligned.
; Load the args into registers
	mov	rDstAddr, dstAddr
	les	rSrcAddr, srcAddr
	mov	rWCount, srcPitch	; this is already in units of dwords

	cmp	rWCount, 8
	jbe	slbiff486LastLoop	; do them 1 at a time if the count
					; is less than or equal to 8

; There are at least 9 words left in the line, do a block of 8
slbiff486Loop8x:
	mov	eax, ES:[rSrcAddr]
	bswap	eax
	not	eax
	mov	[rDstAddr], eax
	mov	eax, ES:[rSrcAddr+1*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+1*4], eax
	mov	eax, ES:[rSrcAddr+2*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+2*4], eax
	mov	eax, ES:[rSrcAddr+3*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+3*4], eax
	mov	eax, ES:[rSrcAddr+4*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+4*4], eax
	mov	eax, ES:[rSrcAddr+5*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+5*4], eax
	mov	eax, ES:[rSrcAddr+6*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+6*4], eax
	mov	eax, ES:[rSrcAddr+7*4]
	bswap	eax
	not	eax
	mov	[rDstAddr+7*4], eax
	add	rSrcAddr, 4*8
	add	rDstAddr, 4*8
	sub	rWCount, 8
	cmp	rWCount, 8
	ja	slbiff486Loop8x		; do another block of 8 if there are
					; another 8 words ready

; There are 8 or fewer words left, do them 1 at a time
slbiff486LastLoop:
	mov	eax, ES:[rSrcAddr]
	bswap	eax
	not	eax
	mov	[rDstAddr], eax
	add	rSrcAddr, 4
	add	rDstAddr, 4
	dec	rWCount			; this is faster than LOOP!
	jnz	slbiff486LastLoop

;
; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_swapLineBytesInvertedFromFar486 ENDP


_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
; no global initialized data used
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END


IF 0				; These are all comments

COMMENT ^	Just to make sure
/*
 * $Log: pcbswap.asm,v $
 * Revision 1.1.1.1  1999/09/20 18:18:18  jlin
 * Imported sources
 *
 * Revision 1.16  1998/09/25 22:12:00  aber
 * aber on Fri Sep 25 15:10:57 PDT 1998
 *
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Fixed byte-swap bug, caused when no whole words is 0.
 *
 * Revision 1.15  1998/01/13 23:58:01  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * CommonCheck: Add IPState parameter to almost all private routines.
 *
 * Revision 1.14  1996/09/17 00:45:18  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
 * in which it randomly decides that some (always uncommented?) lines
 * are <quote> too long <unquote>.  This behavior seems to depend on the
 * total offset into the file when the end of each line is seen.  In
 * the previous version, the number of characters used for the Win3.1 and
 * WinNT versions were not quite the same.  This caused MASM running on
 * the two platforms to randomly decide different lines were too long.
 * This idiotic workaround pads the number of characters inserted so that
 * the number of characters seen by the assembler is the same in both cases.
 * So long as this is true, MASM is both environments exhibits the same
 * behavior.  The lines it decides are too long can have comments added to
 * them (which seems to fix the problem!!).
 *
 * Revision 1.13  1996/09/16 00:49:39  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
 * on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
 * for NT; 3) ASSUME DS is FLAT when assembling for WinNT.
 *
 * Revision 1.12  1995/10/18 04:04:28  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Add casts and new procedure types to make Visual C++ happy.
 *
# Revision 1.11  1995/09/22  21:25:43  ddavies
# ddavies on Fri Sep 22 14:24:52 PDT 1995
#
# pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
#
# Ooops.  Add the -not- opcode to the ip_swapLineBytesInvertedNear routine
# so that it really inverts.
#
# Revision 1.10  1995/09/15  21:47:24  ddavies
# pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
#
# CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
# IF 0 that was already there.  Why does David Hirsch's version of ML get
# upset about the # signs that RCS puts at the beginnings of line (sometimes)
# while our systems are happy??
#
# Revision 1.9  1995/09/15  21:13:30  ddavies
# ddavies on Fri Sep 15 14:13:26 PDT 1995
#
# pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
#
# Try again.
#
 * Revision 1.8  1995/09/15  21:10:01  ddavies
 * ddavies on Fri Sep 15 14:09:32 PDT 1995
 *
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Try to change # to * to make David Hirsch's version of ML happy.
 *
 * Revision 1.7  1995/09/13  00:18:24  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Changes for NT.  Create a duplicate version of ip_getByteSwapProc that is
 * only assembled when the AsmForNT flag is defined.  In the new version,
 * we return the address of ip_swapLineBytesNear or ip_swapLineBytesInvertedNear
 * depending on the state of the invert flag.  The direction flag is ignored
 * since there are no far pointers in that world.  In the original version
 * of ip_getByteSwapProc, return a pointer to ip_swapLineBytesInvertedNear if
 * the direction is cPixrToPixr (near pointers) and the invert flag is set.
 * This allows us to test ip_swapLineBytesInvertedNear even on plain old
 * Windows 3.1 machines (using i_swapImageBytes with the new invertImage flag
 * set).
 *
 * Revision 1.6  1995/05/19  02:56:56  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Move the code used to determine the processor type into pcproctp.asm.
 * This code now INVOKE's the new ip_getProcessorType.
 *
 * Revision 1.5  1994/01/22  01:11:17  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Change comment to say shrpixr.pub instead of shrio.pub.
 *
 * Revision 1.4  1993/11/05  01:51:46  ddavies
 * pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
 *
 * Add RCS magic to have string specifying this file's name, rev number and
 * last edit date into the object file.  Also add change log at the bottom.

 ^

 */

ENDIF

