;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pccmpgry.asm:  grayscale thresholding.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; This file contains assembly code to compare scaled versions of
; two 8 bit gray lines, producing a binary result.  A bit in the binary
; result is set if image1Pixel; mult1 > image2Pixel; mult2.
;
; This routine is called from i_compareGrayImages in i_cmpgry.c.  It
; performs the same function as ip_compareGrayLineScaleP1 and
; ip_compareGrayLineScaleP2 in micmpgry.c, but has a very different
; structure.
;
; The basic idea here is to read the two source bytes, run one of them
; (the byte from image 1 in the ip_compareGrayLineScaleP1, the byte from
; image 2 in ip_compareGrayLineScaleP2) through a table that computes
; either mult2/mult1 * pixel (...ScaleP2) or mult1/mult2 * pixel (...ScaleP1)
; and compare the results.  The carry bit set by the comparison is
; left shifted into a result word by adding the result word to itself
; with carry.
;
; Since the result word is being shifted left, we process the line left to
; right.  We explicitly write out 8 copies of this sequence to avoid
; incrementing pointers, decrementing counters and testing results too
; often.  We execute the 8x loop as four times in a row, writing out
; a result word at the end.
;
; I started writing out 32 copies of the code that produces a bit (instead
; of 8, but that seemed excessive.
; When there are less than 32 pixels left in a line, we fall back
; into a normal loop that does one pixel at a time.  When it completes,
; it shifts the result word left by (32 - (number of output bits computed))
; before writing it to the output.


_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pccmpgry.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment:
; EAX:	byte from image1, scratch
; EBX:	address to table which scales the bytes
; ECX:	byte from image2, shift count
; EDX:	output word
; EBP:	points to variables on the stack
; ESI:	source address 1
; EDI:	source address 2
; ESP:	stack pointer

rSrcAddr1	EQU esi
rSrcAddr2	EQU edi
rInput1		EQU al
rInput1Hi	EQU ah
rInput1Long	EQU eax
rInput2		EQU cl		; also used as shift count
rInput2Hi	EQU ch
rInput2Long	EQU ecx
rTabAddr	EQU ebx
rOWord		EQU edx


; Parameters:
; srcAddr1:	address of left end of current source line in image 1
; srcAddr2:	address of left end of current source line in image 2
; dstAddr:	address of left end of current dest line
; scaleTable:	address of table used to scale inputs
; imageWidth:	number of 32 bit words in destination image line

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPDWORD TYPEDEF NEAR PTR DWORD

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_compareGrayLineScaleP1.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will process the image from top to bottom.
; We will move from right to left within each line of the image.


;---------------------------------------------------------------------------
;	ip_compareGrayLineScaleP1
;---------------------------------------------------------------------------

ip_compareGrayLineScaleP1Start::
ip_compareGrayLineScaleP1 PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr1:DWORD, srcAddr2:DWORD, \
	dstAddr:DWORD, scaleTable:DWORD, \
	imageWidth:DWORD

; make the assembler happy
	mov	eax, pIPState

; load up the addresses
	mov	rTabAddr, scaleTable
	mov	rSrcAddr1, srcAddr1
	mov	rSrcAddr2, srcAddr2

; If there are more than 32 pixels left in the line, process them in the
; 32-pixel-at-a-time section.  Otherwise, go to the 1-pixel-at-a-time loop.
; Note the weird addressing; this is used because we address Little Endian
; memory as if we were reading contiguous 32-bit things instead of
; bytes.  That is, we think that pixels are stored left to right within
; 32-bit words on all platforms.
	mov	rInput1Long, imageWidth
	cmp	rInput1Long, 32
	jb	cmpGLScaleP1BitLoopSetup

; START OF LOOP OVER THE ROW

cmpGLScaleP1RowLoop:
	mov	rInput2Hi, 4			; number of pixels/word
	xor	rInput1Long, rInput1Long	; clear input registers

; This block of code fills an output byte
cmpGLScaleP1WordLoop:
	mov	rInput1, [rSrcAddr1 + 3]
	mov	rInput2, [rSrcAddr2 + 3]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 0

; compute bit 1
	mov	rInput1, [rSrcAddr1 + 2]
	mov	rInput2, [rSrcAddr2 + 2]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 1

; compute bit 2
	mov	rInput1, [rSrcAddr1 + 1]
	mov	rInput2, [rSrcAddr2 + 1]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 2

; compute bit 3
	mov	rInput1, [rSrcAddr1 + 0]
	mov	rInput2, [rSrcAddr2 + 0]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 3

; compute bit 4
	mov	rInput1, [rSrcAddr1 + 7]
	mov	rInput2, [rSrcAddr2 + 7]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 4

; compute bit 5
	mov	rInput1, [rSrcAddr1 + 6]
	mov	rInput2, [rSrcAddr2 + 6]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 5

; compute bit 6
	mov	rInput1, [rSrcAddr1 + 5]
	mov	rInput2, [rSrcAddr2 + 5]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 6

; compute bit 7
	mov	rInput1, [rSrcAddr1 + 4]
	mov	rInput2, [rSrcAddr2 + 4]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 7

; Increment the addresses, decrement the word count and see if we should
; compute another output byte or write the word we've got.
	add	rSrcAddr1, 8
	add	rSrcAddr2, 8
	dec	rInput2Hi				; did another byte
	jnz	cmpGLScaleP1WordLoop			; loop if more to do

; At this point, rOWord is full, so write it to memory and return to the
; top of the loop.
	mov	rInput1Long, dstAddr			; get dest address
	mov	[rInput1Long], rOWord			; write output word
	add	rInput1Long, 4				; point to next word
	mov	dstAddr, rInput1Long			; keep it safe
	mov	rInput1Long, imageWidth
	sub	rInput1Long, 32				; decrement image width
	cmp	rInput1Long, 32				; still >= 32?
	mov	imageWidth, rInput1Long
	jae	cmpGLScaleP1RowLoop			; yes, do another
							; block of 32 pix.

; This loop is used when we have less than 32 pixels to compute.  We remember
; the number of pixels remaining so we know how much to shift the result to
; align it.
cmpGLScaleP1BitLoopSetup:
	cmp	rInput1Long, 0				; are we done now?
	je	cmpGLScaleP1Fin				; yes, quit
	mov	rInput2Hi, rInput1			; remember the count
	xor	rInput1Long, rInput1Long		; clear input registers
	jmp	cmpGLScaleP1BitLoopEntry

cmpGLScaleP1BitLoop:
	add	rSrcAddr1, 4
	add	rSrcAddr2, 4
cmpGLScaleP1BitLoopEntry:
	mov	rInput1, [rSrcAddr1 + 3]
	mov	rInput2, [rSrcAddr2 + 3]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput2Hi
	jz	cmpGLScaleP1BitLoopFin			; quit if that was it

	mov	rInput1, [rSrcAddr1 + 2]
	mov	rInput2, [rSrcAddr2 + 2]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput2Hi
	jz	cmpGLScaleP1BitLoopFin			; quit if that was it

	mov	rInput1, [rSrcAddr1 + 1]
	mov	rInput2, [rSrcAddr2 + 1]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput2Hi
	jz	cmpGLScaleP1BitLoopFin			; quit if that was it

	mov	rInput1, [rSrcAddr1 + 0]
	mov	rInput2, [rSrcAddr2 + 0]
	mov	rInput1, [rTabAddr + rInput1Long]	; scale input 1
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput2Hi
	jnz	cmpGLScaleP1BitLoop			; keep going if there's
							; more

; We're done with the last line.  Shift the bits into position and
; write them out.  We must put the shift count into rInput2 since it's
; cl, which is the only legal place for shift counts.
cmpGLScaleP1BitLoopFin:
	mov	rInput2, BYTE PTR imageWidth
	sub	rInput2, 32
	neg	rInput2					; create a positive
							; shift count
	shl	rOWord, rInput2				; align the output bits
	mov	rInput1Long, dstAddr			; get the dest address
	mov	[rInput1Long], rOWord			; write the last word

; That's it!  Just return
cmpGLScaleP1Fin:
	ret

ip_compareGrayLineScaleP1 ENDP



;---------------------------------------------------------------------------
;	ip_compareGrayLineScaleP2
;---------------------------------------------------------------------------

ip_compareGrayLineScaleP2Start::
ip_compareGrayLineScaleP2 PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr1:DWORD, srcAddr2:DWORD, \
	dstAddr:DWORD, scaleTable:DWORD, \
	imageWidth:DWORD

; make the assembler happy
	mov	eax, pIPState

; load up the addresses
	mov	rTabAddr, scaleTable
	mov	rSrcAddr1, srcAddr1
	mov	rSrcAddr2, srcAddr2

; If there are more than 32 pixels left in the line, process them in the
; 32-pixel-at-a-time section.  Otherwise, go to the 1-pixel-at-a-time loop.
; Note the weird addressing; this is used because we address Little Endian
; memory as if we were reading contiguous 32-bit things instead of
; bytes.  That is, we think that pixels are stored left to right within
; 32-bit words on all platforms.
	mov	rInput2Long, imageWidth
	cmp	rInput2Long, 32
	jb	cmpGLScaleP2BitLoopSetup

; START OF LOOP OVER THE ROW

cmpGLScaleP2RowLoop:
	mov	rInput1Hi, 4			; number of pixels/word
	xor	rInput2Long, rInput2Long	; clear input registers

; This block of code fills an output byte
cmpGLScaleP2WordLoop:
	mov	rInput1, [rSrcAddr1 + 3]
	mov	rInput2, [rSrcAddr2 + 3]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 0

; compute bit 1
	mov	rInput1, [rSrcAddr1 + 2]
	mov	rInput2, [rSrcAddr2 + 2]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 1

; compute bit 2
	mov	rInput1, [rSrcAddr1 + 1]
	mov	rInput2, [rSrcAddr2 + 1]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 2

; compute bit 3
	mov	rInput1, [rSrcAddr1 + 0]
	mov	rInput2, [rSrcAddr2 + 0]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 3

; compute bit 4
	mov	rInput1, [rSrcAddr1 + 7]
	mov	rInput2, [rSrcAddr2 + 7]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 4

; compute bit 5
	mov	rInput1, [rSrcAddr1 + 6]
	mov	rInput2, [rSrcAddr2 + 6]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 5

; compute bit 6
	mov	rInput1, [rSrcAddr1 + 5]
	mov	rInput2, [rSrcAddr2 + 5]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 6

; compute bit 7
	mov	rInput1, [rSrcAddr1 + 4]
	mov	rInput2, [rSrcAddr2 + 4]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord				; insert bit 7

; Increment the addresses, decrement the word count and see if we should
; compute another output byte or write the word we've got.
	add	rSrcAddr1, 8
	add	rSrcAddr2, 8
	dec	rInput1Hi				; did another byte
	jnz	cmpGLScaleP2WordLoop			; loop if more to do

; At this point, rOWord is full, so write it to memory and return to the
; top of the loop.
	mov	rInput2Long, dstAddr			; get dest address
	mov	[rInput2Long], rOWord			; write output word
	add	rInput2Long, 4				; point to next word
	mov	dstAddr, rInput2Long			; keep it safe
	mov	rInput2Long, imageWidth
	sub	rInput2Long, 32				; decrement image width
	cmp	rInput2Long, 32				; still >= 32?
	mov	imageWidth, rInput2Long
	jae	cmpGLScaleP2RowLoop			; yes, do another
							; block of 32 pix.

; This loop is used when we have less than 32 pixels to compute.  We remember
; the number of pixels remaining so we know how much to shift the result to
; align it.
cmpGLScaleP2BitLoopSetup:
	cmp	rInput2Long, 0				; are we done now?
	je	cmpGLScaleP2Fin				; yes, quit
	mov	rInput1Hi, rInput2			; remember the count
	xor	rInput2Long, rInput2Long		; clear input registers
	jmp	cmpGLScaleP2BitLoopEntry

cmpGLScaleP2BitLoop:
	add	rSrcAddr1, 4
	add	rSrcAddr2, 4
cmpGLScaleP2BitLoopEntry:
	mov	rInput1, [rSrcAddr1 + 3]
	mov	rInput2, [rSrcAddr2 + 3]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput1Hi
	jz	cmpGLScaleP2BitLoopFin			; quit if that was it

	mov	rInput1, [rSrcAddr1 + 2]
	mov	rInput2, [rSrcAddr2 + 2]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput1Hi
	jz	cmpGLScaleP2BitLoopFin			; quit if that was it

	mov	rInput1, [rSrcAddr1 + 1]
	mov	rInput2, [rSrcAddr2 + 1]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput1Hi
	jz	cmpGLScaleP2BitLoopFin			; quit if that was it

	mov	rInput1, [rSrcAddr1 + 0]
	mov	rInput2, [rSrcAddr2 + 0]
	mov	rInput2, [rTabAddr + rInput2Long]	; scale input 2
	cmp	rInput2, rInput1
	adc	rOWord, rOWord
	dec	rInput1Hi
	jnz	cmpGLScaleP2BitLoop			; keep going if there's
							; more

; We're done with the last line.  Shift the bits into position and
; write them out.  We must put the shift count into rInput2 since it's
; cl, which is the only legal place for shift counts.
cmpGLScaleP2BitLoopFin:
	mov	rInput2, BYTE PTR imageWidth
	sub	rInput2, 32
	neg	rInput2					; create a positive
							; shift count
	shl	rOWord, rInput2				; align the output bits
	mov	rInput1Long, dstAddr			; get the dest address
	mov	[rInput1Long], rOWord			; write the last word

; That's it!  Just return
cmpGLScaleP2Fin:
	ret

ip_compareGrayLineScaleP2 ENDP



_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END


IF 0				; These are all comments

COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pccmpgry.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.7  1998/01/13 23:58:03  ddavies
pccmpgry.asm -> /product/ipcore/ipshared/src/RCS/pccmpgry.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.6  1996/09/17 00:45:21  ddavies
pccmpgry.asm -> /product/ipcore/ipshared/src/RCS/pccmpgry.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.5  1996/09/16 00:49:54  ddavies
pccmpgry.asm -> /product/ipcore/ipshared/src/RCS/pccmpgry.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.4  1995/09/15 21:47:28  ddavies
pccmpgry.asm -> /product/ipcore/ipshared/src/RCS/pccmpgry.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

# Revision 1.3  1993/11/05  01:51:53  ddavies
# pccmpgry.asm -> /product/ipcore/ipshared/src/RCS/pccmpgry.asm,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

^

ENDIF

