;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcdepth.asm:  grayscale thresholding.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The 386/486 processors can reload source words quickly.

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; In the SPARC algorithms, we went to a great deal of trouble to use double
; word (64 bit) write operations.  Writes on an SS2 are slow, so it was
; beneficial.  The 386/486 has relatively fast 32 bit writes and no 64 bit
; writes, so we will discard that complication.

_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcdepth.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment:
; EAX:	scratch
; EBX:	scratch 
; ECX:	address to table which unscrambles the bits
; EDX:	scrambled output word
; EBP:	points to variables on the stack
; ESI:	source address
; EDI:	destination address
; ESP:	stack pointer

rSrcAddr	EQU esi
rDstAddr	EQU edi
; for use by the 4 bit/pixel pass filter code:
rRepThresh	EQU edi
rScratch	EQU eax
rScratchH	EQU ah
rScratchL	EQU al
rScratch1	EQU ebx
rTabAddr	EQU ecx
rOWord		EQU edx
rOWordH		EQU dh
rOWordL		EQU dl

; Constant definitions:
cMSBitMask	EQU 080808080H
cnMSBitMask	EQU 07F7F7F7FH
cMS2BitMask	EQU 0C0C0C0C0H
cMSBitMask2bpp	EQU 0AAAAAAAAH
cMSBitMask4bpp	EQU 088888888H
cnMSBitMask4bpp	EQU 077777777H
cByteMask	EQU 0FFH
cLSBitMask4bpp	EQU 011111111H
cnLSBitMask4bpp	EQU 0EEEEEEEEH
cLSBitCarry4bpp	EQU 022222222H

; Parameters:
; srcAddr:	address of left end of current source line
; destAddr:	address of left end of current dest line
; dstOffset:	offset of rightmost dest word
; srcOffset:	offset of rightmost src word
; srcLimit:	address of word containing rightmost
; tabByte0:	table to directly unscramble MS byte
; tabByte1:	table to directly unscramble MS byte -1
; tabByte2:	table to directly unscramble MS byte - 2
; tabByte3:	table to directly unscramble LS byte
; thresh:	four copies of the only threshold for toBinaryThreshLine
;		or the lower threshold for Line
; upperLower:	four copies of (upper threshold - lower threshold), used
;		only in toBinaryBandPassLine
; Note that any of these tables can be used if the output bit is
; shifted to the appropriate position

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPDWORD TYPEDEF NEAR PTR DWORD

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_toBinary128.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will process the image from top to bottom.
; We will move from right to left within each line of the image.

ip_toBinary128LineStart::
ip_toBinary128Line	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tabByte0:DWORD, tabByte1:DWORD, tabByte2:DWORD, tabByte3:DWORD
LOCAL	lineEnd:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tabByte0
	mov	eax, tabByte1
	mov	eax, tabByte2

; load the real parameters
	mov	rTabAddr, tabByte3
	mov	rSrcAddr, srcAddr
	mov	rDstAddr, dstAddr

; at this point rSrcAddr should point to the first source word on the
; left side of the image just inside the frame.
; at this point rDstAddr should point to the first dest word on the
; left side of the image just inside the frame.
	mov	rOWord, 0		; initialize the output word

; do the last destination word first because it may require less than 8
; source words.
	mov	lineEnd, rSrcAddr
	mov	rScratch1, rSrcAddr	; point to the first that is not
	add	rScratch1, srcOffset	; used to compose the last
					; dest word
	add	rSrcAddr, srcLimit	; point to the last source word
	add	rDstAddr, dstOffset
tb128LastWordLoop:
	shr	rOWord, 1		; make room for next bit
	mov	rScratch, [rSrcAddr]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 4		; next source word
	cmp	rSrcAddr, rScratch1	; are we done yet?
	jae	tb128LastWordLoop

	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

tb128WriteLastWord:
	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

; now do all the full 8 word source to 1 word dest
tb128InnerLoop:

; now read 8 source words and put the MS bits of each pixel into rOWord
	mov	rOWord, [rSrcAddr]
	and	rOWord, cMSBitMask	; get the most signif bits
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-4]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-8]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-12]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-16]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-20]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-24]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-28]
	and	rScratch, cMSBitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 32		; next source chunk

; have all the destination bits.  now unscramble them.
	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

	cmp	rSrcAddr, lineEnd	; are we done with the line?
	ja	tb128InnerLoop		; no, do the next dest word.

	ret				; yes, return
ip_toBinary128Line ENDP

ip_toBinaryFrom2bppLineStart::
ip_toBinaryFrom2bppLine	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tabByte0: DWORD, tabByte1: DWORD, tabByte2: DWORD, tabByte3: DWORD
LOCAL	lineEnd:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tabByte0
	mov	eax, tabByte1
	mov	eax, tabByte2

; load the real parameters
	mov	rTabAddr, tabByte3
	mov	rSrcAddr, srcAddr
	mov	rDstAddr, dstAddr


; at this point rSrcAddr should point to the first source word on the
; left side of the image just inside the frame.
; at this point rDstAddr should point to the first dest word on the
; left side of the image just inside the frame.
	mov	rOWord, 0		; initialize the output word

; do the last destination word first because it may require less than 2
; source words.
	mov	lineEnd, rSrcAddr
	mov	rScratch1, rSrcAddr	; point to the first that is not
	add	rScratch1, srcOffset	; used to compose the last
					; dest word
	add	rSrcAddr, srcLimit	; point to the last source word
	add	rDstAddr, dstOffset
tbF2LastWordLoop:
	shr	rOWord, 1		; make room for next bit
	mov	rScratch, [rSrcAddr]
	and	rScratch, cMSBitMask2bpp ; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 4		; next source word
	cmp	rSrcAddr, rScratch1	; are we done yet?
	jae	tbF2LastWordLoop

	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	shl	rScratch1, 4		; shift the unscrambled word
					; left by 4
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	shl	rScratch1, 4		; shift the unscrambled word
					; left by 4
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	shl	rScratch1, 4		; shift the unscrambled word
					; left by 4
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

tbF2WriteLastWord:
	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

; now do all the full 2 word source to 1 word dest
tbF2InnerLoop:

; now read 2 source words and put the MS bits of each pixel into rOWord
	mov	rOWord, [rSrcAddr]
	and	rOWord, cMSBitMask2bpp	; get the most signif bits
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-4]
	and	rScratch, cMSBitMask2bpp ; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 8		; next source chunk

; have all the destination bits.  now unscramble them.
	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	shl	rScratch1, 4		; shift the unscrambled word
					; left by r
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	shl	rScratch1, 4		; shift the unscrambled word
					; left by 4
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	shl	rScratch1, 4		; shift the unscrambled word
					; left by 4
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

	cmp	rSrcAddr, lineEnd	; are we done with the line?
	ja	tbF2InnerLoop		; no, do the next dest word.

	ret				; Done with the line, so return
ip_toBinaryFrom2bppLine ENDP

ip_toBinaryFrom4bppLineStart::
ip_toBinaryFrom4bppLine	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tabByte0: DWORD, tabByte1: DWORD, tabByte2: DWORD, tabByte3: DWORD
LOCAL	lineEnd:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tabByte0
	mov	eax, tabByte1
	mov	eax, tabByte2

; load the real parameters
	mov	rTabAddr, tabByte3
	mov	rSrcAddr, srcAddr
	mov	rDstAddr, dstAddr

; at this point rSrcAddr should point to the first source word on the
; left side of the image just inside the frame.
; at this point rDstAddr should point to the first dest word on the
; left side of the image just inside the frame.
	mov	rOWord, 0		; initialize the output word

; do the last destination word first because it may require less than 4
; source words.
	mov	lineEnd, rSrcAddr
	mov	rScratch1, rSrcAddr	; point to the first that is not
	add	rScratch1, srcOffset	; used to compose the last
					; dest word
	add	rSrcAddr, srcLimit	; point to the last source word
	add	rDstAddr, dstOffset
tbF4LastWordLoop:
	shr	rOWord, 1		; make room for next bit
	mov	rScratch, [rSrcAddr]
	and	rScratch, cMSBitMask4bpp ; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 4		; next source word
	cmp	rSrcAddr, rScratch1	; are we done yet?
	jae	tbF4LastWordLoop

	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

tbF4WriteLastWord:
	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

; now do all the full 4 word source to 1 word dest
tbF4InnerLoop:

; now read 4 source words and put the MS bits of each pixel into rOWord
	mov	rOWord, [rSrcAddr]
	and	rOWord, cMSBitMask4bpp	; get the most signif bits
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-4]
	and	rScratch, cMSBitMask4bpp ; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-8]
	and	rScratch, cMSBitMask4bpp ; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-12]
	and	rScratch, cMSBitMask4bpp ; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 16		; next source chunk

; have all the destination bits.  now unscramble them.
	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

	cmp	rSrcAddr, lineEnd	; are we done with the line?
	ja	tbF4InnerLoop		; no, do the next dest word.

	ret				; yes, done with the line
ip_toBinaryFrom4bppLine ENDP

ip_toTwoBitsPerPixelLineStart::
ip_toTwoBitsPerPixelLine PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tabByte0: DWORD, tabByte1: DWORD, tabByte2: DWORD, tabByte3: DWORD
LOCAL	lineEnd:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tabByte0
	mov	eax, tabByte1
	mov	eax, tabByte2

; load the real parameters
	mov	rTabAddr, tabByte3
	mov	rSrcAddr, srcAddr
	mov	rDstAddr, dstAddr

; at this point rSrcAddr should point to the first source word on the
; left side of the image just inside the frame.
; at this point rDstAddr should point to the first dest word on the
; left side of the image just inside the frame.
	mov	rOWord, 0		; initialize the output word

; do the last destination word first because it may require less than 4
; source words.
	mov	lineEnd, rSrcAddr
	mov	rScratch1, rSrcAddr	; point to the first that is not
	add	rScratch1, srcOffset	; used to compose the last
					; dest word
	add	rSrcAddr, srcLimit	; point to the last source word
	add	rDstAddr, dstOffset
t2bppLastWordLoop:
	shr	rOWord, 2		; make room for next bit
	mov	rScratch, [rSrcAddr]
	and	rScratch, cMS2BitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 4		; next source word
	cmp	rSrcAddr, rScratch1	; are we done yet?
	jae	t2bppLastWordLoop

	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

t2bppWriteLastWord:
	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

; now do all the full 4 word source to 1 word dest
t2bppInnerLoop:

; now read 4 source words and put the MS bits of each pixel into rOWord
	mov	rOWord, [rSrcAddr]
	and	rOWord, cMS2BitMask	; get the most signif bits
	shr	rOWord, 2

	mov	rScratch, [rSrcAddr-4]
	and	rScratch, cMS2BitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 2

	mov	rScratch, [rSrcAddr-8]
	and	rScratch, cMS2BitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 2

	mov	rScratch, [rSrcAddr-12]
	and	rScratch, cMS2BitMask	; get the most signif bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 16		; next source chunk

; have all the destination bits.  now unscramble them.
	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

	cmp	rSrcAddr, lineEnd	; are we done with the line?
	ja	t2bppInnerLoop		; no, do the next dest word.

	ret				; yes, done with the line
ip_toTwoBitsPerPixelLine ENDP

ip_toBinaryThreshLineStart::
ip_toBinaryThreshLine	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tabByte0: DWORD, tabByte1: DWORD, tabByte2: DWORD, tabByte3: DWORD, \
	repThresh:DWORD
LOCAL	lineEnd:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tabByte0
	mov	eax, tabByte1
	mov	eax, tabByte2

; load the real parameters
	mov	rTabAddr, tabByte3
	mov	rSrcAddr, srcAddr
	mov	rDstAddr, dstAddr

; at this point rSrcAddr should point to the first source word on the
; left side of the image just inside the frame.
; at this point rDstAddr should point to the first dest word on the
; left side of the image just inside the frame.
tbtImageLoop:
	mov	rOWord, 0		; initialize the output word

; do the last destination word first because it may require less than 8
; source words.
	mov	lineEnd, rSrcAddr
	mov	rScratch1, rSrcAddr	; point to the first that is not
	add	rScratch1, srcOffset	; used to compose the last
					; dest word
	add	rSrcAddr, srcLimit	; point to the last source word
	add	rDstAddr, dstOffset
tbtLastWordLoop:
	shr	rOWord, 1		; make room for next bit
	mov	rScratch, [rSrcAddr]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, repThresh
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 4		; next source word
	cmp	rSrcAddr, rScratch1	; are we done yet?
	jae	tbtLastWordLoop

	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

tbtWriteLastWord:
	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

; now do all the full 8 word source to 1 word dest
tbtInnerLoop:
	mov	rScratch1, repThresh

; now read 8 source words and put the MS bits of each pixel into rOWord
	mov	rOWord, [rSrcAddr]
	shr	rOWord, 1		; get rid of LS bit
	or	rOWord, cMSBitMask	; preset carryout
	sub	rOWord, rScratch1
	and	rOWord, cMSBitMask
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-4]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, rScratch1
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-8]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, rScratch1
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-12]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, rScratch1	; was TOO LONG!
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-16]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, rScratch1
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-20]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, rScratch1
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-24]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, rScratch1
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-28]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, rScratch1
	and	rScratch, cMSBitMask
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 32		; next source chunk

; have all the destination bits.  now unscramble them.
	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

	cmp	rSrcAddr, lineEnd	; are we done with the line?
	ja	tbtInnerLoop		; no, do the next dest word.

	ret				; yes, done with the line
ip_toBinaryThreshLine ENDP


; We really need another register for the low bit of the threshold
; so we'll use EBP.  We'll only use it in the image loop and carefully
; save it before use and restore it before leaving the procedure.
; The stack grows from high to low memory and the stack pointer points to
; the last element pushed onto the stack.  At the stack of the image
; loop we have edi, esi, edx, ecs, ebx, eax, lastEnd, savedEBP, lineEnd,
; srcLimit, srcOffset, dstOffset, EBP of calling procedure, 
; return address, parameters where edi is on the top of the stack.

rRepThreshLow	EQU	EBP
; saved edi, esi, edx, ecx, ebx, eax occupy [ESP+0] through[ESP+20]
sLastEnd	EQU	[ESP+24]
sSavedEBP	EQU	[ESP+28]
sLineEnd	EQU	[ESP+32]
;caller's EBP and return address occupy [ESP+36 and ESP+40]
sIPState	EQU	[ESP+44]
sSrcAddr	EQU	[ESP+48]
sDstAddr	EQU	[ESP+52]
sDstOffset	EQU	[ESP+56]
sSrcOffset	EQU	[ESP+60]
sSrcLimit	EQU	[ESP+64]
sTabByte0	EQU	[ESP+68]
sTabByte1	EQU	[ESP+72]
sTabByte2	EQU	[ESP+76]
sTabByte3	EQU	[ESP+80]
sThresh		EQU	[ESP+84]
sThreshLow	EQU	[ESP+88]

ip_toBinaryFrom4bppThreshLineStart::
ip_toBinaryFrom4bppThreshLine PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tabByte0: DWORD, tabByte1: DWORD, tabByte2: DWORD, tabByte3: DWORD, \
	thresh:DWORD, threshLow:DWORD
LOCAL	lineEnd:DWORD, \
	savedEBP:DWORD, lastEnd:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tabByte0
	mov	eax, tabByte1
	mov	eax, tabByte2
	mov	eax, dstAddr
	mov	eax, lineEnd
	mov	eax, savedEBP
	mov	eax, lastEnd
	mov	eax, dstOffset
	mov	eax, srcOffset
	mov	eax, srcLimit

; load the real parameters
	mov	rTabAddr, tabByte3
	mov	rSrcAddr, srcAddr
	mov	rRepThresh, thresh

	mov	sSavedEBP, EBP
	mov	rRepThreshLow, threshLow	; load EBP with lower thresh

; at this point rSrcAddr should point to the first source word on the
; left side of the image just inside the frame.
; at this point dstAddr should point to the first dest word on the
; left side of the image just inside the frame.
	mov	rOWord, 0		; initialize the output word

; do the last destination word first because it may require less than 4
; source words.
	mov	sLineEnd, rSrcAddr
	mov	rScratch, rSrcAddr	; point to the first that is not
	add	rScratch, sSrcOffset	; used to compose the last
	mov	sLastEnd, rScratch	; dest word
	add	rSrcAddr, sSrcLimit	; point to the last source word
	mov	rScratch, sDstAddr
	add	rScratch, sDstOffset
	mov	sDstAddr, rScratch
tbtF4LastWordLoop:
	shr	rOWord, 1		; make room for next output bit
	mov	rScratch, [rSrcAddr]	; get the next input word
	mov	rScratch1, rScratch	; get the LS bits into another
	and	rScratch1, cLSBitMask4bpp ; register
	or	rScratch1, cLSBitCarry4bpp ; set the carry-out bit
	sub	rScratch1, rRepThreshLow ; subtract the threshold
	not	rScratch1		; 
	and	rScratch1, cLSBitCarry4bpp ; keep only the carry-out bit
	shr	rScratch1, 1		; shift into position
	shr	rScratch, 1		; shift 3 MS bits into position
	or	rScratch, cMSBitMask4bpp ; set the carry-out bit
	sub	rScratch, rRepThresh	; subtract the threshold
	sub	rScratch, rScratch1	; subtract off the LS carry
	and	rScratch, cMSBitMask4bpp ; keep only the carry-out bit
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 4		; next source word
	cmp	rSrcAddr, sLastEnd	; are we done yet?
	jae	tbtF4LastWordLoop

	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	mov	rScratch, sDstAddr
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

tbtF4WriteLastWord:
	mov	[rScratch], rScratch1
	sub	rScratch, 4		; point to next word
	mov	sDstAddr, rScratch

; now do all the full 4 word source to 1 word dest
tbtF4InnerLoop:

; now read 4 source words and put the MS bits of each pixel into rOWord
	mov	rOWord, [rSrcAddr]
	mov	rScratch1, rOWord	; get the LS bits into another
	and	rScratch1, cLSBitMask4bpp ; register
	or	rScratch1, cLSBitCarry4bpp ; set the carry-out bit
	sub	rScratch1, rRepThreshLow ; subtract the threshold
	not	rScratch1		; 
	and	rScratch1, cLSBitCarry4bpp ; keep only the carry-out bit
	shr	rScratch1, 1		; shift into position
	shr	rOWord, 1		; shift 3 MS bits into position
	or	rOWord, cMSBitMask4bpp 	; set the carry-out bit
	sub	rOWord, rRepThresh	; subtract the threshold
	sub	rOWord, rScratch1	; subtract off the LS carry
	and	rOWord, cMSBitMask4bpp	 ; keep only the carry-out bit
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-4]
	mov	rScratch1, rScratch	; get the LS bits into another
	and	rScratch1, cLSBitMask4bpp ; register
	or	rScratch1, cLSBitCarry4bpp ; set the carry-out bit
	sub	rScratch1, rRepThreshLow ; subtract the threshold
	not	rScratch1		; 
	and	rScratch1, cLSBitCarry4bpp ; keep only the carry-out bit
	shr	rScratch1, 1		; shift into position
	shr	rScratch, 1		; shift 3 MS bits into position
	or	rScratch, cMSBitMask4bpp ; set the carry-out bit
	sub	rScratch, rRepThresh	; subtract the threshold
	sub	rScratch, rScratch1	; subtract off the LS carry
	and	rScratch, cMSBitMask4bpp ; keep only the carry-out bit
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-8]
	mov	rScratch1, rScratch	; get the LS bits into another
	and	rScratch1, cLSBitMask4bpp ; register
	or	rScratch1, cLSBitCarry4bpp ; set the carry-out bit
	sub	rScratch1, rRepThreshLow ; subtract the threshold
	not	rScratch1		; 
	and	rScratch1, cLSBitCarry4bpp ; keep only the carry-out bit
	shr	rScratch1, 1		; shift into position
	shr	rScratch, 1		; shift 3 MS bits into position
	or	rScratch, cMSBitMask4bpp ; set the carry-out bit
	sub	rScratch, rRepThresh	; subtract the threshold
	sub	rScratch, rScratch1	; subtract off the LS carry
	and	rScratch, cMSBitMask4bpp ; keep only the carry-out bit
	or	rOWord, rScratch	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-12]
	mov	rScratch1, rScratch	; get the LS bits into another
	and	rScratch1, cLSBitMask4bpp ; register
	or	rScratch1, cLSBitCarry4bpp ; set the carry-out bit
	sub	rScratch1, rRepThreshLow ; subtract the threshold
	not	rScratch1		; 
	and	rScratch1, cLSBitCarry4bpp ; keep only the carry-out bit
	shr	rScratch1, 1		; shift into position
	shr	rScratch, 1		; shift 3 MS bits into position
	or	rScratch, cMSBitMask4bpp ; set the carry-out bit
	sub	rScratch, rRepThresh	; subtract the threshold
	sub	rScratch, rScratch1	; subtract off the LS carry
	and	rScratch, cMSBitMask4bpp ; keep only the carry-out bit
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 16		; next source chunk

; have all the destination bits.  now unscramble them.
	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
	add	rScratch1, rScratch1	; left by 2
	mov	rScratch, sDstAddr
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

	mov	[rScratch], rScratch1
	sub	rScratch, 4		; point to next word
	mov	sDstAddr, rScratch

	cmp	rSrcAddr, sLineEnd	; are we done with the line?
	ja	tbtF4InnerLoop		; no, do the next dest word.


	mov	EBP, sSavedEBP		; yes, restore EBP and return

	ret
ip_toBinaryFrom4bppThreshLine ENDP


ip_toBinaryBandPassLineStart::
ip_toBinaryBandPassLine	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tabByte0: DWORD, tabByte1: DWORD, tabByte2: DWORD, tabByte3: DWORD, \
	lowerThresh:DWORD, upperThresh:DWORD
LOCAL	lineEnd:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tabByte0
	mov	eax, tabByte1
	mov	eax, tabByte2

; load the real parameters
	mov	rTabAddr, tabByte3
	mov	rSrcAddr, srcAddr
	mov	rDstAddr, dstAddr

; at this point rSrcAddr should point to the first source word on the
; left side of the image just inside the frame.
; at this point rDstAddr should point to the first dest word on the
; left side of the image just inside the frame.
tbbpImageLoop:
	mov	rOWord, 0		; initialize the output word

; do the last destination word first because it may require less than 8
; source words.
	mov	lineEnd, rSrcAddr
	mov	rScratch1, rSrcAddr	; point to the first that is not
	add	rScratch1, srcOffset	; used to compose the last
					; dest word
	add	rSrcAddr, srcLimit	; point to the last source word
	add	rDstAddr, dstOffset
tbbpLastWordLoop:
	shr	rOWord, 1		; make room for next bit
	mov	rScratch, [rSrcAddr]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	neg	rScratch		; do upperThresh - rScratch
	add	rScratch, upperThresh
	and	rScratch, cMSBitMask	; get carry bits
	or	rOWord, rScratch	; put them in the dest word
	sub	rSrcAddr, 4		; next source word
	cmp	rSrcAddr, rScratch1	; are we done yet?
	jae	tbbpLastWordLoop

	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

tbbpWriteLastWord:
	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

; now do all the full 8 word source to 1 word dest
tbbpInnerLoop:

; now read 8 source words and put the MS bits of each pixel into rOWord
	mov	rScratch, [rSrcAddr]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	mov	rOWord, rScratch1	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-4]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	or	rOWord, rScratch1	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-8]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	or	rOWord, rScratch1	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-12]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	or	rOWord, rScratch1	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-16]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	or	rOWord, rScratch1	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-20]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	or	rOWord, rScratch1	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-24]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	or	rOWord, rScratch1	; put them in the dest word
	shr	rOWord, 1

	mov	rScratch, [rSrcAddr-28]
	shr	rScratch, 1		; get rid of LS bit
	or	rScratch, cMSBitMask	; preset carryout
	sub	rScratch, lowerThresh	; subtract lower threshold
	and	rScratch, cnMSBitMask	; set carry bits to 0
	mov	rScratch1, upperThresh
	sub	rScratch1, rScratch	; subtract from upper
	and	rScratch1, cMSBitMask	; get carry bits
	or	rOWord, rScratch1	; put them in the dest word
	sub	rSrcAddr, 32		; next source chunk

; have all the destination bits.  now unscramble them.
	mov	rScratch, 0
	rol	rOWord, 16		; get MS 2 bytes
	mov	rScratchL, rOWordH	; get the first byte
	mov	rScratch1, [rTabAddr+rScratch*4] ; unscramble it

	mov	rScratchL, rOWordL	; get the second byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	rol	rOWord, 16		; get LS 2 bytes
	mov	rScratchL, rOWordH	; get third byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rScratch*4] ; get next set of 
					; unscrambled bits

	and	rOWord, cByteMask	; get final byte
	add	rScratch1, rScratch1	; shift the unscrambled word
					; left by 1
	or	rScratch1, [rTabAddr+rOWord*4] ; get next set of 
					; unscrambled bits

	mov	[rDstAddr], rScratch1
	sub	rDstAddr, 4		; point to next word

	cmp	rSrcAddr, lineEnd	; are we done with the line?
	ja	tbbpInnerLoop		; no, do the next dest word.

	ret				; yes, done with the line

ip_toBinaryBandPassLine ENDP




_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcdepth.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.5  1998/01/13 23:58:05  ddavies
pcdepth.asm -> /product/ipcore/ipshared/src/RCS/pcdepth.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.4  1996/09/17 00:45:24  ddavies
pcdepth.asm -> /product/ipcore/ipshared/src/RCS/pcdepth.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.3  1996/09/16 00:50:11  ddavies
pcdepth.asm -> /product/ipcore/ipshared/src/RCS/pcdepth.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.2  1995/09/15 21:47:32  ddavies
pcdepth.asm -> /product/ipcore/ipshared/src/RCS/pcdepth.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

# Revision 1.1  1994/01/28  04:23:20  ddavies
# Initial revision
#
# Revision 1.5  1993/11/16  03:02:57  ddavies
# pcbinarz.asm -> /product/ipcore/ipshared/src/RCS/pcbinarz.asm,v
#
# All of these functions had been writing into the right frame.  All had been
# using the source bytes/line to find the rightmost source word and the
# dest bytes/line to find the rightmost dest word.  This doesn't work because
# bytes/line - 2*frame = image + pad, not just image.  All have been changed
# to use the image width in pixels to calculate source and destination
# word offsets.
#
# Revision 1.4  1993/11/05  01:51:37  ddavies
# pcbinarz.asm -> /product/ipcore/ipshared/src/RCS/pcbinarz.asm,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

^

ENDIF

