;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcdibpix.asm:	Code to convert between RGB DIBs and 3 8-bit
;			images in our format.

	.386		; compile code for a 486 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; There seem to be two popular file formats on PCs.  In one, the leftmost
; byte in an image is in the leftmost byte of a 32-bit word.  That's the
; format we use.  In the other, the leftmost image byte is in byte 0
; of a 32-bit word.  That's the same as our format on big endian machines,
; but different on little endian machines.  The first format has the
; advantage that left and right shifts within 32-bit words work correctly.
; The second format has the advantage that moving from left to right
; always means incrementing memory addresses.
;
; In pcbswap.asm, we have code that just does byteswaps between the formats.
; This module contains code that translates between color Microsoft DIBs
; and 3 8-bit images.  Not only must the bytes be reordered, we must also
; pack convert between 3 individual images and one 24-bit image.
;
; A C implementation of this function is in pcdibpix.c.  We are doing
; assembly to take advantage of the 386's ability to move individual
; bytes in registers.  The general idea is that we read 4-byte things,
; transfer the bytes as appropriate, and write 4-byte things.  There's more
; shifting this way (as opposed to just reading and writing each byte),
; but we hope to avoid clogging up the 386/486's write buffers.
;
; The loop transfers between 3 of our image words and three DIB image words.
; Since there is no guarantee that the total number of pixels is a
; multiple of 4, there is a make-up loop at the bottom that transfers
; one RGB pixel at a time to transfer the remaining pixels.
;
; The order of pixels in the RGB data words is (R0 is leftmost Red pixel):
; Leftmost word:	B1 R0 G0 B0
;			G2 B2 R1 G1
; Rightmost word:	R3 G3 B3 R2
;
; When transfering from images to DIB, I was originally going to read
; the three source words and move the bytes one at a time.  However, it
; makes more sense to move the bytes directly from the individual color
; images into the RGB register.  This lets us keep their addresses in
; registers and avoids the necessity of shifting them.
;
; When moving data from the RGB image to the individual color images,
; things about the same, but the loading is trickier.  We have three color
; image pointers, one RGB pointer and one register for accumulating data.
; Instead of using the register to accumulate RGB data, we use it to
; accumulate a word of red, then a word of green, then a word of blue
; in turn.
;
; FEATURE (?) ADDITION:  In the NT and Win95 worlds, there's no such thing
; as a far pointer.  In those cases, we want the code to simply take a
; near pointer off the stack and not use the ES register.  We do this 
; using conditional assembly.  If the symbol AsmForNT is defined, we
; use near pointers and do not use the ES register.


_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcdibpix.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment (byte swap routines):
; EAX:	Accumulating data word
; EBX:	Blue image pointer 
; ECX:	line counter
; EDX:	Green image pointer
; EBP:	Addresses locals and parameters on stack
; ESI:	Red image pointer
; EDI:	RGB image pointer
; ESP:	stack pointer
; ES:   segment for DIB

rData		EQU eax
rDataL		EQU al
rDataH		EQU ah
rRedAddr	EQU esi
rBlueAddr	EQU ebx
rGreenAddr	EQU edx
rDIBAddr	EQU edi
rCount		EQU ecx

; Local Variables:
; none

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPDWORD TYPEDEF NEAR PTR DWORD
FPBYTE TYPEDEF FAR PTR BYTE
FPDWORD TYPEDEF FAR PTR DWORD


; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF


; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Verions of all routines for 16-bit machines with far pointers
; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
IFNDEF AsmForNT


; *********** PROCEDURE TO TRANSFER FROM SEPARATE COLORS TO RGB DIB *********


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_combine3ToDIBLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers
; The arguments are declared in the order specified by pcdibpix.c
ip_combine3ToDIBLineStart::
ip_combine3ToDIBLine	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI ES,\
	pIPState:DWORD, pDIB:FPDWORD, pRed:DWORD, pGreen:DWORD, pBlue:DWORD,\
	count:DWORD
; All registers used are also pushed on the stack automatically.

	mov	eax, pIPState		; keep assembler happy

; Load the registers
	mov	rRedAddr, pRed		; pointer to red image
	mov	rGreenAddr, pGreen	; pointer to green image
	mov	rBlueAddr, pBlue	; pointer to blue image
	les	rDIBAddr, pDIB		; pointer to DIB image
	mov	rCount, count		; number of pixels to transfer

; While the rCount is greater than or equal to 4 (the number of complete
; RGB pixels in three words), do the fancy loop.
	cmp	rCount, 4
	jl	last3Combine3ToDIB	; too small already, so just do
					; the last few pixels

; bias the count down by 4 so we can avoid an extra comparison with
; 4 at the end of the loop.
	sub	rCount, 4

; The main procesing loop
; Remember the originals are stored as R0 R1 R2 R3 (big endian order)
combine3ToDIBLoop:
	mov	rDataL, [rRedAddr   + 3]	; xx xx xx R0
	mov	rDataH, [rBlueAddr  + 2]	; xx xx B1 R0
	rol	rData, 16			; B1 R0 xx xx
	mov	rDataL, [rBlueAddr  + 3]	; B1 R0 xx B0
	mov	rDataH, [rGreenAddr + 3]	; B1 R0 G0 B0
	mov	ES:[rDIBAddr + 0], rData

	mov	rDataL, [rBlueAddr  + 1]	; xx xx xx B2
	mov	rDataH, [rGreenAddr + 1]	; xx xx G2 B2
	rol	rData, 16			; G2 B2 xx xx
	mov	rDataL, [rGreenAddr + 2]	; G2 B2 xx G1
	mov	rDataH, [rRedAddr   + 2]	; G2 B2 R1 G1
	mov	ES:[rDIBAddr + 4], rData

	mov	rDataL, [rGreenAddr + 0]	; xx xx xx G3
	mov	rDataH, [rRedAddr   + 0]	; xx xx R3 G3
	rol	rData, 16			; R3 G3 xx xx
	mov	rDataL, [rRedAddr   + 1]	; R3 G3 xx R2
	mov	rDataH, [rBlueAddr  + 0]	; R3 G3 B3 R2
	mov	ES:[rDIBAddr + 8], rData

; Are we done with blocks of 4 yet?
	add	rRedAddr, 4
	add	rGreenAddr, 4
	add	rBlueAddr, 4
	add	rDIBAddr, 12
	sub	rCount, 4
	jge	combine3ToDIBLoop		; no, do another block

; remove the bias from rCount.  If there were exactly a multiple of
; 4 pixels in the line, this will make rCount equal 0.  If the number
; of pixels in the line was not a multiple of 4, the remainder will
; appear in rCount here.
	add	rCount, 4

; There are three or fewer pixels remaining.  Do them one at a time.
; REMEMBER to keep the PIXR addresses straight!  We go 3, 2, 1, 0, not
; 0, 1, 2, 3!!
last3Combine3ToDIB:
	jz	finCombine3ToDIB		; quit if we're done now
	mov	rData, rBlueAddr
	xor	rData, 3
	mov	rDataL, [rData]
	mov	ES:[rDIBAddr + 0], rDataL
	mov	rData, rGreenAddr
	xor	rData, 3
	mov	rDataL, [rData]
	mov	ES:[rDIBAddr + 1], rDataL
	mov	rData, rRedAddr
	xor	rData, 3
	mov	rDataL, [rData]
	mov	ES:[rDIBAddr + 2], rDataL
	inc	rBlueAddr
	inc	rGreenAddr
	inc	rRedAddr
	add	rDIBAddr, 3
	dec	rCount
	jg	last3Combine3ToDIB

; At this point, we're finished so return
finCombine3ToDIB:
	ret
ip_combine3ToDIBLine  ENDP


; ********* PROCEDURE TO TRANSFER FROM FROM RGB DIB to SEPARATE COLOR ********


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_combine3ToDIBLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers
; The arguments are declared in the order specified by pcdibpix.c
ip_combineDIBTo3LineStart::
ip_combineDIBTo3Line	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI ES, \
	pIPState:DWORD, pDIB:FPDWORD, pRed:DWORD, pGreen:DWORD, pBlue:DWORD,\
	count:DWORD
; All registers used are also pushed on the stack automatically.

	mov	eax, pIPState		; keep assembler happy

; Load the registers
	mov	rRedAddr, pRed		; pointer to red image
	mov	rGreenAddr, pGreen	; pointer to green image
	mov	rBlueAddr, pBlue	; pointer to blue image
	les	rDIBAddr, pDIB		; pointer to DIB image
	mov	rCount, count		; number of pixels to transfer

; While the rCount is greater than or equal to 4 (the number of complete
; RGB pixels in three words), do the fancy loop.
	cmp	rCount, 4
	jl	last3CombineDIBTo3	; too small already, so just do
					; the last few pixels

; bias the count down by 4 so we can avoid an extra comparison with
; 4 at the end of the loop.
	sub	rCount, 4

; The main procesing loop.
combineDIBTo3Loop:
	mov	rDataH, ES:[rDIBAddr +  2]	; xx xx R0 xx
	mov	rDataL, ES:[rDIBAddr +  5]	; xx xx R0 R1
	rol	rData, 16			; R0 R1 xx xx
	mov	rDataH, ES:[rDIBAddr +  8]	; R0 R1 R2 xx
	mov	rDataL, ES:[rDIBAddr + 11]	; R0 R1 R2 R3
	mov	[rRedAddr], rData

	mov	rDataH, ES:[rDIBAddr +  1]	; xx xx G0 xx
	mov	rDataL, ES:[rDIBAddr +  4]	; xx xx G0 G1
	rol	rData, 16			; G0 G1 xx xx
	mov	rDataH, ES:[rDIBAddr +  7]	; G0 G1 G2 xx
	mov	rDataL, ES:[rDIBAddr + 10]	; G0 G1 G2 G3
	mov	[rGreenAddr], rData

	mov	rDataH, ES:[rDIBAddr +  0]	; xx xx B0 xx
	mov	rDataL, ES:[rDIBAddr +  3]	; xx xx B0 B1
	rol	rData, 16			; B0 B1 xx xx
	mov	rDataH, ES:[rDIBAddr +  6]	; B0 B1 B2 xx
	mov	rDataL, ES:[rDIBAddr +  9]	; B0 B1 B2 B3
	mov	[rBlueAddr], rData

; Are we done with blocks of 4 yet?
	add	rRedAddr, 4
	add	rGreenAddr, 4
	add	rBlueAddr, 4
	add	rDIBAddr, 12
	sub	rCount, 4
	jge	combineDIBTo3Loop	; no, do another block

; remove the bias from rCount.  If there were exactly a multiple of
; 4 pixels in the line, this will make rCount equal 0.  If the number
; of pixels in the line was not a multiple of 4, the remainder will
; appear in rCount here.
	add	rCount, 4

; There are three or fewer pixels remaining.  Do them one at a time.
; REMEMBER to keep the PIXR addresses straight!  We go 3, 2, 1, 0, not
; 0, 1, 2, 3!!
last3CombineDIBTo3:
	jz	finCombineDIBTo3	; quit if we're done now
	mov	rDataL, ES:[rDIBAddr + 0]
	xor	rBlueAddr, 3
	mov	[rBlueAddr], rDataL
	xor	rBlueAddr, 3

	mov	rDataL, ES:[rDIBAddr + 1]
	xor	rGreenAddr, 3
	mov	[rGreenAddr], rDataL
	xor	rGreenAddr, 3

	mov	rDataL, ES:[rDIBAddr + 2]
	xor	rRedAddr, 3
	mov	[rRedAddr], rDataL
	xor	rRedAddr, 3

	inc	rBlueAddr
	inc	rGreenAddr
	inc	rRedAddr
	add	rDIBAddr, 3
	dec	rCount
	jg	last3CombineDIBTo3

; At this point, we're finished so return
finCombineDIBTo3:
	ret
ip_combineDIBTo3Line  ENDP


; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Verions of all routines for 32-bit machines with NO far pointers
; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
ELSE	


; *********** PROCEDURE TO TRANSFER FROM SEPARATE COLORS TO RGB DIB *********


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_combine3ToDIBLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers
; The arguments are declared in the order specified by pcdibpix.c
ip_combine3ToDIBLineStart::
ip_combine3ToDIBLine	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI ES,\
	pIPState:DWORD, pDIB:DWORD, pRed:DWORD, pGreen:DWORD, pBlue:DWORD,\
	count:DWORD
; All registers used are also pushed on the stack automatically.

	mov	eax, pIPState		; keep assembler happy

; Load the registers
	mov	rRedAddr, pRed		; pointer to red image
	mov	rGreenAddr, pGreen	; pointer to green image
	mov	rBlueAddr, pBlue	; pointer to blue image
	mov	rDIBAddr, pDIB		; pointer to DIB image
	mov	rCount, count		; number of pixels to transfer

; While the rCount is greater than or equal to 4 (the number of complete
; RGB pixels in three words), do the fancy loop.
	cmp	rCount, 4
	jl	last3Combine3ToDIB	; too small already, so just do
					; the last few pixels

; bias the count down by 4 so we can avoid an extra comparison with
; 4 at the end of the loop.
	sub	rCount, 4

; The main procesing loop
; Remember the originals are stored as R0 R1 R2 R3 (big endian order)
combine3ToDIBLoop:
	mov	rDataL, [rRedAddr   + 3]	; xx xx xx R0
	mov	rDataH, [rBlueAddr  + 2]	; xx xx B1 R0
	rol	rData, 16			; B1 R0 xx xx
	mov	rDataL, [rBlueAddr  + 3]	; B1 R0 xx B0
	mov	rDataH, [rGreenAddr + 3]	; B1 R0 G0 B0
	mov	[rDIBAddr + 0], rData

	mov	rDataL, [rBlueAddr  + 1]	; xx xx xx B2
	mov	rDataH, [rGreenAddr + 1]	; xx xx G2 B2
	rol	rData, 16			; G2 B2 xx xx
	mov	rDataL, [rGreenAddr + 2]	; G2 B2 xx G1
	mov	rDataH, [rRedAddr   + 2]	; G2 B2 R1 G1
	mov	[rDIBAddr + 4], rData

	mov	rDataL, [rGreenAddr + 0]	; xx xx xx G3
	mov	rDataH, [rRedAddr   + 0]	; xx xx R3 G3
	rol	rData, 16			; R3 G3 xx xx
	mov	rDataL, [rRedAddr   + 1]	; R3 G3 xx R2
	mov	rDataH, [rBlueAddr  + 0]	; R3 G3 B3 R2
	mov	[rDIBAddr + 8], rData

; Are we done with blocks of 4 yet?
	add	rRedAddr, 4
	add	rGreenAddr, 4
	add	rBlueAddr, 4
	add	rDIBAddr, 12
	sub	rCount, 4
	jge	combine3ToDIBLoop		; no, do another block

; remove the bias from rCount.  If there were exactly a multiple of
; 4 pixels in the line, this will make rCount equal 0.  If the number
; of pixels in the line was not a multiple of 4, the remainder will
; appear in rCount here.
	add	rCount, 4

; There are three or fewer pixels remaining.  Do them one at a time.
; REMEMBER to keep the PIXR addresses straight!  We go 3, 2, 1, 0, not
; 0, 1, 2, 3!!
last3Combine3ToDIB:
	jz	finCombine3ToDIB		; quit if we're done now
	mov	rData, rBlueAddr
	xor	rData, 3
	mov	rDataL, [rData]
	mov	[rDIBAddr + 0], rDataL
	mov	rData, rGreenAddr
	xor	rData, 3
	mov	rDataL, [rData]
	mov	[rDIBAddr + 1], rDataL
	mov	rData, rRedAddr
	xor	rData, 3
	mov	rDataL, [rData]
	mov	[rDIBAddr + 2], rDataL
	inc	rBlueAddr
	inc	rGreenAddr
	inc	rRedAddr
	add	rDIBAddr, 3
	dec	rCount
	jg	last3Combine3ToDIB

; At this point, we're finished so return
finCombine3ToDIB:
	ret
ip_combine3ToDIBLine  ENDP


; ********* PROCEDURE TO TRANSFER FROM FROM RGB DIB to SEPARATE COLOR ********


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_combine3ToDIBLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers
; The arguments are declared in the order specified by pcdibpix.c
ip_combineDIBTo3LineStart::
ip_combineDIBTo3Line	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI ES, \
	pIPState:DWORD, pDIB:DWORD, pRed:DWORD, pGreen:DWORD, pBlue:DWORD,\
	count:DWORD
; All registers used are also pushed on the stack automatically.

	mov	eax, pIPState		; keep assembler happy

; Load the registers
	mov	rRedAddr, pRed		; pointer to red image
	mov	rGreenAddr, pGreen	; pointer to green image
	mov	rBlueAddr, pBlue	; pointer to blue image
	mov	rDIBAddr, pDIB		; pointer to DIB image
	mov	rCount, count		; number of pixels to transfer

; While the rCount is greater than or equal to 4 (the number of complete
; RGB pixels in three words), do the fancy loop.
	cmp	rCount, 4
	jl	last3CombineDIBTo3	; too small already, so just do
					; the last few pixels

; bias the count down by 4 so we can avoid an extra comparison with
; 4 at the end of the loop.
	sub	rCount, 4

; The main procesing loop.
combineDIBTo3Loop:
	mov	rDataH, [rDIBAddr +  2]	; xx xx R0 xx
	mov	rDataL, [rDIBAddr +  5]	; xx xx R0 R1
	rol	rData, 16			; R0 R1 xx xx
	mov	rDataH, [rDIBAddr +  8]	; R0 R1 R2 xx
	mov	rDataL, [rDIBAddr + 11]	; R0 R1 R2 R3
	mov	[rRedAddr], rData

	mov	rDataH, [rDIBAddr +  1]	; xx xx G0 xx
	mov	rDataL, [rDIBAddr +  4]	; xx xx G0 G1
	rol	rData, 16			; G0 G1 xx xx
	mov	rDataH, [rDIBAddr +  7]	; G0 G1 G2 xx
	mov	rDataL, [rDIBAddr + 10]	; G0 G1 G2 G3
	mov	[rGreenAddr], rData

	mov	rDataH, [rDIBAddr +  0]	; xx xx B0 xx
	mov	rDataL, [rDIBAddr +  3]	; xx xx B0 B1
	rol	rData, 16			; B0 B1 xx xx
	mov	rDataH, [rDIBAddr +  6]	; B0 B1 B2 xx
	mov	rDataL, [rDIBAddr +  9]	; B0 B1 B2 B3
	mov	[rBlueAddr], rData

; Are we done with blocks of 4 yet?
	add	rRedAddr, 4
	add	rGreenAddr, 4
	add	rBlueAddr, 4
	add	rDIBAddr, 12
	sub	rCount, 4
	jge	combineDIBTo3Loop	; no, do another block

; remove the bias from rCount.  If there were exactly a multiple of
; 4 pixels in the line, this will make rCount equal 0.  If the number
; of pixels in the line was not a multiple of 4, the remainder will
; appear in rCount here.
	add	rCount, 4

; There are three or fewer pixels remaining.  Do them one at a time.
; REMEMBER to keep the PIXR addresses straight!  We go 3, 2, 1, 0, not
; 0, 1, 2, 3!!
last3CombineDIBTo3:
	jz	finCombineDIBTo3	; quit if we're done now
	mov	rDataL, [rDIBAddr + 0]
	xor	rBlueAddr, 3
	mov	[rBlueAddr], rDataL
	xor	rBlueAddr, 3

	mov	rDataL, [rDIBAddr + 1]
	xor	rGreenAddr, 3
	mov	[rGreenAddr], rDataL
	xor	rGreenAddr, 3

	mov	rDataL, [rDIBAddr + 2]
	xor	rRedAddr, 3
	mov	[rRedAddr], rDataL
	xor	rRedAddr, 3

	inc	rBlueAddr
	inc	rGreenAddr
	inc	rRedAddr
	add	rDIBAddr, 3
	dec	rCount
	jg	last3CombineDIBTo3

; At this point, we're finished so return
finCombineDIBTo3:
	ret
ip_combineDIBTo3Line  ENDP

ENDIF	; Not AsmForNT



_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
; no global initialized data used
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END


IF 0				; These are all comments

COMMENT	^			David Hirsch's version of ML doesn't
				believe these are if'd out, so we
				need this too.  ?????
$Log: pcdibpix.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.9  1998/01/13 23:58:06  ddavies
pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.8  1996/09/17 00:45:27  ddavies
pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.7  1996/09/16 00:50:24  ddavies
pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.6  1995/09/15 21:47:36  ddavies
pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

# Revision 1.5  1995/09/12  01:13:03  ddavies
# ddavies on Mon Sep 11 18:11:29 PDT 1995
#
# pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v
#
# Add new versions of the routines ip_combineDIBTo3Line and ip_combine3ToDIBLine.
# The new versions are assembled when the compile time flag AsmForNT is
# defines.  In the new versions, the procedures expect the DIB pointers to
# be NEAR (4 bytes) instead of FAR (6 bytes).
#
# Revision 1.4  1994/10/20  22:14:21  ddavies
# ddavies on Thu Oct 20 15:13:34 PDT 1994
#
# pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v
#
# Save ES register in ip_combineDIBTo3Line.  It was being smashed without being saved
# and restored.  Thanks Bill Streeper!
#
# Revision 1.3  1994/04/20  23:54:18  ddavies
# ddavies on Wed Apr 20 16:51:24 PDT 1994
#
# pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v
#
# Fixed bug that was common to ip_combineDIBTo3Line and ip_combine3ToDIBLine.
# When the width of the image in pixels was not a multiple of 4, the last
# pixels on the right side of each destination line would be (sort of)
# byteswapped.  I forgot to XOR the PIXR addresses with 3 before using them.
# Thanks to Macduff for finding this one!
#
# Revision 1.2  1993/11/05  01:52:00  ddavies
# pcdibpix.asm -> /product/ipcore/ipshared/src/RCS/pcdibpix.asm,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

 ^

ENDIF

