;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcexpand.asm.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The 386/486 processors can reload source words quickly.

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; In the SPARC algorithms, we went to a great deal of trouble to use double
; word (64 bit) write operations.  Writes on an SS2 are slow, so it was
; beneficial.  The 386/486 has relatively fast 32 bit writes and no 64 bit
; writes, so we will discard that complication.

; The SPARC algorithms broke the entire image into blocks, processing
; each block vertically.  This was a good compromise between efficient
; processing (processing the image vertically minimizes memory writes of
; partial results) and cache utilization (minimized by horizontal image
; reads).  In the 386/486 world, there is only one partial result.
; Therefore, the algorithm can be constructed to go through the image
; either horizontally or vertically.  Since horizontal motion maximizes
; cache utilization, we'll process the image horizontally.


_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcexpand.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment:
; EAX:	scratch,  output word
; EBX:	scratch, source word 
; ECX:	scratch
; EDX:	address of first destination line (expand2), scratch for others
; EBP:	points to variables on the stack
; ESI:	source offset
; EDI:	destination offset
; ESP:	stack pointer

rDstOffset	EQU edi
rSrcOffset	EQU esi
rDstAddr1	EQU edx
rCount		EQU edx
exp2TabAddrMask	EQU 01FEH
exp4TabAddrMask	EQU 03FCH
exp8TabAddrMask	EQU 03FCH

; Local Variables:
; srcAddr:	address of left end of current source line
; dstAddr:	address of left end of current dest line
; imageHeight:	number of lines in source image, not including frame
; imageWidth:	width in pixels, not including frame, of source
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; dstPitch:	number of bytes to be added to address to move down one line
;		in destination (output) image
; blockPitch: 	number of bytes to be added to dest address to move down
;		2 lines
; tabExp2:	pointer to 2x expansion table
; tabExp4:	pointer to 4x expansion table
; tabExp8:	pointer to 8x expansion table
; tabExp16:	pointer to 16x expansion table

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPWORD TYPEDEF NEAR PTR WORD
NPDWORD TYPEDEF NEAR PTR DWORD
WriteOpProto TYPEDEF PROTO NEAR
WriteOpPtr TYPEDEF PTR WriteOpProto

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_expand2Asm.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by fexpand.c

expand2AsmStart::
ip_expand2Asm	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, \
	dstAddr:DWORD, dstPitch:DWORD,\
	imageWidth:DWORD, imageHeight:DWORD, tabExp2:DWORD
LOCAL	blockPitch:DWORD

; All registers used are also pushed on the stack automatically.

	mov	eax, pIPState		; keep assembly happy

; assume that both srcAddr and destAddr are at least 32-bit aligned
; and point to the image NOT the frame.
; assume imageHeight is at least 1.

; dstAddr1 points to the line below the primary destination line so that we
; can get at both rows of each 2x2 expanded square

	mov	eax, dstPitch
	mov	rDstAddr1, dstAddr
	add	rDstAddr1, eax		; get 2nd dest line

; get dest address increment
	shl	eax, 1
	mov	blockPitch, eax

; imageWidth will be compared to dstOffset to see if the end of the line
; has been reached.  imageWidth is the width of the source in bits.
; Convert it to the byte offset of the byte just beyond the last one to
; be written by ((2*imageWidth+31)/32)*4.
	mov	eax, imageWidth
	shl	eax, 1			; width of the expanded image
	add	eax, 31			; get ready for rounding
; do these next 2 shifts independently so that truncating happens
	shr	eax, 5			; get number of 32-bit words
	shl	eax, 2			; get number of bytes
	mov	imageWidth, eax

expand2xImageLoop:
	xor	rSrcOffset, rSrcOffset	; clear offsets
	xor	rDstOffset, rDstOffset
	
; The source line probably really has fewer valid bits that could fit in  
; the source line.  For example, the width may be 33 but looking at
; sBpl would allow 64 bits because of the pad.  As a result we cannot
; blindly expand the last word on the line since it could result in 
; more destination bits than we have room for.  In the 33 bits example,
; we should only write 66 bits (33 X 2) not 128 bits (64 x 2).
; The case of expansion by 2 is simple.  we'll simply expand up 16 source
; bits at a time and expand them since we know that we can always write 32 
; bits of destination (although we may be writing into the frame).
expand2xRowLoop:
	mov	ebx, srcAddr		; get source address
	mov	ebx, [ebx+rSrcOffset]	; get first 32 bit src word
	add	rSrcOffset, 4		; and point to the next one
	
; table points to words so need a word offset
	push	esi
	mov	ecx, ebx		; keep source word for later use
	shr	ebx, 23			; get word offset into table
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	esi, tabExp2
	mov	ax, [esi+ebx] 		; get next 16 bit dest
	shl	eax, 16			; put output into place
	mov	ebx, ecx		; get original source word
	shr	ebx, 15			; make into word offset
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	ax, [esi+ebx]		; get next 16 bit dest

	mov	ebx, dstAddr		; get first line of destination
	mov	[ebx+rDstOffset], eax 	; write the dest words
	mov	[rDstAddr1+rDstOffset], eax
	add	rDstOffset, 4		; point to the next dest

	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expand2xFinRow		; no, go to the next line

	mov	ebx, ecx		; get source word again
	shr	ebx, 7			; get word offset into table
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	esi, tabExp2
	mov	ax, [esi+ebx] 		; get next 16 bit dest
	shl	eax, 16			; put output into place
	mov	ebx, ecx		; get original source word
	shl	ebx, 1			; make into word offset
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	ax, [esi+ebx]		 ; get next 16 bit dest
	pop	esi

	mov	ecx, dstAddr		; get first line of destination
	mov	[ecx+rDstOffset], eax 	; write the dest words
	mov	[rDstAddr1+rDstOffset], eax
	add	rDstOffset, 4		; point to the next dest

	cmp	rDstOffset, imageWidth	; can we write the next word?
	jb	expand2xRowLoop		; yes, return to top of loop
	jmp	expand2xFinRow1

expand2xFinRow:
	pop	esi
	mov	ecx, dstAddr
expand2xFinRow1:
	add	ecx, blockPitch		; point to next dest lines
	mov	dstAddr, ecx
	add	rDstAddr1, blockPitch
	mov	eax, srcPitch		; point to next src lines
	add	srcAddr, eax
	dec	imageHeight
	jnz	expand2xImageLoop
	ret

ip_expand2Asm ENDP


; -------------------------------------------------------------------------
; The 2x Expand and Mask routine is identical to the 2x Expansion routine
; except the "mov" ops that write the destination are replaced with
; "and" ops.
; -------------------------------------------------------------------------

expandAndMask2AsmStart::
ip_expandAndMask2Asm	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, \
	dstAddr:DWORD, dstPitch:DWORD,\
	imageWidth:DWORD, imageHeight:DWORD, tabExp2:DWORD
LOCAL	blockPitch:DWORD

; All registers used are also pushed on the stack automatically.

	mov	eax, pIPState		; keep assembly happy

; assume that both srcAddr and destAddr are at least 32-bit aligned
; and point to the image NOT the frame.
; assume imageHeight is at least 1.

; dstAddr1 points to the line below the primary destination line so that we
; can get at both rows of each 2x2 expanded square

	mov	eax, dstPitch
	mov	rDstAddr1, dstAddr
	add	rDstAddr1, eax		; get 2nd dest line

; get dest address increment
	shl	eax, 1
	mov	blockPitch, eax

; imageWidth will be compared to dstOffset to see if the end of the line
; has been reached.  imageWidth is the width of the source in bits.
; Convert it to the byte offset of the byte just beyond the last one to
; be written by ((2*imageWidth+31)/32)*4.
	mov	eax, imageWidth
	shl	eax, 1			; width of the expanded image
	add	eax, 31			; get ready for rounding
; do these next 2 shifts independently so that truncating happens
	shr	eax, 5			; get number of 32-bit words
	shl	eax, 2			; get number of bytes
	mov	imageWidth, eax

expandAM2xImageLoop:
	xor	rSrcOffset, rSrcOffset	; clear offsets
	xor	rDstOffset, rDstOffset
	
; The source line probably really has fewer valid bits that could fit in  
; the source line.  For example, the width may be 33 but looking at
; sBpl would allow 64 bits because of the pad.  As a result we cannot
; blindly expand the last word on the line since it could result in 
; more destination bits than we have room for.  In the 33 bits example,
; we should only write 66 bits (33 X 2) not 128 bits (64 x 2).
; The case of expansion by 2 is simple.  we'll simply expand up 16 source
; bits at a time and expand them since we know that we can always write 32 
; bits of destination (although we may be writing into the frame).
expandAM2xRowLoop:
	mov	ebx, srcAddr		; get source address
	mov	ebx, [ebx+rSrcOffset]	; get first 32 bit src word
	add	rSrcOffset, 4		; and point to the next one
	
; table points to words so need a word offset
	push	esi
	mov	ecx, ebx		; keep source word for later use
	shr	ebx, 23			; get word offset into table
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	esi, tabExp2
	mov	ax, [esi+ebx] 		; get next 16 bit dest
	shl	eax, 16			; put output into place
	mov	ebx, ecx		; get original source word
	shr	ebx, 15			; make into word offset
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	ax, [esi+ebx]		; get next 16 bit dest

	mov	ebx, dstAddr		; get first line of destination
	and	[ebx+rDstOffset], eax 	; mask the dest words
	and	[rDstAddr1+rDstOffset], eax
	add	rDstOffset, 4		; point to the next dest

	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expandAM2xFinRow	; no, go to the next line

	mov	ebx, ecx		; get source word again
	shr	ebx, 7			; get word offset into table
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	esi, tabExp2
	mov	ax, [esi+ebx] 		; get next 16 bit dest
	shl	eax, 16			; put output into place
	mov	ebx, ecx		; get original source word
	shl	ebx, 1			; make into word offset
	and	ebx, exp2TabAddrMask	; get only the necessary bits
	mov	ax, [esi+ebx]		 ; get next 16 bit dest
	pop	esi

	mov	ecx, dstAddr		; get first line of destination
	and	[ecx+rDstOffset], eax 	; mask the dest words
	and	[rDstAddr1+rDstOffset], eax
	add	rDstOffset, 4		; point to the next dest

	cmp	rDstOffset, imageWidth	; can we write the next word?
	jb	expandAM2xRowLoop	; yes, return to top of loop
	jmp	expandAM2xFinRow1

expandAM2xFinRow:
	pop	esi
	mov	ecx, dstAddr
expandAM2xFinRow1:
	add	ecx, blockPitch		; point to next dest lines
	mov	dstAddr, ecx
	add	rDstAddr1, blockPitch
	mov	eax, srcPitch		; point to next src lines
	add	srcAddr, eax
	dec	imageHeight
	jnz	expandAM2xImageLoop
	ret

ip_expandAndMask2Asm ENDP


;---------------------------------------------------------------------------

expand4AsmStart::
ip_expand4Asm	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, \
	dstAddr:DWORD, dstPitch:DWORD,\
	imageWidth:DWORD, imageHeight:DWORD, tabExp4:DWORD
LOCAL	blockPitch:DWORD, destAddr1:DWORD, destAddr2:DWORD, destAddr3:DWORD

	mov	eax, pIPState		; keep assembly happy

; assume that both srcAddr and destAddr are at least 32-bit aligned
; and point to the image NOT the frame.
; assume imageHeight is at least 1.

; dstAddr1 points to the line below the primary destination line so that we
; can get at both rows of each 4x4 expanded square

	mov	eax, dstPitch
	mov	ebx, dstAddr
	add	ebx, eax		; point to 2nd dest line
	mov	destAddr1, ebx
	add	ebx, eax		; point to the 3rd dest line
	mov	destAddr2, ebx
	add	ebx, eax		; point to the 4th dest line
	mov	destAddr3, ebx

	shl	eax, 2			; loop skips 4 lines
	mov	blockPitch, eax

; imageWidth will be compared to dstOffset to see if the end of the line
; has been reached.  imageWidth is the width of the source in bits.
; Convert it to the byte offset of the byte just beyond the last one to
; be written by ((4*imageWidth+31)/32)*4.
	mov	eax, imageWidth
	shl	eax, 2			; width of the expanded image
	add	eax, 31			; get ready for rounding
; do these next 2 shifts independently so that truncating happens
	shr	eax, 5			; get number of 32-bit words
	shl	eax, 2			; get number of bytes
	mov	imageWidth, eax

; the last source word (32 bits) may cause from 4 to 128 bits to be
; written. These may not all be available to be written.  Since we read 32
; source bits at a time we must check after each 8 bits to make sure
; that we can write the next destination word.

expand4xImageLoop:
	xor	rSrcOffset, rSrcOffset	; clear offsets
	xor	rDstOffset, rDstOffset

expand4xRowLoop:
	mov	ebx, srcAddr		; get source address
	mov	ecx, [ebx+rSrcOffset]	; get source word
	add	rSrcOffset, 4		; point to the next word

	mov	edx, ecx		; save for later use
	shr	ecx, 22			; get bits [31..24]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]		; get expanded bits
	mov	ebx, dstAddr		; get first line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr1		; get second line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr2		; get third line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expand4xFinRow		; no, go to the next line

					; yes, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 14			; get bits [23..16]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]
	mov	ebx, dstAddr		; get first line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr1		; get second line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr2		; get third line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expand4xFinRow		; no, go to the next line

					; yes, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 6			; get bits [15..8]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]
	mov	ebx, dstAddr		; get first line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr1		; get second line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr2		; get third line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expand4xFinRow		; no, go to the next line

					; yes, continue
	mov	ecx, edx		; get source word back
	shl	ecx, 2			; get bits [7..0]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]
	mov	ebx, dstAddr		; get first line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr1		; get second line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr2		; get third line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	mov	[ebx+rDstOffset], eax	; write expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jb	expand4xRowLoop		; yes, continue with line

					; no, done with line
expand4xFinRow:
	mov	ecx, blockPitch		; point to next dest lines
	add	dstAddr, ecx
	add	destAddr1, ecx
	add 	destAddr2, ecx
	add	destAddr3, ecx
	mov	eax, srcPitch		; point to next source line
	add	srcAddr, eax
	dec 	imageHeight		; are we done with the image?
	jnz	expand4xImageLoop	; no, continue with next line
	ret				; yes, get outta here
ip_expand4Asm ENDP


; -------------------------------------------------------------------------
; The 4x Expand and Mask routine is identical to the 4x Expansion routine
; except the "mov" ops that write the destination are replaced with
; "and" ops.
; -------------------------------------------------------------------------

expandAndMask4AsmStart::
ip_expandAndMask4Asm	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, \
	dstAddr:DWORD, dstPitch:DWORD,\
	imageWidth:DWORD, imageHeight:DWORD, tabExp4:DWORD
LOCAL	blockPitch:DWORD, destAddr1:DWORD, destAddr2:DWORD, destAddr3:DWORD

	mov	eax, pIPState		; keep assembly happy

; assume that both srcAddr and destAddr are at least 32-bit aligned
; and point to the image NOT the frame.
; assume imageHeight is at least 1.

; dstAddr1 points to the line below the primary destination line so that we
; can get at both rows of each 4x4 expanded square

	mov	eax, dstPitch
	mov	ebx, dstAddr
	add	ebx, eax		; point to 2nd dest line
	mov	destAddr1, ebx
	add	ebx, eax		; point to the 3rd dest line
	mov	destAddr2, ebx
	add	ebx, eax		; point to the 4th dest line
	mov	destAddr3, ebx

	shl	eax, 2			; loop skips 4 lines
	mov	blockPitch, eax

; imageWidth will be compared to dstOffset to see if the end of the line
; has been reached.  imageWidth is the width of the source in bits.
; Convert it to the byte offset of the byte just beyond the last one to
; be written by ((4*imageWidth+31)/32)*4.
	mov	eax, imageWidth
	shl	eax, 2			; width of the expanded image
	add	eax, 31			; get ready for rounding
; do these next 2 shifts independently so that truncating happens
	shr	eax, 5			; get number of 32-bit words
	shl	eax, 2			; get number of bytes
	mov	imageWidth, eax

; the last source word (32 bits) may cause from 4 to 128 bits to be
; written. These may not all be available to be written.  Since we read 32
; source bits at a time we must check after each 8 bits to make sure
; that we can write the next destination word.

expandAM4xImageLoop:
	xor	rSrcOffset, rSrcOffset	; clear offsets
	xor	rDstOffset, rDstOffset

expandAM4xRowLoop:
	mov	ebx, srcAddr		; get source address
	mov	ecx, [ebx+rSrcOffset]	; get source word
	add	rSrcOffset, 4		; point to the next word

	mov	edx, ecx		; save for later use
	shr	ecx, 22			; get bits [31..24]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]		; get expanded bits
	mov	ebx, dstAddr		; get first line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr1		; get second line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr2		; get third line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expandAM4xFinRow	; no, go to the next line

					; yes, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 14			; get bits [23..16]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]
	mov	ebx, dstAddr		; get first line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr1		; get second line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr2		; get third line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expandAM4xFinRow	; no, go to the next line

					; yes, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 6			; get bits [15..8]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]
	mov	ebx, dstAddr		; get first line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr1		; get second line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr2		; get third line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jae	expandAM4xFinRow	; no, go to the next line

					; yes, continue
	mov	ecx, edx		; get source word back
	shl	ecx, 2			; get bits [7..0]
	and	ecx, exp4TabAddrMask
	mov	ebx, tabExp4		; expansion table address
	mov	eax, [ebx+ecx]
	mov	ebx, dstAddr		; get first line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr1		; get second line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr2		; get third line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	mov	ebx, destAddr3		; get fourth line of destination
	and	[ebx+rDstOffset], eax	; mask with expanded word
	add	rDstOffset, 4		; point to next dest word
	cmp	rDstOffset, imageWidth	; can we write the next word?
	jb	expandAM4xRowLoop	; yes, continue with line

					; no, done with line
expandAM4xFinRow:
	mov	ecx, blockPitch		; point to next dest lines
	add	dstAddr, ecx
	add	destAddr1, ecx
	add 	destAddr2, ecx
	add	destAddr3, ecx
	mov	eax, srcPitch		; point to next source line
	add	srcAddr, eax
	dec 	imageHeight		; are we done with the image?
	jnz	expandAM4xImageLoop	; no, continue with next line
	ret				; yes, get outta here
ip_expandAndMask4Asm ENDP


;-----------------------------------------------------------------------
; The 8x Expand and Mask routine uses the 8x expansion routine.  It does
; point expWrRoutine to ExpandAM8WriteDst instead of Expand8WriteDst.
; The former does and ops instead of memory writes.
;
; WARNING NOTE!!
;   Both ip_expandAndMask8Asm and ip_expand8Asm are called with
;   EXACTLY the same parameters, and have EXACTLY the same local variables,
;   EXACTLY in the same order.  This is _not_ coincidental.  The local
;   variable expWrRoutine tells the body below which write routine to call,
;   but we get away without passing the local variable because we cheat and
;   fake the execution out by having them in exactly the same place on the
;   stack and therefore it is just like passing each procedures local
;   variables into expand8Entry!    -ejaquith
;-----------------------------------------------------------------------

expandAndMask8AsmStart::
ip_expandAndMask8Asm	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, \
	dstAddr:DWORD, dstPitch:DWORD,\
	imageWidth:DWORD, imageHeight:DWORD, tabExp8:DWORD
LOCAL	expWrRoutine:DWORD, blockPitch:DWORD, destAddr1:DWORD


; Point to the proper writing routine for ip_expandAndMask8Asm.
	mov	eax, pIPState		; keep assembly happy
	mov	expWrRoutine, OFFSET ExpandAM8WriteDst
	mov	eax, srcAddr		; touch all the vars to keep
	mov	eax, srcPitch		; the assembler happy
	mov	eax, dstAddr
	mov	eax, dstPitch
	mov	eax, imageWidth
	mov	eax, imageHeight
	mov	eax, tabExp8
	mov	eax, blockPitch
	mov	eax, destAddr1
	jmp	expand8Entry
	ret				; fake return for the sake of
					; the assembler
ip_expandAndMask8Asm	ENDP

;-----------------------------------------------------------------------
; WARNING NOTE!!
;   See the warning note above.
expand8AsmStart::
ip_expand8Asm	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, \
	dstAddr:DWORD, dstPitch:DWORD,\
	imageWidth:DWORD, imageHeight:DWORD, tabExp8:DWORD
LOCAL	expWrRoutine:DWORD, blockPitch:DWORD, destAddr1:DWORD
; Point to the proper writing routine for ip_expand8Asm.
	mov	expWrRoutine, OFFSET Expand8WriteDst

	mov	eax, pIPState		; keep assembly happy

; assume that both srcAddr and destAddr are at least 32-bit aligned
; and point to the image NOT the frame.
; assume imageHeight is at least 1.

; point at all lines below the primary destination line so that we
; can easily get at all rows of each 8x8 expanded square
expand8Entry:
	mov	ebx, dstAddr
	mov	destAddr1, ebx		; point to 1st dest line
	mov	eax, dstPitch
	shl	eax, 3			; loop skips 8 lines
	mov	blockPitch, eax

; imageWidth will be compared to dstOffset to see if the end of the line
; has been reached.  imageWidth is the width of the source in bits.
; Convert it to the byte offset of the byte just beyond the last one to
; be written by ((8*imageWidth+31)/32)*4.
	mov	eax, imageWidth
	shl	eax, 3			; width of the expanded image
	add	eax, 31			; get ready for rounding
; do these next 2 shifts independently so that truncating happens
	shr	eax, 5			; get number of 32-bit words
	shl	eax, 2			; get number of bytes
	mov	imageWidth, eax

; the last source word (32 bits) may cause from 8 to 256 bits to be
; written. These may not all be available to be written.  Since we read 32
; source bits at a time we must check after each 4 bits to make sure
; that we can write the next destination word.

expand8xImageLoop:
	xor	rSrcOffset, rSrcOffset	; clear offsets
	xor	rDstOffset, rDstOffset

expand8xRowLoop:
	mov	ebx, srcAddr		; get source address
	mov	ecx, [ebx+rSrcOffset]	; get source word
	add	rSrcOffset, 4		; point to the next word

	mov	edx, ecx		; save for later use
	shr	ecx, 22			; get bits [31..24]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8
	add	ebx, 256*4		; hi nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand8xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 22			; get bits [27..24]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8		; low nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand8xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 14			; get bits [23..20]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8
	add	ebx, 256*4		; hi nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand8xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 14			; get bits [19..16]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8		; low nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand8xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 6			; get bits [15..12]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8
	add	ebx, 256*4		; hi nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand8xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source word back
	shr	ecx, 6			; get bits [11..8]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8		; low nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand8xFinRow		; done, go to the next line

					; not done, continue

	mov	ecx, edx		; get source word back
	shl	ecx, 2			; get bits [7..4]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8
	add	ebx, 256*4		; hi nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand8xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source word back
	shl	ecx, 2			; get bits [3..0]
	and	ecx, exp8TabAddrMask
	mov	ebx, tabExp8		; low nibble expansion table address
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	expWrRoutine		; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jb	expand8xRowLoop		; not done, continue with line

					; done with line
expand8xFinRow:
	mov	ecx, blockPitch		; point to next dest line
	add	destAddr1, ecx
	mov	eax, srcPitch		; point to next source line
	add	srcAddr, eax
	dec 	imageHeight		; are we done with the image?
	jnz	expand8xImageLoop	; no, continue with next line
	ret				; yes, get outta here
ip_expand8Asm ENDP

; Procedure: Expand8WriteDst
; This procedure assumes that:
;   eax contains the expanded word
;   ebx contains the first line of the destination to write to
;   ecx contains the destination pitch increment
; It will write the 8 destination words at the current destination offset
; found in rDstOffset.
; Smashes ebx.
; Increments rDstOffset before returning.
Expand8WriteDst PROC NEAR
	add	ebx, rDstOffset		; save us time and avoid offsets below
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get second line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get third line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get fourth line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get first line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get second line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get third line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get fourth line of destination
	mov	[ebx], eax		; write expanded word
	add	rDstOffset, 4		; point to next dest word
	ret
Expand8WriteDst ENDP

;-------------------------------------------------------------------------

; Procedure: ExpandAM8WriteDst
; This is the same as Expand8WriteDst except it ANDs the expanded data
; with the dest instead of writing it there.
; This procedure assumes that:
;   eax contains the expanded word
;   ebx contains the first line of the destination to write to
;   ecx contains the destination pitch increment
; It will write the 8 destination words at the current destination offset
; found in rDstOffset.
; Smashes ebx.
; Increments rDstOffset before returning.
ExpandAM8WriteDst PROC NEAR
	add	ebx, rDstOffset		; save us time and avoid offsets below
	and	[ebx], eax		; mask with expanded word
	add	ebx, ecx		; get second line of destination
	and	[ebx], eax		; mask with expanded word
	add	ebx, ecx		; get third line of destination
	and	[ebx], eax		; mask with expanded word
	add	ebx, ecx		; get fourth line of destination
	and	[ebx], eax		; mask with expanded word
	add	ebx, ecx		; get first line of destination
	and	[ebx], eax		; mask with expanded word
	add	ebx, ecx		; get second line of destination
	and	[ebx], eax		; mask with expanded word
	add	ebx, ecx		; get third line of destination
	and	[ebx], eax		; mask with expanded word
	add	ebx, ecx		; get fourth line of destination
	and	[ebx], eax		; mask with expanded word
	add	rDstOffset, 4		; point to next dest word
	ret
ExpandAM8WriteDst ENDP


;-------------------------------------------------------------------------


expand16AsmStart::
ip_expand16Asm	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, \
	dstAddr:DWORD, dstPitch:DWORD,\
	imageWidth:DWORD, imageHeight:DWORD, tabExp16:DWORD
LOCAL	blockPitch:DWORD, destAddr1:DWORD

	mov	eax, pIPState		; keep assembly happy

; assume that both srcAddr and destAddr are at least 32-bit aligned
; and point to the image NOT the frame.
; assume imageHeight is at least 1.

; points at all lines below the primary destination line so that we
; can easily get at all rows of each 16x16 expanded square

	mov	ebx, dstAddr
	mov	destAddr1, ebx		; point to 1st dest line
	mov	eax, dstPitch
	shl	eax, 4			; loop skips 16 lines
	mov	blockPitch, eax

; imageWidth will be compared to dstOffset to see if the end of the line
; has been reached.  imageWidth is the width of the source in bits.
; Convert it to the byte offset of the byte just beyond the last one to
; be written by ((16*imageWidth+31)/32)*4.
	mov	eax, imageWidth
	shl	eax, 4			; width of the expanded image
	add	eax, 31			; get ready for rounding
; do these next 2 shifts independently so that truncating happens
	shr	eax, 5			; get number of 32-bit words
	shl	eax, 2			; get number of bytes
	mov	imageWidth, eax

; the last source byte (8 bits) may cause from 16 to 128 bits to be
; written. These may not all be available to be written.  Since we read 8
; source bits at a time we must check after each 2 bits to make sure
; that we can write the next destination word.

expand16xImageLoop:
	xor	rSrcOffset, rSrcOffset	; clear offsets
	xor	rDstOffset, rDstOffset

expand16xRowLoop:
	mov	ebx, srcAddr		; get source address
	lea	ecx, [ebx+rSrcOffset]	; get the address with
					; the byte order reversed within
					; words
	xor	ecx, 3			; fix the byte order
	movzx	ecx, BYTE PTR [ecx]	; get source byte
	add	rSrcOffset, 1		; point to the next byte
	shl	ecx, 2			; make it index words
	mov	edx, ecx		; save for later use

	add	ecx, 3*256*4		; point to word 3 table
	mov	ebx, tabExp16		; expansion table pointer for write
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	Expand16WriteDst	; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand16xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source byte back
	add	ecx, 2*256*4		; point to word 2 table
	mov	ebx, tabExp16		; expansion table pointer for write
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	Expand16WriteDst	; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand16xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source byte back
	add	ecx, 256*4		; point to word 1 table
	mov	ebx, tabExp16		; expansion table pointer for write
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	Expand16WriteDst	; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jae	expand16xFinRow		; done, go to the next line

					; not done, continue
	mov	ecx, edx		; get source byte back
					; already pointing to word 0 table
	mov	ebx, tabExp16		; expansion table pointer for write
	mov	eax, [ebx+ecx]		; get expanded word and set for write
	mov	ebx, destAddr1		; get 1st dest line and set for write
	mov	ecx, dstPitch		; set pitch for write
	call	Expand16WriteDst	; write the destination word
	cmp	rDstOffset, imageWidth	; are we done with this line?
	jb	expand16xRowLoop	; not done, continue with line

					; done with line
expand16xFinRow:
	mov	ecx, blockPitch		; point to next dest lines
	add	destAddr1, ecx
	mov	eax, srcPitch		; point to next source line
	add	srcAddr, eax
	dec 	imageHeight		; are we done with the image?
	jnz	expand16xImageLoop	; no, continue with next line
	ret				; yes, get outta here
ip_expand16Asm ENDP

; Procedure: Expand16WriteDst
; This procedure assumes that:
;   eax contains the expanded word
;   ebx contains the first line of the destination to write to
;   ecx contains the destination pitch increment
; It will write the 16 destination words at the current destination offset
; found in rDstOffset.
; Smashes ebx.
; Increments rDstOffset before returning.

Expand16WriteDst PROC NEAR
	add	ebx, rDstOffset		; save us time and avoid offsets below
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get second line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get third line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get fourth line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get fifth line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get sixth line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get seventh line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get eighth line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get ninth line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get tenth line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get eleventh line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get 12th line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get 13th line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get 14th line of destination
	mov	[ebx], eax		; write expanded word
	add	ebx, ecx		; get 15th line of destination
	mov	[ebx], eax		; write expanded word
 	add	ebx, ecx		; get 16th line of destination
	mov	[ebx], eax		; write expanded word
	add	rDstOffset, 4		; point to next dest word
	ret
Expand16WriteDst ENDP


;-------------------------------------------------------------------------

_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
; no global initialized data used
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcexpand.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.13  1998/01/13 23:58:08  ddavies
pcexpand.asm -> /product/ipcore/ipshared/src/RCS/pcexpand.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.12  1996/09/17 00:45:31  ddavies
pcexpand.asm -> /product/ipcore/ipshared/src/RCS/pcexpand.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.11  1996/09/16 00:50:37  ddavies
pcexpand.asm -> /product/ipcore/ipshared/src/RCS/pcexpand.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.10  1995/09/15 21:47:40  ddavies
pcexpand.asm -> /product/ipcore/ipshared/src/RCS/pcexpand.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

Revision 1.9  1994/08/03  17:21:48  ejaquith
ejaquith on Wed Aug 3 10:16:56 PDT 1994

pcexpand.asm -> /product/ipcore/ipshared/src/RCS/pcexpand.asm,v

Reentrancy!
Got rid of last globals in this file, by adding local variables to
each function, and by passing pointers like expWrRoutine cleverly
between functions.  Removed the 8 and 16 destination line pointers and
now instead calculate each as an offset from the previous.  Changed
interface to internal WriteDst functions and sped them up.

Revision 1.8  1994/07/27  15:53:51  ejaquith
pcexpand.asm -> /product/ipcore/ipshared/src/RCS/pcexpand.asm,v

Reentrancy!
Removed globals and made functions be called with pointers to expansion
tables instead.

Revision 1.7  1993/11/05  01:52:07  ddavies
pcexpand.asm -> /product/ipcore/ipshared/src/RCS/pcexpand.asm,v

Add RCS magic to have string specifying this file's name, rev number and
last edit date into the object file.  Also add change log at the bottom.

^

ENDIF

