;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcfloyd.asm:  template for assembly code.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The 386/486 processors can reload source words quickly.

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.


_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcfloyd.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Definitions
; EAX:	scratch
; EBX:	binary destination word 
; ECX:	gray word line 0
; EDX:	gray word line 1
; EBP:	points to variables on the stack
; ESI:	pointer to source image
; EDI:	pointer to gray buffer
; ESP:	stack pointer

rDLine0		EQU ebx
rGLine0		EQU ecx
rGLine0Lower	EQU cl
rGLine1		EQU edx
rGLine1Lower	EQU dl
rSrcAddr	EQU esi
rGrayBuf	EQU edi

; constants
DBLOCK_HEIGHT	EQU 1

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPWORD TYPEDEF NEAR PTR WORD
NPDWORD TYPEDEF NEAR PTR DWORD


; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name is ip_ditherFloydLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by ffloyd.c

; Local Variables:
; srcAddr:	address of left end of current source line
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; dstAddr:	address of left end of current destination line
; dstPitch:	number of bytes to be added to address to move down one line
;		in the destination (input) image
; grayBuf:	address of buffer holding pixels from previous passes
; imageWidth:	width of dithered image (in pixels)
; hdeTable:	address of horizontal error table
; vdeTable:	address of vertical error table
; dtTable:	address of pixel truncation table
; drcTable:	address of dynamic range correction table
; hdeTable, vdeTable, and dtTable are unused passed in parameters since we
; have copied these tables to known offsets in memory.  However, the function
; gets called with them when the C code version is used, so we have to accept
; them and then ignore them.
; Unlike the sparc assembly version or the C code version, this code
; process only one line at a time.  This version needs more local variables
; than there are registers, but a version that writes more than one 
; destination line at a time would use even more.  Since we would be writing
; these local variables to memory anyway, we might as well only process
; one line and write to the gray buffer (already in memory).
; At some point (somewhere around 8 lines, give or take one), the 
; advantage actually lies with the multiple line version.  It seems that
; the added complexity is not worth it.

; This single line dithering procedure will jump right to the
; ip_ditherFloydLine procedure (since it is a single line dither after all!)
; WARNING NOTE!  The parameter list and local variable declarations _must_
; be the same for this to work.
ip_ditherFloydSingleLine	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, dstAddr:DWORD, \
	dstPitch:DWORD, grayBuf:DWORD, imageWidth:DWORD, blockHt:DWORD,\
	hdeTable:DWORD, vdeTable:DWORD, dtTable:DWORD, drcTable:DWORD
LOCAL	loopCount: DWORD, lastWritten: DWORD

; keep the assembler from complaining
	mov	eax, pIPState
	mov	eax, srcAddr
	mov	eax, srcPitch
	mov	eax, dstAddr
	mov	eax, dstPitch
	mov	eax, grayBuf
	mov	eax, imageWidth
	mov	eax, blockHt
	mov	eax, hdeTable
	mov	eax, vdeTable
	mov	eax, dtTable
	mov	eax, drcTable
	mov	eax, loopCount
	mov	eax, lastWritten
	jmp	ditherFloydSingleLine
	ret	; keep assembler happy
ip_ditherFloydSingleLine ENDP


ditherFloydLineStart::
ip_ditherFloydLine	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, dstAddr:DWORD, \
	dstPitch:DWORD, grayBuf:DWORD, imageWidth:DWORD, blockHt:DWORD,\
	hdeTable:DWORD, vdeTable:DWORD, dtTable:DWORD, drcTable:DWORD
LOCAL	loopCount: DWORD, lastWritten: DWORD

; keep the assembler from complaining
	mov	eax, pIPState
	mov	eax, srcPitch
	mov	eax, dstPitch
	mov	eax, hdeTable
	mov	eax, vdeTable
	mov	eax, dtTable
	mov	eax, blockHt

ditherFloydSingleLine:
; load data from gray buffer into rGLine0.
	mov	rGrayBuf, grayBuf	; keep the pointer to grayBuf in a register
	mov	rGLine0, 0		; make sure the entire word is 0
	mov	rGLine0Lower, [rGrayBuf+DBLOCK_HEIGHT+0]	; read the first byte
	mov	ebx, 0			; clear the word since the move won't
	mov	bl, [rGrayBuf+DBLOCK_HEIGHT+1]	; read the second byte
	shl	rGLine0, 22		; get them in place
	shl	ebx, 11
	or	rGLine0, ebx		; first and second bytes
	mov	rGLine0Lower, [rGrayBuf+DBLOCK_HEIGHT+2]	; and the third byte

; load data from image into rGLine1
	mov	ebx, drcTable
	mov	rSrcAddr, srcAddr	; keep the pointer to the source in a register

; if the original request was to only do a rectangle of the image
; the source address will not be a multiple of 4 bytes.  In order to make
; the inner loop as fast as possible, it is assumed that the offset 
; for the next source address is a well known constant 0 through 4.
; This means that we must enter the loop at different place (and get
; the first 3 source words from different offsets).
;
; least sig.	offsets to read		enter loop this
; 2 bits of	(after addr masked)	# of reads from
; source	to init rGLine1		the top
; 00		-4, +3, +2		0 (srcAddr+1)
; 01		+3, +2, +1		1 (srcAddr+0) (inc src after this)
; 10		+2, +1, 0		2 (srcAddr+3)
; 11		+1, +0, (inc addr) +3	3 (srcAddr+2)

	mov	eax, rSrcAddr		; get the bottom 2 bits  
	and	eax, 3			; 
	and	rSrcAddr, 0FFFFFFFCH	; src addr is now a mult of 4 bytes

; compute the number of times we go through the loop
; the number of times that we go through the loop is the width of the
; image plus some pad (the distance from the top of the loop) to make
; sure that we process enough bits.  Why the pad?  We always want to
; decrement count by 4 at the bottom of the loop (if we don't we would 
; have to decrement the count and possibly jump after each souce word).
; Without the pad, if the loop wasn't entered at the top we might not
; do enough source words.  Consider souce address with 11 in the bottom
; bits and a width of 2.  If we just used the width, we would jump
; out of the loop at the bottom and miss one word; hence we add the pad
; (which is the number of words that we didn't write the first time
; through the loop but will decrement for).
	mov	rGLine1, imageWidth
	mov	lastWritten, rGLine1
	add	rGLine1, eax
	mov	loopCount, rGLine1
	jmp	dFloydReadSrc[eax*4]

; original source had 01 in the bottom bits
dFloydLoadSrc1:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+3] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+2]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+1]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] ; and complete rGLine1

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

	jmp	dFloydSrcPlus0

; original source had 10 in the bottom bits
dFloydLoadSrc2:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+2] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+1]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+0]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] ; and complete rGLine1

	add	rSrcAddr, 4		; point to the next source word

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

	jmp	dFloydSrcPlus3

; original source had 11 in the bottom bits
dFloydLoadSrc3:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+1] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+0]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	add	rSrcAddr, 4		; point to the next source word
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+3]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] ; and complete rGLine1

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

	jmp	dFloydSrcPlus3

; original source had 00 in the bottom bits
dFloydLoadSrc0:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr-4] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+3]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+2]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] ; and complete rGLine1

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

;	jmp	dFloydSrcPlus1


; ***** MAIN LOOP *****

; do 4 source bytes at a time so we now the offset from srcAddr at each
; point in the loop and this way don't have to calculate the byte swapped
; address on the fly.

dFloydRowLoop:


dFloydSrcPlus1:
; get threshold pixel and error values
	mov	eax, rGLine0		; Line was TOO LONG!
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable[eax] ; apply 7/16 error
	adc	rDLine0, rDLine0	; shift data into binary output
	jc	dFloydWrite0		; is it time to do a write?

; shift new pixel into gray line 0
dFloydThresh0:
	shl	rGLine0, 11		; make room for next source byte
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+1] 	; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]


dFloydSrcPlus0:
; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable[eax] ; apply 7/16 error
	adc	rDLine0, rDLine0	; shift data into binary output
	jc	dFloydWrite1		; is it time to do a write?

; shift new pixel into gray line 0
dFloydThresh1:
	shl	rGLine0, 11		; make room for next source byte
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+0]	; get next source pixel
	add	rSrcAddr, 4		; point to next source word
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]


dFloydSrcPlus3:
; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable[eax] ; apply 7/16 error
	adc	rDLine0, rDLine0	; shift data into binary output
	jc	dFloydWrite2		; is it time to do a write?

; shift new pixel into gray line 0
dFloydThresh2:
	shl	rGLine0, 11		; make room for next source byte
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+3]	; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]


dFloydSrcPlus2:
; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable[eax] ; apply 7/16 error
	adc	rDLine0, rDLine0	; shift data into binary output
	jc	dFloydWrite3		; is it time to do a write?

; shift new pixel into gray line 0
dFloydThresh3:
	shl	rGLine0, 11		; make room for next source byte
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+2]	; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]


; we're finished with one column of data.  should we keep going?
	sub	loopCount, 4
	jg	dFloydRowLoop		; are we at the end yet?

; no, we shouldn't put out anymore dithered data, but we should finish
; filling the gray buffer.  The top line is DBLOCK_HEIGHT bytes ahead of
; us now.  We need to convert at least DBLOCK_HEIGHT more bytes, then
; put out all the data in rGLine1.
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it

	mov	eax, 0			; make sure this is clear
	mov	al, [rSrcAddr]		; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]

; now write out all the data in rGLine1
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf+1], al	; insert it

	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 11			;
	and	eax, 03FFH		; get rid of high order pixel 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf+2], al	; insert it

	and	rGLine1, 03FFH		; grab last output pixel
	mov	al, dTruncTable[rGLine1] ; truncate it
	mov	[rGrayBuf+3], al	; insert it

; we may have been left with some number of bits in the destination
; register.  Shift those bits into the most significant bits and then
; mask to get only the correct number (the inner loop could have 
; written 0-3 bits too many.

	mov	ecx, 32			; find shift count
	mov	eax, lastWritten	; number of real bits remaining
					; to be written
	sub	eax, loopCount		; but we overwrote by the negative
					; of loopCount so get the
					; the number of bits really in
					; rDLine0
	jz	dFloydLeave		; nothing to write
	sub	ecx, eax		; the shift count
	shl	rDLine0, cl		; put into place

dFloydWriteFinal:
	mov	eax, dstAddr		; get destination address
	not	rDLine0			; invert dithered data
	mov	ecx, imageWidth		; get mask
	and	ecx,  1FH
	mov	ecx, dFloydWriteMask[ecx*4]
	and	rDLine0, ecx		; get rid of the extra bits
	mov	[eax], rDLine0

dFloydLeave:
	ret

dFloydWrite0:
	mov	eax, dstAddr		; get destination address
	not	rDLine0			; invert dithered data
	mov	[eax], rDLine0
	mov	rDLine0, 1		; insert marker bit
	add	eax, 4			; now point to the next dest word
	mov	dstAddr, eax
	sub	lastWritten, 32		; 32 more bits written
	jmp	dFloydThresh0

dFloydWrite1:
	mov	eax, dstAddr		; get destination address
	not	rDLine0			; invert dithered data
	mov	[eax], rDLine0
	mov	rDLine0, 1		; insert marker bit
	add	eax, 4			; now point to the next dest word
	mov	dstAddr, eax
	sub	lastWritten, 32		; 32 more bits written
	jmp	dFloydThresh1

dFloydWrite2:
	mov	eax, dstAddr		; get destination address
	not	rDLine0			; invert dithered data
	mov	[eax], rDLine0
	mov	rDLine0, 1		; insert marker bit
	add	eax, 4			; now point to the next dest word
	mov	dstAddr, eax
	sub	lastWritten, 32		; 32 more bits written
	jmp	dFloydThresh2

dFloydWrite3:
	mov	eax, dstAddr		; get destination address
	not	rDLine0			; invert dithered data
	mov	[eax], rDLine0
	mov	rDLine0, 1		; insert marker bit
	add	eax, 4			; now point to the next dest word
	mov	dstAddr, eax
	sub	lastWritten, 32		; 32 more bits written
	jmp	dFloydThresh3

ip_ditherFloydLine ENDP


ditherFloydLine2Start::
ip_ditherFloydLine2	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, dstAddr:DWORD, \
	dstPitch:DWORD, grayBuf:DWORD, imageWidth:DWORD, blockHt:DWORD,\
	hdeTable:DWORD, vdeTable:DWORD, dtTable:DWORD, drcTable:DWORD

LOCAL	loopCount: DWORD, lastWritten: DWORD

; keep the assembler from complaining
	mov	eax, pIPState
	mov	eax, srcPitch
	mov	eax, dstPitch
	mov	eax, blockHt
	mov	eax, hdeTable
	mov	eax, vdeTable
	mov	eax, dtTable

; load data from gray buffer into rGLine0.
	mov	rGrayBuf, grayBuf	; keep the pointer to grayBuf in a register
	mov	rGLine0, 0		; make sure the entire word is 0
	mov	rGLine0Lower, [rGrayBuf+DBLOCK_HEIGHT+0]	; read the first byte
	mov	ebx, 0			; clear the word since the move won't
	mov	bl, [rGrayBuf+DBLOCK_HEIGHT+1]	; read the second byte
	shl	rGLine0, 22		; get them in place
	shl	ebx, 11
	or	rGLine0, ebx		; first and second bytes
	mov	rGLine0Lower, [rGrayBuf+DBLOCK_HEIGHT+2]	; and the third byte


; load data from image into rGLine1
	mov	ebx, drcTable
	mov	rSrcAddr, srcAddr	; keep the pointer to the source in a register

; if the original request was to only do a rectangle of the image
; the source address will not be a multiple of 4 bytes.  In order to make
; the inner loop as fast as possible, it is assumed that the offset 
; for the next source address is a well known constant 0 through 4.
; This means that we must enter the loop at different place (and get
; the first 3 source words from different offsets).

; least sig.	offsets to read		enter loop this
; 2 bits of	(after addr masked)	# of reads from
; source	to init rGLine1		the top
; 00		-4, +3, +2		0 (srcAddr+1)
; 01		+3, +2, +1		1 (srcAddr+0) (inc src after this)
; 10		+2, +1, 0		2 (srcAddr+3)
; 11		+1, +0, (inc addr) +3	3 (srcAddr+2)

	mov	eax, rSrcAddr		; get the bottom 2 bits  
	and	eax, 3			; 
	and	rSrcAddr, 0FFFFFFFCH	; src addr is now a mult of 4 bytes

; compute the number of times we go through the loop
; the number of times that we go through the loop is the width of the
; image plus some pad (the distance from the top of the loop) to make
; sure that we process enough bits.  Why the pad?  We always want to
; decrement count by 4 at the bottom of the loop (if we don't we would 
; have to decrement the count and possibly jump after each souce word).
; Without the pad, if the loop wasn't entered at the top we might not
; do enough source words.  Consider souce address with 11 in the bottom
; bits and a width of 2.  If we just used the width, we would jump
; out of the loop at the bottom and miss one word; hence we add the pad.
	mov	rGLine1, imageWidth
	mov	lastWritten, rGLine1
	add	rGLine1, eax
	mov	loopCount, rGLine1
	jmp	dFloyd2ReadSrc[eax*4]


; original source had 01 in the bottom bits
dFloyd2LoadSrc1:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+3] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+2]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+1]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] ; and complete rGLine1

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

	jmp	dFloyd2SrcPlus0

; original source had 10 in the bottom bits
dFloyd2LoadSrc2:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+2] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+1]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+0]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] ; and complete rGLine1

	add	rSrcAddr, 4		; point to the next source word

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

	jmp	dFloyd2SrcPlus3

; original source had 11 in the bottom bits
dFloyd2LoadSrc3:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+1] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+0]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	add	rSrcAddr, 4		; point to the next source word
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+3]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] ; and complete rGLine1

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

	jmp	dFloyd2SrcPlus3

; original source had 00 in the bottom bits
dFloyd2LoadSrc0:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr-4] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+3]	; read the second byte
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	rGLine1, 22		; get then into place
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+2]	; read the third byte
	mov	rGLine1Lower, [ebx+eax] 	; and complete rGLine1

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

;	jmp	dFloyd2SrcPlus1


; ***** MAIN LOOP *****

; do 4 source bytes at a time so we now the offset from srcAddr at each
; point in the loop and this way don't have to calculate the byte swapped
; address on the fly.

dFloyd2RowLoop:

; ***** FIRST SOURCE BYTE *****

dFloyd2SrcPlus1:
; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable2[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable2[eax] ; apply 7/16 error
; insert dithered data into 2bpp output register
	add	rGLine0, rGLine0	; put first new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	add	rGLine0, rGLine0	; put second new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	jc	dFloyd2Write0		; do we need to write dest word?

; shift new pixel into gray line 0
dFloyd2Thresh0:
	shl	rGLine0, 9		; make room for next source byte
					; shift count is less 2 because
					; of method of inserting dest bits
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+1] 	; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]

; ***** SECOND SOURCE BYTE *****

dFloyd2SrcPlus0:
; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable2[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable2[eax] ; apply 7/16 error
; insert dithered data into 2bpp output register
	add	rGLine0, rGLine0	; put first new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	add	rGLine0, rGLine0	; put second new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	jc	dFloyd2Write1		; do we need to write dest word?

; shift new pixel into gray line 0
dFloyd2Thresh1:
	shl	rGLine0, 9		; make room for next source byte
					; shift count is less 2 because
					; of method of inserting dest bits
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+0]	; get next source pixel
	add	rSrcAddr, 4		; point to next source word
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]

; ***** THIRD SOURCE BYTE *****

dFloyd2SrcPlus3:
; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable2[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable2[eax] ; apply 7/16 error
; insert dithered data into 2bpp output register
	add	rGLine0, rGLine0	; put first new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	add	rGLine0, rGLine0	; put second new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	jc	dFloyd2Write2		; do we need to write dest word?

; shift new pixel into gray line 0
dFloyd2Thresh2:
	shl	rGLine0, 9		; make room for next source byte
					; shift count is less 2 because
					; of method of inserting dest bits
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+3]	; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]

; ***** FOURTH SOURCE BYTE *****

dFloyd2SrcPlus2:
; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable2[eax] ; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable2[eax] ; apply 7/16 error
; insert dithered data into 2bpp output register
	add	rGLine0, rGLine0	; put first new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	add	rGLine0, rGLine0	; put second new bit in carry
	adc	rDLine0, rDLine0	; shift data into output
	jc	dFloyd2Write3		; do we need to write dest word?

; shift new pixel into gray line 0
dFloyd2Thresh3:
	shl	rGLine0, 9		; make room for next source byte
					; shift count is less 2 because
					; of method of inserting dest bits
	mov	rGLine0Lower, [rGrayBuf+4]	; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it
	inc	rGrayBuf		; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr+2]	; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]


; we're finished with one column of data.  should we keep going?
	sub	loopCount, 4
	jg	dFloyd2RowLoop		; are we at the end yet?

; no, we shouldn't put out anymore dithered data, but we should finish
; filling the gray buffer.  The top line is DBLOCK_HEIGHT bytes ahead of
; us now.  We need to convert at least DBLOCK_HEIGHT more bytes, then
; put out all the data in rGLine1.
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it

	mov	eax, 0
	mov	al, [rSrcAddr]		; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]

; now write out all the data in rGLine1
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf+1], al	; insert it

	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 11			; 
	and	eax, 03FFH		; get rid of high order pixel
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf+2], al	; insert it

	and	rGLine1, 03FFH		; grab last output pixel
	mov	al, dTruncTable[rGLine1] ; truncate it
	mov	[rGrayBuf+3], al	; insert it


; we may have been left with some number of bits in the destination
; register.  Shift those bits into the most significant bits and then
; mask to get only the correct number (the inner loop could have 
; written 0-3 bits too many.

	mov	ecx, 32			; find shift count
	mov	eax, lastWritten	; number of real pixels remaining
					; to be written
	sub	eax, loopCount		; but we overwrote by the negative
					; of loopCount so get the
					; the number of bits really in
					; rDLine0
	jz	dFloyd2Leave		; nothing to write
	add	eax, eax		; acount for 2bits per pixel
	sub	ecx, eax		; the shift count
	shl	rDLine0, cl		; put into place

dFloyd2WriteFinal:
        mov     eax, dstAddr            ; get destination address
	mov	ecx, imageWidth		; get mask
	add	ecx, ecx		; mult by 2 because of 2bpp
	and	ecx,  1FH
	mov	ecx, dFloydWriteMask[ecx*4]
	and	rDLine0, ecx		; get rid of the extra bits
        mov     [eax], rDLine0        	; write final word

dFloyd2Leave:
	ret

dFloyd2Write0:
        mov     eax, dstAddr            ; get destination address
        add     eax, 4                  ; now point to the next dest word
        mov     dstAddr, eax
        mov     [eax-4], rDLine0        ; since we already inc'd it
        mov     rDLine0, 1              ; insert marker bit
	sub	lastWritten, 16		; 32 more bits written
	jmp	dFloyd2Thresh0

dFloyd2Write1:
        mov     eax, dstAddr            ; get destination address
        add     eax, 4                  ; now point to the next dest word
        mov     dstAddr, eax
        mov     [eax-4], rDLine0        ; since we already inc'd it
        mov     rDLine0, 1              ; insert marker bit
	sub	lastWritten, 16		; 32 more bits written
	jmp	dFloyd2Thresh1

dFloyd2Write2:
        mov     eax, dstAddr            ; get destination address
        add     eax, 4                  ; now point to the next dest word
        mov     dstAddr, eax
        mov     [eax-4], rDLine0        ; since we already inc'd it
        mov     rDLine0, 1              ; insert marker bit
	sub	lastWritten, 16		; 32 more bits written
	jmp	dFloyd2Thresh2

dFloyd2Write3:
        mov     eax, dstAddr            ; get destination address
        add     eax, 4                  ; now point to the next dest word
        mov     dstAddr, eax
        mov     [eax-4], rDLine0        ; since we already inc'd it
        mov     rDLine0, 1              ; insert marker bit
	sub	lastWritten, 16		; 32 more bits written
	jmp	dFloyd2Thresh3
ip_ditherFloydLine2 ENDP



dFloyd4DoPixel	MACRO srcOffset, nibble, drcTable
;; get threshold pixel and error values
	mov	eax, rGLine0
	shr	eax, 20			;; convert to a word index
	and	eax, 0FFFFFFFCH
	add	rGLine1, vdErrorTable[eax] ;; apply {3/16 5/16 1/16} error
	add	rGLine0, hdErrorTable[eax] ;; apply 7/16 error
	adc	rDLine0, rDLine0	;; shift data into binary output
	jc	dFloyd4to1Write&nibble&srcOffset	;; is it time to do a write?

dFloyd4to1Thresh&nibble&srcOffset&:
; shift new pixel into gray line 0
	shl	rGLine0, 11		;; make room for next source byte
	mov	rGLine0Lower, [rGrayBuf+4]	;; insert next source pixel

; calculate the output gray data for the gray buffer
	mov	eax, rGLine1		;; grab next output pixel
	shr	eax, 22			;; 
	mov	al, dTruncTable[eax]	;; truncate it
	mov	[rGrayBuf], al		;; insert it
	inc	rGrayBuf		;; point to next pixel

	mov	eax, 0
	mov	al, [rSrcAddr + srcOffset ] 	;; get next source pixel
	add	eax, drcTable
	shl	rGLine1, 11		;; make room for next pixel
	mov	rGLine1Lower, [eax]
	ENDM

dFloyd4DoWrite	MACRO	srcOffset, nibble
	mov	eax, dstAddr		; get destination address
	not	rDLine0			; invert dithered data
	mov	[eax], rDLine0
	mov	rDLine0, 1		; insert marker bit
	add	eax, 4			; now point to the next dest word
	mov	dstAddr, eax
	sub	lastWritten, 32		; 32 more bits written
	jmp	dFloyd4to1Thresh&nibble&srcOffset
	ENDM


; This single line dithering procedure will jump right to the
; ip_ditherFloydLine4to1 procedure (since it is a single line dither after
; all!)
; WARNING NOTE!  The parameter list and local variable declarations _must_
; be the same for this to work.
ip_ditherFloydSingleLine4to1	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, dstAddr:DWORD, \
	dstPitch:DWORD, grayBuf:DWORD, imageWidth:DWORD, blockHt:DWORD,\
	hdeTable:DWORD, vdeTable:DWORD, dtTable:DWORD, pHDRCTable:DWORD, \
	pLDRCTable:DWORD, pixel:DWORD
LOCAL	loopCount: DWORD, lastWritten: DWORD

; keep the assembler from complaining
	mov	eax, pIPState
	mov	eax, srcAddr
	mov	eax, srcPitch
	mov	eax, dstAddr
	mov	eax, dstPitch
	mov	eax, grayBuf
	mov	eax, imageWidth
	mov	eax, blockHt
	mov	eax, hdeTable
	mov	eax, vdeTable
	mov	eax, dtTable
	mov	eax, pHDRCTable
	mov	eax, pLDRCTable
	mov	eax, pixel
	mov	eax, loopCount
	mov	eax, lastWritten
	jmp	ditherFloydSingleLine4to1
	ret	; keep assembler happy
ip_ditherFloydSingleLine4to1 ENDP


	PUBLIC	ip_ditherFloydLine4to1
ip_ditherFloydLine4to1	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, dstAddr:DWORD, \
	dstPitch:DWORD, grayBuf:DWORD, imageWidth:DWORD, blockHt:DWORD,\
	hdeTable:DWORD, vdeTable:DWORD, dtTable:DWORD, pHDRCTable:DWORD, \
	pLDRCTable:DWORD, pixel:DWORD
LOCAL	loopCount: DWORD, lastWritten: DWORD

; keep the assembler from complaining
	mov	eax, pIPState
	mov	eax, srcPitch
	mov	eax, dstPitch
	mov	eax, blockHt
	mov	eax, hdeTable
	mov	eax, vdeTable
	mov	eax, dtTable

ditherFloydSingleLine4to1:
; load data from gray buffer into rGLine0.
	mov	rGrayBuf, grayBuf	; keep the pointer to grayBuf in a register
	mov	rGLine0, 0		; make sure the entire word is 0
	mov	rGLine0Lower, [rGrayBuf+DBLOCK_HEIGHT+0] 
					; read the first byte
	mov	ebx, 0			; clear the word since the move won't
	mov	bl, [rGrayBuf+DBLOCK_HEIGHT+1]	
					; read the second byte
	shl	rGLine0, 22		; get them in place
	shl	ebx, 11
	or	rGLine0, ebx		; first and second bytes
	mov	rGLine0Lower, [rGrayBuf+DBLOCK_HEIGHT+2]
					; and the third byte

	cmp	pixel, 1
	je	dFloyd4to1LowStart

dFloyd4to1HighStart:
; load data from image into rGLine1
	mov	ebx, pLDRCTable
	mov	rSrcAddr, srcAddr	; keep the pointer to the source in a register

; if the original request was to only do a rectangle of the image
; the source address will not be a multiple of 4 bytes.  In order to make
; the inner loop as fast as possible, it is assumed that the offset 
; for the next source address is a well known constant 0 through 4.
; This means that we must enter the loop at different place (and get
; the first 3 source words from different offsets).
;
; least sig.	offsets to read		enter loop this
; 2 bits of	(after addr masked)	# of reads from
; source	to init rGLine1		the top
; 00		-4, +3			0 (srcAddr+2)
; 01		+3, +2			1 (srcAddr+1)
; 10		+2, +1			2 (srcAddr+0) (inc src after this)
; 11		+1, +0 (inc addr)	3 (srcAddr+3)

	mov	eax, rSrcAddr		; get the bottom 2 bits  
	and	eax, 3			; 
	and	rSrcAddr, 0FFFFFFFCH	; src addr is now a mult of 4 bytes

; compute the number of times we go through the loop
; the number of times that we go through the loop is the width of the
; image plus some pad (the distance from the top of the loop) to make
; sure that we process enough bits.  Why the pad?  We always want to
; decrement count by 8 at the bottom of the loop (if we don't we would 
; have to decrement the count and possibly jump after each souce word).
; Without the pad, if the loop wasn't entered at the top we might not
; do enough source words.  Consider souce address with 11 in the bottom
; bits and a width of 2.  If we just used the width, we would jump
; out of the loop at the bottom and miss one word; hence we add the pad
; (which is the number of words that we didn't write the first time
; through the loop but will decrement for).
	mov	rGLine1, imageWidth
	mov	lastWritten, rGLine1
	add	eax, eax		; since 2 pixels per byte
	add	rGLine1, eax
	mov	loopCount, rGLine1
	jmp	dFloyd4to1ReadSrc[eax*4]

; original source had 01 in the bottom bits
dFloyd4to1LoadSrc1:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+3] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+2]	; read the second byte
	shl	rGLine1, 22		; get first pixel into place
	mov	rGLine1Lower, [ebx+eax] ; get byte from drc table
	mov	ebx, pHDRCTable
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus1 refers the fact that the next source to
; read is srcAddr+1 not that this is one read from the top of the loop.
	jmp	dFloyd4to1SrcPlus1

; original source had 10 in the bottom bits
dFloyd4to1LoadSrc2:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+2] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+1]	; read the second byte
	shl	rGLine1, 22		; get first pixel into place
	mov	rGLine1Lower, [eax+ebx] ; get byte from drc table
	mov	ebx, pHDRCTable
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus0 refers the fact that the next source to
; read is srcAddr+0 not that this is zero reads from the top of the loop.
	jmp	dFloyd4to1SrcPlus0

; original source had 11 in the bottom bits
dFloyd4to1LoadSrc3:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr+1] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+0]	; read the second byte
	shl	rGLine1, 22		; get first pixel into place
	mov	rGLine1Lower, [eax+ebx] ; get lower pixel from drc table
	mov	ebx, pHDRCTable
	add	rSrcAddr, 4		; point to the next source word
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus3 refers the fact that the next source to
; read is srcAddr+3 not that this is three reads from the top of the loop.
	jmp	dFloyd4to1SrcPlus3

; original source had 00 in the bottom bits
dFloyd4to1LoadSrc0:
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	rGLine1Lower, [rSrcAddr-4] ; read the first byte
	mov	rGLine1Lower, [ebx+rGLine1] ; get byte from drc table
	mov	eax, 0			; clear the word since the move won't
	mov	al, [rSrcAddr+3]	; read the second byte
	shl	rGLine1, 22		; get first pixel into place
	mov	rGLine1Lower, [eax+ebx] ; get lower pixel from drc table
	mov	ebx, pHDRCTable
	mov	al, [ebx+eax] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus2 refers the fact that the next source to
; read is srcAddr+2 not that this is two reads from the top of the loop.
;	jmp	dFloyd4to1SrcPlus2


; ***** MAIN LOOP *****

; do 4 source bytes at a time so we now the offset from srcAddr at each
; point in the loop and this way don't have to calculate the byte swapped
; address on the fly.

dFloyd4to1RowLoop:


dFloyd4to1SrcPlus2:
	dFloyd4DoPixel	2, High, pHDRCTable
dFloyd4to1SrcPlus2Low:
	dFloyd4DoPixel	2, Low, pLDRCTable

dFloyd4to1SrcPlus1:
	dFloyd4DoPixel	1, High, pHDRCTable
dFloyd4to1SrcPlus1Low:
	dFloyd4DoPixel	1, Low, pLDRCTable

dFloyd4to1SrcPlus0:
	dFloyd4DoPixel	0, High, pHDRCTable
dFloyd4to1SrcPlus0Low:
	dFloyd4DoPixel	0, Low, pLDRCTable
	add	rSrcAddr, 4		; point to next source word

dFloyd4to1SrcPlus3:
	dFloyd4DoPixel	3, High, pHDRCTable
dFloyd4to1SrcPlus3Low:
	dFloyd4DoPixel	3, Low, pLDRCTable

; we're finished with one column of data.  should we keep going?
	sub	loopCount, 8
	jg	dFloyd4to1RowLoop		; are we at the end yet?

; no, we shouldn't put out anymore dithered data, but we should finish
; filling the gray buffer.  The top line is DBLOCK_HEIGHT 
; pixel ahead of us now.  We need to convert at least  
; DBLOCK_HEIGHT more pixels, then put out all the 
; data in rGLine1.
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf], al		; insert it

	mov	eax, 0			; make sure this is clear
	mov	al, [rSrcAddr]		; get next source pixel
	add	eax, pLDRCTable
	shl	rGLine1, 11		; make room for next pixel
	mov	rGLine1Lower, [eax]

; now write out all the data in rGLine1
	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 22			; 
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf+1], al	; insert it

	mov	eax, rGLine1		; grab next output pixel
	shr	eax, 11			; 
	and	eax, 03FFH		; get rid of high order pixel
	mov	al, dTruncTable[eax]	; truncate it
	mov	[rGrayBuf+2], al	; insert it

	and	rGLine1, 03FFH		; grab last output pixel
	mov	al, dTruncTable[rGLine1] ; truncate it
	mov	[rGrayBuf+3], al	; insert it

; we may have been left with some number of bits in the destination
; register.  Shift those bits into the most significant bits and then
; mask to get only the correct number (the inner loop could have 
; written 0-7 bits too many.
; Unfortunately, it could be that all of the bits in this destination
; are not used and that the *previous* destination word was really the
; last word.  As a result we do not mask the last destination word
; here (in order to clear the right frame) but leave it to the caller.
; (As an example, consider the case of a width of 31 with the bottom 2
; bits of the source address being 01.  Dest word - 1 has 31 valid bits
; and 1 extra bit.  Dest word has 6 extra bits.  We go ahead and shift
; 'cause its too hard to tell when the last dest word in valid.  This
; may write some bogus bits into the frame but no more that 7 so when
; the caller clears the right frame everything will be alright - well
; except for some extra time.)

	mov	ecx, 32			; find shift count
	mov	eax, lastWritten	; number of real bits remaining
					; to be written
	sub	eax, loopCount		; but we overwrote by the negative
					; of loopCount so get the
					; the number of bits really in
					; rDLine0
	jz	dFloyd4to1Leave		; nothing to write
	sub	ecx, eax		; the shift count
	shl	rDLine0, cl		; put into place

dFloyd4to1WriteFinal:
	mov	eax, dstAddr		; get destination address
	not	rDLine0			; invert dithered data
	mov	[eax], rDLine0

dFloyd4to1Leave:
	ret

dFloyd4to1WriteHigh0:
	dFloyd4DoWrite	0, High
dFloyd4to1WriteLow0:
	dFloyd4DoWrite	0, Low

dFloyd4to1WriteHigh1:
	dFloyd4DoWrite	1, High
dFloyd4to1WriteLow1:
	dFloyd4DoWrite	1, Low

dFloyd4to1WriteHigh2:
	dFloyd4DoWrite	2, High
dFloyd4to1WriteLow2:
	dFloyd4DoWrite	2, Low

dFloyd4to1WriteHigh3:
	dFloyd4DoWrite	3, High
dFloyd4to1WriteLow3:
	dFloyd4DoWrite	3, Low


dFloyd4to1LowStart:
; load data from image into rGLine1
	mov	ebx, pHDRCTable
	mov	rSrcAddr, srcAddr	; keep the pointer to the source in a register

; if the original request was to only do a rectangle of the image
; the source address will not be a multiple of 4 bytes.  In order to make
; the inner loop as fast as possible, it is assumed that the offset 
; for the next source address is a well known constant 0 through 4.
; This means that we must enter the loop at different place (and get
; the first 3 source words from different offsets).
;
; least sig.	offsets to read		enter loop this
; 2 bits of	(after addr masked)	# of reads from
; source	to init rGLine1		the top
; 00		+3, +2			0 (srcAddr+1)
; 01		+2, +1			1 (srcAddr+0) (inc src after this)
; 10		+1, +0			2 (srcAddr+3)
; 11		+0, (inc addr) +3	3 (srcAddr+2)

	mov	eax, rSrcAddr		; get the bottom 2 bits  
	and	eax, 3			; 
	and	rSrcAddr, 0FFFFFFFCH	; src addr is now a mult of 4 bytes

; compute the number of times we go through the loop
; the number of times that we go through the loop is the width of the
; image plus some pad (the distance from the top of the loop) to make
; sure that we process enough bits.  Why the pad?  We always want to
; decrement count by 8 at the bottom of the loop (if we don't we would 
; have to decrement the count and possibly jump after each souce word).
; Without the pad, if the loop wasn't entered at the top we might not
; do enough source words.  Consider souce address with 11 in the bottom
; bits and a width of 2.  If we just used the width, we would jump
; out of the loop at the bottom and miss one word; hence we add the pad
; (which is the number of words that we didn't write the first time
; through the loop but will decrement for).
	mov	rGLine1, imageWidth
	mov	lastWritten, rGLine1
	add	eax, eax		; since 2 pixels per byte
	inc	eax			; since the first pixel we use
					; is the low pixel
	add	rGLine1, eax
	mov	loopCount, rGLine1
	jmp	dFloyd4to1ReadLowSrc[eax*4]

; original source had 01 in the bottom bits
dFloyd4to1LoadLowSrc1:
	mov	eax, 0			; clear the word since the move won't
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	al, [rSrcAddr+2] 	; read the first byte
	mov	rGLine1Lower, [eax+ebx] ; get byte from drc table
	shl	rGLine1, 22		; get first pixel into place
	mov	al, [rSrcAddr+1]	; read the second byte
	mov	rGLine1Lower, [eax+ebx] ; get byte from drc table
	mov	ebx, pLDRCTable
	mov	al, [rSrcAddr+2] 	; read the first byte again
	mov	al, [eax+ebx] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus1 refers the fact that the next source to
; read is srcAddr+1 not that this is one read from the top of the loop.
	jmp	dFloyd4to1SrcPlus1Low

; original source had 10 in the bottom bits
dFloyd4to1LoadLowSrc2:
	mov	eax, 0			; clear the word since the move won't
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	al, [rSrcAddr+1] 	; read the first byte
	mov	rGLine1Lower, [eax+ebx] ; get byte from drc table
	shl	rGLine1, 22		; get first pixel into place
	mov	al, [rSrcAddr+0]	; read the second byte
	mov	rGLine1Lower, [eax+ebx] ; get byte from drc table
	mov	ebx, pLDRCTable
	mov	al, [rSrcAddr+1] 	; read the first byte again
	add	rSrcAddr, 4		; point to the next source word
	mov	al, [eax+ebx] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus0 refers the fact that the next source to
; read is srcAddr+0 not that this is zero reads from the top of the loop.
	jmp	dFloyd4to1SrcPlus0Low

; original source had 11 in the bottom bits
dFloyd4to1LoadLowSrc3:
	mov	eax, 0			; clear the word since the move won't
	mov	rGLine1, 0		; make sure the entire word is 0
	mov	al, [rSrcAddr+0] 	; read the first byte
	mov	rGLine1Lower, [eax+ebx] ; get byte from drc table
	shl	rGLine1, 22		; get first pixel into place
	mov	al, [rSrcAddr+7]	; read the second byte
	mov	rGLine1Lower, [eax+ebx] ; get lower pixel from drc table
	mov	ebx, pLDRCTable
	mov	al, [rSrcAddr+0] 	; read the first byte again
	add	rSrcAddr, 4		; point to the next source word
	mov	al, [eax+ebx] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus3 refers the fact that the next source to
; read is srcAddr+3 not that this is three reads from the top of the loop.
	jmp	dFloyd4to1SrcPlus3Low

; original source had 00 in the bottom bits
dFloyd4to1LoadLowSrc0:
	mov	eax, 0			; make sure the entire word is 0
	mov	rGLine1, 0		; clear the word since the move won't
	mov	al, [rSrcAddr+3] 	; read the first byte
	mov	rGLine1Lower, [eax+ebx] ; get byte from drc table
	shl	rGLine1, 22		; get first pixel into place
	mov	al, [rSrcAddr+2]	; read the second byte
	mov	rGLine1Lower, [eax+ebx] ; get lower pixel from drc table
	mov	ebx, pLDRCTable
	mov	al, [rSrcAddr+3] 	; read the first byte again
	mov	al, [eax+ebx] 		; get byte from drc table
	shl	eax, 11
	or 	rGLine1, eax		; first and second bytes

; initialize destination word by inserting the marker bit
; when the marker bit falls into the carry, it's time to write the
; destination word.
	mov	rDLine0, 1

; unlike the 8 to 1 code Plus2 refers the fact that the next source to
; read is srcAddr+2 not that this is two reads from the top of the loop.
	jmp	dFloyd4to1SrcPlus2Low


ip_ditherFloydLine4to1 ENDP


_TEXT	ENDS				; End the segment

	PUBLIC	vdErrorTable
	PUBLIC	hdErrorTable
	PUBLIC	vdErrorTable2
	PUBLIC	hdErrorTable2
	PUBLIC	dTruncTable
_DATA	SEGMENT PUBLIC 'DATA'
dFloydReadSrc	DWORD	OFFSET dFloydLoadSrc0, OFFSET dFloydLoadSrc1, OFFSET dFloydLoadSrc2, OFFSET dFloydLoadSrc3
dFloyd2ReadSrc	DWORD	OFFSET dFloyd2LoadSrc0, OFFSET dFloyd2LoadSrc1, OFFSET dFloyd2LoadSrc2, OFFSET dFloyd2LoadSrc3
dFloyd4to1ReadSrc DWORD	OFFSET dFloyd4to1LoadSrc0, OFFSET dFloyd4to1LoadSrc0, OFFSET dFloyd4to1LoadSrc1, OFFSET dFloyd4to1LoadSrc1, OFFSET dFloyd4to1LoadSrc2, OFFSET dFloyd4to1LoadSrc2, OFFSET dFloyd4to1LoadSrc3, OFFSET dFloyd4to1LoadSrc3
dFloyd4to1ReadLowSrc DWORD	OFFSET dFloyd4to1LoadLowSrc0, OFFSET dFloyd4to1LoadLowSrc0, OFFSET dFloyd4to1LoadLowSrc1, OFFSET dFloyd4to1LoadLowSrc1, OFFSET dFloyd4to1LoadLowSrc2, OFFSET dFloyd4to1LoadLowSrc2, OFFSET dFloyd4to1LoadLowSrc3, OFFSET dFloyd4to1LoadLowSrc3

dFloydWriteMask	DWORD	000000000H, 080000000H, 0C0000000H, 0E0000000H, 0F0000000H, 0F8000000H, 0FC000000H, 0FE000000H, 0FF000000H, 0FF800000H, 0FFC00000H, 0FFE00000H, 0FFF00000H, 0FFF80000H, 0FFFC0000H, 0FFFE0000H, 0FFFF0000H, 0FFFF8000H, 0FFFFC000H, 0FFFFE000H, 0FFFFF000H, 0FFFFF800H, 0FFFFFC00H, 0FFFFFE00H, 0FFFFFF00H, 0FFFFFF80H, 0FFFFFFC0H, 0FFFFFFE0H, 0FFFFFFF0H, 0FFFFFFF8H, 0FFFFFFFCH, 0FFFFFFFEH
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
vdErrorTable	DWORD	1024 DUP (?)
hdErrorTable	DWORD	1024 DUP (?)
dTruncTable	BYTE	1024 DUP (?)
vdErrorTable2	DWORD	1024 DUP (?)
hdErrorTable2	DWORD	1024 DUP (?)
_BSS	ENDS

	END


IF 0				; These are all comments

COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcfloyd.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.14  1998/01/13 23:58:11  ddavies
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.13  1996/09/17 00:45:34  ddavies
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.12  1996/09/16 00:50:51  ddavies
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.11  1995/09/15 21:47:45  ddavies
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

Revision 1.10  1994/08/25  17:39:18  ejaquith
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

Added extra redundant parameter, blockHt, to the ditherFloydLine routines,
which is needed in the C/slow code versions.

Revision 1.9  1994/08/22  18:18:51  ejaquith
ejaquith on Mon Aug 22 11:18:05 PDT 1994

pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

Added mov instructions to avoid compiler warnings.

Revision 1.8  1994/08/09  17:25:48  ejaquith
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

Added procedures ip_ditherFloydSingleLine and ip_ditherFloydSingleLine4to1
which in turn call ip_ditherFloydLine and ip_ditherFloydLine4ot1
respectively, as these two latter procedures _are_ single line dithering
procedures!

Revision 1.7  1994/07/27  15:42:31  ejaquith
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

Reentrancy!
Removed globals and made functions be called with pointers to error
diffusion tables instead.

Revision 1.6  1993/11/05  01:52:15  ddavies
pcfloyd.asm -> /product/ipcore/ipshared/src/RCS/pcfloyd.asm,v

Add RCS magic to have string specifying this file's name, rev number and
last edit date into the object file.  Also add change log at the bottom.

^

ENDIF

