;*****************************************************************
;*  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcglyph.asm:  grayscale thresholding (for 600dpi images with glyphs).

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small

_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcglyph.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment:
; EAX:	scratch
; EBX:	scratch 
; ECX:	current threshold
; EDX:	current source word
; EBP:	points to variables on the stack
; ESI:	source address
; EDI:	destination word
; ESP:	stack pointer

rSrcAddr	EQU esi
rOWord		EQU edi
rCurrThresh	EQU edx
rCurrThreshByte EQU dl
rCurr		EQU ecx
rCurrByteHi	EQU ch
rCurrByteLo	EQU cl

; Constant definitions:
cMSBitMask	EQU 080808080H
cnMSBitMask	EQU 07F7F7F7FH
cMS6BitMask	EQU 0FCFCFCFCH
cByteMask	EQU 0FFH

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_gThreshLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will process the image from top to bottom.
; We will move from right to left within each line of the image.

; Parameters:
; srcAddr:	address of left end of current source line
; srcWidth	source width
; sBpl		source bytes per line
; lastSrcAddr:	address of word after last source word referenced
; dstAddr:	address of left end of current dest line
; lastWordMask:	mask for last destination word
; repThreshTable: table with thresholds repeated in all 4 bytes
; thresh:	current threshold
;
; Returns modified threshold

ip_glyphThreshLineStart::
ip_glyphThreshLine	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	sBpl:DWORD, lastSrcAddr:DWORD,  \
	dstAddr:DWORD, lastWordMask:DWORD, repThreshTable:DWORD, \
	thresh:BYTE
LOCAL	next:DWORD, prev:DWORD, sBpl2:DWORD, sOffset:DWORD

; make assembler happy
	mov	eax, pIPState

; get parameters into registers
	mov	rSrcAddr, srcAddr
	xor	rCurrThresh, rCurrThresh	; convert current threshold
	mov	rCurrThreshByte, thresh		; into a 32-bit quantity

; initialize loop variables
	xor	rOWord, rOWord			; clear destination word
	mov	rCurr, [rSrcAddr]		; get first source word
	and	rCurr, cMS6BitMask		; get rid of the least
						; sig. 2 bits.
	mov	prev, rCurr			; use it as the previous
						; source word as well
	mov	eax, [rSrcAddr+4]		; the next source word
	and	eax, cMS6BitMask		; get rid of the least
	mov	next, eax			; sig. 2 bits.
	mov	eax, sBpl			; keep 2*sBpl for use
	add	eax, eax			; when getting the words
	mov	sBpl2, eax			; above and below the 
						; current source word
	mov	sOffset, 1

gThreshLoop:
	cmp	rSrcAddr, lastSrcAddr		; are we done with the
	jge	gThreshMaskLastWord		; line yet?

; check to see if any of the bytes in the source word are greater than 192.
; a source byte is greater than 192 if the 2 most significant bits are
; on so and the 2 ms bits together and check to see if any are on.
	lea	eax, [rCurr*2]			; curr << 1
	and	eax, rCurr			; curr & (curr << 1)
	and	eax, cMSBitMask			; are the ms bits on?
	jne	gThreshDoNotAdjust

; all of the source bytes are less than 192 so adjust the threshold.
	mov	ebx, sBpl2			; offset to other lines 
	mov	eax, [rSrcAddr+ebx]		; get word below current
	neg	ebx				; get word above current
	and	eax, cMS6BitMask		; get rid of the least 2
	shr	eax, 2				; sig bits
	mov	ebx, [rSrcAddr+ebx]
	and	ebx, cMS6BitMask		; get rid of the least 2
	shr	ebx, 2				; sig bits
; compute the average of the surrounding pixels: 
; (prev + next + above + below) /4.  ignore the 2 least significant bits
; of each pixel so that we can compute the average of the each of the
; current source pixels at one time.  each source pixel is actually
; divided by 4 before being added together.
	add	eax, ebx			; above+below
	mov	ebx, prev			; add the prev pixel
	shl	ebx, 22				; into the most significant
	add	eax, ebx			; pixel 
	mov	ebx, next			; add the next pixel
	shr	ebx, 26				; into the least
	add 	eax, ebx			; significant pixel
	mov	ebx, rCurr			; the previous pixels for
	shr	ebx, 10				; the 3 most sig pixels
	add 	eax, ebx			; are in the current word
	mov	ebx, rCurr			; the next pixels for
	shl	ebx, 6				; the 3 least sig pixels
	add	eax, ebx			; are in the current word

; we have the background average.  now adjust the threshold for each
; pixel.
	add	rOWord, rOWord			; get ready for the next
						; dest pixel.
	mov	ebx, eax			; shift ms avg pixel into
	shr	ebx, 24				; ls byte
	sub	ebx, rCurrThresh		; distant to curr thresh
	sar	ebx, 2				; go one quarter the
	add	rCurrThresh, ebx		; distance
	mov	ebx, rCurr			; threshold the ms pixel
	shr	ebx, 24
	sub	ebx, rCurrThresh
	jge	gThreshWhite1
	or	rOWord, 1			; black pixel
gThreshWhite1:
	add	rOWord, rOWord			; get ready for the next
						; dest pixel.
	mov	ebx, eax			; shift next avg pixel into
	shr	ebx, 16				; ls byte
	and	ebx, cByteMask
	sub	ebx, rCurrThresh		; distant to curr thresh
	sar	ebx, 2				; go one quarter the
	add	rCurrThresh, ebx		; distance
	mov	ebx, rCurr			; threshold the pixel
	shr	ebx, 16
	and	ebx, cByteMask
	sub	ebx, rCurrThresh
	jge	gThreshWhite2
	or	rOWord, 1			; black pixel
gThreshWhite2:
	add	rOWord, rOWord			; get ready for the next
						; dest pixel.
	mov	bl, ah				; shift next avg pixel into
	and	ebx, cByteMask			; ls byte
	sub	ebx, rCurrThresh		; distant to curr thresh
	sar	ebx, 2				; go one quarter the
	add	rCurrThresh, ebx		; distance
	mov	bl, rCurrByteHi			; threshold the pixel
	and	ebx, cByteMask
	sub	ebx, rCurrThresh
	jge	gThreshWhite3
	or	rOWord, 1			; black pixel
gThreshWhite3:
	add	rOWord, rOWord			; get ready for the next
						; dest pixel.
	and	eax, cByteMask			; get ls byte
	sub	eax, rCurrThresh		; distant to curr thresh
	sar	eax, 2				; go one quarter the
	add	rCurrThresh, eax		; distance
	mov	ebx, rCurr			; threshold the pixel
	and	ebx, cByteMask
	sub	ebx, rCurrThresh
	jge	gThreshAdjustWrite
	or	rOWord, 1			; black pixel

gThreshAdjustWrite:
	mov	eax, sOffset			; is it time to write
	test	eax, 7				; a destinaion word yet?
	jnz	gThreshAdjustNextWord
	mov	ebx, dstAddr			; yes, write it
	add	ebx, 4				; and update the dest
	mov	dstAddr, ebx			; addr
	mov	[ebx-4], rOWord
gThreshAdjustNextWord:
	inc	eax				; update sOffset
	mov	sOffset, eax
	add	rSrcAddr, 4			; next source word
	mov	prev, rCurr			; prev = curr
	mov	rCurr, next			; curr = next
	mov	eax, [rSrcAddr+4]		; get next source word
	and	eax, cMS6BitMask
	mov	next, eax
	jmp	gThreshLoop


; at least one of the source bytes is greater than 192 so do not adjust
; the threshold.
gThreshDoNotAdjust:
	mov	eax, repThreshTable		; get the threshold
	mov	ebx, [rCurrThresh*4][eax]	; repeated in 4 bytes
; in order to do all the bytes in the source word at once, we need the
; threshold repeated in each of the bytes and to ignore the
; least sig bit of the current threshold.  since the current pixels
; have had the 2 least sig bits removed this shouldn't matter.  note that
; the entries in the table have a 1 in their most sig bits so that the 
; subtract won't bother the previous pixel
	mov	eax, rCurr			; ingore least sig bit
	shr	eax, 1				; of curr pixels
	sub	ebx, eax			; the dest pixels are the
						; ms bit of each byte

	shl	rOWord, 4			; get ready for 4 pixels
	mov	eax, ebx
	shr	eax, 28				; the ms pixel
	and	eax, 8				; into position
	or	rOWord, eax			; and into the output
	mov	eax, ebx			; get the next
	shr	eax, 21				; destination pixel
	and	eax, 4				; into position
	or	rOWord, eax			; and into the output
	mov	eax, ebx			; get the next
	shr	eax, 14				; destination pixel
	and	eax, 2				; into position
	or	rOWord, eax			; and into the output
	shr	ebx, 7				; the last pixel
	and	ebx, 1
	or	rOWord, ebx
	jmp	gThreshAdjustWrite

; we have written all the destination words.  the last word written
; may have extras bits in the frame so we need to mask them out

gThreshMaskLastWord:
	mov	eax, srcWidth
	and	eax, 31
	jnz	gThreshWriteLastWord
	mov	eax, dstAddr
	mov	ebx, [eax-4]
	and	ebx, lastWordMask
	mov	[eax-4], ebx
	mov	eax, rCurrThresh
	ret

; if we had only put the correct number of destination bits into rOWord,
; the shift would be 32 - (w & 0x1F).  however, we put a multiple of
; 4 bits into the destination word
gThreshWriteLastWord:
	add	eax, 3			; round up to find the number of
	and	eax, 31			; bits actually in the dest reg
	mov	ecx, 32	
	sub	ecx, eax		; get the shift count
	shl	rOWord, cl
	mov	ebx, dstAddr
	and	rOWord, lastWordMask
	mov	[ebx], rOWord
	mov	eax, rCurrThresh
	ret
ip_glyphThreshLine ENDP

_TEXT   ENDS                            ; End the segment

_BSS    SEGMENT PUBLIC 'BSS'
_BSS    ENDS

	END

IF 0


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcglyph.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.5  1998/01/13 23:58:13  ddavies
pcglyph.asm -> /product/ipcore/ipshared/src/RCS/pcglyph.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.4  1996/09/17 00:45:37  ddavies
pcglyph.asm -> /product/ipcore/ipshared/src/RCS/pcglyph.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.3  1996/09/16 00:51:05  ddavies
pcglyph.asm -> /product/ipcore/ipshared/src/RCS/pcglyph.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.2  1995/09/15 21:47:49  ddavies
pcglyph.asm -> /product/ipcore/ipshared/src/RCS/pcglyph.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

;; Revision 1.1  1995/03/28  00:03:24  jcraig
;; Initial revision
;;

^

ENDIF
