;*****************************************************************
;*  Copyright (c) 1997, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcgrybnm.asm:  gray selection routines using MMX opcodes.

	.586		; compile code for a Pentium processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; This code uses MMX opcodes, which have 64-bit accesses.  That's the
; main source of our speedup.  These routines are only called if
; the caller detects MMX code.  We don't have to detect it here.
; First, include the file that gives us the MMX macros.

INCLUDE	iammx.inc

_DATA	SEGMENT PUBLIC 'DATA'

RCSInfo	BYTE	"$RCSfile: pcgrybnm.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS


; First, we do the four routines used for binary selectors.  The basic
; scheme in these is to read a byte from the selector image, use a mask
; to expand it into 8 bytes (each bit turns into a byte), and use that
; mask to blend the foreground and background pixels together.
; To maintain reasonable use of the pipeline, we process four groups of
; 8 bytes (32 in all) in parallel.  These groups are horizontally
; adjacent.  Doing two groups that are vertically adjacent uses too
; many registers for all the bytes/line values.

;Register Assignment:
; EAX	holds selector bytes
; EBX	holds selector bytes
; ECX	pointer to foreground gray image
; EDX	pointer to background gray image
; EBP	pts to variables on the stack
; ESI	points to selector image
; EDI	points to dest image
; ESP	stack pointer

; MM0	foreground, lower address
; MM1	background, lower address
; MM2	foreground, upper address (lower + 8)
; MM3	background, upper address
; MM4	
; MM5	
; MM6	
; MM7	holds all 1's for inverting selector mask.

rFore		EQU	ecx
rBack		EQU	edx
rSel		EQU	esi
rDest		EQU	edi

; Parameters
; pBack		addr of left end of source background line
; pFore		addr of left end of source foreground line
; pSel		addr of left end of source selector line
; pDest		addr of left end of dest line
; imgWidth	width of source line in pixels (bytes in this case)


; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_grayBinarySelLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_grayBinarySelLineLongStart::
ip_grayBinarySelLineLong  PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, pBack:DWORD, pFore:DWORD, pDest:DWORD, pSel:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 32 bytes, so we'll point to the last multiple of 32
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	mov	rBack, pBack
	dec	eax			; really want (width - 1)

	mov	rSel, pSel
	and	eax, NOT 01Fh		; get ((width - 1)/32) * 32

	mov	rFore, pFore
	add	rDest, eax		; point to last dest ddword

	add	rBack, eax		; point to last background ddword
	add	rFore, eax		; point to last foreground ddword

	shr	eax, 3			; get ((width - 1)/32) * 4
	pcmpeqb	mm7, mm7		; mm7 <- all 1's for inversion

	add	rSel, eax		; point to last selector word
	pxor	mm3, mm3		; prepare for first write to frame


; In this loop, we process four selector bytes at a time.  We only
; have to test for the end of the loop (beginning of the line) once
; in each iteration.
; In the first iteration, we may write a word of 0's into the pad of
; this scan line or the left frame of the next scanline.  Big deal.
; At the top of the loop, we test a full word of selector to see if its
; filled with 0's.  If so, we just move 4*8 background bytes to the
; dest.
grayBinSelLoop:
	test	DWORD PTR [pSel], 0FFFFFFFFh
	jnz	grayBinSelLongWay	; if there are any foreground bits,
					; do it the long way.

; The entire selector is filled with 0's, so move 32 bytes from
; background to foreground.

	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty				; write mm3 from last loop iteration

	movq	mm1, [pBack]		; get 1st background dword
	; empty

	movq	mm3, [pBack + 8]	; get 2nd background dword
	; empty

	movq	[rDest], mm1		; write it
	; empty

	movq	[rDest + 8], mm3	; write it
	; empty

	movq	mm1, [rBack + 16]	; get 3rd background dword
	; empty

	movq	mm3, [rBack + 24]	; get 4th background dword
	; empty

	movq	[rDest + 16], mm1	; write it
	; empty

	sub	rFore, 32		; adjust foreground ptr
	sub	rBack, 32		; adjust background ptr

	sub	rDest, 32		; adjust dest ptr
	sub	rSel, 4			; adjust selector ptr

	cmp	rSel, pSel		; back at beginning of line yet?
	jae	grayBinSelLoop		; nope, loop around

; We're done, so exit.
	; empty
	jmp	finGrayBinSel


; Here's where we do the loop the long way.  Start with the first
; two selector bytes.  These are the ones on the left edge of the word.

grayBinSelLongWay:
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 3]		; read eax without delay on P6
	mov	bl, [rSel + 2]

	movq	mm0, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	;empty

	movq	mm2, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	movq	mm1, mm0		; copy expanded selector A for invert

	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	movq	mm3, mm2		; copy expanded selector B for invert

	pand	mm0, [rBack]		; AND selector A with background
	pxor	mm1, mm7		; invert selector A

	pand	mm2, [rBack + 8]	; AND selector B with background
	pxor	mm3, mm7		; invert selector B

	pand	mm1, [rFore]		; AND inverted selector A with foregnd
	; empty

	pand	mm3, [rFore + 8]	; AND inverted selector B with foregnd
	por	mm1, mm0		; combine A foreground and background

; process two remaining selector bytes in the selector word
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 1]		; read eax without delay on P6
	mov	bl, [rSel + 0]

	movq	[rDest], mm1		; store combined A's after cycle delay
	por	mm3, mm2		; combine B foreground and background

	movq	mm0, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	; empty

	movq	mm2, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	movq	mm1, mm0		; copy expanded selector A for invert

	movq	[rDest + 8], mm3	; write B from prev loop iteration
	movq	mm3, mm2		; copy expanded selector B for invert

	pand	mm0, [rBack + 16]	; AND selector A with background
	pxor	mm1, mm7		; invert selector A

	pand	mm2, [rBack + 24]	; AND selector B with background
	pxor	mm3, mm7		; invert selector B

	pand	mm1, [rFore + 16]	; AND inverted selector A with foregnd
	; empty

	pand	mm3, [rFore + 24]	; AND inverted selector B with foregnd
	por	mm1, mm0		; combine A foreground and background

	sub	rFore, 32		; adjust foreground ptr
	sub	rBack, 32		; adjust background ptr

	movq	[rDest + 16], mm1	; store combined A's
	por	mm3, mm2		; combine B foreground and background

	sub	rDest, 32		; adjust dest ptr
	sub	rSel, 4			; adjust selector ptr

	cmp	rSel, pSel		; back at beginning of line yet?
	jae	grayBinSelLoop		; nope, loop around

; We're done with the loop.  Write the last dword and return
finGrayBinSel:
	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration

	ret

ip_grayBinarySelLineLong  ENDP




; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_grayBinarySelLineLongNoFG.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_grayBinarySelLineLongNoFGStart::
ip_grayBinarySelLineLongNoFG  PROC NEAR C PUBLIC USES EAX EBX EDX ESI EDI, \
	pIPState:DWORD, pBack:DWORD, pDest:DWORD, pSel:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 32 bytes, so we'll point to the last multiple of 32
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	mov	rBack, pBack
	dec	eax			; really want (width - 1)

	mov	rSel, pSel
	and	eax, NOT 01Fh		; get ((width - 1)/32) * 32

	add	rDest, eax		; point to last dest ddword
	add	rBack, eax		; point to last background ddword

	shr	eax, 3			; get ((width - 1)/32) * 4
	; empty

	add	rSel, eax		; point to last selector word
	pxor	mm3, mm3		; prepare for first write to frame


; In this loop, we process four selector bytes at a time.  We only
; have to test for the end of the loop (beginning of the line) once
; in each iteration.
; In the first iteration, we may write a word of 0's into the pad of
; this scan line or the left frame of the next scanline.  Big deal.
; At the top of the loop, we test a full word of selector to see if its
; filled with 0's.  If so, we just move 4*8 background bytes to the
; dest.
grayBinSelNoFGLoop:
	test	DWORD PTR [pSel], 0FFFFFFFFh
	jnz	grayBinSelNoFGLongWay	; if there are any foreground bits,
					; do it the long way.

; The entire selector is filled with 0's, so move 32 bytes from
; background to foreground.

	movq	[rDest + 24 + 32], mm3	; write 4th from prev loop iteration
	; empty				; write mm3 from last loop iteration

	movq	mm1, [rBack]		; get 1st background dword
	; empty

	movq	mm3, [rBack + 8]	; get 2nd background dword
	; empty

	movq	[rDest], mm1		; write 1st
	; empty

	movq	[rDest + 8], mm3	; write 2nd
	; empty

	movq	mm1, [rBack + 16]	; get 3rd background dword
	; empty

	movq	mm3, [rBack + 24]	; get 4th background dword
	; empty

	movq	[rDest + 16], mm1	; write 3rd
	; empty

	sub	rBack, 32		; adjust background ptr
	sub	rSel, 4			; adjust selector ptr

	sub	rDest, 32		; adjust dest ptr
	cmp	rSel, pSel		; back at beginning of line yet?

	; empty
	jae	grayBinSelNoFGLoop	; nope, loop around

; We're done, so exit.
	; empty
	jmp	finGrayBinSelNoFG


; Here's where we do the loop the long way.  Start with the first
; two selector bytes.  These are the ones on the left edge of the word.

grayBinSelNoFGLongWay:
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 3]		; read eax without delay on P6
	mov	bl, [rSel + 2]

	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty

	movq	mm1, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	;empty

	movq	mm3, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	; empty

	pand	mm1, [rBack]		; AND selector A with backgnd
	; empty

	pand	mm3, [rBack + 8]	; AND selector B with backgnd
	; empty


; process two remaining selector bytes in the selector word
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 1]		; read eax without delay on P6
	mov	bl, [rSel + 0]

	movq	[rDest], mm1		; store combined A's
	; empty

	movq	[rDest + 8], mm3	; store combined B's
	; empty

	movq	mm1, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	;empty

	movq	mm3, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	; empty

	pand	mm1, [rBack + 16]	; AND selector A with backgnd
	; empty

	pand	mm3, [rBack + 24]	; AND selector B with backgnd
	; empty

	movq	[rDest + 16], mm1	; store combined A's
	; empty

	sub	rBack, 32		; adjust background ptr
	sub	rSel, 4			; adjust selector ptr

	sub	rDest, 32		; adjust dest ptr
	cmp	rSel, pSel		; back at beginning of line yet?

	; empty
	jae	grayBinSelNoFGLoop	; nope, loop around

; We're done with the loop.  Write the last dword and return
finGrayBinSelNoFG:
	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty

	ret

ip_grayBinarySelLineLongNoFG  ENDP



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_grayBinarySelLineLongNoBG.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_grayBinarySelLineLongNoBGStart::
ip_grayBinarySelLineLongNoBG  PROC NEAR C PUBLIC USES EAX EBX ECX ESI EDI, \
	pIPState:DWORD, pFore:DWORD, pDest:DWORD, pSel:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 32 bytes, so we'll point to the last multiple of 32
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	mov	rFore, pFore
	dec	eax			; really want (width - 1)

	mov	rSel, pSel
	and	eax, NOT 01Fh		; get ((width - 1)/32) * 32

	add	rDest, eax		; point to last dest ddword
	add	rFore, eax		; point to last foreground ddword

	shr	eax, 3			; get ((width - 1)/32) * 4
	pcmpeqb	mm7, mm7		; mm7 <- all 1's for inversion

	add	rSel, eax		; point to last selector word
	pxor	mm3, mm3		; prepare for first write to frame


; In this loop, we process four selector bytes at a time.  We only
; have to test for the end of the loop (beginning of the line) once
; in each iteration.
; In the first iteration, we may write a word of 0's into the pad of
; this scan line or the left frame of the next scanline.  Big deal.
; At the top of the loop, we test a full word of selector to see if its
; filled with 0's.  If so, we just move 4*8 background bytes to the
; dest.
grayBinSelNoBGLoop:
	test	DWORD PTR [pSel], 0FFFFFFFFh
	jnz	grayBinSelNoBGLongWay	; if there are any foreground bits,
					; do it the long way.

; The entire selector is filled with 0's, so move 32 bytes from
; background to foreground.

	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty				; write mm3 from last loop iteration

	movq	[rDest], mm7		; write all 1's (white)
	; empty

	movq	[rDest + 8], mm7	; write all 1's
	; empty

	movq	[rDest + 16], mm7	; write all 1's
	movq	mm3, mm7		; prepare for write in next loop

	sub	rFore, 32		; adjust foreground ptr
	sub	rSel, 4			; adjust selector ptr

	sub	rDest, 32		; adjust dest ptr
	cmp	rSel, pSel		; back at beginning of line yet?

	; empty
	jae	grayBinSelNoBGLoop	; nope, loop around

; We're done, so exit.
	; empty
	jmp	finGrayBinSelNoBG


; Here's where we do the loop the long way.  Start with the first
; two selector bytes.  These are the ones on the left edge of the word.

grayBinSelNoBGLongWay:
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 3]		; read eax without delay on P6
	mov	bl, [rSel + 2]

	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty

	movq	mm1, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	;empty

	movq	mm3, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	; empty

	por	mm1, [rFore]		; OR selector A with foregnd
	; empty

	por	mm3, [rFore + 8]	; OR selector B with foregnd
	; empty


; process two remaining selector bytes in the selector word
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 1]		; read eax without delay on P6
	mov	bl, [rSel + 0]

	movq	[rDest], mm1		; store combined A's
	; empty

	movq	[rDest + 8], mm3	; store combined B's
	; empty

	movq	mm1, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	;empty

	movq	mm3, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	; empty

	por	mm1, [rFore + 16]	; OR selector A with foregnd
	; empty

	por	mm3, [rFore + 24]	; OR selector B with foregnd
	; empty

	movq	[rDest + 16], mm1	; store combined A's
	; empty

	sub	rFore, 32		; adjust foreground ptr
	sub	rSel, 4			; adjust selector ptr

	sub	rDest, 32		; adjust dest ptr
	cmp	rSel, pSel		; back at beginning of line yet?

	; empty
	jae	grayBinSelNoBGLoop	; nope, loop around

; We're done with the loop.  Write the last dword and return
finGrayBinSelNoBG:
	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty

	ret

ip_grayBinarySelLineLongNoBG  ENDP



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_grayBinarySelLineLongNoFGNoBG.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_grayBinarySelLineLongNoFGNoBGStart::
ip_grayBinarySelLineLongNoFGNoBG  PROC NEAR C PUBLIC USES EAX EBX ESI EDI, \
	pIPState:DWORD, pDest:DWORD, pSel:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 32 bytes, so we'll point to the last multiple of 32
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	mov	rSel, pSel
	dec	eax			; really want (width - 1)

	and	eax, NOT 01Fh		; get ((width - 1)/32) * 32
	; empty

	add	rDest, eax		; point to last dest ddword
	shr	eax, 3			; get ((width - 1)/32) * 4

	pcmpeqb	mm7, mm7		; mm7 <- all 1's for inversion
	add	rSel, eax		; point to last selector word

	pxor	mm3, mm3		; prepare for first write to frame
	; empty


; In this loop, we process four selector bytes at a time.  We only
; have to test for the end of the loop (beginning of the line) once
; in each iteration.
; In the first iteration, we may write a word of 0's into the pad of
; this scan line or the left frame of the next scanline.  Big deal.
; At the top of the loop, we test a full word of selector to see if its
; filled with 0's.  If so, we just move 4*8 background bytes to the
; dest.
grayBinSelNoFGNoBGLoop:
	test	DWORD PTR [pSel], 0FFFFFFFFh
	jnz	grayBinSelNoFGNoBGLongWay	; if there are any foreground bits,
					; do it the long way.

; The entire selector is filled with 0's, so move 32 bytes from
; background to foreground.

	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty				; write mm3 from last loop iteration

	movq	[rDest], mm7		; write all 1's (white)
	; empty

	movq	[rDest + 8], mm7	; write all 1's
	; empty

	movq	[rDest + 16], mm7	; write all 1's
	movq	mm3, mm7		; prepare for write in next loop

	sub	rDest, 32		; adjust dest ptr
	sub	rSel, 4			; adjust selector ptr

	cmp	rSel, pSel		; back at beginning of line yet?
	jae	grayBinSelNoFGNoBGLoop	; nope, loop around

; We're done, so exit.
	; empty
	jmp	finGrayBinSelNoFGNoBG


; Here's where we do the loop the long way.  Start with the first
; two selector bytes.  These are the ones on the left edge of the word.

grayBinSelNoFGNoBGLongWay:
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 3]		; read eax without delay on P6
	mov	bl, [rSel + 2]

	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty

	movq	mm1, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	;empty

	movq	mm3, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	; empty


; process two remaining selector bytes in the selector word
	xor	eax, eax		; this is how you write al and
	xor	ebx, ebx		; read second selector byte

	mov	al, [rSel + 1]		; read eax without delay on P6
	mov	bl, [rSel + 0]

	movq	[rDest], mm1		; store combined A's
	; empty

	movq	[rDest + 8], mm3	; store combined B's
	; empty

	movq	mm1, MMWORD PTR binExp8Table[eax*8] ; load expanded selector A
	;empty

	movq	mm3, MMWORD PTR binExp8Table[ebx*8] ; load expanded selector B
	; empty

	movq	[rDest + 16], mm1	; store combined A's
	; empty

	sub	rDest, 32		; adjust dest ptr
	sub	rSel, 4			; adjust selector ptr

	cmp	rSel, pSel		; back at beginning of line yet?
	jae	grayBinSelNoFGNoBGLoop	; nope, loop around

; We're done with the loop.  Write the last dword and return
finGrayBinSelNoFGNoBG:
	movq	[rDest + 24 + 32], mm3	; write B from prev loop iteration
	; empty

	ret

ip_grayBinarySelLineLongNoFGNoBG  ENDP



; -----------------------------------------------------------------------
; ------------------------- 4 bit Selectors -----------------------------
; -----------------------------------------------------------------------

;Register Assignment:
; EAX	holds selector bytes
; EBX	holds selector bytes
; ECX	pointer to foreground gray image
; EDX	pointer to background gray image
; EBP	pts to variables on the stack
; ESI	points to selector image
; EDI	points to dest image
; ESP	stack pointer

; MM0	foreground, lower word
; MM1	background, lower word
; MM2	foreground, upper word (lower + 4)
; MM3	background, upper word and dest register
; MM4	selector
; MM5	
; MM6	holds all 0's for unpacking
; MM7	holds mask for inverting selector mask.



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_gray4bppSelLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_gray4bppSelLineLongStart::
ip_gray4bppSelLineLong  PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, pBack:DWORD, pFore:DWORD, pSel:DWORD, pDest:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 8 bytes, so we'll point to the last multiple of 8
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	mov	rBack, pBack
	dec	eax			; really want (width - 1)

	mov	rSel, pSel
	and	eax, NOT 07h		; get ((width - 1)/8) * 8

	mov	rFore, pFore
	add	rDest, eax		; point to last dest ddword

	add	rBack, eax		; point to last background ddword
	add	rFore, eax		; point to last foreground ddword

	shr	eax, 1			; get ((width - 1)/8) * 4
	pxor	mm0, mm0

	add	rSel, eax		; point to last selector word
	pxor	mm1, mm1

; We really want to invert only the lower bytes of each halfword in the
; mask, so fill mm7 with 0x00FF00FF00FF00FF
	movq	mm7, MMWORD PTR grayInvMask
	pxor	mm6, mm6		; mm6 <- all 0's for unpack

; We need to decide which type of loop to begin with.  Each loop writes
; stuff from the previous loop iteration off to the right.  When we begin
; the loop, we'll be writing into the right frame.  This should always
; be written with 0's.
; The simple loop writes the contents of mm0, so we should set that to 0.
; The transition from the simple to complex loop expects the value
; read previously to be both mm0 and mm1, so we set them both to 0.
; We now decide which loop to start.

	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does rightmost selector
					; word select all background?
	jne	gray4bppSimpleToComplexSelLoop  ; nope, do complex loop.

	; empty
	jmp	gray4bppSimpleSelLoop	; otherwise, go to simple loop


; Here are the inner loops.  There are four sections:
; 1. Fast, simple inner loop that just copies background to dest;
; 2. Transition to complex inner loop;
; 3. Slower, complex inner loop that combines foreground and background,
; 4. Transition to simple inner loop.
;
; We can loop indefinitely in either of the inner loops.  The transition
; sections are used when we need to go from one to the other.
;
; Why all this weirdness?  Well, fairly large parts of the selector mask
; select background (there's lots of white space between text).  It's
; fast and easy to just copy the background to the output, so we make
; that a special case.
;
; Ok, but why do we have two completely different loops instead of
; two cases inside one loop as is done in the C code?
; As is often the case in MMX-land, we try to keep the pipelines full.
; To do this, one should try to keep two threads going all the time.
; At the beginning of a loop, there's usually only one thing to do.
; At the end of the loop, there's usually one thing to do while the
; results are being prepared.  To keep the processor busy, we overlap
; the beginning and end of the loop.  This overlapping stuff is why
; we have two loops.  The beginnings and ends of the two loops
; are sufficiently different that it's hard to have common code that's
; efficient.
;
; Since we actually have two loops, we need some transition code so
; that each loop's entry conditions are satisified, i.e. the pipeline
; has the right stuff in the right places.
;
; There's another very important point to remember!  The 64-bit dwords
; read from and written to memory are word-swapped.  That is, the 32-bit
; word in the bottom half of the dword register actually appears to the
; left of the 32-bit word in the upper half of the dword.  Also remember
; that selector bytes within a word are read in this order: 3,2,1,0.
; That's left to right.
;
; All this means that we combine selector byte 3 with bits [31:16] of
; the dword, and selector byte 2 with bits [15:00], selector byte 1
; with bits [63:48] and selector byte 0 with bits [47:32] (indexing
; from the LS end of the dword).  To do the combining, we use the MMX
; multiply opcode, which only works in units of 16 bits.  Therefore,
; we have to expand each selector byte out into two 16-bit things and
; expand each 4-byte word inside a dword out into 4 halfwords inside
; a 64-bit dword.  Gack!
;
; Here's the top of the simple loop.  We test to see if the next
; four selector bytes are filled with 0FFFFFFFFh (select background)
; at the bottom of the loop since that's where we decide whether to
; continue with the simple loop or move to the complex loop.
;
; At the beginning of this loop:
; rBack points to next background dword to be read.
; rFore points to next foreground dword to be read.
; rSel points to current selector dword.  It's already
;      been read and found to be all 1's, which is why we're here.
; rDest points to the dest dword that corresponds to the background
;      dword pointed to by rBack, i.e. rDest, rFore and rBack are all in sync.
; mm0 contains the dword found in the previous trip through the loop.
;
; For the sake of speed, we unroll this loop into four copies.  That
; means we only have to decrement the pointers once for all four copies.
; However, we do have to have four exit conditions so the pointers are
; updated correctly after we leave.
; We also take advantage of the fact that the left frame of the selector
; image is not equal to 0FFFFFFFFh (its equal to 0), so we don't have
; to check to see if we're at the beginning of the scan line in each
; unrolled loop - only at the exits.

; But first, the transition from complex loop to simple loop.
; The simple loop is really pretty simple and the code required to complete
; the complex loop is sort of long.
gray4bppComplexToSimpleSelLoop:
	paddw	mm2, mm3		; compute high order dest results
	; empty				; from prev loop iteration.

	psrlw	mm2, 8			; move high results to 4 LS bytes
	; empty

	packuswb mm0, mm2		; assemble low order bytes from mm0
	; empty				; and high order bytes from mm2
	

gray4bppSimpleSelLoop:
	movq	mm1, [rBack]		; read the background bytes
	; empty

	movq	[rDest + 8], mm0	; write prev output
	movq	mm0, mm1		; prepare for next write

	cmp	DWORD PTR [rSel - 4], 0FFFFFFFFh ; does next word select
					; all background?
	jne	gray4bppSimpleSelLoopExit1	; exit after one loop iteration

; loop iteration 2
	movq	mm1, [rBack - 8]	; read the background bytes
	; empty

	movq	[rDest], mm0		; write prev output
	movq	mm0, mm1		; prepare for next write

	cmp	DWORD PTR [rSel - 8], 0FFFFFFFFh ; does next word select
					; all background?
	jne	gray4bppSimpleSelLoopExit2 ; exit after two loop iterations

; loop iteration 3
	movq	mm1, [rBack - 16]	; read the background bytes
	; empty

	movq	[rDest - 8], mm0	; write prev output
	movq	mm0, mm1		; prepare for next write

	cmp	DWORD PTR [rSel - 12], 0FFFFFFFFh	; does next word select
					; all background?
	jne	gray4bppSimpleSelLoopExit3 ; exit after three loop iterations

; loop iteration 4
	movq	mm1, [rBack - 24]	; read the background bytes
	; empty

	sub	rBack, 32		; adjust pointers
	sub	rFore, 32

	movq	[rDest - 16], mm0	; write prev output
	movq	mm0, mm1		; prepare for next write

	sub	rDest, 32
	sub	rSel, 16

gray4bppSimpleSelLoopFin:
	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does next word select
					; all background?
	je	gray4bppSimpleSelLoop	; yep, go do the simple thing again

; At this point, all the pointers are updated.  All we have to do is
; check to see if we've already processed the word at the beginning
; of the line.
	cmp	rSel, pSel
	jb	gray4bppSimpleSelLoopExit4

; This is the transition code from the simple loop to the complex loop.
; The complex loop begins by adding the foreground and background values
; for four of the final answers, shifting this right by 8, then
; packing it with the other four output values, then diddling around,
; then writing the output.  To prepare for this, we unpack the upper
; half of mm0 into one register and the lower half into another to simulate
; the state at the end of the complex loop.

gray4bppSimpleToComplexSelLoop:
	punpcklbw mm0, mm6		; put low word bytes into LS
					; bytes of each 16-bit thing
	movq	mm3, mm1		; get high background

	punpckhbw mm3, mm3		; put hi backgnd word bytes into
					; MS bytes of each halfword.
	pxor	mm2, mm2		; clear hi word foreground reg


; This is the complex loop.  It starts by simultaneously finishing off
; the last complex operation and reading selector bytes from the next
; word.  At the beginning of this loop:
; rBack	points to next background dword to be read.
; rFore	points to next foreground dword to be read.
; rSel	points to current selector dword.  It's already
;	been read as a 32-bit word and found to be NOT all 1's,
;	which is why we're here.  It will now be read as 4 individual
;	bytes, each of which contains 2 nibbles.  Since we want to do
;	a 16-bit multiply with each nibble, we expand each byte to
;	32 bits.  Note that the table used to do the expansion is
;	filled with 8-byte things (top 4 bytes are 0000) so that we
;	can OR one expanded byte from the table with another that's already
;	in a register.	
; rDest	points to the dest dword that corresponds to the source dwords
;	pointed to by rBack & rFore, i.e. rDest, rFore and rBack are
;	all in sync.
; mm0	contains four 16-bit halfwords.  The bytes that will make up
;	the low order 4 bytes of the dest dword are in the 4 LS bytes of the
;	halftwords.
; mm2	contains four 16-bit weighted foreground pixels.  The interesting
;	8 bits in each pixel are in the MS bytes of the halfwords.
; mm3	contains four 16-bit weighted background pixels.  The interesting
;	8 bits in each pixel are in the MS bytes of the halfwords.

gray4bppComplexSelLoop:
	xor	eax, eax		; prepare to read sel byte
	paddw	mm2, mm3		; compute high order dest results
					; from prev loop iteration.

	mov	al, [rSel+3]		; get next selector byte
	psrlw	mm2, 8			; move high results to 4 LS bytes

	movq	mm1, [rBack]		; get next background dword
					; while waiting for eax's AGI
	packuswb mm0, mm2		; assemble low order bytes from mm0
					; and high order bytes from mm2

; while waiting for mm0 to percolate ...
	movq	mm4, MMWORD PTR graySelMask[eax*8] ; get expanded byte 3
	movq	mm3, mm1		; copy background dword so we can
					; operate on both 32-bit words
					; in different registers.

; TA-DA!  The result of the whole loop!
	movq	[rDest + 8], mm0	; write assembled dest dword.
	; empty

	xor	ebx, ebx		; prepare to read sel byte
	; empty				; It ought to be possible to
					; save a clock cycle here, but I can't
					; find it.

	mov	bl, [rSel+2]		; get selector byte 2
	punpcklbw mm1, mm6		; put bytes from low word of background
					; into LS bytes of mm1

	movq	mm0, [rFore]		; read foreground dword
	psllq	mm4, 32			; shifted expanded byte 3 to upper word

	por	mm4, MMWORD PTR graySelMask[ebx*8] ; OR in with expanded byte 2
	movq	mm2, mm0		; copy foreground dword so we can
					; operate on both 32-bit words
					; in different registers.

	pmullw	mm1, mm4		; multiply low four background bytes
					; by first two selector nibbles.
	punpcklbw mm0, mm6		; put bytes from low word of foreground
					; into LS bytes of mm0

	pxor	mm4, mm7		; invert the selector so it selects
					; foreground
	xor	eax, eax		; prepare to read selector byte 1

	pmullw	mm0, mm4		; multiply low four foreground bytes
	mov	al, [rSel+1]		; get selector byte 1

	xor	ebx, ebx		; prepare to read selector byte 0
	punpckhbw mm3, mm6		; put bytes from hi word of background
					; into LS bytes of 4 halfwords in mm3

	mov	bl, [rSel+0]		; read selector byte 0
	punpckhbw mm2, mm6		; put bytes from hi word of foreground
					; into LS bytes of 4 halfwords in mm2

	movq	mm4, MMWORD PTR graySelMask[eax*8] ; get expanded byte 1
	paddw	mm0, mm1		; weighted fore + weighted back for
					; low four pixels

	psllq	mm4, 32			; shift expanded byte 1 into position
	sub	rFore, 8		; decrement foreground ptr for next
					; time through loop

	por	mm4, MMWORD PTR graySelMask[ebx*8] ; OR in expanded byte 0
					; to create selector for hi four pixels
	psrlw	mm0, 8			; put useful product bits into
					; LS bytes of each halfword.

	pmullw	mm3, mm4		; multiply hi four background bytes
	pxor	mm4, mm7		; invert the selector so we can
					; multiply foreground bytes

	pmullw	mm2, mm4		; multiply hi four foreground bytes
	sub	rBack, 8		; decrement background ptr

	sub	rDest, 8		; decrement dest ptr
	sub	rSel, 4			; decrement selector ptr

; Test to see if we should go to the complex loop first.  This is significantly
; more likely than being done with the the line, which is why we test it first.
	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does next selector select
						; all background?
	je	gray4bppComplexToSimpleSelLoop	; yes, go to the simple loop.

; The selector is not 0FFFFFFFFh.  We might be reading the selector image's
; frame.  See if that's happening.  If not, go to the beginning of the
; complex loop.
	cmp	rSel, pSel		; past beginning of line yet?
	jae	gray4bppComplexSelLoop	; nope, continue with complex loop

; This is now we finish up after the complex loop.  It's harder.
	paddw	mm2, mm3		; compute high order dest results
	; empty				; from prev loop iteration.

	psrlw	mm2, 8			; move high results to 4 LS bytes
	; empty

	packuswb mm0, mm2		; assemble low order bytes from mm0
	; empty				; and high order bytes from mm2

; while waiting for mm0 to percolate ...
	; empty
	; empty

	movq	[rDest + 8], mm0	; write assembled dest dword.
	ret



; SIMPLE LOOP FINISHING CODE

; This is where we go when we exit the simple loop after one iteration.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelLoopExit1:
	sub	rBack, 8		; adjust pointers
	sub	rFore, 8

	sub	rDest, 8
	sub	rSel, 4

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm0
	ret


; This is where we go when we exit the simple loop after two iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelLoopExit2:
	sub	rBack, 16		; adjust pointers
	sub	rFore, 16

	sub	rDest, 16
	sub	rSel, 8

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm0
	ret


; This is where we go when we exit the simple loop after three iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelLoopExit3:
	sub	rBack, 24		; adjust pointers
	sub	rFore, 24

	sub	rDest, 24
	sub	rSel, 12

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm0
	ret

; This is where we go when we exit the simple loop after four iterations.
; At this point, we're really done with the scan line, so just write the
; last word and exit.
gray4bppSimpleSelLoopExit4:
	movq	[rDest + 8], mm0
	ret

ip_gray4bppSelLineLong	ENDP



; **************************************************************************



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_gray4bppSelLineLongNoFG.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_gray4bppSelLineLongNoFGStart::
ip_gray4bppSelLineLongNoFG  PROC NEAR C PUBLIC USES EAX EBX EDX ESI EDI, \
	pIPState:DWORD, pBack:DWORD, pSel:DWORD, pDest:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 8 bytes, so we'll point to the last multiple of 8
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	mov	rBack, pBack
	dec	eax			; really want (width - 1)

	mov	rSel, pSel
	and	eax, NOT 07h		; get ((width - 1)/8) * 8

	add	rDest, eax		; point to last dest ddword
	add	rBack, eax		; point to last background ddword

	shr	eax, 1			; get ((width - 1)/8) * 4
	pxor	mm1, mm1		; clear to write into right frame

	add	rSel, eax		; point to last selector word
	pxor	mm6, mm6		; mm6 <- all 0's for unpack

	pxor	mm3, mm3		; loops assume partial results
	; empty				; are in mm3

; We need to decide which type of loop to begin with.  Each loop writes
; stuff from the previous loop iteration off to the right.  When we begin
; the loop, we'll be writing into the right frame.  This should always
; be written with 0's, which are now in mm1.
; The simple loop writes the contents of mm1.
; The transition from the simple to complex loop expects the value
; read previously to be only in mm1, so that's cool too.
; We now decide which loop to start.

	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does rightmost selector
					; word select all background?
	jne	gray4bppSimpleToComplexSelNoFGLoop  ; nope, do complex loop.
					; otherwise, fall through to simple
					; loop


; Here are the inner loops.  There are four sections:
; 1. Fast, simple inner loop that just copies background to dest;
; 2. Transition to complex inner loop;
; 3. Slower, complex inner loop that combines foreground and background,
; 4. Transition to simple inner loop.
;
; We can loop indefinitely in either of the inner loops.  The transition
; sections are used when we need to go from one to the other.
;
; Why all this weirdness?  Well, fairly large parts of the selector mask
; select background (there's lots of white space between text).  It's
; fast and easy to just copy the background to the output, so we make
; that a special case.
;
; Ok, but why do we have two completely different loops instead of
; two cases inside one loop as is done in the C code?
; As is often the case in MMX-land, we try to keep the pipelines full.
; To do this, one should try to keep two threads going all the time.
; At the beginning of a loop, there's usually only one thing to do.
; At the end of the loop, there's usually one thing to do while the
; results are being prepared.  To keep the processor busy, we overlap
; the beginning and end of the loop.  This overlapping stuff is why
; we have two loops.  The beginnings and ends of the two loops
; are sufficiently different that it's hard to have common code that's
; efficient.
;
; Since we actually have two loops, we need some transition code so
; that each loop's entry conditions are satisified, i.e. the pipeline
; has the right stuff in the right places.
;
; There's another very important point to remember!  The 64-bit dwords
; read from and written to memory are word-swapped.  That is, the 32-bit
; word in the bottom half of the dword register actually appears to the
; left of the 32-bit word in the upper half of the dword.  Also remember
; that selector bytes within a word are read in this order: 3,2,1,0.
; That's left to right.
;
; All this means that we combine selector byte 3 with bits [31:16] of
; the dword, and selector byte 2 with bits [15:00], selector byte 1
; with bits [63:48] and selector byte 0 with bits [47:32] (indexing
; from the LS end of the dword).  To do the combining, we use the MMX
; multiply opcode, which only works in units of 16 bits.  Therefore,
; we have to expand each selector byte out into two 16-bit things and
; expand each 4-byte word inside a dword out into 4 halfwords inside
; a 64-bit dword.  Gack!
;
; Here's the top of the simple loop.  We test to see if the next
; four selector bytes are filled with 0FFFFFFFFh (select background)
; at the bottom of the loop since that's where we decide whether to
; continue with the simple loop or move to the complex loop.
;
; At the beginning of this loop:
; rBack points to next background dword to be read.
; rSel points to current selector dword.  It's already
;      been read and found to be all 1's, which is why we're here.
; rDest points to the dest dword that corresponds to the background
;      dword pointed to by rBack, i.e. rDest and rBack are all in sync.
; mm1 contains the dword found in the previous trip through the loop.
;
; For the sake of speed, we unroll this loop into four copies.  That
; means we only have to decrement the pointers once for all four copies.
; However, we do have to have four exit conditions so the pointers are
; updated correctly after we leave.
; We also take advantage of the fact that the left frame of the selector
; image is not equal to 0FFFFFFFFh (its equal to 0), so we don't have
; to check to see if we're at the beginning of the scan line in each
; unrolled loop - only at the exits.

; But first, the transition from complex loop to simple loop.
; The simple loop is really pretty simple and the code required to complete
; the complex loop is sort of long.
gray4bppComplexToSimpleSelNoFGLoop:
	psrlw	mm3, 8			; move high results to 4 LS bytes
	; empty

	packuswb mm1, mm3		; assemble low order bytes from mm1
	; empty

gray4bppSimpleSelNoFGLoop:
	movq	mm3, [rBack]		; read the background bytes
	; empty

	movq	[rDest + 8], mm1	; write prev output
	movq	mm1, mm3		; prepare for next write

	cmp	DWORD PTR [rSel - 4], 0FFFFFFFFh ; does next word select
					; all background?
	jne	gray4bppSimpleSelNoFGLoopExit1	; exit after one loop iteration

; loop iteration 2
	movq	mm3, [rBack - 8]	; read the background bytes
	; empty

	movq	[rDest], mm1		; write prev output
	movq	mm1, mm3		; prepare for next write

	cmp	DWORD PTR [rSel - 8], 0FFFFFFFFh ; does next word select
					; all background?
	jne	gray4bppSimpleSelNoFGLoopExit2 ; exit after two loop iterations

; loop iteration 3
	movq	mm3, [rBack - 16]	; read the background bytes
	; empty

	movq	[rDest - 8], mm1	; write prev output
	movq	mm1, mm3		; prepare for next write

	cmp	DWORD PTR [rSel - 12], 0FFFFFFFFh	; does next word select
					; all background?
	jne	gray4bppSimpleSelNoFGLoopExit3 ; exit after 3 loop iterations

; loop iteration 4
	movq	mm3, [rBack - 24]	; read the background bytes
	; empty

	sub	rBack, 32		; adjust pointers
	; empty				; sorry

	movq	[rDest - 16], mm1	; write prev output
	movq	mm1, mm3		; prepare for next write

	sub	rDest, 32
	sub	rSel, 16

gray4bppSimpleSelNoFGLoopFin:
	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does next word select
					; all background?
	je	gray4bppSimpleSelNoFGLoop ; yep, go do the simple thing again

; We might be done with the entire scan line.  Test and see.
	cmp	rSel, pSel
	jb	gray4bppSimpleSelNoFGLoopExit4

; This is the transition code from the simple loop to the complex loop.
; The complex loop begins by shifting weighted background right by 8, then
; packing it with the other four output values, then diddling around,
; then writing the output.  To prepare for this, we unpack the upper
; half of mm1 into one register and the lower half into another to simulate
; the state at the end of the complex loop.

gray4bppSimpleToComplexSelNoFGLoop:
	punpcklbw mm1, mm6		; put low word bytes into LS
	; empty				; bytes of each 16-bit thing

	punpckhbw mm3, mm3		; put hi backgnd word bytes into
					; MS bytes of each halfword.
	xor	eax, eax		; get ready to read selector byte 3


; This is the complex loop.  It starts by simultaneously finishing off
; the last complex operation and reading selector bytes from the next
; word.  At the beginning of this loop:
; rBack	points to next background dword to be read.
; rSel	points to current selector dword.  It's already
;	been read as a 32-bit word and found to be NOT all 1's,
;	which is why we're here.  It will now be read as 4 individual
;	bytes, each of which contains 2 nibbles.  Since we want to do
;	a 16-bit multiply with each nibble, we expand each byte to
;	32 bits.  Note that the table used to do the expansion is
;	filled with 8-byte things (top 4 bytes are 0000) so that we
;	can OR one expanded byte from the table with another that's already
;	in a register.	
; rDest	points to the dest dword that corresponds to the source dwords
;	pointed to by rBack, i.e. rDest and rBack are in sync.
; mm1	contains four 16-bit background pixels.  The bytes that will make up
;	the low order 4 bytes of the dest dword are in the 4 LS bytes of the
;	halftwords.
; mm3	contains four 16-bit weighted background pixels.  The interesting
;	8 bits in each pixel are in the MS bytes of the halfwords.

gray4bppComplexSelNoFGLoop:
	mov	al, [rSel+3]		; get next selector byte
	xor	ebx, ebx		; prepare to read sel byte 2

	mov	bl, [rSel+2]		; get selector byte 2
	psrlw	mm3, 8			; move high results to 4 LS bytes

	movq	mm4, MMWORD PTR graySelMask[eax*8] ; get expanded byte 3
	packuswb mm1, mm3		; assemble low order bytes from mm1
					; and high order bytes from mm3

; while waiting for mm1 to percolate ...
	movq	mm3, [rBack]		; get next background dword
	psllq	mm4, 32			; shifted expanded byte 3 to upper word

; TA-DA!  The result of the whole loop!
	movq	[rDest + 8], mm1	; write assembled dest dword.
	movq	mm1, mm3		; copy background dword so we can
					; operate on both 32-bit words
					; in different registers.

	xor	eax, eax		; prepare to read selector byte 1
	punpcklbw mm1, mm6		; put bytes from low word of background
					; into LS bytes of mm1

	por	mm4, MMWORD PTR graySelMask[ebx*8] ; OR in with expanded byte 2
	punpckhbw mm3, mm6		; put bytes from hi word of background
					; into LS bytes of 4 halfwords in mm3

	mov	al, [rSel+1]		; get selector byte 1
	pmullw	mm1, mm4		; multiply low four background bytes
					; by first two selector nibbles.

	xor	ebx, ebx		; prepare to read selector byte 0
	sub	rBack, 8		; decrement background ptr

	movq	mm4, MMWORD PTR graySelMask[eax*8] ; get expanded byte 1
	; empty

	mov	bl, [rSel+0]		; read selector byte 0
	psllq	mm4, 32			; shift expanded byte 1 into position

	sub	rSel, 4			; decrement selector ptr
	xor	eax, eax		; prepare to read sel byte 3 next time

	por	mm4, MMWORD PTR graySelMask[ebx*8] ; OR in expanded byte 0
					; to create selector for hi four pixels
	psrlw	mm1, 8			; put useful product bits into
					; LS bytes of each halfword.

	pmullw	mm3, mm4		; multiply hi four background bytes
	sub	rDest, 8		; decrement dest ptr

; Test to see if we should go to the complex loop first.  This is significantly
; more likely than being done with the the line, which is why we test it first.
	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does next selector select
						; all background?
	je	gray4bppComplexToSimpleSelNoFGLoop ; yes, go to the simple loop.

; The selector is not 0FFFFFFFFh.  We might be reading the selector image's
; frame.  See if that's happening.  If not, go to the beginning of the
; complex loop.
	cmp	rSel, pSel		; past beginning of line yet?
	jae	gray4bppComplexSelNoFGLoop ; nope, continue with complex loop

; This is now we finish up after the complex loop.  It's harder.
	psrlw	mm3, 8			; move high results to 4 LS bytes
	; empty

	packuswb mm1, mm3		; assemble low order bytes from mm0
	; empty				; and high order bytes from mm2

; while waiting for mm1 to percolate ...
	; empty
	; empty

	movq	[rDest + 8], mm1	; write assembled dest dword.
	ret


; SIMPLE LOOP FINISHING CODE

; This is where we go when we exit the simple loop after one iteration.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoFGLoopExit1:
	sub	rBack, 8		; adjust pointers
	; empty

	sub	rDest, 8
	sub	rSel, 4

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoFGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm1
	ret


; This is where we go when we exit the simple loop after two iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoFGLoopExit2:
	sub	rBack, 16		; adjust pointers
	; empty

	sub	rDest, 16
	sub	rSel, 8

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoFGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm1
	ret


; This is where we go when we exit the simple loop after three iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoFGLoopExit3:
	sub	rBack, 24		; adjust pointers
	; empty

	sub	rDest, 24
	sub	rSel, 12

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoFGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm1
	ret

; This is where we go when we exit the simple loop after four iterations.
; At this point, we're really done with the scan line, so just write the
; last word and exit.
gray4bppSimpleSelNoFGLoopExit4:
	movq	[rDest + 8], mm1
	ret

ip_gray4bppSelLineLongNoFG	ENDP



; **************************************************************************



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_gray4bppSelNoBGLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_gray4bppSelLineLongNoBGStart::
ip_gray4bppSelLineLongNoBG  PROC NEAR C PUBLIC USES EAX EBX ECX ESI EDI, \
	pIPState:DWORD, pFore:DWORD, pSel:DWORD, pDest:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 8 bytes, so we'll point to the last multiple of 8
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	mov	rFore, pFore
	dec	eax			; really want (width - 1)

	mov	rSel, pSel
	and	eax, NOT 07h		; get ((width - 1)/8) * 8

	add	rDest, eax		; point to last dest ddword
	add	rFore, eax		; point to last foreground ddword

	shr	eax, 1			; get ((width - 1)/8) * 4
	pxor	mm0, mm0

	add	rSel, eax		; point to last selector word
	pxor	mm1, mm1

; We really want to invert only the lower bytes of each halfword in the
; mask, so fill mm7 with 0x00FF00FF00FF00FF
	movq	mm7, MMWORD PTR grayInvMask
	pxor	mm6, mm6		; mm6 <- all 0's for unpack

; We need to decide which type of loop to begin with.  Each loop writes
; stuff from the previous loop iteration off to the right.  When we begin
; the loop, we'll be writing into the right frame.  This should always
; be written with 0's.
; The simple loop writes the contents of mm0, so we should set that to 0.
; The transition from the simple to complex loop expects the value
; read previously to be both mm0 and mm1, so we set them both to 0.
; We now decide which loop to start.

	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does rightmost selector
					; word select all background?
	jne	gray4bppSimpleToComplexSelNoBGLoop  ; nope, do complex loop.

	; empty
	jmp	gray4bppSimpleSelNoBGLoop	; otherwise, go to simple loop


; Here are the inner loops.  There are four sections:
; 1. Fast, simple inner loop that just copies background to dest;
; 2. Transition to complex inner loop;
; 3. Slower, complex inner loop that combines foreground and background,
; 4. Transition to simple inner loop.
;
; We can loop indefinitely in either of the inner loops.  The transition
; sections are used when we need to go from one to the other.
;
; Why all this weirdness?  Well, fairly large parts of the selector mask
; select background (there's lots of white space between text).  It's
; fast and easy to just copy the background to the output, so we make
; that a special case.
;
; Ok, but why do we have two completely different loops instead of
; two cases inside one loop as is done in the C code?
; As is often the case in MMX-land, we try to keep the pipelines full.
; To do this, one should try to keep two threads going all the time.
; At the beginning of a loop, there's usually only one thing to do.
; At the end of the loop, there's usually one thing to do while the
; results are being prepared.  To keep the processor busy, we overlap
; the beginning and end of the loop.  This overlapping stuff is why
; we have two loops.  The beginnings and ends of the two loops
; are sufficiently different that it's hard to have common code that's
; efficient.
;
; Since we actually have two loops, we need some transition code so
; that each loop's entry conditions are satisified, i.e. the pipeline
; has the right stuff in the right places.
;
; There's another very important point to remember!  The 64-bit dwords
; read from and written to memory are word-swapped.  That is, the 32-bit
; word in the bottom half of the dword register actually appears to the
; left of the 32-bit word in the upper half of the dword.  Also remember
; that selector bytes within a word are read in this order: 3,2,1,0.
; That's left to right.
;
; All this means that we combine selector byte 3 with bits [31:16] of
; the dword, and selector byte 2 with bits [15:00], selector byte 1
; with bits [63:48] and selector byte 0 with bits [47:32] (indexing
; from the LS end of the dword).  To do the combining, we use the MMX
; multiply opcode, which only works in units of 16 bits.  Therefore,
; we have to expand each selector byte out into two 16-bit things and
; expand each 4-byte word inside a dword out into 4 halfwords inside
; a 64-bit dword.  Gack!
;
; Here's the top of the simple loop.  We test to see if the next
; four selector bytes are filled with 0FFFFFFFFh (select background)
; at the bottom of the loop since that's where we decide whether to
; continue with the simple loop or move to the complex loop.
;
; At the beginning of this loop:
; rFore points to next foreground dword to be read.
; rSel points to current selector dword.  It's already
;      been read and found to be all 1's, which is why we're here.
; rDest points to the dest dword that corresponds to the background
;      dword pointed to by rFore, i.e. rDest and rFore are in sync.
; mm0 contains the dword found in the previous trip through the loop.
;
; For the sake of speed, we unroll this simple loop into four copies.  That
; means we only have to decrement the pointers once for all four copies.
; However, we do have to have four exit conditions so the pointers are
; updated correctly after we leave.
; We also take advantage of the fact that the left frame of the selector
; image is not equal to 0FFFFFFFFh (its equal to 0), so we don't have
; to check to see if we're at the beginning of the scan line in each
; unrolled loop - only at the exits.

; But first, the transition from complex loop to simple loop.
; The simple loop is really pretty simple and the code required to complete
; the complex loop is sort of long.
gray4bppComplexToSimpleSelNoBGLoop:
	psrlw	mm2, 8			; move high results to 4 LS bytes
	; empty

	paddw	mm2, mm3		; compute high order dest results
	; empty				; from prev loop iteration.

	packuswb mm0, mm2		; assemble low order bytes from mm0
	; empty				; and high order bytes from mm2
	

gray4bppSimpleSelNoBGLoop:
	movq	[rDest + 8], mm0	; write prev output
	pcmpeqb	mm0, mm0		; prepare virtual backgnd for next write

	cmp	DWORD PTR [rSel - 4], 0FFFFFFFFh ; does next word select
					; all background?
	jne	gray4bppSimpleSelNoBGLoopExit1	; exit after one loop iteration

; loop iteration 2
	movq	[rDest], mm0		; write prev output
	; empty

	cmp	DWORD PTR [rSel - 8], 0FFFFFFFFh ; does next word select
					; all background?
	jne	gray4bppSimpleSelNoBGLoopExit2 ; exit after two loop iterations

; loop iteration 3
	movq	[rDest - 8], mm0	; write prev output
	; empty

	cmp	DWORD PTR [rSel - 12], 0FFFFFFFFh	; does next word select
					; all background?
	jne	gray4bppSimpleSelNoBGLoopExit3 ; exit after 3 loop iterations

; loop iteration 4
	movq	[rDest - 16], mm0	; write prev output
	; empty

	sub	rFore, 32		; adjust pointers
	sub	rDest, 32

	sub	rSel, 16
	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does next word select
					; all background?
	; empty
	je	gray4bppSimpleSelNoBGLoop ; yep, go do the simple thing again

; At this point, all the pointers are updated.  All we have to do is
; check to see if we've already processed the word at the beginning
; of the line.
	cmp	rSel, pSel
	jb	gray4bppSimpleSelNoBGLoopExit4

; This is the transition code from the simple loop to the complex loop.
; The complex loop begins by shifting the foreground values right by 8
; for four of the final answers, adding this to the virtual background
; values, then packing it with the other four output values, then
; diddling around, then writing the output.  To prepare for this,
; we unpack the upper half of mm0 into one register and the lower
; half into another to simulate the state at the end of the complex loop.

gray4bppSimpleToComplexSelNoBGLoop:
	movq	mm3, mm0		; get high virtual background
					; NOTE, mm0 normally contains all
					; 1's at this point, but will
					; contains 0's if this is how
					; we enter the loop.
	punpcklbw mm0, mm6		; put low word bytes into LS
					; bytes of each 16-bit thing
					

	punpckhbw mm3, mm6		; put hi backgnd word bytes into
					; LS bytes of each halfword.
	pxor	mm2, mm2		; clear hi word foreground reg


; This is the complex loop.  It starts by simultaneously finishing off
; the last complex operation and reading selector bytes from the next
; word.  At the beginning of this loop:
; rFore	points to next foreground dword to be read.
; rSel	points to current selector dword.  It's already
;	been read as a 32-bit word and found to be NOT all 1's,
;	which is why we're here.  It will now be read as 4 individual
;	bytes, each of which contains 2 nibbles.  Since we want to do
;	a 16-bit multiply with each nibble, we expand each byte to
;	32 bits.  Note that the table used to do the expansion is
;	filled with 8-byte things (top 4 bytes are 0000) so that we
;	can OR one expanded byte from the table with another that's already
;	in a register.	
; rDest	points to the dest dword that corresponds to the source dwords
;	pointed to by rFore, i.e. rDest and rFore are in sync.
; mm0	contains four 16-bit halfwords.  The bytes that will make up
;	the low order 4 bytes of the dest dword are in the 4 LS bytes of the
;	halftwords.
; mm2	contains four 16-bit weighted foreground pixels.  The interesting
;	8 bits in each pixel are in the MS bytes of the halfwords.
; mm3	contains four 16-bit weighted virtual background pixels.
;	The interesting 8 bits in each pixel are in the LS bytes
;	of the halfwords.

gray4bppComplexSelNoBGLoop:
	xor	eax, eax		; prepare to read sel byte
	xor	ebx, ebx		; prepare to read sel byte

	mov	al, [rSel+3]		; get next selector byte
	psrlw	mm2, 8			; move high results to 4 LS bytes

	mov	bl, [rSel+2]		; get selector byte 2
	paddw	mm2, mm3		; compute high order dest results
					; from prev loop iteration.

	movq	mm4, MMWORD PTR graySelMask[eax*8] ; get expanded byte 3
	packuswb mm0, mm2		; assemble low order bytes from mm0
					; and high order bytes from mm2

; while waiting for mm0 to percolate ...
	movq	mm2, [rFore]		; read foreground dword
	psllq	mm4, 32			; shifted expanded byte 3 to upper word

; TA-DA!  The result of the whole loop!
	movq	[rDest + 8], mm0	; write assembled dest dword.
	movq	mm0, mm2		; copy foreground dword so we can
					; operate on both 32-bit words
					; in different registers.

	por	mm4, MMWORD PTR graySelMask[ebx*8] ; OR in with expanded byte 2
	punpcklbw mm0, mm6		; put bytes from low word of foreground
					; into LS bytes of mm0

	movq	mm1, mm4		; get virtual background bytes
	pxor	mm4, mm7		; invert the selector so it selects
					; foreground
	pmullw	mm0, mm4		; multiply low four foreground bytes
	xor	ebx, ebx		; prepare to read selector byte 0

	xor	eax, eax		; prepare to read selector byte 1
	mov	bl, [rSel+0]		; read selector byte 0

	mov	al, [rSel+1]		; get selector byte 1
	punpckhbw mm2, mm6		; put bytes from hi word of foreground
					; into LS bytes of 4 halfwords in mm2

	movq	mm4, MMWORD PTR graySelMask[ebx*8] ; get expanded byte 0
	psrlw	mm0, 8			; put useful product bits into
					; LS bytes of each halfword.

	sub	rDest, 8		; decrement dest ptr
	sub	rFore, 8		; decrement foreground ptr for next
					; time through loop

	punpckldq mm4, MMWORD PTR graySelMask[eax*8] ; insert in expanded byte 1
					; to create selector for hi four pixels
	paddw	mm0, mm1		; weighted fore + weighted back for
					; low four pixels

	movq	mm3, mm4		; create hi four virtual
					; background bytes
	pxor	mm4, mm7		; invert the selector so we can
					; multiply foreground bytes

	pmullw	mm2, mm4		; multiply hi four foreground bytes
	sub	rSel, 4			; decrement selector ptr

; Test to see if we should go to the complex loop first.  This is significantly
; more likely than being done with the the line, which is why we test it first.
	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does next selector select
						; all background?
	je	gray4bppComplexToSimpleSelNoBGLoop ; yes, go to the simple loop.

; The selector is not 0FFFFFFFFh.  We might be reading the selector image's
; frame.  See if that's happening.  If not, go to the beginning of the
; complex loop.
	cmp	rSel, pSel		; past beginning of line yet?
	jae	gray4bppComplexSelNoBGLoop ; nope, continue with complex loop

; This is now we finish up after the complex loop.  It's harder.
	psrlw	mm2, 8			; move high results to 4 LS bytes
	; empty

	paddw	mm2, mm3		; compute high order dest results
	; empty				; from prev loop iteration.

	packuswb mm0, mm2		; assemble low order bytes from mm0
	; empty				; and high order bytes from mm2

; while waiting for mm0 to percolate ...
	; empty
	; empty

	movq	[rDest + 8], mm0	; write assembled dest dword.
	ret



; SIMPLE LOOP FINISHING CODE

; This is where we go when we exit the simple loop after one iteration.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoBGLoopExit1:
	sub	rFore, 8	; adjust pointers
	; empty

	sub	rDest, 8
	sub	rSel, 4

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoBGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm0
	ret


; This is where we go when we exit the simple loop after two iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoBGLoopExit2:
	sub	rFore, 16		; adjust pointers
	; empty

	sub	rDest, 16
	sub	rSel, 8

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoBGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm0
	ret


; This is where we go when we exit the simple loop after three iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoBGLoopExit3:
	sub	rFore, 24		; adjust pointers
	; empty

	sub	rDest, 24
	sub	rSel, 12

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoBGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm0
	ret

; This is where we go when we exit the simple loop after four iterations.
; At this point, we're really done with the scan line, so just write the
; last word and exit.
gray4bppSimpleSelNoBGLoopExit4:
	movq	[rDest + 8], mm0
	ret

ip_gray4bppSelLineLongNoBG	ENDP



; **************************************************************************



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_gray4bppSelNoFGNoBGLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_gray4bppSelLineLongNoFGNoBGStart::
ip_gray4bppSelLineLongNoFGNoBG  PROC NEAR C PUBLIC USES EAX EBX ECX ESI EDI, \
	pIPState:DWORD, pSel:DWORD, pDest:DWORD, \
	imgWidth:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  We want to process the line from right to
; left in units of 8 bytes, so we'll point to the last multiple of 8
; bytes in the foreground, background and dest.  We point to the last
; multiple of 4 bytes in the selector.
	mov	rDest, pDest
	mov	eax, imgWidth

	dec	eax			; really want (width - 1)
	mov	rSel, pSel

	and	eax, NOT 07h		; get ((width - 1)/8) * 8
	pxor	mm6, mm6		; mm6 <- all 0's for unpack

	add	rDest, eax		; point to last dest ddword
	shr	eax, 1			; get ((width - 1)/8) * 4

	add	rSel, eax		; point to last selector word
	pxor	mm1, mm1

; We need to decide which type of loop to begin with.  Each loop writes
; stuff from the previous loop iteration off to the right.  When we begin
; the loop, we'll be writing into the right frame.  This should always
; be written with 0's.
; The simple loop writes the contents of mm0, so we should set that to 0.
; The transition from the simple to complex loop expects the value
; read previously to be both mm0 and mm1, so we set them both to 0.
; We now decide which loop to start.

	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does rightmost selector
					; word select all background?
	jne	gray4bppSimpleToComplexSelNoFGNoBGLoop  ; nope, do complex loop.

	; empty
	jmp	gray4bppSimpleSelNoFGNoBGLoop	; otherwise, go to simple loop


; Here are the inner loops.  There are four sections:
; 1. Fast, simple inner loop that just copies background to dest;
; 2. Transition to complex inner loop;
; 3. Slower, complex inner loop that combines foreground and background,
; 4. Transition to simple inner loop.
;
; We can loop indefinitely in either of the inner loops.  The transition
; sections are used when we need to go from one to the other.
;
; Why all this weirdness?  Well, fairly large parts of the selector mask
; select background (there's lots of white space between text).  It's
; fast and easy to just copy the background to the output, so we make
; that a special case.
;
; Ok, but why do we have two completely different loops instead of
; two cases inside one loop as is done in the C code?
; As is often the case in MMX-land, we try to keep the pipelines full.
; To do this, one should try to keep two threads going all the time.
; At the beginning of a loop, there's usually only one thing to do.
; At the end of the loop, there's usually one thing to do while the
; results are being prepared.  To keep the processor busy, we overlap
; the beginning and end of the loop.  This overlapping stuff is why
; we have two loops.  The beginnings and ends of the two loops
; are sufficiently different that it's hard to have common code that's
; efficient.
;
; Since we actually have two loops, we need some transition code so
; that each loop's entry conditions are satisified, i.e. the pipeline
; has the right stuff in the right places.
;
; There's another very important point to remember!  The 64-bit dwords
; read from and written to memory are word-swapped.  That is, the 32-bit
; word in the bottom half of the dword register actually appears to the
; left of the 32-bit word in the upper half of the dword.  Also remember
; that selector bytes within a word are read in this order: 3,2,1,0.
; That's left to right.
;
; All this means that we combine selector byte 3 with bits [31:16] of
; the dword, and selector byte 2 with bits [15:00], selector byte 1
; with bits [63:48] and selector byte 0 with bits [47:32] (indexing
; from the LS end of the dword).  To do the combining, we use the MMX
; multiply opcode, which only works in units of 16 bits.  Therefore,
; we have to expand each selector byte out into two 16-bit things and
; expand each 4-byte word inside a dword out into 4 halfwords inside
; a 64-bit dword.  Gack!
;
; Here's the top of the simple loop.  We test to see if the next
; four selector bytes are filled with 0FFFFFFFFh (select background)
; at the bottom of the loop since that's where we decide whether to
; continue with the simple loop or move to the complex loop.
;
; At the beginning of this loop:
; rFore points to next foreground dword to be read.
; rSel points to current selector dword.  It's already
;      been read and found to be all 1's, which is why we're here.
; rDest points to the dest dword that corresponds to the background
;      dword pointed to by rFore, i.e. rDest and rFore are in sync.
; mm0 contains the dword found in the previous trip through the loop.
;
; For the sake of speed, we unroll this simple loop into four copies.  That
; means we only have to decrement the pointers once for all four copies.
; However, we do have to have four exit conditions so the pointers are
; updated correctly after we leave.
; We also take advantage of the fact that the left frame of the selector
; image is not equal to 0FFFFFFFFh (its equal to 0), so we don't have
; to check to see if we're at the beginning of the scan line in each
; unrolled loop - only at the exits.

; But first, the transition from complex loop to simple loop.
; The simple loop is really pretty simple and the code required to complete
; the complex loop is sort of long.
gray4bppComplexToSimpleSelNoFGNoBGLoop:
	packuswb mm1, mm4		; assemble low order bytes from mm1
	; empty				; and high order bytes from mm3
	

gray4bppSimpleSelNoFGNoBGLoop:
	movq	[rDest + 8], mm1	; write prev output
	pcmpeqb	mm1, mm1		; prepare virtual backgnd for next write

	cmp	DWORD PTR [rSel - 4], 0FFFFFFFFh	; does next word select
							; all background?
	jne	gray4bppSimpleSelNoFGNoBGLoopExit1	; exit after one loop
							; iteration

; loop iteration 2
	movq	[rDest], mm1				; write prev output
	; empty

	cmp	DWORD PTR [rSel - 8], 0FFFFFFFFh	; does next word select
							; all background?
	jne	gray4bppSimpleSelNoFGNoBGLoopExit2	; exit after two loop
							; iterations

; loop iteration 3
	movq	[rDest - 8], mm1			; write prev output
	; empty

	cmp	DWORD PTR [rSel - 12], 0FFFFFFFFh	; does next word select
							; all background?
	jne	gray4bppSimpleSelNoFGNoBGLoopExit3	; exit after 3 loop
							; iterations

; loop iteration 4
	movq	[rDest - 16], mm1			; write prev output
	; empty

	sub	rDest, 32
	sub	rSel, 16

gray4bppSimpleSelNoFGNoBGLoopFin:
	cmp	DWORD PTR [rSel], 0FFFFFFFFh		; does next word select
							; all background?
	je	gray4bppSimpleSelNoFGNoBGLoop 		; yep, go do the
							; simple thing again

; At this point, all the pointers are updated.  All we have to do is
; check to see if we've already processed the word at the beginning
; of the line.
	cmp	rSel, pSel
	jb	gray4bppSimpleSelNoFGNoBGLoopExit4

; This is the transition code from the simple loop to the complex loop.
; The complex loop begins by shifting the foreground values right by 8
; for four of the final answers, adding this to the virtual background
; values, then packing it with the other four output values, then
; diddling around, then writing the output.  To prepare for this,
; we unpack the upper half of mm0 into one register and the lower
; half into another to simulate the state at the end of the complex loop.

gray4bppSimpleToComplexSelNoFGNoBGLoop:
	movq	mm4, mm1		; get high virtual background
					; NOTE, mm1 normally contains all
					; 1's at this point, but will
					; contains 0's if this is how
					; we enter the loop.
	punpcklbw mm1, mm6		; put low word bytes into LS
					; bytes of each 16-bit thing
					

	punpckhbw mm4, mm6		; put hi backgnd word bytes into
					; LS bytes of each halfword.
	xor	ebx, ebx		; for load of bl following


; This is the complex loop.  It starts by simultaneously finishing off
; the last complex operation and reading selector bytes from the next
; word.  At the beginning of this loop:
; rSel	points to current selector dword.  It's already
;	been read as a 32-bit word and found to be NOT all 1's,
;	which is why we're here.  It will now be read as 4 individual
;	bytes, each of which contains 2 nibbles.  Since we want to do
;	a 16-bit multiply with each nibble, we expand each byte to
;	32 bits.  Note that the table used to do the expansion is
;	filled with 8-byte things (top 4 bytes are 0000) so that we
;	can OR one expanded byte from the table with another that's already
;	in a register.	
; rDest	points to the dest dword that corresponds to the source dwords
;	pointed to be written with data produced in this loop.
; mm1	contains four 16-bit weighted virtual background pixels.
;	The bytes that will make up the low order 4 bytes of the
;	dest dword are in the 4 LS bytes of the halftwords.
; mm4	contains four 16-bit weighted virtual background pixels.
;	The interesting 8 bits in each pixel are in the LS bytes
;	of the halfwords.

gray4bppComplexSelNoFGNoBGLoop:
	xor	eax, eax		; prepare to read sel byte
	mov	bl, [rSel+2]		; get selector byte 2

	mov	al, [rSel+3]		; get next selector byte
	xor	ecx, ecx		; for selector byte 0

	movq	mm5, MMWORD PTR graySelMask[ebx*8] ; get expanded byte 2
	packuswb mm1, mm4		; assemble low order bytes from mm1
					; and high order bytes from mm4

; while waiting for mm1 to percolate ...
	punpckldq mm5, MMWORD PTR graySelMask[eax*8] ; insert expanded byte 3
	; empty

; TA-DA!  The result of the whole loop!
	movq	[rDest + 8], mm1	; write assembled dest dword.
	movq	mm1, mm5		; copy for low order bytes

	sub	rDest, 8		; decrement dest ptr
	xor	ebx, ebx		; prepare to read selector byte 0

	xor	eax, eax		; prepare to read selector byte 1
	mov	cl, [rSel+0]		; read selector byte 0

	mov	al, [rSel+1]		; get selector byte 1
	sub	rSel, 4			; decrement selector ptr

	movq	mm4, MMWORD PTR graySelMask[ecx*8] ; get expanded byte 0
	; empty

	punpckldq mm4, MMWORD PTR graySelMask[eax*8] ; insert in expanded byte 1
					; to create selector for hi four pixels
	; empty

; Test to see if we should go to the complex loop first.  This is significantly
; more likely than being done with the the line, which is why we test it first.
	cmp	DWORD PTR [rSel], 0FFFFFFFFh	; does next selector select
						; all background?
	je	gray4bppComplexToSimpleSelNoFGNoBGLoop	; yes, go to the
							; simple loop.

; The selector is not 0FFFFFFFFh.  We might be reading the selector image's
; frame.  See if that's happening.  If not, go to the beginning of the
; complex loop.
	cmp	rSel, pSel			; past beginning of line yet?
	jae	gray4bppComplexSelNoFGNoBGLoop	; nope, continue with
						; complex loop

; This is now we finish up after the complex loop.  It's harder.
	packuswb mm1, mm4		; assemble low order bytes from mm1
	; empty				; and high order bytes from mm3

; while waiting for mm1 to percolate ...
	; empty
	; empty

	movq	[rDest + 8], mm1	; write assembled dest dword.
	ret



; SIMPLE LOOP FINISHING CODE

; This is where we go when we exit the simple loop after one iteration.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoFGNoBGLoopExit1:
	sub	rDest, 8
	sub	rSel, 4

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoFGNoBGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm1
	ret


; This is where we go when we exit the simple loop after two iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoFGNoBGLoopExit2:
	sub	rDest, 16
	sub	rSel, 8

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoFGNoBGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm1
	ret


; This is where we go when we exit the simple loop after three iterations.
; We need to update the pointers and see if we're done with the line yet.
gray4bppSimpleSelNoFGNoBGLoopExit3:
	sub	rDest, 24
	sub	rSel, 12

	cmp	rSel, pSel
	jae	gray4bppSimpleToComplexSelNoFGNoBGLoop

; If we've done the whole line, write the last word and quit.
	movq	[rDest + 8], mm1
	ret

; This is where we go when we exit the simple loop after four iterations.
; At this point, we're really done with the scan line, so just write the
; last word and exit.
gray4bppSimpleSelNoFGNoBGLoopExit4:
	movq	[rDest + 8], mm1
	ret

ip_gray4bppSelLineLongNoFGNoBG	ENDP





; Segment for initialized variables, like masks
_DATA	SEGMENT PUBLIC 'DATA'

; This is the table used to expand each binary selector byte into 8 bytes.
; Note that since the order of words within MMX double words is reversed,
; the order of the words within each dword entry is reversed.
; Also, since we want to combine with background using an uninverted
; table entry (this is the common case), we invert the table entries so
; that a 0 in the selector turns into FF in the table entry.
		ALIGN	8
binExp8Table	QWORD	0FFFFFFFFFFFFFFFFh
		QWORD	0FFFFFF00FFFFFFFFh
		QWORD	0FFFF00FFFFFFFFFFh
		QWORD	0FFFF0000FFFFFFFFh
		QWORD	0FF00FFFFFFFFFFFFh
		QWORD	0FF00FF00FFFFFFFFh
		QWORD	0FF0000FFFFFFFFFFh
		QWORD	0FF000000FFFFFFFFh
		QWORD	000FFFFFFFFFFFFFFh
		QWORD	000FFFF00FFFFFFFFh
		QWORD	000FF00FFFFFFFFFFh
		QWORD	000FF0000FFFFFFFFh
		QWORD	00000FFFFFFFFFFFFh
		QWORD	00000FF00FFFFFFFFh
		QWORD	0000000FFFFFFFFFFh
		QWORD	000000000FFFFFFFFh
		QWORD	0FFFFFFFFFFFFFF00h
		QWORD	0FFFFFF00FFFFFF00h
		QWORD	0FFFF00FFFFFFFF00h
		QWORD	0FFFF0000FFFFFF00h
		QWORD	0FF00FFFFFFFFFF00h
		QWORD	0FF00FF00FFFFFF00h
		QWORD	0FF0000FFFFFFFF00h
		QWORD	0FF000000FFFFFF00h
		QWORD	000FFFFFFFFFFFF00h
		QWORD	000FFFF00FFFFFF00h
		QWORD	000FF00FFFFFFFF00h
		QWORD	000FF0000FFFFFF00h
		QWORD	00000FFFFFFFFFF00h
		QWORD	00000FF00FFFFFF00h
		QWORD	0000000FFFFFFFF00h
		QWORD	000000000FFFFFF00h
		QWORD	0FFFFFFFFFFFF00FFh
		QWORD	0FFFFFF00FFFF00FFh
		QWORD	0FFFF00FFFFFF00FFh
		QWORD	0FFFF0000FFFF00FFh
		QWORD	0FF00FFFFFFFF00FFh
		QWORD	0FF00FF00FFFF00FFh
		QWORD	0FF0000FFFFFF00FFh
		QWORD	0FF000000FFFF00FFh
		QWORD	000FFFFFFFFFF00FFh
		QWORD	000FFFF00FFFF00FFh
		QWORD	000FF00FFFFFF00FFh
		QWORD	000FF0000FFFF00FFh
		QWORD	00000FFFFFFFF00FFh
		QWORD	00000FF00FFFF00FFh
		QWORD	0000000FFFFFF00FFh
		QWORD	000000000FFFF00FFh
		QWORD	0FFFFFFFFFFFF0000h
		QWORD	0FFFFFF00FFFF0000h
		QWORD	0FFFF00FFFFFF0000h
		QWORD	0FFFF0000FFFF0000h
		QWORD	0FF00FFFFFFFF0000h
		QWORD	0FF00FF00FFFF0000h
		QWORD	0FF0000FFFFFF0000h
		QWORD	0FF000000FFFF0000h
		QWORD	000FFFFFFFFFF0000h
		QWORD	000FFFF00FFFF0000h
		QWORD	000FF00FFFFFF0000h
		QWORD	000FF0000FFFF0000h
		QWORD	00000FFFFFFFF0000h
		QWORD	00000FF00FFFF0000h
		QWORD	0000000FFFFFF0000h
		QWORD	000000000FFFF0000h
		QWORD	0FFFFFFFFFF00FFFFh
		QWORD	0FFFFFF00FF00FFFFh
		QWORD	0FFFF00FFFF00FFFFh
		QWORD	0FFFF0000FF00FFFFh
		QWORD	0FF00FFFFFF00FFFFh
		QWORD	0FF00FF00FF00FFFFh
		QWORD	0FF0000FFFF00FFFFh
		QWORD	0FF000000FF00FFFFh
		QWORD	000FFFFFFFF00FFFFh
		QWORD	000FFFF00FF00FFFFh
		QWORD	000FF00FFFF00FFFFh
		QWORD	000FF0000FF00FFFFh
		QWORD	00000FFFFFF00FFFFh
		QWORD	00000FF00FF00FFFFh
		QWORD	0000000FFFF00FFFFh
		QWORD	000000000FF00FFFFh
		QWORD	0FFFFFFFFFF00FF00h
		QWORD	0FFFFFF00FF00FF00h
		QWORD	0FFFF00FFFF00FF00h
		QWORD	0FFFF0000FF00FF00h
		QWORD	0FF00FFFFFF00FF00h
		QWORD	0FF00FF00FF00FF00h
		QWORD	0FF0000FFFF00FF00h
		QWORD	0FF000000FF00FF00h
		QWORD	000FFFFFFFF00FF00h
		QWORD	000FFFF00FF00FF00h
		QWORD	000FF00FFFF00FF00h
		QWORD	000FF0000FF00FF00h
		QWORD	00000FFFFFF00FF00h
		QWORD	00000FF00FF00FF00h
		QWORD	0000000FFFF00FF00h
		QWORD	000000000FF00FF00h
		QWORD	0FFFFFFFFFF0000FFh
		QWORD	0FFFFFF00FF0000FFh
		QWORD	0FFFF00FFFF0000FFh
		QWORD	0FFFF0000FF0000FFh
		QWORD	0FF00FFFFFF0000FFh
		QWORD	0FF00FF00FF0000FFh
		QWORD	0FF0000FFFF0000FFh
		QWORD	0FF000000FF0000FFh
		QWORD	000FFFFFFFF0000FFh
		QWORD	000FFFF00FF0000FFh
		QWORD	000FF00FFFF0000FFh
		QWORD	000FF0000FF0000FFh
		QWORD	00000FFFFFF0000FFh
		QWORD	00000FF00FF0000FFh
		QWORD	0000000FFFF0000FFh
		QWORD	000000000FF0000FFh
		QWORD	0FFFFFFFFFF000000h
		QWORD	0FFFFFF00FF000000h
		QWORD	0FFFF00FFFF000000h
		QWORD	0FFFF0000FF000000h
		QWORD	0FF00FFFFFF000000h
		QWORD	0FF00FF00FF000000h
		QWORD	0FF0000FFFF000000h
		QWORD	0FF000000FF000000h
		QWORD	000FFFFFFFF000000h
		QWORD	000FFFF00FF000000h
		QWORD	000FF00FFFF000000h
		QWORD	000FF0000FF000000h
		QWORD	00000FFFFFF000000h
		QWORD	00000FF00FF000000h
		QWORD	0000000FFFF000000h
		QWORD	000000000FF000000h
		QWORD	0FFFFFFFF00FFFFFFh
		QWORD	0FFFFFF0000FFFFFFh
		QWORD	0FFFF00FF00FFFFFFh
		QWORD	0FFFF000000FFFFFFh
		QWORD	0FF00FFFF00FFFFFFh
		QWORD	0FF00FF0000FFFFFFh
		QWORD	0FF0000FF00FFFFFFh
		QWORD	0FF00000000FFFFFFh
		QWORD	000FFFFFF00FFFFFFh
		QWORD	000FFFF0000FFFFFFh
		QWORD	000FF00FF00FFFFFFh
		QWORD	000FF000000FFFFFFh
		QWORD	00000FFFF00FFFFFFh
		QWORD	00000FF0000FFFFFFh
		QWORD	0000000FF00FFFFFFh
		QWORD	00000000000FFFFFFh
		QWORD	0FFFFFFFF00FFFF00h
		QWORD	0FFFFFF0000FFFF00h
		QWORD	0FFFF00FF00FFFF00h
		QWORD	0FFFF000000FFFF00h
		QWORD	0FF00FFFF00FFFF00h
		QWORD	0FF00FF0000FFFF00h
		QWORD	0FF0000FF00FFFF00h
		QWORD	0FF00000000FFFF00h
		QWORD	000FFFFFF00FFFF00h
		QWORD	000FFFF0000FFFF00h
		QWORD	000FF00FF00FFFF00h
		QWORD	000FF000000FFFF00h
		QWORD	00000FFFF00FFFF00h
		QWORD	00000FF0000FFFF00h
		QWORD	0000000FF00FFFF00h
		QWORD	00000000000FFFF00h
		QWORD	0FFFFFFFF00FF00FFh
		QWORD	0FFFFFF0000FF00FFh
		QWORD	0FFFF00FF00FF00FFh
		QWORD	0FFFF000000FF00FFh
		QWORD	0FF00FFFF00FF00FFh
		QWORD	0FF00FF0000FF00FFh
		QWORD	0FF0000FF00FF00FFh
		QWORD	0FF00000000FF00FFh
		QWORD	000FFFFFF00FF00FFh
		QWORD	000FFFF0000FF00FFh
		QWORD	000FF00FF00FF00FFh
		QWORD	000FF000000FF00FFh
		QWORD	00000FFFF00FF00FFh
		QWORD	00000FF0000FF00FFh
		QWORD	0000000FF00FF00FFh
		QWORD	00000000000FF00FFh
		QWORD	0FFFFFFFF00FF0000h
		QWORD	0FFFFFF0000FF0000h
		QWORD	0FFFF00FF00FF0000h
		QWORD	0FFFF000000FF0000h
		QWORD	0FF00FFFF00FF0000h
		QWORD	0FF00FF0000FF0000h
		QWORD	0FF0000FF00FF0000h
		QWORD	0FF00000000FF0000h
		QWORD	000FFFFFF00FF0000h
		QWORD	000FFFF0000FF0000h
		QWORD	000FF00FF00FF0000h
		QWORD	000FF000000FF0000h
		QWORD	00000FFFF00FF0000h
		QWORD	00000FF0000FF0000h
		QWORD	0000000FF00FF0000h
		QWORD	00000000000FF0000h
		QWORD	0FFFFFFFF0000FFFFh
		QWORD	0FFFFFF000000FFFFh
		QWORD	0FFFF00FF0000FFFFh
		QWORD	0FFFF00000000FFFFh
		QWORD	0FF00FFFF0000FFFFh
		QWORD	0FF00FF000000FFFFh
		QWORD	0FF0000FF0000FFFFh
		QWORD	0FF0000000000FFFFh
		QWORD	000FFFFFF0000FFFFh
		QWORD	000FFFF000000FFFFh
		QWORD	000FF00FF0000FFFFh
		QWORD	000FF00000000FFFFh
		QWORD	00000FFFF0000FFFFh
		QWORD	00000FF000000FFFFh
		QWORD	0000000FF0000FFFFh
		QWORD	0000000000000FFFFh
		QWORD	0FFFFFFFF0000FF00h
		QWORD	0FFFFFF000000FF00h
		QWORD	0FFFF00FF0000FF00h
		QWORD	0FFFF00000000FF00h
		QWORD	0FF00FFFF0000FF00h
		QWORD	0FF00FF000000FF00h
		QWORD	0FF0000FF0000FF00h
		QWORD	0FF0000000000FF00h
		QWORD	000FFFFFF0000FF00h
		QWORD	000FFFF000000FF00h
		QWORD	000FF00FF0000FF00h
		QWORD	000FF00000000FF00h
		QWORD	00000FFFF0000FF00h
		QWORD	00000FF000000FF00h
		QWORD	0000000FF0000FF00h
		QWORD	0000000000000FF00h
		QWORD	0FFFFFFFF000000FFh
		QWORD	0FFFFFF00000000FFh
		QWORD	0FFFF00FF000000FFh
		QWORD	0FFFF0000000000FFh
		QWORD	0FF00FFFF000000FFh
		QWORD	0FF00FF00000000FFh
		QWORD	0FF0000FF000000FFh
		QWORD	0FF000000000000FFh
		QWORD	000FFFFFF000000FFh
		QWORD	000FFFF00000000FFh
		QWORD	000FF00FF000000FFh
		QWORD	000FF0000000000FFh
		QWORD	00000FFFF000000FFh
		QWORD	00000FF00000000FFh
		QWORD	0000000FF000000FFh
		QWORD	000000000000000FFh
		QWORD	0FFFFFFFF00000000h
		QWORD	0FFFFFF0000000000h
		QWORD	0FFFF00FF00000000h
		QWORD	0FFFF000000000000h
		QWORD	0FF00FFFF00000000h
		QWORD	0FF00FF0000000000h
		QWORD	0FF0000FF00000000h
		QWORD	0FF00000000000000h
		QWORD	000FFFFFF00000000h
		QWORD	000FFFF0000000000h
		QWORD	000FF00FF00000000h
		QWORD	000FF000000000000h
		QWORD	00000FFFF00000000h
		QWORD	00000FF0000000000h
		QWORD	0000000FF00000000h
		QWORD	00000000000000000h

; This is the mask used to invert the lower byte of each halfword in
; the selector mask.  We don't want to invert the upper bytes, that
; would turn them into negative numbers!
grayInvMask	QWORD	000FF00FF00FF00FFh

; Here is the table used to expand each nibble of a byte into a byte, then
; expand each byte into 2 bytes.  This is used to turn 4 bit selectors
; into 16-bit multipliers.
; Yeah, you could get the same numbers with a smaller table and more
; operations.  This is a 2KB table.  Big deal.
graySelMask	QWORD	00000000000000000h
		QWORD	00000000000000011h
		QWORD	00000000000000022h
		QWORD	00000000000000033h
		QWORD	00000000000000044h
		QWORD	00000000000000055h
		QWORD	00000000000000066h
		QWORD	00000000000000077h
		QWORD	00000000000000088h
		QWORD	00000000000000099h
		QWORD	000000000000000AAh
		QWORD	000000000000000BBh
		QWORD	000000000000000CCh
		QWORD	000000000000000DDh
		QWORD	000000000000000EEh
		QWORD	000000000000000FFh
		QWORD	00000000000110000h
		QWORD	00000000000110011h
		QWORD	00000000000110022h
		QWORD	00000000000110033h
		QWORD	00000000000110044h
		QWORD	00000000000110055h
		QWORD	00000000000110066h
		QWORD	00000000000110077h
		QWORD	00000000000110088h
		QWORD	00000000000110099h
		QWORD	000000000001100AAh
		QWORD	000000000001100BBh
		QWORD	000000000001100CCh
		QWORD	000000000001100DDh
		QWORD	000000000001100EEh
		QWORD	000000000001100FFh
		QWORD	00000000000220000h
		QWORD	00000000000220011h
		QWORD	00000000000220022h
		QWORD	00000000000220033h
		QWORD	00000000000220044h
		QWORD	00000000000220055h
		QWORD	00000000000220066h
		QWORD	00000000000220077h
		QWORD	00000000000220088h
		QWORD	00000000000220099h
		QWORD	000000000002200AAh
		QWORD	000000000002200BBh
		QWORD	000000000002200CCh
		QWORD	000000000002200DDh
		QWORD	000000000002200EEh
		QWORD	000000000002200FFh
		QWORD	00000000000330000h
		QWORD	00000000000330011h
		QWORD	00000000000330022h
		QWORD	00000000000330033h
		QWORD	00000000000330044h
		QWORD	00000000000330055h
		QWORD	00000000000330066h
		QWORD	00000000000330077h
		QWORD	00000000000330088h
		QWORD	00000000000330099h
		QWORD	000000000003300AAh
		QWORD	000000000003300BBh
		QWORD	000000000003300CCh
		QWORD	000000000003300DDh
		QWORD	000000000003300EEh
		QWORD	000000000003300FFh
		QWORD	00000000000440000h
		QWORD	00000000000440011h
		QWORD	00000000000440022h
		QWORD	00000000000440033h
		QWORD	00000000000440044h
		QWORD	00000000000440055h
		QWORD	00000000000440066h
		QWORD	00000000000440077h
		QWORD	00000000000440088h
		QWORD	00000000000440099h
		QWORD	000000000004400AAh
		QWORD	000000000004400BBh
		QWORD	000000000004400CCh
		QWORD	000000000004400DDh
		QWORD	000000000004400EEh
		QWORD	000000000004400FFh
		QWORD	00000000000550000h
		QWORD	00000000000550011h
		QWORD	00000000000550022h
		QWORD	00000000000550033h
		QWORD	00000000000550044h
		QWORD	00000000000550055h
		QWORD	00000000000550066h
		QWORD	00000000000550077h
		QWORD	00000000000550088h
		QWORD	00000000000550099h
		QWORD	000000000005500AAh
		QWORD	000000000005500BBh
		QWORD	000000000005500CCh
		QWORD	000000000005500DDh
		QWORD	000000000005500EEh
		QWORD	000000000005500FFh
		QWORD	00000000000660000h
		QWORD	00000000000660011h
		QWORD	00000000000660022h
		QWORD	00000000000660033h
		QWORD	00000000000660044h
		QWORD	00000000000660055h
		QWORD	00000000000660066h
		QWORD	00000000000660077h
		QWORD	00000000000660088h
		QWORD	00000000000660099h
		QWORD	000000000006600AAh
		QWORD	000000000006600BBh
		QWORD	000000000006600CCh
		QWORD	000000000006600DDh
		QWORD	000000000006600EEh
		QWORD	000000000006600FFh
		QWORD	00000000000770000h
		QWORD	00000000000770011h
		QWORD	00000000000770022h
		QWORD	00000000000770033h
		QWORD	00000000000770044h
		QWORD	00000000000770055h
		QWORD	00000000000770066h
		QWORD	00000000000770077h
		QWORD	00000000000770088h
		QWORD	00000000000770099h
		QWORD	000000000007700AAh
		QWORD	000000000007700BBh
		QWORD	000000000007700CCh
		QWORD	000000000007700DDh
		QWORD	000000000007700EEh
		QWORD	000000000007700FFh
		QWORD	00000000000880000h
		QWORD	00000000000880011h
		QWORD	00000000000880022h
		QWORD	00000000000880033h
		QWORD	00000000000880044h
		QWORD	00000000000880055h
		QWORD	00000000000880066h
		QWORD	00000000000880077h
		QWORD	00000000000880088h
		QWORD	00000000000880099h
		QWORD	000000000008800AAh
		QWORD	000000000008800BBh
		QWORD	000000000008800CCh
		QWORD	000000000008800DDh
		QWORD	000000000008800EEh
		QWORD	000000000008800FFh
		QWORD	00000000000990000h
		QWORD	00000000000990011h
		QWORD	00000000000990022h
		QWORD	00000000000990033h
		QWORD	00000000000990044h
		QWORD	00000000000990055h
		QWORD	00000000000990066h
		QWORD	00000000000990077h
		QWORD	00000000000990088h
		QWORD	00000000000990099h
		QWORD	000000000009900AAh
		QWORD	000000000009900BBh
		QWORD	000000000009900CCh
		QWORD	000000000009900DDh
		QWORD	000000000009900EEh
		QWORD	000000000009900FFh
		QWORD	00000000000AA0000h
		QWORD	00000000000AA0011h
		QWORD	00000000000AA0022h
		QWORD	00000000000AA0033h
		QWORD	00000000000AA0044h
		QWORD	00000000000AA0055h
		QWORD	00000000000AA0066h
		QWORD	00000000000AA0077h
		QWORD	00000000000AA0088h
		QWORD	00000000000AA0099h
		QWORD	00000000000AA00AAh
		QWORD	00000000000AA00BBh
		QWORD	00000000000AA00CCh
		QWORD	00000000000AA00DDh
		QWORD	00000000000AA00EEh
		QWORD	00000000000AA00FFh
		QWORD	00000000000BB0000h
		QWORD	00000000000BB0011h
		QWORD	00000000000BB0022h
		QWORD	00000000000BB0033h
		QWORD	00000000000BB0044h
		QWORD	00000000000BB0055h
		QWORD	00000000000BB0066h
		QWORD	00000000000BB0077h
		QWORD	00000000000BB0088h
		QWORD	00000000000BB0099h
		QWORD	00000000000BB00AAh
		QWORD	00000000000BB00BBh
		QWORD	00000000000BB00CCh
		QWORD	00000000000BB00DDh
		QWORD	00000000000BB00EEh
		QWORD	00000000000BB00FFh
		QWORD	00000000000CC0000h
		QWORD	00000000000CC0011h
		QWORD	00000000000CC0022h
		QWORD	00000000000CC0033h
		QWORD	00000000000CC0044h
		QWORD	00000000000CC0055h
		QWORD	00000000000CC0066h
		QWORD	00000000000CC0077h
		QWORD	00000000000CC0088h
		QWORD	00000000000CC0099h
		QWORD	00000000000CC00AAh
		QWORD	00000000000CC00BBh
		QWORD	00000000000CC00CCh
		QWORD	00000000000CC00DDh
		QWORD	00000000000CC00EEh
		QWORD	00000000000CC00FFh
		QWORD	00000000000DD0000h
		QWORD	00000000000DD0011h
		QWORD	00000000000DD0022h
		QWORD	00000000000DD0033h
		QWORD	00000000000DD0044h
		QWORD	00000000000DD0055h
		QWORD	00000000000DD0066h
		QWORD	00000000000DD0077h
		QWORD	00000000000DD0088h
		QWORD	00000000000DD0099h
		QWORD	00000000000DD00AAh
		QWORD	00000000000DD00BBh
		QWORD	00000000000DD00CCh
		QWORD	00000000000DD00DDh
		QWORD	00000000000DD00EEh
		QWORD	00000000000DD00FFh
		QWORD	00000000000EE0000h
		QWORD	00000000000EE0011h
		QWORD	00000000000EE0022h
		QWORD	00000000000EE0033h
		QWORD	00000000000EE0044h
		QWORD	00000000000EE0055h
		QWORD	00000000000EE0066h
		QWORD	00000000000EE0077h
		QWORD	00000000000EE0088h
		QWORD	00000000000EE0099h
		QWORD	00000000000EE00AAh
		QWORD	00000000000EE00BBh
		QWORD	00000000000EE00CCh
		QWORD	00000000000EE00DDh
		QWORD	00000000000EE00EEh
		QWORD	00000000000EE00FFh
		QWORD	00000000000FF0000h
		QWORD	00000000000FF0011h
		QWORD	00000000000FF0022h
		QWORD	00000000000FF0033h
		QWORD	00000000000FF0044h
		QWORD	00000000000FF0055h
		QWORD	00000000000FF0066h
		QWORD	00000000000FF0077h
		QWORD	00000000000FF0088h
		QWORD	00000000000FF0099h
		QWORD	00000000000FF00AAh
		QWORD	00000000000FF00BBh
		QWORD	00000000000FF00CCh
		QWORD	00000000000FF00DDh
		QWORD	00000000000FF00EEh
		QWORD	00000000000FF00FFh

_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no initialized variables
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcgrybnm.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.2  1998/01/13 23:58:15  ddavies
pcgrybnm.asm -> /product/ipcore/ipshared/src/RCS/pcgrybnm.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.1  1997/06/06 04:22:00  ddavies
Initial revision




^

ENDIF


