;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcgsrot.asm:  template for assembly code.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The 386/486 processors can reload source words quickly.

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.


_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcgsrot.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:19 $"
	BYTE	0
_DATA	ENDS

; Register Definitions
; EAX:	scratch
; EBX:	scratch 
; ECX:	gray word
; EDX:	source bytes per line
; EBP:	points to variables on the stack
; ESI:	pointer to source image
; EDI:	y position of current input pixel in output's frame of reference
; ESP:	stack pointer

rSrcAddr	EQU esi
rGray		EQU ecx
rGrayLower	EQU cl
rSrcPitch	EQU edx
ryIOCoord	EQU edi

; constants
XF_MASK		EQU 3800H
XI_INC		EQU 4000H
YF_MASK		EQU 700H
YI_INC		EQU 800H
PIX_MASK	EQU 0FEH
WORD_MASK	EQU 0FFFFH

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPWORD TYPEDEF NEAR PTR WORD
NPDWORD TYPEDEF NEAR PTR DWORD


; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_rotateGrayLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by ffloyd.c

; Local Variables:
; srcAddr:	address of left end of current source line
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; dstAddr:	address of left end of current destination line
; dstPitch:	number of bytes to be added to address to move down one line
;		in the destination (output) image
; imageWidth:	width of stipe (in pixels)
; sinRotAngle:	sin of rotation angle
; xIOCoord:	x position of current input pixel in output's frame of ref
; yIOCoord:	y position of current input pixel in output's frame of ref
; xCount:	increment this count at each move towards the right
; yCount:	increment this count at each move towards the bottom
; grayRotTable: table for gray rotation
; hS		height of source image.  This is a late addition.

; See the comments in migsrot.c for a description of the
; algorithm

ip_rotateGrayLineStart::
ip_rotateGrayLine	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, dstAddr:DWORD, \
	imageWidth:DWORD, xVInc:DWORD, sinRotAngle:DWORD, \
	xIOCoord:DWORD, yIOCoord:DWORD, xCount:DWORD, yCount:DWORD, \
	grayRotTable:DWORD, hS:DWORD
LOCAL	col:DWORD, pTopLineTab:DWORD, pBotLineTab:DWORD

; make the assembler happy
	mov	eax, pIPState

; From migsrot.c:
; Just because it's important in what's coming up, let's set the
; format of the table lookup index:
; |   13   12   11    |  10     9      8  | 7 6 5 4 3 2 1 | 0 |
; | 3b xIOCoord (fff) | 3b yIOCoord (fff) |  7b pixel val | 0 |
;
; The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
; yIOCoord bits are its 3 MS fraction bits.  The 7 pixel bits are
; the 7 MS pixel bits.
; The 0 at the end makes this an index into a table of short ints.
; We'd like to have the three MS fraction bits in xIOC aligned with
; their position in the table lookup.

; align the three MS fraction bits with their position in the table lookup
	sar	xIOCoord, 2
	mov	ryIOCoord, yIOCoord
	sar	ryIOCoord, 5
	mov	eax, xVInc
	sar	eax, 2
	mov	xVInc, eax
; align sinRotAngle with yIOCoord since they are added together in the 
; inner loop.  Since yIOCoord is shifted by 5, sinRotAngle should be
; shifted by 5.
	mov	eax, sinRotAngle
	sar	eax, 5	
	mov	sinRotAngle, eax

; initialize the reg making up the gray box to 0 since no
; input pixels have been inserted yet
	mov	rGray, 0

; initialize loop count
	mov	col, 0

; and finally setup the source information
	mov	rSrcAddr, srcAddr
	mov	rSrcPitch, srcPitch

gsrotRowLoop:

; Here's where we check to see if only part of the block lies outside the
; input image.  If so, we'll have to look at each input pixel individually.
; We test to see if we're begining to the left/right of the image, beginning
; above/below the image or ending above/below the image.
; The conditions for entering the checking loop are:
; 1. yCount - 3 is negative (might rise up above image in this block)
; 2. yCount + 2 is greater than or equal hS (normally we go down 1,
;		if we backup 1, we might go down 2).  Since we do an
;		unsigned comparison, this also catches all of the cases where
;		yCount is negative.
; 3. xCount - 1 is negative (starting to the left of image when potential
;		backup is taken into account)
; 4. xCount + 10 is greater than wS (sample could end to right of image)

	mov	eax, yCount		; check to see if yCount + 2
	add	eax, 2			; is negative or greater than hS
	cmp	eax, hS
	jae	gsrotCheckBlock		; yes, go do it the hard way

	add	eax, (-2-3)
	cmp	eax, hS
	jae	gsrotCheckBlock		; yes, go do it the hard way

	mov	eax, xCount		; check xCount+10
	add	eax, 10
	cmp	eax, imageWidth		; too big?
	jae	gsrotCheckBlock		; yes, go do it the hard way

	add	eax, (-10-1)		; check xCount-1
	cmp	eax, imageWidth		; is out of bounds?
	jb	gsrotNoCheckBlock	; all ok, use the no-checking loop

; at least part of the input is outside the input image
gsrotCheckBlock:
	mov	eax, grayRotTable	; set up pBotLineTab	
	mov	ebx, xIOCoord
	and	ebx, XF_MASK
	add	eax, ebx
	mov	pBotLineTab, eax

	add	eax, YF_MASK		; setup pTopLineTab
	mov	pTopLineTab, eax

	mov	eax, xIOCoord
	cmp	eax, 0			; are we in column 0?
	js	gsrotCheckColM1		; no, we're in column -1

	sub	eax, XI_INC		; xIOCoord -= xInc
	mov	xIOCoord, eax
	dec	rSrcAddr		; rSrcAddr--
	dec	xCount			; xCount--
	sub	ryIOCoord, sinRotAngle	; yIOCoord -= sinRotAngle
	mov	ebx, ryIOCoord		; is yIOCoord + YI_INC < 0?
	add	ebx, YI_INC
	jns	gsrotCheckColM1
; we are above row -1
	mov	ryIOCoord, ebx		; yIOCoord += YI_INC
	mov	eax, xIOCoord		; xIOCoord -= xVInc
	sub	eax, xVInc
	mov	xIOCoord, eax
	add	rSrcAddr, rSrcPitch	; point to row 0
	inc	yCount			; the common case

; column -1
gsrotCheckColM1:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrotM1TopOffImage	; yes, use white space
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrotM1TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrotM1FinishTopPix

; top pixel is off the image so use max white value
gsrotM1TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrotM1FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx

gsrotM1BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrotM1BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gsrotM1BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrotM1FinishBotPix

; bottom pixel is off the image so use max white value
gsrotM1BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrotM1FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckColM1Row0	; yes, go do some fixup

; column 0
gsrotCheckCol0:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot0TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot0TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot0FinishTopPix

; top pixel is off the image so use max white value
gsrot0TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot0FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

gsrot0BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot0BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot0BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot0FinishBotPix

; bottom pixel is off the image so use max white value
gsrot0BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot0FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckCol0Row0	; yes, go do some fixup

; column 1
gsrotCheckCol1:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot1TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot1TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot1FinishTopPix

; top pixel is off the image so use max white value
gsrot1TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot1FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

gsrot1BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot1BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot1BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot1FinishBotPix

; bottom pixel is off the image so use max white value
gsrot1BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot1FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckCol1Row0	; yes, go do some fixup

; column 2
gsrotCheckCol2:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot2TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot2TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot2FinishTopPix

; top pixel is off the image so use max white value
gsrot2TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot2FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

gsrot2BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot2BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot2BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot2FinishBotPix

; bottom pixel is off the image so use max white value
gsrot2BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot2FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckCol2Row0	; yes, go do some fixup

; column 3
gsrotCheckCol3:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot3TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot3TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot3FinishTopPix

; top pixel is off the image so use max white value
gsrot3TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot3FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	push	ebx			; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx

gsrot3BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot3BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot3BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot3FinishBotPix

; bottom pixel is off the image so use max white value
gsrot3BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot3FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	mov	eax, ebx		; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add 	rGray, ebx

	mov	ebx, dstAddr
	shl	eax, 24			; align gray value from bottom line
	mov	[ebx], rGray		; write first word
	pop	rGray			; get gray value from top line
	shl	rGray, 24		; align gray value
	add	rGray, eax		; apply gray value

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckCol3Row0	; yes, go do some fixup

; column 4
gsrotCheckCol4:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot4TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot4TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot4FinishTopPix

; top pixel is off the image so use max white value
gsrot4TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot4FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

gsrot4BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot4BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot4BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot4FinishBotPix

; bottom pixel is off the image so use max white value
gsrot4BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot4FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckCol4Row0	; yes, go do some fixup

; column 5
gsrotCheckCol5:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot5TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot5TopOffImage	; yes, use white space


; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot5FinishTopPix

; top pixel is off the image so use max white value
gsrot5TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot5FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

gsrot5BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot5BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot5BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot5FinishBotPix

; bottom pixel is off the image so use max white value
gsrot5BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot5FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckCol5Row0	; yes, go do some fixup

; column 6
gsrotCheckCol6:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot6TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot6TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot6FinishTopPix

; top pixel is off the image so use max white value
gsrot6TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot6FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

gsrot6BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot6BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot6BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot6FinishBotPix

; bottom pixel is off the image so use max white value
gsrot6BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot6FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotCheckCol6Row0	; yes, go do some fixup

; column 7
gsrotCheckCol7:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; increment xCount
	mov	xCount, eax
	cmp	eax, imageWidth		; off on left/right?
	jae	gsrot7TopOffImage	
	cmp	ebx, hS			; off on top/bottom?
	jae	gsrot7TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK		; get fractional part
	and	ebx, PIX_MASK		; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot7FinishTopPix

; top pixel is off the image so use max white value
gsrot7TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + PIX_MASK]	; get gray value
gsrot7FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx

gsrot7BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gsrot7BotOffImage	; yes, use white space
	mov	ebx, yCount		; is bottom pixel on input image?
	inc	ebx
	cmp	ebx, hS			; off image on top/bottom?
	jae	gsrot7BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gsrot7FinishBotPix

; bottom pixel is off the image so use max white value
gsrot7BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + PIX_MASK]	; get gray value
gsrot7FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add 	rGray, ebx
	jmp	gsrotRowComplete

; Note that we didn't try to advance the pointer.
; We're already in the right position to start the next sweep

; this is the normal loop case.  all of the input is inside the input image
; Since we're not near the bottom of the image, we don't have to check to
; see if yCount is crossing the border when we change it.
gsrotNoCheckBlock:
	mov	eax, grayRotTable	; set up pBotLineTab	
	mov	ebx, xIOCoord
	and	ebx, XF_MASK
	add	eax, ebx
	mov	pBotLineTab, eax

	add	eax, YF_MASK		; setup pTopLineTab
	mov	pTopLineTab, eax

	mov	eax, xIOCoord
	cmp	eax, 0			; are we in column 0?
	js	gsrotNoCheckColM1	; no, we're in column -1

	sub	eax, XI_INC		; xIOCoord -= xInc
	mov	xIOCoord, eax
	dec	rSrcAddr		; rSrcAddr--
	dec	xCount			; xCount--
	sub	ryIOCoord, sinRotAngle	; yIOCoord -= sinRotAngle
	mov	ebx, ryIOCoord		; is yIOCoord + YI_INC < 0?
	add	ebx, YI_INC
	jns	gsrotNoCheckColM1
; we are above row -1
	mov	ryIOCoord, ebx		; yIOCoord += YI_INC
	mov	eax, xIOCoord		; xIOCoord -= xVInc
	sub	eax, xVInc
	mov	xIOCoord, eax
	add	rSrcAddr, rSrcPitch	; point to row 0
	inc	yCount			; yCount++

; column -1
gsrotNoCheckColM1:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotNoCheckColM1Row0	; yes, go do some fixup

; column 0
gsrotNoCheckCol0:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotNoCheckCol0Row0	; yes, go do some fixup

; column 1
gsrotNoCheckCol1:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotNoCheckCol1Row0	; yes, go do some fixup

; column 2
gsrotNoCheckCol2:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0
	jns	gsrotNoCheckCol2Row0	; yes, go do some fixup

; column 3
gsrotNoCheckCol3:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	push	ebx			; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	mov	eax, ebx		; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	ebx, dstAddr
	shl	eax, 24			; align gray value from bottom line
	mov	[ebx], rGray		; write first word
	pop	rGray			; get gray value from top line
	shl	rGray, 24		; align gray value
	add 	rGray, eax

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotNoCheckCol3Row0	; yes, go do some fixup

; column 4
gsrotNoCheckCol4:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotNoCheckCol4Row0	; yes, go do some fixup

gsrotNoCheckCol5:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotNoCheckCol5Row0	; yes, go do some fixup

gsrotNoCheckCol6:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gsrotNoCheckCol6Row0	; yes, go do some fixup

gsrotNoCheckCol7:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	ebx, PIX_MASK		; mask off LS bit
	and	eax, YF_MASK		; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	and	ebx, PIX_MASK		; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

; Note that we dind't try to advance the pointers.
; They're already in the right position to start the next sweep.

	mov	eax, xCount
	add	eax, 8
	mov	xCount, eax

; We've completed the formation of this row.  Write it out.  Note that
; we always write gray 0.  If the line ends after that don't write gray 1.
gsrotRowComplete:
	mov	eax, col		; get loop count
	add	eax, 4			; we've done 4 bytes
	cmp	eax, imageWidth		; do we need to write gray 1
	jge	gsrotDone

gsrotNextBlock:
	mov	ebx, dstAddr
	add	eax, 4			; we actually write 8 bytes each
					; loop so add 4 more to col
	mov	[ebx+4], rGray		; write destination word
	cmp	eax, imageWidth		; are we done with the line?
	jge	gsrotDone		; yes, just go quit

	mov	col, eax		; update loop count
	add	ebx, 8			; update destination address
	mov	dstAddr, ebx
	mov	rGray, 0		; clear out gray word
	jmp	gsrotRowLoop

gsrotDone:
	ret

gsrotNoCheckColM1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol0

gsrotNoCheckCol0Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol1

gsrotNoCheckCol1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol2

gsrotNoCheckCol2Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol3

gsrotNoCheckCol3Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol4

gsrotNoCheckCol4Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol5

gsrotNoCheckCol5Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol6

gsrotNoCheckCol6Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gsrotNoCheckCol7



gsrotCheckColM1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol0

gsrotCheckCol0Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol1

gsrotCheckCol1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol2

gsrotCheckCol2Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol3

gsrotCheckCol3Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol4

gsrotCheckCol4Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol5

gsrotCheckCol5Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol6

gsrotCheckCol6Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC	; yIOCoord -= YI_INC
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gsrotCheckCol7

ip_rotateGrayLine ENDP


_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcgsrot.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:19  jlin
Imported sources

Revision 1.11  1998/01/13 23:58:19  ddavies
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.10  1996/10/05 02:26:59  ddavies
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

Change the semantics of xCount and yCount.  Instead of trying to arrange
for them to be negative outside the image, we just to unsigned compares of
 xCount >= width and yCount >= height.  If these are true, we're off the image.
Also get rid of the fixup of xCount at the left boundary.

Revision 1.9  1996/09/17 00:45:40  ddavies
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.8  1996/09/16 00:51:17  ddavies
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.7  1995/12/13 01:53:57  ddavies
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

In keeping with the new, resolution-independent, method representing skew
angles, use two versions of sinRotAngle - horizSinRotAngle is used to adjust
y coords when moving horizontally and vertSinRotAngle is used to adjust
x coords when moving vertically.  Also fix a bug in which xCount was
being adjusted for use to detect the right edge of the image when it was
equal to 0.  In this case, it may still be necessary to read at pixel with
x coord = -1, but we wouldn't substitute a white pixel since xCount would
have been changed to be relative to the right edge of the image instead of
being relative to the left edge.

Revision 1.6  1995/09/15  21:47:53  ddavies
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

Revision 1.5  1994/07/27  16:02:54  ejaquith
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

Reentrancy!
Removed globals and passed in gray rotation table pointers instead.

Revision 1.4  1993/11/05  01:52:42  ddavies
pcgsrot.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot.asm,v

Add RCS magic to have string specifying this file's name, rev number and
last edit date into the object file.  Also add change log at the bottom.

^

ENDIF

