;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcgsrot.asm:  template for assembly code.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The 386/486 processors can reload source words quickly.

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.


_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcgsrot1.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:19 $"
	BYTE	0
_DATA	ENDS

; Register Definitions
; EAX:	scratch
; EBX:	scratch 
; ECX:	gray word
; EDX:	source bytes per line
; EBP:	points to variables on the stack
; ESI:	pointer to source image
; EDI:	y position of current input pixel in output's frame of reference
; ESP:	stack pointer

rSrcAddr		EQU esi
rGray			EQU ecx
rGrayH			EQU ch
rGrayL			EQU cl
rSrcPitch		EQU edx
ryIOCoord		EQU edi

; constants
XF_MASK_4BPP		EQU 700H
XI_INC_4BPP		EQU 800H
YF_MASK_4BPP		EQU 0E0H
YI_INC_4BPP		EQU 100H
MAX_WHITE_4BPP		EQU 1EH
WORD_MASK		EQU 0FFFFH

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPWORD TYPEDEF NEAR PTR WORD
NPDWORD TYPEDEF NEAR PTR DWORD


; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_rotate4bppGrayLine.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by ffloyd.c

; Local Variables:
; srcAddr:	address of left end of current source line
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; dstAddr:	address of left end of current destination line
; dstPitch:	number of bytes to be added to address to move down one line
;		in the destination (output) image
; imageWidth:	width of stipe (in pixels)
; sinRotAngle:	sin of rotation angle
; xIOCoord:	x position of current input pixel in output's frame of ref
; yIOCoord:	y position of current input pixel in output's frame of ref
; xCount:	increment this count at each move towards the right
; yCount:	increment this count at each move towards the bottom
; nibble	0 => high nibble of byte; 1 => low nibble of byte
; pGrayTable	pointer to the rotation table
; hS		height of source image.  This is a late addition.

; See the comments in sgsrot.c for a description of the
; algorithm

ip_rotate4bppGrayLineStart::
	PUBLIC	ip_rotate4bppGrayLine
ip_rotate4bppGrayLine	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcPitch:DWORD, dstAddr:DWORD, \
	imageWidth:DWORD, xVInc:DWORD, sinRotAngle:DWORD, \
	xIOCoord:DWORD, yIOCoord:DWORD, xCount:DWORD, yCount:DWORD, \
	nibble: DWORD, pGrayTable:DWORD, hS:DWORD
LOCAL	col:DWORD, pTopLineTab:DWORD, pBotLineTab:DWORD

; make the assembler happy
	mov	eax, pIPState

; From migsrot.c:
; Just because it's important in what's coming up, let's set the
; format of the table lookup index:
; |  10    9   8      |  7     6      5   | 4 3 2 1 | 0
; | 3b xIOCoord (fff) | 3b yIOCoord (fff) | pixel   |
;
; The 3 bits of xIOCoord contain the 3 MS bits of fraction.  The 3
; yIOCoord bits are its 3 MS fraction bits.
; We'd like to have the three MS fraction bits in xIOC aligned with
; their position in the table lookup.

; align the three MS fraction bits with their position in the table lookup
	sar	xIOCoord, 5
	mov	ryIOCoord, yIOCoord
	sar	ryIOCoord, 8
	mov	eax, xVInc
	sar	eax, 5
	mov	xVInc, eax
; align sinRotAngle with yIOCoord since they are added together in the 
; inner loop.  Since the yIOCoord is shifted by 8, sinRotAngle should
; be shifted by 8.
	mov	eax, sinRotAngle
	sar	eax, 8	
	mov	sinRotAngle, eax

; initialize the reg making up the gray box to 0 since no
; input pixels have been inserted yet
	mov	rGray, 0

; initialize loop count
	mov	col, 0

; and finally setup the source information
	mov	rSrcAddr, srcAddr
	mov	rSrcPitch, srcPitch

gs4rotRowLoop:

; Here's where we check to see if only part of the block lies outside the
; input image.  If so, we'll have to look at each input pixel individually.
; We test to see if we're begining to the left/right of the image, beginning
; above/below the image or ending above/below the image.
; The conditions for entering the checking loop are:
; 1. yCount - 3 is negative (might rise up above image in this block)
; 2. yCount + 2 is greater than or equal hS (normally we go down 1,
;		if we backup 1, we might go down 2).  Since we do an
;		unsigned comparison, this also catches all of the cases where
;		yCount is negative.
; 3. xCount - 1 is negative (starting to the left of image when potential
;		backup is taken into account)
; 4. xCount + 10 is greater than wS (sample could end to right of image)

	mov	eax, yCount		; check to see if yCount + 2
	add	eax, 2			; is negative or greater than hS
	cmp	eax, hS
	jae	gs4rotCheckBlock	; yes, go do it the hard way

	add	eax, (-2-3)
	cmp	eax, hS
	jae	gs4rotCheckBlock	; yes, go do it the hard way

	mov	eax, xCount		; check xCount+10
	add	eax, 10
	cmp	eax, imageWidth		; too big?
	jae	gs4rotCheckBlock	; yes, go do it the hard way

	add	eax, (-10-1)		; check xCount-1
	cmp	eax, imageWidth		; is out of bounds?
	jb	gs4rotNoCheckBlock	; all ok, use the no-checking loop


; at least part of the input is outside the input image
gs4rotCheckBlock:
	mov	eax, pGrayTable		; set up pBotLineTab	
	mov	ebx, xIOCoord
	and	ebx, XF_MASK_4BPP
	add	eax, ebx
	mov	pBotLineTab, eax

	add	eax, YF_MASK_4BPP	; setup pTopLineTab
	mov	pTopLineTab, eax

	mov	eax, xIOCoord
	cmp	eax, 0			; are we in column 0?
	js	gs4rotCheckNibble	; no, we're in column -1

	sub	eax, XI_INC_4BPP	; xIOCoord -= xInc
	mov	xIOCoord, eax
	dec	xCount			; xCount--
	mov	eax, nibble		; different nibble in the byte
	xor 	eax, 1
	mov	nibble, eax
	jz	gs4rotCheckRow
	dec	rSrcAddr		; rSrcAddr--
gs4rotCheckRow:
	sub	ryIOCoord, sinRotAngle	; yIOCoord -= sinRotAngle
	mov	ebx, ryIOCoord		; is yIOCoord + YI_INC_4BPP < 0?
	add	ebx, YI_INC_4BPP
	jns	gs4rotCheckNibble
; we are above row -1
	mov	ryIOCoord, ebx		; yIOCoord += YI_INC_4BPP
	mov	eax, xIOCoord		; xIOCoord -= xVInc
	sub	eax, xVInc
	mov	xIOCoord, eax
	add	rSrcAddr, rSrcPitch	; point to row 0
	inc	yCount			; the common case

gs4rotCheckNibble:
	cmp	nibble, 0		; high or low nibble?
	jnz	gs4rotCheckLowColM1

; column -1
gs4rotCheckColM1:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotM1TopOffImage	; yes, use white space
	cmp	ebx, hS			; off on top/bottom?
	jae	gs4rotM1TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; get high nibble
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotM1FinishTopPix

; top pixel is off the image so use max white value
gs4rotM1TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotM1FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx

gs4rotM1BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotM1BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotM1BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; high nibble
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotM1FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotM1BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotM1FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckColM1Row0	; yes, go do some fixup

; column 0
gs4rotCheckCol0:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot0TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot0TopOffImage	; yes, use white space


; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot0FinishTopPix

; top pixel is off the image so use max white value
gs4rot0TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot0FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

gs4rot0BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot0BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot0BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot0FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot0BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot0FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckCol0Row0	; yes, go do some fixup

; column 1
gs4rotCheckCol1:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot1TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot1TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot1FinishTopPix

; top pixel is off the image so use max white value
gs4rot1TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot1FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

gs4rot1BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot1BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot1BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot1FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot1BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot1FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckCol1Row0	; yes, go do some fixup

; column 2
gs4rotCheckCol2:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot2TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot2TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot2FinishTopPix

; top pixel is off the image so use max white value
gs4rot2TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot2FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

gs4rot2BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot2BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot2BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot2FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot2BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot2FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckCol2Row0	; yes, go do some fixup

; column 3
gs4rotCheckCol3:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot3TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot3TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot3FinishTopPix

; top pixel is off the image so use max white value
gs4rot3TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot3FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	push	ebx			; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx

gs4rot3BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot3BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot3BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot3FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot3BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot3FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	mov	eax, ebx		; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add 	rGray, ebx

	add	rGray, 08080808H	; round to next pixel value
	and	rGray, 0F0F0F0F0H	; clear unused bits
	mov	ebx, rGray		; get into temp
	shl	ebx, 4			; pair nibbles into bytes
	or	rGray, ebx
	mov	ebx, dstAddr		; get ready for write
	shl	eax, 24			; align gray value from bottom line
	mov	[ebx+2], rGrayH		; write first byte
					; watch out for byte swapping
	shr	rGray, 16		; get next nibble into position
	mov	[ebx+3], rGrayH		; write second byte
					; watch out for byte swapping
	pop	rGray			; get gray value from top line
	shl	rGray, 24		; align gray value
	add	rGray, eax		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckCol3Row0	; yes, go do some fixup

; column 4
gs4rotCheckCol4:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot4TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot4TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot4FinishTopPix

; top pixel is off the image so use max white value
gs4rot4TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot4FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

gs4rot4BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot4BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot4BotOffImage	; yes, use white space


; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot4FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot4BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot4FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckCol4Row0	; yes, go do some fixup

; column 5
gs4rotCheckCol5:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot5TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot5TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot5FinishTopPix

; top pixel is off the image so use max white value
gs4rot5TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot5FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

gs4rot5BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot5BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot5BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot5FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot5BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot5FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckCol5Row0	; yes, go do some fixup

; column 6
gs4rotCheckCol6:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot6TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot6TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot6FinishTopPix

; top pixel is off the image so use max white value
gs4rot6TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot6FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

gs4rot6BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot6BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot6BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot6FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot6BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot6FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotCheckCol6Row0	; yes, go do some fixup

; column 7
gs4rotCheckCol7:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rot7TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rot7TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot7FinishTopPix

; top pixel is off the image so use max white value
gs4rot7TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rot7FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx

gs4rot7BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rot7BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rot7BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rot7FinishBotPix

; bottom pixel is off the image so use max white value
gs4rot7BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rot7FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add 	rGray, ebx
	jmp	gs4rotRowComplete

; Note that we didn't try to advance the pointer.
; We're already in the right position to start the next sweep



; The is the code for checking but the first pixel is in the low nibble
; of the source byte
gs4rotCheckLowColM1:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLowM1TopOffImage	; yes, use white space
	cmp	ebx, hS			; off on top/bottom?
	jae	gs4rotLowM1TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; get low nibble
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLowM1FinishTopPix

; top pixel is off the image so use max white value
gs4rotLowM1TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLowM1FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx

gs4rotLowM1BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLowM1BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLowM1BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; get low nibble
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLowM1FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLowM1BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLowM1FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckColM1Row0	; yes, go do some fixup

; column 0
gs4rotLowCheckCol0:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow0TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow0TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow0FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow0TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow0FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

gs4rotLow0BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow0BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow0BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow0FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow0BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow0FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckCol0Row0	; yes, go do some fixup

; column 1
gs4rotLowCheckCol1:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow1TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow1TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow1FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow1TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow1FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

gs4rotLow1BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow1BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow1BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow1FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow1BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow1FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckCol1Row0	; yes, go do some fixup

; column 2
gs4rotLowCheckCol2:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow2TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow2TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow2FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow2TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow2FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

gs4rotLow2BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow2BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow2BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow2FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow2BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow2FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckCol2Row0	; yes, go do some fixup

; column 3
gs4rotLowCheckCol3:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow3TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow3TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow3FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow3TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP	; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow3FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	push	ebx			; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx

gs4rotLow3BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow3BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow3BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow3FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow3BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow3FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	mov	eax, ebx		; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add 	rGray, ebx

	add	rGray, 08080808H	; round to next pixel value
	and	rGray, 0F0F0F0F0H	; clear unused bits
	mov	ebx, rGray		; get into temp
	shl	ebx, 4			; pair nibbles into bytes
	or	rGray, ebx
	mov	ebx, dstAddr		; get ready for write
	shl	eax, 24			; align gray value from bottom line
	mov	[ebx+2], rGrayH		; write first byte
					; watch out for byte swapping
	shr	rGray, 16		; get next nibble into position
	mov	[ebx+3], rGrayH		; write second byte
					; watch out for byte swapping
	pop	rGray			; get gray value from top line
	shl	rGray, 24		; align gray value
	add	rGray, eax		; apply gray value

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckCol3Row0	; yes, go do some fixup

; column 4
gs4rotLowCheckCol4:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow4TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow4TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow4FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow4TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow4FinishTopPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

gs4rotLow4BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow4BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow4BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow4FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow4BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow4FinishBotPix:
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckCol4Row0	; yes, go do some fixup

; column 5
gs4rotLowCheckCol5:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow5TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow5TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow5FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow5TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow5FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

gs4rotLow5BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow5BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow5BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow5FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow5BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow5FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx

	inc	rSrcAddr		; point to next pixel just in case

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckCol5Row0	; yes, go do some fixup

; column 6
gs4rotLowCheckCol6:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow6TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow6TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow6FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow6TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow6FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

gs4rotLow6BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow6BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow6BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow6FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow6BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow6FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowCheckCol6Row0	; yes, go do some fixup

; column 7
gs4rotLowCheckCol7:
	mov	eax, xCount		; is top pixel on input image?
	mov	ebx, yCount
	inc	eax			; next pixel
	cmp	ebx, hS			; off on top/bottom?
	mov	xCount, eax
	jae	gs4rotLow7TopOffImage	; yes, use white space
	cmp	eax, imageWidth		; off on left/right?
	jae	gs4rotLow7TopOffImage	; yes, use white space

; top pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	and	eax, YF_MASK_4BPP	; get fractional part
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow7FinishTopPix

; top pixel is off the image so use max white value
gs4rotLow7TopOffImage:
	mov	eax, ryIOCoord		; get ready for table access
	and	eax, YF_MASK_4BPP		; get fractional part
	mov	ebx, eax		; save for use by bottom pixel
	neg	ebx
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx + MAX_WHITE_4BPP]	; get gray value
gs4rotLow7FinishTopPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx

gs4rotLow7BotPix:
	mov	ebx, xCount		; is bottom pixel on input image?
	cmp	ebx, imageWidth		; off on left/right?
	jae	gs4rotLow7BotOffImage	; yes, use white space
	mov	ebx, yCount
	inc	ebx
	cmp	ebx, hS			; bottom off image?
	jae	gs4rotLow7BotOffImage	; yes, use white space

; bottom pixel is on the image
	xor	rSrcAddr, 3		; byte swapping
	mov	bl, [rSrcAddr + rSrcPitch]	; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; byte swapping
	mov	bx, [ebx]		; get gray value
	jmp	gs4rotLow7FinishBotPix

; bottom pixel is off the image so use max white value
gs4rotLow7BotOffImage:
	add	eax, pBotLineTab	; get table address of bottom line
	mov	bx, [eax + MAX_WHITE_4BPP]	; get gray value
gs4rotLow7FinishBotPix:
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add 	rGray, ebx
	jmp	gs4rotRowComplete


; ***************************** NORMAL LOOP **************************

; this is the normal loop case.  all of the input is inside the input image
gs4rotNoCheckBlock:
	mov	eax, pGrayTable		; set up pBotLineTab	
	mov	ebx, xIOCoord
	and	ebx, XF_MASK_4BPP
	add	eax, ebx
	mov	pBotLineTab, eax

	add	eax, YF_MASK_4BPP	; setup pTopLineTab
	mov	pTopLineTab, eax

	mov	eax, xIOCoord
	cmp	eax, 0			; are we in column 0?
	js	gs4rotNoCheckNibble	; no, we're in column -1

	sub	eax, XI_INC_4BPP	; xIOCoord -= xInc
	mov	xIOCoord, eax
	dec	xCount			; xCount--
	mov	eax, nibble		; different nibble in the byte
	xor 	eax, 1
	mov	nibble, eax
	jz	gs4rotNoCheckRow
	dec	rSrcAddr		; rSrcAddr--
gs4rotNoCheckRow:
	sub	ryIOCoord, sinRotAngle	; yIOCoord -= sinRotAngle
	mov	ebx, ryIOCoord		; is yIOCoord + YI_INC_4BPP < 0?
	add	ebx, YI_INC_4BPP
	jns	gs4rotNoCheckNibble
; we are above row -1
	mov	ryIOCoord, ebx		; yIOCoord += YI_INC_4BPP
	mov	eax, xIOCoord		; xIOCoord -= xVInc
	sub	eax, xVInc
	mov	xIOCoord, eax
	add	rSrcAddr, rSrcPitch	; point to row 0
	inc	yCount			; yCount++

gs4rotNoCheckNibble:
	cmp	nibble, 0		; high nibble?
	jnz	gs4rotLowNoChkColM1

; column -1
gs4rotNoCheckColM1:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotNoCheckColM1Row0	; yes, entered row 0

; column 0
gs4rotNoCheckCol0:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotNoCheckCol0Row0	; yes, entered row 0

; column 1
gs4rotNoCheckCol1:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotNoCheckCol1Row0	; yes, go do some fixup

; column 2
gs4rotNoCheckCol2:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0
	jns	gs4rotNoCheckCol2Row0	; yes, go do some fixup

; column 3
gs4rotNoCheckCol3:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	push	ebx			; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	mov	eax, ebx		; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	add	rGray, 08080808H	; round to next pixel value
	and	rGray, 0F0F0F0F0H	; clear unused bits
	mov	ebx, rGray		; get into temp
	shl	ebx, 4			; pair nibbles into bytes
	or	rGray, ebx
	mov	ebx, dstAddr		; get ready for write
	shl	eax, 24			; align gray value from bottom line
	mov	[ebx+2], rGrayH		; write first byte
					; watch out for byte swapping
	shr	rGray, 16		; get next nibble into position
	mov	[ebx+3], rGrayH		; write second byte
					; watch out for byte swapping
	pop	rGray			; get gray value from top line
	shl	rGray, 24		; align gray value
	add 	rGray, eax

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotNoCheckCol3Row0	; yes, go do some fixup

; column 4
gs4rotNoCheckCol4:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get the low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	add	ebx, ebx		; get the low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotNoCheckCol4Row0	; yes, go do some fixup

gs4rotNoCheckCol5:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotNoCheckCol5Row0	; yes, go do some fixup

gs4rotNoCheckCol6:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotNoCheckCol6Row0	; yes, go do some fixup

gs4rotNoCheckCol7:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

; Note that we dind't try to advance the pointers.
; They're already in the right position to start the next sweep.

	mov	eax, xCount
	add	eax, 8
	mov	xCount, eax
	jmp	gs4rotRowComplete

; column -1
gs4rotLowNoChkColM1:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 24			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowNoChkColM1Row0	; yes, do some fixup

; column 0
gs4rotLowNoChkCol0:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off LS bit
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowNoChkCol0Row0	; yes, go do some fixup

; column 1
gs4rotLowNoChkCol1:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off high nibble
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowNoChkCol1Row0	; yes, go do some fixup

; column 2
gs4rotLowNoChkCol2:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch] ; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0
	jns	gs4rotLowNoChkCol2Row0	; yes, go do some fixup

; column 3
gs4rotLowNoChkCol3:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	push	ebx			; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	mov	eax, ebx		; save pixel for use in next output
					; word
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	add	rGray, 08080808H	; round to next pixel value
	and	rGray, 0F0F0F0F0H	; clear unused bits
	mov	ebx, rGray		; get into temp
	shl	ebx, 4			; pair nibbles into bytes
	or	rGray, ebx
	mov	ebx, dstAddr		; get ready for write
	shl	eax, 24			; align gray value from bottom line
	mov	[ebx+2], rGrayH		; write first byte
					; watch out for byte swapping
	shr	rGray, 16		; get next nibble into position
	mov	[ebx+3], rGrayH		; write second byte
					; watch out for byte swapping
	pop	rGray			; get gray value from top line
	shl	rGray, 24		; align gray value
	add 	rGray, eax

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowNoChkCol3Row0	; yes, go do some fixup

; column 4
gs4rotLowNoChkCol4:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
;	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 16			; align gray value
	add	rGray, ebx		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowNoChkCol4Row0	; yes, go do some fixup

gs4rotLowNoChkCol5:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shl	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	inc	rSrcAddr		; point to next pixel

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowNoChkCol5Row0	; yes, go do some fixup

gs4rotLowNoChkCol6:
	xor	rSrcAddr, 3		; byte swapping
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	shr	ebx, 3			; get high nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
;	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	mov	bx, [eax+ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
;	shl	ebx, 0			; align gray value
	add	rGray, ebx		; apply gray value

	add	ryIOCoord, sinRotAngle	; entered row 0?
	jns	gs4rotLowNoChkCol6Row0	; yes, go do some fixup

gs4rotLowNoChkCol7:
	mov	eax, ryIOCoord		; get ready for table access
	mov	bl, [rSrcAddr]		; get the next byte
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	and	eax, YF_MASK_4BPP	; get fractional part
	sub	ebx, eax		; table offset
	add	ebx, pTopLineTab	; get table address of top line
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

	mov	bl, [rSrcAddr+rSrcPitch]	; get bottom source pixel
	add	ebx, ebx		; get low nibble into position
	and	ebx, MAX_WHITE_4BPP	; mask off unused bits
	add	ebx, eax		; table offset
	add	ebx, pBotLineTab	; get table address of bottom line
	xor	rSrcAddr, 3		; un-byte swap
	mov	bx, [ebx]		; get gray value
	and	ebx, WORD_MASK		; get rid of top 16 bits
	shr	ebx, 8			; align gray value
	add	rGray, ebx		; apply gray value

; Note that we dind't try to advance the pointers.
; They're already in the right position to start the next sweep.

	mov	eax, xCount
	add	eax, 8
	mov	xCount, eax

gs4rotRowComplete:
gs4rotNextBlock:
	add	rGray, 08080808H	; round to next pixel value
	and	rGray, 0F0F0F0F0H	; clear unused bits
	mov	ebx, rGray		; get into temp
	shl	ebx, 4			; pair nibbles into bytes
	or	rGray, ebx
	mov	ebx, dstAddr		; get ready for write
	mov	eax, col		; get loop count
	mov	[ebx+0], rGrayH		; write first byte
					; watch out for byte swapping
	shr	rGray, 16		; get next nibble into position
	mov	[ebx+1], rGrayH		; write second byte
					; watch out for byte swapping

	add	ebx, 4			; increment dest pointer
	mov	dstAddr, ebx

	add	eax, 8			; we've done 8 pixels
	cmp	eax, imageWidth		; are we done with the line?
	jge	gs4rotDone		; yes, just go quit

	mov	col, eax		; update loop count
	mov	rGray, 0		; clear out gray word
	jmp	gs4rotRowLoop

gs4rotDone:
	ret


; ********************* Fixup blocks ************************

gs4rotNoCheckColM1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol0

gs4rotNoCheckCol0Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol1

gs4rotNoCheckCol1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol2

gs4rotNoCheckCol2Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol3

gs4rotNoCheckCol3Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol4

gs4rotNoCheckCol4Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol5

gs4rotNoCheckCol5Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol6

gs4rotNoCheckCol6Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotNoCheckCol7

gs4rotLowNoChkColM1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol0

gs4rotLowNoChkCol0Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol1

gs4rotLowNoChkCol1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol2

gs4rotLowNoChkCol2Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol3

gs4rotLowNoChkCol3Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol4

gs4rotLowNoChkCol4Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol5

gs4rotLowNoChkCol5Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol6

gs4rotLowNoChkCol6Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; yCount--
	jmp	gs4rotLowNoChkCol7



gs4rotCheckColM1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol0

gs4rotCheckCol0Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol1

gs4rotCheckCol1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol2

gs4rotCheckCol2Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol3

gs4rotCheckCol3Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol4

gs4rotCheckCol4Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol5

gs4rotCheckCol5Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol6

gs4rotCheckCol6Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount			; the common case
	jmp	gs4rotCheckCol7



gs4rotLowCheckColM1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol0

gs4rotLowCheckCol0Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol1

gs4rotLowCheckCol1Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol2

gs4rotLowCheckCol2Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol3

gs4rotLowCheckCol3Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol4

gs4rotLowCheckCol4Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol5

gs4rotLowCheckCol5Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol6

gs4rotLowCheckCol6Row0:
	sub	rSrcAddr, rSrcPitch	; move up one input row
	sub	ryIOCoord, YI_INC_4BPP	; yIOCoord -= YI_INC_4BPP
	mov	ebx, xIOCoord		; xIOCoord += xVInc
	add	ebx, xVInc
	mov	xIOCoord, ebx
	dec	yCount
	jmp	gs4rotLowCheckCol7

ip_rotate4bppGrayLine ENDP


_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcgsrot1.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:19  jlin
Imported sources

Revision 1.9  1998/01/13 23:58:22  ddavies
pcgsrot1.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot1.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.8  1996/10/05 02:27:19  ddavies
pcgsrot1.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot1.asm,v

Change the semantics of xCount and yCount.  Instead of trying to arrange
for them to be negative outside the image, we just to unsigned compares of
xCount >= width and yCount >= height.  If these are true, we're off the image.
Also get rid of the fixup of xCount at the left boundary.

Revision 1.7  1996/09/17 00:45:43  ddavies
pcgsrot1.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot1.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.6  1996/09/16 00:51:30  ddavies
pcgsrot1.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot1.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.5  1995/12/13 01:54:34  ddavies
pcgsrot1.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot1.asm,v

In keeping with the new, resolution-independent, method representing skew
angles, use two versions of sinRotAngle - horizSinRotAngle is used to adjust
y coords when moving horizontally and vertSinRotAngle is used to adjust
x coords when moving vertically.  Also fix a bug in which xCount was
being adjusted for use to detect the right edge of the image when it was
equal to 0.  In this case, it may still be necessary to read at pixel with
x coord = -1, but we wouldn't substitute a white pixel since xCount would
have been changed to be relative to the right edge of the image instead of
being relative to the left edge.

# Revision 1.4  1995/09/15  21:47:57  ddavies
# pcgsrot1.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot1.asm,v
#
# CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
# IF 0 that was already there.  Why does David Hirsch's version of ML get
# upset about the # signs that RCS puts at the beginnings of line (sometimes)
# while our systems are happy??
#
# Revision 1.3  1993/11/05  01:52:51  ddavies
# pcgsrot1.asm -> /product/ipcore/ipshared/src/RCS/pcgsrot1.asm,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

^

ENDIF

