;*****************************************************************
;*  Copyright (c) 1996, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcredm.asm:  2x threshold reduction using MMX opcodes.

	.586		; compile code for a Pentium processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; This code uses MMX opcodes, which have 64-bit accesses.  That's the
; main source of our speedup.  These routines are only called if
; the caller detects MMX code.  We don't have to detect it here.
; First, include the file that gives us the MMX macros.

INCLUDE	iammx.inc

_DATA	SEGMENT PUBLIC 'DATA'
; We need to access a mask that's used to get rid of pad bits.
EXTERN	mmxByteEdgeMask:QWORD

RCSInfo	BYTE	"$RCSfile: pcmatvec.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; The text for this routine looks like:
;	for (row = numRows; row > 0; row--)
;	{
;	    sum = 0;
;	    pLocalInVec = pInVector;
;	    while (pLocalInVec < pEndInVec)
;	    {
;		sum += (Int32)(*pMatrix) * (*pLocalInVec);
;		pMatrix++;
;		pLocalInVec++;
;	    }
;    
;		/* move to next row starting on an 8 byte boundary */
;	    pMatrix += colsPad;
;	    *pOutVector++ = sum;
;	}
;
; Note that the pMatrix elements are SIGNED bytes.  Rich Smith says that
; the input Int32's really only have 9 significant bits, so we can safely
; represent them as 16-bit signed integers.  This is good since the only
; multiply ops allowed in the MMX instruction set are 16 bit multiplies.
; The one we use multiplies four 16 bit things in one register by
; 4 16-bit things in another register.  It adds the two 32-bit products
; formed from multipliers/multiplicands in the upper 32 bits of the
; dword, putting the sum in the upper half.  It does the same for the
; 32-bit products in the lower half of the dword.  We'll add these
; sums of products to a running sum of products.  At the end of the
; row, we'll add the two running sums together and store that.
;
; Rich Smith has modified his code so that the row weights start on
; 8-byte boundaries.  The input and output vectors also start on
; 8-byte boundaries.  This means we can just read them happily.
;
; The inner loop read 8 matrix elements in one dword.  It unpacks the lower
; half into another dword.  It reads two 32-bit input vector words into
; a dword and packs it, together with the next two 32-bit input vector words
; into a single dword.  The dword containing 4 matrix halfword weights is
; multiplied by the dword containing 4 input vector halfwords and the
; two sums of products are added to the two running sums.  We then
; unpack the upper for matrix elements (sign-extending them during the
; unpack) and pack four more input vector words and do another multiply.
; Again the sums of products are added to the running sums.
;
; The loop goes from "right to left", i.e. from the end of the input
; vector to the beginning.  This makes dealing with the end conditions
; a little easier.  It's basically all taken care of by figuring out
; how to enter the main loop.
;
; The main loop has two parts.  The first half prepares dwords for the
; second half and does multiplies on dwords prepared by the second half.
; The second half is similar in that it prepares words for the first half
; and multiples the words prepared by the first half.  Since there
; are 8 multiplies done in each half, there are theoretically 16 possible
; entry points.  If we believed that the clients padded the input vectors
; out to 8-byte boundaries, we could limit this to 8 entry points, but
; we can't guarantee that.  So we really do have 16 entry points.
;
; Can't we limit the number of entry points by processing whatever bytes
; are present in the rightmost dword, then jumping into the second half
; of the loop?  Well, not if we want to keep the loop exit condition simple.
; The loop is set up so that when it gets done processing 16 multiplies,
; it will be exactly at the beginning of the input vector.  That means
; we need to divide the length of the input vector by 16 to see how many
; times we go through the loop.  The entry points take care of processing
; the rightmost numCols MOD 16 multiplies.


;Register Assignment:
; EAX	scratch
; EBX	address of loop entry code
; ECX	number of matrix rows left
; EDX	pointer to matrix Int8s
; EBP	pts to variables on the stack
; ESI	pointer to input vector Int32s
; EDI	pointer to output vector Int32s
; ESP	stack pointer

; MM0	holds first 4 unpacked matrix halfwords
; MM1	holds second 4 unpacked matrix halfwords
; MM2	holds third 4 unpacked matrix halfwords
; MM3	holds fourth 4 unpacked matrix halfwords
; MM4	holds 2 input vector words / four packed input halfwords
; MM5	holds 2 input vector words / four packed input halfwords
; MM6	holds 2 words of running sum for current column
; MM7	

rScratch	EQU	eax
rLoopEntry	EQU	ebx
rRowsLeft	EQU	ecx
rMatrix		EQU	edx
rInVector	EQU	esi
rOutVector	EQU	edi


; Parameters
; pMatrix	addr of first matrix byte, 8-byte aligned.
; pInVector	address of first input vector Int32, 8-byte aligned
; pOutVector	address of first output vector Int32, 8-byte aligned
; numRows	number of rows in matrix = length of output vector
; numCols	number of cols in matrix = length of input vector
; colsPad	number added to address of weight matrix element that's
;		just beyond the last one in a column to get to the
;		first weight matrix element in the next column.

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_longOpMathLibLinked.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses no registers

; This routine simply returns cTrue to indicate that the long operations are
; linked in.

ip_longOpMathLibLinkedStart::
ip_longOpMathLibLinked	PROC NEAR C PUBLIC, \
	pIPState:DWORD

; make the assembler happy
	mov	eax, pIPState

	mov	eax, 1		; value of cTrue
	ret

ip_longOpMathLibLinked ENDP


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_multMatrix8Vector32.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_multMatrix8Vector32Start::
ip_multMatrix8Vector32	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, pMatrix:DWORD, pInVector:DWORD, \
	pOutVector:DWORD, numRows:DWORD, numCols:DWORD

LOCAL	matToNextLine:DWORD, inVecStart:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; Load the parameters.  We will process each matrix row from the end
; to the beginning in units of 16 multiplies.  This is 2 weight matrix dwords
; or 8 input matrix dwords.  Doing the multiplies in units of 16 makes the
; inner loop fit together efficiently.  It's all that pipeline nonsense.
; Processing from right to left makes dealing with the end conditions a little
; easier since they're all dealt with in the loop entrance code.
;
; To do this, we need to know what the first matrix address
; is, what the first input vector offset is, and what number we need
; to add to the matrix address at the left end of one row to get to
; the correct matrix address at the right end of the next row.
;
; The offset from the beginning of a matrix row to the beginning of
; the last doubleword pair in the row is ((numCols - 1) / 16) * 16.
;
; The offset from the beginning of one matrix row to the beginning of
; the next matrix row is (numCols + 7) / 8 * 8.
;
; Therefore, the distance from the beginning of one row to the last
; doubleword of the next row is
; (((numCols + 7) / 8) * 8) + (((numCols - 1) / 16) * 16).
;
; We also need to know how to enter the inner loop.  There are
; 16 possible entries, corresponding to the 16 bytes of a doubleword pair.
; The rightmost dword pair might have 1 valid byte, 2 valid bytes, ..., 16
; valid bytes.  The number of valid bytes in the last dword pair equals
; ((numCols - 1) MOD 16) + 1.  The funnyness with the addition and subtraction
; converts numbers that are equal to 0 MOD 16 into 16, which is what we want.
; Since we're dealing with the input vector in units of 16, we want to
; point to input vector element (numCols / 16) * 16 at the top of the
; inner loop.
; HOWEVER, remember that we exit the loop when we've subtracted 16 too much
; from rMatrix.  The loop exit condition happens when we've gone past the
; beginning of the row.  Therefore, we need to add 16 back in when computing
; the number we add to rMatrix to get to the end of the next line.

; make the assembler happy
	mov	eax, pIPState

	mov	rScratch, numCols		; start to calc offsets
	mov	rMatrix, pMatrix

	mov	rLoopEntry, numCols
	add	rScratch, 7 + 16		; (numCols + 7) + 16

	dec	rLoopEntry			; (numCols - 1)
	mov	rInVector, pInVector

	and	rLoopEntry, NOT(0Fh)		; ((numCols - 1) / 16) * 16
	and	rScratch, NOT(07h)		; ((numCols + 7 + 16) / 8) * 8

	add	rMatrix, rLoopEntry		; pt to last doubleword pair in
						; first matrix row.
	add	rScratch, rLoopEntry		; offset from beginning of one
						; row to end of next.

	lea	rInVector, [rInVector + 4 * rLoopEntry]	; get addr of last
						; input vector Int32.
	mov	matToNextLine, rScratch		; store line offset

	mov	rLoopEntry, numCols		; prepare to compute entry pt.
	mov	inVecStart, rInVector		; remember where to start

; We don't really compute ((numCols - 1) MOD 16) + 1 explicitly, we just
; use (numCols MOD 16) to look the address of the loop entry point up in
; a table that does the -1+1 thing.
	and	rLoopEntry, 0Fh			; now we want lower 4 bits
	pxor	mm0, mm0			; start clearing mmx regs

; Wait for rLoopEntry to filter through pipeline.
	mov	rRowsLeft, numRows
	mov	rOutVector, pOutVector		; load output vec

	mov	rLoopEntry, DWORD PTR mvEntryTable[4 * rLoopEntry]
	pxor	mm1, mm1

	pxor	mm2, mm2
	pxor	mm3, mm3

	pxor	mm4, mm4
	pxor	mm5, mm5

	pxor	mm6, mm6
	mov	rScratch, pInVector		; for faster reg compare

	; empty
	jmp	NEAR PTR rLoopEntry


; ------------------------ MAIN LOOP ENTRY ---------------------------

; Here's where we decide how to enter the main loop.
matvecNewRow:
	movd	[rOutVector - 4], mm6	; store result sum after delay
	pxor	mm6, mm6

; Clear mm4 and mm5 in case we jump directly to matvecMult16 or matvecMult8
	pxor	mm4, mm4
	pxor	mm5, mm5

	; empty
	jmp	NEAR PTR rLoopEntry


; There are 15 valid bytes in the last dword pair.
matvecLoopEntry14:
	punpckhbw mm3, [rMatrix + 8]	; matrix bytes [15:12] -> halfwords

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords

	movq	mm5, [rInVector + 12*4]	; read input vector [13:12]
	psraw	mm3, 8			; sign-extend matrix bytes [15:12]

	movq	mm4, [rInVector + 8*4]	; read input vector [9:8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	movd	mm1, [rInVector + 14*4] ; don't read last input vector element

	packssdw mm5, mm1		; pack input [x:14:13:12]

	packssdw mm4, [rInVector + 10*4] ; pack input [11:10:9:8]
	jmp	matvecMult8


; There are 14 valid bytes in the last dword pair.
matvecLoopEntry13:
	punpckhbw mm3, [rMatrix + 8]	; matrix bytes [15:12] -> halfwords

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords
	pxor	mm1, mm1		; don't read input vec [15:14]

	movq	mm5, [rInVector + 12*4]	; read input vector [13:12]
	psraw	mm3, 8			; sign-extend matrix bytes [15:12]

	movq	mm4, [rInVector + 8*4]	; read input vector [9:8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	packssdw mm5, mm1		; pack input [x:x:13:12]

	packssdw mm4, [rInVector + 10*4] ; pack input [11:10:9:8]
	jmp	matvecMult8


; There are 13 valid bytes in the last dword pair.
matvecLoopEntry12:
	punpckhbw mm3, [rMatrix + 8]	; matrix bytes [15:12] -> halfwords

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords
	pxor	mm1, mm1		; don't read input vec [15:14]

	movd	mm5, [rInVector + 12*4]	; read input vector [12]
	psraw	mm3, 8			; sign-extend matrix bytes [15:12]

	movq	mm4, [rInVector + 8*4]	; read input vector [9:8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	packssdw mm5, mm1		; pack input [x:x:x:12]

	packssdw mm4, [rInVector + 10*4] ; pack input [11:10:9:8]
	jmp	matvecMult8


; There are 12 valid bytes in the last dword pair.
matvecLoopEntry11:
	pxor	mm3, mm3		; don't read matrix bytes [15:12]

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords
	pxor	mm5, mm5		; don't read input vector [15:12]

	movq	mm4, [rInVector + 8*4]	; read input vector [9:8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	packssdw mm4, [rInVector + 10*4] ; pack input [11:10:9:8]
	jmp	matvecMult8


; There are 11 valid bytes in the last dword pair.
matvecLoopEntry10:
	movd	mm1, [rInVector + 10*4] ; don't read input vec [11]
	pxor	mm3, mm3		; don't read matrix bytes [15:12]

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords
	pxor	mm5, mm5		; don't read input vector [15:12]

	movq	mm4, [rInVector + 8*4]	; read input vector [9:8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	packssdw mm4, mm1		 ; pack input [x:10:9:8]
	jmp	matvecMult8


; There are 10 valid bytes in the last dword pair.
matvecLoopEntry9:
	pxor	mm1, mm1		; don't read input vec [11:10]
	pxor	mm3, mm3		; don't read matrix bytes [15:12]

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords
	pxor	mm5, mm5		; don't read input vector [15:12]

	movq	mm4, [rInVector + 8*4]	; read input vector [9:8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	packssdw mm4, mm1		 ; pack input [x:x:9:8]
	jmp	matvecMult8


; There are 9 valid bytes in the last dword pair.
matvecLoopEntry8:
	pxor	mm1, mm1		; don't read input vec [11:10]
	pxor	mm3, mm3		; don't read matrix bytes [15:12]

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords
	pxor	mm5, mm5		; don't read input vector [15:12]

	movd	mm4, [rInVector + 8*4]	; read input vector [8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	packssdw mm4, mm1		 ; pack input [x:x:x:8]
	jmp	matvecMult8


; The case of there being only 8 bytes in the last dword pair was taken
; care of by simply branching to matvecMult8 directly.


; There are 7 valid bytes in the last dword pair.
matvecLoopEntry6:
	punpckhbw mm1, [rMatrix + 0]	; matrix bytes [7:4] -> halfwords

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords

	movq	mm5, [rInVector + 4*4]	; read input vector [5:4]
	psraw	mm1, 8			; sign-extend matrix bytes [7:4]

	movq	mm4, [rInVector + 0*4]	; read input vector [1:0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	movd	mm3, [rInVector + 6*4]	; don't read last input vector element

	packssdw mm5, mm3		; pack input [x:6:5:4]

	packssdw mm4, [rInVector + 2*4] ; pack input [3:2:1:0]
	jmp	matvecAdjAddrs


; There are 6 valid bytes in the last dword pair.
matvecLoopEntry5:
	punpckhbw mm1, [rMatrix + 0]	; matrix bytes [7:4] -> halfwords

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords
	pxor	mm3, mm3		; don't read input vec [7:6]

	movq	mm5, [rInVector + 4*4]	; read input vector [5:4]
	psraw	mm1, 8			; sign-extend matrix bytes [7:4]

	movq	mm4, [rInVector + 0*4]	; read input vector [1:0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	packssdw mm5, mm3		; pack input [x:x:5:4]

	packssdw mm4, [rInVector + 2*4] ; pack input [3:2:1:0]
	jmp	matvecAdjAddrs


; There are 5 valid bytes in the last dword pair.
matvecLoopEntry4:
	punpckhbw mm1, [rMatrix + 0]	; matrix bytes [7:4] -> halfwords

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords
	pxor	mm3, mm3		; don't read input vec [7:6]

	movd	mm5, [rInVector + 4*4]	; read input vector [4]
	psraw	mm1, 8			; sign-extend matrix bytes [7:4]

	movq	mm4, [rInVector + 0*4]	; read input vector [1:0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	packssdw mm5, mm3		; pack input [x:x:5:4]

	packssdw mm4, [rInVector + 2*4] ; pack input [3:2:1:0]
	jmp	matvecAdjAddrs


; There are 4 valid bytes in the last dword pair.
matvecLoopEntry3:
	pxor	mm1, mm1		; don't read matrix bytes [7:4]

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords
	pxor	mm5, mm5		; don't read input vec [7:4]

	movq	mm4, [rInVector + 0*4]	; read input vector [1:0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	packssdw mm4, [rInVector + 2*4] ; pack input [3:2:1:0]
	jmp	matvecAdjAddrs


; There are 3 valid bytes in the last dword pair.
matvecLoopEntry2:
	movd	mm3, [rInVector + 2*4]	; don't read input vec [3]
	pxor	mm1, mm1		; don't read matrix bytes [7:4]

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords
	pxor	mm5, mm5		; don't read input vec [7:4]

	movq	mm4, [rInVector + 0*4]	; read input vector [1:0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	packssdw mm4, mm3		; pack input [x:2:1:0]
	jmp	matvecAdjAddrs


; There are 2 valid bytes in the last dword pair.
matvecLoopEntry1:
	pxor	mm3, mm3		; don't read input vec [3:2]
	pxor	mm1, mm1		; don't read matrix bytes [7:4]

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords
	pxor	mm5, mm5		; don't read input vec [7:4]

	movq	mm4, [rInVector + 0*4]	; read input vector [1:0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	packssdw mm4, mm3		; pack input [x:x:1:0]
	jmp	matvecAdjAddrs


; There is 1 valid byte in the last dword pair.
matvecLoopEntry0:
	pxor	mm3, mm3		; don't read input vec [3:2]
	pxor	mm1, mm1		; don't read matrix bytes [7:4]

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords
	pxor	mm5, mm5		; don't read input vec [7:4]

	movd	mm4, [rInVector + 0*4]	; read input vector [0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	packssdw mm4, mm3		; pack input [x:x:x:0]
	jmp	matvecAdjAddrs


; There are never 0 valid bytes in the last dword pair.  In that case,
; we jump back by a dword pair where there are 16 valid bytes.



; ----------------------- DO  2 x 8 MULTIPLIES ---------------------
;
; This loop does 16 multiplies.  Each pmaddwd opcode multiplies four
; matrix halfwords with four input vector halfwords.  The two products
; in the upper half of the result dword are added together and the result
; is put into the upper 32 bits of the result dword.  The same thing
; happens with the lower two products.
;
; The trick is that the matrix elements are signed bytes and the
; input vector elements are signed 32-bit words.  The reason the loop
; is so long is that we spend time converting the bytes to 16-bit halfwords
; and converting the words into halfwords.  The punpckhbw opcodes puts
; the signed bytes into the upper byte of each halfword.  We then do
; an arithmetic right shift by 8 to fill the upper byte of each halfword
; with the sign bit and put the data byte into the lower byte of each
; halfword where it belongs.
; We're allowed to convert the signed words into halfwords because we know
; from Rich Smith that there are really only 9 useful bits in the
; input vector numbers.  The packssdw opcode basically just chops off the
; upper 16 bits of each word since those bits only contain sign bits.
; If some input vector integer has an absolute value bigger than 32K,
; the corresponding output halfword will be pegged at +32K or -32K.
;
; Note the extraordinary cleverness of this loop.  We use memory ops to
; reduce the number of opcodes (just operate on the thing in memory instead
; of reading it first).  However, all MMX memory opcodes can only be in
; the U pipe and can only pair with other MMX opcodes.  Furthermore, there's
; only one MMX shifter and the pack/unpack opcodes use it.  So we have
; 6 memory-using MMX opcodes, 4 of which also need the shifter.  Luckily,
; we have 6 more opcodes, only 2 of which need the shifter and we can
; just fit things together like a jigsaw.  Also note that we had to
; assign the MMX registers so that we had enough.  Basically, that means
; doing the multiplies early in the cycle (so the 4 committed registers
; become 2 committed registers) and using 4 registers for weight matrix
; halfwords and 2 registers for input vector halfwords.  Using 4 registers
; for input vector and 2 for weight matrix doesn't let us start by
; reading new weight matrix values and multiplying with the previous
; weight matrix values.

matvecMult16:
	punpckhbw mm3, [rMatrix + 8]	; matrix bytes [15:12] -> halfwords
	pmaddwd	mm1, mm5		; compute matrix[7:4] * input[7:4]
					; These are from the previous trip
					; through the loop.

	punpcklbw mm2, [rMatrix + 8]	; matrix bytes [11:8] -> halfwords
	pmaddwd	mm0, mm4		; compute matrix[3:0] * input [3:0].

	movq	mm5, [rInVector + 12*4]	; read input vector [13:12]
	psraw	mm3, 8			; sign-extend matrix bytes [15:12]

	movq	mm4, [rInVector + 8*4]	; read input vector [9:8]
	psraw	mm2, 8			; sign-extend matrix bytes [11:8]

	packssdw mm5, [rInVector + 14*4] ; pack input [15:14:13:12]
	paddd	mm6, mm1		; accumulate both [7:4] products

	packssdw mm4, [rInVector + 10*4] ; pack input [11:10:9:8]
	paddd	mm6, mm0		; accumulate both [3:0] products

; Second half of the loop
matvecMult8:
	punpckhbw mm1, [rMatrix + 0]	; matrix bytes [7:4] -> halfwords
	pmaddwd	mm3, mm5		; compute matrix[15:12] * input[15:12]
					; These are from the previous trip
					; through the loop.

	punpcklbw mm0, [rMatrix + 0]	; matrix bytes [3:0] -> halfwords
	pmaddwd	mm2, mm4		; compute matrix[11:8] * input [11:8].

	movq	mm5, [rInVector + 4*4]	; read input vector [5:4]
	psraw	mm1, 8			; sign-extend matrix bytes [7:4]

	movq	mm4, [rInVector + 0*4]	; read input vector [1:0]
	psraw	mm0, 8			; sign-extend matrix bytes [3:0]

	packssdw mm5, [rInVector + 6*4] ; pack input [7:6:5:4]
	paddd	mm6, mm3		; accumulate both [7:4] products

	packssdw mm4, [rInVector + 2*4] ; pack input [3:2:1:0]
	paddd	mm6, mm2		; accumulate both [3:0] products

matvecAdjAddrs:
	sub	rMatrix, 16		; point to prev matrix dword pair
	sub	rInVector, 16*4		; point to prev Input vec octet-pair

	cmp	rInVector, rScratch	; back to beginning of input vec yet?
	jge	matvecMult16		; nope, do it again.


; --------------------- EXIT FOR MAIN LOOP --------------------------

; At this point, we've just read the last 8 weight and input values, so
; multiply them.  Also get ready for storing the results.

	pmaddwd	mm1, mm5		; compute matrix[7:4] * input[7:4]
					; These are from the previous trip
					; through the loop.
	add	rMatrix, matToNextLine	; move to end of next line of matrix

	pmaddwd	mm0, mm4		; compute matrix[3:0] * input [3:0].
	add	rOutVector, 4		; point to next output loc

; wait for multiply to percolate.  Since there's only one multiplier,
; we can only start one multiply per cycle.  Drat.
	pxor	mm3, mm3		; clear for next loop entry
	pxor	mm2, mm2

	paddd	mm6, mm1		; add in last two sums
	pxor	mm1, mm1

	paddd	mm6, mm0		; get 2 sums
	pxor	mm0, mm0

	movq	mm4, mm6		; make a copy of the 2 sums
	psrlq	mm6, 32			; align sum in LS 32 bits

	paddd	mm6, mm4		; put sum in LS 32 bits
	mov	rInVector, inVecStart

; wait for mm6 to percolate
	dec	rRowsLeft		; decrement count of rows left
	jg	matvecNewRow		; start a new row if there are more

; We're really done, so return after writing final output word
matvecFinish:
	movd	[rOutVector - 4], mm6	; store result sum
	ret

ip_multMatrix8Vector32	ENDP



; Segment for initialized variables, like masks
_DATA	SEGMENT PUBLIC 'DATA'

; This is the table of loop entry point addresses.  When numCols MOD 16
; equals 0, we want to start at byte offset 15 so we read 16 bytes
; (16, 15,..., 1, 0).  When numCols MOD 16 equals 1, we want to read 1 byte
; and so on.
mvEntryTable	DWORD	matvecMult16,
			matvecLoopEntry0,
			matvecLoopEntry1,
			matvecLoopEntry2,
			matvecLoopEntry3,
			matvecLoopEntry4,
			matvecLoopEntry5,
			matvecLoopEntry6,
			matvecMult8,
			matvecLoopEntry8,
			matvecLoopEntry9,
			matvecLoopEntry10,
			matvecLoopEntry11,
			matvecLoopEntry12,
			matvecLoopEntry13,
			matvecLoopEntry14

_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no initialized variables
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcmatvec.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.2  1998/01/13 23:58:24  ddavies
pcmatvec.asm -> /product/ipcore/ipshared/src/RCS/pcmatvec.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.1  1996/10/30 05:56:45  ddavies
Initial revision



^

ENDIF



