;*****************************************************************
;*  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcpackd.asm:  assembly code to decompress one PackBits row.

	.386		; compile code for a 486 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

INCLUDE	pcproctp.prv	; contains definition of constants used to
			; determine processor type (386 vs 486)

INCLUDE pcpackd.prv	; contains definition of PackBitsDecompressInfo struct

; This file contains x86 assembly representations of the routines in
; mipackd.c.  We do assembly implementations because we should be able
; to make the transfers of literal strings more efficient.  We'll basically
; follow the structure of the routines in mipackd.c, so you should be
; familiar with that module before reading this one.


_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcpackd.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment (PackBits Decompression):
; EAX:	scratch
; EBX:	current output word
; CH:	byteCount of current run
; CL:	offset of next vacant byte in current output word
; EDX:	Address of image word on which run ends	
; EBP:	Addresses locals and parameters on stack
; ESI:	address of current compressed data byte
; EDI:	address of current image word
; ESP:	stack pointer

rOutWord	EQU ebx
rOutWordLo	EQU bl
rOutWordHi	EQU bh
rByteCntNxtVacByte EQU ecx
rByteCount	EQU ch
rNextVacantByte	EQU cl		; also used as shift specifier!!
rRunEnd		EQU edx
rRunEndLo	EQU dl
rCurImage	EQU edi
rCompData	EQU esi

; Parameters:
; pPBDecompInfo:	address of private decompressed data structure
; pImage:		address of leftmost image word
; pImageEnd:		address of byte just beyond rightmost image byte
; ppCompData:		address of pointer to compressed data
; pCompDataEnd:		address of byte just beyond the end of the compressed
;			data buffer

; Local Variables:
; leftMask		used to insert data at beginnings and ends of runs
; numRemainingBytes	holds number of literal bytes that didn't fit into
;			the current literal run
; tempNextVacantByte	holds next value of nextVacantByte

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPDWORD TYPEDEF NEAR PTR DWORD
FPBYTE TYPEDEF FAR PTR BYTE
FPDWORD TYPEDEF FAR PTR DWORD

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF
;**************************************************************************
; * ip_getPackBitsCompProc
; *
; * See the prototype for ip_getPackBitsCompProc in shrcmprs.prv to get
; * lots of comments on the interface
; *.
; *************************************************************************

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_getPackBitsDecompProc.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers
; The arguments are declared in the order specified by this routine's
; definition in mipackd.c

ip_getPackBitsDecompProcStart::
ip_getPackBitsDecompProc	PROC NEAR C PUBLIC USES EAX,\
	pIPState:DWORD, ppProc:DWORD
; All registers used are also pushed on the stack automatically.
	LOCAL	procType:DWORD

; make the assembler happy
	mov	eax, pIPState

; Figure out which type of processor we're running on
	INVOKE	ip_getProcessorType, pIPState, ADDR procType
	cmp	procType, cProcIs386
	je	pbdGetProcIs386
; This is a 486 or above
	mov	eax, ppProc
; DEBUGGING ONLY!!
	mov	DWORD PTR [eax], OFFSET ip_decompPackBitsRow386
	jmp	pbdGetProcFin

; This is a 386
pbdGetProcIs386:
	mov	eax, ppProc
	mov	DWORD PTR [eax], OFFSET ip_decompPackBitsRow386

; The assembler turns this into a pile of code
pbdGetProcFin:
	ret
ip_getPackBitsDecompProc ENDP



; Here are the restart states for decompression.
; cPBDecompNewLine is already state 0.  It is used when the next
; thing to do is read a bytecount.  pPBDecompInfo->nextVacantByte
; says where the first vacancy is in the output word.

cPBDecompReadData	EQU	1	

; *************************************************************************
; * ip_decompPackBitsRow386
; *
; * See the prototype for ip_decompPackBitsRow in shrcmprs.prv to get
; * lots of comments on the interface for the flavors of that routine
; * encoded below.
; *************************************************************************

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_decompPackBitsRow386.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers, returns a value in eax
; The arguments are declared in the order specified by mipackd.c

ip_decompPackBitsRow386Start::
ip_decompPackBitsRow386	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI,\
	pIPState:DWORD, pPBDecompInfo:DWORD, pImage:DWORD, pImageEnd:DWORD,\
	ppCompData:DWORD, pCompDataEnd:DWORD
LOCAL	leftMask:DWORD, tempNextVacantByte:DWORD, numRemainingBytes:BYTE

; All registers used are also pushed on the stack automatically.

; make the assembler happy
	mov	eax, pIPState

; first, initialize our local variables
	mov	eax, pPBDecompInfo		; get pointer to decomp info
	mov	rCurImage, pImage		; get pointer to image
; add in offset
	add	rCurImage, (PackBitsDecompressionInfo PTR [eax]).imageWordOffset

	mov	rCompData, ppCompData
	mov	rCompData, [rCompData]		; get ptr to compressed data

; get current image word
	mov	rOutWord, (PackBitsDecompressionInfo PTR [eax]).imageWord
	mov	rByteCntNxtVacByte, 0		; clear upper 16 bits of reg
; load remaining byte count.  This should work even though the structure
; entry is a DWORD and rByteCount is a byte register because this
; is a little endian machine.
	mov	rByteCount, \
		BYTE PTR (PackBitsDecompressionInfo PTR [eax]).runByteCount
; ditto for next vacant byte number
	mov	rNextVacantByte, \
		BYTE PTR (PackBitsDecompressionInfo PTR [eax]).nextVacantByte

	mov	numRemainingBytes, 0		; all strings fit in the
						; compressed data buffer
						; so far
	mov	tempNextVacantByte, 0		; clear MS bytes of this var.

; initialize the left mask value.  This is used to select which
; bytes on the left end of a new run are used.  We want leftMask to be
; (0xffffffff >> (nextVacantByte * 8)).  nextVacantByte counts
; from MS byte (00) to LS byte (3).
	mov	eax, 0FFFFFFFFh			; includes all words
	shl	rNextVacantByte, 3		; multiply shift dist by 8
						; because there are 8 bits/byte
	shr	eax, rNextVacantByte		; works because rNextVacantByte
						; is cl !!
	shr	rNextVacantByte, 3		; restore rNextVacantByte
	mov	leftMask, eax

; Now figure out which state to start or restart in.  Compare
; pPBDecompInfo->decompState to the two states, jumping to the correct one
	mov	eax, pPBDecompInfo
	cmp	(PackBitsDecompressionInfo PTR [eax]).decompState, 0		; read a new byte count?
	jne	pbdReadData386			; no, go read data

; Jump to the bottom of the main loop to see if we're already done.
; It's weird if we are, but it should be possible.  Before we get
; there, rByteCount should be equal 0 so rByteCntNxtVacByte is equivalent
; to rNextVacantByte.  rByteCount probably already is 0, but you never know.
	xor	rByteCount, rByteCount
	jmp	pbdEndLoopCheck386


; *************** MAIN LOOP FOR 386 *******************************

; In this loop, we read a bytecount, decide whether it's for a literal
; string or a replica string and fill in the string.

pbdMainLoop386:

; check to see if the byte exists.  It doesn't exist if we've run
; off the end of the compresed data buffer.
	cmp	rCompData, pCompDataEnd
	jae	pbdExhaustedCompDataReadByteCount386

; We know the bytecount is inside the buffer, so read it.
	mov	rByteCount, [rCompData]
	inc	rCompData

; If the bytecount was the last thing in the compressed data buffer,
; stop here.  This is also where we restart if we reached the end of
; the compressed data in the midst of a run of literals.  We can't reach
; the end of the compressed data in the midst of a run of replicas
; since thre's only one replica byte to read.  We either fail here before
; reading it or fill in the entire run with it.
pbdReadData386:
	cmp	rCompData, pCompDataEnd
	jae	pbdExhaustedCompDataReadData386

; Now test the bytecount.  If it's greater than or equal to 0, we need
; to copy a string of literals.  Otherwise, we'll expand a replica byte
	cmp	rByteCount, 0
	jl	pbdExpandReplicas386


; +++++++++++++++++++ COPY LITERAL STRINGS +++++++++++++++++++++++++

; first, increment the 	bytecount since we always encode the length of
; the literal string minus 1.  The largest positive number is 7Fh, so the
; largest value after the increment is 80h.  All further references to
; bytecount are unsigned.
	inc	rByteCount

; If the run extends beyond the end of the compressed data buffer,
; shorten the byte count.  At the end of this computation, rRunEnd
; points to the location in the compressed data just after the end
; of the literal string.
	xor	rRunEnd, rRunEnd
	mov	rRunEndLo, rByteCount
	add	rRunEnd, rCompData
	cmp	rRunEnd, pCompDataEnd
	jb	pbdLiteralStringFits386

; Ooops, the literal string ran off the end of the compressed data buffer,
; so shorten it.  It should be set to the difference between pCompDataEnd
; and rCompData.  numRemainingBytes should be set to the difference between
; the original value of rByteCount and the new value.
	mov	eax, pCompDataEnd
	sub	eax, rCompData		; know the diff fits in a byte
	sub	rByteCount, al
	mov	numRemainingBytes, rByteCount
	mov	rByteCount, al
	mov	rRunEnd, pCompDataEnd	; set rRunEnd to point just beyond
					; the last byte of the run

; Decrement rRunEnd so it points to the last byte of the literal string
pbdLiteralStringFits386:
	dec	rRunEnd

; Calculate the address of the nextVacantByte once the
; run has been transferred.
	add	rByteCount, rNextVacantByte
	dec	rByteCount			; compute byte position
						; occupied by last byte of
						; literal string after it's
						; been copied to rOutWord.
	and	rByteCount, 3
	mov	BYTE PTR tempNextVacantByte, rByteCount	; save count in memory

	xor	rByteCount, rByteCount		; clear upper part of
						; of rByteCntNxtVacByte.
						; nextVacantByte is in lowest
						; byte.  This is done so
						; we can use rByteCntNxtVacByte
						; as an index below

; See if we can read at least 4 bytes beyond the end of the string
; from the compressed data stream. This is almost always true.
; It lets us always fill up the output word from the compressed data
; stream without fear of running off the end.
	mov	eax, pCompDataEnd
	sub	eax, rRunEnd
	cmp	eax, 4				; are at least 4 bytes left?
	jl	pbdLiteralStringAtEnd386 	; no, be slow and careful

; Cool, we can just blindly fill rOutWord and mask the unneeded stuff off
; later.
	jmp	DWORD PTR pbdStartLiteralStringTable386[rByteCntNxtVacByte * 4]

; Here's a table of starting addresses.  There are only four valid
; starting positions in a word: 0 (fill all four bytes), 1, 2, and 3 (fill
; only LS byte).
pbdStartLiteralStringTable386:
	DWORD	OFFSET pbdStartLiteralStringByte0386
	DWORD	OFFSET pbdStartLiteralStringByte1386
	DWORD	OFFSET pbdStartLiteralStringByte2386
	DWORD	OFFSET pbdStartLiteralStringByte3386

; There are no bytes in rOutWord, so fill up the whole thing
pbdStartLiteralStringByte0386:
	mov	rOutWordHi, [rCompData]		; HA!, we know how to make
						; the lines shorter when
						; MASM complains they're tool
						; long!
	mov	rOutWordLo, [rCompData+1]
	shl	rOutWord, 16			; align 2 MS bytes
	mov	rOutWordHi, [rCompData+2]
	mov	rOutWordLo, [rCompData+3]
	add	rCompData, 4
	jmp	pbdLiteralLoopCheck386

; There is one byte in the MS byte of rOutWord
pbdStartLiteralStringByte1386:
	shr	rOutWord, 16			; put MS byte into rOutWordHi
	mov	rOutWordLo, [rCompData]
	shl	rOutWord, 16			; align 2 MS bytes
	mov	rOutWordHi, [rCompData+1]
	mov	rOutWordLo, [rCompData+2]
	add	rCompData, 3
	jmp	pbdLiteralLoopCheck386

; There are already 2 bytes in the upper 2 MS bytes of rOutWord
pbdStartLiteralStringByte2386:
	mov	rOutWordHi, [rCompData]
	mov	rOutWordLo, [rCompData+1]
	add	rCompData, 2
	jmp	pbdLiteralLoopCheck386

; There are already 3 bytes in the upper 3 MS bytes of rOutWord
pbdStartLiteralStringByte3386:
	mov	rOutWordLo, [rCompData]
	add	rCompData, 1
	jmp	pbdLiteralLoopCheck386

; This is the loop in which we transfer whole words from the compressed
; data stream to the image.  Any extra bytes loaded into the word will
; be masked off by the right mask at the end of the transfer.  Loading
; the extra bytes seems like a more efficient thing to do than to carefully
; check before loading each byte to see if we're at the end.
; Note that we do the write at the top of the loop because we only want
; to write a full word if there's another word coming.  We want to
; exit the loop with the last bytes in rOutWord, but not yet written
; to memory.
pbdLiteralLoop386:
	mov	[rCurImage], rOutWord
	add	rCurImage, 4

	mov	rOutWordHi, [rCompData]
	mov	rOutWordLo, [rCompData+1]
	shl	rOutWord, 16			; align 2 MS bytes
	mov	rOutWordHi, [rCompData+2]
	mov	rOutWordLo, [rCompData+3]
	add	rCompData, 4

pbdLiteralLoopCheck386:
	cmp	rRunEnd, rCompData
	jae	pbdLiteralLoop386

; rOutWord now has the last bytes in it.  rCompData points to some piece
; of trash.  rRunEnd has the address of the last word in the string, though
; the address points to the compressed data stream, not the image word.
; Fix up rCompData.
	mov	rCompData, rRunEnd
	inc	rCompData

; tempNextVacantByte has the address of the last valid byte of the string
; in rOutWord.  Put this into rNextVacantByte and make sure rByteCount is
; cleared.
	mov	rByteCntNxtVacByte, tempNextVacantByte

; We're now at the end of the body of code that's used when there is no
; danger of the literal string being near the end of the compressed data
; buffer.  So, just jump to the end of the main loop where we'll apply
; the right mask to rOutWord or store it if it's full.
	jmp	pbdEndLoop386

;------------  End of normal code used to copy literal strings ---------
;---------  Start of special case code used to copy literal strings ----

; This code is used when we're near the end of the compressed data buffer.
; It doesn't have be particularly efficient since it is seldom used.
; In this code, we shift the bytes into rOutWord one at a time.  In
; preparation, we shift rOutWord right so the rightmost valid byte
; is in the rightmost byte.  This is a shift right by (4 - nextVacantByte)
; bytes.
; Before each shift, we test to see if rOutWord is full.  If so, we write it
; to the image.
pbdLiteralStringAtEnd386:
	sub	rNextVacantByte, 4		; -(shift dist in bytes)
	neg	rNextVacantByte			; shift distance in bytes
	shl	rNextVacantByte, 3		; shift dist in bits
	shr	rOutWord, rNextVacantByte	; works because rNextVacantByte
						; is cl
	shr	rNextVacantByte, 3		; equals number of vacant
						; bytes in rOutWord

pbdSlowLiteralLoop:
	test	rNextVacantByte, 0FFh		; is rOutWord full?
	jnz	pbdSlowLiteralLoopNextByte	; no put another byte in
	mov	[rCurImage], rOutWord		; yes, write it
	add	rCurImage, 4
	mov	rNextVacantByte, 4		; rOutWord is empty now
pbdSlowLiteralLoopNextByte:
	shl	rOutWord, 8			; make room for new byte
	mov	rOutWordLo, [rCompData]		; load it
	dec	rNextVacantByte
	inc	rCompData
	cmp	rCompData, rRunEnd		; at end of run yet?
	jbe	pbdSlowLiteralLoop		; nope, keep going


; When we get here, rNextVacantByte holds the number of vacant bytes
; at the TOP of rOutWord.  Realign rOutWord.  Since the loop exits
; immediately after putting a byte into rOutWord, it's always got
; at least one byte in it at this point.
	shl	rNextVacantByte, 3		; # of vacant bits
	shl	rOutWord, rNextVacantByte	; rNextVacantByte is cl !

; Put the address of the next vacant byte into rNextVacantByte and
; make sure rByteCount is cleared.
	mov	rByteCntNxtVacByte, tempNextVacantByte

; See if the run stopped early because we ran out of data.  If
; so, we can almost just quit now.  If it stopped early, but it also
; happened to stop on an image word boundary, write the image word now
; since we won't do it when we restart.
	test	numRemainingBytes, 0FFh
	jz	pbdEndLoop386

; The string ended with bytes left.  Did we happen to fill an image word?
; If so, write it to the image.
	cmp	rNextVacantByte, 3
	jne	pbdEarlyTermMaskWord386		; no, just mask rOutWord

	mov	[rCurImage], rOutWord		; yes, write the word
	add	rCurImage, 4			; and update the pointer
	xor	rNextVacantByte, rNextVacantByte ; next vacant byte is MS byte
	xor	rOutWord, rOutWord		; and output word is empty
	jmp	pbdEarlyTermFin

; The loop terminate early.  Get the mask corresponding to rNextVacantByte,
; and apply it to rOutWord before quitting.  Also updata rNextVacantByte
; so it really does point to the next vacant byte in rOutWord.
; Note that since ch (rBytecount) is still clear, we can index the
; mask table with rByteCntNxtVacByte directly.  That register's lowest
; byte contains rNextVacantByte and the rest of the register is 0.
pbdEarlyTermMaskWord386:
	and	rOutWord, pbdRightMask[rByteCntNxtVacByte * 4]
	inc	rNextVacantByte

; Now load rByteCount with the right count since it's ok to set
; ch at this point.  Also decrement it as explained above.
pbdEarlyTermFin:
	mov	rByteCount, numRemainingBytes
	dec	rByteCount
	jmp	pbdExhaustedCompDataReadData386

; ++++++++++++++++++ END OF COPYING LITERAL STRINGS +++++++++++++++++++++

; ++++++++++++++++++ EXPAND REPLICA STRINGS ++++++++++++++++++++++++++++

; In this code, we first do nothing if the bytecount equals 128 (same as
; -128).  That's a Noop.
pbdExpandReplicas386:
	cmp	rByteCount, 128
	je	pbdMainLoop386

; Now we convert the bytecount, a negative number, to the correct positive
; number.  This is -byteCount + 1.  However, the only thing we do with
; this quantity is to compute the address of the last image byte affected
; by the replication.  To do that, we'll add (rByteCount - 1), so just
; skip the increment/decrement pair.
	neg	rByteCount

; Here's the code used to replicate a byte.  We know
; that the byte is available in the compressed data
; stream because we checked in the main loop.
;
; The code follows these steps:
; 1. Set rRunEnd equal to rCurImage + nextVacantByte +
;	  byteCount - 1.  This is the full address of the last
;	  byte of the string.  Note that since we didn't increment
;	  rByteCount above, we don't have to do the "-1" here.
; 2. Set rNextVacantByte = rRunEnd & 0x3.  This is the
;	  offset of the last byte in the run relative to
;	  its word.
; 3. Set pRunEnd equal to pRunEnd & ~0x3.  This makes
;	  it point to the last word to be filled.
; 4. Replicate the byte into all four bytes of
;	  eax.
; 5. Insert eax & leftMask into rOutWord.
; 6. If rCurImage equals rRunEnd, skip to step 9.
; 7. Write rOutWord to [rCurImage], load rOutWord with
;	  eax and make rCurImage point to the next word.
; 8. While rCurImage does not equal rRunEnd, write rOutWord
;	  to [rCurImage], and increment pCurImage.  We will
;	  actually do this loop in slightly larger chunks
;	  (2 words) for efficiency.
	add	rNextVacantByte, rByteCount	; sum in cl
	xor	rByteCount, rByteCount		; rByteCntNxtVacByte now
						; contains the sum
	add	rByteCntNxtVacByte, rCurImage	; address of last image byte
	mov	rRunEnd, rByteCntNxtVacByte	; where it belongs
	and	rByteCntNxtVacByte, 3		; gets offset of last
						; valid byte in image since
						; rNextVacantByte == lower
						; byte of rByteCntNxtVacByte
	and	rRunEnd, NOT 3			; gets add of last image
						; word to be written

; replicate all the bytes into eax. Most of the time, the byte being
; loaded is 0, but a test for this takes more time than just loading
; it..
	mov	al, [rCompData]
	mov	ah, al
	shl	eax, 16
	mov	al, [rCompData]
	mov	ah, al
	inc	rCompData			; point to next bytecount

	push	eax				; save for rOutWord later
	and	eax, leftMask			; get rid of part that doesn't
						; belong in the first word
	or	rOutWord, eax			; compute first word
	pop	eax				; get back replicated word

; If we've just formed the last (and first) word in the string,
; skip to the end where we`ll mask it.
	cmp	rRunEnd, rCurImage
	je	pbdEndLoop386

; The first word isn't also the last, so write it and make another word
	mov	[rCurImage], rOutWord
	add	rCurImage, 4
	mov	rOutWord, eax

; here's the big replica loop.  We will write 8 bytes at a time, so
; decrement rRunEnd by 8 to make the tests easier.  It's important to
; be able to write lots of stuff quickly to fill in white space.
	sub	rRunEnd, 8
	jmp	pbdReplicaLoop8Check386

pbdReplicaLoop8386:
	mov	[rCurImage], rOutWord
	mov	[rCurImage+4], rOutWord
	add	rCurImage, 8
pbdReplicaLoop8Check386:
	cmp	rCurImage, rRunEnd
	jbe	pbdReplicaLoop8386

; fix rRunEnd back up
	add	rRunEnd, 8

; At this point, rCurImage equals rRunEnd-4 or rRunEnd.  The farthest
; forward we could have written is rRunEnd-8+4, or the word just before
; rRunEnd's word.  That only happens if rCurImage now equals rRunEnd.
; If we haven't written that word, write it
	cmp	rCurImage, rRunEnd
	je	pbdEndLoop386
	mov	[rCurImage], rOutWord		; write 2nd-to-last word
	add	rCurImage, 4

; ++++++++++++++++ END OF EXPANDING REPLICA STRINGS +++++++++++++++++++++++++

; +++++++++++++++++++++++++ BOTTOM OF MAIN LOOP +++++++++++++++++++++++++++++

; At this point, rOutWord has the last word, and some extra, in it.
; rNextVacantByte has the number of the last valid byte in rOutWord.
; rByteCount contains 0, so rByteCntNxtVacByte as a whole equals
; rNextVacantByte.
; See if rOutWord is full.  If so, store it and initialize
; rNextVacantByte and leftMask.  If not, fetch the new mask value,
; use it on rOutWord, complement it and save it into leftMask.
pbdEndLoop386:
	cmp	rNextVacantByte, 3		; is rOutWord full?
	jne	pbdEndLoopNoWrite386

; rOutWord is full
	mov	[rCurImage], rOutWord
	add	rCurImage, 4
	mov	rNextVacantByte, 0
	mov	leftMask, 0FFFFFFFFh		; all bytes of next string are
						; useful
	xor	rOutWord, rOutWord		; we OR things into this
	jmp	pbdEndLoopCheck386

; rOutWord is not full
pbdEndLoopNoWrite386:
	mov	eax, pbdRightMask[rByteCntNxtVacByte * 4] ; get the right mask
	and	rOutWord, eax			; use it
	not	eax				; invert it to become the
						; new left mask
	mov	leftMask, eax			; save it to become the
						; new left mask
	inc	rNextVacantByte			; point to next vacant byte
						; in rOutWord

; Well, here we are at the bottom of the main loop.  If rCurImage +
; nextVacantByte < pImageEnd, keep going.  Otherwise, fall through to
; finish up.
pbdEndLoopCheck386:
	mov	eax, rCurImage
	add	eax, rByteCntNxtVacByte		; rCurImage + nextVacantByte
	cmp	eax, pImageEnd
	jb	pbdMainLoop386			; not done yet

; ************************** END OF MAIN LOOP ***************************

; We get here after happily completing our task.  The last bytes
; of the scan line may still be in rOutWord, so we should write
; them.  Only do this if rNextVacantByte points to the middle of
; a word.  If it doesn't point to the middle of a word, pCurImage
; is pointing off the end of the scan line.  It's ok to write there in a
; framed image, but some weirdo may want to use this on unframed images.
	test	rNextVacantByte, 0FFh
	jz	pbdCleanup386
	mov	[rCurImage], rOutWord

; Clear out the saved structures so we know that we're done with the line.
pbdCleanup386:
	mov	eax, pPBDecompInfo
	mov	(PackBitsDecompressionInfo PTR [eax]).nextVacantByte, 0
	mov	(PackBitsDecompressionInfo PTR [eax]).imageWordOffset, 0
	mov	(PackBitsDecompressionInfo PTR [eax]).imageWord, 0

; Set the state of the decompression buffer to say that the next thing to
; do is read a bytecount.
	mov	(PackBitsDecompressionInfo PTR [eax]).decompState, 0

; store the value of the next place from which we'll read compressed data
	mov	eax, ppCompData
	mov	[eax], rCompData

; We're happy, return the successful error code
	mov	eax, 0
	ret



; *********************** EXHAUSTED COMPRESSION BUFFER *********************

pbdExhaustedCompDataReadByteCount386:
	mov	eax, pPBDecompInfo
; restart by reading a new line
	mov	(PackBitsDecompressionInfo PTR [eax]).decompState, 0
	jmp	pbdExhaustedCompData386

pbdExhaustedCompDataReadData386:
	mov	eax, pPBDecompInfo
; restart by reading data
	mov	(PackBitsDecompressionInfo PTR [eax]).decompState, \
		cPBDecompReadData

pbdExhaustedCompData386:
	mov	(PackBitsDecompressionInfo PTR [eax]).imageWord, rOutWord
	mov	BYTE PTR (PackBitsDecompressionInfo PTR [eax]).nextVacantByte, \
		rNextVacantByte
	sub	rCurImage, pImage
	mov	(PackBitsDecompressionInfo PTR [eax]).imageWordOffset, rCurImage
	mov	BYTE PTR (PackBitsDecompressionInfo PTR [eax]).runByteCount, \
		rByteCount
	mov	eax, ppCompData			; get address of returned
						; compressed data ptr
	mov	rCompData, pCompDataEnd
	mov	[eax], rCompData		; say we're done
;
; We're done, so return.  The assembler turns this into a PILE of code!
	mov	eax, 0				; return ia_successful
	ret
ip_decompPackBitsRow386 ENDP


_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
; This is the array of masks used to get rid of crud off the end of
; the right side of each run.  It is indexed with the offset of the
; last byte in the run relative to the left side of the word.
; The array is repeated because we put a termination condition bit
; into bit 2 of the index.
; Why are entries 3 and 7 set to 0?  Shouldn't they be all 1's?
; Well, we don't really use the rightmask to mask anything if the
; run ends on a word boundary.  However, we do initialize the subsequent
; left mask (bits of the next run that are valid) as the complement
; of the previous rightmask.  All bits in subsequent image word are
; new, so we set the table up so that when we complement 0, we get
; all 1's for the next leftmask.

pbdRightMask	DWORD 	0FF000000h, 0FFFF0000h, 0FFFFFF00h, 00000000h,
			0FF000000h, 0FFFF0000h, 0FFFFFF00h, 00000000h

_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcpackd.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.7  1998/01/13 23:58:26  ddavies
pcpackd.asm -> /product/ipcore/ipshared/src/RCS/pcpackd.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.6  1996/09/17 00:45:47  ddavies
pcpackd.asm -> /product/ipcore/ipshared/src/RCS/pcpackd.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.5  1996/09/16 00:51:44  ddavies
pcpackd.asm -> /product/ipcore/ipshared/src/RCS/pcpackd.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.4  1995/09/15 21:48:02  ddavies
pcpackd.asm -> /product/ipcore/ipshared/src/RCS/pcpackd.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

;; Revision 1.3  1995/05/26  01:10:12  ddavies
;; pcpackd.asm -> /product/ipcore/ipshared/src/RCS/pcpackd.asm,v
;;
;; Added/fixed a bunch of comments.
;;
;; Revision 1.2  1995/05/19  18:13:45  ddavies
;; ddavies on Fri May 19 11:13:04 PDT 1995
;;
;; pcpackd.asm -> /product/ipcore/ipshared/src/RCS/pcpackd.asm,v
;;
;; Add a completely meaningless (and perhaps petty) comment to line 333 so
;; that MASM won't complain that line 333 is too long.  I love this thing.
;;
;; Revision 1.1  1995/05/19  02:55:35  ddavies
;; Initial revision
;;
# Revision 1.5  1994/01/22  01:11:17  ddavies
# pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
#
# Change comment to say shrpixr.pub instead of shrio.pub.
#
# Revision 1.4  1993/11/05  01:51:46  ddavies
# pcbswap.asm -> /product/ipcore/ipshared/src/RCS/pcbswap.asm,v
#
# Add RCS magic to have string specifying this file's name, rev number and
# last edit date into the object file.  Also add change log at the bottom.
#

^

ENDIF

