;*****************************************************************
;*  Copyright (c) 1995, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcproctp.asm:  assemble code used to distinguish 386's from 486's and above.

	.586		; compile code for a Pentium processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The 486 and new processors have a couple of new opcodes.  The byteswap
; opcode is particularly useful.  Code in this module is used to determine
; the processor type.  It's cribbed from the Intel Microprocessor manual.
; The trick is to try flipping a bit in the extended flags register that's
; there in the 486 (and later) but not in the 386.

INCLUDE	pcproctp.prv	; contains flags used to indentify processor types
INCLUDE iammx.inc	; to get definition of MMX's EMMS opcode

; Other routines need to access a table listed below:
PUBLIC	mmxByteEdgeMask

_DATA	SEGMENT PUBLIC 'DATA'
; We need access to this global variable.  We'll declare it in a C
; file so that debuggers (Microsoft's in particular) will feel good
; about looking at it.

RCSInfo	BYTE	"$RCSfile: pcproctp.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment (byte swap routines):
; EAX:	data word
; EBX:	storage for ESP 
; ECX:	old value of flags
; EDX:	
; EBP:	Addresses locals and parameters on stack
; ESI:	
; EDI:	
; ESP:	stack pointer

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPDWORD TYPEDEF NEAR PTR DWORD
FPBYTE TYPEDEF FAR PTR BYTE
FPDWORD TYPEDEF FAR PTR DWORD


; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_getProcessorType which starts at global label
; ip_getProcessorTypeStart.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses a few registers

; This procedure takes a pointer to a double word in which the procedure
; type will be recorded.  The word is filled with cProcIs386, cProcIs486,
; cProcIsPentium or cProcIsPentiumPro depending on the processor type.
; 
ip_getProcessorTypeStart::
ip_getProcessorType	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI,\
	pIPState:DWORD, pProc:LOCALNPDWORD
; All registers used are also pushed on the stack automatically.

; make the assembler happy
	mov	eax, pIPState

	mov	esi, esp	; save stack pointer
	and	esp, NOT 3	; align stack to avoid an alignment check
				; fault
	pushfd			; push the 32-bit flag register
	pop	eax		; get flags into eax
	mov	ecx, eax	; save them in ecx so we can restore them
				; when we're finished
	xor	eax, cACBit	; flip the AC flag bit
	push	eax		; put it back into the flags reg
	popfd
	pushfd
	pop	eax		; get the flags back again
	xor	eax, ecx	; is the AC bit still flipped?  If so, this
				; is a 486.  If not, there was no register in
				; that flag bit so this is a 386.
	je	procTypeIs386

; This is a 486 or above.  See if it has the cpuid opcode, which
; is required to learn more.
	mov	eax, ecx	; restore the AC flag bit.
	xor	eax, cIDBit	; flip the ID bit.
	push	eax		; put it back into the flags reg
	popfd
	pushfd
	pop	eax		; get the flags back again
	xor	eax, ecx	; is the ID bit still flipped?  If so, this
				; is a processor with the cpuid opcode.
				; If not, this is an early 486..
	je	procTypeIs486

; This processor, which might be a 486, Pentium or Pentium Pro ( or later?)
; supports the cpuid opcode.  This opcode destroys ecx, so save it.
	push	ecx

; The way you use the cpuid opcode, is to call it with eax = 0.  On return,
; eax contains the highest valid input argument to the cpuid instruction.
; ebx, ecx and edx contain some ASCII characters that one could use to
; identify the processor's manufacturer.
	xor	eax, eax	; create first operand to cpuid opcode
	cpuid
	cmp	eax, 0		; is 0 the only legal input?
	je	procTypeIsAdvanced486

; Go get the feature vector
	mov	eax, 1
	cpuid

; The processor family (486, Pentium, ...) is in the processor signature
; in eax
	and	eax, cFamilyMask ; get just the family bits
	shr	eax, cFamilyPos	; align them at the bottom of the word
	mov	ebx, pProc	; point to return loc
	mov	DWORD PTR [ebx], eax	; store processor type.
	pop	ecx		; restore ecx
	jmp	procTypeFin	; and we're done.


; This is a 486 with a boring cpuid opcode, restore ecx and say so.
procTypeIsAdvanced486:
	pop	ecx
; This is an early 486
procTypeIs486:
	mov	eax, pProc	; get address of returned variable
	mov	DWORD PTR [eax], cProcIs486	; is a 486
	jmp	procTypeFin

; This is a 386
procTypeIs386:
	mov	eax, pProc	; get address of returned variable
	mov	DWORD PTR [eax], cProcIs386	; is a 386

procTypeFin:
	push	ecx		; restore the flags
	popfd
	mov	esp, esi	; restore sp

; We're done, so return.  The assembler turns this into a PILE of code!
	ret
ip_getProcessorType ENDP




; This procedure takes a pointer to a double word in which the longest
; legal access (in bytes) for this type of processor will be recorded.
; The word will be filled with 4 or 8 depending on the processor type.
; This procedure is only called when we're running Windows 3.1.

; Plain old ip_getLongestAccess is now a C routine.  It makes use of
; the Visual C/C++ structured exception handling mechanism to see if
; MMX opcodes exist.  That's because reading cr0 is illegal in NT.

ip_getLongestWin31AccessStart::
ip_getLongestWin31Access	PROC NEAR C PUBLIC USES EAX EBX ECX EDX,\
	pIPState:DWORD, pLength:LOCALNPDWORD
; All registers used are also pushed on the stack automatically.
	LOCAL	procType:DWORD

; make the assembler happy
	mov	eax, pIPState

; We don't have to check to see if long accesses are enabled.  If they
; weren't enabled, ip_getLongestAccess wouldn't have called this routine.

; Figure out which type of processor we're running on
	INVOKE	ip_getProcessorType, pIPState, ADDR procType

; If this isn't a 586 or above, say longest access is 4 bytes
	cmp	procType, cProcIsPentium
	jl	lengthIs4Bytes

; We know this is a Pentium or above, so it could have MMX.  Check
; the Floating point emulation bit of CR0.  If this bit is set,
; MMX instructions will cause traps, so we wouldn't want to do
; MMX stuff in that case.
; This routine is only called when we're running under Win31, where
; reading cr0 is a perfectly valid operation.
	mov	eax, cr0		; get the emulation bit
	test	eax, cEMBit		; is it on?
	jnz	lengthIs4Bytes		; yes, dont' issue MMX opcodes.

; Well, now see if this Pentium or Pentium Pro actually has MMX
; opcodes.
	mov	eax, 1
	cpuid				; go get the feature flags

	test	edx, cMMXAvailBit	; are the MMX instructions available?
	jz	lengthIs4Bytes		; no, set long access length to
					; 4 bytes.

; MMX opcodes are available.  Set the access length to 8 bytes.
	mov	eax, pLength		; get ptr to returned length
	mov	DWORD PTR [eax], 8
	jmp	getLongWin31AccessFin

lengthIs4Bytes:
	mov	eax, pLength		; get ptr to returned length
	mov	DWORD PTR [eax], 4

; We're done, so return.  The assembler turns this into a PILE of code!
getLongWin31AccessFin:
	ret
ip_getLongestWin31Access ENDP


_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'

; Well, this isn't really used to figure out whether we have MMX,
; but it's general MMX stuff.  Many MMX routines have to mask off
; bits from beyond the frame when the read at the right edge of
; images.  These masks help them do that.
; This is really an array of masks.  It's addresses by the source
; width in bytes MOD 8.  It gets rid of everything past the last
; source byte with something in it.  We assume the frame takes care of
; that byte.
	ALIGN	8
mmxByteEdgeMask	QWORD	0FFFFFFFFFFFFFFFFh	; byteWidth MOD 8 == 0
		QWORD	000000000FF000000h	; byteWidth MOD 8 == 1
		QWORD	000000000FFFF0000h	; byteWidth MOD 8 == 2
		QWORD	000000000FFFFFF00h	; byteWidth MOD 8 == 3
		QWORD	000000000FFFFFFFFh	; byteWidth MOD 8 == 4
		QWORD	0FF000000FFFFFFFFh	; byteWidth MOD 8 == 5
		QWORD	0FFFF0000FFFFFFFFh	; byteWidth MOD 8 == 6
		QWORD	0FFFFFF00FFFFFFFFh	; byteWidth MOD 8 == 7

_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END

IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcproctp.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.9  1998/01/21 21:51:19  ddavies
pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

Remove check of ipVar_enableLongAccesses.  If long (> 32 bit) accesses
weren't allowed, we wouldn't be in ip_getLongestWin31Access.

Revision 1.8  1998/01/13 23:58:27  ddavies
pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.7  1996/10/24 19:54:39  ddavies
pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

This is assembly code that determines processor type, including whether
the processor supports MMX opcodes.  There's also an implementation of
ip_getLongestWin31Access here.  This routine uses the processor type and
the presence of MMX hardware and the fact that the floating point emulation
bit in Control Register 0 is not set to decide that one can make 8-byte
memory accesses (and other long ops).  This routine is only called in
Win31 where you can read CR0 without a trap.

Revision 1.6  1996/09/24 18:30:19  ddavies
ddavies on Tue Sep 24 11:29:56 PDT 1996

pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

Reference unused variable procType.  Who would think assemblers would
check?

Revision 1.5  1996/09/24 02:43:58  ddavies
pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

Update ip_getProcessorType and ip_getLongestAccess so they really deal
with the cpuid opcode and MMX machines.  ip_getProcessorType now figures
out whether the processor has a cpuid opcode and uses it if so.  It returns
the processor family.  ip_getLongestAccess figures out whether this is a
processor that could have MMX, detects whether it does have MMX, and sets
the longest memory access to 8 bytes if so.  Otherwise, the longest access
is set to 4 bytes.
Intel says that if the floating point emulation bit in control register 0
(CR0) is set, MMX opcodes will cause unimplemented opcode traps.  On the
other hand, reading CR0 at our priority ring also causes a trap.
They're still trying to decide what to do.

Revision 1.4  1996/09/17 00:45:50  ddavies
pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.3  1996/09/16 00:51:58  ddavies
pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.2  1995/09/15 21:48:06  ddavies
pcproctp.asm -> /product/ipcore/ipshared/src/RCS/pcproctp.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

;; Revision 1.1  1995/05/19  02:57:32  ddavies
;; Initial revision
;;

^

ENDIF

