;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcredcm.asm:  2x horizontal reduction.

	.586		; compile code for a Pentium processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; This code uses MMX opcodes, which have 64-bit accesses.  That's the
; main source of our speedup.  These routines are only called if
; the caller detects MMX code.  We don't have to detect it here.
; First, include the file that gives us the MMX macros.

INCLUDE	iammx.inc

_DATA	SEGMENT PUBLIC 'DATA'
; We need to access a mask that's used to get rid of pad bits.
EXTERN	mmxByteEdgeMask:QWORD

RCSInfo	BYTE	"$RCSfile: pcredcm.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

;Register Assignment:
; EAX	scratch
; EBX	scratch
; ECX	scratch
; EDX	final source address
; EBP	pts to variables on the stack
; ESI	source address
; EDI	dest address
; ESP	stack pointer

; MM0	word of left source cell
; MM1	scratch
; MM2	word of right source cell
; MM3	scratch
; MM4	final dest word
; MM5	
; MM6	
; MM7	

rScratch1	TEXTEQU	<eax>
rScratch2	TEXTEQU	<ebx>
rScratch3	TEXTEQU	<ecx>
rSrcFin		TEXTEQU	<edx>
rSrcAddr	TEXTEQU	<esi>
rDstAddr	TEXTEQU	<edi>

; Parameters
; srcAddr	addr of left end of source line
; srcWidth	width of source line in bits
; dstAddr	addr of left end of dest line
; dstOffset	offset in bytes to rightmost dest word
; srcOffset	offset in bytes to leftmost byte in rightmost
;		group of source bits that wil form the rightmost
;		dest word
; srcLimit	offset in bytes to rightmost source byte (often
;		to the right of byte at srcOffset)

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF


; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_compressColsOrLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_compressColsOrLineLongStart::
ip_compressColsOrLineLong PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword pair in the line.
; This pair may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
; rSrcFin will point to the leftmost doubleword in the source line.
	mov	rSrcFin, srcAddr	; pt to 1st source dword
	mov	rSrcAddr, rSrcFin

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first left/right dword pair.
					; We do this because we're really
					; always going to go after the
					; even doubleword in an even/odd
					; pair.  If the rightmost dest
					; dword gets bits from two source
					; dwords, we use the regular loop.
					; If the rightmost dest dword gets
					; bits from only 1 source dword, we
					; use the code just below.
	mov	rDstAddr, dstAddr	; pt to leftmost dest doubleword

	add	rDstAddr, dstOffset	; pt to rightmost dest doubleword
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 1.  In reducing by 2, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 2,
; 8N + 3, etc.  So, we only add 6, not 7.
	mov	rScratch2, srcWidth	; get source width in bits.

	add	rScratch2, 6		; round up to next byte
	mov	rScratch1, srcOffset

; The first dest doubleword could be composed of one or two source
; doublewords.  By the way, we know that it's safe to read doublewords
; because the caller checked to see that all the source and dest lines
; begin on multiples of 8 and have bytes/line that are multiples of 8.
; If the first dest word required one doubleword, do the special code
; below.  Otherwise, skip into the main loop.
	and	rScratch2, 038h		; = (src width in bytes MOD 8) * 8
	cmp	rScratch1, srcLimit	; do these point to the same word?
	je	redColsOrRightDWord	; yes, then the first dest dword
					; contains data from only one source
					; dword.  Go encode that.


; This is the preamble to the regular loop that encodes a dest dword from
; two source dwords.  It looks like the top of the regular loop, except
; it doesn't write out a previous dest dword.
	movq	mm0, [rSrcAddr]		; get left doubleword

	movq	mm2, [rSrcAddr + 8]	; get right doubleword
	movq	mm1, mm0		; get ready to combine left columns

	pand	mm2, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	psrlq	mm1, 1			; move to lower end of each bit pair

	movq	mm3, mm2		; get ready to combine right columns
	por	mm0, mm1		; get left result in lower bits

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get final mask.
	jmp	redColsOrLongLoopEntry	; go to rest of the loop



; The first dest doubleword is composed of only one source doubleword and
; a word of 0's.  Compute this below.  Remember that rSrcAddr already
; points to the right dword.
redColsOrRightDWord:
	movq	mm0, [rSrcAddr]		; get doubleword
	sub	rSrcAddr, 16		; pt to previous doubleword pair

; get rid of pad bits
	pand	mm0, [MMWORD PTR mmxByteEdgeMask + rScratch2]

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get final mask.
	movq	mm1, mm0		; get ready to combine columns

	psrlq	mm1, 1			; move to lower end of each bit pair

	por	mm0, mm1		; get result in lower bits

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just those bits

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1

	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	movq	mm1, mm0		; get ready to make clumps of 2 bits

	psrlq	mm1, 1

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off

	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2

	por	mm0, mm1

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte

	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4

	por	mm4, mm0
	sub	rDstAddr, 8		; prepare for next loop

	pand	mm4, mm7		; got them.
	pxor	mm3, mm3		; clear "other half word"

; the byte packing and writing are done in the main loop.
	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jl	redColsOrLongFin	; quit if we're done


; ----------------------------------------------------------------
; MAIN LOOP
; ----------------------------------------------------------------

; On entering this loop, rSrcAddr should be pointing 16 bytes to the
; left of the first doubleword of a doubleword pair,
; rDstAddr should be pointing to first doubleword to be written,
; rSrcFin should be pointing to leftmost doubleword of the source line.
; mm3 should have bytes from right dword and mm4 should have bits from
; left dword.
; Pack left bits into the *lower* half of a word.  Remember that the lower
; half is the left half in this wacky world.
redColsOrLongLoop:
	movq	mm0, [rSrcAddr]		; get left doubleword
	packuswb mm4, mm3		; pack the bytes

	movq	mm2, [rSrcAddr + 8]	; get right doubleword
	movq	mm1, mm0		; get ready to combine left columns

; This clock cycle takes place while we're waiting for mm4 to progress
; through the pipeline.
	movq	mm3, mm2		; get ready to combine right columns
	psrlq	mm1, 1			; move to lower end of each bit pair

	movq	[rDstAddr + 8], mm4	; write the previous result
	por	mm0, mm1		; get left result in lower bits

; We enter the loop here, with the proper preamble, if the first
; dest doubleword contains bits from two source dwords.
redColsOrLongLoopEntry:
	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just left bits
	psrlq	mm3, 1			; move to lower end of each bit pair

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1
	por	mm2, mm3		; get right result in lower bits

	pand	mm2, [MMWORD PTR lsBitMask1Bit]	; select just right bits
	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	punpckldq mm3, mm2		; put low halfword of mm2 into
					; high halfword of mm3
	movq	mm1, mm0		; get ready to make clumps of 2 bits

	punpckhdq mm2, mm3		; complete the word swap
	sub	rSrcAddr, 16		; pt to previous doubleword pair

	psrlq	mm1, 1
	movq	mm3, mm2		; get ready to make clumps of 2 bits

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble
	psrlq	mm3, 1

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off
	por	mm2, mm3		; make clumps in lower 2 bits of
					; each nibble

	pand	mm2, [MMWORD PTR lsBitMask2Bits] ; mask bits in right dword
	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2
	movq	mm3, mm2		; get ready to make clumps of 4 bits

	por	mm0, mm1
	psrlq	mm3, 2

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	por	mm2, mm3

	pand	mm2, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4
	movq	mm3, mm2		; get ready to make clumps of bytes

	por	mm4, mm0
	psrlq	mm3, 4

	por	mm3, mm2
	pand	mm4, mm7		; got bytes in left dword.

; finish right dword at top of loop
	sub	rDstAddr, 8		; get ready for next trip through loop
	pand	mm3, mm7		; got bytes in right dword.

	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jge	redColsOrLongLoop		; no, keep going

; We're done with that line.
redColsOrLongFin:
	packuswb mm4, mm3		; pack the bytes

	movq	[rDstAddr + 8], mm4	; write the last word
redColsOrLongRet:
	ret

ip_compressColsOrLineLong	ENDP



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_compressColsAndLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_compressColsAndLineLongStart::
ip_compressColsAndLineLong   PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword pair in the line.
; This pair may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
; rSrcFin will point to the leftmost doubleword in the source line.
	mov	rSrcFin, srcAddr	; pt to 1st source dword
	mov	rSrcAddr, rSrcFin

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first left/right dword pair.
					; We do this because we're really
					; always going to go after the
					; even doubleword in an even/odd
					; pair.  If the rightmost dest
					; dword gets bits from two source
					; dwords, we use the regular loop.
					; If the rightmost dest dword gets
					; bits from only 1 source dword, we
					; use the code just below.
	mov	rDstAddr, dstAddr	; pt to leftmost dest doubleword

	add	rDstAddr, dstOffset	; pt to rightmost dest doubleword
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 1.  In reducing by 2, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 2,
; 8N + 3, etc.  So, we only add 6, not 7.
	mov	rScratch2, srcWidth	; get source width in bits.

	add	rScratch2, 6		; round up to next byte
	mov	rScratch1, srcOffset

; The first dest doubleword could be composed of one or two source
; doublewords.  By the way, we know that it's safe to read doublewords
; because the caller checked to see that all the source and dest lines
; begin on multiples of 8 and have bytes/line that are multiples of 8.
; If the first dest word required one doubleword, do the special code
; below.  Otherwise, skip into the main loop.
	and	rScratch2, 038h		; = (src width in bytes MOD 8) * 8
	cmp	rScratch1, srcLimit	; do these point to the same word?
	je	redColsAndRightDWord	; yes, then the first dest dword
					; contains data from only one source
					; dword.  Go encode that.


; This is the preamble to the regular loop that encodes a dest dword from
; two source dwords.  It looks like the top of the regular loop, except
; it doesn't write out a previous dest dword.
	movq	mm0, [rSrcAddr]		; get left doubleword

	movq	mm2, [rSrcAddr + 8]	; get right doubleword
	movq	mm1, mm0		; get ready to combine left columns

	pand	mm2, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	psrlq	mm1, 1			; move to lower end of each bit pair

	movq	mm3, mm2		; get ready to combine right columns
	pand	mm0, mm1		; get left result in lower bits

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get final mask.
	jmp	redColsAndLongLoopEntry	; go to rest of the loop



; The first dest doubleword is composed of only one source doubleword and
; a word of 0's.  Compute this below.  Remember that rSrcAddr already
; points to the right dword.
redColsAndRightDWord:
	movq	mm0, [rSrcAddr]		; get doubleword
	sub	rSrcAddr, 16		; pt to previous doubleword pair

; get rid of pad bits
	pand	mm0, [MMWORD PTR mmxByteEdgeMask + rScratch2]

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get final mask.
	movq	mm1, mm0		; get ready to combine columns

	psrlq	mm1, 1			; move to lower end of each bit pair

	pand	mm0, mm1		; get result in lower bits

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just those bits

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1

	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	movq	mm1, mm0		; get ready to make clumps of 2 bits

	psrlq	mm1, 1

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off

	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2

	por	mm0, mm1

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte

	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4

	por	mm4, mm0
	sub	rDstAddr, 8		; prepare for next loop

	pand	mm4, mm7		; got them.
	pxor	mm3, mm3		; clear "other half word"

; the byte packing and writing are done in the main loop.
	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jl	redColsAndLongFin	; quit if we're done


; ----------------------------------------------------------------
; MAIN LOOP
; ----------------------------------------------------------------

; On entering this loop, rSrcAddr should be pointing 16 bytes to the
; left of the first doubleword of a doubleword pair,
; rDstAddr should be pointing to first doubleword to be written,
; rSrcFin should be pointing to leftmost doubleword of the source line.
; mm3 should have bytes from right dword and mm4 should have bits from
; left dword.
; Pack left bits into the *lower* half of a word.  Remember that the lower
; half is the left half in this wacky world.
redColsAndLongLoop:
	movq	mm0, [rSrcAddr]		; get left doubleword
	packuswb mm4, mm3		; pack the bytes

	movq	mm2, [rSrcAddr + 8]	; get right doubleword
	movq	mm1, mm0		; get ready to combine left columns

; This clock cycle takes place while we're waiting for mm4 to progress
; through the pipeline.
	movq	mm3, mm2		; get ready to combine right columns
	psrlq	mm1, 1			; move to lower end of each bit pair

	movq	[rDstAddr + 8], mm4	; write the previous result
	pand	mm0, mm1		; get left result in lower bits

; We enter the loop here, with the proper preamble, if the first
; dest doubleword contains bits from two source dwords.
redColsAndLongLoopEntry:
	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just left bits
	psrlq	mm3, 1			; move to lower end of each bit pair

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1
	pand	mm2, mm3		; get right result in lower bits

	pand	mm2, [MMWORD PTR lsBitMask1Bit]	; select just right bits
	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	punpckldq mm3, mm2		; put low halfword of mm2 into
					; high halfword of mm3
	movq	mm1, mm0		; get ready to make clumps of 2 bits

	punpckhdq mm2, mm3		; complete the word swap
	sub	rSrcAddr, 16		; pt to previous doubleword pair

	psrlq	mm1, 1
	movq	mm3, mm2		; get ready to make clumps of 2 bits

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble
	psrlq	mm3, 1

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off
	por	mm2, mm3		; make clumps in lower 2 bits of
					; each nibble

	pand	mm2, [MMWORD PTR lsBitMask2Bits] ; mask bits in right dword
	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2
	movq	mm3, mm2		; get ready to make clumps of 4 bits

	por	mm0, mm1
	psrlq	mm3, 2

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	por	mm2, mm3

	pand	mm2, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4
	movq	mm3, mm2		; get ready to make clumps of bytes

	por	mm4, mm0
	psrlq	mm3, 4

	por	mm3, mm2
	pand	mm4, mm7		; got bytes in left dword.

; finish right dword at top of loop
	sub	rDstAddr, 8		; get ready for next trip through loop
	pand	mm3, mm7		; got bytes in right dword.

	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jge	redColsAndLongLoop		; no, keep going

; We're done with that line.
redColsAndLongFin:
	packuswb mm4, mm3		; pack the bytes

	movq	[rDstAddr + 8], mm4	; write the last word
redColsAndLongRet:
	ret

ip_compressColsAndLineLong	ENDP




; -----------------------  4x Column Reduction  ---------------------------

; This code looks like the 2x reduction code above.  We sample four
; columns, each containing one doubleword, producing 16 useful bits
; from the 64 original bits.
;
; This is based on a 4x reduction routine in pcredmn.asm.  That one was
; rewritten several times.  Here, we basically remove the reads of words
; in other rows since we don't need them.
;
; We'll use the MMX pack opcode, packuswb.  It converts two
; dwords, each containg bits in every other byte (the odd bytes when
; numbering from the MS end of the word), into one dword with every
; byte filled.  The odd bytes from one dword end up in the upper half
; of the dest dword and the odd bytes from the other other end up in the
; lower half.  What a deal.
;
; Since we're doing 4x reduction, the basic idea is to fairly quickly
; get 2 bits per byte (8/4 = 2), then get 4 bits in every other byte,
; then pack two such dwords.  This gives us a dword with 4 bits in every
; byte.  We repeat this process for two more dwords, producing another
; dword with 4 bits in every byte.  We then shift/OR each of these dwords
; (it's nicer to do two things in parallel), producing two words with
; 8 bits in every other byte.  These two words are packed to produce
; the final dest word.
;
; To compute compressColsOr, we OR together all the source bits to
; produce one bit.  To perform compressColsAnd, we AND them.
; ORing and ANDing horizontally is messy.
; The standard way to do this is to use shifts and ORs.  First,
; you combine each pixel with its neighbor on the left.  This
; produces results in the even bit positions. Then you combine
; each of these pixels with the pixel that 2 to the left.
; That produces result pixels at the locations that are 0 MOD 4.  Then
; we have to clump the two result bits together, so there's another
; shift/OR cycle.  Each shift/combine cycle takes 3 MMX opcodes.  We
; also have to mask the results at one stage, so we need (3 * 3) + 1 = 10
; opcodes to get to the two-bits-clumped-at-the-end-of-each-byte stage.
; We need (4 * 3) + 2 mask steps = 14 cycles to get to the 4-bits-clumped-
; at-the-end-of-every-other-byte stage.
;
; Well, maybe not.  I thought about a cute scheme that uses the the
; packed compare feature to set all the bits of a byte if any of the bits
; are one.  Unfortunately, you have to separately test each nibble, then
; combine the results and put them in the right place.  A byte shift (one
; that breaks the shift chain at byte boundaries) would have been nice for
; this, but there's no such thing in MMX-land.  After a lot of fooling
; around, I managed to get something that was the same speed as the standard
; shift/Or method described above.  Defeat.
;
; Well, except maybe not.  The latest scheme uses byte adds, which do exist.
; First we make a copy of the input register and AND the copy with 0777...7h.
; This turns off the high order bit of each nibble.  We then add 0777...7h to
; this.  If there are any 1's in the lower three bits of the nibble, the
; MS bit of the nibble is now turned on.  We OR this with the original
; value.  If there were any 1's in a nibble, that nibble's MS bit is now
; a 1.  In 4 opcodes, we've done the work of 6 above.
;
; That worked for compressColsOr.  The steps for compressColsAnd are almost the
; same.  We do the same copy and the same AND with 07777...7h.  We then
; add 0111...1h.  If there were any 0's in the lower three bits of a nibble,
; the MS bit of that nibble is 0, otherwise it's a 1.  We AND this with
; the original.  If there were any 0's in a nibble, its MS bit is now 0,
; which is what we want.  The only hiccup here is that we don't have enough
; MMX registers to hold both masks, so we have to use a memory cycle to
; do the add of 01111....1h.  Too bad.
;
; But wait, we're not done.  We AND the result obtained above with 08888...8h.
; This gets rid of everything except the bits we really want.  We then
; add 038383838...38h to the word.  This "carries" (nerd pun) the lower
; nibble result in bit 3 up to bit 6, right next to the upper nibble result.
; With these two opcodes, we've clumped the two bits for a byte together.
;
; Unfortunately, you can't use the carry trick to move two bits, so we need
; to use the move-shift-or sequence to move the clumps of two bits together
; in every other byte.  Since the add of 0383838...38h left dreck behind,
; we need to make with 0CCCC...CCh before doing that.
;
; When we're done with this step, we have four bits at the top of the even
; numbered bytes (byte 0 is MS byte) and dreck at the top of the odd
; numbered bytes.  We do a halfword shift (shift chain broken at 16-bit
; boundaries) right by 12 to put these in the lower bytes where the pack
; opcode will expect to find them.
;
; The total of all this sillyness is 11 opcodes, compared to 14 for the
; standard way.  Since all of this happens 4 times for each dest dword,
; we save 4 * (14-11) = 12 opcodes for each dest dword, which is significant.
;
;
;
; Wordswapping raises it ugly head here.  The x86 processors are little
; endian machines.  We have a convention that images are accessed in 32-bit
; chunks, so left-to-right within the image corresponds to MS-to-LS within
; a 32-bit word.  Unfortunately, the 32-bit word that's on the left in
; an image is at the LS (right) end of a 64-bit dword.  Grump.
;
; There are two places we have to worry about this - when reading source
; dwords and when writing dest dwords.   Wordswapping the dest dwords is
; easy, we just flip the order of the arguments to the final pack opcode.
; It takes 2 opcodes (unpacklo, unpackhi) to wordswap a dword register
; containing source args.  It takes 4 opcodes (movq, shiftL, shiftR, por)
; to swap bytes within halfwords or halfwords within words.  So swapping
; the original source dwords is a good idea, right?  Wrong.  We'd need
; 2 * 4 = 8 opcodes to swap all four source dwords and only 4 * 1 = 4
; opcodes to byteswap the final dword, so we'll swap the final one.
;
; As with almost any MMX task, it helps if you try to do at least two things
; at once.  This makes it easier to fill the empty slots in the instruction
; pairs.  The basic loop structure will be:
;
; 1. Fetch bits from dword column 3.
; 2. Fetch bits from dword column 2.
; 3. Starting in step 1, clump together the bits from columns 1 and 0.
;    These were read in the previous loop iteration.
; 4. Pack the clumped bits from columns 1 and 0 with the previously clumped
;    bits from columns 3 and 2 (from the previous trip through the loop).
;    Byteswap and write out the result dword.
; 5. Check to see if we're done with the loop. If so, jump to the end.
; 6. Fetch the bits from dword column 1.
; 7. Fetch the bits from dword column 0.
; 8. Starting in step 6, clump the bits from columns 3 and 2 together.
; 9. Decrement the source addresses.
; 10. If there are more source columns to read, loop back to step 1.
; 11. If we're at the left end of the scan line (alternative  to step 9),
;     do the consolidation steps that would be done in parallel with steps 1
;     and 2 and jump back to step 3.  The check in step 5 will exit if
;     we're done.
;
; On entering the routine, we test to see if the rightmost dword requires
; 1, 2, 3 or 4 source dword columns.  In each case, we read the column
; required and mask the dword to make sure that no pad bits creep into
; the reduced image.  If we're reading columns 3 or 1, we continue on
; to read columns 0 and 2 respectively before leaping into the loop.
; This is because the code in the loop for reading columns 3 and 2 is
; a unit, as is the code for reading columns 1 and 0.
;

;
;Register Assignment:
; EAX	scratch
; EBX	scratch
; ECX	scratch
; EDX	scratch
; EBP	pts to variables on the stack
; ESI	source address, src line 0
; EDI	dest address
; ESP	stack pointer

; There are four columns of doublewords in each 4x4 doubleword square.
; We use mm0 to accumulate the results for column 0, mm1 for column 1
; and so on.
; MM0	used to accumulate data from doubleword col 0
; MM1	used to accumulate data from doubleword col 1
; MM2	used to accumulate data from doubleword col 2
; MM3	used to accumulate data from doubleword col 3
; MM4	used to consolidate data from even doubleword cols
; MM5	used to consolidate data from odd doubleword cols
; MM6   holds all bits seen so far in loop
; MM7	holds 07777777777777777h, used in horizontal reduction

rScratch	TEXTEQU	<eax>
rScratch1	TEXTEQU	<ebx>
rScratch2	TEXTEQU	<ecx>
rScratch3	TEXTEQU	<edx>
rSrcAddr	TEXTEQU	<esi>
rDstAddr	TEXTEQU	<edi>

; Parameters
; srcAddr	addr of left end of source line
; srcWidth	width of source image in bits
; dstAddr	addr of left end of dest line
; dstOffset	offset in bytes to rightmost dest word
; srcOffset	offset in bytes to leftmost byte in rightmost
;		group of source bits that will form the rightmost
;		dest word
; srcLimit	offset in bytes to rightmost source byte (often
;		to the right of byte at srcOffset)



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_compress4ColsOrLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_compress4ColsOrLineLongStart::
ip_compress4ColsOrLineLong PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword quad in the line.
; This quad may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	rSrcAddr, srcAddr	; pt to 1st source dword
	mov	rDstAddr, dstAddr	; pt to 1st dest dword

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first dword quad.
					; We do this because we're really
					; always going to go after the
					; first doubleword in a quad.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first quad.
	add	rDstAddr, dstOffset	; point to right dest dword

; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword quad)
; can be 0, 8, 16 or 24.  If it's 0, we enter the loop where it reads
; one source dword before writing the dest.  If it's 8, we enter the loop
; where it reads 2 source dwords and so on.
	mov	rScratch, srcLimit	; get srcOffset
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 3.  In reducing by 4, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 4,
; 8N + 5, etc.  So, we only add 4, not 7.
	mov	rScratch1, srcWidth	; get source width in bits.

	add	rScratch1, 4		; round up to next byte
	pxor	mm0, mm0		; clear registers

	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8
	sub	rScratch, srcOffset	; get difference

; Initialize all the MMX registers in a wad.  Since none of these
; operations affect the condition codes, the codes from the "sub"
; above will be the ones the "je" below looks at.
	pxor	mm1, mm1
	pxor	mm2, mm2

; We mask each source nibble with 7, then add 7 to set MS bit if there
; are any 1's in the lower three bits.
	movq	mm7, [MMWORD PTR all7sMask]
	pxor	mm3, mm3

; Load the mask for the rightmost dword in the line
	movq	mm4, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	pxor	mm5, mm5		; mm5 gets used at red4ColsOrLoopW1

	pxor	mm6, mm6 
	je	red4ColsOrLoopW0Entry	; if there's only one dword in
					; the rightmost quad, process it.

	cmp	rScratch, 8		; if there are two dwords in the
	je	red4ColsOrLoopW1Entry	; rightmost quad, process them

	cmp	rScratch, 16		; if there are three dwords in the
	je	red4ColsOrLoopW2Entry	; rightmost quad, process them



; There are four words in the loop.  Read the cols 3 and 2,
; then branch into the loop to read cols 1 and 0.
red4ColsOrLoopW3Entry:
	movq	mm3, [rSrcAddr + 24]	; get next word 3
	; empty

	movq	mm2, [rSrcAddr + 16]	; get next word 2
	pand	mm3, mm4		; get rid of bits of end of image

	movq	mm5, mm3		; get ready for step 1, col 3
	jmp	red4ColsOrLoopW1	; do next two dword cols

; Here's where we start when dword 2 is the rightmost word in the scanline.
red4ColsOrLoopW2Entry:
	movq	mm2, [rSrcAddr + 16]	; get next word 2
	; empty

	pand	mm2, mm4		; get rid of bits of end of image
	; empty

	; empty
	jmp	red4ColsOrLoopW1	; do next two dword cols

; Here's where we start when dword 1 is the rightmost word in the scanline.
red4ColsOrLoopW1Entry:
	movq	mm1, [rSrcAddr + 8]	; get next word 1
	; empty

	movq	mm0, [rSrcAddr + 0]	; get next word 0
	pand	mm1, mm4		; get rid of bits off edge of image

	movq	mm5, mm1		; get ready for step 1, col 1
	jmp	red4ColsOrLoopW3

red4ColsOrLoopW0Entry:
	movq	mm0, [rSrcAddr + 0]	; get next word 0
	; empty

	pand	mm0, mm4		; get rid of bits off edge of image
	jmp	red4ColsOrLoopW3	; adjust addresses and
					; do next two dword cols


; --------------------- MAIN LOOP -----------------------------

; --------------------- Read Words 3 & 2 ---------------------------

; This is the start of the main loop.  Do horizontal reduction of data in dword
; columns 0 and 1.  At the end of this half of the loop, read dwords in cols
; 3 and 2.
; On loop entry, mm5 holds contents of mm3 and rSrcAddr is pointing
; to the dword quad to the right of the dword quad we want to read next.
red4ColsOrLoopW3:
	movq	mm3, [MMWORD PTR all8sMask] ; nib result mask
	movq	mm4, mm0		; make copy of col 0 bits

	pand	mm4, mm7		; get rid of MS bit in each nibble
	pand	mm5, mm7		; MS bit of col 1 nibbles

	paddb	mm4, mm7		; set MS bit if any 1's in LS 3 bits
	paddb	mm5, mm7		; set nib result bits if low bits set

	por	mm0, mm4		; set MS bit if any 1's in nibble
	por	mm1, mm5		; set MS bit if any 1's in col 1 nibble

	movq	mm2, [MMWORD PTR all38sMask] ; carries bit 3 to bit 6
	pand	mm0, mm3		; select just nib result bits

	pand	mm1, mm3		; select just nib result bits
	paddb	mm0, mm2		; carry LS nib result bit up to
					; bit 6 of each byte.

	movq	mm3, [MMWORD PTR allC0Mask]
	paddb	mm1, mm2		; carry LS nib result bit up to
					; bit 6 of each byte.

	pand	mm0, mm3		; get paired result bits for col 0
	pand	mm1, mm3

	movq	mm4, mm0		; get ready to copy col 0 bit pairs
	movq	mm5, mm1		; get ready to copy col 1 bit pairs

	sub	rDstAddr, 8		; decrement dest address
	psllq	mm4, 6			; put clumps in top end of upper bytes

	psllq	mm5, 6			; put clumps in top end of upper bytes
	por	mm0, mm4		; col 0 quad results in odd bytes

	por	mm1, mm5		; col 1 quad results in odd bytes
	psrlw	mm0, 8			; col 0 res in bits 7:4 of even bytes

	psrlw	mm1, 8			; col 1 res in bits 7:4 of even bytes
	sub	rSrcAddr, 32		; point to curent dword quad

	packuswb mm1, mm0		; produce col0:1 bits in bits 7:4 of
					; all bytes
	movq	mm4, mm6		; prepare to compact col2:col3 bits

	movq	mm5, mm1		; prepare to compace col0:col1 bits
	psllq	mm4, 4			; put 8 col2:col3 bits in the odd bytes

	psllq	mm5, 4			; put 8 col0:col1 bits in the odd bytes
	por	mm1, mm5		; combine shifted col0:1 bits

	psrlw	mm1, 8			; 8 col0:1 bits in even bytes
	por	mm6, mm4		; combine shifted col2:3 bits

	psrlw	mm6, 8			; 8 col2:3 bits in even bytes
	cmp	rSrcAddr, srcAddr	; back at beginning of line yet?

; Here's where we wordswap the final dest word by flipping the order
; of the args to the pack command.  The dest appears in the LS word and
; the source appears in the MS word.  Therefore, we get 2:3:0:1, which
; is what we want.
	packuswb mm1, mm6
	jl	red4ColsOrExit		; quit if that was last dest word

; Now we need to byteswap the bytes in here.  Remember that each source word
; produced 16 bits, 8 from the upper word and 8 from the lower word.  We
; now need to swap the bytes, which corresponds to swapping the words in
; the original.
	movq	mm4, mm1		; make a copy
	psllw	mm1, 8			; shift low bytes to high bytes

	movq	mm3, [rSrcAddr + 24]	; get top word 3
	psrlw	mm4, 8			; shift high bytes to low bytes

	movq	mm2, [rSrcAddr + 16]	; get top word 2
	por	mm1, mm4		; combine swapped bytes

; wait for mm1 to percolate through pipeline
	; empty
	; empty

	movq	[rDstAddr + 8], mm1	; store dest word
	movq	mm5, mm3		; get ready for step 1, col 3


; ------------------- Get Bits from Words 1 & 0 ----------------------

; We can enter the loop at this point, so the source addresses must
; be consistent.
red4ColsOrLoopW1:
	movq	mm1, [MMWORD PTR all8sMask] ; nib result mask
	movq	mm4, mm2		; make copy of col 2 bits

	pand	mm4, mm7		; get rid of MS bit in each nibble
	pand	mm5, mm7		; MS bit of col 3 nibbles

	paddb	mm4, mm7		; set MS bit if any 1's in LS 3 bits
	paddb	mm5, mm7		; set nib result bits if low bits set

	por	mm2, mm4		; set MS bit if any 1's in nibble
	por	mm3, mm5		; set MS bit if any 1's in col 3 nibble

	movq	mm0, [MMWORD PTR all38sMask] ; carries bit 3 to bit 6
	pand	mm2, mm1		; select just nib result bits

	pand	mm3, mm1		; select just nib result bits
	paddb	mm2, mm0		; carry LS nib result bit up to
					; bit 6 of each byte.

	movq	mm1, [MMWORD PTR allC0Mask] ; get paired result bit mask
	paddb	mm3, mm0		; carry LS nib result bit up to
					; bit 6 of each byte.

	pand	mm3, mm1		; select just col 3 nib result bits
	pand	mm2, mm1		; select just col 2 nib result bits

	movq	mm4, mm2		; get ready to copy col 2 bit pairs
	movq	mm6, mm3		; get ready to copy col 3 bit pairs

	movq	mm0, [rSrcAddr]		; get next word 0
	psllq	mm4, 6			; put clumps in top end of upper bytes

	psllq	mm6, 6			; put clumps in top end of upper bytes
	por	mm2, mm4		; col 2 quad results in odd bytes

; Now we finish reading the col 0 inputs while col 3 consolidation catches up
	por	mm6, mm3		; col 3 quad results in odd bytes
	psrlw	mm2, 8			; col 2 res in bits 7:4 of even bytes

	movq	mm1, [rSrcAddr + 8] ; get next word 1
	psrlw	mm6, 8			; col 3 res in bits 7:4 of even bytes

	packuswb mm6, mm2		; produce col2:3 bits in bits 7:4 of
					; all bytes
	movq	mm5, mm1		; get ready for step 1, col 1

	; empty
	jmp	red4ColsOrLoopW3	; nope, do another quad

; yeah, if I were even nerdier that I am, I'd have the conditional jump
; exits the loop at the bottom instead of in the middle.  But figuring
; out how the loop was structured would be a pain in that case, so I'm
; leaving it this way.

; ----------------------------- LOOP EXIT -------------------------------

; At this point, we've almost finished creating the last dest dword.
; the only thing left is to byteswap it.  Do that and finish.
red4ColsOrExit:
	movq	mm4, mm1		; make a copy
	psllw	mm1, 8			; shift low bytes to high bytes

	psrlw	mm4, 8			; shift high bytes to low bytes

	por	mm1, mm4		; combine swapped bytes

; wait for mm1 to percolate through pipeline
	; empty
	; empty

	movq	[rDstAddr + 8], mm1	; store dest word
	ret

ip_compress4ColsOrLineLong ENDP




; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_compress4ColsAndLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_compress4ColsAndLineLongStart::
ip_compress4ColsAndLineLong PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword quad in the line.
; This quad may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	rSrcAddr, srcAddr	; pt to 1st source dword
	mov	rDstAddr, dstAddr	; pt to 1st dest dword

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first dword quad.
					; We do this because we're really
					; always going to go after the
					; first doubleword in a quad.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first quad.
	add	rDstAddr, dstOffset	; point to right dest dword

; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword quad)
; can be 0, 8, 16 or 24.  If it's 0, we enter the loop where it reads
; one source dword before writing the dest.  If it's 8, we enter the loop
; where it reads 2 source dwords and so on.
	mov	rScratch, srcLimit	; get srcOffset
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 3.  In reducing by 4, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 4,
; 8N + 5, etc.  So, we only add 4, not 7.
	mov	rScratch1, srcWidth	; get source width in bits.

	add	rScratch1, 4		; round up to next byte
	pxor	mm0, mm0		; clear registers

	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8
	sub	rScratch, srcOffset	; get difference

; Initialize all the MMX registers in a wad.  Since none of these
; operations affect the condition codes, the codes from the "sub"
; above will be the ones the "je" below looks at.
	pxor	mm1, mm1
	pxor	mm2, mm2

; We mask each source nibble with 7, then add 7 to set MS bit if there
; are any 1's in the lower three bits.
	movq	mm7, [MMWORD PTR all7sMask]
	pxor	mm3, mm3

; Load the mask for the rightmost dword in the line
	movq	mm4, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	pxor	mm5, mm5		; mm5 gets used at red4ColsAndLoopW1

	pxor	mm6, mm6 
	je	red4ColsAndLoopW0Entry	; if there's only one dword in
					; the rightmost quad, process it.

	cmp	rScratch, 8		; if there are two dwords in the
	je	red4ColsAndLoopW1Entry	; rightmost quad, process them

	cmp	rScratch, 16		; if there are three dwords in the
	je	red4ColsAndLoopW2Entry	; rightmost quad, process them



; There are four words in the loop.  Read the cols 3 and 2,
; then branch into the loop to read cols 1 and 0.
red4ColsAndLoopW3Entry:
	movq	mm3, [rSrcAddr + 24]	; get next word 3
	; empty

	movq	mm2, [rSrcAddr + 16]	; get next word 2
	pand	mm3, mm4		; get rid of bits of end of image

	; empty
	jmp	red4ColsAndLoopW1	; do next two dword cols

; Here's where we start when dword 2 is the rightmost word in the scanline.
red4ColsAndLoopW2Entry:
	movq	mm2, [rSrcAddr + 16]	; get next word 2
	; empty

	pand	mm2, mm4		; get rid of bits of end of image
	jmp	red4ColsAndLoopW1	; do next two dword cols

; Here's where we start when dword 1 is the rightmost word in the scanline.
red4ColsAndLoopW1Entry:
	movq	mm1, [rSrcAddr + 8]	; get next word 1
	; empty

	movq	mm0, [rSrcAddr + 0]	; get next word 0
	pand	mm1, mm4		; get rid of bits off edge of image

	; empty
	jmp	red4ColsAndLoopW3

red4ColsAndLoopW0Entry:
	movq	mm0, [rSrcAddr + 0]	; get next word 0
	; empty

	pand	mm0, mm4		; get rid of bits off edge of image
	jmp	red4ColsAndLoopW3	; adjust addresses and
					; do next two dword cols


; --------------------- MAIN LOOP -----------------------------

; --------------------- Read Words 3 & 2 ---------------------------

; This is the start of the main loop.  Do horizontal reduction of data in dword
; columns 0 and 1.  At the end of this half of the loop, read dwords in cols
; 3 and 2.
; On loop entry, mm5 holds contents of mm3 and rSrcAddr is pointing
; to the dword quad to the right of the dword quad we want to read next.
; The big difference between ANDing columns and ORing columns is that we
; want to detect any 0's instead of any 1's.  We do this by masking each
; nibble with 7 as before, then adding 1 instead of 7.  If the result is
; 0, there was at least one 0 in the lower three bits.  We AND this result
; with the original word.  The resulting MS nibble will be 0 if any of the
; nibble's bits were 0.
red4ColsAndLoopW3:
	movq	mm3, [MMWORD PTR all8sMask] ; nib result mask
	movq	mm4, mm0		; make copy of col 0 bits

	movq	mm2, [MMWORD PTR all1sMask] ; add this to LS 3 nib bits to
					; detect any 0's
	movq	mm5, mm1

	pand	mm4, mm7		; get rid of MS bit in each nibble
	pand	mm5, mm7		; MS bit of col 1 nibbles

	paddb	mm4, mm2		; set MS bit if any 1's in LS 3 bits
	paddb	mm5, mm2		; set nib result bits if low bits set

	pand	mm0, mm4		; set MS bit if all 1's in nibble
	pand	mm1, mm5		; set MS bit if all 1's in col 1 nibble

	movq	mm2, [MMWORD PTR all38sMask] ; carries bit 3 to bit 6
	pand	mm0, mm3		; select just nib result bits

	pand	mm1, mm3		; select just nib result bits
	paddb	mm0, mm2		; carry LS nib result bit up to
					; bit 6 of each byte.

	movq	mm3, [MMWORD PTR allC0Mask]
	paddb	mm1, mm2		; carry LS nib result bit up to
					; bit 6 of each byte.

	pand	mm0, mm3		; get paired result bits for col 0
	pand	mm1, mm3

	movq	mm4, mm0		; get ready to copy col 0 bit pairs
	movq	mm5, mm1		; get ready to copy col 1 bit pairs

	sub	rDstAddr, 8		; decrement dest address
	psllq	mm4, 6			; put clumps in top end of upper bytes

	psllq	mm5, 6			; put clumps in top end of upper bytes
	por	mm0, mm4		; col 0 quad results in odd bytes

	por	mm1, mm5		; col 1 quad results in odd bytes
	psrlw	mm0, 8			; col 0 res in bits 7:4 of even bytes

	psrlw	mm1, 8			; col 1 res in bits 7:4 of even bytes
	sub	rSrcAddr, 32		; point to curent dword quad

	packuswb mm1, mm0		; produce col0:1 bits in bits 7:4 of
					; all bytes
	movq	mm4, mm6		; prepare to compact col2:col3 bits

	movq	mm5, mm1		; prepare to compace col0:col1 bits
	psllq	mm4, 4			; put 8 col2:col3 bits in the odd bytes

	psllq	mm5, 4			; put 8 col0:col1 bits in the odd bytes
	por	mm1, mm5		; combine shifted col0:1 bits

	psrlw	mm1, 8			; 8 col0:1 bits in even bytes
	por	mm6, mm4		; combine shifted col2:3 bits

	psrlw	mm6, 8			; 8 col2:3 bits in even bytes
	cmp	rSrcAddr, srcAddr	; back at beginning of line yet?

; Here's where we wordswap the final dest word by flipping the order
; of the args to the pack command.  The dest appears in the LS word and
; the source appears in the MS word.  Therefore, we get 2:3:0:1, which
; is what we want.
	packuswb mm1, mm6
	jl	red4ColsAndExit		; quit if that was last dest word

; Now we need to byteswap the bytes in here.  Remember that each source word
; produced 16 bits, 8 from the upper word and 8 from the lower word.  We
; now need to swap the bytes, which corresponds to swapping the words in
; the original.
	movq	mm4, mm1		; make a copy
	psllw	mm1, 8			; shift low bytes to high bytes

	movq	mm3, [rSrcAddr + 24]	; get top word 3
	psrlw	mm4, 8			; shift high bytes to low bytes

	movq	mm2, [rSrcAddr + 16]	; get top word 2
	por	mm1, mm4		; combine swapped bytes

; wait for mm1 to percolate through pipeline
	; empty
	; empty

	movq	[rDstAddr + 8], mm1	; store dest word
	; empty


; ------------------- Get Bits from Words 1 & 0 ----------------------

; We can enter the loop at this point, so the source addresses must
; be consistent.
red4ColsAndLoopW1:
	movq	mm1, [MMWORD PTR all8sMask] ; nib result mask
	movq	mm4, mm2		; make copy of col 2 bits

	movq	mm0, [MMWORD PTR all1sMask] ; add this to LS 3 nib bits to
					; detect any 0's
	movq	mm5, mm3		; make copy of col 3 bits

	pand	mm4, mm7		; get rid of MS bit in each nibble
	pand	mm5, mm7		; MS bit of col 3 nibbles

	paddb	mm4, mm0		; set MS bit if any 1's in LS 3 bits
	paddb	mm5, mm0		; set nib result bits if low bits set

	pand	mm2, mm4		; set MS bit if all 1's in nibble
	pand	mm3, mm5		; set MS bit if all 1's in col 3 nibble

	movq	mm0, [MMWORD PTR all38sMask] ; carries bit 3 to bit 6
	pand	mm2, mm1		; select just nib result bits

	pand	mm3, mm1		; select just nib result bits
	paddb	mm2, mm0		; carry LS nib result bit up to
					; bit 6 of each byte.

	movq	mm1, [MMWORD PTR allC0Mask] ; get paired result bit mask
	paddb	mm3, mm0		; carry LS nib result bit up to
					; bit 6 of each byte.

	pand	mm3, mm1		; select just col 3 nib result bits
	pand	mm2, mm1		; select just col 2 nib result bits

	movq	mm4, mm2		; get ready to copy col 2 bit pairs
	movq	mm6, mm3		; get ready to copy col 3 bit pairs

	movq	mm0, [rSrcAddr]		; get next word 0
	psllq	mm4, 6			; put clumps in top end of upper bytes

	psllq	mm6, 6			; put clumps in top end of upper bytes
	por	mm2, mm4		; col 2 quad results in odd bytes

; Now we finish reading the col 0 inputs while col 3 consolidation catches up
	por	mm6, mm3		; col 3 quad results in odd bytes
	psrlw	mm2, 8			; col 2 res in bits 7:4 of even bytes

	movq	mm1, [rSrcAddr + 8] ; get next word 1
	psrlw	mm6, 8			; col 3 res in bits 7:4 of even bytes

	packuswb mm6, mm2		; produce col2:3 bits in bits 7:4 of
					; all bytes
	jmp	red4ColsAndLoopW3	; nope, do another quad

; yeah, if I were even nerdier that I am, I'd have the conditional jump
; exits the loop at the bottom instead of in the middle.  But figuring
; out how the loop was structured would be a pain in that case, so I'm
; leaving it this way.

; ----------------------------- LOOP EXIT -------------------------------

; At this point, we've almost finished creating the last dest dword.
; the only thing left is to byteswap it.  Do that and finish.
red4ColsAndExit:
	movq	mm4, mm1		; make a copy
	psllw	mm1, 8			; shift low bytes to high bytes

	psrlw	mm4, 8			; shift high bytes to low bytes

	por	mm1, mm4		; combine swapped bytes

; wait for mm1 to percolate through pipeline
	; empty
	; empty

	movq	[rDstAddr + 8], mm1	; store dest word
	ret

ip_compress4ColsAndLineLong ENDP





; *************************************************************************
; ***************************** 8x REDUCTION ******************************
; *************************************************************************

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; In our world, 8x reduction is a lot like 4x reduction.  But, didn't we
; say that using a table to unscramble interleaved bits was a win at
; 8x reduction?  Well, we did, but we lied.  The MMX pack opcodes can be
; used to our advantage for combining partial results.  Think of a tree.
; We use 4 pack opcodes to convert 8 dwords into 4 dwords.  We use two
; more pack opcodes to convert these to two dwords and one more to produce
; the final dword.
;
; In 8x reduction, we want to turn each byte into one bit.  Fairly quickly,
; we want to put 2 bits into every other byte so we can use the pack
; opcode.  This pack op will put 2 bits from 2 dwords into every byte.
; We can then put 4 bits into every other byte using shifts and ORs.
; We pack this with another dword that contains bits from 2 source dwords,
; producing a dword with 4 bits in every byte.  This dword contains the
; bits from 4 source dwords.  We repeat the entire process for four more
; dwords, ending up with two dwords, each containing the bits from 4
; source dwords.  We compact each of these so they contain 8 bits in
; every other byte, and pack them together to get the final result.
; This whole process is shown pictorially below.
;
; How do we deal with wordswapping?  Remember that the LS 32 bits of an
; dword read from the image contain the bits on the *left*.  As with
; 4x reduction, wordswapping the dest dword is easy, we just swap the
; arguments on the final pack opcode.  Dealing with swapping the source
; dword bits is a little trickier.
;
; Each source dword is reduced to 8 bits in 8x reduction (64/8 = 8).
; Each byte produces 1 bit.  During consolidation, there's a stage where
; we have a dword containing 4 bits/byte and we want to convert this to
; 8 bits in every other byte.  This is going from step 5 to step 6 below.
; The four bits in each byte all belong to the same 32-bit source word.
; We want to swap them with the 4 bits belonging to the other source word
; in the original source dword.  We do this by doing a right shift by
; 12 instead of a left shift by 4 as shown in step 6 below:
;
; In step 1, we combine the rows of the rightmost dword column of the
; 8-dword group of columns that will go into the next dest dword.
; We set each byte to all 1's or all 0's depending on the
; reduced value for that byte.  This is done with the pcmpeqb (packed
; compare for equality in units of bytes) opcode.
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 1:   0000000011111111222222223333333344444444555555556666666677777777
;
; In step 2, we consolidate 2 bits into the bits 2:3 of the odd (as counted
; from the MS end of the word) bytes.  This is done with a shift and a mask op.
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 2:   ..........01..............23..............45..............67....
;
; Steps 1 and 2 are repeated for the next dword to the left.  We pack
; the two resulting dwords into one dword.
; src dword   66      66      66      66      77      77      77      77
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 3:   ..01......23......45......67......01......23......45......67....
;
; Steps 1, 2 and 3 are repeated for 2 more source dwords.  We then shift
; and OR the two result dwords so they each have 4 bits at the MS end of each
; odd byte.  This is done by shifting the a copy of the dword *right* by
; 6 and ORing, producing ..01....|0123....| in the top bytes.  We then
; mask this with 00F0 00F0 00F0 00F0, which leaves the correct bits.
; src dword         6666            6666            7777            7777
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 4:   ........0123............4567............0123............4567....
;
; These two dwords are then packed together using the pack opcode to produce
; a dword with 4 bits in every byte.
; src dword 4444    4444    5555    5555    6666    6666    7777    7777    
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 5:   0123....4567....0123....4567....0123....4567....0123....4567....
;
; Steps 1 through 5 are repeated for the next 4 source dwords on the left
; producing another such dword containing bits from dwords 0, 1, 2 and 3.
; Here's the trick.  We have to consolidate each of these dwords so they
; have 8 bits in every other byte.  We do this by shift a copy right by 12
; and ORing with the original and selecting only the odd bytes, which
; produces:
; src dword         44444444        55555555        66666666        77777777
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 6:   ........45670123........45670123........45670123........45670123
;
; The bits from the LS word of the source dwords are now in the MS bits of
; the consolidated bytes, effectively word swapping the source dwords.  We
; do this operation at this stage since we had to do this combining op
; anyway.  We then pack two such source words, putting the 32 bit from the
; left in the image on the right in the dest dword.
; src dword 4444444455555555666666667777777700000000111111112222222233333333
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 7:   4567012345670123456701234567012345670123456701234567012345670123

; There's one more trick.  For ColOr reduction, we'd like to produce
; a 1 in the output if there are any 1's in the input.  The way we produce
; the original masks in step 1 is to use a packed compare MMX opcode.
; The closest we can come to comparing for any 1's is to compare for all
; 0's.  However, this produces a 1 when the input byte was all 0's, which
; is the inverse of what we want.  We fix it at the very end, by inverting
; the dest word before it's written.  For ColAnd reduction, we combine
; the bits of each byte by ANDing them.  If there are any 0's, the
; resulting byte will have some 0's.  We compare it with all 1's.  If there
; were any 0's, the result is 0, which is what we want.  Therefore, we do not
; need to invert the final dest word for ColAnd reduction.
;
; For ColOr, it would be swell to use an unsigned packed compare
; for greater than to detect any 1's, but there's no such operation.
; MMX only has signed packed compares.  Any negative number is less than 0x00
; even though it has bits turned on.  Drat.
;

; Parameters
; srcAddr	addr of left end of source line
; srcWidth	width of source image in bits
; dstAddr	addr of left end of dest line
; dstOffset	offset in bytes to rightmost dest word
; srcOffset	offset in bytes to leftmost byte in rightmost
;		group of source bits that will form the rightmost
;		dest word
; srcLimit	offset in bytes to rightmost source byte (often
;		to the right of byte at srcOffset)


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_compress8ColsOrLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_compress8ColsOrLineLongStart::
ip_compress8ColsOrLineLong PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword octet in the line.
; This octet may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	rSrcAddr, srcAddr	; pt to 1st source dword
	mov	rDstAddr, dstAddr

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first dword octet.
					; We do this because we're really
					; always going to go after the
					; first doubleword in an octet.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first octet.
	add	rDstAddr, dstOffset	; get addr of rightmost dest dword

; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; NOTE that we don't set the "results" from previous dwords here.  Since
; we want to pretend that they contain the results from reading 0's and
; the result of reading 0's is all 1's, we need to set 1's.  However,
; different loop entry points need different sets of 1's, so we set
; them in the loop entry code.
	mov	rScratch1, srcWidth	; get source width in bits.
; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword octet)
; can be 0, 8, 16, 24, 32, 40, 48 or 56.  If it's 0, we enter the loop
; where it reads one source dword before writing the dest.  If it's 8,
; we enter the loop where it reads 2 source dwords and so on.
	mov	rScratch, srcLimit	; get srcLimit

	and	rScratch1, 038h		; = (srcWidth in bytes MOD 8) * 8
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; is between 8N + 1 and 8N + 7.  In reducing by 8, we only look at 8N bits.
; Therefore, we never want to round up to the next byte!  So skip the
; rounding step and just subtract srcOffset.
	sub	rScratch, srcOffset

; Get the offset into the table of pointers to jump entry locations:
	shr	rScratch, 1		; [0, 8, 16, ...] -> [0, 4, 8, ...]
	; empty

; Simulate reading all 0's from cols 6,7.
	movq	mm4, [MMWORD PTR all30Mask]
	; empty

; Get the mask used to get rid of pad bits.
	movq	mm5, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	; empty

; Simulate reading all 0's from cols 4,5,6,7.
	movq	mm6, [MMWORD PTR allF0Mask]
	; empty

	movq	mm7, [MMWORD PTR all0030Mask]
	; empty

	; empty
	jmp	red8ColsOrJumpTable[rScratch]


;------------------------- LOOP ENTRIES  -----------------------------

; This is the loop entry taken when we want to read 8 columns of dwords
; to form the first dest dword.
red8ColsOrEntryW7:
	movq	mm3, [rSrcAddr + 7*8]			; col 7
	; empty

	movq	mm2, [rSrcAddr + 6*8]			; col 6
	pand	mm3, mm5

	pxor	mm5, mm5				; prepare for compare

	pcmpeqb	mm2, mm5				; any 1's?
	pcmpeqb	mm3, mm5				; any 1's?

	; empty
	jmp	red8ColsOrLoopEntryW54


; This is the loop entry taken when we want to read 7 columns of dwords
; to form the first dest dword.
red8ColsOrEntryW6:
	movq	mm2, [rSrcAddr + 6*8]			; col 6
	pcmpeqb	mm3, mm3				; simulate compared
							; col 7 data

	pand	mm2, mm5
	pxor	mm5, mm5				; prepare for compare

	pcmpeqb	mm2, mm5				; any 1's?
	jmp	red8ColsOrLoopEntryW54


; This is the loop entry taken when we want to read 6 columns of dwords
; to form the first dest dword.
red8ColsOrEntryW5:
	movq	mm1, [rSrcAddr + 5*8]			; col 5
	; empty

	movq	mm0, [rSrcAddr + 4*8]			; col 4
	pand	mm1, mm5

	movq	mm3, [MMWORD PTR all30Mask]		; simulate col 6,7 data
	pxor	mm5, mm5				; prepare for compare

	pcmpeqb	mm0, mm5				; any 1's in dword 4?
	jmp	red8ColsOrLoopEntryW32


; This is the loop entry taken when we want to read 5 columns of dwords
; to form the first dest dword.
red8ColsOrEntryW4:
	movq	mm0, [rSrcAddr + 4*8]			; col 4
	pxor	mm1, mm1				; simulate raw
							; col 5 data

	pand	mm0, mm5
	pxor	mm5, mm5				; prepare for compare

	movq	mm3, [MMWORD PTR all30Mask]		; simulate col 6,7 data
	pcmpeqb	mm0, mm5				; any 1's in dword 4?

	; empty
	jmp	red8ColsOrLoopEntryW32


; This is the loop entry taken when we want to read 4 columns of dwords
; to form the first dest dword.
red8ColsOrEntryW3:
	movq	mm3, [rSrcAddr + 3*8]			; col 3
	; empty

	movq	mm2, [rSrcAddr + 2*8]			; col 2
	pand	mm3, mm5

	pxor	mm5, mm5				; prepare for compare
	; empty

	pcmpeqb	mm2, mm5				; any 1's in dword 2?
	pcmpeqb	mm3, mm5				; any 1's in dword 3?

	; empty
	jmp	red8ColsOrLoopEntryW10


; This is the loop entry taken when we want to read 3 columns of dwords
; to form the first dest dword.
red8ColsOrEntryW2:
	movq	mm2, [rSrcAddr + 2*8]			; col 2
	pcmpeqb	mm3, mm3				; simulate compared
							; col 3 data

	pand	mm2, mm5
	pxor	mm5, mm5				; prepare for compare

	pcmpeqb	mm2, mm5				; any 1's in dword 2?
	jmp	red8ColsOrLoopEntryW10


; This is the loop entry taken when we want to read 2 columns of dwords
; to form the first dest dword.
red8ColsOrEntryW1:
	movq	mm1, [rSrcAddr + 1*8]			; col 1

	movq	mm0, [rSrcAddr + 0*8]			; col 0
	pand	mm1, mm5

	movq	mm3, [MMWORD PTR all30Mask]		; simulate col 2,3 data
	pxor	mm5, mm5				; prepare for compare

	sub	rSrcAddr, 64				; decrement addr reg
	jmp	red8ColsOrLoopEnd


; This is the loop entry taken when we want to read 1 column of dwords
; to form the first dest dword.
red8ColsOrEntryW0:
	movq	mm0, [rSrcAddr + 0*8]			; col 0
	pxor	mm1, mm1				; simulate raw
							; col 1 data

	pand	mm0, mm5
	pxor	mm5, mm5				; prepare for compare

	movq	mm3, [MMWORD PTR all30Mask]		; simulate col 2,3 data
	; empty

	sub	rSrcAddr, 64				; decrement addr reg
	jmp	red8ColsOrLoopEnd


; ------------------------- MAIN LOOP ------------------------------

; This is the main loop for 8x, ColsOr reduction.
;
; The basic loop structure is:
;  A. Read words 7,6 of a new dword octet and consolidate data from
;     dword 0 of the dword octet to the right.
;	mm0 = raw dword 0 data from dword octet to the right
;	mm1 = raw dword 1 data from dword octet to the right
;	mm2 = accumulating new raw dword 6 data
;	mm3 = accumulating new raw dword 7 data
;	mm4 = dword 2,3 data from dword octet to the right
;	mm5 = used to accumulate dword 0, 1 data
;	mm6 = dword 4,5,6,7 data from dword octet to the right
;  B. Combine the consolidated dword 0 data with that for dword 1.
;  C. Combine the dword 0,1 dword with dword 2,3 dword.
;  D. Combine dword 0,1,2,3 dword with dword 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
;  E. Write the result out to [rDstAddr] and decrement rDstAddr by 8.
;  F. If that was the last dest word, quit.
;  G. Read words 5,4 of this dword octet and consolidate data from
;     dwords 7.6 of this dword octet.
;	mm1 = accumulating raw dword 5 data
;	mm0 = accumulating raw dword 4 data
;	mm2 = raw dword 6 data
;	mm3 = raw dword 7 data
;	mm5 = used to consolidate dword 6,7 data
;  H. Combine consolidated dword 7 and consolidated dword 6 data.
;  I. Read words 3,2 of this dword octet and consolidate data from
;     dwords 5,4 of this dword octet.
;	mm0 = raw dword 4 data
;	mm1 = raw dword 5 data
;	mm2 = used to accumulate raw dword 2 data
;	mm3 = used to accumulate raw dword 3 data
;	mm4 = dword 6,7 data
;	mm5 = used to consolidate dword 4,5 data
;  J. Combine consolidated dword 5 and consolidated dword 4 data.
;  K. Combine consolidated dword 4,5 and consolidated dword 6,7 data.
;  L. Read dwords 1,0 of this dword octet and consolidate data from
;     dwords 3,2 of this dword octet.
;	mm0 = accumulating raw dword 0 data
;	mm1 = accumulating raw dword 1 data
;	mm2 = raw dword 2 data
;	mm3 = raw dword 3 data
;	mm5 = used to consolidate dword 2,3 data
;	mm6 = dword 4,5,6,7 data
;  M. Combine consolidated dword 2 and consolidated dword 3 data.
;  N. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
;  O. Since that was the last dword octet, do the consolidate of dword 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
;


; Here's the real loop code.  The comments above have been copied
; to the top of each section.
; In all sections:
;	mm5 = usually contains 0, sometimes used to shift stuff, then
;	      restored to 0.
;	mm7 = mask containing 0030 0030 0030 0030

;  A. Read words 7,6 of a new dword octet and consolidate data from
;     dword 0 of the dword octet to the right.
;	mm0 = raw dword 0 data from dword octet to the right
;	mm1 = raw dword 1 data from dword octet to the right
;	mm2 = accumulating new raw dword 6 data
;	mm3 = dword 2,3 data from dword octet to the right
;	mm4 = accumulating new raw dword 7 data
;	mm5 = used to accumulate dword 0, 1 data
;	mm6 = dword 4,5,6,7 data from dword octet to the right
red8ColsOrLoopEntryW76:
	pcmpeqb	mm0, mm5				; any 1's?
	pcmpeqb	mm1, mm5				; any 1's?

	movq	mm4, [rSrcAddr + 7*8]			; dword 7
	psrlq	mm0, 3					; res in odd bytes

	movq	mm2, [rSrcAddr + 6*8]			; dword 6
	psrlq	mm1, 3					; res in odd bytes

; We loop back here on the very last (leftmost) dword of the line to
; finish up.
red8ColsOrLoopFinish:
	pand	mm0, mm7				; select col 0 bits
	pand	mm1, mm7				; select col 1 bits

;  B. Combine the consolidated col 0 data with that for col 1.
	packuswb mm1, mm0				; col 0 bits on the left
;  C. Combine the col 0,1 dword with col 2,3 dword.
	movq	mm0, mm3				; prepare cols 2,3

	movq	mm5, mm1				; prepare cols 0,1
	psrlw	mm0, 6					; pack to bottom bytes

	psrlw	mm5, 6					; pack to bottom bytes
	por	mm3, mm0				; combine col 2,3 data

	pand	mm3, [MMWORD PTR all00F0Mask]		; mask off extra bits
	por	mm1, mm5				; combine col 0,1 data

	pand	mm1, [MMWORD PTR all00F0Mask]		; mask off extra bits
	; empty

;  Step C ends with packuswb below.
;  D. Combine col 0,1,2,3 dword with col 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
	packuswb mm3, mm1				; pack col 0,1,2,3 data
	movq	mm0, mm6				; get ready to align
							; col 4,5,6,7 data

	movq	mm5, mm3				; get ready to align
							; col 0,1,2,3 data
	psrlw	mm0, 12					; pack and "wordswap"
							; results from cols
							; 4,5,6,7.

	psrlw	mm5, 12					; pack and "wordswap"
							; results from cols
							; 0,1,2,3
	por	mm6, mm0				; combine results

	pand	mm6, [MMWORD PTR all00FFMask]		; get bytes we want
	por	mm3, mm5				; combine 0123 results

	pand	mm3, [MMWORD PTR all00FFMask]		; get bytes we want
	pcmpeqb	mm0, mm0				; get all 1's mask

; Note that we intentionally put the mm6 bits (cols 4,5,6,7) in the UPPER
; half of the result dword.  This accounts for the wordswapping inherent
; in this little endian platform.
	packuswb mm3, mm6				; get inverted dest dw
	pxor	mm5, mm5				; get ready to compare

	pxor	mm3, mm0				; invert dest word
							; because all result
							; bits were inverted
							; from original
							; pcmpeqb that made them
	sub	rDstAddr, 8				; point to prev word

; wait for mm3 and rDstAddr to percolate through pipeline
	pcmpeqb	mm2, mm5				; any 1's in dword 6?
	pcmpeqb	mm4, mm5				; any 1's in dword 7?

;  E. Write the result out to [dstAddr] and decrement dstAddr by 8.
	movq	[rDstAddr + 8], mm3			; write result dword
	movq	mm3, mm4				; where next code
							; wants dword 7

;  F. If that was the last dest word, quit.
	cmp	rSrcAddr, srcAddr
	jl	red8ColsOrExit		; quit if that was last dest word


;  G. Read words 5,4 of this dword octet and consolidate data from
;     dwords 7.6 of this dword octet.
;	mm1 = accumulating raw dword 5 data
;	mm0 = accumulating raw dword 4 data
;	mm2 = compared dword 6 data
;	mm3 = compared dword 7 data
;	mm5 = used to consolidate dword 6,7 data
red8ColsOrLoopEntryW54:
	movq	mm1, [rSrcAddr + 5*8]			; dword 5
	psrlq	mm2, 3					; res in odd bytes

	movq	mm0, [rSrcAddr + 4*8]			; dword 4
	psrlq	mm3, 3					; res in odd bytes

	pand	mm2, mm7				; select col 6 bits
	pand	mm3, mm7				; select col 7 bits

;  H. Combine consolidated dword 7 and consolidated dword 6 data.
	packuswb mm3, mm2				; col 6 bits on the left
	pcmpeqb	mm0, mm5				; any 1's in dword 4?

;  I. Read words 3,2 of this dword octet and consolidate data from
;     dwords 5,4 of this dword octet.
;	mm0 = compared dword 4 data
;	mm1 = raw dword 5 data
;	mm2 = used to accumulate raw dword 2 data
;	mm3 = dword 6,7 data / used to accumulate raw dword 3 data
;	mm4 = gets dword 6,7 data
;	mm5 = used to consolidate dword 4,5 data
red8ColsOrLoopEntryW32:
	movq	mm4, mm3				; move dword 6,7 data
	pcmpeqb	mm1, mm5				; any 1's in dword 5?

	movq	mm3, [rSrcAddr + 3*8]			; dword 3
	psrlq	mm0, 3					; res in odd bytes

	movq	mm2, [rSrcAddr + 2*8]			; dword 2
	psrlq	mm1, 3					; res in odd bytes

	pand	mm0, mm7				; select dword 4 bits
	pand	mm1, mm7				; select dword 5 bits

;  J. Combine consolidated dword 5 and consolidated dword 4 data.
	packuswb mm1, mm0				; col 4 bits on the left
;  K. Combine consolidated dword 4,5 and consolidated dword 6,7 data.
	movq	mm6, mm4				; prepare dwords 6,7

	movq	mm0, mm1				; prepare dwords 4,5
	psrlw	mm6, 6					; pack to bottom bytes

	psrlw	mm0, 6					; pack to bottom bytes
	por	mm6, mm4				; combine dword 6,7 data

	pand	mm6, [MMWORD PTR all00F0Mask]		; mask off extra bits
	por	mm1, mm0				; combine dword 4,5 data

	pand	mm1, [MMWORD PTR all00F0Mask]		; mask off extra bits
	pcmpeqb	mm2, mm5				; any 1's in dword 2?

	packuswb mm6, mm1				; create 4,5,6,7 data
	pcmpeqb	mm3, mm5				; any 1's in dword 3?

;  L. Read dwords 1,0 of this dword octet and consolidate data from
;     dwords 3,2 of this dword octet.
;	mm0 = accumulating raw dword 0 data
;	mm1 = accumulating raw dword 1 data
;	mm2 = compared dword 2 data
;	mm3 = compared dword 3 data
;	mm5 = used to consolidate dword 2,3 data
;	mm6 = dword 4,5,6,7 data
red8ColsOrLoopEntryW10:
	movq	mm1, [rSrcAddr + 1*8]			; dword 1
	psrlq	mm2, 3					; res in odd bytes

	movq	mm0, [rSrcAddr + 0*8]			; dword 0
	psrlq	mm3, 3					; res in odd bytes

	pand	mm2, mm7				; select col 6 bits
	pand	mm3, mm7				; select col 7 bits

;  M. Combine consolidated dword 2 and consolidated dword 3 data.
	packuswb mm3, mm2				; col 2 bits on the left
;  N. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
	sub	rSrcAddr, 64

; This is the loop entry point when the rightmost dword octet has 1 or 2
; dwords in it.  We do some setup, then jump in here.
red8ColsOrLoopEnd:
	cmp	rSrcAddr, srcAddr			; done src dwords yet?
	jge	red8ColsOrLoopEntryW76			; no, do another octet


;  O. Since that was the last dword octet, do the consolidate of col 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
	pcmpeqb	mm0, mm5				; any 1's in col 0?
	pcmpeqb	mm1, mm5				; any 1's in col 1?

	psrlq	mm0, 3					; res in odd bytes

	psrlq	mm1, 3					; res in odd bytes
	jmp	red8ColsOrLoopFinish


; We get here after the last (leftmost) dword was written (Step F)
red8ColsOrExit:
	ret

ip_compress8ColsOrLineLong ENDP




; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_compress8ColsAndLineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_compress8ColsAndLineLongStart::
ip_compress8ColsAndLineLong PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword octet in the line.
; This octet may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	rSrcAddr, srcAddr	; pt to 1st source dword
	mov	rDstAddr, dstAddr

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first dword octet.
					; We do this because we're really
					; always going to go after the
					; first doubleword in an octet.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first octet.
	add	rDstAddr, dstOffset	; get addr of rightmost dest dword

; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; NOTE that we don't set the "results" from previous dwords here.  Since
; we want to pretend that they contain the results from reading 0's and
; the result of reading 0's is all 1's, we need to set 1's.  However,
; different loop entry points need different sets of 1's, so we set
; them in the loop entry code.
	mov	rScratch1, srcWidth	; get source width in bits.
; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword octet)
; can be 0, 8, 16, 24, 32, 40, 48 or 56.  If it's 0, we enter the loop
; where it reads one source dword before writing the dest.  If it's 8,
; we enter the loop where it reads 2 source dwords and so on.
	mov	rScratch, srcLimit	; get srcLimit

	and	rScratch1, 038h		; = (srcWidth in bytes MOD 8) * 8
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; is between 8N + 1 and 8N + 7.  In reducing by 8, we only look at 8N bits.
; Therefore, we never want to round up to the next byte!  So skip the
; rounding step and just subtract srcOffset.
	sub	rScratch, srcOffset

; Get the offset into the table of pointers to jump entry locations:
	shr	rScratch, 1		; [0, 8, 16, ...] -> [0, 4, 8, ...]
	; empty

; Simulate reading all 0's from cols 6,7.
	pxor	mm4, mm4
	; empty

; Get the mask used to get rid of pad bits.
	movq	mm5, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	; empty

; Simulate reading all 0's from cols 4,5,6,7.
	pxor	mm6, mm6
	; empty

	movq	mm7, [MMWORD PTR all0030Mask]
	; empty

	; empty
	jmp	red8ColsAndJumpTable[rScratch]


;------------------------- LOOP ENTRIES  -----------------------------

; This is the loop entry taken when we want to read 8 columns of dwords
; to form the first dest dword.
red8ColsAndEntryW7:
	movq	mm3, [rSrcAddr + 7*8]			; col 7
	; empty

	movq	mm2, [rSrcAddr + 6*8]			; col 6
	pand	mm3, mm5

	pcmpeqb	mm5, mm5				; prepare for compare

	pcmpeqb	mm2, mm5				; all 1's?
	pcmpeqb	mm3, mm5				; all 1's?

	; empty
	jmp	red8ColsAndLoopEntryW54


; This is the loop entry taken when we want to read 7 columns of dwords
; to form the first dest dword.
red8ColsAndEntryW6:
	movq	mm2, [rSrcAddr + 6*8]			; col 6
	pxor	mm3, mm3				; simulate compared
							; col 7 data

	pand	mm2, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	pcmpeqb	mm2, mm5				; all 1's?
	jmp	red8ColsAndLoopEntryW54


; This is the loop entry taken when we want to read 6 columns of dwords
; to form the first dest dword.
red8ColsAndEntryW5:
	movq	mm1, [rSrcAddr + 5*8]			; col 5
	; empty

	movq	mm0, [rSrcAddr + 4*8]			; col 4
	pand	mm1, mm5

	pxor	mm3, mm3				; simulate col 6,7 data
	pcmpeqb	mm5, mm5				; prepare for compare

	pcmpeqb	mm0, mm5				; all 1's in dword 4?
	jmp	red8ColsAndLoopEntryW32


; This is the loop entry taken when we want to read 5 columns of dwords
; to form the first dest dword.
red8ColsAndEntryW4:
	movq	mm0, [rSrcAddr + 4*8]			; col 4
	pxor	mm1, mm1				; simulate raw
							; col 5 data

	pand	mm0, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	pxor	mm3, mm3				; simulate col 6,7 data
	pcmpeqb	mm0, mm5				; all 1's in dword 4?

	; empty
	jmp	red8ColsAndLoopEntryW32


; This is the loop entry taken when we want to read 4 columns of dwords
; to form the first dest dword.
red8ColsAndEntryW3:
	movq	mm3, [rSrcAddr + 3*8]			; col 3
	; empty

	movq	mm2, [rSrcAddr + 2*8]			; col 2
	pand	mm3, mm5

	pcmpeqb	mm5, mm5				; prepare for compare
	; empty

	pcmpeqb	mm2, mm5				; all 1's in dword 2?
	pcmpeqb	mm3, mm5				; all 1's in dword 3?

	; empty
	jmp	red8ColsAndLoopEntryW10


; This is the loop entry taken when we want to read 3 columns of dwords
; to form the first dest dword.
red8ColsAndEntryW2:
	movq	mm2, [rSrcAddr + 2*8]			; col 2
	pxor	mm3, mm3				; simulate compared
							; col 3 data

	pand	mm2, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	pcmpeqb	mm2, mm5				; all 1's in dword 2?
	jmp	red8ColsAndLoopEntryW10


; This is the loop entry taken when we want to read 2 columns of dwords
; to form the first dest dword.
red8ColsAndEntryW1:
	movq	mm1, [rSrcAddr + 1*8]			; col 1

	movq	mm0, [rSrcAddr + 0*8]			; col 0
	pand	mm1, mm5

	pxor	mm3, mm3				; simulate col 2,3 data
	pcmpeqb	mm5, mm5				; prepare for compare

	sub	rSrcAddr, 64				; decrement addr reg
	jmp	red8ColsAndLoopEnd


; This is the loop entry taken when we want to read 1 column of dwords
; to form the first dest dword.
red8ColsAndEntryW0:
	movq	mm0, [rSrcAddr + 0*8]			; col 0
	pxor	mm1, mm1				; simulate raw
							; col 1 data

	pand	mm0, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	pxor	mm3, mm3				; simulate col 2,3 data
	sub	rSrcAddr, 64				; decrement addr reg

	; empty
	jmp	red8ColsAndLoopEnd


; ------------------------- MAIN LOOP ------------------------------

; This is the main loop for 8x, ColsAnd reduction.
;
; The basic loop structure is:
;  A. Read words 7,6 of a new dword octet and consolidate data from
;     dword 0 of the dword octet to the right.
;	mm0 = raw dword 0 data from dword octet to the right
;	mm1 = raw dword 1 data from dword octet to the right
;	mm2 = accumulating new raw dword 6 data
;	mm3 = accumulating new raw dword 7 data
;	mm4 = dword 2,3 data from dword octet to the right
;	mm5 = used to accumulate dword 0, 1 data
;	mm6 = dword 4,5,6,7 data from dword octet to the right
;  B. Combine the consolidated dword 0 data with that for dword 1.
;  C. Combine the dword 0,1 dword with dword 2,3 dword.
;  D. Combine dword 0,1,2,3 dword with dword 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
;  E. Write the result out to [rDstAddr] and decrement rDstAddr by 8.
;  F. If that was the last dest word, quit.
;  G. Read words 5,4 of this dword octet and consolidate data from
;     dwords 7.6 of this dword octet.
;	mm1 = accumulating raw dword 5 data
;	mm0 = accumulating raw dword 4 data
;	mm2 = raw dword 6 data
;	mm3 = raw dword 7 data
;	mm5 = used to consolidate dword 6,7 data
;  H. Combine consolidated dword 7 and consolidated dword 6 data.
;  I. Read words 3,2 of this dword octet and consolidate data from
;     dwords 5,4 of this dword octet.
;	mm0 = raw dword 4 data
;	mm1 = raw dword 5 data
;	mm2 = used to accumulate raw dword 2 data
;	mm3 = used to accumulate raw dword 3 data
;	mm4 = dword 6,7 data
;	mm5 = used to consolidate dword 4,5 data
;  J. Combine consolidated dword 5 and consolidated dword 4 data.
;  K. Combine consolidated dword 4,5 and consolidated dword 6,7 data.
;  L. Read dwords 1,0 of this dword octet and consolidate data from
;     dwords 3,2 of this dword octet.
;	mm0 = accumulating raw dword 0 data
;	mm1 = accumulating raw dword 1 data
;	mm2 = raw dword 2 data
;	mm3 = raw dword 3 data
;	mm5 = used to consolidate dword 2,3 data
;	mm6 = dword 4,5,6,7 data
;  M. Combine consolidated dword 2 and consolidated dword 3 data.
;  N. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
;  O. Since that was the last dword octet, do the consolidate of dword 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
;


; Here's the real loop code.  The comments above have been copied
; to the top of each section.
; In all sections:
;	mm5 = usually contains 0, sometimes used to shift stuff, then
;	      restored to 0.
;	mm7 = mask containing 0030 0030 0030 0030

;  A. Read words 7,6 of a new dword octet and consolidate data from
;     dword 0 of the dword octet to the right.
;	mm0 = raw dword 0 data from dword octet to the right
;	mm1 = raw dword 1 data from dword octet to the right
;	mm2 = accumulating new raw dword 6 data
;	mm3 = dword 2,3 data from dword octet to the right
;	mm4 = accumulating new raw dword 7 data
;	mm5 = used to accumulate dword 0, 1 data
;	mm6 = dword 4,5,6,7 data from dword octet to the right
red8ColsAndLoopEntryW76:
	pcmpeqb	mm0, mm5				; all 1's?
	pcmpeqb	mm1, mm5				; all 1's?

	movq	mm4, [rSrcAddr + 7*8]			; dword 7
	psrlq	mm0, 3					; res in odd bytes

	movq	mm2, [rSrcAddr + 6*8]			; dword 6
	psrlq	mm1, 3					; res in odd bytes

; We loop back here on the very last (leftmost) dword of the line to
; finish up.
red8ColsAndLoopFinish:
	pand	mm0, mm7				; select col 0 bits
	pand	mm1, mm7				; select col 1 bits

;  B. Combine the consolidated col 0 data with that for col 1.
	packuswb mm1, mm0				; col 0 bits on the left
;  C. Combine the col 0,1 dword with col 2,3 dword.
	movq	mm0, mm3				; prepare cols 2,3

	movq	mm5, mm1				; prepare cols 0,1
	psrlw	mm0, 6					; pack to bottom bytes

	psrlw	mm5, 6					; pack to bottom bytes
	por	mm3, mm0				; combine col 2,3 data

	pand	mm3, [MMWORD PTR all00F0Mask]		; mask off extra bits
	por	mm1, mm5				; combine col 0,1 data

	pand	mm1, [MMWORD PTR all00F0Mask]		; mask off extra bits
	; empty

;  Step C ends with packuswb below.
;  D. Combine col 0,1,2,3 dword with col 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
	packuswb mm3, mm1				; pack col 0,1,2,3 data
	movq	mm0, mm6				; get ready to align
							; col 4,5,6,7 data

	movq	mm5, mm3				; get ready to align
							; col 0,1,2,3 data
	psrlw	mm0, 12					; pack and "wordswap"
							; results from cols
							; 4,5,6,7.

	psrlw	mm5, 12					; pack and "wordswap"
							; results from cols
							; 0,1,2,3
	por	mm6, mm0				; combine results

	pand	mm6, [MMWORD PTR all00FFMask]		; get bytes we want
	por	mm3, mm5				; combine 0123 results

; Since the way we generated the result bit from each 8 bits produced
; a 1 if ALL the bits in the square are on, (0 if any of the input bits
; was 0) we don't need to invert the result, so don't do it.
	pand	mm3, [MMWORD PTR all00FFMask]		; get bytes we want
	pcmpeqb	mm5, mm5				; get ready to compare

; Note that we intentionally put the mm6 bits (cols 4,5,6,7) in the UPPER
; half of the result dword.  This accounts for the wordswapping inherent
; in this little endian platform.
	packuswb mm3, mm6				; get inverted dest dw
	sub	rDstAddr, 8				; point to prev word

; wait for mm3 and rDstAddr to percolate through pipeline
	pcmpeqb	mm2, mm5				; all 1's in dword 6?
	pcmpeqb	mm4, mm5				; all 1's in dword 7?

;  E. Write the result out to [dstAddr] and decrement dstAddr by 8.
	movq	[rDstAddr + 8], mm3			; write result dword
	movq	mm3, mm4				; where next code
							; wants dword 7

;  F. If that was the last dest word, quit.
	cmp	rSrcAddr, srcAddr
	jl	red8ColsAndExit		; quit if that was last dest word


;  G. Read words 5,4 of this dword octet and consolidate data from
;     dwords 7.6 of this dword octet.
;	mm1 = accumulating raw dword 5 data
;	mm0 = accumulating raw dword 4 data
;	mm2 = compared dword 6 data
;	mm3 = compared dword 7 data
;	mm5 = used to consolidate dword 6,7 data
red8ColsAndLoopEntryW54:
	movq	mm1, [rSrcAddr + 5*8]			; dword 5
	psrlq	mm2, 3					; res in odd bytes

	movq	mm0, [rSrcAddr + 4*8]			; dword 4
	psrlq	mm3, 3					; res in odd bytes

	pand	mm2, mm7				; select col 6 bits
	pand	mm3, mm7				; select col 7 bits

;  H. Combine consolidated dword 7 and consolidated dword 6 data.
	packuswb mm3, mm2				; col 6 bits on the left
	pcmpeqb	mm0, mm5				; all 1's in dword 4?

;  I. Read words 3,2 of this dword octet and consolidate data from
;     dwords 5,4 of this dword octet.
;	mm0 = compared dword 4 data
;	mm1 = raw dword 5 data
;	mm2 = used to accumulate raw dword 2 data
;	mm3 = dword 6,7 data / used to accumulate raw dword 3 data
;	mm4 = gets dword 6,7 data
;	mm5 = used to consolidate dword 4,5 data
red8ColsAndLoopEntryW32:
	movq	mm4, mm3				; move dword 6,7 data
	pcmpeqb	mm1, mm5				; all 1's in dword 5?

	movq	mm3, [rSrcAddr + 3*8]			; dword 3
	psrlq	mm0, 3					; res in odd bytes

	movq	mm2, [rSrcAddr + 2*8]			; dword 2
	psrlq	mm1, 3					; res in odd bytes

	pand	mm0, mm7				; select dword 4 bits
	pand	mm1, mm7				; select dword 5 bits

;  J. Combine consolidated dword 5 and consolidated dword 4 data.
	packuswb mm1, mm0				; col 4 bits on the left
;  K. Combine consolidated dword 4,5 and consolidated dword 6,7 data.
	movq	mm6, mm4				; prepare dwords 6,7

	movq	mm0, mm1				; prepare dwords 4,5
	psrlw	mm6, 6					; pack to bottom bytes

	psrlw	mm0, 6					; pack to bottom bytes
	por	mm6, mm4				; combine dword 6,7 data

	pand	mm6, [MMWORD PTR all00F0Mask]		; mask off extra bits
	por	mm1, mm0				; combine dword 4,5 data

	pand	mm1, [MMWORD PTR all00F0Mask]		; mask off extra bits
	pcmpeqb	mm2, mm5				; all 1's in dword 2?

	packuswb mm6, mm1				; create 4,5,6,7 data
	pcmpeqb	mm3, mm5				; all 1's in dword 3?

;  L. Read dwords 1,0 of this dword octet and consolidate data from
;     dwords 3,2 of this dword octet.
;	mm0 = accumulating raw dword 0 data
;	mm1 = accumulating raw dword 1 data
;	mm2 = compared dword 2 data
;	mm3 = compared dword 3 data
;	mm5 = used to consolidate dword 2,3 data
;	mm6 = dword 4,5,6,7 data
red8ColsAndLoopEntryW10:
	movq	mm1, [rSrcAddr + 1*8]			; dword 1
	psrlq	mm2, 3					; res in odd bytes

	movq	mm0, [rSrcAddr + 0*8]			; dword 0
	psrlq	mm3, 3					; res in odd bytes

	pand	mm2, mm7				; select col 6 bits
	pand	mm3, mm7				; select col 7 bits

;  M. Combine consolidated dword 2 and consolidated dword 3 data.
	packuswb mm3, mm2				; col 2 bits on the left
;  N. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
	sub	rSrcAddr, 64

; This is the loop entry point when the rightmost dword octet has 1 or 2
; dwords in it.  We do some setup, then jump in here.
red8ColsAndLoopEnd:
	cmp	rSrcAddr, srcAddr			; done src dwords yet?
	jge	red8ColsAndLoopEntryW76			; no, do another octet


;  O. Since that was the last dword octet, do the consolidate of col 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
	pcmpeqb	mm0, mm5				; all 1's in col 0?
	pcmpeqb	mm1, mm5				; all 1's in col 1?

	psrlq	mm0, 3					; res in odd bytes

	psrlq	mm1, 3					; res in odd bytes
	jmp	red8ColsAndLoopFinish


; We get here after the last (leftmost) dword was written (Step F)
red8ColsAndExit:
	ret

ip_compress8ColsAndLineLong ENDP








; Segment for initialized variables, like masks
_DATA	SEGMENT PUBLIC 'DATA'

		ALIGN	8
lsBitMask1Bit	QWORD	05555555555555555h
lsBitMask2Bits	QWORD	03333333333333333h
lsBitMask4Bits	QWORD	00F0F0F0F0F0F0F0Fh
lsBitMask8Bits	QWORD	000FF00FF00FF00FFh

; masks used for 4x reduction
		ALIGN	8
all1sMask	QWORD	01111111111111111h
all7sMask	QWORD	07777777777777777h
all8sMask	QWORD	08888888888888888h
allC0Mask	QWORD	0C0C0C0C0C0C0C0C0h
all38sMask	QWORD	03838383838383838h

; masks used for 8x reduction
all30Mask	QWORD	03030303030303030h
all0030Mask	QWORD	00030003000300030h
allF0Mask	QWORD	0F0F0F0F0F0F0F0F0h
all00F0Mask	QWORD	000F000F000F000F0h
all00FFMask	QWORD	000FF00FF00FF00FFh

; These are the jump tables used to get to the correct loop
; entry point for 8x reduction.
red8ColsOrJumpTable DWORD	red8ColsOrEntryW0,
				red8ColsOrEntryW1,
				red8ColsOrEntryW2,
				red8ColsOrEntryW3,
				red8ColsOrEntryW4,
				red8ColsOrEntryW5,
				red8ColsOrEntryW6,
				red8ColsOrEntryW7

red8ColsAndJumpTable DWORD	red8ColsAndEntryW0,
				red8ColsAndEntryW1,
				red8ColsAndEntryW2,
				red8ColsAndEntryW3,
				red8ColsAndEntryW4,
				red8ColsAndEntryW5,
				red8ColsAndEntryW6,
				red8ColsAndEntryW7

_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no initialized variables
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcredcm.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.2  1998/01/13 23:58:29  ddavies
pcredcm.asm -> /product/ipcore/ipshared/src/RCS/pcredcm.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.1  1996/11/04 21:49:08  ddavies
Initial revision




^

ENDIF




