;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcredm.asm:  2x threshold reduction using MMX opcodes.

	.586		; compile code for a Pentium processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; This code uses MMX opcodes, which have 64-bit accesses.  That's the
; main source of our speedup.  These routines are only called if
; the caller detects MMX code.  We don't have to detect it here.
; First, include the file that gives us the MMX macros.

INCLUDE	iammx.inc

_DATA	SEGMENT PUBLIC 'DATA'
; We need to access a mask that's used to get rid of pad bits.
EXTERN	mmxByteEdgeMask:QWORD

RCSInfo	BYTE	"$RCSfile: pcredm.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

;Register Assignment:
; EAX	scratch
; EBX	scratch
; ECX	pointer to lower source line
; EDX	final source address
; EBP	pts to variables on the stack
; ESI	source address
; EDI	dest address
; ESP	stack pointer

; MM0	upper/remaining word of left source cell
; MM1	lower word of left source cell / scratch
; MM2	upper/remaining word of right source cell
; MM3	lower word of right source cell / scratch
; the other registers are mainly used when computing thresholds
; 2 and 3.  There, we have to do independent checks for pixels in
; two rows and pixels in two columns of each 2x2 square.
; MM4	upper/remaining word for left source cell for col check
; MM5	lower word of left source cell / scratch for col check
; MM6	upper/remaining word of right source cell for col check
; MM7	lower word of right source cell / scratch for col check

rScratch1	EQU	eax
rScratch2	EQU	ebx
rSrcAddrLo	EQU	ecx
rSrcFin		EQU	edx
rSrcAddr	EQU	esi
rDstAddr	EQU	edi

; Parameters
; srcAddr	addr of left end of source line
; srcWidth	width of source line in bits
; srcBpl	source bytes/line
; dstAddr	addr of left end of dest line
; dstOffset	offset in bytes to rightmost dest word
; srcOffset	offset in bytes to leftmost byte in rightmost
;		group of source bits that wil form the rightmost
;		dest word
; srcLimit	offset in bytes to rightmost source byte (often
;		to the right of byte at srcOffset)
; tab2to1Permute addr of table used to unscramble reduced bits.
;		This is not used here.

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_longOpScaleLibLinked.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses no registers

; This routine simply returns cTrue to indicate that the long operations are
; linked in.

ip_longOpScaleLibLinkedStart::
ip_longOpScaleLibLinked	PROC NEAR C PUBLIC,\
	pIPState:DWORD

; make the assembler happy
	mov	eax, pIPState

	mov	eax, 1		; value of cTrue
	ret
ip_longOpScaleLibLinked ENDP



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduceT1LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduceT1LineLongStart::
ip_reduceT1LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab2to1Permute:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tab2to1Permute

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword pair in the line.
; This pair may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
; rSrcFin will point to the leftmost doubleword in the source line.
	mov	rSrcFin, srcAddr	; pt to 1st source dword
	mov	rSrcAddr, rSrcFin

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first left/right dword pair.
					; We do this because we're really
					; always going to go after the
					; even doubleword in an even/odd
					; pair.  If the rightmost dest
					; dword gets bits from two source
					; dwords, we use the regular loop.
					; If the rightmost dest dword gets
					; bits from only 1 source dword, we
					; use the code just below.
	mov	rSrcAddrLo, rSrcAddr

	add	rSrcAddrLo, srcBpl	; point to second line
	mov	rDstAddr, dstAddr	; pt to leftmost dest doubleword

	add	rDstAddr, dstOffset	; pt to rightmost dest doubleword
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 1.  In reducing by 2, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 2,
; 8N + 3, etc.  So, we only add 6, not 7.
	mov	rScratch2, srcWidth	; get source width in bits.

	add	rScratch2, 6		; round up to next byte
	mov	rScratch1, srcOffset

; The first dest doubleword could be composed of one or two source
; doublewords.  By the way, we know that it's safe to read doublewords
; because the caller checked to see that all the source and dest lines
; begin on multiples of 8 and have bytes/line that are multiples of 8.
; If the first dest word required one doubleword, do the special code
; below.  Otherwise, skip into the main loop.
	and	rScratch2, 038h		; = (src width in bytes MOD 8) * 8
	cmp	rScratch1, srcLimit	; do these point to the same word?
	je	redT1RightDWord		; yes, then the first dest dword
					; contains data from only one source
					; dword.  Go encode that.

; This is the preamble to the regular loop that encodes a dest dword from
; two source dwords.  It looks like the top of the regular loop, except
; it doesn't write out a previous dest dword.
	movq	mm0, [rSrcAddr]		; get upper left doubleword

	movq	mm1, [rSrcAddrLo]	; get lower left doubleword

	movq	mm2, [rSrcAddr+ 8 ]	 ; get upper right doubleword
	por	mm0, mm1		; combine left rows

	movq	mm1, mm0		; get ready to combine left columns
	; empty

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	psrlq	mm1, 1			; move to lower end of each bit pair

	sub	rSrcAddr, 16		; pt to previous doubleword pair
	por	mm2, mm3		; combine right rows

	pand	mm2, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	jmp	redT1LongLoopEntry	; go to rest of the loop



; The first dest doubleword is composed of only one source doubleword and
; a word of 0's.  Compute this below.  Remember that rSrcAddr and
; rSrcAddrLo already point to the right dwords.
redT1RightDWord:
	movq	mm0, [rSrcAddr]		; get upper doubleword
	sub	rSrcAddr, 16		; pt to previous doubleword pair

	movq	mm1, [rSrcAddrLo]	; get lower doubleword
	sub	rSrcAddrLo, 16

	por	mm0, mm1		; combine rows

; get rid of pad bits
	pand	mm0, [MMWORD PTR mmxByteEdgeMask + rScratch2]

	movq	mm1, mm0		; get ready to combine columns

	psrlq	mm1, 1			; move to lower end of each bit pair

	por	mm0, mm1		; get result in lower bits

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just those bits

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1

	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	movq	mm1, mm0		; get ready to make clumps of 2 bits

	psrlq	mm1, 1

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off

	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2

	por	mm0, mm1

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte

	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4

	por	mm4, mm0
	sub	rDstAddr, 8		; prepare for next loop

	pand	mm4, [MMWORD PTR lsBitMask8Bits] ; got them.
	pxor	mm2, mm2		; clear "other half word"

; the byte packing and writing are done in the main loop.
	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jl	redT1LongFin		; quit if we're done


; ----------------------------------------------------------------
; MAIN LOOP
; ----------------------------------------------------------------

; On entering this loop, rSrcAddr should be pointing 16 bytes to the
; left of the first doubleword of a doubleword pair,
; rDstAddr should be pointing to first doubleword to be written,
; rSrcFin should be pointing to leftmost doubleword of the source line.
; mm0 should already be loaded with the left doubleword of the pair
; and mm1 should be loaded with the word below that.
redT1LongLoop:
	movq	mm0, [rSrcAddr]		; get upper left doubleword
	pand	mm2, mm7		; got bytes in right dword.

; Pack left bits into the *lower* half of a word.  Remember that the lower
; half is the left half in this wacky world.
	movq	mm1, [rSrcAddrLo]	; get lower left doubleword
	packuswb mm4, mm2		; pack the bytes

; This clock cycle takes place while we're waiting for mm4 to progress
; through the pipeline.
	movq	mm2, [rSrcAddr+ 8 ]	 ; get upper right doubleword
	por	mm0, mm1		; combine left rows

	movq	[rDstAddr + 8], mm4	; write the previous result
	movq	mm1, mm0		; get ready to combine left columns

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	psrlq	mm1, 1			; move to lower end of each bit pair

	sub	rSrcAddr, 16		; pt to previous doubleword pair
	por	mm2, mm3		; combine right rows

; We enter the loop here, with the proper preamble, if the first
; dest doubleword contains bits from two source dwords.
redT1LongLoopEntry:
	por	mm0, mm1		; get left result in lower bits
	movq	mm3, mm2		; get ready to combine right columns

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just left bits
	psrlq	mm3, 1			; move to lower end of each bit pair

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1
	por	mm2, mm3		; get right result in lower bits

	pand	mm2, [MMWORD PTR lsBitMask1Bit]	; select just right bits
	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	sub	rSrcAddrLo, 16		; point to prev doubleword pair
	punpckldq mm3, mm2		; put low halfword of mm2 into
					; high halfword of mm3

	movq	mm1, mm0		; get ready to make clumps of 2 bits
	punpckhdq mm2, mm3		; complete the word swap

	psrlq	mm1, 1
	movq	mm3, mm2		; get ready to make clumps of 2 bits

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble
	psrlq	mm3, 1

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off
	por	mm2, mm3		; make clumps in lower 2 bits of
					; each nibble

	pand	mm2, [MMWORD PTR lsBitMask2Bits] ; mask bits in right dword
	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2
	movq	mm3, mm2		; get ready to make clumps of 4 bits

	por	mm0, mm1
	psrlq	mm3, 2

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	por	mm2, mm3

	pand	mm2, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4
	movq	mm3, mm2		; get ready to make clumps of bytes

	por	mm4, mm0
	psrlq	mm3, 4

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get final mask.
	por	mm2, mm3

; finish right dword at top of loop
	pand	mm4, mm7		; got bytes in left dword.
	sub	rDstAddr, 8		; get ready for next trip through loop

	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jge	redT1LongLoop		; no, keep going

; We're done with that line.
redT1LongFin:
	pand	mm2, mm7		; got bytes in right dword.

	packuswb mm4, mm2		; pack the bytes

	movq	[rDstAddr + 8], mm4	; write the last word
redT1LongRet:
	ret

ip_reduceT1LineLong	ENDP


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduceT2LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduceT2LineLongStart::
ip_reduceT2LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab2to1Permute:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tab2to1Permute

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword pair in the line.
; This pair may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
; rSrcFin will point to the leftmost doubleword in the source line.
	mov	rSrcFin, srcAddr	; pt to 1st source dword
	mov	rSrcAddr, rSrcFin

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first left/right dword pair.
					; We do this because we're really
					; always going to go after the
					; even doubleword in an even/odd
					; pair.  If the rightmost dest
					; dword gets bits from two source
					; dwords, we use the regular loop.
					; If the rightmost dest dword gets
					; bits from only 1 source dword, we
					; use the code just below.
	mov	rSrcAddrLo, rSrcAddr

	add	rSrcAddrLo, srcBpl	; point to second line
	mov	rDstAddr, dstAddr	; pt to leftmost dest doubleword

	add	rDstAddr, dstOffset	; pt to rightmost dest doubleword
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 1.  In reducing by 2, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 2,
; 8N + 3, etc.  So, we only add 6, not 7.
	mov	rScratch2, srcWidth	; get source width in bits.

	add	rScratch2, 6		; round up to next byte
	mov	rScratch1, srcOffset

; The first dest doubleword could be composed of one or two source
; doublewords.  By the way, we know that it's safe to read doublewords
; because the caller checked to see that all the source and dest lines
; begin on multiples of 8 and have bytes/line that are multiples of 8.
; If the first dest word required one doubleword, do the special code
; below.  Otherwise, skip into the main loop.
	and	rScratch2, 038h		; = (src width in bytes MOD 8) * 8
	cmp	rScratch1, srcLimit	; do these point to the same word?
	je	redT2RightDWord		; yes, then the first dest dword
					; contains data from only one source
					; dword.  Go encode that.

; This is the preamble to the regular loop that encodes a dest dword from
; two source dwords.  It looks like the top of the regular loop, except
; it doesn't write out a previous dest dword.
	movq	mm0, [rSrcAddr]		; get upper left doubleword

	movq	mm1, [rSrcAddrLo]	; get lower left doubleword

	movq	mm2, [rSrcAddr + 8 ]	; get upper right doubleword
	movq	mm5, mm1		; get ready for left test2

	movq	mm7, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	movq	mm4, mm0		; get ready for left test2

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	pand	mm2, mm7		; get rid of pad bits

	pand	mm3, mm7		; get rid of pad bits
	movq	mm6, mm2		; get ready for right test2

	sub	rSrcAddr, 16		; pt to previous doublword pair
	movq	mm7, mm3		; get ready for right test2

	; empty
	jmp	redT2LongLoopEntry	; jump into loop



; The first dest doubleword is composed of only one source doubleword and
; a word of 0's.  Compute this below.  Remember that rSrcAddr and
; rSrcAddrLo already point to the right dwords.
; The detection for at least 2 bits on in each 2x2 square is kind of cute.
; We do two tests and combine the results.  The first tests detects at least
; one bit in each column.  The second test detects at least one bit in each
; row.  One, or both, of these tests must be true for there to be at least
; two bits in the 2x2 square.  The first test is done by ORing the samples
; vertically, then shifting by 1 and ANDing.  The second test is done
; by ANDing vertically, shifting and ORing the results.  We combine the
; the test results by ORing them together (i.e. at least one of the tests
; found at least two bits).
redT2RightDWord:
	movq	mm0, [rSrcAddr]		; get upper doubleword
	; empty

	movq	mm7, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	; empty

	movq	mm1, [rSrcAddrLo]	; get lower doubleword
	pand	mm0, mm7		; get rid of pad bits

	pand	mm1, mm7		; get rid of pad bits
	movq	mm4, mm0		; get ready for test2

	por	mm0, mm1		; test1: OR vertically
	movq	mm5, mm1		; get ready for test2

	movq	mm1, mm0		; test1: get ready to combine columns
	pand	mm4, mm5		; test2: AND vertically

	psrlq	mm1, 1			; test1: shift col test data
	movq	mm5, mm4		; test2: prepare to shift row test data

	pand	mm0, mm1		; test1: AND horizontally
	psrlq	mm5, 1			; test2: shift row test data

	sub	rSrcAddrLo, 16
	por	mm4, mm5		; test2: OR horizontally

	sub	rSrcAddr, 16		; pt to previous doubleword pair
	por	mm0, mm4		; combine tests 1 and 2

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just those bits

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1

	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	movq	mm1, mm0		; get ready to make clumps of 2 bits

	psrlq	mm1, 1

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off

	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2

	por	mm0, mm1

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte

	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4

	por	mm4, mm0
	sub	rDstAddr, 8		; prepare for entrance to main loop


	pand	mm4, [MMWORD PTR lsBitMask8Bits] ; got them.
	pxor	mm2, mm2		; clear "other half word"

; the byte packing and writing are done in the main loop.
	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jl	redT2LongFin		; quit if we're done

; ----------------------------------------------------------------
; MAIN LOOP
; ----------------------------------------------------------------

; On entering this loop, rSrcAddr should be pointing 16 bytes to the
; left of the first doubleword of a doubleword pair,
; rDstAddr should be pointing to first doubleword to be written,
; rSrcFin should be pointing to leftmost doubleword of the source line.
; mm0 should already be loaded with the left doubleword of the pair
; and mm1 should be loaded with the word below that.
; See the comments above on our method for detecting that there are at
; least 2 bits on in every 2x2 input square.
redT2LongLoop:
	movq	mm0, [rSrcAddr]		; get upper left doubleword
	pand	mm2, mm7		; got bytes in right dword.

; Pack left bits into the *lower* half of a word.  Remember that the lower
; half is the left half in this wacky world.
	movq	mm1, [rSrcAddrLo]	; get lower left doubleword
	packuswb mm4, mm2		; pack the bytes

; This clock cycle takes place while we're waiting for mm4 to progress
; through the pipeline.
	movq	mm2, [rSrcAddr + 8 ]	; get upper right doubleword
	movq	mm5, mm1		; get ready for left test2

	movq	[rDstAddr + 8], mm4	; write the previous result
	movq	mm4, mm0		; get ready for left test2

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	movq	mm6, mm2		; get ready for right test2

	sub	rSrcAddr, 16		; pt to previous doublword pair
	movq	mm7, mm3		; get ready for right test2

; We enter the loop here, with the proper preamble, if the first
; dest doubleword contains bits from two source dwords.
redT2LongLoopEntry:
	por	mm0, mm1		; Left Test1: OR rows
	por	mm2, mm3		; Right Test1: OR rows

	pand	mm4, mm5		; Left Test2: AND rows
	pand	mm6, mm7		; Right Test2: AND rows

	movq	mm1, mm0		; Left Test1: prepare to combine cols
	movq	mm3, mm2		; Right Test1: prepare to combine cols

	psrlq	mm1, 1			; Left Test1: shift cols
	movq	mm5, mm4		; Right Test2: prepare to combine cols

	psrlq	mm3, 1			; Right Test1: shift cols
	movq	mm7, mm6		; Right Test2: prepare to combine cols

	psrlq	mm5, 1			; Left Test2: shift cols
	pand	mm0, mm1		; Left Test1: AND horizontally

	psrlq	mm7, 1			; Right Test2: shift cols
	pand	mm2, mm3		; Right Test1, AND horozontally

	por	mm4, mm5		; Left Test2: OR horizontally
	por	mm6, mm7		; Right Test2: OR horizontally

	movq	mm7, [MMWORD PTR lsBitMask1Bit]	; get LS bit mask
	por	mm0, mm4		; Left: Combine Test1 OR Test2

	por	mm2, mm6		; Right: Combine Test1 OR Test2
	pand	mm0, mm7		; Left: get just LS bits of each pair



; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1
	pand	mm2, mm7		; Right: get LS bits of each pair

	movq	mm7, [MMWORD PTR lsBitMask2Bits] ; get next mask
	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	sub	rSrcAddrLo, 16		; point to prev doubleword pair
	punpckldq mm3, mm2		; put low halfword of mm2 into
					; high halfword of mm3

	movq	mm1, mm0		; get ready to make clumps of 2 bits
	punpckhdq mm2, mm3		; complete the word swap

	psrlq	mm1, 1
	movq	mm3, mm2		; get ready to make clumps of 2 bits

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble
	psrlq	mm3, 1

	pand	mm0, mm7		; mask them off
	por	mm2, mm3		; make clumps in lower 2 bits of
					; each nibble

	pand	mm2, mm7		; mask bits in right doubleword
	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2
	movq	mm3, mm2		; get ready to make clumps of 4 bits

	por	mm0, mm1
	psrlq	mm3, 2

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	por	mm2, mm3

	pand	mm2, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4
	movq	mm3, mm2		; get ready to make clumps of bytes

	por	mm4, mm0
	psrlq	mm3, 4

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get last mask.
	por	mm2, mm3

; finish right dword at top of loop
	pand	mm4, mm7		; got bytes in left dword.
	sub	rDstAddr, 8		; get ready for next trip through loop

	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jge	redT2LongLoop		; no, keep going

; We're done with that line.
redT2LongFin:
	pand	mm2, mm7		; got bytes in right dword.

	packuswb mm4, mm2		; pack the bytes

	movq	[rDstAddr + 8], mm4	; write the last word
redT2LongRet:
	ret

ip_reduceT2LineLong	ENDP


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduceT1LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduceT3LineLongStart::
ip_reduceT3LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab2to1Permute:DWORD

; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tab2to1Permute

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword pair in the line.
; This pair may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
; rSrcFin will point to the leftmost doubleword in the source line.
	mov	rSrcFin, srcAddr	; pt to 1st source dword
	mov	rSrcAddr, rSrcFin

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first left/right dword pair.
					; We do this because we're really
					; always going to go after the
					; even doubleword in an even/odd
					; pair.  If the rightmost dest
					; dword gets bits from two source
					; dwords, we use the regular loop.
					; If the rightmost dest dword gets
					; bits from only 1 source dword, we
					; use the code just below.
	mov	rSrcAddrLo, rSrcAddr

	add	rSrcAddrLo, srcBpl	; point to second line
	mov	rDstAddr, dstAddr	; pt to leftmost dest doubleword

	add	rDstAddr, dstOffset	; pt to rightmost dest doubleword
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 1.  In reducing by 2, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 2,
; 8N + 3, etc.  So, we only add 6, not 7.
	mov	rScratch2, srcWidth	; get source width in bits.

	add	rScratch2, 6		; round up to next byte
	mov	rScratch1, srcOffset

; The first dest doubleword could be composed of one or two source
; doublewords.  By the way, we know that it's safe to read doublewords
; because the caller checked to see that all the source and dest lines
; begin on multiples of 8 and have bytes/line that are multiples of 8.
; If the first dest word required one doubleword, do the special code
; below.  Otherwise, skip into the main loop.
	and	rScratch2, 038h		; = (src width in bytes MOD 8) * 8
	cmp	rScratch1, srcLimit	; do these point to the same word?
	je	redT3RightDWord		; yes, then the first dest dword
					; contains data from only one source
					; dword.  Go encode that.

; This is the preamble to the regular loop that encodes a dest dword from
; two source dwords.  It looks like the top of the regular loop, except
; it doesn't write out a previous dest dword.
	movq	mm0, [rSrcAddr]		; get upper left doubleword

	movq	mm1, [rSrcAddrLo]	; get lower left doubleword

	movq	mm2, [rSrcAddr + 8 ]	; get upper right doubleword
	movq	mm5, mm1		; get ready for left test2

	movq	mm7, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	movq	mm4, mm0		; get ready for left test2

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	pand	mm2, mm7		; get rid of pad bits

	pand	mm3, mm7		; get rid of pad bits
	movq	mm6, mm2		; get ready for right test2

	sub	rSrcAddr, 16		; pt to previous doublword pair
	movq	mm7, mm3		; get ready for right test2

	; empty
	jmp	redT3LongLoopEntry	; go to rest of the loop



; The first dest doubleword is composed of only one source doubleword and
; a word of 0's.  Compute this below.  Remember that rSrcAddr and
; rSrcAddrLo already point to the right dwords.
; The detection for at least 2 bits on in each 2x2 square is kind of cute.
; We do two tests and combine the results.  The first tests detects at least
; one bit in each column.  The second test detects at least one bit in each
; row.  One, or both, of these tests must be true for there to be at least
; two bits in the 2x2 square.  The first test is done by ORing the samples
; vertically, then shifting by 1 and ANDing.  The second test is done
; by ANDing vertically, shifting and ORing the results.  We combine the
; the test results by ORing them together (i.e. at least one of the tests
; found at least two bits).
redT3RightDWord:
	movq	mm0, [rSrcAddr]		; get upper doubleword
	; empty

	movq	mm7, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	; empty

	movq	mm1, [rSrcAddrLo]	; get lower doubleword
	pand	mm0, mm7		; get rid of pad bits

	pand	mm1, mm7		; get rid of pad bits
	movq	mm4, mm0		; get ready for test2

	por	mm0, mm1		; test1: OR vertically
	movq	mm5, mm1		; get ready for test2

	movq	mm1, mm0		; test1: get ready to combine columns
	pand	mm4, mm5		; test2: AND vertically

	psrlq	mm1, 1			; test1: shift col test data
	movq	mm5, mm4		; test2: prepare to shift row test data

	pand	mm0, mm1		; test1: AND horizontally
	psrlq	mm5, 1			; test2: shift row test data

	sub	rSrcAddrLo, 16
	por	mm4, mm5		; test2: OR horizontally

	sub	rSrcAddr, 16		; pt to previous doubleword pair
	pand	mm0, mm4		; combine tests 1 and 2

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just those bits

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1

	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	movq	mm1, mm0		; get ready to make clumps of 2 bits

	psrlq	mm1, 1

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off

	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2

	por	mm0, mm1

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte

	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4

	por	mm4, mm0
	sub	rDstAddr, 8		; prepare for entrance to main loop


	pand	mm4, [MMWORD PTR lsBitMask8Bits] ; got them.
	pxor	mm2, mm2		; clear "other half word"

; the byte packing and writing are done in the main loop.
	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jl	redT3LongFin		; quit if we're done

; ----------------------------------------------------------------
; MAIN LOOP
; ----------------------------------------------------------------

; On entering this loop, rSrcAddr should be pointing 16 bytes to the
; left of the first doubleword of a doubleword pair,
; rDstAddr should be pointing to first doubleword to be written,
; rSrcFin should be pointing to leftmost doubleword of the source line.
; mm0 should already be loaded with the left doubleword of the pair
; and mm1 should be loaded with the word below that.
; See the comments above on our method for detecting that there are at
; least 2 bits on in every 2x2 input square.
redT3LongLoop:
	movq	mm0, [rSrcAddr]		; get upper left doubleword
	pand	mm2, mm7		; got bytes in right dword.

; Pack left bits into the *lower* half of a word.  Remember that the lower
; half is the left half in this wacky world.
	movq	mm1, [rSrcAddrLo]	; get lower left doubleword
	packuswb mm4, mm2		; pack the bytes

; This clock cycle takes place while we're waiting for mm4 to progress
; through the pipeline.
	movq	mm2, [rSrcAddr + 8 ]	; get upper right doubleword
	movq	mm5, mm1		; get ready for left test2

	movq	[rDstAddr + 8], mm4	; write the previous result
	movq	mm4, mm0		; get ready for left test2

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	movq	mm6, mm2		; get ready for right test2

	sub	rSrcAddr, 16		; pt to previous doublword pair
	movq	mm7, mm3		; get ready for right test2


; We enter the loop here, with the proper preamble, if the first
; dest doubleword contains bits from two source dwords.
redT3LongLoopEntry:
	por	mm0, mm1		; Left Test1: OR rows
	por	mm2, mm3		; Right Test1: OR rows

	pand	mm4, mm5		; Left Test2: AND rows
	pand	mm6, mm7		; Right Test2: AND rows

	movq	mm1, mm0		; Left Test1: prepare to combine cols
	movq	mm3, mm2		; Right Test1: prepare to combine cols

	psrlq	mm1, 1			; Left Test1: shift cols
	movq	mm5, mm4		; Right Test2: prepare to combine cols

	psrlq	mm3, 1			; Right Test1: shift cols
	movq	mm7, mm6		; Right Test2: prepare to combine cols

	psrlq	mm5, 1			; Left Test2: shift cols
	pand	mm0, mm1		; Left Test1: AND horizontally

	psrlq	mm7, 1			; Right Test2: shift cols
	pand	mm2, mm3		; Right Test1, AND horozontally

	por	mm4, mm5		; Left Test2: OR horizontally
	por	mm6, mm7		; Right Test2: OR horizontally

	movq	mm7, [MMWORD PTR lsBitMask1Bit]	; get LS bit mask
	pand	mm0, mm4		; Left: Combine Test1 AND Test2

	pand	mm2, mm6		; Right: Combine Test1 AND Test2
	pand	mm0, mm7		; Left: get just LS bits of each pair



; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1
	pand	mm2, mm7		; Right: get LS bits of each pair

	movq	mm7, [MMWORD PTR lsBitMask2Bits] ; get next mask
	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	sub	rSrcAddrLo, 16		; point to prev doubleword pair
	punpckldq mm3, mm2		; put low halfword of mm2 into
					; high halfword of mm3

	movq	mm1, mm0		; get ready to make clumps of 2 bits
	punpckhdq mm2, mm3		; complete the word swap

	psrlq	mm1, 1
	movq	mm3, mm2		; get ready to make clumps of 2 bits

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble
	psrlq	mm3, 1

	pand	mm0, mm7		; mask them off
	por	mm2, mm3		; make clumps in lower 2 bits of
					; each nibble

	pand	mm2, mm7		; mask bits in right doubleword
	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2
	movq	mm3, mm2		; get ready to make clumps of 4 bits

	por	mm0, mm1
	psrlq	mm3, 2

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	por	mm2, mm3

	pand	mm2, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4
	movq	mm3, mm2		; get ready to make clumps of bytes

	por	mm4, mm0
	psrlq	mm3, 4

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get last mask.
	por	mm2, mm3

; finish right dword at top of loop
	pand	mm4, mm7		; got bytes in left dword.
	sub	rDstAddr, 8		; get ready for next trip through loop

	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jge	redT3LongLoop		; no, keep going

; We're done with that line.
redT3LongFin:
	pand	mm2, mm7		; got bytes in right dword.

	packuswb mm4, mm2		; pack the bytes

	movq	[rDstAddr + 8], mm4	; write the last word
redT3LongRet:
	ret

ip_reduceT3LineLong	ENDP

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduceT4LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduceT4LineLongStart::
ip_reduceT4LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab2to1Permute:DWORD


; Make the silly assembly happy
	mov	eax, pIPState
	mov	eax, tab2to1Permute

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword pair in the line.
; This pair may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
; rSrcFin will point to the leftmost doubleword in the source line.
	mov	rSrcFin, srcAddr	; pt to 1st source dword
	mov	rSrcAddr, rSrcFin

	add	rSrcAddr, srcOffset	; pt to left source dword of
					; first left/right dword pair.
					; We do this because we're really
					; always going to go after the
					; even doubleword in an even/odd
					; pair.  If the rightmost dest
					; dword gets bits from two source
					; dwords, we use the regular loop.
					; If the rightmost dest dword gets
					; bits from only 1 source dword, we
					; use the code just below.
	mov	rSrcAddrLo, rSrcAddr

	add	rSrcAddrLo, srcBpl	; point to second line
	mov	rDstAddr, dstAddr	; pt to leftmost dest doubleword

	add	rDstAddr, dstOffset	; pt to rightmost dest doubleword
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 1.  In reducing by 2, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 2,
; 8N + 3, etc.  So, we only add 6, not 7.
	mov	rScratch2, srcWidth	; get source width in bits.

	add	rScratch2, 6		; round up to next byte
	mov	rScratch1, srcOffset

; The first dest doubleword could be composed of one or two source
; doublewords.  By the way, we know that it's safe to read doublewords
; because the caller checked to see that all the source and dest lines
; begin on multiples of 8 and have bytes/line that are multiples of 8.
; If the first dest word required one doubleword, do the special code
; below.  Otherwise, skip into the main loop.
	and	rScratch2, 038h		; = (src width in bytes MOD 8) * 8
	cmp	rScratch1, srcLimit	; do these point to the same word?
	je	redT4RightDWord		; yes, then the first dest dword
					; contains data from only one source
					; dword.  Go encode that.

; This is the preamble to the regular loop that encodes a dest dword from
; two source dwords.  It looks like the top of the regular loop, except
; it doesn't write out a previous dest dword..
	movq	mm0, [rSrcAddr]		; get upper left doubleword

	movq	mm1, [rSrcAddrLo]	; get lower left doubleword

	movq	mm2, [rSrcAddr+ 8 ]	 ; get upper right doubleword
	pand	mm0, mm1		; combine left rows

	movq	mm1, mm0		; get ready to combine left columns
	; empty

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	psrlq	mm1, 1			; move to lower end of each bit pair

	sub	rSrcAddr, 16		; pt to previous doubleword pair
	pand	mm2, mm3		; combine right rows

	pand	mm2, [MMWORD PTR mmxByteEdgeMask + rScratch2]
	jmp	redT4LongLoopEntry	; go to rest of the loop



; The first dest doubleword is composed of only one source doubleword and
; a word of 0's.  Compute this below.  Remember that rSrcAddr and
; rSrcAddrLo already point to the right dwords.
redT4RightDWord:
	movq	mm0, [rSrcAddr]		; get upper doubleword
	sub	rSrcAddr, 16		; pt to previous doubleword pair

	movq	mm1, [rSrcAddrLo]	; get lower doubleword
	sub	rSrcAddrLo, 16

	pand	mm0, mm1		; combine rows

; get rid of pad bits
	pand	mm0, [MMWORD PTR mmxByteEdgeMask + rScratch2]

	movq	mm1, mm0		; get ready to combine columns

	psrlq	mm1, 1			; move to lower end of each bit pair

	pand	mm0, mm1		; get result in lower bits

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just those bits

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1

	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	movq	mm1, mm0		; get ready to make clumps of 2 bits

	psrlq	mm1, 1

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off

	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2

	por	mm0, mm1

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte

	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4

	por	mm4, mm0
	sub	rDstAddr, 8		; prepare for next loop

	pand	mm4, [MMWORD PTR lsBitMask8Bits]	; got them.
	pxor	mm2, mm2		; clear "other half word"

; the byte packing and writing are done in the main loop.
	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jl	redT4LongFin		; quit if we're done


; ----------------------------------------------------------------
; MAIN LOOP
; ----------------------------------------------------------------

; On entering this loop, rSrcAddr should be pointing 16 bytes to the
; left of the first doubleword of a doubleword pair,
; rDstAddr should be pointing to first doubleword to be written,
; rSrcFin should be pointing to leftmost doubleword of the source line.
; mm0 should already be loaded with the left doubleword of the pair
; and mm1 should be loaded with the word below that.
redT4LongLoop:
	movq	mm0, [rSrcAddr]		; get upper left doubleword
	pand	mm2, mm7		; got bytes in right dword.

; Pack left bits into the *lower* half of a word.  Remember that the lower
; half is the left half in this wacky world.
	movq	mm1, [rSrcAddrLo]	; get lower left doubleword
	packuswb mm4, mm2		; pack the bytes

; This clock cycle takes place while we're waiting for mm4 to progress
; through the pipeline.
	movq	mm2, [rSrcAddr+ 8 ]	 ; get upper right doubleword
	pand	mm0, mm1		; combine left rows

	movq	[rDstAddr + 8], mm4	; write the previous result
	movq	mm1, mm0		; get ready to combine left columns

	movq	mm3, [rSrcAddrLo + 8]	; get lower right doubleword
	psrlq	mm1, 1			; move to lower end of each bit pair

	sub	rSrcAddr, 16		; pt to previous doubleword pair
	pand	mm2, mm3		; combine right rows

; We enter the loop here, with the proper preamble, if the first
; dest doubleword contains bits from two source dwords.
redT4LongLoopEntry:
	pand	mm0, mm1		; get left result in lower bits
	movq	mm3, mm2		; get ready to combine right columns

	pand	mm0, [MMWORD PTR lsBitMask1Bit]	; select just left bits
	psrlq	mm3, 1			; move to lower end of each bit pair

; word swap this value so that the reduced halfwords will end up in the
; correct order.  Remember that all double words have the leftmost word
; in the *lower* half.
	punpckldq mm1, mm0		; put low halfword of mm0 into
					; high halfword of mm1
	pand	mm2, mm3		; get right result in lower bits

	pand	mm2, [MMWORD PTR lsBitMask1Bit]	; select just right bits
	punpckhdq mm0, mm1		; put high halfword of mm0 into
					; low halfword of mm0 and put
					; low halfword of mm0 (now in high
					; halfword of mm1) into high
					; halfword of mm0.  Voila!

	sub	rSrcAddrLo, 16		; point to prev doubleword pair
	punpckldq mm3, mm2		; put low halfword of mm2 into
					; high halfword of mm3

	movq	mm1, mm0		; get ready to make clumps of 2 bits
	punpckhdq mm2, mm3		; complete the word swap

	psrlq	mm1, 1
	movq	mm3, mm2		; get ready to make clumps of 2 bits

	por	mm0, mm1		; make clumps in lower 2 bits of
					; each nibble
	psrlq	mm3, 1

	pand	mm0, [MMWORD PTR lsBitMask2Bits] ; mask them off
	por	mm2, mm3		; make clumps in lower 2 bits of
					; each nibble

	pand	mm2, [MMWORD PTR lsBitMask2Bits] ; mask bits in right dword
	movq	mm1, mm0		; get ready to make clumps of 4 bits

	psrlq	mm1, 2
	movq	mm3, mm2		; get ready to make clumps of 4 bits

	por	mm0, mm1
	psrlq	mm3, 2

	pand	mm0, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	por	mm2, mm3

	pand	mm2, [MMWORD PTR lsBitMask4Bits] ; got lower nibble of each byte
	movq	mm4, mm0		; get ready to make clumps of bytes

	psrlq	mm4, 4
	movq	mm3, mm2		; get ready to make clumps of bytes

	por	mm4, mm0
	psrlq	mm3, 4

	movq	mm7, [MMWORD PTR lsBitMask8Bits] ; get final mask.
	por	mm2, mm3

	pand	mm4, mm7		; got bytes in left dword.
					; finish right dword at top of loop
	sub	rDstAddr, 8		; get ready for next trip through loop

	cmp	rSrcAddr, rSrcFin	; are we at the end of the row?
	jge	redT4LongLoop		; no, keep going

; We're done with that line.
redT4LongFin:
	pand	mm2, mm7		; got bytes in right dword.

	packuswb mm4, mm2		; pack the bytes

	movq	[rDstAddr + 8], mm4	; write the last word
redT4LongRet:
	ret

ip_reduceT4LineLong	ENDP



; Segment for initialized variables, like masks
_DATA	SEGMENT PUBLIC 'DATA'

		ALIGN	8
lsBitMask1Bit	QWORD	05555555555555555h
msBitMask1Bit	QWORD	0AAAAAAAAAAAAAAAAh
lsBitMask2Bits	QWORD	03333333333333333h
msBitMask2Bits	QWORD	0CCCCCCCCCCCCCCCCh
lsBitMask4Bits	QWORD	00F0F0F0F0F0F0F0Fh
msBitMask4Bits	QWORD	0F0F0F0F0F0F0F0F0h
lsBitMask8Bits	QWORD	000FF00FF00FF00FFh
msBitMask8Bits	QWORD	0FF00FF00FF00FF00h

_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no initialized variables
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcredm.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.2  1998/01/13 23:58:31  ddavies
pcredm.asm -> /product/ipcore/ipshared/src/RCS/pcredm.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.1  1996/10/24 19:16:51  ddavies
Initial revision



^

ENDIF


