;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcredmn.asm:  4x, 8x threshold reduction using MMX opcodes.

	.586		; compile code for a Pentium processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

; This code uses MMX opcodes, which have 64-bit accesses.  That's the
; main source of our speedup.  These routines are only called if
; the caller detects MMX code.  We don't have to detect it here.
; First, include the file that gives us the MMX macros.

INCLUDE	iammx.inc

_DATA	SEGMENT PUBLIC 'DATA'
EXTERN	mmxByteEdgeMask:QWORD
RCSInfo	BYTE	"$RCSfile: pcredmn.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; This code looks like the 2x reduction code in pcredm.asm.  We sample
; four doublewords in a column, producing 16 useful bits from the 256
; original bits (each 4 bit x 4 bit square in the original turns into
; 1 output bit).
;
; This is the fourth rewrite of this code.  I'm learning a lot, I hope.
; This time we'll use the MMX pack opcode, packuswb.  It converts two
; dwords, each containg bits in every other byte (the even bytes when
; numbering from the LS end of the word), into one dword with every
; byte filled.  The even bytes from one dword end up in the upper half
; of the dest dword and the even bytes from the other other end up in the
; lower half.  What a deal.
;
; Since we're doing 4x reduction, the basic idea is to fairly quickly
; get 2 bits per byte (8/4 = 2), then get 4 bits in every other byte,
; then pack two such dwords.  This gives us a dword with 4 bits in every
; byte.  We repeat this process for two more dwords, producing another
; dword with 4 bits in every byte.  We then shift/OR each of these dwords
; (it's nicer to do two things in parallel), producing two words with
; 8 bits in every other byte.  These two words are packed to produce
; the final dest word.
;
; To compute a threshold 1 reduction, we OR together all the bits in
; each 4x4 square to produce one bit.  To produce a threshold 4 reduction,
; we AND them.  The ORing and ANDing vertically is easy, we just combine
; dwords from adjacent lines.  ORing and ANDing horizontally is messy.
; The standard way to do this is to use shifts and ORs.  First,
; you combine each pixel with its neighbor on the left.  This
; produces results in the even bit positions. Then you combine
; each of these pixels with the pixel that 2 to the left.
; That produces result pixels at the locations that are 0 MOD 4.  Then
; we have to clump the two result bits together, so there's another
; shift/OR cycle.  Each shift/combine cycle takes 3 MMX opcodes.  We
; also have to mask the results at one stage, so we need (3 * 3) + 1 = 10
; opcodes to get to the two-bits-clumped-at-the-end-of-each-byte stage.
; We need (4 * 3) + 2 mask steps = 14 cycles to get to the 4-bits-clumped-
; at-the-end-of-every-other-byte stage.
;
; Well, maybe not.  I thought about a cute scheme that uses the the
; packed compare feature to set all the bits of a byte if any of the bits
; are one.  Unfortunately, you have to separately test each nibble, then
; combine the results and put them in the right place.  A byte shift (one
; that breaks the shift chain at byte boundaries) would have been nice for
; this, but there's no such thing in MMX-land.  After a lot of fooling
; around, I managed to get something that was the same speed as the standard
; method described above.  Defeat.
;
; Well, except maybe not.  The latest scheme uses byte adds, which do exist.
; First we make a copy of the input register and AND the copy with 0777...7h.
; This turns off the high order bit of each nibble.  We then add 0777...7h to
; this.  If there are any 1's in the lower three bits of the nibble, the
; MS bit of the nibble is now turned on.  We OR this with the original
; value.  If there were any 1's in a nibble, that nibble's MS bit is now
; a 1.  In 4 opcodes, we've done the work of 6 above.
;
; That worked for threshold 1.  The steps for threshold 4 are almost the
; same.  We do the same copy and the same AND with 07777...7h.  We then
; add 0111...1h.  If there were any 0's in the lower three bits of a nibble,
; the MS bit of that nibble is 0, otherwise it's a 1.  We AND this with
; the original.  If there were any 0's in a nibble, it's MS bit is now 0,
; which is what we want.  The only hiccup here is that we don't have enough
; MMX registers to hold both masks, so we have to use a memory cycle to
; do the add of 01111....1h.  Too bad.
;
; But, we're not done.  We AND the result obtained above with 08888...8h.
; This gets rid of everything except the bits we really want.  We then
; add 038383838...38h to the word.  This "carries" (nerd pun) the lower
; nibble result in bit 3 up to bit 6, right next to the upper nibble result.
; With these two opcodes, we've clumped the two bits for a byte together.
;
; Unfortunately, you can't use the carry trick to move two bits, so we need
; to use the move-shift-or sequence to move the clumps of two bits together
; in every other byte.  Since the add of 0383838...38h left dreck behind,
; we need to make with 0CCCC...CCh before doing that.
;
; When we're done with this step, we have four bits at the top of the odd
; numbered bytes and dreck at the top of the even numbered bytes.  We do
; a halfword shift (shift chain broken at 16-bit boundaries) right by 12
; to put these in the lower bytes where the pack opcode will expect to find
; them.
;
; The total of all this sillyness is 11 opcodes, compared to 14 for the
; standard way.  Since all of this happens 4 times for each dest dword,
; we save 4 * (14-11) = 12 cycles for each dword, which is significant.
;
;
;
; Wordswapping raises it ugly head here.  The x86 processors are little
; endian machines.  We have a convention that images are accessed in 32-bit
; chunks, so left-to-right within the image corresponds to MS-to-LS within
; a 32-bit word.  Unfortunately, the 32-bit word that's on the left in
; an image is at the LS (right) end of a 64-bit dword.  Grump.
;
; There are two places we have to worry about this, when reading source
; dwords and when writing dest dwords.   Wordswapping the dest dwords is
; easy, we just flip the order of the arguments to the final pack opcode.
; It takes 2 opcodes (unpacklo, unpackhi) to wordswap a dword register
; containing source args.  It takes 4 opcodes (movq, shiftL, shiftR, or)
; to swap bytes within halfwords or halfwords within words.  So swapping
; the original source dwords is a good idea, right?  Wrong.  We'd need
; 2 * 4 = 8 opcodes to swap all four source dwords and only 4 * 1 = 4
; opcodes to byteswap the final dword, so we'll swap the final one.
;
; As with almost any MMX task, it helps if you try to do at least two things
; at once.  This makes it easier to fill the empty slots in the instruction
; pairs.  The basic loop structure will be:
;
; 1. Sample bits from dword column 3.
; 2. Sample bits from dword column 2.
; 3. Starting in step 1, clump together the bits from columns 0 and 1.
;    These were read in the previous loop iteration.
; 4. Pack the clumped bits from columns 0 and 1 with the previously clumped
;    bits from columns 2 and 3 (from the previous trip through the loop).
;    Byteswap and write out the result dword.
; 5. Check to see if we're done with the loop. If so, jump to the end.
; 6. Sample the bits from dword column 1.
; 7. Sample the bits from dword column 0.
; 8. Starting in step 6, clump the bits from columns 2 and 3 together.
; 9. Decrement the source addresses.
; 10. If there are more source columns to read, loop back to step 1.
; 11. If we're at the left end of the scan line (alternative  to step 9),
;     do the consolidation steps that would be done in parallel with steps 1
;     and 2 and jump back to step 3.  The check in step 5 will exit if
;     we're done.
;
; On entering the routine, we test to see if the rightmost dword requires
; 1, 2, 3 or 4 source dword columns.  In each case, we read the column
; required and mask the dword to make sure that no pad bits creep into
; the reduced image.  If we're reading columns 3 or 1, we continue on
; to read columns 0 and 2 respectively before leaping into the loop.
; This is because the code in the loop for reading columns 3 and 2 is
; a unit, as is the code for reading columns 1 and 0.
;
;
;
;Register Assignment:
; EAX	scratch / dest address
; EBX	source address, src line 1
; ECX	source address, src line 2
; EDX	source address, src line 3
; EBP	pts to variables on the stack
; ESI	source address, src line 0
; EDI	scratch
; ESP	stack pointer

; There are four columns of doublewords in each 4x4 doubleword square.
; We use mm0 to accumulate the results for column 0, mm1 for column 1
; and so on.
; MM0	used to accumulate data from doubleword col 0
; MM1	used to accumulate data from doubleword col 1
; MM2	used to accumulate data from doubleword col 2
; MM3	used to accumulate data from doubleword col 3
; MM4	used to consolidate data from even doubleword cols
; MM5	used to consolidate data from odd doubleword cols
; MM6   holds all bits seen so far in loop
; MM7	holds 07777777777777777h, used in horizontal reduction

rScratch	TEXTEQU	<eax>
rSrcAddr1	TEXTEQU	<ebx>
rSrcAddr2	TEXTEQU	<ecx>
rSrcAddr3	TEXTEQU	<edx>
rSrcAddr0	TEXTEQU	<esi>
rScratch1	TEXTEQU	<edi>

; Parameters
; srcAddr	addr of left end of source line
; srcWidth	width of source image in bits
; srcBpl	source bytes/line
; dstAddr	addr of left end of dest line
; dstOffset	offset in bytes to rightmost dest word
; srcOffset	offset in bytes to leftmost byte in rightmost
;		group of source bits that will form the rightmost
;		dest word
; srcLimit	offset in bytes to rightmost source byte (often
;		to the right of byte at srcOffset)
; tab4to1Permute addr of table used to unscramble reduced bits.
;		 Unused in this code.

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduce4T1LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduce4T1LineLongStart::
ip_reduce4T1LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab4to1Permute:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; Make the assembler happy
	mov	eax, pIPState
	mov	eax, tab4to1Permute

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword quad in the line.
; This quad may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	rSrcAddr0, srcAddr	; pt to 1st source dword
	mov	rScratch1, srcBpl	; fill empty slot

	add	rSrcAddr0, srcOffset	; pt to left source dword of
					; first dword quad.
					; We do this because we're really
					; always going to go after the
					; first doubleword in a quad.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first quad.
	; empty

	lea	rSrcAddr1, [rSrcAddr0 + rScratch1] ; point to src line 1
	lea	rSrcAddr2, [rSrcAddr0 + 2*rScratch1] ; point to src line 2

	mov	rSrcAddr3, rSrcAddr2	; prepare to point to line 3
	mov	rScratch, dstOffset	; get offset to rightmost dest dword

; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword quad)
; can be 0, 8, 16 or 24.  If it's 0, we enter the loop where it reads
; one source dword before writing the dest.  If it's 8, we enter the loop
; where it reads 2 source dwords and so on.
	add	dstAddr, rScratch	; pt to rightmost dest doubleword
	mov	rScratch, srcLimit	; get srcOffset

	add	rSrcAddr3, rScratch1	; pt to line 3
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 3.  In reducing by 4, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 4,
; 8N + 5, etc.  So, we only add 4, not 7.
	mov	rScratch1, srcWidth	; get source width in bits.

	add	rScratch1, 4		; round up to next byte
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	sub	rScratch, srcOffset	; get difference
	pxor	mm0, mm0		; clear registers

; Initialize all the MMX registers in a wad.  Since none of these
; operations affect the condition codes, the codes from the "sub"
; above will be the ones the "je" below looks at.
	pxor	mm1, mm1
	pxor	mm2, mm2

	movq	mm7, [MMWORD PTR all7sMask]
	pxor	mm3, mm3

; We mask each source nibble with 7, then add 7 to set MS bit if there
; are any 1's in the lower three bits.
	pxor	mm4, mm4
	pxor	mm5, mm5		; mm5 gets used at red4T1LoopW1

	pxor	mm6, mm6 
	je	red4T1LoopW0Entry	; if there's only one dword in
					; the rightmost quad, process it.

	cmp	rScratch, 8		; if there are two dwords in the
	je	red4T1LoopW1Entry	; rightmost quad, process them

	cmp	rScratch, 16		; if there are three dwords in the
	je	red4T1LoopW2Entry	; rightmost quad, process them



; There are four words in the loop.  Read the cols 3 and 2,
; then branch into the loop to read cols 1 and 0.
red4T1LoopW3Entry:
	movq	mm3, [rSrcAddr0 + 24]	; get top word 3

	por	mm3, [rSrcAddr1 + 24]	; and with next word 3

	por	mm3, [rSrcAddr2 + 24]	; and next word 3

	por	mm3, [rSrcAddr3 + 24]	; and final word 3

	pand	mm3, [MMWORD PTR mmxByteEdgeMask + rScratch1]

; Read dword 2 and DO NOT mask it
	movq	mm2, [rSrcAddr0 + 16]	; get top word 2

	por	mm2, [rSrcAddr1 + 16]	; and with next word 2

	por	mm2, [rSrcAddr2 + 16]	; and next word 2
	movq	mm5, mm3		; get ready for step 1, col 3

	por	mm2, [rSrcAddr3 + 16]	; and final word 2

	jmp	red4T1LoopW1		; do next two dword cols

; Here's where we start when dword 2 is the rightmost word in the scanline.
red4T1LoopW2Entry:
	movq	mm2, [rSrcAddr0 + 16]	; get top word 2
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	por	mm2, [rSrcAddr1 + 16]	; and with next word 2

	por	mm2, [rSrcAddr2 + 16]	; and next word 2

	por	mm2, [rSrcAddr3 + 16]	; and final word 2

	pand	mm2, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	jmp	red4T1LoopW1		; do next two dword cols

; Here's where we start when dword 1 is the rightmost word in the scanline.
red4T1LoopW1Entry:
	movq	mm1, [rSrcAddr0 + 8]	; get top word 1
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	por	mm1, [rSrcAddr1 + 8]	; and with next word 1

	por	mm1, [rSrcAddr2 + 8]	; and next word 1

	por	mm1, [rSrcAddr3 + 8]	; and final word 1

	pand	mm1, [MMWORD PTR mmxByteEdgeMask + rScratch1]

; Read dword 0 and DO NOT mask it
	movq	mm0, [rSrcAddr0]	; get top word 0

	por	mm0, [rSrcAddr1]	; and with next word 0

	por	mm0, [rSrcAddr2]	; and next word 0
	movq	mm5, mm1		; get ready for step 1, col 1

	por	mm0, [rSrcAddr3]	; and final word 0

	jmp	red4T1LoopBottom	; adjust addresses and
					; do next two dword cols

red4T1LoopW0Entry:
	movq	mm0, [rSrcAddr0]	; get top word 0
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	por	mm0, [rSrcAddr1]	; and with next word 0

	por	mm0, [rSrcAddr2]	; and next word 0
	movq	mm5, mm1		; get ready for step 1, col 1

	por	mm0, [rSrcAddr3]	; and final word 0

	pand	mm0, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	jmp	red4T1LoopBottom	; adjust addresses and
					; do next two dword cols


; --------------------- MAIN LOOP -----------------------------

; --------------------- Read Word 3 ---------------------------

; This is the start of the main loop.  Read and accumulate results from
; dword column 3.  Start doing horizontal reduction of data in word
; columns 0 and 1.
; Note that we only get here from the bottom of the loop, so we can
; depend on mm4 already holding the contents of mm0.
red4T1LoopW3:
	movq	mm3, [rSrcAddr0 + 24]	; get top word 3
	movq	mm4, mm0		; make copy of col 0 bits

	por	mm3, [rSrcAddr1 + 24]	; get word 3, line 1
	pand	mm4, mm7		; get rid of MS bit in each nibble

	por	mm3, [rSrcAddr2 + 24]	; get word 3, line 2
	paddb	mm4, mm7		; set MS bit if any 1's in LS 3 bits

	por	mm3, [rSrcAddr3 + 24]	; get word 3, line 3
	por	mm0, mm4		; set MS bit if any 1's in nibble

; We need to use more masks and don't have the registers to hold them
; at this point, so we just do it now.  We'll save the reading of mm2
; for later when we need to let the other consolidation channel (mm1's)
; catch up.
	pand	mm0, [MMWORD PTR all8sMask] ; select just nib result bits
	pand	mm5, mm7		; MS bit of col 1 nibbles

	paddb	mm0, [MMWORD PTR all38sMask] ; carry LS nib result bit up to
					; bit 6 of each byte.
	paddb	mm5, mm7		; set nib result bits if low bits set

	pand	mm0, [MMWORD PTR allC0Mask] ; get paired result bits for col 0
	por	mm1, mm5		; set MS bit if any 1's in col 1 nibble

; Now do the stuff with col 1 that requires the extra masks
	pand	mm1, [MMWORD PTR all8sMask] ; select just col 1 nib result bits
	movq	mm4, mm0		; get ready to copy col 0 bit pairs

	paddb	mm1, [MMWORD PTR all38sMask] ; put col 1 results in bits 7,6
	psllq	mm4, 6			; put clumps in top end of upper bytes

	pand	mm1, [MMWORD PTR allC0Mask] ; get paired result bits for col 1
	por	mm0, mm4		; col 0 quad results in odd bytes

; Now we finish reading the col 2 inputs while col 1 consolidation catches up
	movq	mm2, [rSrcAddr0 + 16]	; get top word 2
	movq	mm5, mm1		; get ready to copy col 1 bit pairs

	por	mm2, [rSrcAddr1 + 16]	; get word 2, line 1
	psllq	mm5, 6			; put clumps in top end of upper bytes

	por	mm2, [rSrcAddr2 + 16]	; get word 2, line 2
	por	mm1, mm5		; col 1 quad results in odd bytes

	por	mm2, [rSrcAddr3 + 16]	; get word 2, line 3
	psrlw	mm0, 8			; col 0 res in bits 7:4 of even bytes

; We jump in here when consolidating the very last source word read.
red4T1FinishConsolidating:
	psrlw	mm1, 8			; col 1 res in bits 7:4 of even bytes
	mov	rScratch, dstAddr	; read dest address

	packuswb mm1, mm0		; produce col0:1 bits in bits 7:4 of
					; all bytes
	movq	mm4, mm6		; prepare to compact col2:col3 bits

	movq	mm5, mm1		; prepare to compace col0:col1 bits
	psllq	mm4, 4			; put 8 col2:col3 bits in the odd bytes

	psllq	mm5, 4			; put 8 col0:col1 bits in the odd bytes
	por	mm1, mm5		; combine shifted col0:1 bits

	psrlw	mm1, 8			; 8 col0:1 bits in even bytes
	por	mm6, mm4		; combine shifted col2:3 bits

	psrlw	mm6, 8			; 8 col2:3 bits in even bytes
	sub	rScratch, 8		; decrement dest address

; Here's where we wordswap the final dest word by flipping the order
; of the args to the pack command.  The dest appears in the LS word and
; the source appears in the MS word.  Therefore, we get 2:3:0:1, which
; is what we want.
	packuswb mm1, mm6
	mov	dstAddr, rScratch	; update stored version

; Now we need to byteswap the bytes in here.  Remember that each source word
; produced 16 bits, 8 from the upper word and 8 from the lower word.  We
; now need to swap the bytes, which corresponds to swapping the words in
; the original.
	movq	mm4, mm1		; make a copy
	psllw	mm1, 8			; shift low bytes to high bytes

	psrlw	mm4, 8			; shift high bytes to low bytes
	; empty

	por	mm1, mm4		; combine swapped bytes
	; empty

; wait for mm0 to percolate through pipeline
	; empty
	; empty

	movq	[rScratch + 8], mm1	; store dest word
	movq	mm5, mm3		; get ready for step 1, col 3

	cmp	rSrcAddr0, srcAddr	; back at beginning of line yet?
	jl	red4T1Exit		; quit if that was last dest word

; ------------------- Get Bits from Words 1 & 0 ----------------------

; We can enter the loop at this point, so the source addresses must
; be consistent.
red4T1LoopW1:
	movq	mm1, [rSrcAddr0 + 8]	; get top word 1
	movq	mm4, mm2		; make copy of col 2 bits

	por	mm1, [rSrcAddr1 + 8]	; get word 1, line 1
	pand	mm4, mm7		; get rid of MS bit in each nibble

	por	mm1, [rSrcAddr2 + 8]	; get word 1, line 2
	paddb	mm4, mm7		; set MS bit if any 1's in LS 3 bits

	por	mm1, [rSrcAddr3 + 8]	; get word 1, line 3
	por	mm2, mm4		; set MS bit if any 1's in nibble

; We need to use more masks and don't have the registers to hold them
; at this point, so we just do it now.  We'll save the reading of mm0
; for later when we need to let the other consolidation channel (mm3's)
; catch up.
	pand	mm2, [MMWORD PTR all8sMask] ; select just nib result bits
	pand	mm5, mm7		; MS bit of col 3 nibbles

	paddb	mm2, [MMWORD PTR all38sMask] ; carry LS nib result bit up to
					; bit 6 of each byte.
	paddb	mm5, mm7		; set nib result bits if low bits set

	pand	mm2, [MMWORD PTR allC0Mask] ; get paired result bits for col 2
	por	mm3, mm5		; set MS bit if any 1's in col 3 nibble

; Now do the stuff with col 3 that requires the extra masks
	pand	mm3, [MMWORD PTR all8sMask] ; select just col 3 nib result bits
	movq	mm4, mm2		; get ready to copy col 2 bit pairs

	paddb	mm3, [MMWORD PTR all38sMask] ; put col 3 results in bits 7,6
	psllq	mm4, 6			; put clumps in top end of upper bytes

	pand	mm3, [MMWORD PTR allC0Mask] ; get paired result bits for col 3
	por	mm2, mm4		; col 2 quad results in odd bytes

; Now we finish reading the col 0 inputs while col 3 consolidation catches up
	movq	mm0, [rSrcAddr0]	; get top word 0
	movq	mm6, mm3		; get ready to copy col 3 bit pairs

	por	mm0, [rSrcAddr1]	; get word 0, line 1
	psllq	mm6, 6			; put clumps in top end of upper bytes

	por	mm0, [rSrcAddr2]	; get word 0, line 2
	por	mm6, mm3		; col 3 quad results in odd bytes

	por	mm0, [rSrcAddr3]	; get word 0, line 3
	psrlw	mm2, 8			; col 2 res in bits 7:4 of even bytes

	psrlw	mm6, 8			; col 3 res in bits 7:4 of even bytes
	; empty

	packuswb mm6, mm2		; produce col2:3 bits in bits 7:4 of
					; all bytes

; We jump into the loop here at the right edge of the scan line.
red4T1LoopBottom:
	sub	rSrcAddr1, 32		; update rSrcAddr1 by 4*length of dword

	sub	rSrcAddr0, 32		; update rSrcAddr0
	movq	mm5, mm1		; get ready for step 1, col 1

	sub	rSrcAddr2, 32		; update rSrcAddr2
	sub	rSrcAddr3, 32		; and rSrcAddr3

	cmp	rSrcAddr0, srcAddr	; done all src words yet?
	jge	red4T1LoopW3		; nope, do another quad

; ----------------------------- LOOP EXIT -------------------------------

; At this point, we've just read the leftmost word of the last quad.
; We need to consolidate it, merge it with the other results, unscramble
; them and write out the result.
; We could have copied all the word 3, word 2 and writing code down here, but
; that's overboard even for me.  Instead, we just do the consolidation
; part of the word 3 and word 2 code (no reads allowed) and jump back
; to the code just after we're done reading.  After it finishes writing
; the dest word, there's a test to see if rSrcAddr0 is less than srcAddr.
; If so, it exits.
	movq	mm4, mm0		; make copy of col 0 bits

	pand	mm4, mm7		; get rid of MS bit in each nibble

	paddb	mm4, mm7		; set MS bit if any 1's in LS 3 bits

	por	mm0, mm4		; set MS bit if any 1's in nibble

; We need to use more masks and don't have the registers to hold them
; at this point, so we just do it now.  We'll save the reading of mm2
; for later when we need to let the other consolidation channel (mm1's)
; catch up.
	pand	mm0, [MMWORD PTR all8sMask] ; select just nib result bits
	pand	mm5, mm7		; MS bit of col 1 nibbles

	paddb	mm0, [MMWORD PTR all38sMask] ; carry LS nib result bit up to
					; bit 6 of each byte.
	paddb	mm5, mm7		; set nib result bits if low bits set

	pand	mm0, [MMWORD PTR allC0Mask] ; get paired result bits for col 0
	por	mm1, mm5		; set MS bit if any 1's in col 1 nibble

; Now do the stuff with col 1 that requires the extra masks
	pand	mm1, [MMWORD PTR all8sMask] ; select just col 1 nib result bits
	movq	mm4, mm0		; get ready to copy col 0 bit pairs

	paddb	mm1, [MMWORD PTR all38sMask] ; put col 1 results in bits 7,6
	psllq	mm4, 6			; put clumps in top end of upper bytes

	pand	mm1, [MMWORD PTR allC0Mask] ; get paired result bits for col 1
	por	mm0, mm4		; col 0 quad results in odd bytes

; Now we finish reading the col 2 inputs while col 1 consolidation catches up
	movq	mm5, mm1		; get ready to copy col 1 bit pairs

	psllq	mm5, 6			; put clumps in top end of upper bytes

	por	mm1, mm5		; col 1 quad results in odd bytes
	psrlw	mm0, 8			; col 0 res in bits 7:4 of even bytes

	; empty
	jmp	red4T1FinishConsolidating


; We get here after the last (leftmost) dword was written
red4T1Exit:
	ret

ip_reduce4T1LineLong ENDP



; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduce4T4LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduce4T4LineLongStart::
ip_reduce4T4LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab4to1Permute:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; Make the assembler happy
	mov	eax, pIPState
	mov	eax, tab4to1Permute

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword quad in the line.
; This quad may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	rSrcAddr0, srcAddr	; pt to 1st source dword
	mov	rScratch1, srcBpl	; fill empty slot

	add	rSrcAddr0, srcOffset	; pt to left source dword of
					; first dword quad.
					; We do this because we're really
					; always going to go after the
					; first doubleword in a quad.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first quad.
	; empty

	lea	rSrcAddr1, [rSrcAddr0 + rScratch1] ; point to src line 1
	lea	rSrcAddr2, [rSrcAddr0 + 2*rScratch1] ; point to src line 2

	mov	rSrcAddr3, rSrcAddr2	; prepare to point to line 3
	mov	rScratch, dstOffset	; get offset to rightmost dest dword

; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword quad)
; can be 0, 8, 16 or 24.  If it's 0, we enter the loop where it reads
; one source dword before writing the dest.  If it's 8, we enter the loop
; where it reads 2 source dwords and so on.
	add	dstAddr, rScratch	; pt to rightmost dest doubleword
	mov	rScratch, srcLimit	; get srcOffset

	add	rSrcAddr3, rScratch1	; pt to line 3
; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; equals 8N + 3.  In reducing by 4, we only look at 8N bits.  Therefore,
; we only want to round up to the next byte if we're looking at 8N + 4,
; 8N + 5, etc.  So, we only add 4, not 7.
	mov	rScratch1, srcWidth	; get source width in bits.

	add	rScratch1, 4		; round up to next byte
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	sub	rScratch, srcOffset	; get difference
	pxor	mm0, mm0		; clear registers

; Initialize all the MMX registers in a wad.  Since none of these
; operations affect the condition codes, the codes from the "sub"
; above will be the ones the "je" below looks at.
	pxor	mm1, mm1
	pxor	mm2, mm2

	movq	mm7, [MMWORD PTR all7sMask]
	pxor	mm3, mm3

; We mask each source nibble with 7, then add 7 to set MS bit if there
; are any 1's in the lower three bits.
	pxor	mm4, mm4
	pxor	mm5, mm5		; mm5 gets used at red4T4LoopW1

	pxor	mm6, mm6 
	je	red4T4LoopW0Entry	; if there's only one dword in
					; the rightmost quad, process it.

	cmp	rScratch, 8		; if there are two dwords in the
	je	red4T4LoopW1Entry	; rightmost quad, process them

	cmp	rScratch, 16		; if there are three dwords in the
	je	red4T4LoopW2Entry	; rightmost quad, process them



; There are four words in the loop.  Read the cols 3 and 2,
; then branch into the loop to read cols 1 and 0.
red4T4LoopW3Entry:
	movq	mm3, [rSrcAddr0 + 24]	; get top word 3
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	pand	mm3, [rSrcAddr1 + 24]	; and with next word 3

	pand	mm3, [rSrcAddr2 + 24]	; and next word 3

	pand	mm3, [rSrcAddr3 + 24]	; and final word 3

	pand	mm3, [MMWORD PTR mmxByteEdgeMask + rScratch1]

; Read dword 2 and DO NOT mask it
	movq	mm2, [rSrcAddr0 + 16]	; get top word 2

	pand	mm2, [rSrcAddr1 + 16]	; and with next word 2

	pand	mm2, [rSrcAddr2 + 16]	; and next word 2
	movq	mm5, mm3		; get ready for step 1, col 3

	pand	mm2, [rSrcAddr3 + 16]	; and final word 2

	jmp	red4T4LoopW1		; do next two dword cols

; Here's where we start when dword 2 is the rightmost word in the scanline.
red4T4LoopW2Entry:
	movq	mm2, [rSrcAddr0 + 16]	; get top word 2
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	pand	mm2, [rSrcAddr1 + 16]	; and with next word 2

	pand	mm2, [rSrcAddr2 + 16]	; and next word 2

	pand	mm2, [rSrcAddr3 + 16]	; and final word 2

	pand	mm2, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	jmp	red4T4LoopW1		; do next two dword cols

; Here's where we start when dword 1 is the rightmost word in the scanline.
red4T4LoopW1Entry:
	movq	mm1, [rSrcAddr0 + 8]	; get top word 1
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	pand	mm1, [rSrcAddr1 + 8]	; and with next word 1

	pand	mm1, [rSrcAddr2 + 8]	; and next word 1

	pand	mm1, [rSrcAddr3 + 8]	; and final word 1

	pand	mm1, [MMWORD PTR mmxByteEdgeMask + rScratch1]

; Read dword 0 and DO NOT mask it
	movq	mm0, [rSrcAddr0]	; get top word 0

	pand	mm0, [rSrcAddr1]	; and with next word 0

	pand	mm0, [rSrcAddr2]	; and next word 0
	movq	mm5, mm1		; get ready for step 1, col 1

	pand	mm0, [rSrcAddr3]	; and final word 0

	jmp	red4T4LoopBottom	; do next two dword cols

; Here's where we start when dword 0 is the rightmost word in the scanline.
red4T4LoopW0Entry:
	movq	mm0, [rSrcAddr0]	; get top word 0
	and	rScratch1, 038h		; = (src width in bytes MOD 8) * 8

	pand	mm0, [rSrcAddr1]	; and with next word 0

	pand	mm0, [rSrcAddr2]	; and next word 0

	pand	mm0, [rSrcAddr3]	; and final word 0

	pand	mm0, [MMWORD PTR mmxByteEdgeMask + rScratch1]
	jmp	red4T4LoopBottom	; do next two dword cols



; --------------------- MAIN LOOP -----------------------------

; --------------------- Read Word 3 ---------------------------

; This is the start of the main loop.  Read and accumulate results from
; dword column 3.  Start doing horizontal reduction of data in word
; columns 0 and 1.
; Note that we only get here from the bottom of the loop, so we can
; depend on mm4 already holding the contents of mm0.
red4T4LoopW3:
	movq	mm3, [rSrcAddr0 + 24]	; get top word 3
	movq	mm4, mm0		; make copy of col 0 bits

	pand	mm3, [rSrcAddr1 + 24]	; get word 3, line 1
	pand	mm4, mm7		; get rid of MS bit in each nibble

	pand	mm3, [rSrcAddr2 + 24]	; get word 3, line 2
	; empty

	paddb	mm4, [MMWORD PTR all1sMask] ; set MS bit if nib == x111
	; empty

	pand	mm3, [rSrcAddr3 + 24]	; get word 3, line 3
	pand	mm0, mm4		; set MS bit if all 1's in nibble

; We need to use more masks and don't have the registers to hold them
; at this point, so we just do it now.  We'll save the reading of mm2
; for later when we need to let the other consolidation channel (mm1's)
; catch up.
	pand	mm0, [MMWORD PTR all8sMask] ; select just nib result bits
	pand	mm5, mm7		; MS bit of col 1 nibbles

	paddb	mm0, [MMWORD PTR all38sMask] ; carry LS nib result bit up to
	; empty				; bit 6 of each byte.

	paddb	mm5, [MMWORD PTR all1sMask] ; set MS bit if nib == x111
	; empty

	pand	mm0, [MMWORD PTR allC0Mask] ; get paired result bits for col 0
	pand	mm1, mm5		; set MS bit if all 1's in col 1 nibble

; Now do the stuff with col 1 that requires the extra masks
	pand	mm1, [MMWORD PTR all8sMask] ; select just col 1 nib result bits
	movq	mm4, mm0		; get ready to copy col 0 bit pairs

	paddb	mm1, [MMWORD PTR all38sMask] ; put col 1 results in bits 7,6
	psllq	mm4, 6			; put clumps in top end of upper bytes

	pand	mm1, [MMWORD PTR allC0Mask] ; get paired result bits for col 1
	por	mm0, mm4		; col 0 quad results in odd bytes

; Now we finish reading the col 2 inputs while col 1 consolidation catches up
	movq	mm2, [rSrcAddr0 + 16]	; get top word 2
	movq	mm5, mm1		; get ready to copy col 1 bit pairs

	pand	mm2, [rSrcAddr1 + 16]	; get word 2, line 1
	psllq	mm5, 6			; put clumps in top end of upper bytes

	pand	mm2, [rSrcAddr2 + 16]	; get word 2, line 2
	por	mm1, mm5		; col 1 quad results in odd bytes

	pand	mm2, [rSrcAddr3 + 16]	; get word 2, line 3
	psrlw	mm0, 8			; col 0 res in bits 7:4 of even bytes

; We jump in here when consolidating the very last source word read.
red4T4FinishConsolidating:
	psrlw	mm1, 8			; col 1 res in bits 7:4 of even bytes
	mov	rScratch, dstAddr	; read dest address

	packuswb mm1, mm0		; produce col0:1 bits in bits 7:4 of
					; all bytes
	movq	mm4, mm6		; prepare to compact col2:col3 bits

	movq	mm5, mm1		; prepare to compace col0:col1 bits
	psllq	mm4, 4			; put 8 col2:col3 bits in the odd bytes

	psllq	mm5, 4			; put 8 col0:col1 bits in the odd bytes
	por	mm1, mm5		; combine shifted col0:1 bits

	psrlw	mm1, 8			; 8 col0:1 bits in even bytes
	por	mm6, mm4		; combine shifted col2:3 bits

	psrlw	mm6, 8			; 8 col2:3 bits in even bytes
	sub	rScratch, 8		; decrement dest address

; Here's where we wordswap the final dest word by flipping the order
; of the args to the pack command.  The dest appears in the LS word and
; the source appears in the MS word.  Therefore, we get 2:3:0:1, which
; is what we want.
	packuswb mm1, mm6
	mov	dstAddr, rScratch	; update stored version

; Now we need to byteswap the bytes in here.  Remember that each source word
; produced 16 bits, 8 from the upper word and 8 from the lower word.  We
; now need to swap the bytes, which corresponds to swapping the words in
; the original.
	movq	mm4, mm1		; make a copy
	psllw	mm1, 8			; shift low bytes to high bytes

	psrlw	mm4, 8			; shift high bytes to low bytes
	; empty

	por	mm1, mm4		; combine swapped bytes
	; empty

; wait for mm0 to percolate through pipeline
	; empty
	; empty

	movq	[rScratch + 8], mm1	; store dest word
	movq	mm5, mm3		; get ready for step 1, col 3

	cmp	rSrcAddr0, srcAddr	; back at beginning of line yet?
	jl	red4T4Exit		; quit if that was last dest word

; ------------------- Get Bits from Words 1 & 0 ----------------------

; We can enter the loop at this point, so the source addresses must
; be consistent.
red4T4LoopW1:
	movq	mm1, [rSrcAddr0 + 8]	; get top word 1
	movq	mm4, mm2		; make copy of col 2 bits

	pand	mm1, [rSrcAddr1 + 8]	; get word 1, line 1
	pand	mm4, mm7		; get rid of MS bit in each nibble

	pand	mm1, [rSrcAddr2 + 8]	; get word 1, line 2
	; empty

	paddb	mm4, [MMWORD PTR all1sMask] ; set MS bit if nib == x111
	; empty

	pand	mm1, [rSrcAddr3 + 8]	; get word 1, line 3
	pand	mm2, mm4		; set MS bit if all 1's in nibble

; We need to use more masks and don't have the registers to hold them
; at this point, so we just do it now.  We'll save the reading of mm0
; for later when we need to let the other consolidation channel (mm3's)
; catch up.
	pand	mm2, [MMWORD PTR all8sMask] ; select just nib result bits
	pand	mm5, mm7		; MS bit of col 3 nibbles

	paddb	mm2, [MMWORD PTR all38sMask] ; carry LS nib result bit up to
					; bit 6 of each byte.
	; empty

	paddb	mm5, [MMWORD PTR all1sMask] ; set MS bit if nib == x111
	; empty

	pand	mm2, [MMWORD PTR allC0Mask] ; get paired result bits for col 2
	pand	mm3, mm5		; set MS bit if all 1's in col 3 nibble

; Now do the stuff with col 3 that requires the extra masks
	pand	mm3, [MMWORD PTR all8sMask] ; select just col 3 nib result bits
	movq	mm4, mm2		; get ready to copy col 2 bit pairs

	paddb	mm3, [MMWORD PTR all38sMask] ; put col 3 results in bits 7,6
	psllq	mm4, 6			; put clumps in top end of upper bytes

	pand	mm3, [MMWORD PTR allC0Mask] ; get paired result bits for col 3
	por	mm2, mm4		; col 2 quad results in odd bytes

; Now we finish reading the col 0 inputs while col 3 consolidation catches up
	movq	mm0, [rSrcAddr0]	; get top word 0
	movq	mm6, mm3		; get ready to copy col 3 bit pairs

	pand	mm0, [rSrcAddr1]	; get word 0, line 1
	psllq	mm6, 6			; put clumps in top end of upper bytes

	pand	mm0, [rSrcAddr2]	; get word 0, line 2
	por	mm6, mm3		; col 3 quad results in odd bytes

	pand	mm0, [rSrcAddr3]	; get word 0, line 3
	psrlw	mm2, 8			; col 2 res in bits 7:4 of even bytes

	psrlw	mm6, 8			; col 3 res in bits 7:4 of even bytes
	; empty

	packuswb mm6, mm2		; produce col2:3 bits in bits 7:4 of
					; all bytes
; We jump into the loop here at the right edge of the scan line.
red4T4LoopBottom:
	sub	rSrcAddr1, 32		; update rSrcAddr1 by 4*length of dword

	sub	rSrcAddr0, 32		; update rSrcAddr0
	movq	mm5, mm1		; get ready for step 1, col 1

	sub	rSrcAddr2, 32;		; update rSrcAddr2
	sub	rSrcAddr3, 32		; and rSrcAddr3

	cmp	rSrcAddr0, srcAddr	; done all src words yet?
	jge	red4T4LoopW3		; nope, do another quad

; ----------------------------- LOOP EXIT -------------------------------

; At this point, we've just read the leftmost word of the last quad.
; We need to consolidate it, merge it with the other results, unscramble
; them and write out the result.
; We could have copied all the word 3, word 2 and writing code down here, but
; that's overboard even for me.  Instead, we just do the consolidation
; part of the word 3 and word 2 code (no reads allowed) and jump back
; to the code just after we're done reading.  After it finishes writing
; the dest word, there's a test to see if rSrcAddr0 is less than srcAddr.
; If so, it exits.
	movq	mm4, mm0		; make copy of col 0 bits
	; empty

	pand	mm4, mm7		; get rid of MS bit in each nibble
	; empty

	paddb	mm4, [MMWORD PTR all1sMask] ; set MS bit if nib == x111
	; empty

	pand	mm0, mm4		; set MS bit if all 1's in nibble
	; empty

; We need to use more masks and don't have the registers to hold them
; at this point, so we just do it now.  We'll save the reading of mm2
; for later when we need to let the other consolidation channel (mm1's)
; catch up.
	pand	mm0, [MMWORD PTR all8sMask] ; select just nib result bits
	pand	mm5, mm7		; MS bit of col 1 nibbles

	paddb	mm0, [MMWORD PTR all38sMask] ; carry LS nib result bit up to
					; bit 6 of each byte.
	; empty

	paddb	mm5, [MMWORD PTR all1sMask] ; set MS bit if nib == x111
	; empty

	pand	mm0, [MMWORD PTR allC0Mask] ; get paired result bits for col 0
	pand	mm1, mm5		; set MS bit if all 1's in col 1 nibble

; Now do the stuff with col 1 that requires the extra masks
	pand	mm1, [MMWORD PTR all8sMask] ; select just col 1 nib result bits
	movq	mm4, mm0		; get ready to copy col 0 bit pairs

	paddb	mm1, [MMWORD PTR all38sMask] ; put col 1 results in bits 7,6
	psllq	mm4, 6			; put clumps in top end of upper bytes

	pand	mm1, [MMWORD PTR allC0Mask] ; get paired result bits for col 1
	por	mm0, mm4		; col 0 quad results in odd bytes

; Now we finish reading the col 2 inputs while col 1 consolidation catches up
	movq	mm5, mm1		; get ready to copy col 1 bit pairs

	psllq	mm5, 6			; put clumps in top end of upper bytes

	por	mm1, mm5		; col 1 quad results in odd bytes
	psrlw	mm0, 8			; col 0 res in bits 7:4 of even bytes

	; empty
	jmp	red4T4FinishConsolidating


; We get here after the last (leftmost) dword was written
red4T4Exit:
	ret

ip_reduce4T4LineLong ENDP





; *************************************************************************
; ***************************** 8x REDUCTION ******************************
; *************************************************************************

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; In our world, 8x reduction is a lot like 4x reduction.  But, didn't we
; say that using a table to unscramble interleaved bits was a win at
; 8x reduction?  Well, we did, but we lied.  It's all the fault of those
; pesky pairing restrictions on Pentiae.
;
; See, you can only pair a MMX opcode that accesses memory (or an integer
; register) with another MMX opcode.  That other MMX opcode can't access
; memory or an integer register.  In 8x reduction, we have to combine 8
; columns of dwords, each containing 8 rows of dwords.  While we're ORing
; or ANDing the rows of dwords in each column, we're accessing memory.
; There's not a lot to do in the empty slots of these 8 memory accesses
; besides consolidate the previous column's data.  We'll do that.  Forget
; the unscrambling tables.
;
; In 8x reduction, we want to turn each byte (produced from ORing or ANDing
; 8 vertically adjacent bytes) into one bit.  Fairly quickly,
; we want to put 2 bits into every other byte so we can use the pack
; opcode.  This will put 2 bits from 2 dwords into every byte.
; We can then put 4 bits into every other byte using shifts and ORs.
; We pack this with another dword that contains bits from 2 source dwords,
; producing a dword with 4 bits in every byte.  This dword contains the
; bits from 4 source dwords.  We repeat the entire process for four more
; dwords, ending up with two dwords, each containing the bits from 4
; source dwords.  We compact each of these so they contain 8 bits in
; every other byte, and pack them together to get the final result.
; This whole process is shown pictorially below.
;
; How do we deal with wordswapping?  Remember that the LS 32 bits of an
; dword read from the image contain the bits on the *left*.  As with
; 4x reduction, wordswapping the dest dword is easy, we just swap the
; arguments on the final pack opcode.  Dealing with swapping the source
; dword bits is a little trickier.
;
; Each source dword column is reduced to 8 bits in 8x reduction (64/8 = 8).
; Each byte produces 1 bit.  During consolidation, there's a stage where
; we have a dword containing 4 bits/byte and we want to convert this to
; 8 bits in every other byte.  This is going from step 5 to step 6 below.
; The four bits in each byte all belong to the same 32-bit source word.
; We want to swap them with the 4 bits belonging to the other source word
; in the original source dword.  We do this by doing a right shift by
; 12 instead of a left shift by 4 as shown in step 6 below:
;
; In step 1, we combine the rows of the rightmost dword column of the
; 8-dword group of columns that will go into the next dest dword.
; We set each byte to all 1's or all 0's depending on the
; reduced value for that byte.  This is done with the pcmpeqb (packed
; compare for equality in units of bytes) opcode.
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 1:   0000000011111111222222223333333344444444555555556666666677777777
;
; In step 2, we consolidate 2 bits into the bits 2:3 of the odd (as counted
; from the MS end of the word) bytes.  This is done with a shift and a mask op.
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 2:   ..........01..............23..............45..............67....
;
; Steps 1 and 2 are repeated for the next dword to the left.  We pack
; the two resulting dwords into one dword.
; src dword 66      66      66      66      77      77      77      77
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 3:   ..01......23......45......67......01......23......45......67....
;
; Steps 1, 2 and 3 are repeated for 2 more source dwords.  We then shift
; and OR the two result dwords so they each have 4 bits at the MS end of each
; odd byte.  This is done by shifting the a copy of the dword *right* by
; 6 and ORing, producing ..01....|0123....| in the top bytes.  We then
; mask this with 00F0 00F0 00F0 00F0, which leaves the correct bits.
; src dword         6666            6666            7777            7777
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 4:   ........0123............4567............0123............4567....
;
; These two dwords are then packed together using the pack opcode to produce
; a dword with 4 bits in every byte.
; src dword 4444    4444    5555    5555    6666    6666    7777    7777    
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 5:   0123....4567....0123....4567....0123....4567....0123....4567....
;
; Steps 1 through 5 are repeated for the next 4 source dwords on the left
; producing another such dword containing bits from dwords 0, 1, 2 and 3.
; Here's the trick.  We have to consolidate each of these dwords so they
; have 8 bits in every other byte.  We do this by shift a copy right by 12
; and ORing with the original and selecting only the odd bytes, which
; produces:
; src dword         44444444        55555555        66666666        77777777
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 6:   ........45670123........45670123........45670123........45670123
;
; The bits from the LS word of the source dwords are now in the MS bits of
; the consolidated bytes, effectively word swapping the source dwords.  We
; do this operation at this stage since we had to do this combining op
; anyway.  We then pack two such source words, putting the 32 bit from the
; left in the image on the right in the dest dword.
; src dword 4444444455555555666666667777777700000000111111112222222233333333
; position: 0123456701234567012345670123456701234567012345670123456701234567
; step 7:   4567012345670123456701234567012345670123456701234567012345670123

; There's one more trick.  For threshold 1 reduction, we'd like to produce
; a 1 in the output if there are any 1's in the input.  The way we produce
; the original masks in step 1 is to use a packed compare MMX opcode.
; The closest we can come to comparing for any 1's is to compare for all
; 0's.  However, this produces a 1 when the input byte was all 0's, which
; is the inverse of what we want.  We fix it at the very end, by inverting
; the dest word before it's written.  For threshold 4 reduction, we combine
; the rows of each 8x8 square by ANDing them.  If there are any 0's, the
; resulting byte will have some 0's.  We compare it with all 1's.  If there
; were any 0's, the result is 0, which is what we want.  Therefore, we do not
; need to invert the final dest word for threshold 4.
;
; For threshold 1, it would be swell to use an unsigned packed compare
; for greater than to detect any 1's, but there's no such thing.
; MMX only has signed compares.  Any negative number is less than 0x00
; even though it has bits turned on.  Drat.
;
; Addressing the source images a little trick.  We want to have
; address registers set up so we can just read 8 lines from the source
; images.  But we don't have 8 address registers.  In fact, we're having
; trouble putting 4 or 5 together.  This will work though.  We use five integer
; registers as follows:
; rSrcAddr0:	pointer to line 0 (top line)
; rSrcAddr1:	pointer to line 3
; rSrcAddr2:	pointer to line 4
; rSrcAddr3: 	pointer to line 7
; rSrcBpl:	source bytes/line
;
; We address the 8 lines as follows:
; line 0:	[rSrcAddr0 + colOffset]
; line 1:	[rSrcAddr0 + rSrcBpl + colOffset]
; line 2: 	[rSrcAddr0 + 2*rSrcBpl + colOffset]
; line 3:	[rSrcAddr1 + colOffset]
; line 4:	[rSrcAddr2 + colOffset]
; line 5:	[rSrcAddr2 + rSrcBpl + colOffset]
; line 6: 	[rSrcAddr2 + 2*rSrcBpl + colOffset]
; line 7:	[rSrcAddr3 + colOffset]
;
; This means we only have one scratch register, but life is tough all over.

rScratch1	TEXTEQU	<>		; get rid of this definition so we'll
					; know if we use it by mistake
rSrcBpl		TEXTEQU <edi>		; new alias for edi.

; Parameters
; srcAddr	addr of left end of source line
; srcWidth	width of source image in bits
; srcBpl	source bytes/line
; dstAddr	addr of left end of dest line
; dstOffset	offset in bytes to rightmost dest word
; srcOffset	offset in bytes to leftmost byte in rightmost
;		group of source bits that will form the rightmost
;		dest word
; srcLimit	offset in bytes to rightmost source byte (often
;		to the right of byte at srcOffset)
; tab4to1Permute addr of table used to unscramble reduced bits.
;		 Unused in this code.


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduce8T1LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduce8T1LineLongStart::
ip_reduce8T1LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab8to1Permute:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword octet in the line.
; This octet may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	eax, tab8to1Permute	; Make the assembler happy
	mov	rSrcAddr0, srcAddr	; pt to 1st source dword

	mov	rSrcBpl, srcBpl		; fill empty slot
	add	rSrcAddr0, srcOffset	; pt to left source dword of
					; first dword octet.
					; We do this because we're really
					; always going to go after the
					; first doubleword in an octet.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first octet.

; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
; NOTE that we don't set the "results" from previous dwords here.  Since
; we want to pretend that they contain the results from reading 0's and
; the result of reading 0's is all 1's, we need to set 1's.  However,
; different loop entry points need different sets of 1's, so we set
; them in the loop entry code.
	mov	rScratch, srcWidth	; get source width in bits.
	lea	rSrcAddr2, [rSrcAddr0 + 4*rSrcBpl] ; point to src line 4

	and	rScratch, 038h		; = (srcWidth in bytes MOD 8) * 8
	lea	rSrcAddr3, [rSrcAddr0 + 8*rSrcBpl] ; point to src line 8

; Get the mask used to get rid of pad bits.
	movq	mm5, [MMWORD PTR mmxByteEdgeMask + rScratch]
	; empty

	mov	rScratch, dstOffset	; get offset to rightmost dest dword
	sub	rSrcAddr3, rSrcBpl	; point to line 7

	add	dstAddr, rScratch	; pt to rightmost dest doublewor
; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword octet)
; can be 0, 8, 16, 24, 32, 40, 48 or 56.  If it's 0, we enter the loop
; where it reads one source dword before writing the dest.  If it's 8,
; we enter the loop where it reads 2 source dwords and so on.
	mov	rScratch, srcLimit	; get srcLimit

	mov	rSrcAddr1, rSrcAddr2	; point to line 4
; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; is between 8N + 1 and 8N + 7.  In reducing by 8, we only look at 8N bits.
; Therefore, we never want to round up to the next byte!  So skip the
; rounding step and just subtract srcOffset.
	sub	rScratch, srcOffset

	sub	rSrcAddr1, rSrcBpl	; point to line 3
; Get the offset into the table of pointers to jump entry locations:
	shr	rScratch, 1		; [0, 8, 16, ...] -> [0, 4, 8, ...]

; Get the mask used in step 4 above - used after putting nibbles into MS
; half of odd bytes (numbered from the left).
	movq	mm4, [MMWORD PTR all00F0Mask]
	; empty

; Simulate reading all 0's from cols 4,5,6,7.
	movq	mm6, [MMWORD PTR allF0Mask]
	; empty

	movq	mm7, [MMWORD PTR all0030Mask]
	; empty

	; empty
	jmp	red8T1JumpTable[rScratch]


;------------------------- LOOP ENTRIES  -----------------------------

; This is the loop entry taken when we want to read 8 columns of dwords
; to form the first dest dword.
red8T1EntryW7:
	movq	mm3, [rSrcAddr0 + 7*8]			; line 0

	por	mm3, [rSrcAddr0 + rSrcBpl + 7*8]	; line 1

	por	mm3, [rSrcAddr0 + 2*rSrcBpl + 7*8]	; line 2

	por	mm3, [rSrcAddr1 + 7*8]			; line 3

	por	mm3, [rSrcAddr2 + 7*8]			; line 4

	por	mm3, [rSrcAddr2 + rSrcBpl + 7*8]	; line 5

	por	mm3, [rSrcAddr2 + 2*rSrcBpl + 7*8]	; line 6

	por	mm3, [rSrcAddr3 + 7*8]			; line 7

	pand	mm3, mm5
	pxor	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T1LoopEntryW6


; This is the loop entry taken when we want to read 7 columns of dwords
; to form the first dest dword.
red8T1EntryW6:
	movq	mm2, [rSrcAddr0 + 6*8]			; line 0

	por	mm2, [rSrcAddr0 + rSrcBpl + 6*8]	; line 1

	por	mm2, [rSrcAddr0 + 2*rSrcBpl + 6*8]	; line 2

	por	mm2, [rSrcAddr1 + 6*8]			; line 3

	por	mm2, [rSrcAddr2 + 6*8]			; line 4

	por	mm2, [rSrcAddr2 + rSrcBpl + 6*8]	; line 5

	por	mm2, [rSrcAddr2 + 2*rSrcBpl + 6*8]	; line 6

	por	mm2, [rSrcAddr3 + 6*8]			; line 7
	movq	mm3, mm7				; simulate consolidated
							; col 7 data

	pand	mm2, mm5
	pxor	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T1LoopEntryW5


; This is the loop entry taken when we want to read 6 columns of dwords
; to form the first dest dword.
red8T1EntryW5:
	movq	mm1, [rSrcAddr0 + 5*8]			; line 0

	por	mm1, [rSrcAddr0 + rSrcBpl + 5*8]	; line 1

	por	mm1, [rSrcAddr0 + 2*rSrcBpl + 5*8]	; line 2

	por	mm1, [rSrcAddr1 + 5*8]			; line 3

	por	mm1, [rSrcAddr2 + 5*8]			; line 4

	por	mm1, [rSrcAddr2 + rSrcBpl + 5*8]	; line 5

	por	mm1, [rSrcAddr2 + 2*rSrcBpl + 5*8]	; line 6

	por	mm1, [rSrcAddr3 + 5*8]			; line 7

	pand	mm1, mm5
	pxor	mm5, mm5				; prepare for compare

	movq	mm2, [MMWORD PTR all30Mask]		; simulate col 6,7 data
	jmp	red8T1LoopEntryW4


; This is the loop entry taken when we want to read 5 columns of dwords
; to form the first dest dword.
red8T1EntryW4:
	movq	mm0, [rSrcAddr0 + 4*8]			; line 0

	por	mm0, [rSrcAddr0 + rSrcBpl + 4*8]	; line 1

	por	mm0, [rSrcAddr0 + 2*rSrcBpl + 4*8]	; line 2

	por	mm0, [rSrcAddr1 + 4*8]			; line 3

	por	mm0, [rSrcAddr2 + 4*8]			; line 4

	por	mm0, [rSrcAddr2 + rSrcBpl + 4*8]	; line 5

	por	mm0, [rSrcAddr2 + 2*rSrcBpl + 4*8]	; line 6

	por	mm0, [rSrcAddr3 + 4*8]			; line 7
	movq	mm1, mm7				; simulate consolidated
							; col 5 data

	pand	mm0, mm5
	pxor	mm5, mm5				; prepare for compare

	movq	mm2, [MMWORD PTR all30Mask]		; simulate col 6,7 data
	jmp	red8T1LoopEntryW3


; This is the loop entry taken when we want to read 4 columns of dwords
; to form the first dest dword.
red8T1EntryW3:
	movq	mm3, [rSrcAddr0 + 3*8]			; line 0

	por	mm3, [rSrcAddr0 + rSrcBpl + 3*8]	; line 1

	por	mm3, [rSrcAddr0 + 2*rSrcBpl + 3*8]	; line 2

	por	mm3, [rSrcAddr1 + 3*8]			; line 3

	por	mm3, [rSrcAddr2 + 3*8]			; line 4

	por	mm3, [rSrcAddr2 + rSrcBpl + 3*8]	; line 5

	por	mm3, [rSrcAddr2 + 2*rSrcBpl + 3*8]	; line 6

	por	mm3, [rSrcAddr3 + 3*8]			; line 7

	pand	mm3, mm5
	pxor	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T1LoopEntryW2


; This is the loop entry taken when we want to read 3 columns of dwords
; to form the first dest dword.
red8T1EntryW2:
	movq	mm2, [rSrcAddr0 + 2*8]			; line 0

	por	mm2, [rSrcAddr0 + rSrcBpl + 2*8]	; line 1

	por	mm2, [rSrcAddr0 + 2*rSrcBpl + 2*8]	; line 2

	por	mm2, [rSrcAddr1 + 2*8]			; line 3

	por	mm2, [rSrcAddr2 + 2*8]			; line 4

	por	mm2, [rSrcAddr2 + rSrcBpl + 2*8]	; line 5

	por	mm2, [rSrcAddr2 + 2*rSrcBpl + 2*8]	; line 6

	por	mm2, [rSrcAddr3 + 2*8]			; line 7
	movq	mm3, mm7				; simulate consolidated
							; col 3 data

	pand	mm2, mm5
	pxor	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T1LoopEntryW1


; This is the loop entry taken when we want to read 2 columns of dwords
; to form the first dest dword.
red8T1EntryW1:
	movq	mm1, [rSrcAddr0 + 1*8]			; line 0

	por	mm1, [rSrcAddr0 + rSrcBpl + 1*8]	; line 1

	por	mm1, [rSrcAddr0 + 2*rSrcBpl + 1*8]	; line 2

	por	mm1, [rSrcAddr1 + 1*8]			; line 3

	por	mm1, [rSrcAddr2 + 1*8]			; line 4

	por	mm1, [rSrcAddr2 + rSrcBpl + 1*8]	; line 5

	por	mm1, [rSrcAddr2 + 2*rSrcBpl + 1*8]	; line 6

	por	mm1, [rSrcAddr3 + 1*8]			; line 7

	pand	mm1, mm5
	pxor	mm5, mm5				; prepare for compare

	movq	mm2, [MMWORD PTR all30Mask]		; simulate col 2,3 data
	jmp	red8T1LoopEntryW0


; This is the loop entry taken when we want to read 1 column of dwords
; to form the first dest dword.
red8T1EntryW0:
	movq	mm0, [rSrcAddr0 + 0*8]			; line 0

	por	mm0, [rSrcAddr0 + rSrcBpl + 0*8]	; line 1

	por	mm0, [rSrcAddr0 + 2*rSrcBpl + 0*8]	; line 2

	por	mm0, [rSrcAddr1 + 0*8]			; line 3

	por	mm0, [rSrcAddr2 + 0*8]			; line 4

	por	mm0, [rSrcAddr2 + rSrcBpl + 0*8]	; line 5

	por	mm0, [rSrcAddr2 + 2*rSrcBpl + 0*8]	; line 6

	por	mm0, [rSrcAddr3 + 0*8]			; line 7
	movq	mm1, mm7				; simulate consolidated
							; col 1 data

	pand	mm0, mm5
	pxor	mm5, mm5				; prepare for compare

	movq	mm2, [MMWORD PTR all30Mask]		; simulate col 2,3 data
	jmp	red8T1LoopEnd


; ------------------------- MAIN LOOP ------------------------------

; This is the main loop for 8x, threshold 1 reduction.
;
; The basic loop structure is:
;  A. Read word 7 of a new dword octet and consolidate data from
;     col 0 of the dword octet to the right.
;	mm0 = raw col0 data from dword octet to the right
;	mm1 = consolidated col1 data from dword octet to the right
;	mm2 = col 2,3 data from dword octet to the right
;	mm3 = accumulating new raw col 7 data
;	mm5 = used to accumulate col0 data
;	mm6 = col 4,5,6,7 data from dword octet to the right
;  B. Combine the consolidated col 0 data with that for col 1.
;  C. Combine the col 0,1 dword with col 2,3 dword.
;  D. Combine col 0,1,2,3 dword with col 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
;  E. Write the result out to [dstAddr] and decrement dstAddr by 8.
;  F. If that was the last dest word, quit.
;  G. Read word 6 of this dword octet and consolidate data from
;     col 7 of this dword octet.
;	mm2 = accumulating raw data from col 6
;	mm3 = raw col 7 data
;	mm5 = used to accumulate col 7 data
;  H. Read word 5 of this dword octet and consolidate data from
;     col 6 of this dword octet.
;	mm1 = accumulating raw col 5 data
;	mm2 = raw col 6 data
;	mm3 = consolidated col7 data
;	mm5 = used to consolidate col 6 data
;  I. Combine consolidated col7 and consolidated col6 data.
;  J. Read word 4 of this dword octet and consolidate data from
;     col 5 of this dword octet.
;	mm0 = accumulating raw data from col 4
;	mm1 = raw col 5 data
;	mm2 = col 6,7 data
;	mm5 = used to accumulate col 5 data
;  K. Read word 3 of this dword octet and consolidate data from
;     col 4 of this dword octet.
;	mm0 = raw col4 data
;	mm1 = consolidated col 5 data
;	mm2 = col 6,7 data
;	mm3 = used to accumulate raw col3 data
;	mm5 = used to consolidate col 4 data
;  L. Combine consolidated col5 and consolidated col4 data.
;  M. Combine consolidated col 4,5 and consolidated col 6,7 data.
;  N. Read word 2 of this dword octet and consolidate data from
;     col 3 of this dword octet.
;	mm2 = accumulating raw data from col 2
;	mm3 = raw col 3 data
;	mm5 = used to accumulate col 3 data
;	mm6 = col 4,5,6,7 data
;  O. Read word 1 of this dword octet and consolidate data from
;     col 2 of this dword octet.
;	mm1 = accumulating raw col 1 data
;	mm2 = raw col 2 data
;	mm3 = consolidated col3 data
;	mm5 = used to consolidate col 2 data
;	mm6 = col 4,5,6,7 data
;  P. Combine consolidated col 2 and consolidated col 3 data.
;  Q. Read word 0 of this dword octet and consolidate data from
;     col 1 of this dword octet.
;	mm0 = accumulating raw data from col 0
;	mm1 = raw col 1 data
;	mm2 = col 2,3 data
;	mm5 = used to accumulate col 1 data
;	mm6 = col 4,5,6,7 data
;  R. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
;  S. Since that was the last dword octet, do the consolidate of col 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
;


; Here's the real loop code.  The comments above have been copied
; to the top of each section.
; In all sections:
;	mm4 = mask containing 00F0 00F0 00F0 00F0
;	mm5 = usually contains 0, sometimes used to shift stuff, then
;	      restored to 0.
;	mm7 = mask containing 0300 0300 0300 0300

;  A. Read word 7 of a new dword octet and consolidate data from
;     col 0 of the dword octet to the right.
;	mm0 = raw col0 data from dword octet to the right
;	mm1 = consolidated col1 data from dword octet to the right
;	mm2 = col 2,3 data from dword octet to the right
;	mm3 = accumulating new raw col 7 data
;	mm5 = used to accumulate col0 data
;	mm6 = col 4,5,6,7 data from dword octet to the right
red8T1LoopEntryW7:
	movq	mm3, [rSrcAddr0 + 7*8]			; line 0
	pcmpeqb	mm0, mm5				; any 1's?

	por	mm3, [rSrcAddr0 + rSrcBpl + 7*8]	; line 1
	psrlq	mm0, 3					; res in odd bytes

	por	mm3, [rSrcAddr0 + 2*rSrcBpl + 7*8]	; line 2
	pand	mm0, mm7				; select col 0 bits

;  B. Combine the consolidated col 0 data with that for col 1.
	por	mm3, [rSrcAddr1 + 7*8]			; line 3
	packuswb mm1, mm0				; col 0 bits on the left

;  C. Combine the col 0,1 dword with col 2,3 dword.
	por	mm3, [rSrcAddr2 + 7*8]			; line 4
	movq	mm0, mm2				; prepare cols 2,3

	por	mm3, [rSrcAddr2 + rSrcBpl + 7*8]	; line 5
	movq	mm5, mm1				; prepare cols 0,1

	por	mm3, [rSrcAddr2 + 2*rSrcBpl + 7*8]	; line 6
	psrlw	mm0, 6					; pack to bottom bytes

	por	mm3, [rSrcAddr3 + 7*8]			; line 7
	psrlw	mm5, 6					; pack to bottom bytes

; This is where we jump to when we've just finished consolidating the
; leftmost source dword in the scan line (step S).
red8T1LoopFinish:
	por	mm2, mm0				; combine col 2,3 data
	por	mm1, mm5				; combine col 0,1 data

	pand	mm2, mm4				; mask off extra bits
	pand	mm1, mm4				; mask off extra bits

;  Step C ends with packuswb below.
;  D. Combine col 0,1,2,3 dword with col 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
	packuswb mm2, mm1				; pack col 0,1,2,3 data
	movq	mm0, mm6				; get ready to align
							; col 4,5,6,7 data

	movq	mm5, mm2				; get ready to align
							; col 0,1,2,3 data
	psrlw	mm0, 12					; pack and "wordswap"
							; results from cols
							; 4,5,6,7.

	psrlw	mm5, 12					; pack and "wordswap"
							; results from cols
							; 0,1,2,3
	por	mm6, mm0				; combine results

	pand	mm6, [MMWORD PTR all00FFMask]		; get bytes we want
	por	mm2, mm5				; combine 0123 results

	pand	mm2, [MMWORD PTR all00FFMask]		; get bytes we want
	pcmpeqb	mm0, mm0				; get all 1's mask

; Note that we intentionally put the mm6 bits (cols 4,5,6,7) in the UPPER
; half of the result dword.  This accounts for the wordswapping inherent
; in this little endian platform.
	packuswb mm2, mm6				; get inverted dest dw
	mov	rScratch, dstAddr			; load dest address

	pxor	mm2, mm0				; invert dest word
							; because all result
							; bits were inverted
							; from original
							; pcmpeqb that made them
	sub	rScratch, 8				; point to prev word

; wait for mm2 and rScratch to percolate through pipeline
	mov	dstAddr, rScratch			; pt to prev src word
	; empty

;  E. Write the result out to [dstAddr] and decrement dstAddr by 8.
	movq	[rScratch + 8], mm2			; write result dword
	pxor	mm5, mm5				; get ready to compare

;  F. If that was the last dest word, quit.
	cmp	rSrcAddr0, srcAddr
	jl	red8T1Exit		; quit if that was last dest word


;  G. Read word 6 of the new dword octet and consolidate data from
;     col 7 of this dword octet.
;	mm2 = accumulating raw data from col 6
;	mm3 = raw col 7 data
;	mm5 = used to accumulate col 7 data
red8T1LoopEntryW6:
	movq	mm2, [rSrcAddr0 + 6*8]			; line 0
	pcmpeqb	mm3, mm5				; any 1's?

	por	mm2, [rSrcAddr0 + rSrcBpl + 6*8]	; line 1
	psrlq	mm3, 3					; res in odd bytes

	por	mm2, [rSrcAddr0 + 2*rSrcBpl + 6*8]	; line 2
	pand	mm3, mm7				; select col 7 bits

	por	mm2, [rSrcAddr1 + 6*8]			; line 3

	por	mm2, [rSrcAddr2 + 6*8]			; line 4

	por	mm2, [rSrcAddr2 + rSrcBpl + 6*8]	; line 5

	por	mm2, [rSrcAddr2 + 2*rSrcBpl + 6*8]	; line 6

	por	mm2, [rSrcAddr3 + 6*8]			; line 7


;  H. Read word 5 of the new dword octet and consolidate data from
;     col 6 of this dword octet.
;	mm1 = accumulating raw col 5 data
;	mm2 = raw col 6 data
;	mm3 = consolidated col7 data
;	mm5 = used to consolidate col 6 data
red8T1LoopEntryW5:
	movq	mm1, [rSrcAddr0 + 5*8]			; line 0
	pcmpeqb	mm2, mm5				; any 1's?

	por	mm1, [rSrcAddr0 + rSrcBpl + 5*8]	; line 1
	psrlq	mm2, 3					; res in odd bytes

	por	mm1, [rSrcAddr0 + 2*rSrcBpl + 5*8]	; line 2
	pand	mm2, mm7				; select col 6 bits

;  I. Combine consolidated col7 and consolidated col6 data.
	por	mm1, [rSrcAddr1 + 5*8]			; line 3
	packuswb mm3, mm2				; col 6 bits on the left

	por	mm1, [rSrcAddr2 + 5*8]			; line 4
	movq	mm2, mm3				; where we want it

	por	mm1, [rSrcAddr2 + rSrcBpl + 5*8]	; line 5

	por	mm1, [rSrcAddr2 + 2*rSrcBpl + 5*8]	; line 6

	por	mm1, [rSrcAddr3 + 5*8]			; line 7


;  J. Read word 4 of the new dword octet and consolidate data from
;     col 5 of this dword octet.
;	mm0 = accumulating raw data from col 4
;	mm1 = raw col 5 data
;	mm2 = col 6,7 data
;	mm5 = used to accumulate col 5 data
red8T1LoopEntryW4:
	movq	mm0, [rSrcAddr0 + 4*8]			; line 0
	pcmpeqb	mm1, mm5				; any 1's?

	por	mm0, [rSrcAddr0 + rSrcBpl + 4*8]	; line 1
	psrlq	mm1, 3					; res in odd bytes

	por	mm0, [rSrcAddr0 + 2*rSrcBpl + 4*8]	; line 2
	pand	mm1, mm7				; select col 5 bits

	por	mm0, [rSrcAddr1 + 4*8]			; line 3

	por	mm0, [rSrcAddr2 + 4*8]			; line 4

	por	mm0, [rSrcAddr2 + rSrcBpl + 4*8]	; line 5

	por	mm0, [rSrcAddr2 + 2*rSrcBpl + 4*8]	; line 6

	por	mm0, [rSrcAddr3 + 4*8]			; line 7


;  K. Read word 3 of the new dword octet and consolidate data from
;     col 4 of this dword octet.
;	mm0 = raw col4 data
;	mm1 = consolidated col 5 data
;	mm2 = col 6,7 data
;	mm3 = used to accumulate raw col3 data
;	mm5 = used to consolidate col 4 data
red8T1LoopEntryW3:
	movq	mm3, [rSrcAddr0 + 3*8]			; line 0
	pcmpeqb	mm0, mm5				; any 1's?

	por	mm3, [rSrcAddr0 + rSrcBpl + 3*8]	; line 1
	psrlq	mm0, 3					; res in odd bytes

	por	mm3, [rSrcAddr0 + 2*rSrcBpl + 3*8]	; line 2
	pand	mm0, mm7				; select col 4 bits

;  L. Combine consolidated col5 and consolidated col4 data.
	por	mm3, [rSrcAddr1 + 3*8]			; line 3
	packuswb mm1, mm0				; col 4 bits on the left

;  M. Combine consolidated col 4,5 and consolidated col 6,7 data.
	por	mm3, [rSrcAddr2 + 3*8]			; line 4
	movq	mm6, mm2				; prepare cols 6,7

	por	mm3, [rSrcAddr2 + rSrcBpl + 3*8]	; line 5
	movq	mm5, mm1				; prepare cols 4,5

	por	mm3, [rSrcAddr2 + 2*rSrcBpl + 3*8]	; line 6
	psrlw	mm6, 6					; pack to bottom bytes

	por	mm3, [rSrcAddr3 + 3*8]			; line 7
	psrlw	mm5, 6					; pack to bottom bytes

	por	mm6, mm2				; combine col 6,7 data
	por	mm1, mm5				; combine col 4,5 data

	pand	mm6, mm4				; mask off extra bits
	pand	mm1, mm4				; mask off extra bits

	packuswb mm6, mm1				; get col 4,5,6,7 data
	pxor	mm5, mm5				; prepare for compare

;  N. Read word 2 of the new dword octet and consolidate data from
;     col 3 of this dword octet.
;	mm2 = accumulating raw data from col 2
;	mm3 = raw col 3 data
;	mm5 = used to accumulate col 3 data
;	mm6 = col 4,5,6,7 data
red8T1LoopEntryW2:
	movq	mm2, [rSrcAddr0 + 2*8]			; line 0
	pcmpeqb	mm3, mm5				; any 1's?

	por	mm2, [rSrcAddr0 + rSrcBpl + 2*8]	; line 1
	psrlq	mm3, 3					; res in odd bytes

	por	mm2, [rSrcAddr0 + 2*rSrcBpl + 2*8]	; line 2
	pand	mm3, mm7				; select col 3 bits

	por	mm2, [rSrcAddr1 + 2*8]			; line 3

	por	mm2, [rSrcAddr2 + 2*8]			; line 4

	por	mm2, [rSrcAddr2 + rSrcBpl + 2*8]	; line 5

	por	mm2, [rSrcAddr2 + 2*rSrcBpl + 2*8]	; line 6

	por	mm2, [rSrcAddr3 + 2*8]			; line 7

;  O. Read word 1 of the new dword octet and consolidate data from
;     col 2 of this dword octet.
;	mm1 = accumulating raw col 1 data
;	mm2 = raw col 2 data
;	mm3 = consolidated col3 data
;	mm5 = used to consolidate col 2 data
;	mm6 = col 4,5,6,7 data
red8T1LoopEntryW1:
	movq	mm1, [rSrcAddr0 + 1*8]			; line 0
	pcmpeqb	mm2, mm5				; any 1's?

	por	mm1, [rSrcAddr0 + rSrcBpl + 1*8]	; line 1
	psrlq	mm2, 3					; res in odd bytes

	por	mm1, [rSrcAddr0 + 2*rSrcBpl + 1*8]	; line 2
	pand	mm2, mm7				; select col 2 bits

;  P. Combine consolidated col 2 and consolidated col 3 data.
	por	mm1, [rSrcAddr1 + 1*8]			; line 3
	packuswb mm3, mm2				; col 2 bits on the left

	por	mm1, [rSrcAddr2 + 1*8]			; line 4
	movq	mm2, mm3				; where we want it

	por	mm1, [rSrcAddr2 + rSrcBpl + 1*8]	; line 5

	por	mm1, [rSrcAddr2 + 2*rSrcBpl + 1*8]	; line 6

	por	mm1, [rSrcAddr3 + 1*8]			; line 7

;  Q. Read word 0 of the new dword octet and consolidate data from
;     col 1 of this dword octet.
;	mm0 = accumulating raw data from col 0
;	mm1 = raw col 1 data
;	mm2 = col 2,3 data
;	mm5 = used to accumulate col 1 data
;	mm6 = col 4,5,6,7 data
red8T1LoopEntryW0:
	movq	mm0, [rSrcAddr0 + 0*8]			; line 0
	pcmpeqb	mm1, mm5				; any 1's?

	por	mm0, [rSrcAddr0 + rSrcBpl + 0*8]	; line 1
	psrlq	mm1, 3					; res in odd bytes

	por	mm0, [rSrcAddr0 + 2*rSrcBpl + 0*8]	; line 2
	pand	mm1, mm7				; select col 1 bits

	por	mm0, [rSrcAddr1 + 0*8]			; line 3

	por	mm0, [rSrcAddr2 + 0*8]			; line 4

	por	mm0, [rSrcAddr2 + rSrcBpl + 0*8]	; line 5

	por	mm0, [rSrcAddr2 + 2*rSrcBpl + 0*8]	; line 6

	por	mm0, [rSrcAddr3 + 0*8]			; line 7

;  R. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
;     Note that we couldn't do this in the empty slots above since
;     you can't pair an integer opcode with an MMX opcode that references
;     memory or integer registers.  Bleah!
red8T1LoopEnd:
	sub	rSrcAddr3, 64			; decrement by size of octet
	sub	rSrcAddr2, 64			; which is 8 dwords across
						; times 8 bytes/dword

	sub	rSrcAddr1, 64
	sub	rSrcAddr0, 64

	cmp	rSrcAddr0, srcAddr		; done all src words yet?
	jge	red8T1LoopEntryW7		; nope, do another octet


;  S. Since that was the last dword octet, do the consolidate of col 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
	pcmpeqb	mm0, mm5				; any 1's?

	psrlq	mm0, 3					; res in odd bytes

	pand	mm0, mm7				; select col 0 bits

;  B'. Combine the consolidated col 0 data with that for col 1.
	packuswb mm1, mm0				; mm0 bits on the left
;  C'. Combine the col 0,1 dword with col 2,3 dword.
	movq	mm0, mm2				; prepare cols 2,3

	movq	mm5, mm1				; prepare cols 0,1
	psrlw	mm0, 6					; pack to bottom bytes

	psrlw	mm5, 6					; pack to bottom bytes
	jmp	red8T1LoopFinish


; We get here after the last (leftmost) dword was written (Step F)
red8T1Exit:
	ret

ip_reduce8T1LineLong ENDP


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_reduce8T4LineLong.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers

; We will move from right to left within each line of the image.

ip_reduce8T4LineLongStart::
ip_reduce8T4LineLong	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI, \
	pIPState:DWORD, srcAddr:DWORD, srcWidth:DWORD, \
	srcBpl:DWORD, dstAddr:DWORD, \
	dstOffset:DWORD, srcOffset:DWORD, srcLimit:DWORD, \
	tab8to1Permute:DWORD

; NOTE: When I say dword or doubleword, I mean 64-bit thing.  Yeah, this
; isn't what Intel means, they think a 64-bit thing is a quadword.  Tough.

; make the assembler happy
	mov	eax, pIPState

; Load the real parameters.  After we add srcAddr to srcOffset, rSrcAddr
; will point to the left dword of the rightmost doubleword octet in the line.
; This octet may not be complete, in that part of it may be off the image.
; rDstAddr will point to the rightmost doubleword in the destination line.
	mov	eax, tab8to1Permute	; Make the assembler happy
	mov	rSrcAddr0, srcAddr	; pt to 1st source dword

	mov	rSrcBpl, srcBpl		; fill empty slot
	add	rSrcAddr0, srcOffset	; pt to left source dword of
					; first dword octet.
					; We do this because we're really
					; always going to go after the
					; first doubleword in an octet.
					; We jump into the reduction loop
					; at the place that picks up the
					; right number of dwords to form
					; the first octet.

; Get ready to mask off the extra bits from the rightmost dword sample.
; srcWidth is the width of the source in bits.  We divide by 8 and get
; rid of all but the next 3 bits (compute width in bytes MOD 8).
; This number will be used to address the mask table for the rightmost
; words read.
	mov	rScratch, srcWidth	; get source width in bits.
	pxor	mm0, mm0		; clear MMX registers

	and	rScratch, 038h		; = (srcWidth in bytes MOD 8) * 8
	lea	rSrcAddr2, [rSrcAddr0 + 4*rSrcBpl] ; point to src line 4

; Get the mask used to get rid of pad bits.
	movq	mm5, [MMWORD PTR mmxByteEdgeMask + rScratch]
	pxor	mm1, mm1

	mov	rScratch, dstOffset	; get offset to rightmost dest dword
	lea	rSrcAddr3, [rSrcAddr0 + 8*rSrcBpl] ; point to src line 8

	sub	rSrcAddr3, rSrcBpl	; point to line 7
	add	dstAddr, rScratch	; pt to rightmost dest doubleword

; Get the mask used in step 4 above - used after putting nibbles into MS
; half of odd bytes (numbered from the left).
	movq	mm4, [MMWORD PTR all00F0Mask]
	pxor	mm6, mm6		; clear result from words 4,5,6,7

; Here's where we decide how to enter the reducing loop.  The difference
; between srcLimit (offset of rightmost source dword with something in it)
; and srcOffset (offset of leftmost dword of rightmost source dword octet)
; can be 0, 8, 16, 24, 32, 40, 48 or 56.  If it's 0, we enter the loop
; where it reads one source dword before writing the dest.  If it's 8,
; we enter the loop where it reads 2 source dwords and so on.
	mov	rScratch, srcLimit	; get srcLimit
	pxor	mm3, mm3

; The step of rounding up to the next byte boundary is kind of weird.
; You'd think we just add 7 to the width.  Well, say that the width
; is between 8N + 1 and 8N + 7.  In reducing by 8, we only look at 8N bits.
; Therefore, we never want to round up to the next byte!  So skip the
; rounding step and just subtract srcOffset.
	sub	rScratch, srcOffset
	mov	rSrcAddr1, rSrcAddr2	; point to line 4

; Get the offset into the table of pointers to jump entry locations:
	shr	rScratch, 1		; [0, 8, 16, ...] -> [0, 4, 8, ...]
	sub	rSrcAddr1, rSrcBpl	; point to line 3

; We mask every step 1 source halfword with 00C0.  See step 2 above.
; Put this here to avoid address generation interlock (using rScratch as
; an address reg in the cycle after setting it).
	movq	mm7, [MMWORD PTR all0030Mask]
	pxor	mm2, mm2

	; empty
	jmp	red8T4JumpTable[rScratch]


;------------------------- LOOP ENTRIES  -----------------------------

; This is the loop entry taken when we want to read 8 columns of dwords
; to form the first dest dword.
red8T4EntryW7:
	movq	mm3, [rSrcAddr0 + 7*8]			; line 0

	pand	mm3, [rSrcAddr0 + rSrcBpl + 7*8]	; line 1

	pand	mm3, [rSrcAddr0 + 2*rSrcBpl + 7*8]	; line 2

	pand	mm3, [rSrcAddr1 + 7*8]			; line 3

	pand	mm3, [rSrcAddr2 + 7*8]			; line 4

	pand	mm3, [rSrcAddr2 + rSrcBpl + 7*8]	; line 5

	pand	mm3, [rSrcAddr2 + 2*rSrcBpl + 7*8]	; line 6

	pand	mm3, [rSrcAddr3 + 7*8]			; line 7

	pand	mm3, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEntryW6


; This is the loop entry taken when we want to read 7 columns of dwords
; to form the first dest dword.
red8T4EntryW6:
	movq	mm2, [rSrcAddr0 + 6*8]			; line 0

	pand	mm2, [rSrcAddr0 + rSrcBpl + 6*8]	; line 1

	pand	mm2, [rSrcAddr0 + 2*rSrcBpl + 6*8]	; line 2

	pand	mm2, [rSrcAddr1 + 6*8]			; line 3

	pand	mm2, [rSrcAddr2 + 6*8]			; line 4

	pand	mm2, [rSrcAddr2 + rSrcBpl + 6*8]	; line 5

	pand	mm2, [rSrcAddr2 + 2*rSrcBpl + 6*8]	; line 6

	pand	mm2, [rSrcAddr3 + 6*8]			; line 7

	pand	mm2, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEntryW5


; This is the loop entry taken when we want to read 6 columns of dwords
; to form the first dest dword.
red8T4EntryW5:
	movq	mm1, [rSrcAddr0 + 5*8]			; line 0

	pand	mm1, [rSrcAddr0 + rSrcBpl + 5*8]	; line 1

	pand	mm1, [rSrcAddr0 + 2*rSrcBpl + 5*8]	; line 2

	pand	mm1, [rSrcAddr1 + 5*8]			; line 3

	pand	mm1, [rSrcAddr2 + 5*8]			; line 4

	pand	mm1, [rSrcAddr2 + rSrcBpl + 5*8]	; line 5

	pand	mm1, [rSrcAddr2 + 2*rSrcBpl + 5*8]	; line 6

	pand	mm1, [rSrcAddr3 + 5*8]			; line 7

	pand	mm1, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEntryW4


; This is the loop entry taken when we want to read 5 columns of dwords
; to form the first dest dword.
red8T4EntryW4:
	movq	mm0, [rSrcAddr0 + 4*8]			; line 0

	pand	mm0, [rSrcAddr0 + rSrcBpl + 4*8]	; line 1

	pand	mm0, [rSrcAddr0 + 2*rSrcBpl + 4*8]	; line 2

	pand	mm0, [rSrcAddr1 + 4*8]			; line 3

	pand	mm0, [rSrcAddr2 + 4*8]			; line 4

	pand	mm0, [rSrcAddr2 + rSrcBpl + 4*8]	; line 5

	pand	mm0, [rSrcAddr2 + 2*rSrcBpl + 4*8]	; line 6

	pand	mm0, [rSrcAddr3 + 4*8]			; line 7

	pand	mm0, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEntryW3


; This is the loop entry taken when we want to read 4 columns of dwords
; to form the first dest dword.
red8T4EntryW3:
	movq	mm3, [rSrcAddr0 + 3*8]			; line 0

	pand	mm3, [rSrcAddr0 + rSrcBpl + 3*8]	; line 1

	pand	mm3, [rSrcAddr0 + 2*rSrcBpl + 3*8]	; line 2

	pand	mm3, [rSrcAddr1 + 3*8]			; line 3

	pand	mm3, [rSrcAddr2 + 3*8]			; line 4

	pand	mm3, [rSrcAddr2 + rSrcBpl + 3*8]	; line 5

	pand	mm3, [rSrcAddr2 + 2*rSrcBpl + 3*8]	; line 6

	pand	mm3, [rSrcAddr3 + 3*8]			; line 7

	pand	mm3, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEntryW2


; This is the loop entry taken when we want to read 3 columns of dwords
; to form the first dest dword.
red8T4EntryW2:
	movq	mm2, [rSrcAddr0 + 2*8]			; line 0

	pand	mm2, [rSrcAddr0 + rSrcBpl + 2*8]	; line 1

	pand	mm2, [rSrcAddr0 + 2*rSrcBpl + 2*8]	; line 2

	pand	mm2, [rSrcAddr1 + 2*8]			; line 3

	pand	mm2, [rSrcAddr2 + 2*8]			; line 4

	pand	mm2, [rSrcAddr2 + rSrcBpl + 2*8]	; line 5

	pand	mm2, [rSrcAddr2 + 2*rSrcBpl + 2*8]	; line 6

	pand	mm2, [rSrcAddr3 + 2*8]			; line 7

	pand	mm2, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEntryW1


; This is the loop entry taken when we want to read 2 columns of dwords
; to form the first dest dword.
red8T4EntryW1:
	movq	mm1, [rSrcAddr0 + 1*8]			; line 0

	pand	mm1, [rSrcAddr0 + rSrcBpl + 1*8]	; line 1

	pand	mm1, [rSrcAddr0 + 2*rSrcBpl + 1*8]	; line 2

	pand	mm1, [rSrcAddr1 + 1*8]			; line 3

	pand	mm1, [rSrcAddr2 + 1*8]			; line 4

	pand	mm1, [rSrcAddr2 + rSrcBpl + 1*8]	; line 5

	pand	mm1, [rSrcAddr2 + 2*rSrcBpl + 1*8]	; line 6

	pand	mm1, [rSrcAddr3 + 1*8]			; line 7

	pand	mm1, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEntryW0


; This is the loop entry taken when we want to read 1 column of dwords
; to form the first dest dword.
red8T4EntryW0:
	movq	mm0, [rSrcAddr0 + 0*8]			; line 0

	pand	mm0, [rSrcAddr0 + rSrcBpl + 0*8]	; line 1

	pand	mm0, [rSrcAddr0 + 2*rSrcBpl + 0*8]	; line 2

	pand	mm0, [rSrcAddr1 + 0*8]			; line 3

	pand	mm0, [rSrcAddr2 + 0*8]			; line 4

	pand	mm0, [rSrcAddr2 + rSrcBpl + 0*8]	; line 5

	pand	mm0, [rSrcAddr2 + 2*rSrcBpl + 0*8]	; line 6

	pand	mm0, [rSrcAddr3 + 0*8]			; line 7

	pand	mm0, mm5
	pcmpeqb	mm5, mm5				; prepare for compare

	; empty
	jmp	red8T4LoopEnd


; ------------------------- MAIN LOOP ------------------------------

; This is the main loop for 8x, threshold 1 reduction.
;
; The basic loop structure is:
;  A. Read word 7 of a new dword octet and consolidate data from
;     col 0 of the dword octet to the right.
;	mm0 = raw col0 data from dword octet to the right
;	mm1 = consolidated col1 data from dword octet to the right
;	mm2 = col 2,3 data from dword octet to the right
;	mm3 = accumulating new raw col 7 data
;	mm5 = used to accumulate col0 data
;	mm6 = col 4,5,6,7 data from dword octet to the right
;  B. Combine the consolidated col 0 data with that for col 1.
;  C. Combine the col 0,1 dword with col 2,3 dword.
;  D. Combine col 0,1,2,3 dword with col 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
;  E. Write the result out to [dstAddr] and decrement dstAddr by 8.
;  F. If that was the last dest word, quit.
;  G. Read word 6 of this dword octet and consolidate data from
;     col 7 of this dword octet.
;	mm2 = accumulating raw data from col 6
;	mm3 = raw col 7 data
;	mm5 = used to accumulate col 7 data
;  H. Read word 5 of this dword octet and consolidate data from
;     col 6 of this dword octet.
;	mm1 = accumulating raw col 5 data
;	mm2 = raw col 6 data
;	mm3 = consolidated col7 data
;	mm5 = used to consolidate col 6 data
;  I. Combine consolidated col7 and consolidated col6 data.
;  J. Read word 4 of this dword octet and consolidate data from
;     col 5 of this dword octet.
;	mm0 = accumulating raw data from col 4
;	mm1 = raw col 5 data
;	mm2 = col 6,7 data
;	mm5 = used to accumulate col 5 data
;  K. Read word 3 of this dword octet and consolidate data from
;     col 4 of this dword octet.
;	mm0 = raw col4 data
;	mm1 = consolidated col 5 data
;	mm2 = col 6,7 data
;	mm3 = used to accumulate raw col3 data
;	mm5 = used to consolidate col 4 data
;  L. Combine consolidated col5 and consolidated col4 data.
;  M. Combine consolidated col 4,5 and consolidated col 6,7 data.
;  N. Read word 2 of this dword octet and consolidate data from
;     col 3 of this dword octet.
;	mm2 = accumulating raw data from col 2
;	mm3 = raw col 3 data
;	mm5 = used to accumulate col 3 data
;	mm6 = col 4,5,6,7 data
;  O. Read word 1 of this dword octet and consolidate data from
;     col 2 of this dword octet.
;	mm1 = accumulating raw col 1 data
;	mm2 = raw col 2 data
;	mm3 = consolidated col3 data
;	mm5 = used to consolidate col 2 data
;	mm6 = col 4,5,6,7 data
;  P. Combine consolidated col 2 and consolidated col 3 data.
;  Q. Read word 0 of this dword octet and consolidate data from
;     col 1 of this dword octet.
;	mm0 = accumulating raw data from col 0
;	mm1 = raw col 1 data
;	mm2 = col 2,3 data
;	mm5 = used to accumulate col 1 data
;	mm6 = col 4,5,6,7 data
;  R. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
;  S. Since that was the last dword octet, do the consolidate of col 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
;


; Here's the real loop code.  The comments above have been copied
; to the top of each section.
; In all sections:
;	mm4 = mask containing 00F0 00F0 00F0 00F0
;	mm5 = usually contains all 1's, sometimes used to shift stuff, then
;	      restored to all 1's.
;	mm7 = mask containing 0300 0300 0300 0300

;  A. Read word 7 of a new dword octet and consolidate data from
;     col 0 of the dword octet to the right.
;	mm0 = raw col0 data from dword octet to the right
;	mm1 = consolidated col1 data from dword octet to the right
;	mm2 = col 2,3 data from dword octet to the right
;	mm3 = accumulating new raw col 7 data
;	mm5 = used to accumulate col0 data
;	mm6 = col 4,5,6,7 data from dword octet to the right
red8T4LoopEntryW7:
	movq	mm3, [rSrcAddr0 + 7*8]			; line 0
	pcmpeqb	mm0, mm5				; all 1's?

	pand	mm3, [rSrcAddr0 + rSrcBpl + 7*8]	; line 1
	psrlq	mm0, 3					; res in odd bytes

	pand	mm3, [rSrcAddr0 + 2*rSrcBpl + 7*8]	; line 2
	pand	mm0, mm7				; select col 0 bits

;  B. Combine the consolidated col 0 data with that for col 1.
	pand	mm3, [rSrcAddr1 + 7*8]			; line 3
	packuswb mm1, mm0				; col 0 bits on the left

;  C. Combine the col 0,1 dword with col 2,3 dword.
	pand	mm3, [rSrcAddr2 + 7*8]			; line 4
	movq	mm0, mm2				; prepare cols 2,3

	pand	mm3, [rSrcAddr2 + rSrcBpl + 7*8]	; line 5
	movq	mm5, mm1				; prepare cols 0,1

	pand	mm3, [rSrcAddr2 + 2*rSrcBpl + 7*8]	; line 6
	psrlw	mm0, 6					; pack to bottom bytes

	pand	mm3, [rSrcAddr3 + 7*8]			; line 7
	psrlw	mm5, 6					; pack to bottom bytes

; This is where we jump to when we've just finished consolidating the
; leftmost source dword in the scan line (step S).
red8T4LoopFinish:
	por	mm2, mm0				; combine col 2,3 data
	por	mm1, mm5				; combine col 0,1 data

	pand	mm2, mm4				; mask off extra bits
	pand	mm1, mm4				; mask off extra bits

;  Step C ends with packuswb below.
;  D. Combine col 0,1,2,3 dword with col 4,5,6,7 dword after
;     nibble-swapping both (see step 6 above).
	packuswb mm2, mm1				; pack col 0,1,2,3 data
	movq	mm0, mm6				; get ready to align
							; col 4,5,6,7 data

	movq	mm5, mm2				; get ready to align
							; col 0,1,2,3 data
	psrlw	mm0, 12					; pack and "wordswap"
							; results from cols
							; 4,5,6,7.

	psrlw	mm5, 12					; pack and "wordswap"
							; results from cols
							; 0,1,2,3
	por	mm6, mm0				; combine results

	pand	mm6, [MMWORD PTR all00FFMask]		; get bytes we want
	por	mm2, mm5				; combine 0123 results

	pand	mm2, [MMWORD PTR all00FFMask]		; get bytes we want
	; empty

; Note that we intentionally put the mm6 bits (cols 4,5,6,7) in the UPPER
; half of the result dword.  This accounts for the wordswapping inherent
; in this little endian platform.
	packuswb mm2, mm6				; get inverted dest dw
	mov	rScratch, dstAddr			; load dest address

; Since the way we generated the result bit from each 8x8 square produced
; a 1 if ALL the bits in the square are on, we don't need to invert the
; result, so don't do it.
	; empty
	sub	rScratch, 8				; point to prev word

; wait for mm2 and rScratch to percolate through pipeline
	mov	dstAddr, rScratch			; pt to prev src word
	; empty

;  E. Write the result out to [dstAddr] and decrement dstAddr by 8.
	movq	[rScratch + 8], mm2			; write result dword
	pcmpeqb	mm5, mm5				; get ready to compare

;  F. If that was the last dest word, quit.
	cmp	rSrcAddr0, srcAddr
	jl	red8T4Exit		; quit if that was last dest word


;  G. Read word 6 of the new dword octet and consolidate data from
;     col 7 of this dword octet.
;	mm2 = accumulating raw data from col 6
;	mm3 = raw col 7 data
;	mm5 = used to accumulate col 7 data
red8T4LoopEntryW6:
	movq	mm2, [rSrcAddr0 + 6*8]			; line 0
	pcmpeqb	mm3, mm5				; all 1's?

	pand	mm2, [rSrcAddr0 + rSrcBpl + 6*8]	; line 1
	psrlq	mm3, 3					; res in odd bytes

	pand	mm2, [rSrcAddr0 + 2*rSrcBpl + 6*8]	; line 2
	pand	mm3, mm7				; select col 7 bits

	pand	mm2, [rSrcAddr1 + 6*8]			; line 3

	pand	mm2, [rSrcAddr2 + 6*8]			; line 4

	pand	mm2, [rSrcAddr2 + rSrcBpl + 6*8]	; line 5

	pand	mm2, [rSrcAddr2 + 2*rSrcBpl + 6*8]	; line 6

	pand	mm2, [rSrcAddr3 + 6*8]			; line 7


;  H. Read word 5 of the new dword octet and consolidate data from
;     col 6 of this dword octet.
;	mm1 = accumulating raw col 5 data
;	mm2 = raw col 6 data
;	mm3 = consolidated col7 data
;	mm5 = used to consolidate col 6 data
red8T4LoopEntryW5:
	movq	mm1, [rSrcAddr0 + 5*8]			; line 0
	pcmpeqb	mm2, mm5				; all 1's?

	pand	mm1, [rSrcAddr0 + rSrcBpl + 5*8]	; line 1
	psrlq	mm2, 3					; res in odd bytes

	pand	mm1, [rSrcAddr0 + 2*rSrcBpl + 5*8]	; line 2
	pand	mm2, mm7				; select col 6 bits

;  I. Combine consolidated col7 and consolidated col6 data.
	pand	mm1, [rSrcAddr1 + 5*8]			; line 3
	packuswb mm3, mm2				; col 6 bits on the left

	pand	mm1, [rSrcAddr2 + 5*8]			; line 4
	movq	mm2, mm3				; where we want it

	pand	mm1, [rSrcAddr2 + rSrcBpl + 5*8]	; line 5

	pand	mm1, [rSrcAddr2 + 2*rSrcBpl + 5*8]	; line 6

	pand	mm1, [rSrcAddr3 + 5*8]			; line 7


;  J. Read word 4 of the new dword octet and consolidate data from
;     col 5 of this dword octet.
;	mm0 = accumulating raw data from col 4
;	mm1 = raw col 5 data
;	mm2 = col 6,7 data
;	mm5 = used to accumulate col 5 data
red8T4LoopEntryW4:
	movq	mm0, [rSrcAddr0 + 4*8]			; line 0
	pcmpeqb	mm1, mm5				; all 1's?

	pand	mm0, [rSrcAddr0 + rSrcBpl + 4*8]	; line 1
	psrlq	mm1, 3					; res in odd bytes

	pand	mm0, [rSrcAddr0 + 2*rSrcBpl + 4*8]	; line 2
	pand	mm1, mm7				; select col 5 bits

	pand	mm0, [rSrcAddr1 + 4*8]			; line 3

	pand	mm0, [rSrcAddr2 + 4*8]			; line 4

	pand	mm0, [rSrcAddr2 + rSrcBpl + 4*8]	; line 5

	pand	mm0, [rSrcAddr2 + 2*rSrcBpl + 4*8]	; line 6

	pand	mm0, [rSrcAddr3 + 4*8]			; line 7


;  K. Read word 3 of the new dword octet and consolidate data from
;     col 4 of this dword octet.
;	mm0 = raw col4 data
;	mm1 = consolidated col 5 data
;	mm2 = col 6,7 data
;	mm3 = used to accumulate raw col3 data
;	mm5 = used to consolidate col 4 data
red8T4LoopEntryW3:
	movq	mm3, [rSrcAddr0 + 3*8]			; line 0
	pcmpeqb	mm0, mm5				; all 1's?

	pand	mm3, [rSrcAddr0 + rSrcBpl + 3*8]	; line 1
	psrlq	mm0, 3					; res in odd bytes

	pand	mm3, [rSrcAddr0 + 2*rSrcBpl + 3*8]	; line 2
	pand	mm0, mm7				; select col 4 bits

;  L. Combine consolidated col5 and consolidated col4 data.
	pand	mm3, [rSrcAddr1 + 3*8]			; line 3
	packuswb mm1, mm0				; col 4 bits on the left

;  M. Combine consolidated col 4,5 and consolidated col 6,7 data.
	pand	mm3, [rSrcAddr2 + 3*8]			; line 4
	movq	mm6, mm2				; prepare cols 6,7

	pand	mm3, [rSrcAddr2 + rSrcBpl + 3*8]	; line 5
	movq	mm5, mm1				; prepare cols 4,5

	pand	mm3, [rSrcAddr2 + 2*rSrcBpl + 3*8]	; line 6
	psrlw	mm6, 6					; pack to bottom bytes

	pand	mm3, [rSrcAddr3 + 3*8]			; line 7
	psrlw	mm5, 6					; pack to bottom bytes

	por	mm6, mm2				; combine col 6,7 data
	por	mm1, mm5				; combine col 4,5 data

	pand	mm6, mm4				; mask off extra bits
	pand	mm1, mm4				; mask off extra bits

	packuswb mm6, mm1				; get col 4,5,6,7 data
	pcmpeqb	mm5, mm5				; prepare for compare

;  N. Read word 2 of the new dword octet and consolidate data from
;     col 3 of this dword octet.
;	mm2 = accumulating raw data from col 2
;	mm3 = raw col 3 data
;	mm5 = used to accumulate col 3 data
;	mm6 = col 4,5,6,7 data
red8T4LoopEntryW2:
	movq	mm2, [rSrcAddr0 + 2*8]			; line 0
	pcmpeqb	mm3, mm5				; all 1's?

	pand	mm2, [rSrcAddr0 + rSrcBpl + 2*8]	; line 1
	psrlq	mm3, 3					; res in odd bytes

	pand	mm2, [rSrcAddr0 + 2*rSrcBpl + 2*8]	; line 2
	pand	mm3, mm7				; select col 3 bits

	pand	mm2, [rSrcAddr1 + 2*8]			; line 3

	pand	mm2, [rSrcAddr2 + 2*8]			; line 4

	pand	mm2, [rSrcAddr2 + rSrcBpl + 2*8]	; line 5

	pand	mm2, [rSrcAddr2 + 2*rSrcBpl + 2*8]	; line 6

	pand	mm2, [rSrcAddr3 + 2*8]			; line 7

;  O. Read word 1 of the new dword octet and consolidate data from
;     col 2 of this dword octet.
;	mm1 = accumulating raw col 1 data
;	mm2 = raw col 2 data
;	mm3 = consolidated col3 data
;	mm5 = used to consolidate col 2 data
;	mm6 = col 4,5,6,7 data
red8T4LoopEntryW1:
	movq	mm1, [rSrcAddr0 + 1*8]			; line 0
	pcmpeqb	mm2, mm5				; all 1's?

	pand	mm1, [rSrcAddr0 + rSrcBpl + 1*8]	; line 1
	psrlq	mm2, 3					; res in odd bytes

	pand	mm1, [rSrcAddr0 + 2*rSrcBpl + 1*8]	; line 2
	pand	mm2, mm7				; select col 2 bits

;  P. Combine consolidated col 2 and consolidated col 3 data.
	pand	mm1, [rSrcAddr1 + 1*8]			; line 3
	packuswb mm3, mm2				; col 2 bits on the left

	pand	mm1, [rSrcAddr2 + 1*8]			; line 4
	movq	mm2, mm3				; where we want it

	pand	mm1, [rSrcAddr2 + rSrcBpl + 1*8]	; line 5

	pand	mm1, [rSrcAddr2 + 2*rSrcBpl + 1*8]	; line 6

	pand	mm1, [rSrcAddr3 + 1*8]			; line 7

;  Q. Read word 0 of the new dword octet and consolidate data from
;     col 1 of this dword octet.
;	mm0 = accumulating raw data from col 0
;	mm1 = raw col 1 data
;	mm2 = col 2,3 data
;	mm5 = used to accumulate col 1 data
;	mm6 = col 4,5,6,7 data
red8T4LoopEntryW0:
	movq	mm0, [rSrcAddr0 + 0*8]			; line 0
	pcmpeqb	mm1, mm5				; all 1's?

	pand	mm0, [rSrcAddr0 + rSrcBpl + 0*8]	; line 1
	psrlq	mm1, 3					; res in odd bytes

	pand	mm0, [rSrcAddr0 + 2*rSrcBpl + 0*8]	; line 2
	pand	mm1, mm7				; select col 1 bits

	pand	mm0, [rSrcAddr1 + 0*8]			; line 3

	pand	mm0, [rSrcAddr2 + 0*8]			; line 4

	pand	mm0, [rSrcAddr2 + rSrcBpl + 0*8]	; line 5

	pand	mm0, [rSrcAddr2 + 2*rSrcBpl + 0*8]	; line 6

	pand	mm0, [rSrcAddr3 + 0*8]			; line 7

;  R. Decrement address registers.  If there's another dword octet to the
;     right, loop back to step A.
;     Note that we couldn't do this in the empty slots above since
;     you can't pair an integer opcode with an MMX opcode that references
;     memory or integer registers.  Bleah!
red8T4LoopEnd:
	sub	rSrcAddr3, 64			; decrement by size of octet
	sub	rSrcAddr2, 64			; which is 8 dwords across
						; times 8 bytes/dword

	sub	rSrcAddr1, 64
	sub	rSrcAddr0, 64

	cmp	rSrcAddr0, srcAddr		; done all src words yet?
	jge	red8T4LoopEntryW7		; nope, do another octet


;  S. Since that was the last dword octet, do the consolidate of col 0
;     data normally done in step A and jump to step B.  We'll exit from
;     step F.
	pcmpeqb	mm0, mm5				; all 1's?

	psrlq	mm0, 3					; res in odd bytes

	pand	mm0, mm7				; select col 0 bits

;  B'. Combine the consolidated col 0 data with that for col 1.
	packuswb mm1, mm0				; mm0 bits on the left
;  C'. Combine the col 0,1 dword with col 2,3 dword.
	movq	mm0, mm2				; prepare cols 2,3

	movq	mm5, mm1				; prepare cols 0,1
	psrlw	mm0, 6					; pack to bottom bytes

	psrlw	mm5, 6					; pack to bottom bytes
	jmp	red8T4LoopFinish

; We get here after the last (leftmost) dword was written (Step F)
red8T4Exit:
	ret

ip_reduce8T4LineLong ENDP



; Segment for initialized variables, like masks
; We don't need a mask for every horizontal consolidation step since
; sometimes we're ORing into stuff that was already 0.
_DATA	SEGMENT PUBLIC 'DATA'

; Jump tables for 8x reduction.  I originally tried to follow
; the example in the MASM Programmer's Guide and use a weird type for
; these things, but DWORD seems to work fine..
red8T1JumpTable	DWORD	red8T1EntryW0,
			red8T1EntryW1,
			red8T1EntryW2,
			red8T1EntryW3,
			red8T1EntryW4,
			red8T1EntryW5,
			red8T1EntryW6,
			red8T1EntryW7

red8T4JumpTable	DWORD	red8T4EntryW0,
			red8T4EntryW1,
			red8T4EntryW2,
			red8T4EntryW3,
			red8T4EntryW4,
			red8T4EntryW5,
			red8T4EntryW6,
			red8T4EntryW7

; masks used for 4x reduction
		ALIGN	8
all1sMask	QWORD	01111111111111111h
all7sMask	QWORD	07777777777777777h
all8sMask	QWORD	08888888888888888h
allC0Mask	QWORD	0C0C0C0C0C0C0C0C0h
all38sMask	QWORD	03838383838383838h

; masks used for 8x reduction
all30Mask	QWORD	03030303030303030h
all0030Mask	QWORD	00030003000300030h
allF0Mask	QWORD	0F0F0F0F0F0F0F0F0h
all00F0Mask	QWORD	000F000F000F000F0h
all00FFMask	QWORD	000FF00FF00FF00FFh


_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no initialized variables
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcredmn.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.2  1998/01/13 23:58:34  ddavies
pcredmn.asm -> /product/ipcore/ipshared/src/RCS/pcredmn.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.1  1996/10/24 19:19:29  ddavies
Initial revision



^

ENDIF





