/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * pcscgarb:	Scale-to-gray, arbitrary downsampling, PC version
 *		which makes and calls on-the-fly-generated code
 *
 * Exports:
 *	Int32	    ip_scaleToGrayArb4bppLine
 *	Int32	    ip_scaleToGrayArbLineTabBytes
 *	Int32	    ip_makeScaleToGrayArbLineTab
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"
#include "memory.pub"

	/* prototypes */
#include "shrscgr.pub"
#include "shrscgr.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: pcscgarb.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $");

/* Labels in the code */
extern Int8 sgStart, sgClearOutputWord, sgPixelStart, sgPixelStartAddr,
	    sgAccum1, sgAccum1Source, sgAccum1Mask,
	    sgAccum1Shift, sgAccum1Filter,
	    sgAccum1Pair, sgAccum1PairSource1, sgAccum1PairSource2,
	    sgAccum1PairMask1, sgAccum1PairMask2,
	    sgAccum1PairShift1, sgAccum1PairShift2,
	    sgAccum1PairFilter1, sgAccum1PairFilter2,
	    sgGotoBottom, sgGotoBottomAddr, sgAccum2, sgAccum2Source1,
	    sgAccum2Source2, sgAccum2Mask, sgAccum2Filter,
	    sgSavePixel, sgStoreWord, sgStoreWordAddr, sgStoreWordTableLimit,
	    sgReturn;

/* Lengths of routines */
extern UInt32	sgStartLen,sgPixelStartLen, sgGotoBottomLen,
		sgAccum1Len, sgAccum2Len, sgAccum1PairLen, sgSavePixelLen, 
		sgStoreWordLen, sgReturnLen;

#define SET_CODE_LONG(ptr, base, offset, value) \
    (*((UInt32 *) (ptr + ((UInt32) &offset - (UInt32) &base - 4))) = value)

#define SET_CODE_BYTE(ptr, base, offset, value) \
    (*((UInt8 *) (ptr + ((UInt32) &offset - (UInt32) &base - 1))) = value)

#define GET_LABEL_ADDR(ptr, base, offset) \
    ((UInt32 *) (ptr + ((UInt32) &offset - (UInt32) &base - 4)))

/* Forward refs */
static void CDECL 
ip_genScgPrologue (
    IPState    *pIPState,
    UInt8     **codePtr);

static Int32 * CDECL 
ip_genScgPixelStart (
    IPState    *pIPState,
    UInt8     **codePtr);

static void CDECL 
ip_genScgAccum1 (
    IPState    *pIPState,
    UInt8     **codePtr,
    Int32	source,
    Int32	filterOffset,
    Int32	mask);

static void CDECL 
ip_genScgAccum1Pair (
    IPState    *pIPState,
    UInt8     **codePtr,
    Int32	source1,
    Int32	source2,
    Int32	filterOffset1,
    Int32	filterOffset2,
    UInt32	mask);

static void CDECL 
ip_genScgAccum2 (
    IPState    *pIPState,
    UInt8     **codePtr,
    Int32	source,
    Int32	filterOffset,
    Int32	mask);

static Int32 * CDECL 
ip_genScgGotoBottom (
    IPState    *pIPState,
    UInt8     **codePtr);

static void CDECL 
ip_genScgSavePixel (
    IPState    *pIPState,
    UInt8     **codePtr);

static Int32 * CDECL 
ip_genScgStoreWord (
    IPState    *pIPState,
    UInt8     **codePtr,
    UInt32	tableLimit);

static void CDECL 
ip_genScgEpilogue (
    IPState    *pIPState,
    UInt8     **codePtr);

static void CDECL 
ip_patchJump (
    IPState    *pIPState,
    Int32      *patchAddr,
    UInt8      *destAddr);

typedef void (CDECL *tLineCode) (
    IPState    *pIPState,
    UInt32     *dest,
    UInt32     *src,
    Int32      *filterRow);

/***********************************************************************
 * ip_scaleToGrayArb4bppLine
 *
 *	Inner loop of scale-to-gray. See the comments in i_scgarb.c
 *	to find out what's going on.
 *
 * Arguments:
 *	destP	    -	    Destination image
 *	sourceP	    -	    Source image
 *	sWpl	    -	    32-bit word distance between lines
 *	filterHeight-	    Height of filter
 *	wD	    -	    Number of output pixels to generate
 *	pLineTab    -	    The line table, which actually contains the code
 *	filterRow   -	    Pointer to the current filter row
 *	filterNextRow-	    Longword offset for next filter row
 *
 * Returns:
 *	nothing
 ***********************************************************************/

void CDECL 
ip_scaleToGrayArb4bppLine (
    IPState    *pIPState,
    UInt32     *destP,
    UInt32     *sourceP,
    Int32	sWpl,
    Int32	filterWidth,
    Int32	filterHeight,
    Int32	wD,
    void       *pLineTab,
    Int32      *filterRow,
    Int32	filterNextRow)
{
    PROLOGUE;
    Int32 paddedWidth;
    LineCode codePtr;
    SGLineTab *lineTab = pLineTab;
    UInt32 padBits;
    paddedWidth = (wD + 7) & ~7;
	/* In Visual C++, the compiler complains about casting data
	 * pointers to function pointers.  That's exactly what we
	 * want to do here, so turn off the warning.
	 */
#ifdef _MSC_VER
#pragma warning ( disable : 4055 )
#endif
    codePtr = (LineCode) ((UInt8 *) pLineTab + (paddedWidth * sizeof(SGLineTab)));
#ifdef _MSC_VER
#pragma warning ( default : 4055 )
#endif

    /* Shut the compiler up */
    pIPState = pIPState;
    __error__ = __error__;
    sWpl = sWpl;
    filterHeight = filterHeight;
    wD = wD;
    filterNextRow = filterNextRow;
    filterWidth = filterWidth;

    /* Call that code! Blammo! */
    (*codePtr)(destP, sourceP, filterRow, lineTab);

    /* Crush the pad in the last output word, if needed */
    if (wD & 7)
    {
	padBits = (1 << (4 * (8 - (wD & 7)))) - 1;
	destP[wD/8] &= ~padBits;
    }
}

/***********************************************************************
 *
 * ip_scaleToGrayArbLineTabBytes
 *
 *	Computes the size, in bytes, of block that must be allocated for
 *	i_makeScaleToGray4bppArbLineTab. (It's done this way to support
 *	applications like the DAE which must do their own allocation)
 *	This version figures out how much code it's going to generate.
 * 
 *	filter	-	    The filter being used, created by above
 *      xD      -           X-offset into "virtual" image for dest image
 *	wD	-	    width (or height) of dest image
 *      wVirt   -           Width of "virtual" image
 *	wS	-	    Width of source rectangle
 *	sBpl	-	    inter-line distance, in bytes, in source image
 *
 * Returns:
 *	number of bytes, 0 = error 
 *
 ***********************************************************************/

Int32 CDECL 
ip_scaleToGrayArbLineTabBytes (
    IPState    *pIPState,
    void       *filter,
    Int32	xD,
    Int32	wD,
    Int32	wVirt,
    Int32	wS,
    Int32	sBpl)
{
    PROLOGUE;
    Int32	 codeCount, pixelCount;
    FilterDesc	*fDesc = (FilterDesc *) filter;

    /* Shut the compiler up */
    pIPState = pIPState;
    __error__ = __error__;
    sBpl = sBpl;
    wS = wVirt = xD;

    /* First, we get the line table itself. */
    codeCount = (wD + ((8 - (wD & 7)) & 7)) * sizeof (SGLineTab);

    /* Generate the procedure entry */
    codeCount += sgStartLen ;

    /* Count for a pixel */
    pixelCount = sgPixelStart;
    if (fDesc->height / 2)
	pixelCount += (fDesc->height / 2) * sgAccum1PairLen;
    if (fDesc->height & 1)
	pixelCount += sgAccum1Len;
    pixelCount += sgGotoBottomLen;
    pixelCount += fDesc->height * sgAccum2Len;

    /* We unroll an output word pixel's worth along with a store */
    codeCount += 8 * pixelCount + sgStoreWordLen;

    /* And then, the code is done */
    codeCount += sgReturnLen;

    return codeCount;
}

/***********************************************************************
 *
 * ip_makeScaleToGrayArbLineTab
 *
 *	Generate some assembly code for the scale-to-gray inner loop
 *	This replaces the "line table" in the "C" version with code
 *	that does the same thing. The file "scgtemp.asm" contains
 *	code templates, which will be assembled and linked. This
 *	routine copies pieces of that code out and modifies some of the
 *	values. Since we're not writing an assembler, the way the
 *	values get modified is a bit hokey: we put a label AFTER the
 *	instruction with the value in question, and since the value
 *	is ALWAYS (in this case--make sure that's still true if you
 *	add new instructions) at the end of the instruction, we move
 *	back the right amount, either 1 byte or 4 bytes, and store
 *	the value.
 *
 * ---
 *	The speed-up on a 66mhz DX2 over the "C" code wasn't all that 
 *	exciting, maybe 1.5x. Adding up cycle times showed that it was
 *	a lot slower than it "should" have been, the reason being that
 *	it was all in-line, and therefore not cached, and on a machine
 *	with such a slow (33mhz) external bus, the time to load all
 *	of the instructions overwhelmed everything else (duh). 
 *	So, I'm rewriting this to generate a loop for each output word, 
 *	with inline code for each pixel in an output word. This requires
 *	that we pass in shift counts and source buffer offsets as a table.
 *	
 * Arguments:
 *	filter  -	filter LUT returned by i_makeScaleToGrayArbFilter
 *      xD      -       X-offset into "virtual" image for dest image (no frame)
 *	wD	-	width (or height) of dest image
 *      wVirt   -       Width of "virtual" image
 *	wS	-	Width of source rectangle
 *	sBpl	-	Source image distance between lines, in bytes
 *	lineTab	-	Line table, which in this case is the line table
 *			followed by some code which we generate to
 *			compute the line.
 *			This should be allocated by the caller, using
 *			a length computed by i_makeScaleToGrayArbLineTabBytes
 * Returns:
 *	ia_successful or error code
 ***********************************************************************/

Int32 CDECL 
ip_makeScaleToGrayArbLineTab (
    IPState    *pIPState,
    void       *filter,
    Int32	xD,
    Int32	wD,
    Int32	wVirt,
    Int32	wS,
    Int32	sBpl,
    void       *lTab)
{
    PROLOGUE;
    UInt8	*codePtr, *codeBase;
    SGLineTab	*lineTab = (SGLineTab *) lTab;
    Int32	 i, filterOffset, startPixel, startBit, startWord, shift1,
		 row, shift2, prevStartWord, paddedWidth;
    Float64	 sourcePixel, sourceInc;
    UInt32	 mask, filterWidth, filterNextRow;
    Int32	*jump2WordAddress, *jumpDoneAddress, *jumpTopAddress;
    UInt8	*topCodePtr;
    FilterDesc	*fDesc = (FilterDesc *) filter;

    /* filterOffset is the number of source pixels from the center
       to the left edge of the filter */
    filterOffset = fDesc->width / 2;
    filterWidth = fDesc->width;
    sourceInc = (Float64) wS / (Float64) wVirt;
    sourcePixel = xD * sourceInc + cFrameBits;
    filterNextRow = 1 << filterWidth;
    mask = filterNextRow - 1;
    paddedWidth = (wD + 7) & ~7;
    codePtr = (UInt8 *) lTab + paddedWidth * sizeof(SGLineTab);
    codeBase = codePtr;

    /* Generate the line table. We generate up to 7 extra entries
       for the pad pixels, which we'll process and then mask out */
    prevStartWord = 0;
    for (i = 0; i < paddedWidth; i++)
    {
	startPixel = (Int32) ((sourcePixel - filterOffset) + 0.01);
	startBit = 32 - (startPixel & 31);
	startWord = startPixel >> 5;

	/* The offset is from the previous one, so we can use an
	   incrementing register */
	lineTab[i].offset = (startWord - prevStartWord) * 4;
	prevStartWord = startWord;

	/* Does it straddle a word? */
	if (startBit >= fDesc->width)
	{
	    shift1 = startBit - fDesc->width;
	    lineTab[i].shift.packedShift = shift1;
	}
	else
	{
	    /* Straddles a word, so make two shifts/masks. The first word
	       is shifted left, the second is shifted right*/
	    shift1 = fDesc->width - startBit;
	    shift2 = 32 - (fDesc->width - startBit);
	    lineTab[i].shift.packedShift  = (shift2 << 24) | shift1;
	}

	sourcePixel += sourceInc;
    }

    /* Now, generate the code. */

    /* Generate the procedure entry */
    ip_genScgPrologue (pIPState, &codePtr);

    /* Generate code for 8 pixels, unrolled, followed by a storeOutputWord,
       which also updates the loop count (which is the table pointer)
       and loops */

    /* Save this address; we'll need to jump to it at the bottom */
    topCodePtr = codePtr;

    for (i = 0; i < 8; i++)
    {
	/* Clear the accumulator and branch to the 2-word case if needed */
	jump2WordAddress = ip_genScgPixelStart (pIPState, &codePtr);

	/* One-word case. Generate pairs first */
	for (row = 0; row < (fDesc->height & (~1)); row += 2)
	{
	    ip_genScgAccum1Pair (pIPState, &codePtr,
				 row * sBpl, (row + 1) * sBpl,
				 row * filterNextRow * 4,
				 (row + 1) * filterNextRow * 4,
				 mask);
	}

	/* Then generate the rest */
	for (; row < fDesc->height; row++)
	{
	    ip_genScgAccum1 (pIPState, &codePtr, row * sBpl, 
			     row * filterNextRow * 4, mask);
	}

	/* Generate the branch around the 2-pixel case */
	jumpDoneAddress = ip_genScgGotoBottom (pIPState, &codePtr);

	/* Patch the jump for 2 word case to here */
	ip_patchJump (pIPState, jump2WordAddress, codePtr);

	/* Generate the 2-word case */
	for (row = 0; row < fDesc->height; row++)
	{
	    ip_genScgAccum2 (pIPState, &codePtr, row * sBpl, 
			     row * filterNextRow * 4, mask);
	}

	/* Patch the jump to here from the one-pixel case */
	ip_patchJump (pIPState, jumpDoneAddress, codePtr);

	/* Save that pixel into the output word */
	ip_genScgSavePixel (pIPState, &codePtr);
    }


    /* Write that output word, and loop */
    jumpTopAddress = ip_genScgStoreWord (pIPState, &codePtr, (UInt32) codeBase);

    /* Patch the jump back up to the top */
    ip_patchJump (pIPState, jumpTopAddress, topCodePtr);

    /* Generate the epilogue */
    ip_genScgEpilogue (pIPState, &codePtr);

    /* Bitchen */
    NORMAL_RETURN;
}


/* 
 * Code generators--they just copy code into the buffer. Since
 * we compute the buffer size, we're (allegedly) guaranteed that it
 * doesn't overflow. Look at scgtemp.asm to see what code gets generated.
 *
 * When one returns a "patchAddr", it's the address of a jump inside the
 * code fragment that needs to be patched. (Actually, it's the address
 * of the jump offset, which is at the end of the instruction). 
 */

static void CDECL 
ip_genScgPrologue (
    IPState    *pIPState,
    UInt8     **codePtr)
{
    PROLOGUE;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgStart, sgStartLen);
    *codePtr += sgStartLen;
}

static Int32 * CDECL 
ip_genScgPixelStart (
    IPState    *pIPState,
    UInt8     **codePtr)
{
    PROLOGUE;
    Int32 *patchAddr;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgPixelStart, sgPixelStartLen);

    patchAddr = (Int32 *) GET_LABEL_ADDR (*codePtr, sgPixelStart, sgPixelStartAddr);

    *codePtr += sgPixelStartLen;
    return patchAddr;
}

static void CDECL 
ip_genScgAccum1 (
    IPState    *pIPState,
    UInt8     **codePtr,
    Int32	source,
    Int32	filterOffset,
    Int32	mask)
{
    PROLOGUE;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgAccum1, sgAccum1Len);

    SET_CODE_LONG(*codePtr, sgAccum1, sgAccum1Source, source);
    SET_CODE_LONG(*codePtr, sgAccum1, sgAccum1Filter, filterOffset);
    SET_CODE_LONG(*codePtr, sgAccum1, sgAccum1Mask, mask);
	
   *codePtr += sgAccum1Len;
}

static void CDECL 
ip_genScgAccum1Pair (
    IPState    *pIPState,
    UInt8     **codePtr,
    Int32	source1,
    Int32	source2,
    Int32	filterOffset1,
    Int32	filterOffset2,
    UInt32	mask)

{
    PROLOGUE;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgAccum1Pair, sgAccum1PairLen);

    SET_CODE_LONG(*codePtr, sgAccum1Pair, sgAccum1PairSource1, source1);
    SET_CODE_LONG(*codePtr, sgAccum1Pair, sgAccum1PairSource2, source2);
    SET_CODE_LONG(*codePtr, sgAccum1Pair, sgAccum1PairFilter1, filterOffset1);
    SET_CODE_LONG(*codePtr, sgAccum1Pair, sgAccum1PairFilter2, filterOffset2);
    SET_CODE_LONG(*codePtr, sgAccum1Pair, sgAccum1PairMask1, mask);
    SET_CODE_LONG(*codePtr, sgAccum1Pair, sgAccum1PairMask2, mask);
	
   *codePtr += sgAccum1PairLen;
}

static Int32 * CDECL 
ip_genScgGotoBottom (
    IPState    *pIPState,
    UInt8     **codePtr)
{
    PROLOGUE;
    Int32 *patchAddr;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgGotoBottom, sgGotoBottomLen);
    patchAddr = (Int32 *) GET_LABEL_ADDR (*codePtr, sgGotoBottom, sgGotoBottomAddr);

   *codePtr += sgGotoBottomLen;
    return patchAddr;
}

static void CDECL 
ip_genScgAccum2 (
    IPState    *pIPState,
    UInt8     **codePtr,
    Int32	source,
    Int32	filterOffset,
    Int32	mask)
{
    PROLOGUE;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgAccum2, sgAccum2Len);

    SET_CODE_LONG(*codePtr, sgAccum2, sgAccum2Source1, source);
    SET_CODE_LONG(*codePtr, sgAccum2, sgAccum2Source2, source + 4);
    SET_CODE_LONG(*codePtr, sgAccum2, sgAccum2Filter, filterOffset);
    SET_CODE_LONG(*codePtr, sgAccum2, sgAccum2Mask, mask);
	
   *codePtr += sgAccum2Len;
}

static void CDECL 
ip_genScgSavePixel (
    IPState    *pIPState,
    UInt8     **codePtr)
{
    PROLOGUE;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgSavePixel, sgSavePixelLen);
   *codePtr += sgSavePixelLen;
}

static Int32 * CDECL 
ip_genScgStoreWord (
    IPState    *pIPState,
    UInt8     **codePtr,
    UInt32	tableLimit)
{
    PROLOGUE;
    Int32 *patchAddr;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgStoreWord, sgStoreWordLen);

    SET_CODE_LONG(*codePtr, sgStoreWord, sgStoreWordTableLimit, tableLimit);
    patchAddr = (Int32 *) GET_LABEL_ADDR (*codePtr, sgStoreWord, sgStoreWordAddr);

   *codePtr += sgStoreWordLen;
    return patchAddr;
}

static void CDECL 
ip_genScgEpilogue (
    IPState    *pIPState,
    UInt8     **codePtr)
{
    PROLOGUE;
	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    memcpy ((char *) *codePtr, (char *) &sgReturn, sgReturnLen);
    *codePtr += sgReturnLen;
}

/*
 * ip_patchJump
 *
 *	Patch a jump instruction once we know where to jump. Jumps are 
 *	relative, so we compute the distance between the dest. 
 *	address and the instruction AFTER the patch address.
 *
 * Arguments:
 *	patchAddr   -	The place in the code which holds the jump offset
 *	destAddr    -	The absolute location to which to jump
 */

static void CDECL 
ip_patchJump (
    IPState    *pIPState,
    Int32      *patchAddr,
    UInt8      *destAddr)
{
    PROLOGUE;
    UInt8   *ptr;

	/* Keep the compiler happy */
    pIPState = pIPState;
    __error__ = __error__;

    ptr = ((UInt8 *) patchAddr) + 4;

    /* patchAddr +4 because it's relative to the PC after the inst. */
    *patchAddr = destAddr - ptr;
}

