;*****************************************************************
;*  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  pcskew.asm:  template for assembly code.

	.386		; compile code for a 386 processor

;	OPTION	LANGUAGE:C
	.MODEL 	small
	OPTION  NOSCOPED

; The algorithms used here are described in gory detail in sskew.c.  The
; SPARC assembly code, which this code mimics fairly closely, is described
; in ssskew.s.

; The WATCOM Linker's PHARLAP format causes the linker to produce code in
; the FLAT memory model.  All that stuff in the Microsoft Assembler manual
; (and in the WATCOM compiler manual and in the flags we give to the 
; WATCOM compiler) about small models is inoperative.  So, that makes
; things easy.  We write code as though we had the small model and
; everything's peachy.

_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: pcskew.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register Assignment for robot:
; EAX:	scratch, source word
; EBX:	source address 
; ECX:	robot state struct pointer
; EDX:	search count
; EBP:	points to variables on the stack
; ESI:	scratch
; EDI:	source pitch
; ESP:	stack pointer

; Register Assignment for ip_findDiffPixelsSum
; EAX	scratch
; EBX:	source offset 
; ECX:	scratch
; EDX:	current run score
; EBP:	points to variables on the stack
; ESI:	srcAddr
; EDI:	srcAddr1
; ESP:	stack pointer

; for use by robot
rSrcPitch	EQU edi
rSearchCount	EQU edx
rState		EQU ecx
; for use by findDiffPixelssum
rSrcOffset	EQU ebx
rSrcAddr	EQU esi
rSrcAddr1	EQU edi
rRunScore	EQU edx

; constants
; if cMaskSample is changed be sure to change cNMaskSample
; used by ip_findDiffPixelsSum when edge differences.
cSampleMask		EQU 07F7F7F7FH
cNSampleMask		EQU NOT cSampleMask
cScoreShiftDist		EQU 4
; used by robot:
cSampleSpacing		EQU 8	; sample every 8th vertical line when
				; looking for the first edge
cLogSampleSpacing 	EQU 3
cMaxSpaceCount		EQU 2	; max amount of white space that can be
				; skipped over
cMaxSpaceCountPlus1	EQU cMaxSpaceCount+1 ; error: too much white space
cMinValidRunLen		EQU 2	; ignore runs shorter than this
cMinSampleLength	EQU cMinValidRunLen+1 
				; ignore edges shorter than this
cBigStepValue		EQU 3	; if the edge changes by more than this in
				; one sample, it's a "big step".  Track it,
				; but remember it might not be well
				; behaved
cBigStepValuePlus1	EQU cBigStepValue+1 ; error: step too big
cLogHeightToSkip	EQU 4
vCountOffset		EQU 4	; offset from start of slope structure to
				; vCount
hCountOffset		EQU 0	; offset from start of slope structure to
				; hCount
ia_successful		EQU 0	; the "everything was fine" return code



; This is a structure for the robot, so we can pass a pointer to this
; information from one procedure to the other (otherwise, we'd need more
; registers --laugh-- or have to pass all these fields as parameters at
; every call) and use the fields just like they were registers, etc.
; Note also to avoid making any variables of this struct global we also
; CAN'T give the fields initializers, since local instances are used and
; their initializers do not get called.
ROBOTSTATE	STRUCT
	hCount		DWORD	?	; count of horizontal movement
	vCount		DWORD	?	; count of vertical movement
	currentSpaces	DWORD	?	; # white words found
	maxSlopeTrack	DWORD	?	; max vert distance to look
	minLegalAddr	DWORD	?
	maxLegalAddr	DWORD	?
	runLengthBound	DWORD	?	; number of samples we can take
					; before deciding edge is long enough
ROBOTSTATE	ENDS

; type declarations and external references
NPBYTE TYPEDEF NEAR PTR BYTE
NPWORD TYPEDEF NEAR PTR WORD
NPDWORD TYPEDEF NEAR PTR DWORD


; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_findDiffPixelsSum.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by fskew.c

; Assumes that numVSamp and numHSamp are at least 1

; Local Variables:
; srcAddr:	address of left end of current source line
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; numVSamp:	number of vertical samples to take
; numHSamp:	number of horizontal samples to take for each vertical samp
; srcVAddrInc:	amount to increment to get to next source line to sample
; srcHAddrInc:	amount to increment to get to next word within source line
; scrPitch:	source bytes per line
; fracVector:	bit vector for factional line displacements
; pResultLo:	Pointer to word that will contain the low word of
;		the result count.
; pResultHi:	Pointer to word that will contain the high word of
;		the result count.

ip_findDiffPixelsSumStart::
ip_findDiffPixelsSum	PROC NEAR C PUBLIC USES EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, numVSamp:DWORD, numHSamp:DWORD, \
	srcVAddrInc:DWORD, srcHAddrInc:DWORD, srcPitch:DWORD, \
	fracVector: DWORD, pResultLo:DWORD, pResultHi:DWORD
LOCAL	vSamp: DWORD, hSamp: DWORD, totalScoreLo:DWORD, totalScoreHi:DWORD, curFracVec: DWORD

; make the assembler happy
	mov	eax, pIPState

	mov	totalScoreLo, 0		; initialize the score
	mov	totalScoreHi, 0
	mov	rSrcAddr, srcAddr
	mov	rSrcAddr1, rSrcAddr	; point to line 2 below current one
	mov	ecx, srcPitch
	add	rSrcAddr1, ecx
	add	rSrcAddr1, ecx
	mov	eax, numVSamp		; init number of sample rows
	mov	vSamp, eax		; assumes that this is at least 1

findDiffPixelsImageLoop:
	mov	rSrcOffset, 0		; offset into image 
	mov	rRunScore, 0		; score of this line
	mov	eax, fracVector	
	mov	curFracVec, eax
	mov	eax, numHSamp		; init number of horizontal samples
	mov	hSamp, eax		; assumes that this is at least 1

findDiffPixelsRowLoop:
	mov	eax, [rSrcAddr+rSrcOffset] ; get input samples
	mov	ecx, [rSrcAddr1+rSrcOffset]
	add	rSrcOffset, srcHAddrInc	; point to next source word
	rol	curFracVec, 1		; do we need to jump a line?
	jnc	findDiffPixelsComputeDiff ; no, just compute the difference
	add	rSrcOffset, srcPitch	; yes, increment the src offset to
					; take care of the jump

findDiffPixelsComputeDiff:
	cmp	eax, ecx		; are there any edges at all?
	jz	findDiffPixelsCheckRow	; no, see if we're at the end of the
					; row
; there was at least one edge so compute differences
	and	eax, cSampleMask	; get rid of MS bit of each byte
	and	ecx, cSampleMask
	add	eax, cSampleMask	; MS bit is set if any of the lower
	add	ecx, cSampleMask	; 7 were on
	xor	eax, ecx		; edge bits in MS bit
	and	eax, cNSampleMask	; get just edge bits
; rRunScore is really 4 individual counts which are combined only at the
; end of the line.
	shr	eax, 7			; add the 4 individual counts
	add	rRunScore, eax		; into the current total

findDiffPixelsCheckRow:
	dec	hSamp			; are we done with this row?
	jnz	findDiffPixelsRowLoop	; no, continue with the row

; yes, we're at the end of the row.  
; There are 4 edge counts in rRunScore.  Add them together.  We know
; that there are no more than 127 horizontal samples, so none of
; the counts can be bigger than 127.  This lets us do the first
; set of adds without masking anything since there can be no overflow
; at this stage.
	mov	eax, rRunScore
	shr	eax, 16			; get 2 scores from 4
	add	eax, rRunScore
	mov	rRunScore, eax		
	shr	eax, 8			; get 1 score from 2
	and	eax, 0FFH		; get rid of trash
	and	rRunScore, 0FFH
	add	eax, rRunScore

; Now, do a crude non-linear operation, so that big scores (which you
; would get if you were sampling along an edge) really stand out.
; Note that this shift may cause some of the score bits to be shifted
; to the upper score word.  Take care of that case by shifting the
; score to the right by (32-shiftCount) and adding the result to the
; upper word.  Since the 386 and above processors only look at the
; ls 5 bits of the shiftcount, (32-shiftcount) = -shiftCount for our
; purposes.  HOWEVER, if shiftCount = 0, then -shiftCount also equals
; 0, not 32.  In this case, we want to add 0 to totalScoreHi, so
; just skip the add.
	mov	rRunScore, eax		; save the score for adding
					; to the low score word
	mov	ecx, eax
	shr	ecx, cScoreShiftDist	; compute shiftCount
	neg	cl			; compute 32-shiftCount
	jz	findDiffPixelsSumUpdateLowScore
	shr	eax, cl			; amount we add to upper score word
	add	totalScoreHi, eax

; do the lower score word
	neg	cl			; restore shiftCount
	shl	rRunScore, cl
findDiffPixelsSumUpdateLowScore:
	add	totalScoreLo, rRunScore	; add total to low word
	adc	totalScoreHi, 0		; add carry to high word

	add	rSrcAddr, srcVAddrInc	; point to next source lines
	add	rSrcAddr1, srcVAddrInc
	dec	vSamp			; are we done with the image
	jnz	findDiffPixelsImageLoop	; no, go do the next line

; At this point, the total score is in totalScoreHi:totalScoreLo.
; We want to adjust it so that the MS bit of totalScoreLo is shifted
; into totalScoreHi and cleared in totalScoreLo.
	mov	eax, totalScoreLo
	mov	ecx, totalScoreHi
	shld	ecx, eax, 1		; adjust totalScoreHi
	and	eax, 07FFFFFFFH		; clear MS bit of totalScoreLo
	mov	ebx, pResultLo		; point to lower result word
	mov	[ebx], eax		; store lower result word
	mov	ebx, pResultHi		; point to upper result word
	mov	[ebx], ecx		; store upper result word
	mov	eax, ia_successful	; return value
	ret
ip_findDiffPixelsSum ENDP


; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name ip_doEdgeSample.
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by fskew.c

; Local Variables:
; srcAddr:	address of left end of current source line
; srcPitch:	number of bytes to be added to address to move down one line
;		in the source (input) image
; slopeAddr:	addr of structure that countains horizontal and vertical
;		movement counter
; maxVertSearch: maximum amount to look vertically for an edge
; legalAddrMin:	lowest legal address still in image
; legalAddrMax:	highest legal address still in image
; hRunEnd:	amt of hor. movement for this run
; vRunEnd:	amt of vert. movement for this run

ip_doEdgeSampleStart::
ip_doEdgeSample	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI,\
	pIPState:DWORD, srcAddr:DWORD, srcBpl:DWORD, slopeAddr:DWORD, \
	maxVertSearch:DWORD, legalAddrMin:DWORD, legalAddrMax:DWORD
; declared a couple local variables and
; declare a local structure variable "state" (fields will not be initialized)
LOCAL	hRunEnd:DWORD, vRunEnd:DWORD, runLength:DWORD, state:ROBOTSTATE

; keep the assembler happy
	mov	eax, pIPState

; keep maxSlopeTrack, minLegalAddr, and maxLegalAddr in the data
; segment since ip_doFindEdgeTakeStep can't get to them without either having
; them pushed onto the stack in its frame or using ebp and the correct
; offset given that a new procedure has been called.  For the sake of
; clarity, just put them in the data segment.
	mov	eax, maxVertSearch
	mov	state.maxSlopeTrack, eax
	mov	eax, legalAddrMin
	mov	state.minLegalAddr, eax
	mov	eax, legalAddrMax
	mov	state.maxLegalAddr, eax
	mov	rSrcPitch, srcBpl

; Put the maximum legal edge length into runLengthBound.  This is decremented
; and checked in doFindEdgeTakeStep in the same way we check for the
; edge ending in white space.
	mov	eax, rSrcPitch
	shr	eax, 1			; 2 bytes/sample
	mov	state.runLengthBound, eax

; assume that we're in a white place and look down for black
	mov	ecx, rSrcPitch		; calc offset between src lines
	shl	ecx, cLogSampleSpacing

doEdgeSampleFindBlack:
	mov	ebx, srcAddr		; get source address
	mov	eax, [ebx]		; get source word
	or	eax, 0			; is this the black word that we want?
	jnz	doEdgeSampleFindWhite	; yes, go look for black edge

	add	ebx, ecx		; point to next src word
	mov	eax, [ebx]		; get source word
	or	eax, 0			; is this the black word that we want?
	jnz	doEdgeSampleFindWhite	; yes, go look for black edge

	add	ebx, ecx		; point to next src word
	mov	eax, [ebx]		; get source word
	or	eax, 0			; is this the black word that we want?
	jz	finishDoEdgeSample	; no, give up and leave

; We have found some black, now look downwards for white so that we can
; get a black/white horizontal edge to track.
doEdgeSampleFindWhite:
	add	ebx, ecx		; point to next src word
	mov	eax, [ebx]		; get source word
	or	eax, 0			; is this the white word that we want?
	jz	doEdgeSampleFoundEdge	; yes, go refine the white edge

	add	ebx, ecx		; point to next src word
	mov	eax, [ebx]		; get source word
	or	eax, 0			; is this the white word that we want?
	jz	doEdgeSampleFoundEdge	; yes, go refine the white edge

	add	ebx, ecx		; point to next src word
	mov	eax, [ebx]		; get source word
	or	eax, 0			; is this the white word that we want?
	jnz	finishDoEdgeSample	; no, give up and leave

; We've almost found a black/white edge.  The edge is somewhere between
; ebx and ebx-ecx(lineOffset).  Refine the location using a linear
; search.
doEdgeSampleFoundEdge:
	sub	ebx, rSrcPitch		; point back
	mov	eax, [ebx]		; get sample word
	or	eax, 0			; is this black word that we want?
	jz	doEdgeSampleFoundEdge	; no keep looking
; when we exit this loop ebx will point to the black word

; here comes the infamous tracking state machine.
; In the initial state, hCount and vCount are cleared.
; We return here as long as one of the first cMinValidRunLen steps of the
; run is big.

	mov	state.currentSpaces, cMaxSpaceCount
	mov	hRunEnd, 0
	mov	vRunEnd, 0
	xor	ebx, 2			; we'll deal with 16 bit words
					; from now on so accomodate word
					; swapping

doEdgeSampleTrackInit:
	mov	state.hCount, 0
	mov	state.vCount, 0
	mov	runLength, cMinValidRunLen
	lea	rState, state		; load structure address of robot
					; state to register for use in
					; ip_doFindEdgeTakeStep
doEdgeSampleFind1stRun:
	call	ip_doFindEdgeTakeStep	; try to take one step
	cmp	eax, cBigStepValue 	; was one little step taken?
	jg	doEdgeSampleTrackInit	; no, try again.
	dec	runLength		; yes. is this a valid run?
	jnz	doEdgeSampleFind1stRun	; not yet, try to take another step

; Found a(nother) valid run.  Record its location and take a step.
doEdgeSampleFoundRun:
	mov	eax, state.hCount	; yes, update hRunEnd, vRunEnd now
	mov	hRunEnd, eax
	mov	eax, state.vCount	; that we're in a run
	mov 	vRunEnd, eax
	call	ip_doFindEdgeTakeStep	; try to take another step
	cmp	eax, cBigStepValue 	; was one little step taken?
	jle	doEdgeSampleFoundRun	; yes, take another

; no, a big step was taken, so that run is over.
; See if we can find another run.

doEdgeSampleFindLaterRunInit:
	mov	runLength, cMinValidRunLen
doEdgeSampleFindLaterRun:
	call	ip_doFindEdgeTakeStep	; try to take one step
	cmp	eax, cBigStepValue 	; was one little step taken?
	jg	doEdgeSampleFindLaterRunInit	; no, try again.
	dec	runLength		; yes. is this a valid run
	jnz	doEdgeSampleFindLaterRun ; not yet, try to take another step

; Have we gone too far trying to skip over weird stuff?  This can happen
; if we follow a long diagonal line, for example.  If the difference
; between the current value of hCount and the last value of hRunEnd
; is larger than the amount of white space we're allowed to skip plus
; the length of the run required to exit the loop, just quit without
; recording the new value of hCount and vCount.
	mov	eax, state.hCount
	sub	eax, hRunEnd		; length of run
	cmp	eax, cMaxSpaceCount+cMinValidRunLen+1
	jle	doEdgeSampleFoundRun	; this is a valid run.  go process it.

; This state machine is exited when ip_doFindEdgeTakeStep is unable to
; find an edge or when we've tracked a long alien object.

doEdgeSamplePostResults:
	cmp	hRunEnd, cMinSampleLength ; was that edge big enough to record?
	jle	finishDoEdgeSample	; no, just get outta here
	mov	ebx, slopeAddr
	mov	eax, vRunEnd		; get vertical movement
	add	[ebx+vCountOffset], eax	; and calc new value
	mov	eax, hRunEnd		; get horizontal movement
	add	[ebx+hCountOffset], eax	; and calc new value

finishDoEdgeSample:
	ret
ip_doEdgeSample ENDP

; This procedure takes one step to the right and tries to follow the edge
; On entry assumes that ebx contains the source address xor'd with 2
; On return ebx (source address) will point to hCount over and vCount
; up from entry.
; On return eax will hold the size of the vertical step
ip_doFindEdgeTakeStepStart::
; if this procedure ever push registers on the stack or is no longer near,
; the code at doTakeStepErrorFinish will need to be modified.
ip_doFindEdgeTakeStep	PROC NEAR C PRIVATE
	cmp	ebx, (ROBOTSTATE PTR [rState]).minLegalAddr
					; have we gone off the top?
	jle	doTakeStepErrorFinish
	cmp	ebx, (ROBOTSTATE PTR [rState]).maxLegalAddr
					; have we gone off the bottom?
	jge	doTakeStepErrorFinish

; Move 2 bytes to the right and take a sample.  If it's black, look
; down one line for white.  If it's white, start looking up for
; the black edge.
	xor	ebx, 2			; move 2 bytes to the right and
	add	ebx, 2			; take a sample
	xor	ebx, 2
	mov	ax, [ebx]		; get the input sample
	or	ax, 0			; is this sample still black?
	jz	doTakeStepWhite

doTakeStepBlack:
	mov	esi, rSrcPitch		; look down one line
	mov	ax, [ebx+esi]
	or	ax, 0			; is this sample white?
	jz	doTakeStepHorizRet
; the sample to the right was black and the sample below that is black
; so the edge is slanting down
	mov	rSearchCount, (ROBOTSTATE PTR [rState]).maxSlopeTrack
	shl	rSearchCount, 1
doTakeStepBlackLoop:
	dec	rSearchCount
	jz	doTakeStepErrorFinish	; ended without finding edge
	add	esi, rSrcPitch		; look down at the next line
	mov	ax, [ebx+esi]
	or	ax, 0			; is this sample still black?
	jnz	doTakeStepBlackLoop	; yes, look down another line
; we've found the white edge
; Number of lines edge went down (pos dir) = 2*maxSlopeTrace-rSearchCount
; Number of halfwords we went over = 1
; address of lowest black line = ebx+(esi-rSrcPitch)
	sub	esi, rSrcPitch		; get addr of lowest black line
	add	ebx, esi
	inc	(ROBOTSTATE PTR [rState]).hCount
					; one halfword over
	mov	eax, (ROBOTSTATE PTR [rState]).maxSlopeTrack
					; get # lines went down
	shl	eax, 1			; 2*maxSlopeTrack
	sub	eax, rSearchCount	; eax holds # lines went down
	add	(ROBOTSTATE PTR [rState]).vCount, eax
	mov	(ROBOTSTATE PTR [rState]).currentSpaces, cMaxSpaceCount
					; we're not in white space
	dec	(ROBOTSTATE PTR [rState]).runLengthBound
					; is the edge too long?
	jz	doTakeStepErrorFinish	; yes, quit
;vertical step size in eax
	ret

; the sample to the right was white, so the edge has moved up.
doTakeStepWhite:
	mov	esi, 0			; clear offset
	mov	eax, (ROBOTSTATE PTR [rState]).maxSlopeTrack
					; init rSearchCount
	mov	rSearchCount, eax
doTakeStepWhiteLoop:
	dec	rSearchCount
	jz	doTakeStepWhiteDesert
	sub	esi, rSrcPitch
	mov	ax, [ebx+esi]
	or	ax, 0			; is this sample still white?
	jz	doTakeStepWhiteLoop	; yes, look up another line

; We have found the black edge.
; Number of lines edge went up (neg dir) = maxSlopeTrack-rSearchCount
; Number of halfwords we went over = 1
; address of lowest black line = ebx+esi
	add	ebx, esi		; get addr of lowest black line
	inc	(ROBOTSTATE PTR [rState]).hCount
					; one halfword over
	mov	eax, (ROBOTSTATE PTR [rState]).maxSlopeTrack
					; get # lines went up
	sub	eax, rSearchCount
	sub	(ROBOTSTATE PTR [rState]).vCount, eax
	mov	(ROBOTSTATE PTR [rState]).currentSpaces, cMaxSpaceCount
					; white space has ended
	dec	(ROBOTSTATE PTR [rState]).runLengthBound
					; is the edge too long?
	jz	doTakeStepErrorFinish	; yes, quit
					; vertical step size in eax
	ret

; The edge has vanished.
; Number of halfwords we went over = 1
; Number of lines edge went up (neg dir) = 0
; source address = ebx
doTakeStepWhiteDesert:
	inc	(ROBOTSTATE PTR [rState]).hCount
					; one halfword over
	dec	(ROBOTSTATE PTR [rState]).currentSpaces
					; found white space
	jz	doTakeStepErrorFinish	; if we've found to much exit back
					; to ip_doEdgeSample's caller
	dec	(ROBOTSTATE PTR [rState]).runLengthBound
					; is the edge too long?
	jz	doTakeStepErrorFinish	; yes, quit
	mov	eax, cBigStepValuePlus1	; vertical step size
	ret

doTakeStepErrorFinish:
; we want to return to ip_doEdgeSample's caller.  First we have to get rid
; of the stuff on the stack that is doTakeStep's.  Then we can jump to
; the end of ip_doEdgeSample.  If we do it this way, we eliminate compares
; in ip_doEdgeSample of currentSpaces against 0.  But doing the compare
; here means that we can't just return when currentSpaces is 0 since
; there is no check in ip_doEdgeSample's loop anymore; hence, the fancy
; stuff with the stack
	pop	eax			; the return address
	jmp	doEdgeSamplePostResults	; the end of doEdgeSample	
	
doTakeStepHorizRet:
	inc	(ROBOTSTATE PTR [rState]).hCount
	mov	(ROBOTSTATE PTR [rState]).currentSpaces, cMaxSpaceCount
					; we're not in white space
	mov	eax, 0			; vertical step size
	dec	(ROBOTSTATE PTR [rState]).runLengthBound
					; is the edge too long?
	jz	doTakeStepErrorFinish	; yes, quit
	ret

ip_doFindEdgeTakeStep ENDP


_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
; no global initialized data used
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END


IF 0				; These are all comments


COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: pcskew.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.11  1998/01/13 23:58:37  ddavies
pcskew.asm -> /product/ipcore/ipshared/src/RCS/pcskew.asm,v

CommonCheck: Add IPState parameter to almost all private routines.

Revision 1.10  1996/09/17 00:45:53  ddavies
pcskew.asm -> /product/ipcore/ipshared/src/RCS/pcskew.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.9  1996/09/16 00:52:12  ddavies
pcskew.asm -> /product/ipcore/ipshared/src/RCS/pcskew.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

Revision 1.8  1995/09/15 21:48:09  ddavies
pcskew.asm -> /product/ipcore/ipshared/src/RCS/pcskew.asm,v

CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
IF 0 that was already there.  Why does David Hirsch's version of ML get
upset about the # signs that RCS puts at the beginnings of line (sometimes)
while our systems are happy??

Revision 1.7  1995/02/24  02:33:49  ddavies
pcskew.asm -> /product/ipcore/ipshared/src/RCS/pcskew.asm,v

Change ip_findDiffPixelsSum so that it computes the total score into a
double word instead of into a single 32-bit word.  This is used to
prevent overflows.  This routine assumes that there are at most 127
horizontal samples per line, which is guaranteed by ip_detectSkewFine.

Revision 1.6  1994/08/03  17:25:52  ejaquith
pcskew.asm -> /product/ipcore/ipshared/src/RCS/pcskew.asm,v

Reentrancy!
Removed globals by adding local variables to some functions.  The Big Change
is that there were so many globals that get passed to ip_doFindEdgeTakeStep
that I created a robot structure in one procedure, then pass that pointer
into the doFindEdgeTakeStep procedure.  This was hard to come by as it must
be coded fairly particularly to avoid globals.

Revision 1.5  1993/11/05  01:53:14  ddavies
pcskew.asm -> /product/ipcore/ipshared/src/RCS/pcskew.asm,v

Add RCS magic to have string specifying this file's name, rev number and
last edit date into the object file.  Also add change log at the bottom.

^

ENDIF

