;*****************************************************************
;*  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
;*  Copyright protection claimed includes all forms and matters  *
;*  of copyrightable material and information now allowed by     *
;*  statutory or judicial law or hereafter granted, including    *
;*  without limitation, material generated from the software     *
;*  programs which are displayed on the screen such as icons,    *
;*  screen display looks, etc.                                   *
;*****************************************************************/

;*  scgtemp.asm:  templates for scale-to-gray assembly code.

	.486		; compile code for a 386 processor

;	OPTION	LANGUAGE:C

; Constants that are easily visible from a debugger
Big	equ	12345678h
Small	equ	5h

_DATA	SEGMENT PUBLIC 'DATA'
RCSInfo	BYTE	"$RCSfile: scgtemp.asm,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	BYTE	0
_DATA	ENDS

; Register/local Assignment:
; EAX:		convolution accumulator
; EBX:		tmp1
; ECX:		shiftCount
; EDX:		tmp2
; local:	outputWord
; local:	destPtr
; ESI:		srcPtr
; EDI:		filterPtr
; EBP:		tablePtr

; Parameters:
; destP:	Address of destination line
; srcP:		Address of start of source line 
; filterP:	Pointer to column of filters
; tableP	Address of (shift,shift,offset) table
;

; The following code is VERY bizarre.  The MASM assemblers seem to
; randomly decide that lines are too long.  This randomness seems to
; be a function of the total number of characters encountered.  So,
; we try to make sure that the total number of characters encountered
; in both the NT and Windows3.1 version are the same.  That way, fixing
; it on NT should also fix it on Win3.1.
IFNDEF AsmForNT
DGROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ELSE
;GROUP	GROUP	_DATA, _BSS	; DGROUP is a group of segments containing
				; (at least) _DATA and _BSS.  We'll use this
				; declartion in the ASSUME below.
ENDIF

; Declare the code segment.  It's byte aligned, should be combined with other
; segments of the same name, contains code that does 32 bit ops by default
; and should belong to the class CODE.  This last thinger means it is grouped
; with all the other 'CODE' segments at load time.
; Note that the NT line has one more TAB char, so needs one fewer comment char.

_TEXT	SEGMENT	BYTE PUBLIC USE32 'CODE'
	ASSUME	CS:_TEXT	; the default anyway
IFNDEF AsmForNT
	ASSUME	DS:DGROUP	; points to group containing data seg.
ELSE
	ASSUME	DS:FLAT		; points to flat data seg. More chars
ENDIF

; To get the Microsoft assembler to believe the following labels are PUBLIC,
; we have to declare them with double colons, e.g. "scgStart::".
	PUBLIC sgStart, sgStartLen
	PUBLIC sgPixelStart, sgPixelStartAddr, sgPixelStartLen
	PUBLIC sgAccum1, sgAccum1Source, sgAccum1Mask, sgAccum1Filter,
	       sgAccum1Len
	PUBLIC sgGotoBottom, sgGotoBottomAddr, sgGotoBottomLen
	PUBLIC sgAccum2, sgAccum2Source1, sgAccum2Source2, sgAccum2Mask
	PUBLIC sgAccum2Filter, sgAccum2Len
	PUBLIC sgAccum1Pair, sgAccum1PairSource1, sgAccum1PairSource2
	PUBLIC sgAccum1PairMask1, sgAccum1PairMask2
	PUBLIC sgAccum1PairFilter1, sgAccum1PairFilter2
	PUBLIC sgAccum1PairLen
	PUBLIC sgSavePixel, sgSavePixelLen
	PUBLIC sgStoreWord, sgStoreWordAddr, sgStoreWordTableLimit, 
	       sgStoreWordLen
	PUBLIC sgReturn, sgReturnLen

; Declare the procedure using the Microsoft assembler's fancy features:
; procedure name scgProc
; It's a near proc.
; It is called using C language protocol (first arg pushed last, _ prepended
;   to proc name)
; It's a public name so the Linker can find it
; It uses lots of registers
; The arguments are declared in the order specified by pcgenscg.c

sgStart::
	PUBLIC	sgProc
sgProc	PROC NEAR C PUBLIC USES EAX EBX ECX EDX ESI EDI EBP,\
	destP:DWORD, srcP:DWORD, filterP:DWORD, tableP:DWORD

; All registers used are also pushed on the stack automatically.
; NOTE that we don't use the MASM scheme for generating locals, since 
; that wastes BP, and we'd really rather have the extra register
; So define some local offsets from the stack
OutWord		equ	0h
DestPtr		equ	4h
FrameSize	equ	8h		; Size of this local frame

; Constants for dealing with the shift, offset table
TableShift	equ	0h		; shift count (s)
TableDelta	equ	4h		; Source pointer delta
TableEntrySize	equ	8h		; Size of a table entry

; Load the registers
	mov	esi,srcP		; Load pointer to filter row
	mov	edi,filterP		; Load output line pointer
	mov	ebx,destP		; Save destP in temp reg
	mov	ebp,tableP		; Load table pointer

; Move down the stack to make our local frame
	sub	esp,FrameSize
	mov	[esp+DestPtr],ebx	; save destPtr
sgStartEnd::

;
; Clear the convolution sum and shift up the output word.
; Decide whether this is a one- or two-word case, and branch accordingly.
; The branch will be filled in later, since we're generating
; each branch's code depending on the filter height.
;
sgPixelStart::
	sub	eax,eax
	shl	DWORD PTR[esp+OutWord],4
	add	esi,[ebp+TableDelta]	; Increment the source pointer
	mov	ecx,[ebp+TableShift]	; Get the shift count(s)
	cmp	ecx,100h
	jae	sgReturn		; If >100, there's more than one
sgPixelStartAddr::
sgPixelStartEnd::

;
; Process a filter row, one-word case. In this case, bits to process are
; contained in a single source word, so we only need to do one load, shift,
; and mask. 
; All of the constants we use are either 32-bit constants or 8-bit constants,
; and are always at the end of each instruction, so we put a label *after* the
; instruction for the constant, which we will offset by -4 or -1
; before we stuff the constant.

sgAccum1::
	mov	ebx,[esi+Big]		; Get the word, address to be filled in
sgAccum1Source::
	shr	ebx,cl			; Shift it down
	and	ebx,Big			; Mask the relevant bits
sgAccum1Mask::
	add	eax,[edi+ebx*4+Big]	; Look up the bits and accumulate
sgAccum1Filter::
sgAccum1End::

;
; Finished with one-word case processing; go to the bottom of the loop
;

sgGotoBottom::
	jmp	foo
sgGotoBottomAddr::
sgGotoBottomEnd::

;
; Process a filter row, two-word case. In this case, bits to process
; straddle two words, so we need to do two load-mask-shift operations
; and OR them together. The shift counts are stored as x:0:0:x in ecx

sgAccum2::
	mov	ebx,[esi+Big]		; Get word, address to be filled in
sgAccum2Source1::			; 
	shl	ebx,cl			; Shift it up
;	bswap	ecx			; Get the other shift count into cl
	rol	ecx,8
	mov	edx,[esi+Big]		; Get next word
sgAccum2Source2::		 
	shr	edx,cl			; Shift those bits down
	or	ebx,edx
	and	ebx,Big
sgAccum2Mask::
;	bswap	ecx			; Swap back shift count
	ror	ecx,8
	add	eax,[edi+ebx*4+Big]	; Look up the bits and accumulate
sgAccum2Filter::
sgAccum2End::

IF 0
;
; Process a filter row, two-word case. In this case, bits to process
; straddle two words, so we need to do two load-mask-shift operations
; and OR them together. The shift counts are stored as x:0:0:x in ecx
; This version doesn't do bswaps, i.e. usable on 386s

sgAccum2N::
	mov	ebx,[esi+Big]		; Get word, address to be filled in
sgAccum2Source1N::			; 
	shl	ebx,cl			; Shift it up
	rol	ecx,8			; Get the other shift count into cl
	mov	edx,[esi+Big]		; Get next word
sgAccum2Source2N::		 
	shr	edx,cl			; Shift those bits down
	or	ebx,edx
	and	ebx,Big
sgAccum2MaskN::
	ror	ecx,8			; Swap back CL
	add	eax,[edi+ebx*4+Big]	; Look up the bits and accumulate
sgAccum2FilterN::
sgAccum2EndN::
ENDIF
;
; Process two filter rows, one-word case. In this case, bits to process are
; contained in a single source word, so we only need to do one load, mask,
; and shift. 

sgAccum1Pair::
	mov	ebx,[esi+Big]		; Get the word, address to be filled in
sgAccum1PairSource1::
	mov	edx,[esi+Big]		; Get the next word
sgAccum1PairSource2::			; 
	shr	ebx,cl			; Shift it down
	and	ebx,Big		; Mask the relevant bits
sgAccum1PairMask1::
	shr	edx,cl			; Shift it down
	and	edx,Big		; Mask the relevant bits
sgAccum1PairMask2::
	add	eax,[edi+ebx*4+Big]	; Look up the bits and accumulate
sgAccum1PairFilter1::
	add	eax,[edi+edx*4+Big]	; Look up the bits and accumulate
sgAccum1PairFilter2::
sgAccum1PairEnd::


;
; So we've finished a filter row. Store it into the output word,
; increment the source pointer, and increment the table pointer
; The ouput word was already shifted when we cleared the input word.
;
sgSavePixel::
	shr	eax,18h			; Get the integer portion from 8.24
	or	[esp+OutWord],eax	; Stick them bits in
	add	ebp,TableEntrySize	; Move to the next table entry
sgSavePixelEnd::

;
; The output word is full, so store it and loop. We finish when
; tableP reaches tableEnd.
; This loop is only done for pixels which completely fill an output word;
; the remaining pixels are done in-line.
;
sgStoreWord::
	mov	edx,[esp+OutWord]	; Get output word
	mov	ebx,[esp+DestPtr]	; Get the pointer
	mov	[ebx],edx		; Store it
	add	ebx,4 			; Increment the output word ptr
	mov	[esp+DestPtr],ebx	; Save back the output word ptr
	cmp	ebp,Big			; Is tablePtr at the limit?
sgStoreWordTableLimit::
	jnae	sgStart			; If not, keep going
sgStoreWordAddr::
sgStoreWordEnd::

;
; Epilogue
; 	This restores SP back so that BP can be restored, and the
;	environment can be happy. Everything but restoring our frame is
;	generated by the assembler.

sgReturn::
	add	esp,FrameSize		; Free our frame
	ret
sgReturnEnd::


sgProc ENDP

; Lengths of the code blocks.  Put these down here so the disassembler
; output is easier to read (starts with code at offset 0).
sgStartLen		DWORD	sgStartEnd - sgStart
sgPixelStartLen		DWORD	sgPixelStartEnd - sgPixelStart
sgAccum1Len		DWORD	sgAccum1End - sgAccum1
sgGotoBottomLen		DWORD	sgGotoBottomEnd - sgGotoBottom
sgAccum2Len		DWORD	sgAccum2End - sgAccum2
sgAccum1PairLen		DWORD	sgAccum1PairEnd - sgAccum1Pair
sgSavePixelLen		DWORD	sgSavePixelEnd - sgSavePixel
sgStoreWordLen		DWORD	sgStoreWordEnd - sgStoreWord
sgReturnLen		DWORD	sgReturnEnd - sgReturn

foo:
_TEXT	ENDS				; End the segment

_DATA	SEGMENT PUBLIC 'DATA'
; all locals are on the stack
_DATA	ENDS

_BSS	SEGMENT PUBLIC 'BSS'
; no global uninitialized data used
_BSS	ENDS

	END

IF 0				; These are all comments

COMMENT ^			David Hirsch's version of ML doesn't
				think this part is IF'd out for some
				reason ????
$Log: scgtemp.asm,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.6  1996/09/17 00:45:56  ddavies
scgtemp.asm -> /product/ipcore/ipshared/src/RCS/scgtemp.asm,v

CommonCheck: These files have some AsmForNT sections.  MASM has a stupid bug (forever)
in which it randomly decides that some (always uncommented?) lines
are <quote> too long <unquote>.  This behavior seems to depend on the
total offset into the file when the end of each line is seen.  In
the previous version, the number of characters used for the Win3.1 and
WinNT versions were not quite the same.  This caused MASM running on
the two platforms to randomly decide different lines were too long.
This idiotic workaround pads the number of characters inserted so that
the number of characters seen by the assembler is the same in both cases.
So long as this is true, MASM is both environments exhibits the same
behavior.  The lines it decides are too long can have comments added to
them (which seems to fix the problem!!).

Revision 1.5  1996/09/16 00:52:25  ddavies
scgtemp.asm -> /product/ipcore/ipshared/src/RCS/scgtemp.asm,v

Bunch of changes to make MASM 6.11d happy: 1) remove alignment restriction
on _DATA and _BSS segments; 2) don't declare DGROUP when assembling
for NT; 3) ASSUME DS is FLAT when assembling for WinNT.

;; Revision 1.4  1995/09/15  21:48:13  ddavies
;; scgtemp.asm -> /product/ipcore/ipshared/src/RCS/scgtemp.asm,v
;;
;; CommonCheck: Enclose the log entries in the multiline comment chars IN ADDITION TO THE
;; IF 0 that was already there.  Why does David Hirsch's version of ML get
;; upset about the # signs that RCS puts at the beginnings of line (sometimes)
;; while our systems are happy??
;;
;; Revision 1.3  1995/01/13  02:14:50  macleod
;; macleod on Thu Jan 12 18:12:06 PST 1995
;;
;; scgtemp.asm -> /product/ipcore/ipshared/src/RCS/scgtemp.asm,v
;;
;; Changed a BSWAP to a rotate, because the 386 doesn't have BSWAP.
;; This did add a bit of time back (about 3% in one test I did).
;; At some point, I'll have to check the CPU type, and generate the
;; appropriate instruction for the CPU.
;;
;; Revision 1.2  1995/01/13  02:06:33  macleod
;; scgtemp.asm -> /product/ipcore/ipshared/src/RCS/scgtemp.asm,v
;;
;; Rewrote the code so that it is generated in a loop rather than inline.
;;
;; Revision 1.1  1995/01/04  01:18:44  macleod
;; Initial revision
;;

^

ENDIF

