/*****************************************************************
 *  Copyright (c) 1994, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *  ssflush.c -- assembly code that attempts to flush the instruction
 *		 cache using the iflush opcode.
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssflush.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	.align	4

/* We get weird crashes on SPARC 5 machines, possibly associated with
 * executing generated assembly code.  A possible cause of these crashes
 * is deleting a buffer containing generated code, allocating a new code
 * buffer and getting the same memory, filling it with new stuff, having
 * some of the old stuff remain in the I cache, then try to execute the
 * new stuff.  We execute some old stuff and some new stuff, getting trash.
 *
 * This module attempts the brute force method of cache flush - it just
 * executes a bunch of opcodes.  Let's try for 8KB of junk.
 * Our opcode sequence is a bunch of jumps, each one 16 bytes ahead.  This
 * should touch all the cache lines in this block.
 */

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define rIPState	%i0
#define	rSrcAddr	%i1		/* current flush address */
#define rBufLength	%i2		/* current ICache flush count */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	srcAddrOffset	0x48
#define	bufLengthOffset	0x4c

/* magical C Frame constant (a few local variables) */
#define	LF26		-80

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(i_flushICache)
MK_GLOBAL_NAME(i_flushICache):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  As stated above
! we also use some of this stuff.
	st	rIPState,   [%fp+ipStateOffset]
	st	rSrcAddr,   [%fp+srcAddrOffset]
	st	rBufLength, [%fp+bufLengthOffset]

! Just to make sure we're clearing enough, add 32 to the buffer length.
! See, if the buffer ends in the middle of a cache line, we'd probably
! decrement bufLength, decide that it was negative and not flush the
! last line.
	add	rBufLength, 0x20, rBufLength
ficLoop:
	iflush	rSrcAddr		! flush address contained in rSrcAddr
					! from instruction cache
	subcc	rBufLength, 0x10, rBufLength
	bge	ficLoop
	add	rSrcAddr, 0x10, rSrcAddr

#if 0
! if that doesn't work, we can try this mess.
! Just start jumping			  ADDRESS
1:	ba	0f			! 0x0000
	nop
	nop
	nop
0:	ba	1f			! 0x0010
	nop
	nop
	nop
1:	ba	0f			! 0x0020
	nop
	nop
	nop
0:	ba	1f			! 0x0030
	nop
	nop
	nop
1:	ba	0f			! 0x0040
	nop
	nop
	nop
0:	ba	1f			! 0x0050
	nop
	nop
	nop
1:	ba	0f			! 0x0060
	nop
	nop
	nop
0:	ba	1f			! 0x0070
	nop
	nop
	nop
1:	ba	0f			! 0x0080
	nop
	nop
	nop
0:	ba	1f			! 0x0090
	nop
	nop
	nop
1:	ba	0f			! 0x00A0
	nop
	nop
	nop
0:	ba	1f			! 0x00B0
	nop
	nop
	nop
1:	ba	0f			! 0x00C0
	nop
	nop
	nop
0:	ba	1f			! 0x00D0
	nop
	nop
	nop
1:	ba	0f			! 0x00E0
	nop
	nop
	nop
0:	ba	1f			! 0x00F0
	nop
	nop
	nop

1:	ba	0f			! 0x0100
	nop
	nop
	nop
0:	ba	1f			! 0x0110
	nop
	nop
	nop
1:	ba	0f			! 0x0120
	nop
	nop
	nop
0:	ba	1f			! 0x0130
	nop
	nop
	nop
1:	ba	0f			! 0x0140
	nop
	nop
	nop
0:	ba	1f			! 0x0150
	nop
	nop
	nop
1:	ba	0f			! 0x0160
	nop
	nop
	nop
0:	ba	1f			! 0x0170
	nop
	nop
	nop
1:	ba	0f			! 0x0180
	nop
	nop
	nop
0:	ba	1f			! 0x0190
	nop
	nop
	nop
1:	ba	0f			! 0x01A0
	nop
	nop
	nop
0:	ba	1f			! 0x01B0
	nop
	nop
	nop
1:	ba	0f			! 0x01C0
	nop
	nop
	nop
0:	ba	1f			! 0x01D0
	nop
	nop
	nop
1:	ba	0f			! 0x01E0
	nop
	nop
	nop
0:	ba	1f			! 0x01F0
	nop
	nop
	nop

1:	ba	0f			! 0x0200
	nop
	nop
	nop
0:	ba	1f			! 0x0210
	nop
	nop
	nop
1:	ba	0f			! 0x0220
	nop
	nop
	nop
0:	ba	1f			! 0x0230
	nop
	nop
	nop
1:	ba	0f			! 0x0240
	nop
	nop
	nop
0:	ba	1f			! 0x0250
	nop
	nop
	nop
1:	ba	0f			! 0x0260
	nop
	nop
	nop
0:	ba	1f			! 0x0270
	nop
	nop
	nop
1:	ba	0f			! 0x0280
	nop
	nop
	nop
0:	ba	1f			! 0x0290
	nop
	nop
	nop
1:	ba	0f			! 0x02A0
	nop
	nop
	nop
0:	ba	1f			! 0x02B0
	nop
	nop
	nop
1:	ba	0f			! 0x02C0
	nop
	nop
	nop
0:	ba	1f			! 0x02D0
	nop
	nop
	nop
1:	ba	0f			! 0x02E0
	nop
	nop
	nop
0:	ba	1f			! 0x02F0
	nop
	nop
	nop

1:	ba	0f			! 0x0300
	nop
	nop
	nop
0:	ba	1f			! 0x0310
	nop
	nop
	nop
1:	ba	0f			! 0x0320
	nop
	nop
	nop
0:	ba	1f			! 0x0330
	nop
	nop
	nop
1:	ba	0f			! 0x0340
	nop
	nop
	nop
0:	ba	1f			! 0x0350
	nop
	nop
	nop
1:	ba	0f			! 0x0360
	nop
	nop
	nop
0:	ba	1f			! 0x0370
	nop
	nop
	nop
1:	ba	0f			! 0x0380
	nop
	nop
	nop
0:	ba	1f			! 0x0390
	nop
	nop
	nop
1:	ba	0f			! 0x03A0
	nop
	nop
	nop
0:	ba	1f			! 0x03B0
	nop
	nop
	nop
1:	ba	0f			! 0x03C0
	nop
	nop
	nop
0:	ba	1f			! 0x03D0
	nop
	nop
	nop
1:	ba	0f			! 0x03E0
	nop
	nop
	nop
0:	ba	1f			! 0x03F0
	nop
	nop
	nop

1:	ba	0f			! 0x0400
	nop
	nop
	nop
0:	ba	1f			! 0x0410
	nop
	nop
	nop
1:	ba	0f			! 0x0420
	nop
	nop
	nop
0:	ba	1f			! 0x0430
	nop
	nop
	nop
1:	ba	0f			! 0x0440
	nop
	nop
	nop
0:	ba	1f			! 0x0450
	nop
	nop
	nop
1:	ba	0f			! 0x0460
	nop
	nop
	nop
0:	ba	1f			! 0x0470
	nop
	nop
	nop
1:	ba	0f			! 0x0480
	nop
	nop
	nop
0:	ba	1f			! 0x0490
	nop
	nop
	nop
1:	ba	0f			! 0x04A0
	nop
	nop
	nop
0:	ba	1f			! 0x04B0
	nop
	nop
	nop
1:	ba	0f			! 0x04C0
	nop
	nop
	nop
0:	ba	1f			! 0x04D0
	nop
	nop
	nop
1:	ba	0f			! 0x04E0
	nop
	nop
	nop
0:	ba	1f			! 0x04F0
	nop
	nop
	nop

1:	ba	0f			! 0x0500
	nop
	nop
	nop
0:	ba	1f			! 0x0510
	nop
	nop
	nop
1:	ba	0f			! 0x0520
	nop
	nop
	nop
0:	ba	1f			! 0x0530
	nop
	nop
	nop
1:	ba	0f			! 0x0540
	nop
	nop
	nop
0:	ba	1f			! 0x0550
	nop
	nop
	nop
1:	ba	0f			! 0x0560
	nop
	nop
	nop
0:	ba	1f			! 0x0570
	nop
	nop
	nop
1:	ba	0f			! 0x0580
	nop
	nop
	nop
0:	ba	1f			! 0x0590
	nop
	nop
	nop
1:	ba	0f			! 0x05A0
	nop
	nop
	nop
0:	ba	1f			! 0x05B0
	nop
	nop
	nop
1:	ba	0f			! 0x05C0
	nop
	nop
	nop
0:	ba	1f			! 0x05D0
	nop
	nop
	nop
1:	ba	0f			! 0x05E0
	nop
	nop
	nop
0:	ba	1f			! 0x05F0
	nop
	nop
	nop

1:	ba	0f			! 0x0600
	nop
	nop
	nop
0:	ba	1f			! 0x0610
	nop
	nop
	nop
1:	ba	0f			! 0x0620
	nop
	nop
	nop
0:	ba	1f			! 0x0630
	nop
	nop
	nop
1:	ba	0f			! 0x0640
	nop
	nop
	nop
0:	ba	1f			! 0x0650
	nop
	nop
	nop
1:	ba	0f			! 0x0660
	nop
	nop
	nop
0:	ba	1f			! 0x0670
	nop
	nop
	nop
1:	ba	0f			! 0x0680
	nop
	nop
	nop
0:	ba	1f			! 0x0690
	nop
	nop
	nop
1:	ba	0f			! 0x06A0
	nop
	nop
	nop
0:	ba	1f			! 0x06B0
	nop
	nop
	nop
1:	ba	0f			! 0x06C0
	nop
	nop
	nop
0:	ba	1f			! 0x06D0
	nop
	nop
	nop
1:	ba	0f			! 0x06E0
	nop
	nop
	nop
0:	ba	1f			! 0x06F0
	nop
	nop
	nop

1:	ba	0f			! 0x0700
	nop
	nop
	nop
0:	ba	1f			! 0x0710
	nop
	nop
	nop
1:	ba	0f			! 0x0720
	nop
	nop
	nop
0:	ba	1f			! 0x0730
	nop
	nop
	nop
1:	ba	0f			! 0x0740
	nop
	nop
	nop
0:	ba	1f			! 0x0750
	nop
	nop
	nop
1:	ba	0f			! 0x0760
	nop
	nop
	nop
0:	ba	1f			! 0x0770
	nop
	nop
	nop
1:	ba	0f			! 0x0780
	nop
	nop
	nop
0:	ba	1f			! 0x0790
	nop
	nop
	nop
1:	ba	0f			! 0x07A0
	nop
	nop
	nop
0:	ba	1f			! 0x07B0
	nop
	nop
	nop
1:	ba	0f			! 0x07C0
	nop
	nop
	nop
0:	ba	1f			! 0x07D0
	nop
	nop
	nop
1:	ba	0f			! 0x07E0
	nop
	nop
	nop
0:	ba	1f			! 0x07F0
	nop
	nop
	nop

1:	ba	0f			! 0x0800
	nop
	nop
	nop
0:	ba	1f			! 0x0810
	nop
	nop
	nop
1:	ba	0f			! 0x0820
	nop
	nop
	nop
0:	ba	1f			! 0x0830
	nop
	nop
	nop
1:	ba	0f			! 0x0840
	nop
	nop
	nop
0:	ba	1f			! 0x0850
	nop
	nop
	nop
1:	ba	0f			! 0x0860
	nop
	nop
	nop
0:	ba	1f			! 0x0870
	nop
	nop
	nop
1:	ba	0f			! 0x0880
	nop
	nop
	nop
0:	ba	1f			! 0x0890
	nop
	nop
	nop
1:	ba	0f			! 0x08A0
	nop
	nop
	nop
0:	ba	1f			! 0x08B0
	nop
	nop
	nop
1:	ba	0f			! 0x08C0
	nop
	nop
	nop
0:	ba	1f			! 0x08D0
	nop
	nop
	nop
1:	ba	0f			! 0x08E0
	nop
	nop
	nop
0:	ba	1f			! 0x08F0
	nop
	nop
	nop

1:	ba	0f			! 0x0900
	nop
	nop
	nop
0:	ba	1f			! 0x0910
	nop
	nop
	nop
1:	ba	0f			! 0x0920
	nop
	nop
	nop
0:	ba	1f			! 0x0930
	nop
	nop
	nop
1:	ba	0f			! 0x0940
	nop
	nop
	nop
0:	ba	1f			! 0x0950
	nop
	nop
	nop
1:	ba	0f			! 0x0960
	nop
	nop
	nop
0:	ba	1f			! 0x0970
	nop
	nop
	nop
1:	ba	0f			! 0x0980
	nop
	nop
	nop
0:	ba	1f			! 0x0990
	nop
	nop
	nop
1:	ba	0f			! 0x09A0
	nop
	nop
	nop
0:	ba	1f			! 0x09B0
	nop
	nop
	nop
1:	ba	0f			! 0x09C0
	nop
	nop
	nop
0:	ba	1f			! 0x09D0
	nop
	nop
	nop
1:	ba	0f			! 0x09E0
	nop
	nop
	nop
0:	ba	1f			! 0x09F0
	nop
	nop
	nop

1:	ba	0f			! 0x0A00
	nop
	nop
	nop
0:	ba	1f			! 0x0A10
	nop
	nop
	nop
1:	ba	0f			! 0x0A20
	nop
	nop
	nop
0:	ba	1f			! 0x0A30
	nop
	nop
	nop
1:	ba	0f			! 0x0A40
	nop
	nop
	nop
0:	ba	1f			! 0x0A50
	nop
	nop
	nop
1:	ba	0f			! 0x0A60
	nop
	nop
	nop
0:	ba	1f			! 0x0A70
	nop
	nop
	nop
1:	ba	0f			! 0x0A80
	nop
	nop
	nop
0:	ba	1f			! 0x0A90
	nop
	nop
	nop
1:	ba	0f			! 0x0AA0
	nop
	nop
	nop
0:	ba	1f			! 0x0AB0
	nop
	nop
	nop
1:	ba	0f			! 0x0AC0
	nop
	nop
	nop
0:	ba	1f			! 0x0AD0
	nop
	nop
	nop
1:	ba	0f			! 0x0AE0
	nop
	nop
	nop
0:	ba	1f			! 0x0AF0
	nop
	nop
	nop

1:	ba	0f			! 0x0B00
	nop
	nop
	nop
0:	ba	1f			! 0x0B10
	nop
	nop
	nop
1:	ba	0f			! 0x0B20
	nop
	nop
	nop
0:	ba	1f			! 0x0B30
	nop
	nop
	nop
1:	ba	0f			! 0x0B40
	nop
	nop
	nop
0:	ba	1f			! 0x0B50
	nop
	nop
	nop
1:	ba	0f			! 0x0B60
	nop
	nop
	nop
0:	ba	1f			! 0x0B70
	nop
	nop
	nop
1:	ba	0f			! 0x0B80
	nop
	nop
	nop
0:	ba	1f			! 0x0B90
	nop
	nop
	nop
1:	ba	0f			! 0x0BA0
	nop
	nop
	nop
0:	ba	1f			! 0x0BB0
	nop
	nop
	nop
1:	ba	0f			! 0x0BC0
	nop
	nop
	nop
0:	ba	1f			! 0x0BD0
	nop
	nop
	nop
1:	ba	0f			! 0x0BE0
	nop
	nop
	nop
0:	ba	1f			! 0x0BF0
	nop
	nop
	nop

1:	ba	0f			! 0x0C00
	nop
	nop
	nop
0:	ba	1f			! 0x0C10
	nop
	nop
	nop
1:	ba	0f			! 0x0C20
	nop
	nop
	nop
0:	ba	1f			! 0x0C30
	nop
	nop
	nop
1:	ba	0f			! 0x0C40
	nop
	nop
	nop
0:	ba	1f			! 0x0C50
	nop
	nop
	nop
1:	ba	0f			! 0x0C60
	nop
	nop
	nop
0:	ba	1f			! 0x0C70
	nop
	nop
	nop
1:	ba	0f			! 0x0C80
	nop
	nop
	nop
0:	ba	1f			! 0x0C90
	nop
	nop
	nop
1:	ba	0f			! 0x0CA0
	nop
	nop
	nop
0:	ba	1f			! 0x0CB0
	nop
	nop
	nop
1:	ba	0f			! 0x0CC0
	nop
	nop
	nop
0:	ba	1f			! 0x0CD0
	nop
	nop
	nop
1:	ba	0f			! 0x0CE0
	nop
	nop
	nop
0:	ba	1f			! 0x0CF0
	nop
	nop
	nop

1:	ba	0f			! 0x0D00
	nop
	nop
	nop
0:	ba	1f			! 0x0D10
	nop
	nop
	nop
1:	ba	0f			! 0x0D20
	nop
	nop
	nop
0:	ba	1f			! 0x0D30
	nop
	nop
	nop
1:	ba	0f			! 0x0D40
	nop
	nop
	nop
0:	ba	1f			! 0x0D50
	nop
	nop
	nop
1:	ba	0f			! 0x0D60
	nop
	nop
	nop
0:	ba	1f			! 0x0D70
	nop
	nop
	nop
1:	ba	0f			! 0x0D80
	nop
	nop
	nop
0:	ba	1f			! 0x0D90
	nop
	nop
	nop
1:	ba	0f			! 0x0DA0
	nop
	nop
	nop
0:	ba	1f			! 0x0DB0
	nop
	nop
	nop
1:	ba	0f			! 0x0DC0
	nop
	nop
	nop
0:	ba	1f			! 0x0DD0
	nop
	nop
	nop
1:	ba	0f			! 0x0DE0
	nop
	nop
	nop
0:	ba	1f			! 0x0DF0
	nop
	nop
	nop

1:	ba	0f			! 0x0E00
	nop
	nop
	nop
0:	ba	1f			! 0x0E10
	nop
	nop
	nop
1:	ba	0f			! 0x0E20
	nop
	nop
	nop
0:	ba	1f			! 0x0E30
	nop
	nop
	nop
1:	ba	0f			! 0x0E40
	nop
	nop
	nop
0:	ba	1f			! 0x0E50
	nop
	nop
	nop
1:	ba	0f			! 0x0E60
	nop
	nop
	nop
0:	ba	1f			! 0x0E70
	nop
	nop
	nop
1:	ba	0f			! 0x0E80
	nop
	nop
	nop
0:	ba	1f			! 0x0E90
	nop
	nop
	nop
1:	ba	0f			! 0x0EA0
	nop
	nop
	nop
0:	ba	1f			! 0x0EB0
	nop
	nop
	nop
1:	ba	0f			! 0x0EC0
	nop
	nop
	nop
0:	ba	1f			! 0x0ED0
	nop
	nop
	nop
1:	ba	0f			! 0x0EE0
	nop
	nop
	nop
0:	ba	1f			! 0x0EF0
	nop
	nop
	nop

1:	ba	0f			! 0x0F00
	nop
	nop
	nop
0:	ba	1f			! 0x0F10
	nop
	nop
	nop
1:	ba	0f			! 0x0F20
	nop
	nop
	nop
0:	ba	1f			! 0x0F30
	nop
	nop
	nop
1:	ba	0f			! 0x0F40
	nop
	nop
	nop
0:	ba	1f			! 0x0F50
	nop
	nop
	nop
1:	ba	0f			! 0x0F60
	nop
	nop
	nop
0:	ba	1f			! 0x0F70
	nop
	nop
	nop
1:	ba	0f			! 0x0F80
	nop
	nop
	nop
0:	ba	1f			! 0x0F90
	nop
	nop
	nop
1:	ba	0f			! 0x0FA0
	nop
	nop
	nop
0:	ba	1f			! 0x0FB0
	nop
	nop
	nop
1:	ba	0f			! 0x0FC0
	nop
	nop
	nop
0:	ba	1f			! 0x0FD0
	nop
	nop
	nop
1:	ba	0f			! 0x0FE0
	nop
	nop
	nop
0:	ba	1f			! 0x0FF0
	nop
	nop
	nop

! *********************** 4KB ***************************************

1:	ba	0f			! 0x1000
	nop
	nop
	nop
0:	ba	1f			! 0x1010
	nop
	nop
	nop
1:	ba	0f			! 0x1020
	nop
	nop
	nop
0:	ba	1f			! 0x1030
	nop
	nop
	nop
1:	ba	0f			! 0x1040
	nop
	nop
	nop
0:	ba	1f			! 0x1050
	nop
	nop
	nop
1:	ba	0f			! 0x1060
	nop
	nop
	nop
0:	ba	1f			! 0x1070
	nop
	nop
	nop
1:	ba	0f			! 0x1080
	nop
	nop
	nop
0:	ba	1f			! 0x1090
	nop
	nop
	nop
1:	ba	0f			! 0x10A0
	nop
	nop
	nop
0:	ba	1f			! 0x10B0
	nop
	nop
	nop
1:	ba	0f			! 0x10C0
	nop
	nop
	nop
0:	ba	1f			! 0x10D0
	nop
	nop
	nop
1:	ba	0f			! 0x10E0
	nop
	nop
	nop
0:	ba	1f			! 0x10F0
	nop
	nop
	nop

1:	ba	0f			! 0x1100
	nop
	nop
	nop
0:	ba	1f			! 0x1110
	nop
	nop
	nop
1:	ba	0f			! 0x1120
	nop
	nop
	nop
0:	ba	1f			! 0x1130
	nop
	nop
	nop
1:	ba	0f			! 0x1140
	nop
	nop
	nop
0:	ba	1f			! 0x1150
	nop
	nop
	nop
1:	ba	0f			! 0x1160
	nop
	nop
	nop
0:	ba	1f			! 0x1170
	nop
	nop
	nop
1:	ba	0f			! 0x1180
	nop
	nop
	nop
0:	ba	1f			! 0x1190
	nop
	nop
	nop
1:	ba	0f			! 0x11A0
	nop
	nop
	nop
0:	ba	1f			! 0x11B0
	nop
	nop
	nop
1:	ba	0f			! 0x11C0
	nop
	nop
	nop
0:	ba	1f			! 0x11D0
	nop
	nop
	nop
1:	ba	0f			! 0x11E0
	nop
	nop
	nop
0:	ba	1f			! 0x11F0
	nop
	nop
	nop

1:	ba	0f			! 0x1200
	nop
	nop
	nop
0:	ba	1f			! 0x1210
	nop
	nop
	nop
1:	ba	0f			! 0x1220
	nop
	nop
	nop
0:	ba	1f			! 0x1230
	nop
	nop
	nop
1:	ba	0f			! 0x1240
	nop
	nop
	nop
0:	ba	1f			! 0x1250
	nop
	nop
	nop
1:	ba	0f			! 0x1260
	nop
	nop
	nop
0:	ba	1f			! 0x1270
	nop
	nop
	nop
1:	ba	0f			! 0x1280
	nop
	nop
	nop
0:	ba	1f			! 0x1290
	nop
	nop
	nop
1:	ba	0f			! 0x12A0
	nop
	nop
	nop
0:	ba	1f			! 0x12B0
	nop
	nop
	nop
1:	ba	0f			! 0x12C0
	nop
	nop
	nop
0:	ba	1f			! 0x12D0
	nop
	nop
	nop
1:	ba	0f			! 0x12E0
	nop
	nop
	nop
0:	ba	1f			! 0x12F0
	nop
	nop
	nop

1:	ba	0f			! 0x1300
	nop
	nop
	nop
0:	ba	1f			! 0x1310
	nop
	nop
	nop
1:	ba	0f			! 0x1320
	nop
	nop
	nop
0:	ba	1f			! 0x1330
	nop
	nop
	nop
1:	ba	0f			! 0x1340
	nop
	nop
	nop
0:	ba	1f			! 0x1350
	nop
	nop
	nop
1:	ba	0f			! 0x1360
	nop
	nop
	nop
0:	ba	1f			! 0x1370
	nop
	nop
	nop
1:	ba	0f			! 0x1380
	nop
	nop
	nop
0:	ba	1f			! 0x1390
	nop
	nop
	nop
1:	ba	0f			! 0x13A0
	nop
	nop
	nop
0:	ba	1f			! 0x13B0
	nop
	nop
	nop
1:	ba	0f			! 0x13C0
	nop
	nop
	nop
0:	ba	1f			! 0x13D0
	nop
	nop
	nop
1:	ba	0f			! 0x13E0
	nop
	nop
	nop
0:	ba	1f			! 0x13F0
	nop
	nop
	nop

1:	ba	0f			! 0x1400
	nop
	nop
	nop
0:	ba	1f			! 0x1410
	nop
	nop
	nop
1:	ba	0f			! 0x1420
	nop
	nop
	nop
0:	ba	1f			! 0x1430
	nop
	nop
	nop
1:	ba	0f			! 0x1440
	nop
	nop
	nop
0:	ba	1f			! 0x1450
	nop
	nop
	nop
1:	ba	0f			! 0x1460
	nop
	nop
	nop
0:	ba	1f			! 0x1470
	nop
	nop
	nop
1:	ba	0f			! 0x1480
	nop
	nop
	nop
0:	ba	1f			! 0x1490
	nop
	nop
	nop
1:	ba	0f			! 0x14A0
	nop
	nop
	nop
0:	ba	1f			! 0x14B0
	nop
	nop
	nop
1:	ba	0f			! 0x14C0
	nop
	nop
	nop
0:	ba	1f			! 0x14D0
	nop
	nop
	nop
1:	ba	0f			! 0x14E0
	nop
	nop
	nop
0:	ba	1f			! 0x14F0
	nop
	nop
	nop

1:	ba	0f			! 0x1500
	nop
	nop
	nop
0:	ba	1f			! 0x1510
	nop
	nop
	nop
1:	ba	0f			! 0x1520
	nop
	nop
	nop
0:	ba	1f			! 0x1530
	nop
	nop
	nop
1:	ba	0f			! 0x1540
	nop
	nop
	nop
0:	ba	1f			! 0x1550
	nop
	nop
	nop
1:	ba	0f			! 0x1560
	nop
	nop
	nop
0:	ba	1f			! 0x1570
	nop
	nop
	nop
1:	ba	0f			! 0x1580
	nop
	nop
	nop
0:	ba	1f			! 0x1590
	nop
	nop
	nop
1:	ba	0f			! 0x15A0
	nop
	nop
	nop
0:	ba	1f			! 0x15B0
	nop
	nop
	nop
1:	ba	0f			! 0x15C0
	nop
	nop
	nop
0:	ba	1f			! 0x15D0
	nop
	nop
	nop
1:	ba	0f			! 0x15E0
	nop
	nop
	nop
0:	ba	1f			! 0x15F0
	nop
	nop
	nop

1:	ba	0f			! 0x1600
	nop
	nop
	nop
0:	ba	1f			! 0x1610
	nop
	nop
	nop
1:	ba	0f			! 0x1620
	nop
	nop
	nop
0:	ba	1f			! 0x1630
	nop
	nop
	nop
1:	ba	0f			! 0x1640
	nop
	nop
	nop
0:	ba	1f			! 0x1650
	nop
	nop
	nop
1:	ba	0f			! 0x1660
	nop
	nop
	nop
0:	ba	1f			! 0x1670
	nop
	nop
	nop
1:	ba	0f			! 0x1680
	nop
	nop
	nop
0:	ba	1f			! 0x1690
	nop
	nop
	nop
1:	ba	0f			! 0x16A0
	nop
	nop
	nop
0:	ba	1f			! 0x16B0
	nop
	nop
	nop
1:	ba	0f			! 0x16C0
	nop
	nop
	nop
0:	ba	1f			! 0x16D0
	nop
	nop
	nop
1:	ba	0f			! 0x16E0
	nop
	nop
	nop
0:	ba	1f			! 0x16F0
	nop
	nop
	nop

1:	ba	0f			! 0x1700
	nop
	nop
	nop
0:	ba	1f			! 0x1710
	nop
	nop
	nop
1:	ba	0f			! 0x1720
	nop
	nop
	nop
0:	ba	1f			! 0x1730
	nop
	nop
	nop
1:	ba	0f			! 0x1740
	nop
	nop
	nop
0:	ba	1f			! 0x1750
	nop
	nop
	nop
1:	ba	0f			! 0x1760
	nop
	nop
	nop
0:	ba	1f			! 0x1770
	nop
	nop
	nop
1:	ba	0f			! 0x1780
	nop
	nop
	nop
0:	ba	1f			! 0x1790
	nop
	nop
	nop
1:	ba	0f			! 0x17A0
	nop
	nop
	nop
0:	ba	1f			! 0x17B0
	nop
	nop
	nop
1:	ba	0f			! 0x17C0
	nop
	nop
	nop
0:	ba	1f			! 0x17D0
	nop
	nop
	nop
1:	ba	0f			! 0x17E0
	nop
	nop
	nop
0:	ba	1f			! 0x17F0
	nop
	nop
	nop

1:	ba	0f			! 0x1800
	nop
	nop
	nop
0:	ba	1f			! 0x1810
	nop
	nop
	nop
1:	ba	0f			! 0x1820
	nop
	nop
	nop
0:	ba	1f			! 0x1830
	nop
	nop
	nop
1:	ba	0f			! 0x1840
	nop
	nop
	nop
0:	ba	1f			! 0x1850
	nop
	nop
	nop
1:	ba	0f			! 0x1860
	nop
	nop
	nop
0:	ba	1f			! 0x1870
	nop
	nop
	nop
1:	ba	0f			! 0x1880
	nop
	nop
	nop
0:	ba	1f			! 0x1890
	nop
	nop
	nop
1:	ba	0f			! 0x18A0
	nop
	nop
	nop
0:	ba	1f			! 0x18B0
	nop
	nop
	nop
1:	ba	0f			! 0x18C0
	nop
	nop
	nop
0:	ba	1f			! 0x18D0
	nop
	nop
	nop
1:	ba	0f			! 0x18E0
	nop
	nop
	nop
0:	ba	1f			! 0x18F0
	nop
	nop
	nop

1:	ba	0f			! 0x1900
	nop
	nop
	nop
0:	ba	1f			! 0x1910
	nop
	nop
	nop
1:	ba	0f			! 0x1920
	nop
	nop
	nop
0:	ba	1f			! 0x1930
	nop
	nop
	nop
1:	ba	0f			! 0x1940
	nop
	nop
	nop
0:	ba	1f			! 0x1950
	nop
	nop
	nop
1:	ba	0f			! 0x1960
	nop
	nop
	nop
0:	ba	1f			! 0x1970
	nop
	nop
	nop
1:	ba	0f			! 0x1980
	nop
	nop
	nop
0:	ba	1f			! 0x1990
	nop
	nop
	nop
1:	ba	0f			! 0x19A0
	nop
	nop
	nop
0:	ba	1f			! 0x19B0
	nop
	nop
	nop
1:	ba	0f			! 0x19C0
	nop
	nop
	nop
0:	ba	1f			! 0x19D0
	nop
	nop
	nop
1:	ba	0f			! 0x19E0
	nop
	nop
	nop
0:	ba	1f			! 0x19F0
	nop
	nop
	nop

1:	ba	0f			! 0x1A00
	nop
	nop
	nop
0:	ba	1f			! 0x1A10
	nop
	nop
	nop
1:	ba	0f			! 0x1A20
	nop
	nop
	nop
0:	ba	1f			! 0x1A30
	nop
	nop
	nop
1:	ba	0f			! 0x1A40
	nop
	nop
	nop
0:	ba	1f			! 0x1A50
	nop
	nop
	nop
1:	ba	0f			! 0x1A60
	nop
	nop
	nop
0:	ba	1f			! 0x1A70
	nop
	nop
	nop
1:	ba	0f			! 0x1A80
	nop
	nop
	nop
0:	ba	1f			! 0x1A90
	nop
	nop
	nop
1:	ba	0f			! 0x1AA0
	nop
	nop
	nop
0:	ba	1f			! 0x1AB0
	nop
	nop
	nop
1:	ba	0f			! 0x1AC0
	nop
	nop
	nop
0:	ba	1f			! 0x1AD0
	nop
	nop
	nop
1:	ba	0f			! 0x1AE0
	nop
	nop
	nop
0:	ba	1f			! 0x1AF0
	nop
	nop
	nop

1:	ba	0f			! 0x1B00
	nop
	nop
	nop
0:	ba	1f			! 0x1B10
	nop
	nop
	nop
1:	ba	0f			! 0x1B20
	nop
	nop
	nop
0:	ba	1f			! 0x1B30
	nop
	nop
	nop
1:	ba	0f			! 0x1B40
	nop
	nop
	nop
0:	ba	1f			! 0x1B50
	nop
	nop
	nop
1:	ba	0f			! 0x1B60
	nop
	nop
	nop
0:	ba	1f			! 0x1B70
	nop
	nop
	nop
1:	ba	0f			! 0x1B80
	nop
	nop
	nop
0:	ba	1f			! 0x1B90
	nop
	nop
	nop
1:	ba	0f			! 0x1BA0
	nop
	nop
	nop
0:	ba	1f			! 0x1BB0
	nop
	nop
	nop
1:	ba	0f			! 0x1BC0
	nop
	nop
	nop
0:	ba	1f			! 0x1BD0
	nop
	nop
	nop
1:	ba	0f			! 0x1BE0
	nop
	nop
	nop
0:	ba	1f			! 0x1BF0
	nop
	nop
	nop

1:	ba	0f			! 0x1C00
	nop
	nop
	nop
0:	ba	1f			! 0x1C10
	nop
	nop
	nop
1:	ba	0f			! 0x1C20
	nop
	nop
	nop
0:	ba	1f			! 0x1C30
	nop
	nop
	nop
1:	ba	0f			! 0x1C40
	nop
	nop
	nop
0:	ba	1f			! 0x1C50
	nop
	nop
	nop
1:	ba	0f			! 0x1C60
	nop
	nop
	nop
0:	ba	1f			! 0x1C70
	nop
	nop
	nop
1:	ba	0f			! 0x1C80
	nop
	nop
	nop
0:	ba	1f			! 0x1C90
	nop
	nop
	nop
1:	ba	0f			! 0x1CA0
	nop
	nop
	nop
0:	ba	1f			! 0x1CB0
	nop
	nop
	nop
1:	ba	0f			! 0x1CC0
	nop
	nop
	nop
0:	ba	1f			! 0x1CD0
	nop
	nop
	nop
1:	ba	0f			! 0x1CE0
	nop
	nop
	nop
0:	ba	1f			! 0x1CF0
	nop
	nop
	nop

1:	ba	0f			! 0x1D00
	nop
	nop
	nop
0:	ba	1f			! 0x1D10
	nop
	nop
	nop
1:	ba	0f			! 0x1D20
	nop
	nop
	nop
0:	ba	1f			! 0x1D30
	nop
	nop
	nop
1:	ba	0f			! 0x1D40
	nop
	nop
	nop
0:	ba	1f			! 0x1D50
	nop
	nop
	nop
1:	ba	0f			! 0x1D60
	nop
	nop
	nop
0:	ba	1f			! 0x1D70
	nop
	nop
	nop
1:	ba	0f			! 0x1D80
	nop
	nop
	nop
0:	ba	1f			! 0x1D90
	nop
	nop
	nop
1:	ba	0f			! 0x1DA0
	nop
	nop
	nop
0:	ba	1f			! 0x1DB0
	nop
	nop
	nop
1:	ba	0f			! 0x1DC0
	nop
	nop
	nop
0:	ba	1f			! 0x1DD0
	nop
	nop
	nop
1:	ba	0f			! 0x1DE0
	nop
	nop
	nop
0:	ba	1f			! 0x1DF0
	nop
	nop
	nop

1:	ba	0f			! 0x1E00
	nop
	nop
	nop
0:	ba	1f			! 0x1E10
	nop
	nop
	nop
1:	ba	0f			! 0x1E20
	nop
	nop
	nop
0:	ba	1f			! 0x1E30
	nop
	nop
	nop
1:	ba	0f			! 0x1E40
	nop
	nop
	nop
0:	ba	1f			! 0x1E50
	nop
	nop
	nop
1:	ba	0f			! 0x1E60
	nop
	nop
	nop
0:	ba	1f			! 0x1E70
	nop
	nop
	nop
1:	ba	0f			! 0x1E80
	nop
	nop
	nop
0:	ba	1f			! 0x1E90
	nop
	nop
	nop
1:	ba	0f			! 0x1EA0
	nop
	nop
	nop
0:	ba	1f			! 0x1EB0
	nop
	nop
	nop
1:	ba	0f			! 0x1EC0
	nop
	nop
	nop
0:	ba	1f			! 0x1ED0
	nop
	nop
	nop
1:	ba	0f			! 0x1EE0
	nop
	nop
	nop
0:	ba	1f			! 0x1EF0
	nop
	nop
	nop

1:	ba	0f			! 0x1F00
	nop
	nop
	nop
0:	ba	1f			! 0x1F10
	nop
	nop
	nop
1:	ba	0f			! 0x1F20
	nop
	nop
	nop
0:	ba	1f			! 0x1F30
	nop
	nop
	nop
1:	ba	0f			! 0x1F40
	nop
	nop
	nop
0:	ba	1f			! 0x1F50
	nop
	nop
	nop
1:	ba	0f			! 0x1F60
	nop
	nop
	nop
0:	ba	1f			! 0x1F70
	nop
	nop
	nop
1:	ba	0f			! 0x1F80
	nop
	nop
	nop
0:	ba	1f			! 0x1F90
	nop
	nop
	nop
1:	ba	0f			! 0x1FA0
	nop
	nop
	nop
0:	ba	1f			! 0x1FB0
	nop
	nop
	nop
1:	ba	0f			! 0x1FC0
	nop
	nop
	nop
0:	ba	1f			! 0x1FD0
	nop
	nop
	nop
1:	ba	0f			! 0x1FE0
	nop
	nop
	nop
0:	nop				! 0x1FF0
	nop
	nop
	nop


! *********************** 8KB ***************************************

1:	ba	0f			! 0x2000
	nop
	nop
	nop
0:	ba	1f			! 0x2010
	nop
	nop
	nop
1:	ba	0f			! 0x2020
	nop
	nop
	nop
0:	ba	1f			! 0x2030
	nop
	nop
	nop
1:	ba	0f			! 0x2040
	nop
	nop
	nop
0:	ba	1f			! 0x2050
	nop
	nop
	nop
1:	ba	0f			! 0x2060
	nop
	nop
	nop
0:	ba	1f			! 0x2070
	nop
	nop
	nop
1:	ba	0f			! 0x2080
	nop
	nop
	nop
0:	ba	1f			! 0x2090
	nop
	nop
	nop
1:	ba	0f			! 0x20A0
	nop
	nop
	nop
0:	ba	1f			! 0x20B0
	nop
	nop
	nop
1:	ba	0f			! 0x20C0
	nop
	nop
	nop
0:	ba	1f			! 0x20D0
	nop
	nop
	nop
1:	ba	0f			! 0x20E0
	nop
	nop
	nop
0:	ba	1f			! 0x20F0
	nop
	nop
	nop

1:	ba	0f			! 0x2100
	nop
	nop
	nop
0:	ba	1f			! 0x2110
	nop
	nop
	nop
1:	ba	0f			! 0x2120
	nop
	nop
	nop
0:	ba	1f			! 0x2130
	nop
	nop
	nop
1:	ba	0f			! 0x2140
	nop
	nop
	nop
0:	ba	1f			! 0x2150
	nop
	nop
	nop
1:	ba	0f			! 0x2160
	nop
	nop
	nop
0:	ba	1f			! 0x2170
	nop
	nop
	nop
1:	ba	0f			! 0x2180
	nop
	nop
	nop
0:	ba	1f			! 0x2190
	nop
	nop
	nop
1:	ba	0f			! 0x21A0
	nop
	nop
	nop
0:	ba	1f			! 0x21B0
	nop
	nop
	nop
1:	ba	0f			! 0x21C0
	nop
	nop
	nop
0:	ba	1f			! 0x21D0
	nop
	nop
	nop
1:	ba	0f			! 0x21E0
	nop
	nop
	nop
0:	ba	1f			! 0x21F0
	nop
	nop
	nop

1:	ba	0f			! 0x2200
	nop
	nop
	nop
0:	ba	1f			! 0x2210
	nop
	nop
	nop
1:	ba	0f			! 0x2220
	nop
	nop
	nop
0:	ba	1f			! 0x2230
	nop
	nop
	nop
1:	ba	0f			! 0x2240
	nop
	nop
	nop
0:	ba	1f			! 0x2250
	nop
	nop
	nop
1:	ba	0f			! 0x2260
	nop
	nop
	nop
0:	ba	1f			! 0x2270
	nop
	nop
	nop
1:	ba	0f			! 0x2280
	nop
	nop
	nop
0:	ba	1f			! 0x2290
	nop
	nop
	nop
1:	ba	0f			! 0x22A0
	nop
	nop
	nop
0:	ba	1f			! 0x22B0
	nop
	nop
	nop
1:	ba	0f			! 0x22C0
	nop
	nop
	nop
0:	ba	1f			! 0x22D0
	nop
	nop
	nop
1:	ba	0f			! 0x22E0
	nop
	nop
	nop
0:	ba	1f			! 0x22F0
	nop
	nop
	nop

1:	ba	0f			! 0x2300
	nop
	nop
	nop
0:	ba	1f			! 0x2310
	nop
	nop
	nop
1:	ba	0f			! 0x2320
	nop
	nop
	nop
0:	ba	1f			! 0x2330
	nop
	nop
	nop
1:	ba	0f			! 0x2340
	nop
	nop
	nop
0:	ba	1f			! 0x2350
	nop
	nop
	nop
1:	ba	0f			! 0x2360
	nop
	nop
	nop
0:	ba	1f			! 0x2370
	nop
	nop
	nop
1:	ba	0f			! 0x2380
	nop
	nop
	nop
0:	ba	1f			! 0x2390
	nop
	nop
	nop
1:	ba	0f			! 0x23A0
	nop
	nop
	nop
0:	ba	1f			! 0x23B0
	nop
	nop
	nop
1:	ba	0f			! 0x23C0
	nop
	nop
	nop
0:	ba	1f			! 0x23D0
	nop
	nop
	nop
1:	ba	0f			! 0x23E0
	nop
	nop
	nop
0:	ba	1f			! 0x23F0
	nop
	nop
	nop

1:	ba	0f			! 0x2400
	nop
	nop
	nop
0:	ba	1f			! 0x2410
	nop
	nop
	nop
1:	ba	0f			! 0x2420
	nop
	nop
	nop
0:	ba	1f			! 0x2430
	nop
	nop
	nop
1:	ba	0f			! 0x2440
	nop
	nop
	nop
0:	ba	1f			! 0x2450
	nop
	nop
	nop
1:	ba	0f			! 0x2460
	nop
	nop
	nop
0:	ba	1f			! 0x2470
	nop
	nop
	nop
1:	ba	0f			! 0x2480
	nop
	nop
	nop
0:	ba	1f			! 0x2490
	nop
	nop
	nop
1:	ba	0f			! 0x24A0
	nop
	nop
	nop
0:	ba	1f			! 0x24B0
	nop
	nop
	nop
1:	ba	0f			! 0x24C0
	nop
	nop
	nop
0:	ba	1f			! 0x24D0
	nop
	nop
	nop
1:	ba	0f			! 0x24E0
	nop
	nop
	nop
0:	ba	1f			! 0x24F0
	nop
	nop
	nop

1:	ba	0f			! 0x2500
	nop
	nop
	nop
0:	ba	1f			! 0x2510
	nop
	nop
	nop
1:	ba	0f			! 0x2520
	nop
	nop
	nop
0:	ba	1f			! 0x2530
	nop
	nop
	nop
1:	ba	0f			! 0x2540
	nop
	nop
	nop
0:	ba	1f			! 0x2550
	nop
	nop
	nop
1:	ba	0f			! 0x2560
	nop
	nop
	nop
0:	ba	1f			! 0x2570
	nop
	nop
	nop
1:	ba	0f			! 0x2580
	nop
	nop
	nop
0:	ba	1f			! 0x2590
	nop
	nop
	nop
1:	ba	0f			! 0x25A0
	nop
	nop
	nop
0:	ba	1f			! 0x25B0
	nop
	nop
	nop
1:	ba	0f			! 0x25C0
	nop
	nop
	nop
0:	ba	1f			! 0x25D0
	nop
	nop
	nop
1:	ba	0f			! 0x25E0
	nop
	nop
	nop
0:	ba	1f			! 0x25F0
	nop
	nop
	nop

1:	ba	0f			! 0x2600
	nop
	nop
	nop
0:	ba	1f			! 0x2610
	nop
	nop
	nop
1:	ba	0f			! 0x2620
	nop
	nop
	nop
0:	ba	1f			! 0x2630
	nop
	nop
	nop
1:	ba	0f			! 0x2640
	nop
	nop
	nop
0:	ba	1f			! 0x2650
	nop
	nop
	nop
1:	ba	0f			! 0x2660
	nop
	nop
	nop
0:	ba	1f			! 0x2670
	nop
	nop
	nop
1:	ba	0f			! 0x2680
	nop
	nop
	nop
0:	ba	1f			! 0x2690
	nop
	nop
	nop
1:	ba	0f			! 0x26A0
	nop
	nop
	nop
0:	ba	1f			! 0x26B0
	nop
	nop
	nop
1:	ba	0f			! 0x26C0
	nop
	nop
	nop
0:	ba	1f			! 0x26D0
	nop
	nop
	nop
1:	ba	0f			! 0x26E0
	nop
	nop
	nop
0:	ba	1f			! 0x26F0
	nop
	nop
	nop

1:	ba	0f			! 0x2700
	nop
	nop
	nop
0:	ba	1f			! 0x2710
	nop
	nop
	nop
1:	ba	0f			! 0x2720
	nop
	nop
	nop
0:	ba	1f			! 0x2730
	nop
	nop
	nop
1:	ba	0f			! 0x2740
	nop
	nop
	nop
0:	ba	1f			! 0x2750
	nop
	nop
	nop
1:	ba	0f			! 0x2760
	nop
	nop
	nop
0:	ba	1f			! 0x2770
	nop
	nop
	nop
1:	ba	0f			! 0x2780
	nop
	nop
	nop
0:	ba	1f			! 0x2790
	nop
	nop
	nop
1:	ba	0f			! 0x27A0
	nop
	nop
	nop
0:	ba	1f			! 0x27B0
	nop
	nop
	nop
1:	ba	0f			! 0x27C0
	nop
	nop
	nop
0:	ba	1f			! 0x27D0
	nop
	nop
	nop
1:	ba	0f			! 0x27E0
	nop
	nop
	nop
0:	ba	1f			! 0x27F0
	nop
	nop
	nop

1:	ba	0f			! 0x2800
	nop
	nop
	nop
0:	ba	1f			! 0x2810
	nop
	nop
	nop
1:	ba	0f			! 0x2820
	nop
	nop
	nop
0:	ba	1f			! 0x2830
	nop
	nop
	nop
1:	ba	0f			! 0x2840
	nop
	nop
	nop
0:	ba	1f			! 0x2850
	nop
	nop
	nop
1:	ba	0f			! 0x2860
	nop
	nop
	nop
0:	ba	1f			! 0x2870
	nop
	nop
	nop
1:	ba	0f			! 0x2880
	nop
	nop
	nop
0:	ba	1f			! 0x2890
	nop
	nop
	nop
1:	ba	0f			! 0x28A0
	nop
	nop
	nop
0:	ba	1f			! 0x28B0
	nop
	nop
	nop
1:	ba	0f			! 0x28C0
	nop
	nop
	nop
0:	ba	1f			! 0x28D0
	nop
	nop
	nop
1:	ba	0f			! 0x28E0
	nop
	nop
	nop
0:	ba	1f			! 0x28F0
	nop
	nop
	nop

1:	ba	0f			! 0x2900
	nop
	nop
	nop
0:	ba	1f			! 0x2910
	nop
	nop
	nop
1:	ba	0f			! 0x2920
	nop
	nop
	nop
0:	ba	1f			! 0x2930
	nop
	nop
	nop
1:	ba	0f			! 0x2940
	nop
	nop
	nop
0:	ba	1f			! 0x2950
	nop
	nop
	nop
1:	ba	0f			! 0x2960
	nop
	nop
	nop
0:	ba	1f			! 0x2970
	nop
	nop
	nop
1:	ba	0f			! 0x2980
	nop
	nop
	nop
0:	ba	1f			! 0x2990
	nop
	nop
	nop
1:	ba	0f			! 0x29A0
	nop
	nop
	nop
0:	ba	1f			! 0x29B0
	nop
	nop
	nop
1:	ba	0f			! 0x29C0
	nop
	nop
	nop
0:	ba	1f			! 0x29D0
	nop
	nop
	nop
1:	ba	0f			! 0x29E0
	nop
	nop
	nop
0:	ba	1f			! 0x29F0
	nop
	nop
	nop

1:	ba	0f			! 0x2A00
	nop
	nop
	nop
0:	ba	1f			! 0x2A10
	nop
	nop
	nop
1:	ba	0f			! 0x2A20
	nop
	nop
	nop
0:	ba	1f			! 0x2A30
	nop
	nop
	nop
1:	ba	0f			! 0x2A40
	nop
	nop
	nop
0:	ba	1f			! 0x2A50
	nop
	nop
	nop
1:	ba	0f			! 0x2A60
	nop
	nop
	nop
0:	ba	1f			! 0x2A70
	nop
	nop
	nop
1:	ba	0f			! 0x2A80
	nop
	nop
	nop
0:	ba	1f			! 0x2A90
	nop
	nop
	nop
1:	ba	0f			! 0x2AA0
	nop
	nop
	nop
0:	ba	1f			! 0x2AB0
	nop
	nop
	nop
1:	ba	0f			! 0x2AC0
	nop
	nop
	nop
0:	ba	1f			! 0x2AD0
	nop
	nop
	nop
1:	ba	0f			! 0x2AE0
	nop
	nop
	nop
0:	ba	1f			! 0x2AF0
	nop
	nop
	nop

1:	ba	0f			! 0x2B00
	nop
	nop
	nop
0:	ba	1f			! 0x2B10
	nop
	nop
	nop
1:	ba	0f			! 0x2B20
	nop
	nop
	nop
0:	ba	1f			! 0x2B30
	nop
	nop
	nop
1:	ba	0f			! 0x2B40
	nop
	nop
	nop
0:	ba	1f			! 0x2B50
	nop
	nop
	nop
1:	ba	0f			! 0x2B60
	nop
	nop
	nop
0:	ba	1f			! 0x2B70
	nop
	nop
	nop
1:	ba	0f			! 0x2B80
	nop
	nop
	nop
0:	ba	1f			! 0x2B90
	nop
	nop
	nop
1:	ba	0f			! 0x2BA0
	nop
	nop
	nop
0:	ba	1f			! 0x2BB0
	nop
	nop
	nop
1:	ba	0f			! 0x2BC0
	nop
	nop
	nop
0:	ba	1f			! 0x2BD0
	nop
	nop
	nop
1:	ba	0f			! 0x2BE0
	nop
	nop
	nop
0:	ba	1f			! 0x2BF0
	nop
	nop
	nop

1:	ba	0f			! 0x2C00
	nop
	nop
	nop
0:	ba	1f			! 0x2C10
	nop
	nop
	nop
1:	ba	0f			! 0x2C20
	nop
	nop
	nop
0:	ba	1f			! 0x2C30
	nop
	nop
	nop
1:	ba	0f			! 0x2C40
	nop
	nop
	nop
0:	ba	1f			! 0x2C50
	nop
	nop
	nop
1:	ba	0f			! 0x2C60
	nop
	nop
	nop
0:	ba	1f			! 0x2C70
	nop
	nop
	nop
1:	ba	0f			! 0x2C80
	nop
	nop
	nop
0:	ba	1f			! 0x2C90
	nop
	nop
	nop
1:	ba	0f			! 0x2CA0
	nop
	nop
	nop
0:	ba	1f			! 0x2CB0
	nop
	nop
	nop
1:	ba	0f			! 0x2CC0
	nop
	nop
	nop
0:	ba	1f			! 0x2CD0
	nop
	nop
	nop
1:	ba	0f			! 0x2CE0
	nop
	nop
	nop
0:	ba	1f			! 0x2CF0
	nop
	nop
	nop

1:	ba	0f			! 0x2D00
	nop
	nop
	nop
0:	ba	1f			! 0x2D10
	nop
	nop
	nop
1:	ba	0f			! 0x2D20
	nop
	nop
	nop
0:	ba	1f			! 0x2D30
	nop
	nop
	nop
1:	ba	0f			! 0x2D40
	nop
	nop
	nop
0:	ba	1f			! 0x2D50
	nop
	nop
	nop
1:	ba	0f			! 0x2D60
	nop
	nop
	nop
0:	ba	1f			! 0x2D70
	nop
	nop
	nop
1:	ba	0f			! 0x2D80
	nop
	nop
	nop
0:	ba	1f			! 0x2D90
	nop
	nop
	nop
1:	ba	0f			! 0x2DA0
	nop
	nop
	nop
0:	ba	1f			! 0x2DB0
	nop
	nop
	nop
1:	ba	0f			! 0x2DC0
	nop
	nop
	nop
0:	ba	1f			! 0x2DD0
	nop
	nop
	nop
1:	ba	0f			! 0x2DE0
	nop
	nop
	nop
0:	ba	1f			! 0x2DF0
	nop
	nop
	nop

1:	ba	0f			! 0x2E00
	nop
	nop
	nop
0:	ba	1f			! 0x2E10
	nop
	nop
	nop
1:	ba	0f			! 0x2E20
	nop
	nop
	nop
0:	ba	1f			! 0x2E30
	nop
	nop
	nop
1:	ba	0f			! 0x2E40
	nop
	nop
	nop
0:	ba	1f			! 0x2E50
	nop
	nop
	nop
1:	ba	0f			! 0x2E60
	nop
	nop
	nop
0:	ba	1f			! 0x2E70
	nop
	nop
	nop
1:	ba	0f			! 0x2E80
	nop
	nop
	nop
0:	ba	1f			! 0x2E90
	nop
	nop
	nop
1:	ba	0f			! 0x2EA0
	nop
	nop
	nop
0:	ba	1f			! 0x2EB0
	nop
	nop
	nop
1:	ba	0f			! 0x2EC0
	nop
	nop
	nop
0:	ba	1f			! 0x2ED0
	nop
	nop
	nop
1:	ba	0f			! 0x2EE0
	nop
	nop
	nop
0:	ba	1f			! 0x2EF0
	nop
	nop
	nop

1:	ba	0f			! 0x2F00
	nop
	nop
	nop
0:	ba	1f			! 0x2F10
	nop
	nop
	nop
1:	ba	0f			! 0x2F20
	nop
	nop
	nop
0:	ba	1f			! 0x2F30
	nop
	nop
	nop
1:	ba	0f			! 0x2F40
	nop
	nop
	nop
0:	ba	1f			! 0x2F50
	nop
	nop
	nop
1:	ba	0f			! 0x2F60
	nop
	nop
	nop
0:	ba	1f			! 0x2F70
	nop
	nop
	nop
1:	ba	0f			! 0x2F80
	nop
	nop
	nop
0:	ba	1f			! 0x2F90
	nop
	nop
	nop
1:	ba	0f			! 0x2FA0
	nop
	nop
	nop
0:	ba	1f			! 0x2FB0
	nop
	nop
	nop
1:	ba	0f			! 0x2FC0
	nop
	nop
	nop
0:	ba	1f			! 0x2FD0
	nop
	nop
	nop
1:	ba	0f			! 0x2FE0
	nop
	nop
	nop
0:	ba	1f			! 0x2FF0
	nop
	nop
	nop

! *********************** 12KB ***************************************

1:	ba	0f			! 0x3000
	nop
	nop
	nop
0:	ba	1f			! 0x3010
	nop
	nop
	nop
1:	ba	0f			! 0x3020
	nop
	nop
	nop
0:	ba	1f			! 0x3030
	nop
	nop
	nop
1:	ba	0f			! 0x3040
	nop
	nop
	nop
0:	ba	1f			! 0x3050
	nop
	nop
	nop
1:	ba	0f			! 0x3060
	nop
	nop
	nop
0:	ba	1f			! 0x3070
	nop
	nop
	nop
1:	ba	0f			! 0x3080
	nop
	nop
	nop
0:	ba	1f			! 0x3090
	nop
	nop
	nop
1:	ba	0f			! 0x30A0
	nop
	nop
	nop
0:	ba	1f			! 0x30B0
	nop
	nop
	nop
1:	ba	0f			! 0x30C0
	nop
	nop
	nop
0:	ba	1f			! 0x30D0
	nop
	nop
	nop
1:	ba	0f			! 0x30E0
	nop
	nop
	nop
0:	ba	1f			! 0x30F0
	nop
	nop
	nop

1:	ba	0f			! 0x3100
	nop
	nop
	nop
0:	ba	1f			! 0x3110
	nop
	nop
	nop
1:	ba	0f			! 0x3120
	nop
	nop
	nop
0:	ba	1f			! 0x3130
	nop
	nop
	nop
1:	ba	0f			! 0x3140
	nop
	nop
	nop
0:	ba	1f			! 0x3150
	nop
	nop
	nop
1:	ba	0f			! 0x3160
	nop
	nop
	nop
0:	ba	1f			! 0x3170
	nop
	nop
	nop
1:	ba	0f			! 0x3180
	nop
	nop
	nop
0:	ba	1f			! 0x3190
	nop
	nop
	nop
1:	ba	0f			! 0x31A0
	nop
	nop
	nop
0:	ba	1f			! 0x31B0
	nop
	nop
	nop
1:	ba	0f			! 0x31C0
	nop
	nop
	nop
0:	ba	1f			! 0x31D0
	nop
	nop
	nop
1:	ba	0f			! 0x31E0
	nop
	nop
	nop
0:	ba	1f			! 0x31F0
	nop
	nop
	nop

1:	ba	0f			! 0x3200
	nop
	nop
	nop
0:	ba	1f			! 0x3210
	nop
	nop
	nop
1:	ba	0f			! 0x3220
	nop
	nop
	nop
0:	ba	1f			! 0x3230
	nop
	nop
	nop
1:	ba	0f			! 0x3240
	nop
	nop
	nop
0:	ba	1f			! 0x3250
	nop
	nop
	nop
1:	ba	0f			! 0x3260
	nop
	nop
	nop
0:	ba	1f			! 0x3270
	nop
	nop
	nop
1:	ba	0f			! 0x3280
	nop
	nop
	nop
0:	ba	1f			! 0x3290
	nop
	nop
	nop
1:	ba	0f			! 0x32A0
	nop
	nop
	nop
0:	ba	1f			! 0x32B0
	nop
	nop
	nop
1:	ba	0f			! 0x32C0
	nop
	nop
	nop
0:	ba	1f			! 0x32D0
	nop
	nop
	nop
1:	ba	0f			! 0x32E0
	nop
	nop
	nop
0:	ba	1f			! 0x32F0
	nop
	nop
	nop

1:	ba	0f			! 0x3300
	nop
	nop
	nop
0:	ba	1f			! 0x3310
	nop
	nop
	nop
1:	ba	0f			! 0x3320
	nop
	nop
	nop
0:	ba	1f			! 0x3330
	nop
	nop
	nop
1:	ba	0f			! 0x3340
	nop
	nop
	nop
0:	ba	1f			! 0x3350
	nop
	nop
	nop
1:	ba	0f			! 0x3360
	nop
	nop
	nop
0:	ba	1f			! 0x3370
	nop
	nop
	nop
1:	ba	0f			! 0x3380
	nop
	nop
	nop
0:	ba	1f			! 0x3390
	nop
	nop
	nop
1:	ba	0f			! 0x33A0
	nop
	nop
	nop
0:	ba	1f			! 0x33B0
	nop
	nop
	nop
1:	ba	0f			! 0x33C0
	nop
	nop
	nop
0:	ba	1f			! 0x33D0
	nop
	nop
	nop
1:	ba	0f			! 0x33E0
	nop
	nop
	nop
0:	ba	1f			! 0x33F0
	nop
	nop
	nop

1:	ba	0f			! 0x3400
	nop
	nop
	nop
0:	ba	1f			! 0x3410
	nop
	nop
	nop
1:	ba	0f			! 0x3420
	nop
	nop
	nop
0:	ba	1f			! 0x3430
	nop
	nop
	nop
1:	ba	0f			! 0x3440
	nop
	nop
	nop
0:	ba	1f			! 0x3450
	nop
	nop
	nop
1:	ba	0f			! 0x3460
	nop
	nop
	nop
0:	ba	1f			! 0x3470
	nop
	nop
	nop
1:	ba	0f			! 0x3480
	nop
	nop
	nop
0:	ba	1f			! 0x3490
	nop
	nop
	nop
1:	ba	0f			! 0x34A0
	nop
	nop
	nop
0:	ba	1f			! 0x34B0
	nop
	nop
	nop
1:	ba	0f			! 0x34C0
	nop
	nop
	nop
0:	ba	1f			! 0x34D0
	nop
	nop
	nop
1:	ba	0f			! 0x34E0
	nop
	nop
	nop
0:	ba	1f			! 0x34F0
	nop
	nop
	nop

1:	ba	0f			! 0x3500
	nop
	nop
	nop
0:	ba	1f			! 0x3510
	nop
	nop
	nop
1:	ba	0f			! 0x3520
	nop
	nop
	nop
0:	ba	1f			! 0x3530
	nop
	nop
	nop
1:	ba	0f			! 0x3540
	nop
	nop
	nop
0:	ba	1f			! 0x3550
	nop
	nop
	nop
1:	ba	0f			! 0x3560
	nop
	nop
	nop
0:	ba	1f			! 0x3570
	nop
	nop
	nop
1:	ba	0f			! 0x3580
	nop
	nop
	nop
0:	ba	1f			! 0x3590
	nop
	nop
	nop
1:	ba	0f			! 0x35A0
	nop
	nop
	nop
0:	ba	1f			! 0x35B0
	nop
	nop
	nop
1:	ba	0f			! 0x35C0
	nop
	nop
	nop
0:	ba	1f			! 0x35D0
	nop
	nop
	nop
1:	ba	0f			! 0x35E0
	nop
	nop
	nop
0:	ba	1f			! 0x35F0
	nop
	nop
	nop

1:	ba	0f			! 0x3600
	nop
	nop
	nop
0:	ba	1f			! 0x3610
	nop
	nop
	nop
1:	ba	0f			! 0x3620
	nop
	nop
	nop
0:	ba	1f			! 0x3630
	nop
	nop
	nop
1:	ba	0f			! 0x3640
	nop
	nop
	nop
0:	ba	1f			! 0x3650
	nop
	nop
	nop
1:	ba	0f			! 0x3660
	nop
	nop
	nop
0:	ba	1f			! 0x3670
	nop
	nop
	nop
1:	ba	0f			! 0x3680
	nop
	nop
	nop
0:	ba	1f			! 0x3690
	nop
	nop
	nop
1:	ba	0f			! 0x36A0
	nop
	nop
	nop
0:	ba	1f			! 0x36B0
	nop
	nop
	nop
1:	ba	0f			! 0x36C0
	nop
	nop
	nop
0:	ba	1f			! 0x36D0
	nop
	nop
	nop
1:	ba	0f			! 0x36E0
	nop
	nop
	nop
0:	ba	1f			! 0x36F0
	nop
	nop
	nop

1:	ba	0f			! 0x3700
	nop
	nop
	nop
0:	ba	1f			! 0x3710
	nop
	nop
	nop
1:	ba	0f			! 0x3720
	nop
	nop
	nop
0:	ba	1f			! 0x3730
	nop
	nop
	nop
1:	ba	0f			! 0x3740
	nop
	nop
	nop
0:	ba	1f			! 0x3750
	nop
	nop
	nop
1:	ba	0f			! 0x3760
	nop
	nop
	nop
0:	ba	1f			! 0x3770
	nop
	nop
	nop
1:	ba	0f			! 0x3780
	nop
	nop
	nop
0:	ba	1f			! 0x3790
	nop
	nop
	nop
1:	ba	0f			! 0x37A0
	nop
	nop
	nop
0:	ba	1f			! 0x37B0
	nop
	nop
	nop
1:	ba	0f			! 0x37C0
	nop
	nop
	nop
0:	ba	1f			! 0x37D0
	nop
	nop
	nop
1:	ba	0f			! 0x37E0
	nop
	nop
	nop
0:	ba	1f			! 0x37F0
	nop
	nop
	nop

1:	ba	0f			! 0x3800
	nop
	nop
	nop
0:	ba	1f			! 0x3810
	nop
	nop
	nop
1:	ba	0f			! 0x3820
	nop
	nop
	nop
0:	ba	1f			! 0x3830
	nop
	nop
	nop
1:	ba	0f			! 0x3840
	nop
	nop
	nop
0:	ba	1f			! 0x3850
	nop
	nop
	nop
1:	ba	0f			! 0x3860
	nop
	nop
	nop
0:	ba	1f			! 0x3870
	nop
	nop
	nop
1:	ba	0f			! 0x3880
	nop
	nop
	nop
0:	ba	1f			! 0x3890
	nop
	nop
	nop
1:	ba	0f			! 0x38A0
	nop
	nop
	nop
0:	ba	1f			! 0x38B0
	nop
	nop
	nop
1:	ba	0f			! 0x38C0
	nop
	nop
	nop
0:	ba	1f			! 0x38D0
	nop
	nop
	nop
1:	ba	0f			! 0x38E0
	nop
	nop
	nop
0:	ba	1f			! 0x38F0
	nop
	nop
	nop

1:	ba	0f			! 0x3900
	nop
	nop
	nop
0:	ba	1f			! 0x3910
	nop
	nop
	nop
1:	ba	0f			! 0x3920
	nop
	nop
	nop
0:	ba	1f			! 0x3930
	nop
	nop
	nop
1:	ba	0f			! 0x3940
	nop
	nop
	nop
0:	ba	1f			! 0x3950
	nop
	nop
	nop
1:	ba	0f			! 0x3960
	nop
	nop
	nop
0:	ba	1f			! 0x3970
	nop
	nop
	nop
1:	ba	0f			! 0x3980
	nop
	nop
	nop
0:	ba	1f			! 0x3990
	nop
	nop
	nop
1:	ba	0f			! 0x39A0
	nop
	nop
	nop
0:	ba	1f			! 0x39B0
	nop
	nop
	nop
1:	ba	0f			! 0x39C0
	nop
	nop
	nop
0:	ba	1f			! 0x39D0
	nop
	nop
	nop
1:	ba	0f			! 0x39E0
	nop
	nop
	nop
0:	ba	1f			! 0x39F0
	nop
	nop
	nop

1:	ba	0f			! 0x3A00
	nop
	nop
	nop
0:	ba	1f			! 0x3A10
	nop
	nop
	nop
1:	ba	0f			! 0x3A20
	nop
	nop
	nop
0:	ba	1f			! 0x3A30
	nop
	nop
	nop
1:	ba	0f			! 0x3A40
	nop
	nop
	nop
0:	ba	1f			! 0x3A50
	nop
	nop
	nop
1:	ba	0f			! 0x3A60
	nop
	nop
	nop
0:	ba	1f			! 0x3A70
	nop
	nop
	nop
1:	ba	0f			! 0x3A80
	nop
	nop
	nop
0:	ba	1f			! 0x3A90
	nop
	nop
	nop
1:	ba	0f			! 0x3AA0
	nop
	nop
	nop
0:	ba	1f			! 0x3AB0
	nop
	nop
	nop
1:	ba	0f			! 0x3AC0
	nop
	nop
	nop
0:	ba	1f			! 0x3AD0
	nop
	nop
	nop
1:	ba	0f			! 0x3AE0
	nop
	nop
	nop
0:	ba	1f			! 0x3AF0
	nop
	nop
	nop

1:	ba	0f			! 0x3B00
	nop
	nop
	nop
0:	ba	1f			! 0x3B10
	nop
	nop
	nop
1:	ba	0f			! 0x3B20
	nop
	nop
	nop
0:	ba	1f			! 0x3B30
	nop
	nop
	nop
1:	ba	0f			! 0x3B40
	nop
	nop
	nop
0:	ba	1f			! 0x3B50
	nop
	nop
	nop
1:	ba	0f			! 0x3B60
	nop
	nop
	nop
0:	ba	1f			! 0x3B70
	nop
	nop
	nop
1:	ba	0f			! 0x3B80
	nop
	nop
	nop
0:	ba	1f			! 0x3B90
	nop
	nop
	nop
1:	ba	0f			! 0x3BA0
	nop
	nop
	nop
0:	ba	1f			! 0x3BB0
	nop
	nop
	nop
1:	ba	0f			! 0x3BC0
	nop
	nop
	nop
0:	ba	1f			! 0x3BD0
	nop
	nop
	nop
1:	ba	0f			! 0x3BE0
	nop
	nop
	nop
0:	ba	1f			! 0x3BF0
	nop
	nop
	nop

1:	ba	0f			! 0x3C00
	nop
	nop
	nop
0:	ba	1f			! 0x3C10
	nop
	nop
	nop
1:	ba	0f			! 0x3C20
	nop
	nop
	nop
0:	ba	1f			! 0x3C30
	nop
	nop
	nop
1:	ba	0f			! 0x3C40
	nop
	nop
	nop
0:	ba	1f			! 0x3C50
	nop
	nop
	nop
1:	ba	0f			! 0x3C60
	nop
	nop
	nop
0:	ba	1f			! 0x3C70
	nop
	nop
	nop
1:	ba	0f			! 0x3C80
	nop
	nop
	nop
0:	ba	1f			! 0x3C90
	nop
	nop
	nop
1:	ba	0f			! 0x3CA0
	nop
	nop
	nop
0:	ba	1f			! 0x3CB0
	nop
	nop
	nop
1:	ba	0f			! 0x3CC0
	nop
	nop
	nop
0:	ba	1f			! 0x3CD0
	nop
	nop
	nop
1:	ba	0f			! 0x3CE0
	nop
	nop
	nop
0:	ba	1f			! 0x3CF0
	nop
	nop
	nop

1:	ba	0f			! 0x3D00
	nop
	nop
	nop
0:	ba	1f			! 0x3D10
	nop
	nop
	nop
1:	ba	0f			! 0x3D20
	nop
	nop
	nop
0:	ba	1f			! 0x3D30
	nop
	nop
	nop
1:	ba	0f			! 0x3D40
	nop
	nop
	nop
0:	ba	1f			! 0x3D50
	nop
	nop
	nop
1:	ba	0f			! 0x3D60
	nop
	nop
	nop
0:	ba	1f			! 0x3D70
	nop
	nop
	nop
1:	ba	0f			! 0x3D80
	nop
	nop
	nop
0:	ba	1f			! 0x3D90
	nop
	nop
	nop
1:	ba	0f			! 0x3DA0
	nop
	nop
	nop
0:	ba	1f			! 0x3DB0
	nop
	nop
	nop
1:	ba	0f			! 0x3DC0
	nop
	nop
	nop
0:	ba	1f			! 0x3DD0
	nop
	nop
	nop
1:	ba	0f			! 0x3DE0
	nop
	nop
	nop
0:	ba	1f			! 0x3DF0
	nop
	nop
	nop

1:	ba	0f			! 0x3E00
	nop
	nop
	nop
0:	ba	1f			! 0x3E10
	nop
	nop
	nop
1:	ba	0f			! 0x3E20
	nop
	nop
	nop
0:	ba	1f			! 0x3E30
	nop
	nop
	nop
1:	ba	0f			! 0x3E40
	nop
	nop
	nop
0:	ba	1f			! 0x3E50
	nop
	nop
	nop
1:	ba	0f			! 0x3E60
	nop
	nop
	nop
0:	ba	1f			! 0x3E70
	nop
	nop
	nop
1:	ba	0f			! 0x3E80
	nop
	nop
	nop
0:	ba	1f			! 0x3E90
	nop
	nop
	nop
1:	ba	0f			! 0x3EA0
	nop
	nop
	nop
0:	ba	1f			! 0x3EB0
	nop
	nop
	nop
1:	ba	0f			! 0x3EC0
	nop
	nop
	nop
0:	ba	1f			! 0x3ED0
	nop
	nop
	nop
1:	ba	0f			! 0x3EE0
	nop
	nop
	nop
0:	ba	1f			! 0x3EF0
	nop
	nop
	nop

1:	ba	0f			! 0x3F00
	nop
	nop
	nop
0:	ba	1f			! 0x3F10
	nop
	nop
	nop
1:	ba	0f			! 0x3F20
	nop
	nop
	nop
0:	ba	1f			! 0x3F30
	nop
	nop
	nop
1:	ba	0f			! 0x3F40
	nop
	nop
	nop
0:	ba	1f			! 0x3F50
	nop
	nop
	nop
1:	ba	0f			! 0x3F60
	nop
	nop
	nop
0:	ba	1f			! 0x3F70
	nop
	nop
	nop
1:	ba	0f			! 0x3F80
	nop
	nop
	nop
0:	ba	1f			! 0x3F90
	nop
	nop
	nop
1:	ba	0f			! 0x3FA0
	nop
	nop
	nop
0:	ba	1f			! 0x3FB0
	nop
	nop
	nop
1:	ba	0f			! 0x3FC0
	nop
	nop
	nop
0:	ba	1f			! 0x3FD0
	nop
	nop
	nop
1:	ba	0f			! 0x3FE0
	nop
	nop
	nop
0:	nop				! 0x3FF0
	nop
	nop
	nop


! *********************** 16KB ***************************************
#endif


	mov	0,%i0
	ret
	restore

