/*****************************************************************
 *  Copyright (c) 1992, Xerox Corporation.  All rights reserved. *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* File: ssframe.s
 *
 * This file contains assembly code to add and remove frames from pixrects
 * in place.  The routine removeFrameFromLine is also used to copy pixrects.
 *
 * This file also contains code to clear the right frame of an image.
 *
 *
 */

#include "glblname.prv"

RCSInfo:
	.asciz	"$RCSfile: ssframe.s,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:18:18 $"
	.align	4

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPState	%i0		/* current addr in dest image */
#define	rDstAddr	%i1		/* current addr in dest image */
#define rSrcAddr	%i2		/* current addr in source image */
#define rLength		%i3		/* number of image bytes to transfer */
#define rFrameWidth	%i4		/* width of frame in bytes */
#define rRightMask	%i5		/* used to clear frame in
					 * addFrameToLine */

#define	rCount		%g1		/* used to clear left & right frames */
#define	rOut0		%g2		/* output register pair */
#define rOut4		%g3
#define	rSrcAddrP4	%g4		/* points to rSrcAddr + 4 */
#define	rOffset		%g5		/* address offset */

#define rLeftShift	%o4		/* left shift distance */
#define rRightShift	%o5		/* right shift distance */

#define rInput0		%l0		/* input word 0 */
#define rInput1		%l1		/* input word 1 */

#define	rRetAddr	%o7		/* defined by SPARC's call inst */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	dstAddrOffset	0x48
#define	srcAddrOffset	0x4c
#define	lengthOffset	0x50
#define	frmWidthOffset	0x54
#define rightMaskOffset 0x58

/* magical C Frame constant (no local variables) */
#define	LF26		-64

LL0:					! does this really do anything?

! The ip_addFrameToLine routine originally assumed that all images were
! padded to 32 bit boundaries.  While this is true of our images, it
! may not be true of images received from scanners.  There's an HP scanner
! which refuses to pad it's 2550 pixel images out to 320 bytes.  It
! insists on 319.  So, this routine must be able to handle source images
! that are not packed on 4 byte boundaries.  This means it must be
! able to shift bytes into position during the transfer.  Ouch.
!
! We'll have two copies of the code.  Each copy has four subcases for the
! combinations of even/odd dest address and even/odd dest pitch.
! The first copy, the original, does no shifting.  It's executed when
! rSrcAddr is a multiple of 4.  The second copy is selected when
! rSrcAddr is not a multiple 4.  In it, we will shift the source data
! left by (rScrAddr & 0x3) before putting it into the dest address.
! Since we're doing a left shift, we have to read the source word to the
! right of the one containing the byte that will go into the MS byte of the
! dest word.  This makes thing more complicated - that's why we have two
! copies of the code.


! -----------------------------------------------------------------------------
! _ip_addFrameToLine Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_addFrameToLine)
MK_GLOBAL_NAME(ip_addFrameToLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState, [%fp+ipStateOffset]
	st	rDstAddr, [%fp+dstAddrOffset]
	st	rSrcAddr, [%fp+srcAddrOffset]
	st	rLength,  [%fp+lengthOffset]
	st	rFrameWidth, [%fp+frmWidthOffset]
	st	rRightMask,  [%fp+rightMaskOffset]

! load the frame width and clear the frame
	mov	0, rOut0
	mov	rFrameWidth, rCount

addFrameLeftFrameLoop:
	subcc	rCount, 0x4, rCount
	bnz	addFrameLeftFrameLoop
	st	rOut0, [rDstAddr + rCount]

! point to left edge of destination image
	add	rDstAddr, rFrameWidth, rDstAddr
	add	rDstAddr, 0x8, rDstAddr		! used in loop pipelining

! and set up rSrcAddrP4.  These will be used to read
! one of the pair of source words, or two of the triplet of shifted
! source words to be stored in each double write.
	add	rSrcAddr, 0x4, rSrcAddrP4
	and	rSrcAddrP4, ~0x3, rSrcAddrP4	! we'll fix rScrAddr later

! Find the offset of the word which contains the rightmost destination
! byte.  We will use this offset plus 4 as the the offset of the source
! word containing the rightmost source byte when shifting is requried.
! Otherwise we use just this offset to find the rightmost source byte.
! This works because the source offset is applied after the original
! source poiner has been rounded down to the previous word boundary.
	sub	rLength, 0x1, rOffset
	and	rOffset, ~0x3, rOffset		! round down to nearest word

! Decide whether there will be shifting in the loops.  If not, just
! execute the original fast code.  If so, execute the slower, shiftier code.
	andcc	rSrcAddr, 0x3, rLeftShift	! are either of the
						! lower bits set?
	bnz	addFrameWithShift		! no, do the fast case
	sll	rLeftShift, 0x3, rLeftShift	! get shift distance in bits


! ------------- ADD FRAME, DO NOT SHIFT INPUTS ----------------------------

! We want to use double word stores and check for the end of the line as
! seldom as possible.  To minimize operations, we will use one register for
! both the address offset and the count.  That means we'll store from right
! to left across the image.  There are four cases:
!  1. Rightmost dest address is even, line length is even:
!	Do one single store first, check for the end after reading word
!	that will be stored at odd address.  Finish with a single store.
!  2. Rightmost dest address is even, line length is odd:
!	Do on single store first, check for the end after first single
!	store and after every double store thereafter.
!  3. Rightmost dest address is odd, line length is even:
!	Do double store first, check for end after every double store.
!  4. Rightmost dest address is odd, line length is odd:
!	Check for end before doing every double store.  End with a single store

! Is rightmost word address even or odd?  Even though we are really
! calculating rDstAddr+rOffset+8, the odd vs. even choice comes out the same
addFrameNoShift:
	add	rDstAddr, rOffset, rCount
	andcc	rCount, 0x4, %g0
	bnz	addFrameOddDestLine
	andcc	rOffset, 0x4, %g0

! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the original offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
	bz	addFrameEvenAddrOddLength
	ld	[rSrcAddr + rOffset], rOut0

! ------------------------------------------------

! EVEN ADDRESS, EVEN LENGTH

! rightmost dest address is even, line length is even
addFrameEvenAddrEvenLength:
	subcc	rOffset, 0x8, rOffset		! point to even addr to left
						! Uses INTERLOCK cycle.
	st	rOut0, [rDstAddr + rOffset]	! rDstAddr is 8 too big, so
						! this works ok.
	bneg	addFrameEEFin			! quit if this is the last
						! word to be read
	ld	[rSrcAddrP4 + rOffset], rOut4	! read word for odd store

addFrameEELoop:
	ld	[rSrcAddr + rOffset], rOut0	! read word for even store
	subcc	rOffset, 0x8, rOffset		! pt to even addr to left
	std	rOut0, [rDstAddr + rOffset]
	bge	addFrameEELoop			! more to do?
	ld	[rSrcAddrP4 + rOffset], rOut4	! load odd word

! We get here having just read the last word into rOut4.  Fix up rDstAddr
! and store the word into [rDstAddr].
addFrameEEFin:
	sub	rDstAddr, 0x8, rDstAddr
	ba	addFrameRightFrame		! go add the right frame
	st	rOut4, [rDstAddr]


! ------------------------------------------------

! EVEN ADDRESS, ODD LENGTH

! We do a load and store, but check for the end of line "after" doing the
! store.  Well, we really do the store in the delay slot of the check, so
! rOffset has already been decremented.  To make things work out right, we've
! already pre-incremented rDstAddr by 8.
addFrameEvenAddrOddLength:
	subcc	rOffset, 0x8, rOffset		! at end of line yet?
	bneg	addFrameEOFin			! yes, fix up rDstAddr
	st	rOut0, [rDstAddr + rOffset]

addFrameEOLoop:
	ld	[rSrcAddr + rOffset], rOut0	! load even word
	ld	[rSrcAddrP4 + rOffset], rOut4	! load odd word
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	addFrameEOLoop			! no, keep going
	std	rOut0, [rDstAddr + rOffset]

addFrameEOFin:
	ba	addFrameRightFrame		! go add the right frame
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------
! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the oringinal offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
addFrameOddDestLine:
	bz	addFrameOddAddrOddLength
	subcc	rOffset, 0x4, rOffset		! point to even addr

! ------------------------------------------------

! ODD ADDRESS, EVEN LENGTH

! This case is like even addr, odd length in that we do the check AFTER
! the store.  As above, rDstAddr has been modified to make things work
! out right.
addFrameOELoop:
	ld	[rSrcAddr + rOffset], rOut0
	ld	[rSrcAddrP4 + rOffset], rOut4
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	addFrameOELoop			! no, keep going
	std	rOut0, [rDstAddr + rOffset]

addFrameOEFin:
	ba	addFrameRightFrame		! go add the right frame
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------

! ODD ADDRESS, ODD LENGTH

! This case is like the even address, even length case in that we quit
! after reading the word at an odd address.
addFrameOddAddrOddLength:
	bneg	addFrameOOFin
	ld	[rSrcAddrP4 + rOffset], rOut4	! get odd word

addFrameOOLoop:
	ld	[rSrcAddr + rOffset], rOut0	! get even word
	subcc	rOffset, 0x8, rOffset		! point to even wd to the left
						! Uses INTERLOCK cycle.
	std	rOut0, [rDstAddr + rOffset]
	bge	addFrameOOLoop			! finish rest of line
	ld	[rSrcAddrP4 + rOffset], rOut4

! We've just loaded the last word.  Fix up rDstAddr and store the last word
addFrameOOFin:
	sub	rDstAddr, 0x8, rDstAddr
	ba	addFrameRightFrame
	st	rOut4, [rDstAddr]



! ----------------------------------------------------
! ------------- ADD FRAME, SHIFT INPUTS ---------------
! ----------------------------------------------------

! This is similar to the code above except each double word is composed
! from bytes from three source words.  The word on the right
! comes from [rSrcAddr+rOffset] << rLeftShift | [rSrcAddrP4+rOffset] >>
! rRightShift.  The word on the left comes from
! [rSrcAddr+rOffset-8] << rLeftShift | [rSrcAddr+rOffset] >> rRightShift.
! We rotate input registers so each input word is read only once.

! When we get here, rShiftLeft contains the left shift distance in
! bits.  We compute the right shift distance by subtracting this from
! 32.  We also need to compute a mask for the rightmost dest word.
! That's because it may not end on a word boundary, so may contain
! weird bits.

! Is rightmost word address even or odd?  Even though we are really
! calculating rDstAddr+rOffset+8, the odd vs. even choice comes out the same
addFrameWithShift:
	and	rSrcAddr, ~0x3, rSrcAddr	! get rid of lower bits
	sub	rLeftShift, 32, rRightShift	! equals -right shift distance
	sub	%g0, rRightShift, rRightShift	! get the proper sign

	add	rDstAddr, rOffset, rCount
	andcc	rCount, 0x4, %g0
	bnz	addFrameShiftOddDestLine
	andcc	rOffset, 0x4, %g0

! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the original offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
	bz	addFrameShiftEvenAddrOddLength
	ld	[rSrcAddrP4 + rOffset], rInput1	! get right word

! ------------------------------------------------

! EVEN ADDRESS, EVEN LENGTH

! rightmost dest address is even, line length is even
addFrameShiftEvenAddrEvenLength:
	ld	[rSrcAddr + rOffset], rInput0	! get middle word
	srl	rInput1, rRightShift, rOut0	! position bit on the right
	sll	rInput0, rLeftShift, rInput1	! position the bits on left
	or	rOut0, rInput1, rOut0		! compose even dest word
	subcc	rOffset, 0x8, rOffset		! point to even addr to left
	st	rOut0, [rDstAddr + rOffset]	! rDstAddr is 8 too big, so
						! this works ok.
	bneg	addFrameShiftEEFin		! quit if this is the last
						! word to be read
	ld	[rSrcAddrP4 + rOffset], rInput1	! read lower word for odd store

addFrameShiftEELoop:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! get low word for even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! pt to even addr to left
	std	rOut0, [rDstAddr + rOffset]
	bge	addFrameShiftEELoop		! more to do?
	ld	[rSrcAddrP4 + rOffset], rInput1	! load low word, odd word

! We get here having just read the low part of the last word into
! rInput1.  Fix up rDstAddr and store the word into [rDstAddr].
addFrameShiftEEFin:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	sub	rDstAddr, 0x8, rDstAddr
	ba	addFrameRightFrame		! go add the right frame
	st	rOut4, [rDstAddr]


! ------------------------------------------------

! EVEN ADDRESS, ODD LENGTH

! We do a load and store, but check for the end of line "after" doing the
! store.  Well, we really do the store in the delay slot of the check, so
! rOffset has already been decremented.  To make things work out right, we've
! already pre-incremented rDstAddr by 8.
addFrameShiftEvenAddrOddLength:
	ld	[rSrcAddr + rOffset], rInput0	! get middle word
	srl	rInput1, rRightShift, rOut0	! position bit on the right
	sll	rInput0, rLeftShift, rInput1	! position the bits on left
	or	rOut0, rInput1, rOut0		! compose even dest word
	subcc	rOffset, 0x8, rOffset		! at end of line yet?
	bneg	addFrameShiftEOFin		! yes, fix up rDstAddr
	st	rOut0, [rDstAddr + rOffset]

addFrameShiftEOLoop:
	ld	[rSrcAddrP4 + rOffset], rInput1	! load low word, odd store
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! load low word, even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	addFrameShiftEOLoop		! no, keep going
	std	rOut0, [rDstAddr + rOffset]

addFrameShiftEOFin:
	ba	addFrameRightFrame		! go add the right frame
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------
! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the oringinal offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
addFrameShiftOddDestLine:
	ld	[rSrcAddrP4 + rOffset], rInput0	! get high word, odd store
	bz	addFrameShiftOddAddrOddLength
	subcc	rOffset, 0x4, rOffset		! point to even addr

! ------------------------------------------------

! ODD ADDRESS, EVEN LENGTH

! This case is like even addr, odd length in that we do the check AFTER
! the store.  As above, rDstAddr has been modified to make things work
! out right.
addFrameShiftOELoop:
	ld	[rSrcAddrP4 + rOffset], rInput1	! get low word, odd store
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! get low word, even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	addFrameShiftOELoop		! no, keep going
	std	rOut0, [rDstAddr + rOffset]

addFrameShiftOEFin:
	ba	addFrameRightFrame		! go add the right frame
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------

! ODD ADDRESS, ODD LENGTH

! This case is like the even address, even length case in that we quit
! after reading the word at an odd address.
addFrameShiftOddAddrOddLength:
	bneg	addFrameShiftOOFin
	ld	[rSrcAddrP4 + rOffset], rInput1	! get low word, odd store

addFrameShiftOOLoop:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! get low word, even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! point to even wd to the left
	std	rOut0, [rDstAddr + rOffset]
	bge	addFrameShiftOOLoop		! finish rest of line
	ld	[rSrcAddrP4 + rOffset], rInput1	! get low word, odd store

! We've just loaded the last word.  Fix up rDstAddr and store the last word
addFrameShiftOOFin:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	sub	rDstAddr, 0x8, rDstAddr
!	ba	addFrameRightFrame
	st	rOut4, [rDstAddr]


! ------------------------------------------------
! At this point, all the image data has been transferred.  We need to clear
! the right frame.  Note that the right frame begins in the rightmost
! word of the line.  We clear it hear instead of when writing that word
! because breaking up those loops was a pain.
addFrameRightFrame:
	add	rLength, 3, rLength		! round up to next word
	and	rLength, ~0x3, rLength
	add	rDstAddr, rLength, rDstAddr	! point to right frame
	ld	[rDstAddr - 4], rInput0		! get last image word
	mov	0, rOut0			! fill interlock
	and	rInput0, rRightMask, rInput0	! clear right frame
	st	rInput0, [rDstAddr - 4]

addFrameRightFrameLoop:
	subcc	rFrameWidth, 0x4, rFrameWidth
	bnz	addFrameRightFrameLoop
	st	rOut0, [rDstAddr + rFrameWidth]


! END OF THIS ROW

! We've put out the right number of words, so quit
addFrameToLineFin:
	mov	0,%i0
	ret
	restore




! -----------------------------------------------------------------------------
! _ip_removeFrameFromLine Routine
! -----------------------------------------------------------------------------

! This routine is also used to copy pixrects, so we'll work a little harder
! to make it fast.  We also have the luxury of having a frame on the source
! image, so can read a bit off the ends.
! The old version of this code used to read 8 source words and do four
! double word writes in row on the theory that it would help the SS2's
! micro-TLB.  That thing only holds 1 virtual->real address.  Blah.
! Since we now have to shift data back and forth during the transfer
! and none of the microTLB stuff seems to matter anyway, we'll just
! read two words and do a double write like the addFrame code does
! (sort of).  The sort-of part is that the source is guaranteed to be
! word aligned and the dest is not.

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_removeFrameFromLine)
MK_GLOBAL_NAME(ip_removeFrameFromLine):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.  We also load the addresses of the permutation table here.
	st	rIPState, [%fp+ipStateOffset]
	st	rDstAddr, [%fp+dstAddrOffset]
	st	rSrcAddr, [%fp+srcAddrOffset]
	st	rLength,  [%fp+lengthOffset]

! and set up rSrcAddrP4.  These will be used to read
! one of the pair of source words, or two of the triplet of shifted
! source words to be stored in each double write.
	add	rSrcAddr, 0x4, rSrcAddrP4
	add	rDstAddr, 0x8, rDstAddr		! used in loop pipelining

! Find the offset of the word which contains the rightmost source
! byte.  This is also the offset of the rightmost dest byte relative to
! rDstAddr+4.  We will process the line from right to left, then do the
! leftmost dest word using a merge with the old data.
	sub	rLength, 0x1, rOffset
	and	rOffset, ~0x3, rOffset		! round down to nearest word


! Decide whether there will be shifting in the loops.  If not, just
! execute the original fast code.  If so, execute the slower, shiftier code.
	andcc	rDstAddr, 0x3, rRightShift	! are either of the
						! lower bits set?
	bnz	rmvFrameWithShift		! no, do the fast case
	sll	rRightShift, 0x3, rRightShift	! get shift distance in bits


! ------------- REMOVE FRAME, DO NOT SHIFT INPUTS ----------------------------

! We want to use double word stores and check for the end of the line as
! seldom as possible.  To minimize operations, we will use one register for
! both the address offset and the count.  That means we'll store from right
! to left across the image.  There are four cases:
!  1. Rightmost dest address is even, line length is even:
!	Do one single store first, check for the end after reading word
!	that will be stored at odd address.  Finish with a single store.
!  2. Rightmost dest address is even, line length is odd:
!	Do on single store first, check for the end after first single
!	store and after every double store thereafter.
!  3. Rightmost dest address is odd, line length is even:
!	Do double store first, check for end after every double store.
!  4. Rightmost dest address is odd, line length is odd:
!	Check for end before doing every double store.  End with a single store

! Is rightmost word address even or odd?  Even though we are really
! calculating rDstAddr+rOffset+8, the odd vs. even choice comes out the same
rmvFrameNoShift:
	add	rDstAddr, rOffset, rCount
	andcc	rCount, 0x4, %g0
	bnz	rmvFrameOddDestLine
	andcc	rOffset, 0x4, %g0

! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the original offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
	bz	rmvFrameEvenAddrOddLength
	ld	[rSrcAddr + rOffset], rOut0

! ------------------------------------------------

! EVEN ADDRESS, EVEN LENGTH

! rightmost dest address is even, line length is even
rmvFrameEvenAddrEvenLength:
	subcc	rOffset, 0x8, rOffset		! point to even addr to left
						! Uses INTERLOCK cycle.
	st	rOut0, [rDstAddr + rOffset]	! rDstAddr is 8 too big, so
						! this works ok.
	bneg	rmvFrameEEFin			! quit if this is the last
						! word to be read
	ld	[rSrcAddrP4 + rOffset], rOut4	! read word for odd store

rmvFrameEELoop:
	ld	[rSrcAddr + rOffset], rOut0	! read word for even store
	subcc	rOffset, 0x8, rOffset		! pt to even addr to left
	std	rOut0, [rDstAddr + rOffset]
	bge	rmvFrameEELoop			! more to do?
	ld	[rSrcAddrP4 + rOffset], rOut4	! load odd word

! We get here having just read the last word into rOut4.  Fix up rDstAddr
! and store the word into [rDstAddr].
rmvFrameEEFin:
	sub	rDstAddr, 0x8, rDstAddr
	ba	rmvFrameFromLineFin		! line is finished
	st	rOut4, [rDstAddr]


! ------------------------------------------------

! EVEN ADDRESS, ODD LENGTH

! We do a load and store, but check for the end of line "after" doing the
! store.  Well, we really do the store in the delay slot of the check, so
! rOffset has already been decremented.  To make things work out right, we've
! already pre-incremented rDstAddr by 8.
rmvFrameEvenAddrOddLength:
	subcc	rOffset, 0x8, rOffset		! at end of line yet?
	bneg	rmvFrameEOFin			! yes, fix up rDstAddr
	st	rOut0, [rDstAddr + rOffset]

rmvFrameEOLoop:
	ld	[rSrcAddr + rOffset], rOut0	! load even word
	ld	[rSrcAddrP4 + rOffset], rOut4	! load odd word
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	rmvFrameEOLoop			! no, keep going
	std	rOut0, [rDstAddr + rOffset]

rmvFrameEOFin:
	ba	rmvFrameFromLineFin		! line is finished
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------
! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the original offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
rmvFrameOddDestLine:
	bz	rmvFrameOddAddrOddLength
	subcc	rOffset, 0x4, rOffset		! point to even addr

! ------------------------------------------------

! ODD ADDRESS, EVEN LENGTH

! This case is like even addr, odd length in that we do the check AFTER
! the store.  As above, rDstAddr has been modified to make things work
! out right.
rmvFrameOELoop:
	ld	[rSrcAddr + rOffset], rOut0
	ld	[rSrcAddrP4 + rOffset], rOut4
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	rmvFrameOELoop			! no, keep going
	std	rOut0, [rDstAddr + rOffset]

rmvFrameOEFin:
	ba	rmvFrameFromLineFin		! line is finished
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------

! ODD ADDRESS, ODD LENGTH

! This case is like the even address, even length case in that we quit
! after reading the word at an odd address.
rmvFrameOddAddrOddLength:
	bneg	rmvFrameOOFin
	ld	[rSrcAddrP4 + rOffset], rOut4	! get odd word

rmvFrameOOLoop:
	ld	[rSrcAddr + rOffset], rOut0	! get even word
	subcc	rOffset, 0x8, rOffset		! point to even wd to the left
						! Uses INTERLOCK cycle.
	std	rOut0, [rDstAddr + rOffset]
	bge	rmvFrameOOLoop			! finish rest of line
	ld	[rSrcAddrP4 + rOffset], rOut4

! We've just loaded the last word.  Fix up rDstAddr and store the last word
rmvFrameOOFin:
	sub	rDstAddr, 0x8, rDstAddr
	ba	rmvFrameFromLineFin
	st	rOut4, [rDstAddr]



! ----------------------------------------------------
! ------------- REMOVE FRAME, SHIFT INPUTS ---------------
! ----------------------------------------------------

! This is similar to the code above except each double word is composed
! from bytes from three source words.  The source is word aligned, but
! the dest is not.  We must merge the last dest word on the *left* with
! the previous contents of that word.
! Here are the comments from miframe.c:
	    /* process the line from right to left to keep the inner
	     * loop small.  Since we're generally going to shift
	     * right and we want to use a single offset for both
	     * source and dest words AND we have to do special merging
	     * with the leftmost dest word, make the dest pointer
	     * point to the next higher word boundary.  Each dest
	     * word will be formed from (oldOutputWord |
	     * sourceWord << leftShift) and oldOutputWord =
	     * sourceWord >> rightShift.  Like this:
	     *
	     *	dest |----|----|----|----|----|----|----|----|
	     *		 ^dPtr[0]			  ^dPtr[imageBytes-1]
	     *  src  |----|----|----|----|----|----|----|----|
	     *	      ^sPtr[0]			      ^sPtr[imageBytes-1]
	     * rSrcAddr 0    P4   0    P4   0    P4   0    P4
	     *
	     * leftShift = 8, rightShift = 24; dPtr is set to point
	     * to the second word in the line called dest and we
	     * process from right to left, i.e. from the dest word
	     * containing dPtr[imageBytes-1] and the source word containing
	     * sPtr[imageBytes-1] down to the left.  We stop before touching
	     * the word containing dPtr[0].  That word is processed by
	     * merging the shifted source word with any previous dest
	     * bytes in that word.
	     * Note that since we start with oldInputWord set to 0,
	     * the rightmost word is padded with 0's.
	     */

! We rotate input registers so each input word is read only once.

! When we get here, rShiftRight contains the right shift distance in
! bits.  We compute the left shift distance by subtracting this from
! 32.

! Is rightmost word address even or odd?  Even though we are really
! calculating rDstAddr+rOffset+8, the odd vs. even choice comes out the same
rmvFrameWithShift:
	and	rDstAddr, ~0x3, rDstAddr	! get rid of lower bits
	add	rDstAddr, 4, rDstAddr		! point to higher word
	sub	rRightShift, 32, rLeftShift	! equals -(left shift distance)
	sub	%g0, rLeftShift, rLeftShift	! get the proper sign
	mov	0, rInput0			! pad rightmost dest word
						! with 0's

	add	rDstAddr, rOffset, rCount
	andcc	rCount, 0x4, %g0
	bnz	rmvFrameShiftOddDestLine
	andcc	rOffset, 0x4, %g0

! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the original offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
	bz	rmvFrameShiftEvenAddrOddLength
	mov	0, rInput1			! get right word

! ------------------------------------------------

! EVEN ADDRESS, EVEN LENGTH

! rightmost dest address is even, line length is even
rmvFrameShiftEvenAddrEvenLength:
	ld	[rSrcAddr + rOffset], rInput0	! get middle word
	mov	rInput1, rOut0			! position bit on the right
	sll	rInput0, rLeftShift, rInput1	! position the bits on left
	or	rOut0, rInput1, rOut0		! compose even dest word
	subcc	rOffset, 0x8, rOffset		! point to even addr to left
	st	rOut0, [rDstAddr + rOffset]	! rDstAddr is 8 too big, so
						! this works ok.
	bneg	rmvFrameShiftEEFin		! quit if this is the last
						! word to be read
	ld	[rSrcAddrP4 + rOffset], rInput1	! read lower word for odd store

rmvFrameShiftEELoop:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! get low word for even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! pt to even addr to left
	std	rOut0, [rDstAddr + rOffset]
	bge	rmvFrameShiftEELoop		! more to do?
	ld	[rSrcAddrP4 + rOffset], rInput1	! load low word, odd word

! We get here having just read the low part of the last word into
! rInput1.  Fix up rDstAddr and store the word into [rDstAddr].
rmvFrameShiftEEFin:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	sub	rDstAddr, 0x8, rDstAddr
	mov	rInput1, rInput0		! for consistency
	ba	rmvFrameFin			! line is finished
	st	rOut4, [rDstAddr]


! ------------------------------------------------

! EVEN ADDRESS, ODD LENGTH

! We do a load and store, but check for the end of line "after" doing the
! store.  Well, we really do the store in the delay slot of the check, so
! rOffset has already been decremented.  To make things work out right, we've
! already pre-incremented rDstAddr by 8.
rmvFrameShiftEvenAddrOddLength:
	ld	[rSrcAddr + rOffset], rInput0	! get middle word
	srl	rInput1, rRightShift, rOut0	! position bit on the right
	sll	rInput0, rLeftShift, rInput1	! position the bits on left
	or	rOut0, rInput1, rOut0		! compose even dest word
	subcc	rOffset, 0x8, rOffset		! at end of line yet?
	bneg	rmvFrameShiftEOFin		! yes, fix up rDstAddr
	st	rOut0, [rDstAddr + rOffset]

rmvFrameShiftEOLoop:
	ld	[rSrcAddrP4 + rOffset], rInput1	! load low word, odd store
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! load low word, even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	rmvFrameShiftEOLoop		! no, keep going
	std	rOut0, [rDstAddr + rOffset]

rmvFrameShiftEOFin:
	ba	rmvFrameFin			! line is finished
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------
! If the original offset is 0, 8, 16, etc, the line has an ODD
! number of words in it.  If the original offset is 4, 12, 20, etc
! line has an EVEN number of words.  This is confusing - draw a picture:
! |  0  |  4  |  8  | 12 | ...
rmvFrameShiftOddDestLine:
	mov	0, rInput0			! get high word, odd store
	bz	rmvFrameShiftOddAddrOddLength
	subcc	rOffset, 0x4, rOffset		! point to even addr

! ------------------------------------------------

! ODD ADDRESS, EVEN LENGTH

! This case is like even addr, odd length in that we do the check AFTER
! the store.  As above, rDstAddr has been modified to make things work
! out right.
rmvFrameShiftOELoop:
	ld	[rSrcAddrP4 + rOffset], rInput1	! get low word, odd store
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! get low word, even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! should this be the last store?
	bge	rmvFrameShiftOELoop		! no, keep going
	std	rOut0, [rDstAddr + rOffset]

rmvFrameShiftOEFin:
	ba	rmvFrameFin			! line is finished
	sub	rDstAddr, 0x8, rDstAddr		! repair rDstAddr


! ------------------------------------------------

! ODD ADDRESS, ODD LENGTH

! This case is like the even address, even length case in that we quit
! after reading the word at an odd address.
rmvFrameShiftOddAddrOddLength:
	bneg	rmvFrameShiftOOFin
	ld	[rSrcAddrP4 + rOffset], rInput1	! get low word, odd store

rmvFrameShiftOOLoop:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	ld	[rSrcAddr + rOffset], rInput0	! get low word, even store
	srl	rInput1, rRightShift, rOut0	! form low part, even word
	sll	rInput0, rLeftShift, rInput1	! form hi part, even word
	or	rOut0, rInput1, rOut0		! form even word
	subcc	rOffset, 0x8, rOffset		! point to even wd to the left
	std	rOut0, [rDstAddr + rOffset]
	bge	rmvFrameShiftOOLoop		! finish rest of line
	ld	[rSrcAddrP4 + rOffset], rInput1	! get low word, odd store

! We've just loaded the last word.  Fix up rDstAddr and store the last word
rmvFrameShiftOOFin:
	srl	rInput0, rRightShift, rOut4	! form bottom of odd word
	sll	rInput1, rLeftShift, rInput0	! form top of odd word
	or	rInput0, rOut4, rOut4		! form odd word
	sub	rDstAddr, 0x8, rDstAddr
	mov	rInput1, rInput0
!	ba	rmvFrameFin
	st	rOut4, [rDstAddr]


! ------------------------------------------------
! At this point, all the image data except the leftmost data word has
! been transferred.  The leftover data for that word is in rInput0.
! We read the data from the address pointed to by rDstAddr-4, mask it
! appropriately, or in a shifted version of rInput0 and write it back.
rmvFrameFin:
	ld	[rDstAddr - 4], rInput1		! get leftmost image word
	mov	-1, rOut0			! generate mask
	sll	rOut0, rLeftShift, rOut0
	and	rOut0, rInput1, rOut0		! get old data
	srl	rInput0, rRightShift, rInput0	! shift previous data
	or	rInput0, rOut0, rOut0		! combine old and new data
	st	rOut0, [rDstAddr - 4]

! END OF THIS ROW

! We've put out the right number of words, so quit
rmvFrameFromLineFin:
	mov	0,%i0
	ret
	restore




! ***********************************************************************
! ip_clearRightFrame
!
! This routine cruises down the right frame of an image, looking for frame
! bits to clear.  It doesn't just blindly clear all frame bits to minimize
! write ops.  This is important on the SS2 where writes are expensive.  It
! will probably be important on future SPARC procs also since writes tend to
! cause bus activity.

! We'll get most of our speed by unrolling the loop.  We'll have 8 copies of
! the code.  If we do an extra few lines, it's ok since we'll only be
! clearing stuff in the frame at the bottom of the image.

! We maintain 7 address offsets, one for each of the lines being inspected
! below the top line.  We'll also maintain two temporary data registers so
! we can overlap loading one word with writing out the previous one.

! Undefine previous register definitions to prevent collisions
#undef	rDstAddr
#undef	rSrcAddr
#undef	rLength
#undef	rFrameWidth

#undef	rCount
#undef	rOut0
#undef	rOut4
#undef	rSrcAddrP4
#undef	rOffset

#undef	rSrcAddrM4
#undef	rSrcAddrM8
#undef	rSrcAddrM12
#undef	rSrcAddrM16
#undef	rSrcAddrM20
#undef	rSrcAddrM24

#undef	rOutM8
#undef	rOutM4
#undef	rOutM16
#undef	rOutM12
#undef	rOutM24
#undef	rOutM20

#undef	rRetAddr

/* Boilerplate constants */
#undef	dstAddrOffset
#undef	srcAddrOffset
#undef	lengthOffset
#undef	frmWidthOffset

! REGISTER USAGE (after new register window appears):
! DON'T TOUCH i6 (frame pointer) or i7 (return address), g0 (always 0),
! g6 (??), g7 (??), o6 (stack pointer).
! NOTE: cpp requires c-style comments
#define	rIPState	%i0		/* ptr to source/dest image */
#define	rSrcAddr	%i1		/* ptr to source/dest image */
#define	rSrcPitch	%i2		/* number of bytes/line in image */
#define	rSrcHeight	%i3		/* number of lines to clear */
#define	rImageMask	%i4		/* used to preserve image bits */
#define	rFrameMask	%i5		/* used to detect frame bits */

#define	rSrcPitch2	%g1		/* pts 2 lines below rSrcAddr */
#define	rSrcPitch3	%g2		/* pts 3 lines below rSrcAddr */
#define	rSrcPitch4	%g3		/* pts 4 lines below rSrcAddr */
#define	rSrcPitch5	%g4		/* pts 5 lines below rSrcAddr */
#define	rSrcPitch6	%g5		/* pts 6 lines below rSrcAddr */
#define	rSrcPitch7	%o0		/* pts 7 lines below rSrcAddr */
#define rSrcPitch8	%o1		/* pts 8 lines below rSrcAddr */

/* we don't need this many input registers, but it makes the code
 * easy to read */
#define	rInput0		%l0		/* first input reg */
#define	rInput1		%l1		/* second input reg */
#define	rInput2		%l2		/* third input reg */
#define	rInput3		%l3		/* first input reg */
#define	rInput4		%l4		/* second input reg */
#define	rInput5		%l5		/* third input reg */
#define	rInput6		%l6		/* second input reg */
#define	rInput7		%l7		/* third input reg */

/* Boilerplate constants */
#define	ipStateOffset	0x44
#define	srcAddrOffset	0x48
#define	srcPitchOffset	0x4c
#define	srcHeightOffset	0x50
#define	imageMaskOffset	0x54
#define	frameMaskOffset	0x58


! -----------------------------------------------------------------------------
! _ip_clearRightFrameBlock Routine
! -----------------------------------------------------------------------------

	.seg	"data"
	.seg	"text"
	.global	MK_GLOBAL_NAME(ip_clearRightFrameBlock)
MK_GLOBAL_NAME(ip_clearRightFrameBlock):
	!#PROLOGUE# 0
	sethi	%hi(LF26),%g1		! get new frame ptr and register window
	add	%g1,%lo(LF26),%g1
	save	%sp,%g1,%sp
! at this point in the normal C code, we'd store the incoming vars.  This
! might be some system magic, so we'll keep doing it.  Sometimes they come
! in handy.
	st	rIPState,   [%fp+ipStateOffset]
	st	rSrcAddr,   [%fp+srcAddrOffset]
	st	rSrcPitch,  [%fp+srcPitchOffset]
	st	rSrcHeight, [%fp+srcHeightOffset]
	st	rImageMask, [%fp+imageMaskOffset]
	st	rFrameMask, [%fp+frameMaskOffset]

! Load the source pitch registers
	add	rSrcPitch, rSrcPitch, rSrcPitch2
	add	rSrcPitch2, rSrcPitch, rSrcPitch3
	add	rSrcPitch3, rSrcPitch, rSrcPitch4
	add	rSrcPitch4, rSrcPitch, rSrcPitch5
	add	rSrcPitch5, rSrcPitch, rSrcPitch6
	add	rSrcPitch6, rSrcPitch, rSrcPitch7

! We want to do the 8-at-a-time loop only when there are at least 8 lines
! left.  Yes, there should always be a frame below the end of the image,
! but in reduction, we use frameless intermediate images so that io_reduce
! can share the same code.  In that case, we'll be clearing the pad bits,
! which is ok so long as we don't go below the bottom of the image.
! So, before starting, subtract 8 from the height.  We only do the
! 8-at-a-time loop while the modified height is still less than 8.
	subcc	rSrcHeight, 8, rSrcHeight
	ble	cfbLast8Lines
	add	rSrcPitch7, rSrcPitch, rSrcPitch8	! delay slot

! start the loop.  We read and test each word to see if there are bits in
! the frame.  If so, the frame bits are cleared.  If not, we go on to the
! next word.  Note that we actually go on to the next word while deciding
! about the current word.
cfbLoop:
	ld	[rSrcAddr], rInput0		! read 0th line
	ld	[rSrcAddr+rSrcPitch], rInput1	! read 1st line, fill interlock
	andcc	rInput0, rFrameMask, %g0	! anything in 0th frame?
	bz	cfbTest1			! no, test first line
	ld	[rSrcAddr+rSrcPitch2], rInput2	! load 2nd line in delay slot

! there was something to clear in 0th line
	and	rInput0, rImageMask, rInput0	! clear frame part
	st	rInput0, [rSrcAddr]

cfbTest1:
	andcc	rInput1, rFrameMask, %g0	! anything in 1st frame?
	bz	cfbTest2			! no, test 2nd line
	ld	[rSrcAddr+rSrcPitch3], rInput3	! load 3rd line in delay slot

! there was something to clear in 1st line
	and	rInput1, rImageMask, rInput1	! clear frame part
	st	rInput1, [rSrcAddr+rSrcPitch]

cfbTest2:
	andcc	rInput2, rFrameMask, %g0	! anything in 2nd frame?
	bz	cfbTest3			! no, test 3rd line
	ld	[rSrcAddr+rSrcPitch4], rInput4	! load 4th line in delay slot

! there was something to clear in 2nd line
	and	rInput2, rImageMask, rInput2	! clear frame part
	st	rInput2, [rSrcAddr+rSrcPitch2]

cfbTest3:
	andcc	rInput3, rFrameMask, %g0	! anything in 3rd frame?
	bz	cfbTest4			! no, test 4th line
	ld	[rSrcAddr+rSrcPitch5], rInput5	! load 4th line in delay slot

! there was something to clear in 3rd line
	and	rInput3, rImageMask, rInput3	! clear frame part
	st	rInput3, [rSrcAddr+rSrcPitch3]

cfbTest4:
	andcc	rInput4, rFrameMask, %g0	! anything in 4th frame?
	bz	cfbTest5			! no, test 5th line
	ld	[rSrcAddr+rSrcPitch6], rInput6	! load 6th line in delay slot

! there was something to clear in 4th line
	and	rInput4, rImageMask, rInput4	! clear frame part
	st	rInput4, [rSrcAddr+rSrcPitch4]

cfbTest5:
	andcc	rInput5, rFrameMask, %g0	! anything in 5th frame?
	bz	cfbTest6			! no, test 6th line
	ld	[rSrcAddr+rSrcPitch7], rInput7	! load 7th line in delay slot

! there was something to clear in 5th line
	and	rInput5, rImageMask, rInput5	! clear frame part
	st	rInput5, [rSrcAddr+rSrcPitch5]

cfbTest6:
	andcc	rInput6, rFrameMask, %g0	! anything in 6th frame?
	bz	cfbTest7			! no, test 7th line
	andcc	rInput7, rFrameMask, %g0	! anything in 7th frame?

! there was something to clear in 6th line
	and	rInput6, rImageMask, rInput6	! clear frame part
	st	rInput6, [rSrcAddr+rSrcPitch6]

cfbTest7:
	bz	cfbLoopBot			! no, see if we're done
	subcc	rSrcHeight, 8, rSrcHeight	! decrement line count

! there was something to clear in 7th line
	and	rInput7, rImageMask, rInput7	! clear frame part
	st	rInput7, [rSrcAddr+rSrcPitch7]

cfbLoopBot:
	bg	cfbLoop
	add	rSrcAddr, rSrcPitch8, rSrcAddr	! point to next block


! process the last 7 or fewer lines one at a time.  First, fix up rSrcHeight.
cfbLast8Lines:
	addcc	rSrcHeight, 8, rSrcHeight	! compensate for initial
	bz	cfbFin				! decrement, quit if given a
						! initial height of 0
! do 1 line
	ld	[rSrcAddr], rInput0		! read 0th line
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr]

! do 2 lines
	ld	[rSrcAddr+rSrcPitch], rInput0	! read 1st line, fill interlock
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr+rSrcPitch]

! do 3 lines
	ld	[rSrcAddr+rSrcPitch2], rInput0	! read 1st line, fill interlock
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr+rSrcPitch2]

! do 4 lines
	ld	[rSrcAddr+rSrcPitch3], rInput0	! read 1st line, fill interlock
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr+rSrcPitch3]

! do 5 lines
	ld	[rSrcAddr+rSrcPitch4], rInput0	! read 1st line, fill interlock
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr+rSrcPitch4]

! do 6 lines
	ld	[rSrcAddr+rSrcPitch5], rInput0	! read 1st line, fill interlock
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr+rSrcPitch5]

! do 7 lines
	ld	[rSrcAddr+rSrcPitch6], rInput0	! read 1st line, fill interlock
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr+rSrcPitch6]

! do 8 lines
	ld	[rSrcAddr+rSrcPitch7], rInput0	! read 1st line, fill interlock
	deccc	rSrcHeight			! all done?
	and	rInput0, rImageMask, rInput0	! clear frame part
	bz	cfbFin				! yes, quit after this store
	st	rInput0, [rSrcAddr+rSrcPitch7]

! We've cleared the block, so quit
cfbFin:
	mov	0,%i0
	ret
	restore


/*

$Log: ssframe.s,v $
Revision 1.1.1.1  1999/09/20 18:18:18  jlin
Imported sources

Revision 1.11  1998/01/13 23:57:43  ddavies
ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v

CommonCheck: Add IPState parameter to almost all private routines.

! Revision 1.10  1994/07/15  22:35:35  ddavies
! ddavies on Fri Jul 15 15:33:00 PDT 1994
!
! ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v
!
! Fix "bug" that appears when you have an image that does not have a frame.
! In that case, ip_clearRightFrameBlock should really clear exactly the pad
! bits in the rightmost image word.  It was clearing in units of 8 lines, so
! it cleared too much below the bottom of the image if there was no
! bottom frame.  This happened when doing reduction with intermediate
! frameless images.
!
! Revision 1.9  1994/07/11  23:16:42  ddavies
! ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v
!
! Replace a calculation of rightMask with a variable rightMask that's passed down
! from i_addFrame to ip_addFrameToLine.  This allows the latter routine to
! correctly clear exactly the frame bits in the rightmost image word.  We
! had been assuming that unframed images are padded on the right to byte
! boundaries with 0's.  This is not true of images that have been rotated
! in place.
!
! Revision 1.8  1994/01/04  03:12:33  ddavies
! ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v
!
! Fixed two bugs in ip_removeFrameFromLine that were left over from the
! pre-Holiday rush.  The routine was modifying the rightmost word of the
! *previous* scan line when no shifting was done.  It was also writing
! the leftmost word of each scan line with the wrong data when shifting
! was required.  It carefully computed rOut0, then stored rInput0.
!
! Revision 1.7  1993/12/22  04:59:51  ddavies
! ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v
!
! THIS DOESN'T WORK, unless it's pretty close.  The implementation of
! ip_addFrameToLine should be ok, it hasn't changed.  The implementation of
! ip_removeFrameFromLine is known to be slightly off on the right edge of
! an 8bpp image.  I never got to try it on 1 bpp images.  I'll fix it after
! Christmas.
!
! Revision 1.6  1993/12/14  21:27:38  ddavies
! ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v
!
! Changed ip_addFrameToLine so that it deals with frameless images whose
! bytes/line is not a multiple of 4.  This involves shifting and masking
! during the data copying step.  Did not touch ip_removeFrameFromLine
! since no one wanted removed images to have scan line pitches that aren't
! a multiple of 4 (yet).
!
! Revision 1.5  1993/12/04  04:59:59  ddavies
! ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v
!
! Make all global variables use the MK_GLOBAL_NAME macro.  See glblname.prv.
!
! Revision 1.4  1993/11/05  02:24:08  ddavies
! ssframe.s -> /product/ipcore/ipshared/src/RCS/ssframe.s,v
!
! Add RCS magic to have string specifying this file's name, rev number and
! last edit date into the object file.  Also add change log at the bottom.
!

*/

