/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

/*
 *  bitacces.h
 *
 *         Inlines for bit access within an array of (unsigned) chars
 *                        TEST_CHAR_ARRAY_BIT
 *                        GET_CHAR_ARRAY_BIT
 *                        SET_CHAR_ARRAY_BIT
 *                        CLEAR_CHAR_ARRAY_BIT
 */

#ifndef BITACCES_H
#define BITACCES_H

#ifdef _MSC_VER
#pragma pack(push, bitacces_h, 1)
#endif

IP_RCSINFO(bitacces_h_RCSInfo, "$RCSfile: bitacces.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:10 $ */


/******************************************************************
 *****   Inlines for bit access within an array of chars      *****
 *****************************************************************/
/*
 *  TEST_CHAR_ARRAY_BIT:  takes a ptr to an array of bytes and a bit index n
 *                unsigned char   *data
 *                int              n
 *            returns 0 if the n-th bit is clear; nonzero if set.
 */
#define TEST_CHAR_ARRAY_BIT(data, n) (*((data) + ((n) >> 3)) & (128 >> ((n) & 7)))


/*
 *  GET_CHAR_ARRAY_BIT:   takes a ptr to an array of bytes and a bit index n
 *                unsigned char   *data
 *                int              n
 *            returns 0 if the n-th bit is clear; 1 if set.
 */
#define  GET_CHAR_ARRAY_BIT(data, n)  ((*((data) + ((n) >> 3)) & (128 >> ((n) & 7))) ? 1 : 0)
 

/*
 *  SET_CHAR_ARRAY_BIT:  takes a ptr to an array of bytes and a bit index n
 *                unsigned char   *data
 *                int              n
 *            sets the referenced bit to 1
 */
#define  SET_CHAR_ARRAY_BIT(data, n)   (*((data) + ((n) >> 3)) |= (128 >> ((n) & 7)))


/*
 *  CLEAR_CHAR_ARRAY_BIT:  takes a ptr to an array of bytes and a bit index n
 *                unsigned char   *data
 *                int              n
 *            sets the referenced bit to 0
 */
#define  CLEAR_CHAR_ARRAY_BIT(data, n)   (*((data) + ((n) >> 3)) &= (255 ^ (128 >> ((n) & 7))))

#ifdef _MSC_VER
#pragma pack(pop, bitacces_h)
#endif

#endif /* bitacces_h */

/*

$Log: bitacces.h,v $
Revision 1.1.1.1  1999/09/20 18:19:10  jlin
Imported sources

Revision 1.10  1998/10/05 19:12:34  aber
aber on Mon Oct 5 12:12:34 PDT 1998

bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v

CommonCheck: Fix module name in all-encompassing ifdef to be upper-case, to avoid
naming conflict with packing pragma name.

Revision 1.9  1998/10/02 18:24:41  aber
bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v

CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
using libraries and includes.

Revision 1.8  1995/12/13 04:04:14  ejaquith
bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v

CommonCheck: Added ";" to IP_RCSINFO().
Removed function comment blocks from before prototypes and
added special prototype comment.

 * Revision 1.7  1995/08/31  22:02:44  ejaquith
 * bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1994/09/23  00:46:19  jcraig
 * jcraig on Thu Sep 22 17:45:52 PDT 1994
 *
 * bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v
 *
 * remove macros which are duplicated in imageref.pub
 *
 * Revision 1.5  1994/09/08  00:32:15  lperry
 * bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.4  1994/01/15  03:45:26  jcraig
 * jcraig on Fri Jan 14 19:44:44 PST 1994
 *
 * bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v
 *
 * add macros GET_IMAGE_BITS, SET_IMAGE_1BIT, SET_IMAGE_2BITS,
 * SET_IMAGE_4BITS and SET_IMAGE_BITS.
 *
 * Revision 1.3  1993/12/03  01:09:43  ddavies
 * bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  02:59:34  ddavies
 * bitacces.h -> /product/ipcore/ipwrap/include/RCS/bitacces.h,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

*/

