/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


#ifndef _DESCRIP_H_INCLUDED_
#define _DESCRIP_H_INCLUDED_

#ifdef _MSC_VER
#pragma pack(push, descrip_h, 1)
#endif

IP_RCSINFO(descrip_h_RCSInfo, "$RCSfile: descrip.h,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:10 $");



#define DESCRIP_MAGIC           ((UInt16)0xfcf0)
#define DESCRIP_BYTES_HEAD_SIZE 8
#define CURRENT_DESCRIP_VERSION 2
#define NULL_CHAR               '\0'

/* The immediate concerns are to make our current form registration package
 * more accurate and faster.  There are certain kinds of information that
 * we would like to add in to the compiling package and save with the FCF
 * that will help towards this end.  Additionally, none of the modules
 * (landmark, Hausdorff, glyph) talk to each other very well and none allow
 * 2-way communication.  We can use shared information with these modules
 * that will also help in our speed and accuracy quest.
 * I will work at this task for Smartpaper to get these mentioned changes
 * implemented.  More specifically, I will be modifying the FCF format as
 * mentioned later.  Additionally, I will be changing the structure of
 * both the registration and compilation modules to accomodate these
 * changes.  Finally, I will create a very flexible and more intuitive
 * routine at the ipwrap level that I will refer to as the "matching module".
 * First, let's look at some terminology, along with how the structures
 * and FCF file will look.
 */


/* WARNING!!
 * Cannot change the ordering to this AND cannot remove any of the
 * entries.  The ONLY thing you can do is ADD more entry types to the
 * [now "virtual"!] enumerated list.
 */
#define BADLMTYPE			0
#define BIGHORIZONTAL			1
#define BIGVERTICAL			2
#define GLYPHSLM			3
#define TEXTBOXES			4
#define TEXTSEGMENTS 			5
typedef Int32 landmarktype;

typedef struct textSegmentsType
{
    Int32      iterations;
    Int32      nLandmarks;
    RECTANGL  *landmarks;
} textSegmentsType;

typedef struct unknownLM
{
    UInt32     nBytesInBuffer;
    UInt8     *buffer;
} unknownLM;

typedef struct landmarksSlot
{
    landmarktype     type;  /* Type of landmarks in this slot */
    void            *lms;   /* Ptr to landmarks in this slot
			     * (Must be cast to one of above struct ptrs) */
} landmarksSlot;


typedef unsigned long SLOTTYPEIO;



/* Each potential type of mark on the page will be referred to as a class.
 * These are the types of classes that we will use.
 * The first 9 will be renderable checkmarks that can use the Hausdorff
 * module for matching.  (Note that if a filled in rectangular mark is
 * desired, it can be done using a very thick line).
 * The next 6 are "pre-checked" marks that may appear (and are also renderable
 * and can use the Hausdorff module for matching).
 * The next 4 are other kinds of classes that may appear on the page
 * that can be used to add information to the matcher and help in calculation
 * of movements from checkmarks from their ideal location on the ideal page
 * (the page the user drafted up and made an fcf from) to their actual
 * location on the marked up page we register.
 * Finally, 2 more kinds are added for potential future uses of Doug Ferrell's
 * circle detection code and similar kinds of marks.
 * 
 * WARNING!!  Once this code is released, the order of these MAY NOT CHANGE!!
 *  And further NONE CAN BE DELETED!!  (Otherwise, files already written out
 *  with this list would be read in and get interpreted to be the wrong class)
 *  Adding to the end of this list is fine.
 */
#define RECTANGLE			0
#define LINE				1
#define BUBBLE				2
#define CHECKBOX_BOTRITE		3
#define CHECKBOX_TOPRITE		4
#define CHECKBOX_BOTLEFT		5
#define CHECKBOX_TOPLEFT		6
#define CLIPREGION			7
#define BOXSIGNATURE			8
#define RECTANGLE_CHECKED		9
#define BUBBLE_CHECKED			10
#define CHECKBOX_BOTRITE_CHECKED	11
#define CHECKBOX_TOPRITE_CHECKED	12
#define CHECKBOX_BOTLEFT_CHECKED	13
#define CHECKBOX_TOPLEFT_CHECKED	14
#define GLYPHS				15
#define LAND_MARK			16
#define AREA				17
#define TILE				18
#define DRAWN_CIRCLE			19
#define DRAWN_RECTANGLE			20
#define BAD				21
typedef Int32 classKind;


/* This will be used for each instance to allow the matching module to
 * know the state of each instance.  For example, we might match 20% of
 * instances on the page using Hausdorff methods, not be able to find
 * matches for 5% of the instances when we tried using Hausdorff methods,
 * have the caller have already found some instances, and interpolate
 * the remaining instances that were not found by any of the mentioned
 * methods.
 */
#define USER_GUESS			0
#define USER_FOUND			1
#define INTERPOLATED			2
#define HAUSDORFF_FOUND			3
#define HAUSDORFF_MISSED		4
typedef Int32 instStatus;


/* WART for HAUSDORFF_MISSED will need to set some flag in internal stuff */

/* Each potential type of mark on the page will be referred to as a class.
 * Not every field will be used with each class.
 */
typedef struct classType
{
    classKind    kind;   /* Kind of class this is */
    Int32        w;      /* Width of bitmap, w/o frame or pad */
    Int32        h;      /* Height of bitmap, w/o frame or pad */
    Int32        wThick; /* Thickness of vertical bitmap pieces */
    Int32        hThick; /* Thickness of horizontal bitmap pieces */

    modelType   *model;  /* Model with bitmap,etc of this class */
    PIXR        *srcPr;  /* Ptr to source pixr where bitmap came from/is */
    void        *extra;  /* Extra field client may use */
} classType;

/* Each actual mark on the page will be an instance.  Instances will be
 * specific locations of classes.  (see below description).
 */
typedef struct instanceType
{
    Float64      percentAtoB;     /* Hausdorff percents filled if/when a */
    Float64      percentBtoA;     /*   Hausdorff match is found */

    Float64      guessScalex;     /* Passed in best guess at scale in */
    Float64      guessScaley;     /*   x and y. */
    Float64      scalex;          /* Scale determined via interopolation or */
    Float64      scaley;          /*   Hausdorff match */
    
    instStatus   status;          /* State this instance is in */

    Int32        guessx;          /* Location where user believes instance */
    Int32        guessy;          /*   should be in ideal page */
    
    Int32        rectx;           /* Maximal subrect around above guess */
    Int32        recty;           /*   location that we should search for a */
    Int32        rectw;           /*   match in with Hausdorff techniques */
    Int32        recth;           /*   (rectx <= guessx, recty <= guessy
				   *    rectw >= class' w, recth >= class' h)
				   */

    char        *name;            /* Optional ptr to name string of instance */
    void        *extra;           /* Extra field client may use */

    Int16        classNo;         /* Class this instance belongs to */
    
    Int16        guessTransx;     /* Passed in best guess at translation in */
    Int16        guessTransy;     /*   x and y. (relative to guessx,guessy) */
    Int16        transx;          /* Translation determined via interpolation*/
    Int16        transy;          /*   or Hausdorff match
				   *   (relative to guessx,guessy) */

    UInt16       sampleScore;     /* A higher score indicates this instance
				   *   would make a better sample when finding
				   *   a sampling of instances on the page.
				   */
    UInt8        rowColRims;      /* If this ANDed w/ (..) isn't 0 then (..)
				   * 0x80  start of row and/or column
				   * 0x40  part of row of instances
				   * 0x20  part of column of instances
				   * 0x10  end of row and/or column
				   * If this ANDed w/ (..) isn't 0 then (..)
				   * 0x08  top
				   * 0x04  bottom
				   * 0x02  left
				   * 0x01  right
				   * rim around instance is at most 10% sparse
				   */
} instanceType;

typedef struct groupType
{
    Int16      nInstances;  /* Number of instances in this group(ing) */
    Int16     *instances;   /* An array of instance indices in instance array*/
    Int16      minChecks;   /* User expected to check AT LEAST this many and */
    Int16      maxChecks;   /* AT MOST this many instances in group. */
} groupType;



typedef struct pageDataType 
{
    classType           *classes;
    instanceType        *instances;
    groupType           *groups;
    landmarksSlot       *lmSlots;
    struct pageDataType *pNext;

    Int32                nClasses, nInstances, nGroups, nSlots;

    /* Page value which is what user calls it.  Eg: they first generate a
     * description, and want to refer to it as page #77.
     * The _user_ can then set pageIndex to 77.
     */
    Int16                pageIndex;

    /* Resolution that the classes bitmaps were rendered with;
     * that any himetric-->pixel conversion gets made with.
     */
    Int32                xres, yres;
} pageDataType;


typedef struct correlateType
{
    Int32              nSlots;
    Int32              estScaleX, estScaleY;
    Int32            **maparray;
} correlateType;

#ifdef _MSC_VER
#pragma pack(pop, descrip_h)
#endif

#endif /* _DESCRIP_H_INCLUDED_ */


/*
  $Log: descrip.h,v $
  Revision 1.1.1.1  1999/09/20 18:19:10  jlin
  Imported sources

  Revision 1.4  1998/10/02 18:25:06  aber
  descrip.h -> /product/ipcore/ipwrap/include/RCS/descrip.h,v

  CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
  using libraries and includes.

  Revision 1.3  1996/03/27 03:11:35  ejaquith
  ejaquith on Tue Mar 26 19:07:42 PST 1996

  descrip.h -> /product/ipcore/ipwrap/include/RCS/descrip.h,v

  Removed usage of enumerated lists for various kinds of information
  like what class a special (check)mark is, what kind of landmarks a
  set of landmarks are, and the result/status of a special (check)mark
  on a page (found, not found, guessed, interpolated).
  This should have never made it in.  I, uh, must have been snowboarding.

  Revision 1.2  1996/02/25 01:22:49  ejaquith
  descrip.h -> /product/ipcore/ipwrap/include/RCS/descrip.h,v

  Oops.  Hadn't added RCS info and log (until now).


  */
