/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef GFIOPRO_H
#define GFIOPRO_H


/*
 * gfiopro.h - generic file I/O platform independent prototypes.
 *
$Log: gfiopro.h,v $
Revision 1.1.1.1  1999/09/20 18:19:11  jlin
Imported sources

Revision 1.4  1998/10/05 19:12:40  aber
gfiopro.h -> /product/ipcore/ipwrap/include/RCS/gfiopro.h,v

CommonCheck: Fix module name in all-encompassing ifdef to be upper-case, to avoid
naming conflict with packing pragma name.

Revision 1.3  1998/10/02 18:25:38  aber
gfiopro.h -> /product/ipcore/ipwrap/include/RCS/gfiopro.h,v

CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
using libraries and includes.

Revision 1.2  1995/07/31 17:59:53  ejaquith
gfiopro.h -> /product/ipcore/ipwrap/include/RCS/gfiopro.h,v

Standardizing changes -- first pass.

 * Revision 1.1  1994/10/21  17:19:48  ejaquith
 * Initial revision
 *
 * 
 *    Rev 1.1   17 Sep 1994 13:08:14   SMARTIN
 * Added GFIO_Error() 
 * 
 *    Rev 1.0   13 Aug 1994 17:55:20   EHOPPE
 * Initial revision.
 * 
 *    Rev 1.4   25 May 1994 23:16:10   SMARTIN
 * No change.
 * 
 *    Rev 1.3   24 May 1994 15:02:32   smartin
 * 
 *    Rev 1.2   19 May 1994 13:22:12   smartin
 * 
 *    Rev 1.1   18 May 1994 13:33:38   smartin
 * No change.
 * 
 *    Rev 1.0   16 May 1994 12:48:16   smartin
 * Initial revision.
 * Revision 1.3  1994/05/04  23:01:09  dferrell
 * dferrell on Wed May 4 16:00:05 PDT 1994
 *
 * gfiopro.h -> /product/amber/gfio/include/RCS/gfiopro.h,v
 *
 * Corrected the comment header for GFIO_OpenFile.  Specifically,
 * the function returns GFIO_FAILURE if the file was already open.
 *
 * Revision 1.2  1994/04/26  19:06:03  dferrell
 * dferrell on Tue Apr 26 12:05:48 PDT 1994
 *
 * gfiopro.h -> /product/amber/gfio/include/RCS/gfiopro.h,v
 *
 * Added prototype for GFIO_FileSize ().
 *
 * Revision 1.1  1994/04/18  18:47:13  dferrell
 * Initial revision
 *
 */

#ifdef _MSC_VER
#pragma pack(push, gfiopro_h, 1)
#endif

/* These definitions might should be put in a different header...
 */

/* file open modes */
#define  GFIO_READ  1
#define  GFIO_WRITE 2
#define  GFIO_UPDATE 4

/* error returns */
#define  GFIO_SUCCESS  0
#define  GFIO_FAILURE -1
#define  GFIO_EOF     -2

/*************************************************************
 * GFIO_CreateTempFile () - Create a temporary file.
 *             
 * Inputs:
 *             None
 *
 * Returns:
 *             A pointer to the file token for the new file, or
 *             NULL if the file could not be opened.
 *
 * Notes:
 *             The platform specific information for files
 *             is stored in the file tokens which are passed
 *             in to GFIO.
 *
 * See Also: 
 *             GFIO_DestroyTempFile ()
 */
void * CDECL GFIO_CreateTempFile (void);

/*************************************************************
 * GFIO_DestroyTempFile () - Remove a temp file from the file 
 *             system.
 *             
 * Inputs:
 *             ft - the file token associated with the file to
 *                  be removed.
 *
 * Returns:
 *             Nothing
 *
 * See Also: 
 *             GFIO_CreateTempFile ()
 */
void CDECL GFIO_DestroyTempFile (void *ft);

/*************************************************************
 * GFIO_OpenFile () - Open a file.
 *             
 * Inputs:
 *             ft   - the file token associated with the file to
 *                    be removed.
 *             mode - GFIO_READ, GFIO_WRITE, or both (ORing)
 *
 * Returns:
 *             GFIO_SUCCESS - file opened correctly
 *             GFIO_FAILURE - error occured while opening the
 *                            file or the file was already open
 *
 * Notes:
 *             The current position in the file is 0 upon return.
 * See Also: 
 *             GFIO_CloseFile ()
 */
int CDECL  GFIO_OpenFile (void *ft, int mode);

/*************************************************************
 * GFIO_CloseFile () - Close a file.
 *             
 * Inputs:
 *             ft   - the file token associated with the file to
 *                    be removed.
 *             
 * Returns:
 *             GFIO_SUCCESS - file closed properly
 *             GFIO_Failure - error closing file
 *
 * See Also: 
 *             GFIO_OpenFile ()
 */
int CDECL GFIO_CloseFile (void *ft);

/*************************************************************
 * GFIO_Read () - Read from a file to a buffer.
 *             
 * Inputs:
 *             ft        - the file token associated with the file
 *                         from which to read
 *             buffer    - array to read bytes into, must be at
 *                         least byteCount characters in length
 *             byteCount - number of bytes to read
 *
 * Returns:
 *             number of bytes actually read into the buffer, or
 *             GFIO_EOF - attempted read at EOF
 *             GFIO_FAILURE - other error occured (file not open)
 *
 * Notes:
 *             Reads data starting at the current position in the
 *             file and increments the position by byteCount bytes.
 *             If during the read EOF is encountered, returns the
 *             number of bytes read up to that point.  I already at
 *             EOF when read specified, returns GFIO_EOF.
 *
 * See Also: 
 *             GFIO_Write ()
 */
long int CDECL GFIO_Read (void *ft, char *buffer, unsigned long byteCount);

/*************************************************************
 * GFIO_Write () - Write a buffer of bytes to a file.
 *             
 * Inputs:
 *             ft        - the file token associated with the file
 *                         in which to write
 *             buffer    - array of characters to be written, must
 *                         be at least byteCount characters in length
 *             byteCount - number of bytes to write
 *             
 * Returns:
 *             number of characters written into the file, or
 *             GFIO_FAILURE - the file is not open, not
 *                            writeable, or an error occured.
 *
 * Notes:
 *             Writes data starting at the current position in the
 *             file and increments the position by byteCount bytes.
 *
 * See Also: 
 *             
 */
long int CDECL GFIO_Write (void *ft, char *buffer, unsigned long byteCount);

/*************************************************************
 * GFIO_Seek () - Seek to an arbitrary position in a file.
 *             
 * Inputs:
 *             ft     - the file token associated with the file
 *                      in which to seek
 *             offset - number of bytes from the beginning of
 *                      the file to set the current position
 *
 * Returns:
 *             how far the current position offset was actually
 *             set from the beginning of the file, or
 *             GFIO_EOF - attempt to seek past EOF
 *             GFIO_FAILURE - another error occured
 *
 * Notes:
 *             If the file is not open, returns GFIO_FAILURE.
 *
 * See Also: 
 *             
 */
long int CDECL GFIO_Seek (void *ft, unsigned long offset);

/**************************************************************
 * GFIO_FileSize() - Find out the number of bytes in a file
 *
 * Inputs:
 *             ft     - the file token associated with the file
 *                      in which to get the size
 *
 * Returns:
 *             the size of the file in bytes or GFIO_FAILURE
 *
 */
long int CDECL GFIO_FileSize (void *ft);

/**************************************************************
 * GFIO_Error() - checks if a file error occurred
 *
 * Inputs:
 *             ft     - the file token associated with the file
 *                      in which to get the error
 *
 * Returns:
 *             return 0 if no error, non-zero if error
 *
 */
long int CDECL GFIO_Error (void *ft);


#ifdef _MSC_VER
#pragma pack(pop, gfiopro_h)
#endif

#endif /* GFIOPRO_H */
