/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef IMAGEIO_H
#define IMAGEIO_H

#ifdef _MSC_VER
#pragma pack(push, imageio_h, 1)
#endif

/* imageio.c
 *
 * $Header: /project/icons/cvsroot/ipcore/src/ipwrap/include/imageio.h,v 1.2 2000/02/29 00:24:37 jlin Exp $
 *
 * DESCRIPTION
 *   Declarations for the non-TIFF file reading/writing interface, imageio.
 *
 * $Log: imageio.h,v $
 * Revision 1.2  2000/02/29 00:24:37  jlin
 * Add strong arm related changes
 *
 * Revision 1.1.1.1  1999/09/20  18:19:11  jlin
 * Imported sources
 *
 * Revision 1.4  1998/10/02 18:25:57  aber
 * imageio.h -> /product/ipcore/ipwrap/include/RCS/imageio.h,v
 *
 * CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
 * using libraries and includes.
 *
 * Revision 1.3  1997/06/30 23:17:21  ddavies
 * imageio.h -> /product/ipcore/ipwrap/include/RCS/imageio.h,v
 *
 * Comment out all prototypes.  They're not used anywhere and aren't valid.
 * SOMEDAY, someone should clean up this pcx/imageio mess.
 *
 * Revision 1.2  1995/07/31 18:00:02  ejaquith
 * imageio.h -> /product/ipcore/ipwrap/include/RCS/imageio.h,v
 *
 * Standardizing changes -- first pass.
 *
 * Revision 1.1  1994/10/21  17:19:51  ejaquith
 * Initial revision
 *
 * 
 *    Rev 1.11   18 Oct 1994 18:46:48   EHOPPE
 * 
 * Added debug_file stream for DEBUG printf's in imageio.
 * 
 *    Rev 1.10   26 Sep 1994 15:29:16   EHOPPE
 * Added use of resolution information.  Removed localized reverse contrast
 * kludges.  Fixed THE reverse contrast problem in imageio.  Still need to add
 * 'invert' flag for other clients.  Fixed gray images read as color (GIF, BMP
 * Added byte swapping code and removed nonportable language features.
 */

/*
 * INCLUDES
 */

/*
 * CONSTANTS
 */

/* Use these values for nonTIFF, too. --EHoppe */
/* TIFF equates for Tag names and Field types */
#define II 0x4949 /* Byte orders */
#define MM 0x4D4D

/**************************************
 * This is changed by Ron Federic for ARM platform
 *This fixes a bug in figuring out enddianness on Win32 systems, and also adds
 *"arm" as another little-endian system. The symbol "arm" is predefined by the
 *compiler -- I used it instead of something like _ARM_ARCH_ because the other
 *symbols like __WATCOMC__ it was already using were predefined compiler symbols
 *as well.
 *#ifdef __WATCOMC__
 *#define NATIVEORDER II
 *#else
 *#define NATIVEORDER MM
 *#endif
 **************************************/
 
#if defined(__WATCOMC__) || defined(_WIN32) || defined(WIN32) || defined(arm)
#define NATIVEORDER II
#else
#define NATIVEORDER MM
#endif

#define IMAGEIO_PCX     1
#define IMAGEIO_BMP     2
#define IMAGEIO_DCX     3
#define IMAGEIO_TIFF    4
#define IMAGEIO_JFIF    5
#define IMAGEIO_PHOTOCD 6
#define IMAGEIO_GIF     7
#define IMAGEIO_EPS     8
#define IMAGEIO_UNKNOWN 9

#define IMAGEIO_NOERROR         0   
#define IMAGEIO_NOMORE          6   
#define IMAGEIO_ERROR_MEMORY    0x5001
#define IMAGEIO_ERROR_FILE      0x5002
#define IMAGEIO_ERROR           0x5005
#define IMAGEIO_ERROR_PARAMETER 0x5006
#define IMAGEIO_ERROR_NOSUPPORT 0x5007

/* these are the high level types of images */
#define IMAGEIO_BINARY     1
#define IMAGEIO_GRAY16     4
#define IMAGEIO_PALETTE16  5
#define IMAGEIO_GRAY256    8
#define IMAGEIO_PALETTE256 9
#define IMAGEIO_FULLCOLOR  24

#define IMAGEIO_ORIGIN_TOPLEFT      1
#define IMAGEIO_ORIGIN_BOTTOMLEFT   2

/*
 * MACROS
 */

/*
 * TYPEDEFS
 */

/*
 * ENUMS
 */

/*
 * GLOBAL VARIABLE DECLARATIONS
 */

/*
 * EXTERNAL DECLARATIONS
 */
 
#ifdef DEBUG
extern FILE *debug_file;
#endif

/*
 * FUNCTION PROTOTYPES
 * We don't use any of these to implement pcx.  Gag, what a mess.
 */

#if 0
void swap(Int8* x);
void swapl(Int8* x);
void swapd(Int8* x);

GDATA* CDECL imageioAllocGData();
void   CDECL imageioFreeGData( GDATA *pGData );
UInt32 CDECL imageioOpenRead( UInt8 format, GDATA *pGData, void *gfioToken );
UInt32 CDECL imageioOpenWrite( UInt8 format, GDATA *pGData, void *gfioToken );
UInt32 CDECL imageioReadNextScanline( GDATA *pGData, char *buffer );
UInt32 CDECL imageioReadNextScanlineConverted( GDATA *pGData, char *buffer, UInt16 type );
UInt32 CDECL imageioWriteNextScanline( GDATA *pGData, char *buffer );
UInt32 CDECL imageioGotoNextPage( GDATA *pGData );
UInt32 CDECL imageioClose( GDATA *pGData );

UInt32 CDECL imageioGetPalette( GDATA *pGData, char *palette_buffer );
UInt32 CDECL imageioSetPalette( GDATA *pGData, char *palette_buffer );
UInt32 CDECL imageioGetFormatOrigin( UInt8 format );

UInt8  CDECL imageioGetFileType( void *gfioToken );
UInt32 CDECL imageioGetDepth( GDATA *pGData );
UInt32 CDECL imageioGetBPL( GDATA *pGData );
#endif

#ifdef _MSC_VER
#pragma pack(pop, imageio_h)
#endif

#endif
