/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


#ifndef  IMAGEREP_H_INCLUDED
#define  IMAGEREP_H_INCLUDED

#ifdef _MSC_VER
#pragma pack(push, imagerep_h, 1)
#endif

IP_RCSINFO(imagerep_h_RCSInfo, "$RCSfile: imagerep.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:11 $ */

/*
 *  image file format types
 */
#define    NO_TYPE         	0 
#define    SUNRAS_COMP  	1
#define    SUNRAS       	2
#define    PS           	4
#define    FF_BMP		5
#define	   PNM			6
#define	   PCX			7

#define    TIFF_DEFAULT		30  /* RW    G3 if pixr depth==1, else LZW */
#define	   TIFF_UNCOMP		31  /* RW    No compression */
#define	   TIFF_LZW		32  /* RW    Lempel-Ziv & Welch */
#define	   TIFF_G3		33  /* RW 1  CCITT Group 3 fax encoding */
#define	   TIFF_G4		34  /* RW 1  CCITT Group 4 fax encoding */
#define	   TIFF_PACKBITS	35  /* RW    Macintosh Run Length Encoding */
#define	   TIFF_JPEG		36  /*       JPEG compression */
#define	   TIFF_THUNDERSCAN	37  /* R     ThunderScan RLE */
#define	   TIFF_NEXT		38  /* R     NeXT 2-bit RLE */
#define	   TIFF_RLE		39  /* RW 1  Modified Huffman RLE */

#define    TIFF_G3_2D		40  /* RW 1  CCITT Group 3 with 2D option */
#define    TIFF_G3_PAD		41  /* RW 1  CCITT G3 with pad before EOL */
#define    TIFF_G3_2D_PAD	42  /* RW 1  CCITT G3 with 2D and pad */
#define    TIFF_LOSSYDITHER	43  /* RW 1  Lossy Dither encoding */

/* TYPES DEFINED FOR TESTING ONLY */
/* These types are used in testing to read the file in one scanline
 * at a time.
 */
#define	   TIFF_RLE_1LINE	45
#define	   TIFF_G3_1LINE	46
#define	   TIFF_G4_1LINE	47

#define    TIFF_DEFAULT_SS	50  /*       As above, but write the entire */
#define	   TIFF_UNCOMP_SS	51  /*       file out in a  Single  Strip. */
#define	   TIFF_LZW_SS		52
#define	   TIFF_G3_SS		53
#define	   TIFF_G4_SS		54
#define	   TIFF_PACKBITS_SS	55
#define	   TIFF_JPEG_SS		56
#define	   TIFF_THUNDERSCAN_SS	57
#define	   TIFF_NEXT_SS		58
#define	   TIFF_RLE_SS		59
#define    TIFF_LOSSYDITHER_SS	60

    /* For TIFF library compression schemes, the comments contain what kind
     * of support is available.  For example, "RW 1" means you may read
     * or write a file with this compression scheme, but are restricted to
     * images of depth 1.  "R" would mean you may only read a file with this
     * compression scheme, writing is not supported, and the image may be
     * any usual depth.
     */

/*******************************************************************
   FileHeader: contains some information about the image file
*******************************************************************/
struct FileHeader
{
    char      *filename;  /* full pathname */
    Int32      type;      /* e.g., SUNRAS */
    void      *header;    /* ptr. to internal data structure */
    Int32      width, height, Wpl, Bpl, totalBytes;
    Int32      filePages;	/* number of pages in the file */
    Int32      pageNumber;	/* number of page for which this
				   information was extracted */
};
typedef struct FileHeader FH;


 
/**************************************************************
     BMP_MAGIC
***************************************************************/
#define  BMP_MAGIC     0x4d42		/* ('M' << 8) | 'B' */
 
/**************************************************************
     PS_MAGIC
***************************************************************/
#define  PS_MAGIC        0x2521         /*  %!  */

/**************************************************************
     PCX_MAGIC
***************************************************************/
#define  PCX_MAGICv0        0x000A
#define  PCX_MAGICv2        0x020A
#define  PCX_MAGICv3        0x030A
#define  PCX_MAGICv5        0x050A

/**************************************************************
     TIFF_MAGIC
***************************************************************/
#define	TIFF_BIGENDIAN		0x4d4d
#define	TIFF_LITTLEENDIAN	0x4949

#ifdef _MSC_VER
#pragma pack(pop, imagerep_h)
#endif

#endif  /* IMAGEREP_H_INCLUDED */

/*

$Log: imagerep.h,v $
Revision 1.1.1.1  1999/09/20 18:19:11  jlin
Imported sources

Revision 1.21  1998/10/02 18:25:58  aber
imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v

CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
using libraries and includes.

Revision 1.20  1997/06/30 23:17:32  ddavies
imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v

Remove all references to AIS.

Revision 1.19  1997/03/17 19:58:11  ddavies
imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v

Add TIFF_LOSSYDITHER and TIFF_LOSSYDITHER_SS file types.

Revision 1.18  1996/05/16 17:13:26  ddavies
ddavies on Thu May 16 10:12:19 PDT 1996

imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v

Added new file types used for testing.  These are TIFF_RLE_1LINE, TIFF_G3_1LINE and
TIFF_G4_1LINE.
These cause the test routines to ask that files be read in one scanline at a time.

Revision 1.17  1996/05/11 01:15:24  ddavies
imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v

Add image types TIFF_G3_2D (Group 3 with 2D optional encoding),
TIFF_G3_PAD (Group 3 with pad inserted before EOL to make each EOL end
on a byte boundary) and TIFF_G3_2D_PAD (both).  These are primarily for
testing, but do allow one to write files in these formats.  The k factor
of a file depends on its resolution, so that should be set before
writing it.  Files with vertical resolution less than 150 lpi will hve
a K factor or 2.  Files with greater resolution will have a k factor of 4.

Revision 1.16  1995/12/13 04:08:23  ejaquith
imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v

CommonCheck: Added ";" to IP_RCSINFO().
Removed function comment blocks from before prototypes and
added special prototype comment.

 * Revision 1.15  1995/10/10  03:16:05  ejaquith
 * ejaquith on Mon Oct 9 20:15:46 PDT 1995
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Yanked LZ_COMP i/o type (because we never actually supported it anyway!)
 *
 * Revision 1.14  1995/10/03  20:52:39  ejaquith
 * ejaquith on Tue Oct 3 13:52:17 PDT 1995
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Deleted all obsolete FF_TIFF.. forms as all files are now converted
 *
 * Revision 1.13  1995/09/29  00:28:17  ejaquith
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Added TIFF_* variants for file i/o.
 * Added TIFF_*_SS variants for file i/o when writing only a single strip TIFF
 * image.
 *
 * Revision 1.12  1995/09/12  18:31:16  ejaquith
 * ejaquith on Tue Sep 12 11:30:05 PDT 1995
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Removed type DJFAX (and redefined types following where it used to be).
 * Removed djVersion, djHRes, djVRes, and djOrientation fields from
 * FileHeader internal struct. (since only DJFAX stuff used them).
 *
 * Revision 1.11  1995/08/31  22:03:23  ejaquith
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.10  1995/07/31  18:00:06  ejaquith
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Standardizing changes -- first pass.
 *
 * Revision 1.9  1995/03/21  04:36:03  ddavies
 * ddavies on Mon Mar 20 20:35:30 PST 1995
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Added FF_TIFF_RLE format, allowing us to write Modified Huffman format in TIFF.
 *
 * Revision 1.8  1994/11/02  01:45:55  ejaquith
 * ejaquith on Tue Nov 1 17:44:57 PST 1994
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Updated FF_TIFF* macro values.  Added some comments to help user
 * Updated FF_TIFF* macro values.  Added some comments to help user
 * sift through TIFF compression types.
 *
 * Revision 1.7  1994/10/21  01:22:18  ejaquith
 * ejaquith on Thu Oct 20 18:21:49 PDT 1994
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Changed PCX_MAGIC macros to byte-swapped correct values.
 *
 * Revision 1.6  1994/10/10  17:23:28  ejaquith
 * ejaquith on Mon Oct 10 10:21:13 PDT 1994
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Changed NONE macro to NO_TYPE for reading and writing as the new TIFF
 * library will need this and it is more descriptive anyway.  Added the
 * PCX macro format type, the FF_TIFF_JPEG macro format type, and changed
 * the PNM macro format type to allow the TIFF macros to be together.
 * Added PCX_MAGIC* macros (four of them) to allow detecting PCX format
 * files for reading.
 * Anticipate addition of new TIFF library and PCX reading/writing routines
 * 11/94.
 *
 * Revision 1.5  1994/09/08  00:33:50  lperry
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.4  1994/08/16  22:01:03  macleod
 * macleod on Tue Aug 16 15:00:09 PDT 1994
 *
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Added "PNM" file type, supporting the PBMPLUS "PGM" and "PBM" file
 * formats. Note that we only support the "P4" and "P5" formats, which
 * are packed bits/bytes rather than 7-bit ASCII.
 *
 * Revision 1.3  1993/12/03  01:10:21  ddavies
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:00:35  ddavies
 * imagerep.h -> /product/ipcore/ipwrap/include/RCS/imagerep.h,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

*/

