/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef LISTS_H_INCLUDED
#define LISTS_H_INCLUDED

#ifdef _MSC_VER
#pragma pack(push, lists_h, 1)
#endif

IP_RCSINFO(lists_h_RCSInfo, "$RCSfile: lists.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:12 $ */



#ifndef  UNDEF
#define  UNDEF      -1
#endif  /* UNDEF */

/***************************************************************************

    This code allows user-defined structures to be combined in doubly
    linked rings.  Each user-defined structure has two fields,
    next and prev, that point to the address of other user-defined
    structures of the same type.  The macros below are supposed to
    allow manipulation of these lists without ever accessing these
    internal fields.

    To use this generic list facility:

        1. Insert

		this_struct_type      *next;
		this_struct_type      *prev;

	   anywhere within the user-defined structure.

	2. Add to list.

	   When adding to an empty list, the next and prev must
	   point to the structure itself.  This is the condition
	   where there is only one element in the list.

        3. Use the the following macros to alter or maintain the list(s).
            INSERT_BEFORE
            INSERT_AFTER
            ADD_TO_HEAD
            ADD_TO_TAIL
	    GET_NEXT
	    GET_PREV
	    GET_TAIL
	    GET_NEXT_LINEAR
	    GET_PREV_LINEAR
	    BEGIN_FOREACH_FORWARD
	    BEGIN_FOREACH_REVERSE
	    END_FOREACH
            REMOVE_FROM_LIST
            REMOVE_FROM_HEAD
            REMOVE_FROM_TAIL
            ITERATE_REVERSE
            ITERATE_FORWARD
	    REVERSE_LIST
	    CONCAT_LISTS

	4. INIT_LINKS is a very simple macro is used internally to
	   initialize the list.  This needs only be done at the time
	   when the first element is placed on the list, and the
	   macros above treat this special case invisibly to the user.
	   Users can also do the initialization themselves on any
	   data structure, simply by writing
	       a->next = a->prev = a;

	5. Use the following macros to get list count and list position.
	    GET_COUNT
	    GET_POSITION_IN_LIST
	    GET_AT_POSITION

	6. Example use:

	    typedef struct Foo
	    {
		struct Foo   *next, *prev;
		struct Bar   *firstbar;   // of Bars
	    } FOO;

	    typedef struct Bar
	    {
		struct Bar   *next, *prev;
		int           label;
	    } BAR;

	    struct Foo *
	    createFoo()
	    {
		FOO   *a;
		a = (FOO *)calloc(1, sizeof(struct Foo));
		a->next = a->prev = a;    // just to be safe
		return a;
	    }
		
	    struct Bar *
	    createBar()
	    {
		BAR   *a;
		a = (BAR *)calloc(1, sizeof(struct Bar));
		a->next = a->prev = a;    // just to be safe
		return a;
	    }
		

	    struct Bar *
	    destroyBar(b)
	    BAR    *b;
	    {
		if (!b)
		    return NULL;

		free(b);
		return NULL;
	    }

	    struct Foo *
	    destroyFoo(a)
	    FOO      *a;
	    {
	    BAR      *b;

		BEGIN_FOREACH_REVERSE(a->firstbar, b, BAR)
		{
		    REMOVE_FROM_LIST(a->firstbar, b)
		    destroyBar(b);
		}
		END_FOREACH();
		a->firstbar = NULL;
		free(a);
		return NULL;
	    }
		

	    main()
	    {
	    FOO    *firstfoo = NULL;
	    FOO    *a, *currentfoo;
	    BAR    *b, *currentbar, *firstbar;
	    int     i, j, k;

		    // make a linked list of Foos, and attach two Bars to each
		for (i = 0, j = 1; i < 10; i++)
		{
		    a = createFoo();
		    ADD_TO_TAIL(firstfoo, a)
		    for (k = 0; k < 4; k++)
		    {
			b = createBar();
			b->label = j++;
			ADD_TO_TAIL(a->firstbar, b)
		    }
		}

		    // go down the list of Foos and print out what there is
		BEGIN_FOREACH_FORWARD(firstfoo, currentfoo, FOO)
		{
		    firstbar = currentfoo->firstbar;
		    printf("\n  Foo address %d\n", currentfoo);
		    BEGIN_FOREACH_FORWARD(firstbar, currentbar, BAR)
		    {
			printf("       bar(%d): %d\n",
				    currentbar, currentbar->label);
		    }
		    END_FOREACH()
		}
		END_FOREACH()


		    // go down the list of Foos and destroy all the bars
		BEGIN_FOREACH_FORWARD(firstfoo, currentfoo, FOO)
		{
		    REMOVE_FROM_LIST(firstfoo, currentfoo)
		    destroyFoo(currentfoo);
		}
		END_FOREACH()

		    // go down the list of Foos and print out what's there
		BEGIN_FOREACH_FORWARD(firstfoo, currentfoo, FOO)
		{
		    firstbar = currentfoo->firstbar;
		    printf("\n  Foo address %d\n", currentfoo);
		    BEGIN_FOREACH_FORWARD(firstbar, currentbar, BAR)
		    {
			printf("       bar(%d): %d\n",
				currentbar, currentbar->label);
		    }
		    END_FOREACH()
		}
		END_FOREACH()


	    }


****************************************************************************/

    /*
     *  Initially, we have a pointer to a NULL list.  After the first
     *  item is made, as it is being placed on the list, using one
     *  of several macros such as ADD_TO_HEAD, ADD_TO_TAIL, INSERT_BEFORE,
     *  or INSERT_AFTER, the internal links are initialized to point to itself.
     */
#define INIT_LINKS(pFirst) \
    { \
	if (pFirst) { \
            (pFirst)->next = (pFirst)->prev = (pFirst); \
	} \
    }


    /*
     *  INSERT_BEFORE inserts an item before any member of the list.
     *  If the ptr to the first item in the list is NULL, it is an empty list
     *  and we must initialize the internal list pointers to point to
     *  the member itself.
     */
#define INSERT_BEFORE(pInlist, pNew) \
    { \
	if (pNew) { \
	    if ((pInlist) == NULL) { \
		(pInlist) = (pNew); \
		INIT_LINKS(pNew) \
	    } \
	    else { \
		(pNew)->next = (pInlist); \
		(pNew)->prev = (pInlist)->prev; \
		(pInlist)->prev->next = (pNew); \
		(pInlist)->prev = (pNew); \
	    } \
	} \
    }


    /*
     *  INSERT_AFTER inserts an item after any member of the list.
     *  If the ptr to the first item in the list is NULL, it is an empty list
     *  and we must initialize the internal list pointers to point to
     *  the member itself.
     */
#define INSERT_AFTER(pInlist, pNew) \
    { \
	if (pNew) { \
	    if ((pInlist) == NULL) { \
		(pInlist) = (pNew); \
		INIT_LINKS(pNew) \
	    } \
	    else { \
		(pNew)->next = (pInlist)->next; \
		(pNew)->prev = (pInlist); \
		(pInlist)->next->prev = (pNew); \
		(pInlist)->next = (pNew); \
	    } \
	} \
    }


    /*
     *  ADD_TO_HEAD is equivalent to INSERT_BEFORE the head of the list,
     *  followed by a reasignment of the first element.
     */
#define  ADD_TO_HEAD(pFirst, pNew) \
    { \
	if (pNew) { \
	    if (pFirst == NULL) { \
		(pFirst) = (pNew); \
		INIT_LINKS(pNew) \
	    } \
	    else { \
		(pNew)->next = (pFirst); \
		(pNew)->prev = (pFirst)->prev; \
		(pFirst)->prev->next = (pNew); \
		(pFirst)->prev = (pNew); \
		(pFirst) = (pNew); \
	    } \
	} \
    }


    /*
     *  ADD_TO_TAIL is equivalent to INSERT_BEFORE the head of the list.
     */
#define  ADD_TO_TAIL(pFirst, pNew)    INSERT_BEFORE(pFirst, pNew)

    
#define  GET_NEXT(pInlist) \
    (((pInlist) == NULL) ? NULL : (pInlist)->next)
    
#define  GET_PREV(pInlist) \
    (((pInlist) == NULL) ? NULL : (pInlist)->prev)
    
#define  GET_NEXT_LINEAR(pFirst, pInlist) \
    ((!(pFirst) || !(pInlist) || (pInlist)->next == (pFirst)) \
	? NULL : (pInlist)->next)
    
#define  GET_PREV_LINEAR(pFirst, pInlist) \
    ((!(pFirst) || !(pInlist) || (pInlist) == (pFirst)) \
	? NULL : (pInlist)->prev)
    

    /*
     *  GET_TAIL is equivalent to GET_PREV applied to the head of the list.
     */
#define  GET_TAIL(pFirst) \
    (((pFirst) == NULL) ? NULL : (pFirst)->prev)
    


    /*
     *  BEGIN_FOREACH_FORWARD takes a "first" element, which
     *  is not necessarily at the head of the list, a dummy pointer,
     *  previously declared, to any list element, and the type
     *  of the element (not including the *).  It computes a pointer
     *  to each object in the list, in succession in the forward direction,
     *  that can be used within the body of the loop.   It is a bit
     *  tricky, because we must store the ptr to the structure before
     *  the "first" element; this will be the "last" to be processed.
     *  Now, if within the body of the for statement we are deleting
     *  these elements, then the "first" will be destroyed and updated at
     *  each iteration.  How do we stop the iteration?  By also saving
     *  the "prev" pointer at each iteration, and testing to find the
     *  second time that the "prev" is the same as the "last", which was
     *  stored at the beginning.  Note also that pFirst is tested for
     *  non-NULLness at each iteration; this is done because, yes,
     *  it can be changed by an operation within the loop body.
     *  This implementation is analagous to the BEGIN_FOREACH_REVERSE macro.
     * 
     *  The following version is NG because if the first element is
     *  destroyed more than once, it quits:

	#define BEGIN_FOREACH_FORWARD(pFirst, pCurrent, type) \
	    if (pFirst) { \
	      type     *pNext; \
	      Int32     firstpass = 1; \
		for ((pCurrent) = (pFirst), pNext = (pCurrent)->next; \
		    (pFirst) && \
		    ((pCurrent) != (pFirst)) || \
		    (firstpass ? ((firstpass = 0), 1) : 0); \
		(pCurrent) = pNext, pNext = pNext->next) {`

     *
     *  The following version is NG because the update part uses the
     *  current pointer, which may be set to NULL within the body!

	#define BEGIN_FOREACH_FORWARD(pFirst, pCurrent, type) \
	    if (pFirst) { \
	      type     *pLast; \
	      type     *pPrev; \
	      type     *pNext; \
	      Int32     firstpass = 1; \
		for ((pCurrent) = (pFirst), pNext = (pCurrent)->next, \
		     pLast = (pFirst)->prev, pPrev = pLast; \
		    (pFirst) && \
		    ((pPrev) != (pLast)) || \
		    (firstpass ? ((firstpass = 0), 1) : 0); \
		    pPrev = (pCurrent), (pCurrent) = pNext, pNext = pNext->next){
     */
#define BEGIN_FOREACH_FORWARD(pFirst, pCurrent, type) \
    if (pFirst) { \
        type     *_i_pNext; \
        type     *_i_pLast; \
        Int32     _i_notdone = 1; \
	for ((pCurrent) = (pFirst), _i_pNext = (pCurrent)->next, \
	      _i_pLast = (pFirst)->prev; \
	    (pFirst) && \
	    (_i_notdone) && \
	    ((pCurrent) != _i_pLast) || (_i_notdone ? ((_i_notdone = 0), 1) : 0); \
	    (pCurrent) = _i_pNext, _i_pNext = _i_pNext->next) {



    /*
     *  BEGIN_FOREACH_REVERSE takes a "first" element, which
     *  is not necessarily at the head of the list, a dummy pointer,
     *  previously declared, to any list element, and the type
     *  of the element (not including the *).  It computes a pointer
     *  to each object in the list, in succession in the reverse direction,
     *  that can be used within the body of the loop.   This is a bit
     *  tricky because we need to stop AFTER we've processed the "first"
     *  element.  We maintain the variable pLast for the test, and quit
     *  the second time it is equal to the "first".
     *
     *  The following version is NG because the update part uses the
     *  current pointer, which may be set to NULL within the body!
     *
	#define BEGIN_FOREACH_REVERSE(pFirst, pCurrent, type) \
	    if (pFirst) { \
	      type    *pPrev; \
	      type    *pLast; \
	      Int32    firstpass = 1; \
		for (pLast = (pFirst), (pCurrent) = (pFirst)->prev, \
			    pPrev = (pCurrent)->prev; \
		    (pFirst) && \
		    (pLast != (pFirst)) || \
		    (firstpass ? ((firstpass = 0), 1) : 0); \
		    pLast = (pCurrent), (pCurrent) = pPrev, \
		    pPrev = pPrev->prev) {
     *
     */
#define BEGIN_FOREACH_REVERSE(pFirst, pCurrent, type) \
    if (pFirst) { \
        type    *_i_pPrev; \
        Int32    _i_notdone = 1; \
	for ((pCurrent) = (pFirst)->prev, _i_pPrev = (pCurrent)->prev; \
	    (pFirst) && \
	    (_i_notdone) && \
	    ((pCurrent) != (pFirst)) || (_i_notdone ? ((_i_notdone = 0), 1) : 0); \
	    (pCurrent) = _i_pPrev, _i_pPrev = _i_pPrev->prev) {


     /*
      *  END_FOREACH ends either type of BEGIN_FOREACH_* loop
      */
#define END_FOREACH() }  }


    /*
     *  REMOVE_FROM_LIST unlinks the member from the ring.
     *  It does not free any memory.
     *  If the link points to itself, this must be the only element in the
     *  list, and it can be removed by simply setting the list ptr to NULL.
     *  Otherwise, the element is removed, and if it was the head, a
     *  new element is assigned to be the head.
     */
#define REMOVE_FROM_LIST(pFirst, pInlist) \
    { \
	if ((pFirst) && (pInlist)) { \
	    if ((pInlist)->prev == (pInlist)) \
		(pFirst) = NULL; \
	    else { \
		(pInlist)->prev->next = (pInlist)->next; \
		(pInlist)->next->prev = (pInlist)->prev; \
		if ((pInlist) == (pFirst)) \
		    (pFirst) = (pInlist)->next; \
	    } \
	} \
    }


    /*
     *  REMOVE_FROM_HEAD unlinks the head from the ring.
     *  It does not free any memory.
     *  The second argument is a dummy ptr to a list element.  It saves
     *  a ptr to the removed item.
     *  If the link points to itself, this must be the only element in the
     *  list, and it can be removed by simply setting the list ptr to NULL.
     *  Otherwise, the element is removed and the head is reassigned
     *  to the next element.
     */
#define REMOVE_FROM_HEAD(pFirst, pCatch) \
    { \
	(pCatch) = (pFirst); \
	if (pFirst) { \
	    if ((pFirst)->prev == (pFirst)) \
		(pFirst) = NULL; \
	    else { \
		(pFirst)->prev->next = (pFirst)->next; \
		(pFirst)->next->prev = (pFirst)->prev; \
		(pFirst) = (pFirst)->next; \
	    } \
	} \
    }


    /*
     *  REMOVE_FROM_TAIL unlinks the tail from the ring.
     *  It does not free any memory.
     *  The second argument is a dummy ptr to a list element.  It saves
     *  a ptr to the removed item.
     *  If the link points to itself, this must be the only element in the
     *  list, and it can be removed by simply setting the list ptr to NULL.
     *  Otherwise, the element is removed, and the head is not changed.
     */
#define REMOVE_FROM_TAIL(pFirst, pCatch) \
    { \
	(pCatch) = NULL; \
	if (pFirst) { \
	    (pCatch) = (pFirst)->prev; \
	    if ((pFirst)->prev == (pFirst)) \
		(pFirst) = NULL; \
	    else { \
		(pFirst)->prev->prev->next = (pFirst); \
		(pFirst)->prev = (pFirst)->prev->prev; \
	    } \
	} \
    }


    /*
     *  ITERATE_REVERSE works backward from the first element given,
     *  calling pFunc once for each member of the list, and passing
     *  the function a ptr to the member.  The third parameter is the
     *  datatype of each member of the list (without the *).
     *  By going in the reverse direction, we are assured that
     *  pCurrent and pFirst are different until the last element is chosen.
     *  Thus, if the function destroys the member, the head of the list
     *  will not be altered.
     */
#define ITERATE_REVERSE(pFirst, pFunc, type) \
    { \
    type    *_i_pCurrent; \
	BEGIN_FOREACH_REVERSE(pFirst, _i_pCurrent, type) \
	    (*(pFunc))(_i_pCurrent); \
	END_FOREACH() \
    }

    /*
     *  ITERATE_FORWARD works forward from the first element given,
     *  calling pFunc once for each member of the list, and passing
     *  the function a ptr to the member.  The third parameter is the
     *  datatype of each member of the list (without the *).
     *  The ITERATE_REVERSE macro should be used for cases where the
     *  function destroys the selected member.
     */
#define ITERATE_FORWARD(pFirst, pFunc, type) \
    { \
    type    *_i_pCurrent; \
	BEGIN_FOREACH_FORWARD(pFirst, _i_pCurrent, type) \
	    (*(pFunc))(_i_pCurrent); \
	END_FOREACH() \
    }


    /*
     *  REVERSE_LIST reverses the order of a list, in place
     *  Note that we end by re-assigning the head of the list
     *  to the element that was initially the last, and after
     *  the re-arrangement, has become the second.
     */
#define  REVERSE_LIST(pFirst, type) \
    { \
    type   *_i_pCurrent; \
    type   *_i_pTemp; \
	BEGIN_FOREACH_FORWARD(pFirst, _i_pCurrent, type) \
	    _i_pTemp = _i_pCurrent->next; \
	    _i_pCurrent->next = _i_pCurrent->prev; \
	    _i_pCurrent->prev = _i_pTemp; \
	END_FOREACH() \
	if (pFirst) \
	    (pFirst) = (pFirst)->next; \
    }


    /*
     *  CONCAT_LISTS concatenates two circular lists together, adding the
     *  second to the tail of the first.  The head of the
     *  first remains the head of the concatenated lists, and the tail
     *  of the second is joined to the head of the first.
     */
#define  CONCAT_LISTS(pFirst1, pFirst2) \
    { \
    void   *_p_pTemp; \
	if ((pFirst1) && (pFirst2)) { \
	    (pFirst2)->prev->next = (pFirst1); \
	    (pFirst1)->prev->next = (pFirst2); \
	    _p_pTemp = (pFirst2)->prev; \
	    (pFirst2)->prev = (pFirst1)->prev; \
	    (pFirst1)->prev = _p_pTemp; \
	} \
    }



/*------------------------------------------------------------------------*
 *                        Count and Position                              *
 *------------------------------------------------------------------------*/
    /*
     *  GET_COUNT gives the number of elements in the list
     *      "pFirst" is a pointer to the head of the list;
     *      each pointer is to a user-defined structure of type "type";
     *      "count" is a pre-defined integer into which the result placed.
     *
     *  I wanted to have the macro return a value, rather than side-effecting
     *  count, but a block definition is required, and it appears that
     *  an rval with braces, other than as a list, is illegal syntax.
     */
#define  GET_COUNT(pFirst, type, count) \
    { \
    type   *_i_pCurrent; \
        (count) = 0; \
	BEGIN_FOREACH_FORWARD(pFirst, _i_pCurrent, type) \
	    (count)++; \
	END_FOREACH() \
    }


    /*
     *  GET_POSITION_IN_LIST finds the integer position within the
     *  list of the given element,
     *  starting with 0 for the position of the head of the list.
     *  The result is assigned to the variable "position".
     *  If the given element is not in the list (an error), the value
     *  of "position" is set to UNDEF.
     */
#define GET_POSITION_IN_LIST(pFirst, pElement, type, position) \
    { \
    Int32    _i_count; \
    type    *_i_pCurrent; \
	(position) = UNDEF; \
	_i_count = 0; \
	if ((pFirst) && (pElement)) { \
	    BEGIN_FOREACH_FORWARD(pFirst, _i_pCurrent, type) \
		if ((pElement) == _i_pCurrent) { \
		    (position) = _i_count; \
		    break; \
		} \
		else \
		    _i_count++; \
	    END_FOREACH() \
	} \
    }


    /*
     *  GET_AT_POSITION finds the element that occurs at the given
     *  position within the list,
     *  starting with 0 for the position of the head of the list.
     *  The result is assigned to the pointer variable "pElement".
     *  If there is no element at the given position (an error), the
     *  value of "pElement" is set to NULL.
     */
#define GET_AT_POSITION(pFirst, pElement, type, position) \
    { \
    Int32    _i_count; \
    type    *_i_pCurrent; \
	(pElement) = NULL; \
	GET_COUNT(pFirst, type, _i_count) \
	if ((pFirst) && ((position) < _i_count)) { \
	    _i_count = 0; \
	    BEGIN_FOREACH_FORWARD(pFirst, _i_pCurrent, type) \
		if ((position) == _i_count) { \
		    (pElement) = _i_pCurrent; \
		    break; \
		} \
		else \
		    _i_count++; \
	    END_FOREACH() \
	} \
    } 


#ifdef _MSC_VER
#pragma pack(pop, lists_h)
#endif

#endif   /* LISTS_H_INCLUDED */


/*

$Log: lists.h,v $
Revision 1.1.1.1  1999/09/20 18:19:12  jlin
Imported sources

Revision 1.9  1998/10/02 18:26:32  aber
lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v

CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
using libraries and includes.

Revision 1.8  1995/12/13 04:10:18  ejaquith
lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v

CommonCheck: Added ";" to IP_RCSINFO().
Removed function comment blocks from before prototypes and
added special prototype comment.

 * Revision 1.7  1995/08/31  22:03:36  ejaquith
 * lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/01  20:43:00  ejaquith
 * lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v
 *
 * Handy script munched part of header -- restored now.
 *
 * Revision 1.5  1995/07/31  17:33:30  macleod
 * macleod on Mon Jul 31 10:33:30 PDT 1995
 *
 * lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v
 *
 * Changed prototypes to conform to new standards.
 *
 * Revision 1.4  1994/09/08  00:34:51  lperry
 * lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.3  1993/12/03  01:11:12  ddavies
 * lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:01:41  ddavies
 * lists.h -> /product/ipcore/ipwrap/include/RCS/lists.h,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

*/

