/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


#ifndef _MATCH_H_INCLUDED_
#define _MATCH_H_INCLUDED_

#ifdef _MSC_VER
#pragma pack(push, match_h, 1)
#endif

IP_RCSINFO(match_h_RCSInfo, "$RCSfile: match.h,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:12 $");



#define DEBUG_SPARSE     0  /* 1 to write out a PIXR with rects around all
			     * the active object rims.
			     */



    /* Amount to allow increasing the search subrect in the Hausdorff
     * searching routines, in inches */
#define MAX_TRANS_INCR_INCHES      0.1F
      
    /* Make the mimimum required Hausdorff percent for a match be based upon
     * the size of the object's dimensions (min and max) in inches.
     * It spits out a number in range [0.6, 0.9].
     * If the min dimension is less than 0.05, make it 0.9 (too flat so make
     * the matching strict)
     * O.W. If max dimen is: you get:
     * 0.2" --> 0.9
     * 0.6" --> 0.75
     * 1"   --> 0.6 */
#define MIN_HAUS_PERCENT(mind, maxd) \
    (((mind) <= 0.05 || (maxd) <= 0.2) ? 0.9 : ((maxd) >= 1. ? 0.6 : 0.975 - 0.375 * (maxd)))
 
    /* This is the lowest Hausdorff percent that we'll find acceptable as
     * a match when we have expanded our search region from our initial guess
     * as far out as we will allow.  Thus, we want to accept anything in this
     * last/biggest search region that could be close to a match */
#define ANY_MATCH   0.5
    /* Normally, I use 50%.  But in some cases, the model is large and only
     * off by a bit in terms of the render, so it is "nicer" to allow for
     * down to 30% in these cases.
     */


    /* unused now */
#if 0
    /* This is used for checking if
     * (w) > SPARSE_CHECK * (xthick),
     * (h) > SPARSE_CHECK * (ythick)
     * where (w) and (h) are the width and height of the model we will render
     * and (xthick) and (ythick) are the thickness dimensions of the model
     * we will render.
     * We would want to check this to approximate whether the rendered symbol
     * will be "sparse".  We approximate each symbol's "denseness" by adding
     * the thickness of two bars, (xthick) or (ythick) thick, running across
     * the model bitmap.
     * A model would be sparse when:
     *   (((xthick) or (ythick)) * 2 / ((w) or (h))) < 10%         -or-
     *   20 * ((xthick) or (ythick)) < ((w) or (h))                        */
#define SPARSE_CHECK        20
#endif


    /* In sparseRimVerify() make the rims we check be this number times the
     * respective (x/y)thickness in the symbol */
#define RIM_THICKNESS_MULTIPLE    3
#define SKIP_THICKNESS_MULTIPLE   1
    /* In sparseRimVerify() all four rims must be at most this percent dense */
#define SPARSE_RIM_MAX_PERCENT    0.2F

    /* If a model bitmap has very large dimensions, then most likely the
     * regular Hausdorff matching took care of positive false matches, so
     * this is verifiable based simply on its dimensions. */
#define VERIFIABLY_BIG(w,h)   (((w) > 70) && ((h) > (70)))
    


#ifdef _MSC_VER
#pragma pack(pop, match_h)
#endif

#endif /* _MATCH_H_INCLUDED_ */

/*
  $Log: match.h,v $
  Revision 1.1.1.1  1999/09/20 18:19:12  jlin
  Imported sources

  Revision 1.3  1998/10/02 18:26:34  aber
  match.h -> /product/ipcore/ipwrap/include/RCS/match.h,v

  CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
  using libraries and includes.

  Revision 1.2  1996/02/26 20:06:01  ddavies
  ddavies on Mon Feb 26 12:05:31 PST 1996

  match.h -> /product/ipcore/ipwrap/include/RCS/match.h,v

  Add the magic F to floating const declarations to make Visual C/C++ happy.

  Revision 1.1  1996/02/25 01:22:20  ejaquith
  Initial revision


  */
