/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


/* 
 *  morph.h:  defined constants, STREL, and STAR
 */

#ifndef MORPH_H_INCLUDED
#define MORPH_H_INCLUDED

#ifdef _MSC_VER
#pragma pack(push, morph_h, 1)
#endif

IP_RCSINFO(morph_h_RCSInfo, "$RCSfile: morph.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:12 $ */

#define  DONT_CARE        0
#define  HIT              1
#define  MISS             2

#define  OFF_PIXELS       0
#define  ON_PIXELS        1

#define  PIXELS           0
#define  WORDS            1

#define  HORIZ            1
#define  VERT             2
#define  BOTH             3

#define  LEFT             0
#define  RIGHT            1
#define  TOP              2
#define  BOTTOM           3

#define  UP               0
#define  DOWN             1

#define  FIRST            1
#define  LAST             2

#define  FIXED_ITERATIONS 1
#define  TO_COMPLETION    2

#define  TO_ABLATION      1
#define  TO_ONE_PIXEL     2
#define  FOUR_CONNECT     3
#define  EIGHT_CONNECT    4

/*
 *  These are used for the fast 1D seed fill operations.  They're also
 *  defined in ipshared/include/shrip.pub.  They're included here so
 *  that applications that use fillClipSweep2*() routines need
 *  not include shrip.pub.
 */
#define  FILL_RIGHT       1
#define  FILL_LEFT        2
#define  FILL_RIGHT_LEFT  3
#define  FILL_DOWN        1
#define  FILL_UP          2
#define  FILL_UP_DOWN     3


/* 
 *  The following def'ns flag the components of STRELs to be used
 *  with hit/miss convolutions and rank order operations
 */
#define  HITS_ONLY        1
#define  MISSES_ONLY      2
#define  HITS_AND_MISSES  3


    /* strel types */
#define  H_1D_STREL       1     /* hor, 1D, hits only */
#define  V_1D_STREL       2     /* vert, 1D, hits only */
#define  HMD_1D_STREL     3     /* 1D, hits & miss and/or dc */
#define  S_2D_STREL       4     /* separable, 2D, hits only */
#define  HMD_S_2D_STREL   5     /* separable, 2D, hits & miss/dc */
#define  NS_2D_STREL      6     /* non-separable 2D */


    /*  star types */
#define  NB_STAR          1     /* narrow bandwidth star */
#define  WB_STAR          2     /* wide bandwidth star */


/*-------------------------------------------------------------*
		Structuring Element Definitions
 *-------------------------------------------------------------*/
struct structElem
{
    Int32  	     nx;		/* width of SE */
    Int32  	     ny;		/* height of SE */
    Int32  	     cx;		/* x location of SE center */
    Int32  	     cy;		/* y location of SE center */
    StrelID	     ID;		/* ID number for "named" strels */
    Int16	   **data;		/* DONT_CARE, HIT, MISS values,
					 * in row-major matrix order
					 * (all values for row 0 from left
					 *  to right, all values for row
					 *  1, etc). */
    ErodeProc	     erodeOp;		/* assembly code for erosion */
    LongErodeProc    longErodeOp;	/* assembly code for erosion with
					 * long (>= 64 bit) operations */
    DilateProc	     dilateOp;		/* assembly code for dilation */
    LongDilateProc   longDilateOp;	/* assembly code for dilation with
					 * long (>= 64 bit) operations */
    Int32	     accelFlag;		/* set to 1 if can use machine code
					 * generated at runtime. */
    Int32	     blockWidth;	/* used by runtime machine code */
    Int32	     longBlockWidth;	/* used by long op runtime code */
};
typedef struct structElem STREL;


struct StrelArray
{
    Int32               n;         /* number of strels in the array */
    Int32               nalloc;    /* number of pointers alloc in each array*/
    char              **name;      /* array of strel names */
    struct structElem **strels;    /* array of pointers to strels  */
};
typedef struct StrelArray STAR;


/****************************************************
*  Data structures for managing the lists of strels
****************************************************/

/*
  A strel reference data type includes the
  strel pointer and a count of the number
  of references to it, for use in deallocation.
*/
typedef struct {
    STREL *strel;
    Int32 refCount;
} StrelRef;

/*
* Define the structure for the StrelInfo state 
* information. For now, it only contains
* the hash table and a strel count.
*/

struct strelInfoStruct {
    struct strelInfoStruct *globalStrelInfo;  /* Link to the global one */
    Int32 userCount;
    Bool strel1InitsDone;    
    Bool strel11InitsDone;  
    Bool strel12InitsDone; 
    StrelRef *strelArray;
};
typedef struct strelInfoStruct StrelInfo;

#ifdef _MSC_VER
#pragma pack(pop, morph_h)
#endif

#endif  /* MORPH_H_INCLUDED */

/*

$Log: morph.h,v $
Revision 1.1.1.1  1999/09/20 18:19:12  jlin
Imported sources

Revision 1.14  1998/10/02 18:26:39  aber
morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v

CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
using libraries and includes.

Revision 1.13  1996/12/18 23:14:16  ddavies
ddavies on Wed Dec 18 15:12:32 PST 1996

morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v

Add longErodeOp, longDilateOp and longBlockWidthInBytes parameters to the
STREL structure.  These are used for machine code generated at runtime that
uses long (>= 64 bit) operations.  Also improve comments.

Revision 1.12  1995/12/13 04:10:33  ejaquith
morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v

CommonCheck: Added ";" to IP_RCSINFO().
Removed function comment blocks from before prototypes and
added special prototype comment.

 * Revision 1.11  1995/10/03  00:37:05  ddavies
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Replace the ambiguous declarations of for pointers to machine code inside
 * the STREL with the new ErodeProc and DilateProc procedures.  These contain
 * the complete list of args so 1) they are checked properly, and 2) the
 * C++ compilers don't whine.
 *
 * Revision 1.10  1995/08/31  22:03:41  ejaquith
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/07/29  04:26:30  ddavies
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Update to new style.
 *
 * Revision 1.8  1995/01/26  22:39:28  lperry
 * lperry on Thu Jan 26 14:37:51 PST 1995
 *
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Remove the STREL structure members crc and refCount,
 * which are now obsolete. Change the type of the
 * ID field to StrelID (an enumerated type) instead of
 * UInt32.
 *
 * Revision 1.7  1995/01/24  22:36:47  lperry
 * lperry on Tue Jan 24 14:35:45 PST 1995
 *
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Get rid of all the linked list strel management structs.
 * Define a new StrelInfo struct which tracks strels on
 * an array with reference counting.
 *
 * Revision 1.6  1994/09/08  00:35:09  lperry
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.5  1994/08/18  00:31:29  lperry
 * lperry on Wed Aug 17 17:28:28 PDT 1994
 *
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Add a reference count field to the strel node structure,
 * so the strel manager can keep track of how many references
 * each user has to a strel, instead of just the raw total.
 *
 * Revision 1.4  1994/08/10  22:27:41  lperry
 * lperry on Wed Aug 10 15:26:19 PDT 1994
 *
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Add crc, ID and refCount fields to strel definition
 * for use with strel manager function.
 * Also add data structure definitions for strel manager.
 *
 * Revision 1.3  1993/12/03  01:11:20  ddavies
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:01:55  ddavies
 * morph.h -> /product/ipcore/ipwrap/include/RCS/morph.h,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

*/

