/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


/* 
 *  morphg.h
 */

#ifndef MORPHG_H_INCLUDED
#define MORPHG_H_INCLUDED

#ifdef _MSC_VER
#pragma pack(push, morphg_h, 1)
#endif

IP_RCSINFO(morphg_h_RCSInfo, "$RCSfile: morphg.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:12 $ */

    /* a LOOP is a closed path; a CHAIN is an open path */
#define CHAIN_TYPE   0
#define LOOP_TYPE    1


/* This should probably go in some other include file */
/* Note: type DIRECTION *must* be signed */
typedef char   DIRECTION;


#define  SMLVAL      0
#define  BIGVAL8     0xFF
#define  BIGVAL16    0x7FFF    /* 32.7K */


/*
 *                                               3 2 1
 * Encoding of directions for square grids:      4 * 0
 *                                               5 6 7
 *
 *                                                2 1
 * Encoding of directions for hexagonal grids:   3 * 0
 *                                                4 5
 */
	

    /* point data structure definition. */
struct point {
    Int16 x;
    Int16 y;
};

    /* the path type */
typedef struct Path {
    struct point origine; /* origin pixel */
    DIRECTION *liste;     /* list of directions */
    Int32 longueur;        /* number of DIRECTIONS */
    Int32 type;           /* LOOP_TYPE or CHAIN_TYPE */
    Int32 label;          /* usually label of initial object */
    Int32 parity;          /* for hexagonal paths. this is the
			     parity of line on which origin is.
			     This field has absolutely NO meaning
			     for non 6-connected chains */
} PATH;


    /* defines useful in contour tracking stuff */
#define INSIDE_VAL  -31999
#define OUTSIDE_VAL -31998


    /* structure defined on top of previous one so that
       one can keep track of the array of directions even
       when there are no paths. Not used currently */
/*
struct list_of_loops {
    DIRECTION    *direction_array;
    struct Path  *path_array;
}
*/


    /* global arrays defined in gipdef.c */
#if 0
extern Int32  deltax[],  deltay[];
extern Int32 delta6x[], delta6y[];
#endif

    /*-------------------------------------------------------------------*
     *   Structuring element structures for grayscale morphology only    *
     *-------------------------------------------------------------------*/

struct side_se {          /* a side of a structuring element */
    Int32          size;  /* number of pixels of this side   */
    struct  point *pts;   /* pointer to the pixels           */
    };

typedef struct se {       /* structuring element structure    */
    Int32 center_location;  /* connected component in which center
			     of element is located. 0 if center
			     does not belong to element       */
    Int32 parity;           /* parity of lines on which element
			     can be mapped (hexagonal case)   */
    Int32 nb_cc;            /* number of connected components   */
    Int16 minx;            /* smallest x coord of pixels of SE */
    Int16 miny;            /* smallest y coord of pixels of SE */
    Int16 maxx;            /* largest  x coord of pixels of SE */
    Int16 maxy;            /* largest  y coord of pixels of SE */
    struct point *origins;/* array of one pixel per component */
    Int32 size;             /* total number of pixels           */
    struct point *pts;    /* array of all pixels              */
    struct side_se *sides;/* pointer to the (8) sides of the
			     structuring element              */
    } SE;



struct Strelptr
{
    Int32         *rowLocs;      /* array of offsets giving row position
				  * relative to SE center of rows in the SE
				  * that contain nonzero SE elements */
    Int32          rows;         /* number of such row offsets */
    Int32        **elemLocs;     /* array of offsets within each row for the
				  * nonzero SE elements, relative to the 
				  * SE center  */
    Int32         *elems;        /* number of offsets within each row */
};
typedef struct Strelptr STRELPTR;


#ifdef _MSC_VER
#pragma pack(pop, morphg_h)
#endif

#endif  /* MORPHG_H_INCLUDED */
