/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


/*
 *  pixr.h
 */

#ifndef  PIXR_H_INCLUDED
#define  PIXR_H_INCLUDED

#ifdef _MSC_VER
#pragma pack(push, pixr_h, 1)
#endif


IP_RCSINFO(pixr_h_RCSInfo, "$RCSfile: pixr.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:12 $ */


 /*------------------------------------------------------------------------*
  *                                                                        *
  *      The image has Frame, Pad, and Image regions in the                *
  *      following configuration (symbolically):                           *
  *                                                                        *
  *                                                                        *
  *           FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFP                *
  *           FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIFFP                *
  *           FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFP                *
  *           FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFP                *
  *                                                                        *
  *                                                                        *
  *    See the notes at the end of this file for constraints on the        *
  *    sizes of these regions.                                             *
  *                                                                        *
  *------------------------------------------------------------------------*/



 /*------------------------------------------------------------------------*
  *                           Structures                                   *
  *------------------------------------------------------------------------*/

/*
 * A "color record" is four packed bytes. It's used in color maps.
 */
#define	    cMaxColorPlanes	4

typedef UInt8 ColorRecord[cMaxColorPlanes];

/* A Chrominance is an (x,y) pair. It's used to specify white points, etc. */
typedef struct
{
    /* These are usually CIE (x,y). */
    Float32 x, y;
} Chrominance;

/* 
 * ColorType defines the supported color spaces.
 * Avoid enumerated types since no two compilers can agree on the
 * number of bytes in a variable of this type.
 */
typedef Int32	ColorType;

#define	cColorUnknown	0	/* Unknown color space */
#define	cColorRGB	1	/* Red, green, blue, usually unknown */
#define	cColorCMYK	2	/* Cyan, magenta, yellow, black */
#define	cColorCIELAB	3
#define	cColorCIELUV	4
#define	cColorGray	5
#define	cColorNone	6	/* No color, i.e. transparent */

/*
 * a ColorSpec includes a color record, a color type, and a chrominance
 * for white point (if needed)
 */


/* psm 7/31/96: took this out--never used, and collides with Linotype
 * color manager 
 */
#if 0 
typedef struct
{
    ColorType    type;         /* Type of color */
    ColorRecord  color;        /* The color itself */
    Chrominance  white;        /* The white point, in case it's needed */
} ColorSpec;
#endif /* 0 */

/* Standard Plane definitions, also offsets into a color record */
#define	    cRedPlane		0
#define	    cGreenPlane		1
#define	    cBluePlane		2

#define	    cPseudoPlane	0	/* The pseudo-color plane */

#define	    cCyanPlane		0
#define	    cMagentaPlane	1
#define	    cYellowPlane	2
#define	    cBlackPlane		3

#define	    cLuminancePlane	0	/* Luminance */
#define	    cXPlane		1	/* first chroma axis--a* or u* or x */
#define	    cYPlane		2	/* second chroma axis--b* or v* or y */

/*
 * A color map
 */
typedef struct
{
    Int16	    count;		/* # of colors in map */
    ColorType	    type;		/* Type of colors in map */
    ColorRecord	    *map;		/* The map itself */
} ColorMap;


/*
 * psm 4/6/94: adding support for more than one image plane in a PIXR. 
 * The way we do this is to support "hierarchical" PIXRs, where the parent
 * PIXR has a channel count of >1, which is the discriminant for the union
 * that follows. 
 */

 /*
  * A single plane. This structure is placed in a union with the "childPlanes"
  */
typedef struct PlaneStruct	*PlanePtr;
typedef struct PlaneStruct
{
    UInt32		 w;		/* width in pixels                  */
    UInt32		 h;		/* height in pixels                 */
    UInt32		 d;		/* depth in bits                    */
    UInt32		 bpl;		/* bytes/line                       */
    UInt32		 frame;		/* pixels of frame placed           */
					/* uniformly around image           */
    UInt32		 byteOrder;	/* order of bytes in image words    */
    ScaledInt1616	 xRes;		/* X-resolution, in DPI	            */
    ScaledInt1616	 yRes;		/* Y-resolution, in DPI	            */
    ColorType		 type;		/* Type  of image, see ColorType    */
    ColorMap		*colorMap;	/* Image's color map                */
    Chrominance		 white;		/* Image's white point, if relevant */
    UInt8		*imageBuffer;	/* pointer to beginning of image
					 * buffer, which may be different
					 * from "image" because of alignment
					 * constraints. */
    UInt8		*image;		/* the image data for this plane    */
    PlanePtr		 pPrevOwner;	/* pointer to previous owner of same
					 * image data block.		    */
    PlanePtr		 pNextOwner;	/* pointer to next owner of same
					 * image data block.		    */
} Plane;

struct Pixr
{
    /*
     * As mentioned above, if the number of color channels is 1, then
     * we use the "Plane" structure of the pixrData union; otherwise,
     * we use the childPlanes pointers as pointers to the planes, which
     * are themselves PIXRs with channels==1. This should be more-or-less
     * transparent to the users of the accessor functions, (with the
     * exception that some users want to be able to compact the heap)
     */
      
    UInt16		 channels;			/* Number of color channels */

    union
    {
	struct Pixr     *childPlanes[cMaxColorPlanes];	/* Child pointers */
	Plane		 plane;				/* The plane itself */
    } pixrData;
};

typedef struct Pixr PIXR;

/* These are now no longer going to be used, since we are changing
   the way everything is called to use error returns, and pass in
   the destination pixr pointer.
*/
#if 0

/*-----------------------------------------------------------------------*
 *      Useful (and suggested) macros for image creation/destruction     *
 *-----------------------------------------------------------------------*/

/* 
 * These always create one-channel, type cColorGray, undefined resolution
 * PIXRs of the depth passed.
 */
	/* usual case: standard frame */
#define  CREATE_PIXR(w, h, d)                createPixr(w, h, d, 32)
	/* special case: no frame */
#define  CREATE_PIXR_NF(w, h, d)             createPixr(w, h, d, 0)
        /* destroys and nulls the ptr */
#define  DESTROY_PIXR(pixr)                  ((pixr) = destroyPixr(pixr))

/*
 * Might as well have some color versions of the CREATE_PIXR, etc. macros
 * Also includes color type field, unlike CREATE_PIXR, and channel count. 
 * The depth is per-channel in the case of RGB. Does not create a color map.
 * Resolution will default to "cResolutionUndefined"
 */

/* Usual case: standard frame */
#define CREATE_COLOR_PIXR(w, h, d, c, t)	createPixrExtended(w, h, d, 32, c, t, NULL)

/* Special case: no frame */
#define CREATE_COLOR_PIXR_NF(w, h, d, c, t)	createPixrExtended(w, h, d, 0, c, t, NULL)

#endif

/*-----------------------------------------------------------------------*
 *
 *   Notes:
 *
 *       (1) The w and h fields give the size of the "proper" image, in
 *           pixels.  They do not include pad or frame pixels.
 *
 *       (2) These Pixr image structs can have a frame and line padding.
 *           For greatest ultimate simplicity, we will always
 *           always take the origin of the coordinate system to be the
 *           upper-left corner of the proper image (the part that is
 *           inside the frame).  Thus, no pixels in the proper image
 *           have negative values for any of their coordinates.
 *
 *       (3) If a frame exists, its width must be a multiple of 32 pixels.
 *           This also guarantees that each raster line of the proper
 *           image (sans frame) is aligned on a 32-bit word boundary.
 *           The frame pixels have the same depth as the image pixels.
 *
 *       (4) Pad pixels are placed to the right of the framed image when
 *           the width of the entire image (image plus frame) is not a
 *           multiple of 32.  They insure that the width of the image
 *           is an integral number of 32-bit words.
 *
 *       (7) For accessing subimages, the x_offset and y_offset fields
 *           can be set.  The x_offset must be restricted to allow
 *           word-alignment of the sub-image.   These offsets can be
 *           used in both owned and "secondary" Pixrs, the latter
 *           created with pixrPoint().
 *
 *       (8) Pixr image data is to be interpreted in 32-bit units in both 
 *           big-endian and little-endian worlds.  In all cases, we
 *           assume that the image data is placed in memory in strict
 *           left-->right order, with the leftmost image pixel being
 *           stored in the Most Significant bit/2-bit/nibble/byte/etc
 *           of the 32-bit word.  This order may differ for various
 *           file formats, which will have their own conventions
 *           [e.g., sunraster is the same ("big-endian");
 *           djfax is opposite ("little-endian")], but the readers and
 *           writers will conditionally reverse bytes as required so
 *           that the image data in memory is in the order given above.
 *
 *	 (9) The resolution fields I just added (3/30/94) are used
 *	     to represent the X- and Y- resolution of the PIXR in
 *	     spots per inch. This is useful when running "heuristics"
 *	     which need some idea of the scale of the thing passed.
 *	     The reason for having two independent axes is that some
 *	     devices (notably a lot of printers on the market) have
 *	     different X- and Y- resolutions, e.g. 600x300.
 *
 *-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*
  The following section contains definitions that have been excised from
  pixrect.h.  They are needed by almost all files that contain calls to
  pixr_rop().  We might want to move them to another file eventually, but
  for now they are here for convenience.
 *-----------------------------------------------------------------------*/
/*
 * Several of the above operations return a common, distinguished value when
 * an error arises.  That value is defined as follows:
 */
#define   PIX_ERR   -1

/*
 * Operations.  The 'op' in 'rasterop' may be any binary Boolean function, 
 * encoded as an integer from 0 to 15 (the op code) shifted left by one bit.  
 * The function is applied per-pixel.  
 *
 * The following permit the op to be expressed as Boolean combinations
 * of the two inputs 'src' and 'dst'.  Thus oring the source and destination
 * together is written as PIX_SRC|PIX_DST, while xoring the source with the
 * destination is PIX_SRC^PIX_DST.  Since ~op would set the color and clip
 * bits, the macro PIX_NOT is provided for use in place of ~.
 */
#define   PIX_SRC      (0xC << 1)
#define   PIX_DST      (0xA << 1)
#define   PIX_NOT(op)   ((op) ^ 0x1E)
#define   PIX_CLR      (0x0 << 1)
#define   PIX_SET      (0xF << 1)

/* macros which tell whether a rasterop needs SRC or DST values */
#define   PIXOP_NEEDS_DST(op)   ((((op)>>1)^(op)) & PIX_NOT(PIX_DST))
#define   PIXOP_NEEDS_SRC(op)   ((((op)>>2)^(op)) & PIX_NOT(PIX_SRC))

/* macros for encoding and extracting color field */
#define   PIX_COLOR(c)   ((c)<<5)
#define   PIX_OPCOLOR(op)   ((op)>>5)
#define   PIX_OP_CLIP(op)   ((op)&0x1f)
#define   PIX_OP(op)   ((op)&0x1e)

/*
 * The pseudo-operation PIX_DONTCLIP specifies that clipping should not
 * be performed.  PIX_CLIP is also provided, although unnecessary.
 */
#define   PIX_DONTCLIP      0x1
#define   PIX_CLIP      0x0

/* resolution constants */
#define	  cResolutionUndefined	    0

/* Possible values of order of bytes within an image word */
#define	  cBigEndianByteOrder		0
#define	  cLittleEndianByteOrder	1

/*
 * These are used to pass PIXR validations to the pixrValidate function.
 */

/* Depth flags: one of these should be set  */
#define cDepth1             1
#define cDepth2             2
#define cDepth4             4
#define cDepth8             8
#define cDepth10           16
#define cDepth12           32
#define cDepth16           64
#define cDepthMask        127
#define cDepthAny         127

/* Channel flag encondings: one of these should be set */
#define cChannels1        128
#define cChannels3        256
#define cChannelsMask     384
#define cChannelsAny      384

/* Byte order encodings */
#define cByteOrderBig     512        /* Big-endian: MSB at lowest addr */
#define cByteOrderLittle 1024        /* Little-endian: LSB at lowest addr */
#define cByteOrderMask   1536
#define cByteOrderAny    1536

/* Frame flags: # of pixels in frame */
#define cFrame0          2048
#define cFrame32         4096
#define cFrameMask       6144
#define cFrameAny        6144

#ifdef _MSC_VER
#pragma pack(pop, pixr_h)
#endif

#endif  /* PIXR_H_INCLUDED */

/*

$Log: pixr.h,v $
Revision 1.1.1.1  1999/09/20 18:19:12  jlin
Imported sources

Revision 1.18  1998/10/02 18:26:52  aber
pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v

CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
using libraries and includes.

Revision 1.17  1996/09/13 19:13:13  ddavies
ddavies on Fri Sep 13 11:42:10 PDT 1996

pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v

Add the imageBuffer entry to the PIXR data structure.  This records the
location of the first byte of the memory buffer holding an image.  This might
not be first byte of the frame because we may need to align the image
so that long memory accesses (8 byte?) can be performed.

Revision 1.16  1996/07/31 20:51:37  macleod
macleod on Wed Jul 31 13:50:08 PDT 1996

pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v

Disabled definition of ColorSpec type, because we never use it,
and it collides with the name of a structure used by Linotype's
color management system. (Perhaps we should go further in limiting
namespace pollution by prefixing ALL our names, including types,
structures, etc.)

Revision 1.15  1995/12/13 04:11:20  ejaquith
pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v

CommonCheck: Added ";" to IP_RCSINFO().
Removed function comment blocks from before prototypes and
added special prototype comment.

 * Revision 1.14  1995/10/28  01:21:23  ddavies
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Add pointers used to save the ring of owners of a particular array of
 * pixels to the Plane structure.  This allows us to have multiple PIXR
 * structures owning the same pixel array.  Also add the byteOrder flag
 * to the Plane structure since it was missing!
 *
 * Revision 1.13  1995/09/13  21:48:31  ddavies
 * ddavies on Wed Sep 13 14:47:59 PDT 1995
 *
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Change the resolutions stored in the image from Int16's to ScaleInt1616's so
 * we can avoid roundoff errors when scaling.
 *
 * Revision 1.12  1995/08/31  22:03:49  ejaquith
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.11  1995/08/30  23:56:39  ddavies
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Replace enumerated type ColorType with typedef of ColorType (as an Int32) and
 * #defines for its values.  This was done because the number of bytes requried
 * to represent a variable of type enum depends on the compiler used.
 *
 * Revision 1.10  1995/08/17  00:59:17  macleod
 * macleod on Wed Aug 16 17:58:51 PDT 1995
 *
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * For the pixr type flags, added "any" versions, i.e. cChannelsAny,
 * cDepthAny, etc.
 *
 * Revision 1.9  1995/07/31  17:33:39  macleod
 * macleod on Mon Jul 31 10:33:39 PDT 1995
 *
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Changed prototypes to conform to new standards.
 *
 * Revision 1.8  1995/06/12  14:53:13  macleod
 * macleod on Mon Jun 12 07:52:23 PDT 1995
 *
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Ooops. Forgot about this one (never checked it out).
 * Has a new type, ColorSpec, which combines color type, a color record,
 * and a white point.
 * Also added a color type for "none" (i.e. transparent)
 *
 * Revision 1.7  1994/09/08  00:35:26  lperry
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.6  1994/04/12  21:44:49  macleod
 * macleod on Tue Apr 12 14:44:26 PDT 1994
 *
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Changed CREATE_COLOR_PIXR and CREATE_COLOR_PIXR_NF to call newer
 * functions
 *
 * Revision 1.5  1994/04/12  01:10:44  macleod
 * macleod on Mon Apr 11 18:09:13 PDT 1994
 *
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Changed the PIXR structure again, this time quite a bit, to support
 * color PIXRs. Basically, the PIXR structure uses a "channels" field
 * as a discriminant for a union. If "channels" == 1, then the PIXR
 * has all of the usual PIXR stuff like size, pointer to image data, etc.
 * If "channels" is not 1, then the PIXR has up to four pointers to
 * PIXR structures, each of which has "channels" == 1. I've then modified
 * the accessor functions to deal with these structures. This eliminates
 * the need for a separate ColorPixr structure.
 *
 * Revision 1.4  1994/04/05  00:37:57  macleod
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Added ColorPixr type, which is an array of PIXR pointers, and a couple
 * of of other things. Also added resolution fields to the PIXR structure
 *
 * Revision 1.3  1993/12/03  01:11:30  ddavies
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:02:13  ddavies
 * pixr.h -> /product/ipcore/ipwrap/include/RCS/pixr.h,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

*/

