/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


/*
 * timing.h
 *
 * Macros to allow timing measurements and printouts, if the
 * compile-time flag TIMING has been defined and is 1.

 Usage example:

 foo()
 {
   Timer innerloop, outerloop;
   ...
   START_TIMER(outerloop);
   for(i=0; i < 9; i++)
   {
     compute1();
     START_TIMER(innerloop);
     for(j=0; j < 9999999; j++)
       compute2();
     PRINT_TIME("The innerloop took %.2f seconds\n", innerloop);
   }
   PRINT_TIME("The total time was %.2f seconds\n", outerloop);
   ...
 }
*/

#ifndef  _TIMING_H
#define  _TIMING_H


IP_RCSINFO(timing_h_RCSInfo, "$RCSfile: timing.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:13 $ */


/*--------*/
#if defined(TIMING) && TIMING

    /*--------*/
typedef clock_t Timer;

#ifdef __GNUC__
#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC          1000000	/* only our gcc includes on the Sun
					 * seem to be missing this */
#endif
extern	clock_t clock(void);	/* this is here because gcc decided this is
				 * only a cplusplus thingie.  Weird.
				 */
#endif	/* __GNUC__ */

#ifdef _MSC_VER
#pragma pack(push, timing_h, 1)
#endif

#define START_TIME(time) (time) = clock()
#define STOP_TIME(time)  (float)(clock() - (time))/CLOCKS_PER_SEC
#define ELAPSED_TIME(time) (time) = clock() - (time)

  	/* The SUN timer has a resolution of 1/60 second. Round it to 1/100.
           Other timers assumed to have resolution >= 1/100, so no rounding. */
    /*--------*/
#ifdef _MSC_VER
#pragma pack(pop, timing_h)
#endif

#ifdef _UNIX_
	/*--------*/
#define PRINT_TIME(printProc, printInfo, message, time) \
  i_printf(pIPWState, printProc, printInfo, message, (float)((long)((STOP_TIME(time) * 100) + 0.5)) / 100.0)
#define PRINT_ELAPSED(printProc, printInfo, message, time) \
  i_printf(pIPWState, printProc, printInfo, message, (float)((long)(((float)(time)/CLOCKS_PER_SEC * 100) + 0.5)) / 100.0)

    /*--------*/
#else
	/*--------*/
#define PRINT_TIME(printProc, printInfo, message, time) \
         i_printf(pIPWState, printProc, printInfo, message, STOP_TIME(time))
#define PRINT_ELAPSED(printProc, printInfo, message, time) \
         i_printf(pIPWState, printProc, printInfo, message, (float)(time)/CLOCKS_PER_SEC)
    /*--------*/
#endif   /* _UNIX_ */

/*--------*/
#else

    /*--------*/
typedef int Timer;
#define START_TIME(time)
#define STOP_TIME(time)
#define ELAPSED_TIME(time)
#define PRINT_TIME(printProc, printInfo, message, value)
#define PRINT_ELAPSED(printProc, printInfo, message, time)

/*--------*/
#endif /* TIMING */

#endif /* _timing_h */

/*

$Log: timing.h,v $
Revision 1.1.1.1  1999/09/20 18:19:13  jlin
Imported sources

Revision 1.14  1998/10/05 19:12:46  aber
timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v

CommonCheck: Fix module name in all-encompassing ifdef to be upper-case, to avoid
naming conflict with packing pragma name.

Revision 1.13  1998/10/02 18:27:31  aber
timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v

CommonCheck: Updated copyright notice, and added pragma packing of 1 for clients
using libraries and includes.

Revision 1.12  1998/01/14 00:09:42  ddavies
timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v

CommonCheck: Add IPState parameter to all callback routine definitions.

Revision 1.11  1995/12/13 04:13:29  ejaquith
timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v

CommonCheck: Added ";" to IP_RCSINFO().
Removed function comment blocks from before prototypes and
added special prototype comment.

 * Revision 1.10  1995/10/07  02:49:20  ddavies
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * Redefine the PRINT_TIME and PRINT_ELAPSED macros so they take a printing
 * procedure and private data for that procedure instead of an error stream.
 *
 * Revision 1.9  1995/08/31  22:04:25  ejaquith
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/17  22:23:58  jcraig
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * ipwrap house cleaning, take 1.
 *
 * Revision 1.7  1995/07/29  02:03:55  jcraig
 * jcraig on Fri Jul 28 19:03:48 PDT 1995
 *
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * first rev of cleanup change
 *
 * Revision 1.6  1995/07/17  23:41:24  jcraig
 * jcraig on Mon Jul 17 16:40:09 PDT 1995
 *
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * add macro which calculates elapsed time.
 * add macro which prints elapsed time (instead of macro which
 * calls STOP_TIME and prints).
 *
 * Revision 1.5  1994/09/08  00:36:56  lperry
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * Remove Date from RCS macro. Test for redundant types.pub includes.
 *
 * Revision 1.4  1994/05/05  01:05:53  ddavies
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * Define CLOCKS_PER_SEC if it isn't already defined.  Add prototype for
 * clock since gcc's time.h thinks it's a C++ thing.
 *
 * Revision 1.3  1993/12/03  01:12:24  ddavies
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:04:16  ddavies
 * timing.h -> /product/ipcore/ipwrap/include/RCS/timing.h,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

*/

