/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef XIFCONV_H_INCLUDED
#define XIFCONV_H_INCLUDED

IP_RCSINFO(xifconv_hRCSInfo, "$RCSfile: xifconv.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:13 $ */

#ifdef __cplusplus
extern "C"
{
#endif

#ifdef _MSC_VER
#pragma pack(push, xifconv_h, 1)
#endif

/* xifconv.h
 *
 */

/*
 * INCLUDES
 */


/*
 * CONSTANTS
 */
#define XIF_CONV_INVERT		0x00000001	/* 0 = white -> 0 = black */

/*
 * MACROS
 */

/*
 * TYPEDEFS
 */

/*
 * ENUMS
 */

/*
 * GLOBAL VARIABLE DECLARATIONS
 */

/*
 * FUNCTION PROTOTYPES
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do internal image processing to convert from TIFF format to PIXR
 *
 * LongDescription:::
 *  w_XifConvertImage() does minimal image processing required to put
 *  image data into an expected PIXR state (binary: 0 = white, non-paletted:
 *  0 = black).
 *
 * Input:::
 *  xImg::			Valid image
 *  pPixr::			Image data structure
 *  dwFlags::			Processes (see xifconv.h XIF_CONV_xxx)
 *  rowOffset::			Offset of first row to process
 *  rowCount::			Number of rows to process
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifConvertImage(XImg *xImg, PIXR *pPixr, UInt32 dwFlags, UInt32 rowOffset,
	UInt32 rowCount);

/*****************************************************************************
 * Begindocumentation:::
 *
 * OneLineDescription:::
 *  Convert the byte ordering of arbitrary imagery to PIXR order.
 *
 * LongDescription:::
 *  wp_PCtoPIXR() converts the byte ordering of arbitrary imagery to PIXR order.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  inImg::			Image to convert
 *  inBPL::			Bytes per image line in inImg
 *
 * Output:::
 *  outImg::			Image to receive the converted data
 *  outBPL::			Bytes per image line in outImg
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_PCtoPIXR(XImg *xImg, UInt8 **inImg, UInt32 inBPL, UInt8 **outImg,
	 UInt32 outBPL);

#ifdef _MSC_VER
#pragma pack(pop, xifconv_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* XIFCONV_H_INCLUDED */
