/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


#ifndef XIFF_H_INCLUDED
#define XIFF_H_INCLUDED

IP_RCSINFO(xiff_h_RCSInfo, "$RCSfile: xiff.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:13 $ */

/* xiff.h
 *
 * DESCRIPTION
 *	 Declarations for the XIFF Toolkit API.
 *
 */


#ifdef __cplusplus
extern "C"
{
#endif

#ifdef _MSC_VER
#pragma pack(push, xiff_h, 1)
#endif

/*
 * INCLUDES
 */

/*
 * CONSTANTS
 */

/* Current XIFF version: 4.0
 */
#define CURRENT_XIFF_VERSION	3	/* Major version number */
#define CURRENT_XIFF_REVISION	0	/* Minor version number */

#define XIFF_KEY_2_0	"XEROX DIFF"	/* 10-char key for 2.0 files */
#define XIFF_KEY_3_0	" eXtended "	/* 10-char key */
#define XIFF_KEY_LENGTH	10		/* Does not include NULL terminator */

#define XIFF_KEY_OFFSET	8	/* 8 bytes into file, follows TIFF header */
#define XIFF_EXTS_OFFSET	(XIFF_KEY_OFFSET \
				 + sizeof(XIFF_KEY_3_0) - 1 /* key */\
				 + 1 /* version */ \
				 + 1 /* revision */)
#define XIFF_SW_2_0	\
	"v1.0 Xerox Image Format"	/* Software tag that
					 * must appear on 2.0 primary IFDs */

/* Doc tags (XIFF extension tags)
 */
#define DOCTAG_NULL		0	/* Placeholder for future tag */
#define DOCTAG_PAGETABLE	1	/* First page table location */
#define DOCTAG_AUTHOR		2	/* Author block */
#define DOCTAG_REFERENCEPAGE	3	/* Reference page */
#define DOCTAG_NEXTEXT		4	/* Next set of document tags */
#define DOCTAG_SYMLIBTABLE	5	/* First symbol library */
#define DOCTAG_COUNT		6	/* One more than last one */

typedef unsigned long int	DocTags;

/* Annotation flags */
#define ANNOT_NOTVIEWABLE	0x00000001	/* Not viewable */
#define ANNOT_HYPERLINK		0x00000002	/* Annotation also hyperlinks
						 * to either another spot in
						 * the document (a page), or
						 * an URL. */


/*
 * MACROS
 */
/* Bits per pixel */
#define XIF_BPP(coltype)	((coltype) == XIF_COLORTYPE_BINARY? 1:	\
				 (coltype) == XIF_COLORTYPE_COLOR24? 24:\
				 ((coltype) == XIF_COLORTYPE_GRAY8 ||	\
				  (coltype) == XIF_COLORTYPE_COLOR8)? 8:\
				 ((coltype) == XIF_COLORTYPE_GRAY4 ||	\
				  (coltype) == XIF_COLORTYPE_COLOR4)? 4: 0)

/* Bits per pixel element */
#define XIF_BPPE(coltype)	((coltype) == XIF_COLORTYPE_BINARY? 1:	\
				 (coltype) == XIF_COLORTYPE_COLOR24? 8:	\
				 ((coltype) == XIF_COLORTYPE_GRAY8 ||	\
				  (coltype) == XIF_COLORTYPE_COLOR8)? 8:\
				 ((coltype) == XIF_COLORTYPE_GRAY4 ||	\
				  (coltype) == XIF_COLORTYPE_COLOR4)? 4: 0)

/* Channels per pixel */
#define XIF_CHANNELS(coltype)	((coltype) == XIF_COLORTYPE_BINARY? 1:	\
				 (coltype) == XIF_COLORTYPE_COLOR24? 3:	\
				 ((coltype) == XIF_COLORTYPE_GRAY8 ||	\
				  (coltype) == XIF_COLORTYPE_COLOR8)? 1:\
				 ((coltype) == XIF_COLORTYPE_GRAY4 ||	\
				  (coltype) == XIF_COLORTYPE_COLOR4)? 1: 0)

/*
 * TYPEDEFS
 */
#ifndef _TYPES_PUB_INCLUDED
typedef unsigned char      	UInt8;
typedef unsigned short int 	UInt16;
typedef unsigned long int  	UInt32;
typedef signed char		Int8;
typedef short int		Int16;
typedef long int		Int32;
typedef unsigned long int	Bool;
typedef unsigned long int	ErrCode;
#endif


/*
 * XIF basic types
 *
 * For portability, the XIF library uses platform-independent type names.
 * The definitions of the type will be in a single platform-dependent
 * section here.
 */
#if defined(sparc)
#define signed
#endif


/*
 * XIF derived types
 */

/* Handle to client's instance of library */
typedef UInt32 XIF_INSTHANDLE;

/* Handle to open document */
typedef UInt32 XIF_DOCHANDLE;

/* Handle to an active page */
typedef UInt32 XIF_PAGEHANDLE;

/* Handle to an active image */
typedef UInt32 XIF_IMGHANDLE;


/* Return codes (seemingly out of sequence since they were borrowed from
 * a compression/decompression package which this library sits on top of). */
#define XIF_NOERROR		0	/* Normal return */
#define XIF_NOMEMORY		1	/* Ran out of memory */
#define XIF_BADPARAMETER	2	/* Invalid parameter to XIF_function */
#define XIF_NOSUPPORT		6	/* Unsupported feature or condition */
#define XIF_BUFFEREXCEEDED	7	/* Data buffer filled and will
					 * overflow */
#define XIF_INTERNAL_ERROR	12	/* Unexpected internal state (shouldn't
					 * happen) */
#define XIF_CLIENTABORT		14	/* Client requested abort */
#define XIF_TIFFEMPTY		18	/* 0-page TIFF (okay for XIFF) */
#define XIF_NOTTIFF		19	/* File does not comform to TIFF */
#define XIF_NOTXIFF		20	/* File does not comform to XIFF */
#define XIF_SEQUENCE		21	/* Function called out-of-sequence */
#define XIF_BADFILEMODE		22	/* File not open in BINARY/READ/WRITE
					 * mode */
#define XIF_BADFILEFORMAT	23	/* File format/data is corrupted */
#define XIF_READ_ERR		2000	/* Error reading */
#define XIF_WRITE_ERR		2001	/* Error writing */
#define XIF_SEEK_ERR		2002	/* Error seeking */
#define XIF_FILELEN_ERR		2003	/* Error seeking to EOF */


/* Types of image */
#define XIF_COLORTYPE_NONE	0	/* None specified */
#define XIF_COLORTYPE_BINARY	1	/* Binary image */
#define XIF_COLORTYPE_GRAY4	2	/* 4 bpp grayscale (values: 0-15) */
#define XIF_COLORTYPE_GRAY8	3	/* 8 bpp grayscale (values: 0-255) */
#define XIF_COLORTYPE_COLOR4	4	/* 4 bpp color (paletted indices: 0-15)
					 */
#define XIF_COLORTYPE_COLOR8	5	/* 8 bpp color (paletted indices:
					 * 0-255) */
#define XIF_COLORTYPE_COLOR24	6	/* 24 bpp color (8 bpp for each of RGB)
					 */

typedef unsigned long int	XIF_COLORTYPE;


/*
 * The type of the image with respect to the image/subimage
 * hierarchy. _PRIMARY indicates text plane/page image, which
 * is binary, unless no other images are attached to this page.
 * These image types can be divided into two classes, those
 * with image content and those that modify image content.
 * For example, an XIF_IMGTYPE_MASK depends on and modifes
 * the image content of a non-binary parent image.  An 
 * XIF_IMGTYPE_MASK would never be displayed independently of
 * its parent.
 */
#define XIF_IMGTYPE_UNKNOWN	0	/* None */
#define XIF_IMGTYPE_PRIMARY	1	/* Binary text - used as a FG/BG
					 * selector, Image layer 2 */
#define XIF_IMGTYPE_LAYER	2	/* Belonging to an image layer
					 * other than 2 */
#define XIF_IMGTYPE_MASK	3	/* Transparency mask */

typedef unsigned long int	XIF_IMGTYPE;


/*
 * XIF general I/O functions
 *
 * The XIF client is responsible for managing the interface to the
 * local file system and must provide functions for basic file
 * I/O. To the XIF library, a file is represented by set of I/O
 * functions, and a file ID. This file ID (specified by the client),
 * is sent as a parameter to each of the I/O functions when the library
 * moves about the file, or accesses the data.
 *
 * As an example, let's say I have an implementation of the functions
 * myRead(), myWrite(), mySeek(), and myFilesize(). I setup a file
 * interface after a XIFF instance is established:
 *	fd = open(filename, _O_READ | _O_BINARY);
 *	XIF_OpenDocumentRead(xInst, (Int32)fd, myRead, myWrite, mySeek,
 *		mySize, &docHandle);
 * Now the docHandle is associated with the file via the client ID (file
 * descriptor), and I/O functions.
 *
 * The following is the format of the functions:
 */
typedef Int32 (CDECL *XIF_READFUNC)(UInt32 dwClientFileID,
				    Int8 *pBuf,
				    Int32 dwByteCount
);
typedef Int32 (CDECL *XIF_WRITEFUNC)(UInt32 dwClientFileID,
			       const Int8 *pBuf,
				     Int32 dwByteCount
);
typedef Int32 (CDECL *XIF_SEEKFUNC)(UInt32 dwClientFileID,
				    Int32 dwOffset
);
typedef Int32 (CDECL *XIF_SIZEFUNC)(UInt32 dwClientFileID
);


/*
 * Compression methods
 */
#define XIF_COMP_UNKNOWN	0x0000	/* No valid compression */
#define XIF_NOC			0x0001	/* No compression, leave as raw data */
#define XIF_G4			0x0002	/* CCITT group-4 (ITU-T T.6) */
#define XIF_JPEG		0x0004	/* JPEG */
#define XIF_WAVELET		0x0008	/* Xerox Wavelet */
#define XIF_TOKENIZATION	0x0010	/* Tokenization */
#define XIF_LLTOKENIZATION	0x0020	/* Lossless Tokenization */
#define XIF_LOSSYDITHER		0x0040	/* Xerox Lossy Dither */
#define XIF_COLORFORTOKENS	0x0080	/* Color plane for later color
					 * tokenization */

typedef unsigned long int	XIF_COMPRESSION;


/*
 ******************************************************************************
 *				FUNCTION PROTOTYPES			      *
 ******************************************************************************
 */

/*
 * XIF progress function. The progress function is selected by the client.
 * It is called periodically by XIF functions to report progress and offer
 * the client an opportunity to cancel. The meaning of the progress function
 * depends on the function being carried out. If not otherwise documented,
 * the progress value will be -1, indicating no meaningful progress to report, 
 * or a number from 0 to 100, indicating percent complete.
 */
typedef Int32
(*XIF_PROGRESSFUNC)(
	UInt32 dwClientID,
	Int32 dwProgress);

ErrCode CDECL
XIF_SetProgressFunc(
	XIF_INSTHANDLE xInstHandle,
	XIF_PROGRESSFUNC xProgress);


/*
 * XIF memory allocation function. These function allows the client to
 * override XIF's memory allocation functions. It is not required, and
 * may not be available on all platforms.
 *
 * Once the functions have been overridden, they can be restored to
 * their defaults by passing NULL for each function.
 *
 * 'hasPtrs' is non-zero when allocating an array of pointers. On some
 * systems, this indicates that the memory management must update the
 * contents of this array, should the memory pointed to by this array
 * move. It is a garbage collection feature available to the client.
 * >>> However, none of the calls made by any of the functions make use
 * >>> of this (hasPtrs always equals 0).
 */
typedef void * (CDECL *XIF_MALLOCFUNC)(size_t dwBytes,
				       Bool hasPtrs,
				       void *pClientMallocID);
typedef void (CDECL *XIF_FREEFUNC)(void *pBuf,
				   size_t dwBytes,		/* Ignored */
				   void *pClientMallocID);

/*
 * XIF_SetMallocFuncs
 */
ErrCode CDECL
XIF_SetMallocFuncs(
	XIF_INSTHANDLE xInstHandle,
	void *dwClientMallocID,
	XIF_MALLOCFUNC xMalloc,
	XIF_FREEFUNC xFree
);


/*
 * XIF_InitInstance
 */
ErrCode CDECL
XIF_InitInstance(
	UInt32 dwClientInstID,
	XIF_INSTHANDLE *pxInst
);


/*
 * XIF_GetInstance
 */
ErrCode CDECL
XIF_GetInstance(
    XIF_DOCHANDLE xDocHandle,
    XIF_INSTHANDLE *pxInstHandle
);


/*
 * XIF_EndInstance
 */
ErrCode CDECL
XIF_EndInstance(
	XIF_INSTHANDLE xInstHandle
);


/*
 * XIF_GetClientID
 */
ErrCode CDECL
XIF_GetClientID(
	XIF_INSTHANDLE xInstHandle,
	UInt32 *dwClientInstID
);


/*
 * XIF_IsXiff
 */
ErrCode CDECL
XIF_IsXiff(
	XIF_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	XIF_READFUNC xReadFunc,
	XIF_SEEKFUNC xSeekFunc,
	XIF_SIZEFUNC xSizeFunc,
	UInt8 *pVersion,
	UInt8 *pRevision
);


/*
 * XIF_OpenDocumentRead
 */
ErrCode CDECL
XIF_OpenDocumentRead(
	XIF_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	XIF_READFUNC xReadFunc,
	XIF_WRITEFUNC xWriteFunc,
	XIF_SEEKFUNC xSeekFunc,
	XIF_SIZEFUNC xSizeFunc,
	XIF_DOCHANDLE *pxDocHandle
);


/*
 * XIF_CloseDocument
 */
ErrCode CDECL
XIF_CloseDocument(
    XIF_DOCHANDLE xDocHandle
);


/*
 * XIF_GetDocInfo
 */
ErrCode CDECL
XIF_GetDocInfo(
	XIF_DOCHANDLE xDocHandle,	/* Input document handle */
/* Returned values */
	UInt8 *pVersion,		/* Version of XIFF */
	UInt8 *pRevision,		/* Revision of this version of XIFF */
	UInt16 *pwNPages,		/* Number of pages in file */
	const Int8 **ppAuthorBlock,	/* Author string (NULL if none) */
	UInt16 *pwAuthorBlockBytes,	/* Size of author string */
	Int32 *pdwReferencePageOffset	/* Reference page offset (-1 if none) */
);


/*
 * XIF_GetPageInfo
 */
ErrCode CDECL
XIF_GetPageInfo(
	XIF_DOCHANDLE xDocHandle,	/* Input document handle */
	Int32 dwPageOffset,		/* Page offset (starting with 0) */
/* Returned values */
	UInt32 **ppdwImageCounts,	/* Number of images within each layer
					 */
	UInt32 *pdwLayersPresent,	/* Number of layers present (although
					 * some layers may be missing) */
	UInt16 *pwAnnotationCount,	/* Number of annotations on page */
	XIF_PAGEHANDLE *pxPageHandle	/* Page handle, used to identify this
					 * page in the page level routines */
);


/*
 * XIF_GetImageInfo
 */
ErrCode CDECL
XIF_GetImageInfo(
	XIF_PAGEHANDLE xPageHandle,	/* Page image handle */
	UInt32 dwImgLayer,		/* Image layer in the MRC */
	UInt32 dwImgOffset,		/* Image offset (for multiple images)
					 */
/* Returned values */
	UInt32 *pdwXOffset,		/* X location in page coordinates */
	UInt32 *pdwYOffset,		/* Y location in page coordinates */
	UInt32 *pdwXResolution,		/* X Resolution of image in dpm */
	UInt32 *pdwYResolution,		/* Y Resolution of image in dpm */
	UInt32 *pdwWidth,		/* Width in image's X resolution units
					 */
	UInt32 *pdwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 *pdwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	XIF_COMPRESSION *pxCompression,	/* Compression type for src/dst */
	XIF_COLORTYPE *pxColorType,	/* Color type (see XIF_COLORTYPE) */
	UInt8 *pPalette,		/* Color palette if XIF_COLORTYPE is:
					 * XIF_COLORTYPE_COLOR4 or
					 * XIF_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	UInt32 *pdwMaskImages,		/* Number of mask images */
	XIF_IMGHANDLE *pxImgHandle	/* Returned image handle */
);


/*
 * XIF_GetImageStats
 */
ErrCode CDECL
XIF_GetImageStats(
	XIF_IMGHANDLE xImgHandle,	/* Image handle */
/* Returned values */
	UInt32 *pdwCmpSize,		/* Compressed image byte total */
	UInt32 *pdwDictCount,		/* Number of token dictionaries */
	UInt32 **ppdwDictIDs,		/* Dictionary ID list (if tokenized) */
	Int32 **ppdwDictSizes		/* Dictionary byte list (tokenized) */
);


/*
 * XIF_GetMaskImageInfo
 */
ErrCode CDECL
XIF_GetMaskImageInfo(
	XIF_IMGHANDLE xParentImgHandle,	/* Parent image handle */
	UInt32 dwImgOffset,		/* Image offset (for multiple images)
					 */
/* Returned values */
	UInt32 *pdwXOffset,		/* X location in page coordinates */
	UInt32 *pdwYOffset,		/* Y location in page coordinates */
	UInt32 *pdwXResolution,		/* X Resolution of image in dpm */
	UInt32 *pdwYResolution,		/* Y Resolution of image in dpm */
	UInt32 *pdwWidth,		/* Width in image's X resolution units
					 */
	UInt32 *pdwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 *pdwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	XIF_COMPRESSION *pxCompression,	/* Compression type for src/dst */
	XIF_IMGHANDLE *pxImgHandle	/* Returned image handle */
);


#define XIF_IMGFLAGS_RGBOrder		(0x00)
#define XIF_IMGFLAGS_BGROrder		(0x01)		/* NIMP */
#define XIF_IMGFLAGS_TopToBottom	(0x00)
#define XIF_IMGFLAGS_BottomToTop	(0x02)		/* NIMP */
#define XIF_IMGFLAGS_WhiteIsZero	(0x00)
#define XIF_IMGFLAGS_BlackIsZero	(0x04)		/* NIMP */
#define XIF_IMGFLAGS_ColorChunky	(0x00)		/* NIMP */
#define XIF_IMGFLAGS_ColorPlanar	(0x08)
#define XIF_IMGFLAGS_ByteOrder		(0x00)		/* NIMP */
#define XIF_IMGFLAGS_DwordOrder		(0x10)
#define XIF_IMGFLAGS_Uncompressed	(0x00)
#define XIF_IMGFLAGS_Compressed		(0x20)		/* NIMP */

#define XIF_IMGFLAGS_ColorOrder		(0x01)
#define XIF_IMGFLAGS_ScanlineOrder	(0x02)
#define XIF_IMGFLAGS_PhotoInterp	(0x04)
#define XIF_IMGFLAGS_ColorFormat	(0x08)
#define XIF_IMGFLAGS_ByteOrdering	(0x10)
#define XIF_IMGFLAGS_Compression	(0x20)

/*
 * XIF_ImageReadStart
 */
ErrCode CDECL
XIF_ImageReadStart(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwFlags
);


/*
 * XIF_GetCompressedStripSize
 */
ErrCode CDECL
XIF_GetCompressedStripSize(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImageRowOffset,
	UInt32 *pdwBytes
);


/*
 * XIF_CompressedReadStrip
 */
ErrCode CDECL
XIF_CompressedReadStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppCmpBuffer,
	UInt32 *pdwCmpBufsize,
	UInt32 *pdwBytesFilled
);


/*
 * XIF_ImageReadStrip
 */
ErrCode CDECL
XIF_ImageReadStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 *pdwRowsFilled
);


/*
 * XIF_ImageReadFinish
 */
ErrCode CDECL
XIF_ImageReadFinish(
	XIF_IMGHANDLE xImgHandle
);


/*
 * XIF_GetLayerInfo
 */
ErrCode CDECL
XIF_GetLayerInfo(
    XIF_PAGEHANDLE xPageHandle,
    UInt32 dwImgLayer,
    XIF_COMPRESSION *pxCompression,
    UInt32 *pdwXOffset,
    UInt32 *pdwYOffset,
    UInt32 *pdwXResolution,
    UInt32 *pdwYResolution,
    UInt32 *pdwWidth,
    UInt32 *pdwHeight,
    XIF_IMGHANDLE *pxImgHandle
);


/*
 * XIF_ImageReadLayer
 */
ErrCode CDECL
XIF_ImageReadLayer(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImageHeight,
	UInt8 *pImageBuffer,
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
);


/*
 * XIF_GetAnnotationInfo
 */
ErrCode CDECL
XIF_GetAnnotationInfo(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 wAnnotOffset,		    /* Offset of annotation for this
					     * page */
	UInt16 *pwAnnotType,		    /* Type of annotation */
	UInt32 *pdwXOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 *pdwYOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 *pdwWidth,		    /* Width in page coordinates */
	UInt32 *pdwHeight,		    /* Height in page coordinates */
	UInt32 *pdwFlags,		    /* Annotation flags (see ANNOT_xxx
					     * above */
	UInt32 *pdwAnnotLength		    /* Length of annotation data, in
					     * bytes */
);


/*
 * XIF_GetAnnotationData
 */
ErrCode CDECL
XIF_GetAnnotationData(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 wAnnotOffset,
	UInt32 dwBytes,
	UInt8 *pData
);


/*
 * **********						********** 
 * **********	BEGIN WRITE API		********** 
 * **********						********** 
 */

/*
 * XIF_CreateDocument
 */
ErrCode CDECL
XIF_CreateDocument(
	XIF_INSTHANDLE xInstHandle,
	Int32 dwClientID,		/* Input client ID */
	UInt8 version,			/* Version of XIFF format to create */
	XIF_READFUNC xReadFunc,		/* Input function */
	XIF_WRITEFUNC xWriteFunc,	/* Input function */
	XIF_SEEKFUNC xSeekFunc,		/* Input function */
	XIF_SIZEFUNC xSizeFunc,		/* Input function */
	XIF_DOCHANDLE *pxDocHandle	/* Returned */
);


/*
 * XIF_OpenDocumentReadWrite
 */
ErrCode CDECL
XIF_OpenDocumentReadWrite(
	XIF_INSTHANDLE xInstHandle,
	Int32 dwClientID,		/* Input client ID */
	UInt8 version,			/* Version of XIFF format to create */
	XIF_READFUNC xReadFunc,		/* Input function */
	XIF_WRITEFUNC xWriteFunc,	/* Input function */
	XIF_SEEKFUNC xSeekFunc,		/* Input function */
	XIF_SIZEFUNC xSizeFunc,		/* Input function */
	XIF_DOCHANDLE *pxDocHandle	/* Returned */
);


/*
 * XIF_CopyImage
 */
ErrCode CDECL
XIF_CopyImage(
	XIF_IMGHANDLE xSrcImgHandle,
	XIF_IMGHANDLE xParentImgHandle,
	XIF_PAGEHANDLE xDstPageHandle,
	UInt32 dwImgLayer,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwBestStripHeight,
	XIF_IMGHANDLE *pxDstImgHandle
);


/*
 * XIF_CopyPage
 */
ErrCode CDECL
XIF_CopyPage(
	XIF_DOCHANDLE xSrcDocHandle,
	XIF_DOCHANDLE xDstDocHandle,
	Int32 dwSrcPageOffset,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
);


/*
 * XIF_AddPageStart
 */
ErrCode CDECL
XIF_AddPageStart(
	XIF_DOCHANDLE xDocHandle,
	Int32 dwPageOffset,		    /* Page offset */
	XIF_PAGEHANDLE *pxPageHandle	    /* Page handle returned */
);


/*
 * XIF_AddPageFinish
 */
ErrCode CDECL
XIF_AddPageFinish(
	XIF_PAGEHANDLE xPageHandle
);


/*
 * XIF_DeletePage
 */
ErrCode CDECL
XIF_DeletePage(
	XIF_PAGEHANDLE xPageHandle
);


/*
 * XIF_CheckImageInfo()
 */
ErrCode CDECL
XIF_CheckImageInfo(
	XIF_PAGEHANDLE xPageHandle,
	UInt32 *pdwImgLayer,
	UInt32 *pdwImgOrder,
	XIF_COLORTYPE *pxColorType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	UInt8 *pPalette,
	XIF_COMPRESSION *pxCompression
);


/*
 * XIF_AddImage
 */
ErrCode CDECL
XIF_AddImage(
	XIF_PAGEHANDLE xPageHandle,	/* Page to attach to */
	UInt32 dwImgLayer,		/* Image layer in the MRC */
	XIF_COLORTYPE xColorType,	/* Color type (XIF_COLORTYPE_BINARY
					 * ...) */
	UInt32 dwXOffset,		/* X location in page coordinates */
	UInt32 dwYOffset,		/* Y location in page coordinates */
	UInt32 dwXResolution,		/* X Resolution of image in dpm */
	UInt32 dwYResolution,		/* Y Resolution of image in dpm */
	UInt32 dwWidth,			/* Width in image's X resolution units
					 */
	UInt32 dwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 dwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	UInt8 *pPalette,		/* Color palette if XIF_COLORTYPE is:
					 * XIF_COLORTYPE_COLOR4 or
					 * XIF_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	XIF_IMGHANDLE *pxImgHandle	/* Returned image handle for image to
					 * be added */
);


/*
 * XIF_DeleteImage
 */
ErrCode CDECL
XIF_DeleteImage(
	XIF_IMGHANDLE xImgHandle
);


/*
 * XIF_AddMaskImage
 */
ErrCode CDECL
XIF_AddMaskImage(
	XIF_IMGHANDLE xParentImgHandle,	/* Parent image to attach to */
	UInt32 dwBestStripHeight,	/* For best performance, write strips
					 * with height a multiple of this. */
	XIF_IMGHANDLE *pImgHandle	/* Returned image handle for image to
					 * be added */
);


/*
 * XIF_ImageWriteStart
 *
 * Initiate an image write, to be followed with a sequence of one or more calls
 * to XIF_ImageWriteStrip() or XIF_CompressedWriteStrip() or
 * XIF_TokenizeImage().
 * In the case of XIF_TokenizeImage(), if using color tokenization, then
 * the color image must have been previously passed, and marked for compression
 * using compression type: XIF_COLORFORTOKENS.
 */
ErrCode CDECL
XIF_ImageWriteStart(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwFlags, 
	XIF_COMPRESSION *pxCompression, /* XIF_COMP_UNKNOWN yields default */
	UInt32 *pdwJPEGQuality	/* 0-100:
				 *   0 - default used and passed back
				 *  75   good quality, best compression
				 *  90   better quality, good compression
				 * 100   best quality, okay compression */
);	  

	
/*
 * XIF_CompressedWriteStrip
 */
ErrCode CDECL
XIF_CompressedWriteStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 *pCmpBuffer,
	UInt32 dwCmpBytes,
	UInt32 *pdwBytesStored
);

	
/*
 * XIF_ImageWriteStrip
 */
ErrCode CDECL
XIF_ImageWriteStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pRowsFilled
);

	
/*
 * XIF_ImageWriteFinish
 */
ErrCode CDECL
XIF_ImageWriteFinish(
	XIF_IMGHANDLE xImgHandle
);
	

/* Tokenization flags */	
#define TOKENIZE_OFF			(0x00)	/* No tokenization */
#define TOKENIZE_ON			(0x01)	/* Tokenization enabled */
#define TOKENIZE_LOSSLESS	    	(0x02)	/* Tokenization in lossless
						 * mode */
#define TOKENIZE_BINARYSCAN		(0x04)	/* Image was scanned with a
						 * binary scanner, indicating
						 * that halftones may be
						 * removed and lossy dither
						 * compressed before
						 * tokenization */
#define TOKENIZE_COLORTOO	    	(0x08)	/* Color tokenize: use
						 * previously added FG image
						 * to extract colors from */

/*
 * XIF_TokenizeImage()
 */
ErrCode CDECL
XIF_TokenizeImage(
	XIF_IMGHANDLE xImgHandle,
	UInt8 *pImgData,	/* One channel only */
	UInt32 dwBytesPerRow,	/* Bytes per row */
	UInt32 dwFlags		/* Flags: Lossless Tokenization/Binary Scan
				 * (see TOKEN_FLAGS) */
);


/*
 * XIF_WriteTokenImages()
 */
ErrCode CDECL
XIF_WriteTokenImages(
	XIF_DOCHANDLE xDocHandle
);

/*
 * XIF_WriteTokenImage()
 */
ErrCode CDECL
XIF_WriteTokenImage(
	XIF_IMGHANDLE xImgHandle
);


/*
 * XIF_WriteTokenDictionary()
 */
ErrCode CDECL
XIF_WriteTokenDictionary(
	XIF_DOCHANDLE xDocHandle
);


/*
 * XIF_ShareColorTokenImage()
 */
ErrCode CDECL
XIF_ShareColorTokenImage(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwBinaryImageBytes,	/* Bytes shared by binary image */
	XIF_COLORTYPE xColorType	/* Color image color type */
);


/*
 * XIF_SetAuthor
 */
ErrCode CDECL
XIF_SetAuthor(
	XIF_DOCHANDLE xDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
);


/*
 * XIF_SetReferencePage
 */
ErrCode CDECL
XIF_SetReferencePage(
	XIF_DOCHANDLE xDocHandle,
	Int32 dwReferencePageOffset
);


/*
 * XIF_SetAnnotTableEntryCount
 */
ErrCode CDECL
XIF_SetAnnotTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * XIF_SetPageTableEntryCount
 */
ErrCode CDECL
XIF_SetPageTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * XIF_SetSymlibTableEntryCount
 */
ErrCode CDECL
XIF_SetSymlibTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * XIF_AddAnnotation
 */
ErrCode CDECL
XIF_AddAnnotation(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 wAnnotType,		    /* Type of annotation */
	UInt32 dwXOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 dwYOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 dwWidth,			    /* Width in page coordinates */
	UInt32 dwHeight,		    /* Height in page coordinates */
	UInt32 dwFlags,			    /* Annotation flags (see ANNOT_xxx
					     * above) */
	UInt8 *pBuf,			    /* Annotation data */
	UInt32 dwAnnotLength,		    /* Length of annotation data, in
					     * bytes */
	UInt16 *pAnnotID		    /* ID given to this annotation
					     * (offset) */
); 


/*
 * XIF_DeleteAnnotation
 */
ErrCode CDECL
XIF_DeleteAnnotation(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 dwAnnotOffset
);


/*
 * XIF_Compress
 */
ErrCode CDECL
XIF_Compress(
	XIF_IMGHANDLE xImgHandle,	/* Image handle */
	XIF_COMPRESSION xCompression,	/* See XIF_COMPRESSION types */
	UInt32 *pdwQuality,		/* Quality to use, actual returned */
	XIF_COLORTYPE xColorType,	/* Type of color */
	UInt32 dwFlags,			/* XIF_IMGFLAGS_xxx combo */
	UInt8 *pSrc[3],			/* Source buffer channels */
	UInt32 dwBytesPerRow,		/* Bytes per row in image buffer */
	UInt32 dwRowOffset,		/* Row offset in image of data to
					 * compress */
	UInt32 dwWidth,			/* Columns of COLORTYPE in src image */
	UInt32 dwHeight,		/* Rows in src image */
	UInt8 **ppDst,			/* Pointer to destination compressed
					 * data buffers */
	UInt32 *pdwBytesInDst,		/* Pointer to byte count of destination
					 * buffer */
	UInt32 *pdwBytesFilled		/* Actual bytes in destination filled
					 */
);


/*
 * XIF_Decompress
 */
ErrCode CDECL
XIF_Decompress(
	XIF_IMGHANDLE xImgHandle,	/* Image handle */
	XIF_COMPRESSION xCompression,	/* See XIF_COMPRESSION types */
	XIF_COLORTYPE xColorType,	/* Type of color */
	UInt32 dwFlags,			/* XIF_IMGFLAGS_xxx bit combo */
	UInt8 *pSrc,			/* Source buffer */
	UInt32 dwBytesInSrc,		/* Byte count of source buffer */
	UInt32 dwRowOffset,		/* Destination row offset */
	UInt32 dwWidth,			/* Columns of COLORTYPE in dst image */
	UInt32 dwHeight,		/* Rows in dst image */
	UInt8 *pDst[3],			/* Destination image buffer channels */
	UInt32 dwBytesPerRow,		/* Bytes per row in image buffer */
	UInt32 *pdwRowsFilled		/* Number of rows successfully filled
					 */
);


#ifdef _MSC_VER
#pragma pack(pop, xiff_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* XIFF_H_INCLUDED */
