/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef XIFFILE_H_INCLUDED
#define XIFFILE_H_INCLUDED

IP_RCSINFO(xiffile_hRCSInfo, "$RCSfile: xiffile.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:13 $ */

#ifdef __cplusplus
extern "C"
{
#endif

/* xiffile.h
 *
 * DESCRIPTION
 *	 Declarations for XIFF file level structures (and above, like instance
 *	 handles).
 *
 */

/*
 * INCLUDES
 */
#include "xiff.h"

#ifdef _MSC_VER
#pragma pack(push, xiffile_h, 1)
#endif

/*
 * CONSTANTS
 */
#define XIFF_MAGIC_NUM	0x123abc00		/* Instance identifier */

/*
 * MACROS
 */

#define XERR_GEN(cond, err)\
do { \
    if (cond) \
    {\
        return err;\
    }\
} while(0)

/*
 * XERR_GEN
 *  Generate an error. "cond" is some condition (probably, it's just tested
 *  for non-zero) and "err" is the error code to generate 
 */ 

#define XERR_GEN(cond, err)\
do { \
    if (cond) \
    {\
        return err;\
    }\
} while(0)

/*
 * XERR_GEN_CLEANUP
 *   Error generation: instead of returning, as in XERR_GEN, go to a cleanup
 *   action at the end of the function. 
 */

#define XERR_GEN_CLEANUP(cond, err)\
do {\
    if (cond)\
    {\
        __error__ = err;\
        goto ERR_CLEANUP;\
    }\
} while(0)

/*
 * XERR_PASS
 *   This is an error check on a function call. It assumes that the
 *   function returns an error code, and if the error code is non-zero,
 *   returns from the function (after possibly printing an error message).
 */

#define XERR_PASS(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != cSuccessful)\
    {\
        return __error__;\
    }\
} while (0)

/*
 * XERR_PASS_CLEANUP
 *   This is an error check on a function call. It is like XERR_PASS except
 *   that rather than returning from the function, it sets the local
 *   "__error__" variable, and does a "goto" to some cleanup code. The
 *   cleanup code is labeled "ERR_CLEANUP". The cleanup should have an
 *   invocation of the "ERROR_RETURN" macro at the end of it.
 */

#define XERR_PASS_CLEANUP(functionCall)\
do {\
    __error__ = functionCall;\
    if (__error__ != cSuccessful)\
    {\
        goto ERR_CLEANUP;\
    }\
} while (0)



#define XIF_FILETOKEN(xFile)			(&(xFile)->fileToken)
#define XIF_READ(xFile, buffer, bytes)		\
	if (XIF_FILETOKEN(xFile)->FileRead(XIF_FILETOKEN(xFile)->dwClientFileID, \
	    (Int8 *)(buffer), bytes) != (bytes)) {	\
	    __error__ = XIF_READ_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define XIF_WRITE(xFile, buffer, bytes)		\
	if (XIF_FILETOKEN(xFile)->FileWrite(XIF_FILETOKEN(xFile)->dwClientFileID, \
	    (const Int8 *)(buffer), bytes) != (bytes)) {	\
	    __error__ = XIF_WRITE_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define XIF_SEEK(xFile, offset)			\
	if (XIF_FILETOKEN(xFile)->FileSeek(XIF_FILETOKEN(xFile)->dwClientFileID, \
	    offset) != (Int32)(offset)) {	\
	    __error__ = XIF_SEEK_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define XIF_FILESIZE(xFile, size)		\
	if (((size) = XIF_FILETOKEN(xFile)->FileSize(XIF_FILETOKEN(xFile)->\
		dwClientFileID)) < 0) {		\
	    __error__ = XIF_FILELEN_ERR;		\
	    goto ERR_CLEANUP;			\
	}

/* Malloc and free, must have a local variable *instData which is active. */
#define XIF_MALLOC(bytes)			\
	instData->malloc(bytes, 0, instData->mallocID)

#define XIF_FREE(buffer)			\
	instData->free(buffer, 0, instData->mallocID)

#define CHECK_MALLOC(call)			\
	if ((call) == NULL) {			\
	    __error__ = XIF_NOMEMORY;		\
	    goto ERR_CLEANUP;			\
	}


#define CheckHandleValidity(structPtr)		\
	if (structPtr == 0 ||			\
		((Instance *)structPtr)->handleID != (UInt32)structPtr) {  \
		__error__ = XIF_BADPARAMETER;	\
		ERR_RETURN;			\
	}
#define CheckStructureSize(structPtr)		\
	if (structPtr == NULL ||		\
		structPtr->dwSize != sizeof(*structPtr)) {	\
		__error__ = XIF_BADPARAMETER;	\
		ERR_RETURN;			\
	}

#define SwapS(p)		wp_swaps((Int8 *)p)
#define SwapL(p)		wp_swapl((Int8 *)p)
#define SwapD(p)		wp_swapd((Int8 *)p)

/* Conditional swap, depends on file token local variable: xFile */
#define SwapShort(xFile, p)	\
	if (XIF_FILETOKEN(xFile)->bSwapBytes) { \
		SwapS(p); \
	}
#define SwapLong(xFile, p)	\
	if (XIF_FILETOKEN(xFile)->bSwapBytes) { \
		SwapL(p); \
	}
#define SwapDouble(xFile, p)	\
	if (XIF_FILETOKEN(xFile)->bSwapBytes) { \
		SwapD(p); \
	}

/*
 * TYPEDEFS
 */

/* Document token
 */
typedef struct XIF_TOKEN
{
    UInt32 dwSize;		/* Size of this structure */
    Int32 dwClientFileID;	/* Client's file identifier (fd, fp, structure, etc.) */
    XIF_READFUNC FileRead;	/* Read function, set by client */
    XIF_WRITEFUNC FileWrite;	/* Write function, set by client */
    XIF_SEEKFUNC FileSeek;	/* Seek function, set by client */
    XIF_SIZEFUNC FileSize;	/* File size function, set by client */
    UInt32 dwCurPos;		/* Current file position */
    Bool bSwapBytes;		/* File must be byte swapped for endian differences */
} XIF_TOKEN;


typedef struct XFile {
    UInt32 handleID;			/* Handle ID, consistency check */
    XIF_TOKEN fileToken;		/* File token (see xiff.prv) */
    struct Instance *instance;		/* XIFF instance */
} XFile;

typedef struct Instance {
    UInt32 handleID;			/* Handle ID, consistency check */
    UInt32 nextDocID;			/* Unique ID assigned to each doc */
    UInt32 xiffMagicNum;		/* XIFF instance identifier */
    UInt32 clientInstID;		/* Client passed instance identifier */
    XIF_MALLOCFUNC malloc;		/* Client malloc function */
    XIF_FREEFUNC free;			/* Client free function */
    void *mallocID;			/* Malloc ID sent to malloc/free */
    void *defaultMallocID;		/* Reserved default Malloc ID */
    UInt32 pIPWState;			/* Handle for IPCore IPWrap layer */
    UInt32 myIPWState;			/* Boolean for private (vs. client)
					 * IPWrap handle */
} Instance;

/*
 * PROTOTYPES
 */
ErrCode CDECL
wp_XifFileLock(XIF_TOKEN *fileToken);
ErrCode CDECL
wp_XifFileUnLock(XIF_TOKEN *fileToken);

void wp_swaps(Int8 *x);
void wp_swapl(Int8 *x);
void wp_swapd(Int8 *x);


#ifdef _MSC_VER
#pragma pack(pop, xiffile_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* XIFFILE_H_INCLUDED */

