/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef XIFIMAGE_H_INCLUDED
#define XIFIMAGE_H_INCLUDED

IP_RCSINFO(xifimage_hRCSInfo, "$RCSfile: xifimage.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:13 $ */

#ifdef __cplusplus
extern "C"
{
#endif

/* xifimage.h
 *
 * DESCRIPTION
 *	 Declarations for the XIFF Toolkit API.
 *
 */

/*
 * INCLUDES
 */
#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "pixr.h"

#include "xifpage.h"
#include "xifinfo.h"
#include "xiftiff.h"

#ifdef _MSC_VER
#pragma pack(push, xifimage_h, 1)
#endif

/*
 * CONSTANTS
 */
#define MAX_EXPECTED_TAG_COUNT	22	/* Most tags an individual image
				         * is expected to have */

/* Calling sequence status */
#define XIMG_SEQ_GETINFO	0x00000001	/* Get image info called */
#define XIMG_SEQ_READSTART	0x00000010	/* Read start called */
#define XIMG_SEQ_READSTRIP	0x00000020	/* Read strip called */
#define XIMG_SEQ_READFINISH	0x00000000	/* Read finish called */
#define XIMG_SEQ_READSTATUS	0x000000f0	/* Read status bits */
#define XIMG_SEQ_WRITESTART	0x00000100	/* Write start called */
#define XIMG_SEQ_WRITESTRIP	0x00000200	/* Write strip called */
#define XIMG_SEQ_WRITEFINISH	0x00000000	/* Write finish called */
#define XIMG_SEQ_WRITESTATUS	0x00000f00	/* Write status bits */
#define XIMG_SEQ_TOKENSTART	0x00001000	/* Image is in a tokenized state
						 */
#define XIMG_SEQ_MASKADDED	0x00010000	/* Image has had a mask added */



/*
 * MACROS
 */
#define XIF_IMGFILE(xImg)		XIF_PAGEFILE(xImg->xPage)
#define XIF_IMGINST(xImg)		XIF_IMGFILE(xImg)->instance
#define XIF_IMGFILETOKEN(xImg)		XIF_FILETOKEN(XIF_IMGFILE(xImg))

/*
 * TYPEDEFS
 */

/* A linked list of miscellaneous memory, used to store arbitrary information
 * reused between strips or pages */
typedef struct MISC_MEM {
    struct MISC_MEM *next;
    UInt32 bytes;
    void *mem;
} MISC_MEM;

typedef struct XImg {	
    UInt32 handleID;		    /* Handle ID, consistency check */
    struct XImg *prev;		    /* Previous (or last) image of this type
				     * in linked list */
    struct XImg *next;		    /* Next image of this type in linked list
				     */
    XPage *xPage;		    /* Page structure of parent */
    UInt32 seqStatus;		    /* Calling sequence status */
    Int32 ifdOffset;		    /* Offset of IFD */
    XIF_IMGINFO imgInfo;	    /* Image information */
    UInt32 reservedFieldCount;	    /* Number of fields that can be written
				     * at IFD location */
    UInt32 imgFlags;		    /* As described during XIF_Image(Read/
				     * Write)Start */
    UInt32 JPEGquality;		    /* JPEG quality, if JPEG compressing */
    struct XImg *xMaskImg;	    /* Image handle to next (only)
				     * transparency mask image */
    struct XImg *lastSubImg;	    /* Last subimage attached in SubIFDs list
				     */
    /* Field Information */
    TiffFields fields;		    /* TIFF fields */
    Int32 offsetsLocation;	    /* Strip offsets array location */
    Int32 byteCountsLocation;	    /* Strip byte counts array location */
    Int32 interchangeLocation;	    /* JPEG interchange field location */
    Int32 interchangeLengthLocation;/* Interchange length field location */
    Int32 subIFDsTagOffset;	    /* SubIFDs field value offset */
    Int32 annotationOffsetsCountOffset; /* AnnotationOffsets field count offset */
    Int32 annotationOffsetsDataOffset;/* AnnotationOffsets data array offset */
    Int32 maskIFDsTagOffset;	    /* MaskIFDs field value offset */
    Int32 imageLayerLocation;	    /* ImageLayer field value location */
    Int32 ILContentTagOffset;	    /* ImageLayerContent field value offset */
    Int32 ILContentDataOffset;	    /* ImageLayerContent data offset */
    /* Intermediate buffering */
    MISC_MEM *tmpCmpData;	    /* Temporary compression info */
    UInt32 minBytesPerRow;	    /* Physical bytes required per row */
    UInt8 *cmpBuffer;		    /* Intermediate compressed buffer */
    UInt32 cmpBufsize;		    /* Compressed buffer capacity */
    UInt32 cmpBytes;		    /* Number of compressed bytes in the
				     * buffer */
    PIXR *tmpPixr;		    /* Intermediate PIXR buffer (if needed) */
    UInt32 tmpPixrRow;		    /* Image row offset of image in tmpPixr */
    UInt32 tmpPixrRows;		    /* Number of rows currently in tmpPixr */
    UInt32 posSize;		    /* Position block size of Tundra data */
    UInt32 *dictList;		    /* List of token dictionaries used in
				     * by this image */
    UInt32 dictCount;		    /* Number of dictionaries required by
				     * this image */
    Int32 *dictSizes;		    /* Dictionary sizes (passed back to client,
				     * but preserved here for memory mgmnt) */
} XImg;


/*
 * PROTOTYPES
 */

#ifdef _MSC_VER
#pragma pack(pop, xifimage_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* XIFIMAGE_H_INCLUDED */

