/*
 * Trade secret of ScanSoft, Inc.
 * Copyright (c) 1991-1998, ScanSoft, Inc.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef XIFPAGE_H_INCLUDED
#define XIFPAGE_H_INCLUDED

IP_RCSINFO(xifpage_hRCSInfo, "$RCSfile: xifpage.h,v $; $Revision: 1.1.1.1 $");
/* $Date: 1999/09/20 18:19:13 $ */

#ifdef __cplusplus
extern "C"
{
#endif

/* xifpage.h
 *
 * DESCRIPTION
 *	 Declarations for XIFF page level structures.
 *
 */

/*
 * INCLUDES
 */
#include "xifdoc.h"
#include "xifinfo.h"

#ifdef _MSC_VER
#pragma pack(push, xifpage_h, 1)
#endif

/*
 * CONSTANTS
 */

/* Calling sequence status */
#define XPAGE_SEQ_GETINFO	0x0001	/* Get image info called */
#define XPAGE_SEQ_READSTART	0x0010	/* Read start called */
#define XPAGE_SEQ_READFINISH	0x0000	/* Read finish called */
#define XPAGE_SEQ_READSTATUS	0x00f0	/* Read status bits */
#define XPAGE_SEQ_WRITESTART	0x0100	/* Write start called */
#define XPAGE_SEQ_WRITEPRIMARY	0x0200	/* Primary image added */
#define XPAGE_SEQ_WRITEPENDING	0x0400	/* Write finish called, waiting on
					 * tokenization */
#define XPAGE_SEQ_WRITEFINISH	0x0000	/* Page is completely written */
#define XPAGE_SEQ_WRITESTATUS	0x0f00	/* Write status bits */
#define XPAGE_SEQ_TOKENSTART	0x1000	/* Reference page set */
#define XPAGE_SEQ_TKCOLORADDED	0x2000	/* Color for tokenization added */
#define XPAGE_SEQ_TOKENDICTSAVED 0x4000	/* Token dictionary saved */

/*
 * MACROS
 */
#define XIF_PAGEFILE(xPage)		XIF_DOCFILE(xPage->xDoc)
#define XIF_PAGEINST(xPage)		XIF_PAGEFILE(xPage)->instance
#define XIF_PAGEFILETOKEN(xImg)		XIF_FILETOKEN(XIF_PAGEFILE(xPage))

/*
 * TYPEDEFS
 */

/* Annotation list table header */
typedef struct AnnotTableHeader {
    UInt16 annotCount;		    /* Number of annotations in page (total) */
    UInt16 entriesPerTable;	    /* Number of annotation table entries per
				     * table */
    UInt16 bytesPerEntry;	    /* Number of bytes per table entries */
    UInt16 lastOrdinal;		    /* Ordinal (1, 2, ...) of last table */
    UInt32 firstTableOffset;	    /* File offset of first table */
} AnnotTableHeader;

/* Annotation table entry (describes an individual annotation) */
typedef struct AnnotTableEntry {
    UInt16 id;			    /* Unique ID of this annotation (offset) */
    UInt16 type;		    /* Annotation type (see xiff.h) */
    UInt32 xOffset;		    /* X offset in page coordinates */
    UInt32 yOffset;		    /* Y offset in page coordinates */
    UInt32 width;		    /* Width in page coordinates */
    UInt32 height;		    /* Height in page coordinates */
    UInt32 flags;		    /* Annotation flags (see ANNOT_xxx above)
				     */
    UInt32 bytes;		    /* Number of bytes in the data */
    Int32 offset;		    /* File offset of the annotation data */
    UInt8 *data;		    /* Annotation data */
} AnnotTableEntry;

/* Individual annotation table, linked if more than one */
typedef struct AnnotTable {
    struct AnnotTable *next;	    /* Link to next table (cached) */
    UInt16 ordinal;		    /* Ordinal (1, 2, ...) of this table */
    UInt16 validAnnotations;	    /* Number of valid annotations in table */
    AnnotTableEntry *entries;	    /* Allocated entry list */
    Int32 nextTableOffset;	    /* File offset of next annotation table */
} AnnotTable;

/* Page handle
 */
typedef struct XPage {
    UInt32 handleID;		    /* Handle ID, consistency check */
    struct XPage *prev;		    /* Previous page in linked list */
    struct XPage *next;		    /* Next page in linked list */
    XDoc *xDoc;			    /* Document structure */
    UInt32 seqStatus;		    /* Calling sequence status */
    XIF_PAGEINFO pageInfo;	    /* Page information */
    UInt32 *extImageCounts;	    /* Image counts exposed to client (avoids
				     * true count getting corrupted by client)
				     */
    struct XImg *xImgs;		    /* Image list (linked) */
    Int32 annotTableHeaderOffset;   /* Offset to annotation table header
				     * (identified by the AnnotationOffsets
				     * field) */
    struct AnnotTableHeader annotTableHeader;/* Annotation list table header */
    struct AnnotTable *firstAnnotTable;	/* Annotation list table for first */
    struct XImg *fgImg;		    /* Foreground image when attaching color
				     * to tokenization */
    UInt32 tokenPageID;		    /* Page ID assigned during tokenization */
    UInt32 tokenDictID;		    /* Symbol library dictionary ID of page
				     * specific symbol dictionary */
} XPage;



/*
 * PROTOTYPES
 */

#ifdef _MSC_VER
#pragma pack(pop, xifpage_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* XIFPAGE_H_INCLUDED */

