/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * aie.c
 *
 *	Port of Automated Image Enhancement
 *
 * Functions defined:
 *
 *	w_aieAnalyzeRGB ()    -	Analyze an image for the purposes
 *				of enhancing it
 *	w_aieProcessRGB ()    -	Enhance an image given the results of aieAnalyze
 *	w_aieAnalyzeGray ()   -	Analyze an image for the purposes of
 *				enhancing it
 *	w_aieProcessGray ()   -	Enhance an image given the results of aieAnalyze
 *	w_aieFinish ()	      -	IPW_FREE data structure(s) from analyzer
 *	w_aieSuggestGamma ()  -	Stand-alone module to analyze gamma
 *
 *  The reason for separate analyze and process functions is to allow one
 *  to subsample the image passed to aieAnalyze. According to the AIE people
 *  at WRC, the analysis works on images subsampled as small as 256^2,
 *  but was tested mainly on images 512^2.
 */

#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "frames.pub"
#include "geomadt.h"
#include "pixr.h"
#include "aiedefs.pub"

/* PROTOTYPES */
#include "props.pub"
#include "props.prv"
#include "color.pub"
#include "aie.pub"
#include "shrip.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: aie.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:13 $");
/*
 * aieSharpTab
 *
 *	The standard sharpening table
 */
#if 0
Float64 aieSharpTab[] =
{
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, -1.000, 
    -1.000, -1.000, -1.000, -1.000,  3.000,  2.983,  2.967,  2.950,  
     2.933,  2.917,  2.900,  2.883,  2.867,  2.850,  2.833,  2.817,  
     2.800,  2.783,  2.767,  2.750,  2.733,  2.717,  2.700,  2.683,  
     2.667,  2.650,  2.633,  2.617,  2.600,  2.583,  2.567,  2.550,  
     2.533,  2.517,  2.500,  2.483,  2.467,  2.450,  2.433,  2.417,  
     2.400,  2.383,  2.367,  2.350,  2.333,  2.317,  2.300,  2.283,  
     2.267,  2.250,  2.233,  2.217,  2.200,  2.183,  2.167,  2.150,  
     2.133,  2.117,  2.100,  2.083,  2.067,  2.050,  2.033,  2.017,  
     2.000,  1.983,  1.967,  1.950,  1.933,  1.917,  1.900,  1.883,  
     1.867,  1.850,  1.833,  1.817,  1.800,  1.783,  1.767,  1.750,  
     1.733,  1.717,  1.700,  1.683,  1.667,  1.650,  1.633,  1.617,  
     1.600,  1.583,  1.567,  1.550,  1.533,  1.517,  1.500,  1.483,  
     1.467,  1.450,  1.433,  1.417,  1.400,  1.383,  1.367,  1.350,  
     1.333,  1.317,  1.300,  1.283,  1.267,  1.250,  1.233,  1.217,  
     1.200,  1.183,  1.167,  1.150,  1.133,  1.117,  1.100,  1.083,  
     1.067,  1.050,  1.033,  1.017,  1.000,  1.000,  1.000,  1.000,  
     1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  
     1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  
     1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  
     1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000,  1.000
};
#endif

/*
 * aieStandardDefaults 
 *
 *	This structure contains a set of standard default values for the AIE analysis 
 *	module.
 */

aAieDefaults aieStandardDefaults =
{
    0,		    /* Black */
    255,	    /* White */
    255,	    /* Table length */

    /* The scale defaults */
    {
	5, 5,	    /* top/left frame percentages */
	90, 90	    /* width/height image percentages */
    },

    /* The exposure defaults */
    {
	25, 230,    /* dark90: acceptable dark/bright range for 90% */
	25, 230,    /* dark100: acceptable dark/bright range for 100% */
	0.5, 2.0,   /* gamma90: acceptable range for 90% gamma value */
	0.5, 2.0,   /* gamma100: acceptable range for 100% gamma value */
	0.5	    /* gammaConfidence: confidence range for gamma */
    },

    /* The color balance defaults */
    {
	0.50,	    /* maxWhiteOff: maximum % that white can be corrected */
	0.50,	    /* maxBlackOff: not actually used */
	0.50,	    /* maxGrayOff: maximum % that gray can be corrected */
    },

    /* The contrast enhancement defaults */
    {
	4, 4,	    /* selectX, Y: Number of local regions (x,y) to analyze */
	10,	    /* enhanceAverage: average histogram points with window this wide */
	1,	    /* histogramMinValid: min # local histos valid to use */
	2.0,	    /* histogramVariance: variance fudge factor? */
	0.2,	    /* autoPower: standard power function value */
	50.0,	    /* maxGlobalVariance: max. variance, above which we do nothing */
	100.0,	    /* maxLocal1Variance: variance below which we use power1 */
	0.2,	    /* autoPower1: power to use if local variances < above */
	200.0,	    /* maxLocal2Variance: variance below which we use power2 */
	0.1,	    /* autoPower2: power to use if local variances < above */
	0.0,	    /* autoPower3: default power if all else fails */
	1.000,	    /* factor3: another fudge factor? */
    },

    /* The sharpness defaults */
    {
	1.000,	    /* minEdge: minimum edge sharpening value */
	3.0,	    /* maxEdge: maximum edge sharpening value */
	1.000,	    /* multiplier: fudge factor to multiply by */
	0.0,	    /* offset: fudge bias to add in */
	256,	    /* levels: number of levels in histogram */
	/* The sharpening table is defined above */
	NULL		/* was: &aieSharpTab[0] */
    }
};

/*
 * This structure is used to return a set of values back as a void 
 * so that they are hidden, but the processing module can get at them.
 */

typedef struct
{
    aTable  *red, *green, *blue, *gray;
    Float64 sharpness;
} aAnalysisResults;


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Analyze the appearance of a color image.
 *
 * LongDescription:::
 *  The folks at Webster Research Center have come up with lots of heuristics
 *  to improve the appearance of color images.  They perk up the brightness,
 *  contrast, color balance and anything else that seems interesting.
 *  This routine's job is to analyze the color image.  It puts the 
 *  results in a private structure that will be used by the code 
 *  that processes the image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pDefaults::      Pointer to structure containing default constants
 *		     used to tune the analysis.  Default values are contained
 *		     in the structure aieStandardDefaults.
 *  prS::            Pointer to the image to be analyzed.
 *
 * Output:::
 *  ppAnalysis::     Pointer to Pointer to the private analysis structure.
 *		     *ppAnalysis must be NULL.  The structure is always
 *		     created here.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrPtrNotNull
 *  if *ppAnalysis is not NULL on entry.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeRGB (
    IPWState           *pIPWState,
    aAieDefaults       *pDefaults,
    PIXR               *prS,
    void              **ppAnalysis)
{
    PROLOGUE;
    aAnalysisResults  *pRes = NULL;
    PIXR              *pLuminance = NULL, *subR, *subG, *subB;
    Int32              w, h, bpl;
    UInt8             *pR, *pG, *pB, *pL;

    ERR_GEN(pDefaults == NULL || prS == NULL || ppAnalysis == NULL,
	    cErrNullPtr);
    ERR_GEN(*ppAnalysis != NULL, cErrPtrNotNull);

	/* Remember that the depth is the depth of ONE channel */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);

    pRes = (aAnalysisResults *) IPW_MALLOC (sizeof (aAnalysisResults), 
					    cAllocatorData);
    ERR_GEN(pRes == NULL, cErrNomem);

    pRes->red = pRes->green = pRes->blue = pRes->gray = NULL;

    /* Make a luminance channel, use DITC's formula */
    ERR_PASS( w_RGBToY2 (pIPWState, prS, &pLuminance));

    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, pLuminance, &pL));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &w));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &h));
    

    /* Do the analysis */
    ERR_PASS_CLEANUP( i_aieAnalyzeRGB (pIPWState, pDefaults, pR, pG, pB,
				       pL, bpl, w, h,
				       &(pRes->red),
				       &(pRes->green),
				       &(pRes->blue),
				       &(pRes->sharpness)));

    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &pLuminance));

    *ppAnalysis = (void *)pRes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pRes != NULL)
    {
	if (pRes->red)
	    i_aieDestroyTable (pIPWState, &pRes->red);

	if (pRes->green)
	    i_aieDestroyTable (pIPWState, &pRes->green);

	if (pRes->blue)
	    i_aieDestroyTable (pIPWState, &pRes->blue);

	if (pRes->gray)
	    i_aieDestroyTable (pIPWState, &pRes->gray);

	IPW_FREE (pRes, cAllocatorData);
    }
    w_destroyPixr(pIPWState, &pLuminance);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generate state block used to analyze color image in strips.
 *
 * LongDescription:::
 *  w_aieAnalyzeRGB analyzes an entire color image.  The results are used
 *  to improve contrast, brightness, etc.  However, color images are
 *  very large.  It is useful to be able to analyze a color image
 *  a strip at a time and accumulate the results.  This procedure allocates
 *  a state structure that will be used by w_aieAnalyzeRGBProcess to
 *  analyze a block of image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pDefaults::      Pointer to structure containing default constants
 *		     used to tune the analysis.  Default values are contained
 *		     in the structure aieStandardDefaults.
 *  w::              width of image to be analyzed.
 *  h::              total height of image to be analyzed.
 *  bpl::            bytes/line of each plane of the image to be analyzed.
 *
 * Output:::
 *  ppThing::        Pointer to pointer to private state block that will
 *		     be used by w_aieAnalyzeRGBProcess.  *ppThing must be NULL.
 *		     The private state block is always created here.
 *
 * Returns:::
 *  In addition to the normal checks, this routine returns cErrPtrNotNull
 *  if *ppThing is not NULL on entry.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeRGBInit (
    IPWState           *pIPWState,
    aAieDefaults       *pDefaults,
    Int32               w,
    Int32               h,
    Int32               bpl,
    void              **ppThing)
{
    PROLOGUE;
    aIntermediates *pR;
    
    ERR_GEN(pDefaults == NULL || ppThing == NULL, cErrNullPtr);
    ERR_GEN(*ppThing != NULL, cErrPtrNotNull);

    ERR_PASS( i_aieAnalyzeRGB_init (pIPWState, pDefaults, w, h, bpl, &pR));
    *ppThing = (void *)pR;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Analyze the appearance of one strip of a color image.
 *
 * LongDescription:::
 *  The folks at Webster Research Center have come up with lots of heuristics
 *  to improve the appearance of color images.  They perk up the brightness,
 *  contrast, color balance and anything else that seems interesting.
 *  This routine is used to analyze the color image.  It puts the 
 *  results in a private structure that will be used by the code 
 *  that processes the image.
 *
 *  However, color images are huge, so it's nice to be able to analyze
 *  them a strip at a time.  This routine does that, accumulating the
 *  results in a private state structure.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pThing::         Pointer to private strip-based state block that
 *		     accumulates the partial results of the analysis.
 *  prS::            Pointer to band of image to be analyzed.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeRGBProcess (
    IPWState           *pIPWState,
    void               *pThing,
    PIXR               *prS)
{
    PROLOGUE;
    aIntermediates *pData;
    PIXR           *pLuminance = NULL, *subR, *subG, *subB;
    UInt8          *pR, *pG, *pB, *pL;
    Int32           h, w, bpl;

    ERR_GEN(prS == NULL || pThing == NULL, cErrNullPtr);

	/* Remember that the depth is the depth of ONE channel */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);

    pData = (aIntermediates *) pThing;

    /* Compute luminance */
    ERR_PASS( w_RGBToY2 (pIPWState, prS, &pLuminance));

    ERR_PASS_CLEANUP( w_pixrGetWidth (pIPWState, prS, &w));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prS, &bpl));
    pData->w = w;
    pData->bpl = bpl;

    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subR));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subG));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subR, &pR));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subG, &pG));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, pLuminance, &pL));
    ERR_PASS_CLEANUP( w_pixrGetHeight (pIPWState, prS, &h));
    pData->h = h;

    /* Pass the band to the analyzers */
    ERR_PASS_CLEANUP( i_aieAnalyzeRGB_data (pIPWState, pData,
					    pR, pG, pB, pL, h));

    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &pLuminance));
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &pLuminance);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finish strip-based analysis of a color image.
 *
 * LongDescription:::
 *  w_aieAnalyzeRGBProcess is used to analyze a series of strips of a color
 *  image.  This routine flushes the results out of the analyzers, transferring
 *  them from the private strip-based state block to the standard AIE
 *  state block.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  ppThing::        Pointer to pointer to private strip-based state block.
 *		     This block was created by w_aieAnalyzeRGBInit.
 *		     **ppThing will be freed and *ppThing will be set to
 *		     NULL on completion.
 *
 * Output:::
 *  ppAnalysis::     Pointer to pointer to AIE's private state block that
 *		     holds the results of the analysis.  On entry, *ppAnalysis
 *		     must always be NULL.
 *
 * Returns:::
 *  In addition to the normal checks, this routine returns cErrPtrNotNull
 *  if *ppAnalysis is not NULL on entry.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeRGBFinish (
    IPWState           *pIPWState,
    void              **ppThing,
    void              **ppAnalysis)
{
    PROLOGUE;
    aAnalysisResults    *pRes = NULL;

    ERR_GEN(ppAnalysis == NULL || ppThing == NULL, cErrNullPtr);
    ERR_GEN(*ppAnalysis != NULL, cErrPtrNotNull);

    pRes = (aAnalysisResults *) IPW_MALLOC (sizeof (aAnalysisResults), 
					    cAllocatorData);
    ERR_GEN(pRes == NULL, cErrNomem);

    pRes->red = pRes->green = pRes->blue = pRes->gray = NULL;

    ERR_PASS_CLEANUP( i_aieAnalyzeRGB_finish (pIPWState,
					      (aIntermediates **) ppThing,
					      &(pRes->red),
					      &(pRes->green),
					      &(pRes->blue),
					      &(pRes->sharpness)));

    *ppAnalysis = pRes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pRes != NULL)
    {
	if (pRes->red)
	    i_aieDestroyTable (pIPWState, &pRes->red);

	if (pRes->green)
	    i_aieDestroyTable (pIPWState, &pRes->green);

	if (pRes->blue)
	    i_aieDestroyTable (pIPWState, &pRes->blue);

	if (pRes->gray)
	    i_aieDestroyTable (pIPWState, &pRes->gray);

	IPW_FREE (pRes, cAllocatorData);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Analyze the appearance of a gray image.
 *
 * LongDescription:::
 *  The folks at Webster Research Center have come up with lots of heuristics
 *  to improve the appearance of gray images.  They perk up the brightness,
 *  contrast and anything else that seems interesting.
 *  This routine's job is to analyze the gray image.  It puts the 
 *  results in a private structure that will be used by the code 
 *  that processes the image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pDefaults::      Pointer to structure containing default constants
 *		     used to tune the analysis.  Default values are contained
 *		     in the structure aieStandardDefaults.
 *  prS::            Pointer to the image to be analyzed.
 *
 * Output:::
 *  ppAnalysis::     Pointer to Pointer to the private analysis structure.
 *		     *ppAnalysis must be NULL.  The structure is always
 *		     created here.
 *
 * Returns:::
 *  In addition to the normal stuff, this routine returns cErrPtrNotNull
 *  if *ppAnalysis is not NULL on entry.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeGray (
    IPWState           *pIPWState,
    aAieDefaults       *pDefaults,
    PIXR               *prS,
    void              **ppAnalysis)
{
    PROLOGUE;
    aAnalysisResults     *pRes = NULL;
    Int32                 bpl, w, h;
    UInt8                *pS;

    ERR_GEN(pDefaults == NULL || prS == NULL || ppAnalysis == NULL,
	    cErrNullPtr);
    ERR_GEN(*ppAnalysis != NULL, cErrPtrNotNull);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    pRes = (aAnalysisResults *) IPW_MALLOC (sizeof (aAnalysisResults), 
					    cAllocatorData);
    ERR_GEN(pRes == NULL, cErrNomem);

    pRes->red = pRes->green = pRes->blue = pRes->gray = NULL;

    ERR_PASS_CLEANUP( w_pixrGetImageAll(pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &w));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &h));

    /* Do the analysis */
    ERR_PASS_CLEANUP( i_aieAnalyzeGray (pIPWState, pDefaults, pS, bpl, w, h,
				        &(pRes->gray),
				        &(pRes->sharpness)));

    *ppAnalysis = pRes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pRes != NULL)
    {
	if (pRes->red)
	    i_aieDestroyTable (pIPWState, &pRes->red);

	if (pRes->green)
	    i_aieDestroyTable (pIPWState, &pRes->green);

	if (pRes->blue)
	    i_aieDestroyTable (pIPWState, &pRes->blue);

	if (pRes->gray)
	    i_aieDestroyTable (pIPWState, &pRes->gray);

	IPW_FREE (pRes, cAllocatorData);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generate state block used to analyze gray image in strips.
 *
 * LongDescription:::
 *  w_aieAnalyzeRGB analyzes an entire gray image.  The results are used
 *  to improve contrast, brightness, etc.  However, gray images are
 *  very large.  It is useful to be able to analyze a gray image
 *  a strip at a time and accumulate the results.  This procedure allocates
 *  a state structure that will be used by w_aieAnalyzeGrayProcess to
 *  analyze a block of image.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pDefaults::      Pointer to structure containing default constants
 *		     used to tune the analysis.  Default values are contained
 *		     in the structure aieStandardDefaults.
 *  w::              width of image to be analyzed.
 *  h::              total height of image to be analyzed.
 *  bpl::            bytes/line of each plane of the image to be analyzed.
 *
 * Output:::
 *  ppThing::        Pointer to pointer to private state block that will
 *		     be used by w_aieAnalyzeGrayProcess.  *ppThing must be NULL.
 *		     The private state block is always created here.
 *
 * Returns:::
 *  In addition to the normal checks, this routine returns cErrPtrNotNull
 *  if *ppThing is not NULL on entry.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeGrayInit (
    IPWState           *pIPWState,
    aAieDefaults       *pDefaults,
    Int32               w,
    Int32               h,
    Int32               bpl,
    void              **ppThing)
{
    PROLOGUE;
    aIntermediates *pR;

    ERR_GEN(pDefaults == NULL || ppThing == NULL,
	    cErrNullPtr);
    ERR_GEN(*ppThing != NULL, cErrPtrNotNull);

    ERR_PASS( i_aieAnalyzeGray_init (pIPWState, pDefaults, w, h, bpl, &pR));
    *ppThing = (void *)pR;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Analyze the appearance of one strip of a gray image.
 *
 * LongDescription:::
 *  The folks at Webster Research Center have come up with lots of heuristics
 *  to improve the appearance of gray images.  They perk up the brightness,
 *  contrast and anything else that seems interesting.
 *  This routine is used to analyze the gray image.  It puts the 
 *  results in a private structure that will be used by the code 
 *  that processes the image.
 *
 *  However, gray images are huge, so it's nice to be able to analyze
 *  them a strip at a time.  This routine does that, accumulating the
 *  results in a private state structure.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pThing::         Pointer to private strip-based state block that
 *		     accumulates the partial results of the analysis.
 *  prS::            Pointer to band of image to be analyzed.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeGrayProcess (
    IPWState           *pIPWState,
    void               *pThing,
    PIXR               *prS)
{
    PROLOGUE;
    aIntermediates *pData;
    Int32           h, w, bpl;
    UInt8          *pS;

    ERR_GEN(prS == NULL || pThing == NULL, cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    pData = (aIntermediates *) pThing;
    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &w));
    ERR_PASS( w_pixrGetBpl (pIPWState, prS, &bpl));
    pData->w = w;
    pData->bpl = bpl;

    ERR_PASS( w_pixrGetImage (pIPWState, prS, &pS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &h));
    pData->h = h;

    /* Pass the band to the analyzers */
    ERR_PASS( i_aieAnalyzeGray_data (pIPWState, pData, pS, h));
				    
    NORMAL_RETURN;
}
		

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finish strip-based analysis of a gray image.
 *
 * LongDescription:::
 *  w_aieAnalyzeGrayProcess is used to analyze a series of strips of a gray
 *  image.  This routine flushes the results out of the analyzers, transferring
 *  them from the private strip-based state block to the standard AIE
 *  state block.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  ppThing::        Pointer to pointer to private strip-based state block.
 *		     This block was created by w_aieAnalyzeGrayInit.
 *		     **ppThing will be freed and *ppThing will be set to
 *		     NULL on completion.
 *
 * Output:::
 *  ppAnalysis::     Pointer to pointer to AIE's private state block that
 *		     holds the results of the analysis.  On entry, *ppAnalysis
 *		     must always be NULL.
 *
 * Returns:::
 *  In addition to the normal checks, this routine returns cErrPtrNotNull
 *  if *ppAnalysis is not NULL on entry.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAnalyzeGrayFinish (
    IPWState           *pIPWState,
    void              **ppThing,
    void              **ppAnalysis)
{
    PROLOGUE;
    aAnalysisResults    *pRes = NULL;

    ERR_GEN(ppAnalysis == NULL || ppThing == NULL, cErrNullPtr);
    ERR_GEN(*ppAnalysis != NULL, cErrPtrNotNull);

    pRes = (aAnalysisResults *) IPW_MALLOC (sizeof (aAnalysisResults), 
					    cAllocatorData);
    ERR_GEN(pRes == NULL, cErrNomem);

    pRes->red = pRes->green = pRes->blue = pRes->gray = NULL;

    ERR_PASS_CLEANUP( i_aieAnalyzeGray_finish (pIPWState,
					       (aIntermediates **) ppThing,
					       &(pRes->gray),
					       &(pRes->sharpness)));

	/* BUG, BUG, BUG!!  No one really frees *ppThing!!!.  It was
	 * allocated in ipshared and freed there.  For now, we'll
	 * just set the point to NULL as a reminder that Peter
	 * should fix it sometime.
	 */
    *ppThing = NULL;

    *ppAnalysis = pRes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pRes != NULL)
    {
	if (pRes->red)
	    i_aieDestroyTable (pIPWState, &pRes->red);

	if (pRes->green)
	    i_aieDestroyTable (pIPWState, &pRes->green);

	if (pRes->blue)
	    i_aieDestroyTable (pIPWState, &pRes->blue);

	if (pRes->gray)
	    i_aieDestroyTable (pIPWState, &pRes->gray);

	IPW_FREE (pRes, cAllocatorData);
    }
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Apply results of image analysis to a color image.
 *
 * LongDescription:::
 *  Either w_aieAnalyzeRGB or the strip-based w_aieAnalyzeRGBProcess can
 *  generate tables that could be used to improve the appearance of a
 *  color image.  The specify how contrast, color balance, brightness, etc
 *  could be changed.  This routine improves the color image by applying the
 *  tables.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to color image that will be spruced up.
 *  pAnalysis::      Pointer to private AIE state block that contains
 *		     the results of the analysis.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  If no dest
 *		     image is provided, one will be created.  If one is
 *		     provided, its dimensions, resolution, depth, etc will
 *		     be checked.  As usual, if *pPrD == prS, the operation
 *		     will be done in place.
 *
 * Notes:::
 *  The source and dest images may in fact be slices of the full source
 *  and dest images.  This allows processing by strips if desired.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieProcessRGB (
    IPWState           *pIPWState,
    PIXR               *prS,
    void               *pAnalysis,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    aAnalysisResults	*pRes;
    Int32		 wS, hS, bplS, bplD;
    ScaledInt1616	 xSdpi, ySdpi;
    PIXR		*subRS, *subGS, *subBS, *subRD, *subGD, *subBD;
    UInt8		*pRS, *pGS, *pBS, *pRD, *pGD, *pBD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pAnalysis == NULL || pPrD == NULL,
	    cErrNullPtr);

	/* Remember that the depth is the depth of ONE channel */
    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels3, cFrame32);

    pRes = (aAnalysisResults *) pAnalysis;

    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrExtendedConditionalNoInit (pIPWState, wS, hS, 8,
						      xSdpi, ySdpi, cFrameBits,
						      3, cColorRGB, NULL, &prD,
						      &rightFrameColsToClear,
						      &bottomFrameRowsToClear,
						      &allocatedHere));

    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cRedPlane, &subRS));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cGreenPlane, &subGS));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prS, cBluePlane, &subBS));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cRedPlane, &subRD));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cGreenPlane, &subGD));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prD, cBluePlane, &subBD));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subRS, &pRS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subGS, &pGS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subBS, &pBS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subRD, &pRD));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subGD, &pGD));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, subBD, &pBD));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));

    /* Do the processing */
    ERR_PASS_CLEANUP( i_aieProcessRGB (pIPWState, pRS, pGS, pBS, bplS,
				       pRD, pGD, pBD, bplD,
				       wS, hS,
				       pRes->red,
				       pRes->green,
				       pRes->blue,
				       pRes->sharpness));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Apply results of image analysis to a gray image.
 *
 * LongDescription:::
 *  Either w_aieAnalyzeGray or the strip-based w_aieAnalyzeGrayProcess can
 *  generate tables that could be used to improve the appearance of a
 *  gray image.  The specify how contrast, brightness, etc
 *  could be changed.  This routine improves the gray image by applying the
 *  tables.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  prS::            Pointer to gray image that will be spruced up.
 *  pAnalysis::      Pointer to private AIE state block that contains
 *		     the results of the analysis.
 *
 * Output:::
 *  pPrD::           Pointer to pointer to destination image.  If no dest
 *		     image is provided, one will be created.  If one is
 *		     provided, its dimensions, resolution, depth, etc will
 *		     be checked.  As usual, if *pPrD == prS, the operation
 *		     will be done in place.
 *
 * Notes:::
 *  The source and dest images may in fact be slices of the full source
 *  and dest images.  This allows processing by strips if desired.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieProcessGray (
    IPWState           *pIPWState,
    void               *pAnalysis,
    PIXR               *prS,
    PIXR              **pPrD)
{
    PROLOGUE;
    PIXR		*prD = NULL;
    aAnalysisResults	*pRes;
    ScaledInt1616	 xSdpi, ySdpi;
    Int32		 wS, hS, bplS, bplD;
    UInt8		 *pS, *pD;
    Bool		 allocatedHere = cFalse;
    Int32		 rightFrameColsToClear, bottomFrameRowsToClear;

    ERR_GEN(prS == NULL || pAnalysis == NULL || pPrD == NULL,
	    cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    pRes = (aAnalysisResults *) pAnalysis;
    ERR_PASS( w_pixrGetWidth (pIPWState, prS, &wS));
    ERR_PASS( w_pixrGetHeight (pIPWState, prS, &hS));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, prS, &xSdpi, &ySdpi));

	/* if we already have a dest pixr, make sure it's the right size.
	 * otherwise, create one.
	 */
    prD = *pPrD;
    ERR_PASS( wp_createPixrConditionalNoInit (pIPWState, wS, hS, 8,
					      xSdpi, ySdpi, cFrameBits,
					      &prD,
					      &rightFrameColsToClear,
					      &bottomFrameRowsToClear,
					      &allocatedHere));

    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prS, &bplS));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, prD, &bplD));
    ERR_PASS_CLEANUP( w_pixrFramedWidth (pIPWState, prS, &wS));
    ERR_PASS_CLEANUP( w_pixrFramedHeight(pIPWState, prS, &hS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prS, &pS));
    ERR_PASS_CLEANUP( w_pixrGetImageAll (pIPWState, prD, &pD));

    /* Do the processing */
    ERR_PASS_CLEANUP( i_aieProcessGray (pIPWState, pS, bplS,
				        pD, bplD,
				        wS, hS, pRes->gray, pRes->sharpness));

	/* If we need to clear any parts of the frame, do so */
    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	ERR_PASS_CLEANUP( wp_clearRightBottomFrames( pIPWState, prD,
						     rightFrameColsToClear,
						     bottomFrameRowsToClear));

	/* it all worked, so return a pointer to the dest pixr. */
    *pPrD = prD;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (allocatedHere)
	w_destroyPixr(pIPWState, &prD);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Suggest a gamma correction value for an image.
 *
 * LongDescription:::
 *  To make a gray or color image look better, one often stretches out
 *  the distance between dark intensity values and crams together the distance
 *  between light intensity values.  This makes it easier to see small
 *  differences in dark areas.  A curve that does this is x**(1/gamma).
 *  For example, if gamma = 2, this is the square root curve.  This routine
 *  suggests an optimal value for gamma given an image.
 *
 * Input::: 
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pDefaults::      Pointer to structure containing constants used to
 *		     tune the image analysis.  Default values are contained
 *		     in the structure aieStandardDefaults.
 *  prS::            Pointer to the image to be analyzed.
 *
 * Output:::
 *  pGamma::         Pointer to returned gamma value.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, histogram, color, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieSuggestGamma (
    IPWState           *pIPWState,
    aAieDefaults       *pDefaults,
    PIXR               *prS,
    Float64            *pGamma)
{
    PROLOGUE;
    Int32     arr100[4];
    Float64   variance;
    Int32     w, h, bpl;
    UInt8    *pS;
    Int32     channels;
    UInt32     uniqueLevels = 0;

    ERR_GEN(prS == NULL || pDefaults == NULL || pGamma == NULL,
	    cErrNullPtr);

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepth8, cChannels1, cFrame32);

    ERR_PASS( w_pixrGetBpl(pIPWState, prS, &bpl));
    ERR_PASS( w_pixrFramedWidth (pIPWState, prS, &w));
    ERR_PASS( w_pixrFramedHeight(pIPWState, prS, &h));
    ERR_PASS( w_pixrGetImageAll (pIPWState, prS, &pS));
    ERR_PASS( w_pixrGetChannelCount(pIPWState, prS, &channels));

    /* Count the number of colors or grays */
    switch (channels) {
    case 1:
	ERR_PASS( i_countGrays (pIPWState, pS, w, h, bpl, &uniqueLevels));
	break;
    case 3:
	/* Approximation: we should really convert to luminance, and
	 * based on the approach by DITC, we should use their values
	 * in the conversion, then call i_countColors() here instead. */
	ERR_PASS( i_countGrays (pIPWState, pS, w, h, bpl, &uniqueLevels));
	break;
    }


    /* Do the analysis */
    ERR_PASS( i_aieExposure (pIPWState, &(pDefaults->exposure), pS, w, h, bpl,
			     NULL, &(arr100[0]), &variance, pGamma,
			     uniqueLevels));

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deallocate all private AIE structures.
 *
 * LongDescription:::
 *  The AIE analyzers and processing functions (w_aieAnalyze*, w_aieProcess*)
 *  use a private AIE state block.  This routine should be called by
 *  clients after completing an analysis and processing task to deallocate
 *  all the pieces of the private state structure.
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *
 * Output:::
 *  ppAnalysis::     Pointer to Pointer to private AIE state block.
 *		     The stuff in **ppAnalysis will be freed and
 *		     *ppAnalysis will be set to NULL on completion.
 *
 * Notes:::
 *  Clients should use this routine instead of simply freeing *ppAnalysis
 *  since this routine knows about all the little memory blocks attached
 *  to the private state structure.  Simply freeing the state structure
 *  would result in leaking all these blocks.
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, histogram, color, gray
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  W. Fuss, R. Eschbach (WRC), P. Macleod
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieFinish (
    IPWState           *pIPWState,
    void              **ppAnalysis)
{
    PROLOGUE;
    aAnalysisResults *pRes;

    ERR_GEN(ppAnalysis == NULL, cErrNullPtr);

    if (*ppAnalysis != NULL) {
	pRes = (aAnalysisResults *) *ppAnalysis;

	if (pRes->red)
	{
	    ERR_PASS (i_aieDestroyTable (pIPWState, &pRes->red));
	}
	if (pRes->green)
	{
	    ERR_PASS (i_aieDestroyTable (pIPWState, &pRes->green));
	}
	if (pRes->blue)
	{
	    ERR_PASS (i_aieDestroyTable (pIPWState, &pRes->blue));
	}
	if (pRes->gray)
	{
	    ERR_PASS (i_aieDestroyTable (pIPWState, &pRes->gray));
	}
	IPW_FREE (pRes, cAllocatorData);
	*ppAnalysis = NULL;
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get sharpness value.
 *
 * LongDescription:::
 *  Analysis structure is private structure. When user want get the aie
 *  sharpness value, use this routine to get it. 
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pAnalysis::      Pointer to private analysis results
 *
 * Output:::
 *  sharpness::      Pointer to sharpness value
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieGetSharpness (
    IPWState           *pIPWState,
    void               *pAnalysis,
    Float64            *sharpness)
{
    PROLOGUE;
    aAnalysisResults  *pRes = NULL;

    ERR_GEN(pAnalysis == NULL, cErrNullPtr);

    pRes = (aAnalysisResults *) pAnalysis;

    *sharpness = pRes->sharpness;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get TRC ( mapping table ).
 *
 * LongDescription:::
 *  Analysis structure is private structure. When user want access the aie
 *  anaylysis TRCs, use this routine to access them. Be rememeber, the tables
 *  are managed inside and will be freed in w_aieFinish().
 *
 * Input:::
 *  pIPWState::      Pointer to standard IPWrap state block.
 *  pAnalysis::      Pointer to private analysis results
 *  channels::       Gray: channels=1, color: channels =3
 *
 * Output:::
 *  ppTable::        Pointer to pointer of TRC (order: RGB )
 *  pLength::        Pointer to length of TRC (order: RGB)
 *
 * FunctionCategory:::
 *  automatic enhancement
 *
 * Keywords:::
 *  automatic enhancement, pixel mapping, histogram, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_aieAccessTRC (
    IPWState           *pIPWState,
    void               *pAnalysis,
    Int32               channels,
    Int32             **ppTable,
    Int32              *pLength) 
{
    PROLOGUE;
    aAnalysisResults  *pRes = NULL;

    ERR_GEN(pAnalysis == NULL || ppTable == NULL || pLength == NULL, 
	    cErrNullPtr);
    ERR_GEN( (channels != 1) && (channels != 3), cErrChannelsNotEqual);

    pRes = (aAnalysisResults *) pAnalysis;

/* Here, we just pass out the table pointer. Suppose users just access
   them and not modify them. They will be available after analysis 
   and freed in w_aieFinish(). */
   
    if( channels == 3)
    {
	*pLength = pRes->red->length;
	*(pLength +1) = pRes->green->length;
	*(pLength +2) = pRes->blue->length;
	*ppTable = &pRes->red->table[0];
	*(ppTable+1) = &pRes->green->table[0];
	*(ppTable+2) = &pRes->blue->table[0];
    }
    else
    {
	*pLength = pRes->gray->length;
	*ppTable = &pRes->gray->table[0];
    }
 
    NORMAL_RETURN;
}

/*
 $Log: aie.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:13  jlin
 Imported sources

 Revision 1.39  1998/01/14 00:10:36  ddavies
 ddavies on Tue Jan 13 16:10:36 PST 1998

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.38  1997/12/04 00:37:03  xzhu
 xzhu on Wed Dec 3 16:36:16 PST 1997

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 Fixed problem of supporting analyze segmented image data.

 Revision 1.37  1997/12/02 00:31:18  xzhu
 xzhu on Mon Dec 1 16:28:08 PST 1997

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 Assign width in w_aieAnalyzeRGBProcess, since prSs  may be different width and heigh in different time.

 Revision 1.36  1997/11/13 22:52:43  aber
 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 Add changes in thresholds and code to support new DITC AIE algorithm(s).
 See changes in i_aieana.c in ipshared for a list.

 Revision 1.35  1997/10/07 20:47:21  xzhu
 xzhu on Tue Oct 7 13:46:22 PDT 1997

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 Created two new finctions w_aiegetSharpness and w_aieAccessTRC based on Andrew's suggestion.

 Revision 1.34  1997/09/24 20:42:31  xzhu
 xzhu on Wed Sep 24 13:38:44 PDT 1997

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 Let strip by strip analysis also work in segment by segment ( means: input sub-images can be different width, bpl and height).

 Revision 1.33  1997/04/15 20:06:20  ddavies
 ddavies on Tue Apr 15 13:05:21 PDT 1997

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 Substitute NoInit version of pixr creation routines where possible.

 Revision 1.32  1996/03/19 00:10:53  macleod
 macleod on Mon Mar 18 16:10:30 PST 1996

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 Call i_aieDestroyTable rather than IPW_FREE to destroy TRC tables.
 Doing the latter was an evil thing.

 Revision 1.31  1996/01/04 20:20:53  ejaquith
 ejaquith on Thu Jan 4 12:20:52 PST 1996

 aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.30  1995/12/13  04:04:33  ejaquith
 * ejaquith on Tue Dec 12 20:04:33 PST 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/11/22  17:32:52  macleod
 * macleod on Wed Nov 22 09:30:54 PST 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * w_aieAnalyzeRGBProcess was passing the framed height of the strip
 * rather than the unframed. Duh. How it ever worked is beyond me.
 * Amazingly, it passed the regression test if the aie test was run
 * alone, but not when run with everything else. (And, of course,
 * purify screamed)
 *
 * Revision 1.28  1995/11/22  01:53:42  macleod
 * macleod on Tue Nov 21 17:53:01 PST 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Changed calls to i_aieAnalyzeRGB_finish and i_aieAnalyzeGrayFinish.
 * They should now free their internal data structures.
 *
 * Revision 1.27  1995/11/13  20:09:12  ejaquith
 * ejaquith on Mon Nov 13 12:09:11 PST 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.26  1995/10/13  21:58:18  ejaquith
 * ejaquith on Fri Oct 13 14:58:17 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.25  1995/10/05  04:30:33  ddavies
 * ddavies on Wed Oct 4 21:30:33 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Fill in autodoc comment blocks.
 *
 * Revision 1.24  1995/09/21  22:51:09  ejaquith
 * ejaquith on Thu Sep 21 15:51:08 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.23  1995/09/20  21:30:33  ejaquith
 * ejaquith on Wed Sep 20 14:29:57 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Made cleanup on error handling actually cleanup (had leaks on errors).
 *
 * Revision 1.22  1995/09/15  18:01:55  ddavies
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Fix the way dest images are conditionally generated.  Re-indent the local
 * variable declarations.
 *
 * Revision 1.21  1995/09/11  22:12:27  ddavies
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Add error checking for bad PIXRs, invalid Rects and null pointers.
 *
 * Revision 1.20  1995/09/10  00:34:27  ddavies
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Reorder parameters, add error checking of input pointers and input pixrs,
 * combine the Pixr/Pr routines, make creation of a destination PIXR optional
 * in most cases (created if client doesn't supply one).
 *
 * Revision 1.19  1995/09/07  01:29:19  ejaquith
 * ejaquith on Wed Sep 6 18:29:18 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.18  1995/09/06  20:16:54  ejaquith
 * ejaquith on Wed Sep 6 13:16:53 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.17  1995/09/01  01:14:57  ddavies
 * ddavies on Thu Aug 31 18:14:11 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Put minimum number of include files back in.
 *
 * Revision 1.16  1995/08/31  21:10:46  ejaquith
 * ejaquith on Thu Aug 31 14:10:45 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.15  1995/08/25  01:34:48  ejaquith
 * ejaquith on Thu Aug 24 18:34:04 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Fixed warnings
 *
 * Revision 1.14  1995/08/17  22:03:45  ejaquith
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.13  1995/07/29  04:34:26  ddavies
 * ddavies on Fri Jul 28 21:33:46 PDT 1995
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Checkin preliminary changes to new style - mostly making decls agree with protos.
 *
 * Revision 1.12  1994/10/07  19:19:14  lperry
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Remove use of xpros.h files. Substitute .pub headers.
 *
 * Revision 1.11  1994/10/05  22:22:19  macleod
 * macleod on Wed Oct 5 15:20:56 PDT 1994
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Added support for banded aie, analyzer phase. Note that the processing can
 * already be done banded. The only change needed was that the process functions
 * will only allocate the result PIXR (passed in as pointer-to-pointer) if the
 * pointer is NULL, otherwise the process functions just use the PIXR as passed.
 *
 * Revision 1.10  1994/07/12  00:46:55  macleod
 * macleod on Mon Jul 11 17:46:45 PDT 1994
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Added aieFinish(), which is used to IPW_FREE whatever an aie analyzer allocated
 *
 * Revision 1.9  1994/06/20  22:54:46  macleod
 * macleod on Mon Jun 20 15:54:26 PDT 1994
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * aieAnalyzeRGB was making a temporary Luminance PIXR and not IPW_FREEing it.
 *
 * Revision 1.8  1994/06/08  04:06:27  ddavies
 * ddavies on Tue Jun 7 21:05:44 PDT 1994
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Remove include of obsolete ipdef.h.
 *
 * Revision 1.7  1994/04/12  01:12:33  macleod
 * macleod on Mon Apr 11 18:12:26 PDT 1994
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.6  1994/04/05  00:39:14  macleod
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Color AIE routines now take a ColorPixr instead of three PIXRs.
 *
 * Revision 1.5  1994/03/23  02:56:58  macleod
 * macleod on Tue Mar 22 18:55:00 PST 1994
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Something was returning NULL for an error code when it should have
 * returned non-zero.
 *
 * Revision 1.4  1993/12/21  23:21:22  macleod
 * macleod on Tue Dec 21 15:20:48 PST 1993
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * added aieSuggestGamma to suggest a gamma for an image
 *
 * Revision 1.3  1993/12/02  23:19:17  ddavies
 * ddavies on Thu Dec 2 15:18:45 PST 1993
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:09:37  ddavies
 * ddavies on Fri Nov 5 19:09:25 PST 1993
 *
 * aie.c -> /product/ipcore/ipwrap/src/RCS/aie.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

