/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * appsutil.c
 *	      
 *   Applications utilities. Why are these here? Because linkers are
 *   usually stupid, and want libraries in order. The routines in here
 *   are here because they depend on things that are higher in the food
 *   chain than one might think.
 *
 * Exports:
 *   w_copyGrayMasked
 */

#include <stdio.h>
#include <math.h>
#include "ipwcore.prv"
#include "types.pub"
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "frames.pub"
#include "iaerror.pub"
#include "appsinc.h"
#include "logicops.h"
#include "imagerep.h"
#include "io.pub"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "shrscal.pub"
#include "prxform.pub"
#include "shrip.pub"
#include "reduce.pub"
#include "rectops.pub"
#include "cut.pub"
#include "coords.pub"
#include "clean.pub"
#include "iw_scale.pub"
#include "resconv.pub"
#include "grbinop.pub"
#include "logicops.pub"
#include "intops.pub"
#include "appsutil.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: appsutil.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:13 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Copy one gray pixr onto another, through a mask
 *
 * LongDescription:::
 *    This routine is used to copy one gray PIXR into another, using
 *    a mask to determine which bits to copy. ACHTUNG! This uses
 *    opGrayBinPrIP to pick out the bits and to cut a hole in the
 *    dest PIXR, and it is therefore somewhat sub-optimal! If you actually
 *    need to use this in production code, make me re-write it! (It's
 *    not really that hard..)
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  xS::             Offset into source image
 *  yS::             
 *  prSrc::          Source image
 *  xM::             Offset into mask image
 *  yM::             
 *  prMask::         Mask image 
 *  xD::             Rectangle to copy into in destination image
 *  yD::             
 *  wD::             
 *  hD::             
 *
 * Output:::
 *  prDest::         Destination image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  The reason this is here is due to stupid libarary dependencies.
 *  This function should be re-written so that it doesn't depend on
 *  the iw_scale, etc. functions.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_copyGrayMasked (
    IPWState           *pIPWState,
    Int32               xS,
    Int32               yS,
    PIXR               *prSrc,
    Int32               xM,
    Int32               yM,
    PIXR               *prMask,
    Int32               xD,
    Int32               yD,
    Int32               wD,
    Int32               hD,
    PIXR               *prDest)
{
    PROLOGUE;
    PIXR     *prDestCrop = NULL, *prSrcCrop = NULL, *prMaskCrop = NULL,
	     *prConv = NULL, *prPlane;
    PIXR     *planes[cMaxColorPlanes];
    Bool      cropDest = cFalse, cropSrc = cFalse, cropMask = cFalse;
    Int32     channels, srcDepth, destDepth, srcChannels, destW, 
	      destH, srcW, srcH, maskW, maskH, i;
    ScaledInt1616     xDPI, yDPI;
    ColorType type;
    ColorMap *map;

    ERR_GEN_PIXR (prDest, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1|cChannels3, cFrame32);
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1|cChannels3, cFrame32);
    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetType (pIPWState, prDest, &type));
    ERR_PASS (w_pixrGetColorMap (pIPWState, prDest, &map));
    ERR_PASS (w_pixrGetDepth (pIPWState, prDest, &destDepth));
    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &srcDepth));
   
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prDest, &channels));
    ERR_PASS (w_pixrGetChannelCount (pIPWState, prSrc, &srcChannels));
    ERR_GEN (channels != srcChannels, cErrChannelsNotEqual);
    ERR_PASS (w_pixrGetWidth (pIPWState, prDest, &destW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prDest, &destH));
    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &srcW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &srcH));
    ERR_PASS (w_pixrGetWidth (pIPWState, prMask, &maskW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prMask, &maskH));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prDest, &xDPI, &yDPI));

    /* If we are doing sub-rects, we need to make copies */
    if (xD == 0 && yD == 0 && wD == destW && hD == destH)
    {
	cropDest = cFalse;
	prDestCrop = prDest;
    }
    else
    {
	cropDest = cTrue;

	ERR_PASS (w_createPixrExtendedNoInit (pIPWState, wD, hD, destDepth,
					      xDPI, yDPI, cFrameBits,
					      channels, type, map,
					      &prDestCrop));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, xD, yD, prDest, PIX_SRC,
				      0, 0, wD, hD, prDestCrop));
    }
    

    
    if (xS == 0 && yS == 0 && wD == srcW && hD == srcH)
    {
	cropSrc = cFalse;
	prSrcCrop = prSrc;
    }
    else
    {
	cropSrc = cTrue;
	ERR_PASS_CLEANUP (w_createPixrExtendedNoInit (pIPWState, wD, hD,
						      srcDepth, xDPI, yDPI, 
						      cFrameBits, channels,
						      type, map, &prSrcCrop));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, xS, yS, prSrc, PIX_SRC,
				      0, 0, wD, hD, prSrcCrop));
    }

    if (xM == 0 && yM == 0 && wD == maskW && hD == maskH)
    {
	cropMask = cFalse;
	prMaskCrop = prMask;
    }
    else
    {
	cropMask = cTrue;
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, wD, hD, 1, xDPI, yDPI, 
					cFrameBits, &prMaskCrop));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, xM, yM, prMask, PIX_SRC,
				      0, 0, wD, hD, prMaskCrop));
    }
    
    if (channels > 1)
    {

	/* "paint" the dest planes */
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prDestCrop, i, 
						&prPlane));
	    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prPlane, prMaskCrop, 
					   PIX_SRC | PIX_DST, &prPlane));
	}

	/* "paint" the source planes */
	ERR_PASS_CLEANUP (w_invert (pIPWState, prMaskCrop, &prMaskCrop));
	for (i = 0; i < channels; i++)
	{
	    
	    ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prSrcCrop, i, 
					        &prPlane));
	    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prPlane, prMaskCrop, 
					   PIX_SRC | PIX_DST, &prPlane));
	}
	ERR_PASS_CLEANUP (w_invert (pIPWState, prMaskCrop, &prMaskCrop));
	
    }
    else
    {
	/* "paint" the dest, forcing all in mask to be 0 */
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prDestCrop, prMaskCrop, 
				       PIX_SRC | PIX_DST, &prDestCrop));

	/* "paint" the src, i.e. make all non-src image 0 */
	ERR_PASS_CLEANUP (w_invert (pIPWState, prMaskCrop, &prMaskCrop));
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prSrcCrop, prMaskCrop,
				       PIX_SRC | PIX_DST, &prSrcCrop));
	ERR_PASS_CLEANUP (w_invert (pIPWState, prMaskCrop, &prMaskCrop));
    }

    
    /* Convert depths, if necessary, and copy it in */
    if (srcDepth != destDepth)
    {
	prConv = NULL;
	
	if (channels > 1)
	{
	    for (i = 0; i < channels; i++)
	    {
		ERR_PASS_CLEANUP (w_pixrGetChannel(pIPWState, prSrcCrop, i,
						   &prPlane));
		ERR_PASS_CLEANUP (w_depthConvertGray (pIPWState, prPlane, 
						      destDepth, cFalse, 
						      IP_HI_QUALITY,
						      &(planes[i])));
	    }
	    ERR_PASS_CLEANUP (w_pixrGetType (pIPWState, prSrc, &type));
	    ERR_PASS_CLEANUP (w_createPixrComposite (pIPWState, planes,
						     channels, type, &prConv));

	}
	else
	{
	    ERR_PASS_CLEANUP (w_depthConvertGray (pIPWState, prSrcCrop,
						  destDepth, cFalse, 
						  IP_HI_QUALITY, &prConv));
	}

	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prConv, OR, prDestCrop));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prConv));
	
    }
    else
    {
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prSrcCrop, OR, prDestCrop));
    }
    
    /* Copy back in, if necessary */
    if (cropDest)
    {
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prDestCrop, PIX_SRC,
				      xD, yD, wD, hD, prDest));
	
    }

    /* Free stuff up if needed */
    if (cropDest)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDestCrop));
    if (cropSrc) 
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSrcCrop));
    if (cropMask)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMaskCrop));
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (cropDest)
	w_destroyPixr (pIPWState, &prDestCrop);
    if (cropSrc)
	w_destroyPixr (pIPWState, &prSrcCrop);
    if (cropMask)
	w_destroyPixr (pIPWState, &prMaskCrop);
    
    ERR_RETURN;
}

/*
$Log: appsutil.c,v $
Revision 1.1.1.1  1999/09/20 18:19:13  jlin
Imported sources

Revision 1.3  1997/06/26 20:00:07  ddavies
ddavies on Thu Jun 26 13:00:06 PDT 1997

appsutil.c -> /product/ipcore/ipwrap/src/RCS/appsutil.c,v

CommonCheck: Add inclusion of new file rast.pub.

Revision 1.2  1997/04/15 20:07:47  ddavies
appsutil.c -> /product/ipcore/ipwrap/src/RCS/appsutil.c,v

Substitute NoInit (uninitialized) version of pixr creation routines where
possible.

Revision 1.1  1996/09/04 22:58:44  macleod
Initial revision

*/
