/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * autoproc.c
 *	      
 * autoscan contains lots image processing, all of processings are
 * organized in the piper lines. It is hard to fit the requirement
 * of "interactive way", such as get results in the middle of 
 * prcessing, set back modified results. Therefore we take out
 * image processing and packed them in several functions. API,
 * autoscan or others can shared them together.
 *
 * Exports: w_autoProcSegmentPage()
 *          w_autoProcCuttingPage()
 *          w_autoProcDetections()
 *          w_autoProcOrientImage()
 *          w_autoProcBlastMasks()
 *          w_autoProcPastePict()
 *          w_autoProcCleanText()
 *          w_autoProcAIE()
 *          w_autoProcScaledCrop()
 *          w_autoProcDeskewImage()
 *
 */

#include <stdio.h>
#include <math.h>
#include "ipwcore.prv"
#include "types.pub"
#include "defines.pub"
#include "geomadt.h"
#include "morph.h"
#include "morphg.h"
#include "pixr.h"
#include "thrshold.h"
#include "frames.pub"
#include "iaerror.pub"
#include "appsinc.h"
#include "logicops.h"
#include "ansiprot.h"
#include "arithops.h"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "thrshold.pub"
#include "fthresh.pub"
#include "shrscal.pub"
#include "page.pub"
#include "prxform.pub"
#include "shrip.pub"
#include "shrcc.pub"
#include "skew.pub"
#include "crop.pub"
#include "reduce.pub"
#include "expand.pub"
#include "rectops.pub"
#include "cut.pub"
#include "rotatop.pub"
#include "rescale.pub"
#include "fill.pub"
#include "halftone.pub"
#include "binarize.pub"
#include "coords.pub"
#include "imginfo.pub"
#include "clean.pub"
#include "imginfo.pub"
#include "iw_scale.pub"
#include "aie.pub"
#include "clean.pub"
#include "resconv.pub"
#include "morph.pub"
#include "morphg.pub"
#include "strel.pub"
#include "grbinop.pub"
#include "logicops.pub"
#include "appsutil.pub"
#include "cntr.pub"
#include "imageref.pub"
#include "arithops.pub"
#include "intops.pub"
#include "autoscan.pub"
#include "autoproc.pub"

#undef DEBUG
#ifdef DEBUG
#endif


IP_RCSINFO(autoproc_c_RCSInfo, "$RCSfile: autoproc.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Perform segmentation on a page and output masks.
 *
 * LongDescription:::
 *   Find the images in the image plane by calling w_segmentPagePixrs, 
 *   then output masks. The image plane is low resolution image (100dpi).
 *
 * Input:::
 *  pIPWState::      State structure
 *  pImagePlane::    Point to input image to be segmented.
 *  
 *  
 * Output:::
 *  ppMasks::        Point to points of masks
 *  pmaskCount::      Point to Number of masks
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcSegmentPage(
    IPWState       *pIPWState,
    PIXR           *pImagePlane,
    PixrRef       **ppMasks,
    Int32          *pmaskCount)
{
    PROLOGUE;
    Int32          dpi, textThresh, backThresh, depth;
    Int32          redFactor, ruleWidth, minWidth, minHeight;
    PIXR          *prPrimary = NULL;
    Bool           invert, probablyText;
    ScaledInt1616  xRes, yRes;

    ERR_GEN (pImagePlane == NULL, cErrNullPtr);
    ERR_GEN (ppMasks == NULL || *ppMasks != NULL, cErrNullPtr);

    ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, pImagePlane, 
				       &prPrimary));
    ERR_PASS (w_pixrGetDepth (pIPWState, prPrimary,&depth));

    if( depth == 8 )
    {
	ERR_PASS (w_findTextThreshold (pIPWState, prPrimary, &textThresh, 
				       &invert));
	ERR_PASS (w_findBgThreshAroundText (pIPWState, prPrimary, NULL, 
					    textThresh,	&backThresh));
    }
    else
    {
	ERR_PASS (w_findThresholds (pIPWState, prPrimary,
				    NULL, cFalse, 
				    &backThresh, &textThresh, 
				    &invert, &probablyText, NULL));
    }

    /* Do the segmentation.  These factors affect it, have not
     * really been tuned very much, and should be.
     */
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, pImagePlane, &xRes, &yRes));
    dpi = MAX (ScaledInt1616_TO_Int32(xRes), ScaledInt1616_TO_Int32(yRes));

    if (dpi >= 300)
    {
	redFactor = 4;
	ruleWidth = 8;
	minWidth = 99;
	minHeight = 15;
    }
    else if (dpi >= 200)
    {
	redFactor = 2;
	ruleWidth = 4;
	minWidth = 66;
	minHeight = 10;
    }
    else
    {
	redFactor = 1;
	ruleWidth = 2;
	minWidth = 33;
	minHeight = 5;
    }

    ERR_PASS (w_segmentPagePixrs (pIPWState, prPrimary,
				  textThresh, backThresh, 
				  redFactor, ruleWidth, 1,
				  minWidth, minHeight, 1,
				  ppMasks, pmaskCount));

    NORMAL_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Perform cutting on image plane according to input masks.
 *
 * LongDescription:::
 *   After page segmentation and modification, use this routine to
 *   get segmented image.  
 *
 * Input:::
 *  pIPWState::      State structure
 *  pImagePlane::    Point to input image to be cut.
 *  pMasks::         Point to points of masks
 *  maskCount::      Number of masks
 *  skew::           Detected skew angle
 *  
 * Output:::
 *  ppImages::       Point to point of segmented images
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcCuttingPage(
    IPWState       *pIPWState,
    PIXR           *pImagePlane,
    PixrRef        *pMasks,
    Int32           maskCount,
    Float32         skew,
    PixrRef       **ppImages)
{
    PROLOGUE;
    Int32          i;
    Float64        cosTheta;
    RECTANGL       rect;
    PixrRef       *pImages = NULL;

    ERR_GEN (pImagePlane == NULL, cErrNullPtr);
    ERR_GEN (((maskCount > 0) && (pMasks == NULL)), cErrNullPtr);
    ERR_GEN (ppImages == NULL || *ppImages != NULL, cErrNullPtr);

    /*
     * Make a list of PIXRs for the images in the image plane
     * corresponding to the masks in the mask list.
     */
    pImages = IPW_MALLOC (maskCount * sizeof(PixrRef), cAllocatorData);
    ERR_GEN (!pImages, ia_nomem);
    cosTheta = cos(skew);

    for (i = 0; i < maskCount; i++)
    {
	rect.x = pMasks[i].xOffset;
	rect.y = pMasks[i].yOffset;
	ERR_PASS (w_pixrGetWidth (pIPWState, pMasks[i].pPixr,
				  &rect.w));
	ERR_PASS (w_pixrGetHeight (pIPWState, pMasks[i].pPixr,
				   &rect.h));
	
	if (skew < cLargestSmallAngle &&
	    skew > cBinaryAdjustCutoff)
	{
	    pImages[i].xOffset = (Int32) ((Float64) rect.x * cosTheta);
	    pImages[i].yOffset = (Int32) ((Float64) rect.y * cosTheta);
	}
	else
	{
	    pImages[i].xOffset = rect.x;
	    pImages[i].yOffset = rect.y;
	}
	    
	pImages[i].pPixr = NULL;
	ERR_PASS (w_rectOf (pIPWState, pImagePlane, &rect,
				    &(pImages[i].pPixr)));
    }

    *ppImages = pImages;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Detect bounding box, skew angle and orientation..
 *
 * LongDescription:::
 *   According to options (crop, deskew, orient), detect bounding box, 
 *   orientation and sekw angle. 
 *
 * Input:::
 *  pIPWState::      State structure
 *  prTextPlane::    Binary image for detecting
 *  processingFlags:: See autoscan processing flag in autoscan.pub
 *  
 * Output:::
 *  pBoundingBox1::  Point to init bounding box
 *  pOrientation::   Point to orientation
 *  pSkew::          Point to skew angle
 *  pConfidence::    Point to skew detection confidence
 *  pBoundingBox2::  Point to bounding box 
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcDetections (
   IPWState       *pIPWState,
   PIXR           *prTextPlane,
   UInt32          processingFlags,
   RECTANGL       *pBoundingBox1,
   Int32          *pOrientation,
   Float32        *pSkew,
   Int32          *pConfidence,
   RECTANGL       *pBoundingBox2)
{
    PROLOGUE; 
    Int32          dpi, margin = 0, w, h, len;
    PIXR          *prText = NULL, *prTmp = NULL; 
    RECTANGL      *tmpRect = NULL;
    Bool           fillEdges = cFalse;
    ScaledInt1616  xRes, yRes;    
    Bool           canInPlace;

    ERR_GEN (!pBoundingBox1, cErrNullPtr);
    ERR_GEN (!pBoundingBox2, cErrNullPtr);
    ERR_GEN (!pOrientation, cErrNullPtr);
    ERR_GEN (!pSkew, cErrNullPtr);
    ERR_GEN (!pConfidence, cErrNullPtr);
    ERR_GEN (!prTextPlane, cErrNullPtr);

    *pOrientation = 0;
    *pSkew = 0.0F;
    *pConfidence = 0;
    ERR_PASS (w_copy (pIPWState, prTextPlane, &prText));

    pBoundingBox1->x = pBoundingBox1->y =0;
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prText, &pBoundingBox1->w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prText, &pBoundingBox1->h));
    *pBoundingBox2 = *pBoundingBox1;

    if ( (processingFlags & cAutoOption_crop) && 
	 ((processingFlags & cAutoOption_deskew) || 
	  (processingFlags & cAutoOption_orient)))
    {
	/* First, find an "initial" bounding box. The reason we
	 * do this is so that we can crop, orient deskew, and then
	 * crop again, thus potentially reducing the number of
	 * pixels we need to deskew, which should make things
	 * go faster on average. We always leave a margin on this
	 * bounding box.
	 */
	ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, 
						  prText,
						  &xRes, &yRes));
	    
	margin = MAX (ScaledInt1616_TO_Int32 (xRes),
		      ScaledInt1616_TO_Int32 (yRes)) / 2;
	fillEdges = (processingFlags & cAutoOption_fillEdges) ? cTrue : cFalse;
	
	ERR_PASS_CLEANUP (
	    w_documentBoundingBox (pIPWState, prTextPlane,
				   NULL, margin, fillEdges,
				   pBoundingBox1));

	ERR_PASS_CLEANUP (w_clipRectToPixr (pIPWState, 
					    prText,
					    pBoundingBox1,
					    &tmpRect));

	*pBoundingBox1 = *tmpRect;
	ERR_PASS_CLEANUP (w_destroyRect (pIPWState, &tmpRect));


	/* Crop the text plane in-place */
	ERR_PASS_CLEANUP (
	    w_cutRectIP(pIPWState, 
			pBoundingBox1->x,
			pBoundingBox1->y,
			pBoundingBox1->w, 
			pBoundingBox1->h, 
			prText));

    }

    /* Check orientation */
    if ( processingFlags & cAutoOption_orient )
    {
	if ((processingFlags & cAutoOption_orient4))
	{
	    ERR_PASS_CLEANUP (w_orient4 (pIPWState, prText, pOrientation));
	}
	else
	{
	    ERR_PASS_CLEANUP (w_orient (pIPWState, prText, pOrientation));
	}

	if (*pOrientation != 0)
	{  
	    /* check again, some table will cause mistakes */
	    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, 
						      prText,
						      &xRes, &yRes));

	    dpi = MAX(ScaledInt1616_TO_Int32 (xRes), 
		      ScaledInt1616_TO_Int32 (yRes));
	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prText, &w));
	    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prText, &h));

	    if( dpi >= 200 )
	    {
		ERR_PASS( w_createPixr(pIPWState, (w+1)/2, (h+1)/2, 1,
				       xRes/2, yRes/2, cFrameBits, &prTmp));
		ERR_PASS (w_reduceio2 (pIPWState, prText, 2, prTmp));
		dpi /= 2;
		len = MAX( (w+1)/2, (h+1)/2 );
		ERR_PASS_CLEANUP (w_removeHVRules (pIPWState, prTmp, dpi/2, len,
						   1, (dpi/28)+1, &prTmp));
	    }
	    else
	    {
		len = MAX(w, h);
		ERR_PASS_CLEANUP (w_removeHVRules (pIPWState, prText, 
						   dpi/2, len,
						   1, (dpi/28)+1, &prTmp));
	    }

	    if (processingFlags & cAutoOption_orient4)
	    {
		ERR_PASS_CLEANUP (w_orient4 (pIPWState, prTmp, pOrientation));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_orient (pIPWState, prTmp, pOrientation));
	    }
	    ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &prTmp));
	    prTmp = NULL;
	    
	}

	if ( *pOrientation )
	{
	    canInPlace=(processingFlags & cAutoOption_moreMemory)? cFalse:cTrue;
	    ERR_PASS_CLEANUP (w_autoProcOrientImage (pIPWState, *pOrientation, 
					      canInPlace, &prText));
	}

    }

    /* skew */
    if (processingFlags & cAutoOption_deskew)
    {
	/* 
	 * Deskew. If we oriented, do a skew-detect again.
	 * This might not be necessary.
	 */
	ERR_PASS_CLEANUP (w_findSkewSample (pIPWState, prText, 2, 
					    pConfidence, pSkew));
	    
	if ((processingFlags & cAutoOption_crop) &&
	    fabs(*pSkew) > cSmallestRotatingAngle)
	{
	    ERR_PASS_CLEANUP (w_rotateArb (pIPWState, prText, *pSkew, 
					   cTrue, cFalse, &prTmp));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prText));
	    prText = prTmp;
	    prTmp = NULL;
	}
    }

    /* Find bounding box */
    if (processingFlags & cAutoOption_crop)
    {
	if (processingFlags & cAutoOption_boundingBoxNoMargin)
	{
	    margin = 0;
	}

	ERR_PASS_CLEANUP (w_documentBoundingBox (pIPWState, prText,
						 NULL, margin, 
						 fillEdges, 
						 pBoundingBox2));

	ERR_PASS_CLEANUP (w_clipRectToPixr (pIPWState, 
					    prText,
					    pBoundingBox2,
					    &tmpRect));

	*pBoundingBox2 = *tmpRect;
	ERR_PASS_CLEANUP (w_destroyRect (pIPWState, &tmpRect));

    }
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prText));

    NORMAL_RETURN;

ERR_CLEANUP:

   if (prText) w_destroyPixr (pIPWState, &prText);
   if (prTmp) w_destroyPixr (pIPWState, &prTmp);
   if (tmpRect) w_destroyRect (pIPWState, &tmpRect);

   ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Given an orientation angle (0, 90, 180, or 270) orient image passed
 *
 * LongDescription:::
 *   Internal routine: Given an orientation angle (0, 90, 180, or 270) 
 *   orient the image passed. If the inPlace flag is set, and the image
 *   is high enough, do the orientation in-place. Otherwise, make a copy
 *   and destroy the original.
 *
 * Input:::
 *  pIPWState::      State structure
 *  deg::            Orientation, 0, 90, 180 or 270
 *  canInPlace::     If cTrue, can do it in place
 *  
 *  
 * Output:::
 *  ppPrImage::      Image to orient, replaced with oriented image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcOrientImage (
    IPWState *pIPWState,
    Int32     deg,
    Bool      canInPlace,
    PIXR    **ppPrImage)
{
    PROLOGUE;
    Int32          w, h, depth;
    ScaledInt1616  xRes, yRes;
    PIXR          *prTransfer = NULL, *prTmp = NULL;

    /* NULL image is a do-nothing, not an error */
    if (*ppPrImage)
    {
	if (canInPlace && (deg == 90 || deg == 270))
	{
	    /* Create a transfer buffer for the RGB */
	    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, 
					      *ppPrImage,
					      &depth));
	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, 
					      *ppPrImage, &w));
	    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, 
					       *ppPrImage, &h));
	    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, *ppPrImage, 
						      &xRes, &yRes));
	    
	    if (h > 256)
	    {
		ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, 256, w,
						      depth, xRes, yRes,
						      cFrameBits,
						      &prTransfer));
	    }
	}
	
	switch (deg)
	{
	case 0:
	    break;
	    
	case 90:
	    if (prTransfer)
	    {
		ERR_PASS_CLEANUP (w_rotate270InPlace (pIPWState, *ppPrImage, 
						      prTransfer));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_rotate270 (pIPWState, *ppPrImage, &prTmp));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, ppPrImage));
		*ppPrImage = prTmp;
		prTmp = NULL;
	    }
	    break;

	case 180:
	    if (prTransfer)
	    {
		ERR_PASS_CLEANUP (w_rotate180InPlace (pIPWState, *ppPrImage));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_rotate180 (pIPWState, *ppPrImage, &prTmp));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, ppPrImage));
		*ppPrImage = prTmp;
		prTmp = NULL;
	    }
	    break;

	case 270:
	    if (prTransfer)
	    {
		ERR_PASS_CLEANUP (w_rotate90InPlace (pIPWState, *ppPrImage, 
						      prTransfer));
	    }
	    else
	    {
		ERR_PASS_CLEANUP (w_rotate90 (pIPWState, *ppPrImage, &prTmp));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, ppPrImage));
		*ppPrImage = prTmp;
		prTmp = NULL;
	    }
	    break;

	default:
	    ERR_GEN_CLEANUP (pIPWState, ia_invalidParm);
	    break;
	}

	if (prTransfer)
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTransfer));
    }

    NORMAL_RETURN;
	
ERR_CLEANUP:
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if (prTransfer)
	w_destroyPixr (pIPWState, &prTransfer);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Blast out segmented image from text, or background, or
 *   foreground plane.
 *
 * LongDescription:::
 *   After page segmentation and modification, use this routine to
 *   do blasting out masks from text, foreground and background plane.
 *
 * Input:::
 *  pIPWState::       State structure
 *  prTextPlane::     Point to text 
 *  prForeground::    Point to foreground plane, if no foreground, input NULL
 *  prBackground::    Point to background plane, if no background, input NULL
 *  pImages::         Point to points of images
 *  pMasks::          Point to points of masks
 *  maskCount::       Number of masks or images
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcBlastMasks(
    IPWState       *pIPWState,
    PIXR           *prTextPlane,
    PIXR           *prForeground,
    PIXR           *prBackground,
    PixrRef        *pImages,
    PixrRef        *pMasks,
    Int32           maskCount)
{
    PROLOGUE;
    Int32          w, h, i;
    Int32          expandedX = 0, expandedY = 0, expandedW = 0, expandedH = 0;
    PIXR          *prTmp = NULL, *prStrip = NULL;
    Float32        pictureRes, textRes;
    Float32        fgPictRatio = (Float32) 0, bgPictRatio = (Float32) 0;
    STREL         *pTheStrel = NULL;
    ScaledInt1616  pictXRes, pictYRes, textXRes, textYRes;
    ScaledInt1616  fgXRes, fgYRes, bgXRes, bgYRes;

    if( maskCount < 1 ) NORMAL_RETURN;

    ERR_GEN (((maskCount > 0) && (pMasks == NULL)), cErrNullPtr);

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, pMasks[0].pPixr, 
					      &pictXRes, &pictYRes));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prTextPlane, 
					      &textXRes, &textYRes));
    pictureRes = ScaledInt1616_TO_Float32(pictXRes);
    textRes = ScaledInt1616_TO_Float32 (textXRes);

    if( prForeground )
    {
	ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prForeground, 
					      &fgXRes, &fgYRes));
	fgPictRatio = ScaledInt1616_TO_Float32 (fgXRes) / 
	               ScaledInt1616_TO_Float32 (pictXRes);
    }
    if( prBackground )
    {
	ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prBackground, 
					      &bgXRes, &bgYRes));
	bgPictRatio = ScaledInt1616_TO_Float32 (bgXRes) / 
	               ScaledInt1616_TO_Float32 (pictXRes);	
    }
    /* After scaling up and down, background, foreground, selector,
       segmented image and mask will have size error, the accumulated 
       error will be greater than 1. But 3*3 erosion just can cover
       error 0 or 1 in four directions. So we change to 5*5 erosion, 
       it will cover the errors in four directions up to 2.   */

    ERR_PASS (w_initStrelByName(pIPWState, strel5));
    ERR_PASS (w_getStrel(pIPWState, strel5, &pTheStrel)); 
   
    /* Blast away the masks in the text plane */
    for (i = 0; i < maskCount; i++)
    {
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, 
					  pMasks[i].pPixr, &w));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, 
					  pMasks[i].pPixr, &h));

	expandedW = (Int32) ((Float32) w * textRes / pictureRes );
	expandedH = (Int32) ((Float32) h * textRes / pictureRes );

	expandedX = (Int32) ((Float32) pMasks[i].xOffset *
			     textRes / pictureRes );
	expandedY = (Int32) ((Float32) pMasks[i].yOffset * 
			     textRes / pictureRes );
	prTmp = NULL;
	ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, pMasks[i].pPixr, 
					 expandedW, expandedH, 
					 &prTmp));
	if( prForeground || prBackground )
	    ERR_PASS_CLEANUP (w_erode (pIPWState, prTmp, pTheStrel, &prTmp)); 
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prTmp, 
				      PIX_NOT(PIX_SRC) & PIX_DST, 
				      expandedX, expandedY, 
				      expandedW, expandedH, 
				      prTextPlane));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	if( prForeground || prBackground )
	{
	    if( prForeground )
	    {
		expandedW = (Int32)((Float32) w * fgPictRatio);
		expandedH = (Int32)((Float32) h * fgPictRatio);
		expandedX = (Int32)((Float32) pMasks[i].xOffset * 
				fgPictRatio );
		expandedY = (Int32)((Float32) pMasks[i].yOffset * 
				fgPictRatio );
		prTmp = prStrip = NULL;
		ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, 
					     pMasks[i].pPixr, 
					     expandedW, expandedH, 
					     &prTmp));
		ERR_PASS_CLEANUP (w_erode (pIPWState, prTmp, pTheStrel, 
					   &prTmp));
		ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, 
					     pImages[i].pPixr, 
					     expandedW, expandedH, 
					     &prStrip));
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prStrip, 
					      PIX_CLR, 
					      0, 0, expandedW, expandedH, 
					      prStrip));
		ERR_PASS (w_pixrSetResolutionDPI(pIPWState, prTmp,
						 fgXRes, fgYRes));
		ERR_PASS (w_pixrSetResolutionDPI(pIPWState, prStrip,
						 fgXRes, fgYRes));
		ERR_PASS_CLEANUP (w_copyGrayMasked (pIPWState, 
						    0, 0, prStrip, 
						    0, 0, prTmp,
						    expandedX, expandedY, 
						    expandedW, expandedH, 
						    prForeground));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prStrip));
	    }
	    if (prBackground)
	    {
		expandedW = (Int32)((Float32) w * bgPictRatio );
		expandedH = (Int32)((Float32) h * bgPictRatio );
		expandedX = (Int32)((Float32) pMasks[i].xOffset * 
				bgPictRatio );
		expandedY = (Int32)((Float32) pMasks[i].yOffset * 
				bgPictRatio );
		prTmp = prStrip = NULL;
		ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, 
					     pMasks[i].pPixr, 
					     expandedW, expandedH, 
					     &prTmp));
		ERR_PASS_CLEANUP (w_erode (pIPWState, prTmp, pTheStrel, 
					   &prTmp));
		ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, 
					     pImages[i].pPixr, 
					     expandedW, expandedH, 
					     &prStrip));
		ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prStrip, 
					      PIX_SET, 
					      0, 0, expandedW, expandedH, 
					      prStrip));
		ERR_PASS_CLEANUP (w_pixrSetResolutionDPI(pIPWState, prTmp,
							 bgXRes, bgYRes));
		ERR_PASS_CLEANUP (w_pixrSetResolutionDPI(pIPWState, prStrip,
							 bgXRes, bgYRes));
		ERR_PASS_CLEANUP (w_copyGrayMasked (pIPWState, 
						    0, 0, prStrip, 
						    0, 0, prTmp,
						    expandedX, expandedY, 
						    expandedW, expandedH, 
						    prBackground));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prStrip));
	    }
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);
    if (prStrip)
	w_destroyPixr (pIPWState, &prStrip);

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Paste segmented image to background plane.
 *
 * LongDescription:::
 *   After page segmentation and modification, use this routine to
 *   paste segmented images to background map.
 *
 * Input:::
 *  pIPWState::       State structure
 *  prBackground::    Point to background plane
 *  pImages::         Point to points of images
 *  pMasks::          Point to points of masks
 *  maskCount::       Number of masks or images
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcPastePict(
    IPWState       *pIPWState,
    PIXR           *prBackground,
    PixrRef        *pImages,
    PixrRef        *pMasks,
    Int32           maskCount)
{
    PROLOGUE;
    Int32     i, w, h;

    ERR_GEN (((maskCount > 0) && (pImages == NULL)), cErrNullPtr);
    ERR_GEN (((maskCount > 0) && (pMasks == NULL)), cErrNullPtr);
    ERR_GEN ((prBackground == NULL), cErrNullPtr);

    /* Paste the segmented images onto the background plane.*/
    for (i = 0; i < maskCount; i++)
    {
	ERR_PASS ( w_pixrGetWidth (pIPWState, pImages[i].pPixr, &w));
	ERR_PASS ( w_pixrGetHeight (pIPWState, pImages[i].pPixr, &h));

	ERR_PASS ( w_copyGrayMasked (pIPWState, 
				     0, 0, pImages[i].pPixr, 
				     0, 0, pMasks[i].pPixr, 
				     pImages[i].xOffset, 
				     pImages[i].yOffset, 
				     w, h, prBackground));
	
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Perform "document drycleaning" on the text plane
 *
 * LongDescription:::
 *   Internal routine: Given that we have a text plane, perform requested
 *   "document drycleaning" functions on it. At the moment, we assume
 *   that we did, if requested, do pepper removal during binarization,
 *   so what we need to do here is stipple (shading) and rule removal, unless
 *   we got a binary input, in which case we didn't do any binarization.
 *
 * Input:::
 *  pIPWState::       State structure
 *  ppTextPlane::     Point to point of binary image, 
 *  processingFlags:: See autoscan processing flag in autoscan.pub
 *  
 * Output:::
 *                    
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoProcCleanText (
    IPWState      *pIPWState,
    PIXR         **ppTextPlane,
    UInt32         processingFlags)
{
    PROLOGUE;
    PIXR          *prTmp = NULL;
    Int32          reduction, textDPI;
    ScaledInt1616  xRes, yRes;

    ERR_GEN (ppTextPlane == NULL, cErrNullPtr);
    ERR_GEN (*ppTextPlane == NULL, cErrNullPtr);
    ERR_GEN_PIXR (*ppTextPlane, cByteOrderBig, cDepth1, cChannels1, cFrame32);

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, *ppTextPlane, &xRes, &yRes));
    textDPI= MAX(ScaledInt1616_TO_Int32 (xRes),ScaledInt1616_TO_Int32 (yRes));

    /* Stipple removal */
    if ( processingFlags & cAutoOption_shadingRemove )
    {
	if (textDPI >= 300)
	    reduction = 4;
	else if (textDPI >= 200)
	    reduction = 2;
	else
	    reduction = 1;
  
	ERR_PASS_CLEANUP (w_cleanStipple (pIPWState, *ppTextPlane, 
					  reduction, &prTmp));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, ppTextPlane));
	*ppTextPlane = prTmp;
	prTmp = NULL;
    }

    if ( processingFlags & cAutoOption_lineRemove )
    {
	/* Remove rules */
	/* These parameters are kind of random. &&&FIND OUT what PerfectScan
	   did, and do that. */

	ERR_PASS_CLEANUP (w_removeHVRules (pIPWState, *ppTextPlane,
				   textDPI/2, textDPI * 30, 
				   1, textDPI/10, &prTmp));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, ppTextPlane));
	*ppTextPlane = prTmp;
	prTmp = NULL;
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp)
	w_destroyPixr (pIPWState, &prTmp);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Enhance image by use AIE
 *
 * LongDescription:::
 *   Use AIE to enhance gray or color image. If input signle image, do
 *   one image AIE. If input segmented images, analyze them and use one
 *   average results to enhance all of them.
 *
 * Input:::
 *  pIPWState::      State structure
 *  imageCount::     Number of images
 *  ppImage::        Pointer to pointer of image to be enhanced. it must
 *                   be NULL, if ppImages != NULL.
 *  ppImages::       Point to point of images to be enhanced. It must be 
 *                   NULL, if ppImage != NULL.
 *  
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *   The input image must be gray or color image.
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  X. Zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcAIE (
    IPWState            *pIPWState,
    Int32                imageCount,
    PIXR               **ppImage,
    PixrRef            **ppImages)
{
    PROLOGUE;
    Int32        channels;
    void        *pAnalysis = NULL;
    PIXR        *prImageToAIE;
    extern aAieDefaults aieStandardDefaults;

    ERR_GEN ( (*ppImage == NULL && *ppImages == NULL), cErrNullPtr); 

    if( *ppImage != NULL )
    {
	prImageToAIE = *ppImage;
	ERR_PASS (w_pixrGetChannelCount (pIPWState, 
					 prImageToAIE, 
					 &channels));


	if (channels == 3)
	{
	    /* Analyze */
	    ERR_PASS_CLEANUP (w_aieAnalyzeRGB (pIPWState, 
					       &aieStandardDefaults, 
					       prImageToAIE, 
					       &pAnalysis));

	    /* Correct */
	    ERR_PASS_CLEANUP (
		w_aieProcessRGB (pIPWState, prImageToAIE, pAnalysis, 
				 &(prImageToAIE)));

	    /* Free the analysis thing */
	    ERR_PASS_CLEANUP (w_aieFinish (pIPWState, &pAnalysis));
	}
	else if (channels == 1)
	{
	    /* Analyze */
	    ERR_PASS_CLEANUP (w_aieAnalyzeGray (pIPWState, 
						&aieStandardDefaults, 
						prImageToAIE, 
						&pAnalysis));

	    /* Correct */
	    /* NOTE: the parameters are reversed from aieProcessRGB.
	       I should probably fix this. */
	    ERR_PASS_CLEANUP ( w_aieProcessGray (pIPWState, 
						 pAnalysis, prImageToAIE,
						 &(prImageToAIE) ));
	
	    /* Free the analysis thing */
	    ERR_PASS_CLEANUP (w_aieFinish (pIPWState, &pAnalysis));
	}
	else 
	{
	    ERR_GEN (-1, cErrChannelsNotSupported);
	}

	*ppImage = prImageToAIE;
    }
    else if( imageCount > 0 )
    {
	Int32          i;
	PixrRef       *pImages;
	void          *bandedThing = NULL;

	pImages = *ppImages;
	ERR_PASS (w_pixrGetChannelCount (pIPWState, 
					 pImages[0].pPixr, 
					 &channels));
	if( channels == 1)
	{
	    ERR_PASS( w_aieAnalyzeGrayInit (pIPWState, &aieStandardDefaults,
				    0, 0, 0, &bandedThing));
	    for(i=0; i<imageCount; ++i)
	    {
		ERR_PASS( w_aieAnalyzeGrayProcess (pIPWState, bandedThing, 
						   pImages[i].pPixr));
	    }
	    ERR_PASS (w_aieAnalyzeGrayFinish (pIPWState, &bandedThing, 
					      &pAnalysis));
	    for(i=0; i<imageCount; ++i)
	    {
		prImageToAIE = pImages[i].pPixr;
		ERR_PASS_CLEANUP (w_aieProcessGray (pIPWState, pAnalysis, 
						    prImageToAIE,
						    &prImageToAIE));
	    }
	    ERR_PASS_CLEANUP (w_aieFinish (pIPWState, &pAnalysis));	    
	}
	else if( channels ==  3)
	{
	    ERR_PASS( w_aieAnalyzeRGBInit (pIPWState, &aieStandardDefaults,
				     0, 0, 0, &bandedThing));
	    for(i=0; i<imageCount; ++i)
	    {
		ERR_PASS (w_aieAnalyzeRGBProcess (pIPWState, bandedThing, 
						   pImages[i].pPixr));
	    }
	    ERR_PASS( w_aieAnalyzeRGBFinish (pIPWState, &bandedThing, 
					     &pAnalysis));
	    for(i=0; i<imageCount; ++i)
	    {
		prImageToAIE = pImages[i].pPixr;
		ERR_PASS_CLEANUP (w_aieProcessRGB (pIPWState, pAnalysis, 
						   prImageToAIE,
						   &prImageToAIE));
	    }
	    ERR_PASS_CLEANUP (w_aieFinish (pIPWState, &pAnalysis));
	}
	else
	{
	    ERR_GEN (-1, cErrChannelsNotSupported);
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    if (pAnalysis)
	w_aieFinish (pIPWState, &pAnalysis);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Crop an image in-place, given a bounding rectangle and a scaling
 *
 * LongDescription:::
 *   Given a bounding rectangle and scale factor,
 *   scale the rectangle by the scale factor, clip it in bounds, and
 *   then cut the image in-place to the scaled region.
 *
 * Input:::
 *  pIPWState::         State structure
 *  pRect::             Rectangle to use
 *  xScale::            X-scaling, 1.0 = no scaling
 *  yScale::            Y-scaling, 1.0 = no scaling
 *  
 * Output:::
 *  pImage::            Image to cut
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *   Moved from autoscan.c.
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoProcScaledCrop (
    IPWState        *pIPWState,
    RECTANGL        *pRect,
    Float32          xScale,
    Float32          yScale,
    PIXR            *pImage)
{
    PROLOGUE;
    RECTANGL scaledRect, *clippedRect = NULL;

    ERR_GEN( pRect == NULL, cErrNullPtr); 
    ERR_GEN( pImage == NULL, cErrNullPtr); 

    scaledRect.x = (Int32)((Float32)pRect->x * xScale); 
    scaledRect.y = (Int32)((Float32)pRect->y * yScale); 
    scaledRect.w = (Int32)((Float32)pRect->w * xScale); 
    scaledRect.h = (Int32)((Float32)pRect->h * yScale); 

    ERR_PASS (w_clipRectToPixr (pIPWState, pImage, &scaledRect, 
				&clippedRect));

    ERR_PASS_CLEANUP (w_cutRectIP (pIPWState, 
				   clippedRect->x,
				   clippedRect->y,
				   clippedRect->w,
				   clippedRect->h,
				   pImage));

    ERR_PASS_CLEANUP (w_destroyRect (pIPWState, &clippedRect));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (clippedRect)
	w_destroyRect (pIPWState, &clippedRect);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Given a skew angle, deskew an image, in place if possible.
 *
 * LongDescription:::
 *   Given a skew angle, deskew the image passed.
 *   If the angle is small enough, do it in place, otherwise rotate the
 *   image into a copy and destroy the original.
 *<BR>
 *   The interpolated and non-interpolated routines yield somewhat
 *   different results, and won't line up, which is the reason for
 *   the flag allowing one to select whether or not to interpolate.
 *
 * Input:::
 *  pIPWState::      State structure
 *  skew::           Skew angle, in radians
 *  interpolate::    Should we interpolate? If non-zero, and we're not
 *                   doing binary, do interpolated rotate.
 *  
 * Output:::
 *  pPrImage::         Image to deskew, replaced with deskewed image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  Moved from autoscan.c.
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/

ErrCode CDECL
w_autoProcDeskewImage (
    IPWState *pIPWState,
    Float32   skew,
    Int32     interpolate, 
    PIXR    **pPrImage)
{
    PROLOGUE;
    Int32 depth;
    PIXR *prRotated = NULL;

    /* We don't raise an error if a NULL image is passed */
    if (*pPrImage != NULL)
    {
	ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, *pPrImage, &depth));
	if (depth == 1)
	{
	    /* We can always do binary in-place, and never interpolate */
	    ERR_PASS_CLEANUP (w_rotateArb (pIPWState, *pPrImage, skew,
					   cTrue, cTrue, pPrImage));
	}
	else
	{
	    if (fabs(skew) > cLargestSmallAngle)
	    {
		/* Well, if you give THAT big a skew, you deserve to
		   use a lot of memory */
		prRotated = NULL;
		if (interpolate && (fabs(skew) < cLargestGrayAngle))
		{
		    ERR_PASS_CLEANUP (w_rotateGray (pIPWState, *pPrImage,
						    skew,  &prRotated));
		}
		else
		{
		    ERR_PASS_CLEANUP (w_rotateArb (pIPWState, *pPrImage, skew,
						   cTrue, cTrue, &prRotated));
		}
		    
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, pPrImage));
	    
		*pPrImage = prRotated;
		prRotated = NULL;
	    }
	    else
	    {
		/* In-place gray rotate */
		if (interpolate)
		{
		    ERR_PASS_CLEANUP (w_rotateGrayIP (pIPWState, skew, 
						      *pPrImage));
		}
		else
		{
		    ERR_PASS_CLEANUP (w_rotateArb (pIPWState, *pPrImage, skew,
						   cTrue, cTrue, pPrImage));
		}
	    }
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (prRotated)
	w_destroyPixr (pIPWState, &prRotated);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Post process high resolution binary image for OCR segmentation.
 *
 * LongDescription:::
 *  Post process high resolution binary image for improving OCR segmentation.
 *  Adaptive threshold is good to recognition of character, but the picture
 *  will be more fragmented. Here we try to find pictures and make them
 *  more connected.  
 *
 * Input:::
 *  pIPWState::      State structure
 *  pTextPlane::     Point to high resolution binary image
 *  pImagePlane::    Point to low resolution color or gray image
 *  options::        Option flag  (in autoproc.pub)
 *  
 *  
 * Output:::
 *  ppText::         Point to pointer of processed image (TextPlane size)
 *  ppMask::         Point to pointer of mask image (ImagePlane size)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  auto scan
 *
 * Keywords:::
 *  auto scan
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_autoProcTextplaneForOCRSeg(
    IPWState       *pIPWState,
    PIXR           *pTextPlane,
    PIXR           *pImagePlane,
    Int32           options,
    PIXR          **ppText,
    PIXR          **ppMask)
{
    PROLOGUE;
    Int32          dpi, textThresh, backThresh, depth;
    PIXR          *prPrimary = NULL, *pBack = NULL;
    Int32          redFactor, ruleWidth, minWidth, minHeight;
    Bool           invert, probablyText;
    ScaledInt1616  xRes, yRes;
    PIXR          *pMaskedText = NULL, *pTextMask = NULL, *pText = NULL;
    PIXR          *pTmp = NULL, *pReveMask = NULL, *pPictMask = NULL;
    Int32          width, height, bits;
    STREL         *pStrel = NULL;
    RECTANGL	  *textRects = NULL;

    ERR_GEN (pImagePlane == NULL, cErrNullPtr);
    if( (options & cAutoProc_FullBlacken) ||
	(options & cAutoProc_MakeBlackenMask) )
	ERR_GEN (pTextPlane == NULL, cErrNullPtr);

    ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, pImagePlane, 
				       &prPrimary));
    ERR_PASS (w_pixrGetDepth (pIPWState, prPrimary,&depth));

    if( depth == 8 )
    {
	ERR_PASS (w_findTextThreshold (pIPWState, prPrimary, &textThresh, 
				       &invert));
	ERR_PASS (w_findBgThreshAroundText (pIPWState, prPrimary, NULL, 
					    textThresh,	&backThresh));
    }
    else
    {
	ERR_PASS (w_findThresholds (pIPWState, prPrimary,
				    NULL, cFalse, 
				    &backThresh, &textThresh, 
				    &invert, &probablyText, NULL));
    }

    /* Do the segmentation.  These factors affect it, have not
     * really been tuned very much, and should be.
     */
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, pImagePlane, &xRes, &yRes));
    dpi = MAX (ScaledInt1616_TO_Int32(xRes), ScaledInt1616_TO_Int32(yRes));

    if (dpi >= 300)
    {
	redFactor = 4;
	ruleWidth = 8;
	minWidth = 99;
	minHeight = 15;
    }
    else if (dpi >= 200)
    {
	redFactor = 2;
	ruleWidth = 4;
	minWidth = 66;
	minHeight = 10;
    }
    else
    {
	redFactor = 1;
	ruleWidth = 2;
	minWidth = 33;
	minHeight = 5;
    }

    pTmp = NULL;
    if(options & cAutoProc_MakePictureMask) 
    {
	ERR_GEN ((options & cAutoProc_MakeBlackenMask), cErrTypeNotSupported);
	ERR_GEN (ppMask == NULL, cErrNullPtr);
	ERR_PASS (w_segmentPageMask (pIPWState, prPrimary, 
				     textThresh, backThresh,
				     redFactor, ruleWidth, 1,
				     minWidth, minHeight, 1,
				     NULL, &pPictMask));
	if( *ppMask == NULL )
	    *ppMask = pPictMask;
	else
	{
	    ERR_PASS_CLEANUP( w_copy(pIPWState, pPictMask, ppMask));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPictMask));
	}
	pPictMask = NULL;
    }
    
    if( (options & cAutoProc_FullBlacken) ||
	(options & cAutoProc_MakeBlackenMask) )
    {
	ERR_PASS (w_segmentPagePaintingMask (pIPWState, prPrimary,
				  textThresh, backThresh, 
				  redFactor, ruleWidth, 1,
				  minWidth, minHeight,
				  NULL, &pPictMask));
    }

    if( pPictMask != NULL )
    {
	ERR_PASS_CLEANUP (w_reduce2 (pIPWState, pTextPlane, 1, &pText));

	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pText, &width));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pText, &height));
	ERR_PASS_CLEANUP (w_clonePixr (pIPWState, pText, &pMaskedText));
	ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, pPictMask, width, height, 
					 &pMaskedText));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPictMask));

	if( options & cAutoProc_FullBlacken )
	{
	    pTmp = NULL;
	    ERR_PASS_CLEANUP (w_invert (pIPWState, pText, &pTmp));
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pMaskedText, AND, pTmp));
	    ERR_PASS_CLEANUP (w_textMask (pIPWState, pTmp, &pReveMask, 1));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pTmp));

	    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, pTextPlane, 
					      &xRes, &yRes));
	    dpi = MAX (ScaledInt1616_TO_Int32(xRes), 
		       ScaledInt1616_TO_Int32(yRes));

	    if ((dpi/2) >= 300)
	    {
		redFactor = 4;
		ruleWidth = 8;
		minWidth = 99;
		minHeight = 15;
	    }
	    else if ((dpi/2) >= 200)
	    {
		redFactor = 2;
		ruleWidth = 4;
		minWidth = 66;
		minHeight = 10;
	    }
	    else
	    {
		redFactor = 1;
		ruleWidth = 2;
		minWidth = 33;
		minHeight = 5;
	    }

	    ERR_PASS_CLEANUP (w_halftoneThresh (pIPWState, pText, redFactor,
					    ruleWidth, 1, minWidth, 
					    minHeight, &textRects, &pBack));
	    if( pBack)
	    {
		ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &textRects));
	    }
	}
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, pMaskedText, AND, pText));
	ERR_PASS_CLEANUP (w_textMask (pIPWState, pText, &pTextMask, 0));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pText));

	if( pReveMask != NULL )
	{
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pReveMask, OR, pTextMask));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pReveMask));
	}

	ERR_PASS_CLEANUP (w_numberBits(pIPWState, pTextMask, &bits));
	if( bits > 10 )
	{
	    ERR_PASS_CLEANUP (w_reduce2 (pIPWState, pTextMask, 1, &pTmp));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pTextMask)); 

	    pStrel = NULL;
	    ERR_PASS_CLEANUP (w_initStrelByName (pIPWState, strel15v));
	    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel15v, &pStrel));
	    ERR_PASS_CLEANUP (w_close (pIPWState, pTmp, pStrel, &pTmp));

	    pStrel = NULL;
	    ERR_PASS_CLEANUP (w_initStrelByName (pIPWState, strel20h));
	    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel20h, &pStrel));
	    ERR_PASS_CLEANUP (w_dilate (pIPWState, pTmp, pStrel, &pTmp));

	    pStrel = NULL;
	    ERR_PASS_CLEANUP (w_initStrelByName (pIPWState, strel10v));
	    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel10v, &pStrel));
	    ERR_PASS_CLEANUP (w_dilate (pIPWState, pTmp, pStrel, &pTmp));

	    ERR_PASS_CLEANUP (w_expand (pIPWState, pTmp, 2, &pTextMask));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pTmp));

	    ERR_PASS_CLEANUP (w_logicOp(pIPWState, pTextMask, SUBTRACT, 
					pMaskedText));
	}
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pTextMask));
	if( pBack )
	{
	    pStrel = NULL;
	    ERR_PASS_CLEANUP (w_initStrelByName (pIPWState, strel5));
	    ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel5, &pStrel));
	    ERR_PASS_CLEANUP (w_erode (pIPWState, pBack, pStrel, &pBack));
	    
	    ERR_PASS_CLEANUP (w_logicOp(pIPWState, pBack, SUBTRACT, 
					pMaskedText));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pBack));
	}

	if( options & cAutoProc_FullBlacken)
	{
	    ERR_PASS_CLEANUP(w_expand (pIPWState, pMaskedText, 2, &pPictMask));
	    if( pTextPlane != *ppText )
		ERR_PASS_CLEANUP (w_copy (pIPWState, pTextPlane, ppText));
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, pPictMask, OR, *ppText));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pPictMask));
	}
	if( options & cAutoProc_MakeBlackenMask )
	{
	    ERR_GEN (ppMask == NULL, cErrNullPtr);
	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, pImagePlane, &width));
	    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, pImagePlane, 
					       &height));
	    ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, pMaskedText, width, 
					     height, ppMask));
	}
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pMaskedText)); 
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pTmp) w_destroyPixr (pIPWState, &pTmp);
    if (pText) w_destroyPixr (pIPWState, &pText);
    if (pBack) w_destroyPixr (pIPWState, &pBack);
    if (pTextMask) w_destroyPixr (pIPWState, &pTextMask);
    if (pReveMask) w_destroyPixr (pIPWState, &pReveMask);
    if (pPictMask) w_destroyPixr (pIPWState, &pPictMask);
    if (pMaskedText) w_destroyPixr (pIPWState, &pMaskedText);

    ERR_RETURN;
}
	
/* 
 * $Log: autoproc.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:22  jlin
 * Imported sources
 *
 * Revision 1.14  1998/10/16 16:43:17  xzhu
 * xzhu on Fri Oct 16 09:41:50 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * When option is cAutoPorc_MakePictureMask, text pixr will not be used, it can be NULL.
 *
 * Revision 1.13  1998/10/14 00:44:20  xzhu
 * xzhu on Tue Oct 13 17:43:08 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * w_autoProcTextplaneForOCRSeg can output painting mask, or painted text plane, or picture mask according to options.
 *
 * Revision 1.12  1998/10/09 17:34:29  xzhu
 * xzhu on Fri Oct 9 10:34:17 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Fix the bug reported from Andrew. The hafltone mask may be NULL if it returns from unwork. We need to check it before doing painting.
 *
 * Revision 1.11  1998/10/02 22:13:52  xzhu
 * xzhu on Fri Oct 2 15:13:42 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Added new function.
 *
 * Revision 1.10  1998/07/28 23:12:43  xzhu
 * xzhu on Tue Jul 28 16:12:06 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Moved ScaledCrop and DeskewImage from autoscan to autoproc.
 *
 * Revision 1.9  1998/07/24 19:46:26  xzhu
 * xzhu on Fri Jul 24 12:45:55 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Implemented new function for AIE.
 *
 * Revision 1.8  1998/06/30 22:15:11  xzhu
 * xzhu on Tue Jun 30 15:14:25 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Using findBgThreshAroundText replace findBackgroundThresh.
 *
 * Revision 1.7  1998/06/23 18:13:04  xzhu
 * xzhu on Tue Jun 23 11:11:58 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Use findBgThreshAroundText to fix segmentation problem reported by Tim.
 *
 * Revision 1.6  1998/06/18 19:25:18  xzhu
 * xzhu on Thu Jun 18 12:24:14 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Moved in clean text function from autoscan.
 *
 * Revision 1.5  1998/06/12 00:48:23  xzhu
 * xzhu on Thu Jun 11 17:47:17 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Added new functions to blaste masks and paste pictures.
 *
 * Revision 1.4  1998/05/26 23:38:46  xzhu
 * xzhu on Tue May 26 16:37:29 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Check both maskCount and point of masks before cutting image.
 *
 * Revision 1.3  1998/05/26 16:30:41  xzhu
 * xzhu on Tue May 26 09:29:56 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Assign contant as float.
 *
 * Revision 1.2  1998/05/22 22:28:18  xzhu
 * xzhu on Fri May 22 15:28:04 PDT 1998
 *
 * autoproc.c -> /product/ipcore/ipwrap/src/RCS/autoproc.c,v
 *
 * Added new function.
 *
 * Revision 1.1  1998/05/19 17:57:11  xzhu
 * Initial revision
 * 
 */
