/*<HTML><HEAD><TITLE>backgrnd.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* 
 * backgrnd.c
 *
 * Functions defined:
 *	w_backgroundToWhite	given the background plane of a color
 *				image, attempt to fixup the background
 *				so that it is white
 *	w_backgroundChromThresh	given the background plane of a color
 *				image, attempt to fixup the background
 *				so that it is white by thresholding
 *				based on chrominance
 *	w_backgroundLighten	given the background plane of a color,
 *				image, attempt to lighten the background
 *				by modifying the luminance channel
 */

#include "ipwcore.prv"
#include "defines.pub"
#include "thrshold.h"
#include "imageref.pub"
#include "geomadt.h"
#include "morphg.h"
#include "logicops.h"
#include "frames.pub"

#include "thrshold.pub"
#include "props.pub"
#include "rast.pub"
#include "curves.pub"
#include "grbinop.pub"
#include "math.h"
#include "prxform.pub"
#include "backgrnd.pub"
#include "strel.pub"
#include "morph.pub"
#include "color.pub"
#include "shrcc.pub"
#include "fill.pub"
#include "logicops.pub"
#include "curves.pub"
#include "bctable.pub"
#include "floyd.pub"
#include "rectops.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: backgrnd.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:14 $");

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Attempt to fixup the background so that it is max white.
 *
 * LongDescription:::
 *  given the background plane of a color image, attempt to fixup the 
 *  background so that it is max white.  this is done by trying to find
 *  the max chrominance of the background and thresholding at this value.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		State structure
 *  bgPixr::		background image
 *  chromThreshold::	if given, pixels with chroma values above this
 *			value with be considered colored (i.e. tint
 *			regions) and those below will be background.
 *			if -1, a threshold will be automatically 
 *			determined
 *  
 * Output:::
 *  dPixr::		background image with background set to white
 *			if *dPixr is NULL and image will be allocated.
 *			*dPixr may be equal to bgPixr.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  Assumes that the background has be images blasted away (i.e. they
 *  are max white).
 *
 * FunctionCategory:::
 *   automatic enhancement
 *
 * Keywords:::
 *   color, multi-channel
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  jcraig:
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_backgroundChromThresh (
    IPWState		*pIPWState,
    PIXR		*bgPixr,
    Int32		 chromThreshold,
    PIXR		**dPixr)
{
PROLOGUE;
void		*pState = NULL;	/* starte block for rgb to lab conversion */
Int32		 red, green, blue, index;
Float64		 L, a, b;
PIXR		*chromaPixr = NULL, *maskPixr = NULL, *fillPixr = NULL;
PIXR		*tPixr = NULL, *t1Pixr = NULL, *newBGPixr = NULL;
Int32		*chromaHisto = NULL;
Int32		 chromaVal, chromaMode;
Int32		 width, height, w, h, bpl;
ScaledInt1616	xdpi, ydpi;
UInt8		*pR, *pG, *pB, *pD;
UInt8		*chromaTable;
STREL		*strel = NULL;

    ERR_GEN_PIXR(bgPixr, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    if (*dPixr != NULL)
	ERR_GEN_PIXR(*dPixr, cByteOrderBig, cDepth8, cChannels3, cFrame32);

    /* make a table to convert RGB to Lab.  
     * what should the white point be set to?  first, note that the
     * background image has some max white regions where the images were
     * removed.  we could remember to ignore them everywhere but that is
     * a pain.  second, using the max value has seemed to give the best
     * range in the test images that we have.  this, of course, could
     * be wrong for other images.  In addition, since we may have an 
     * uncalibrated scanner, we'll just go with these values.
     */
    ERR_PASS(w_RGBToLabInit(pIPWState, 255.0, 255.0, 255.0, &pState));
    chromaTable = (UInt8 *)IPW_MALLOC(4096, cAllocatorData);
    ERR_GEN(chromaTable == NULL, cErrNomem);
    for (red = 0; red < 256; red += 0x10)
	for (green = 0; green < 256; green += 0x10)
	    for (blue = 0; blue < 256; blue += 0x10)
	    {
		index = (red << 4) | green | (blue >> 4);
		ERR_PASS_CLEANUP(w_RGBToLabProcess(pIPWState, red, green,  
					           blue, pState, &L, &a,
						   &b));
		chromaVal = ((Int32)sqrt(a * a + b * b));
		chromaTable[index] = MIN(chromaVal, 255);
	    }
    ERR_PASS(w_RGBToLabFinish(pIPWState, &pState));

    /* make the chroma image */
    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, bgPixr, &width));
    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, bgPixr, &height));
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, bgPixr, &bpl));
    ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, bgPixr, &xdpi,
					    &ydpi));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cRedPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pR));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cGreenPlane, 
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pG));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cBluePlane, 
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pB));
    ERR_PASS_CLEANUP(w_createPixrNoInit(pIPWState, width, height, 8,
				        xdpi, ydpi, cFrameBits, 
					&chromaPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, chromaPixr, &pD));

    for (h = 0; h < height; h++, pR += bpl, pG += bpl, pB += bpl, pD += bpl)
	for (w = 0; w < width; w++)
	    UCHAR_ACCESS(&pD[w]) = 
		chromaTable[((UCHAR_ACCESS(&pR[w]) & 0xF0) << 4) | 
			    (UCHAR_ACCESS(&pG[w]) & 0xF0) |
			    (UCHAR_ACCESS(&pB[w]) >> 4)];

    /* if the caller did not specify a threshold, find one.  otherwise,
     * just use the threshold from the caller.
     */
    if (chromThreshold == -1)
    {
   	 /* analyze the histogram to try to find the background */
    	ERR_PASS_CLEANUP(w_histo(pIPWState, chromaPixr, NULL, 
				 &chromaHisto));
    	/* the region that corresponds to images has been to the maximum
    	 * white which results in a chroma of 0.  so, don't look for a
     	* maximum at 0.
    	 */
    	ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, chromaHisto, 1, 255,
				    	     &chromaMode));
#if 0
    	ERR_PASS_CLEANUP(w_findLocalMinimum(pIPWState, chromaHisto, 256,   
				      	    chromaMode, 255, 1, 5,
				            &chromThreshold));
#endif
    	ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, chromaHisto,    
				          chromaMode, 255, 1, 0.05, 5, 1.5,
				          &chromThreshold));
    }

    /* now threshold the image at the max point of the background peak. */
    ERR_PASS_CLEANUP(w_threshold(pIPWState, chromaPixr, 0, chromThreshold,
			         THRESH_INVERSE, &maskPixr));

    /* and try to fill in any holes in the mask.
     * first try to close up any small holes that might cause some
     * leakage during the edge fill. then use the edge fill to get a 
     * solid region.
     */
    ERR_PASS_CLEANUP(w_initStrelByName(pIPWState, strel7h));
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, strel7h, &strel));
    ERR_PASS_CLEANUP(w_close(pIPWState, maskPixr, strel, &maskPixr));
    strel = NULL;
    ERR_PASS_CLEANUP(w_initStrelByName(pIPWState, strel7v));
    ERR_PASS_CLEANUP(w_getStrel(pIPWState, strel7v, &strel));
    ERR_PASS_CLEANUP(w_close(pIPWState, maskPixr, strel, &maskPixr));
    strel = NULL;
    ERR_PASS_CLEANUP(w_edgeFillSweep4(pIPWState, maskPixr, 100, &fillPixr));
    ERR_PASS_CLEANUP(w_invert(pIPWState, fillPixr, &fillPixr));
    ERR_PASS_CLEANUP(w_logicOp(pIPWState, fillPixr, UNION, maskPixr));

    /* now blast the background to white */
    if (*dPixr == NULL)
	ERR_PASS_CLEANUP(w_createPixrExtendedNoInit(pIPWState, width,
						    height, 8, xdpi,
						    ydpi, cFrameBits, 3, 
						    cColorRGB, NULL,
						    &newBGPixr));
    else newBGPixr = *dPixr;
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cRedPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cRedPlane,
				      &t1Pixr));
    ERR_PASS_CLEANUP(w_opGrayBin(pIPWState, tPixr, maskPixr, 
				 PIX_SRC & PIX_DST, &t1Pixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cGreenPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cGreenPlane, 
				     &t1Pixr));
    ERR_PASS_CLEANUP(w_opGrayBin(pIPWState, tPixr, maskPixr, 
				 PIX_SRC & PIX_DST, &t1Pixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cBluePlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cBluePlane,
				      &t1Pixr));
    ERR_PASS_CLEANUP(w_opGrayBin(pIPWState, tPixr, maskPixr, 
				 PIX_SRC & PIX_DST, &t1Pixr));

    /* cleanup and get outta here */
    if (*dPixr == NULL)
	*dPixr = newBGPixr;
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &chromaPixr));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &maskPixr));
    if (chromaTable != NULL) IPW_FREE(chromaTable, cAllocatorData);
    if (chromaHisto != NULL) IPW_FREE(chromaHisto, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &chromaPixr);
    w_destroyPixr(pIPWState, &maskPixr);
    w_destroyPixr(pIPWState, &newBGPixr);
    if (chromaTable != NULL) IPW_FREE(chromaTable, cAllocatorData);
    if (chromaHisto != NULL) IPW_FREE(chromaHisto, cAllocatorData);
    ERR_RETURN;
}

#define c114Offset	0
#define c886Offset	c114Offset + 256 * sizeof(UInt8)
#define c299Offset	c886Offset + 256 * sizeof(UInt8)
#define c701Offset	c299Offset + 256 * sizeof(UInt8)
#define c413Offset	c701Offset + 256 * sizeof(UInt8)
#define c587Offset	c413Offset + 256 * sizeof(UInt8)
/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Make tables used in w_backgroundLighten()
 *
 * LongDescription:::
 *  w_backgroundLighten converts an RGB image to luminance (YCrCb space),
 *  and remaps the luminance values.  it then needs to take generate
 *  another RGB background image.  so it needs the new luminance values
 *  and CrCb to convert back.  there is no need to convert completely
 *  to CrCb.  we can compute the new RGB values from Y and the original
 *  values.  furthermore, we do not want to multiply in the inner loop.
 *  this procedure produces 6 tables concatenated together which have
 *  the appropriate constants times RGBfor all possible RGB values.
 *  Note that these tables are UInt8s so we have lost some precision
 *  in the conversion.  The RGBToY conversion did the same so it didn't
 *  seem appropriate to be extremely accurate here.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		State structure
 *  
 * Output:::
 *  convTable::		table of constants mulitplied by all possible
 *			RGB values.  Must be NULL on entry.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   automatic enhancement
 *
 * Keywords:::
 *   color, multi-channel
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  jcraig:
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
static ErrCode CDECL
wp_makeBackgroundLightenTables(
    IPWState		*pIPWState,
    UInt8		**convTable)
{
PROLOGUE;
UInt8	*table;
Int32	 i;
UInt8	 entry;

    /* 6 tables with 256 entries of Uint8 */
    ERR_GEN(*convTable != NULL, cErrInvalidParm);
    table = (UInt8 *)IPW_MALLOC(6 * 256 * sizeof(UInt8),
	        cAllocatorData);
    ERR_GEN(table == NULL, cErrNomem);

    for (i = 0; i < 256; i++)
    {
	entry = (UInt8)(0.11400 * i + 0.5);
	table[i+c114Offset] = entry;
	table[i+c886Offset] = i - entry;

	entry = (UInt8)(0.29900 * i + 0.5);
	table[i+c299Offset] = entry;
	table[i+c701Offset] = i - entry;

	entry = (UInt8)(0.41300 * i + 0.5);
	table[i+c413Offset] = entry;
	table[i+c587Offset] = i - entry;
    }

    *convTable = table;
    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Attempt to lighten the background.
 *
 * LongDescription:::
 *  given the background plane of a color image, attempt to fixup the 
 *  is much lighter.  this is done by trying to find the 
 *  luminance of the background and stretching the luminance values so
 *  that pixels above that value have their luminance set to the max
 *  and those below that value have a linear stretch done.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::		State structure
 *  bgPixr::		background image
 *  lumThreshold::	if given, pixels with luminance values above this
 *			value will be considered background and their
 *			luminance set to the max while those pixels with
 *			luminance below this value will have a linear
 *			stretch done.
 *			if -1, a threshold will be automatically 
 *			determined
 *  
 * Output:::
 *  dPixr::		background image with background lightened

 *			if *dPixr is NULL and image will be allocated.
 *			*dPixr may be equal to bgPixr.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  Assumes that the background has be images blasted away (i.e. they
 *  are max white).
 *
 * FunctionCategory:::
 *   automatic enhancement
 *
 * Keywords:::
 *   color, multi-channel
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  jcraig:
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_backgroundLighten(
    IPWState		*pIPWState,
    PIXR		*bgPixr,
    Int32		 lumThreshold,
    PIXR		**dPixr)
{
PROLOGUE;
PIXR		*lumPixr = NULL, *newBGPixr = NULL, *tPixr = NULL;
ScaledInt1616	 xdpi, ydpi;
UInt8		*pR, *pG, *pB, *pL, *pNewR, *pNewB, *pNewG;
Int32		*lumHisto = NULL;
UInt8		*trc = NULL;
Int32		 h, w, width, height, bpl, newBpl;
UInt8		*convTable = NULL;

    ERR_GEN_PIXR(bgPixr, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    if (*dPixr != NULL)
	ERR_GEN_PIXR(*dPixr, cByteOrderBig, cDepth8, cChannels3, cFrame32);

    /* we need to convert to luminance.  the easiest/fastest way to
     * do this is to use the YCrCb color space.  
     * Note that because the tables used in this procedure are
     * composed of UInt8's max white will have a luminance of 254.
     * this is important because the images have been blasted to 255
     * and we need to ignore these when analyzing the histogram.
     */
    ERR_PASS_CLEANUP(w_RGBToY(pIPWState, bgPixr, &lumPixr));

    if (lumThreshold == -1)
    {

    	ERR_PASS_CLEANUP(w_histo(pIPWState, lumPixr, NULL, &lumHisto));
    	ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, lumHisto, 0, 253,
				    	     &lumThreshold));
#if 0
    	ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, lumHisto,    
				          lumMode, 0, -1, 0.10, 5,
				          &lumThreshold));
#endif
    }

    ERR_PASS_CLEANUP(w_makeLinearTable(pIPWState, 256, 0, lumThreshold,
				       &trc));

    /* now get the new RGB image. */
    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, bgPixr, &width));
    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, bgPixr, &height));
    if (*dPixr == NULL)
     {
    	ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(pIPWState, bgPixr, &xdpi,
					        &ydpi));
	ERR_PASS_CLEANUP(w_createPixrExtendedNoInit(pIPWState, width,
						    height, 8, xdpi,
						    ydpi, cFrameBits, 3, 
						    cColorRGB, NULL,
						    &newBGPixr));
    }
    else newBGPixr = *dPixr;

    /* get pointers to all the images */
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cRedPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pR));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cGreenPlane, 
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pG));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cBluePlane, 
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pB));

    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cRedPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pNewR));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cGreenPlane, 
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pNewG));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cBluePlane, 
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tPixr, &pNewB));
    ERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, lumPixr, &pL));

    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, bgPixr, &bpl));
    /* note that if we allocated this the bpl's should be the same
     * but if the pixr was already allocated it may be different.
     */
    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, newBGPixr, &newBpl));
    ERR_PASS_CLEANUP(wp_makeBackgroundLightenTables(pIPWState, &convTable));
    for (h = 0; h < height; 
	 h++, pR += bpl, pG += bpl, pB += bpl, pL += bpl,
	 pNewR += newBpl, pNewG += newBpl, pNewB += newBpl)
	for (w = 0; w < width; w++)
    {
    UInt8	 Y, red, green, blue;

	red = UCHAR_ACCESS(&pR[w]);
	green = UCHAR_ACCESS(&pG[w]);
	blue = UCHAR_ACCESS(&pB[w]);
	Y = UCHAR_ACCESS(&pL[w]);

	/* start with the equations:
	 * Cb = -0.16874 * R - 0.33126 * G + 0.50000 * B
	 * Cr =  0.50000 * R - 0.41869 * G - 0.08131 * B
	 *
	 * R' = Y                + 1.40200 * Cr
	 * G' = Y - 0.34414 * Cb - 0.71414 * Cr
	 * B' = Y + 1.77200 * Cb
	 *
	 * so
	 * R' = Y + 1.40200 (0.50000 * R - 0.41869 * G - 0.08131 * B)
	 *    = Y + 0.70100 * R - 0.58700 * G - .11400 * B
	 *
	 * G' = Y - 0.34414 (-0.16874 * R - 0.33126 * G + 0.50000 * B) 
		  - 0.71414 (0.50000 * R - 0.41869 * G - 0.08131 * B)
	 *    = Y + 0.05807 * R + 0.11400 * G - 0.17207 * B
		  - 0.35707 * R + 0.29900 * G + 0.05807 * B
	 *    = Y - 0.29900 * R + 0.41300 * G - 0.11400 * B
	 *
	 * B' = Y + 1.77200 (-0.16874 * R - 0.33126 * G + 0.50000 * B)
	 *    = Y - 0.29900 * R - 0.58700 * G + 0.88600 * B
	 */

	/* now find the new values and put them in place.
	 * note that since Y has been modified we need to make sure
	 * that a pixel value does not exceed 255.  this is the
	 * slow way but it works for now....
	 */
	UCHAR_ACCESS(&pNewR[w]) = 
	    MIN(trc[Y] + convTable[c701Offset+red] -
	          convTable[c587Offset+green] - convTable[c114Offset+blue],
		255);
	UCHAR_ACCESS(&pNewG[w]) = 
	    MIN(trc[Y] - convTable[c299Offset+red] +
	          convTable[c413Offset+green] - convTable[c114Offset+blue],
		255);
	UCHAR_ACCESS(&pNewB[w]) = 
	    MIN(trc[Y] - convTable[c299Offset+red] -
	          convTable[c587Offset+green] + convTable[c886Offset+blue],
		255);
    }

    if (*dPixr == NULL) *dPixr = newBGPixr;

    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &lumPixr));
    if (trc != NULL) IPW_FREE(trc, cAllocatorData);
    if (lumHisto != NULL) IPW_FREE(lumHisto, cAllocatorData);
    if (convTable != NULL) IPW_FREE(convTable, cAllocatorData);
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr(pIPWState, &lumPixr);
    IPW_FREE(trc, cAllocatorData);
    IPW_FREE(lumHisto, cAllocatorData);
    IPW_FREE(convTable, cAllocatorData);
    ERR_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Attempt to fixup the background so that it is max white.
 *
 * LongDescription:::
 *  given the background plane of a color image, attempt to find the 
 *  background and make it white.  if no thresholds are given, this 
 *  procedure attempts to analyze a histogram and find one.  it assumes
 *  that the background is the largest part of the page.  if this is
 *  not the case, e.g. a tint region occupies most of the page, the
 *  tint region will be recognized as background and blasted away.
 *
 * PrivateDescription:::
 *  dealing with 3 planes of an RGB image is a pain.  it would be nice
 *  if we could convert the RGB image into a gray image.  previous
 *  attempts at this function have used chrominance and luminance.  
 *  these have not worked too well since the background and the tint
 *  regions are easily distinguished.  this version attempts something
 *  different.  it converts the image to gray by dithering the RGB
 *  image to a color mapped image where the color map is constructed
 *  to be a "nice gray".  dithering seems to both smooth and quantize
 *  the image.  this helps determine the thresholds from the histogram
 *  and hence find the tint regions. 
 *
 * Input:::
 *  pIPWState::		State structure
 *  bgPixr::		background image
 *  imPixr::		image mask (binary)
 *  selPixr::		selector (binary - with images removed)
 *  bwTables::		pointer to structure which contains tables
 *			used by this routine.  if NULL, tables will
 *			be made and destroyed by this routine.  otherwise,
 *			make tables with w_makeBackgroundToWhiteTables()
 *			and destroy with w_destroyBackgroundToWhiteTables().
 *			these tables will be the same for every call to
 *			this procedure.
 *  backLowThresh::	lower threshold 
 *			if -1, a threshold will be automatically 
 *			determined
 *  backHighThresh::	upper threshold 
 *			if -1, a threshold will be automatically 
 *			determined
 *  ccWidth::		minimum width of a feature to consider it to be tint
 *  ccHeight::		minimum height of feature to consider it to be tint
 *  
 * Output:::
 *  dPixr::		background image with background lightened
 *			if *dPixr is NULL and image will be allocated.
 *			*dPixr may be equal to bgPixr.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  Assumes that the background has the images blasted away (i.e. they
 *  are max white).
 *  imPixr may be NULL if there are no pictures in the background.  or,
 *  it may be a completely white.
 *  Assumes that the background, selector and image mask are all the 
 *  same resolution.
 *  ccWidth and ccHeight should be fairly small.  My initial tests have
 *  been done with ccWidth = ccHeight = 5.
 *
 * FunctionCategory:::
 *   automatic enhancement
 *
 * Keywords:::
 *   color, multi-channel
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  jcraig:
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_backgroundToWhite(
    IPWState		*pIPWState,
    PIXR		*bgPixr,
    PIXR		*imPixr,
    PIXR		*selPixr,
    BackgroundWhiteInfo	*bwTables,
    Int32		 backLowThresh,
    Int32		 backHighThresh,
    UInt32		 ccWidth,
    UInt32		 ccHeight,
    PIXR		**dPixr)
{
PROLOGUE;
Int32		*dHisto = NULL, *rectHisto = NULL;
Int32		 smallHisto[256], rSmallHisto[256];
Int32		 histoIndex[256], rHistoIndex[256];
PIXR		*ditherPixr = NULL, *threshPixr = NULL, *newBGPixr = NULL;
PIXR		*rPixr = NULL, *tPixr = NULL, *t1Pixr = NULL, *sPixr = NULL;
Int32		 backMode, rectMode, i, redMode, greenMode, blueMode; 
Int32		 nextEntry, rNextEntry, totalPixels, peakPixels;
RECTANGL	*cRect = NULL, *rectList = NULL;
Float64		 backHue;
Bool		 allocatedHere = cFalse;
PIXR            *prMask = NULL;

    ERR_GEN_PIXR(bgPixr, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    if (*dPixr != NULL)
	ERR_GEN_PIXR(*dPixr, cByteOrderBig, cDepth8, cChannels3, cFrame32);
    if (imPixr != NULL)
    {
	ERR_GEN_PIXR(imPixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
	ERR_PASS_CLEANUP (w_invert (pIPWState, imPixr, &prMask));
    }
    else /* generate mask image */
    {
	Int32 i;
	UInt8      map[256];

	for(i=0; i<256; ++i)
	    map[i] = (UInt8)0 ;
	map[255] = (UInt8) 128;

	ERR_PASS_CLEANUP (w_map(pIPWState, bgPixr, &(map[0]), &sPixr));

	ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, sPixr, cRedPlane,
		   		      &tPixr));	
        ERR_PASS_CLEANUP(w_threshold(pIPWState, tPixr, 128, 128, 
				     THRESH_INVERSE, &prMask));

	ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, sPixr, cGreenPlane,
		   		      &tPixr));	
        ERR_PASS_CLEANUP(w_threshold(pIPWState, tPixr, 128, 128, 
				     THRESH_INVERSE, &t1Pixr));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, t1Pixr, OR, prMask));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t1Pixr));

	ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, sPixr, cBluePlane,
		   		      &tPixr));	
        ERR_PASS_CLEANUP(w_threshold(pIPWState, tPixr, 128, 128, 
				     THRESH_INVERSE, &t1Pixr));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, t1Pixr, OR, prMask));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &t1Pixr));

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &sPixr));
    }

    /* create a gray image by dithering.  the gives us an image which
     * is fairly smooth and the pixel values have been quantized.
     */
    if (bwTables == NULL)
    {
	allocatedHere = cTrue;
	ERR_PASS_CLEANUP(w_makeBackgroundToWhiteTables(pIPWState,
						       &bwTables));
    }
    ERR_PASS_CLEANUP(w_ditherFloyd24ToCMap(pIPWState, bgPixr, NULL,
                          		   bwTables->drcTable,
					   bwTables->drcTable,
					   bwTables->drcTable,
                          		   bwTables->colorMap,
					   bwTables->colorLut, 0, 8,
					   &ditherPixr));

    /* if the left and right edges of the background peak have not been
     * given, go determine them.
     */
    if ((backLowThresh == -1) || (backHighThresh == -1))
    {
    	ERR_PASS_CLEANUP(w_histoMask(pIPWState, ditherPixr, prMask, 
				     NULL, &dHisto));

        /* compact the histogram into a smaller one because there are many
         * empty entries in the larger one and w_findPeakExtent() will 
         * stop if it finds a zero entry.
         */
        nextEntry = 0;
	totalPixels = 0;
        for (i = 0; i < 256; i++)
	    if (dHisto[i])
	    {
	        smallHisto[nextEntry] = dHisto[i];
	        histoIndex[nextEntry] = i;
	        nextEntry++;
		totalPixels += dHisto[i];
 	    }
        IPW_FREE(dHisto, cAllocatorData);

	smallHisto[nextEntry] = 0;
	histoIndex[nextEntry] = 241;

	/* No background to be whiten */
	if( nextEntry < 2)
	{
	    ERR_PASS_CLEANUP(w_copy(pIPWState, bgPixr, dPixr));
	    if (allocatedHere == cTrue)
		ERR_PASS_CLEANUP(w_destroyBackgroundToWhiteTables(pIPWState,
						          &bwTables));
	    if( prMask  ) 
		ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prMask));
	    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &ditherPixr));
	    NORMAL_RETURN;
	}

	/* the last entry in smallHisto/dHisto corresponds to max white.
	 * we do not want to include that in our count of total pixels
	 * in the area because images are set to max white.  so remove
	 * that number of pixels from the count.
	 */
	/* 4-14-98, generate mask first, the image has be masked out,
	   here, we don't need to skip last entry */
	/*
	totalPixels -= smallHisto[nextEntry-1];
	*/

        /* now analyze it.  note that nextEntry points one beyond the last
         * valid entry.  in addition, the last entry corresponds to white
         * so we actually want to stop at nextEntry-2 since we do not
         * wish to consider white (because the images have been blasted to 
         * white.
         */
        ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, smallHisto, 0,
		     			     nextEntry-1, &backMode));
	ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, smallHisto, backMode,
				          0, -1, 0.25, 3, 1.5,
					  &backLowThresh));
	ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, smallHisto, backMode,
				          nextEntry-1, 1, 0.25, 3, 1.5, 
				          &backHighThresh));
	/* check to make sure that thresholds have been found */
	if (backHighThresh == -1)
	    backHighThresh = nextEntry;
	if (backLowThresh == -1)
	    backLowThresh = 0;

	/* magic number time....  assume that the background contains
	 * more pixels than any tint region (which does cause some 
	 * images to fail).  however, if a tint region is very smooth
	 * the global maximum may be a tint region instead of the
	 * background especially since the background tents to have
	 * lots of bleedthru and so is dispersed amount many pixel
	 * intensity values.  so check total number of pixels in the peak
	 * and determine if enough are included to be considered background.
    	 */
	for (i = backLowThresh + 1, peakPixels = 0; i < backHighThresh; i++)
	    peakPixels += smallHisto[i];
	if (peakPixels < (totalPixels / 2))
	{
	Int32	 modeLeft, modeRight, newLowThresh, newHighThresh;
	Int32	 newPeakPixels;

	    /* then we need to find a new peak.  look for a new peak on
	     * both the left and the right of the current one.  choose
	     * the highest one.
	     */
	    if (backMode > 0)
	        ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, smallHisto, 
			    			     0, backMode-1,
						     &modeLeft));
	    else 
		modeLeft = -1;
	    if (backMode < (nextEntry-1))
	        ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, smallHisto, 
			    			     backMode+1,
						     nextEntry-1,
						     &modeRight));
	    else
		modeRight = -1;
	    if (((modeLeft != -1) && (modeRight != -1) &&
		 (smallHisto[modeLeft] > smallHisto[modeRight])) ||
		((modeLeft != -1) && (modeRight == -1)))
	    {
		/* check the left peak */
		ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, smallHisto,
						  modeLeft, 0, 
		     	     			  -1, 0.25, 3, 1.5, 
						  &newLowThresh));
		ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, smallHisto,
						  modeLeft, backMode-1, 
		     	     			  1, 0.25, 3, 1.5,
						  &newHighThresh));
		if (newHighThresh == -1)
		    newHighThresh = backMode-1;
		if (newLowThresh == -1)
		    newLowThresh = 0;
		/* check to make sure that the peak that we have found
		 * is better than the previous one.  if not, just go
		 * with the previous guess.
		 */
		for (i = newLowThresh + 1, newPeakPixels = 0; 
		     i < newHighThresh; i++)
	    	    newPeakPixels += smallHisto[i];
		if (newPeakPixels > peakPixels)
		{
		    backHighThresh = newHighThresh;
		    backLowThresh = newLowThresh;
	            backMode = modeLeft;
		}
	    }
	    /* just to make sure that the case of modeLeft and
	     * modeRight both equal -1 gets handled correctly.  this
	     * image would be really strange....
	     */
	    else if (((modeLeft != -1) && (modeRight != -1) &&
		      (smallHisto[modeLeft] <= smallHisto[modeRight])) ||
		     ((modeLeft == -1) && (modeRight != -1)))
	    {
		/* check the right peak */
    	        ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, smallHisto,
						  modeRight, backMode+1, 
		     	     			  -1, 0.25, 3, 1.5,
						  &newLowThresh));
    	        ERR_PASS_CLEANUP(w_findPeakExtent(pIPWState, smallHisto,
						  modeRight, nextEntry-1, 
		     	     			  1, 0.25, 3, 1.5,
						  &newHighThresh));
		if (newHighThresh == -1)
		    newHighThresh = nextEntry-1;
		if (newLowThresh == -1)
		    newLowThresh = backMode+1;
		/* check to make sure that the peak that we have found
		 * is better than the previous one.  if not, just go
		 * with the previous guess.
		 */
		for (i = newLowThresh + 1, newPeakPixels = 0; 
		     i < newHighThresh; i++)
	    	    newPeakPixels += smallHisto[i];
		if (newPeakPixels > peakPixels)
		{
		    backHighThresh = newHighThresh;
		    backLowThresh = newLowThresh;
	            backMode = modeRight;
		}
	    }
	}
        ERR_PASS_CLEANUP(w_threshold(pIPWState, ditherPixr, 
				     histoIndex[backLowThresh]+2, 
				     histoIndex[backHighThresh], 
				     THRESH_INVERSE, &threshPixr));
    }
    else	/* just use the thresholds passed in */
        ERR_PASS_CLEANUP(w_threshold(pIPWState, ditherPixr, 
				     backLowThresh, backHighThresh,
				     THRESH_INVERSE, &threshPixr));

    if( prMask ) 
	ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &prMask));

    /* and get rid of the images */
    if (imPixr != NULL)
        ERR_PASS_CLEANUP(w_logicOp(pIPWState, imPixr, SUBTRACT, 
				   threshPixr));

    /* get the selector ready in case it is needed.
     * why is the selector used?  sometimes regions show up in the
     * background when the selector says get it from the foreground.
     * the most common case of this is reverse video.  the background
     * usually has the reverse video tint region and the selector 
     * indicates that only the text should come from the background.
     * now if the text is not white or other regions touch the reverse
     * video, we may ignore them because the reverse video overwhelms
     * the statistics.  so, we do not want to consider it when we are
     * taking the histogram.  hence, we need the selector plane.
     * why not just blast it from the thresholded pixr?  if we did that
     * we would remove the text as well.  this would leave holes in the
     * text on tint region.  this might cause them to fail the test 
     * which determines if they are close enough to rectangular and 
     * fills in the entire rectangle if so.
     * this could be solved in 2 ways both of which take time and so
     * have not yet been implemented.  it may become necessary later.
     * 1) look for all the text in the selector and remove it.  
     * 2) before determining if the region is rectangular, count the
     * bits in the region in the selector and don't consider them when
     * counting pixels.  we may also have to do this for the image mask
     * since an image in the midddle of the tint region can make the
     * region appear non-rectangular since the counts are off.
     */
    ERR_PASS_CLEANUP(w_invert(pIPWState, selPixr, &sPixr));
    if (imPixr != NULL)
	ERR_PASS_CLEANUP(w_logicOp(pIPWState, imPixr, XOR, sPixr));

    /* now determine the hue of the background.  we may need this value
     * when filtering connected components.  note that we are using 
     * the hue calculation from the HSV color space ("Compute Graphics
     * Principles and Practice" by Foley and VanDam).  I guess that we
     * should really use hue calculated from the L*a*b* or L*u*v*
     * color space but this one is much easier to compute....
     * first, find the mode of each of the RGB planes.  Note that we
     * are just reusing the variable dHisto instead of having yet 
     * another one around.
     */
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cRedPlane,
		   		      &tPixr));
    ERR_PASS_CLEANUP(w_histo(pIPWState, tPixr, NULL, &dHisto));
    ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, dHisto, 0, 254, 
					 &redMode));
    IPW_FREE(dHisto, cAllocatorData);

    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cGreenPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_histo(pIPWState, tPixr, NULL, &dHisto));
    ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, dHisto, 0, 254, 
					 &greenMode));
    IPW_FREE(dHisto, cAllocatorData);

    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cBluePlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_histo(pIPWState, tPixr, NULL, &dHisto));
    ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, dHisto, 0, 254, 
					 &blueMode));
    IPW_FREE(dHisto, cAllocatorData);

    if ((redMode >= greenMode) && (redMode >= blueMode))
    {
	if (greenMode == blueMode)
	    backHue = 0;
	else if (greenMode > blueMode)
	    backHue = (Float64)(greenMode - blueMode) / 
		      (Float64)(redMode - blueMode);
	else
	    backHue = (Float64)(greenMode - blueMode) / 
		      (Float64)(redMode - greenMode);
    }
    else if ((greenMode >= redMode) && (greenMode >= blueMode))
    {
	if (redMode == blueMode)
	    backHue = 0;
	else if (redMode > blueMode)
	    backHue = (Float64)(blueMode - redMode) / 
		      (Float64)(greenMode - blueMode) + 2;
	else 
	    backHue = (Float64)(blueMode - redMode) / 
		      (Float64)(greenMode - redMode) + 2;
    }
    else
    {
	if (redMode == greenMode)
	    backHue = 0;
	else if (redMode > greenMode)
	    backHue = (Float64)(redMode - greenMode) /
		       (Float64)(blueMode - redMode) + 4;
	else
	    backHue = (Float64)(redMode - greenMode) /
		       (Float64)(blueMode - redMode) + 4;
    }
    backHue = backHue * 60;
    if (backHue < 0)
	backHue += 360;

    /* now look at each of the connected components and determine if
     * we want to keep them in the final mask.
     * assume that tint regions are usually rectangular and may have
     * small holes that need to get filled in.
     * first, determine if a connected component is mostly rectangular
     * and filled in.  if so, fill it in entirely in the final mask.
     * if it is not, look at the histogram.   if the mode is mostly
     * the same as the background, then ignore it. if the mode
     * falls exactly on the edges of the background range, look
     * at the hue and compare it to the background hue.
     * note this is useful since splotchy but connected regions will
     * include lots of the background in a rectangular area.  hopefully, 
     * this will overwhelm the connected component and so the mode will
     * indicate that it is background.
     */
    ERR_PASS_CLEANUP(w_rabb(pIPWState, threshPixr, ccWidth, ccHeight,
			    &rectList));
    ERR_PASS_CLEANUP(w_clonePixr(pIPWState, threshPixr, &rPixr));
    for (cRect = rectList; cRect != NULL; cRect = cRect->next)
    {
	if (cRect->mass >= (0.90 * (cRect->w * cRect->h)))
	{
	    /* then this has a good chance of being a rectangular tint
	     * region.
	     */
	    ERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, NULL, PIX_SET,
		       			cRect->x, cRect->y, cRect->w,
					cRect->h, rPixr));
	}
	else
	{
	    /* this may not be a tint region.  check out the histogram
	     * and see if the mode is in the background range.
	     */
	    ERR_PASS_CLEANUP(w_histoMask(pIPWState, ditherPixr, sPixr,
					 cRect, &rectHisto));
	    /* compress because of missing entries */
	    rNextEntry = 0;
    	    for (i = 0; i < 256; i++)
	        if (rectHisto[i])
	        {
	    	    rSmallHisto[rNextEntry] = rectHisto[i];
	    	    rHistoIndex[rNextEntry] = i;
	    	    rNextEntry++;
	        }
	    ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, rSmallHisto, 0,
			        		 rNextEntry-2, &rectMode));
	    if ((rHistoIndex[rectMode] < histoIndex[backLowThresh]) ||
		(rHistoIndex[rectMode] > histoIndex[backHighThresh]))
	    {
	        /* then this has a good chance of being a tint region. */
		ERR_PASS_CLEANUP(w_pixrPutPixel(pIPWState, rPixr, 
                                 		cRect->px, cRect->py, 1));
		ERR_PASS_CLEANUP(w_fillClipSweep8(pIPWState, rPixr,
						  threshPixr, 250,
						  &rPixr));
	    }
	    else if ((rHistoIndex[rectMode] == histoIndex[backLowThresh]) ||
		(rHistoIndex[rectMode] == histoIndex[backHighThresh]))
	    {
	    Float64	 rHue;

    		ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr,
						  cRedPlane, &tPixr));
    		ERR_PASS_CLEANUP(w_histo(pIPWState, tPixr, cRect,
					 &dHisto));
   		ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, dHisto, 0,
					 	     254, &redMode));
    		IPW_FREE(dHisto, cAllocatorData);

    		ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr,
						  cGreenPlane, &tPixr));
    		ERR_PASS_CLEANUP(w_histo(pIPWState, tPixr, cRect,
					 &dHisto));
    		ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, dHisto, 0,
					 	     254, &greenMode));
    		IPW_FREE(dHisto, cAllocatorData);

    		ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr,
						  cBluePlane, &tPixr));
    		ERR_PASS_CLEANUP(w_histo(pIPWState, tPixr, cRect,
					 &dHisto));
    		ERR_PASS_CLEANUP(w_findGlobalMaximum(pIPWState, dHisto, 0,
					 	     254, &blueMode));
    		IPW_FREE(dHisto, cAllocatorData);

		if ((redMode >= greenMode) && (redMode >= blueMode))
		{
		    if (greenMode == blueMode)
	    		rHue = 0;
		    else if (greenMode > blueMode)
	    		rHue = (Float64)(greenMode - blueMode) / 
		      	       (Float64)(redMode - blueMode);
		    else
	    		rHue = (Float64)(greenMode - blueMode) / 
		      	       (Float64)(redMode - greenMode);
		}
    		else if ((greenMode >= redMode) && (greenMode >= blueMode))
		{
		    if (redMode == blueMode)
	    		rHue = 0;
		    else if (redMode > blueMode)
	    		rHue = (Float64)(blueMode - redMode) / 
		      	       (Float64)(greenMode - blueMode) + 2;
		    else 
	    		rHue = (Float64)(blueMode - redMode) / 
		      	       (Float64)(greenMode - redMode) + 2;
		}
    		else
		{
		    if (redMode == greenMode)
			rHue = 0;
		    else if (redMode > greenMode)
	    		rHue = (Float64)(redMode - greenMode) /
		       	       (Float64)(blueMode - redMode) + 4;
		   else
	    		rHue = (Float64)(redMode - greenMode) /
		       	       (Float64)(blueMode - redMode) + 4;
		}
    		rHue = rHue * 60;
    		if (rHue < 0)
		    rHue += 360;

		/* 30 < backHue < 330 */
		if (backHue > 330)
		{
		    if ((rHue < (backHue - 30)) && 
		        (rHue > (30 - (360 - backHue))))
		    {
	                /* then call this a tint region. */
		        ERR_PASS_CLEANUP(w_pixrPutPixel(pIPWState, rPixr, 
                                 		        cRect->px, 
						        cRect->py, 1));
		        ERR_PASS_CLEANUP(w_fillClipSweep8(pIPWState, rPixr,
						          threshPixr, 250,
						          &rPixr));
		    }
		}
		else if (backHue < 30)
		{
		    if ((rHue > (backHue + 30)) &&
		        (rHue < (360 - (30 - backHue))))
		    {
		        ERR_PASS_CLEANUP(w_pixrPutPixel(pIPWState, rPixr, 
                                 		        cRect->px, 
						        cRect->py, 1));
		        ERR_PASS_CLEANUP(w_fillClipSweep8(pIPWState, rPixr,
						          threshPixr, 250,
						          &rPixr));
		    }
		}
		else if ((rHue > (backHue + 30)) || 
			 (rHue < (backHue - 30)))

		{
	            /* then call this a tint region. */
		    ERR_PASS_CLEANUP(w_pixrPutPixel(pIPWState, rPixr, 
                                 		    cRect->px, cRect->py,
						    1));
		    ERR_PASS_CLEANUP(w_fillClipSweep8(pIPWState, rPixr,
						      threshPixr, 250,
						      &rPixr));
		}
	    }
	    IPW_FREE(rectHisto, cAllocatorData);
	}
    }
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &sPixr));

    /* now blast the background to white */
    if (*dPixr == NULL)
	ERR_PASS_CLEANUP(w_clonePixrNoInit(pIPWState, bgPixr, &newBGPixr));
    else 
	newBGPixr = *dPixr;

    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cRedPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cRedPlane,
				      &t1Pixr));
    ERR_PASS_CLEANUP(w_opGrayBin(pIPWState, t1Pixr, rPixr, 
				 PIX_SRC & PIX_DST, &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cGreenPlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cGreenPlane,
				      &t1Pixr));
    ERR_PASS_CLEANUP(w_opGrayBin(pIPWState, t1Pixr, rPixr, 
				 PIX_SRC & PIX_DST, &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, newBGPixr, cBluePlane,
				      &tPixr));
    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, bgPixr, cBluePlane,
				      &t1Pixr));
   ERR_PASS_CLEANUP(w_opGrayBin(pIPWState, t1Pixr, rPixr, 
				 PIX_SRC & PIX_DST, &tPixr));

    /* cleanup and get outta here */
    if (*dPixr == NULL)
	*dPixr = newBGPixr;
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &sPixr));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &ditherPixr));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &threshPixr));
    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &rPixr));
    ERR_PASS_CLEANUP(w_destroyRectList(pIPWState, &rectList));
    IPW_FREE(dHisto, cAllocatorData);
    if (allocatedHere == cTrue)
	ERR_PASS_CLEANUP(w_destroyBackgroundToWhiteTables(pIPWState,
						          &bwTables));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (*dPixr == NULL) w_destroyPixr(pIPWState, &newBGPixr);
    if( prMask ) w_destroyPixr(pIPWState, &prMask);
    w_destroyPixr(pIPWState, &sPixr);
    w_destroyPixr(pIPWState, &ditherPixr);
    w_destroyPixr(pIPWState, &threshPixr);
    w_destroyPixr(pIPWState, &rPixr);
    w_destroyRectList(pIPWState, &rectList);
    IPW_FREE(dHisto, cAllocatorData);
    IPW_FREE(rectHisto, cAllocatorData);
    if (allocatedHere == cTrue)
	w_destroyBackgroundToWhiteTables(pIPWState, &bwTables);
    ERR_RETURN;
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Make tables used by w_backgroundToWhite().
 *
 * LongDescription:::
 *
 * Input:::
 *  pIPWState::		State structure
 *  
 * Output:::
 *  bwInfo::		pointer to pointer to structure for tables.
 *
 * Returns:::
 *  error status
 *
 * FunctionCategory:::
 *   automatic enhancement
 *
 * Keywords:::
 *   color, multi-channel
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  jcraig:
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_makeBackgroundToWhiteTables(
    IPWState		*pIPWState,
    BackgroundWhiteInfo	**bwInfo)

{
PROLOGUE;
BackgroundWhiteInfo	*bwTables;
Int32			 i;

    ERR_GEN(*bwInfo != NULL, cErrPtrNotNull);

    bwTables = 
	(BackgroundWhiteInfo *) IPW_CALLOC(1, sizeof(BackgroundWhiteInfo),
					   cAllocatorData);
    ERR_GEN(bwTables == NULL, cErrNomem);

    ERR_PASS_CLEANUP(w_createColorMap(pIPWState, 256, &bwTables->colorMap));

    bwTables->drcTable = (UInt8 *) IPW_MALLOC(256 * sizeof(UInt8),
					      cAllocatorData);
    ERR_GEN_CLEANUP(bwTables->drcTable == NULL, cErrNomem);

    for (i = 0; i < 256; i++)
    {
        bwTables->colorMap->map[i][cRedPlane] = (UInt8)i;
        bwTables->colorMap->map[i][cGreenPlane] = (UInt8)i;
        bwTables->colorMap->map[i][cBluePlane] = (UInt8)i;
	bwTables->drcTable[i] =  (UInt8)i;
    }

    ERR_PASS_CLEANUP(w_cMapToLut(pIPWState, bwTables->colorMap, cFalse,
				 &bwTables->colorLut));
    *bwInfo = bwTables;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(bwTables->colorLut, cAllocatorData);
    w_destroyColorMap(pIPWState, &bwTables->colorMap);
    IPW_FREE(bwTables->drcTable, cAllocatorData);
    ERR_RETURN;
}


/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Destroy tables made by w_makeBackgroundToWhiteTables().
 *
 * LongDescription:::
 *
 * Input:::
 *  pIPWState::		State structure
 *  
 * Output:::
 *  bwInfo::		pointer to pointer to structure for tables.
 *
 * Returns:::
 *  error status
 *
 * FunctionCategory:::
 *   automatic enhancement
 *
 * Keywords:::
 *   color, multi-channel
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  jcraig:
 *
 * Citations:::
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_destroyBackgroundToWhiteTables(
    IPWState		*pIPWState,
    BackgroundWhiteInfo	**bwInfo)

{
PROLOGUE;
BackgroundWhiteInfo	*bwTables;

    ERR_GEN(*bwInfo == NULL, cErrNullPtr);
    bwTables = *bwInfo;

    IPW_FREE(bwTables->colorLut, cAllocatorData);
    IPW_FREE(bwTables->drcTable, cAllocatorData);
    ERR_PASS_CLEANUP(w_destroyColorMap(pIPWState, &bwTables->colorMap));
    IPW_FREE(bwTables, cAllocatorData);

    *bwInfo = NULL;
    NORMAL_RETURN;

ERR_CLEANUP:
    bwTables->colorLut = NULL;
    bwTables->drcTable = NULL;
    ERR_RETURN;
}


/*
 *$Log: backgrnd.c,v $
 *Revision 1.1.1.1  1999/09/20 18:19:14  jlin
 *Imported sources
 *
 *Revision 1.12  1998/07/28 21:50:53  aber
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *CommonCheck: Warning fixes for gcc v2.8x.
 *
 *Revision 1.11  1998/04/16 00:08:46  xzhu
 *xzhu on Wed Apr 15 16:57:08 PDT 1998
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *Using histMask instead of taking out most right entry. If there is input mask, use input mask, else generate mask.
 *
 *Revision 1.10  1997/11/26 19:30:54  xzhu
 *xzhu on Wed Nov 26 11:29:37 PST 1997
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *Added boolean input argument in calling w_cMapToLut, since its prototype was changed.
 *
 *Revision 1.9  1997/07/16 02:59:07  jcraig
 *jcraig on Tue Jul 15 19:58:13 PDT 1997
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *fix bug in w_backgroundToWhite to use correct endpoints when checking
 *for a new background peak to the right of the first one chosen.
 *
 *Revision 1.8  1997/07/12 01:49:04  jcraig
 *jcraig on Fri Jul 11 18:42:52 PDT 1997
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *add new procedures, w_makeBackgroundToWhiteTables and
 *w_destroyBackgroundToWhiteTables, which will make and destroy the tables
 *used by w_backgroundToWhite.  These tables are the same for each
 *call to w_backgroundToWhite so that can be made once and used
 *several times.
 *add parameters to w_backgroundToWhite.  1) the selector - this is
 *used so that areas which are actually in the foreground (like the
 *tint of reverse video) can be eliminated and won't influence the
 *histograms when analyzing each connected component.  2) pointer
 *to structure which contains the tables used by this routine.  make
 *them once and use them several times.  if null, we make and destroy
 *the tables.  3-4) minimum width and height of a connected component
 *to consider it to be tint.
 *fix bug encountered if the mode of the histogram was just below
 *maximum white.
 *
 *Revision 1.7  1997/06/26 20:00:16  ddavies
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *CommonCheck: Add inclusion of new file rast.pub.
 *
 *Revision 1.6  1997/06/24 23:43:54  jcraig
 *jcraig on Tue Jun 24 16:33:39 PDT 1997
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *chnages to w_backgroundToWhite.  check pixrs for byte order, depth, etc.
 *if index from w_findPeakExtent returns -1, try setting  it to
 *some "reasonable" value (0 if looking for the left edge, the max
 *index if looking for the right edge).
 *when looking for second peak (cause first one didn't contain enought
 *pixels), only take its values if that peak contains more pixels than
 *the first.
 *when filtering connected components, if the mode of the bounding
 *rectangle falls exactly on the edges of the background range, look
 *at the hue and compare it to the background hue.
 *
 *Revision 1.5  1997/06/17 01:36:46  jcraig
 *jcraig on Mon Jun 16 18:33:34 PDT 1997
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *rename w_backgroundToWhite to w_backgroundChromThresh.
 *w_backgroundToWhite now has a new implementation based on color
 *dithering.
 *
 *Revision 1.4  1997/05/26 21:07:43  jcraig
 *jcraig on Mon May 26 14:07:02 PDT 1997
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *Get rid of compiler warnings (really I did a build before I checked
 *this in on Friday - I have no idea how I missed these!)
 *
 *Revision 1.3  1997/05/24 01:26:37  jcraig
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *fix byteswapping bug in w_backgroundToWhite.
 *add new procedure, w_backgroundLighten, which attempts to lighten
 *the background.
 *
 *Revision 1.2  1997/05/21 16:59:00  jcraig
 *jcraig on Wed May 21 09:58:37 PDT 1997
 *
 *backgrnd.c -> /product/ipcore/ipwrap/src/RCS/backgrnd.c,v
 *
 *remove size parameter from call to w_findPeakExtent
 *
 *Revision 1.1  1997/05/21 02:01:38  jcraig
 *Initial revision
 *
 */

/*</PRE></BODY>*/
