/*<HTML><HEAD><TITLE>binarize.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  binarize.c
 *
 * Binarization
 *
 * Exports:
 *        w_binarizeInit ()
 *        w_binarizeBand ()
 *        w_binarizeFinish ()
 *        w_binarize ()
 *        w_binarize2 ()
 *        w_binarize2Init()
 *        w_binarize2Band()
 *        w_binarize2Finish()
 *        w_binarize3 ()
 */


#include <stdio.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "alplib.h"
#include "thrshold.h"
#include "morphg.h"
#include "defines.pub"
#include "frames.pub"
#include "logicops.h"
#include "imagerep.h"
#include "imageref.pub"
#include "iaerror.pub"
#include "ansiprot.h"

/* For debugging */
#include "plotdefs.h"
#include "utils.pub"
#include "shrpixr.pub"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "floyd.pub"
#include "thrshold.pub"
#include "rescale.pub"
#include "shrscal.pub"
#include "strel.pub"
#include "clean.pub"
#include "morph.pub"
#include "morphg.pub"
#include "reduce.pub"
#include "intops.pub"
#include "shrcc.pub"
#include "shrip.pub"
#include "fill.pub"
#include "logicops.pub"
#include "binarize.pub"
#include "rectops.pub"
#include "resconv.pub"
#include "halftone.pub"
#include "fthresh.pub"
#include "io.pub"
#include "bctable.pub"
#include "ccthresh.pub"
#include "gradient.pub"
#include "prxform.pub"
#include "grbinop.pub"
#include "sharpen.pub"
#include "expand.pub"
#include "binarize.prv"

IP_RCSINFO(RCSInfo, "$RCSfile: binarize.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");

/* For debugging */
/*#define dumpBits(p) w_numberBits(pIPWState, p, &bits);\
fprintf(stderr, "%ld: %ld bits\n", step++, bits);
*/
#define dumpBits(p) /* */

/*
 * The idea behind banded binarization is to take bands of 4- or 8-bit
 * data and produce a binary version which is adaptively thresholded.
 * There are two modes: text-only, where we don't care what happens to
 * the pictures but would like the text to look good and eliminate
 * text-on-tints as much as possible, and text-and-picture, where we want
 * to preserve and dither the pictures, adaptively threshold the text, and
 * select between dithered and thresholded based on whether we're near a
 * text edge.
 *
 * Methods:
 *
 * Our adaptive thresholding algorithm is to take 64x64 tiles and
 * find a threshold for each individually. The first thresholding method
 * tried, namely connected-component-based thresholding, didn't work
 * very well when the tile size got small--it was too unstable. The
 * thresholding method we're using now is the one in findEdgeThreshold,
 * which uses a slope-of-edges based algorithm (see that function
 * for more details). We do a reduction by either 2 or 4, depending
 * on the resolution, before we send any image data to findEdgeThreshold.
 * This is done for two reasons: first, findEdgeThreshold is a bit
 * slow, and reducing the data by 1/4 or 1/16 is a help. Second,
 * the results are a bit more robust since noise is reduced.
 *
 * The binarize function is written to take 64 lines at a time of
 * gray data. We threshold them adaptively, and then put the results
 * in a buffer called the "strip buffer". A strip is currently 384
 * lines high. If the parameters call for it, we also dither into
 * a separate buffer, also 384 lines high. The reason we make this
 * 384 line buffer is so that we can run halftoneThresh on it to
 * segment text from pictures. On a 64-line buffer, halftoneThresh
 * doesn't seem to have enough vertical stuff to work very well.
 *
 * In "text-only" mode, we either threshold adaptively only, or
 * dither the pictures and use halftoneThresh to preserve them, but
 * attempt to use only the thresholded image in regions where
 * there is not picture, so that we get rid of tints behind text.
 * Once we've segmented, we use the resulting mask to decide whether
 * to take the dithered or the thresholded version if we're doing
 * "textOnly, dithered" mode. If we're not dithering, we don't
 * segment, and just take the adaptively thresholded text (in fact,
 * we could certainly dispense with the 384 line buffer, but don't
 * right now). Why 384 lines? It's historical, actually--a previous
 * algorithm did x3 reduction, and I wanted something greater than
 * 256 that is divisible by three.. It could probably be 256 or 512
 * just as easily, but I've done quite a bit of testing with 384,
 * so I'll just leave it that way for now.
 *
 * When doing "text and pictures" mode, we do something a little
 * different, inspired by some schemes used in the Fuji Xerox "AColor"
 * 635. The idea is that in general, dithering is the right
 * thing to reproduce a page, except at the edges of text, where we'd
 * like sharper edges. So what we do is find the edges in the image
 * using a morphological filter (difference between a dilate and an
 * erode, in binary) and then use the results of halftoneThresh to
 * get rid of the edges that are contained in the pictures. We then
 * use that edge map (dilated slightly) to blast away the dithered
 * version, and replace it with the thresholded version.
 *
 * 5/1/95: It has been requested that this have an option to expand
 * while binarizing, i.e. upconvert the gray image before thresholding.
 * This becomes yet another ball-of-string function, but the advantage,
 * I suppose, is that we can get away with creating only a single expanded
 * gray tile. This means, of course, that the "strip" height will
 * increase with the expansion factor. 
 *
 * 5/13/95: There has been a problem with the thresholding of the edges
 * of paper, etc. The issue is that a tile which lands on paper's edge
 * will naturally have only the paper as the "large transition", therefore
 * thresholding one side of the tile to black. This has been deemed
 * "unacceptable", so I'm making an attempt to fix it. A related issue is
 * that on noisy originals, we can sometimes toggle a bit too much
 * between a couple of thresholds, resulting in uneven thresholding. 
 * The solution I'm attempting in order to solve both problems is 
 * to try to find, for each row of tiles, regions of adjacent tiles with
 * nearly the same threshold. Tiles which are not part of a "sufficiently
 * large" region will be considered "unknown", and will be thresholded
 * using the nearest "known" threshold. This should eliminate the edge
 * problem, and should also reduce problems around heavily tinted regions,
 * where we sometimes leave a "halo" when there are tiles with no
 * edge information.
 * I've made a couple of previous attempts and chucked them:
 * (1) have a simple threshold of contrast 
 * to decide whether a tile was "text" or "unknown". This worked most of the
 * time, but I did find cases where there was text below the threshold.
 * (2) take an average of the last few thresholds. This does eliminate some
 * of the edge problems, but causes an annoying effect on heavy tints, where
 * the tint bleeds into the text as the average moves towards the correct text
 * threshold.
 * How do we find these regions? Finding the optimal partitions is probably
 * a reasonable goal, but this must run reasonably quickly, and I'm guessing
 * that a fairly naive scheme will work well enough, since the flat regions
 * are usually going to be quite large and isolated.
 * My first attempt is just a one-way walk through the tile thresholds,
 * keeping a candiate "start" position and searching forwards until we
 * either find a difference too large. If the distance between the start
 * position and the tile before the offending difference is large enough,
 * we accept it as a region and move our candidate "start" position to
 * the offending tile. How to measure the maximum distance is a bit tricky--
 * is it distance from the running average? From the first tile? The latter
 * has the advantage of not allowing slow gradients to go too far.
 * Instead of the "first" fit method described, we could do a "best fit"
 * method, finding *all* candidate regions, and then taking the
 * largest ones out until all the tiles which were classified in regions
 * are covered by a region. I'm guessing that such complications will be
 * unnecessary.
 * 
 */

/* The state structure */
typedef struct
{
    Int32      width, height;        /* Size of image */
    Int32      eWidth, eHeight;      /* Size of expanded image */
    Int32      tileSize;             /* Size of a tile */
    Int32      eTileW, eTileH;       /* Size of expanded tile */
    ScaledInt1616 dpi;               /* Image DPI */
    Int32      depth;                /* Bit depth */
    Int32      reductions;           /* # of reductions to do for threshold */
    Int32      row;                  /* Current band row */
    Int32      stripRow;             /* Base row of current strip */
    Int32      expansion;            /* Expansion factor before binarizing:
                                        0 = none, 1 = 1.5, 2 = 2x */
    Int32      stripHeight;          /* The "strip height", with expansion */
    Int32      bandHeight;           /* Height of an expanded band */
    Bool       textOnly;	     /* If true, use text-only alg. */
    Bool       dithering;	     /* If true, dither pics always */
    Bool       zapPepper;	     /* If true, remove pepper in text */
    PIXR      *prDitheredStrip;      /* Current strip, dithered */
    PIXR      *prThresholdedStrip;   /* Current strip, thresholded */
    PIXR      *prResult;             /* Full output binary image */
    PIXR      *prExpandedBand;       /* If dithering, the expanded band */
    UInt32    *pResConTab;           /* Resolution conversion table */
    void      *ditherState;          /* Current dither state */
    UInt8     *pTRC;                 /* Tone-reproduction curve */    
    Int32      linesPer;             /* # of lines dithered at once */
    Int32      stripOffset;          /* Current strip offset */
} BinState;

/*****************************************************************************
 * BeginDocumentation::: "w_binarizeInit"
 *
 * OneLineDescription:::
 *  Intialize banded binarization
 *
 * LongDescription:::
 *  The idea behind banded binarization is to take bands of 4- or 8-bit
 *  data and produce a binary version which is adaptively thresholded.
 *  There are two modes: text-only, where we don't care what happens to
 *  the pictures but would like the text to look good and eliminate
 *  text-on-tints as much as possible, and text-and-picture, where we want
 *  to preserve and dither the pictures, adaptively threshold the text, and
 *  select between dithered and thresholded based on whether we're near a
 *  text edge.
 *  <BR>
 *  This function initializes banded binarization. Creates a state structure 
 *  and associated stuff, including the output binary image. The 
 *  input is to be processed in bands of cBandHeight in height
 *  (the final band can be less).
 *
 * PrivateDescription:::
 *  Our adaptive thresholding algorithm is to take 64x64 tiles and
 *  find a threshold for each individually. The first thresholding method
 *  tried, namely connected-component-based thresholding, didn't work
 *  very well when the tile size got small--it was too unstable. The
 *  thresholding method we're using now is the one in findEdgeThreshold,
 *  which uses a slope-of-edges based algorithm (see that function
 *  for more details). We do a reduction by either 2 or 4, depending
 *  on the resolution, before we send any image data to findEdgeThreshold.
 *  This is done for two reasons: first, findEdgeThreshold is a bit
 *  slow, and reducing the data by 1/4 or 1/16 is a help. Second,
 *  the results are a bit more robust since noise is reduced.
 *  <BR>
 *  The binarize function is written to take 64 lines at a time of
 *  gray data. We threshold them adaptively, and then put the results
 *  in a buffer called the "strip buffer". A strip is currently 384
 *  lines high. If the parameters call for it, we also dither into
 *  a separate buffer, also 384 lines high. The reason we make this
 *  384 line buffer is so that we can run halftoneThresh on it to
 *  segment text from pictures. On a 64-line buffer, halftoneThresh
 *  doesn't seem to have enough vertical stuff to work very well.
 *  <BR>
 *  In "text-only" mode, we either threshold adaptively only, or
 *  dither the pictures and use halftoneThresh to preserve them, but
 *  attempt to use only the thresholded image in regions where
 *  there is not picture, so that we get rid of tints behind text.
 *  Once we've segmented, we use the resulting mask to decide whether
 *  to take the dithered or the thresholded version if we're doing
 *  "textOnly, dithered" mode. If we're not dithering, we don't
 *  segment, and just take the adaptively thresholded text (in fact,
 *  we could certainly dispense with the 384 line buffer, but don't
 *  right now). Why 384 lines? It's historical, actually--a previous
 *  algorithm did x3 reduction, and I wanted something greater than
 *  256 that is divisible by three.. It could probably be 256 or 512
 *  just as easily, but I've done quite a bit of testing with 384,
 *  so I'll just leave it that way for now.
 *  <BR>
 *  When doing "text and pictures" mode, we do something a little
 *  different, inspired by some schemes used in the Fuji Xerox "AColor"
 *  635. The idea is that in general, dithering is the right
 *  thing to reproduce a page, except at the edges of text, where we'd
 *  like sharper edges. So what we do is find the edges in the image
 *  using a morphological filter (difference between a dilate and an
 *  erode, in binary) and then use the results of halftoneThresh to
 *  get rid of the edges that are contained in the pictures. We then
 *  use that edge map (dilated slightly) to blast away the dithered
 *  version, and replace it with the thresholded version.
 *  <BR>
 *  5/1/95: It has been requested that this have an option to expand
 *  while binarizing, i.e. upconvert the gray image before thresholding.
 *  This becomes yet another ball-of-string function, but the advantage,
 *  I suppose, is that we can get away with creating only a single expanded
 *  gray tile. This means, of course, that the "strip" height will
 *  increase with the expansion factor. 
 *  <BR>
 *  5/13/95: There has been a problem with the thresholding of the edges
 *  of paper, etc. The issue is that a tile which lands on paper's edge
 *  will naturally have only the paper as the "large transition", therefore
 *  thresholding one side of the tile to black. This has been deemed
 *  "unacceptable", so I'm making an attempt to fix it. A related issue is
 *  that on noisy originals, we can sometimes toggle a bit too much
 *  between a couple of thresholds, resulting in uneven thresholding. 
 *  The solution I'm attempting in order to solve both problems is 
 *  to try to find, for each row of tiles, regions of adjacent tiles with
 *  nearly the same threshold. Tiles which are not part of a "sufficiently
 *  large" region will be considered "unknown", and will be thresholded
 *  using the nearest "known" threshold. This should eliminate the edge
 *  problem, and should also reduce problems around heavily tinted regions,
 *  where we sometimes leave a "halo" when there are tiles with no
 *  edge information.
 *  I've made a couple of previous attempts and chucked them:
 *  (1) have a simple threshold of contrast 
 *  to decide whether a tile was "text" or "unknown". This worked most of the
 *  time, but I did find cases where there was text below the threshold.
 *  <BR>
 *  (2) take an average of the last few thresholds. This does eliminate some
 *  of the edge problems, but causes an annoying effect on heavy tints, where
 *  the tint bleeds into the text as the average moves towards the correct text
 *  threshold.
 *  <BR>
 *  How do we find these regions? Finding the optimal partitions is probably
 *  a reasonable goal, but this must run reasonably quickly, and I'm guessing
 *  that a fairly naive scheme will work well enough, since the flat regions
 *  are usually going to be quite large and isolated.
 *  My first attempt is just a one-way walk through the tile thresholds,
 *  keeping a candiate "start" position and searching forwards until we
 *  either find a difference too large. If the distance between the start
 *  position and the tile before the offending difference is large enough,
 *  we accept it as a region and move our candidate "start" position to
 *  the offending tile. How to measure the maximum distance is a bit tricky--
 *  is it distance from the running average? From the first tile? The latter
 *  has the advantage of not allowing slow gradients to go too far.
 *  Instead of the "first" fit method described, we could do a "best fit"
 *  method, finding *all* candidate regions, and then taking the
 *  largest ones out until all the tiles which were classified in regions
 *  are covered by a region. I'm guessing that such complications will be
 *  unnecessary.
 *
 * Input:::
 *  pIPWState::      State structure
 *  width::          Width of source image, without frame
 *  height::         Height of source image, without frame
 *  depth::          Depth, in bits, of source image
 *  dpi::            DPI of the image, used to determine reductions, &c.
 *  expand::         If zero, don't expand the image. Otherwise, expand
 *                   the image before thresholding: 1 = 1.5x, 2 = 2x
 *  pTRC::           Tone-correction curve, if dithering
 *  textOnly::       If true, do a text-only binarization,
 *                   otherwise do a text/picture binarization.
 *  dithering::      If true, always dither even in text-only mode
 *  zapPepper::      If true, remove pepper in the text part. This
 *                   is only applicable if doing textOnly.
 *  pPrResult::      If non-NULL, use this as a result PIXR, otherwise
 *                   create one
 *
 * Output:::
 *  ppBinState::     State structure to pass to the band and finish 
 *                   routines, or NULL if error
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_binarizeInit"></A>
 *****************************************************************************/
ErrCode CDECL
w_binarizeInit (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       dpi,
    Int32               expand,
    UInt8              *pTRC,
    Bool                textOnly,
    Bool                dithering,
    Bool                zapPepper,
    PIXR              **pPrResult,
    void              **ppBinState)
{
    PROLOGUE;
    BinState    *s = NULL;
    Int32       unscaledDpi;
    UInt32      resconTableSize;
    ERR_GEN_CLEANUP (expand < 0 || expand > 2, 
		     cErrExpansionNotSupported);

    ERR_GEN_CLEANUP (depth != 4 && depth != 8, cErrPixr_depth);
    
    /* Initialize the strels we'll need */
    ERR_PASS_CLEANUP (w_initStrelByName (pIPWState, strel3));

    /* Allocate the state structure */
    s = (BinState *) IPW_CALLOC (1, sizeof (BinState), cAllocatorData);
    ERR_GEN_CLEANUP (s == NULL, cErrNomem);

    s->width = width;
    s->height = height;
    s->dpi = dpi;
    unscaledDpi = ScaledInt1616_TO_Int32 (dpi);
    s->depth = depth;
    s->row = 0;
    s->stripRow = 0;
    s->stripOffset = 0;
    s->textOnly = textOnly;
    s->dithering = dithering;
    s->pTRC = pTRC;
    if (s->textOnly == cFalse)
	s->dithering = cTrue;
    
    s->zapPepper = zapPepper;
    if (unscaledDpi >= 300)
	s->reductions = 4;
    else if (unscaledDpi >= 200)
	s->reductions = 2;
    else
	s->reductions = 2;

    s->ditherState = NULL;
    s->expansion = expand;
    if (unscaledDpi < 200)
	s->tileSize = cTile / 2;
    else
	s->tileSize = cTile;

    if (s->expansion == 1)
    {
	s->stripHeight = cStripHeight * 3 / 2;
	s->eWidth = width * 3 / 2;
	s->eHeight = height * 3 / 2;
	s->eTileW = s->eTileH = s->tileSize * 3 / 2;
	s->bandHeight = cBandHeight * 3 / 2;
	s->pResConTab = NULL;
	i_getResconTableSize (pIPWState, 2, 3, s->depth, &resconTableSize);
	if (resconTableSize)
	{
	    s->pResConTab = (UInt32 *)IPW_MALLOC (resconTableSize, 
						  cAllocatorData);
	    ERR_GEN (s->pResConTab == NULL, ia_nomem);
	    ERR_PASS_CLEANUP( i_make2To3ResconTable (pIPWState, &s->pResConTab, 
						     s->depth));
	}
    }
    else if (s->expansion)
    {
	s->stripHeight = cStripHeight * s->expansion;
	s->eWidth = width * s->expansion;
	s->eHeight = height * s->expansion;
	s->eTileW = s->eTileH = s->tileSize * s->expansion;
	s->bandHeight = cBandHeight * s->expansion;
	s->pResConTab = NULL;
	i_getResconTableSize (pIPWState, 1, 2, s->depth, &resconTableSize);
	if (resconTableSize)
	{
	    s->pResConTab = (UInt32 *)IPW_MALLOC (resconTableSize,
						  cAllocatorData);
	    ERR_GEN (s->pResConTab == NULL, ia_nomem);
	    ERR_PASS_CLEANUP (i_make1To2ResconTable (pIPWState, &s->pResConTab, 
						     s->depth));
	}
    }
    else
    {
	s->stripHeight = cStripHeight;
	s->eWidth = width;
	s->eHeight = height;
	s->eTileW = s->eTileH = s->tileSize;
	s->bandHeight = cBandHeight;
    }

    if (s->dithering)
    {
	s->prDitheredStrip = NULL;
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState,
					      s->eWidth, s->stripHeight, 
					      1, dpi, dpi,
					      cFrameBits,
					      &(s->prDitheredStrip)));
	if (s->expansion)
	{
	    s->prExpandedBand = NULL;
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, s->eWidth, 
						  s->bandHeight,
						  s->depth, dpi, dpi,
						  cFrameBits,
						  &(s->prExpandedBand)));
	}

    }
   
    s->prThresholdedStrip = NULL;
    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, s->eWidth, s->stripHeight,
					  1, dpi, dpi, cFrameBits,
					  &(s->prThresholdedStrip)));


    if (*pPrResult != NULL)
    {
	s->prResult = *pPrResult;
    }
    else
    {
	/* This has to be initialized because we OR the text into it..
	 */
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, s->eWidth, s->eHeight,
					1, dpi, dpi, cFrameBits,
					&(s->prResult)));
    }

    *ppBinState = (void *) s;
    *pPrResult = s->prResult;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (s)
    {
	if (s->prDitheredStrip) w_destroyPixr (pIPWState, &s->prDitheredStrip);
	if (s->prThresholdedStrip) w_destroyPixr (pIPWState, &s->prThresholdedStrip);
	if (s->prExpandedBand) w_destroyPixr (pIPWState, &s->prExpandedBand);
	if (s->prResult) w_destroyPixr (pIPWState, &s->prResult);
	if (s->pResConTab) IPW_FREE (s->pResConTab, cAllocatorData);
	IPW_FREE (s, cAllocatorData);
    }
    
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation::: "wp_binarizeStripCopy"
 *
 * OneLineDescription:::
 *  Process a strip of binarized image and copy it to the destined image.
 *
 * LongDescription:::
 *  Before we copy binarized strip image ( including several bands ),
 *  we will do some processing, then copy strip image into the destined 
 *  image buffer.
 *
 * PrivateDescription:::
 *  We moved this part out from w_binarizeBand() to let it work
 *  without the condition image band height = cBandHeight. Whenever we
 *  find the remained strip buffer is less than input image band, we copy
 *  and empty strip image buffer.
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  state::          Binarization state structure, as created by
 *                   w_binarizeInit()
 *
 * Output:::
 *
 * Returns:::
 *  Error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod/zhu::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_binarizeBand"></A>
 *****************************************************************************/
static ErrCode CDECL
wp_binarizeStripCopy (
    IPWState           *pIPWState,
    void               *state)
{
    PROLOGUE;
    BinState      *s = (BinState *) state;
    PIXR          *prMask = NULL;
    PIXR     	  *prTmp1 = NULL, *prTmp2 = NULL;
    Int32         stripH, unscaledDpi;
    ScaledInt1616 dpi;
    STREL         *pStrel3 = NULL;
    RECTANGL      *rects = NULL;

    stripH = s->stripOffset;
    dpi = s->dpi;
    unscaledDpi = ScaledInt1616_TO_Int32 (dpi);
	
    if (s->dithering)
    {
	prMask = NULL;
	if (unscaledDpi >= 300)
	{
	    ERR_PASS_CLEANUP (
		w_halftoneThresh (pIPWState, s->prThresholdedStrip,  
				  4, 6, 1, 90, 15, &rects, &prMask));
	}
	else if (unscaledDpi >= 200)
	{
	    ERR_PASS_CLEANUP (
		w_halftoneThresh (pIPWState, s->prThresholdedStrip, 
				  2, 3, 1, 45, 8, &rects, &prMask));
	}
	else
	{
	    ERR_PASS_CLEANUP (
		w_halftoneThresh (pIPWState, s->prThresholdedStrip, 
				  1, 2, 1, 33, 5, &rects, &prMask));
	}
	
	ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &rects));
	rects = NULL;
    }

    if (s->textOnly == cFalse)
    {
	    /*
	     * We're doing the text-and-pictures case. What we do here
	     * is find edges using a morphological filter, and then use the 
	     * mask from halftoneThresh to blow away the edges that were
	     * inside the picture. Then, we get rid of these edges in 
	     * the dithered version, and copy the edges out of the 
	     * thresholded version. The idea is that we preserve the
	     * dithered version everywhere EXCEPT where there's a text
	     * edge. 
	     */

	    /* Find edges by taking the difference of a dilation and
	       an erosion. */
	ERR_PASS_CLEANUP (w_getStrel (pIPWState, strel3, &pStrel3));
	prTmp1 = NULL;
	ERR_PASS_CLEANUP (w_erode (pIPWState, s->prThresholdedStrip, 
				   pStrel3, &prTmp1));
	prTmp2 = NULL;
	ERR_PASS_CLEANUP (w_dilate (pIPWState, s->prThresholdedStrip, 
				    pStrel3, &prTmp2));
	
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTmp1, XOR, prTmp2));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));

	/* Dilate it a bit to get the centers of characters better */
	ERR_PASS_CLEANUP (w_dilate (pIPWState, prTmp2, pStrel3, &prTmp2));
	
	/* Mask off the pictures in the edge mask */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prMask,
				      PIX_NOT(PIX_SRC) & PIX_DST, 
				      0, 0, s->eWidth, stripH, prTmp2));

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMask));
	
	/* Get rid of the edges in the dithered version */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prTmp2,
				      PIX_NOT(PIX_SRC) & PIX_DST,
				      0, 0, s->eWidth, stripH, 
				      s->prDitheredStrip));

	/* Copy the dithered into the result */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      0, 0, s->prDitheredStrip, PIX_SRC,
				      0, s->stripRow, s->eWidth, stripH, 
				      s->prResult));

	/* Blow away the non-edges in binarized version */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prTmp2,
				      PIX_SRC & PIX_DST, 
				      0, 0,	s->eWidth, stripH, 
				      s->prThresholdedStrip));
	
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp2));
	
	/* And copy the text in */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState,
				      0, 0, s->prThresholdedStrip,
				      PIX_SRC | PIX_DST,
				      0, s->stripRow, s->eWidth, stripH, 
				      s->prResult));
    }
    else
    {
	    /* This is the text-only case. We just copy the thresholded
	     * version if we weren't dithering, otherwise we use the
	     * halftoneThresh mask to decide which pixels are picture 
	     * (and should come from the dithered version) 
	     */
	       
	if (s->dithering)
	{
	    /* Get rid of the non-pictures in the dithered version */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prMask,
					  PIX_SRC & PIX_DST,
					  0, 0, s->eWidth, stripH,
					  s->prDitheredStrip));

	    /* Copy the dithered into the result */
	    ERR_PASS_CLEANUP (
		w_rasterop (pIPWState,
			    0, 0, s->prDitheredStrip, PIX_SRC, 
			    0, s->stripRow, s->eWidth, stripH, 
			    s->prResult));

	    /* Blow away the pictures in binarized version */
	    ERR_PASS_CLEANUP (
		w_rasterop (pIPWState, 0, 0, prMask,
			    PIX_NOT(PIX_SRC) & PIX_DST,
			    0, 0, s->eWidth, stripH, 
			    s->prThresholdedStrip));

	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMask));
		
	    /* And maybe take out pepper */
	    if (s->zapPepper)
	    {
		ERR_PASS_CLEANUP (w_removePepper (
		    pIPWState, s->prThresholdedStrip, 4, 
		    &s->prThresholdedStrip)); 
	    }
		
	    /* And copy the text in */
	    ERR_PASS_CLEANUP (
		w_rasterop (pIPWState, 0, 0, s->prThresholdedStrip,
			    PIX_SRC | PIX_DST, 
			    0, s->stripRow, s->eWidth, stripH, 
			    s->prResult));
	}
	else
	{
	    if (s->zapPepper)
	    {
		ERR_PASS_CLEANUP (
		    w_removePepper (pIPWState,
				    s->prThresholdedStrip, 2, 
				    &s->prThresholdedStrip)); 
	    }
	
	    /* And copy the text in */
	    ERR_PASS_CLEANUP (
		w_rasterop (pIPWState, 0, 0, s->prThresholdedStrip,
			    PIX_SRC | PIX_DST, 
			    0, s->stripRow, s->eWidth, stripH, 
			    s->prResult));
	}
    }

    s->stripRow += stripH;

    /* Clean it for next time */
    s->stripOffset = 0;
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
				  0, 0, s->eWidth, s->stripHeight, 
				  s->prThresholdedStrip));
	
    
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Process a band for binarization
 *
 * LongDescription:::
 *  The idea behind banded binarization is to take bands of 4- or 8-bit
 *  data and produce a binary version which is adaptively thresholded.
 *  There are two modes: text-only, where we don't care what happens to
 *  the pictures but would like the text to look good and eliminate
 *  text-on-tints as much as possible, and text-and-picture, where we want
 *  to preserve and dither the pictures, adaptively threshold the text, and
 *  select between dithered and thresholded based on whether we're near a
 *  text edge.
 *  <BR>
 *  This function processes a band. The band should be cBandHeight lines
 *  high, and as wide as we said in the initialization. If this is the 
 *  last strip, it can be fewer than cBandHeight lines high.
 *  <BR>
 *  What we do is to threshold into the "thresholdedStrip" buffer,
 *  dither into the "ditheredStrip" buffer if called for, and if 
 *  either the buffer is full or this is the last band, segment the
 *  thresholded thing and produce the resulting strip.
 *
 * PrivateDescription:::
 *  01/07/98
 *  It is hard to qurantee the band height = cBandHeight. For example, 
 *  orignal image band is scaled or reduced. Therefore we add stripOffset to 
 *  remember the current strip offset and judge whether copy to
 *  result or not. The band height can be value which is around cBandHeight
 *  and  less than cStripHeight.
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  state::          Binarization state structure, as created by
 *                   w_binarizeInit()
 *  prBand::         The band to process
 *
 * Output:::
 *
 * Returns:::
 *  Error status 
 *
 * Notes:::
 *  If the process fails, all data structures are de-allocated, so one
 *  should not call w_binarizeFinish after an error.
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_binarizeBand (
    IPWState           *pIPWState,
    void               *state,
    PIXR               *prBand)
{
    PROLOGUE;
    BinState *s = (BinState *) state;
    Int32     depth, stripOffset, nextDitheredRow, tileW, tileH, i, col, eCol,
	textThresh, eTileH, eTileW, range, t0, t1, j,
#if 0
	      *histo, sum1, sum2, sum3, lo, hi, ix, r3,
	      whitePoint, whiteCount, darkPoint, darkCount,
#endif
	      thresholds[cMaxTilesPerLine], tile, tiles, threshSum, w,
	      candidate, maxDiff, lastRegionEnd, th0, th1, midPoint, minLength,
              wB, hB, bplB, wE, hE, bplE, bandH, wT, hT, bplT, unscaledDpi;
    PIXR     *prTile = NULL, *prThresh = NULL, *prReduced = NULL, 
	     *prTmp1 = NULL,
             *prEBand = NULL, *prExpandedTile = NULL, *prETile = NULL;
    UInt8    *pB = NULL, *pE = NULL, *pT = NULL;
    RECTANGL  hRect;
    ScaledInt1616 dpi;
/*    Int32 bits, step = 0; */
    
    ERR_GEN (s == NULL, cErrNullPtr);

    ERR_GEN_PIXR (prBand, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1, cFrame32);
    
    ERR_PASS (w_pixrGetHeight (pIPWState, prBand, &tileH));

    ERR_PASS (w_pixrGetWidth (pIPWState, prBand, &w));
    ERR_GEN (((w != s->width) || (tileH > cStripHeight)), cErrImageTooWide);

    if (s->expansion)
	eTileH = (s->expansion == 1) ? tileH * 3 / 2 : tileH * s->expansion;
    else
	eTileH = tileH;

    ERR_PASS (w_pixrGetDepth (pIPWState, prBand, &depth));
    
    maxDiff = (depth == 4) ? cMaxDiff4 : cMaxDiff8;
    t0 = t1 = -1;
    range = (1 << depth);
    if ((s->stripHeight - s->stripOffset) < eTileH )
    {
	ERR_PASS_CLEANUP( wp_binarizeStripCopy(pIPWState, state));
    }
    stripOffset = s->stripOffset;
    minLength = 4;

dumpBits(prBand);
    /* If necessary, dither the band */
    if (s->dithering)
    {
	ERR_PASS (w_pixrGetImageAll (pIPWState, prBand, &pB));
	ERR_PASS (w_pixrFramedWidth (pIPWState, prBand, &wB));
	ERR_PASS (w_pixrFramedHeight (pIPWState, prBand, &hB));
	ERR_PASS (w_pixrGetBpl (pIPWState, prBand, &bplB));
	if (s->expansion)
	{
	    ERR_PASS (w_pixrGetImageAll (pIPWState, s->prExpandedBand, &pE));
	    ERR_PASS (w_pixrFramedWidth (pIPWState, s->prExpandedBand, &wE));
	    ERR_PASS (w_pixrFramedHeight (pIPWState, s->prExpandedBand, &hE));
	    ERR_PASS (w_pixrGetBpl (pIPWState, s->prExpandedBand, &bplE));
	}
	    
	if (s->expansion == 1)
	{
	    ERR_PASS( i_rescon2To3 (pIPWState, pB, wB, hB, bplB, depth,
				    pE, wE, hE, bplE, depth, s->pResConTab));
	    prEBand = s->prExpandedBand;
	}
	else if (s->expansion == 2)
	{
	    ERR_PASS( i_rescon1To2 (pIPWState, pB, wB, hB, bplB, depth,
				    pE, wE, hE, bplE, depth, s->pResConTab));
	    prEBand = s->prExpandedBand;
	}
	else
	{
	    prEBand = prBand;
	}

	if (!s->ditherState)
	{
	    /* Since ditherFloydInit requires a PIXR, we need to call it
	       here rather than in the init routine */
	    s->linesPer = s->bandHeight;

	    ERR_PASS (w_pixrGetWidth (pIPWState, prEBand, &w));
	    ERR_PASS (w_ditherFloydInit (pIPWState, prEBand, 0, 0, w, s->pTRC, 
				       &s->linesPer, &s->ditherState));
	}

	/* do linesPer lines */
	nextDitheredRow = 0;
	ERR_PASS (w_pixrGetHeight (pIPWState, prEBand, &bandH));
	if (bandH >= s->linesPer)
	{
	    ERR_PASS (w_ditherFloydProcess (pIPWState, 
					  stripOffset, prEBand, 0, 0, 
					  s->ditherState, s->prDitheredStrip));
	    nextDitheredRow = s->linesPer;
	}
	
	/* do the rest line-at-a-time */
	for (i = nextDitheredRow; i < bandH; i++)
	{
	    ERR_PASS (w_ditherFloydSingleProcess (pIPWState,
						  stripOffset + i, prBand,
						  0, i, s->ditherState, 
						  s->prDitheredStrip));
	}
    }

    /* Create a tile and thresholded tile */
    prTile = NULL;
    dpi = s->dpi;
    unscaledDpi = ScaledInt1616_TO_Int32 (dpi);
    ERR_PASS (w_createPixr (pIPWState, s->tileSize, tileH, depth,
			    dpi, dpi, cFrameBits, &prTile));
    if (s->expansion)
    {
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, s->eTileW,
					      eTileH, depth, dpi, dpi, 
					      cFrameBits, &prExpandedTile));
    }

    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, s->eTileW, eTileH, 1, 
					  dpi, dpi,
					  cFrameBits,  &prThresh));

    /* Reduce the strip */
    if (s->reductions == 2)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prBand, 200, 100,
				       &prReduced));
    }
    else if (s->reductions == 4)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prBand, 200, 100, 
				       &prTmp1));
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prTmp1, 200, 100, 
				       &prReduced));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
    }
    else
    {
	ERR_PASS_CLEANUP (w_copy(pIPWState, prBand, &prReduced));
    }

    hRect.x = hRect.y = 0;
    hRect.h = tileH;
    eCol = 0;

    /* 
     * First, compute the threshold for each tile
     */
    tile = 0;
    for (col = 0; col < s->width; col += s->tileSize)
    {
	tileW = MIN (s->width - col, s->tileSize);
	hRect.w = tileW;
#if 0	
	/* Copy in the tile */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, col, 0, prBand, PIX_SRC,
				      0, 0, tileW, tileH, prTile));
#endif
	/* Create a reduced tile for finding the threshold */
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, tileW / s->reductions, 
					      tileH / s->reductions,
					      depth, dpi, dpi,
					      cFrameBits, &prTmp1));

	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      col / s->reductions, 0, prReduced,
				      PIX_SRC, 0, 0, 
				      tileW / s->reductions, 
				      tileH / s->reductions, prTmp1));
    
#if 0
	hRect.w = tileW;
	hRect.w = tileW;
	
	/* Collect a histogram for the tile */
	if (depth == 4)
	    histo = histoRect4Pr (pIPWState, prTile, &hRect);
	else
	    histo = histoRectPr (pIPWState, prTile, &hRect);
	    
	if (histo == NULL)
	    return abortI("no histo?", procName, 1);

	lo = 0; hi = range - 1;
	/* Bound the non-zero range */
	for (i = 0; i < range; i++)
	{
	    if (histo[i])
	    {
		lo = i;
		break;
	    }
	}
	    
	for (i = range - 1; i >= 0; i--)
	{
	    if (histo[i])
	    {
		hi = i;
		break;
	    }
	}

	sum1 = sum2 = sum3 = 0;
	r3 = (hi - lo + 1) / 3;
	ix = lo;
	    
	/* Compute a sum for thirds of the histogram */
	for (i = 0; i < r3; i++)
	{
	    sum1 += histo[ix++];
	}
	for (i = 0; i < r3; i++)
	{
	    sum2 += histo[ix++];
	}

	while (ix <= hi)
	{
	    sum3 += histo[ix++];
	}

	FREE(histo);
#endif
	/* Find the threshold on the reduced version using edge-based model */
	ERR_PASS (w_findEdgeThreshold (pIPWState, prTmp1, (RECTANGL *) NULL,
				       (unscaledDpi < 200) ? 28 : 64,
				       &textThresh));

	/*
	 * This is a RIDICULOUS hack. The idea is that at very low
	 * resolutions (e.g. 100dpi) the edge-based thresholding stuff
	 * finds what might be an OK threshold if the font is fairly
	 * heavy, but if the font has strokes that are considerably
	 * thinner than the pixel size, the thin strokes come out so
	 * light that a threshold which preserves the thickness of thinner
	 * strokes will necessarily break the characters at the thin
	 * strokes. Fact is, there just aren't enough pixels.. Anyway,
	 * the tendency is to make things too light. This hack, which
	 * works better than any scheme I've come up with, is to increase
	 * the threshold by some percentage, thus making everything
	 * a little darker. Right now, I'm using 125%, arrived at
	 * empirically from only about 30 documents. It's far from
	 * perfect, but if you want decent results, ask yourself why
	 * you ever thought scanning 100dpi was a good idea in the
	 * first place. I've tried connected-components-based thresholders
	 * instead, but when doing tiled thresholding, they never
	 * seem to be quite stable enough.
	 */
	if (unscaledDpi < 200)
	{
	    if (textThresh != -1)
	    {
		textThresh = (textThresh * 133) / 100; 
		textThresh = MIN(textThresh, ((1 << depth) - 1));
	    }
		    
	}

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp1));
        thresholds[tile++] = textThresh;
    }

    tiles = tile;

    /*
     * Find regions that are roughly the same by picking a starting point,
     * searching ahead, looking for tiles which are too different than the
     * starting point to be part of the same region, and if the distance
     * is great enough, filling in the region with the mean threshold
     * value
     */
    if (1) /* Was: (s->dpi >= 200) */ 
    {
	lastRegionEnd = -1;
	
	for (candidate = 0; candidate < tiles - 1; )
	{
	    threshSum = thresholds[candidate];

	    for (i = candidate + 1; i < tiles; i++)
	    {
		if (ABS(thresholds[i] - thresholds[candidate]) > maxDiff)
		{
		    /* This threshold too far away; do we have a region? */
		    if (i - candidate >= minLength)
		    {
			/* Fill in the region */
			textThresh = threshSum / (i - candidate);
			for (j = candidate; j < i; j++)
			{
			    thresholds[j] = textThresh;
			}

			/* Candidate is now the ith tile */
			candidate = i;
		    }
		    else
		    {
			thresholds[candidate] = -1;
			candidate++;
		    }
		    break;
		}

		threshSum += thresholds[i];
	    }

	    /* Handle falling off the end with a valid region */
	    if ((i == tiles) && (i - candidate >= minLength))
	    {
		/* Fill in the region */
		textThresh = threshSum / (i - candidate);
		for (j = candidate; j < i; j++)
		{
		    thresholds[j] = textThresh;
		}

		candidate = i;
	    }
	    else
		thresholds[candidate++] = -1;
	}


	/* Fill in the gaps */
	lastRegionEnd = 0;
	i = 0;
 	th0 = (1 << s->depth) / 2; /* 4? */
	while (i < tiles)
	{
	    if (thresholds[i] == -1)
	    {
		for (j = i; j < tiles && thresholds[j] == -1; j++)
		    ;
		if (j == tiles)
		{
		    /* fill to end */
		    for (j = i; j < tiles; j++)
			thresholds[j] = th0;
		    i = tiles;
		    break;
		}
		else
		{
		    midPoint = (i + j) / 2;
		    th1 = thresholds[j];
		    if (i == 0)
			th0 = th1;
		    for (tile = i; tile < j; tile++)
		    {
			thresholds[tile] = (tile < midPoint) ? th0 : th1;
		    }
		}
		i = j;
	    }
	    else
	    {
		th0 = thresholds[i];
		i++;
	    }
	}
    }
    /* We've computed the thresholds, now expand, if necessary, and
       threshold */

    for (col = 0, tile = 0; col < s->width; col += s->tileSize, tile++)
    {
	tileW = MIN (s->width - col, s->tileSize);

	/* Copy in the tile */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, col, 0, prBand, PIX_SRC,
				      0, 0, tileW, tileH, prTile));

	ERR_PASS (w_pixrGetImageAll (pIPWState, prTile, &pT));
	ERR_PASS (w_pixrFramedWidth (pIPWState, prTile, &wT));
	ERR_PASS (w_pixrFramedHeight (pIPWState, prTile, &hT));
	ERR_PASS (w_pixrGetBpl (pIPWState, prTile, &bplT));
	if (s->expansion >= 1)
	{
	    ERR_PASS (w_pixrGetImageAll (pIPWState, prExpandedTile, &pE));
	    ERR_PASS (w_pixrFramedWidth (pIPWState, prExpandedTile, &wE));
	    ERR_PASS (w_pixrFramedHeight (pIPWState, prExpandedTile, &hE));
	    ERR_PASS (w_pixrGetBpl (pIPWState, prExpandedTile, &bplE));
	}

	/* Expand the tile */
	if (s->expansion == 1)
	{
	    ERR_PASS_CLEANUP( i_rescon2To3 (pIPWState, pT, wT, hT, bplT, depth,
					    pE, wE, hE, bplE, depth,
					    s->pResConTab));
	    prETile = prExpandedTile;
	    eTileW = tileW * 3 / 2;
	}
	else if (s->expansion == 2)
	{
	    ERR_PASS_CLEANUP( i_rescon1To2 (pIPWState, pT, wT, hT, bplT, depth,
					    pE, wE, hE, bplE, depth,
					    s->pResConTab));
	    prETile = prExpandedTile;
	    eTileW = tileW * s->expansion;
	}
	else
	{
	    prETile = prTile;
	    eTileW = tileW;
    
	}

	/* Threshold that tile */
	ERR_PASS_CLEANUP (w_threshold (pIPWState, prETile, 0, 
				       thresholds[tile], THRESH_NORM,
				       &prThresh));
    
	/* Copy it into the thresholded strip */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prThresh, PIX_SRC,
				      eCol, stripOffset, eTileW, eTileH,
				      s->prThresholdedStrip));

	eCol += eTileW;
    }
    s->stripOffset += eTileH;

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTile));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prThresh));

    if (s->expansion)
    {
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prExpandedTile));
    }

    /* If we've filled the strip, or are on the last band,
       do some segmentation and/or copying.  */
    if (((s->stripHeight - s->stripOffset) < eTileH) ||
	(s->row + eTileH >= s->eHeight) )
    {
	ERR_PASS_CLEANUP( wp_binarizeStripCopy(pIPWState, state));
    }
    
    s->row += eTileH;
    
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp1) w_destroyPixr (pIPWState, &prTmp1);
    if (prThresh) w_destroyPixr (pIPWState, &prThresh);
    if (prExpandedTile) w_destroyPixr (pIPWState, &prExpandedTile);
    if (prTile) w_destroyPixr (pIPWState, &prTile);

    /* Finish up stuff, and assume that the caller won't call the "finish"
       routine */
    if (s->dithering)
    {
	Int32 w, h;
	w_pixrGetWidth (pIPWState, s->prDitheredStrip, &w);
	w_pixrGetHeight (pIPWState, s->prDitheredStrip, &h);
	w_ditherFloydFinish (pIPWState, w, h, &s->ditherState, 
			     s->prDitheredStrip);
	if (s->prDitheredStrip) w_destroyPixr (pIPWState, &s->prDitheredStrip);
    }
    
    if (s->prThresholdedStrip) 
	w_destroyPixr (pIPWState, &s->prThresholdedStrip);
    if (s->prExpandedBand) w_destroyPixr (pIPWState, &s->prExpandedBand);
    if (s->pResConTab) IPW_FREE (s->pResConTab, cAllocatorData);
    IPW_FREE (s, cAllocatorData);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Free everything and return the resulting image.
 *
 * LongDescription:::
 *    We've completed our binarization. Free everything and return the
 *    resulting image.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  ppBinState::     The binarization state strucure, which is zeroed on exit
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_binarizeFinish (
    IPWState           *pIPWState,
    void              **ppBinState)
{
    PROLOGUE;
    Int32     w, h;
    BinState *s = (BinState *) *ppBinState;
    
    ERR_GEN (s == NULL, cErrNullPtr); 

    if ( s->stripOffset >0 )
    {
	ERR_PASS (wp_binarizeStripCopy(pIPWState, *ppBinState ));
    }

    if (s->dithering)
    {
	ERR_PASS (w_pixrGetWidth (pIPWState, s->prDitheredStrip, &w));
	ERR_PASS (w_pixrGetHeight (pIPWState, s->prDitheredStrip, &h));
	ERR_PASS (w_ditherFloydFinish (pIPWState, w, h, &s->ditherState,
				       s->prDitheredStrip));
	
	if (s->prDitheredStrip) 
	    ERR_PASS (w_destroyPixr (pIPWState, &s->prDitheredStrip));
    }
    
    if (s->prThresholdedStrip) 
	ERR_PASS (w_destroyPixr (pIPWState, &s->prThresholdedStrip));

    if (s->prExpandedBand) 
	ERR_PASS (w_destroyPixr (pIPWState, &s->prExpandedBand));

    if (s->pResConTab) IPW_FREE (s->pResConTab, cAllocatorData);
    IPW_FREE (s, cAllocatorData);
    *ppBinState = NULL;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::: "w_binarizeGetCurImgH"
 *
 * OneLineDescription:::
 *  Get current binarized image (accumated strip by strip ) height.
 *
 * LongDescription:::
 *  In autoscan, we need current binarized image height to aligment gray 
 *  image. binarizeBand will not give this imformation out. This function
 *  output current binarized image height.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  state::          Binarization state structure, as created by
 *                   w_binarizeInit()
 *
 * Output:::
 *  height::         Current binarized image height
 *
 * Returns:::
 *  Error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  zhu::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_binarizeBand"></A>
 *****************************************************************************/
ErrCode CDECL
w_binarizeGetCurImgH (
    IPWState           *pIPWState,
    void               *state,
    Int32              *height)
{
    PROLOGUE;
    BinState *s = (BinState *) state;

    pIPWState = pIPWState; /* make compiler happy */
    *height = s->row;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation::: "w_binarize"
 *
 * OneLineDescription:::
 *  Non-banded wrapper to banded binarization
 *
 * LongDescription:::
 *  The whole "banded binarize" stuff was useful, but sometimes you just
 *  want to binarize an image that's in memory. This call does just that.
 *  Note that it's not very efficient: it copies the image into a band
 *  buffer, and calls the banded code with that. It would be better to make
 *  a fake PIXR which points inside the image.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Source image
 *  expand::         If zero, don't expand the image. Otherwise, expand
 *                   the image before thresholding: 1 = 1.5x, 2 = 2x
 *  pTRC::           Tone-correction curve, if dithering
 *  textOnly::       If true, do a text-only binarization,
 *                   otherwise do a text/picture binarization.
 *  dithering::      If true, always dither even in text-only mode
 *  zapPepper::      If true, remove pepper in the text part. This
 *                   is only applicable if doing textOnly.
 *
 * Output:::
 *  pPrResult::      Resulting image.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_binarize"></A>
 *****************************************************************************/
ErrCode CDECL
w_binarize (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32               expand,
    UInt8              *pTRC,
    Bool                textOnly,
    Bool                dithering,
    Bool                zapPepper,
    PIXR              **pPrResult)
{
    PROLOGUE;
    Int32           width, height, depth, row, bandH;
    ScaledInt1616   xRes, yRes, dpi;
    PIXR           *prBand = NULL, *prResult = NULL;
    void           *pBinState;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth4|cDepth8, 
		  cChannels1, cFrame32);
    ERR_GEN (pPrResult == NULL, cErrNullPtr);
    ERR_GEN (*pPrResult != NULL, cErrPtrNotNull);
    ERR_PASS (w_pixrGetWidth (pIPWState, prSrc, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, prSrc, &height));
    ERR_PASS (w_pixrGetDepth (pIPWState, prSrc, &depth));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xRes, &yRes));
    dpi = (ScaledInt1616_TO_Int32(xRes) > ScaledInt1616_TO_Int32(yRes)) ?
	xRes : yRes;

    
    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, width,  cBandHeight,
					  depth, xRes, yRes,
					  cFrameBits, &prBand));

    ERR_PASS_CLEANUP (w_binarizeInit (pIPWState,  width, height, 
				      depth, dpi, expand, pTRC, textOnly, 
				      dithering, zapPepper, 
				      &prResult, &pBinState));

    for (row = 0; row < height; row += cBandHeight)
    {
	bandH = MIN(cBandHeight, height - row);
	if (bandH < cBandHeight)
	{
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBand));
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, width, 
						  bandH, depth,
						  xRes, yRes, 
						  cFrameBits, &prBand));
	}
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      0, row, prSrc,
				      PIX_SRC, 
				      0, 0, width, bandH, 
				      prBand));
	ERR_PASS_CLEANUP (w_binarizeBand (pIPWState, pBinState, prBand));
    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prBand));
    ERR_PASS_CLEANUP (w_binarizeFinish (pIPWState, &pBinState));
    *pPrResult = prResult;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prBand) w_destroyPixr (pIPWState, &prBand);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation::: "w_binarize2"
 *
 * OneLineDescription::: 
 *  Non-banded adaptive binarization using another algorithm, 8-bit only
 *
 * LongDescription:::
 *  This algorithm performs significantly better (as measured in OCR 
 *  accuracy) than the above algorithm, but it's not implemented as
 *  a banded process yet.
 * <BR>
 *  At the moment, this algorithm only supports 8 bit/pixel gray images.
 *
 * PrivateDescription:::
 *  The tile-based algorithm had several shortcomings. If text was near
 *  other higher-contrast features, it could be thresholded poorly.
 *  The tile-neighborhood code didn't work particularly well. This 
 *  algorithm does something a little bit different. The general idea
 *  is to find "candidate threshold" sites, that is, places in the image
 *  where we're pretty sure we can compute a good threshold. We compute
 *  the threshold only in those regions, and then propagate the threshold
 *  to adjacent pixels by calling w_grayPropagate4(). 
 *  <BR>
 *  In a bit more detail, the method is as follows:
 *  <BR>
 *  1. Reduce the image to around 100 dpi.<BR>
 *  2. Create a gradient image by computing the difference between a
 *     Max-reduction and a Min-Reduction.<BR>
 *  3. Threshold this gradient image using a magical gradient threshold
 *     constant. NOTE that this constant hasn't really been tuned much.
 *     If the constant is too high, low-contrast text won't be thresholded
 *     correctly, but the lower the contrast, the more low-contrast crud
 *     shows up as black.<BR>
 *  4. Find the connected components in the thresholded gradient image.<BR>
 *  5. Create a "threshold map" image, which should be set to all zero.<BR>
 *  5. For each connected component, compute a threshold on the bounding
 *     rectangle by searching for largest adjacent pixel difference (in 4 
 *     directions) and computing the threshold as in the code above.<BR>
 *  6. Paint each threshold computed into the threshold map, in the bounding
 *     rectangle of the connected component.<BR>
 *  7. Call w_grayPropagate4() on the threshold map to fill in the zero
 *     pixels with a nearby threshold.  <BR>
 *  8. Use that threshold map to binarize the image.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to threshold
 *  expansion::      Expansion factor, 0 = no expansion, 1 = 1.5x, 2 = 2x
 *  gradThresh::     Gradient threshold to use. Suggest 32?
 *
 * Output:::
 *  pPrThresholded::  Binarized image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_binarize2"></A>
 *****************************************************************************/
ErrCode CDECL
w_binarize2 (
    IPWState  *pIPWState,
    PIXR      *prSrc,
    Int32      expansion,
    Int32      gradThresh,
    PIXR     **pPrThresholded)
{
    PROLOGUE;
    PIXR          *prMin = NULL, *prMax = NULL, *prDiff = NULL, 
	          *prReduced = NULL,*prThresholds = NULL, *prThreshMask = NULL, 
	          *prThresholded = NULL, *prTmp = NULL, *prPrimary = NULL;
    Int32          dpi, thresh, w, h, bpl, depth, maxV;
    Int32         *pHist = NULL, f1, f2, i;
    UInt8         *imageP, *imagePAll /*, *bcTable = NULL*/;
    RECTANGL      *rectList, *rectP, rect;
    ColorRecord    color; 
    ScaledInt1616  xDPI, yDPI;
    Bool           doSmooth, isScaledUp;
    Int32          n1, n2, orgW, orgH, orgBpl, row, col;
    UInt8          prev, curr, *primP, *tmpP;
    Int32         *hTab=NULL, *vTab, *histTab;

/* Macro used below */
#define ACCUMDIFF(offset) \
    otherPixel = UCHAR_ACCESS(offset);\
    delta = pixel - otherPixel;\
    if (delta)\
    {\
	delta = ABS(delta);\
        if (delta > maxDelta)\
        {\
            sum = pixel + otherPixel;\
            count = 1;\
            maxDelta = delta;\
        }\
        else if (delta == maxDelta)\
        {\
            sum += (pixel + otherPixel);\
            count ++;\
        }\
    }

    /* Check stuff */
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth8, cChannels1|cChannels3, 
		  cFrame32);
    ERR_GEN (pPrThresholded == NULL, cErrNullPtr);
    ERR_GEN (*pPrThresholded != NULL, cErrPtrNotNull);

    if( expansion & cIsScaledImage )
    {
	expansion &= 0x3fff;
	isScaledUp = cTrue;
    }
    else
	isScaledUp = cFalse;
	
    /* Get the "primary" channel so we don't operate on RGB for no reason */
    ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, prSrc, &prPrimary));
    ERR_PASS (w_pixrGetWidth (pIPWState, prPrimary, &orgW));
    ERR_PASS (w_pixrGetHeight (pIPWState, prPrimary, &orgH));
    ERR_PASS (w_pixrGetBpl (pIPWState, prPrimary, &orgBpl));
    ERR_PASS (w_pixrGetImage (pIPWState, prPrimary, &primP));

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prPrimary, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    /* Reduce */
    if (dpi > 300)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prPrimary, 200, 100, &prTmp));
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prTmp, 200, 100, &prReduced));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	
    }
    else if (dpi >= 250)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prPrimary, 300, 100, 
				     &prReduced));
    }
    else if (dpi >= 150)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prPrimary, 200, 100, 
				     &prReduced));
    }
    else
    {
	prReduced = prPrimary;
    }

    /*
     * Compute a gradient by taking the difference between max-reduction
     * and min-reduction.
     */

#if 1
    ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prReduced, 1, 
				      cFalse, cFalse, cFalse, cFalse, &prMin));
    ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prReduced, 1, 
				      cTrue, cTrue, cTrue, cTrue, &prMax));
    ERR_PASS_CLEANUP (w_subtractGray (pIPWState, prMax, prMin, &prDiff));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMin));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMax));
#else
    /* Use Doug's gradient finding function instead */
    ERR_PASS_CLEANUP (w_findMagnitudeOfGradient (pIPWState, prReduced,
						 cGradOpMaxL1FromCenter,
						 cGradNeighborhoodDownRight,
						 &prDiff));
#endif		      
    /* Threshold and invert the gradient image */
    ERR_PASS_CLEANUP (w_pixrGetDepth (pIPWState, prReduced, &depth));
    ERR_PASS_CLEANUP (w_invert (pIPWState, prDiff, &prDiff));
    ERR_PASS_CLEANUP (w_ccTextThreshold (pIPWState, prDiff, 20, 
					  cFalse, &thresh));
    maxV = 256;
    if( depth == 4) maxV = 16;
/* low resolution magic number */
    if( thresh > (maxV -28)) thresh = (maxV -28); 
    else if(thresh < (maxV -gradThresh))
    {
	ERR_PASS_CLEANUP (w_histo (pIPWState, prDiff, NULL, &pHist));
	f1 = f2 =0;
	for(i=0; i<maxV; ++i)
	{
	    f1 += pHist[i];
	    if( i <= thresh ) f2 += pHist[i];
	}
	f1 /= 25;  /* at lease 4% */
	while( f2 < f1 && thresh < gradThresh  )
	{
	    ++ thresh;  
	    f2 += pHist[thresh];
	}
	IPW_FREE(pHist, cAllocatorData);
    }
    ERR_PASS_CLEANUP (w_threshold (pIPWState, prDiff, 0, thresh,
				   THRESH_NORM, &prTmp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDiff));

    doSmooth = cFalse;
    if( dpi > 300 && isScaledUp == cFalse)
    {
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prReduced, &w));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prReduced, &h));
	ERR_PASS_CLEANUP (w_histo (pIPWState, prReduced, NULL, &pHist)); 
	f1 =0;
	f2 = (w*h) >>1;
	for(i=0; i<256; ++i)
	{
	    f1 += pHist[i];
	    if(f1 >= f2) break;
	}
	if( i < cMedianThreshold ) 
	{
	    doSmooth = cTrue;
	    ERR_PASS_CLEANUP (w_expand(pIPWState, prTmp, 4, &prMin));
	}
	IPW_FREE(pHist, cAllocatorData);
    }

    /* Find the connected components */
    ERR_PASS_CLEANUP (w_rabb (pIPWState, prTmp, 2, 2, &rectList));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prReduced, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prReduced, &h));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prReduced, &bpl));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prReduced, &imageP));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prReduced, &imagePAll));

    /* Create table for thresholds */
    hTab = (Int32 *) IPW_MALLOC ((h+w+258)*sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (hTab == NULL, cErrNomem);
    vTab = hTab + w +1;
    histTab = vTab + h +1;
    for(i=0; i<=w; ++i)
	hTab[i] = ((i * orgW)/w);

    for(i=0; i<=h; ++i)
	vTab[i] = ((i * orgH)/h);

	/* This *may* be a mistake.  There is clearly some confusion
	 * above in how the gradient image is computed above.  In one case,
	 * it's half the size of prReduced (the reduceMinMax case).
	 * In the other case (w_findMagnitudeOfGradient), it's the
	 * same size as prReduced.
	 * The following code creates threshold and distance images
	 * that are the half the size of prReduced.  We later
	 * paint rectangles in this image whose dimensions have been
	 * scaled by 2x relative to the gradient image, which is the
	 * same size as this image.  If we made this image the same
	 * size as prReduced, then 1) it would work the same as the
	 * strip-by-strip version of binarize2 (in which the threshold
	 * image is the same size as prReduced) and 2) the comments
	 * below about when to scale the rectangle would make sense.
	 * As it is, we're painting double-size rectangles into this
	 * image, which is sort of weird.
	 */
    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w/2, h/2, 8,
					  50<<16, 50<<16,
					  cFrameBits, &prThresholds));
    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w/2, h/2, 1,
				    50<<16, 50<<16,
				    cFrameBits, &prThreshMask));
    /* 
     * For each connected component, compute a threshold by searching
     * for the maximal adjacent pixel differences (h, v, down and left,
     * down and right) and taking the average of those pixels
     */
    if( maxV ==  256 ) maxV -= 50;
    for (rectP = rectList; rectP; rectP = rectP->next)
    {
	/* Reject large stuff; it's probably image. 
	 * This should probably be computed after histogramming
	 * the connected components or something. 
	 */
	rectP->mass = 0;
	if (rectP->w > cMaxThreshCandidateSize ||
	    rectP->h > cMaxThreshCandidateSize)
	    continue;

	/* Ignore things that are very long and thin; probably rules */
	if (rectP->w > rectP->h * 8 || rectP->h > rectP->w * 8)
	    continue;

#if 1
	rect.x = rectP->x * 2;
	rect.y = rectP->y * 2;
	rect.w = rectP->w * 2;
	rect.h = rectP->h * 2;
#else
	/* not reduced if using direct gradient */
	rect.x = rectP->x;
	rect.y = rectP->y;
	rect.w = rectP->w;
	rect.h = rectP->h;
#endif
	/* Dilate the rectangle a bit, and clip */
	if (rect.x < 0) rect.x = 0;
	if (rect.y < 0) rect.y = 0;
	if (rect.x >= w) rect.x = w-1;
	if (rect.y >= h) rect.y = h-1;
	if ( (rect.x + rect.w) > w) rect.w = w - rect.x;
	if ( (rect.y + rect.h) > h) rect.h = h - rect.y;      

#if 0
	/* Compute the gradient on the reduced version */
	p = imageP + ((rect.y * bpl) + rect.x);
	
	maxDelta = 0;
	sum = 0; count = 0;
	for (i = 0; i < rect.h; i++)
	{
	    for (j = 0; j < rect.w; j++)
	    {
		pixel = UCHAR_ACCESS (p + j); /* Was via bctable */
		ACCUMDIFF (p + j + 1);
		ACCUMDIFF (p + j + bpl);
		ACCUMDIFF (p + j + bpl + 1);
		ACCUMDIFF (p + j + bpl - 1);
	    }
	    p += bpl;
	}

	/* Compute the threshold */
	thresh = sum / (count * 2);
#else
	ERR_PASS_CLEANUP (w_findEdge2Thresholds (pIPWState, prReduced, &rect, 
						 1, &f1, &thresh));

	if((thresh - f1 ) > cThresholdsDiff)
	{
	    rect.x = hTab[rect.x] -2;
	    rect.y = vTab[rect.y] -2;
	    rect.w = hTab[rect.w] +2;
	    rect.h = vTab[rect.h] +2;
	    if( rect.x <0) rect.x =0;
	    if( rect.y <0) rect.y =0;
	    
	    MEMSET((void *)histTab, 0, 256*sizeof(Int32));
	    tmpP = primP + rect.y * orgBpl + rect.x;
	    n1 = n2 = 0;
	    for(row = 0; row< rect.h; ++row, tmpP += orgBpl)
	    {
		prev = UCHAR_ACCESS(tmpP);
		++ histTab[prev];
		if( prev < f1 ) ++ n1;
		if( prev < thresh ) ++n2;         
		for(col=1; col< rect.w; ++col)
		{
		    curr = UCHAR_ACCESS(tmpP + col);
		    ++ histTab[curr];
		    if( prev >= f1 && curr < f1 ) ++n1;
		    if( prev >= thresh && curr < thresh ) ++n2;
		    prev = curr;
		}
	    }
	    col = n1 >> 3;
	    if( col < 2) col=2;
	    if( (n2 - n1)>col ) thresh = f1;
	    else
	    {
		n1 = (rect.w * rect.h)/2;
		i = n2 = 0;
		while(i <= f1 ) n2 += histTab[i++];
		if( n2 >= n1 ) thresh = f1;
	    }
	}

#endif
	/* If threshold very high, it may be noise in edge, pain this 
	 * rectagle after propgation. */
	if( thresh >= maxV )
	{
	    rectP->mass = f1;
	    continue;
	}
	/* Seed the threshold in the threshold map */
#if 0
	ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, prThresholds,
					  rect.x + (rect.w / 2),
					  rect.y + (rect.h / 2),
					  thresh));
#else	    
	color[0] = color[1] = color[2] = color[3] = (UInt8) thresh;
	ERR_PASS_CLEANUP (w_paint (pIPWState, prThresholds, rectP, color));
	color[0] = color[1] = color[2] = color[3] = 1;
	ERR_PASS_CLEANUP (w_paint (pIPWState, prThreshMask, rectP, color));
#endif

    }

    IPW_FREE (hTab, cAllocatorData);
    if (prReduced != prPrimary)
    {
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));
    }

    /*
     * Propagate the thresholds to regions of the image that 
     * have no threshold computed. We use a 4-connected propagation
     * method.
     */
    ERR_PASS_CLEANUP (w_grayPropagate4MaskBi (pIPWState, prThreshMask,
					    cTrue, prThresholds));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prThreshMask));

    /* We marked the regions that were very light (had thresh >= maxV)
     * above by setting rectP->mass equal to that thresh.  We didn't
     * put those threshold into the threshod map because we didn`t
     * want those very light thresholds to propagate.  However, we
     * do want to use those thresholds locally, so insert them into
     * the already-propagated threshold map now.
     *
     * Potential problem: if these rectangle overlap existing good
     * thresholds, the existing thresholds will be replaced.  That
     * could be a mistake.
     */
    for (rectP = rectList; rectP; rectP = rectP->next)
    {
	if (rectP->mass == 0)
	    continue;	/* ignores things that are too big, or have
			 * big aspect ratios, or have already been
			 * propagated.
			 */
	/*  
	 * Paint region which was saved because the object was too light.
	 */
	color[0] = color[1] = color[2] = color[3] = (UInt8) rectP->mass;
	ERR_PASS_CLEANUP (w_paint (pIPWState, prThresholds, rectP, color));

    }
    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &rectList));

    /* Do the thresholding */
    if (expansion)
    {
	PIXR *prExpanded = NULL;
	Int32 f1, f2;

	if (expansion == 1)
	{
	    f1 = 200; f2 = 300;
	}
	else
	{
	    f1 = 1; f2 = expansion;
	}

	
	/* Achtung! Uses gobs of memory. Fix if useful.. */
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prPrimary, f1, f2, 
				     &prExpanded));
	ERR_PASS_CLEANUP (w_thresholdByPixel (pIPWState, prExpanded, 
					      prThresholds, 
					      THRESH_NORM, &prThresholded));
	ERR_PASS_CLEANUP (w_destroyPixr(pIPWState, &prExpanded));
	
    }
    else if( doSmooth && dpi > 300)
    {
	Int32   priW, priH;
	PIXR   *prSmooth = NULL;

	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prPrimary, &priW));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prPrimary, &priH));
	prTmp = NULL;
	ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, prMin, 
					 priW, priH, &prTmp));
	ERR_PASS_CLEANUP(w_pixrSetResolutionDPI (pIPWState,
						 prTmp, xDPI, yDPI));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMin));
	ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp, &prMin));
	ERR_PASS_CLEANUP (w_smooth (pIPWState, prPrimary, 3, &prSmooth));
	prMax = NULL;
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prSmooth, prTmp, 
				       PIX_SRC | PIX_DST, &prMax));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSmooth));
	ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prPrimary, prMin, 
				       PIX_SRC | PIX_DST, &prSmooth));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMin));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMax, OR, prSmooth));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMax));
	ERR_PASS_CLEANUP (w_thresholdByPixel (pIPWState, prSmooth, 
					      prThresholds, THRESH_NORM, 
					      &prThresholded));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSmooth));
    }
    else
    {
	ERR_PASS_CLEANUP (w_thresholdByPixel (pIPWState, prPrimary, 
					      prThresholds, THRESH_NORM, 
					      &prThresholded));
    }

    /* That's that */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prThresholds));
    *pPrThresholded = prThresholded;
    NORMAL_RETURN;

ERR_CLEANUP:
    /* Free all of the pixrs that might be lying around */
    if (prMin) w_destroyPixr (pIPWState, &prMin);
    if (prMax) w_destroyPixr (pIPWState, &prMax);
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    if (prDiff) w_destroyPixr (pIPWState, &prDiff);
    if (prReduced) w_destroyPixr (pIPWState, &prReduced);
    if (prThresholds) w_destroyPixr (pIPWState, &prThresholds);
    if (prThreshMask) w_destroyPixr (pIPWState, &prThreshMask);
    if (prThresholded) w_destroyPixr (pIPWState, &prThresholded);
    if (pHist) IPW_FREE (pHist, cAllocatorData);
    if (hTab) IPW_FREE (hTab, cAllocatorData);

    ERR_RETURN;
}

/* The state structure for banded binarize2 */
typedef struct
{
    Int32      width, height;        /* Size of image */
    Int32      eWidth, eHeight;      /* Size of expanded image */
    ScaledInt1616 dpi;               /* Image DPI */
    Int32      depth;                /* Bit depth */
    Int32      reductions;           /* # of reductions to do for threshold */
    Int32      row;                  /* Current band row */
    Int32      expansion;            /* Expansion factor before binarizing:
                                        0 = none, 1 = 1.5, 2 = 2x */
    Int32      bandHeight;           /* Height of an expanded band */
    Int32      gThresh;              /* Gradient threshold to use */
    PIXR      *prResult;             /* Full output binary image */
    PIXR      *prThresholds;         /* Computed thresholds */
    PIXR      *prDistances;          /* Computed thresholds */
    PIXR      *prLastRow;            /* Thresholds for previous row */
    PIXR      *prLastDistances;      /* Distances for previous row */
    UInt32    *pResConTab;           /* Resolution conversion table */
    Bool       isScaledUp;           /* Input image has been scaled */
    PIXR      *prSmoothed;           /* Resolution > 300, binary */
    Int32     *pHist;                /* Image histogram */
    Int32     *hTab;                 /* Scaling table */
    Int32     *vTab;                 /* Scaling table */
} Bin2State;

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Intialize banded binarization, second algorithm
 *
 * LongDescription:::
 *  The idea behind banded binarization is to take bands of 8-bit
 *  data and produce a binary version which is adaptively thresholded.
 *  <BR>
 *  This function initializes banded binarization. Creates a state structure 
 *  and associated stuff, including the output binary image. 
 *
 * PrivateDescription:::
 *  The tile-based algorithm had several shortcomings. If text was near
 *  other higher-contrast features, it could be thresholded poorly.
 *  The tile-neighborhood code didn't work particularly well. This 
 *  algorithm does something a little bit different. The general idea
 *  is to find "candidate threshold" sites, that is, places in the image
 *  where we're pretty sure we can compute a good threshold. We compute
 *  the threshold only in those regions, and then propagate the threshold
 *  to adjacent pixels by calling w_grayPropagate4(). 
 *  <BR>
 *  In a bit more detail, the method is as follows:
 *  <BR>
 *  1. Reduce the image to around 100 dpi.<BR>
 *  2. Create a gradient image by computing the difference between a
 *     Max-reduction and a Min-Reduction. <BR>
 *  3. Threshold this gradient image using a magical gradient threshold
 *     constant. NOTE that this constant hasn't really been tuned much.
 *     If the constant is too high, low-contrast text won't be thresholded
 *     correctly, but the lower the contrast, the more low-contrast crud
 *     shows up as black.<BR>
 *  4. Find the connected components in the thresholded gradient image.<BR>
 *  5. Create a "threshold map" image and a distance image, which should
 *     be set to all ones.<BR>
 *  6. For each connected component, compute a threshold on the bounding
 *     rectangle by searching for largest adjacent pixel difference (in 4 
 *     directions) and computing the threshold as in the code above.<BR>
 *  7. Paint each threshold computed into the threshold map and clear the
 *     corresponding pixels in the distance map, in the bounding
 *     rectangle of the connected component.<BR>
 *  8. Call w_grayPropagate4Mask() on the threshold map and distance map
 *     to fill in the non-seed pixels with a nearby threshold.  <BR>
 *  9. Use that threshold map to binarize the image.
 *
 *  
 * Input:::
 *  pIPWState::      State structure
 *  width::          Width of source image, without frame
 *  height::         Height of source image, without frame
 *  depth::          Depth, in bits, of source image
 *  dpi::            DPI of the image, used to determine reductions, &c.
 *  expand::         If zero, don't expand the image. Otherwise, expand
 *                   the image before thresholding: 1 = 1.5x, 2 = 2x
 *  threshold::      Gradient threshold to use. 64 works well for 300dpi.
 *                   Others should, at this point, be tuned.
 *  pPrResult::      If non-NULL, use this as a result PIXR, otherwise
 *                   create one
 *
 * Output:::
 *  ppBinState::     State structure to pass to the band and finish 
 *                   routines, or NULL if error
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  in development
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_binarize2Init (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    ScaledInt1616       dpi,
    Int32               expand,
    Int32               threshold,
    PIXR              **pPrResult,
    void              **ppBinState)
{
    PROLOGUE;
    Bin2State  *s = NULL;
    UInt32      resconTableSize;
    Int32       unscaledDpi, w, h, i;

    ERR_GEN_CLEANUP ((expand & 0x3fff) < 0 || (expand & 0x3fff) > 2, 
		     cErrExpansionNotSupported);

    ERR_GEN_CLEANUP (depth != 8, cErrPixr_depth);
    
    /* Initialize the strels we'll need */
    ERR_PASS_CLEANUP (w_initStrelByName (pIPWState, strel3));

    /* Allocate the state structure. We use the same structure as
     * the original banded binarize, though we won't use all of
     * the fields.
     */
    s = (Bin2State *) IPW_CALLOC (1, sizeof (Bin2State), cAllocatorData);
    ERR_GEN_CLEANUP (s == NULL, cErrNomem);

    if( expand & cIsScaledImage )
    {
	expand &= 0x3fff;
	s->isScaledUp = cTrue;
    }
    else
	s->isScaledUp = cFalse;

    s->width = width;
    s->height = height;
    s->dpi = dpi;
    s->depth = depth;
    s->row = 0;
    s->prLastRow = NULL;
    s->prLastDistances = NULL;
    s->prThresholds = NULL;
    s->prDistances = NULL;
    s->prSmoothed = NULL;
    s->pHist = NULL;
    s->hTab = NULL;
    s->gThresh = threshold;
    s->expansion = expand;
    unscaledDpi = ScaledInt1616_TO_Int32 (dpi);
    if (unscaledDpi > 300)
    {
	w = width / 4;
	h = height / 4;
    }
    else if (unscaledDpi >= 250)
    {
	w = width / 3;
	h = height / 3;
    }
    else if (unscaledDpi >= 150)
    {
	w = width / 2;
	h = height / 2;
    }
    else
    {
	w = width;
	h = height;
    }
    s->hTab = (Int32 *) IPW_MALLOC ((h+w+4)*sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (s->hTab == NULL, cErrNomem);
    s->vTab = s->hTab + w +2;
    for(i=0; i<(w+2); ++i)
	s->hTab[i] = ((i * width)/w);

    for(i=0; i<(h+2); ++i)
	s->vTab[i] = ((i * height)/h);

    if (s->expansion == 1)
    {
	s->eWidth = width * 3 / 2;
	s->eHeight = height * 3 / 2;
	s->pResConTab = NULL;
	i_getResconTableSize (pIPWState, 2, 3, s->depth, &resconTableSize);
	if (resconTableSize)
	{
	    s->pResConTab = (UInt32 *)IPW_MALLOC (resconTableSize, 
						  cAllocatorData);
	    ERR_GEN (s->pResConTab == NULL, ia_nomem);
	    ERR_PASS_CLEANUP( i_make2To3ResconTable (pIPWState, &s->pResConTab, 
						     s->depth));
	}
    }
    else if (s->expansion)
    {
	s->eWidth = width * s->expansion;
	s->eHeight = height * s->expansion;
	s->pResConTab = NULL;
	i_getResconTableSize (pIPWState, 1, 2, s->depth, &resconTableSize);
	if (resconTableSize)
	{
	    s->pResConTab = (UInt32 *)IPW_MALLOC (resconTableSize,
						  cAllocatorData);
	    ERR_GEN (s->pResConTab == NULL, ia_nomem);
	    ERR_PASS_CLEANUP (i_make1To2ResconTable (pIPWState, &s->pResConTab, 
						     s->depth));
	}
    }
    else
    {
	s->eWidth = width;
	s->eHeight = height;
    }


    if (*pPrResult != NULL)
    {
	s->prResult = *pPrResult;
    }
    else
    {
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, s->eWidth, s->eHeight,
					      1, dpi, dpi,
					      cFrameBits, &(s->prResult)));
    }

    if( ScaledInt1616_TO_Int32 (s->dpi) > 300 && s->isScaledUp == cFalse)
    {
	ERR_PASS_CLEANUP (w_clonePixr (pIPWState, s->prResult, 
				       &(s->prSmoothed)));	
    }

    *ppBinState = (void *) s;
    *pPrResult = s->prResult;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (s)
    {
	if (s->prResult) w_destroyPixr (pIPWState, &s->prResult);
	if (s->pResConTab) IPW_FREE (s->pResConTab, cAllocatorData);
	IPW_FREE (s, cAllocatorData);
    }
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Process a band for binarization
 *
 * LongDescription:::
 *  The idea behind banded binarization is to take bands of 8-bit
 *  data and produce a binary version which is adaptively thresholded.
 *  <BR>
 *  This function processes a band. The band should be as wide as we said in
 *  the initialization. There are no requirements on the height, but it
 *  doesn't work all that well with bands of less than around 64 lines.
 *  128 works quite well.  In particular, there is no requirement that the
 *  bands be the same height.
 *
 * PrivateDescription:::
 *  Making the banded voronoi-based thresholder work requires either
 *  taking a second pass through the data, which we don't want to do
 *  because it adds too many requirements on the host, or using an
 *  approximation. The approximation we are attempting is to compute 
 *  each band's threshold map, and then seeding the previous threshold map
 *  into the next one. The problem with this scheme is that it
 *  means that we don't get thresholds propagated backwards. Another
 *  scheme would be to do a one-band delay on the thresholds.
 *  <BR>---<BR>
 *  The "seed the previous map" solution didn't work all that well,
 *  because it could move the seeds closer, and have bad thresholds
 *  "leak" down into the text. The fix was to use a sequential two-pass
 *  propagation algorithm, where we could take the result of the down-sweep
 *  and use it to seed the next down sweep, but since we kept the distances
 *  as well as the values, the seeds wouldn't "move." This scheme works
 *  reasonably well.
 *  <BR>---<BR>
 *  NOTE that there is an interesting difference between this "by-strip"
 *  thresholding and plain old binarize2.  The plain method reduces the
 *  input image to around 100 dpi, then creates a gradient image at
 *  1/2 this size (in each dimension).  It then creates a threshold image
 *  at the gradient image resolution.  This stripified version also
 *  reduces the input image to around 100 dpi and produces a gradient image
 *  at about 1/2 that size.  However, it produces a threshold image back
 *  at 100 dpi.  This means that the gray propagation op chews on 4 times
 *  (2w * 2h) as many pixels in the strip version.
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  state::          Binarization state structure, as created by
 *                   w_binarizeInit()
 *  prBand::         The band to process
 *
 * Output:::
 *  pThreshRect::    Rectangle in result image (from w_binarize2Init)
 *                   that was just thresholded. This is used if the
 *                   caller needs to get at the thresholded data
 *                   during the banded processing. If NULL, this 
 *                   information is not passed back. The RECTANGL
 *                   must, therefore, be allocated by the caller.
 *
 * Returns:::
 *  Error status 
 *
 * Notes:::
 *  If the process fails, all data structures are de-allocated, so one
 *  should not call w_binarize2Finish after an error.
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_binarize2Band (
    IPWState           *pIPWState,
    void               *state,
    PIXR               *prBand,
    RECTANGL           *pThreshRect)
{
    PROLOGUE;
    Int32       bandH, bandW, w, h, wT, hT, thresh, maxV,
	        unscaledDpi, eBandH, i, f1, f2, threshBpl, distBpl;
    Bin2State  *s = NULL;
    PIXR       *prTmp = NULL, *prDiff = NULL, *prGradThresh = NULL,
	       *prReduced = NULL, *prExpanded = NULL, *prThresholded = NULL,
	       *prMin = NULL, *prMax = NULL;
    Int32      *pHist = NULL;
    RECTANGL    rect, *rectList = NULL, *rectP = NULL;
    UInt8      *threshP = NULL, *distP = NULL;
    ColorRecord color;
    Int32          n1, n2, orgBpl, row, col, rectW, rectH;
    UInt8          prev, curr, *primP, *tmpP;
    Int32         *histTab = NULL;

    s = (Bin2State *) state;
    ERR_GEN (s == NULL, cErrNullPtr);

    ERR_GEN_PIXR (prBand, cByteOrderBig, cDepth8, cChannels1, cFrame32);
    
    ERR_PASS (w_pixrGetHeight (pIPWState, prBand, &bandH));
    ERR_PASS (w_pixrGetWidth (pIPWState, prBand, &bandW));
    ERR_PASS (w_pixrGetBpl (pIPWState, prBand, &orgBpl));
    ERR_PASS (w_pixrGetImage (pIPWState, prBand, &primP));

    ERR_GEN ((bandW != s->width), 
	     cErrImageTooWide);

    if (s->expansion)
	eBandH = (s->expansion == 1) ? bandH * 3 / 2 : bandH * s->expansion;
    else
	eBandH = bandH;

/* 4-24-98 API may not resize last band and just copy partial image in,
    (s->row+eBandH > s->eHeight) causes crash */
    if( eBandH > (s->eHeight - s->row))
	eBandH = s->eHeight - s->row;

    unscaledDpi = ScaledInt1616_TO_Int32 (s->dpi);

    /* Reduce the strip */
    if (unscaledDpi > 300)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prBand, 200, 100, &prTmp));
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prTmp, 200, 100, &prReduced));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	
    }
    else if (unscaledDpi >= 250)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prBand, 300, 100, 
				     &prReduced));
    }
    else if (unscaledDpi >= 150)
    {
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prBand, 200, 100, 
				     &prReduced));
    }
    else
    {
	prReduced = prBand;
    }

    /* Compute the gradient for the band */
#if 0
    ERR_PASS_CLEANUP (w_findMagnitudeOfGradient (pIPWState, prReduced, 
						 cGradOpMaxL1FromCenter, 
						 cGradNeighborhoodDownRight, 
						 &prDiff));
#else
    ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prReduced, 1, 
				      cFalse, cFalse, cFalse, cFalse, &prMin));
    ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prReduced, 1, 
				      cTrue, cTrue, cTrue, cTrue, &prMax));
    ERR_PASS_CLEANUP (w_subtractGray (pIPWState, prMax, prMin, &prDiff));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMin));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMax));

#endif
    /* Threshold and invert the gradient image */
    ERR_PASS_CLEANUP (w_invert (pIPWState, prDiff, &prDiff));
    ERR_PASS_CLEANUP (w_ccTextThreshold (pIPWState, prDiff, 20, 
					  cFalse, &thresh));

	/* We now check the computed threshold.  We adjust it if necessary
	 * so that two conditions are met.
	 * First, the threshold must be at least 28 below the maximum
	 * pixel value.  28 is a magic number.
	 * Second, we adjust it up so that at least 4% of the pixels in
	 * the image have values that are less than the threshold.
	 */
    maxV = 256;
    if (thresh > (maxV - 28))
	thresh = (maxV - 28); 
    else
    if (thresh < (maxV - s->gThresh))
    {
      ERR_PASS_CLEANUP (w_histo (pIPWState, prDiff, NULL, &pHist));
      f1 = f2 =0;
      for (i=0; i < maxV; ++i)
      {
         f1 += pHist[i];
         if (i <= thresh )
	    f2 += pHist[i];
      }
      f1 /= 25;  /* at lease 4% */
      while( f2 < f1 && thresh < s->gThresh  )
      {
         ++ thresh;  
         f2 += pHist[thresh];
      }
      IPW_FREE(pHist, cAllocatorData);
    }
    ERR_PASS_CLEANUP (w_threshold (pIPWState, prDiff, 0, thresh, 
				   THRESH_NORM, &prGradThresh));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDiff));
    if( unscaledDpi > 300 && s->isScaledUp ==  cFalse)
    {
	ERR_PASS_CLEANUP (w_histo (pIPWState, prReduced, NULL, &pHist)); 
	ERR_PASS_CLEANUP (w_expand(pIPWState, prGradThresh, 4, &prMin));
	if( s->pHist == NULL )
	{
	    s->pHist = pHist;
	    pHist = NULL;
	}
	else
	{
	    for(i=0; i<256; ++i)
		s->pHist[i] += pHist[i];
	    IPW_FREE(pHist, cAllocatorData);
	}
    }


    /*
     * Find the connected components in the gradient image. We compute
     * the gradient "around" those connected components.
     */
    ERR_PASS_CLEANUP (w_rabb (pIPWState, prGradThresh, 2, 2, &rectList));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prGradThresh));

    /* 
     * Replicate the boundaries of the reduced images into the frame
     * so that we can ignore edge effects rather than clipping.
     * This is so that w_findEdgeThreshold can safely look around
     * the outside of every object in prReduced without stumbling
     * into the frame.
     * Copy 4 pixels on left/right edge, 2 pixels on top/bottom. We copy
     * the columns first, and then should remember to replicate the
     * copied portions of the row as well, so we don't ignore the
     * corners.
     */
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prReduced, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prReduced, &h));

    /* Left */
    for (i = 0; i < 4; i++)
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prReduced, 
				      PIX_SRC|PIX_DONTCLIP,
				      -1 - i, 0, 1, h, prReduced));
    /* Right */
    for (i = 0; i < 4; i++)
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, w - 1, 0, prReduced, 
				      PIX_SRC|PIX_DONTCLIP,
				      w + i, 0, 1, h, prReduced));
    /* Top */
    for (i = 0; i < 2; i++)
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, -2, 0, prReduced, 
				      PIX_SRC|PIX_DONTCLIP,
				      -2, -1 - i, w + 2, 1, prReduced));
    /* Bottom */
    for (i = 0; i < 2; i++)
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, -2, h - 1, prReduced, 
				      PIX_SRC|PIX_DONTCLIP,
				      -2, h + i, w + 2, 1, prReduced));

    /* Create threshold and distance maps if either they don't exist
     * or are the wrong size.  If the client keeps calling us with
     * band of the same height (as seems likely), we'll only have to
     * allocate these once (and maybe once for the last partial strip).
     */
    if (s->prThresholds == NULL) 
    {
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 8, 
					      s->dpi, s->dpi, 
					      cFrameBits, &(s->prThresholds)));
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, w, h, 16, 
					      s->dpi, s->dpi, 
					      cFrameBits, &(s->prDistances)));
    }
    else	/* the saved image exist, are they the right size? */
    {
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, s->prThresholds, &wT));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, s->prThresholds, &hT));

	if ((w != wT) || (h != hT))
	{
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &(s->prThresholds)));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &(s->prDistances)));
	    ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 8, 
						  s->dpi, s->dpi, 
						  cFrameBits,
						  &(s->prThresholds)));
	    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, w, h, 16, 
						  s->dpi, s->dpi, 
						  cFrameBits,
						  &(s->prDistances)));
	}
    }

    /* Initialize the distance image to all 1's */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_SET,
				  0, 0, w, h, s->prDistances));

    /*
     * Copy in the last row of the previous threshold map, if it exists.
     * This propagates the thresholds down to the next map, so that
     * we get some context. It would be more accurate to use an extra strip
     * of seeds, and re-propagate both strips worth of seeds at once,
     * but that would double the propagation amount, and hopefully won't
     * be too relevant in practice.
     */
    if (s->prLastRow)
    {
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      0, 0, s->prLastRow,
				      PIX_SRC,
				      0, 0, w, 1, s->prThresholds));
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, s->prLastDistances, 
				      PIX_SRC, 0, 0, w, 1, s->prDistances));
    }

    /* 
     * We have all of the connected components for the thresholded gradient.
     * For each connected component, compute a threshold using the edge-based
     * threshold finder (w_findEdgeThreshold) in the extent of each 
     * connected component.
     * But first, set maxV to some magic value below the real maximum
     * value.  This is used to detect regions in which the threshold is so
     * high so as to indicate the presence of light noise and halftones.  We
     * don't spread out these thresholds, we just paint them in after
     * we're done.
     */
    if( maxV == 256 ) maxV -= 30;
    histTab = (Int32 *) IPW_MALLOC (256*sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (histTab == NULL, cErrNomem);
    for (rectP = rectList; rectP; rectP = rectP->next)
    {
	/* Reject stuff that's large; it's probably image. This should
   	 * probably be computed after histogramming the connected
	 * components or something, rather than being ad-hoc. 
	 * Ignore things that are very long and thin; they're 
	 * probably rules.
	 */
	rectP->mass =0;
	if (rectP->w > cMaxThreshCandidateSize ||
	    rectP->h > cMaxThreshCandidateSize ||
	    rectP->w > rectP->h * 10 || 
	    rectP->h > rectP->w * 10)
	{
	    continue;
	}

	/* Scale the rectangle up to the size of s->prThresholds, and
	 * dilate it rectangle a bit.
	 */
	rect.x = rectP->x * 2 - 1;
	rect.y = rectP->y * 2 - 1;
	rect.w = rectP->w * 2 + 1;
	rect.h = rectP->h * 2 + 1;
	if( rect.x < 0) rect.x = 0;
	if( rect.y < 0) rect.y = 0;
	if( rect.x >= w) rect.x = w-1;
	if( rect.y >= h) rect.y = h-1;
	if( (rect.x + rect.w) > w) rect.w = w - rect.x;
	if( (rect.y + rect.h) > h) rect.h = h - rect.y;  

	/* Compute that threshold. The "minDiff" is always small,
	 * since we already decided what to accept in the gradient threshold
	 */
#if 0
	ERR_PASS_CLEANUP (w_findEdgeThreshold (pIPWState, prReduced, &rect, 
						 1, &thresh));
#else
	ERR_PASS_CLEANUP (w_findEdge2Thresholds (pIPWState, prReduced, &rect, 
						 1, &f1, &thresh));
	if( (thresh - f1 ) > cThresholdsDiff)
	{
	    MEMSET((void *)histTab, 0, 256*sizeof(Int32));
	    tmpP = primP +  s->vTab[rect.y] * orgBpl + s->hTab[rect.x];
	    n1 = n2 = 0;
	    rectH = s->vTab[rect.h];
	    rectW = s->hTab[rect.w];
	    for(row = 0; row< rectH; ++row, tmpP += orgBpl)
	    {
		prev = UCHAR_ACCESS(tmpP);
		++ histTab[prev];
		if( prev < f1 ) ++ n1;
		if( prev < thresh ) ++n2;         
		for(col=1; col< rectW; ++col)
		{
		    curr = UCHAR_ACCESS(tmpP + col);
		    ++ histTab[curr];
		    if( prev >= f1 && curr < f1 ) ++n1;
		    if( prev >= thresh && curr < thresh ) ++n2;
		    prev = curr;
		}
	    }
	    col = n1 >> 3;
	    if( col < 2) col=2;
	    if( (n2 - n1)>col ) thresh = f1;
	    else
	    {
		i = n2 =0;
		while(i<=f1) n2 += histTab[i++];
		n1 = rectW * rectH /2;
		if( n2 > n1 ) thresh = f1;
	    }
	}
#endif
	/* if the threshold is very light, don't propagate it.  It's
	 * probably some image thing.  We'll paint it into the
	 * threshold image after propagation.
	 */
	if ( thresh >= maxV )
	{
	    /* save it and paint this region after propgation */
	    rectP->mass = f1;
	    continue;
	}

	/* Paint that threshold into the threshold map */
	color[0] = color[1] = color[2] = color[3] = (UInt8) thresh;
	ERR_PASS_CLEANUP (w_paint (pIPWState, s->prThresholds, &rect, color));
	/* Clear that portion of the distance map to mark these
	 * pixels as seeds.
	 */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
				      rect.x, rect.y, rect.w, rect.h,
				      s->prDistances));
    }

    IPW_FREE (histTab, cAllocatorData);
    if (prReduced != prBand)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));

    /* 
     * Propagate the thresholds to fill the threshold image.
     * We can copy the bottom row of one strip to seed the top row 
     * of the next strip without separating out the down and up sweeps,
     * since the bottom row only depends on the down sweep.  We use
     * the ipshared routine because it allows us to pass in the
     * (seeded) distance image.
     */
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, s->prThresholds, 
					 &threshP));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, s->prThresholds, &threshBpl));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, s->prDistances, &distP));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, s->prDistances, &distBpl));
    ERR_PASS_CLEANUP (i_propagateGray4Mask (pIPWState, threshP,
					    w, h, threshBpl, 8, distP,
					    distBpl, 16));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, s->prThresholds, 
					 &threshP));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, s->prDistances, &distP));
    ERR_PASS_CLEANUP (i_grayMinusDist (pIPWState, threshP, w, h, threshBpl,
					   distP, distBpl));

    /*
     * Copy out the last row of the threshold map to use as seeds next time.
     * Also copy the distances for those thresholds.
     */
    if (s->prLastRow == NULL)
    {
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, w, 1, 8, 
					      s->dpi, s->dpi, 
					      cFrameBits, &(s->prLastRow)));
	ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, w, 1, 16, 
					      s->dpi, s->dpi, cFrameBits, 
					      &(s->prLastDistances)));
    }

    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, h - 1, s->prThresholds,
				  PIX_SRC,
				  0, 0, w, 1, s->prLastRow));
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, h - 1, s->prDistances,
				  PIX_SRC,
				  0, 0, w, 1, s->prLastDistances));

    for (rectP = rectList; rectP; rectP = rectP->next)
    {
	if (rectP->mass == 0)
	    continue;
	/*  
	 * Paint region which you saved before propagating because these
	 * thresholds were too light.
	 * Potential Bug: if these rectangles overlap good thresholds,
	 * then we'll overwrite them here.
	 */

	/* Scale the rectangle up to the size of the threshold map and
	 * dilate it a bit.
	 */
	rect.x = rectP->x * 2 - 1;
	rect.y = rectP->y * 2 - 1;
	rect.w = rectP->w * 2 + 1;
	rect.h = rectP->h * 2 + 1;
	if( rect.x < 0) rect.x =0;
	if( rect.y < 0) rect.y =0;

	/* Paint that threshold into the threshold map */
	color[0] = color[1] = color[2] = color[3] = (UInt8) rectP->mass;
	ERR_PASS_CLEANUP (w_paint (pIPWState, s->prThresholds, &rect, color));

    }
    ERR_PASS_CLEANUP (w_destroyRectList (pIPWState, &rectList));

    /* Expand the band, if necessary */
    if (s->expansion)
    {
	if (s->expansion == 1)
	{
	    f1 = 200; f2 = 300;
	}
	else
	{
	    f1 = 1; f2 = s->expansion;
	}

	ERR_PASS_CLEANUP (w_resConv (pIPWState, prBand, f1, f2, &prExpanded));
	ERR_PASS_CLEANUP (w_thresholdByPixel (pIPWState, prExpanded, 
					      s->prThresholds, THRESH_NORM, 
					      &prThresholded));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prExpanded));
    }
    else
    {
	/* Threshold the band */
	ERR_PASS_CLEANUP (w_thresholdByPixel (pIPWState, prBand, 
					      s->prThresholds,
					      THRESH_NORM,
					      &prThresholded));
	if( unscaledDpi > 300 && s->isScaledUp ==  cFalse )
	{
	    PIXR   *prSmooth = NULL;
	    ScaledInt1616  xDPI, yDPI;

	    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prBand, 
						      &xDPI, &yDPI));
	    prTmp = NULL;
	    ERR_PASS_CLEANUP (w_scaleToSize (pIPWState, prMin, 
					 bandW, bandH, &prTmp));
	    ERR_PASS_CLEANUP(w_pixrSetResolutionDPI (pIPWState,
						     prTmp, xDPI, yDPI));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMin));
	    ERR_PASS_CLEANUP (w_invert (pIPWState, prTmp, &prMin));
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prBand, 
				      PIX_SRC|PIX_DONTCLIP,
				      0, -1, bandW, 1, prBand));
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, bandH - 1, prBand, 
				      PIX_SRC|PIX_DONTCLIP,
				      0, bandH, bandW, 1, prBand));
	    ERR_PASS_CLEANUP (w_smooth (pIPWState, prBand, 3, &prSmooth));
	    prMax = NULL;
	    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prSmooth, prTmp, 
					   PIX_SRC | PIX_DST, &prMax));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSmooth));
	    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prBand, prMin, 
				       PIX_SRC | PIX_DST, &prSmooth));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMin));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prMax, OR, prSmooth));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prMax));
	    ERR_PASS_CLEANUP (w_thresholdByPixel (pIPWState, prSmooth, 
					      s->prThresholds, THRESH_NORM, 
					      &prTmp));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prSmooth));
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prTmp, PIX_SRC,
				  0, s->row, s->eWidth, eBandH,
				  s->prSmoothed));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
	}
    }

    /* Copy that thresholded band in */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, prThresholded,
				  PIX_SRC,
				  0, s->row, s->eWidth, eBandH,
				  s->prResult));

    /* Fill the rectangle if needed */
    if (pThreshRect)
    {
	pThreshRect->x = 0;
	pThreshRect->y = s->row;
	pThreshRect->w = s->eWidth;
	pThreshRect->h = eBandH;
    }

    s->row += eBandH;
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prThresholded));


    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    if (prDiff) w_destroyPixr (pIPWState, &prDiff);
    if (prGradThresh) w_destroyPixr (pIPWState, &prGradThresh);
    if (prReduced) w_destroyPixr (pIPWState, &prReduced);
    if (prThresholded) w_destroyPixr (pIPWState, &prThresholded);
    if (prExpanded) w_destroyPixr (pIPWState, &prExpanded);
    if (s->prThresholds) w_destroyPixr (pIPWState, &s->prThresholds);
    if (s->prLastRow) w_destroyPixr (pIPWState, &s->prLastRow);
    if (s->prDistances) w_destroyPixr (pIPWState, &s->prDistances);
    if (s->prLastDistances) w_destroyPixr (pIPWState, &s->prLastDistances);
    if (histTab) IPW_FREE (histTab, cAllocatorData);

    IPW_FREE (s, cAllocatorData);

    ERR_RETURN;
}
/*****************************************************************************
 * BeginDocumentation::: "w_binarize2Finish"
 *
 * OneLineDescription:::
 *    Free everything and return the resulting image.
 *
 * LongDescription:::
 *    We've completed our binarization. Free everything and return the
 *    resulting image.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      IPWrap state structure
 *  ppBinState::     The binarization state strucure, which is freed on exit
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  in development
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_binarize2Finish"></A>
 *****************************************************************************/
ErrCode CDECL
w_binarize2Finish (
    IPWState           *pIPWState,
    void              **ppBinState)
{
    PROLOGUE;
    Bin2State *s = (Bin2State *) *ppBinState;
    Int32      i, f1, f2;

    ERR_GEN (s == NULL, cErrNullPtr); 

    if (s->prThresholds) 
	ERR_PASS (w_destroyPixr (pIPWState, &s->prThresholds));
    if (s->prLastRow) 
	ERR_PASS (w_destroyPixr (pIPWState, &s->prLastRow));
    if (s->prDistances )
        ERR_PASS (w_destroyPixr (pIPWState, &s->prDistances));
    if (s->prLastDistances)
	ERR_PASS (w_destroyPixr (pIPWState, &s->prLastDistances));
    if (s->hTab)
	IPW_FREE (s->hTab, cAllocatorData); 
    if (s->pHist)
    {
	f1 = f2 =0;
	for(i=0; i<256; ++ i)
	    f2 += s->pHist[i];

	f2 /= 2;
	for(i=0; i<256; ++i)
	{
	    f1 += s->pHist[i];
	    if( f1 >= f2) break;
	}
	if( i < cMedianThreshold )
	{
	    ERR_PASS (w_copy (pIPWState, s->prSmoothed, &s->prResult) );
	}
	IPW_FREE (s->pHist, cAllocatorData);
    }

    if (s->prSmoothed) 
	ERR_PASS (w_destroyPixr (pIPWState, &s->prSmoothed));

    IPW_FREE (s, cAllocatorData);
    *ppBinState = NULL;

    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription::: 
 *  Non-banded adaptive binarization using yet another algorithm, 8-bit only
 *
 * LongDescription:::
 *  This algorithm appears to get the same basic text quality as binarize2,
 *  but does better on line art, tints and pictures.  binarize2 can miss thin
 *  line art and insert splotches in tints and pictures..
 * <BR>
 *  At the moment, this algorithm only supports 8 bit/pixel gray images.
 *
 * PrivateDescription:::
 *  binarize2 works pretty well on text in fairly uniform regions.  However,
 *  in region boundaries, it can get confused and propagate an inappropriate
 *  threshold.  It also makes a global decision about the size of the
 *  gradient that will be called significant, which can mean that small
 *  low-contrast regions may be shortchanged.
 *  <BR>
 *  This method does only very local threshold propagation around regions of
 *  significant contrast.  It uses a globally computed threshold to fill in
 *  the remaining areas.  This can lead to spotty dark regions, but is
 *  generally preferable to the large chunks of mis-thresholded background
 *  occasionally produced by binarize2.
 *  <BR>
 *  In a bit more detail, the method is as follows:
 *  <BR>
 *  1. Make four versions of ~50 dpi image:<BR>
 *	a. The first is made by retaining the maximum pixel value
 *	   at every reduction stage (Max-reduction).<BR>
 *	b. The second is made by retaining the minimum pixel value
 *	   at every reduction stage (Min-reduction).<BR>
 *	c. The third is made using a smoothing reduction (from resolution
 *	   conversion) step for the first 2x reduction and
 *	   Max-reduction thereafter.
 *	d. The last also uses a smoothing reduction for the first 2x
 *	   stage and Min-reduction thereafter.
 *	The difference between the first and third images is an indication
 *	of the amount of noise in the background, which we assume for
 *	the purposes of discussion is always lighter than the foreground.
 *	Similarly, the difference between the second and fourth images
 *	indicates the amount of noise in the foreground.  We try to move
 *	the local threshold down to get under background noise and up to
 *	rise above foreground noise.
 *  2. Use grayscale morphology to compute the real local maxima for
 *     the first (max), second (min) and third (smoothed max) images.
 *     We do two operations on each image, one using strel5h and the
 *     other using strel5v.  The grayscale morphology ops like linear
 *     structure elements.<BR>
 *  3. Subtract image 4 (smoothed min) from the dilated version of
 *     image 3 (smoothed max).  Threshold the difference using an argument
 *     that was passed in as the threshold value.  The resulting mask image
 *     indicates regions where there are significant gradients
 *     (like text, line art and edges of big regions).<BR>
 *  4. Use grayscale erosion on the fourth image using strel5h and strel5v.
 *     This spreads the minima out around these regions.<BR>
 *  5. Dilate the mask image using strel3.  Note that we dilate the
 *     gradient mask by less than we dilate the minima and maxima.  This
 *     guarantees that only the real minima and maxima fit under the mask.<BR>
 *  5. Compute a threshold image (an image that holds local threshold values)
 *     that is equal to: <BR>
 *     (Max+Min)/2 - (Max - SmoothedMax) + (SmoothedMin - Min).<BR>
 *     This moves the ideal threshold ((Max+Min)/2) down to compensate for
 *     background noise (Max - SmoothedMax) and up to account for foreground
 *     noise (SmoothedMin - Min).  However, this threshold image is only
 *     valid in the areas under the gradient mask.
 *     <BR>
 *  6. Run w_ccTextThreshold on the pure max and pure min images.  Compute
 *     a new threshold image that is filled with 3/4 * MIN(the two thresholds
 *     just computed).<BR>
 *  7. Merge the two threshold images using the mask image.  Smooth the
 *     resulting threshold image to minimize "ransom note" effect of locally
 *     changing thresholds.<BR>
 *  8. Use that threshold map to binarize the image.
 *
 * Input:::
 *  pIPWState::      State structure
 *  prSrc::          Image to threshold
 *  gradThresh::     Gradient threshold to use. Suggest 50.
 *
 * Output:::
 *  pPrThresholded::  Pointer to pointer to binarized image.  On input,
 *		      this must be a pointer to a NULL pointer.  On
 *		      successful return, this will be pointer to a
 *		      pointer to a valid PIXR.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  adaptive thresholding
 *
 * Keywords:::
 *  adaptive thresholding
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_binarize3 (
    IPWState  *pIPWState,
    PIXR      *prSrc,
    Int32      gradThresh,
    PIXR     **pPrThresholded)
{
    PROLOGUE;
    PIXR          *prRedMin = NULL, *prRedMax = NULL, *prRedDiff = NULL, 
	          *prRedAvg = NULL, *prRedMaxSm = NULL, *prRedMinSm = NULL, 
	          *prExp = NULL, *prRedMask = NULL, *prPrimary = NULL;
    Int32          dpi;
    Int32	   maxThresh, minThresh, scaleFactor;
    STREL	  *pStrel3 = NULL, *pStrel5h = NULL, *pStrel5v = NULL;
    ColorRecord    color; 
    ScaledInt1616  xDPI, yDPI;

    /* Check stuff */
    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth8, cChannels1|cChannels3, 
		  cFrame32);
    ERR_GEN (pPrThresholded == NULL, cErrNullPtr);
    ERR_GEN (*pPrThresholded != NULL, cErrPtrNotNull);

    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5h));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5h, &pStrel5h));
    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel5v));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel5v, &pStrel5v));
    ERR_PASS_CLEANUP (w_initStrelByName(pIPWState, strel3));
    ERR_PASS_CLEANUP (w_getStrel(pIPWState, strel3, &pStrel3));

    /* Get the "primary" channel so we don't operate on RGB for no reason */
    ERR_PASS (w_pixrGetPrimaryChannel (pIPWState, prSrc, &prPrimary));

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prPrimary, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

	/* Produce the right reductions.  This is a very important part.
	 * We want to preserve indicators of the amount of foreground and
	 * background noise in the original image.  We do this by
	 * comparing a version of the image that's been purely min (or max)
	 * reduced to a version that uses resConv as the first reduction
	 * step.
	 */
    if (dpi > 250)
    {
	    /* Make pure maximum reduced image */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prPrimary, 3,
					  cTrue, cTrue, cTrue, cTrue,
					  &prRedMax));
	    /* Make pure minimum reduced image */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prPrimary, 3,
					  cFalse, cFalse, cFalse, cFalse,
					  &prRedMin));
	    /* Do first step smoothing reduction */
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prPrimary, 200, 100, &prExp));
	    /* Make smoothed max reduced image */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prExp, 2,
					  cTrue, cTrue, cTrue, cTrue,
					  &prRedMaxSm));
	    /* Make smoothed min reduced image */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prExp, 2,
					  cFalse, cFalse, cFalse, cFalse,
					  &prRedMinSm));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prExp));
	scaleFactor = 8;
    }
    else
    {
	    /* this is basically the same as the stuff above with one
	     * fewer reduction stage.
	     */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prPrimary, 2,
					  cTrue, cTrue, cTrue, cTrue,
					  &prRedMax));
	    /* Make pure minimum reduced image */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prPrimary, 2,
					  cFalse, cFalse, cFalse, cFalse,
					  &prRedMin));
	    /* Do first step smoothing reduction */
	ERR_PASS_CLEANUP (w_resConv (pIPWState, prPrimary, 200, 100, &prExp));
	    /* Make smoothed max reduced image */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prExp, 1,
					  cTrue, cTrue, cTrue, cTrue,
					  &prRedMaxSm));
	    /* Make smoothed min reduced image */
	ERR_PASS_CLEANUP (w_reduceMinMax (pIPWState, prExp, 1,
					  cFalse, cFalse, cFalse, cFalse,
					  &prRedMinSm));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prExp));
	scaleFactor = 4;
    }

	/* Improve the pure max, pure min and smoothed max images using
	 * grayscale erosion (makes min) or dilation (makes max) as
	 * appropriate.
	 */
    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, prRedMax, pStrel5h, &prRedMax));
    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, prRedMax, pStrel5v, &prRedMax));

	/* Make a preliminary guess about the threshold that we'll use
	 * to fill in the low-gradient parts of the image.  We get one
	 * guess from the pure-max reduction and one from the pure-min
	 * reduction.
	 */
    ERR_PASS_CLEANUP (w_ccTextThreshold (pIPWState, prRedMax, 10, cTrue,
					 &maxThresh));

	/* Improve the pure min image and guess a threshold from it. */
    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, prRedMin, pStrel5h, &prRedMin));
    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, prRedMin, pStrel5v, &prRedMin));

    ERR_PASS_CLEANUP (w_ccTextThreshold (pIPWState, prRedMin, 10, cTrue,
					 &minThresh));

	/* We really don't know which of these will be bigger, it varies
	 * by image.  We choose the minimum of both, then scale it by
	 * 3/4 to make things a bit lighter.
	 */
    minThresh = 3 * MIN(minThresh, maxThresh) / 4;

    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, prRedMaxSm, pStrel5h,
				    &prRedMaxSm));
    ERR_PASS_CLEANUP (w_dilateGray (pIPWState, prRedMaxSm, pStrel5v,
				    &prRedMaxSm));

    /*
     * Compute a preliminary gradient by taking the difference between
     * the dilated, smoothed max-reduction and the plain smoothed
     * min-reduction.  We dilate the smoothed max reduction first
     * so we get a nice ceiling over all text.  We do NOT dilate the
     * the smoothed min first so that we get a more accurate mask image
     * right where there's high-gradient stuff.
     */
    ERR_PASS_CLEANUP (w_subtractGrayNoClip (pIPWState, prRedMaxSm, prRedMinSm,
					    &prRedDiff));

    ERR_PASS_CLEANUP (w_threshold (pIPWState, prRedDiff, 0, gradThresh,
				   THRESH_INVERSE, &prRedMask));

	/* close tiny holes in the mask, then remove isolated dots in
	 * it.  This generally reduces the amount of spurious noise found.
	 */
    ERR_PASS_CLEANUP (w_close (pIPWState, prRedMask, pStrel3, &prRedMask));
    ERR_PASS_CLEANUP (w_cleanPepper3 (pIPWState, prRedMask, 0, &prRedMask));

	/* Dilate the mask to cover the rest of the high gradient areas. */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prRedMask, pStrel3, &prRedMask));

	/* Use gray erosion to extend the smoothed minima out beyond the
	 * mask.
	 */
    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, prRedMinSm, pStrel5h,
				   &prRedMinSm));
    ERR_PASS_CLEANUP (w_erodeGray (pIPWState, prRedMinSm, pStrel5v,
				   &prRedMinSm));

	/* compute the thresholding image used for the high-gradient areas.
	 * Start with (pureMax + pureMin)/2.
	 */
    ERR_PASS_CLEANUP (w_averageGray (pIPWState, prRedMin, prRedMax, &prRedAvg));

	/* Estimate size of foreground noise */
    ERR_PASS_CLEANUP (w_subtractGrayNoClip (pIPWState, prRedMax, prRedMaxSm,
					    &prRedDiff));
	/* Adjust threshold down by this amount */
    ERR_PASS_CLEANUP (w_subtractGrayNoClip (pIPWState, prRedAvg, prRedDiff,
					    &prRedAvg));

	/* Estimate size of background noise */
    ERR_PASS_CLEANUP (w_subtractGrayNoClip (pIPWState, prRedMinSm, prRedMin,
					    &prRedDiff));

	/* Adjust threshold up by this amount */
    ERR_PASS_CLEANUP (w_addGrayNoClip (pIPWState, prRedAvg, prRedDiff,
				       &prRedAvg));

	/* We're done with a lot of images */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedMax));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedMaxSm));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedMin));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedMinSm));

	/* Now we want to fill in the low-gradient portions of the
	 * low-gradient image.  Create an image that contains the
	 * low-gradient threshold in every pixel.
	 */
    color[0] = color[1] = color[2] = color[3] = (UInt8) minThresh;
    ERR_PASS_CLEANUP (w_paint (pIPWState, prRedDiff, NULL, color));

	/* Paint the portions of this image that correspond to high-gradient
	 * areas to 0.  opGrayBin always paints black stuff to 0.
	 */
    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prRedDiff, prRedMask,
				   (PIX_SRC | PIX_DST), &prRedDiff));

	/* Set the low-gradient portions of the high-gradient threshold
	 * image to 0 by inverting the mask and painting again.
	 */
    ERR_PASS_CLEANUP (w_invert (pIPWState, prRedMask, &prRedMask));
    ERR_PASS_CLEANUP (w_opGrayBin (pIPWState, prRedAvg, prRedMask,
				   (PIX_SRC | PIX_DST), &prRedAvg));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedMask));

	/* Combine the low-gradient thresholds with the high-gradient
	 * thresholds.
	 */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prRedDiff, UNION, prRedAvg));

	/* For the sake of uniformity, smooth out the thresholds.
	 * Note that one PIXR can't be both the source and dest for
	 * this op.
	 */
    ERR_PASS_CLEANUP (w_smooth (pIPWState, prRedAvg, 3, &prRedDiff));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedAvg));

	/* Expand the threshold image back up to the size of the input. */
    ERR_PASS_CLEANUP (w_expand (pIPWState, prRedDiff, scaleFactor, &prExp));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prRedDiff));

	/* Use this to threshold the original image */
    ERR_PASS_CLEANUP (w_thresholdByPixel (pIPWState, prPrimary, prExp,
					  THRESH_NORM, pPrThresholded));

    /* That's that */
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prExp));

    NORMAL_RETURN;

ERR_CLEANUP:
    /* Free all of the pixrs that might be lying around */
    if (prRedMin) w_destroyPixr (pIPWState, &prRedMin);
    if (prRedMax) w_destroyPixr (pIPWState, &prRedMax);
    if (prRedMinSm) w_destroyPixr (pIPWState, &prRedMinSm);
    if (prRedMaxSm) w_destroyPixr (pIPWState, &prRedMaxSm);
    if (prRedDiff) w_destroyPixr (pIPWState, &prRedDiff);
    if (prRedAvg) w_destroyPixr (pIPWState, &prRedAvg);
    if (prRedMask) w_destroyPixr (pIPWState, &prRedMask);
    if (prExp) w_destroyPixr (pIPWState, &prExp);
    ERR_RETURN;
}


/*
 * $Log: binarize.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.66  1998/07/21 22:42:12  xzhu
 * xzhu on Tue Jul 21 15:41:47 PDT 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Increase maxV in strip case.
 *
 * Revision 1.65  1998/06/25 00:09:30  xzhu
 * xzhu on Wed Jun 24 17:09:10 PDT 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Using distance to control propgation value.
 *
 * Revision 1.64  1998/06/11 16:59:42  xzhu
 * xzhu on Thu Jun 11 09:59:07 PDT 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed compiler warning.
 *
 * Revision 1.63  1998/06/10 20:07:29  xzhu
 * xzhu on Wed Jun 10 13:06:00 PDT 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Use new thresholds.
 *
 * Revision 1.62  1998/04/24 17:48:10  xzhu
 * xzhu on Fri Apr 24 10:44:14 PDT 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Check whether the band result will be outside result image buffer to avoid crashing. If input last band image is partial image, the checking will help to avoid crashing.
 *
 * Revision 1.61  1998/04/16 22:32:47  xzhu
 * xzhu on Thu Apr 16 15:28:51 PDT 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Improved binarized image quality for OCR on 400dpi images. After testing the images scanned by Epson, Umax, Hp scanner, the OCR results of images coming from Hp scanner are improved. Keep no change for others.
 *
 * Revision 1.60  1998/03/28 00:19:12  xzhu
 * xzhu on Fri Mar 27 16:17:17 PST 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * If use CreatePixeNoInit for threshold map image and there is no edge to pain in this image, the threshold map will be radom value. So we changed CreatePixr with init. back.
 *
 * Revision 1.59  1998/02/25 03:43:55  ddavies
 * ddavies on Tue Feb 24 19:42:21 PST 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Convert binarize2 and binarize2Band to use the new gray propagation function.
 * Also add a bunch of comments.  Don't create and destroy the threshold and
 * distance images in binarize2Band on each band if the current band requires
 * the same size images as the previous band.
 *
 * Revision 1.58  1998/01/15 18:05:43  xzhu
 * xzhu on Thu Jan 15 10:05:30 PST 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Made compiler happy.
 *
 * Revision 1.57  1998/01/15 01:28:26  xzhu
 * xzhu on Wed Jan 14 17:25:31 PST 1998
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Band height for binarizeBand can be around cBandHeight. Otherwise this condition will cause autoscan crash. Created a new function to get current binary image height during binarizeBand for aligment purpose.
 *
 * Revision 1.56  1998/01/14 00:10:43  ddavies
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.55  1997/10/16 17:24:29  xzhu
 * xzhu on Thu Oct 16 10:24:02 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Deleted extra StartTime().
 *
 * Revision 1.54  1997/07/31 00:54:42  ddavies
 * ddavies on Wed Jul 30 17:53:14 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Add implementation of new adaptive thresholding heuristic, w_binarize3.
 * This one is currently slower than w_binarize2, but seems to get generally
 * better, and more predictable, results in regions containing tints, pictures
 * and weird backgrounds.
 * A banded version will be coming soon, I hope.
 *
 * Revision 1.53  1997/07/30 19:06:51  xzhu
 * xzhu on Wed Jul 30 12:04:38 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed bug in binarize2 where the wrong rectangle was sent to w_pain.
 *
 * Revision 1.52  1997/07/28 18:19:36  xzhu
 * xzhu on Mon Jul 28 11:11:33 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Paint thresholding value two times, one is before propagating and another one is after propagating.
 *
 * Revision 1.51  1997/07/21 18:01:43  xzhu
 * xzhu on Mon Jul 21 10:59:33 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Reverse diff image first, then find pre-thresholds. It improves final results.
 *
 * Revision 1.50  1997/07/01 00:43:14  xzhu
 * xzhu on Mon Jun 30 17:39:31 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fix unefficient function call found by Dan. Use w_threshold( THRESH_INVERSE ) to replace w_threshold( THRESH_NORM ) and w_invert.
 *
 * Revision 1.49  1997/06/26 20:00:20  ddavies
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Add inclusion of new file rast.pub.
 *
 * Revision 1.48  1997/06/17 23:07:17  xzhu
 * xzhu on Tue Jun 17 16:01:23 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * When the rectangle passed into w_paint is outside the image, w_paint will not do correct jod. So we check the rectangle before call w_paint.
 *
 * Revision 1.47  1997/06/12 22:49:07  xzhu
 * xzhu on Thu Jun 12 15:47:53 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Added PIXR_DONTCLIP in w_rasterop for "binarize2". If no this opration, radom image will appear on left side of image.
 *
 * Revision 1.46  1997/05/15 20:50:49  xzhu
 * xzhu on Thu May 15 13:50:00 PDT 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Added algorithm to find gradient image threshold instead of magic number.
 *
 * Revision 1.45  1997/04/15 20:10:30  ddavies
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Substitute NoInit (uninitialized) version of pixr creation routines where
 * possible.
 *
 * Revision 1.44  1997/02/28 00:56:23  xzhu
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Fixed memory leak.
 *
 * Revision 1.43  1997/01/17 23:27:42  xzhu
 * xzhu on Fri Jan 17 15:27:38 PST 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * In w_binarize2Band, we check whether prReduced is equal to prBand. If prReduced is prBand, we keep it. Otherwise we destroy it. If we do not check it, it will cause crashing.
 *
 * Revision 1.42  1997/01/02 15:50:32  macleod
 * macleod on Thu Jan 2 07:50:26 PST 1997
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Backed out of a "fix" that not only broke the regression test, but
 * had occasionally disastrous results.
 *
 * Revision 1.41  1996/12/30 18:40:20  macleod
 * macleod on Mon Dec 30 10:38:34 PST 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * In w_binarizeBand, there is a "default" threshold to use if no thresholds
 * are found in a band, which is probably the cause of the "black band on
 * top" problem, which I've never actually re-created. This hack just reduces
 * that threshold so that it's more likely that if nothing interesting is
 * seen, the band will come out white. A better solution is to use the
 * "binarize2" algorithm, which gives better results, doesn't have tile
 * artifacts, and is even a little faster.
 *
 * Revision 1.40  1996/10/22 02:08:33  macleod
 * macleod on Mon Oct 21 19:08:00 PDT 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed a bug in banded binarization where sometimes the last strip wasn't
 * getting copied into the result image.
 *
 * Revision 1.39  1996/09/12 16:23:44  macleod
 * macleod on Thu Sep 12 09:23:03 PDT 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed bug: when resolution was <200, the threshold was arbitrarily
 * jacked up (an icky hack) but wasn't clipped, so added some clipping.
 *
 * Revision 1.38  1996/07/22 14:12:12  macleod
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed "nag script" warnings--some missing colons, missing ERR_CLEANUP
 * wrappers, misspelled arguments in comment, etc.
 *
 * Revision 1.37  1996/07/19 15:31:23  macleod
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Added a banded version of the "binarize2" algorithm. Doing it in
 * bands was a bit tricky, because the binarize2 algorithm computed a
 * threshold map based on plopping threshold seeds down and computing the
 * voronoi reigons for those seeds using a geodesic propagation algorithm.
 * The problem is that if one does it in bands, one misses seeds in
 * adjacent bands. The solution was to move to a two-pass sequential
 * propagation algorithm, and pass the last line's results for the downward
 * sweep (values and distances) to the first line of the downward sweep
 * of subsequent propagations. Then, the upward sweep can be completed.
 * This passes seeds downwards, so that in each band, we can compute the
 * complete voronoi map given all seeds above. We never look downwards,
 * however, so it does mean we're doing a bit of an approximation. In
 * practice, I've found that not looking downwards hasn't been too much
 * of a problem, but one could conceive of cases where we'll chop off the
 * top of some characters.
 *
 * Revision 1.36  1996/05/30 22:23:39  macleod
 * macleod on Thu May 30 15:23:27 PDT 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed a rather egregious bug in "w_binarize", in which it was
 * copying a strip out 64 times more often than necessary, and
 * trashing the end of the result to boot. This is why 4-bit autoscan
 * took so long...
 *
 * Revision 1.35  1996/03/21 17:08:41  macleod
 * macleod on Thu Mar 21 09:08:05 PST 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * An embarassing oversight--used MALLOC instead of IPW_MALLOC (twice!)
 *
 * Revision 1.34  1996/03/20 22:06:09  macleod
 * macleod on Wed Mar 20 14:05:51 PST 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * (should) fix warning on NT/Watcom
 *
 * Revision 1.33  1996/03/19 22:02:35  macleod
 * macleod on Tue Mar 19 14:01:48 PST 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Don't let ipshared allocate the resolution conversion tables--do it
 * here, since we free them here.
 *
 * Revision 1.32  1996/03/13 01:00:32  macleod
 * macleod on Tue Mar 12 17:00:11 PST 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * w_binarize() was ignoring parameters that it should have passed on
 * to w_binarizeInit().
 *
 * Revision 1.31  1996/03/07 23:21:32  macleod
 * macleod on Thu Mar 7 15:21:18 PST 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Added w_binarize(), which is a non-banded wrapper on top of banded
 * binarization.
 *
 * Revision 1.30  1996/01/08 16:03:28  macleod
 * macleod on Mon Jan 8 08:02:35 PST 1996
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Bug fix: if a color image was passed to w_binarize2, each plane would
 * get binarized, thus taking 200% longer.
 *
 * Revision 1.29  1995/12/07  17:21:39  macleod
 * macleod on Thu Dec 7 09:21:33 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Urg. Fix NT warning.
 *
 * Revision 1.28  1995/12/07  15:52:16  macleod
 * macleod on Thu Dec 7 07:51:58 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed memory leak found by purify, and hopefully fixed the last
 * "nag" script warning.
 *
 * Revision 1.27  1995/12/07  05:45:49  macleod
 * macleod on Wed Dec 6 21:45:26 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Added expansion factor to w_binarize2(); fixed a bug where the
 * gradient threshold passed was ignored.
 *
 * Revision 1.26  1995/12/04  16:11:01  macleod
 * macleod on Mon Dec 4 08:10:27 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed references to other routines in comment header, and
 * fixed EndDocumentationTag for w_binarize2, which had the wrong routine
 * name.
 *
 * Revision 1.25  1995/12/03  00:49:05  macleod
 * macleod on Sat Dec 2 16:47:42 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Added w_binarize2(), which is yet another adaptive thresholding
 * algorithm. This algorithm tries to find "likely" places to threshold,
 * and propagates the thresholds geodesically so that places which don't
 * have a "likely" threshold obtain one from nearby.
 * This only works on 8 bit/pixel images at the moment, and is not
 * banded.
 *
 * Revision 1.24  1995/11/13  20:09:23  ejaquith
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.23  1995/11/06  23:07:33  macleod
 * macleod on Mon Nov 6 15:07:02 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed a couple of bugs in the dithering-while-thresholding option
 * which were introduced during code reorg
 *
 * Revision 1.22  1995/11/04  01:35:44  ejaquith
 * ejaquith on Fri Nov 3 17:35:43 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Fixed auto-doc related warnings like boo-boos, typos, missing ERR_PASS
 * style wrappers, etc..
 *
 * Revision 1.21  1995/11/01  20:28:21  macleod
 * macleod on Wed Nov 1 12:27:53 PST 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Fixed a couple of vile bugs introduced in the code migration; they weren't
 * found because this was not included in the regression test.
 *
 * Revision 1.20  1995/10/13  21:58:28  ejaquith
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.19  1995/10/03  01:59:35  macleod
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.18  1995/09/29  01:52:24  ejaquith
 * ejaquith on Thu Sep 28 18:52:11 PDT 1995
 *
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * Updated w_resConv() invocations.
 *
 * Revision 1.17  1995/09/21  22:51:20  ejaquith
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.16  1995/09/19  19:19:42  macleod
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.15  1995/09/11  21:54:41  macleod
 * binarize.c -> /product/ipcore/ipwrap/src/RCS/binarize.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 * 
 */
    
/*</PRE></BODY>*/
