/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  bmp.c:    routines for reading BMP files into PIXRs and for
 *	      writing PIXRs to BMP files.
 *
 *  Note:  all routines are private (thus, pointers and PIXRs are not checked
 *  at this level)
 */


#include <stdio.h>

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "frames.pub"
#include "alplib.h"
#include "imageref.pub"
#define BMP_VERSION	/* This needed for some grossness in the struct
			 * definitions and some overlapping types.
			 * Basically, ptypes.h and bmp.h were similar
			 * except for 2 fields in one struct, so we
			 * toggle that here.  (Bleah!)  -ejaquith 10nov95
			 */
#include "ptypes.h"
#include "gdefs.h"
#include "pcx.h"

/* prototypes */
#include "io.pub"
#include "io.prv"
#include "props.pub"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: bmp.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


#define BMFHBYTES     14
#define BMP_MAGIC     0x4d42		/* ('M' << 8) | 'B' */



RGBQUAD	     bwmap[2] = { {0,0,0,0}, {255,255,255,0}};



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a BMP file into a PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream.
 *  FileH::          Ptr to file header struct
 *
 * Output:::
 *  pixr::           Ptr to ptr to returned PIXR with image that was read.
 *                   pixr is allocated and filled in.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *  ejaquith::  added RGB capability
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_readBMP (
    IPWState           *pIPWState,
    FILE               *fp,
    FH                 *FileH,
    PIXR              **pixr)
{
    PROLOGUE;
    Int32            width, height, resH, resV;
    UInt16           depth;
    Int32            bytes;
    BITMAPFILEHEADER bmfh;
    BITMAPINFOHEADER bmih;
    Int32            offset;
    Int32            fileBpl;
    Int32            pixrBpl;
    Int32            row;
    UInt8           *imgptr;
    PIXR            *pr = NULL;
    UInt8           *pLineBuf = NULL;

    ERR_GEN(fp == NULL, cErrNullPtr);

    ERR_GEN(fread( &bmfh, BMFHBYTES, 1, fp) != 1, cErrIO_ReadError);

    ERR_GEN(fread( &bmih, sizeof( BITMAPINFOHEADER ), 1, fp) != 1,
	    cErrIO_ReadError);

    /* can process only uncompressed BMP files */
    ERR_PASS( w_getIntel32( pIPWState, (UInt8 *)&bmih.biCompression, &bytes));
    ERR_GEN(bytes != 0, cErrBadType);

    ERR_PASS( w_getIntel32(pIPWState,  (UInt8 *)&bmfh.bfOffBits, &offset));
    ERR_PASS( w_getIntel32(pIPWState,  (UInt8 *)&bmih.biWidth, &width));
    ERR_PASS( w_getIntel32(pIPWState,  (UInt8 *)&bmih.biHeight, &height));
    ERR_PASS( w_getIntel16(pIPWState,  (UInt8 *)&bmih.biBitCount, &depth));
    ERR_PASS( w_getIntel32(pIPWState,  (UInt8 *)&bmih.biXPelsPerMeter, &resH));
    ERR_PASS( w_getIntel32(pIPWState,  (UInt8 *)&bmih.biYPelsPerMeter, &resV));
    ERR_PASS( w_getIntel32(pIPWState,  (UInt8 *)&bmih.biSizeImage, &bytes));

    if (bytes == 0 ) 
    {
	ERR_PASS( w_getIntel32(pIPWState,  (UInt8 *)&bmfh.bfSize, &bytes));
	bytes -= offset;
    }

    /* Oh, hey, here's a new development.
     * If the height is negative, then the height is the ABS val of it,
     * and the image bitmap is NOT upside-down like it normally is.
     * That is, it's vertically flipped from the default.
     */

    /* Since PIXR's don't have color maps (yet), ignore the color
     * map in the bmp.
     * (If image is of depth 24, then it has no colormap in BMP. */

    if (FileH != NULL)
    {
        FileH->width 	     = width;
        FileH->height 	     = ABS(height);
        FileH->Bpl = FileH->Wpl = 0;  /* variable; don't report */
        FileH->totalBytes    = bytes;
    }

    /* seek to the start of the bitmap in the file */
    fseek(fp, offset, 0);

    fileBpl = ((UInt32)(width * depth + 31) >> 5) << 2;

    /* The BMP spec. only allows for these "depth"s! */
    ERR_GEN_CLEANUP(depth != 1  &&  depth != 4  &&  depth != 8  &&  
		    depth != 24, cErrTypeNotSupported);


    if (depth != 24)
    {
	/* In setting resolution of the pixr,  we multiply pels/meter by
	 * 0.01 meters/cm * 2.54 cm/inch to get pels/inch.
	 */
	ERR_PASS( w_createPixrNoInit( pIPWState, width, ABS(height), depth,
				      Int32_TO_ScaledInt1616((Int32)
						       (resH * 0.0254 + 0.5)),
				      Int32_TO_ScaledInt1616((Int32)
						       (resV * 0.0254 + 0.5)),
				      cFrameBits, &pr));

	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pr, &pixrBpl));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, pr, &imgptr));

	if (height >= 0)
	{
	    /* Read the LAST row of the PIXR in first */
	    imgptr += pixrBpl * (height-1); 
	    pixrBpl *= -1; /* We'll then move up a row in the PIXR */
	}
	else
	{
	    /* bitmap was saved right-side up... */
	    height *= -1;
	}

	for ( row = 0; row < height; row++, imgptr += pixrBpl )
	    ERR_GEN_CLEANUP(fread( imgptr, 1, fileBpl, fp ) != fileBpl,
			    cErrIO_ReadError);


	ERR_PASS_CLEANUP( w_flip( pIPWState, pr ));
    
	if ( depth == 1 )
	    ERR_PASS_CLEANUP( w_invert( pIPWState, pr, &pr ));
    }
    else {
	/* 3 channels * 8 depth per plane == "24" */
	PIXR    *subR, *subG, *subB;
	UInt8   *pR, *pG, *pB;
	UInt8   *pCol;
	Int32    x;
	

	/* In setting resolution of the pixr,  we multiply pels/meter by
	 * 0.01 meters/cm * 2.54 cm/inch to get pels/inch.
	 */
	ERR_PASS( w_createPixrExtendedNoInit( pIPWState, width, ABS(height), 8,
					      Int32_TO_ScaledInt1616(
							(Int32) (resH * 0.0254
								+ 0.5)),
					      Int32_TO_ScaledInt1616(
							(Int32) (resV * 0.0254 
								+ 0.5)),
					      cFrameBits,
					      3, cColorRGB, NULL,
					      &pr));

	pLineBuf = (UInt8 *)IPW_CALLOC(fileBpl, sizeof(UInt8),
				       cAllocatorData);
	ERR_GEN(pLineBuf == NULL, cErrNomem);

	ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pr, &pixrBpl));

	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pr, cRedPlane,   &subR));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pr, cGreenPlane, &subG));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pr, cBluePlane,  &subB));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subR, &pR));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subG, &pG));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subB, &pB));

	if (height >= 0)
	{
	    /* Go to the first pixel in the last row */
	    pR += pixrBpl * (height-1);
	    pG += pixrBpl * (height-1);
	    pB += pixrBpl * (height-1);
	    pixrBpl *= -1; /* We'll then move up a row in the PIXR */
	}
	else 
	{
	    /* bitmap was saved right-side up... */
	    height *= -1;
	}

	/* NOTE:  we are assuming the planes are each stored one pixel
	 * per byte in the PIXR.
	 */
	
	for ( row = 0; row < height; row++)
	{
	    /* Line buffer is stored in consecutive bytes of R,G,B */
	    ERR_GEN_CLEANUP(fread (pLineBuf, 1, fileBpl, fp)
			    != fileBpl, cErrIO_ReadError);
	    
	    pCol = pLineBuf;
	    for (x = 0; x < width; x++) {
		UCHAR_ACCESS(&pR[x]) = *(pCol++);
		UCHAR_ACCESS(&pG[x]) = *(pCol++);
		UCHAR_ACCESS(&pB[x]) = *(pCol++);
	    }
	    
	    /* Move up/down to the next row to fill in in the PIXR */
	    pR += pixrBpl;
	    pG += pixrBpl;
	    pB += pixrBpl;
	}

	IPW_FREE(pLineBuf, cAllocatorData);
    }
    

    *pixr = pr;
    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pLineBuf, cAllocatorData);
    w_destroyPixr(pIPWState, &pr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to file stream in BMP format.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to write to.
 *  pixr::           Ptr to PIXR.
 *                   PIXR may be 1-bit binary, 4 or 8 bit gray,
 *                   or 24 bit RGB (3 planes of 8 bits).
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *  ejaquith::  added RGB capability
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_fwriteBMP (
    IPWState           *pIPWState,
    FILE               *fp,
    PIXR               *pixr)
{
    PROLOGUE;
    Int32            imageLen;
    Int32            width, height, depth, channels;
    Int32	     pixrBpl;
    Int32	     fileBpl;
    Int32	     row;
    UInt8	    *imgptr, *pLineBuffer = NULL;
    UInt8	    *cmapaddr = NULL;	/* address of the color map array */
    int  	     cmaplen;	/* number of bytes in the colormap */
    int		     i, colors, val, stepsize;
    BITMAPFILEHEADER bmfh;
    BITMAPINFOHEADER bmih;
    RGBQUAD	    *rgbp;
    ScaledInt1616    xdpi, ydpi;

    ERR_GEN(fp == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &width));
    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &height));
    ERR_PASS( w_pixrGetDepth(pIPWState,  pixr, &depth));
    ERR_PASS( w_pixrGetChannelCount(pIPWState,  pixr, &channels));
    ERR_PASS( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));

    ERR_GEN(depth != 1  &&  depth != 4  &&  depth != 8,
	    cErrDepthNotSupported);

    ERR_GEN((channels == 3  &&  depth != 8) ||
	    (channels != 1  &&  channels != 3), cErrChannelsNotSupported);

    /* Okay, why are we doing this?  Well, the BMP standard allows
     * for 1-bit binary, 4 and 8-bit gray, and 24-bit RGB (and that's all!)
     */

    depth *= channels;

    /* Number of bytes that will be needed to store one image row in file */
    fileBpl = ((width*depth + 31) >> 5) << 2;

    /* Number of bytes to store entire bitmap in single plane
     * (interlaced if RGB)
     */
    imageLen = fileBpl * height;

	/* compose a color map for the bmp file */
    if ( depth == 1 )
    {
	cmaplen  = sizeof( bwmap );
	cmapaddr = (UInt8 *)bwmap;
    }
    else if (depth == 4  ||  depth ==  8)
    {
	colors  = 1 << depth;
	cmaplen = colors * sizeof( RGBQUAD );

	cmapaddr = (UInt8 *)IPW_CALLOC( cmaplen, 1, cAllocatorData);
	ERR_GEN(cmapaddr == NULL, cErrNomem);

	for ( i = 0, val = 0, stepsize = 255/(colors-1),
			      rgbp = (RGBQUAD *)cmapaddr;
	      i < colors;
	      ++i, val += stepsize, ++rgbp )
	{
	    rgbp->rgbBlue = rgbp->rgbGreen = rgbp->rgbRed = val;
	}
    }
    else {
	/* depth == 24 ==> uses no colormap! */
	cmaplen = 0;
    }

    
    /* The BMP file format is as follows:
     * the BITMAPFILEHEADER is written first.
     * then the BITMAPINFOHEADER is written next.
     * A colormap follows.  (for 24-bit, no colormap is written)
     * The image bits are last.
     */

    fseek( fp, 0L, 0 );

    ERR_PASS_CLEANUP( w_putIntel16( pIPWState, BMP_MAGIC,
				    (UInt8 *)&bmfh.bfType ));
    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, 
				    sizeof( BITMAPFILEHEADER) + cmaplen 
				    + sizeof( BITMAPINFOHEADER) + imageLen,
				    (UInt8 *)&bmfh.bfSize));
    
    bmfh.bfReserved1 = 0;
    bmfh.bfReserved2 = 0;
    /* This is the offset from the beginning of the BITMAPFILEHEADER
     * structure to the start of the actual image bits.
     */
    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, sizeof( BITMAPFILEHEADER) + 
				    sizeof( BITMAPINFOHEADER) + cmaplen,
				    (UInt8 *)&bmfh.bfOffBits));

    /* Write out the BITMAPFILEHEADER */
    ERR_GEN_CLEANUP( fwrite( &bmfh, 1, BMFHBYTES, fp) != BMFHBYTES,
		     cErrIO_WriteError);

    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, sizeof( BITMAPINFOHEADER ),
				    (UInt8 *)&bmih.biSize));
    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, width,
				    (UInt8 *)&bmih.biWidth));
    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, height,
				    (UInt8 *)&bmih.biHeight));

    if (ScaledInt1616_TO_Int32(xdpi) > 0 && 
	ScaledInt1616_TO_Int32(ydpi) > 0)
    {
	/* dpi has been set. */

	/* In setting resolution of the BMP,  we multiply
	 * pels/inch * 1 inch/2.54 cm * 100 cm/meter to get pels/meter.
	 */
	ERR_PASS_CLEANUP( w_putIntel32( pIPWState,
					(Int32) ((ScaledInt1616_TO_Int32(xdpi)
						  * 100 / 2.54) + 0.5),
					(UInt8 *)&bmih.biXPelsPerMeter));
	ERR_PASS_CLEANUP( w_putIntel32( pIPWState, 
					(Int32) ((ScaledInt1616_TO_Int32(ydpi)
						  * 100 / 2.54) + 0.5),
					(UInt8 *)&bmih.biYPelsPerMeter));
    }
    else
    {
	bmih.biXPelsPerMeter = 0;
	bmih.biYPelsPerMeter = 0;
    }
    
    

    ERR_PASS_CLEANUP( w_putIntel16( pIPWState, 1, (UInt8 *)&bmih.biPlanes));
    ERR_PASS_CLEANUP( w_putIntel16( pIPWState, (Int16) depth, 
				    (UInt8 *)&bmih.biBitCount));
    bmih.biCompression   = 0;
    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, imageLen,
				    (UInt8 *)&bmih.biSizeImage));
    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, (depth != 24 ? 1 << depth : 0),
				    (UInt8 *)&bmih.biClrUsed)); 
    ERR_PASS_CLEANUP( w_putIntel32( pIPWState, (depth != 24 ? 1 << depth : 0),
				    (UInt8 *)&bmih.biClrImportant)); 

    /* Write out the BITMAPINFOHEADER */
    ERR_GEN_CLEANUP(fwrite( &bmih, 1, sizeof(bmih), fp) != sizeof( bmih),
		    cErrIO_WriteError);

    /* Write out the colormap */
    if (depth != 24)
	ERR_GEN_CLEANUP(fwrite( cmapaddr, 1, cmaplen, fp) != cmaplen,
			cErrIO_WriteError);

    if ( depth == 4  ||  depth == 8 )
	IPW_FREE(cmapaddr, cAllocatorData);


    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pixr, &pixrBpl));

    if (depth != 24)
    {
	if ( depth == 1 ) 
	    ERR_PASS_CLEANUP( w_invert( pIPWState, pixr, &pixr ));
	ERR_PASS_CLEANUP( w_flip(pIPWState, pixr));

	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, pixr, &imgptr));
	imgptr += pixrBpl * (height - 1);
	for ( row = height - 1; row >= 0; --row, imgptr -= pixrBpl ) 
	    ERR_GEN_CLEANUP(fwrite( imgptr, 1, fileBpl, fp) != fileBpl,
			    cErrIO_WriteError);
	
	ERR_PASS_CLEANUP( w_flip(pIPWState, pixr));
	if ( depth == 1 ) 
	    ERR_PASS_CLEANUP( w_invert( pIPWState, pixr, &pixr ));
    }
    else
    {
	/* Write our 3-plane, 8-bit per plane RGB PIXR into 24-bit BMP */
	UInt8 *pR, *pG, *pB;
	Int32  byteDiff, x;
	PIXR  *subR, *subG, *subB;

	/* Set up the buffer to hold 1 row of the image at a time, that is,
	 * 3 bytes of R,G,B data per pixel in the row */
	pLineBuffer = (UInt8 *)IPW_CALLOC(fileBpl,
					  sizeof(UInt8), cAllocatorData);
	ERR_GEN_CLEANUP(pLineBuffer == NULL, cErrNomem);

	ERR_PASS_CLEANUP( w_flip(pIPWState, pixr));

	/* Take the 3 8-bit planes and pack them together line by line,
	 * with byte values for a pixel going left to right R,G,B */
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr,
					   cRedPlane, &subR));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, 
					   cGreenPlane, &subG));
	ERR_PASS_CLEANUP( w_pixrGetChannel(pIPWState, pixr, 
					   cBluePlane, &subB));
	
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subR, &pR));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subG, &pG));
	ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState, subB, &pB));

	/* Move to the last column of the last image row.
	 * NOTE:  we are assuming the planes are each stored one pixel
	 * per byte.
	 */
	pR += pixrBpl * (height - 1) + width-1;
	pG += pixrBpl * (height - 1) + width-1;
	pB += pixrBpl * (height - 1) + width-1;

	byteDiff = pixrBpl - width;

	/* BMP files are stored with the last image row appearing first
	 * in the file (vert. inversion).  Here's what we'll do.
	 * Since the PIXR image data is planar (in 3) and the BMP data
	 * must be written out interlaced (one plane), we'll write each
	 * image line out by stuffing all the pixels into an interlaced
	 * buffer.  Then we just dump the buffer to file.
	 */
	for ( row = height - 1; row >= 0; --row ) 
	{
	    /* Fill in contiguous byte buffer */
	    for (x = 3*width - 1; x >= 0; ) {
		/* Remember we want it to appear in RGB byter order in
		 * left to right order in the interlaced buffer!
		 */
		pLineBuffer[x--] = *(pB--);
		pLineBuffer[x--] = *(pG--);
		pLineBuffer[x--] = *(pR--);
	    }
	    
	    ERR_GEN_CLEANUP(fwrite (pLineBuffer, 1, fileBpl, fp)
			    != fileBpl, cErrIO_WriteError);

	    /* Move to next rows */
	    pR -= byteDiff;
	    pG -= byteDiff;
	    pB -= byteDiff;
	}
	
	ERR_PASS_CLEANUP( w_flip(pIPWState, pixr));

	IPW_FREE(pLineBuffer, cAllocatorData);
    }
    

    NORMAL_RETURN;

ERR_CLEANUP:
    IPW_FREE(pLineBuffer, cAllocatorData);
    if ( depth != 1)
	IPW_FREE(cmapaddr, cAllocatorData);

    ERR_RETURN;
}

/*

 $Log: bmp.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.33  1997/06/30 23:08:47  ddavies
 bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v

 Delete inclusion of Sandpiper's internal.h.

 Revision 1.32  1997/04/15 20:12:08  ddavies
 bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v

 Substitute NoInit (uninitialized) version of pixr creation routines where
 possible.
 Also, use UCHAR_ACCESS to write bytes into bitmap instead of initializing
 entire bitmap to 0 (to make sure the byte-reversed frame is initialized),
 writing bytes in natural byte order, then byte-reversing the entire image.

 Revision 1.31  1996/01/04 20:21:06  ejaquith
 bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.30  1995/12/13  04:04:54  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.29  1995/12/01  19:32:02  ddavies
 * ddavies on Fri Dec 1 11:32:00 PST 1995
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Fix acc compiler warnings.
 *
 * Revision 1.28  1995/11/13  20:04:04  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Added needed additional inclusions of "ptypes.h", "gdefs.h", and "pcx.h".
 * Updated PROLOGUE usage.
 * Removed inclusion of now obsolete file "bmp.h", after moving 2 needed
 * constants from it into this C file.
 * Defined "BMP_VERSION" before "ptypes.h" so as to get the BMP version of
 * the needed structs.
 *
 * Revision 1.27  1995/10/13  21:58:35  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.26  1995/10/13  02:01:47  ejaquith
 * ejaquith on Thu Oct 12 18:59:55 PDT 1995
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Added capability to read BMP files with negative heights.
 * What this means is that the image is stored upside down from the default,
 * which is upside down.  Thus, negative heights ==> right side up!
 *
 * Revision 1.25  1995/10/12  04:23:46  ddavies
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.24  1995/10/10  02:45:13  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Cleaned up the entire BMP i/o routines.
 * Added file depth and channel checking to only allow valid BMP files to
 * be read/written.
 * Added RGB 24-bit i/o capability.  The BMP spec. allows for the following
 * bit depths: 1,4,8, and 24; a binary, two grays, and an RGB.
 * The RGB files are interlaced bytes of R, G, and B.  They are not supposed
 * to use a colormap, so we skip any that may be there on reading.
 * (and don't write one on writing).
 * Added lots of comments and did more error checking.  Now BMP supports
 * the full spec quite nicely and is proud.
 *
 * Revision 1.23  1995/10/02  19:15:43  jcraig
 * jcraig on Mon Oct 2 12:15:00 PDT 1995
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * w_invert takes a pointer to a pointer to a pixr.
 *
 * Revision 1.22  1995/09/29  00:35:28  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.21  1995/09/21  22:51:27  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.20  1995/09/12  16:55:06  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.19  1995/09/08  00:15:10  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.18  1995/09/07  01:29:34  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.17  1995/09/06  20:17:13  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.16  1995/09/01  02:20:50  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.15  1995/08/31  21:11:02  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.14  1995/08/29  22:21:11  ddavies
 * ddavies on Tue Aug 29 15:15:27 PDT 1995
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Remove DEBUG_ON, was defined in debug.h.
 *
 * Revision 1.13  1995/08/29  21:53:12  ddavies
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.12  1995/08/17  22:01:36  ejaquith
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.11  1995/07/26  22:41:40  ddavies
 * ddavies on Wed Jul 26 15:40:38 PDT 1995
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * cmapaddr points to a global color map when the image is binary.  Only free
 * cmapaddr if the image is NOT binary (freeing globals is bad).
 *
 * Revision 1.10  1995/02/03  01:30:37  ddavies
 * ddavies on Thu Feb 2 17:29:51 PST 1995
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Set x and y resolution of pixr read from PelsPerMeter parms in file header.
 *
 * Revision 1.9  1994/10/10  22:32:46  lperry
 * lperry on Mon Oct 10 15:31:39 PDT 1994
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Change to use new header file io.pub instead of iopros.h
 *
 * Revision 1.8  1994/09/16  17:34:24  jcraig
 * jcraig on Fri Sep 16 10:34:18 PDT 1994
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * remove include of bitacces.h which was not needed.
 *
 * Revision 1.7  1994/08/26  01:33:51  lperry
 * lperry on Thu Aug 25 18:30:41 PDT 1994
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Make the FileH file header struct a parameter to achieve reentrancy.
 * Creates an interface change in readBmpPr, but this is only called
 * from readPr/freadPr.
 *
 * Revision 1.6  1994/01/13  21:52:50  dferrell
 * dferrell on Thu Jan 13 13:51:40 PST 1994
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Removed call to setmode () since we now open the file with the correct
 * mode.  Also reformatted some K&R style code.
 *
 * Revision 1.5  1994/01/12  01:16:09  dferrell
 * dferrell on Tue Jan 11 17:15:42 PST 1994
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * #if 0'ed out calls to setmode for the PC.
 *
 * Revision 1.4  1994/01/07  22:11:33  dferrell
 * dferrell on Fri Jan 7 14:10:54 PST 1994
 *
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Replaced #include <malloc.h> with <stdlib.h> for portability.
 *
 * Revision 1.3  1993/12/02  23:19:28  ddavies
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:09:51  ddavies
 * bmp.c -> /product/ipcore/ipwrap/src/RCS/bmp.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

