/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1996, 1997 Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/*
 *  ccsign.c -- Signatures of color image.
 *
 *           w_colorSignature ()
 */

#include <stdio.h>
#include "ipwcore.prv"

/* DEFINITIONS */
#include "defines.pub"
#include "geomadt.h"
#include "pixr.h"
#include "arithops.h"
#include "thrshold.h"
#include "morphg.h"
#include "logicops.h"
#include "imageref.pub"

/* PROTOTYPES */
#include "shrip.pub"
#include "shrcc.pub"
#include "shrrast.pub"
#include "morph.pub"
#include "thrshold.pub"
#include "props.pub"
#include "strel.pub"
#include "morphg.pub"
#include "frame.pub"
#include "logicops.pub"
#include "prxform.pub"
#include "rast.pub"
#include "rectops.pub"
#include "cut.pub"
#include "color.pub"
#include "grbinop.pub"
#include "frames.pub"
#include "cmask.pub"

#include "imagerep.h"
#include "io.pub"

#include "ccsign.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: ccsign.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

#define  cMaxSign     12

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Using color distribution select majority color as features.
 *
 * LongDescription:::
 *   In color distribution find the colors with large number. Then select
 *   these colors as features. 
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  prS::		Pointer to source image Pixr.
 *
 * Output:::
 *  ppFeature::         Potionter to feattures
 *  pnumSig::           Pointer of number of features
 *
 * Notes:::
 *  Based on Yossi Ruber's color signature define and his reference code,
 *  we developed this algorithm to get color features.
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_colorFeatureOne(
    IPWState           *pIPWState,
    PIXR               *prS,
    COLORSIGNATURE    **ppFeature,
    Int32              *pnumSig)
{
    PROLOGUE;
    ColorType       type;
    PIXR           *prLab = NULL;
    CCCLUSTER      *pCluster = NULL;
    COLORSIGNATURE *pResult = NULL, *pTempRes=NULL;
    Int32           number, idx, bpl, width, height, num, minNum;
    PIXR           *subL, *subA, *subB;
    UInt8          *pl, *pa, *pb;
    Int32           maxNum, i, ll, aa, bb;

    ERR_GEN( *ppFeature != NULL, cErrPtrNotNull );
    ERR_GEN( pnumSig == NULL, cErrNullPtr );

    ERR_PASS( w_pixrGetType (pIPWState, prS, &type)); 
    ERR_GEN( type != cColorRGB, cErrInvalidParm);

    prLab = NULL;
    ERR_PASS_CLEANUP( w_RGB2Lab(pIPWState, prS, &prLab));
 
    ERR_PASS_CLEANUP( w_pixrGetWidth (pIPWState, prLab, &width));
    ERR_PASS_CLEANUP( w_pixrGetHeight (pIPWState, prLab, &height));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prLab, &bpl));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prLab, cRedPlane, &subL));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prLab, cGreenPlane, &subA));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prLab, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subL, &pl));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subA, &pa));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pb));

    pCluster = (CCCLUSTER *) IPW_CALLOC(sizeof(CCCLUSTER), TABLENGTH, 
					cAllocatorData);
    ERR_GEN_CLEANUP( pCluster == NULL, cErrNomem);

    ERR_PASS_CLEANUP( i_ccluster(pIPWState, pl, pa, pb, bpl, 0, 0, 
				 width, height, pCluster, &number));

    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prLab));

    pTempRes = (COLORSIGNATURE *) IPW_CALLOC(sizeof(COLORSIGNATURE), number+1, 
					cAllocatorData);
    ERR_GEN_CLEANUP( pTempRes == NULL, cErrNomem);

    minNum = (width * height) >> 8;
    for(idx=0, num=0; idx< TABLENGTH; ++idx)
    {
	if( (pCluster+idx)->num < minNum) continue;
	(pTempRes + num)->ll = (pCluster+idx)->ll;
	(pTempRes + num)->aa = (pCluster+idx)->aa;
	(pTempRes + num)->bb = (pCluster+idx)->bb;
	(pTempRes + num)->num = (pCluster+idx)->num;
	++ num;
    }
    IPW_FREE(pCluster, cAllocatorData);

    /* Sorting from large to small */
    for(idx=0; idx<num; ++idx)
    {
	maxNum = (pTempRes+idx)->num;
	number = idx;
	for(i=idx+1; i<num; ++i)
	{
	    if((pTempRes+i)->num > maxNum)
	    {
		maxNum = (pTempRes+i)->num;
		number = i;
	    }
	}
	*(pTempRes+num) = *(pTempRes+idx);
	*(pTempRes+idx) = *(pTempRes+number);
	*(pTempRes+number) = *(pTempRes+num);
    }

   
    number = (width * height)* 19/20;
    maxNum = 0;
    for(idx=0; idx<num; ++idx)
	maxNum += (pTempRes+idx)->num;

    while( num > cMaxSign && maxNum > number )
    {
	maxNum -= (pTempRes+num-1)->num;
	--num;
    }

    pResult = NULL;
    pResult = (COLORSIGNATURE *) IPW_CALLOC(sizeof(COLORSIGNATURE), num+2, 
					cAllocatorData);
    ERR_GEN_CLEANUP( pResult == NULL, cErrNomem);

    ll = aa = bb =0;
    for(i=0; i<num; ++i)
    {
	*(pResult+i) = *(pTempRes+i);
	ll += ((pResult+i)->num * (Int32)(pResult+i)->ll);
	aa += ((pResult+i)->num * (Int32)(pResult+i)->aa);
	bb += ((pResult+i)->num * (Int32)(pResult+i)->bb);
    }
    /* Save low boundary which will speed up matching in last one */
    (pResult+num)->ll = (UInt8)( ll/ maxNum);
    (pResult+num)->aa = (UInt8)( aa/ maxNum);
    (pResult+num)->bb = (UInt8)( bb/ maxNum);
    (pResult+num)->num = width*height;

    IPW_FREE(pTempRes, cAllocatorData);

    *ppFeature = pResult;
    *pnumSig = num+1;

    NORMAL_RETURN;

ERR_CLEANUP:

    if(pCluster ) IPW_FREE(pCluster, cAllocatorData); 
    if(pResult ) IPW_FREE(pResult, cAllocatorData); 
    if(pTempRes ) IPW_FREE(pTempRes, cAllocatorData);
    if(prLab ) w_destroyPixr(pIPWState, &prLab);

    ERR_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Free color features.
 *
 * LongDescription:::
 *   Free colors features. 
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  ppFeature::         Potionter to feattures
 *
 * Output:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_freeColorFeatureOne(
    IPWState           *pIPWState,
    COLORSIGNATURE    **ppFeature)
{
    PROLOGUE;

    IPW_FREE(*ppFeature, cAllocatorData);

    NORMAL_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Using color distribution select majority color as features.
 *
 * LongDescription:::
 *   In color distribution find the colors with large number. Then select
 *   these colors as features. 
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *  prS::		Pointer to source image Pixr.
 *  xDim::              Cut image xDim parts in horizontal
 *  yDim::              Cut image yDim parts in vertical 
 *
 * Output:::
 *  ppFeature::         Potionter to feattures
 *  pnumSig::           Pointer of number of features
 *
 * Notes:::
 *  Based on Yossi Ruber's color signature define and his reference code,
 *  we developed this algorithm to get color features.
 *  Feature array will have (xDim * yDim +1) feture strings. 
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_colorFeatureArray(
    IPWState           *pIPWState,
    PIXR               *prS,
    Int32               xDim,
    Int32               yDim,
    COLORSIGNATURE    **ppFeature,
    Int32              *pnumSig)
{
    PROLOGUE;
    ColorType       type;
    PIXR           *prLab = NULL;
    CCCLUSTER      *pCluster = NULL;
    COLORSIGNATURE *pResult = NULL, *pTempRes=NULL;
    Int32           number, idx, bpl, width, height, num, minNum;
    PIXR           *subL, *subA, *subB;
    UInt8          *pl, *pa, *pb;
    Int32           maxNum, i, ll, aa, bb, len;
    Int32           x0, y0, w, h, dimen;

    ERR_GEN( *ppFeature != NULL, cErrPtrNotNull );
    ERR_GEN( pnumSig == NULL, cErrNullPtr );
    ERR_GEN( (xDim <1 || yDim <1 || (xDim < 2 && yDim < 2)), 
	     cErrDestDimensIncorrect);

    ERR_PASS( w_pixrGetType (pIPWState, prS, &type)); 
    ERR_GEN( type != cColorRGB, cErrInvalidParm);

    prLab = NULL;
    ERR_PASS_CLEANUP( w_RGB2Lab(pIPWState, prS, &prLab));

    ERR_PASS_CLEANUP( w_pixrGetWidth (pIPWState, prLab, &width));
    ERR_PASS_CLEANUP( w_pixrGetHeight (pIPWState, prLab, &height));
    ERR_PASS_CLEANUP( w_pixrGetBpl (pIPWState, prLab, &bpl));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prLab, cRedPlane, &subL));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prLab, cGreenPlane, &subA));
    ERR_PASS_CLEANUP( w_pixrGetChannel (pIPWState, prLab, cBluePlane, &subB));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subL, &pl));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subA, &pa));
    ERR_PASS_CLEANUP( w_pixrGetImage (pIPWState, subB, &pb));

    x0 = y0 =0;
    w = width;
    h = height;
    dimen = (xDim * yDim) +1;

    for(len=0; len<dimen; ++len)
    {
	if( len  > 0 )
	{
	    if( len  == 1)
	    {
		w = width /xDim;
		h = height /yDim;
	    }
	    else 
		x0 += w;
	    if( x0 >= w*3 )
	    {
		x0 = 0;
		y0 += h;
	    }
	}
	    
	pCluster = (CCCLUSTER *) IPW_CALLOC(sizeof(CCCLUSTER), TABLENGTH, 
					    cAllocatorData);
	ERR_GEN_CLEANUP( pCluster == NULL, cErrNomem);
	
	ERR_PASS_CLEANUP( i_ccluster(pIPWState, pl, pa, pb, bpl, x0, y0, 
				     w, h, pCluster, &number));
	
	pTempRes = (COLORSIGNATURE *) IPW_CALLOC(sizeof(COLORSIGNATURE), 
						 number+1, cAllocatorData);
	ERR_GEN_CLEANUP( pTempRes == NULL, cErrNomem);
	
	minNum = (width * height) >> 8;
	for(idx=0, num=0; idx< TABLENGTH; ++idx)
	{
	    if( (pCluster+idx)->num < minNum) continue;
	    (pTempRes + num)->ll = (pCluster+idx)->ll;
	    (pTempRes + num)->aa = (pCluster+idx)->aa;
	    (pTempRes + num)->bb = (pCluster+idx)->bb;
	    (pTempRes + num)->num = (pCluster+idx)->num;
	    ++ num;
	}
	IPW_FREE(pCluster, cAllocatorData);

	/* Sorting form large to small */
	for(idx=0; idx<num; ++idx)
	{
	    maxNum = (pTempRes+idx)->num;
	    number = idx;
	    for(i=idx+1; i<num; ++i)
	    {
		if((pTempRes+i)->num > maxNum)
		{
		    maxNum = (pTempRes+i)->num;
		    number = i;
		}
	    }
	    *(pTempRes+num) = *(pTempRes+idx);
	    *(pTempRes+idx) = *(pTempRes+number);
	    *(pTempRes+number) = *(pTempRes+num);
	}

	number = (width * height)*19/20;
	maxNum = 0;
	for(idx=0; idx<num; ++idx)
	    maxNum += (pTempRes+idx)->num;
	
	while( num > cMaxSign && maxNum > number )
	{
	    maxNum -= (pTempRes+num-1)->num;
	    --num;
	}

	pResult = NULL;
	pResult = (COLORSIGNATURE *) IPW_CALLOC(sizeof(COLORSIGNATURE), num+2, 
						cAllocatorData);
	ERR_GEN_CLEANUP( pResult == NULL, cErrNomem);

	ll = aa = bb =0;
	for(i=0; i<num; ++i)
	{
	    *(pResult+i) = *(pTempRes+i);
	    ll += ((pResult+i)->num * (Int32)(pResult+i)->ll);
	    aa += ((pResult+i)->num * (Int32)(pResult+i)->aa);
	    bb += ((pResult+i)->num * (Int32)(pResult+i)->bb);
	}
	/* Save low boundary which will speed up matching in last one */
	(pResult+num)->ll = (UInt8)( ll/ maxNum);
	(pResult+num)->aa = (UInt8)( aa/ maxNum);
	(pResult+num)->bb = (UInt8)( bb/ maxNum);
	(pResult+num)->num = width*height;

	IPW_FREE(pTempRes, cAllocatorData);
	
	*(ppFeature+len) = pResult;
	*(pnumSig+len) = num+1;
    }
    ERR_PASS_CLEANUP( w_destroyPixr(pIPWState, &prLab));

    NORMAL_RETURN;

ERR_CLEANUP:

    if(pCluster ) IPW_FREE(pCluster, cAllocatorData); 
    if(pResult ) IPW_FREE(pResult, cAllocatorData); 
    if(pTempRes ) IPW_FREE(pTempRes, cAllocatorData);
    if(prLab ) w_destroyPixr(pIPWState, &prLab);

    ERR_RETURN;
}

/***************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Free color features one dimension array.
 *
 * LongDescription:::
 *   Free colors features one dimension array. 
 *
 * Input:::
 *  pIPWState::		Pointer to standard IPWrap state block.
 *      dimen::         Length of one dimension array
 *  ppFeature::         Potionter to feattures
 *
 * Output:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  segmentation
 *
 * Keywords:::
 *  segmentation, color
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  X. Zhu
 *
 * EndDocumentation:::
 **************************************************************************/
ErrCode CDECL
w_freeColorFeatureArray(
    IPWState           *pIPWState,
    Int32               dimen,
    COLORSIGNATURE    **ppFeature)
{
    PROLOGUE;
    Int32     i;

    for(i=0; i<dimen; ++i)
	IPW_FREE(*(ppFeature+i), cAllocatorData);

    NORMAL_RETURN;
}

/*
 $Log: ccsign.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:22  jlin
 Imported sources

 Revision 1.3  1998/11/13 19:09:54  xzhu
 xzhu on Fri Nov 13 11:09:39 PST 1998

 ccsign.c -> /product/ipcore/ipwrap/src/RCS/ccsign.c,v

 Added new functions.

 Revision 1.2  1998/10/01 17:20:45  xzhu
 xzhu on Thu Oct 1 10:20:05 PDT 1998

 ccsign.c -> /product/ipcore/ipwrap/src/RCS/ccsign.c,v

 Changed percentage from 50% to 90%.

 Revision 1.1  1998/09/02 17:46:52  xzhu
 Initial revision

*/
