/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  coords1.c:  includes subroutines
 *
 *     Creation/annihilation
 *         COORDS  *createCoords():  creates struct and associated arrays
 *         COORDS  *destroyCoords():  frees arrays and returns NULL ptr
 *
 */



#include "ipwcore.prv"
#include "lists.h"
#include "geomadt.h"
#include "pixr.h"

/* Prototypes */
#include "coords.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: coords1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/****************************************************************
 **************       Creation/annihilation      ****************
 ***************************************************************/
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a COORDS object
 *
 * LongDescription:::
 *  Create a COORDS object, which is an array of co-ordinates.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  numPts::         Number of co-ordinates
 *
 * Output:::
 *  ppCoords::       COORDS structure created
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createCoords (
    IPWState           *pIPWState,
    Int32               numPts,
    COORDS            **ppCoords)
{
    PROLOGUE;
COORDS       *coords = NULL;
Int32        *x = NULL, *y = NULL;

    ERR_GEN (!numPts, cErrNullPtr);

    coords = (COORDS *) IPW_CALLOC(1, sizeof(COORDS), 
				   cAllocatorData);
    ERR_GEN (!coords, cErrNomem);
    x = (Int32 *) IPW_CALLOC(numPts, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (!x, cErrNomem);
    y = (Int32 *) IPW_CALLOC(numPts, sizeof(Int32), cAllocatorData);
    ERR_GEN_CLEANUP (!y, cErrNomem);

    coords->x = x;
    coords->y = y;
    coords->n = numPts;

    INIT_LINKS(coords);
    *ppCoords = coords;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (coords)
	IPW_FREE (coords, cAllocatorData);
    if (x)
	IPW_FREE (coords, cAllocatorData);
    if (y)
	IPW_FREE (coords, cAllocatorData);

    ERR_RETURN;

}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroy a COORDS object
 *
 * LongDescription:::
 *  Destroy a COORDS object, freeing the storage used and clearing the
 *  pointer.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure 
 *  ppCoords::       Pointer to COORDS pointer, which is NULL on successful
 *                   completion
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  geometry
 *
 * Keywords:::
 *  geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomber::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_destroyCoords (
    IPWState           *pIPWState,
    COORDS            **ppCoords)
{
    PROLOGUE;
    COORDS *coords = *ppCoords;

    ERR_GEN(!coords, cErrInvalidParm);
    IPW_FREE (coords->x, cAllocatorData);
    IPW_FREE (coords->y, cAllocatorData);
    IPW_FREE (coords, cAllocatorData);

    *ppCoords = NULL;

    NORMAL_RETURN;
}

/*

 $Log: coords1.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 Imported sources

 Revision 1.16  1996/01/04 20:21:35  ejaquith
 coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.15  1995/12/13  04:05:39  ejaquith
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.14  1995/11/13  20:10:06  ejaquith
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.13  1995/10/13  21:59:07  ejaquith
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.12  1995/10/03  21:12:43  macleod
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.11  1995/09/21  22:51:58  ejaquith
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.10  1995/09/06  20:17:44  ejaquith
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.9  1995/09/01  20:43:17  macleod
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * Added back #include files
 *
 * Revision 1.8  1995/08/31  21:11:31  ejaquith
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.7  1995/08/29  21:53:33  ddavies
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.6  1995/08/25  01:35:43  ejaquith
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.5  1995/08/17  21:54:17  macleod
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.4  1994/10/06  20:53:08  lperry
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * Change to use new .pub header file instead of adtpros.h.
 *
 * Revision 1.3  1993/12/03  01:13:39  ddavies
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:10:33  ddavies
 * coords1.c -> /product/ipcore/ipwrap/src/RCS/coords1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

