/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#include "types.pub"
#include "ipscore.prv"
#define	_NEED_REVERSEBYTE_
#include "revbyte.prv"
 

#include "dgcom.prv"

/************************** Local functions ***************************/

/************************* Internal functions *************************/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Computes the CRC-CCITT checksum on an array of bytes of specified length
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  msg::            	data on which to calculate CRC
 *  len::            	number of bytes of data
 *
 * Returns:::
 *  CRC
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
UInt16 CDECL
wp_calcCRC (
    IPWState	       *pIPWState,
    UInt8              *msg,
    UInt32              len)
{
#define cFT 0x8408U		/* the real CRC feedback term (0x1021)
				 * bit reversed. */
    /* Constants used to declare tables */
#define cF00	0
#define cF01	(cFT >> 7)
#define cF10	0
#define	cF11	(cFT >> 6)
#define cF20	0
#define	cF21	(cFT >> 5)
#define cF30	0
#define	cF31	(cFT >> 4)
#define cF40	0
#define	cF41	(cFT >> 3)
#define cF50	0
#define	cF51	(cFT >> 2)
#define cF60	0
#define	cF61	(cFT >> 1)
#define cF70	0
#define	cF71	(cFT >> 0)

    UInt32 i;
    UInt16 acc = 0;

	/* The original code retired 1 bit of data at a time in the
	 * simple way.  If the xor of the MS bits of accumulated CRC
	 * and data was 1, we shifted, then XORed in the feedback term.
	 * If the XOR of the two MS bits was 0, we skipped XORing in the
	 * feedback term.
	 * Can we retire more than one bit at a time?  Well, sure.
	 * Look at two bits.  Note that there's no feedback term
	 * until the fourth bit.  If the XOR of the top two bits are 00,
	 * we just shift left twice.  That's because no feedback
	 * term can affect the decision.  If the XOR of the top two
	 * bits are 10, we shift left by 2 and XOR with feedback shifted
	 * left by 1.  This is the same as shifting left by 1, XORing
	 * the feedback in and shifting the result left by 1 again with
	 * no XOR the second time.  Similarly, if the XOR of the top
	 * two bits is 01, we shift left by 2 and XOR with unshifted
	 * feedback.
	 * This extends to retiring three bits easily since the feedback
	 * can still not affect the decision.  Just for symmetry, let's
	 * try extending to four bits.  Unlike what I thought when I
	 * tried this the first time, it's just like three bits.  Yes,
	 * there is a feedback term in the LS bit of the MS nibble.
	 * However, since we shift BEFORE we insert the feedback term,
	 * it doesn't affect any of the first four bits that make it
	 * to the decision point.
	 * Great, can we do better?  Sure.  Let's try for all 8 bits
	 * at once.  Note that the upper four bits of the XOR term
	 * feedback and affect the lower four bits.  In particular,
	 * feedback.3 is equal to XOR(data.7, acc.31, data.3, acc.27),
	 * feedback.2 is equal to XOR(data.6, acc.30, data.2, acc.26),
	 * and so on.
	 * Can we do even better?  Well, a little.  Note that we
	 * spend time shifting the accumulated data down into place,
	 * which is sort of slow.  We could avoid this if we bit-
	 * reversed the data and accumulator.  That would mean shifting
	 * the accumulator right instead of left, which is ok.
	 * Note that doing this also means that the feedback table
	 * can have 16-bit entries instead of 32-bit entries, which
	 * is cool.
	 * So how do we make the feedback table?  Well, since we're
	 * bit reversed, the bits in the LS nibble of the index
	 * affect the bits in the MS nibble instead of the other
	 * way around.  That means the table has 16 sections.
	 * The lower 4 terms follow the same sequencein each section .
	 * The upper 4 terms copy the lower four terms initially
	 * because of the feedback.  When a bit in the upper nibble
	 * is set and its term would otherwise have been present,
	 * the term vanishes.  This is because the upper nibble bit
	 * switches the sense of the fed back lower nibble bit.
	 * Similarly, when an upper nibble bit is set and the
	 * corresponding lower nibble bit is clear, the upper nibble
	 * bit's term appears instead of vanishing.
	 */
    UInt16 feedback[256] =
    {
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 00 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 01 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 02 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 03 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 04 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 05 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 06 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 07 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 08 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 09 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 0A */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 0B */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 0C */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 0D */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 0E */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 0F */

	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 10 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 11 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 12 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 13 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 14 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 15 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 16 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 17 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 18 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 19 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 1A */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 1B */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 1C */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 1D */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 1E */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 1F */

	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 20 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 21 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 22 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 23 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 24 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 25 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 26 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 27 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 28 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 29 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 2A */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 2B */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 2C */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 2D */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 2E */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 2F */

	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 30 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 31 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 32 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 33 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 34 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 35 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 36 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 37 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 38 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 39 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 3A */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 3B */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 3C */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 3D */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 3E */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 3F */


	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 40 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 41 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 42 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 43 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 44 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 45 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 46 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 47 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 48 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 49 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 4A */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 4B */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 4C */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 4D */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 4E */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 4F */

	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 50 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 51 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 52 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 53 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 54 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 55 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 56 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 57 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 58 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 59 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 5A */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 5B */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 5C */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 5D */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 5E */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 5F */

	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 60 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 61 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 62 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 63 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 64 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 65 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 66 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 67 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 68 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 69 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 6A */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 6B */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 6C */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 6D */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 6E */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 6F */

	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 70 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 71 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 72 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 73 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 74 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 75 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 76 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 77 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 78 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 79 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 7A */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 7B */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 7C */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 7D */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 7E */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 7F */


	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 80 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 81 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 82 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 83 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 84 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 85 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 86 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 87 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 88 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 89 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 8A */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 8B */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 8C */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 8D */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 8E */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 8F */

	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* 90 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* 91 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* 92 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* 93 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* 94 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* 95 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* 96 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* 97 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* 98 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* 99 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* 9A */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* 9B */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* 9C */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* 9D */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* 9E */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* 9F */

	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* A0 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* A1 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* A2 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* A3 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* A4 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* A5 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* A6 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* A7 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* A8 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* A9 */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* AA */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* AB */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* AC */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* AD */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* AE */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* AF */

	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* B0 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* B1 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* B2 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* B3 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* B4 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* B5 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* B6 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* B7 */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* B8 */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* B9 */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* BA */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* BB */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* BC */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* BD */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* BE */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* BF */


	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* C0 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* C1 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* C2 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* C3 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* C4 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* C5 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* C6 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* C7 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* C8 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* C9 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* CA */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* CB */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* CC */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* CD */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* CE */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* CF */

	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* D0 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* D1 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* D2 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* D3 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* D4 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* D5 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* D6 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* D7 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* D8 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* D9 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* DA */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* DB */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* DC */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* DD */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* DE */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* DF */

	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* E0 */
	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* E1 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* E2 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* E3 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* E4 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* E5 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* E6 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* E7 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* E8 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* E9 */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* EA */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* EB */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* EC */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* ED */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* EE */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF01,  /* EF */

	cF71 ^ cF61 ^ cF51 ^ cF41  ^  cF30 ^ cF20 ^ cF10 ^ cF00,  /* F0 */
	cF71 ^ cF61 ^ cF51 ^ cF40  ^  cF30 ^ cF20 ^ cF10 ^ cF01,  /* F1 */
	cF71 ^ cF61 ^ cF50 ^ cF41  ^  cF30 ^ cF20 ^ cF11 ^ cF00,  /* F2 */
	cF71 ^ cF61 ^ cF50 ^ cF40  ^  cF30 ^ cF20 ^ cF11 ^ cF01,  /* F3 */
	cF71 ^ cF60 ^ cF51 ^ cF41  ^  cF30 ^ cF21 ^ cF10 ^ cF00,  /* F4 */
	cF71 ^ cF60 ^ cF51 ^ cF40  ^  cF30 ^ cF21 ^ cF10 ^ cF01,  /* F5 */
	cF71 ^ cF60 ^ cF50 ^ cF41  ^  cF30 ^ cF21 ^ cF11 ^ cF00,  /* F6 */
	cF71 ^ cF60 ^ cF50 ^ cF40  ^  cF30 ^ cF21 ^ cF11 ^ cF01,  /* F7 */
	cF70 ^ cF61 ^ cF51 ^ cF41  ^  cF31 ^ cF20 ^ cF10 ^ cF00,  /* F8 */
	cF70 ^ cF61 ^ cF51 ^ cF40  ^  cF31 ^ cF20 ^ cF10 ^ cF01,  /* F9 */
	cF70 ^ cF61 ^ cF50 ^ cF41  ^  cF31 ^ cF20 ^ cF11 ^ cF00,  /* FA */
	cF70 ^ cF61 ^ cF50 ^ cF40  ^  cF31 ^ cF20 ^ cF11 ^ cF01,  /* FB */
	cF70 ^ cF60 ^ cF51 ^ cF41  ^  cF31 ^ cF21 ^ cF10 ^ cF00,  /* FC */
	cF70 ^ cF60 ^ cF51 ^ cF40  ^  cF31 ^ cF21 ^ cF10 ^ cF01,  /* FD */
	cF70 ^ cF60 ^ cF50 ^ cF41  ^  cF31 ^ cF21 ^ cF11 ^ cF00,  /* FE */
	cF70 ^ cF60 ^ cF50 ^ cF40  ^  cF31 ^ cF21 ^ cF11 ^ cF01}; /* FF */

	/* Keep compiler happy */
    pIPWState = pIPWState;

	/* After all that, here's the big loop.  Since we're retiring
	 * 8 bits at a time, we shift by 8.  Since we're keeping the
	 * accumulated CRC in reverse bit order, we shift right, not
	 * left.  This also means we have to reverse the bits of
	 * each message byte..
	 */
    for (i = 0; i < len; i++)
    {
	acc = (acc >> 8) ^ feedback[(acc & 0xFF) ^ reverseByte[msg[i]]];
    }

	/* Having calculated the CRC bytes, we reverse the bits and
	 * return the 16 bit CRC.
	 */
    return ((reverseByte[acc & 0xFF] << 8) |
	    (reverseByte[acc >> 8]));

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Store CRC into byte stream.
 *
 * LongDescription:::
 * When storing a CRC in a byte stream, we cannot be sure that we
 * are starting on a 16-bit boundary, so store each byte
 * individually. The CRC may be byte-reordered, so that it is
 * guaranteed to be stored as big-endian.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  crc::            	crc to be added to byte stream
 *
 * Output:::
 *  data::           	byte stream
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_storeCRC (
    IPWState	       *pIPWState,
    UInt32              crc,
    UInt8              *data)
{
	/* Keep compiler happy */
    pIPWState = pIPWState;

    data[1] = (UInt8)(crc & 0xFF);
    data[0] = (UInt8)((crc & 0xFF00) >> 8);
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * Get CRC from a byte stream.
 *
 * LongDescription:::
 * When pulling out CRC from a byte stream, we cannot be sure that
 * we are starting on a 16-bit boundary, so must read out each byte
 * individually. The CRC is guaranteed to have been stored as big-
 * endian, and so may be byte-reordered on retrieval.
 *
 * Input:::
 *  pIPWState::		pointer to IPCore state
 *  data::           	byte stream
 *
 * Output:::
 *  crc::            	crc
 *
 * FunctionCategory:::
 *  DataGlyph
 *
 * Keywords:::
 *  DataGlyph
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_retrieveCRC (
    IPWState	       *pIPWState,
    UInt8              *data,
    UInt32             *crc)
{
	/* Keep compiler happy */
    pIPWState = pIPWState;

    *crc = ((UInt32)data[0] << 8) | (UInt32)data[1];
}

/**************************** End of file *****************************/

/*
$Log: crcgen.c,v $
Revision 1.1.1.1  1999/09/20 18:19:15  jlin
Imported sources

Revision 1.20  1998/01/14 00:11:07  ddavies
crcgen.c -> /product/ipcore/ipwrap/src/RCS/crcgen.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.19  1996/03/21 21:28:58  ddavies
crcgen.c -> /product/ipcore/ipwrap/src/RCS/crcgen.c,v

Eliminated wp_hardwareCRC.  Changed the wp_generateCRC so that it
calculates a new CRC roughly 3x faster.

Revision 1.18  1996/01/04 20:21:40  ejaquith
crcgen.c -> /product/ipcore/ipwrap/src/RCS/crcgen.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.17  1995/11/09  01:18:54  jcraig
 * crcgen.c -> /product/ipcore/ipwrap/src/RCS/crcgen.c,v
 *
 * CommonCheck: wrap ERR_IGNORE around all private procs which do not return an
 * ErrCode
 *
 * Revision 1.16  1995/10/13  21:59:14  ejaquith
 * crcgen.c -> /product/ipcore/ipwrap/src/RCS/crcgen.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.15  1995/09/30  00:18:29  ejaquith
 * crcgen.c -> /product/ipcore/ipwrap/src/RCS/crcgen.c,v
 *
 * CommonCheck: Converted private-ish glyph routines to be of "wp_" form and 100% private.
 *
 * Revision 1.14  1995/09/27  23:41:20  jcraig
 * crcgen.c -> /product/ipcore/ipwrap/src/RCS/crcgen.c,v
 *
 * new procedure comments
 *
*/

