/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * display.c
 *
 *	Display images -- mostly for debugging
 *           w_displayImage()
 */

#ifdef SOLARIS2_ENV
/* Stupidity to get popen() when including stdio.h */
#define _XOPEN_SOURCE
#endif /* SOLARIS2_ENV */

#include <stdio.h>

#ifdef SOLARIS2_ENV
/* ansiprot.h includes sys/resource.h.  On some Solaris installations,
 * maybe 2.5 and later, sys.resource.h defines rusage, which requires
 * a type called timeval.  timeval is defined in time.h, but only if
 * _XOPEN_SOURCE is not defined.  Gack.  So undefine it.
 */
#undef _XOPEN_SOURCE
#endif /* SOLARIS2_ENV */

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "ptypes.h"
#include "gdefs.h"
#include "pcx.h"

/* prototypes */
#include "props.pub"
#include "io.pub"
#include "io.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


IP_RCSINFO(RCSInfo, "$RCSfile: display.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:15 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Display an image in a window.
 *
 * LongDescription:::
 *      Display an image in a window. This will only work on Sparcs,
 *      unless someone gets fairly ambitious. WARNING: pbmplus must
 *      be enabled, because this uses "pv". The call returns immediately;
 *      it doesn't wait for the viewer sub-process to die.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prImage::        Ptr to PIXR to display.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_displayImage (
    IPWState           *pIPWState,
    PIXR               *prImage)
{
    PROLOGUE;
#ifdef _UNIX_

#if 0
    char    tmp[L_tmpnam], cmd[L_tmpnam + 32];

    ERR_GEN_PIXR(prImage, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    tmpnam (tmp);
    ERR_PASS( w_write (pIPWState, tmp, prImage, TIFF_UNCOMP));

    sprintf (cmd, "(anytopnm %s | /tool/unixtools/bin/pv ; rm %s) &",
	     tmp, tmp);
    system (cmd);

    NORMAL_RETURN;
#else
    FILE *fp;

    ERR_GEN_PIXR(prImage, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    fp = popen ("/tool/unixtools/bin/pv", "w");

    ERR_GEN(fp == NULL, cErrIO_CreatePipeFail);
    
    ERR_PASS_CLEANUP( wp_fwritePnm (pIPWState, fp, prImage));

    fclose(fp);
    NORMAL_RETURN;

ERR_CLEANUP:
    fclose(fp);
    ERR_RETURN;
#endif	


#else

    /* Not on THIS processor, sorry. */
    /* Shut the compiler up */
    prImage = prImage;

    ERR_GEN(1, cErrNoImpl);

    NORMAL_RETURN;
#endif

}

/* 
 * $Log: display.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:15  jlin
 * Imported sources
 *
 * Revision 1.23  1997/06/30 23:08:58  ddavies
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Delete inclusion of Sandpiper's internal.h.
 *
 * Revision 1.22  1996/03/21 21:33:14  ddavies
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Undefine _XOPEN_SOURCE after including stdio.h so that on Solaris 2.5,
 * where /sys/resource.h defines the rusage struct, timeval will be defined.
 *
 * Revision 1.21  1996/01/04 20:22:00  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.20  1995/12/13  04:06:15  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.19  1995/11/13  20:04:21  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Added needed additional inclusions of "ptypes.h", "gdefs.h", and "pcx.h".
 * Updated PROLOGUE usage.
 *
 * Revision 1.18  1995/10/13  21:59:38  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/10/07  01:21:49  ddavies
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Include stdio.h since ipwcore.prv no longer includes it (will be true soon).
 *
 * Revision 1.16  1995/09/29  00:35:34  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.15  1995/09/21  22:52:28  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.14  1995/09/12  16:55:09  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.13  1995/09/07  01:30:28  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.12  1995/09/06  20:18:16  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.11  1995/09/01  02:20:53  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.10  1995/08/31  21:11:57  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.9  1995/08/29  21:53:55  ddavies
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Removed inclusion of obsolete file debug.h.
 *
 * Revision 1.8  1995/08/22  02:20:51  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.7  1995/08/17  22:01:39  ejaquith
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.6  1995/03/30  19:27:56  macleod
 * macleod on Thu Mar 30 11:27:26 PST 1995
 *
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Under Solaris, one must define "_XOPEN_SOURCE" in order to get
 * stdio.h to define "popen".
 *
 * Revision 1.5  1995/01/27  19:42:30  lperry
 * lperry on Fri Jan 27 11:35:10 PST 1995
 *
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Use explicit path for the display program pv.
 * The newest version of purify has a GUI executable
 * called pv, which can end up earlier in the path.
 *
 * Revision 1.4  1994/10/10  22:32:58  lperry
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Change to use new header file io.pub instead of iopros.h
 *
 * Revision 1.3  1994/08/16  22:22:25  macleod
 * macleod on Tue Aug 16 15:22:02 PDT 1994
 *
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * Parenthesized an assigment in order to avoid gcc's "advice".
 *
 * Revision 1.2  1994/08/16  21:50:59  macleod
 * macleod on Tue Aug 16 14:49:24 PDT 1994
 *
 * display.c -> /product/ipcore/ipwrap/src/RCS/display.c,v
 *
 * displayImage creates a pipe to pv, and writes a pnm file to the
 * pipe. A bit messy: I call fwritePrPnm directly, since fwritePr
 * was unfortunately changed so that it doesn't take a FILE *, but takes
 * a file name instead.
 * Also silenced a warning.
 *
 * Revision 1.1  1994/08/12  00:34:56  macleod
 * Initial revision
 *
 */
