/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* epsutil.c
 *
 * $Log: epsutil.c,v $
 * Revision 1.2  2000/02/29 00:22:54  jlin
 * Add strong arm related changes
 *
 * Revision 1.1.1.1  1999/09/20  18:19:15  jlin
 * Imported sources
 *
 * Revision 1.6  1995/09/29 00:40:06  ejaquith
 * ejaquith on Thu Sep 28 17:39:44 PDT 1995
 *
 * epsutil.c -> /product/ipcore/ipwrap/src/RCS/epsutil.c,v
 *
 * Added macros for TRUE and FALSE.
 * Yanked inclusion of alplib.h.
 *
 * Revision 1.5  1995/09/01  02:23:21  ejaquith
 * epsutil.c -> /product/ipcore/ipwrap/src/RCS/epsutil.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.4  1995/08/31  21:12:08  ejaquith
 * epsutil.c -> /product/ipcore/ipwrap/src/RCS/epsutil.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.3  1995/08/17  22:04:51  ejaquith
 * epsutil.c -> /product/ipcore/ipwrap/src/RCS/epsutil.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.2  1994/11/22  19:01:01  macleod
 * epsutil.c -> /product/ipcore/ipwrap/src/RCS/epsutil.c,v
 *
 * wgetc and wcrgets were declared "static" in the header file, but
 * weren't "static" here.
 *
 * Revision 1.1  1994/10/21  17:18:11  ejaquith
 * Initial revision
 *
 * 
 *    Rev 1.3   24 Jun 1994 18:14:10   MSEIDNER
 * 
 *    Rev 1.2   23 Jun 1994 16:46:36   MSEIDNER
 * 
 * 
 *    Rev 1.1   22 Jun 1994 15:52:16   SMARTIN
 * Compiled for the first time.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "types.pub"
#include "gdefs.h"

/* prototypes */
#include "gfio.h"
#include "gfiopro.h"
#include "epsutil.h"

#define FALSE 0
#define TRUE  1

#define BUFLNGTH  512
static unsigned char  file_buffer[BUFLNGTH];
static int   buffered_lngth;

static int  wcrgets(char *buf, int max, void* gfioToken);
static int  wgetc(void *gfioToken);

/*--ejaquith-- added prototype: */
long int strtol( const char *ptr, char **endptr, int base );

/*--ejaquith-- commented this out
void epsUtilInit()
{
// wbufcnt = 0;
}
*/

int epsHeaderRead( void *gfioToken, Bool *color, UInt8 *depth, UInt32 *width, UInt32 *height )
{
	int      status = 1;
   int      bDepth = FALSE;
   int      bColor = FALSE;
   int      bWidth = FALSE;
   int      bHeight = FALSE;
   int      bDoPic = FALSE;
   char     buf[256];
   char     *p;
  
   while( status && (!bColor || !bDepth || !bWidth || !bHeight ) ) 
   {
      status = wcrgets(buf, 256, gfioToken);

		if (!(strncmp(buf, "/COLOR_EPS", 10))) 
		{
            bColor = TRUE;
            p = buf + 10;
      	    p = strtok(p, " ");
            *color = atoi(p);
        } else 
        if (!(strncmp(buf, "/BITS_PIXEL", 11))) 
        {
            bDepth = TRUE;
            p = buf + 11;
      	    p = strtok(p, " ");
            *depth = atoi(p);
        }else 
        if (!(strncmp(buf, "/width", 6))) 
        {
            bWidth = TRUE;
            p = buf + 6;
      	    p = strtok(p, " ");
            *width = atoi(p);
        } else 
        if (!(strncmp(buf, "/height", 7))) 
        {
            bHeight = TRUE;
            p = buf + 7;
      	    p = strtok(p, " ");
            *height = atoi(p);
        } 
    }

   GFIO_Seek( gfioToken, 0 ); /* seek to the beginning of the file */
   
   while( status && !bDoPic ) 
   {
       status = wcrgets(buf, 256, gfioToken);
	   if (!(strncmp(buf, "/dopic", 6))) 
           bDoPic = TRUE;
   }
   
   /* and do it a second time */
   bDoPic = FALSE;
   while( status && !bDoPic ) 
   {
       status = wcrgets(buf, 256, gfioToken);
	   if (!(strncmp(buf, "dopic", 5))) 
           bDoPic = TRUE;
   }
   buf[0] = wgetc( gfioToken ); /* cleanup the linefeed */
   
   if (!bColor || !bDepth || !bWidth || !bHeight || !bDoPic )
      return -1;
   else
      return 0;
}

int epsScanlineRead( GDATA *pGData, void *gfioToken, char *buffer )
{
   int   i;
   long  bytewidth;
   char  *pDataBuf;
   char  c;
   char num_buf[3];
   int x;
         
   num_buf[2] = 0x0;   /* null terminate */
   /*--ejaquith--added (long)*/
   bytewidth =  ((long)(pGData->x_dim * pGData->depth )) / 8;
   pDataBuf = buffer;

   for ( i=0; i < bytewidth; i++ )
   {
      for( x = 0; x < 2; x++ )
          num_buf[x] = wgetc( gfioToken );
      if( ( c = wgetc( gfioToken ) ) != 0x20 )   /* detect space */
          break;
      *pDataBuf++ = (unsigned char)strtol( &num_buf[0], NULL, 16 );
   } /* End for */
   if( i < bytewidth )
   {
       return( 0 );
   } else
       return( 1 );    
}


static int wcrgets(char *buf, int max, void* gfioToken)
{
  int c;

  max--;

  while ((c = wgetc(gfioToken)) != EOF && c != 0x0d && c != 0x0a && max--)
    if (c != 0x0a)
      *buf++ = c;

  *buf++ = '\n';
  *buf = '\0';

  return (c == EOF ? 0 : 1);
}

/* --ejaquith--changed C++ comments to C */
/*  wgetc
    acts like getc(), just uses your own buffer and counter*/

static int wgetc( void *gfioToken )
{
  return( get_buffered_char( gfioToken ) );
}

unsigned char get_buffered_char( void *fptr )
{
    
    if( buffered_lngth == BUFLNGTH || buffered_lngth == -1 )
    {
	/*--ejaquith--added (char *)*/
        GFIO_Read( fptr, (char *)&file_buffer[0], BUFLNGTH );
        buffered_lngth = 0;
    } /* End if */
    return( file_buffer[buffered_lngth++] );   
}

void init_rd_buffer( void )
{
    buffered_lngth = -1;
}
