/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill6.c:  includes subroutines
 *
 *  -------------  Add new rectangle to the list of rectangles -------------
 *	  Int32	 addCCRectToList()
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"

/* Prototypes */
#include "props.pub"
#include "shrip.pub"
#include "rectops.pub"
#include "fill.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: fill6.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/* This routine is defined here since it is used in both rabb and
 * rabbLite and they are in different modules.
 */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Callback procedure used by rabb and rabbLite
 *
 * LongDescription:::
 *  This is rabb's callback proc that adds the rectangles found by
 *  i_findConnectedComponentBoundingBoxes to the list of rectangles.
 *  We want the coordinates to be relative to the upper left corner of 
 *  the image, not relative to the upper left corner of the frame.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState:: ptr to IPCore state struct
 *  packetV::  void * which is casted into an AddCCPacket structure pointer
 *  x::        X-coordinate of upper-left corner of bounding rectangle      
 *  y::        Y-coordinate of upper-left corner of bounding rectangle 
 *  w::        Width of bounding rectangle      
 *  h::        Height of bounding rectangle
 *  m::        Number of black pixels in connected component
 *  nRuns::    Number of runs in connected component      
 *  px::       X-coordinate of a point on the connected component
 *  py::       Y-coordinate of a point on the connected component
 *
 * Output:::
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_addCCRectToList (
    IPWState	       *pIPWState,
    void *              packetV,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    RECTANGL	*rect = NULL, *rectList;
    AddCCPacket *packet;

    packet = (AddCCPacket *) packetV;

        /*  make a rect */
    ERR_PASS( w_createRect(pIPWState, x - cFrameBits, y - cFrameBits, w, h, 
			 &rect));

	/* set number of active pixels, run and a point that's really
	 * in the connected component. */
    rect->mass = m;
    rect->nRuns = nRuns;
    rect->px = px - cFrameBits;
    rect->py = py - cFrameBits;

    rectList = packet->rectList;
    if (rectList == NULL)
    {
	rectList = rect;
    }
    else
    {
	rect->next = rectList;
	rect->next->prev = rect;
	rectList = rect;
    }
    packet->rectList = rectList;

    NORMAL_RETURN;
}


/*

 $Log: fill6.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.13  1998/01/14 00:11:38  ddavies
 fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.12  1996/01/04 20:22:31  ejaquith
 fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.11  1995/12/13  04:06:58  ejaquith
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.10  1995/11/13  20:11:06  ejaquith
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.9  1995/10/13  22:00:11  ejaquith
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.8  1995/10/03  21:13:02  macleod
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * CommonCheck: Updated function header comments
 *
 * Revision 1.7  1995/09/21  22:52:56  ejaquith
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.6  1995/09/06  20:18:57  ejaquith
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.5  1995/09/01  20:44:17  macleod
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * Added back #include files
 *
 * Revision 1.4  1995/08/31  21:12:38  ejaquith
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.3  1995/08/23  20:19:23  jcraig
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * the packet given to addCCRectToList is a point to a rectangle not
 * a pointer to a pointer to a rectangle.
 *
 * Revision 1.2  1995/08/17  21:55:06  macleod
 * fill6.c -> /product/ipcore/ipwrap/src/RCS/fill6.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.1  1995/07/14  02:54:20  ddavies
 * Initial revision
 *

 */

