/*<HTML><HEAD><TITLE>fill7.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  fill7.c:  includes subroutines
 *
 *      w_grayPropagate () - Geodesic grayscale propagation
 */



#include "ipwcore.prv"
#include "defines.pub"
#include "frames.pub"
#include "pixr.h"
#include "fill.h"
#include "geomadt.h"
#include "shrcc.pub"
#include "imageref.pub"

/* Prototypes */
#include "props.pub"
#include "rast.pub"
#include "shrip.pub"
#include "rectops.pub"
#include "fill.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: fill7.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");

/*
 * The queue contains a pointer to the pixel to which to propagate,
 * and the value to propagate
 */
typedef struct
{
    UInt8  *pPixel;   /* Pointer to image */
    UInt8   val;      /* Pixel value */
} QueueEntry;

/* If the queue fills, double and copy */
#define EXPAND_QUEUE \
q = IPW_REALLOC (q, qLen * 2 * sizeof(QueueEntry), cAllocatorData); \
ERR_GEN(!q, cErrNomem); \
memcpy (&(q[qHead + qLen]), &(q[qHead]), (qLen - qHead) * sizeof(QueueEntry));\
if (qTail > qHead) qTail += qLen;\
qHead += qLen;\
qLen += qLen;\
qNext = qTail + 1;\
if (qNext >= qLen) qNext = 0;


#define PROPAGATE(neighbor) \
if (UCHAR_ACCESS (neighbor) == 0)\
{\
    UCHAR_ACCESS (neighbor) = q[qHead].val; \
    qNext = qTail + 1; \
    if (qNext >= qLen) \
	qNext = 0; \
    if (qNext == qHead) \
    { \
	EXPAND_QUEUE; \
    } \
    q[qTail].pPixel = neighbor;  \
    q[qTail].val = q[qHead].val; \
    qTail = qNext; \
}

#define cInitialQueueLength 2048

/*****************************************************************************
 * BeginDocumentation::: "w_grayPropagate4"
 *
 * OneLineDescription:::
 *  Propagate grayscale values to unfilled neighbors, 4-connected
 *
 * LongDescription:::
 *  Given a grayscale image, propagate all non-zero pixels to (zero)
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly. The neighborhood is defined as
 *  the pixels left, right, above, and below.
 *
 * PrivateDescription:::
 *  This function uses an iterative queue-based algorithm. First,
 *  we scan the whole image and add non-zero pixels to the queue.
 *  Then, we repeatedly take pixels out of the queue. For each pixel,
 *  we first write the propagated value into that pixel. Then,
 *  we look at the neighborhood pixels. If any of them are non-zero,
 *  we add those pixels to the queue. 
 *  <BR>
 *  Note that in the current scheme, we write the initial non-zero pixels
 *  when we don't actually need to. This probably doesn't take much
 *  time, but it is an area to look at for speed improvement.
 *  <BR> 
 *  ---<BR>
 *  This function no longer uses the iterative queue-based algorithm,
 *  but instead calls the sequential algorithm found in ipshared.
 *
 * Input:::
 *  pIPWState::     State structure
 *
 * Output:::
 *  prImage::       Image to dilate
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *  If this becomes useful, the low-level should be moved to ipshared.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  experimental
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *  "Morphological Algorithms" by Luc Vincent, 1991, Harvard Robotics 
 *  Laboratory technical report no. 91-12.
 *
 * EndDocumentation::: <A NAME = "w_grayPropagate4"></A>
 *****************************************************************************/
ErrCode CDECL
w_grayPropagate4 (
    IPWState *pIPWState, 
    PIXR     *prImage)
{
    PROLOGUE;
    Int32 w, h, bplS, bplDist, channels, i;
    PIXR *prDistance = NULL, *prChannel = NULL;
    ScaledInt1616 xDPI, yDPI;
    UInt8 *pS, *pDist;

    ERR_GEN (prImage == NULL, cErrNullPtr);
    ERR_GEN_PIXR (prImage, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);

    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prImage, &channels));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prImage, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prImage, &h));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prImage, &bplS));
    ERR_PASS_CLEANUP (w_pixrGetResolutionDPI (pIPWState, prImage,
						  &xDPI, &yDPI));

    /* Create a distance image */
    ERR_PASS (w_createPixr (pIPWState, w*2, h, 8, xDPI, yDPI, 
			    cFrameBits, &prDistance));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prDistance, &bplDist));
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prDistance, 
					 &pDist));

    for (i = 0; i < channels; i++)
    {
	if (i)
	{
	    /* Clear the distance image */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, NULL, PIX_CLR,
					  0, 0, w*2, h, prDistance));
	}

	prChannel =  NULL;

	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prImage, i, 
					    &prChannel));

	ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, prChannel, &pS));

	/* Do the propagation on the channel. */
	ERR_PASS_CLEANUP (i_propagateGray4 (pIPWState, pS, w, h,
					    bplS, 8, pDist,
					    bplDist));

    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prDistance));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prDistance)
	w_destroyPixr (pIPWState, &prDistance);
    ERR_RETURN;

}



/*****************************************************************************
 * BeginDocumentation::: "w_grayPropagate8"
 *
 * OneLineDescription:::
 *  Propagate grayscale values to unfilled neighbors, 8-connected
 *
 * LongDescription:::
 *  Given a grayscale image, propagate all non-zero pixels to (zero)
 *  neighbors using a geodesic dilation algorithm, such that pixels
 *  expand into neighborhoods evenly. The neighborhood is defined as
 *  all eight surrounding pixels
 *
 * PrivateDescription:::
 *  This function uses an iterative queue-based algorithm. First,
 *  we scan the whole image and add non-zero pixels to the queue.
 *  Then, we repeatedly take pixels out of the queue. For each pixel,
 *  we first write the propagated value into that pixel. Then,
 *  we look at the neighborhood pixels. If any of them are non-zero,
 *  we add those pixels to the queue. 
 *  <BR>
 *  Note that in the current scheme, we write the initial non-zero pixels
 *  when we don't actually need to. This probably doesn't take much
 *  time, but it is an area to look at for speed improvement.
 *
 * Input:::
 *  pIPWState::     State structure
 *
 * Output:::
 *  prImage::       Image to dilate
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *  If this becomes useful, the low-level should be moved to ipshared.
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  morphology, seed filling
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  experimental
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *  "Morphological Algorithms" by Luc Vincent, 1991, Harvard Robotics 
 *  Laboratory technical report no. 91-12.
 *
 * EndDocumentation::: <A NAME = "w_grayPropagate8"></A>
 *****************************************************************************/
ErrCode CDECL
w_grayPropagate8 (
    IPWState *pIPWState, 
    PIXR     *prImage)
{
    PROLOGUE;
    Int32      w, h, bpl, x, y, qHead, qTail, qLen, qNext, i, channels;
    QueueEntry *q = NULL;
    UInt8      *pImg = NULL, *pLine, *pPixel, pixel;
    PIXR       *prChannel = NULL;

    ERR_GEN (prImage == NULL, cErrNullPtr);
    ERR_GEN_PIXR (prImage, cByteOrderBig, cDepth8, cChannelsAny, cFrame32);

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, prImage, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, prImage, &h));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, prImage, &bpl));
    ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prImage, &pImg));
    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, prImage, &channels));

    /* Make a queue */
    qHead = qTail = 0;
    qLen = cInitialQueueLength;
    q = IPW_MALLOC(qLen * sizeof (QueueEntry), cAllocatorData);
    ERR_GEN (!q, cErrNomem);

    for (i = 0; i < channels; i++)
    {
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, prImage, i,
					    &prChannel));

	/* First pass: Set up the queue */
	pLine = pImg;

	for (y = 0; y < h; y++)
	{
	    pPixel = pLine;
	    for (x = 0; x < w; x++, pPixel++)
	    {
		pixel = UCHAR_ACCESS (pPixel);
		if (pixel)
		{
		    qNext = qTail + 1;
		    if (qNext >= qLen)
			qNext = 0;
		    if (qNext == qHead)
		    {
			EXPAND_QUEUE;
		    }
		    q[qTail].val = pixel;
		    q[qTail].pPixel = pPixel;
		    qTail = qNext;
		}
	    }
	    pLine += bpl;
	}
 

	/* Fill the frame with gunk */
	ERR_PASS_CLEANUP (w_setFramePixels (pIPWState, prChannel, 1));

	/* Now, iterate until the queue is empty */
	for (;;)
	{
	    if (qHead == qTail)
		break;

	    pPixel = q[qHead].pPixel;
 	
	    /* See if it propagates -- 8-connected  */
	    PROPAGATE (pPixel - 1);
	    PROPAGATE (pPixel + 1);
	    PROPAGATE (pPixel - bpl- 1);
	    PROPAGATE (pPixel - bpl);
	    PROPAGATE (pPixel - bpl + 1);
	    PROPAGATE (pPixel + bpl - 1);
	    PROPAGATE (pPixel + bpl);
	    PROPAGATE (pPixel + bpl + 1);

	    qHead++;
	    if (qHead >= qLen)
		qHead = 0;
	}

	ERR_PASS_CLEANUP (w_setFramePixels (pIPWState, prChannel, 0));
    }

    IPW_FREE (q, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (q)
	IPW_FREE (q, cAllocatorData);
    
    ERR_RETURN;
}


/*
 $Log: fill7.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:16  jlin
 Imported sources

 Revision 1.9  1998/01/14 00:11:40  ddavies
 fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.8  1997/06/26 20:00:55  ddavies
 fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.7  1996/07/22 14:12:19  macleod
 fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v

 Fixed "nag script" warnings--some missing colons, missing ERR_CLEANUP
 wrappers, misspelled arguments in comment, etc.

 Revision 1.6  1996/07/19 15:33:13  macleod
 fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v

 w_grayPropagate4 uses the ipshared routine, instead of a slower,
 cache-unfriendly iterative FIFO-based algorithm. It is annoying that
 I named the ipshared routine "propagateGray" and the ipwrap routine
 "grayPropagate."

 Revision 1.5  1996/01/08 16:04:47  macleod
 macleod on Mon Jan 8 08:04:10 PST 1996

 fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v

 changed w_grayPropagate4 and w_grayPropagate8 to take multi-channel
 images

 * Revision 1.4  1995/12/13  04:07:02  ejaquith
 * fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.3  1995/12/04  16:12:12  macleod
 * fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v
 *
 * Fixed HTML header, I hope.
 *
 * Revision 1.2  1995/12/04  16:02:08  macleod
 * macleod on Mon Dec 4 08:01:45 PST 1995
 *
 * fill7.c -> /product/ipcore/ipwrap/src/RCS/fill7.c,v
 *
 * Was #including "queue.pub", which I don't use and never checked in.
 * Took out the reference.
 *
 * Revision 1.1  1995/12/03  00:52:18  macleod
 * Initial revision
 *
 */

/*</PRE></BODY>*/
