/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  gesture1.c--Circle gesture detection/extraction processing stuff.
 *
 *			PUBLIC
 *			w_extractCircledRegions ()
 *
 *                      PRIVATE
 *			wp_cg_filterCCData5 ()
 *			wp_cg_filterCCDataPost4 ()
 *			wp_cg_filterCCForMask1 ()
 *			wp_cg_preProcessForCircles3 ()
 *                      wp_cg_filterRuleCCs ()
 *                      wp_cg_breakRules ()
 *                      wp_cg_getCirclesAndParts ()
 *			wp_cg_postProcessForCircles3 ()
 *			wp_cg_closeCircles1 ()
 *			wp_cg_getSubSubMasks ()
 *			wp_cg_isValidMask2 ()
 *                      wp_cg_makePseudoMask ()
 *                      wp_cg_extractCircledRegions ()
 *			wp_cg_freeSubImageList ()
 *
 */

/* TRUE = dump intermediate rule removal pixrs to disk */
#define DEBUG_RR 0

/* TRUE = print status/times to stdout */
#define DEBUG_RR_P 0

/* TRUE = dump intermediate circle extraction pixrs to disk */
#define DEBUG_CE 0

/* TRUE = dump human/machine discrimination data */
#define DEBUG_HM 0

/* TRUE = print human/machine discrimination info */
#define DEBUG_HM_P 0

/* TRUE = dump mask edge pixrs to disk */
#define DEBUG_EDGES 0


/* straight edge to total edge mass ratio for machine drawn descrimination */
/* this value has been experimentally derived */
#define cEdgeRatio 0.4

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "shrcc.pub"
#include "frames.pub"
#include "morphg.h"
#include "alplib.h"
#include "logicops.h"
#include "shrscgr.pub"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "gesture.prv"
#include "gesture.pub"
#include "strel.pub"
#include "morph.pub"
#include "fill.pub"
#include "intops.pub"
#include "logicops.pub"
#include "subsmpl.pub"
#include "reduce.pub"
#include "clean.pub"
#include "shrrast.pub"
#include "cut.pub"
#include "expand.pub"
#include "cntr.pub"

#if DEBUG_CE || DEBUG_RR || DEBUG_EDGES || DEBUG_HM
#include <stdio.h>
#ifdef __GNUC__
#include "ansiprot.h"
#endif
#include "imagerep.h"
#include "io.pub"
#endif

#if DEBUG_HM || DEBUG_HM_P
static char maskCount = 'a';
#endif

#if DEBUG_EDGES || DEBUG_HM
char fileN[32];
Int32 fc = 0;
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: gesture1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:16 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Callback procedure passed into i_findConnectedComponentBoundingBoxes
 *
 * LongDescription:::
 *    Callback procedure passed into i_findConnectedComponentBoundingBoxes
 *    which removes non-circle related CCs.  It will seed CCs 
 *    which are:
 * <BR>
 *     - very dense (sparsity < 1.5);
 *     - nearly single line segments (h == nRuns);
 *     - long skinny things (either vertical or horizontal);
 *     - large and sparse (whole circles);
 *     - very sparse (line segments with gaps/noise).
 * <BR>
 *    This removes most textual components while preserving
 *    things which may be disconnected circle parts.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Pointer to IPCore state
 *  packet::         Generic CC analysis structure
 *  x::              X-location of current CC
 *  y::              Y-location of current CC
 *  w::              Width of current CC
 *  h::              Height of current CC
 *  m::              Mass (number of ON bits) in current CC
 *  nRuns::          Number of runs in current CC
 *  px::             x-location of an ON pixel in the current CC (used
 *                   as a seed)
 *  py::             y-location of an ON pixel in the current CC (used
 *                   as a seed)
 *
 * Output:::
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture, connected component analysis
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  stable
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *  Illegal U-turn, Manteca, CA, May 1992
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_filterCCData5 (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    Int32   area,
            minX,
	    minY,
	    minA;
    PIXR   *outPixr;
    circle_gesture_packet *CCpacket;

    /* to eliminate warnings : */
    x = x;
    y = y;

    CCpacket = (circle_gesture_packet *)packet;
    outPixr = (PIXR *)(CCpacket->data);
    minX = CCpacket->minDimensionX;
    minY = CCpacket->minDimensionY;
    minA = CCpacket->minArea;

    area = w * h;

    /* Component seeding */
    if ((((Float32)area / (Float32)m) < 1.5) ||
        (nRuns == h) ||
	(w > (7 * h)) ||
	(h > (7 * w)) ||
	((area > minA) && ((w > minX) || (h > minY)) && (area > 4 * m)) ||
	(area > 10 * m))
        ERR_PASS (w_pixrPutPixel (pIPWState, outPixr, px - cFrameBits, 
				py - cFrameBits, 1));
    NORMAL_RETURN;

}  /* wp_cg_filterCCData5 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Callback for i_findConnectedComponentBoundingBoxes
 *
 * LongDescription:::
 *  Callback procedure passed into i_findConnectedComponentBoundingBoxes
 *  which filters out CCs which are obviously not circles. It expects
 *  that circles are now connected into 1 component.  The following
 *  CC types are excluded:
 *  <BR>
 *    - Too small or too narrow (w or h),<BR>
 *    - Too dense,<BR>
 *    - Non-enclosing (< 1.7 runs/line) [heuristic!].
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Pointer to IPCore state
 *  packet::         Generic CC analysis structure.
 *  x::              x location of current CC.
 *  y::              y location of current CC.
 *  w::              Width of current CC.
 *  h::              Height of current CC.
 *  m::              Mass (number of ON bits) in
 *	  	     current CC.
 *  nRuns::          Number of runs in current CC.
 *  px::             x location of an ON pixel in the
 *		     current CC (for use as a seed).
 *  py::             y location of an ON pixel in the
 *		     current CC (for use as a seed).
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture, connected component analysis
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_filterCCDataPost4 (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    Int32      area,
               minX,
	       minY,
	       minA;
    sub_mask **list,
              *thisMask;
    circle_gesture_packet *CCpacket;

    CCpacket = (circle_gesture_packet *)packet;
    minX = CCpacket->minDimensionX;
    minY = CCpacket->minDimensionY;
    minA = CCpacket->minArea;
    list = (sub_mask **)(CCpacket->data);

    area = w * h;

    /* Circle CC selection process: */

    if ((area > minA) && ((w > minX) || (h > minY)))
    {
	if ((((Float32)area / (Float32)m) > 3.2))
	{
	    if ((((Float32)nRuns / (Float32)h) > 1.7))
	    {
	        thisMask = (sub_mask *)IPW_MALLOC (sizeof (sub_mask), 
						   cAllocatorData);
		ERR_GEN (thisMask == NULL, cErrNomem);
		thisMask->image = NULL;
		thisMask->x = x - cFrameBits;
		thisMask->y = y - cFrameBits;
		thisMask->w = w;
		thisMask->h = h;
		thisMask->xSeed = px - cFrameBits;
		thisMask->ySeed = py - cFrameBits;

		/* put it in the list */
		thisMask->next = *list;
		*list = thisMask;
	    }
	}
    }

    NORMAL_RETURN;

}  /* wp_cg_filterCCDataPost4 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Create a sub_mask for each CC in mask whose area is >min area
 *
 * LongDescription:::
 *   Creates a sub_mask struct for each CC in a mask whose
 *   area is greater than the minimum allowable area.  Inserts
 *   the structure in the list passed in with the packet.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Pointer to IPCore state
 *  packet::         Generic CC analysis structure.
 *  x::              x location of current CC.
 *  y::              y location of current CC.
 *  w::              Width of current CC.
 *  h::              Height of current CC.
 *  m::              Mass (number of ON bits) in
 *		     current CC.
 *  nRuns::          Number of runs in current CC.
 *  px::             x location of an ON pixel in the
 *		     current CC (for use as a seed).
 *  py::             y location of an ON pixel in the
 *		     current CC (for use as a seed).
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture, connected component analysis
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_filterCCForMask1 (
    IPWState	       *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    Int32      minA;
    sub_mask **list,
              *thisMask;
    circle_gesture_packet *CCpacket;

    /* to eliminate compiler warnings: */
    nRuns = nRuns;

    CCpacket = (circle_gesture_packet *)packet;
    minA = CCpacket->minArea;
    list = (sub_mask **)(CCpacket->data);

    /* Valid circle gesture mask CC selection process: */

    if (m > minA)
    {
	thisMask = (sub_mask *)IPW_MALLOC (sizeof (sub_mask), cAllocatorData);
	ERR_GEN (thisMask == NULL, cErrNomem);
	thisMask->image = NULL;
	thisMask->x = x - cFrameBits;
	thisMask->y = y - cFrameBits;
	thisMask->w = w;
	thisMask->h = h;
	thisMask->xSeed = px - cFrameBits;
	thisMask->ySeed = py - cFrameBits;
#if 0
	thisMask->nRuns = nRuns;
	thisMask->m = m;
#endif

	/* put it in the list */
	thisMask->next = *list;
	*list = thisMask;
    }

    NORMAL_RETURN;

}  /* wp_cg_filterCCForMask1 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Small asymmetric morphological close
 *
 * LongDescription:::
 *   Performs a small asymmetric morphological close on the input 
 *   image.  The operation is intended to close small gaps in
 *   circles and generally reduce the number of CCs to be
 *   subsequently processed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           Input image to be processed.
 *  xres::           Horizontal resolution of the image.
 *  yres::           Vertical resolution of the image.
 *  outPr::          Output image which is created by
 *		     this function if pointer to NULL
 *		     passed in.
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   gesture
 *
 * Keywords:::
 *   gesture, morphology
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_preProcessForCircles3 (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               xres,
    Int32               yres,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32 w, h;
    STREL *sClose = NULL;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);


    if (inPr != (*outPr))
    {
        if ((*outPr) == NULL)
	{
            ERR_PASS (w_clonePixrNoInit (pIPWState, inPr, outPr));
	}
    }

    w = MAX (2, xres / 100);
    h = MAX (2, yres / 100);

    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, w, h, w/2, h/2, &sClose));

    if ((w > 2) && (h > 2))
    {
	sClose->data[0][0] = DONT_CARE;
	sClose->data[0][h-1] = DONT_CARE;
	sClose->data[w-1][0] = DONT_CARE;
	sClose->data[w-1][h-1] = DONT_CARE;
    }

    ERR_PASS_CLEANUP (w_close (pIPWState, inPr, sClose, outPr));

    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &sClose));

    NORMAL_RETURN;
ERR_CLEANUP:
    if (sClose)
	w_destroyStrel (pIPWState, &sClose);

    ERR_RETURN;
} /* cg_preProcessForCircles3 () */


#if 0   /* WARNING -- construction area.  Keep out! */
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Seed only "rule-like" CCs.
 *
 * LongDescription:::
 *    Seed only "rule-like" CCs.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      
 *  packet::         Generic CC analysis structure.
 *  x::              x location of current CC.
 *  y::              y location of current CC.
 *  w::              Width of current CC.
 *  h::              Height of current CC.
 *  m::              Mass (number of ON bits) in current CC.
 *  nRuns::          Number of runs in current CC.
 *  px::             x location of an ON pixel in the 
 *                   current CC (for use as a seed).
 *  py::             y location of an ON pixel in the
 *		     current CC (for use as a seed).
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   gesture
 *
 * Keywords:::
 *   gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_filterRuleCCs (
    IPWState           *pIPWState,
    void               *packet,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    IPWState *pIPWState;
    PIXR *seedPr;

    /* keep compiler quiet */
    x = x;
    y = y;
    w = w;
    h = h;
    m = m;
    nRuns = nRuns;


    seedPr = ((circle_gesture_packet2 *) packet)->prImage;
    pIPWState = ((circle_gesture_packet2 *) packet)->pIPWState;

/*
    if (((w * h) / m)  <= 3)
*/
    ERR_PASS (w_pixrPutPixel (pIPWState, seedPr, px - cFrameBits, 
			    py - cFrameBits, 1));

    NORMAL_RETURN;

}  /* wp_cg_filterRuleCCs () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Attempt to break horizontal and vertical rules
 *
 * LongDescription:::
 *   Attempts to break horizontal and vertical rules by
 *   removing large sections of them.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           Source pixr
 *  xres::           horizontal resolution of source (in dpi)
 *  yres::           vertical resolution of source (in dpi)
 *
 * Output:::
 *  outPr::          pointer to destination pixr pointer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_breakRules (
    IPWState           *pIPWState,
    PIXR               *inPr,
    ScaledInt1616       xres,
    ScaledInt1616       yres,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32     ret, w, h, sw, sh,
              minLinearSize, maxLinearSize,
              minThickness, maxThickness, framedWidth, framedHeight, bpl;
    STREL    *thicken = NULL;  
    UInt8    *pImageAll;
#if 0
    STREL    *connect;
#endif
    PIXR     *hPr = NULL, *vPr = NULL, *tPr = NULL, *rInPr = NULL, 
	     *seedPr = NULL;
    circle_gesture_packet2 packet;

    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &h));

    minLinearSize = MIN (ScaledInt1616_TO_Int32(xres) * 2, 
			 ScaledInt1616_TO_Int32(yres) * 2);
    maxLinearSize = MAX (w, h);
    minThickness = 1;
    maxThickness = MAX ((ScaledInt1616_TO_Int32(xres) / 28), 
			(ScaledInt1616_TO_Int32(yres) / 28));

    ERR_PASS (w_reduce (pIPWState, inPr, 2, 0, 0, 0, &rInPr));

#if DEBUG_RR_P
    fprintf (stdout, "reduction : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "red24", rInPr, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, rInPr, &w));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, rInPr, &h));

    minLinearSize = (minLinearSize + 1) / 2;
    maxLinearSize = (maxLinearSize + 1) / 2;
    minThickness = (minThickness + 1) / 2;
    maxThickness = (maxThickness + 1) / 2;
    sw = MAX (3, ScaledInt1616_TO_Int32(xres)/ 70);
    sh = MAX (3, ScaledInt1616_TO_Int32(yres) / 70);

    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, rInPr, &hPr));
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, rInPr, &vPr));
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, sw, sh, sw / 2, sh / 2, &thicken));

    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, rInPr, 
					      minLinearSize/2,
					      maxLinearSize, RF_PASS, &hPr));

#if DEBUG_RR_P
    fprintf (stdout, "Hrun filter (for H rules) : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "rfhh", hPr, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, rInPr, minThickness, 
					    maxThickness, RF_PASS, &vPr));

#if DEBUG_RR_P
    fprintf (stdout, "Vrun filter (for H rules) : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "rfvh", vPr, TIFF_G3));
#endif

/*
    logicOp (pIPWState, hPr, vPr, AND);
*/
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, hPr, AND, vPr));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, vPr, thicken, &hPr));

#if 0
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 
					ScaledInt1616_TO_Int32(xres)/ 20,
					1,
					ScaledInt1616_TO_Int32(xres)/ 40, 0, 
					&connect));
    ERR_PASS_CLEANUP (w_close (pIPWState, hPr, connect, &hPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &connect));
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "hrules", hPr, TIFF_G3));
#endif

#if DEBUG_RR_P
    fprintf (stdout, "AND, dilate, and H close (for H rules) : %f\n", 
             StopTime ());
#endif

    /* Run CC-BB and eliminate too short horizontal CCs */
    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, rInPr, &seedPr));
    packet.prImage = seedPr;
    ERR_PASS_CLEANUP (w_pixrGetImageAll (pIPWState, hPr, &pImageAll));
    ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, hPr, &bpl));
    ERR_PASS_CLEANUP (w_pixrFramedWidth (pIPWState, hPr, &framedWidth));
    ERR_PASS_CLEANUP (w_pixrFramedHeight (pIPWState, hPr, &framedHeight));

    i_findConnectedComponentBoundingBoxes (pIPWState, pImageAll,
					   bpl,
					   0, 0,
					   framedWidth,
					   framedHeight,
					   minLinearSize, minThickness,
					   1, &packet,
					   cg_filterRuleCCs,
					   &ret);

#if DEBUG_RR_P
    fprintf (stdout, "find CCs (for H rules) : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "seedh", seedPr, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, seedPr, hPr, 2, &seedPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &hPr));
    hPr = seedPr;

#if DEBUG_RR_P
    fprintf (stdout, "8 connected fill (pIPWState, for H rules) : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "fill8h", hPr, TIFF_G3));
#endif

    /* clip! */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, rInPr, AND, hPr));

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "hrclip", hPr, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, rInPr, minLinearSize/2,
					    maxLinearSize, RF_PASS, &vPr));

#if DEBUG_RR_P
    fprintf (stdout, "Vrun filter (for V rules) : %f\n", StopTime ());
#endif

    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, rInPr, &tPr));

    ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, rInPr, minThickness, 
					      maxThickness, RF_PASS, &tPr));
    
#if DEBUG_RR_P
    fprintf (stdout, "Hrun filter (for V rules) : %f\n", StopTime ());
#endif
    
/*
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, tPr, AND, vPr));
*/
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, vPr, AND, tPr));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, tPr, thicken, &vPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &thicken));

#if 0
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 1, ScaledInt1616_TO_Int32(yres) / 20, 0, ScaledInt1616_TO_Int32(yres) / 40,
				      &connect));
    ERR_PASS_CLEANUP (w_close (pIPWState, vPr, connect, &vPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &connect));
#endif

#if DEBUG_RR_P
    fprintf (stdout, "AND and dilate (for V rules) : %f\n", StopTime ());
#endif

    /* Run CC-BB and eliminate too short horizontal CCs */
    ERR_PASS_CLEANUP (clonePixrZero (pIPWState, rInPr, &seedPr));
    packet.prImage = seedPr;
    i_findConnectedComponentBoundingBoxes (pIPWState, 
					   pixrGetImageAll (pIPWState, vPr),
					   pixrGetBpl (pIPWState, vPr),
					   0, 0,
					   pixrFramedWidth (pIPWState, vPr),
					   pixrFramedHeight (pIPWState, vPr),
					   minThickness, minLinearSize,
					   1, &packet,
					   cg_filterRuleCCs,
					   &ret);

#if DEBUG_RR_P
    fprintf (stdout, "find CCs (for V rules) : %f\n", StopTime ());
#endif

    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, seedPr, vPr, 2, &seedPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &vPr));
    vPr = seedPr;

#if DEBUG_RR_P
    fprintf (stdout, "8 connected fill (for V rules) : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "fill8v", vPr, TIFF_G3));
#endif

    /* clip! */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, rInPr, AND, vPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &rInPr));

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "vrclip", vPr, TIFF_G3));
#endif

/*
  ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, hPr, 
  PIX_NOT (PIX_SRC|PIX_DST), 0, 0, w, h, vPr));
*/
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 0, 0, hPr,
				  PIX_NOT (PIX_SRC^PIX_DST), 
				  0, 0, w, h, vPr));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &hPr));

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "rulemask", vPr, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, inPr, outpr));

    ERR_PASS_CLEANUP (w_logicOp (pIPWState, inPr, COPY, *outPr));
    ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, vPr, *outPr));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &vPr));

#if DEBUG_RR_P
    fprintf (stdout, "Final mask steps : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "rrout1", *outPr, TIFF_G3));
#endif

    /* try to reconstruct gaps in circles */
    ERR_PASS_CLEANUP (w_clonePixrNoInit (pIPWState, inPr, &tPr));
/*
  makeBrickStrel (pIPWState, ScaledInt1616_TO_Int32(xres)/ 40, 
  ScaledInt1616_TO_Int32(yres) / 40, ScaledInt1616_TO_Int32(xres)/ 80, ScaledInt1616_TO_Int32(yres) / 80, &thicken);
*/
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 
					ScaledInt1616_TO_Int32(xres)/ 28, 
					ScaledInt1616_TO_Int32(yres) / 28, 
					ScaledInt1616_TO_Int32(xres)/ 56, 
					ScaledInt1616_TO_Int32(yres) / 56,
					&thicken));
    ERR_PASS_CLEANUP (w_dilate (pIPWState, *outPr, thicken, &tPr));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &thicken));

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "gapmask", tPr, TIFF_G3));
#endif

    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, *outPr, inPr, 1, outPr));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, tPr, AND, *outPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tPr));

#if DEBUG_RR_P
    fprintf (stdout, "Circle reconstruction steps : %f\n", StopTime ());
#endif

#if DEBUG_RR
    ERR_PASS_CLEANUP (w_write (pIPWState, "norules", *outPr, TIFF_G3));
#endif

    NORMAL_RETURN;

ERR_CLEANUP:
    if (hPr) w_destroyPixr (pIPWState, &hPr);
    if (vPr) w_destroyPixr (pIPWState, &vPr);
    if (tPr) w_destroyPixr (pIPWState, &tPr);
    if (rInPr) w_destroyPixr (pIPWState, &rInPr);
    if (seedPr) w_destroyPixr (pIPWState, &seedPr);
    if (thicken) w_destroyStrel (pIPWState, &thicken);
    ERR_RETURN;
}  /* wp_cg_breakRules () */
#endif

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Filters out CCs which are not circles or circle parts.
 *
 * LongDescription:::
 *    Filters out CCs which are not circles or potential
 *    disconnected circle parts.  This includes most text.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           source pixr to be filtered
 *  minW::           minimum allowable CC width
 *  minH::           minimum allowable CC height
 *  minA::           minimum allowable CC area
 *
 * Output:::
 *  outPr::          pointer to result pixr pointer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_getCirclesAndParts (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               minW,
    Int32               minH,
    Int32               minA,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32 ret, bpl, width, height;
    UInt8 *pImageAll;
    circle_gesture_packet CCpacket;

    /* Collect seeds for connected components which may
       be parts of circles */
    if ((*outPr) == NULL)
    {
        ERR_PASS (w_clonePixr (pIPWState, inPr, outPr));
    }

    /* set up the callback packet correctly */
    CCpacket.data = (void *)(*outPr);
    CCpacket.minDimensionX = minW;
    CCpacket.minDimensionY = minH;
    CCpacket.minArea = minA;

    ERR_PASS (w_pixrGetImageAll (pIPWState, inPr, &pImageAll));
    ERR_PASS (w_pixrGetBpl (pIPWState, inPr, &bpl));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &height));

    ERR_PASS( i_findConnectedComponentBoundingBoxes (pIPWState,
						     pImageAll,
						     bpl,
						     cFrameBits, cFrameBits,
						     width,
						     height,
						     1, 1, 1, 
						     (void *)(&CCpacket), 
						     wp_cg_filterCCData5, 
						     &ret));


    /*=============> Seed fill the potential circle parts */
    /* For somewhat convoluted circle shapes (or those that intersect 
     * text), the iteration count is very important.  At 5, some circle 
     * parts in the test suite were not completely filled.  This is 
     * particularly troublesome for open circles.  We may want to add 
     * some more slop here, but it does have a relatively high cost.
     */

    ERR_PASS (w_fillClipSweep8 (pIPWState, *outPr, inPr, 6, outPr));

    NORMAL_RETURN;

}  /* wp_cg_getCirclesAndParts () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Dilate with a brick and erode with a "cross".
 *
 * LongDescription:::
 *   Performs an asymmetric 3x3 Close operation.  Dilates
 *   with a brick and Erodes with a "cross."  Used to close
 *   small gaps in circles prior to clipping out candidates
 *   for separate processing.
 *  
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           Input image to be processed.
 *
 * Output:::
 *  outPr::          Output image which is created by
 *		     this function if pointer to NULL 
 *		     passed in.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_postProcessForCircles3 (
    IPWState           *pIPWState,
    PIXR               *inPr,
    PIXR              **outPr)
{
    PROLOGUE;
    PIXR  *tmpPr = NULL;
    STREL *sDilate = NULL;
    STREL *sErode = NULL;

    if ((*outPr) == NULL)
    {
        ERR_PASS (w_clonePixrNoInit (pIPWState, inPr, outPr));
    }

    ERR_PASS (w_makeBrickStrel (pIPWState, 3, 3, 1, 1, &sDilate));
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 3, 3, 1, 1, &sErode));

    /* make the erosion strel a "cross" shape */
    sErode->data[0][0] = DONT_CARE;
    sErode->data[0][2] = DONT_CARE;
    sErode->data[2][0] = DONT_CARE;
    sErode->data[2][2] = DONT_CARE;
    
    ERR_PASS_CLEANUP (w_dilate (pIPWState, inPr, sDilate, &tmpPr));
    ERR_PASS_CLEANUP (w_erode (pIPWState, tmpPr, sErode, outPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));

    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &sDilate));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &sErode));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (sDilate) w_destroyStrel (pIPWState, &sDilate);
    if (sErode) w_destroyStrel (pIPWState, &sErode);
    if (tmpPr) w_destroyPixr (pIPWState, &tmpPr);

    ERR_RETURN;


}  /* wp_cg_postProcessForCircles3 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do a medium brick dilation followed by a cross-shaped erode.
 *
 * LongDescription:::
 *  Performs a medium brick dilation followed by a cross
 *  shaped erode.  This "asymmetric close" will bridge gaps
 *  while preserving diagonal connections made by the dilate.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           Input image
 *  xres::           Horizontal resolution of the image.
 *  yres::           Vertical resolution of the image.
 *
 * Output:::
 *  outPr::          Output image which is created by
 *		     this function.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *   gesture
 *
 * Keywords:::
 *   gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_closeCircles1 (
    IPWState           *pIPWState,
    PIXR               *inPr,
    ScaledInt1616       xres,
    ScaledInt1616       yres,
    PIXR              **outPr)
{
    PROLOGUE;
    Int32 w, h, i, j;
    PIXR  *tmpPr = NULL;
    STREL *sDilate = NULL;
    STREL *sErode = NULL;

    *outPr = NULL;


    /* set the strel size relative to the resolution of the pixr */

    w = MAX (3, (xres / 40));
    h = MAX (3, (yres / 40));
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, w, h, w/2, h/2, &sDilate));

    /* The erosion strel is slightly smaller to help keep the circle
       connected. */
    w = MAX (3, (xres / 50));
    h = MAX (3, (yres / 50));
    ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, w, h, w/2, h/2, &sErode));

    /* make the erosion strel a "cross" shape */
    for (i = 0; i < h; i++)
	for (j = 0; j < w; j++)
	    if ((i != h/2) && (j != w/2))
		sErode->data[i][j] = DONT_CARE;

    ERR_PASS_CLEANUP (w_dilate (pIPWState, inPr, sDilate, &tmpPr));
    ERR_PASS_CLEANUP (w_erode (pIPWState, tmpPr, sErode, outPr));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr));

    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &sDilate));
    ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &sErode));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (sDilate) w_destroyStrel (pIPWState, &sDilate);
    if (sErode) w_destroyStrel (pIPWState, &sErode);
    if (tmpPr) w_destroyPixr (pIPWState, &tmpPr);
    ERR_RETURN;

}  /* wp_cg_closeCircles1 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Separate sub-image candidate into individual alleged gestures
 *
 * LongDescription:::
 *  Since we may have a single sub-image candidate which
 *  contains more than one actual circle gesture (or other
 *  candidate circle-looking-thing to be filtered out later)
 *  we desire to separate them into individual parts to more
 *  easily process them.  This function starts with a single
 *  submask struct which is presumed to contain a mask (sort 
 *  of) for a subsection of the image.  It is also assumed 
 *  that the images processed herein are 200dpi.
 *  <BR>
 *  For images with only one interesting mask region, the
 *  mask is clipped to just slightly larger than its extent
 *  to allow for safety in later processing steps.  For images
 *  with multiple candidate regions, each is clipped out and
 *  appended to the list of submask structures following the
 *  one which was originally passed in.  
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inList::         Node to examine for resizing, splitting,
 *		     or "removal."
 *  minWidth::       Minimum allowable width of mask CC for
 *		     it to be retained.
 *  minHeight::      Minimum allowable height of mask CC for
 *		     it to be retained.
 *  minArea::        Minimum allowable area (mass) of mask CC
 *		     for it to be retained.
 *
 * Output:::
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_getSubSubMasks (
    IPWState           *pIPWState,
    sub_mask           *inList,
    Int32               minWidth,
    Int32               minHeight,
    Int32               minArea)
{
    PROLOGUE;
    Int32 done, ret, bpl, width, height;
    ScaledInt1616 xDPI, yDPI;
    PIXR *tmpMaskImage1 = NULL, *tmpMaskImage2 = NULL;
    UInt8 *pImageAll;
    sub_mask *maskList, *thisMask;
    circle_gesture_packet CCpacket;

    /* check parameters */
    ERR_GEN((inList == NULL) || (minWidth < 0) ||
	    (minHeight < 0) || (minArea < 0), cErrInvalidParm);

    maskList = NULL;

    tmpMaskImage1 = inList->image;

    /* set up the callback packet correctly */
    CCpacket.minDimensionX = minWidth;
    CCpacket.minDimensionY = minHeight;
    CCpacket.minArea = minArea;
    CCpacket.data = (void *)(&maskList);

    /* Construct a list of masks which are "large enough" to
     * be valid circle gestures.
     */
    ERR_PASS (w_pixrGetImageAll (pIPWState, tmpMaskImage1, &pImageAll));
    ERR_PASS (w_pixrGetBpl (pIPWState, tmpMaskImage1, &bpl));
    ERR_PASS (w_pixrGetWidth (pIPWState, tmpMaskImage1, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, tmpMaskImage1, &height));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, tmpMaskImage1, &xDPI, &yDPI));

    ERR_PASS( i_findConnectedComponentBoundingBoxes (
	       pIPWState, 
	       pImageAll, 
	       bpl,
	       cFrameBits, cFrameBits,
	       width,
	       height,
	       minWidth, minHeight, 
	       1, 
	       (void *)(&CCpacket), 
	       wp_cg_filterCCForMask1, 
	       &ret));

    /* Check for empty list which means there were no big mask parts */
    if (maskList == NULL)   
    {
        /* Nothing in the original sub-mask was large enough */
        ERR_PASS (w_destroyPixr (pIPWState, &tmpMaskImage1));
	inList->image = NULL;
    }
    else
    {
        /* check for single element list which is a special case
	 * since we don't want to keep growing the main list with
	 * copies of single masks!  This is the stop growing
	 * condition.
	 */
        if (maskList->next == NULL)
	{
	    /* if the mask is about the same size (> 80% of original), 
	     * don't add it to the main list.
	     */
	    if ((maskList->w > 0.8 * inList->w) &&
                (maskList->h > 0.8 * inList->h))
	    {
	        IPW_FREE (maskList, cAllocatorData);
	    }
	    else
	    {
	        /* the new size is significantly different than the
		 * original size, so make a new PIXR.
		 */
		ERR_CHECK (
		    w_createPixr (pIPWState, maskList->w + (2 * BORDER_WIDTH),
				maskList->h + (2 * BORDER_WIDTH),
				1, xDPI, yDPI, cFrameBits, &(maskList->image)))
		{
		    wp_cg_freeSubImageList (pIPWState, &maskList);
		    ERR_GEN (1, cErrNomem);
                }
		
                ERR_CHECK (w_clonePixr (pIPWState, maskList->image, 
					&tmpMaskImage2))
		{

		    wp_cg_freeSubImageList (pIPWState, &maskList);
		    ERR_GEN (1, cErrNomem);
                }

		/* Copy out the interesting part of the mask 
		 * (including the border!). 
		 */
                ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
					      maskList->x - BORDER_WIDTH,
					      maskList->y - BORDER_WIDTH,
					      tmpMaskImage1,
					      PIX_SRC,
					      0, 0,
					      maskList->w + (2 * BORDER_WIDTH),
					      maskList->h + (2 * BORDER_WIDTH),
					      tmpMaskImage2));

                /* turn on the seed and grow it */
		ERR_PASS_CLEANUP (w_pixrPutPixel (
		    pIPWState, maskList->image, 
		    maskList->xSeed - maskList->x + BORDER_WIDTH, 
		    maskList->ySeed - maskList->y + BORDER_WIDTH, 
		    1));
		    
		ERR_PASS_CLEANUP (w_fillClipSweep4 (
		    pIPWState, maskList->image, tmpMaskImage2, 4, 
		    &maskList->image));

                ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpMaskImage2));

		/* fix the location */
		maskList->x += inList->x - BORDER_WIDTH;
		maskList->y += inList->y - BORDER_WIDTH;

                /* Free the original image */
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpMaskImage1));
		inList->image = NULL;

		/* insert the new thing in the main list */
		maskList->next = inList->next;
		inList->next = maskList;

	    }
	}
	else   /* there are multiple mask components to process */
	{
	    /* Fix up the "relative" coordinates in the list taking into
	     * account the original sub-image's offset and boundary width.
	     * Clip out the sub-sub-mask.
	     */
	    thisMask = maskList;
	    done = cFalse;
	    while (!done)
	    {
	        /* the new size is significantly different than the
		 * original size, so make a new PIXR.
		 */
		ERR_CHECK (w_createPixr (
		    pIPWState, thisMask->w + (2 * BORDER_WIDTH),
		    thisMask->h + (2 * BORDER_WIDTH),
		    1, xDPI, yDPI, cFrameBits, &(thisMask->image)))
		{
		    wp_cg_freeSubImageList (pIPWState, &maskList);
		    ERR_GEN (1, cErrNomem);
                }

                ERR_CHECK (w_clonePixr (
		    pIPWState, thisMask->image, &tmpMaskImage2))
		{
		    wp_cg_freeSubImageList (pIPWState, &maskList);
		    ERR_GEN (1, cErrNomem);
                }

		/* Copy out the interesting part of the mask 
		 * (including the border!). 
		 */
                ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
					      thisMask->x - BORDER_WIDTH,
					      thisMask->y - BORDER_WIDTH,
					      tmpMaskImage1,
					      PIX_SRC,
					      0, 0,
					      thisMask->w + (2 * BORDER_WIDTH),
					      thisMask->h + (2 * BORDER_WIDTH),
					      tmpMaskImage2));
				  
                /* turn on the seed and grow it */
		ERR_PASS_CLEANUP (w_pixrPutPixel (
		    pIPWState, thisMask->image, 
		    thisMask->xSeed - thisMask->x + BORDER_WIDTH, 
		    thisMask->ySeed - thisMask->y + BORDER_WIDTH, 
		    1));

                ERR_PASS_CLEANUP (w_fillClipSweep4 (pIPWState,
						    thisMask->image,
						    tmpMaskImage2, 4, 
						    &thisMask->image));

                ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpMaskImage2));

		/* fix the location */
		thisMask->x += inList->x - BORDER_WIDTH;
		thisMask->y += inList->y - BORDER_WIDTH;

		/* check for running off the end of the list since we
		 * need to keep the pointer to the last element for
		 * insertion into the main list.
		 */
		if (thisMask->next == NULL)
		    done = cTrue;
		else
		    thisMask = thisMask->next;
	    }

	    /* Free the original image */
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpMaskImage1));
	    inList->image = NULL;

	    /* Insert the sublist into the main list */
	    thisMask->next = inList->next;
	    inList->next = maskList;

        }
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (tmpMaskImage1) w_destroyPixr (pIPWState, &tmpMaskImage1);
    if (tmpMaskImage2) w_destroyPixr (pIPWState, &tmpMaskImage2);
   
    ERR_RETURN;

} /* cg_getSubSubMasks () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Determines whether a given sub-image mask is a hand-drawn circle or not.  
 *
 * LongDescription:::
 *   Determines whether a given sub-image mask is a hand-drawn
 *   circle or not.  Checks for "mostly" rectangular things,
 *   things with lots of horizontal and vertical edges, and
 *   things with too many or too small regions.  Attempts to
 *   use aspect ratio to weight importance of vertical &
 *   horizontal edges.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  mask::           The mask image to be evaluated.
 *  xres::           Horizontal resolution of the mask.
 *  yres::           Vertical resolution of the mask.
 *
 * Output:::
 *  pValidFlag::     cTrue if mask edges are irregular,
 *	             cFalse if mask edges are mostly straight.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_isValidMask2 (
    IPWState           *pIPWState,
    PIXR               *mask,
    Int32               xres,
    Int32               yres,
    Int32              *pValidFlag)
{
    PROLOGUE;
    Int32  w, h, l, minW, minH,
           edgeMass, linearMass, maskMass, 
	   vMass, hMass,
	   bbArea, polySize, thinPolySize,
	   colinearDeviation,
	   lastX, lastY, dX, dY, p;
    ipPoint *poly = NULL, *thinPoly = NULL;
    Float32 invTan;
    PIXR  *allEdges = NULL;
    STREL *edgeStrel = NULL;

    if (mask == NULL)
    {
	*pValidFlag = cFalse;
    }
    else
    {
	ERR_PASS_CLEANUP (w_numberBits (pIPWState, mask, &maskMass));

	/* Check for lots of straight edges */
	/* Calc the approximate aspect ratios of the mask */
	/* Compute the size of the vertical & horizontal edge finding strels */

	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, mask, &w));
	w -= (2 * BORDER_WIDTH);
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, mask, &h));
	h -= (2 * BORDER_WIDTH);

	bbArea = w * h;

	/* Erode the mask and XOR it with the original to extract
	 * the edges.
	 */
	ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 3, 3, 1, 1, &edgeStrel));

	/* make the erosion strel a "cross" shape */
	edgeStrel->data[0][0] = DONT_CARE;
	edgeStrel->data[0][2] = DONT_CARE;
	edgeStrel->data[2][0] = DONT_CARE;
	edgeStrel->data[2][2] = DONT_CARE;

	ERR_PASS_CLEANUP (w_erode (pIPWState, mask, edgeStrel, &allEdges));

	ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &edgeStrel));

        ERR_PASS_CLEANUP (w_logicOp (pIPWState, mask, XOR, allEdges));

#if DEBUG_EDGES
	sprintf (fileN, "edges.%ld.tif", fc++);
	ERR_PASS_CLEANUP (w_write (pIPWState, fileN, allEdges, TIFF_G3));
#endif

        /* get the edge mass */
	ERR_PASS_CLEANUP (w_numberBits (pIPWState, allEdges, &edgeMass));

        /* extract the 8-thinned polygon outline of the mask */
	poly = NULL;
	ERR_PASS_CLEANUP (w_contourRegionBinary (pIPWState, allEdges, NULL, 1, 
					       cTrue, &polySize, &poly));

#if DEBUG_HM
	{
	    PIXR *crud1 = NULL;

	    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, allEdges, &crud1));
	    ERR_PASS_CLEANUP (w_drawPolygon (pIPWState, poly, polySize, 
	                             cTrue, cFalse, crud1));
	    sprintf (fileN, "poly1.%c.tif", maskCount);
	    ERR_PASS_CLEANUP (w_write (pIPWState, fileN, crud1, TIFF_G3));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &crud1));
	}
#endif

        /* thin the polygon further by eliminating near-colinear points */
	if ((xres < 145) || (yres < 145))
	    colinearDeviation = 1;
	else
	    colinearDeviation = 2;

	ERR_PASS_CLEANUP (w_thinPolygon (pIPWState, poly, polySize, 
	                                 colinearDeviation,
					 &thinPoly, &thinPolySize));


#if DEBUG_HM
	{
	    PIXR *crud2 = NULL;
	    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, allEdges, &crud2));
	    ERR_PASS_CLEANUP (w_drawPolygon (pIPWState, thinPoly, 
	                          thinPolySize, cTrue, cTrue, crud2));
	    sprintf (fileN, "polyt.%c.tif", maskCount);
	    ERR_PASS_CLEANUP (w_write (pIPWState, fileN, crud2, TIFF_G3));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &crud2));
	}
#endif

	/* Count edge mass in vertical and horizontal edges */
	hMass = 0;
	vMass = 0;
	/* We only want to count relatively long edges */
	minW = MIN (xres / 4, MAX (xres / 8, w / 3));
	minH = MIN (yres / 4, MAX (yres / 8, h / 3));
	lastX = thinPoly[0].x;
	lastY = thinPoly[0].y;
	for (p = 1; p < thinPolySize; p++)
	{
	    dX = ABS (thinPoly[p].x - lastX);
	    dY = ABS (thinPoly[p].y - lastY);

	    /* only consider segments which are relatively long */
	    if ((dX >= minW) || (dY >= minH))
	    {
	        l = MAX (dX, dY);
	        /* handle a small amount of skew */
	        invTan = (Float32)MIN (dX, dY) / (Float32)l;
		if (invTan < 0.035)
		{

                    /* Since long and skinny objects will necessarily
		     * have strong linearity in one dimension, we
		     * weight the the true segment lengths by inverse
		     * aspect ratio in the same dimension.  In other
		     * words, a horizontal segment in a polygon with
		     * a large w:h ratio will have its effective length
		     * reduced in this measurement.  A vertical segment
		     * in the same polygon will have its effective length
		     * increased.
		     */
		    if (dX > dY)   /* horizontal segment */
		        hMass += l;
		    else           /* vertical segment */
		        vMass += l;
		}
	    }

	    lastX = thinPoly[p].x;
	    lastY = thinPoly[p].y;
	}

        /* weight the h and v masses by normalized aspect ratio */
	linearMass = ((hMass * (h * xres)) / (w * yres)) + 
	             ((vMass * (w * yres)) / (h * xres));

        /* we're done with the polygons now */
	IPW_FREE (thinPoly, cAllocatorData);
	IPW_FREE (poly, cAllocatorData);

#if DEBUG_HM_P
	fprintf (stdout, 
		 "Object %c : %ld -> %ld edge points and %ld polygon points, %ld mass, %ld linear.\n",
		 maskCount++, edgeMass, polySize, thinPolySize, maskMass, linearMass);
	fprintf (stdout, "\t%0.1f -> %0.1f => %0.1f\t{%ld}\t|%0.4f|\n",
		 (Float32)edgeMass / polySize,
		 (Float32)polySize / thinPolySize,
		 (Float32)edgeMass / thinPolySize,
		 edgeMass / ((thinPolySize * thinPolySize) * 
	                 (MAX (xres, yres) / MIN (xres, yres))),
		 (Float32)linearMass / (Float32)edgeMass);


#endif

        /* Examine the polygon for shape:
	 * -If it has too few vertices, it is too simple and thus rejected
	 * -If the ratio of edge mass to vertices^2 is too large, it is also
	 *   too simple and thus rejected
	 * -If it has too many long, orthogonal edges it is rejected
	 */

	/* remember that to close the polygons, we have the
	 * first and last point the same...
	 */
	polySize--;
	thinPolySize--;
	if ((thinPolySize < 8) ||
	    ((edgeMass / ((thinPolySize * thinPolySize) * 
	                 (MAX (xres, yres) / MIN (xres, yres)))) >= 5) ||
	    (linearMass > (cEdgeRatio * edgeMass)))
	{
	    *pValidFlag = cFalse;
#if DEBUG_HM_P
	    fprintf (stdout, "NOT HAND DRAWN -----------------------------\n");
#endif
	}
        else
	    /* If it passed thru to here, call it hand-drawn */
	    *pValidFlag = cTrue;

    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &allEdges));

    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyPixr (pIPWState, &allEdges);
    w_destroyStrel (pIPWState, &edgeStrel);
    if (thinPoly != NULL)
	IPW_FREE (thinPoly, cAllocatorData);
    if (poly != NULL)
	IPW_FREE (poly, cAllocatorData);

    ERR_RETURN;

}  /* wp_cg_isValidMask2 () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Construct a "mask" image of valid circles
 *
 * LongDescription:::
 *   Given a binary image with only potential circle gestures
 *   in it, construct a "mask" image of valid circles.  Marks
 *   which do not enclose enough (or any) space or which are
 *   "too square" will not have masked regions in the result.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           source pixr containing circles
 *  xres::           horizontal resolution of source (in dpi)
 *  yres::           vertical resolution of source (in dpi)
 *  minArea::        minimum allowable circle area (in pixels)
 *  minWidth::       minimum allowable circle width (in h pixels)
 *  minHeight::      minimum allowable circle height (in v pixels)
 *
 * Output:::
 *  cCount::         pointer to variable for circle count
 *  maskPr::         pointer to destination pixr pointer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_makePseudoMask (
    IPWState           *pIPWState,
    PIXR               *inPr,
    ScaledInt1616       xres,
    ScaledInt1616       yres,
    Int32               minArea,
    Int32               minWidth,
    Int32               minHeight,
    Int32              *cCount,
    PIXR              **maskPr)
{
    PROLOGUE;
    Int32     ret, bpl, width, height, bits;
    ScaledInt1616     xDPI, yDPI;
    PIXR     *subtmp1 = NULL,
             *subtmp2 = NULL;
    UInt8    *pImageAll;
    sub_mask *imageList,
             *subImage;
    STREL    *pStrel7h = NULL, *pStrel7v = NULL;
    circle_gesture_packet CCpacket;

    /* Inits */
    *cCount = 0;
    imageList = NULL;

    ERR_PASS (w_initStrelByName (pIPWState, strel7h));
    ERR_PASS (w_getStrel (pIPWState, strel7h, &pStrel7h));
    ERR_PASS (w_initStrelByName (pIPWState, strel7v));
    ERR_PASS (w_getStrel (pIPWState, strel7v, &pStrel7v));

    /*=============> Assemble a list of circle-esque components */

    /* set up the callback packet correctly */
    CCpacket.minDimensionX = minWidth;
    CCpacket.minDimensionY = minHeight;
    CCpacket.minArea = minArea;
    CCpacket.data = (void *)(&imageList);
    ERR_PASS (w_pixrGetImageAll (pIPWState, inPr, &pImageAll));
    ERR_PASS (w_pixrGetBpl (pIPWState, inPr, &bpl));
    ERR_PASS (w_pixrGetWidth (pIPWState, inPr, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, inPr, &height));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xDPI, &yDPI));

    ERR_PASS( i_findConnectedComponentBoundingBoxes (
               pIPWState, 
               pImageAll, 
	       bpl,
	       cFrameBits, cFrameBits,
	       width,
	       height,
	       minWidth / 2, minHeight / 2, 1, 
	       (void *)(&CCpacket), 
	       wp_cg_filterCCDataPost4, 
	       &ret));


    if (imageList != NULL)
    {
        if ((*maskPr) == NULL)
	{
	    ERR_CHECK(w_clonePixr (pIPWState, inPr, maskPr))
	    {
		wp_cg_freeSubImageList (pIPWState, &imageList);
		ERR_GEN (1, cErrNomem);
	    }
	}
    }

    /*=============> Process each circle candidate sub-image and create
                     a mask for each valid circle gesture. */

    subImage = imageList;
    while (subImage != NULL)
    {

        /* Check to see if this is one of the original list members
	 * or the result of splitting the mask of one of the original
	 * members in which case we already have the "inverse mask"
	 * image for the component.
	 */
        if (subImage->image == NULL)
	{
	    /* clip the rectangle out of the preprocessed pixr and
	     * seed fill the associated CC.
	     */
	    ERR_PASS_CLEANUP (w_createPixr (pIPWState, subImage->w + (2 * BORDER_WIDTH),
			    subImage->h + (2 * BORDER_WIDTH),
			    1, xDPI, yDPI, cFrameBits, &subtmp1));

    
	    /* snip-snip */
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
					  subImage->x, subImage->y, inPr,
					  PIX_SRC,
					  BORDER_WIDTH, BORDER_WIDTH,
					  subImage->w, subImage->h,
					  subtmp1));

	    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, subtmp1, &subtmp2));

	    /* seed the pixr */
	    ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, subtmp2, 
			      subImage->xSeed - subImage->x + BORDER_WIDTH,
			      subImage->ySeed - subImage->y + BORDER_WIDTH, 
			      1));
    
	    /* grow the seed */
	    ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, subtmp2, 
						subtmp1, 6, &subtmp2));

	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp1));

	    /* OK, damn it!  Looks like we have to give it one more try
	     * to close those stubborn circles--especially 200 dpi
	     * originals which were not preprocessed.
	     */
	    ERR_PASS_CLEANUP (wp_cg_closeCircles1 (pIPWState, subtmp2, 
						  xres, yres, &subtmp1));

	    /*=============> Grab the stuff inside the circle */
    
	    /* we will make what looks like an inverse mask which will
	     * be used to seed fill into the original image producing
	     * the components which should be subtracted from the
	     * clipped original to leave only circled things.
	     */
	    ERR_PASS_CLEANUP (w_edgeFillSweep4 (pIPWState, subtmp1, 4, 
						&subtmp2));

	    /* NOTE: Maybe we should check here for the dual of
	     * the "mostly empty" result which is the mostly
	     * full result.  If we edge fill and the result is
	     * too dense, say it's not a circle and move on.
	     * We use MIN_AREA / 2 so as not to lose borderline
	     * cases where we have increased the density during
	     * preprocessing (connected characters to circles).
	     * This could be particularly troublesome for circles
	     * around single lines of text where we have more
	     * opportunity to fill in the circle accidentally.
	     */
	    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, subtmp2, &width));
	    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, subtmp2, &height));
	    ERR_PASS_CLEANUP (w_numberBits (pIPWState, subtmp2, &bits));
	    if (((width * height) - bits) < (minWidth / 2))
            {
                ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp1));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp2));
		subImage->image = NULL;
	    }
            else
	    {
		/* subtract the circle and stuff it touched from the
		 * black stuff.  This helps us not to accidentally
		 * lose interior components which were connected by
		 * our preprocessing steps.
		 */
		ERR_PASS_CLEANUP (w_logicOp (pIPWState, subtmp1, 
					     SUBTRACT, subtmp2));
	
		/* make the mask positive */
                ERR_PASS_CLEANUP (w_logicOp (pIPWState, subtmp2, INVERT, 
					     subtmp2));

		/* Since the preprocessing operations up to this point have
		 * tended to thicken black marks, we need to erode the
		 * "mask" to make sure we have some seeds on the circle
		 * itself when we re-invert the mask during extraction.
		 * This step also has the happy side-effect of removing
		 * any thickish rules which may have joined larger
		 * mask regions.
		 */
		ERR_PASS_CLEANUP (w_erode (pIPWState, subtmp2, 
					   pStrel7v, &subtmp1));
		ERR_PASS_CLEANUP (w_erode (pIPWState, subtmp1, 
					   pStrel7h, &subtmp2));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp1));

		subImage->image = subtmp2;
                subtmp2 = NULL;

		ERR_PASS_CLEANUP (wp_cg_getSubSubMasks (pIPWState, subImage, minWidth, 
				       minHeight, minArea));
	    }  /* else result of fill not "too dense" */

        }  /* if we had to start from scratch */

	/* OR the sub image into the mask */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      BORDER_WIDTH,
				      BORDER_WIDTH,
				      subImage->image,
				      PIX_SRC | PIX_DST,
				      subImage->x,
				      subImage->y,
				      subImage->w,
				      subImage->h,
				      *maskPr));

	/* increment the count */
	(*cCount)++;

        subImage = subImage->next;
    }

    if (imageList != NULL)
	ERR_PASS_CLEANUP (wp_cg_freeSubImageList (pIPWState, &imageList));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp2));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (imageList != NULL)
	wp_cg_freeSubImageList (pIPWState, &imageList);
    if (subtmp1) w_destroyPixr (pIPWState, &subtmp1);
    if (subtmp2) w_destroyPixr (pIPWState, &subtmp2);

    ERR_RETURN;

}  /* wp_cg_makePseudoMask () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Extract circle contents into a list of sub-image results
 *
 * LongDescription:::
 *   Given the original image and a (possibly reduced) 
 *   "pseudo-mask" image containing black regions where circles 
 *   were detected, extract the circle contents into a list
 *   of sub-image results which are at the original resolution.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  inPr::           original source pixr
 *  maskPr::         pseudo-mask pixr (may be reduced)
 *  xres::           horizontal resolution of source (in dpi)
 *  yres::           vertical resolution of source (in dpi)
 *  hReduction::     horizontal reduction factor
 *  vReduction::     vertical reduction factor
 *  minWidth::       minimum allowable circle width (in h pixels)
 *  minHeight::      minimum allowable circle height (in v pixels)
 *  minArea::        minimum allowable circle area (in pixels)
 *
 * Output:::
 *  cCount::         pointer to variable for extracted circle count
 *  imageList::      pointer to sub-result list head pointer 
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell:
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_extractCircledRegions (
    IPWState           *pIPWState,
    PIXR               *inPr,
    PIXR               *maskPr,
    Int32               xres,
    Int32               yres,
    Float32             hReduction,
    Float32             vReduction,
    Int32               minWidth,
    Int32               minHeight,
    Int32               minArea,
    Int32              *cCount,
    sub_mask          **imageList)
{
    PROLOGUE;
    Int32     retCode,
              ret,
              origMass,
	      bpl,
	      width,
	      height,
              bits,
	      isValid, 
	      subHeight;
    ScaledInt1616     xDPI, yDPI;
    UInt8    *pImageAll;
    PIXR     *subtmp1 = NULL,
             *subtmp2 = NULL,
	     *suborig = NULL;
    sub_mask *maskList = NULL,
             *subMask,
	     *subImage;
    circle_gesture_packet CCpacket;

    /* Inits */
    *imageList = NULL;
    retCode = cSuccessful;

    /*=============> Assemble a list of pseudo-mask regions */

    /* set up the callback packet correctly */
    CCpacket.minDimensionX = minWidth;
    CCpacket.minDimensionY = minHeight;
    CCpacket.minArea = minArea;
    CCpacket.data = (void *)(&maskList);

    ERR_PASS (w_pixrGetImageAll (pIPWState, maskPr, &pImageAll));
    ERR_PASS (w_pixrGetBpl (pIPWState, maskPr, &bpl));
    ERR_PASS (w_pixrGetWidth (pIPWState, maskPr, &width));
    ERR_PASS (w_pixrGetHeight (pIPWState, maskPr, &height));
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, maskPr, &xDPI, &yDPI));

    ERR_PASS( i_findConnectedComponentBoundingBoxes (
               pIPWState,
               pImageAll,
	       bpl,
	       cFrameBits, cFrameBits,
	       width,
	       height,
	       minWidth, minHeight, 1, 
	       (void *)(&CCpacket), 
	       wp_cg_filterCCForMask1, 
	       &ret));


    /*=============> Process each region and extract circle contents. */

    subMask = maskList;
    while (subMask != NULL)
    {
        /* create a PIXR to clip the sub-region into */
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, 
				      subMask->w + (2 * BORDER_WIDTH),
				      subMask->h + (2 * BORDER_WIDTH),
				      1, xDPI, yDPI, cFrameBits, &subtmp2));

	/* snip-snip */
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				      subMask->x, subMask->y, maskPr,
				      PIX_SRC,
				      BORDER_WIDTH, BORDER_WIDTH,
				      subMask->w, subMask->h,
				      subtmp2));

        /* Create a seed PIXR so that we can fill only the interesting
	 * part of the mask we just clipped into subtmp2.
	 */
        ERR_PASS_CLEANUP (w_clonePixr (pIPWState, subtmp2, &subtmp1));

        /* Seed it */
        ERR_PASS_CLEANUP (w_pixrPutPixel (pIPWState, subtmp1, 
			  subMask->xSeed - subMask->x + BORDER_WIDTH,
			  subMask->ySeed - subMask->y + BORDER_WIDTH, 1));

        /* Grow the seed */
        ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, subtmp1, 
					    subtmp2, 8, &subtmp1));
        ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp2));

        /* Here's the _other_ place to check for machine drawn objects */
	ERR_PASS_CLEANUP (wp_cg_isValidMask2 (pIPWState, subtmp1, xres, yres, 
					      &isValid));

	if (!isValid)
	{
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp1));
	    subMask = subMask->next;
	    continue;
	}

	/* We need a negative mask from here on */
        ERR_PASS_CLEANUP (w_logicOp (pIPWState, subtmp1, INVERT, subtmp1));

        /* clip out the corresponding region from the original image
         * being careful about the borders we added to the subimage.
         */
        ERR_PASS_CLEANUP (w_createPixr (pIPWState, 
			(Int32)(hReduction * (subMask->w + (2 * BORDER_WIDTH))),
			(Int32)(vReduction * (subMask->h + (2 * BORDER_WIDTH))),
			1, xDPI, yDPI, cFrameBits, &suborig));

	/* snip-snip */
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, suborig, &width));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, suborig, &height));

	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, suborig, &subHeight));
        ERR_PASS_CLEANUP (w_rasterop (pIPWState,
			(Int32)(hReduction * (subMask->x - BORDER_WIDTH)),
			(Int32)(vReduction * (subMask->y - BORDER_WIDTH)),
				      inPr,
				      PIX_SRC,
				      0, 0,
				      width,
				      subHeight,
				      suborig));
	
	ERR_PASS_CLEANUP (w_numberBits (pIPWState, suborig, &origMass));
	ERR_PASS_CLEANUP (w_copy (pIPWState, suborig, &subtmp2));

	ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, subtmp2, subtmp1, 
					   &subtmp2));

	/* Grow the seeds in subtmp2. */
	ERR_PASS_CLEANUP (w_fillClipSweep8 (pIPWState, subtmp2, 
					    suborig, 4, &subtmp2));

	/* subtract out the circle and stuff outside it */
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, subtmp2, SUBTRACT, suborig));

	/* if this result is empty or "mostly empty", then
	 * back-pedal and do a simple mask with the inverse
	 * mask result in subtmp1.  This will inevitably
	 * include parts of the circle along with stuff
	 * outside the circle in the result.
	 */
	ERR_PASS_CLEANUP (w_numberBits (pIPWState, suborig, &bits));
	if (((origMass + 1) / (bits + 1)) > EMPTY_RESULT_RATIO)
	{
	    /* Add the stuff back in */
	    ERR_PASS_CLEANUP (w_logicOp (pIPWState, subtmp2, UNION, suborig));

	    /* Apply the inverse mask */
	    ERR_PASS_CLEANUP (w_invert (pIPWState, subtmp1, &subtmp1));

	    ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, suborig, subtmp1, 
					       &suborig));
	}

	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp2));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp1));

	/* "Remove" the border */
        ERR_PASS_CLEANUP (w_cutRectIP (pIPWState,
				       (Int32)(hReduction * BORDER_WIDTH), 
				       (Int32)(vReduction * BORDER_WIDTH),
				       (Int32)(hReduction * subMask->w),
				       (Int32)(vReduction * subMask->h),
				       suborig));

        /* Create a new output list node */
	subImage = (sub_mask *)IPW_CALLOC (1, sizeof (sub_mask), 
					   cAllocatorData);

        /* Fill in the data and insert it in the output list */
	subImage->image = suborig;
	suborig = NULL;
	subImage->x = (Int32)(hReduction * subMask->x);
	subImage->y = (Int32)(vReduction * subMask->y);
	subImage->w = (Int32)(hReduction * subMask->w);
	subImage->h = (Int32)(vReduction * subMask->h);
        subImage->next = *imageList;
	*imageList = subImage;

	subMask = subMask->next;
	(*cCount)++;
    }

    ERR_PASS_CLEANUP (wp_cg_freeSubImageList (pIPWState, &maskList));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp2));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &subtmp1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &suborig));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (maskList) wp_cg_freeSubImageList (pIPWState, &maskList);
    if (subtmp2) w_destroyPixr (pIPWState, &subtmp2);
    if (subtmp1) w_destroyPixr (pIPWState, &subtmp1);
    if (suborig) w_destroyPixr (pIPWState, &suborig);

    ERR_RETURN;
}  /* wp_cg_extractCircledRegions () */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free a sub_mask list
 *
 * LongDescription:::
 *  Given a pointer to the head of a list of sub_mask
 *  data structs, frees them and destroys any
 *  associated (non-NULL) PIXRs.  If the  list is empty,
 *  does nothing.  Sets the head pointer to NULL on return.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *
 * Output:::
 *  head::           Head of list to free and set to NULL on exit
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cg_freeSubImageList (
    IPWState           *pIPWState,
    sub_mask          **head)
{
    PROLOGUE;
    sub_mask *thisOne;

    thisOne = *head;
    while (thisOne != NULL)
    {
	sub_mask *tmpMask;

	if (thisOne->image != NULL)
	    ERR_PASS (w_destroyPixr (pIPWState, &thisOne->image));

        /* keep a pointer to this element */
	tmpMask = thisOne;

        /* go to the next element */
	thisOne = thisOne->next;

        /* free the previously visited element */
	IPW_FREE (tmpMask, cAllocatorData);
    }

    *head = NULL;
    
    NORMAL_RETURN;

}  /* wp_cg_freeSubImageList () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finds hand-drawn circle gestures in a PIXR and extracts their contents.
 *
 * LongDescription:::
 *  Finds hand-drawn circle gestures in a PIXR and extracts their contents.
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::           State structure
 *  inPr::                Input PIXR
 *  minWidth::            Minimum width of gesture for
 *			  recognition (in horizontal pixels).
 *  minHeight::           Minimum width of gesture for
 *			  recognition (in vertical pixels).
 *  minArea::             Minimum enclosed area (in pixels at
 *			  input resolution) for a circle
 *			  gesture to be recognized.  If this
 *			  value is too small, enclosing characters
 *			  in text (e.g. o, O, Q, D, etc.) will 
 *			  be recognized as circles.  An area
 *			  of about 0.5 square inch is probably
 *			  the minimum usable value.
 *  separateOutput::      If cTrue, individual result PIXRs
 *			  will be generated for each gesture,
 *			  otherwise, the result will be a single
 *			  PIXR of the same "shape" as inPr
 *			  with all extracted gesture contents
 *			  in it.
 *
 * Output:::
 *  numCirclesExtracted:: Pointer to variable which
 *			  will be set to the number of circle
 *			  gestures recognized and extracted.  This
 *			  is also the length of the RECTANGL
 *			  array for the location of the recognized
 *			  circle gestures.
 *  outPrArrayLength::    Pointer to variable which will be
 *			  set to the actual length of the PIXR
 *			  array (0 if no circles found or error, 
 *			  1 if separate results is cFalse, and 
 *			  >= 1 if separate results is cTrue and
 *			  circles found).
 *  outRectArray[]::      Address of RECTANGL pointer which will
 *			  be set to the address of an allocated
 *			  array of RECTANGLs.  These RECTANGLs
 *			  contain location and size info (and 
 *			  NOTHING else!) for the recognized circle
 *			  gestures.  The length of this array is
 *			  "numCirclesExtracted".
 *  outPrArray[]::        Address of PIXR pointer which will
 *			  be set to an array of PIXR pointers
 *			  where each PIXR is the extracted
 *			  contents of a recognized circle
 *			  gesture found in inPr.  If 
 *			  separateOutput is cTrue, the size
 *			  of this array will be numRegionsExtracted,
 *			  otherwise it will have single entry (a
 *			  PIXR the size of the input PIXR which
 *			  contains only the extracted regions) or
 *			  be NULL (if nothing was found).
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 * Caveats::
 *	The caller is responsible for freeing (using the data free
 *      function passed in at ipwrap initialization) the two arrays 
 *      if they are non-NULL.  The outRectArray is allocated in one 
 *      chunk and can be freed simply.  However, each PIXR in the 
 *      outPrArray must be destroyed (using w_destroyPixr ()) 
 *      individually before the array is freed.  If 
 *      numCirclesExtracted is 0, neither array will be allocated by 
 *      this function.
 * <BR>
 *	minArea should be close to minWidth * minHeight.
 *
 * Bugs::
 *	Circles which intersect both vertical and horizontal
 *	rules will be improperly processed if even detected
 *	(they may sometimes be classified as machine drawn and
 *	ignored).
 *<BR>
 *	Images with dense halftone/dither/stipple may cause
 *	spurious false positives.
 *<BR>
 *	Circles which intersect dense regions may be lost.
 *<BR>
 *	Some types of line art may be recognized as a circle
 *	gesture and be extracted (or partially extracted).
 *	Cartoons are especially troublesome as they are
 *	indeed hand drawn.  Curved geometric figures
 *	(e.g. ellipses, circles, rounded rectangles, etc.)
 *	may also be recognized as circle gestures.
 *
 * FunctionCategory:::
 *  gesture
 *
 * Keywords:::
 *  gesture
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_extractCircledRegions (
    IPWState           *pIPWState,
    PIXR               *inPr,
    Int32               minWidth,
    Int32               minHeight,
    Int32               minArea,
    Int32               separateOutput,
    Int32              *numCirclesExtracted,
    Int32              *outPrArrayLength,
    RECTANGL           *outRectArray[],
    PIXR              **outPrArray[])
{
    PROLOGUE;
    Int32	    retCode,
                    minLDXreduced,
                    minLDYreduced,
                    minAreduced,
                    cCount,
                    cc,
	            i;
    ScaledInt1616   xres1616, 
                    yres1616;
    Int32           xres, yres,
	            xresRed,
	            yresRed;
    Float32         vReduction = 1.0F,
                    hReduction = 1.0F;
    PIXR           *tmpPr1 = NULL,
                   *tmpPr2 = NULL;
    sub_mask       *imageList = NULL,
                   *subImage;

    /* Inits */
    *outRectArray = NULL;
    *outPrArray = NULL;
    *numCirclesExtracted = 0;
    *outPrArrayLength = 0;
    retCode = cSuccessful;

    ERR_GEN_PIXR (inPr, cByteOrderBig, cDepth1, cChannelsAny, cFrame32);

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, inPr, &xres1616, &yres1616));

    xres = ScaledInt1616_TO_Int32 (xres1616);
    yres = ScaledInt1616_TO_Int32 (yres1616);

    ERR_GEN ((xres < 100) || 
             (yres < 100) ||
	     (xres > 600) ||
	     (yres > 600),
	     cErrTypeNotSupported);

    /*=============> Halftone/stipple removal */
    if ((xres > 200) && (yres > 200) &&
	(xres <= 400) && (yres <= 400))
        ERR_PASS (w_createStippleMask (pIPWState, inPr, &tmpPr1));

    if (tmpPr1 != NULL)
    {
	ERR_PASS_CLEANUP (w_copy (pIPWState, inPr, &tmpPr2));
	ERR_PASS_CLEANUP (w_expandAndMask (pIPWState, tmpPr2, tmpPr1, 
					   &tmpPr2));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr1));
#if DEBUG_CE
        ERR_PASS_CLEANUP (w_write (pIPWState, "stip", tmpPr2, TIFF_G3));
#endif
    }
    else
    {
        ERR_PASS_CLEANUP (w_copy (pIPWState, inPr, &tmpPr2));
    }

    /*=============> Reduction : */
    if ((xres >= 290) || 
	(yres >= 290))
    {
        /* NOTE: DAE FAX resolutions can be strange and are often
         * "almost" symmetrical (e.g. 305 x 294, 407 x 392, 204 x 196).
	 */
	if ((MAX (xres, yres) - MIN (xres, yres)) < 16)
	{
	    ERR_PASS_CLEANUP (w_reduce2 (pIPWState, tmpPr2, 1, &tmpPr1));
	    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));
	    hReduction = 2.0F;
	    vReduction = 2.0F;
	}
	else
	{
	    if (xres > yres)
	    {
	        ERR_PASS_CLEANUP (w_reduceCols (pIPWState, tmpPr2, 2, OR, 
						&tmpPr1));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));
		hReduction = 2.0F;
	    }
	    else
	    {
	        ERR_PASS_CLEANUP (w_reduceRows (pIPWState, tmpPr2, 2, OR, 
						&tmpPr1));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));
		vReduction = 2.0F;
	    }
	}

#if DEBUG_CE
        ERR_PASS_CLEANUP (w_write (pIPWState, "red", tmpPr1, TIFF_G3));
#endif

    }
    else  /* image is already "low res" */
        /* do nothing */
	tmpPr1 = tmpPr2;

#if 0
    else  /* image is already "low res" */
    {
	if ((MAX (xres, yres) - MIN (xres, yres)) < 16)
	{
	    /* do nothing */
	    tmpPr1 = tmpPr2;
        }
	else
	{
	    if (xres > yres)
	    {
	        ERR_PASS_CLEANUP (w_expandCols (pIPWState, tmpPr2, 2, &tmpPr1));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));
		hReduction = 0.5;
	    }
	    else
	    {
	        ERR_PASS_CLEANUP (w_expandRows (pIPWState, tmpPr2, 2, &tmpPr1));
		ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));
		vReduction = 0.5;
	    }
	}
    }
#endif
    tmpPr2 = NULL;

    /* calculate reduced area/dimension parameters */
    minLDXreduced = (Int32)(minWidth / hReduction);
    minLDYreduced = (Int32)(minHeight / vReduction);
    minAreduced = (Int32)(minArea / (hReduction * vReduction));
    xresRed = (Int32)(xres / hReduction);
    yresRed = (Int32)(yres / vReduction);

#if 0
    /*=============> Remove those pesky rules :
      Try to get rid of horizontal and vertical rules.  Where these
      intersect circle gestures we will create gaps.  The idea of the
      following code is to close these gaps.
     */
    {
        PIXR *tmpPr3 = NULL;
	STREL *s = NULL;

	ERR_PASS_CLEANUP (w_removeHVRules (pIPWState, tmpPr1, 
				xresRed,
				3000, 
				1, 
				xresRed / 15,
				&tmpPr2));

	ERR_PASS_CLEANUP (w_makeBrickStrel (pIPWState, 
	                         MAX (1, xresRed / 13),
				 MAX (1, xresRed / 13),
	                         MAX (1, xresRed / 25),
				 MAX (1, xresRed / 25),
				 &s));

	ERR_PASS_CLEANUP (w_dilate (pIPWState, tmpPr2, s, &tmpPr3));
	ERR_PASS_CLEANUP (w_logicOp (pIPWState, tmpPr3, AND, tmpPr1));
	ERR_PASS_CLEANUP (w_destroyStrel (pIPWState, &s));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr3));

#if DEBUG_CE
        ERR_PASS_CLEANUP (w_write (pIPWState, "rules", tmpPr1, TIFF_G3));
#endif
    }

#endif

    /*=============> Preprocessing :
      Try to close small gaps and connect circle components which
      may have scanned poorly.  The total number of connected
      components is reduced as well.  Note that this operation also
      has the side effect of closing up remaining halftone/dithered 
      regions and obliterating any lines which passed through these 
      areas.  
     */
    ERR_PASS_CLEANUP (wp_cg_preProcessForCircles3 (pIPWState, tmpPr1, 
                                                   xresRed, yresRed, &tmpPr1));

#if DEBUG_CE
    ERR_PASS_CLEANUP (w_write (pIPWState, "pre", tmpPr1, TIFF_G3));
#endif


    /*=============> Remove rules */
    /* not satisfactorally implemented */

    ERR_PASS_CLEANUP (wp_cg_getCirclesAndParts (pIPWState, tmpPr1, 
                               minLDXreduced, minLDYreduced, 
			       minAreduced, &tmpPr2));
#if DEBUG_CE
    ERR_PASS_CLEANUP (w_write (pIPWState, "parts", tmpPr2, TIFF_G3));
#endif


    /*=============> Do more morphology to connect circle components */
    ERR_PASS_CLEANUP (wp_cg_postProcessForCircles3 (pIPWState, tmpPr2, &tmpPr1));
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));

#if DEBUG_CE
    ERR_PASS_CLEANUP (w_write (pIPWState, "post", tmpPr1, TIFF_G3));
#endif

    /*=============> Create a "peudo-mask" image of circle gestures */
    ERR_PASS_CLEANUP (wp_cg_makePseudoMask (pIPWState, tmpPr1, xresRed, 
					 yresRed, minAreduced, minLDXreduced,
					 minLDYreduced, &cCount, &tmpPr2));

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr1));

#if DEBUG_CE
    fprintf (stdout, "Pseudo-mask has %ld objects\n", cCount);
    ERR_PASS_CLEANUP (w_write (pIPWState, "pmask", tmpPr2, TIFF_G3));
#endif

    if (cCount > 0)
    {
        ERR_PASS_CLEANUP (wp_cg_extractCircledRegions (pIPWState, inPr, tmpPr2,
				      xresRed, yresRed,
				      hReduction, vReduction,
				      minLDXreduced, minLDYreduced,
				      minAreduced, 
				      numCirclesExtracted,
				      &imageList));
    }

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &tmpPr2));

    /*=============> Set up the result pixr array */
    if (imageList != NULL)
    {
        if (separateOutput) /* Fill in the array with sub-results */
	{
	    /* allocate the array of PIXR pointers */
	    *outPrArray = (PIXR **)IPW_CALLOC ((*numCirclesExtracted),
	                                  sizeof (PIXR *), cAllocatorData);
	    ERR_GEN_CLEANUP ((*outPrArray) == NULL, cErrNomem);

	    /* allocate the array of RECTANGLs */
	    *outRectArray = (RECTANGL *)IPW_CALLOC ((*numCirclesExtracted),
						    sizeof (RECTANGL),
						    cAllocatorData);
	    ERR_GEN_CLEANUP ((*outRectArray) == NULL, cErrNomem);
	    *outPrArrayLength = *numCirclesExtracted;

	    cc = 0;
	    subImage = imageList;
	    while (subImage != NULL)
	    {
		if (subImage->image != NULL)
		{
		    (*outPrArray)[cc] = subImage->image;
		    (*outRectArray)[cc].x = subImage->x;
		    (*outRectArray)[cc].y = subImage->y;
		    (*outRectArray)[cc].w = subImage->w;
		    (*outRectArray)[cc].h = subImage->h;
		    
		    /* make sure we don't destroy the pixr later */
		    subImage->image = NULL;
		    cc++;
		}

		subImage = subImage->next;
	    }
	}
	else /* Accumulate the sub-results into one big result image */
	{
	    /* allocate a single PIXR pointer */
	    *outPrArray = (PIXR **)IPW_CALLOC (1, sizeof (PIXR *), 
					       cAllocatorData);
	    ERR_GEN_CLEANUP ((*outPrArray) == NULL, cErrNomem);

	    /* allocate the array of RECTANGLs */
	    *outRectArray = (RECTANGL *)IPW_CALLOC ((*numCirclesExtracted),
						    sizeof (RECTANGL),
						    cAllocatorData);
	    ERR_GEN_CLEANUP ((*outRectArray) == NULL, cErrNomem);

	    /* the output PIXR is the same size as the original input */
	    ERR_PASS_CLEANUP (w_clonePixr (pIPWState, inPr, *outPrArray));
	    *outPrArrayLength = 1;

	    cc = 0;
	    subImage = imageList;
	    while (subImage != NULL)
	    {
		if (subImage->image != NULL)
		{
		    UInt8 *pOutImgAll, *pSubImgAll;
		    Int32  outBpl, subBpl;

		    /* OR in the sub-result */
		    ERR_PASS_CLEANUP (w_pixrGetImageAll (
			pIPWState, (*outPrArray[0]), &pOutImgAll));
		    ERR_PASS_CLEANUP (w_pixrGetBpl (
			pIPWState, (*outPrArray)[0], &outBpl));
		    ERR_PASS_CLEANUP (w_pixrGetImageAll (
			pIPWState, subImage->image, &pSubImgAll));
		    ERR_PASS_CLEANUP (w_pixrGetBpl (
			pIPWState, subImage->image, &subBpl));

		    ERR_PASS_CLEANUP( i_rasterOp (pIPWState, pOutImgAll,
				    cFrameBits + subImage->x, 
				    cFrameBits + subImage->y,
				    subImage->w, 
				    subImage->h, 
				    1,
				    outBpl,
				    PIX_SRC | PIX_DST,
				    pSubImgAll,
				    cFrameBits, 
				    cFrameBits, 
				    1, subBpl));

		    (*outRectArray)[cc].x = subImage->x;
		    (*outRectArray)[cc].y = subImage->y;
		    (*outRectArray)[cc].w = subImage->w;
		    (*outRectArray)[cc].h = subImage->h;
		    cc++;
		}

		subImage = subImage->next;
	    }

        }  /* make combined, single output image */

    } /* circles found? */

    /* clean up the list */
    ERR_PASS_CLEANUP (wp_cg_freeSubImageList (pIPWState, &imageList));

    NORMAL_RETURN;

ERR_CLEANUP:
    if (tmpPr1) w_destroyPixr (pIPWState, &tmpPr1);
    if (tmpPr2) w_destroyPixr (pIPWState, &tmpPr2);
    if (imageList) wp_cg_freeSubImageList (pIPWState, &imageList);
    if (*outPrArray)
    {
	for (i = 0; i < *outPrArrayLength; i++)
	{    
	    if ((*outPrArray)[i]) w_destroyPixr (pIPWState, &((*outPrArray)[i]));
	}
	IPW_FREE (*outPrArray, cAllocatorData);
	*outPrArray = NULL;
	*outPrArrayLength = 0;
    }
    if (*outRectArray)
    {
	IPW_FREE (*outRectArray, cAllocatorData);
	*outRectArray = NULL;
    }
    ERR_RETURN;

}  /* w_extractCircledRegions () */


/*

$Log: gesture1.c,v $
Revision 1.1.1.1  1999/09/20 18:19:16  jlin
Imported sources

Revision 1.40  1998/01/14 00:12:07  ddavies
gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.39  1997/06/26 20:01:05  ddavies
gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v

CommonCheck: Add inclusion of new file rast.pub.

Revision 1.38  1997/04/15 21:51:51  ddavies
ddavies on Tue Apr 15 14:51:31 PDT 1997

gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v

Use non-initialized PIXR creation where possible.

Revision 1.37  1996/09/27 16:43:16  dferrell
dferrell on Fri Sep 27 09:41:31 PDT 1996

gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v

Fixed Visual C++ compiler warnings about constant float assignments.
Added error macros around calls in debug only sections (nag).

Revision 1.36  1996/09/26 21:26:46  dferrell
dferrell on Thu Sep 26 14:25:10 PDT 1996

gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v

Completely rewrote the human vs. machine drawn discrimination function.
The new version is more robust and much less prone to false positives
caused by polygons and "angular" line art (e.g. engineering drawings,
flow charts, etc).  Since the parameters for discrimination are now more
complex, fine tuning at this step has been removed from the interfaces
(trust me).

Revision 1.35  1996/01/04 20:23:01  ejaquith
gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.34  1995/12/13  04:07:46  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.33  1995/11/15  20:10:38  dferrell
 * dferrell on Wed Nov 15 12:10:34 PST 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Fixed function comment blocks.
 *
 * Revision 1.32  1995/11/13  20:11:39  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.31  1995/11/10  02:03:49  ejaquith
 * ejaquith on Thu Nov 9 18:03:48 PST 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.30  1995/10/13  22:00:48  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.29  1995/10/03  20:53:35  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * FF_TIFF_G3 ==> TIFF_G3
 *
 * Revision 1.28  1995/10/03  01:59:32  macleod
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.27  1995/10/02  19:18:06  jcraig
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * w_invert takes a pointer to a pointer to a destination pixr.
 *
 * Revision 1.26  1995/09/29  01:49:21  ejaquith
 * ejaquith on Thu Sep 28 18:49:06 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Changes: TRUE => cTrue, FALSE => cFalse
 *
 * Revision 1.25  1995/09/21  22:53:27  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.24  1995/09/19  19:19:39  macleod
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.23  1995/09/11  21:54:35  macleod
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.22  1995/09/07  01:31:29  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.21  1995/09/06  20:19:26  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.20  1995/09/05  19:00:54  macleod
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Removed calls to GET_STREL macro, instead calling getStrel() function.
 * This required using local variables rather than invoking GET_STREL
 * as an argument to another function, since getStrel can return an error
 * code which should be checked.
 *
 * Revision 1.19  1995/09/01  21:02:51  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.18  1995/08/31  21:13:18  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.17  1995/08/30  00:39:05  dferrell
 * dferrell on Tue Aug 29 17:37:16 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Fixed logic bug where returns had been replaced with setting a parameter's
 * value.  This change allowed the original value to be overwritten (since
 * there was no immediate return) causing all candidate circles to be
 * judged hand-drawn (even line art!).
 *
 * Revision 1.16  1995/08/25  01:36:33  ejaquith
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Fixed warnings
 *
 * Revision 1.15  1995/08/17  21:55:15  macleod
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.14  1995/07/27  17:19:56  dferrell
 * dferrell on Thu Jul 27 10:19:21 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Fixed broken logic which decided whether or not to try to remove stipple.
 *
 * Revision 1.13  1995/07/26  23:57:46  dferrell
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Fixed some silly memory references which were due to not setting freed
 * pixrs pointers to NULL (Doh!).
 *
 * Revision 1.12  1995/07/13  18:55:03  dferrell
 * dferrell on Thu Jul 13 11:54:48 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Turned off all debug flags (whoops).
 *
 * Revision 1.11  1995/07/13  18:01:09  dferrell
 * dferrell on Thu Jul 13 11:00:54 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Fixed problems with abort logic/cleanup.
 *
 * Revision 1.10  1995/06/26  18:47:25  dferrell
 * dferrell on Mon Jun 26 11:44:19 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Removed obsolete function cg_expandAndMask ().
 * Removed unnecessary edgeRatio parameter from cg_makePseudoMask ()
 * interface.
 * Improved clean up operations for several error conditions.
 *
 * Revision 1.9  1995/06/23  23:29:50  dferrell
 * dferrell on Fri Jun 23 16:28:12 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Made many internal flow changes.  Added hooks and prototype code (#if 0ed
 * out) for using polygon heuristics to recognize machine drawn objects as
 * such.
 *
 * Revision 1.8  1995/04/13  18:22:29  dferrell
 * dferrell on Thu Apr 13 11:20:54 PDT 1995
 *
 * gesture1.c -> /product/ipcore/ipwrap/src/RCS/gesture1.c,v
 *
 * Added new private function cg_expandAndMask () which handles all
 * cases of expand and mask functionality in circle gesture extraction.
 * Cleaned up some of the pixr seeding and seed filling calls.
 *

*/
