/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 * gfio.c - generic file I/O functions for use on platforms with
 *          standard C library accessible file systems (UNIX, DOS).
 *
$Log: gfio.c,v $
Revision 1.2  2000/02/29 00:22:54  jlin
Add strong arm related changes

 * Revision 1.1.1.1  1999/09/20  18:19:16  jlin
 * Imported sources
 *
Revision 1.7  1995/09/01 02:23:18  ejaquith
gfio.c -> /product/ipcore/ipwrap/src/RCS/gfio.c,v

CommonCheck: Put minimum number of include files back in.

 * Revision 1.6  1995/08/31  21:13:21  ejaquith
 * gfio.c -> /product/ipcore/ipwrap/src/RCS/gfio.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.5  1995/08/17  22:04:55  ejaquith
 * gfio.c -> /product/ipcore/ipwrap/src/RCS/gfio.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.4  1995/04/01  01:40:55  ddavies
 * ddavies on Fri Mar 31 17:39:28 PST 1995
 *
 * gfio.c -> /product/ipcore/ipwrap/src/RCS/gfio.c,v
 *
 * Use tmpnam and malloc to create a new file name string instead of using
 * the non-ANSI strdup and tmpnam.  This created a warning on gcc/Solaris.
 *
 * Revision 1.3  1995/03/01  22:17:37  ejaquith
 * gfio.c -> /product/ipcore/ipwrap/src/RCS/gfio.c,v
 *
 * Fixed bug where GFIO_DestroyFileToken() was freeing the filetoken
 * twice.  It now (correctly) frees the token's file*name* once and then
 * the token.
 *
 * Revision 1.2  1994/12/08  03:08:21  dferrell
 * dferrell on Wed Dec 7 19:07:51 PST 1994
 *
 * gfio.c -> /product/ipcore/ipwrap/src/RCS/gfio.c,v
 *
 * Added typecaste so that MPW could compile it.
 *
 * Revision 1.1  1994/10/21  17:18:17  ejaquith
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* prototypes */
#include "gfio.h"
#include "gfiopro.h"

#ifdef __WATCOMC__
#endif


/* ejaquith -- added ansi include file to get prototypes */
#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* #define GFIO_DEBUG    ejaquith -- commented this out*/

/* --ejaquith-- added this because not in stdio.h for some reason?! */
int fflush ( FILE *fp );
   

typedef struct 
{
   char *filename;
   FILE *handle;
   int   open;   /* file open flag */
} Token;

/*************************************************************
 * GFIO_CreateFileToken () - Create (malloc) and init a file token
 *             for a file given its name and FILE.
 * Inputs:
 *             file     - An already opened stream handle
 *                        or NULL.
 *             filename - A string containing the file name and
 *                        optionally its path.
 *
 * Returns:
 *             A pointer to the file token for the new file, or
 *             NULL if the file token could not be created
 *
 * Notes:
 *             The platform specific information for files
 *             is stored in the file tokens which are passed
 *             in to GFIO.  If file is NULL, the file open
 *             flag is set to FALSE, otherwise it is TRUE.
 *
 * See Also: 
 *             GFIO_DestroyFileToken ()
 */
void * CDECL GFIO_CreateFileToken (char *filename, FILE *file)
{
   Token* ft;

   if (filename[0] == 0)
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_CreateFileToken: Invalid filename\n");
#endif
      return NULL;
   }

   /* allocate space for a token structure */
   ft= (Token*)malloc(sizeof(Token));
   if (ft==NULL) 
       return NULL;

    ft->filename = (char *)malloc (strlen (filename) + 1);
    if (ft->filename == NULL)
    {
	free (ft);
	return NULL;
    }

   /* fill in the token */
   strncpy (ft->filename, filename, strlen (filename) + 1);
   ft->handle = file;

   if (file == NULL)
       ft->open = 0;
   else
       ft->open = 1;

   return (void*)ft;

}  /* GFIO_CreateFileToken () */


/*************************************************************
 * GFIO_DestroyFileToken () - free resources taken by 
 *             a file token.
 *             
 * Inputs:
 *             void * ft - a file token
 *
 * Returns:
 *
 *             Nothing
 *
 * See Also: 
 *             GFIO_CreateFileToken ()
 */
void CDECL GFIO_DestroyFileToken (void *filetoken)
{
   Token* ft = filetoken;

   if (ft != NULL) 
   {
        if (ft->filename != NULL)
	    free (ft->filename);
        free(filetoken);
   }

}  /* GFIO_DestroyFileToken () */


/*************************************************************
 * GFIO_CreateTempFile () - Create a temporary file.
 *             
 * Inputs:
 *             None
 *
 * Returns:
 *             A pointer to the file token for the new file, or
 *             NULL if the file could not be opened.
 *
 * Notes:
 *             The platform specific information for files
 *             is stored in the file tokens which are passed
 *             in to GFIO.
 *
 * See Also: 
 *             GFIO_DestroyTempFile ()
 */
void* CDECL GFIO_CreateTempFile (void)
{
#ifndef arm
   Token *ft;
   /* FILE  *handle; --ejaquith-- commented this out */

   /* allocate space for a token structure */
   ft = (Token*)malloc (sizeof (Token));
   if (ft == NULL) 
       return NULL;

   /* Avoid use of strdup since it isn't strict ANSI.  Have tmpnam
    * put the new file name directly into ft->filename.
    */
   ft->filename = (char *)malloc(L_tmpnam);
   if (ft->filename == NULL) 
   {
      free(ft);
      return NULL;
   }
   tmpnam(ft->filename);	/* create new file name */
   ft->handle = NULL;
   ft->open = 0;

   return (ft);
#else
   return NULL; /*XXX - implement later if needed */
#endif


}  /* GFIO_CreateTempFile () */


/*************************************************************
 * GFIO_DestroyTempFile () - Remove a temp file from the file 
 *             system.
 *             
 * Inputs:
 *             ft - the file token associated with the file to
 *                  be removed.
 *
 * Returns:
 *             Nothing
 *
 * See Also: 
 *             GFIO_CreateTempFile ()
 */
void CDECL GFIO_DestroyTempFile(void* ft)
{
   Token* ftoken=ft;

   if (ftoken == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_DestroyTempFile: Invalid NULL file token\n");
#endif
      return;
   }

   /* close the file if it was left open */
   if (ftoken->handle) 
       GFIO_CloseFile(ft);

   /* free the file token structure */
   if (ftoken->filename) 
   {
      remove(ftoken->filename);
      free(ftoken->filename);
   }

   free(ftoken);

}  /* GFIO_DestroyTempFile () */


/*************************************************************
 * GFIO_OpenFile () - Open a file.
 *             
 * Inputs:
 *             ft   - the file token associated with the file to
 *                    be removed.
 *             mode - GFIO_READ, GFIO_WRITE, or both (ORing)
 *
 * Returns:
 *             GFIO_SUCCESS - file opened correctly or was
 *                            already open
 *             GFIO_FAILURE - error occured while opening the
 *                            file
 *
 * Notes:
 *             The current position in the file is 0 upon return.
 * See Also: 
 *             GFIO_CloseFile ()
 */
int CDECL GFIO_OpenFile (void* ft, int mode)
{
   Token* ftoken=ft;

   /* check parameters */
   if (ftoken==NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_OpenFile: Invalid NULL file token\n");
#endif
      return GFIO_FAILURE;
   }
   if (ftoken->handle) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_OpenFile: File already open\n");
#endif
      return GFIO_FAILURE;
   }

   if (ftoken->handle != NULL)
       return (GFIO_FAILURE);

   switch (mode) 
   {
        case GFIO_READ:
#ifdef __WATCOMC__
            ftoken->handle = fopen(ftoken->filename,"rb");
#else
            ftoken->handle = fopen(ftoken->filename,"r");
#endif
            break;

        case GFIO_WRITE:
#ifdef __WATCOMC__
            ftoken->handle = fopen(ftoken->filename,"wb");
#else
            ftoken->handle = fopen(ftoken->filename,"w");
#endif
            break;

        case GFIO_READ|GFIO_WRITE:
#ifdef __WATCOMC__
            ftoken->handle = fopen(ftoken->filename,"r+b");
#else
            ftoken->handle = fopen(ftoken->filename,"r+");
#endif
	    /* if the file does not exist, create it */
	    if (ftoken->handle == NULL)
#ifdef __WATCOMC__
                ftoken->handle = fopen(ftoken->filename,"w+b");
#else
                ftoken->handle = fopen(ftoken->filename,"w+");
#endif
            break;

       default:
#ifdef GFIO_DEBUG  /* ejaquith --added this macro wraper */
            fprintf(stderr,"GFIO_OpenFILE: Invalid mode (%d)\n",mode);
#endif
            return GFIO_FAILURE;
   }

   if (ftoken->handle == NULL) 
   {
      return GFIO_FAILURE;
   } 
   else 
   {
      ftoken->open = 1;
      return GFIO_SUCCESS;
   }

}  /* GFIO_OpenFile () */


/*************************************************************
 * GFIO_CloseFile () - Close a file.
 *             
 * Inputs:
 *             ft   - the file token associated with the file to
 *                    be removed.
 *             
 * Returns:
 *             GFIO_SUCCESS - file closed properly
 *             GFIO_Failure - error closing file
 *
 * See Also: 
 *             GFIO_OpenFile ()
 */
int CDECL GFIO_CloseFile (void *ft)
{
   Token* ftoken=ft;

   /* check parameters */
   if (ftoken == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_CloseFile: Invalid NULL file token\n");
#endif
      return GFIO_FAILURE;
   }
   if (ftoken->handle == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_CloseFile: File not open\n");
#endif
      return GFIO_SUCCESS;
   }

   if (fclose(ftoken->handle) == 0) 
   {
      ftoken->handle=NULL;
      ftoken->open = 0;
      return GFIO_SUCCESS;
   }
   else 
      return GFIO_FAILURE;

}  /* GFIO_CloseFile () */


/*************************************************************
 * GFIO_Read () - Read from a file to a buffer.
 *             
 * Inputs:
 *             ft        - the file token associated with the file
 *                         from which to read
 *             buffer    - array to read bytes into, must be at
 *                         least byteCount characters in length
 *             byteCount - number of bytes to read
 *
 * Returns:
 *             number of bytes actually read into the buffer, or
 *             GFIO_EOF - attempted read at EOF
 *             GFIO_FAILURE - other error occured (file not open)
 *
 * Notes:
 *             Reads data starting at the current position in the
 *             file and increments the position by byteCount bytes.
 *             If during the read EOF is encountered, returns the
 *             number of bytes read up to that point.  I already at
 *             EOF when read specified, returns GFIO_EOF.
 *
 * See Also: 
 *             GFIO_Write ()
 */
long int CDECL GFIO_Read (void *ft, char *buffer, unsigned long byteCount)
{
   Token* ftoken=ft;
   long status;

   if (ftoken == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_Read: Invalid NULL file token\n");
#endif
      return GFIO_FAILURE;
   }
   if (buffer == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_Read: Invalid NULL buffer pointer\n");
#endif
      return GFIO_FAILURE;
   }

   if (ftoken->handle == NULL) 
       return GFIO_FAILURE;

   status=fread(buffer, 1, byteCount,ftoken->handle);

   if (status==0) 
   {
      if (feof(ftoken->handle)) 
      {
         return GFIO_EOF;
      } 
      else 
      {
         return GFIO_FAILURE;
      }
   } 
   else 
      return status;

}  /* GFIO_Read () */


/*************************************************************
 * GFIO_Write () - Write a buffer of bytes to a file.
 *             
 * Inputs:
 *             ft        - the file token associated with the file
 *                         in which to write
 *             buffer    - array of characters to be written, must
 *                         be at least byteCount characters in length
 *             byteCount - number of bytes to write
 *             
 * Returns:
 *             number of characters written into the file, or
 *             GFIO_FAILURE - the file is not open, not
 *                            writeable, or an error occured.
 *
 * Notes:
 *             Writes data starting at the current position in the
 *             file and increments the position by byteCount bytes.
 *
 * See Also: 
 *             
 */
long int CDECL GFIO_Write (void *ft, char *buffer, unsigned long byteCount)
{
   Token* ftoken=ft;
   long status;

   if (ftoken == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_Write: Invalid NULL file token\n");
#endif
      return GFIO_FAILURE;
   }
   if (buffer == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf(stderr,"GFIO_Write: Invalid NULL buffer pointer\n");
#endif
      return GFIO_FAILURE;
   }

   if (ftoken->handle == NULL) 
      return GFIO_FAILURE;

   status = fwrite(buffer, 1, byteCount,ftoken->handle);
   if (status == 0) 
       return GFIO_FAILURE;

   /* Watcom C requires this for correct operation when updating (r+b, w+b)
    * a file.  It also seems sensible to be careful on UNIX, so no ifdefs.
    */
   if (fflush (ftoken->handle) != 0) 
       return GFIO_FAILURE;

   return status;

}  /* GFIO_Write () */


/*************************************************************
 * GFIO_Seek () - Seek to an arbitrary position in a file.
 *             
 * Inputs:
 *             ft     - the file token associated with the file
 *                      in which to seek
 *             offset - number of bytes from the beginning of
 *                      the file to set the current position
 *
 * Returns:
 *             how far the current position offset was actually
 *             set from the beginning of the file, or
 *             GFIO_EOF - attempt to seek past EOF
 *             GFIO_FAILURE - another error occured
 *
 * Notes:
 *             If the file is not open, returns GFIO_FAILURE.
 *
 * See Also: 
 *             
 */
long int CDECL GFIO_Seek (void* ft, unsigned long offset)
{
   Token* ftoken=ft;
   long status;

   if (ftoken == NULL) 
   {
#ifdef GFIO_DEBUG
      fprintf (stderr,"GFIO_Seek: Invalid NULL file token\n");
#endif
      return GFIO_FAILURE;
   }

   /* make sure the file is open */
   if (ftoken->handle == NULL) 
       return GFIO_FAILURE;

   status = fseek (ftoken->handle, offset, SEEK_SET);

   if (status < 0) 
   {
      return GFIO_FAILURE;
   } 
   else 
   {
      return ftell(ftoken->handle);
   }

}  /* GFIO_Seek () */


/**************************************************************
 * GFIO_FileSize () - Find out the number of bytes in a file
 *
 * Inputs:
 *             ft     - the file token associated with the file
 *                      in which to get the size
 *
 * Returns:
 *             the size of the file in bytes or GFIO_FAILURE
 *
 */
long int CDECL 
GFIO_FileSize (void *ft)
{
    long original_offset;
    long length;
    Token *ftoken = ft;

    if (ft == NULL)
        return (GFIO_FAILURE);

    if (ftoken->handle == NULL)
        return (GFIO_FAILURE);

    original_offset = ftell (ftoken->handle);
    if (original_offset < 0)
        return (GFIO_FAILURE);

    fseek (ftoken->handle, 0, SEEK_END);
        length = ftell (ftoken->handle);

    fseek (ftoken->handle, original_offset, SEEK_SET);

    return (length);

}  /* GFIO_FileSize () */

