/*<HTML><HEAD><TITLE>imginfo.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/



/* 
 * imginfo.c
 *
 *	Heuristics on images.
 *
 * Functions defined:
 *
 *	orient ()   -	    Figure out which way (0, 90, 180, 270) text is oriented
 *	maskIsRect ()-	    Tests whether the mask surrounded by the bounding
 *			    box passed is approximately a rectangle
 *
 *	textMask () -	    Makes a mask for the regions of a binary
 *			    image which appear to be text-like.
 *	labelColorConnectedComponents ()-   - compute a color for each connected component
 *      coalesceColorCCs () Coalesce the color connected components into
 *			    classes of equivalent colors.
 *	computeInkPlane ()  Given color connected components, make a
 *			    quadtree representing the inks needed to render
 *			    the set of color connected components.
 */


#include <math.h>
#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "logicops.h"
#include "morphg.h"
#include "alplib.h"
#include "defines.pub"
#include "shrscgr.pub"
#include "shrcc.pub"
#include "imginfo.prv"
#include "frames.pub"
#include "imageref.pub"

    /* prototypes */
#include "props.pub"
#include "rast.pub"
#include "strel.pub"
#include "fthresh.pub"
#include "fill.pub"
#include "runhist.pub"
#include "geoutil.pub"
#include "clean.pub"
#include "logicops.pub"
#include "color.pub"
#include "morph.pub"
#include "reduce.pub"
#include "intops.pub"
#include "runfilt.pub"
#include "imginfo.pub"

/*#define DEBUG*/
#define cEdgeMargin 2

IP_RCSINFO(RCSInfo, "$RCSfile: imginfo.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

/*****************************************************************************
 * BeginDocumentation::: "w_orient"
 *
 * OneLineDescription:::
 *  Determine whether a page is oriented vertically or horizontally
 *
 * LongDescription:::
 *	Tries to figure out orientation of an image. This will only
 *	work on images with a fair amount of text and not much halftoned 
 *	stuff, so segment the image before you call it.
 *	The theory this code tests is that there are more vertical strokes than
 *	horizontal strokes in English text.  This is used to distinguish 
 *      landscape from portrait orientation. 
 *
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::     State structure   
 *  prSrc::         Image to check, 1 bit/pixel
 *
 * Output:::
 *  pAngle::        Angle of image, 0 or 90, where 0 means that the
 *                  text goes from left to right, and 90 means that 
 *                  the text goes from top to bottom.
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *	THIS VERSION ONLY DOES V/H ORIENT. Use orient4 if you want all
 *	four directions.
 *
 * FunctionCategory:::
 *  orientation
 *
 * Keywords:::
 *  orientation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod:: From some test code by ddavies
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_orient"></A>
 *****************************************************************************/
ErrCode CDECL
w_orient (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32              *pAngle)
{
    PROLOGUE;
    PIXR                *prReduced = NULL, *prTemp = NULL, *prEnd = NULL;
    Int32		 numPixNorth, numPixEast;
#if 0
    UInt32		 numPixSouth, numPixWest; 
    UInt32		 width, height;
#endif
    Int32		 numPix;
    UInt32		 levelNorth, levelEast, isVertical;
    ScaledInt1616        xDPI, yDPI;
    Int32		 retVal = 0, dpi;
    STREL               *pStrel5 = NULL, *pStrel3h = NULL, *pStrel3v = NULL,
	                *pStrel5h = NULL, *pStrel5v = NULL, *pStrel7h = NULL,
	                *pStrel7v = NULL, *pStrel9h = NULL, *pStrel9v = NULL;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1 | cChannels3,
		  cFrame32);
    ERR_PASS (w_initStrelByName (pIPWState,  strel5));
    ERR_PASS (w_getStrel (pIPWState, strel5, &pStrel5));
    ERR_PASS (w_initStrelByName (pIPWState,  strel3h));
    ERR_PASS (w_getStrel (pIPWState, strel3h, &pStrel3h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel3v));
    ERR_PASS (w_getStrel (pIPWState, strel3v, &pStrel3v));
    ERR_PASS (w_initStrelByName (pIPWState,  strel5h));
    ERR_PASS (w_getStrel (pIPWState, strel5h, &pStrel5h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel5v));
    ERR_PASS (w_getStrel (pIPWState, strel5v, &pStrel5v));
    ERR_PASS (w_initStrelByName (pIPWState,  strel7h));
    ERR_PASS (w_getStrel (pIPWState, strel7h, &pStrel7h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel7v));
    ERR_PASS (w_getStrel (pIPWState, strel7v, &pStrel7v));
    ERR_PASS (w_initStrelByName (pIPWState,  strel9h));
    ERR_PASS (w_getStrel (pIPWState, strel9h, &pStrel9h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel9v));
    ERR_PASS (w_getStrel (pIPWState, strel9v, &pStrel9v));

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    /* make the mask: reduction 2x, thresh 2 */
    if (dpi > 100)
    {
	/* We reduce for speed. Don't do it at lower resolutions,
           because there's not really enough information */
	ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 2, 0, 0, 0, &prReduced));
    }
    else
    {
        ERR_PASS_CLEANUP (w_copy (pIPWState, prSrc, &prReduced));
    }

    ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prReduced, &prTemp));
    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel5, &prTemp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, SUBTRACT, prReduced));

    /* see whether closing horizontally has more effect that closing
     * vertically.  Since intercharacter spaces are smaller than
     * inter-line spaces (almost always). This should normally
     * work.
     */
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prReduced, &numPix));
#ifdef DEBUG
    fprintf(stderr, "The number of pixels is %ld\n", numPix);
#endif
    IP_YIELD;
	/* try the horizontal close */
    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel3h, &prTemp));
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
    levelNorth = 1;
    if ( numPixNorth < 1.2 * numPix)
    {
	ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel5h, &prTemp));
	ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
	levelNorth = 2;
	if (numPixNorth < 1.2 * numPix)
	{
	    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, 
				      pStrel7h, &prTemp));
	    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
	    levelNorth = 3;
	    if (numPixNorth < 1.2 * numPix)
	    {
		ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, 
					  pStrel9h, &prTemp));
		ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
		levelNorth = 4;
	    }
	}
    }

#ifdef DEBUG
    fprintf(stderr, "The number of closes for north or south: %ld\n",
	    levelNorth);
    fprintf(stderr, "The number of pixels indicating up is north or south: %ld\n",
	    numPixNorth - numPix);
#endif
    IP_YIELD;
	/* try the vertical close */
    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel3v, &prTemp));
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
    levelEast = 1;
    if (numPixEast < 1.2 * numPix)
    {
	ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel5v, &prTemp));
	ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
	levelEast = 2;
	if (numPixEast < 1.2 * numPix)
	{
	    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel7v,
				       &prTemp));
	    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
	    levelEast = 3;
	    if (numPixEast < 1.2 * numPix)
	    {
		ERR_PASS_CLEANUP (w_close (pIPWState, prReduced,
					  pStrel9v, &prTemp));
		ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
		levelEast = 4;
	    }
	}
    }

    IP_YIELD;
#ifdef DEBUG
    fprintf(stderr, "The number of closes for east or west: %ld\n",
	    levelEast);
    fprintf(stderr, "The number of pixels indicating up is east or west: %ld\n",
	    numPixEast - numPix);
#endif
    /*
     * Figure out whether it's vertical or horizontal by the # of closes.
     * If that doesn't work (i.e. they were equal) take the max. # of pixels
     */
    if (levelNorth < levelEast)
    	isVertical = 1;
    else if (levelNorth > levelEast)
	isVertical = 0;
    else
	isVertical = (numPixEast - numPix < numPixNorth - numPix);

    retVal = isVertical ? 0 : 90;

    
    if (prEnd) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prEnd));
    if (prTemp) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTemp));
    if (prReduced) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));

    *pAngle = retVal;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prEnd) w_destroyPixr (pIPWState, &prEnd);
    if (prTemp) w_destroyPixr (pIPWState, &prTemp);
    if (prReduced) w_destroyPixr (pIPWState, &prReduced);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation::: "w_orient4"
 *
 * OneLineDescription:::
 *  Determine whether a page is oriented with "up" N, S, W, or E.
 *
 * LongDescription:::
 *  Tries to figure out orientation of an image. This will only
 *  work on images with a fair amount of text and not much halftoned 
 *  stuff, so segment the image before you call it.
 *  <BR>
 *  The theory this code tests is that there are more vertical strokes than
 *  horizontal strokes in English text.  This is used to distinguish landscape
 *  from portrait orientation.  Furthermore, vertical strokes tend to be on
 *  the left, so we can use the difference between the number of left edges
 *  and right edges to decide whether "up" is North, South, East or West in
 *  the current image.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::     State structure   
 *  prSrc::         Image to check
 *
 * Output:::
 *  pAngle::        Angle of image: 0, 90, 180, or 270, where 0 means that the
 *                  text goes from left to right, 90 means that the text
 *                  goes from top to bottom, 180 means the text goes from
 *                  right to left (i.e. upside down) and 270 means
 *                  that the text goes from bottom to top.
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  orientation
 *
 * Keywords:::
 *  orientation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod:: from some test code by ddavies
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_orient4"></A>
 *****************************************************************************/
ErrCode CDECL
w_orient4 (
    IPWState           *pIPWState,
    PIXR               *prSrc,
    Int32              *pAngle)
{
    PROLOGUE;
    PIXR                *prReduced = NULL, *prTemp = NULL, *prEnd = NULL;
    Int32		 numPixNorth, numPixEast;
    Int32		 numPixSouth, numPixWest; 
    Int32		 width, height;
    Int32		 numPix;
    UInt32		 levelNorth, levelEast, isVertical;
    ScaledInt1616        xDPI, yDPI;
    Int32		 retVal = 0, dpi;
    STREL               *pStrel5 = NULL, *pStrel3h = NULL, *pStrel3v = NULL,
	                *pStrel5h = NULL, *pStrel5v = NULL, *pStrel7h = NULL,
	                *pStrel7v = NULL, *pStrel9h = NULL, *pStrel9v = NULL;

    ERR_GEN_PIXR (prSrc, cByteOrderBig, cDepth1, cChannels1 | cChannels3,
		  cFrame32);

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prSrc, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));

    ERR_PASS (w_initStrelByName (pIPWState,  strel5));
    ERR_PASS (w_getStrel (pIPWState, strel5, &pStrel5));
    ERR_PASS (w_initStrelByName (pIPWState,  strel3h));
    ERR_PASS (w_getStrel (pIPWState, strel3h, &pStrel3h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel3v));
    ERR_PASS (w_getStrel (pIPWState, strel3v, &pStrel3v));
    ERR_PASS (w_initStrelByName (pIPWState,  strel5h));
    ERR_PASS (w_getStrel (pIPWState, strel5h, &pStrel5h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel5v));
    ERR_PASS (w_getStrel (pIPWState, strel5v, &pStrel5v));
    ERR_PASS (w_initStrelByName (pIPWState,  strel7h));
    ERR_PASS (w_getStrel (pIPWState, strel7h, &pStrel7h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel7v));
    ERR_PASS (w_getStrel (pIPWState, strel7v, &pStrel7v));
    ERR_PASS (w_initStrelByName (pIPWState,  strel9h));
    ERR_PASS (w_getStrel (pIPWState, strel9h, &pStrel9h));
    ERR_PASS (w_initStrelByName (pIPWState,  strel9v));
    ERR_PASS (w_getStrel (pIPWState, strel9v, &pStrel9v));

    /* make the mask: reduction 2x, thresh 2 */

    if (dpi > 100)
    {
	ERR_PASS_CLEANUP (w_reduce(pIPWState, prSrc, 2, 0, 0, 0, &prReduced));
	    }
    else
    {
	/* Yikes--don't reduce at <= 100dpi, there's not a lot of info. */
	ERR_PASS_CLEANUP (w_copy(pIPWState, prSrc, &prReduced));
    }

    ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prReduced, &prTemp));

    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel5, &prTemp));
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prTemp, SUBTRACT, prReduced));

    /* see whether closing horizontally has more effect that closing
     * vertically.  Since intercharacter spaces are smaller than
     * inter-line spaces (almost always). This should normally
     * work.
     */
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prReduced, &numPix));
#ifdef DEBUG
    fprintf(stderr, "The number of pixels is %ld\n", numPix);
#endif
    IP_YIELD;

	/* try the horizontal close */
    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel3h, &prTemp));
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
    levelNorth = 1;

    if (numPixNorth < 1.2 * numPix)
    {
	ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel5h, &prTemp));
	ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
	levelNorth = 2;
	if (numPixNorth < 1.2 * numPix)
	{
	    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel7h,
				       &prTemp));
	    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
	    levelNorth = 3;
	    if (numPixNorth < 1.2 * numPix)
	    {
		ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, 
					   pStrel9h, &prTemp));
		ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixNorth));
		levelNorth = 4;
	    }
	}
    }

#ifdef DEBUG
    fprintf(stderr, "The number of closes for north or south: %ld\n",
	    levelNorth);
    fprintf(stderr, "The number of pixels indicating up is north or south: %ld\n",
	    numPixNorth - numPix);
#endif
    IP_YIELD;
	/* try the vertical close */
    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel3v, &prTemp));

    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
    levelEast = 1;
    if (numPixEast < 1.2 * numPix)
    {
	ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, pStrel5v, &prTemp));
	ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
	levelEast = 2;
	if (numPixEast < 1.2 * numPix)
	{
	    ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, 
				      pStrel7v, &prTemp));
	    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
	    levelEast = 3;
	    if (numPixEast < 1.2 * numPix)
	    {
		ERR_PASS_CLEANUP (w_close (pIPWState, prReduced, 
					  pStrel9v, &prTemp));
		ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixEast));
		levelEast = 4;
	    }
	}
    }

    IP_YIELD;
#ifdef DEBUG
    fprintf(stderr, "The number of closes for east or west: %ld\n",
	    levelEast);
    fprintf(stderr, "The number of pixels indicating up is east or west: %ld\n",
	    numPixEast - numPix);
#endif
    /*
     * Figure out whether it's vertical or horizontal by the # of closes.
     * If that doesn't work (i.e. they were equal) take the max. # of pixels
     */
    if (levelNorth < levelEast)
    	isVertical = 1;
    else if (levelNorth > levelEast)
	isVertical = 0;
    else
	isVertical = (numPixEast - numPix < numPixNorth - numPix);

    /* Let's see if we can us another way to guess orientation.
     * The only useful, common feature of upper case text,
     * lower case text and numerals I've been able to identify
     * is that horizontal strokes are more likely to be connected
     * to vertical-ish stuff on the left than on the right.
     * We find the horizontal strokes for north/south images
     * and vertical strokes for east/west images.  We isolate the
     * last couple of pixels on both ends.  We shift these up and down
     * (or left and right for the east/west case), AND then with
     * the original image an see how many pixels we get.
     */
    ERR_PASS_CLEANUP (w_clonePixrNoInit(pIPWState, prReduced, &prEnd));

    ERR_PASS_CLEANUP (w_pixrGetWidth(pIPWState, prReduced, &width));
    ERR_PASS_CLEANUP (w_pixrGetHeight(pIPWState, prReduced, &height));
    IP_YIELD;
    /* shift image left, look for pixels indicating up is north.
     * This means the pixels on the left of horizontal lines are
     * connected to stuff above or below them.
     */
	/* get horizontal stuff */
    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel5h, &prTemp));

	/* make it intersect other vertical stuff */
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prTemp, pStrel5v, &prEnd));

	/* keep this around for the right end */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prEnd, COPY, prTemp));
						
    IP_YIELD;

    /* get left end of horizontal stuff.  NOTE, it's not safe to
     * use prEnd and both source and dest since each line is
     * processed from left to right so we would be writing over the
     * source bits before using them.
     */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState,
				  0, 0, prTemp,
				  PIX_NOT(PIX_SRC) & PIX_DST,
				  2, 0, width-2, height, prEnd));

        /* select pixels that touch the end vertically */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReduced, AND, prEnd));
						 
    IP_YIELD;
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prEnd, &numPixNorth));
#ifdef DEBUG
    fprintf(stderr, "The number of pixels indicating up is north: %ld\n",
	    numPixNorth);
#endif
	/* get the right end of horizontal stuff */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 2, 0, prTemp,
				  PIX_NOT(PIX_SRC) & PIX_DST,
				  0, 0, width-2, height, prTemp));
    
	/* select pixels that touch the end vertically */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReduced, AND, prTemp));
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixSouth));
#ifdef DEBUG
    fprintf(stderr, "The number of pixels indicating up is south: %ld\n",
	    numPixSouth);
#endif


	/* Do the same thing for east/west, east first.  When up is east,
	 * there should be more pixels on the upper end of vertical
	 * strokes.
	 */
    /* get vertical stuff */
    ERR_PASS_CLEANUP (w_open (pIPWState, prReduced, pStrel5v, &prTemp));

    /* make it intersect other horizontal stuff */    
    ERR_PASS_CLEANUP (w_dilate (pIPWState, prTemp, pStrel5h, &prEnd));

    /* keep this around for the bottom end */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prEnd, COPY, prTemp));
						 
    IP_YIELD;
    /* get upper end of vertical stuff.  NOTE, we cannot use prEnd
     * as both the source and the dest since the image is processed
     * from top to bottom and would be writing over itself before
     * it was read as source.
     */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				  0, 0, prTemp,
				  PIX_NOT(PIX_SRC) & PIX_DST,
				  0, 2, width, height-2, prEnd));

    /* select pixels that touch the end vertically */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReduced, AND, prEnd));
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prEnd, &numPixEast));
#ifdef DEBUG
    fprintf(stderr, "The number of pixels indicating up is east: %ld\n",
	    numPixEast);
#endif
    IP_YIELD;

	/* get the lower end of vertical stuff */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState,
				  0, 2, prTemp,
				  PIX_NOT(PIX_SRC) & PIX_DST, 
				  0, 0, width, height-2, prTemp));

    /* select pixels that touch the end vertically */
    ERR_PASS_CLEANUP (w_logicOp (pIPWState, prReduced, AND, prTemp));
    ERR_PASS_CLEANUP (w_numberBits(pIPWState, prTemp, &numPixWest));
#ifdef DEBUG
    fprintf(stderr, "The number of pixels indicating up is west: %ld\n",
	    numPixWest);
#endif

    if (isVertical)
    {
	/* Determine whether it's north or south */
	retVal = numPixNorth > numPixSouth ? 0 : 180;
    }
    else
    {
	/* Determine whether it's east or west */
	retVal = numPixEast > numPixWest ? 90 : 270;
    }

    if (prEnd) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prEnd));
    if (prTemp) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTemp));
    if (prReduced) ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prReduced));

    *pAngle = retVal;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prEnd) w_destroyPixr (pIPWState, &prEnd);
    if (prTemp) w_destroyPixr (pIPWState, &prTemp);
    if (prReduced) w_destroyPixr (pIPWState, &prReduced);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::: "w_maskIsRect"
 *
 * OneLineDescription:::
 *  Figure out whether a region of an image mask is rectangular
 *
 * LongDescription:::
 *	The idea behind this function is that when images are segmented, 
 *	approximately rectangular masks can have regions that are "eaten"
 *	into the edge of the mask, such as when a light part of a picture
 *	touches the border of the mask. Here's a (lame) picture of a mask:
 *  
 *	................................
 *	................................
 *	................................
 *	................................
 *	................................
 *	................................
 *	...   ..........................
 *	.     ..........................
 * -->	   .............................
 *	................................
 *	................................
 *	................................
 *	................................
 *	................................
 *	................................
 *	................................
 *
 *  We can see the region pointed to is not part of the mask, but we 
 *  might like it to be.
 * 
 * PrivateDescription:::
 *  Our strategy is quite simple: count the pixels along
 *  the edge of the bounding box. If that number is "close" to the number
 *  of pixels that one would get if the box was completely filled, then
 *  we return true. The "close" is, obviously, a "heuristic" or arbitrary
 *  constant. We assume that the bounding box is "correct" in that it fully
 *  encloses the mask region. This assumption will be true if the bounding
 *  box was generated by w_segmentPage(). Note that this routine won't
 *  handle the case where mask pixels go out beyond the rectangle, 
 *  moving the bounding box out.
 *
 *  --- Well, that didn't work very well, though just counting the
 *  pixels and comparing against the bounding box w * h seems to work ok.
 *  I guess I didn't really need to write a routine for that...
 *
 *  --- Pixel counting is "ok", but sometimes fails when an image is
 *  significantly eaten away inside, but still mostly rectangular.
 *  Another approach would be to count the number of pixels near the edge,
 *  i.e. count the pixels in a strip around the edge, or maybe do a dilate
 *  of that size and then count the pixels on the edge. (Which is what
 *  we ended up doing)
 *
 *  --- New! If image mask is frame, it will be rect and the inside text 
 *  will be deleted. So we check whether there is text in center part.
 *  8-3-98. It created bug in halftone.c, when make image is inside
 *  text mask image, the center is wrong match. So we add xMaskOffse
 *  yMaskOffset to get correct match between mask and text in center part.
 *
 * Input:::
 *  pIPWState::      State structure
 *  box::            Bounding box inside mask image which to look 
 *  prMask::         Mask image 
 *  prText::         Text mask image, if NULL, don't do text check
 *  xMaskOffset::    Mask image x coordinate in text mask image
 *  yMaskOffset::    Mask image y coordinate in text mask image
 *  cutoff::         Percentage of edge pixels above which to call image
 *                   rectangular. 80-90% seems to work fairly well.
 *
 * Output:::
 *  pIsRect::        If cTrue, it's rectangular
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image classification
 *
 * Keywords:::
 *  image classification
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_maskIsRect"></A>
 *****************************************************************************/
ErrCode CDECL
w_maskIsRect (
    IPWState           *pIPWState,
    RECTANGL           *box,
    PIXR               *prMask,
    Int32               xMaskOffset,
    Int32               yMaskOffset,
    PIXR               *prText,
    Int32               cutoff,
    Bool               *pIsRect)
{
    PROLOGUE;
    Int32	percent, bw, bh, count, x, y, w, h, edgeCount;
    RECTANGL	r, cen;
    Int32       h1, w1;

    ERR_GEN_PIXR (prMask, cByteOrderBig, cDepth1, cChannels1 | cChannels3,
		  cFrame32);

    if (!box)
    {
	/* Default to the whole thing */
	r.x = 0;
	r.y = 0;
	ERR_PASS (w_pixrGetWidth (pIPWState, prMask, &(r.w)));
	ERR_PASS (w_pixrGetHeight (pIPWState, prMask, &(r.h)));
	box = &r;
    }
    

    /* Another Scheme: copy out 10% of the edges of the box, dilate them,
       and count # pixels on the edge */
    x = box->x; y = box->y; w = box->w; h = box->h;
    /* STRELs should be 10% */
    bw = MAX(w / 10, 1);
    bh = MAX(h / 10, 1);

    /* Top */
    ERR_PASS_CLEANUP (wp_countEdgePixels (pIPWState, prMask, 
                                          x, y, w, bh, 
                                          bh, &edgeCount));
    count = edgeCount;

    /* Left */
    ERR_PASS_CLEANUP (wp_countEdgePixels (pIPWState, prMask, 
                                          x, y, bw, h,
                                          bw, &edgeCount));
    count += edgeCount;

    /* Right */
    ERR_PASS_CLEANUP (wp_countEdgePixels (pIPWState, prMask, 
                                          x + w - bw, y, bw, h, 
					  bw, &edgeCount));
    count += edgeCount;
    
    /* Bottom */
    ERR_PASS_CLEANUP (wp_countEdgePixels (pIPWState, prMask, 
                                          x, y + h - bh, w, bh,
				          bh, &edgeCount));
    count += edgeCount;

    percent = count * 100 / ((w + h) * 2);

    *pIsRect = (percent > cutoff) ? cTrue : cFalse;

    /* Check center */
    if (prText && *pIsRect && bw >2 && bh >2)
    {
	w1 = w - (bw <<2);
	h1 = h - (bh <<2);
	cen.x = xMaskOffset + x + (bw <<1);
	cen.y = yMaskOffset + y + (bh <<1);
	cen.w = w1;
	cen.h = h1;
	ERR_PASS_CLEANUP (w_numberBitsRect (pIPWState, 
					    prText, &cen,
					    &edgeCount));
	if(edgeCount > (w1*h1/4))
	    *pIsRect = cFalse;		    
    }
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::: "wp_countEdgePixels"
 *
 * OneLineDescription:::
 *   Internal routine: count the number of pixels near an edge 
 *
 * LongDescription:::
 *	Copy out a region, dilate it with some strel,
 *      and count the number of pixels in the region. Divide that count 
 *	by a number passed, which is the width of the region. This is
 *	used to count # of pixels along an edge region within some number
 *	of pixels (i.e. the dilation size). This routine makes and destroys
 *	a PIXR every time, which would be a bad thing if our MALLOC was
 *	slow. Since we destroy it when we leave, it "shouldn't" fragment
 *	the heap.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  prImage::        Image of interest
 *  x::              Region of interest
 *  y::              
 *  w::              
 *  h::              
 *  dw::             Thing to divide by, i.e. strel size
 *
 * Output:::
 *  pPixels::        # of pixels found
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image classification
 *
 * Keywords:::
 *  image classification
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "wp_countEdgePixels"></A>
 *****************************************************************************/
static ErrCode CDECL
wp_countEdgePixels (
    IPWState           *pIPWState,
    PIXR               *prImage,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               dw,
    Int32              *pPixels)
{
    PROLOGUE;
    PIXR       *prTmp = NULL;
    Int32	count;
    ScaledInt1616 xDPI, yDPI;

    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, prImage, &xDPI, &yDPI));
    ERR_PASS (w_createPixrNoInit (pIPWState, w, h, 1, xDPI, yDPI, cFrameBits, 
				  &prTmp));

    /* Copy out */
    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
				  x, y, prImage, 
				  PIX_SRC, 
				  0, 0, w, h, prTmp));

    /* we use run filtering rather than the morphology package to perform
     * the dilation since the "strel" size may be quite large ( > 32).
     * This operation is also more memory efficient than dilation when
     * done in-place.
     */

    if (dw == w)  /* vertical box side (left or right) */
        ERR_PASS_CLEANUP (w_filterHorizontalRuns (pIPWState, NULL, prTmp,
	                                          0, 1, dw, 10000, cPassRuns,
						  &prTmp));

    else  /* horizontal box side (top or bottom) */
        ERR_PASS_CLEANUP (w_filterVerticalRuns (pIPWState, NULL, prTmp,
	                                        0, 1, dw, 10000, cPassRuns,
						&prTmp));


    /* Count */
    ERR_PASS_CLEANUP (w_numberBits (pIPWState, prTmp, &count));
    count /= dw;

    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prTmp));
    *pPixels = count;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prTmp) w_destroyPixr (pIPWState, &prTmp);
    ERR_RETURN;
    
}


/*****************************************************************************
 * BeginDocumentation::: "w_textMask"
 *
 * OneLineDescription:::
 *  Compute a mask of where we find text on a  page
 *
 * LongDescription:::
 *  This function attempts to create a mask of the text on a page.
 *  
 * PrivateDescription:::
 *  Compute a text mask for a page. What we're going to do is find all
 *  of the connected components on the page using rabb, and then run down
 *  the list and try to mark and merge "text-like" regions, and then 
 *  fill those regions in on the resulting mask. rabb gives us back
 *  a linked list of rectangles, which are guaranteed to be in order
 *  of "last scan-line found", i.e. if one had a "j" and an "a" on a line,
 *  the "j" would come first in the list whether or not it was first
 *  on the scan line, because its descender is always below the bottom
 *  of the "a". Stuff is put onto the list in reverse order, so the
 *  bottom right cc's on the page occur first in the list.
 *  <BR>
 *  Our heuristic for finding text is to find adjacent connected components
 *  that are the "right" size and close enough together. Finding these
 *  parameters will require a bit of experimentation, but the important
 *  thing is that they be relative to the size of the connected components
 *  found, not some absolute number (maybe some absolute limits would help 
 *  filter a bit) so that varying text sizes are found correctly.
 *  <BR>
 *  We're going to take the rectangle list that rabb makes and zero out
 *  all of the "mass" values that rabb so carefully computes, and
 *  use them for our own purposes: the mass value will indicate the
 *  number of merged regions that a rectangle represents. The idea is that
 *  once we've gone through and done all of the merging, we go through
 *  and find the connected components that have at least <n> components
 *  merged, where n is probably around 4 or 5. (We're going to punt on
 *  the problem of finding single, isolated short words).
 *  <BR>
 *  One problem that immediately comes up in this scheme is that
 *  if letters merge into words too much, we may fail to find 
 *  enough adjacent connected components. Another problems is that
 *  sometimes we may find connected components made up of letters or
 *  words merged vertically. The moral is that we might want to run this
 *  at 300dpi rather than 100, if we have the choice.
 *  <BR>---<BR>
 *  Images with text-on-tints obviously have a different threshold
 *  for the text-on-background and text-on-tint. In order to get a 
 *  decent text mask for the whole page, I've added an adaptive
 *  threhsold option. Just pass in a grayscale image, and it'll
 *  get adaptively thresholded before the text mask is computed.
 *  <BR>---<BR>
 *  Made a couple of changes. First, use an internal "CCBOX" structure
 *  to hold more info than a RECTANGL, such as the count of merged
 *  boxes and the sum of their heights and centers (for taking accurate
 *  means). Also, match on centers rather than baseline/topline, which
 *  seems a bit more robust.
 *
 * Input:::
 *  pIPWState::      State structure
 *  image::          Binary image to analyze
 *  IsTextSeg::      Bool switch. If want keep more text, set 1, 
 *                   otherwise 0. 
 *
 * Output:::
 *  mask::           Mask computed, under which text lies
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *  This is not robust enough to be used for OCR. It was indended to be
 *  a quick-and-dirty routine used to cut obvious text out of regions
 *  which were determined to be image by a segmenter. It doesn't do
 *  too bad a job, however, and could probably be improved to the
 *  point where when combined with the adaptive thresholding code,
 *  it could be used as an alternative segmenter.
 *
 * FunctionCategory:::
 *  image classification
 *
 * Keywords:::
 *  image classification, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_textMask"></A>
 *****************************************************************************/
ErrCode CDECL
w_textMask (
    IPWState           *pIPWState,
    PIXR               *image,
    PIXR              **mask,
    Bool               IsTextSeg)
{
    PROLOGUE;
    Int32	 w, h, i, x0, x1, depth, rCode, minW, minH,
		 maxSpacing, w1, h1, bpl, dpi;
    UInt8       *pImageAll;
    CCBOX	*boxList, *boxP, *tmpP;
    CCHeader     header;
    PIXR	*binImage = NULL;
    ScaledInt1616 xDPI, yDPI;

/*  new   */
    Int32        *phist =NULL;
    Int32        idx1,idx2, hMin, hMax;
    Int32        clusterNum, charIdx, boxNum, pointLim;
    CCRectHeight *prectH = NULL;
    Int32        smallH, bigH;
    CCBOX        *pimgBox, *pmergeBox, *pbox1, *pbox2, *pbox3;
    Int32        dtop, dbottom, yLim, xLim, breakLim;
    Bool         binside;
    CCBOX        **pplistInImg, **pplistOver;
    Int32        count, left, right, top, bottom, overNum;
    Int32        consisNum, avgH, avgW, lineNum;

    ERR_GEN_PIXR (image, cByteOrderBig, cDepth1 /* | cDepth8 */, 
		  cChannels1 | cChannels3, cFrame32);
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, image, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));
/* dpi may be not ratio */
    if( dpi == 0) dpi = MAX(xDPI, yDPI);

    ERR_GEN( dpi < 10, cErrApps_invalidResolution);

    binImage = image;

    ERR_PASS (w_pixrGetDepth (pIPWState, binImage, &depth));
    ERR_PASS (w_pixrGetWidth (pIPWState, binImage, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, binImage, &h));

/*    if (depth == 8)
    {
	binImage = NULL;
	ERR_PASS (w_adaptiveThreshold (pIPWState, image, dpi, &binImage));
    }
    else
*/

    minW = dpi / 50;
    minH = dpi / 25;
    maxSpacing = dpi / 5;

    if (!(*mask))
    {
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 1,
				       Int32_TO_ScaledInt1616(dpi), 
				       Int32_TO_ScaledInt1616(dpi),  
				      cFrameBits, mask));
    }
    else
    {
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, *mask, &w1));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, *mask, &h1));
	ERR_GEN_CLEANUP (w1 != w || h1 != h, cErrInvalidParm);
    }

    /* Compute the bounding boxes */
#if 0
    if (!(rectList = rabb4Connected (pIPWState, binImage, 2, 4)))
	return abortI("rabb failed", procName, 1);
#else
    header.pBoxList = NULL;
    /* Compute the bounding boxes using our own callback */
    ERR_PASS_CLEANUP (w_pixrGetImageAll(pIPWState, binImage, &pImageAll));
    ERR_PASS_CLEANUP (w_pixrGetBpl(pIPWState, binImage, &bpl));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, binImage, &w1));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, binImage, &h1));

/* for text minW =1, minH=3 */
    minW >>= 1;
    minH >>= 1;
    if(minW <1) minW =1;
    if(minH <3) minH =3;

    ERR_PASS_CLEANUP( i_findConnectedComponentBoundingBoxes (
	    pIPWState, 
	    pImageAll, bpl,  cFrameBits, cFrameBits, w1, h1, minW, minH,
	    0, (void *)&header, wp_addCCBoxToList, &rCode));;
#endif
    boxList = header.pBoxList;
    if(boxList == NULL)
    {
	if(binImage && (binImage != image))
	   ERR_PASS_CLEANUP(w_destroyPixr (pIPWState, &binImage));	   
	NORMAL_RETURN;
    }

/* bounding box height analysis */
    phist = IPW_CALLOC(sizeof(Int32), (h1+2), cAllocatorData);
    ERR_GEN(!phist, cErrNomem);
    prectH = IPW_CALLOC(sizeof(CCRectHeight), (h1>>2), cAllocatorData);
    ERR_GEN(!prectH, cErrNomem);
    boxNum =0;
    boxP = boxList;
    while( boxP != NULL )
    {
	++ *(phist+boxP->h);
	++ *(phist+boxP->h+1);
	++ *(phist+boxP->h-1);
	++ boxNum;
	boxP = boxP->next;
    }
    hMin=0;
    while((hMin < h1) && (*(phist+hMin)==0)) ++hMin;
    hMax = h1+1;
    while((hMax > hMin) && (*(phist+hMax)==0)) --hMax;

    clusterNum =0;
    prectH->left = hMin;
    x0 =0;
    idx1 = hMin;
    while(idx1 <= hMax)
    {
	if( *(phist+idx1) >0)
	{
	    x0 += (*(phist+idx1))*idx1;
	    (prectH+clusterNum)->sunNum += *(phist+idx1);
	    ++ idx1;
	}
	else
	{
	    (prectH+clusterNum)->right = idx1-1;
	    if((prectH+clusterNum)->sunNum >0)
		(prectH+clusterNum)->center = x0/(prectH+clusterNum)->sunNum;
	    while((idx1 <= hMax) && (*(phist+idx1) == 0)) ++ idx1;
	    ++ clusterNum;
	    (prectH+clusterNum)->left = idx1;
	    x0 =0;
	}
    }
    (prectH+clusterNum)->right = idx1-1;
    if((prectH+clusterNum)->sunNum >0)
	(prectH+clusterNum)->center = x0/(prectH+clusterNum)->sunNum;
    ++ clusterNum;    

    x0 =0;
    charIdx =0;
    for(idx1=0; idx1<clusterNum; ++idx1)
    {
	if((prectH+idx1)->sunNum > x0) 
	{
	    x0 = (prectH+idx1)->sunNum;
	    charIdx = idx1;
	}
    }
    if( charIdx > 0) smallH = prectH->center;
    else smallH = prectH->left + (prectH->center - prectH->left)/3;
    
    pointLim = (dpi<<2)/5;
    bigH = (prectH+charIdx)->right +1;
    for(idx1 = clusterNum-1; idx1 > charIdx; --idx1)
    {
	if((prectH+idx1)->left >= pointLim) continue;
	else if((prectH+idx1)->center >= pointLim && 
		(prectH+idx1)->sunNum < 5) continue;
	bigH = (prectH+idx1)->right +1;
	break;
    }
    IPW_FREE(phist, cAllocatorData);

/* filter out bars */
    boxP = boxList;
    while( boxP != NULL )
    {
	if((boxP->w > (w1>>1) && boxP->w > (boxP->h<<3) &&
	       boxP->w > (dpi<<1) && boxP->nRuns < (boxP->w -boxP->h)) ||
	   (boxP->w > dpi && boxP->h < (prectH+charIdx)->right &&
	   ( ( boxP->w > (boxP->h<<3) && boxP->mass*5>= boxP->h*boxP->w*4) || 
	     ( boxP->nRuns< (boxP->w>>1) && boxP->w > (boxP->h<<2) )))|| 
	   (boxP->h < boxP->w && boxP->h > (prectH+charIdx)->right &&
	    boxP->nRuns < (boxP->w>>1) &&
	    (boxP->mass*10 > boxP->h*boxP->w*9 || 
	     boxP->mass*7 < boxP->h*boxP->w)) ||
	    (boxP->h > bigH && (boxP->mass<<3) < (boxP->h *boxP->w)) )  
	{
	    tmpP = boxP->next;
	    if (boxP == boxList)
	    {
		boxList = tmpP;
		if(tmpP) tmpP->prev = NULL;
	    }
	    else
	    {
		boxP->prev->next = tmpP;
		if(tmpP) tmpP->prev = boxP->prev;
	    }
	    IPW_FREE(boxP, cAllocatorData);
	    -- boxNum;
	    boxP = tmpP;
	}
	else boxP = boxP->next;
    }

/*  take out images form box list */
    pimgBox = NULL;
    boxP = boxList;
    while( boxP != NULL )
    {
	tmpP = boxP->next;
	if(boxP->h > bigH) 
	{
	    if (boxP == boxList)
	    {
		boxList = tmpP;
		if(tmpP) tmpP->prev = NULL;
	    }
	    else
	    {
		boxP->prev->next = tmpP;
		if(tmpP) tmpP->prev = boxP->prev;
	    }
	    if(pimgBox == NULL) 
	    {
		pimgBox = boxP;
		boxP->prev = boxP;
		boxP->next = boxP;
	    }
	    else 
	    {
		boxP->prev = pimgBox->prev;
		boxP->next = pimgBox;
		pimgBox->prev = boxP;
		boxP->prev->next = boxP;
	    }
	    -- boxNum;
	}
	boxP = tmpP;
     }

/* Based on normal height merge box */
/* merge loop  */
    x1 = (prectH+charIdx)->right +1;
    if( charIdx >= (clusterNum-1)) x1= (prectH+charIdx)->center;
    if(pimgBox != NULL)
    {
	pimgBox->prev->next = NULL;
	pimgBox->prev = NULL;   
    }
    for(idx1=0; idx1<3; ++idx1)
    {
	if(boxList == NULL) break;
	if(idx1 > 1 && pimgBox == NULL) break;
	for(boxP=boxList; boxP!=NULL; )
	{
	    if(boxP->h < smallH) 
	    {
		boxP = boxP->next;
		continue;
	    }
	    
	    if(idx1 <1)
		breakLim = boxP->yMin - boxP->h;
	    else breakLim = boxP->yMin - (boxP->h<<2);
	    if(idx1 > 1) 
	    {
		if(boxP->h <= x1) 
		{
		    boxP = boxP->next;
		    continue;
		}
		pmergeBox = pimgBox;
	    }
	    else pmergeBox = boxList;
	    
	    while( pmergeBox != NULL )
	    {
		if(pmergeBox->yMax < breakLim) break;
		if(pmergeBox->yMin > boxP->yMax || 
		   pmergeBox->yMax < boxP->yMin ||
		   pmergeBox == boxP ) 
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}
		if (boxP->h <= pmergeBox->h) 
		{
		    hMin = boxP->h;
		    hMax = pmergeBox->h;
		}
		else 
		{
		    hMin = pmergeBox->h;
		    hMax = boxP->h;
		}
		if(idx1 < 1) 
		{
		    yLim = hMin * 3;
		    if(hMax > maxSpacing) yLim -= (hMin >>1);
		}
		else 
		{
		    yLim = hMin <<1;
		    if(pmergeBox->h == hMin && hMin <= smallH) 
			yLim = hMin <<2;
		    else if(pmergeBox->h == hMin && 
			    hMin <= (prectH+charIdx)->center)
			yLim += hMin;
		}

		binside = cFalse;
		if( idx1 > 0 && pmergeBox->h == hMin && 
		    boxP->yMin <= pmergeBox->yMin &&
		    boxP->yMax >= pmergeBox->yMax && 
		    boxP->x <= pmergeBox->x &&
		    (boxP->x+boxP->w) >= (pmergeBox->x+pmergeBox->w))
		    binside = cTrue;

		if( hMax > yLim && binside == cFalse) 
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}
	   
		if( boxP->x < pmergeBox->x) 
		{
		    pbox1 = boxP;
		    pbox2 = pmergeBox;
		}
		else 
		{
		    pbox1 = pmergeBox;
		    pbox2 = boxP;
		}
		if(idx1 >0) xLim = hMin;
		else 
		{
		    xLim = hMax;
		    if( hMax > maxSpacing ) xLim = hMax >>1;
		}

		xLim = MIN(xLim, maxSpacing);

		if ((pbox2->x - (pbox1->x+pbox1->w)) > xLim) 
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}

		dtop = ABS(pbox1->yMin - pbox2->yMin);
		dbottom = ABS(pbox1->yMax - pbox2->yMax);
		if( (binside == cFalse && 
		     (hMax - dtop - dbottom) < (hMax >> 2)) || 
		( (pbox1->yCenter < pbox2->yMin || 
		   pbox1->yCenter > pbox2->yMax) &&
		  (pbox2->yCenter < pbox1->yMin || 
		   pbox2->yCenter > pbox1->yMax) ) ) 
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}

/* Merge them */

		pbox1 = pmergeBox->prev;	   
		pbox2 = pmergeBox->next;
		if( pmergeBox == boxList) 
		{
		    boxList = pbox2;
		    if(pbox2) pbox2->prev = NULL;
		}
		else if( pmergeBox == pimgBox)
		{
		    pimgBox = pbox2;
		    if(pbox2) pbox2->prev = NULL;
		}
		else 
		{
		    if( pbox1 ) pbox1->next = pbox2;
		    if( pbox2 ) pbox2->prev = pbox1;
		}
		if(binside == cFalse)
		{
		    if(pmergeBox->yMin < boxP->yMin) 
			boxP->yMin = pmergeBox->yMin;
		    if(pmergeBox->yMax > boxP->yMax) 
			boxP->yMax = pmergeBox->yMax;
		    boxP->heightSum += pmergeBox->heightSum;
		    boxP->centerSum += pmergeBox->centerSum;
		    boxP->nRuns += pmergeBox->nRuns;
		    boxP->cnt += pmergeBox->cnt;
		    boxP->mass += pmergeBox->mass;
		    boxP->yCenter = (boxP->yMin + boxP->yMax) >>1;
		    boxP->h = boxP->yMax - boxP->yMin;
		    x0 = boxP->x + boxP->w;
		    if( (pmergeBox->x + pmergeBox->w) > x0)
			x0 = pmergeBox->x + pmergeBox->w;
		    if(pmergeBox->x < boxP->x) boxP->x = pmergeBox->x;
		    boxP->w = x0 - boxP->x;
		    boxP->y = boxP->yMin;
		}
		IPW_FREE (pmergeBox, cAllocatorData);
		if(idx1 < 2) -- boxNum;
		pmergeBox = pbox2;
	    }	    
	    boxP = boxP->next;
	}
    }
/* clear small box */
    boxP = boxList;
    while( boxP != NULL )
    {
	if(boxP->h < smallH  ||  boxP->h > (boxP->w<<3) ||
	    (boxP->h <= smallH && boxP->w <= smallH)) 
	{
	    pbox1 = boxP->next;
	    if (boxP == boxList)
	    {
		boxList = pbox1;
		if (pbox1) pbox1->prev = NULL;
	    }
	    else
	    {
		boxP->prev->next = boxP->next;
		if (pbox1)
		    pbox1->prev = boxP->prev;
	    }
	    IPW_FREE (boxP, cAllocatorData);
	    -- boxNum;
	    boxP = pbox1;
	    continue;
 	}
	boxP = boxP->next;
    }

/* classfy frame or halftone image and clear up image box */
/* Whole image may be halftone image */

    if(pimgBox != NULL && boxNum >0) 
    {
	pplistInImg = IPW_CALLOC(sizeof(CCBOX*), boxNum*2, cAllocatorData);
	ERR_GEN(!pplistInImg, cErrNomem);
	pplistOver = pplistInImg + boxNum;
	for(boxP = pimgBox; boxP != NULL; )
	{
	    if( boxList == NULL) 
	    {
		pbox2 = boxP->next;
		if( boxP == pimgBox )
		{
		    pimgBox = pbox2;
		    if(pbox2) pbox2->prev = NULL;
		}
		else
		{
		    boxP->prev->next = pbox2;
		    if(pbox2) pbox2->prev = boxP->prev;
		}
		IPW_FREE(boxP, cAllocatorData);
		boxP = pbox2;
		continue;
	    }
	    count =0;
	    overNum =0;
	    hMax =0;
	    lineNum =0;
	    right = boxP->x;
	    left = boxP->x+boxP->w;
	    top = boxP->yMax;
	    bottom = boxP->yMin;
	    avgH =0;
	    avgW =0;
	    consisNum =0;
	    for(pmergeBox = boxList; pmergeBox!= NULL; )
	    {
		if(boxP->x <= pmergeBox->x && boxP->yMin <= pmergeBox->yMin 
		   && boxP->yMax >= pmergeBox->yMax && 
		   (boxP->x+boxP->w) >= (pmergeBox->x + pmergeBox->w))
		{
		    hMax += pmergeBox->h * pmergeBox->w;
		    avgH += pmergeBox->h;
		    avgW += pmergeBox->w;
		    *(pplistInImg+count) = pmergeBox;
		    ++ count;
		    if(pmergeBox->w > (pmergeBox->h<<1)) ++ lineNum;
		    if(pmergeBox->x < left) left = pmergeBox->x;
		    if((pmergeBox->x+pmergeBox->w) > right)
			right = pmergeBox->x + pmergeBox->w;
		    if(pmergeBox->yMin < top) top = pmergeBox->yMin;
		    if(pmergeBox->yMax > bottom) bottom = pmergeBox->yMax;
		}
		else if(boxP->x < (pmergeBox->x+pmergeBox->w) &&
		    (boxP->x + boxP->w) > pmergeBox->x &&
		    boxP->yMin < pmergeBox->yMax &&
		    boxP->yMax > pmergeBox->yMin)
		{
		    *(pplistOver+overNum) = pmergeBox;
		    ++ overNum;
		}
		pmergeBox = pmergeBox->next;
	    }
	    boxP->heightSum = count;
	    if(count >0) 
	    {
		avgH /= count;
		avgW /= count;
	    }
	    x0 = boxP->h * boxP->w;
	    x1 = (bottom-top)*(right-left);
	    if( count >0 &&  x1 < (x0>>3) && avgW <(avgH<<2))
		binside = cFalse;
	    else if( (boxP->h > (h1>>1) || boxP->w>(w1>>1)) && 
		 boxP->h > (h1>>2) && boxP->w > (w1>>2) &&
		 hMax > (x0 >>4) && x1 > (x0>>1) && 
	         avgW > (avgH << 1) && (hMax > (x0>>2) ||
		 count > (boxNum >>1) ))
		binside = cTrue;   /* frame */
	    else if(lineNum == 0 && count >0) binside = cFalse;
	    else if( avgW >0 && avgW < (avgH<<2) && (hMax <<5) < x0) 
		binside = 20;
	    else if(count >1)
	    {
/* classification of text and halftone image */
/* find consistent neighbors */
		consisNum =0;
		for(idx1=0; idx1<count; ++idx1) 
		{
		    pbox1 = (CCBOX*) *(pplistInImg+idx1);
		    for(idx2=0; idx2<count; ++idx2) 
		    {
			pbox2 = (CCBOX*) *(pplistInImg+idx2);
			if(pbox1 == pbox2) continue;
			if(pbox2->w < (pbox2->h<<1)) continue;
			if(pbox2->h > (pbox1->h<<1) || 
			   pbox1->h > (pbox2->h<<1)) continue;
			if(pbox1->yMin < pbox2->yMax && 
			   pbox1->yMax > pbox2->yMin) 
			{
			    if(pbox1->x > pbox2->x && 
			       (pbox1->x-pbox2->x-pbox2->w) < (pbox1->h*3)) 
			    {
				++ consisNum;
				break;
			    }
			    else if( pbox1->x <= pbox2->x &&
				     (pbox2->x-pbox1->x-pbox1->w)< pbox1->h*3) 
			    {
				++ consisNum;
				break;
			    }
			}
			else if(pbox1->x < (pbox2->x+pbox2->w) &&
				(pbox1->x+pbox1->w) > pbox2->x) 
			{
			    if(pbox1->yMin > pbox2->yMin && 
			       (pbox1->yMin - pbox2->yMax) < (pbox1->h<<1)) 
			    {
				++ consisNum;
				break;
			    }
			    else if(pbox1->yMin <= pbox2->yMin &&
				    (pbox2->yMin - pbox1->yMax)< (pbox1->h<<1)) 
			    {
				++ consisNum;
				break;
			    }
			}
		    }
		}
		if(consisNum < (count >>1) ) binside = cFalse;
		else binside = 10;
	    }
	    else binside = 10;
	    breakLim =0;
	    if(binside == cFalse) 
	    {
		for(idx1=0; idx1<count; ++idx1) 
		{
		    pmergeBox = (CCBOX*) *(pplistInImg+idx1);
		    pbox1 = pmergeBox->prev;	   
		    pbox2 = pmergeBox->next;
		    if( pmergeBox == boxList) 
		    {
			boxList = pbox2;
			if(pbox2) pbox2->prev = NULL;
		    }
		    else 
		    {
			if( pbox1 ) pbox1->next = pbox2;
			if( pbox2 ) pbox2->prev = pbox1;
		    }
		    IPW_FREE (pmergeBox, cAllocatorData);
		    -- boxNum;
		}
	    }
	    else if(binside == 10) 
	    {
		for(idx1=0; idx1<count; ++idx1) 
		{
		    pmergeBox = (CCBOX*) *(pplistInImg+idx1);
		    if(pmergeBox->h >0 && pmergeBox->w > (pmergeBox->h<<1) &&
		       pmergeBox->cnt > 1 )
		    {
			++ breakLim;
			continue;
		    }
		    pbox1 = pmergeBox->prev;	   
		    pbox2 = pmergeBox->next;
		    if( pmergeBox == boxList) 
		    {
			boxList = pbox2;
			if(pbox2) pbox2->prev = NULL;
		    }
		    else 
		    {
			if( pbox1 ) pbox1->next = pbox2;
			if( pbox2 ) pbox2->prev = pbox1;
		    }
		    IPW_FREE (pmergeBox, cAllocatorData);
		    *(pplistInImg+idx1) = NULL;
		    -- boxNum;
		}
	    }
	    else if(binside == 20) 
	    {
		for(idx1=0; idx1<count; ++idx1) 
		{
		    pmergeBox = (CCBOX*) *(pplistInImg+idx1);
		    if(pmergeBox->h >0 &&  pmergeBox->w > pmergeBox->h*4 && 
		       pmergeBox->cnt >= cMinimumBoxCount )
		    {
			++breakLim;
			continue;
		    }
		    pbox1 = pmergeBox->prev;	   
		    pbox2 = pmergeBox->next;
		    if( pmergeBox == boxList) 
		    {
			boxList = pbox2;
			if(pbox2) pbox2->prev = NULL;
		    }
		    else 
		    {
			if( pbox1 ) pbox1->next = pbox2;
			if( pbox2 ) pbox2->prev = pbox1;
		    }
		    IPW_FREE (pmergeBox, cAllocatorData);
		    *(pplistInImg+idx1) = NULL;
		    -- boxNum;
		}
	    }
	    if( (binside == 10 || binside == 20) && breakLim <=3)
	    {
		for(idx1=0; idx1<count; ++idx1) 
		{
		    pmergeBox = (CCBOX*) *(pplistInImg+idx1);
		    if(pmergeBox == NULL) continue;
		    if(pmergeBox->h >0 && 
		       pmergeBox->cnt < (pmergeBox->w/pmergeBox->h)*2)
			continue;
		    pbox1 = pmergeBox->prev;	   
		    pbox2 = pmergeBox->next;
		    if( pmergeBox == boxList) 
		    {
			boxList = pbox2;
			if(pbox2) pbox2->prev = NULL;
		    }
		    else 
		    {
			if( pbox1 ) pbox1->next = pbox2;
			if( pbox2 ) pbox2->prev = pbox1;
		    }
		    IPW_FREE (pmergeBox, cAllocatorData);
		    -- boxNum;
		}
	    }
	    if(overNum >0)
	    {
		lineNum =0;
		for(idx1=0; idx1<overNum; ++idx1)
		{
		    pmergeBox = (CCBOX*) *(pplistOver+idx1);
		    if(pmergeBox->w > (pmergeBox->h<<1) &&
		       pmergeBox->cnt > cMinimumBoxCountWide)
			++ lineNum;
		}
		if( lineNum < overNum/3 || lineNum == 0)
		{
		    for(idx1=0; idx1<overNum; ++idx1)
		    {
			pmergeBox = (CCBOX*) *(pplistOver+idx1);
			pbox1 = pmergeBox->prev;	   
			pbox2 = pmergeBox->next;
			if( pmergeBox == boxList) 
			{
			    boxList = pbox2;
			    if(pbox2) pbox2->prev = NULL;
			}
			else 
			{
			    if( pbox1 ) pbox1->next = pbox2;
			    if( pbox2 ) pbox2->prev = pbox1;
			}
			
			IPW_FREE (pmergeBox, cAllocatorData);
			-- boxNum;
		    }
		}
		else if(lineNum <= (overNum >>1))
		{
		    for(idx1=0; idx1<overNum; ++idx1)
		    {
			pmergeBox = (CCBOX*) *(pplistOver+idx1);
			if(pmergeBox->w > (pmergeBox->h<<1) &&
			   pmergeBox->cnt > cMinimumBoxCountWide)
			    continue;
			pbox1 = pmergeBox->prev;	   
			pbox2 = pmergeBox->next;
			if( pmergeBox == boxList) 
			{
			    boxList = pbox2;
			    if(pbox2) pbox2->prev = NULL;
			}
			else 
			{
			    if( pbox1 ) pbox1->next = pbox2;
			    if( pbox2 ) pbox2->prev = pbox1;
			}

			IPW_FREE (pmergeBox, cAllocatorData);
			-- boxNum;

		    }
		}
	    }
	    boxP = boxP->next;
	}
	IPW_FREE(pplistInImg, cAllocatorData);
    } /* pimgBox != NULL */

    boxP = pimgBox;
    while(boxP != NULL)
    {
	if( boxP->heightSum < 2)
	{
	    consisNum = boxP->cnt;
	    pmergeBox = pimgBox;
	    while(pmergeBox != NULL)
	    {
		if( boxP == pmergeBox || pmergeBox->heightSum >1 ||
		    pmergeBox->yMin > boxP->yCenter ||
		    pmergeBox->yMax < boxP->yCenter)
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}
		yLim =  ((MIN(boxP->h, pmergeBox->h)*3) >>2 );
		if( ABS(boxP->h - pmergeBox->h) > yLim ||
		    ABS(boxP->yMin - pmergeBox->yMin) > yLim ||
		    ABS(boxP->yMax - pmergeBox->yMax) > yLim )
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}
		if(boxP->x < pmergeBox->x)
		{
		    pbox1 = boxP;
		    pbox2 = pmergeBox;
		}
		else
		{
		    pbox1 = pmergeBox;
		    pbox2 = boxP;
		}
		yLim =  (MIN(boxP->h, pmergeBox->h) >>2 );
		yLim = MIN(yLim, maxSpacing);
		if((pbox2->x - (pbox1->x+pbox1->w)) < yLim)
		{
		    pbox1 = pmergeBox->prev;	   
		    pbox2 = pmergeBox->next;
		    if( pmergeBox == pimgBox) 
		    {
			pimgBox = pbox2;
			if(pbox2) pbox2->prev = NULL;
		    }
		    else 
		    {
			if( pbox1 ) pbox1->next = pbox2;
			if( pbox2 ) pbox2->prev = pbox1;
		    }
		    boxP->mass += pmergeBox->mass;
		    boxP->nRuns += pmergeBox->nRuns;
		    if(pmergeBox->yMin < boxP->yMin) 
			boxP->yMin = pmergeBox->yMin;
		    if(pmergeBox->yMax > boxP->yMax) 
			boxP->yMax = pmergeBox->yMax;
		    boxP->cnt += pmergeBox->cnt;
		    boxP->yCenter = (boxP->yMin + boxP->yMax) >>1;
		    boxP->h = boxP->yMax - boxP->yMin;
		    x0 = boxP->x + boxP->w;
		    if( (pmergeBox->x + pmergeBox->w) > x0)
			x0 = pmergeBox->x + pmergeBox->w;
		    if(pmergeBox->x < boxP->x) boxP->x = pmergeBox->x;
		    boxP->w = x0 - boxP->x;
		    boxP->y = boxP->yMin;
		    IPW_FREE (pmergeBox, cAllocatorData);
		    pmergeBox = pbox2;
		}
		else pmergeBox = pmergeBox->next;
	    }
	    if(consisNum >= boxP->cnt) boxP = boxP->next;
	}
	else boxP = boxP->next;
    }

    boxP = pimgBox;
    while( boxP != NULL)
    {
	pbox1 = boxP->next;
	if((boxP->cnt >= cMinimumBoxCount) ||
	    ((boxP->w > boxP->h * cAspectFactor) && 
	     (boxP->cnt >= cMinimumBoxCountWide)))
	{
	    pmergeBox = boxList; 
	    while( pmergeBox != NULL && 
		  pmergeBox->yMax >= boxP->yMax )
	    {
		pmergeBox = pmergeBox->next;
	    }
	    if( pmergeBox != NULL)
	    {
		pbox2 = pmergeBox->next;
		pmergeBox->next = boxP;
		if(pbox2) pbox2->prev = boxP;
		boxP->next = pbox2;
		boxP->prev = pmergeBox;
	    }
	    else 
	    {
		if(boxList != NULL)
		{
		    boxP->next = boxList;
		    boxList->prev = boxP;
		}
		else boxP->next = NULL;
		boxList = boxP;	
		boxP->prev = NULL;
	    }
	    ++ boxNum;
	}
	else IPW_FREE(boxP, cAllocatorData);
	boxP = pbox1;
    }

/* clear up isolated box and noise */

    for(boxP = boxList; boxP != NULL; )
    {
	if(boxList == NULL) break;

	if((boxP->cnt >= cMinimumBoxCount) ||
	    ((boxP->w > boxP->h * cAspectFactor) && 
 	     (boxP->cnt >= cMinimumBoxCountWide)))
	{
	    boxP = boxP->next;
	    continue;
 	}

	x0 = x1 = hMin = 0x7fff;
	i = boxP->x+boxP->w;
	pbox1 = pbox2 = pbox3 = NULL;
	for(pmergeBox = boxList; pmergeBox != NULL; )
	{
	    if(pmergeBox == boxP) 
	    {
		pmergeBox = pmergeBox->next;
		continue;
	    }
	    if(pmergeBox->x < i && 
	       (pmergeBox->x + pmergeBox->w) > boxP->x ) 
	    {
		if(pmergeBox->yMin >= boxP->yMax &&
		   (pmergeBox->yMin - boxP->yMax) < x0) 
		{
		    x0 = pmergeBox->yMin - boxP->yMax;
		    pbox1 = pmergeBox;
		}
		else if(pmergeBox->yMax <= boxP->yMin &&
			(boxP->yMin - pmergeBox->yMax) < x0) 
		{
		    x0 = boxP->yMin - pmergeBox->yMax;
		    pbox1 = pmergeBox;
		}
	    }
	    if(x1 > 0 && pmergeBox->yMin < boxP->yMax && 
	        pmergeBox->yMax > boxP->yMin && 
	       ((boxP->yMin < pmergeBox->yCenter &&
		   boxP->yMax > pmergeBox->yCenter)||
		(pmergeBox->yMin < boxP->yCenter &&
		    pmergeBox->yMax > boxP->yCenter))) 
	    {
		if(pmergeBox->x >= i && (pmergeBox->x - i) < x1) 
		{
		    x1 = pmergeBox->x - i;
		    pbox2 = pmergeBox;
		}
		else if((pmergeBox->x+pmergeBox->w) <= boxP->x &&
			(boxP->x - pmergeBox->x-pmergeBox->w) < x1) 
		{
		    x1 = boxP->x - pmergeBox->x - pmergeBox->w;
		    pbox2 = pmergeBox;
		}
		else if(pmergeBox->x < i && 
			(pmergeBox->x+pmergeBox->w) > boxP->x) 
		{
		    x1 =0;
		    pbox2 = pmergeBox;
		}
	    }
	    if(hMin > 0 && pmergeBox->yMin < boxP->yMax && 
	        pmergeBox->yMax > boxP->yMin && 
		(boxP->h<<1) > pmergeBox->h &&
		(pmergeBox->h<<1) > boxP->h &&
		((boxP->yMin < pmergeBox->yCenter &&
		   boxP->yMax > pmergeBox->yCenter)||
		(pmergeBox->yMin < boxP->yCenter &&
		    pmergeBox->yMax > boxP->yCenter))) 
	    {
		if(pmergeBox->x >= i && (pmergeBox->x - i) < hMin) 
		{
		    hMin = pmergeBox->x - i;
		    pbox3 = pmergeBox;
		}
		else if((pmergeBox->x+pmergeBox->w) <= boxP->x &&
			(boxP->x - pmergeBox->x-pmergeBox->w) < hMin) 
		{
		    hMin = boxP->x - pmergeBox->x - pmergeBox->w;
		    pbox3 = pmergeBox;
		}
		else if(pmergeBox->x < i && 
			(pmergeBox->x+pmergeBox->w) > boxP->x) 
		{
		    hMin =0;
		    pbox3 = pmergeBox;
		}
	    }
	    pmergeBox = pmergeBox->next;
	}
	binside = cFalse;
	if( pbox1 != NULL && x0 < boxP->h*3 && (pbox1->h<<1) > boxP->h &&
	    (boxP->h<<1) > pbox1->h )
	    binside = cTrue;
	if( binside == cFalse && pbox1 != NULL && x0 < boxP->h)
	{
	    idx1 = boxP->x;
	    if(pbox1->x > idx1) idx1 = pbox1->x;
	    idx2 = boxP->x+boxP->w;
	    if( (pbox1->x+pbox1->w) < idx2) idx2 = pbox1->x+pbox1->w;
	    if( ((idx2-idx1) > (boxP->w>>1) || x0 ==0) && 
		( boxP->w > (boxP->h<<1) ||
		x1 < (boxP->h>>1) )) binside = cTrue;
	}
	if( binside == cFalse && pbox3 != NULL && hMin < boxP->h*3 )
	    binside = cTrue;
	if( binside == cTrue && pbox1 != NULL && pbox2 != NULL &&
	    boxP->w < boxP->h &&
	    !( (pbox1->cnt >= cMinimumBoxCount) ||
	    ((pbox1->w > boxP->h * cAspectFactor) && 
	     (pbox1->cnt >= cMinimumBoxCountWide))) &&
	    !( (pbox2->cnt >= cMinimumBoxCount) ||
	    ((pbox2->w > boxP->h * cAspectFactor) && 
	     (pbox2->cnt >= cMinimumBoxCountWide))))
	    binside = cFalse;
        if(binside == cFalse)
	{
	    pbox1 = boxP->next;
	    if(boxP == boxList) 
	    {
		boxList = pbox1;
		if(pbox1) pbox1->prev = NULL;
	    }
	    else 
	    {
		boxP->prev->next = pbox1;
		if(pbox1) pbox1->prev = boxP->prev;
	    }

	    IPW_FREE(boxP, cAllocatorData);
	    -- boxNum;
	    boxP = pbox1;
	}
	else boxP = boxP->next;
    }

/* check all of boxs. if keep text is most important, skip this part. */
    if( !IsTextSeg )
    {
	consisNum =0;
	lineNum =0;
	left = w1;
	top = h1;
	right = bottom =0;
	idx1 =idx2 =0;
	x0 = (prectH+charIdx)->right+2;
	if(charIdx < (clusterNum-1)) x0=(prectH+charIdx+1)->left;
	for(boxP = boxList; boxP != NULL; )
	{
	    if(boxList == NULL) break;
	    if( boxP->w > (boxP->h << 2)) ++ lineNum;
	    if( boxP->x < left) left = boxP->x;
	    if( boxP->yMin < top) top = boxP->yMin;
	    if( boxP->yMax > bottom) bottom = boxP->yMax;
	    if( (boxP->x+boxP->w) > right) right = boxP->x + boxP->w;
	    if (boxP->h < x0) 
	    {
		++idx1;
		if(boxP->w > (boxP->h<<2)) ++ idx2;
	    }
	    for(pmergeBox = boxList; pmergeBox != NULL; )
	    {
		if(pmergeBox == boxP) 
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}
		if(boxP->h >= (pmergeBox->h<<1) || 
		   pmergeBox->h >= (boxP->h<<1) ) 
		{
		    pmergeBox = pmergeBox->next;
		    continue;
		}
		if(boxP->yMin < pmergeBox->yMax && 
		   boxP->yMax > pmergeBox->yMin && 
		   ((boxP->yCenter < pmergeBox->yMax &&
		     boxP->yCenter > pmergeBox->yMin  ) ||
		    (pmergeBox->yCenter < boxP->yMax &&
		     pmergeBox->yCenter > boxP->yMin) )) 
		{
		    if(boxP->x > pmergeBox->x && 
		       (boxP->x-pmergeBox->x-pmergeBox->w) < (boxP->h*3)) 
		    {
			++ consisNum;
			break;
		    }
		    else if( boxP->x <= pmergeBox->x &&
			     (pmergeBox->x-boxP->x-boxP->w)< boxP->h*3) 
		    {
			++ consisNum;
			break;
		    }
		}
		else if(boxP->x < (pmergeBox->x+pmergeBox->w) &&
			(boxP->x+pmergeBox->w) > pmergeBox->x &&
			pmergeBox->w > (pmergeBox->h *3)) 
		{
		    if(boxP->yMin > pmergeBox->yMin && 
		       (boxP->yMin - pmergeBox->yMax) < (boxP->h<<1)) 
		    {
			++ consisNum;
			break;
		    }
		    else if(boxP->yMin <= pmergeBox->yMin &&
			    (pmergeBox->yMin - boxP->yMax)< (boxP->h<<1)) 
		    {
			++ consisNum;
			break;
		    }
		}
		pmergeBox = pmergeBox->next;
	    }

	    boxP = boxP->next;
	}
	binside = cFalse;
	if( (lineNum < (boxNum>>1) && consisNum < (boxNum>>1)) ||
	    (lineNum < (boxNum/5) && consisNum < 
	     ((boxNum<<1)/3) && boxNum <20))
	    binside = cTrue;
	if( boxNum >0 && boxList != NULL && ( ( ( (right-left) < (w1>>1) || 
	    (bottom-top) < (h1>>1) ) && binside == cTrue ) || 
	   (((right-left) < (w1>>2) || (bottom-top) < (h1>>2) || 
	     ((right-left) < (w1>>1) && (bottom-top) <(h1>>1))) && 
		 lineNum == 0 && boxNum < 20) ||
	   ( idx1 > (boxNum <<1)/3 && boxNum > 20 && idx2 < idx1/5 && 
	    lineNum < (boxNum >>2) )) ) 
	{ /* filter out most boxs */
	    for(boxP=boxList; boxP!=NULL; )
	    {
		pbox1 = boxP->next;
		if(boxP->w < (boxP->h<<3) || boxP->cnt <cMinimumBoxCountWide)
		{
		    if (boxP == boxList)
		    {
			boxList = pbox1;
			if (pbox1) pbox1->prev = NULL;
		    }
		    else
		    {
			boxP->prev->next = boxP->next;
			if (pbox1)
			    pbox1->prev = boxP->prev;
		    }

		    IPW_FREE (boxP, cAllocatorData);
		    -- boxNum;		
		}
		boxP = pbox1;
	    }
	}
    } /* IstextSeg */

    /* Go through the boxlist and fill in the boxes */
    for (boxP = boxList; boxP; boxP = boxP->next)
    {
	if(boxP == NULL) break;
	ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
       				  0, 0, NULL,
       				  PIX_SET,
       				  boxP->x, boxP->yMin,
       				  boxP->w, boxP->yMax - boxP->yMin + 1,
				  *mask));
    }

    boxP = boxList;
    while( boxP != NULL )
    {
	tmpP = boxP->next;
	IPW_FREE (boxP, cAllocatorData);
	boxP = tmpP;
    }

    if (binImage != image)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &binImage));

    IPW_FREE(prectH, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (binImage && (binImage != image))
	w_destroyPixr (pIPWState, &binImage);
    if (prectH)     
	IPW_FREE(prectH, cAllocatorData);
    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation::: "w_reversetextMask"
 *
 * OneLineDescription:::
 *  Compute a mask of where we find reverse video text on a  page
 *
 * LongDescription:::
 *  This function attempts to create a mask of the reverse text ( black 
 *  background, white text) on a page.
 *  
 * PrivateDescription:::
 *  This function try to find the white text on the black background and mask
 *  them. It works on binary images. It will not mask reverse text which 
 *  is vertical line, logo, single reverse character, mixed with image, mixed
 *  with normal text and some graphic shaps or inside graphic or image regions.
 *
 * Input:::
 *  pIPWState::      State structure
 *  image::          Binary image to analyze
 *
 * Output:::
 *  mask::           Mask computed, under which text lines
 *
 * Returns:::
 *  error status 
 *
 * Notes:::
 *  This function will just support horizontal reverse text line. 
 *  It just works for binary image.
 *
 * FunctionCategory:::
 *  image classification
 *
 * Keywords:::
 *  image classification, segmentation
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  testing
 *
 * Authors/Modifications:::
 *  xianing::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_reversetextMask"></A>
 *****************************************************************************/
ErrCode CDECL
w_reversetextMask (
    IPWState           *pIPWState,
    PIXR               *image,
    PIXR              **mask)
{
    PROLOGUE;
    Int32	  w, h, x0, x1, depth, rCode, minW, minH,
		  w1, h1, bpl, dpi;
    UInt8         *pImageAll;
    POLYBOX	  *boxList, *boxP, *tmpP;
    POLYHeader    header;
    PIXR	  *binImage = NULL;
    ScaledInt1616 xDPI, yDPI;

/*  new   */
    Int32       maxWidth, minWidth, idx,idx1, height;
    Int32       left, right;
    Int32       size, size1, count, sizeSum;
    Int32       tW, tH;
    Bool        bClear, binside;
    PIXR        *pmask=NULL, *preImg=NULL;
    CCHeader    textHead;
    CCBOX	*rectList = NULL, *rectP, *nextP;
    Int32       *pvProj;

    ERR_GEN_PIXR (image, cByteOrderBig, cDepth1 /* | cDepth8 */, 
		  cChannels1 | cChannels3, cFrame32);
    ERR_PASS (w_pixrGetResolutionDPI (pIPWState, image, &xDPI, &yDPI));
    dpi = MAX (ScaledInt1616_TO_Int32(xDPI), ScaledInt1616_TO_Int32(yDPI));
    if( dpi < 10) dpi = MAX(xDPI, yDPI);

    binImage = image;

    ERR_PASS (w_pixrGetDepth (pIPWState, binImage, &depth));
    ERR_PASS (w_pixrGetWidth (pIPWState, binImage, &w));
    ERR_PASS (w_pixrGetHeight (pIPWState, binImage, &h));

    minW = dpi / 7;
    minH = dpi / 14;

    if (!(*mask))
    {
	ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 1,
				       Int32_TO_ScaledInt1616(dpi), 
				       Int32_TO_ScaledInt1616(dpi),  
				       cFrameBits, mask));
    }
    else
    {
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, *mask, &w1));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, *mask, &h1));
	ERR_GEN_CLEANUP (w1 != w || h1 != h, cErrInvalidParm);
    }

    /* Compute the bounding boxes 8-connection */

    header.pBoxList = NULL;
    header.dpi = dpi;
    header.idx =0;
    for(idx=0; idx<cMaxPrevBox; ++idx)
	header.pallList[idx] = NULL;

    /* Compute the bounding boxes using our own callback */
    ERR_PASS_CLEANUP (w_pixrGetImageAll(pIPWState, binImage, &pImageAll));
    ERR_PASS_CLEANUP (w_pixrGetBpl(pIPWState, binImage, &bpl));
    ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, binImage, &w1));
    ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, binImage, &h1));

    if(minW <1) minW =1;
    if(minH <3) minH =3; 

     ERR_PASS_CLEANUP (w_createPixr (pIPWState, w, h, 1, dpi, dpi, 
				      cFrameBits, &pmask));

     ERR_PASS( i_findConnectedComponentRuns(
				pIPWState,
				pImageAll,
				bpl,               	/* Bpl */
				cFrameBits, cFrameBits,	/* clipX, clipY */
				w1,           	        /* clipW */
				h1,			/* clipH */
				minW, minH,	        /* min width & height
							 * of returned rects */
				0,	   	        /* 4- vs 8-connected */
				(void *)&header,       	/* client's data */
				wp_recognizeBox,	/* callback routine */
				&rCode));   	        /* return code for */
							/* callback proc */

    for(idx=0; idx<cMaxPrevBox; ++idx)
    {
	if(header.pallList[idx] != NULL)
	    IPW_FREE(header.pallList[idx], cAllocatorData);
	header.pallList[idx] = NULL;
    }

/* Clean up vertical bars, small horizontal bars */  
    boxList = header.pBoxList;
    boxP = boxList;
    while( boxP != NULL )
    {
	bClear = cFalse;
	if( boxP->mass < (boxP->size >>1) ||
	    boxP->mass >=  (boxP->size- (minH<<2)) ||
	    boxP->w < minW || boxP->h < minH ||
	    (boxP->w < (minH <<3) && boxP->h < (minH <<3) ) ||
	    boxP->nRuns <= boxP->h )
	    bClear = cTrue;
	else if( boxP->w < dpi && boxP->h > (boxP->w * 6))
	    bClear = cTrue;
	else if( boxP->h > boxP->w)
	{
	    minWidth = 0x7fff;
	    maxWidth =0;
	    x0 =0;
	    while(x0 < boxP->h)
	    {
		if((*(boxP->pright+x0) - *(boxP->pleft+x0)+1)>= (boxP->w>>1))
		    break;
		++ x0;
	    }
	    height = boxP->h -1;
	    while(height >0)
	    {
		if( (*(boxP->pright+height) - *(boxP->pleft+height)+1) >=
		  (boxP->w>>1))
		    break;
		-- height;
	    }
	    if((height - x0+1) < (boxP->h>>1)) bClear = cTrue;
	    for(; x0<=height; x0++)
	    {
		x1 = *(boxP->pright+x0) - *(boxP->pleft+x0)+1;
		if(x1< minWidth) minWidth = x1;
		if(x1> maxWidth) maxWidth = x1;
	    }
	    if( minWidth < (maxWidth>>1) ) bClear = cTrue;
	    else if(boxP->h > (boxP->w <<1) && boxP->w < dpi)
	    {
		left = right =0;
		for(x0=0; x0<boxP->h; ++x0)
		{
		    left += *(boxP->pleft+x0);
		    right += *(boxP->pright+x0);
		}
		left /= boxP->h;
		right /= boxP->h;
		if(ABS(left-boxP->x)> minH || 
		   ABS(right - boxP->x - boxP->w) >minH)
		    bClear = cTrue;		
	    }
	}
	else if(boxP->h < (dpi>>1) && boxP->w <= (boxP->h*3) && boxP->h >0)
	{
	    left = right =0;
	    for(x0=0; x0<boxP->h; ++x0)
	    {
		left += *(boxP->pleft+x0);
		right += *(boxP->pright+x0);
	    }
	    left /= boxP->h;
	    right /= boxP->h;
	    if(ABS(left-boxP->x)> minH || ABS(right - boxP->x - boxP->w) >minH)
		bClear = cTrue;
	}
	if(bClear == cTrue)
	{
	    tmpP = boxP->next;
	    if (boxP == boxList)
	    {
		boxList = tmpP;
		if(tmpP) tmpP->prev = NULL;
	    }
	    else
	    {
		if(boxP->prev) boxP->prev->next = tmpP;
		if(tmpP) tmpP->prev = boxP->prev;
	    }
	    if(boxP->pleft != NULL)
		IPW_FREE(boxP->pleft, cAllocatorData);
	    IPW_FREE(boxP, cAllocatorData);
	    boxP = tmpP;	    
	    continue;
	}
	for(x0=0; x0<boxP->h; x0++)
	{
	    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
       		0, 0, NULL, PIX_SET, *(boxP->pleft+x0), boxP->y+x0,
	       (*(boxP->pright+x0) - *(boxP->pleft+x0)+1), 1, pmask));
	}
	boxP = boxP->next;
    }

    if(boxList != NULL)
    {
/* We inverse reverse video, the do text finding in each reverse video 
   bounding box to confirm whether it is revise video */

	ERR_PASS_CLEANUP (w_logicOp (pIPWState, binImage, AND, pmask));
	ERR_PASS_CLEANUP (w_invert (pIPWState, pmask, &preImg));

	textHead.pBoxList = NULL;
	ERR_PASS_CLEANUP (w_pixrGetImageAll(pIPWState, preImg, &pImageAll));
	ERR_PASS_CLEANUP (w_pixrGetBpl(pIPWState, preImg, &bpl));
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, preImg, &tW));
	ERR_PASS_CLEANUP (w_pixrGetHeight (pIPWState, preImg, &tH));

	minW = dpi / 100;
	minH = dpi / 50;
	if( minW < 1) minW =1;
	if( minH < 3) minH =3;
	ERR_PASS_CLEANUP( i_findConnectedComponentBoundingBoxes(
	    pIPState,
	    pImageAll, bpl,  cFrameBits, cFrameBits, tW, tH, minW, minH,
	    0, (void *)&textHead, wp_addCCBoxToList, &rCode));

	rectList = textHead.pBoxList;
	rectP = rectList;
	while( rectP != NULL)
	{
	    rectP->heightSum = rectP->centerSum * rectP->centerSum;
	    nextP = rectP->next;
	    if( rectP->w >= (rectP->h<<1) && 
		rectP->mass *5 > (rectP->h * rectP->w * 4))
	    {
		if( rectP == rectList)
		{
		    rectList  = nextP;
		    if(rectList) rectList->prev = NULL;
		}
		else
		{
		    if(rectP->prev) rectP->prev->next =nextP;
		    if(nextP) nextP->prev = rectP->prev;
		}
		IPW_FREE (rectP, cAllocatorData);
	    }
	    rectP = nextP;
	}
/* Put text connected components into each reverse video box */
	for(boxP = boxList; boxP; boxP = boxP->next)
	{
	    x0 = boxP->x + boxP->w;
	    x1 = boxP->y + boxP->h;
	    boxP->cnt =0;
	    rectP = rectList;
	    while( rectP != NULL)
	    {
		nextP = rectP->next;
		if(rectP->x >= boxP->x && rectP->yMin >= boxP->y &&
		    rectP->yMax < x1 && (rectP->x + rectP->w) <= x0)
		{
		    if( rectP == rectList)
		    {
			rectList = rectP->next;
			if(rectList) rectList->prev = NULL;
		    }
		    else
		    {
			if(rectP->prev) rectP->prev->next = rectP->next;
			if(rectP->next) rectP->next->prev = rectP->prev;
		    }
		    idx =0;
		    if( rectP->h >= dpi)
		    {
			/* Check the character rect in how many box ? */
			for( tmpP = boxList; tmpP; tmpP = tmpP->next)
			{
			    if( tmpP != boxP && tmpP->x >= rectP->x &&
				tmpP->y >= rectP->yMin &&
				(tmpP->x + tmpP->w) <= (rectP->x + rectP->w) &&
				(tmpP->y + tmpP->h-1) <= rectP->yMax)
			    {
				++ idx;
				break;
			    }
			}
		    }
		    if(idx < 1)
		    {
			rectP->prev = NULL;
			rectP->next = boxP->plist;
			if(rectP->next) rectP->next->prev = rectP;
			boxP->plist = rectP;
			++ boxP->cnt;
		    }
		    else
		    {
			IPW_FREE (rectP, cAllocatorData);
		    }
		}
		rectP = nextP;
	    }
	}
	rectP = rectList;
	while( rectP != NULL)
	{
	    nextP = rectP->next;
	    IPW_FREE (rectP, cAllocatorData);
	    rectP = nextP;
	}
	rectList = NULL;

	pvProj= (Int32 *)IPW_CALLOC(sizeof(Int32),tH,cAllocatorData);
	ERR_GEN (!pvProj, cErrNomem);	
	boxP = boxList;
	while( boxP != NULL)
	{
/* test each box's text distribution and texture feature to recognize
    reverse video  */
	    for(rectP = boxP->plist; rectP; rectP = rectP->next)
	    {
		/* merge overlay rect */
		bClear = cTrue;
		while( bClear == cTrue)
		{
		    bClear = cFalse;
		    nextP = boxP->plist; 
		    while(nextP != NULL)
		    {
			if(nextP == rectP || 
			   nextP->yMin > rectP->yMax ||
			   nextP->yMax < rectP->yMin ) 
			{
			    nextP = nextP->next;
			    continue;
			}
			binside = cFalse;
			left =0;
			if( rectP->yMin <= nextP->yMin &&
			    rectP->yMax >= nextP->yMax &&
			    rectP->x <= nextP->x &&
			    (rectP->x + rectP->w) >= (nextP->x+nextP->w) )
			    binside = cTrue;
			else if((rectP->h >= dpi || nextP->h >= dpi) &&
				( rectP->h > (nextP->h*3) ||
				 nextP->h > (rectP->h*3)) )
			{
			    nextP = nextP->next;
			    continue;
			}
			else if(rectP->x > nextP->x)
			    left = rectP->x - nextP->x - nextP->w;
			else
			    left = nextP->x - rectP->x - rectP->w;
			if((left > rectP->h*2  && left > nextP->h*2 ) ||
			   ( (rectP->yCenter < nextP->yMin || 
			      rectP->yCenter > nextP->yMax) &&
			     (nextP->yCenter < rectP->yMin || 
			      nextP->yCenter > rectP->yMax) ) )
			{
			    nextP = nextP->next;
			    continue;
			}
			if( nextP == boxP->plist)
			{
			    boxP->plist  = nextP->next;
			    if(boxP->plist) boxP->plist->prev = NULL;
			}
			else
			{
			    if(nextP->prev) nextP->prev->next = nextP->next;
			    if(nextP->next) nextP->next->prev = nextP->prev;
			}
			rectP->mass += nextP->mass;
			rectP->nRuns += nextP->nRuns;
			if(binside == cFalse)
			{
			    if(nextP->yMin < rectP->yMin) 
				rectP->yMin = nextP->yMin;
			    if(nextP->yMax > rectP->yMax) 
				rectP->yMax = nextP->yMax;
			    rectP->heightSum += nextP->heightSum;
			    rectP->centerSum += nextP->centerSum;
			    rectP->cnt += nextP->cnt;
			    rectP->yCenter = (rectP->yMin + rectP->yMax) >>1;
			    rectP->h = rectP->yMax - rectP->yMin;
			    x0 = rectP->x + rectP->w;
			    if( (nextP->x + nextP->w) > x0)
				x0 = nextP->x + nextP->w;
			    if(nextP->x < rectP->x) rectP->x = nextP->x;
			    rectP->w = x0 - rectP->x;
			    rectP->y = rectP->yMin;
			}
			IPW_FREE (nextP, cAllocatorData);
			-- boxP->cnt;
			bClear = cTrue;
			break;
		    }
		}
	    }
/* If no text inside, it is not reverse video. Otherwise do text line
    testing */
	    bClear = cFalse;
	    if(boxP->cnt < 1) bClear = cTrue;
	    else if(boxP->w < (boxP->h<<1) &&
		    boxP->h < (boxP->w<<1) &&
		    boxP->cnt >2 &&
		    boxP->w < dpi*3 &&
		    boxP->h < dpi *3)
	    {
		idx = idx1 =0;
		rectP = boxP->plist;
		nextP = rectP->next;
		if( rectP != NULL && nextP != NULL)
		{
		    height = nextP->h + rectP->h;
		    if( nextP->x < (rectP->x + rectP->w) &&
			(nextP->x + nextP->w) > rectP->x &&
			nextP->h < (rectP->h<<1) &&
			rectP->h < (nextP->h<<1) &&
			( nextP->yMin - rectP->yMax) < height)
			++ idx1;
		}
/* If inside text is not majority, it is not revise video */
		while(nextP != NULL)
		{
		    if( nextP->x < (rectP->x + rectP->w) &&
			(nextP->x + nextP->w) > rectP->x &&
			nextP->yMin < rectP->yMax &&
			nextP->yMax > rectP->yMin)
			++ idx;
		    size =0;
		    if( nextP->prev != NULL )
		    {
			height = nextP->h + nextP->prev->h;
			if( nextP->x < (nextP->prev->x + nextP->prev->w) &&
			    (nextP->x + nextP->w) > nextP->prev->x &&
			    nextP->h < (nextP->prev->h<<1) &&
			    nextP->prev->h < (nextP->h<<1) &&
			    (nextP->yMin - nextP->prev->yMax) < height)
			    ++ size;
		    }
		    if( size ==0 && nextP->next != NULL)
		    {
			height = nextP->h + nextP->next->h;
			if( nextP->x < (nextP->next->x + nextP->next->w) &&
			    (nextP->x + nextP->w) > nextP->next->x &&
			    nextP->h < (nextP->next->h<<1) &&
			    nextP->next->h < (nextP->h<<1) &&
			    ( nextP->next->yMin - nextP->yMax) < height)
			    ++ size;
		    }
		    idx1 += size;
		    rectP = nextP;
		    nextP = nextP->next;
		}
		if( idx >= ((boxP->cnt-1)<<1)/3) bClear = cTrue;
		else if( idx1 < (boxP->cnt>>1)) bClear = cTrue;
	    }
/* Statistic text line, noise to filter out non-reverse video box */
	    if(bClear == cFalse && (
		boxP->h < (boxP->w <<1) || boxP->w > dpi))
	    {
		for(idx=0; idx<boxP->h; ++idx)
		    *(pvProj +idx) =0;
		idx = count =0;
		size = size1 = sizeSum =0;
		rectP = boxP->plist;
		while(rectP != NULL)
		{
		    sizeSum += (rectP->h * rectP->w);
		    right = rectP->yMin - boxP->y;
		    for(left=0; left<=rectP->h; ++left, ++right)
		    {
			if( *(pvProj+right) < 1) *(pvProj+right) =1;
		    }
		    if( rectP->h > dpi && ( (rectP->w < rectP->h*2 &&
			rectP->cnt < cMinimumBoxCountWide) ||
			rectP->cnt < cMinimumBoxCount ))
		    {
			idx =0;
			break;
		    }
		    if( rectP->cnt > 1 && rectP->w >= (rectP->h<<1) &&
			(rectP->w * rectP->h *5) >= boxP->size * 3)
		    {
			right = rectP->yMin - boxP->y;
			for(left=0; left<=rectP->h; ++left, ++right)
			    *(pvProj + right) =2;
			size1 += (rectP->h * rectP->w);
			++ idx;
		    }
		    else if( rectP->h >= (minH<<1) && rectP->yMin >= boxP->y &&
			rectP->yMax < (boxP->y + boxP->h) &&
			( rectP->mass * 10 > (rectP->h * rectP->w) || 
			  rectP->w > (rectP->h <<3) ) &&
			( rectP->cnt >= cMinimumBoxCountWide &&
			    rectP->w > rectP->h *3  ))
		    {
			nextP = boxP->plist;
			binside =cFalse;
			x0 = rectP->x + rectP->w;
			while( nextP != NULL)
			{
			    if(nextP != rectP && 
			       rectP->yCenter > nextP->yMin &&
			       rectP->yCenter < nextP->yMax &&
			       rectP->x < (nextP->x + nextP->w) &&
			       x0 > nextP->x)
			    {
				binside = cTrue;
				break;
			    }
			    nextP = nextP->next;
			}
			if( binside == cFalse && 
			    rectP->w < (rectP->h <<3) &&
			    rectP->cnt >1)
			{
			    left = (rectP->heightSum - ((rectP->centerSum *
				   rectP->centerSum)/rectP->cnt))/ rectP->cnt;
			    right = rectP->centerSum/rectP->cnt;
			    idx1 = (rectP->h *rectP->h)/9;
			    if( left > idx1 || 
				ABS(right - rectP->yCenter) > (rectP->h/3))
				binside = cTrue;
			}
 			if(binside == cFalse && boxP->x > 0 &&
			    (boxP->x + boxP->w) < tW)
			{
			    right =rectP->x + rectP->w;
			    idx1 =0;
			    for(left=(rectP->yMin - boxP->y);
				left <= (rectP->yMax - boxP->y); ++ left)
			    {
				if( rectP->x < *(boxP->pleft+left) ||
				    right > *(boxP->pright+left))
				    ++ idx1;
			    }
			    if(idx1>0 && idx1 > (rectP->h /5)) 
				binside = cTrue;
			}
			if( binside == cTrue)
			{
			    rectP = rectP->next;
			    continue;
			}
			right = rectP->yMin - boxP->y;
			for(left=0; left<=rectP->h; ++left, ++right)
			    *(pvProj + right) =2;
			size1 += (rectP->h * rectP->w);
			++idx;
		    }
		    rectP = rectP->next;
		}
		if(idx <1) bClear = cTrue;
		else if( size1 < (sizeSum >>1))  bClear = cTrue;
		else
		{
		    idx1 =1;
		    while(idx1 <boxP->h)
		    {
			if( *(pvProj + idx1) > 0)
			    ++ idx1;
			else if( *(pvProj+idx1-1) >1)
			{
			    left = idx1;
			    while(idx1 < boxP->h && *(pvProj+idx1) <1)
				++ idx1;
			    if(idx1 < boxP->h && *(pvProj+idx1) >1)
			    {
				while(left < idx1)
				{
				    *(pvProj + left) =2;
				    ++ left;
				}
			    }
			}
			else if( *(pvProj+idx1-1) ==1 )
			{
			    left = idx1;
			    while(idx1 < boxP->h && *(pvProj+idx1) <1)
				++ idx1;
			    if(idx1 < boxP->h && *(pvProj+idx1) ==1)
			    {
				while(left < idx1)
				{
				    *(pvProj + left) =1;
				    ++ left;
				}
			    }
			}
			else ++idx1;
		    }
		    left = right =0;
		    for(idx1 =0; idx1<boxP->h; ++idx1)
		    {
			if( *(pvProj +idx1) >0) ++left;
			if( *(pvProj +idx1) >1) ++right;
		    }
		    if(right <= (left>>1)) bClear = cTrue;
		}
	    }
	    tmpP = boxP->next;
	    if(bClear == cTrue)
	    {
		if (boxP == boxList)
		{
		    boxList = tmpP;
		    if(tmpP) tmpP->prev = NULL;
		}
		else
		{
		    if(boxP->prev) boxP->prev->next = tmpP;
		    if(tmpP) tmpP->prev = boxP->prev;
		}
		if(boxP->pleft != NULL)
		    IPW_FREE(boxP->pleft, cAllocatorData);
		rectP = boxP->plist;
		while( rectP != NULL )
		{
		    nextP = rectP->next;
		    IPW_FREE(rectP, cAllocatorData);
		    rectP = nextP;
		}
		IPW_FREE(boxP, cAllocatorData);
	    }
	    boxP = tmpP;
	}
	IPW_FREE(pvProj, cAllocatorData);

/* generate reverse video mask image */
	boxP=boxList;
	while(boxP != NULL)
	{
	    if( boxP->plist != NULL)
	    {
		for(x0=0; x0<boxP->h; x0++)
		{
		    ERR_PASS_CLEANUP (w_rasterop (pIPWState, 
		       0, 0, NULL, PIX_SET, *(boxP->pleft+x0), boxP->y+x0,
		       (*(boxP->pright+x0) - *(boxP->pleft+x0)+1), 1, *mask));
		}
	    }
	    boxP = boxP->next;
	}
    }

    boxP = boxList;
    while( boxP != NULL )
    {
	tmpP = boxP->next;
	if(boxP->pleft != NULL)
	    IPW_FREE(boxP->pleft, cAllocatorData);
	rectP = boxP->plist;
	while( rectP != NULL)
	{
	    nextP = rectP->next;
	    IPW_FREE (rectP, cAllocatorData);
	    rectP = nextP;
	}
	IPW_FREE (boxP, cAllocatorData);
	boxP = tmpP;
    }

    if (binImage != image)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &binImage));
    if(pmask != NULL)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &pmask));
    if(preImg != NULL)
	ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &preImg));

    NORMAL_RETURN;

ERR_CLEANUP:

    if(pmask != NULL)
	w_destroyPixr (pIPWState, &pmask);
    if(preImg != NULL)
	w_destroyPixr (pIPWState, &preImg);
  
    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Callback for i_findConnectedComponentBoundingBoxes, used for CCBOXes
 *
 * LongDescription:::
 * This is an internal version of the a callback function for
 * i_findConnectedComponentBoundingBoxes, , used to make
 * CCBOX rather than box structures. We want the coordinates to be
 * relative to the upper left corner of the image,
 * not relative to the upper left corner of the frame.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  pHeader::        Callback header
 *  x::              Bounding box
 *  y::              
 *  w::              
 *  h::              
 *  m::              Mass (# of black pixels) of connected component
 *  nRuns::          Number of runs in connected component
 *  px::             X-coordinate of a point lying on the connected component
 *  py::             Y-coordinate of a point lying on the connected component
 *
 * Output:::
 *  none; modifies list.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_addCCBoxToList (
    IPWState	       *pIPWState,
    void *              pHeader,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    Int32               px,
    Int32               py)
{
    PROLOGUE;
    CCBOX	*box;
    CCHeader    *header;

    header = (CCHeader *) pHeader;

        /*  make a box */
    box = (CCBOX *)IPW_CALLOC (sizeof (CCBOX), 1, cAllocatorData);
    ERR_GEN (!box, cErrNomem);

	/* Set the box bounds */
    box->x = x - cFrameBits;
    box->y = y - cFrameBits;
    box->w = w;
    box->h = h;

	/* set number of active pixels, run and a point that's really
	 * in the connected component. */
    box->mass = m;
    box->cnt = 1;
    box->nRuns = nRuns;
    box->yCenter = box->y + ((h * 2 + 1) / 4);

	/* The initial sums are just the values themselves, they
	 * get added when they're merged */
    box->centerSum = box->yCenter;
    box->heightSum = h;

	/* The limits are just the values themselves, also */
    box->yMin = box->y;
    box->yMax = box->y + h - 1;


        /* Stick that box into the list */
    if (header->pBoxList == NULL)
    {
	header->pBoxList = box;
    }
    else
    {
	box->next = header->pBoxList;
	box->next->prev = box;
	header->pBoxList = box;
    }

    /* Shut up stupid compiler */
    px = py;

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation::: "wp_recognizeBox"
 *
 * OneLineDescription:::
 *   Callback for i_findConnectedComponentRuns, used for POLYBOXes
 *
 * LongDescription:::
 *  This is an internal version of the a callback function for
 *  i_findConnectedComponentRuns, , used to make
 *  POLYBOX rather than box structures. We want the coordinates to be
 *  relative to the upper left corner of the image,
 *  not relative to the upper left corner of the frame.
 *  We also do feature base classification. We just remain those POLYBOX which
 *  are reverse video text.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  pHeader::        Callback header
 *  x::              Bounding box
 *  y::              
 *  w::              
 *  h::              
 *  m::              Mass (# of black pixels) of connected component
 *  nRuns::          Number of runs in connected component
 *  *pRuns::         Runs position
 *
 * Output:::
 *  none; modifies list.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  connected component analysis
 *
 * Keywords:::
 *  connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xianing 
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_recognizeBox (
    IPState	       *pIPWState,
    void *              pHeader,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               m,
    Int32               nRuns,
    TCScanLineRun      *pRuns)
{
    PROLOGUE;
    POLYBOX	*box;
    POLYHeader   *header;
    LINERUNS     *prectRuns;
    TCScanLineRun   *pcurr, *pnext;
    PREVBOX      *pinBox;
    Int32        idx, idx1, lastL, lastR, left, right, size;
    Int32        minX, maxX, minY, maxY, maxRuns, maxWidth;
    Int32        count, rectRuns, w1, top, bottom, len, lastLen;
    UInt32       maxBRN, maxWRN, brSum, wrSum;
    Int32        blackLen, whiteLen, minAvg, maxAvg, lineNum;
    Bool         bFlag;
    Int32        *pbrHis, *pwrHis, *ptop, *pbottom, *pnRuns, *pmass;
    Int32        *plen;
    Int32        dpi, minH, yLim, xLim;


    header = (POLYHeader *) pHeader;
    dpi = header->dpi;
    minH = dpi / 14;
    yLim = dpi >>2;
    xLim = dpi *4 /5;
    len = nRuns; /* just for compiler */

    /* filter out all vertical line */
    if( w < dpi && h > (w<<2))
    {
	NORMAL_RETURN;
    }

    /* sort runs in line order */
    prectRuns = (LINERUNS *)IPW_CALLOC (sizeof(LINERUNS), h, cAllocatorData);
    ERR_GEN (!prectRuns, cErrNomem);

    pcurr = pRuns;
    size =0;
    for(idx=h-1; idx >= 0 && pcurr != NULL; --idx)
    {
	left = pcurr->runLeft;
	right = pcurr->runRight;
	(prectRuns+idx)->plastRun = pcurr;
	(prectRuns+idx)->pfirstRun = pcurr;
	(prectRuns+idx)->nRuns =1;
	(prectRuns+idx)->mass = pcurr->runRight - pcurr->runLeft +1;
	pnext = pcurr->pNext;
	while( pnext!= NULL && 
	       pnext->runRight < pcurr->runLeft )
	{
	    pcurr = pnext;
	    pnext = pcurr->pNext;
	    left = pcurr->runLeft;
	    (prectRuns+idx)->pfirstRun = pcurr;	
	    (prectRuns+idx)->nRuns ++;
 	    (prectRuns+idx)->mass += (pcurr->runRight - pcurr->runLeft+1);
	}
	pcurr = pcurr->pNext;
	size += (prectRuns+idx)->plastRun->runRight -
	    (prectRuns+idx)->pfirstRun->runLeft +1;
    }

    pinBox = (PREVBOX *)IPW_CALLOC (sizeof (PREVBOX), 1, cAllocatorData);
    ERR_GEN (!pinBox, cErrNomem);

    idx = header->idx;
    header->pallList[idx] = pinBox;
    pinBox->x = x;
    pinBox->y = y;
    pinBox->h = h;
    pinBox->w = w;
    pinBox->size = size;
    pinBox->mass = m;

    /* edge pre-processing */

    if( w < (h<<3) || h > dpi)
    {
    left = dpi >> 3;
    idx = 0;
    while(idx < h)
    {
	count = idx;
	right =0;
 	while(idx < h && (prectRuns+idx)->nRuns >1)
	{
	    pcurr = (prectRuns+idx)->plastRun;
	    pnext = pcurr->pNext;
	    while( pcurr != (prectRuns+idx)->pfirstRun &&
		   (pcurr->runLeft - pnext->runRight) < xLim)
	    {
		pcurr = pnext;
		pnext = pcurr->pNext;
	    }
	    if(pcurr == (prectRuns+idx)->pfirstRun) break;
	    if(right > 0 && ( (prectRuns+idx)->plastRun->runRight < right ||
		  pcurr->runLeft > (prectRuns+idx-1)->plastRun->runRight))
		break;
	    if( ((prectRuns+idx)->plastRun->runRight - pcurr->runLeft)> left ||
		(pcurr->runLeft - pnext->runRight) < xLim )
		break;
	    right = pcurr->runLeft;
	    ++idx;
	}
	if( (idx-count) > yLim )
	{
	    right = 0x7fff;
	    for(len=count; len<idx; ++len)
	    {
		pcurr = (prectRuns+len)->plastRun;
		pnext = pcurr->pNext;
		lastL =1;
		lastR = (pcurr->runRight -pcurr->runLeft+1);
		while( pcurr != (prectRuns+len)->pfirstRun &&
		       (pcurr->runLeft - pnext->runRight) < xLim)
		{
		    pcurr = pnext;
		    lastR += (pcurr->runRight -pcurr->runLeft+1);
		    pnext = pcurr->pNext;
		    ++ lastL;
		}
		if(pcurr->runLeft < right) right = pcurr->runLeft;
		(prectRuns+len)->plastRun = pnext;	
		(prectRuns+len)->nRuns -= (Int16)lastL;
		(prectRuns+len)->mass -= (Int16)lastR;
	    }
	    while(idx <h && (prectRuns+idx)->nRuns >1)
	    {
		pcurr = (prectRuns+idx)->plastRun;
		pnext = pcurr->pNext;
		lastL =1;
		lastR = (pcurr->runRight -pcurr->runLeft+1);
		while( pcurr != (prectRuns+idx)->pfirstRun &&
		       pnext->runRight >= right)
		{
		    pcurr = pnext;
		    lastR += (pcurr->runRight -pcurr->runLeft+1);
		    pnext = pcurr->pNext;
		    ++ lastL;
		}
		if(pcurr == (prectRuns+idx)->pfirstRun ||
		   (prectRuns+idx)->plastRun->runRight -pcurr->runLeft >left ||
		    pcurr->runLeft < (right -3))
		   break;
		(prectRuns+idx)->plastRun = pnext;	
		(prectRuns+idx)->nRuns -= (Int16)lastL;
		(prectRuns+idx)->mass -= (Int16)lastR;
		++ idx;
	    }
	}
	++ idx;
    }

    left = dpi >> 3;
    idx = 0;
    while(idx < h)
    {
	count = idx;
	right =0;
 	while(idx < h && (prectRuns+idx)->nRuns >1)
	{
	    pcurr = (prectRuns+idx)->plastRun;
	    pnext = pcurr->pNext;
	    while( pcurr != (prectRuns+idx)->pfirstRun &&
		   (pcurr->runLeft - pnext->runRight) < xLim &&
		(pnext->runRight - (prectRuns+idx)->pfirstRun->runLeft) > left)
	    {
		pcurr = pnext;
		pnext = pcurr->pNext;
	    }
	    if(pcurr == (prectRuns+idx)->pfirstRun) break;
	    if(right > 0 && ( (prectRuns+idx)->pfirstRun->runLeft > right ||
		  pnext->runRight < (prectRuns+idx-1)->pfirstRun->runLeft))
		break;
	    if((pnext->runRight - (prectRuns+idx)->pfirstRun->runLeft)> left ||
		(pcurr->runLeft - pnext->runRight) < xLim )
		break;
	    right = pnext->runRight;
	    ++idx;
	}
	if( (idx-count) > yLim)
	{
	    for(right=count; right<idx; ++right)
	    {
 		pcurr = (prectRuns+right)->plastRun;
		pnext = pcurr->pNext;
		lastL =1;
		lastR = (pcurr->runRight -pcurr->runLeft+1);
		while( pcurr != (prectRuns+right)->pfirstRun &&
		       (pcurr->runLeft - pnext->runRight) < xLim &&
	         (pnext->runRight-(prectRuns+right)->pfirstRun->runLeft) >left)
		{
		    pcurr = pnext;
		    lastR += (pcurr->runRight -pcurr->runLeft +1);
		    pnext = pcurr->pNext;
		    ++ lastL;
		}
		(prectRuns+right)->pfirstRun = pcurr;	
		(prectRuns+right)->nRuns -= (Int16)lastL;
		(prectRuns+right)->mass -= (Int16)lastR;
	    }
	}
	++ idx;
    }
    }

    /* rectangle out looking */
    pbrHis= (Int32 *)IPW_CALLOC(sizeof(Int32), ((w+2)*4+h*3+6),cAllocatorData);
    ERR_GEN (!pbrHis, cErrNomem);
    pwrHis = pbrHis + w+2;
    ptop = pwrHis +w +2;
    pbottom = ptop +w+2;
    pnRuns = pbottom + w+2;
    pmass = pnRuns + h+2;
    plen = pmass +h +2;
    for(idx=0; idx<h; ++idx)
    {
	*(pmass +idx) = (prectRuns +idx)->mass;
	*(pnRuns +idx) = (prectRuns +idx)->nRuns;
    }

    minX =prectRuns->pfirstRun->runLeft;
    maxX =prectRuns->pfirstRun->runRight;
    minY = maxY = y;
    size = maxX - minX +1;
    count = (Int32)prectRuns->mass;
    rectRuns = (Int32)prectRuns->nRuns;
    maxWidth = maxX - minX +1;
    for(idx=1; idx<h; ++idx)
    {
	lastL = (prectRuns+idx-1)->pfirstRun->runLeft;
	lastR = (prectRuns+idx-1)->plastRun->runRight;
	left = (prectRuns+idx)->pfirstRun->runLeft;
	right = (prectRuns+idx)->plastRun->runRight;
	top = bottom =0;
	if(idx >3 && idx <(h-3))
	{
	    lastLen = (*(pmass+idx-3) + *(pmass +idx -2) + *(pmass +idx-1));
	    len = (*(pmass+idx+1) + *(pmass +idx+2) + *(pmass +idx));
	    top = (*(pnRuns+idx-3) + *(pnRuns+idx-2) + *(pnRuns +idx-1));
	    bottom = (*(pnRuns+idx+1) + *(pnRuns +idx+2) + *(pnRuns+idx));
	    if(top >0) top = (lastLen + (top>>1))/top;
	    if(bottom >0) bottom =(len + (bottom>>1))/bottom;
 	    lastLen /=3;
	    len /= 3;
	}
	else len = lastLen =0;
	bFlag = cFalse;
	if( (right-left) > (lastR - lastL)*3 ||
	    (lastR - lastL) > (right-left) *3 || ( maxWidth > (dpi <<1) &&
    	    ((len > (lastLen << 4) && bottom > (top <<6) && top >0)||
	    (lastLen > (len << 4) && top > (bottom <<6) && bottom >0) ) ) )
	{
	    bFlag = cTrue;
	    if((maxY - minY+1) < minH) ++ bFlag;
	}
	else
	{
	    ++ maxY;
	    if(left < minX) minX = left;
	    if(right > maxX) maxX = right;
	    if((right-left+1) > maxWidth) maxWidth = right-left+1;
	    size += (right - left +1);
	    rectRuns += (Int32)(prectRuns+idx)->nRuns;
	    count += (Int32)(prectRuns+idx)->mass;
	}
	if( idx == (h-1) && (maxY - minY+1) >= minH ) bFlag = cTrue;
	if(bFlag == cTrue) /* check run length distribution */
	{
	    blackLen = whiteLen = 0;
	    for(lastL=0; lastL <= w; lastL++)
	    {
		*(ptop +lastL) =h;
		*(pbottom +lastL) =0;
		*(pbrHis +lastL) = *(pwrHis +lastL) =0;
	    }
	    top = maxY -y;
	    bottom = minY -y;
	    for(lastL= (minY - y); lastL<= (maxY -y); ++lastL)
	    {
		pcurr = (prectRuns + lastL)->plastRun;
		lastR = pcurr->runRight - pcurr->runLeft +1;
		if(lastR >blackLen) blackLen = lastR;
		if( *(pnRuns + lastL) ==1 && lastR > ((maxX - minX)>>1))
		{
		    if(lastL < top) top = lastL;
		    bottom = lastL;
		}
		for(lastR=pcurr->runLeft-x; lastR<=pcurr->runRight-x; ++lastR)
		{
		    *(pbottom+lastR) = lastL;
		    if( *(ptop+lastR) > lastL) *(ptop+lastR) = lastL;
		}
		while(pcurr != (prectRuns +lastL)->pfirstRun)
		{
		    pnext = pcurr->pNext;
		    lastR = pnext->runRight - pnext->runLeft +1;
		    if(lastR >blackLen) blackLen = lastR;
		    for(lastR=pnext->runLeft-x;
			lastR<=pnext->runRight-x; ++lastR)
		    {
			*(pbottom+lastR) = lastL;
			if( *(ptop+lastR) > lastL) *(ptop+lastR) = lastL;
		    }
		    lastR = pcurr->runLeft - pnext->runRight -1;
		    if(lastR >whiteLen) whiteLen = lastR;
		    pcurr = pnext;
		}
	    }

	    maxBRN = maxWRN =0;
	    brSum = wrSum =0;
	    lastR = (maxY - minY +1) >>1;
	    lastL =0;
	    while( lastL < w && (*(pbottom+lastL) - *(ptop+lastL)+1) <=lastR)
		++lastL;
	    w1 = w;
	    while(w1 >= lastL && (*(pbottom+ w1) - *(ptop+ w1)+1) <=lastR) 
		-- w1;
	    for(; lastL<w1; ++lastL)
	    {
		if( (*(pbottom+lastL) - *(ptop+lastL)+1) >lastR)
		{
		    ++ wrSum;
		    maxBRN = 0;
		}
		else 
		{
		    if(maxBRN == 0) 
		    {
			++ brSum;
			if(wrSum > maxWRN) maxWRN = wrSum;
			wrSum =0;
		    }
		    maxBRN =1;
		}
	    }
	    if(wrSum > maxWRN) maxWRN = wrSum;	    
	    if(blackLen < whiteLen || blackLen < (maxWidth >>1)||
		((maxY - minY +1) < (minH<<2) && (maxX - minX +1) <
		    (maxY - minY +1)*6 && blackLen < (maxWidth <<1)/3))
		++ bFlag;
	    else if(brSum > 1 || maxWRN < ((maxX - minX)/3))
		++ bFlag;
	    else if( (maxY - minY +1) > ((maxX - minX+1)<<1) &&
		     (maxX - minX +1) < dpi && (bottom -top) <
		     ((maxY - minY+1)>>1))
		++ bFlag;
	    else if( bottom < top && (maxY - minY+1) > dpi &&
		     (maxX - minX+1) >dpi &&
		     blackLen < (maxWidth *3)/5 &&
		     (maxX - minX+1)< ((maxY - minY+1)<<1) &&
		     (maxY - minY+1)< ((maxX - minX+1)<<1))
		++ bFlag; 
	    else 
	    {
		maxRuns =0;
		lastL = minY -y;
		lastR = maxY -y;
		blackLen =0;
		while(lastL <= lastR)
		{
		    len = (prectRuns+lastL)->plastRun->runRight -
			(prectRuns+lastL)->pfirstRun->runLeft + 1;
		    *(plen + lastL) = len;
		    blackLen += len;
		    if((prectRuns+lastL)->nRuns > maxRuns &&
			(prectRuns+lastL)->mass > (len >>1) )
			maxRuns = (prectRuns+lastL)->nRuns;
		    pcurr = (prectRuns+lastL)->plastRun;
		    while(pcurr != (prectRuns+lastL)->pfirstRun)
		    {
			pnext = pcurr->pNext;
			top = ((pcurr->runLeft + pnext->runRight)>>1) -x;
			if(top <0) top =0;
			if( lastL >= *(ptop+ top) && lastL<= *(pbottom+top) )
			{
			    len = pcurr->runLeft - pnext->runRight -1;
			    ++ *(pwrHis+len);
			    ++ *(pwrHis+len+1);
			    if(len >0) ++ *(pwrHis+len-1);
			}
			pcurr = pnext;
		    }
		    ++ lastL;
		}
		if( (maxY - minY +1) <= (maxX - minX +1)*2 &&
		    maxRuns <= cMinRuns ) ++ bFlag;
		else if( (maxY - minY +1) > (maxX - minX +1)*2 &&
			 maxRuns < 2) ++bFlag;
		else
		{
		    lastL = w-1;
		    while(lastL >0 && *(pwrHis+lastL) <1) --lastL;
		    whiteLen =0;
		    lastR =0;
		    for(len=0; len <=lastL; ++len)
		    {
			if( *(pwrHis+len) > lastR)
			{
			    lastR = *(pwrHis+len);
			    whiteLen = len;
			}
		    }
		    maxWRN =0;
		    if( whiteLen < minH )
			lastR = whiteLen <<3;
		    else
		    {
			lastR = whiteLen *6;
			if((maxX - minX +1) <((maxY - minY+1)<<1))
			{
			    for(len = 2; len <(whiteLen -1); ++len)
			    {
				top = *(pwrHis+len);
				if(top< *(pwrHis+len-1)&&top< *(pwrHis+len-2)&&
				 top< *(pwrHis+len+1)&& top< *(pwrHis+len+2)&&
				   *(pwrHis+len-1) < *(pwrHis+len-2) &&
				   *(pwrHis+len+1) < *(pwrHis+len+2))
				{
				    lastR = (whiteLen <<1) +1;
				    break;
				}
			    }
			}
		    }
		    if( whiteLen < (minH>>1))
			lastR += whiteLen;
		    if( lastR > lastL) lastR = lastL; 
		    for(len= 1; len <lastR; ++len)
		    {
			maxWRN += (UInt32) len * (*(pwrHis+len));
		    }
		    maxBRN =0;
		    for(len = lastR; len <=lastL; ++len)
		    {
		        maxBRN += (UInt32) len * (*(pwrHis+len));
		    }
		    if( (maxX -minX+1) > dpi && (maxX - minX+1)>lastR &&
			(maxY - minY+1) > dpi &&
			( (maxX - minX+1) > (dpi <<2) ||
			  (maxY - minY+1) > (dpi <<2) ) )
		    {
			for(len =0; len <cMaxPrevBox; ++len)
			{
			    if(len == header->idx ||
				header->pallList[len] == NULL) continue;
			    lastR = header->pallList[len]->x +
				 header->pallList[len]->w;
			    if( header->pallList[len]->x > minX &&
				lastR < maxX &&
				header->pallList[len]->y > minY &&
				(header->pallList[len]->y +
				 header->pallList[len]->h) < maxY)
			    {
				lastL = header->pallList[len]->x;
				for(top=minY-y; top <(maxY-y); ++top)
				{
				    if(lastL >
				       (prectRuns+top)->pfirstRun->runLeft &&
				       lastR < 
				       (prectRuns+top)->plastRun->runRight)
					continue;
				    break;
				}
				if( top < (maxY-y)) continue;
				maxBRN -= header->pallList[len]->size;
			    }
			}
				
		    }
		    if((maxX - minX) < ((maxY - minY) <<4) &&
			maxBRN > (maxWRN>>1) ) ++ bFlag;
		}
		if( bFlag == cTrue )
		{
		    len = maxY - minY +1;
		    blackLen = (blackLen + (len >>1)) /len;
		    top = minY - y;
		    bottom = maxY -y;
		    while(top < bottom && *(plen+top) < blackLen) ++ top;
		    while(bottom >top && *(plen + bottom) <blackLen) -- bottom;
		    lastL =0x7fff; /* min */
		    lastR = 0; /* max */
		    len =0; /* avg */
		    for(idx1=top; idx1<= bottom; ++idx1)
		    {
			len += *(pmass + idx1);
			if( *(pmass + idx1) < lastL)
			    lastL = *(pmass +idx1);
			if( *(pmass + idx1) > lastR)
			    lastR = *(pmass +idx1);
		    }
		    minAvg =maxAvg =0;
		    if( (bottom -top +1) >0 )
		    {
			len = (len + ((bottom - top +1)>>1))/(bottom - top +1);
			whiteLen =0;
			lineNum =0;
			for(idx1 =top; idx1<= bottom; ++idx1)
			{
			    if( *(pmass +idx1) >= len)
			    {
				whiteLen += ( *(pmass + idx1) - len);
				maxAvg += *(pmass +idx1);
			    }
			    else
			    {
				whiteLen += (len - *(pmass +idx1));
				minAvg += *(pmass +idx1);
				++ lineNum;
			    }
			}
			idx1 = bottom -top +1;
			whiteLen = (whiteLen +(idx1>>1))/idx1;
			if( lineNum >0) 
			    minAvg = (minAvg + (lineNum >>1))/lineNum;
			else minAvg = lastL;
			if( lineNum < idx1)
			    maxAvg=(maxAvg+((idx1-lineNum)>>1))/(idx1-lineNum);
			else maxAvg = lastR;
			idx1 = MIN( (maxAvg - len), (len - minAvg) );
		    }

		    if( (bottom - top) < 1) ++ bFlag;
		    else if( (maxY - minY +1) > (maxX - minX +1) &&
			(bottom - top) < minH) ++ bFlag;
		    else if( ((maxY - minY+1) > ((maxX - minX+1)<<2) &&
			     (maxX - minX +1) < dpi )||
			     ((maxX - minX+1) > ((maxY - minY+1)<<3) &&
			      (maxY - minY+1) < dpi))
		    {
			if( (lastR - lastL) < minH ||
			    ((maxAvg - len) <(lastR - len)/3 &&
			    (len - minAvg) < (len - lastL)/3)  ||
			    whiteLen < (idx1/3)) ++bFlag;
		    }
		    else
		    {
			if( (maxAvg - len) < ((len -minAvg)<<1) &&
			    (len -minAvg) < ((maxAvg -len)<<1))
			    idx1 >>=1;
			if( (lastR - lastL) < minH ||
			    (maxAvg - len) <(lastR - len)/3 ||
			(len - minAvg) < (len - lastL)/3  ||
			whiteLen < idx1) ++bFlag;
		    }
		}
	    }
	}
	if( bFlag == cTrue  )
	{
	    /*  make a box */
	    box = (POLYBOX *)IPW_CALLOC (sizeof (POLYBOX), 1, cAllocatorData);
	    ERR_GEN (!box, cErrNomem);

	    /* Set the box bounds */
	    box->x = minX - cFrameBits;
	    box->y = minY - cFrameBits;
	    box->w = maxX - minX +1;
	    box->h = maxY - minY +1;

	    box->pleft = (Int32 *)IPW_CALLOC (sizeof(Int32),
			       (box->h*2)+4, cAllocatorData);
	    ERR_GEN (!box->pleft, cErrNomem);
	    box->pright = box->pleft + box->h +2;

	    box->size = size;

	    for(lastL= (minY - y), idx1 =0; lastL<= (maxY -y); ++lastL, ++idx1)
	    {
		*(box->pleft+idx1) = (prectRuns+lastL)->pfirstRun->runLeft -
		     cFrameBits;
		*(box->pright+idx1) = (prectRuns+lastL)->plastRun->runRight -
		     cFrameBits;
	    }

	/* set number of active pixels, run and a point that's really
	 * in the connected component. */
	    box->mass = count;
	    box->cnt = 1;
	    box->nRuns = rectRuns;

        /* Stick that box into the list */
	    if (header->pBoxList == NULL)
	    {
		header->pBoxList = box;
	    }
	    else
	    {
		box->next = header->pBoxList;
		box->next->prev = box;
		header->pBoxList = box;
	    }
	}
	if( bFlag != cFalse)
	{
	    size = right - left +1;
	    minX = left;
	    maxX = right;
	    count = (Int32)(prectRuns+idx)->mass;
	    rectRuns = (Int32)(prectRuns+idx)->nRuns;
	    minY = maxY = y + idx;
	    maxWidth  =right - left+1;
	}
    }

    if(prectRuns != NULL)
	IPW_FREE(prectRuns, cAllocatorData);
    prectRuns = NULL;

    if(pbrHis != NULL)
	IPW_FREE(pbrHis, cAllocatorData);
    pbrHis = NULL;

    ++ header->idx;
    if(header->idx >= cMaxPrevBox) header->idx =0;

    if(header->pallList[header->idx] != NULL)
    {
	IPW_FREE(header->pallList[header->idx], cAllocatorData);
	header->pallList[header->idx] = NULL;
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation::: "w_labelColorCCs"
 *
 * OneLineDescription:::
 *   Label each connected compoent in an image with a color
 *
 * LongDescription:::
 *	Given a binary image and a color image, find each connected 
 *	component in the binary image, and compute the median color
 *	for the corresponding pixels in the color image. Constructs
 *	a linked list of structures which contain the color of the
 *	connected component, along with a list of the runs comprising
 *	the connected component.
 *      <BR>
 *	The structure used for the label is defined in imginfo.pub;
 *	it contains the mean color of the connected component and
 *	a pointer to a list of TCScanLineRun structures, which define
 *	all of the runs in the connected component.
 *	<BR>
 *      The idea is that we'll label the "equivalence classes" of colored
 *	text objects as a post-process after labeling the individual
 *	connected components.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure 
 *  pBinImage::      Binary (thresholded, probably adaptively) version of image
 *  pColorImage::    Color or grayscale image from which to determine color
 *  connectType::    Connectivity type to use, 0=4-connected, 1=8-connected
 *  minWidth::       Min width of a CC to consider
 *  minHeight::      Min height of a CC to consider
 *
 * Output:::
 *  ppLabels::       ColorComponentLabel list computed
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  colored text
 *
 * Keywords:::
 *  colored text, connected component analysis
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_labelColorCCs"></A>
 *****************************************************************************/
ErrCode CDECL
w_labelColorCCs (
    IPWState               *pIPWState,
    PIXR                   *pBinImage,
    PIXR                   *pColorImage,
    UInt32                  connectType,
    UInt32                  minWidth,
    UInt32                  minHeight,
    ColorComponentLabel   **ppLabels)
{
    PROLOGUE;
    Int32		  returnCode;
    LabelCBInfo		  info;
    PIXR                 *prEroded = NULL;
    STREL                *pStrel3 = NULL;

    ERR_GEN_PIXR (pBinImage, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN_PIXR (pColorImage, cByteOrderBig, cDepth8, cChannels1|cChannels3,
		  cFrame32);

    ERR_PASS (w_initStrelByName (pIPWState,  strel3));
    ERR_PASS (w_getStrel (pIPWState, strel3, &pStrel3));

    /* Make an eroded version */
    ERR_PASS (w_erode (pIPWState, pBinImage, pStrel3, &prEroded));
    
    info.pLabels = NULL;
    info.pColorImage = pColorImage;
    info.pErodedImage = prEroded;

    /* call rabbRuns to do the the work; the call-back procedure does
       the color averaging */
    ERR_PASS_CLEANUP (w_rabbRuns (pIPWState, pBinImage, minWidth, minHeight, 
				  connectType, (void *) &info, w_labelCB,
				  &returnCode));
    *ppLabels = info.pLabels;
    ERR_PASS_CLEANUP (w_destroyPixr (pIPWState, &prEroded));

    /* That should be it */
    NORMAL_RETURN;

ERR_CLEANUP:
    if (prEroded) w_destroyPixr (pIPWState, &prEroded);
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Callback to w_rabbRuns to compute a color for a connected component
 *
 * LongDescription:::
 *	Call-back procedure passed to rabbRuns when called by
 *	labelColorConnectedComponents. This call-back takes a list of
 *  	runs, and computes the mean color contained under the runs
 *	in the color version of the image. 
 *
 * PrivateDescription:::
 *	To get a better measure of the color, we're should take
 *	the median color rather than the mean. Because I'm lazy, and
 *	because I want to test this scheme without too much investment,
 *	I'm going to take the mean for now.
 *      <BR>---<BR>
 *      In fact, what I've found is that the median is not as good a measure 
 *      of color as taking the mean of the "center" of the CC. The 
 *      routine above does an "erode" to find this center; however,
 *      it should do a homotopic thinning instead so that it doesn't
 *      break characters. (I've added a thinning routine, so it's just
 *      work to be done).
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  infoVoid::       The "info" structure for the callback
 *  x::              Bounding box of CC
 *  y::              
 *  w::              
 *  h::              
 *  mass::           Mass of CC, i.e. # of black pixles
 *  nRuns::          # of runs
 *  pRuns::          Pointer to list of runs
 *
 * Output:::
 *  none; adds structure to list
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  now that there is a ConnectedComponent data type, this routine
 *  and the routine above should be modified to use it. They aren't,
 *  unfortunately.
 *
 * FunctionCategory:::
 *  colored text
 * Keywords:::
 *  colored text, connected component analysis
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_labelCB (
    IPWState	       *pIPWState,
    void               *infoVoid,
    Int32               x,
    Int32               y,
    Int32               w,
    Int32               h,
    Int32               mass,
    Int32               nRuns,
    TCScanLineRun      *pRuns)
{
    PROLOGUE;
    Int32	     lastx0, x0, x1, wRun, pixels, channels,
		     offset, bpl, i, eBpl, x0Eroded, x1Eroded, eWidth;
    TCScanLineRun   *pCopy = NULL, *pTmp;
    UInt32	     redISum, greenISum, blueISum, redSum, greenSum, blueSum,
	             grayISum, graySum;
    UInt8	    *pR, *pG, *pB, *pRBase, *pGBase, *pBBase, *pEroded, 
                    *pErodedBase;
    LabelCBInfo	    *info = (LabelCBInfo *) infoVoid;
    ColorComponentLabel	*pLabel = NULL;
    PIXR            *prPlane;


    /* Allocate a label thing */
    pLabel = (ColorComponentLabel *) IPW_MALLOC (sizeof (ColorComponentLabel),
						 cAllocatorData);
    ERR_GEN (!pLabel, cErrNomem);
    pLabel->pRuns = NULL;
    pLabel->x = x;
    pLabel->y = y;
    pLabel->w = w;
    pLabel->h = h;
    mass = nRuns;
    /* 
     * The runs only contain x-information. The starting y-coordinate 
     * is implicit in the bottom of the bounding box, and the way we
     * know we've moved up a scanline is that the x-coordinate is
     * greater than the previous one. (They go right->left, bottom->top)
     */

#if 0
    /* First, count the number of pixels in the connected component */
    pTmp = pRuns;
    pixels = 0;

    for (pTmp = pRuns; pTmp; pTmp = pTmp->pNext)
    {
	pixels += pTmp->runRight - pTmp->runLeft + 1;
    }
#endif
    ERR_PASS_CLEANUP (w_pixrGetChannelCount (pIPWState, info->pColorImage, &channels));
    if (channels == 3)
    {
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, info->pColorImage, cRedPlane,
					    &prPlane));
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pRBase));
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, info->pColorImage, 
					    cGreenPlane, &prPlane));
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pGBase));
	ERR_PASS_CLEANUP (w_pixrGetChannel (pIPWState, info->pColorImage, 
					    cBluePlane, &prPlane));
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, prPlane, &pBBase));
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, info->pErodedImage, 
					  &pErodedBase));
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, info->pErodedImage, &eWidth));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, info->pColorImage, &bpl));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, info->pErodedImage, &eBpl));
	redISum = greenISum = blueISum = 0;
	redSum = greenSum = blueSum = 0;
	lastx0 = 0x7fffffff;
	y = (y - cFrameBits) + h - 1;
	pixels = 0;

	/* Label a point actually on a run */
	pLabel->px = pRuns->runLeft - cFrameBits;
	pLabel->py = y;
    
	/* Sum the pixels under the runs and make a copy of the runs list */
	for (; pRuns; pRuns = pRuns->pNext)
	{
	    x0 = pRuns->runLeft - cFrameBits;
	    x1 = pRuns->runRight - cFrameBits;
	    if (x1 >= lastx0)
		y--;
	    lastx0 = x0;
       
	    pEroded = pErodedBase + y * eBpl;
	    if (i_nextBitToTheRight (pIPWState, pEroded, eWidth,
				     x0,  &x0Eroded) ==
		cSuccessful && x0Eroded < x1)
	    {
		if (i_nextBitToTheLeft (pIPWState, pEroded, 0,
					x1,  &x1Eroded) ==
		    cSuccessful && x1Eroded > x0)
		{
		
		    x0 = x0Eroded;
		    x1 = x1Eroded;
		    wRun = x1 - x0 + 1;
		    offset = y * bpl;
		    pR = pRBase + offset;
		    pG = pGBase + offset;
		    pB = pBBase + offset;

		    for (i = 0; i < wRun; i++)
		    {
			/* Sum all pixels */
			redSum += UCHAR_ACCESS(pR + x0);
			greenSum += UCHAR_ACCESS(pG + x0);
			blueSum += UCHAR_ACCESS(pB + x0);
			x0++;
			pixels++;
		    }
		}
	    }

	    /* Copy the run structure */
	    pTmp = (TCScanLineRun *) IPW_MALLOC (sizeof (TCScanLineRun), 
						 cAllocatorData);
	    ERR_GEN_CLEANUP (!pTmp, cErrNomem);

	    *pTmp = *pRuns;
	    if (pCopy)
		pCopy->pNext = pTmp;
	    else
		pLabel->pRuns = pTmp;

	    pCopy = pTmp;
	}

	/*
	 * Compute the mean. We assume that the bin image was eroded enough
	 * to choke out edge pixels.
	 */
	if (pixels)
	{
	    pLabel->color[cRedPlane] = /*redMin*/ (UInt8)(redSum / pixels);
	    pLabel->color[cGreenPlane] = /*greenMin*/ (UInt8) (greenSum / pixels);
	    pLabel->color[cBluePlane] = /*blueMin*/ (UInt8) (blueSum / pixels);
	}
	else
	{
	    /* Paint it black (for diagnostic purposes) */
	    pLabel->color[0] = pLabel->color[1] = pLabel->color[2] = 0;
	}
    }
    else
    {
	/* Grayscale */
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, info->pColorImage, &pGBase));
	
	ERR_PASS_CLEANUP (w_pixrGetImage (pIPWState, info->pErodedImage, 
					  &pErodedBase));
	ERR_PASS_CLEANUP (w_pixrGetWidth (pIPWState, info->pErodedImage, &eWidth));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, info->pColorImage, &bpl));
	ERR_PASS_CLEANUP (w_pixrGetBpl (pIPWState, info->pErodedImage, &eBpl));
	grayISum = 0;
	graySum = 0;
	lastx0 = 0x7fffffff;
	y = (y - cFrameBits) + h - 1;
	pixels = 0;

	/* Label a point actually on a run */
	pLabel->px = pRuns->runLeft - cFrameBits;
	pLabel->py = y;
    
    /* Sum the pixels under the runs and make a copy of the runs list */
	for (; pRuns; pRuns = pRuns->pNext)
	{
	    x0 = pRuns->runLeft - cFrameBits;
	    x1 = pRuns->runRight - cFrameBits;
	    if (x1 >= lastx0)
		y--;
	    lastx0 = x0;
       
	    pEroded = pErodedBase + y * eBpl;
	    if (i_nextBitToTheRight (pIPWState, pEroded, eWidth,
				     x0,  &x0Eroded) ==
		cSuccessful && x0Eroded < x1)
	    {
		if (i_nextBitToTheLeft (pIPWState, pEroded,
					0, x1,  &x1Eroded) ==
		    cSuccessful && x1Eroded > x0)
		{
		
		    x0 = x0Eroded;
		    x1 = x1Eroded;
		    wRun = x1 - x0 + 1;
		    offset = y * bpl;
		    pG = pGBase + offset;

		    for (i = 0; i < wRun; i++)
		    {
			/* Sum all pixels */
			graySum += UCHAR_ACCESS(pG + x0);
			x0++;
			pixels++;
		    }
		}
	    }

	    /* Copy the run structure */
	    pTmp = (TCScanLineRun *) IPW_MALLOC (sizeof (TCScanLineRun), 
						 cAllocatorData);
	    ERR_GEN_CLEANUP (!pTmp, cErrNomem);

	    *pTmp = *pRuns;
	    if (pCopy)
		pCopy->pNext = pTmp;
	    else
		pLabel->pRuns = pTmp;

	    pCopy = pTmp;
	}

	/*
	 * Compute the mean. We assume that the bin image was eroded enough
	 * to choke out edge pixels. In Gray, we're going to fill all the
	 * channels just in case, even though we should only be looking
	 * at the luminance plane in theory.
	 */
	if (pixels)
	{
	    pLabel->color[cLuminancePlane] = (UInt8)(graySum / pixels);
	    pLabel->color[cGreenPlane] =  (UInt8) (graySum / pixels);
	    pLabel->color[cBluePlane] =   (UInt8) (graySum / pixels);
	}
	else
	{
	    /* Paint it black (for diagnostic purposes) */
	    pLabel->color[0] = pLabel->color[1] = pLabel->color[2] = 0;
	}
    }
    /* Stick the label at the head of the list */
    pLabel->pNext = info->pLabels;
    info->pLabels = pLabel;

    NORMAL_RETURN;

ERR_CLEANUP:
    /* Free up stuff */
    if (pLabel)
    {
	if (pLabel->pRuns)
	{
	    pCopy = pLabel->pRuns;
	    while (pCopy)
	    {
		pTmp = pCopy->pNext;
		IPW_FREE (pCopy, cAllocatorData);
		pCopy = pTmp;
	    }
	}
	IPW_FREE (pLabel, cAllocatorData);
    }
    
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation::: "w_coalesceColorCCs"
 *
 * OneLineDescription:::
 *   Partition color-labelled CCs into color equivalence classes
 *
 * LongDescription:::
 *	Assuming that we've labeled our connected components using
 *	labelColorCCs, this procedure partitions those CCs into equivalence
 *	classes based on color distance. Note that we cheat a bit--we
 *	don't check every CC against each other, but rather we check
 *	each CC against the running average color computed for each
 *	class. The reasoning is that we expect a small number of
 *	classes relative to the number of CCs, and thus we should 
 *	improve the running time quite a bit at the expense of some
 *	accuracy. This might also help a bit in reducing problems with
 *      transitive chains of equivalence
 *
 *	We're using L*u*v* as the space to compute color distance;
 *	the L*u*v* conversion is quite slow and uses floating point,
 *	which may be a problem on PCs, since Windows is too stupid
 *	to reliably detect the floating point unit and install an
 *	emulator if there is none.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State structure
 *  pLabel::         The list of ColorComponentLabels structures,
 * 		     as created by labelColorCCs. NOTE that we
 *		     tear this structure apart when we partition it,
 *		     so don't expect it to be a single list when
 *		     done.
 *  maxColorDist::   Max Color distance (du^2 + dv^2)
 *  maxLumDist::     Max Luminance distance (dL^2)
 *
 * Output:::
 *  ppClasses::      The equivalence classes computed
 *  pClassCount::    The number of equivalence classes computed
 *
 * Returns:::
 *  Error status
 *
 * Notes:::
 *  now that there is a ConnectedComponent data type, this routine
 *  and the routine above should be modified to use it. They aren't,
 *  unfortunately.
 *  <BR>
 *  This routine doesn't really work very well yet. It needs more work.
 *
 * FunctionCategory:::
 *  colored text
 *
 * Keywords:::
 *  colored text
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * Citations:::
 *
 * EndDocumentation::: <A NAME = "w_coalesceColorCCs"></A>
 *****************************************************************************/
ErrCode CDECL
w_coalesceColorCCs (
    IPWState                     *pIPWState,
    ColorComponentLabel          *pLabel,
    Int32                         maxColorDist,
    Int32                         maxLumDist,
    ColorComponentEquivalence   **ppClasses,
    Int32                        *pClassCount)
{
    PROLOGUE;
    ColorComponentEquivalence	*pEquivHead = NULL, *pEquiv;
    ColorComponentLabel		*pNextLabel = NULL;
    Bool			 found;
    Int32			 colorDistance, r, g, b,  lumDistance;
    Float64			 L, u, v, L1, u1, v1, du, dv, dL;
    void			*state;

    *pClassCount = 0;
    ERR_PASS (w_RGBToLuvInit (pIPWState, 0.0, 0.0, 0.0, &state));

    /* For each label, look through the equivalence list */
    while (pLabel)
    {
	r = pLabel->color[cRedPlane];
	g = pLabel->color[cGreenPlane];
	b = pLabel->color[cBluePlane];
	ERR_PASS_CLEANUP (w_RGBToLuvProcess (pIPWState, r, g, b, state, &L, &u, &v));

	found = cFalse;
	pNextLabel = pLabel->pNext;

	for (pEquiv = pEquivHead; pEquiv; pEquiv = pEquiv->pNext)
	{
	    ERR_PASS_CLEANUP (w_RGBToLuvProcess (pIPWState, 
					pEquiv->avgColor[cRedPlane],
					pEquiv->avgColor[cGreenPlane],
					pEquiv->avgColor[cBluePlane], 
					state, 
					&L1, &u1, &v1));
	    du = u - u1;
	    dv = v - v1;
	    dL = L - L1;
	    colorDistance = (Int32) ((du * du + dv * dv) * (L*0.01));
	    lumDistance = (Int32) (dL * dL);

	    if ((colorDistance <= maxColorDist) && (lumDistance <= maxLumDist))
	    {
		found = cTrue;
		break;
	    }
	}

	if (found == cTrue)
	{
	    /* Add it to this color class if it was computed from
	     * "interior" pixels. If the first CC found wasn't computed
	     * from "interior" pixels, tough luck, we count it anyway.
	     */
	    if (pLabel->interior > cMinInteriorPixels)
	    {
		/* Accumulate to the color sums */
		pEquiv->colorSums[cRedPlane] += r;
		pEquiv->colorSums[cGreenPlane] += g;
		pEquiv->colorSums[cBluePlane] += b;
		pEquiv->components++;
		pEquiv->avgColor[cRedPlane] = (UInt8) (pEquiv->colorSums[cRedPlane] /
					      pEquiv->components);
		pEquiv->avgColor[cGreenPlane] = (UInt8) (pEquiv->colorSums[cGreenPlane]/
					        pEquiv->components);
		pEquiv->avgColor[cBluePlane] = (UInt8) (pEquiv->colorSums[cBluePlane] /
					       pEquiv->components);
	    }

	    /* Stick it into the equivalence list */
	    pLabel->pNext = pEquiv->pEquivList;
	    pEquiv->pEquivList = pLabel;
	}
	else
	{
	    /* Make a new color class */
	    pEquiv = (ColorComponentEquivalence *)
		     IPW_MALLOC (sizeof (ColorComponentEquivalence), cAllocatorData);
	    ERR_GEN_CLEANUP (!pEquiv, cErrNomem);

	    pEquiv->colorSums[cRedPlane] = 
		pEquiv->avgColor[cRedPlane] = (UInt8) r;
	    pEquiv->colorSums[cGreenPlane] = 
		pEquiv->avgColor[cGreenPlane] = (UInt8) g;
	    pEquiv->colorSums[cBluePlane] = 
		pEquiv->avgColor[cBluePlane] = (UInt8) b;
	    pEquiv->components = 1;

	    /* Put it on the new equivalence list */
	    pLabel->pNext = NULL;
	    pEquiv->pEquivList = pLabel;

	    /* Insert that equivalence class */
	    pEquiv->pNext = pEquivHead;
	    pEquivHead = pEquiv;

	    (*pClassCount)++;
	}

	/* Update the labels pointer */
	pLabel = pNextLabel;
    }
	    
    *ppClasses = pEquivHead;
		
    NORMAL_RETURN;

ERR_CLEANUP:
    /* We don't really know what to do on error. We don't own the
       color component structures, but we tore the list apart. Sorry! */
    ERR_RETURN;
}
	

/*
 * $Log: imginfo.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 * Imported sources
 *
 * Revision 1.78  1998/08/04 00:18:45  xzhu
 * xzhu on Mon Aug 3 17:18:33 PDT 1998
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added input parameters xMaskOffset and yMaskOffset in w_maskIsRect to fix bug when mask image is part of text mask image.
 *
 * Revision 1.77  1998/08/03 19:00:31  xzhu
 * xzhu on Mon Aug 3 11:59:37 PDT 1998
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Changed 1/3 area to 1/4 area as rectagle threshold in w_maskIsRect.
 *
 * Revision 1.76  1998/01/14 00:13:16  ddavies
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.
 *
 * Revision 1.75  1997/06/26 20:04:11  ddavies
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Add inclusion of new file rast.pub.
 *
 * Revision 1.74  1997/06/19 00:58:32  xzhu
 * xzhu on Wed Jun 18 17:57:39 PDT 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fixed bug to cause access violation. (wp_recognizeBox)
 *
 * Revision 1.73  1997/05/16 19:48:27  ddavies
 * ddavies on Fri May 16 12:47:36 PDT 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Removed unused routines w_textRatio and wp_moment (only called by
 * w_textRatio).  Implementations of these still exist in the obsolete
 * directory.
 *
 * Revision 1.72  1997/05/14 21:03:59  xzhu
 * xzhu on Wed May 14 14:03:58 PDT 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: If we just check edge to determine whether the image mask is rectangle, we may delete inside text. So we add something to check whether there is text inside to reduce segmentation errors.
 *
 * Revision 1.71  1997/04/24 23:41:06  ddavies
 * ddavies on Thu Apr 24 16:39:08 PDT 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Delete obsolete routines isText and isTextRegion.  These have been replaced
 * by a routine that finds text using textMask, then does skew detection, then
 * looks at the confidence to see if there's lots of parallel text-like stuff.
 * That's in PerfectScan somewhere.
 *
 * Revision 1.70  1997/04/15 21:56:40  ddavies
 * ddavies on Tue Apr 15 14:56:02 PDT 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Use non-initialized PIXR creation where possible.
 * Fix indentation.
 *
 * Revision 1.69  1997/03/29 01:25:45  xzhu
 * xzhu on Fri Mar 28 17:25:43 PST 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Add a input parameter IsTextSeg in w_textMask(). If want keep more text, turn on this parameter, otherwise set it 0.
 *
 * Revision 1.68  1997/03/28 17:39:59  dferrell
 * dferrell on Fri Mar 28 09:34:34 PST 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Modified wp_countEdgePixels () so that rather than using our morphology
 * package for the dilation step, it now uses run filtering which is more
 * time efficient for the large dilations often performed here.  Run
 * filtering in-place is also more memory efficient than strict dilation
 * in-place.  Changed w_maskIsRect () to reflect these changes.
 *
 * Revision 1.67  1997/01/28 00:40:17  xzhu
 * xzhu on Mon Jan 27 16:30:27 PST 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fix res. setting bugs in w_textMask and w_reversetextMask (Dan found this problem). Added some comments in w_reversetextMask. In w_textmask changed a unreasonable condition for cleaning up noise line inside the image.
 *
 * Revision 1.66  1997/01/10 01:13:36  xzhu
 * xzhu on Thu Jan 9 17:13:21 PST 1997
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * In w_textMask, add error return when dpi is zero. If dpi is zero, textMask will spend long time to run and no output results.
 *
 * Revision 1.65  1996/11/15 00:39:21  xzhu
 * xzhu on Thu Nov 14 16:36:09 PST 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * In w_textMask(), added train reconnection for the image box that will be freed.
 *
 * Revision 1.64  1996/10/23 16:29:18  xzhu
 * xzhu on Wed Oct 23 09:28:54 PDT 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fix core check warning.
 *
 * Revision 1.63  1996/10/22 19:07:31  xzhu
 * xzhu on Tue Oct 22 12:05:12 PDT 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * If input image is white page, do normal return.
 *
 * Revision 1.62  1996/10/02 20:58:39  xzhu
 * xzhu on Wed Oct 2 13:53:03 PDT 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added space to aviod confusing with html format.
 *
 * Revision 1.61  1996/07/19 15:35:40  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * w_labelColorCCs was not allowing gray images to be passed in
 *
 * Revision 1.60  1996/03/26 17:27:58  xzhu
 * xzhu on Tue Mar 26 09:27:40 PST 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fixed coreCheck warning.
 *
 * Revision 1.59  1996/03/25 23:00:56  xzhu
 * xzhu on Mon Mar 25 15:00:19 PST 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fixed PC warning.
 *
 * Revision 1.58  1996/03/25 17:29:07  xzhu
 * xzhu on Mon Mar 25 09:28:22 PST 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * fixed dpi=0 in w_textMask
 * Added new function w_reversetextMask().
 *
 * Revision 1.57  1996/02/15 17:06:30  macleod
 * macleod on Thu Feb 15 09:06:10 PST 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Modified w_labelColorCCs so that it works with grayscale images
 *
 * Revision 1.56  1996/01/22 19:11:51  xzhu
 * xzhu on Mon Jan 22 11:10:20 PST 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fix PC build error to take out memory.h and memset().
 *
 * Revision 1.55  1996/01/20  00:30:39  xzhu
 * xzhu on Fri Jan 19 16:28:47 PST 1996
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Modified w_textMask() and deleted wp_boxedAreAdjacent() .
 *
 * Revision 1.54  1995/12/13  04:09:43  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.53  1995/11/14  22:18:37  ddavies
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Fix documentation bugs found by Eric's scripts.
 *
 * Revision 1.52  1995/11/13  20:13:10  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.51  1995/10/16  22:56:49  dferrell
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.50  1995/10/13  22:02:24  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.49  1995/10/13  02:18:32  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Entered a OneLineDescription for w_orient4() which was missing one.
 *
 * Revision 1.48  1995/10/12  17:44:09  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * added explicit casts (float->int, long->short, long->char, etc) and
 * made floating constants the right size to shut up the NT compiler.
 *
 * Revision 1.47  1995/10/03  20:54:22  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Fixed keyword typo
 *
 * Revision 1.46  1995/10/03  01:59:03  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Edited comment headers. Also fixed a few inconsistencies in interfaces.
 *
 * Revision 1.45  1995/09/21  22:54:01  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.44  1995/09/20  00:06:38  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Oops--while taking out a debugging statement, also took out the statement
 * that followed the next link in a linked list, which caused it to run
 * forever, so I put it back.
 *
 * Revision 1.43  1995/09/19  19:19:12  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Made Int16s into ScaledInt1616s where appropriate; took out dpi parameters
 * where they could be inferred from PIXRs.
 *
 * Revision 1.42  1995/09/11  21:53:55  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Arg order changed, fixed arg order of called functions,
 * took out Pixr and Pr in names, dealt with thrown-away
 * functions, etc.
 *
 * Revision 1.41  1995/09/07  01:32:48  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.40  1995/09/06  20:20:49  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.39  1995/09/05  19:01:06  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Removed calls to GET_STREL macro, instead calling getStrel() function.
 * This required using local variables rather than invoking GET_STREL
 * as an argument to another function, since getStrel can return an error
 * code which should be checked.
 *
 * Revision 1.38  1995/09/01  21:02:20  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Added back in the minimum of include files.
 *
 * Revision 1.37  1995/08/31  21:14:36  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.36  1995/08/29  00:32:38  ejaquith
 * ejaquith on Mon Aug 28 17:32:07 PDT 1995
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Whoops.  Added 3 typos myself in adding the error check wrapping.  Fixed.
 *
 * Revision 1.35  1995/08/29  00:13:56  ejaquith
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * fixed a lot of bugs for "segment" target of ipwtest.
 * call to i_findConnectedComponentBoundingBoxes() was passing in the wrong
 * width and height.
 * The logic in textMask() was incorrect causing an infinite loop.
 * Tons (like 40+) of calls to IPWrap routines were not being error check
 * wrapped.  (which in turn was causing more bugs).
 *
 * Revision 1.34  1995/08/28  22:43:14  jcraig
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * get rid of pc compiler warnings
 *
 * Revision 1.33  1995/08/17  23:00:06  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Needed to comment out the #include of "shros.pub"
 *
 * Revision 1.32  1995/08/17  21:55:25  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * The Grand Code Rework, part I
 *
 * Revision 1.31  1995/06/09  01:11:26  macleod
 * macleod on Thu Jun 8 18:07:05 PDT 1995
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added labelColorCCs, which is used to determine the color under
 * some connected components and label each connected component with
 * its color (which is the mean of the interior of the cc, i.e. the
 * cc eroded, though it should be skeletonized instead). This code
 * is work-in-progress; I'm checking it in so that I have a checkpoint
 * of the state. Also added coalesceColorCCs, which computes the
 * equivalence classes of labeled connected components, in order to
 * reduce the number of colors needed to represent colored text. The
 * equivalence relation is based on color distance. I'm not conviced
 * that this is really the best method for determining the correct
 * small color map; it should probably work locally instead.
 *
 * Revision 1.30  1995/01/24  21:59:53  lperry
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Incorporate changes to use new strel manager interface.
 * Only allocate actual strels required.
 *
 * Revision 1.29  1995/01/11  23:57:10  dferrell
 * dferrell on Wed Jan 11 15:56:55 PST 1995
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Changed include of shrip.pub to shrcc.pub.
 *
 * Revision 1.28  1994/12/05  23:48:33  dferrell
 * dferrell on Mon Dec 5 15:47:42 PST 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added typecaste to CALLOC return value so that MPW C compiler could
 * understand.
 *
 * Revision 1.27  1994/11/23  00:24:46  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Silenced WATCOM warning about unreferenced parameter.
 *
 * Revision 1.26  1994/11/22  18:56:42  macleod
 * macleod on Tue Nov 22 10:56:19 PST 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * orient() and orient4() don't reduce at <= 100dpi;
 * "new improved" text mask finder
 *
 * Revision 1.25  1994/11/04  03:11:26  macleod
 * macleod on Thu Nov 3 19:10:53 PST 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * If you pass a grayscale image to textMask, it will attempt to
 * adaptively threshold it before finding the text.
 *
 * Revision 1.24  1994/10/28  01:58:33  macleod
 * macleod on Thu Oct 27 18:58:21 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Called my_sqrt instead of sqrt
 *
 * Revision 1.23  1994/10/17  20:11:58  macleod
 * macleod on Mon Oct 17 13:10:45 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * maskIsRect had a case where it could request a zero-wide STREL,
 * which caused a floating exception in one case. Pinned it to
 * 1 minimum.
 *
 * Revision 1.22  1994/10/13  17:36:35  macleod
 * macleod on Thu Oct 13 10:35:50 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Forgot a "procName" which kept the debug version from compiling.
 *
 * Revision 1.21  1994/10/12  23:31:51  macleod
 * macleod on Wed Oct 12 16:31:50 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * A better (probably a bit slower) maskIsRect function. What this one
 * does is to go along each of the four edges, and count the number of
 * pixels along the edge for which there is a "1" within 10% the edge
 * (i.e. along the perpindicular, in towards the image). The way this is
 * implemented is to cut out each of the four 10% regions, dilate them by
 * a horizontal or vertical strel perpindicular to the edge axis, count
 * the number of pixels inside, and then divide by the size of the strel.
 *
 * Revision 1.20  1994/10/11  21:46:13  macleod
 * macleod on Tue Oct 11 14:45:42 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added orient4, which does the old 4-direction orient detect, rather
 * than just vertical/horizontal
 *
 * Revision 1.19  1994/10/07  19:06:09  lperry
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Updated to use new .pub/.prv headers rather than xxpros.h
 *
 * Revision 1.18  1994/09/30  16:32:18  lperry
 * lperry on Fri Sep 30 09:31:06 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Convert to use new .pub header files. This is
 * only a partial conversion at this point until
 * all of the headers are in.
 *
 * Revision 1.17  1994/09/30  01:50:29  macleod
 * macleod on Thu Sep 29 18:45:55 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added textMask, which attempts to find a text mask from a binary image.
 * The method is to find all of the connected components, and then
 * attempt to merge them based on size and adjacency. Once everything
 * is merged, take all of the boxes which were merged from more than some
 * constant (currently 5) number of connected components, and call those
 * text. The result is a mask, with all of the text boxes painted in.
 *
 * Revision 1.16  1994/08/29  18:55:03  jcraig
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * change call to numberBitsRect to use RECTANGL and be relative to
 * upper left corner of image.
 *
 * Revision 1.15  1994/08/17  16:13:26  lperry
 * lperry on Wed Aug 17 09:12:04 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added completely redundant parentheses around assignment expresssion
 * tested for truth value to keep gcc happy.
 *
 * Revision 1.14  1994/08/16  21:48:27  macleod
 * macleod on Tue Aug 16 14:47:17 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * The strel initialization stuff had some code "if (retCode == strel1Init..."
 * which should have read "if (retCode = strel1Init..."; the ersult
 * was that the == was always true, so the routine always failed--sort
 * of. Another problem was that the routine was always returning
 * 0, even after failing; it now returns "retCode".
 *
 * Revision 1.13  1994/08/10  21:59:09  lperry
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Changes made to use new strel manager.
 *
 * Revision 1.12  1994/07/21  18:47:38  macleod
 * macleod on Thu Jul 21 11:47:03 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * As requested, auto-orient only does a vertical/horizontal test
 *
 * Revision 1.11  1994/07/12  00:47:36  macleod
 * macleod on Mon Jul 11 17:47:18 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * isTextRegion() allocated a couple of STRELS without freeing them.
 *
 * Revision 1.10  1994/07/01  01:51:04  macleod
 * macleod on Thu Jun 30 18:49:05 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Changed isTextRegion so that if the h- and v-eroded versions have a small
 * number of pixels relative to the original image, they are assumed to be
 * not text. This gets around a problem where occasional small crud in a
 * thresholded image was either vertical or horizontal, and got a high ratio.
 *
 * Revision 1.9  1994/06/21  01:50:22  macleod
 * macleod on Mon Jun 20 18:44:59 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added isTextRegion, which is yet-another-text region classifier.
 * What it does is to reduce the image down to 25-50 dpi, and then does
 * erodes with horizontal and vertical strels, counting the number of
 * bits resulting from each. If the ratio between the number of bits
 * is over about three, the image is assumed to be text. This assumes
 * that we have lots of words which turned into long, thin lines.
 * Note that this won't work for large type (it'll call it picture),
 * and unlike isText, it doesn't magically detect text on a whole page
 * unless there's a lot of it. This routine is intended to be called
 * AFTER segmentation, with a bunch of small regions, which are either
 * text-on-tint regions or images.
 *
 * Revision 1.8  1994/06/01  20:55:53  macleod
 * macleod on Wed Jun 1 13:53:33 PDT 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * After experimenting with more clever schemes for determining whether a mask
 * inside a bounding box was "close enough" to be called a rectangle, the
 * scheme I settled on was just count the pixels and compare them to the
 * number of pixels in the bounding box. Pretty stupid, huh? Well, a cutoff
 * of around 90% seems right. Anyway, the routine is called "maskIsRect".
 *
 * Revision 1.7  1994/05/05  01:41:05  ddavies
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Remove statics from declaration for moment since moment's prototype is
 * included in more than one file.  Change lots of %d's to %ld's to make
 * gcc happy.
 *
 * Revision 1.6  1994/03/24  23:06:00  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * isText and textRatio take a DPI, and are now image-size independent
 * (assuming that the DPI passed is correct)
 * NOTE that the parameters have changed.
 *
 * Revision 1.5  1994/03/10  02:04:41  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Added a bunch of IP_YIELD calls;
 * added a (#if 0'ed out) routine which was intended to be faster
 * than orient(), but is really slower. It'll probably be taken out
 * completely.
 *
 * Revision 1.4  1994/02/15  01:43:39  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * textRatio takes a rectangle
 *
 * Revision 1.3  1994/01/31  20:58:50  macleod
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * added textRatio, which is the heuristic used by isText
 *
 * Revision 1.2  1994/01/07  22:15:24  dferrell
 * dferrell on Fri Jan 7 14:14:02 PST 1994
 *
 * imginfo.c -> /product/ipcore/ipwrap/src/RCS/imginfo.c,v
 *
 * Removed unnecessary #include <malloc.h> line.
 *
 * Revision 1.1  1993/12/15  06:34:36  macleod
 * Initial revision
 *
 */

/*</PRE></BODY>*/
