/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1996, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	iw_chvq.c:	Hierarchical Vector Quantization compression wrappers:
 *
 *	Code to compress PIXRs to HVQ compressed PIXRs:
 *		w_HVQCompress()
 *
 */


#define _IW_CHVQ_C_

#include <stdio.h>
#include <math.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"

/* prototypes */
#include "props.pub"
#include "hvq.pub"

/* static HVQ LUTs */
#include "hvq.h"

IP_RCSINFO(iw_chvq_c_RCSInfo, "$RCSfile: iw_chvq.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

#if defined(macintosh) || defined(__WATCOMC__)

#define LUT_2x1(lut_2x1, i)	((lut_2x1[(i)>>15])[(i)&0x7fff])
#define LUT_2x2(lut_2x2, i)	((lut_2x2[(i)>>15])[(i)&0x7fff])
#define LUT_4x2(lut_4x2, i)	((lut_4x2[(i)>>15])[(i)&0x7fff])
#define LUT_4x4(lut_4x4, i)	((lut_4x4[(i)>>15])[(i)&0x7fff])

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

#define HVQE_2x1(vec, lut_2x1)	(LUT_2x1(lut_2x1, ((vec)[0] << 8) | (vec)[1]))
#define HVQE_2x2(vec, bpl, lut_2x1, lut_2x2)	\
	(LUT_2x2(lut_2x2, ((HVQE_2x1(vec,lut_2x1) << 8) | HVQE_2x1((vec)+bpl,lut_2x1))))
#define HVQE_4x2(vec, bpl, lut_2x1, lut_2x2, lut_4x2)	\
	(LUT_4x2(lut_4x2, ((HVQE_2x2(vec,bpl,lut_2x1,lut_2x2) << 8) | \
		  HVQE_2x2((vec)+2,bpl,lut_2x1,lut_2x2))))
#define HVQE_4x4(vec, bpl, lut_2x1, lut_2x2, lut_4x2, lut_4x4)	\
	(LUT_4x4(lut_4x4, ((HVQE_4x2(vec,bpl,lut_2x1,lut_2x2,lut_4x2) << 8) | \
		  HVQE_4x2((vec)+2*(bpl),bpl,lut_2x1,lut_2x2,lut_4x2))))

#else /* big endian */

#define HVQE_2x1_PC(vec, lut_2x1)	(LUT_2x1(lut_2x1, ((vec)[1] << 8) | (vec)[0]))
#define HVQE_2x2_PC(vec, bpl, lut_2x1, lut_2x2)	\
	(LUT_2x2(lut_2x2, ((HVQE_2x1_PC(vec,lut_2x1) << 8) | \
		  HVQE_2x1_PC((vec)+bpl,lut_2x1))))
#define HVQE_4x2_PC(vec, bpl, lut_2x1, lut_2x2, lut_4x2)	\
	(LUT_4x2(lut_4x2, ((HVQE_2x2_PC((vec)+2,bpl,lut_2x1,lut_2x2) << 8) | \
		  HVQE_2x2_PC(vec,bpl,lut_2x1,lut_2x2))))
#define HVQE_4x4_PC(vec, bpl, lut_2x1, lut_2x2, lut_4x2, lut_4x4)	\
	(LUT_4x4(lut_4x4, ((HVQE_4x2_PC(vec,bpl,lut_2x1,lut_2x2,lut_4x2) << 8) | \
		  HVQE_4x2_PC((vec)+2*(bpl),bpl,lut_2x1,lut_2x2,lut_4x2))))

#endif /* endian */

#else /* macintosh */

#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

#define HVQE_2x1(vec, lut_2x1)	(lut_2x1[((vec)[0] << 8) | (vec)[1]])
#define HVQE_2x2(vec, bpl, lut_2x1, lut_2x2)	\
	(lut_2x2[((HVQE_2x1(vec,lut_2x1) << 8) | HVQE_2x1((vec)+bpl,lut_2x1))])
#define HVQE_4x2(vec, bpl, lut_2x1, lut_2x2, lut_4x2)	\
	(lut_4x2[((HVQE_2x2(vec,bpl,lut_2x1,lut_2x2) << 8) | \
		  HVQE_2x2((vec)+2,bpl,lut_2x1,lut_2x2))])
#define HVQE_4x4(vec, bpl, lut_2x1, lut_2x2, lut_4x2, lut_4x4)	\
	(lut_4x4[((HVQE_4x2(vec,bpl,lut_2x1,lut_2x2,lut_4x2) << 8) | \
		  HVQE_4x2((vec)+2*(bpl),bpl,lut_2x1,lut_2x2,lut_4x2))])

#else /* ! big endian */

#define HVQE_2x1(vec, lut_2x1)	(lut_2x1[((vec)[0] << 8) | (vec)[1]])
#define HVQE_2x2(vec, bpl, lut_2x1, lut_2x2)	\
	(lut_2x2[((HVQE_2x1(vec,lut_2x1) << 8) | HVQE_2x1((vec)+bpl,lut_2x1))])
#define HVQE_4x2(vec, bpl, lut_2x1, lut_2x2, lut_4x2)	\
	(lut_4x2[((HVQE_2x2(vec,bpl,lut_2x1,lut_2x2) << 8) | \
		  HVQE_2x2((vec)+2,bpl,lut_2x1,lut_2x2))])
#define HVQE_4x4(vec, bpl, lut_2x1, lut_2x2, lut_4x2, lut_4x4)	\
	(lut_4x4[((HVQE_4x2(vec,bpl,lut_2x1,lut_2x2,lut_4x2) << 8) | \
		  HVQE_4x2((vec)+2*(bpl),bpl,lut_2x1,lut_2x2,lut_4x2))])

#define HVQE_2x1_PC(vec, lut_2x1)	(lut_2x1[((vec)[1] << 8) | (vec)[0]])
#define HVQE_2x2_PC(vec, bpl, lut_2x1, lut_2x2)	\
	(lut_2x2[((HVQE_2x1_PC(vec,lut_2x1) << 8) | \
		  HVQE_2x1_PC((vec)+bpl,lut_2x1))])
#define HVQE_4x2_PC(vec, bpl, lut_2x1, lut_2x2, lut_4x2)	\
	(lut_4x2[((HVQE_2x2_PC((vec)+2,bpl,lut_2x1,lut_2x2) << 8) | \
		  HVQE_2x2_PC(vec,bpl,lut_2x1,lut_2x2))])
#define HVQE_4x4_PC(vec, bpl, lut_2x1, lut_2x2, lut_4x2, lut_4x4)	\
	(lut_4x4[((HVQE_4x2_PC(vec,bpl,lut_2x1,lut_2x2,lut_4x2) << 8) | \
		  HVQE_4x2_PC((vec)+2*(bpl),bpl,lut_2x1,lut_2x2,lut_4x2))])

#endif /* endian */

#endif /* macintosh */


#if 0
#define hvqe_2x1(vec, lut_2x1)	(lut_2x1[((vec)[0] << 8) | (vec)[1]])


static UInt32
hvqe_2x2(
	UInt8 *vec,
	UInt32 bpl,
	UInt8 *lut_2x1,
	UInt8 *lut_2x2
)
{
    UInt32 top_index, bottom_index;
    UInt32 index_2x2;

    top_index = lut_2x1[(vec[0] << 8) | vec[1]];
    vec += bpl;
    bottom_index = lut_2x1[(vec[0] << 8) | vec[1]];

    index_2x2 = lut_2x2[(top_index << 8) | bottom_index];

    return index_2x2;
}


static UInt32
hvqe_4x2(
	UInt8 *vec,
	UInt32 bpl,
	UInt8 *lut_2x1,
	UInt8 *lut_2x2,
	UInt8 *lut_4x2
)
{
    UInt32 left_index, right_index;
    UInt32 index_4x2;

    left_index = hvqe_2x2(vec, bpl, lut_2x1, lut_2x2);
    right_index = hvqe_2x2(vec + 2, bpl, lut_2x1, lut_2x2);

    index_4x2 = lut_4x2[(left_index << 8) | right_index];

    return index_4x2;
}



static UInt32
hvqe_4x4(
	UInt8 *vec,
	UInt32 bpl,
	UInt8 *lut_2x1,
	UInt8 *lut_2x2,
	UInt8 *lut_4x2,
	UInt8 *lut_4x4
)
{
    UInt32 top_index, bottom_index;
    UInt32 index_4x4;

    top_index = hvqe_4x2(vec, bpl, lut_2x1, lut_2x2, lut_4x2);
    bottom_index = hvqe_4x2(vec + 2 * bpl, bpl, lut_2x1, lut_2x2, lut_4x2);

    index_4x4 = lut_4x4[(top_index << 8) | bottom_index];

    return index_4x4;
}
#endif



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  HVQ Compress gray (or single color channel) 8-bpp image.
 *
 * LongDescription:::
 *  HVQ compress an 8-bpp gray image or single color channel from one PIXR
 *  and store the result as an 8-bpp PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPIXR::        Source 8-bit single channel PIXR to be compressed
 *  hvqLevels::      Number of the four HVQ levels to use. Note: the
 *                   compression achieved per level is: 2:1, 4:1, 8:1, & 16:1
 *
 * Output:::
 *  dstPIXR::       Preallocated destination 8-bit PIXR after compression.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  HVQ
 *
 * Keywords:::
 *  HVQ, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hvqe_channel(
	    IPWState *pIPWState,
	    PIXR *srcPIXR,
	    Int32 hvqLevels,
	    PIXR *dstPIXR
	    )
{
    register UInt8	*vec;		/* Pointer to source vector */
    register UInt32	index;		/* Codebook index resulting from HVQ */
    register Int32	xreduction;	/* Reduction factor in X direction */
#if _ALPACA_IMAGE_FMT_ != cBigEndianFmt
    register Int32	indices;	/* Number of indices buffered up */
#endif
    register Int32	yreduction;	/* Reduction factor in Y direction */
    Int32	src_bpl;	/* Source bytes/line */
    Int32	src_width;	/* Source width */
    Int32	src_height;	/* Source height */
    Int32	src_depth;	/* Source depth */
    UInt8	*src_swath;	/* Source image swath pointer */
    Int32	dst_bpl;	/* Destination bytes/line */
    Int32	dst_width;	/* Destination width */
    Int32	dst_height;	/* Destination height */
    Int32	dst_depth;	/* Destination depth */
    Int32	dst_frame;	/* Destination frame pixels */
    UInt32	vecs_down;	/* How many vectors down the image */
    UInt32	vecs_across;	/* How many vectors along the horizontal */
    UInt32	vec_num;	/* Which vector along the horizontal */
    UInt32	row;		/* Which row */
    UInt8	*dst_row;	/* Pointer to destination row */
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
    UInt8	*dst_ptr;	/* Pointer to item in destination row */
#else
    register UInt32	big_index=0UL;	/* Buffered 32-bit output buffer */
    UInt32	*dst_longptr;	/* Output buffer pointer */
#endif


    PROLOGUE;
    ERR_PASS(w_pixrGetBpl(pIPWState, srcPIXR, &src_bpl));
    ERR_PASS(w_pixrGetWidth(pIPWState, srcPIXR, &src_width));
    ERR_PASS(w_pixrGetHeight(pIPWState, srcPIXR, &src_height));
    ERR_PASS(w_pixrGetDepth(pIPWState, srcPIXR, &src_depth));
    ERR_PASS(w_pixrGetImage(pIPWState, srcPIXR, &src_swath));

    ERR_PASS(w_pixrGetBpl(pIPWState, dstPIXR, &dst_bpl));
    ERR_PASS(w_pixrGetWidth(pIPWState, dstPIXR, &dst_width));
    ERR_PASS(w_pixrGetHeight(pIPWState, dstPIXR, &dst_height));
    ERR_PASS(w_pixrGetDepth(pIPWState, dstPIXR, &dst_depth));
    ERR_PASS(w_pixrGetImage(pIPWState, dstPIXR, &dst_row));
    ERR_PASS(w_pixrGetFrame(pIPWState, dstPIXR, &dst_frame));

    /* Prepare reduction factors based on the number of HVQ levels */
    switch (hvqLevels) {
    case 1:
	xreduction = 2, yreduction = 1;
	break;
    case 2:
	xreduction = 2, yreduction = 2;
	break;
    case 3:
	xreduction = 4, yreduction = 2;
	break;
    case 4:
	xreduction = 4, yreduction = 4;
	break;
    default:
	ERR_RETURN;
    }
#if 1
    vecs_across = (src_width + 3) / xreduction;
#else
    vecs_across = (src_width + xreduction - 1) / xreduction;
#endif
    vecs_down = (src_height + yreduction - 1) / yreduction;

    /* Destination size checking */
    ERR_PASS( dst_depth != 8);
    ERR_PASS( dst_width + dst_frame < vecs_across);
    ERR_PASS( dst_height < vecs_down);

    /* Loop according to compression ratio (number of HVQ levels) */
    switch (hvqLevels) {
    case 1:
	/* Loop through every row of vectors */
	for (row = 0; row < vecs_down; row++, src_swath += src_bpl * yreduction,
	     dst_row += dst_bpl) {
    
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    /* Loop through the vectors within the swath */
	    dst_ptr = dst_row;
	    vec = src_swath;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_2x1(vec, hvqelut_2x1_1);
	    
		/* Buffer out the index */
		*dst_ptr++ = (UInt8)index;

		/* Go to the next vector */
		vec += xreduction;
	    } /* For every vector in the row */

#else /* ! big endian */

	    /* Loop through the vectors within the swath */
	    dst_longptr = (UInt32 *)dst_row;
	    indices = 0;
	    vec = src_swath + 2;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_2x1_PC(vec, hvqelut_2x1_1);
	    
		/* Buffer out the index */
		big_index = (big_index << 8) | index;
		if (++indices == 4) {
		    *dst_longptr++ = big_index;
		    indices = 0;
		}

		/* Go to the next vector */
		if (vec_num & 1) {
		    vec += 6;
		} else {
		    vec -= 2;
		}
	    } /* For every vector in the row */

	    if (indices > 0) {
		big_index <<= 8 * (4 - indices);
		*dst_longptr = big_index;
	    }

#endif /* endian */

	} /* For every row of vectors in the image */
	break;
    case 2:
	/* Loop through every row of vectors */
	for (row = 0; row < vecs_down; row++, src_swath += src_bpl * yreduction,
	     dst_row += dst_bpl) {
    
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

	    /* Loop through the vectors within the swath */
	    dst_ptr = dst_row;
	    vec = src_swath;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_2x2(vec, src_bpl, hvqelut_2x1_1, hvqelut_2x2_1);
	    
		/* Buffer out the index */
		*dst_ptr++ = (UInt8)index;

		/* Go to the next vector */
		vec += xreduction;
	    } /* For every vector in the row */

#else /* ! big endian */

	    /* Loop through the vectors within the swath */
	    dst_longptr = (UInt32 *)dst_row;
	    indices = 0;
	    vec = src_swath + 2;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_2x2_PC(vec, src_bpl, hvqelut_2x1_1, hvqelut_2x2_1);
	    
		/* Buffer out the index */
		big_index = (big_index << 8) | index;
		if (++indices == 4) {
		    *dst_longptr++ = big_index;
		    indices = 0;
		}

		/* Go to the next vector */
		if (vec_num & 1) {
		    vec += 6;
		} else {
		    vec -= 2;
		}
	    } /* For every vector in the row */

	    if (indices > 0) {
		big_index <<= 8 * (4 - indices);
		*dst_longptr = big_index;
	    }

#endif /* endian */

	} /* For every row of vectors in the image */
	break;
    case 3:
	/* Loop through every row of vectors */
	for (row = 0; row < vecs_down; row++, src_swath += src_bpl * yreduction,
	     dst_row += dst_bpl) {
    
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt

	    /* Loop through the vectors within the swath */
	    dst_ptr = dst_row;
	    vec = src_swath;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_4x2(vec, src_bpl, hvqelut_2x1_1, hvqelut_2x2_1,
				hvqelut_4x2_1);
	    
		/* Buffer out the index */
		*dst_ptr++ = (UInt8)index;

		/* Go to the next vector */
		vec += xreduction;
	    } /* For every vector in the row */

#else /* ! big endian */

	    /* Loop through the vectors within the swath */
	    dst_longptr = (UInt32 *)dst_row;
	    indices = 0;
	    vec = src_swath;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_4x2_PC(vec, src_bpl, hvqelut_2x1_1, hvqelut_2x2_1,
				    hvqelut_4x2_1);
	    
		/* Buffer out the index */
		big_index = (big_index << 8) | index;
		if (++indices == 4) {
		    *dst_longptr++ = big_index;
		    indices = 0;
		}

		/* Go to the next vector */
		vec += 4;
	    } /* For every vector in the row */

	    if (indices > 0) {
		big_index <<= 8 * (4 - indices);
		*dst_longptr = big_index;
	    }

#endif /* endian */

	} /* For every row of vectors in the image */
	break;
    case 4:
	/* Loop through every row of vectors */
	for (row = 0; row < vecs_down; row++, src_swath += src_bpl * yreduction,
	     dst_row += dst_bpl) {


#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
	    
	    /* Loop through the vectors within the swath */
	    dst_ptr = dst_row;
	    vec = src_swath;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_4x4(vec, src_bpl, hvqelut_2x1_1, hvqelut_2x2_1,
				hvqelut_4x2_1, hvqelut_4x4_1);
	    
		/* Buffer out the index */
		*dst_ptr++ = (UInt8)index;

		/* Go to the next vector */
		vec += xreduction;
	    } /* For every vector in the row */

#else /* ! big endian */

#if 1

	    /* Loop through the vectors within the swath */
	    dst_longptr = (UInt32 *)dst_row;
	    indices = 0;
	    vec = src_swath;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_4x4_PC(vec, src_bpl, hvqelut_2x1_1, hvqelut_2x2_1,
				    hvqelut_4x2_1, hvqelut_4x4_1);
	    
		/* Buffer out the index */
		big_index = (big_index << 8) | index;
		if (++indices == 4) {
		    *dst_longptr++ = big_index;
		    indices = 0;
		}

		/* Go to the next vector */
		vec += 4;
	    } /* For every vector in the row */

	    if (indices > 0) {
		big_index <<= 8 * (4 - indices);
		*dst_longptr = big_index;
	    }
#else

	    /* Loop through the vectors within the swath */
	    dst_longptr = (UInt32 *)dst_row;
	    vec = src_swath;
	    for (vec_num = 0; vec_num < vecs_across; vec_num++) {

		/* Use the series of quantizations to traverse the vector and
		 * encode */
		index = HVQE_4x4_PC(vec, src_bpl, hvqelut_2x1_1, hvqelut_2x2_1,
				    hvqelut_4x2_1, hvqelut_4x4_1);
	    
		/* Buffer out the index */
		big_index = (big_index << 8) | index;
		if (++indices == 4) {
		    *dst_longptr++ = big_index;
		    indices = 0;
		}

		/* Go to the next vector */
		vec += xreduction;
	    } /* For every vector in the row */

	    if (indices > 0) {
		while (indices < 4) {
		    big_index <<= 8;
		    indices++;
		}
		*dst_longptr = big_index;
	    }
#endif

#endif /* endian */

	} /* For every row of vectors in the image */
	break;
    default:
	ERR_RETURN;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  HVQ Compress gray or RGB PIXR; store in preallocated PIXR.
 *
 * LongDescription:::
 *  Support function for w_HVQCompress(), which allocates the appropriate
 *  PIXR, and then calls this routine. This routine compresses each channel
 *  of the PIXR, and stores the 8-bit HVQ indices in the corresponding
 *  output PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPIXR::        Source 8-bit PIXR to be compressed (gray or RGB)
 *  hvqLevels::      Number of the four HVQ levels to use. Note: the
 *                   compression achieved per level is: 2:1, 4:1, 8:1, & 16:1
 *
 * Output:::
 *  dstPIXR::        Destination 8-bit gray or RGB PIXR after compression.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  HVQ
 *
 * Keywords:::
 *  HVQ, compression, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_hvqe_image(
	  IPWState *pIPWState,
	  PIXR *srcPIXR,
	  Int32 hvqLevels,
	  PIXR *dstPIXR
	  )
{
    Int32	channel, channels;
    PIXR	*src_ch_pixr;
    PIXR	*dst_ch_pixr;

    PROLOGUE;
    ERR_PASS(w_pixrGetChannelCount(pIPWState, srcPIXR, &channels));
    {
	Int32 dst_ch_pixrannels;
	ERR_PASS(w_pixrGetChannelCount(pIPWState, dstPIXR, &dst_ch_pixrannels));
	ERR_PASS(dst_ch_pixrannels != channels);
    }

    for (channel = 0; channel < channels; channel++) {

	/* Get the PIXR for that channel */
	ERR_PASS(w_pixrGetChannel(pIPWState, srcPIXR, channel, &src_ch_pixr));
	ERR_PASS(w_pixrGetChannel(pIPWState, dstPIXR, channel, &dst_ch_pixr));

	/* HVQ that channel */
	ERR_PASS(wp_hvqe_channel(pIPWState, src_ch_pixr, hvqLevels, dst_ch_pixr));
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  HVQ Compress gray or RGB PIXR.
 *
 * LongDescription:::
 *  Compress a gray image in one (or RGB images in three PIXRs) to an (a
 *  set of) output PIXR(s) containing 8-bit HVQ indices. Note that non
 *  destructive image operations may be performed on the output PIXR(s),
 *  like rotation, pixel subsampling, pixel replication, etc.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPIXR::        Source 8-bit PIXR to be compressed (gray or RGB)
 *  hvqLevels::      Number of the four HVQ levels to use. Note: the
 *                   compression achieved per level is: 2:1, 4:1, 8:1, & 16:1
 *
 * Output:::
 *  dstPIXRptr::     Pointer to destination 8-bit PIXR pointer after compression
 *                   (gray or RGB), which is generated and returned.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  HVQ
 *
 * Keywords:::
 *  HVQ, compression, file io, color, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_HVQCompress (
	       IPWState		*pIPWState,
	       PIXR		*srcPIXR,
	       Int32		hvqLevels,
	       PIXR		**dstPIXRptr)
{
    Int32	src_width;	/* Source width */
    Int32	src_height;	/* Source height */
    Int32	src_depth;	/* Source depth */
    ColorType	src_type;	/* Source color type */
    ScaledInt1616  src_xdpi;	/* Source X resolution in dpi */
    ScaledInt1616  src_ydpi;	/* Source Y resolution in dpi */
    PIXR	*dstPixr;	/* Destination PIXR */
    Int32	dst_width;	/* Destination width */
    Int32	dst_height;	/* Destination height */
    Int32	dst_depth;	/* Destination depth */
    ColorType	dst_type;	/* Destination color type */
    ScaledInt1616  dst_xdpi;	/* Destination X resolution in dpi */
    ScaledInt1616  dst_ydpi;	/* Destination Y resolution in dpi */
    Int32	channels;	/* Number of channels in PIXR */
    UInt32	size;		/* Size of input PIXR */
    UInt32	outsize;	/* Size of output PIXR */
    Int32	xreduction;	/* Reduction factor in X direction */
    Int32	yreduction;	/* Reduction factor in Y direction */
    
    PROLOGUE;

    /* Default to return NULL */
    *dstPIXRptr = NULL;

/* Verify arguments */
    ERR_GEN_PIXR(srcPIXR, cByteOrderBig, cDepth8,
		 cChannels1 | cChannels3, cFrame32);

    ERR_PASS(w_pixrGetChannelCount(pIPWState, srcPIXR, &channels));

/* Determine the image size, start address and line pitch */
/* We'll assume here that all of the components are the same */
    ERR_PASS(w_pixrGetWidth(pIPWState, srcPIXR, &src_width));
    ERR_PASS(w_pixrGetHeight(pIPWState, srcPIXR, &src_height));
    ERR_PASS(w_pixrGetDepth(pIPWState, srcPIXR, &src_depth));
    ERR_PASS(w_pixrGetType(pIPWState, srcPIXR, &src_type));
    ERR_PASS(w_pixrGetResolutionDPI(pIPWState, srcPIXR, &src_xdpi, &src_ydpi));

/* Verify that this is an 8-bit image */
    ERR_GEN(! (src_depth==8), cErrDepthNotSupported);

/* Calculate compressed image size */
    switch (hvqLevels) {
    case 1:
	xreduction = 2, yreduction = 1;
	break;
    case 2:
	xreduction = 2, yreduction = 2;
	break;
    case 3:
	xreduction = 4, yreduction = 2;
	break;
    case 4:
	xreduction = 4, yreduction = 4;
	break;
    default:
	ERR_RETURN;
    }
    dst_width = (src_width + xreduction - 1) / xreduction;
    dst_height = (src_height + yreduction - 1) / yreduction;
    dst_depth = 8;
    dst_type = src_type;
    dst_xdpi = ScaledInt1616_TO_Int32(src_xdpi) / xreduction;
    dst_ydpi = ScaledInt1616_TO_Int32(src_ydpi) / yreduction;
 
/* Create a PIXR for the output HVQ image */
    ERR_PASS_CLEANUP(w_createPixrExtendedNoInit(
					pIPWState,
					dst_width, dst_height,
					dst_depth,
					Int32_TO_ScaledInt1616(dst_xdpi),
					Int32_TO_ScaledInt1616(dst_ydpi),
					32, channels, dst_type, NULL,
					&dstPixr));

/* Compress source PIXR into destination PIXR */
    ERR_PASS_CLEANUP(wp_hvqe_image(pIPWState, srcPIXR, hvqLevels, dstPixr));

    size = src_width * src_height * channels;
    outsize = dst_width * dst_height * channels;

    *dstPIXRptr = dstPixr;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (dstPixr) {
	w_destroyPixr(pIPWState, &dstPixr);
    }
	
    ERR_RETURN;
}


#undef _IW_CHVQ_C_

/*
$Log: iw_chvq.c,v $
Revision 1.1.1.1  1999/09/20 18:19:18  jlin
Imported sources

Revision 1.3  1997/04/15 22:12:32  ddavies
iw_chvq.c -> /product/ipcore/ipwrap/src/RCS/iw_chvq.c,v

Use non-initialized PIXR creation where possible.

Revision 1.2  1996/11/23 00:11:45  aber
aber on Fri Nov 22 16:11:45 PST 1996

iw_chvq.c -> /product/ipcore/ipwrap/src/RCS/iw_chvq.c,v

CommonCheck: Fix comments to make coreCheckC happy. Update halftoning for portability.

Revision 1.1  1996/11/21 19:20:02  aber
Initial revision

*/
