/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	iw_cwave.c:	WAVElet compression wrappers:
 *
 *	Code to compress PIXRs to WAVElet compressed files:
 *		w_waveCompress()
 *
 */


#define _IW_CWAVE_C_

#include <stdio.h>
#include <math.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imageref.pub"
#include "defines.pub"
#include "memory.pub"

/* prototypes */
#include "shrpixr.pub"
#include "props.pub"
#include "wavelib.pub"
#include "iw_wave.pub"
#include "iw_wave.prv"

IP_RCSINFO(iw_cwave_c_RCSInfo, "$RCSfile: iw_cwave.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

Int32 f_lowpass[5] = {27941, 12366, -3624, -781, 1239};
Int32 f_highpass[5] = {25837, -13700, -1333, 2114, 0};

extern Int32 x_scan[256], y_scan[256];

UInt8 saveMask[9] = {0, 1, 3, 7, 15, 31, 63, 127, 255};

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get quatization information.
 *
 * LongDescription:::
 *    After wave transform, we find max min value for quatization.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  ration::         Compression ratio
 *
 * Output:::
 *  coefbuf::        Ptr to coef-buffer
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_waveQuatization (
    IPWState           *pIPWState,
    WAVESUBHEAD        *pheader,
    Int16              **coefbuf,
    Int32               ration)
{
    PROLOGUE;
    Int16       *coef[3], *ptmp;
    Int32       chan, row, col, level, bpl;
    Int32       coefh, coefw, halfh, halfw;
    Int16       left, right;

    pIPWState = pIPWState;
    ration = ration;
    for(chan=0; chan<pheader->ncomps; ++chan)
    {
	coef[chan] = *(coefbuf+chan);
	bpl = pheader->coefw[chan][0];

	level = pheader->pyramidN[chan]-1;
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;

	left = 0x7fff;
	for(row=0, ptmp=coef[chan]; row<halfh; ++row, ptmp+=bpl)
	{
	    for(col=0; col<halfw; ++col)
	    {
		if(*(ptmp+col) < left) left = *(ptmp+col);
	    }
	}

	pheader->mindc[chan] = left;
	if(left > 0)
	{
	    for(row=0, ptmp=coef[chan]; row<halfh; ++row, ptmp+=bpl)
	    {
		for(col=0; col<halfw; ++col)
		    *(ptmp+col) -= left;
 	    }
	}

	right =0;
	for(row=0, ptmp=coef[chan]; row<halfh; ++row, ptmp+=bpl)
	{
	    for(col=halfw; col<coefw; ++col)
		if(abs(*(ptmp+col)) > right) right = abs(*(ptmp+col));
	}
	for( ; row<coefh; ++row, ptmp+=bpl)
	{
	    for(col=0; col<coefw; ++col)
		if(abs(*(ptmp+col)) > right) right = abs(*(ptmp+col));
	}

	left =1;
	while((right >>= 1) >0) left <<=1; 	
	pheader->quant[chan] = left;
/*	fprintf(stderr, " Threshold %d ", pheader->quant[chan]);  */
    }
/*    fprintf(stderr, "\n");  */

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Calulate stop point distributions.
 *
 * LongDescription:::
 *   During coding, all of pixel under stop point will be deleted. This reduces
 *   code size a lot. Stop point is the point whose expension pixel in 
 *   following layer are zero.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  coefbuf::        Ptr to coef-buffer
 *  chan::           Channel number
 *
 * Output:::
 *  code::           Ptr to stop point buff.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_rootValue(
    IPWState           *pIPWState,
    WAVESUBHEAD        *pheader,
    Int16              *coefbuf,
    Int32               chan,
    Int16              *code)
{
    PROLOGUE;
    Int16    *coef, *root, value, *coef1, *root1, *root2;
    Int32    bpl, level, rootBpl;
    Int32    coefw, coefh, halfw, halfh;
    Int32    row, col, areaId, row1, col1;
    Int32    nextw, nexth, hw, hh;

    pIPWState = pIPWState;

    bpl = pheader->coefw[chan][0];
    rootBpl = pheader->coefw[chan][1];

    for(level=1; level<pheader->pyramidN[chan]; ++level)
    {
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;
	nextw = pheader->coefw[chan][level-1];
	nexth = pheader->coefh[chan][level-1];
	hw = (nextw+1)>>1;
	hh = (nexth+1)>>1;
	areaId =1;
	coef = coefbuf;
	root = code;
	root1 = code;
	for(row= 0, row1=0; row< halfh; row++, row1+=2, 
		root += rootBpl, coef += bpl*2, root1 += rootBpl*2)
	{
	    coef1 = coef + bpl;
	    root2 = root1 + rootBpl;
	    if((row1+1) >= hh )
	    {
		coef1=coef; 
		root2 = root1;
	    }
	    for(col= halfw, col1 =hw; col < coefw; ++col, col1 +=2)
	    {
		value = abs(coef[col1]) | abs(coef1[col1]);
		if((col1+1) < nextw)
		    value |= abs(*(coef+col1+1)) | abs(*(coef1+col1+1)); 
		root[col] = value;
		if( level > 1)
		{
		    value = root1[col1] | root2[col1];
		    if((col1+1) < nextw)
			value |= root1[col1+1] | root2[col1+1];
		    root[col] |= value;
		}
	    }
	}
	areaId =2;
	coef = coefbuf + hh * bpl;
	root = code + halfh * rootBpl;
	root1 = code + hh*rootBpl;
	for(row= halfh, row1=hh; row< coefh; row++, row1+=2, 
		root += rootBpl, coef += bpl*2, root1 += rootBpl*2)
	{
	    coef1 = coef + bpl;
	    root2 = root1 + rootBpl;
	    if((row1+1) >= nexth )
	    {
		coef1=coef; 
		root2 = root1;
	    }
	    for(col= 0, col1 =0; col < halfw; ++col, col1 +=2)
	    {
		value = abs(coef[col1]) | abs(coef1[col1]);
		if((col1+1) < hw)
		    value |= abs(*(coef+col1+1)) | abs(*(coef1+col1+1)); 
		root[col] = value;
		if( level > 1)
		{
		    value = root1[col1] | root2[col1];
		    if((col1+1) < hw)
			value |= root1[col1+1] | root2[col1+1];
		    root[col] |= value;
		}
	    }
	}
	areaId =3;
	coef = coefbuf + hh*bpl;
	root = code + halfh * rootBpl;
	root1 = code + hh*rootBpl;
	for(row= halfh, row1=hh; row< coefh; row++, row1+=2, 
		root += rootBpl, coef += bpl*2, root1 += rootBpl*2)
	{
	    coef1 = coef + bpl;
	    root2 = root1 + rootBpl;
	    if((row1+1) >= nexth )
	    {
		coef1=coef; 
		root2 = root1;
	    }
	    for(col= halfw, col1 =hw; col < coefw; ++col, col1 +=2)
	    {
		value = abs(coef[col1]) | abs(coef1[col1]);
		if((col1+1) < nextw)
		    value |= abs(*(coef+col1+1)) | abs(*(coef1+col1+1)); 
		root[col] = value;
		if( level > 1)
		{
		    value = root1[col1] | root2[col1];
		    if((col1+1) < nextw)
			value |= root1[col1+1] | root2[col1+1];
		    root[col] |= value;
		}
	    }
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set a passing flag for zero region coefits.
 *
 * LongDescription:::
 *   Set passing flag to coefits in zero region. Once we meet this flag,
 *   we will pass this coefit.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  chan::           Channel number
 *  y::              Row number
 *  x::              Col. number
 *  level::          Lay number
 *  areaId::         Area number
 *
 * Output:::
 *  root::           Ptr pixel flag buff.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_setDeletedCoef(
    IPWState           *pIPWState,
    WAVESUBHEAD        *pheader,
    Int32              chan,
    Int32              y,
    Int32              x,
    Int32              level,
    Int16              areaId,
    UInt8              *root)
{
    PROLOGUE;
    UInt8    *proot;
    Int32    bpl;
    Int32    sum, begRow, endRow, begCol, endCol;
    Int32    coefw, coefh, halfw, halfh, fw, fh;
    Int32    row, col;

    pIPWState = pIPWState;

    bpl = pheader->coefw[chan][0];
    begRow = y;
    begCol = x;
    halfw = (pheader->coefw[chan][level]+1)>>1;
    halfh = (pheader->coefh[chan][level]+1)>>1;
    for(--level,sum=2; level>=0; --level, sum*=2)
    {
	fw = halfw;
	fh = halfh;
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;
 
	switch(areaId)
	{
	case 1:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow *= 2;
	    endRow = MIN((begRow + sum), halfh);
	    endCol = MIN((begCol + sum), coefw);
	    break;
	case 2:
	    begRow = (begRow - fh)*2 +halfh;
	    begCol *= 2;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), halfw);
	    break;
	case 3:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow = (begRow - fh)*2 +halfh;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), coefw);
	    break;
	default:
	    ERR_RETURN;
	}
	proot = root + begRow*bpl + begCol;
	col = endCol-begCol;
	if( col >0)
	{
	    for(row=begRow; row<endRow; ++row, proot+=bpl)
	    {
		MEMSET((void *) proot, 1, col);
	    }
	}
    }
 
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Transform symbols to codes. 
 *
 * LongDescription:::
 *    Transform symbols to codes according to coding principle.
 *
 * Input:::
 *  pIPWState::	    Ptr to IPCore state
 *  symbol::        To be coding
 *  psymbol::       Ptr to symbol coding structure     
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 ****************************************************************************/
ErrCode CDECL
wp_symbolCoding (
    IPWState		*pIPWState,
    UInt8                symbol, 
    SYMBOLSAVE           *psymbol)
{
    PROLOGUE;
    Int16      idx;
    UInt8      code;

    if(symbol == 0x0a)
    {
	code = 1;
	idx = 1;
    }
    else
    {
	code = (symbol & 0x0f)<<1;
	idx = 5;
    }

    if((psymbol->saveBits + idx) <= 8)
    {
	psymbol->saveByte |= (code << psymbol->saveBits);
	psymbol->saveBits += idx;
	if(psymbol->saveBits >= 8)
	{
	    psymbol->WriteCallback(pIPWState,
				   psymbol->passData, &psymbol->saveByte, 1);
	    psymbol->saveByte =0;
	    psymbol->saveBits =0;
	    ++ psymbol->byte_sum;
	}
    }
    else
    {
	psymbol->saveByte |= ((code & saveMask[8-psymbol->saveBits])
			      << psymbol->saveBits);
	psymbol->WriteCallback(pIPWState,
			       psymbol->passData, &psymbol->saveByte, 1);
	++ psymbol->byte_sum;
	psymbol->saveByte = code >> (8 - psymbol->saveBits);
	psymbol->saveBits = idx + psymbol->saveBits -8;
    }

     NORMAL_RETURN;
} 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Symbolize root of tree.
 *
 * LongDescription:::
 *    Symbolize root of tree.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  coef::           Ptr to coef-buffer
 *  map::            Ptr to stop point buff
 *  x::              Col. number
 *  width::          Area width
 *  thre::           Quantization parameter
 *
 * Output:::
 *  byte::           output symbol
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 ****************************************************************************/
ErrCode CDECL
wp_topCoding (
    IPWState           *pIPWState,
    Int16              *coef,
    Int16              *map,
    Int32              x,
    Int32              width,
    Int16              thre,
    UInt8              *byte)
{
    PROLOGUE;
    Int16        pixel, root;

    pIPWState = pIPWState;
    pixel = *(coef + x);
    root = *(map + x);
    *byte =0;
    if( (abs(pixel) & thre) )
    {
	*byte = 1;
	if(pixel < 0) *byte = 3;
    }
    else if((root & thre)==0)
	*byte = 2;

    if( (x+1) < width)
    {
	pixel = *(coef + x +1);
	root = *(map + x +1);
	if( (abs(pixel) & thre) )
	{
	    *byte |= 4;
	    if(pixel < 0) *byte |= 12;
	}
	else if((root & thre) == 0)
	    *byte |= 8;
    }
    else *byte |= 8;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding on the zone.
 *
 * LongDescription:::
 *  Do huffman coding in the assigned zone.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  coefbuf::        Ptr to coef-buffer
 *  root::           Ptr to pixel flag buff
 *  map::            Ptr to stop point buff
 *  chan::           Channel number
 *  y::              Row number
 *  x::              Col. number
 *  level::          Lay number
 *  areaId::         Area number
 *  thre::           Quantization parameter
 *  psymbol::        Ptr to symbol coding structure 
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_zoneCoding(
    IPWState           *pIPWState,
    WAVESUBHEAD        *pheader,
    Int16              ** coefbuf,
    UInt8              *root,
    Int16              *map,
    Int32              chan,
    Int32              y,
    Int32              x,
    Int32              level,
    Int16              areaId,
    Int16              thre,
    SYMBOLSAVE         *psymbol)
{
    PROLOGUE;
    Int16    *coef, *coef1, pixel, value;
    Int32    bpl, idx, num, layer, bpl1;
    Int32    sum, begRow, endRow, begCol, endCol;
    Int32    coefw, coefh, halfw, halfh, fw, fh;
    Int32    row, col, leaf;
    UInt8    *proot, *proot1, byte, bits;
    Int16    *pmap, *pmap1;

    bpl = pheader->coefw[chan][0];
    bpl1 = pheader->coefw[chan][1];

    byte =0;
    bits =0;

    begRow = y;
    begCol = x;
    halfw = (pheader->coefw[chan][level]+1)>>1;
    halfh = (pheader->coefh[chan][level]+1)>>1;
    layer = (Int32) pheader->min_layer[chan];
    leaf  =1;
    for(--level, sum=2; level>=layer && leaf >0; --level, sum*=2)
    {
	fw = halfw;
	fh = halfh;
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;
 
	switch(areaId)
	{
	case 1:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow *= 2;
	    endRow = MIN((begRow + sum), halfh);
	    endCol = MIN((begCol + sum), coefw);
	    coef = *coefbuf + begRow*bpl;
	    proot = root + begRow*bpl;
	    pmap = map + begRow*bpl1;
	    leaf =0;
	    for(col=begCol; col<endCol; ++col)
	    {
		coef1 = coef+col;
		proot1 = proot +col;
		pmap1 = pmap + col;
		for(row=begRow; row<endRow; ++row, coef1 += bpl, 
			proot1 +=bpl, pmap1 += bpl1)
		{
		    if( *proot1 ) continue;
		    ++ leaf;
		    pixel = *coef1;
		    if( (abs(pixel) & thre ))
		    {
			if(pixel <0) 
			{
			    byte |= (CODE_3<<bits);
			    *coef1 = abs(pixel);
			}
			else
			    byte |= (CODE_1<<bits);
		    }
		    else
		    {
			if(level > layer )
			    value = *pmap1 & thre;
			else 
			{
			    byte |= (CODE_2 << bits);
			    value =1;
			}
			if( !value  )
			{
			    ERR_PASS(wp_setDeletedCoef(pIPWState, pheader,
						       chan, row, col, level,
						       areaId, root));
			    byte |= (CODE_2 << bits);
			}
		    }
		    bits += 2;
		    if(bits >= 4)
		    {
			ERR_PASS(wp_symbolCoding (pIPWState, byte, psymbol));
			bits = byte =0;
		    }
		}
	    }
	    break;
	case 2:
	    begRow = (begRow - fh)*2 +halfh;
	    begCol *= 2;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), halfw);
	    coef = *coefbuf + begRow*bpl;
	    proot = root + begRow*bpl;
	    pmap = map + begRow*bpl1;
	    leaf =0;
	    for(row=begRow; row<endRow; ++row, coef += bpl, 
		    proot += bpl, pmap += bpl1)
	    {
		for(col=begCol; col<endCol; ++col)
		{
		    if( proot[col] ) continue;
		    ++ leaf;
		    pixel = coef[col];
		    if( (abs(pixel) & thre ))
		    {
			if(pixel <0) 
			{
			    byte |= (CODE_3<<bits);
			    coef[col] = abs(pixel);
			}
			else
			    byte |= (CODE_1<<bits);
		    }
		    else
		    {
			if(level > layer )
			    value = *(pmap + col) & thre;
			else 
			{
			    byte |= (CODE_2 << bits);
			    value =1;
			}
			if( !value )
			{
			    ERR_PASS(wp_setDeletedCoef(pIPWState, pheader,
						       chan, row, col, level,
						       areaId, root));
			    byte |= (CODE_2 << bits);
			}
		    }
		    bits += 2;
		    if(bits >= 4)
		    {
			ERR_PASS(wp_symbolCoding (pIPWState, byte, psymbol));
			bits = byte =0;
		    }
		}
	    }
	    break;
	case 3:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow = (begRow - fh)*2 +halfh;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), coefw);
	    coef = *coefbuf;
	    leaf =0;
	    for(idx=0, num =0; num<sum*sum; ++idx )
	    {
		if( y_scan[idx] >= sum ||
		    x_scan[idx] >= sum )
		    continue;
		++ num;
		row = begRow + y_scan[idx];
		col = begCol + x_scan[idx];
		if( row >= endRow ||  col >= endCol ) continue;

		if( *(root+col+row*bpl) ) continue;
		++ leaf;
		pixel = *(coef+col+row*bpl);
		if(( abs(pixel) & thre ))
		{
		    if(pixel < 0)
		    {
			byte |= (CODE_3 << bits);
			*(coef+col+row*bpl) = abs(pixel);
		    }
		    else
			byte |= (CODE_1 << bits);
		}
		else
		{
		    if(level > layer )
			value = *(map + row * bpl1 + col) & thre;
		    else 
		    {
			byte |= (CODE_2 << bits);
			value =1;
		    }
		    if( !value )
		    {
			ERR_PASS(wp_setDeletedCoef(pIPWState, pheader,
						   chan, row, col, level,
 						   areaId, root));
			byte |= (CODE_2 <<bits);
		    }
		}
		bits += 2;
		if(bits >= 4)
		{
		    ERR_PASS(wp_symbolCoding (pIPWState, byte, psymbol));
		    bits = byte =0;
		}
	    }
	    break;
	default:
	    ERR_RETURN;
	}

    }
    if(  bits )
    {
	if(byte == CODE_2) byte |= (CODE_2 << 2);
	ERR_PASS(wp_symbolCoding (pIPWState, byte, psymbol));
	bits = byte =0;
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman coding on a area.
 *
 * LongDescription:::
 *  Do huffman coding in the assigned area. The difference between are and
 * zone is that zone can cover several level, but area is in one level. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  coefbuf::        Ptr to coef-buffer
 *  bpl::            Bytes per line
 *  y::               Row number
 *  x::               Col. number
 *  width::           Area width
 *  height::          Area height
 *  thre::            Quantization parameter
 *  psymbol::        Ptr to symbol coding structure 
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_areaCoding (
    IPWState           *pIPWState,
    Int16              *coefbuf,
    Int32              bpl,
    Int32              x,
    Int32              y,
    Int32              width,
    Int32              height,
    Int16              thre,
    SYMBOLSAVE         *psymbol)
{
    PROLOGUE;
    Int32       row, col;
    Int16       *pcoef, pixel;
    UInt8       code, bits;

    pIPWState = pIPWState;
    pcoef = coefbuf + bpl*y +x;
    bits =0;
    code =0;
    for(row=0; row<height; ++row, pcoef += bpl)
    {
	for(col=0; col<width; ++col)
	{
	    pixel = *(pcoef + col);
	    if( (abs(pixel) & thre ) )
	    {
		if( pixel < 0) code |= (CODE_3<<bits);
		else code |= (CODE_1<<bits);
	    }
	    bits += 2;
	    if(bits >= 4)
	    {
		ERR_PASS(wp_symbolCoding (pIPWState, code, psymbol));
		bits =0;
		code =0;
	    }
	}
    }
    if (bits)
    {
	ERR_PASS(wp_symbolCoding (pIPWState, code, psymbol));
    }
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Callback function for writing out compressed image data.
 *
 * LongDescription:::
 *    Callback function for writing out compressed image data.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  userData::       Ptr to FILE ( save compressed data )
 *  buf::            Ptr to byte buffer to write out.
 *  nbytes::         Number of bytes to write.
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 ****************************************************************************/
static ErrCode CDECL
wp_WriteCallback (
    IPWState		 *pIPWState,
    void                 *userData,
    UInt8                *buf,
    UInt32                nbytes)
{
    PROLOGUE;

	/* Keep compiler happy */
    pIPWState = pIPWState;

    if(fwrite((UInt8 *)buf, (size_t)1, (size_t)nbytes, (FILE *)userData) < 1)
    {
	__error__ = cErrIO_WriteError;
	ERR_RETURN;
    }

    NORMAL_RETURN;
} 
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Coding & output wavelet coefficients.
 *
 * LongDescription:::
 *    After wave transform, we coding and output wavelet coefficients.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  coefbuf::        Ptr to coef-buffer
 *  ration::         Compression ratio ( compressed size=orignal size/ration )
 *  WriteCallback::  Call back function for write out compressed data
 *  userData::       Ptr to hook for passing in user data
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_waveCoding (
    IPWState           *pIPWState,
    WAVESUBHEAD      *pheader,
    Int16              **coefbuf,
    Int32              ration,
    ErrCode    (CDECL *WriteCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes),
    void               *userData)
{
    PROLOGUE;
    Int32         chan, coefw, coefh, halfw, halfh, fw, fh;
    Int32         ncomps, row, col, size, bpl, bpl1, level;
    Int32         x, y, bytes, init_sum, dc_len[3];
    Int16         *coef[3], *ptmp, thre, *map, *ptmp1;
    UInt8         ali_code[4] = {0xf3, 0xf3, 0xf3, 0xf0}, *root, byte;
    SYMBOLSAVE    symbol;
    WHUFFINFO     *huff;

    ncomps = pheader->ncomps;
    size = ncomps * pheader->coefw[0][0] * pheader->coefh[0][0];

    huff = (WHUFFINFO *)IPW_CALLOC(sizeof(WHUFFINFO), 1, cAllocatorData); 
    ERR_GEN (huff == NULL, cErrNomem); 
    ERR_PASS(i_wc_waveHuffInit (pIPWState, (void *)userData,
				WriteCallback, huff));

    for(chan=0; chan<ncomps; ++chan)
    {
	coef[chan] = *(coefbuf+chan);
	bpl = pheader->coefw[chan][0];

	level = pheader->pyramidN[chan]-1;

	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
 	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;

	dc_len[chan] = huff->codeLen;
	ERR_PASS( i_wc_set_huffTable (pIPWState, chan, 1, huff));
	for(row=0, ptmp=coef[chan]; row<halfh; ++row, ptmp+=bpl)
	{
	    for(col=0; col<halfw; ++col)
	    {
		ERR_PASS(i_wc_dc_huffCoding (pIPWState,
					     (Int16)(*(ptmp+col)&0xff),
					     huff));
	    }
	}
	for(row=0, ptmp=coef[chan]; row<halfh; ++row, ptmp+=bpl)
	{
	    for(col=0; col<halfw; ++col)
	    {
		ERR_PASS(i_wc_dc_huffCoding (pIPWState,
					     (Int16)(*(ptmp+col)>>8),huff ));
	    }
	}
	dc_len[chan] = huff->codeLen - dc_len[chan];
    }
    ERR_PASS( i_wc_clearZerolen (pIPWState, 0, huff) );

    ERR_PASS(WriteCallback(pIPWState, (void *)userData, (UInt8 *)ali_code, 
			   sizeof(UInt8)*4)); 
    IPW_FREE(huff, cAllocatorData);

    symbol.saveByte = 0;
    symbol.saveBits =0;
    symbol.passData = userData;
    symbol.WriteCallback = WriteCallback;
    symbol.byte_sum =0;

    for(chan=0; chan<ncomps; ++chan)
    {
	coef[chan] = *(coefbuf+chan);
	bpl = pheader->coefw[chan][0];
	bpl1 = pheader->coefw[chan][1];

	root = (UInt8 *)IPW_CALLOC(sizeof(UInt8), 
				  pheader->coefh[chan][0]*bpl, cAllocatorData);
	ERR_GEN (root == NULL, cErrNomem);

	map =(Int16 *)IPW_CALLOC(sizeof(Int16), 
				 pheader->coefh[chan][1]*bpl1, 
				 cAllocatorData);
	ERR_GEN (map == NULL, cErrNomem);

	ERR_PASS(wp_rootValue(pIPWState, pheader, coef[chan], chan, map));

	bytes = bpl*pheader->coefh[chan][0]/ration;
	bytes -= dc_len[chan];
	init_sum = symbol.byte_sum;

	thre = pheader->quant[chan];
	level = pheader->pyramidN[chan]-1;	    
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;
	for(; thre >2; thre>>=1)
	{
	    MEMSET((void *) root, 0, pheader->coefh[chan][0]*bpl);
	    level = pheader->pyramidN[chan]-1;	    
	    coefw = pheader->coefw[chan][level];
	    coefh = pheader->coefh[chan][level];
	    halfw = (coefw+1)>>1;
	    halfh = (coefh+1)>>1;

	    ptmp = coef[chan];
	    ptmp1 = map;
	    for(row=0; row<halfh; ++row, ptmp += bpl, ptmp1+= bpl1)
	    {
		for(col=halfw; col<coefw; col += 2)
		{
		    ERR_PASS(wp_topCoding(pIPWState, ptmp, ptmp1, col, coefw, 
					   thre, &byte));
		    ERR_PASS(wp_symbolCoding (pIPWState, byte, &symbol));
		    if( (byte & 3) != 2)
			ERR_PASS(wp_zoneCoding(pIPWState, pheader, &coef[chan],
					       root, map,chan, row, col, 
					       level, 1, thre, &symbol));

		    if( (byte & 12) != 8)
			ERR_PASS(wp_zoneCoding(pIPWState, pheader, &coef[chan],
					       root, map,chan, row, (col+1),
					       level, 1, thre, &symbol));
		}
	    }
	    ptmp = coef[chan] + halfh*bpl;
	    ptmp1 = map + halfh*bpl1;
	    for(row=halfh; row<coefh; ++row, ptmp += bpl, ptmp1 += bpl1)
	    {
		for(col=0; col<halfw; col += 2)
		{
		    ERR_PASS(wp_topCoding(pIPWState, ptmp, ptmp1, col, halfw, 
					   thre, &byte));
		    ERR_PASS(wp_symbolCoding (pIPWState, byte, &symbol));
		    if( (byte & 3) != 2)
			ERR_PASS(wp_zoneCoding(pIPWState, pheader, &coef[chan],
					       root, map, chan, row, col, 
					       level, 2, thre, &symbol));
		    if( (byte & 12) != 8)
			ERR_PASS(wp_zoneCoding(pIPWState, pheader, &coef[chan],
					       root, map, chan, row, (col+1),
					       level, 2, thre, &symbol));
		}
	    }
	    ptmp = coef[chan] + halfh*bpl;
	    ptmp1 = map + halfh*bpl1;
	    for(row=halfh; row<coefh; ++row, ptmp += bpl, ptmp1 += bpl1)
	    {
		for(col=halfw; col<coefw; col += 2)
		{
		    ERR_PASS(wp_topCoding(pIPWState, ptmp, ptmp1, col, coefw, 
					  thre, &byte));
		    ERR_PASS(wp_symbolCoding (pIPWState, byte, &symbol));
		    if( (byte & 3) != 2)
			ERR_PASS(wp_zoneCoding(pIPWState, pheader, &coef[chan],
					       root,map, chan, row, col, 
					       level, 3, thre, &symbol));
		    if( (byte & 12) != 8)
			ERR_PASS(wp_zoneCoding(pIPWState, pheader, &coef[chan],
					       root,map, chan, row, (col+1),
					       level, 3, thre, &symbol));
		}
	    }
	    level=pheader->pyramidN[chan]-1;
	    x = pheader->coefw[chan][level] -1;
	    y = pheader->coefh[chan][level] -1;
	    halfw = (pheader->coefw[chan][level] +1)>>1;
	    halfh = (pheader->coefh[chan][level] +1)>>1;
	    for(--level, col=2; level>=pheader->min_layer[chan]; 
		--level, col *=2)
	    {
		fw = halfw;
		fh = halfh;
		coefw = pheader->coefw[chan][level];
		coefh = pheader->coefh[chan][level];
		halfw = (coefw +1) >>1;
		halfh = (coefh +1) >>1;
		x = (x - fw)*2 +halfw;
		y = (y - fh)*2 +halfh;
		if( (x+col) < coefw )
		{
		    ERR_PASS( wp_areaCoding (pIPWState, coef[chan], bpl,
					     (x+col), 0, (coefw-x-col), 
					     coefh, thre, &symbol));
		}
		if( (y+col) < coefh )
		{
		    ERR_PASS( wp_areaCoding (pIPWState, coef[chan], bpl, 0, 
					     (y+col), coefw, (coefh-y-col), 
					     thre, &symbol));
		}
	    }
	    if( (symbol.byte_sum - init_sum +10) >= bytes )
	    {
		root[0] = 0;
		ERR_PASS(wp_symbolCoding (pIPWState, root[0], &symbol));
		break;
	    }
	    root[0] = 10;
	    ERR_PASS(wp_symbolCoding (pIPWState, root[0], &symbol));
	}
        IPW_FREE(root, cAllocatorData);	    
        IPW_FREE(map, cAllocatorData);	   
    }

    if( symbol.saveBits )
    {
	ERR_PASS(WriteCallback(pIPWState, (void *)userData,
			       (UInt8 *)&symbol.saveByte, 1));
	++ symbol.byte_sum; 
	symbol.saveBits =0;
	symbol.saveByte =0;
    }
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Seperate image into strip then do wavelet transform.
 *
 * LongDescription:::
 *   Whole page wavelet compression needs large memory for A4 size
 *   images. We split image into strip, then do wavelet compression.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  image::          Ptr to image buffer ptr
 *  w::              Image width
 *  h::              Image height
 *  bpl::            Image bytes per line
 *  ncomps::         Number of image channels 
 *  ration::         Compression ratio ( compressed size=orignal size/ration )
 *  op_flag::        Some compression options
 *  WriteCallback::  Call back function for write out compressed data
 *  userData::       Ptr to userDat passed to call back function
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_cwaveImageSplit (
    IPWState           *pIPWState,
    UInt8              **image,
    Int32              w,
    Int32              h,
    Int32              bpl,
    Int32              ncomps,
    Int32              ration,
    Int32              op_flag,
    ErrCode    (CDECL *WriteCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes),
    void               *userData)
{
    PROLOGUE;
    WAVEIMAGEHEAD      waveHead;
    WAVESUBHEAD        waveSub;
    WIHSAVEFORM        waveHeadSave;
    WSHSAVEFORM        waveSubSave;
    Int32              *lowTab[5], *highTab[4];
    Int16              *userBuf=NULL;
    Int16              *coef[3], layer;
    Int32              row, chan, level, currBpl;
    Int32              coefh, coefw, halfh, halfw;
    Int32              width, height, sub_h;

/*  Init ptr */
    coef[0] = coef[1] = coef[2] = NULL;
    lowTab[0] = NULL;

/* Wave header */
    waveHead.version = WAVE_VERSION;      /* Set current version number */
    waveHead.imageW = (Int16) w;
    waveHead.imageH = (Int16) h;
    waveHead.ncomps = (Int16) ncomps;
    row =1;
    height =  h-512;
    while(height >= 48)
    {
	++row;
	height -= 508;
    }
    waveHead.num_sub = (Int16)row;
/* Byte order */
    waveHeadSave.versionL = (Int8)waveHead.version;
    waveHeadSave.versionH = 0;
    waveHeadSave.ncomps = (Int8)waveHead.ncomps;
    waveHeadSave.num_sub = (Int8)waveHead.num_sub;
    waveHeadSave.imageWL = (UInt8)(waveHead.imageW & 0xff);
    waveHeadSave.imageWH = (UInt8)((waveHead.imageW>>8) & 0xff);
    waveHeadSave.imageHL = (UInt8)(waveHead.imageH & 0xff);
    waveHeadSave.imageHH = (UInt8)((waveHead.imageH>>8) & 0xff);
    ERR_PASS(WriteCallback(pIPWState, (void *)userData, (UInt8 *)&waveHeadSave, 
			   sizeof(WIHSAVEFORM))); 

/* Init filter LUT */
    level = 256 * 8; 
    lowTab[0] = (Int32 *)IPW_CALLOC(sizeof(Int32), level*9, cAllocatorData);
    ERR_GEN_CLEANUP (lowTab[0] == NULL, cErrNomem);
    highTab[0] = lowTab[0] + level*5;
    for(chan=1; chan<4; ++chan)
    {
	lowTab[chan] = lowTab[chan-1]+level;
	highTab[chan] = highTab[chan-1]+level;
    }
    lowTab[4] = lowTab[3] + level;
    ERR_PASS(i_wc_init_filtertbl (pIPWState, f_lowpass, f_highpass, 5,
				  4, level, &lowTab[0], &highTab[0] ));

/* allocate temp  memory */
    row = w + 32;
    if( (h +32) > row) row = h + 32;
    userBuf= (Int16 *)IPW_CALLOC(sizeof(Int16), row, cAllocatorData);
    ERR_GEN_CLEANUP (userBuf == NULL, cErrNomem);

/* copy image to buffer */
    MEMSET((void *)&waveSub, 0, sizeof(WAVESUBHEAD));
    MEMSET((void *)&waveSubSave, 0, sizeof(WSHSAVEFORM));
    waveSub.ncomps = (Int16)ncomps;
    waveSubSave.ncompsL = (UInt8) waveSub.ncomps;

    layer = (Int16)op_flag & cNoZeroLayer;
    if( layer )
    {
	waveSub.min_layer[0] = (Int16)1;
	waveSub.min_layer[1] = (Int16)1;
	waveSub.min_layer[2] = (Int16)1;
	waveSubSave.min_layer[0] = (Int8)1;
	waveSubSave.min_layer[1] = (Int8)1;
	waveSubSave.min_layer[2] = (Int8)1;
    }
    sub_h = 512;
    for(row=0; row<h; row+= sub_h)
    {
	width = w;
	height =512;
	if(row >0) sub_h = 508;
	if( (row+ sub_h) > h)
	{
	    height = h-row;
	    if(row >0) height +=4;
	}
	else if( (h - row -sub_h) < 48)
	{
	    height += (h-row-sub_h);
	    sub_h = h - row;
	} 
	halfw = (width+1)>>1;
	halfh = (height+1)>>1;
	for(chan=0; chan<ncomps; ++chan)
	{
	    if(chan <1)
		coef[chan]=(Int16 *)IPW_CALLOC(sizeof(Int16),
					(width*height),cAllocatorData);
	    else
		coef[chan]=(Int16 *)IPW_CALLOC(sizeof(Int16),(halfw*halfh),
					       cAllocatorData);
	    ERR_GEN_CLEANUP (coef[chan] == NULL, cErrNomem);
	}
	waveSub.pyramidN[0] =waveSub.pyramidN[1] =
		waveSub.pyramidN[2]= PYRAMID;
	waveSubSave.pyramidN[0] =waveSubSave.pyramidN[1] =
		waveSubSave.pyramidN[2]= (Int8)PYRAMID;
	level = row;
	if(row >0) level = row-4;
	ERR_PASS_CLEANUP(i_wc_image2buffer (pIPWState, ncomps, bpl, 0, level, 
					    width, height, image, coef));

/* do wavelet transform */
	for(chan=0; chan<ncomps; ++chan)
	{
	    coefh = height;
	    coefw = width;
	    currBpl = width;
	    if(chan >0)
	    {
		coefh = (height+1)>>1;
		coefw = (width+1)>>1;
		currBpl = (width+1)>>1;
	    }

	    for(level=0; level<PYRAMID; ++level)
	    {
		if(level >1 &&( coefw < 16 || coefh < 16)) break;
		waveSub.coefw[chan][level] = (Int16)coefw;
		waveSub.coefh[chan][level] = (Int16)coefh;
		waveSubSave.coefwL[chan][level] = (UInt8)(coefw & 0xff);
		waveSubSave.coefwH[chan][level] = (UInt8)((coefw>>8) & 0xff);
		waveSubSave.coefhL[chan][level] = (UInt8)(coefh & 0xff);
		waveSubSave.coefhH[chan][level] = (UInt8)((coefh>>8) & 0xff);
		if(level < 2)
		    ERR_PASS_CLEANUP(i_wc_biwaveletTrans_LUT (pIPWState,
							      coefw, coefh,
		   currBpl, &lowTab[0], &highTab[0], &userBuf, &coef[chan] ));
		else
		    ERR_PASS_CLEANUP(i_wc_biwaveletTransform (pIPWState,
							      coefw, coefh,
		     currBpl, f_lowpass, f_highpass, &userBuf, &coef[chan] ));
		coefh = (coefh+1)>>1;
		coefw = (coefw+1)>>1;
	    }
	    if(level < PYRAMID) 
	    {
		waveSub.pyramidN[chan] = (Int16)level;
		waveSubSave.pyramidN[chan] = (Int8)level;
	    }
	}
	ERR_PASS(wp_waveQuatization(pIPWState, &waveSub, coef, ration));
	for(chan=0; chan<ncomps; ++chan)
	{
	    waveSubSave.mindcL[chan] = (UInt8)(waveSub.mindc[chan] & 0xff);
	    waveSubSave.mindcH[chan] =
		(UInt8)((waveSub.mindc[chan]>>8) & 0xff);
	    waveSubSave.quantL[chan] = (UInt8)(waveSub.quant[chan] & 0xff);
	    waveSubSave.quantH[chan]=(UInt8)((waveSub.quant[chan]>>8) & 0xff);
	}
	ERR_PASS(WriteCallback(pIPWState, (void *)userData,
			       (UInt8 *)&waveSubSave, sizeof(WSHSAVEFORM)));    
/* output coef */
	ERR_PASS(wp_waveCoding(pIPWState, &waveSub, coef, ration, 
			       WriteCallback, (void *)userData));
	for(chan=0; chan<ncomps; ++chan)
	    IPW_FREE(coef[chan], cAllocatorData);
    }
    IPW_FREE(lowTab[0], cAllocatorData);
    IPW_FREE(userBuf, cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:

    if(lowTab[0] != NULL)
	IPW_FREE(lowTab[0], cAllocatorData);
    if( userBuf != NULL)
	IPW_FREE(userBuf, cAllocatorData);
    for(chan=0; chan<ncomps; ++chan)
    {
	if(coef[chan] != NULL)
	    IPW_FREE(coef[chan], cAllocatorData);
    }


    ERR_RETURN;

}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  WAVElet Compress gray or RGB PIXR.
 *
 * LongDescription:::
 *    Compress a gray or RGB image in three PIXRs to a file.
 *    The source PIXRs must be either 8 bpp. They must all be
 *    of the same bit depth. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPixr::        Source 4 or 8 bit PIXR to be compressed (gray or RGB)
 *  fdWAVE::         FILE ptr for the compressed file
 *  ration::         ratio:1 compression.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
  *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_waveCompress (
    IPWState           *pIPWState,
    PIXR               *srcPixr,
    FILE               *fdWAVE,
    Int32               ration)
{
    PROLOGUE;
    PIXR               *pixrBand;
    UInt8              *image[3];
    Int32              w, h, d;
    Int32              chan, bpl;
    Int32              ncomps, size;
    Int32              bugetBytes;

/* Verify arguments */
    ERR_GEN(fdWAVE == NULL, cErrNullPtr);
    ERR_GEN_PIXR(srcPixr, cByteOrderBig, cDepth8,
		 cChannels1 | cChannels3, cFrame32);

   ERR_PASS( w_pixrGetChannelCount(pIPWState, srcPixr, &ncomps));

/* Determine the image size, start address and line pitch */
/* We'll assume here that all of the components are the same */
    ERR_PASS( w_pixrGetWidth(pIPWState, srcPixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, srcPixr, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, srcPixr, &bpl));

/* Verify that this is an 8 or 4 bit image */
    ERR_PASS( w_pixrGetDepth(pIPWState, srcPixr, &d));
    ERR_GEN(! (d==8), cErrDepthNotSupported);

/* Calculate compressed image size */
    size = ncomps * w * h;
    bugetBytes = (size + ration -1)/ration;
/*    fprintf(stderr, "Wavelet compression %ld : 1 size = %ld\n",
            ration, bugetBytes); */

/*  Get image */
    for(chan=0; chan<ncomps; ++chan)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, srcPixr, chan, &pixrBand) );
        ERR_GEN( pixrBand == NULL, cErrInvalidParm );
	ERR_PASS( w_pixrGetImage(pIPWState, pixrBand, &(image[chan])) );
    }

    ERR_PASS(wp_cwaveImageSplit (pIPWState, image, w,
				 h, bpl, ncomps, ration, 0,
				 wp_WriteCallback, (void *)fdWAVE));

/*    fprintf(stderr, "image size=%ld  compressed size = %ld\n",size, data_sum); */

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  WAVElet Compress gray or RGB PIXR.
 *
 * LongDescription:::
 *    Compress a gray or RGB image in three PIXRs to a file.
 *    The source PIXRs must be either 8 bpp. They must all be
 *    of the same bit depth. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  srcPixr::        Source 4 or 8 bit PIXR to be compressed (gray or RGB)
 *  WriteCallback::  Call back function for write out compressed data
 *  userData::       Ptr to userDat passed to call back function
 *  option::         Some compression options
 *  ration::         ratio:1 compression.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_waveCompress_option(
    IPWState           *pIPWState,
    PIXR               *srcPixr,
    ErrCode	(CDECL *WriteCallback) (IPWState *pIPWState,
				        void *userData, UInt8 *buf, 
					UInt32 nbytes),
    void               *userData,
    Int32               option,
    Int32               ration)
{
    PROLOGUE;
    PIXR               *pixrBand;
    UInt8              *image[3];
    Int32              w, h, d;
    Int32              chan, bpl;
    Int32              ncomps, size;
    Int32              bugetBytes;

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN_PIXR(srcPixr, cByteOrderBig, cDepth8,
		 cChannels1 | cChannels3, cFrame32);

   ERR_PASS( w_pixrGetChannelCount(pIPWState, srcPixr, &ncomps));

/* Determine the image size, start address and line pitch */
/* We'll assume here that all of the components are the same */
    ERR_PASS( w_pixrGetWidth(pIPWState, srcPixr, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, srcPixr, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, srcPixr, &bpl));

/* Verify that this is an 8 or 4 bit image */
    ERR_PASS( w_pixrGetDepth(pIPWState, srcPixr, &d));
    ERR_GEN(! (d==8), cErrDepthNotSupported);

/* Calculate compressed image size */
    size = ncomps * w * h;
    bugetBytes = (size + ration -1)/ration;
/*    fprintf(stderr, "Wavelet compression %ld : 1 size = %ld\n",
            ration, bugetBytes); */

/*  Get image */
    for(chan=0; chan<ncomps; ++chan)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, srcPixr, chan, &pixrBand) );
        ERR_GEN( pixrBand == NULL, cErrInvalidParm );
	ERR_PASS( w_pixrGetImage(pIPWState, pixrBand, &(image[chan])) );
    }

    ERR_PASS(wp_cwaveImageSplit (pIPWState, image, w,
				 h, bpl, ncomps, ration, option,
				 WriteCallback, userData));

    NORMAL_RETURN;
}

#undef _IW_CWAVE_C_

/*
$Log: iw_cwave.c,v $
Revision 1.1.1.1  1999/09/20 18:19:18  jlin
Imported sources

Revision 1.17  1998/01/15 01:49:37  ddavies
iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Fix WATCOM warnings about unused procedure args.

Revision 1.16  1998/01/14 00:13:33  ddavies
iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.15  1997/12/05 19:50:24  xzhu
xzhu on Fri Dec 5 11:50:23 PST 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Moved x_scan and y_scan from iw_cwave to iw_dwave.

Revision 1.14  1997/10/01 23:33:29  xzhu
xzhu on Wed Oct 1 16:33:29 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Added version info in and check it when decoding.

Revision 1.13  1997/09/22 16:19:32  xzhu
xzhu on Mon Sep 22 09:18:51 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

Fixed PC compiler warnings.

Revision 1.12  1997/09/19 19:09:21  xzhu
xzhu on Fri Sep 19 12:09:20 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Changed wavelet coding method, it improved image quality. It can be progressive decoding.

Revision 1.11  1997/09/17 23:52:18  xzhu
xzhu on Wed Sep 17 16:49:02 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

Allocate Huffman coding info structure memory and pass it to huffman coding function. Then get coding info directly from the structure.

Revision 1.10  1997/07/11 01:59:01  ddavies
iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.9  1997/05/17 01:21:45  xzhu
xzhu on Fri May 16 18:21:43 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: No print out for release version.

Revision 1.8  1997/05/10 00:05:38  xzhu
xzhu on Fri May 9 17:05:38 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Fixed header byte order problem.

Revision 1.7  1997/04/28 16:27:04  xzhu
xzhu on Mon Apr 28 09:27:03 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Added A=A to make compiler happy.

Revision 1.6  1997/04/25 22:07:17  xzhu
xzhu on Fri Apr 25 15:07:16 PDT 1997

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Changed coding algorithm and added new wrap functions.

Revision 1.5  1996/11/08 23:11:42  xzhu
xzhu on Fri Nov 8 15:11:41 PST 1996

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Passed callback function to wp_waveCoding() and wp_wavedecoding().

Revision 1.4  1996/10/16 00:11:48  xzhu
xzhu on Tue Oct 15 17:11:47 PDT 1996

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Ipshared wavelet functions' name has been changed to fit ipshared style. So we did same changes on these functions' name.

Revision 1.3  1996/10/15 18:25:08  xzhu
xzhu on Tue Oct 15 11:24:41 PDT 1996

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

Fixed coreCheck warnings.

Revision 1.2  1996/10/15 18:07:26  xzhu
xzhu on Tue Oct 15 11:07:25 PDT 1996

iw_cwave.c -> /product/ipcore/ipwrap/src/RCS/iw_cwave.c,v

CommonCheck: Fixed coreCheck warnings and WATCOM complier warnings.

*/

