/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*	iw_dwave.c:	WAVElet decompression wrappers:
 *
 *	Code to decompress 3 types of WAVElet compressed files:
 *		w_waveDecompressToGray()
 *		w_waveDecompressToRGB()
 *
 */

#define _IW_DWAVE_C_

#include <stdio.h>

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "jpeglib.h"
#include "defines.pub"
#include "imageref.pub"
#include "frames.pub"

/* prototypes */
#include "shrpixr.pub"
#include "props.pub"
#include "wavelib.pub"
#include "iw_wave.pub"
#include "iw_wave.prv"

IP_RCSINFO(iw_dwave_c_RCSInfo, "$RCSfile: iw_dwave.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");

Int32 r_lowpass[5] = {27941, 13700, -3624, -2114, 1239};
Int32 r_highpass[5] = {25837, -12366, -1333, 781, 0};

static Int32 readLen;

Int32 x_scan[256] = {0, 1, 0, 0, 1, 2, 3, 2, 1, 0, 0, 1, 2, 3, 4,
			    5, 4, 3, 2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 6,
			    5, 4, 3, 2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8,
			    9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 0, 1, 2, 3, 4,
			    5, 6, 7, 8, 9, 10,11,10,9, 8, 7, 6, 5, 4, 3,
			    2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,11,
			    12,13,12,11,10,9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
			    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,11,12,13,14,
			    15,14,13,12,11,10,9, 8, 7, 6, 5, 4, 3, 2, 1,
			    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13, 14,
			    15,15,14,13,12,11,10,9, 8, 7, 6, 5, 4, 3, 2,
			    3, 4, 5, 6, 7, 8, 9, 10,11,12,13,14,15,15,14,
			    13,12,11,10,9, 8, 7, 6, 5, 4, 5, 6, 7, 8, 9,
			    10,11,12,13,14,15,15,14,13,12,11,10,9, 8, 7, 
			    6, 7, 8, 9,10,11,12,13, 14,15,15,14,13,12,11,
			    10,9, 8, 9,10,11, 12,13,14,15,15,14,13,12,11,
			    10,11,12,13,14,15,15,14,13,12,13,14,15,15,14,
			    15};
Int32 y_scan[256] = {0, 0, 1, 2, 1, 0, 0, 1, 2, 3, 4, 3, 2, 1, 0,
			    0, 1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1, 0, 0, 1,
			    2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1, 0,
			    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,9, 8, 7, 6,
			    5, 4, 3, 2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8,
			    9, 10,11,12,11,10,9, 8, 7, 6, 5, 4, 3, 2, 1,
			    0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,11,12,13,
			    14,13,12,11,10,9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
			    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,11,12,13,14,
			    15,15,14,13,12,11,10,9, 8, 7, 6, 5, 4, 3, 2,
			    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,11,12,13,14,15,
			    15,14,13,12,11,10,9, 8, 7, 6, 5, 4, 3, 4, 5,
			    6, 7, 8, 9, 10,11,12,13,14,15,15,14,13,12,11,
			    10,9, 8, 7, 6, 5, 6, 7, 8, 9, 10,11,12,13,14,
			    15,15,14,13,12,11,10,9, 8, 7, 8, 9, 10,11,12,
			    13,14,15,15,14,13,12,11,10,9, 10,11,12,13,14,
			    15,15,14,13,12,11,12,13,14,15,15,14,13,14,15,
			    15};

 
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do dequatization on transformed coefits before reverse transform.
 *
 * LongDescription:::
 *    before reverse wave transform, we do dequatization for every region.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *
 * Output:::
 *  coefbuf::        Ptr to coef-buffer
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_waveDequatization (
    IPWState           *pIPWState,
    WAVESUBHEAD        *pheader,
    Int16              **coefbuf)
{
    PROLOGUE;
    Int16       *coef[3], *ptmp, mindc;
    Int32       chan, row, col, level, bpl;
    Int32       coefh, coefw, halfh, halfw;

    pIPWState = pIPWState;
    for(chan=0; chan<pheader->ncomps; ++chan)
    {
	coef[chan] = *(coefbuf+chan);
	bpl = pheader->coefw[chan][0];

	level = pheader->pyramidN[chan]-1;
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;

	mindc = pheader->mindc[chan];

	if( mindc >0 )
	{
	    for(row=0, ptmp=coef[chan]; row<halfh; ++row, ptmp+=bpl)
	    {
		for(col=0; col<halfw; ++col)
		{
		    *(ptmp+col) += mindc; 
		}
	    }
	}
    }
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set a passing flag for zero region coefits.
 *
 * LongDescription:::
 *   Set passing flag to coefits in zero region. Once we meet this flag,
 *   we will pass this coefit.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  chan::           Channel number
 *  y::               Row number
 *  x::               Col. number
 *  level::           Layer number
 *  areaId::          Area number
 *
 * Output:::
 *  code::           Ptr to passing flag array.
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_setSkipCoef(
    IPWState           *pIPWState,
    WAVESUBHEAD        *pheader,
    Int32              chan,
    Int32              y,
    Int32              x,
    Int32              level,
    Int16              areaId,
    UInt8              *code)
{
    PROLOGUE;
    UInt8    *pcode;
    Int32    bpl;
    Int32    sum, begRow, endRow, begCol, endCol;
    Int32    coefw, coefh, halfw, halfh, fw, fh;
    Int32    row;

    pIPWState = pIPWState;
    bpl = pheader->coefw[chan][0];
    begRow = y;
    begCol = x;
    halfw = (pheader->coefw[chan][level]+1)>>1;
    halfh = (pheader->coefh[chan][level]+1)>>1;
    for(--level,sum=2; level>=pheader->min_layer[chan]; --level, sum*=2)
    {
	fw = halfw;
	fh = halfh;
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;
 
	switch(areaId)
	{
	case 1:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow *= 2;
	    endRow = MIN((begRow + sum), halfh);
	    endCol = MIN((begCol + sum), coefw);
	    break;
	case 2:
	    begRow = (begRow - fh)*2 +halfh;
	    begCol *= 2;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), halfw);
	    break;
	case 3:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow = (begRow - fh)*2 +halfh;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), coefw);
	    break;
	default:
	    ERR_RETURN;
	}
	pcode = code + begRow*bpl + begCol;
	for(row=begRow; row<endRow; ++row, pcode+=bpl)
	{
	    MEMSET((void *) pcode, 1, (endCol-begCol));
	}
    }
 
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set a passing flag for zero region coefits.
 *
 * LongDescription:::
 *   Set passing flag to coefits in zero region. Once we meet this flag,
 *   we will pass this coefit.
 *
 * Input:::
 *  pIPWState::	       Ptr to IPCore state
 *  psymbol::          Ptr to symbol decoding structure
 *
 * Output:::
 *  symbol::           Decoded symbol
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_symbolDecoding (
    IPWState		 *pIPWState,
    UInt8                *symbol,
    SYMBOLREAD           *psymbol)
{
    PROLOGUE;

    if( psymbol->readBits < 1)
    {
	psymbol->ReadCallback(pIPWState, psymbol->passData,
			      &psymbol->readByte, 1);
	psymbol->readBits = 8;
    }
    if( psymbol->readByte & 1)
    {
	*symbol = 0x0a;
	psymbol->readByte >>= 1;
	-- psymbol->readBits;
    }
    else
    {
	psymbol->readByte >>= 1;
	-- psymbol->readBits;
	if( psymbol->readBits < 1)
	{
	    psymbol->ReadCallback(pIPWState, psymbol->passData,
				  &psymbol->readByte, 1);
	    psymbol->readBits = 8;
	}
	if(psymbol->readBits < 4)
	{
	    *symbol = psymbol->readByte;
	    psymbol->ReadCallback(pIPWState, psymbol->passData,
				  &psymbol->readByte,1);
	    *symbol |= (psymbol->readByte << psymbol->readBits);
	    *symbol &= 0x0f;
	    psymbol->readByte >>= (4-psymbol->readBits);
	    psymbol->readBits = 4+ psymbol->readBits;
	}
	else
	{
	    *symbol = psymbol->readByte & 0x0f;
	    psymbol->readBits -= 4;
	    psymbol->readByte >>= 4;
	}
    }

    NORMAL_RETURN;
} 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 * uffman decoding for wavelet coefficients.
 *
 * LongDescription:::
 *    Before wave transform, we decoding the input to get wavelet 
 *  coefficients.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  coefbuf::        Ptr to coef-buffer
 *  root::           Ptr to flag array
 *  chan::           Channel number
 *  y::               Row number
 *  x::               Col. number
 *  areaId::          Area number
 *  thre::            Quantization parameter
 *  psymbol::          Ptr to symbol decoding structure
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_zoneDecoding(
    IPWState           *pIPWState,
    WAVESUBHEAD      *pheader,
    Int16              ** coefbuf,
    UInt8              *root,
    Int32              chan,
    Int32              y,
    Int32              x,
    Int16              areaId,
    Int16              thre,
    SYMBOLREAD         *psymbol)
{
    PROLOGUE;
    Int16    *coef, *coef1;
    Int32    bpl, idx, level, num;
    Int32    sum, begRow, endRow, begCol, endCol;
    Int32    coefw, coefh, halfw, halfh, fw, fh;
    Int32    row, col, layer, leaf;
    UInt8    *pcode, pixel, *pcode1, bits;

    bpl = pheader->coefw[chan][0];

    coef = *coefbuf + y*bpl;
    level= pheader->pyramidN[chan]-1;
 
    bits =4;

    begRow = y;
    begCol = x;
    halfw = (pheader->coefw[chan][level]+1)>>1;
    halfh = (pheader->coefh[chan][level]+1)>>1;
    layer = (Int32) pheader->min_layer[chan];
    leaf =1;
    for(--level,sum=2; level>=layer && leaf >0; --level, sum*=2)
    {
	fw = halfw;
	fh = halfh;
	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;
	switch(areaId)
	{
	case 1:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow *= 2;
	    endRow = MIN((begRow + sum), halfh);
	    endCol = MIN((begCol + sum), coefw);
	    coef = *coefbuf + begRow*bpl;
	    pcode = root + begRow*bpl;
	    leaf =0;
	    for(col=begCol; col<endCol ; ++col)
	    {
		coef1 = coef+col;
		pcode1 = pcode+col;
		for(row=begRow; row<endRow ; 
		    ++row, coef1 += bpl, pcode1 += bpl)
		{
		    if( *pcode1 == 1) continue;
		    ++leaf;
		    if( bits >= 4)
		    {
			ERR_PASS(wp_symbolDecoding(pIPWState, &pixel, psymbol));
			bits =0;
		    }
		    switch(pixel & 0x03)
		    {
		    case 2:
			if(level > layer )
			{
			    ERR_PASS(wp_setSkipCoef(pIPWState,pheader, 
						    chan, row, col, level, 
						    areaId, root));
			}
			break;
		    case 3:
			*coef1 -= thre;
			break;
		    case 1:
			if(*coef1 >= 0)
			    *coef1 += thre;
			else
			    *coef1 -= thre;
			break;
		    default :
			break;
		    }
		    bits += 2;
		    pixel >>= 2;
		}
	    }
	    break;
	case 2:
	    begRow = (begRow - fh)*2 +halfh;
	    begCol *= 2;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), halfw);
	    coef = *coefbuf + begRow*bpl;
	    pcode = root + begRow*bpl;
	    leaf =0;
	    for(row=begRow; row<endRow ; 
		++row, coef += bpl, pcode +=bpl)
	    {
		for(col=begCol; col<endCol ; ++col)
		{
		    if( pcode[col] == 1 ) continue;
		    ++ leaf;
		    if( bits >= 4)
		    {
			ERR_PASS(wp_symbolDecoding(pIPWState, &pixel, psymbol));
			bits =0;
		    }
		    switch(pixel & 0x03)
		    {
		    case 2:
			if(level > layer )
			{
			    ERR_PASS(wp_setSkipCoef(pIPWState,pheader, 
						    chan, row, col, level, 
						    areaId, root));
			}
			break;
		    case 3:
			coef[col] -= thre;
			break;
		    case 1:
			if(coef[col] >= 0)
			    coef[col] += thre;
			else
			    coef[col] -= thre;
			break;
		    default :
			break;
		    }
		    bits += 2;
		    pixel >>= 2;
		}
	    }
	    break;
	case 3:
	    begCol = (begCol - fw)*2 +halfw;
	    begRow = (begRow - fh)*2 +halfh;
	    endRow = MIN((begRow + sum), coefh);
	    endCol = MIN((begCol + sum), coefw);
	    coef = *coefbuf;
	    leaf =0;
	    for(idx=0, num=0; num<sum*sum  ; ++idx)
	    {
		if( y_scan[idx] >= sum ||
		    x_scan[idx] >= sum )
		    continue;
		++ num;
		row = begRow + y_scan[idx];
		col = begCol + x_scan[idx];
		if(row >= endRow || col >= endCol ) 
		    continue;

		if( *(root+col +row*bpl) ==1) continue;

		coef1 = coef + col + row*bpl ;
		++ leaf;
		if( bits >= 4)
		{
		    ERR_PASS(wp_symbolDecoding (pIPWState, &pixel, psymbol));
		    bits =0;
		}
		switch(pixel & 0x03)
		{
		case 2:
		    if(level > layer )
		    {
			ERR_PASS(wp_setSkipCoef(pIPWState,pheader, 
						chan, row, col, level, 
						areaId, root));
		    }
		    break;
		case 3:
		    *coef1 -= thre;
		    break;
		case 1:
		    if(*coef1 >= 0)
			*coef1 += thre;
		    else
			*coef1 -= thre;
		    break;
		default :
		    break;
		}
		bits += 2;
		pixel >>= 2;
	    }
	    break;
	default:
	    ERR_RETURN;
	}  
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *    Callback function for reading compressed image data.
 *
 * LongDescription:::
 *    Callback function for reading compressed image data.
 *
 * Input:::
 *  pIPWState::	     Ptr to IPCore state
 *  userData::       Ptr to FILE ( save compressed data )
 *  buf::            Ptr to byte buffer to write out.
 *  nbytes::         Number of bytes to write.
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_ReadCallback (
    IPWState	     *pIPWState,
    void             *userData,
    UInt8            *buf,
    UInt32           nbytes)
{
    PROLOGUE;
    Int32     idx;
    READINDATA  *pdata;

	/* Keep compiler happy */
    pIPWState = pIPWState;

    pdata = (READINDATA *)userData;
    for(idx=0; idx<nbytes; ++idx)
    {
	if(pdata->currIdx >= BUFFLEN)
	{
	    fread((UInt8 *)pdata->pbuf, (size_t)1, 
		  (size_t)BUFFLEN, pdata->fin);
	    pdata->currIdx = 0;
	}
	*(buf+idx) = *(pdata->pbuf+pdata->currIdx);
	++ pdata->currIdx;
    }
    readLen += idx;
    NORMAL_RETURN;
} 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Huffman decoding on a area.
 *
 * LongDescription:::
 *  Do huffman decoding in the assigned area. The difference between area and
 * zone is that zone can cover several level, but area is in one level. 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  coefbuf::        Ptr to coef-buffer
 *  bpl::            Bytes per line
 *  y::               Row number
 *  x::               Col. number
 *  width::           Area width
 *  height::          Area height
 *  thre::            Quantization parameter
 *  psymbol::         Ptr to symbol decoding structure
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_areaDecoding (
    IPWState           *pIPWState,
    Int16              *coefbuf,
    Int32              bpl,
    Int32              x,
    Int32              y,
    Int32              width,
    Int32              height,
    Int16              thre,
    SYMBOLREAD         *psymbol)
{
    PROLOGUE;
    Int32       row, col;
    Int16       *pcoef, len;
    UInt8       pixel, bits;

    pIPWState = pIPWState;
    pcoef = coefbuf + bpl*y;
    bits =0;
    pixel =0;
    len = -1;
    for(row=0; row<height; ++row, pcoef+=bpl)
    {
	for(col=0; col<width ; ++col)
	{
	    if(len == 0)
	    {
		if( (pixel & 0x03) == 3)
		{
		    *(pcoef+col+x) -= thre;
		}
		else if( (pixel & 0x03) == 1)
		{
		    *(pcoef+col+x) += thre;
		}
		bits += 2;
		pixel >>=2;
		if( bits >= 4)
		{
		    len = -1;
		    bits = pixel =0;
		}
		continue;
	    }
	    ERR_PASS(wp_symbolDecoding (pIPWState, &pixel, psymbol));
	    bits =0;
	    len =0;
	    if( (pixel & 0x03) == 3)
	    {
		*(pcoef+col+x) -= thre;
	    }
	    else if( (pixel & 0x03) == 1)
	    {
		*(pcoef+col+x) += thre;
	    }
	    bits += 2;
	    pixel >>= 2;
	}
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Serch the begining of code in new frame.
 *
 * LongDescription:::
 *  New planet is a new frame. We serch the begining code.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pdata::          Read in buffer structure
 *  readCallback::   Call back function for read compressed data
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  unstable
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_getBegin     (
    IPWState           *pIPWState,
    void               *pdata,
    ErrCode	(CDECL *readCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes))
{
    PROLOGUE;
    Int32     idx = 0;
    UInt8     code;

    pIPWState = pIPWState;
    do
    {
	ERR_PASS(readCallback(pIPWState, (void *)pdata, &code, sizeof(UInt8)));
		 ++idx;
    } while(  code == 0xf3 && idx<4);

    NORMAL_RETURN;
} 

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decoding for wavelet coefficients.
 *
 * LongDescription:::
 *    Before wave transform, we decoding the input to get wavelet 
 *  coefficients.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pheader::        Ptr to wave compression header.
 *  coefbuf::        Ptr to coef-buffer
 *  ReadCallback::   Call back function for reading compressed data 
 *  rData::          Ptr to hook for passing in user data
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_waveDecoding (
    IPWState           *pIPWState,
    WAVESUBHEAD        *pheader,
    Int16             **coefbuf,
    ErrCode	(CDECL *ReadCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes),
    void               *rData)
{
    PROLOGUE;
    Int32         chan, coefw, coefh, halfw, halfh, fw, fh;
    Int32         row, col, level, bpl;
    Int32         ncomps, x, y;
    Int16         *ptmp, index, *coef[3], thre;
    UInt8         *root, pixel;
    SYMBOLREAD   symbol;

    ERR_PASS(i_wd_waveDehuffInit (pIPWState, rData, ReadCallback));

    ncomps = pheader->ncomps;
    for(chan=0; chan<ncomps; ++chan)
    {
	coefw = pheader->coefw[chan][0];
	coefh = pheader->coefh[chan][0];
	*(coefbuf+chan)= (Int16 *)IPW_CALLOC(sizeof(Int16), coefw*coefh, 
					     cAllocatorData);
	ERR_GEN (*(coefbuf+chan) == NULL, cErrNomem);
	bpl = pheader->coefw[chan][0]; 

	level = pheader->pyramidN[chan]-1;

	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;

	ERR_PASS( i_wd_set_huffTable (pIPWState, chan, 1));

	for(row=0, ptmp=*(coefbuf+chan); row<halfh; ++row, ptmp+=bpl)
	{
	    for(col=0; col<halfw; ++col)
	    {
		ERR_PASS(i_wd_dc_decoding (pIPWState,  &index ));
		*(ptmp+col) = index;
	    }
	}
	for(row=0, ptmp=*(coefbuf+chan); row<halfh; ++row, ptmp+=bpl)
	{
	    for(col=0; col<halfw; ++col)
	    {
		ERR_PASS(i_wd_dc_decoding (pIPWState,  &index ));
		*(ptmp+col) |= (index<<8);
	    }
	}
    }

    ERR_PASS( wp_getBegin(pIPWState, rData, ReadCallback));
    symbol.passData = rData;
    symbol.ReadCallback = ReadCallback;
    symbol.readByte =0; 
    symbol.readBits =0;

    for(chan=0; chan<ncomps; ++chan)
    {
	coefw = pheader->coefw[chan][0];
	coefh = pheader->coefh[chan][0];

	bpl = pheader->coefw[chan][0]; 
	root = (UInt8 *)IPW_CALLOC(sizeof(UInt8), 
				  pheader->coefh[chan][0]*bpl, cAllocatorData);
	ERR_GEN (root == NULL, cErrNomem);

	level = pheader->pyramidN[chan]-1;

	coefw = pheader->coefw[chan][level];
	coefh = pheader->coefh[chan][level];
	halfw = (coefw+1)>>1;
	halfh = (coefh+1)>>1;

/* node decoding */
	coef[chan] = *(coefbuf+chan);
	thre = pheader->quant[chan];
	for( ; thre > 2; thre>>=1 )
	{
	    MEMSET((void *) root, 0, pheader->coefh[chan][0]*bpl);
	    level = pheader->pyramidN[chan]-1;

	    coefw = pheader->coefw[chan][level];
	    coefh = pheader->coefh[chan][level];
	    halfw = (coefw+1)>>1;
	    halfh = (coefh+1)>>1;

	    ptmp = coef[chan];
	    for(row=0; row<halfh; ++row, ptmp += bpl)
	    {
		for(col=halfw; col<coefw; col+=2)
		{
		    ERR_PASS(wp_symbolDecoding (pIPWState, &pixel, &symbol));
		    if( (pixel & 0x03) != 2 )
		    {
			if( (pixel & 0x03) == 3)
			    ptmp[col] -= thre;
			else if( (pixel & 0x03) ==1)
			{
			    if( ptmp[col] >= 0)
				ptmp[col] += thre;
			    else
				ptmp[col] -= thre;
			} 

			ERR_PASS(wp_zoneDecoding(pIPWState, pheader, 
						 &coef[chan], root, chan,
						 row,col, 1, thre, &symbol));
		    }
		    if( (col+1) < coefw)
		    {
			if( (pixel & 12) != 8 )
			{
			    if( (pixel & 12) == 12)
				ptmp[col+1] -= thre;
			    else if( (pixel & 0x0c) ==4)
			    {
				if( ptmp[col+1] >= 0)
				    ptmp[col+1] += thre;
				else
				    ptmp[col+1] -= thre;
			    } 
			    ERR_PASS(wp_zoneDecoding(pIPWState, pheader, 
						     &coef[chan],root, chan,
						     row,(col+1), 1, thre,
						     &symbol));
			}
		    }
		}
	    }
	    ptmp = coef[chan] + halfh*bpl;
	    for(row=halfh; row<coefh; ++row, ptmp += bpl)
	    {
		for(col=0; col<halfw; col+=2)
		{
		    ERR_PASS(wp_symbolDecoding (pIPWState, &pixel, &symbol));
		    if( (pixel & 0x03) != 2 )
		    {
			if( (pixel & 0x03) == 3)
			    ptmp[col] -= thre;
			else if( (pixel & 0x03) ==1)
			{
			    if( ptmp[col] >= 0)
				ptmp[col] += thre;
			    else
				ptmp[col] -= thre;
			} 
			ERR_PASS(wp_zoneDecoding(pIPWState, pheader, 
						 &coef[chan], root, chan,
						 row,col, 2, thre, &symbol));
		    }
		    if( (col+1) < halfw)
		    {
			if( (pixel & 0x0c) != 8 )
			{
			    if( (pixel & 0x0c) == 12)
				ptmp[col+1] -= thre;
			    else if( (pixel & 0x0c) ==4)
			    {
				if( ptmp[col+1] >= 0)
				    ptmp[col+1] += thre;
				else
				    ptmp[col+1] -= thre;
			    } 
			    ERR_PASS(wp_zoneDecoding(pIPWState, pheader, 
						     &coef[chan], root, chan,
						     row, (col+1), 2, 
						     thre, &symbol));
			}
		    }
		}
	    }
	    ptmp = coef[chan] + halfh*bpl;
	    for(row=halfh; row<coefh; ++row, ptmp += bpl)
	    {
		for(col=halfw; col<coefw; col += 2)
		{
		    ERR_PASS(wp_symbolDecoding (pIPWState, &pixel, &symbol));
		    if( (pixel & 0x03) != 2 )
		    {
			if( (pixel & 0x03) == 3)
			    ptmp[col] -= thre;
			else if( (pixel & 0x03) ==1)
			{
			    if( ptmp[col] >= 0)
				ptmp[col] += thre;
			    else
				ptmp[col] -= thre;
			} 
			ERR_PASS(wp_zoneDecoding(pIPWState, pheader, 
						 &coef[chan], root, chan,
						 row,col, 3, thre, &symbol));
		    }
		    if( (col+1) < coefw)
		    {
			if( (pixel & 0x0c) != 8 )
			{
			    if( (pixel & 0x0c) == 12)
				ptmp[col+1] -= thre;
			    else if( (pixel & 0x0c) ==4)
			    {
				if( ptmp[col+1] >= 0)
				    ptmp[col+1] += thre;
				else
				    ptmp[col+1] -= thre;
			    } 
			    ERR_PASS(wp_zoneDecoding(pIPWState, pheader, 
						     &coef[chan],root, chan,
						     row,(col+1), 3, 
						     thre, &symbol));
			}
		    }
		}
	    }

	    level=pheader->pyramidN[chan]-1;
	    x = pheader->coefw[chan][level] -1;
	    y = pheader->coefh[chan][level] -1;
	    halfw = (pheader->coefw[chan][level] +1)>>1;
	    halfh = (pheader->coefh[chan][level] +1)>>1;
	    for(--level, col=2; level>=pheader->min_layer[chan]; 
		--level, col *=2)
	    {
		fw = halfw;
		fh = halfh;
		coefw = pheader->coefw[chan][level];
		coefh = pheader->coefh[chan][level];
		halfw = (coefw +1) >>1;
		halfh = (coefh +1) >>1;
		x = (x - fw)*2 +halfw;
		y = (y - fh)*2 +halfh;
		if( (x+col) < coefw )
		{
		    ERR_PASS( wp_areaDecoding (pIPWState, coef[chan], bpl,
					       (x+col), 0, (coefw-x-col), 
					       coefh, thre, &symbol));
		}
		if( (y+col) < coefh )
		{
		    ERR_PASS( wp_areaDecoding (pIPWState, coef[chan], bpl, 
					       0, (y+col), coefw, 
					       (coefh-y-col), thre, &symbol));
		}
	    }
	    ERR_PASS(wp_symbolDecoding (pIPWState, &root[0], &symbol));
	    if( root[0] == 0 ) break; 
	}

        IPW_FREE(root, cAllocatorData);
    }

    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *   Decompress strip image and put them together.
 *
 * LongDescription:::
 *   Compressed image may be one strip or multi-strip. We decompress
 *   strip image then put them together.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  image::          Ptr to image buffer ptr
 *  w::              Image width
 *  h::              Image height
 *  bpl::            Image bytes per line
 *  ncomps::         Number of image channels 
 *  num_sub::        Number of image strips
 *  ReadCallback::  Call back function for reading compressed data
 *  userData::       Ptr to userDat passed to call back function
 *
 * Output:::
 *
 * Returns:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, color, multi-channel,
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_dwaveImageStrip (
    IPWState          *pIPWState,
    UInt8            **image,
    Int32              w,
    Int32              h,
    Int32              bpl,
    Int32              ncomps,
    Int16              num_sub,
    ErrCode    (CDECL *ReadCallback) (IPWState *pIPWState,
				      void *userData, UInt8 *buf, 
				      UInt32 nbytes),
    void               *userData)
{
    PROLOGUE;

    WAVESUBHEAD        waveSub;
    WSHSAVEFORM        waveSubRead;
    Int32              *lowTab[5], *highTab[4];
    Int16              *userBuf = NULL, *coef[3], *clap[3], mincoef, maxcoef;
    UInt8               *tempbuf[3], *ptm1, *ptm2;
    Int32              chan, coefBpl;
    Int32              level, row, coefh, coefw;
    Int32              width, height, bidx, line, col;

/* Init ptr. */
    lowTab[0] =NULL;
    tempbuf[0] = tempbuf[1] = tempbuf[2] = NULL;

/*  allocate buffer for reverse transform */
    row = w + 32;
    if((h +32) > row) row =h + 32; 
    userBuf= (Int16 *)IPW_CALLOC(sizeof(Int16), row, cAllocatorData);
    ERR_GEN (userBuf == NULL, cErrNomem);  

/*  Init table */
    level = 256 *2 +1; 
    lowTab[0] = (Int32 *)IPW_CALLOC(sizeof(Int32), level*9, cAllocatorData);
    ERR_GEN_CLEANUP (lowTab[0] == NULL, cErrNomem);
    highTab[0] = lowTab[0] + level*5;
    for(chan=1; chan<4; ++chan)
    {
	lowTab[chan] = lowTab[chan-1]+level;
	highTab[chan] = highTab[chan-1]+level;
    }
    lowTab[4] = lowTab[3] + level;
    ERR_PASS(i_wd_init_invertTbl (pIPWState, r_lowpass, r_highpass, 5,
				  4, level, &lowTab[0], &highTab[0] ));

    for (chan=0; chan<ncomps; chan++)
    {
	tempbuf[chan] = (UInt8 *)IPW_CALLOC(sizeof(UInt8), bpl*4, 
					    cAllocatorData);
	ERR_GEN_CLEANUP (tempbuf[chan] == NULL, cErrNomem);
    }

    for(bidx=0, row=0; bidx<num_sub; ++bidx)
    {
	ERR_PASS_CLEANUP(ReadCallback(pIPWState, userData,
				      (void *)&waveSubRead, 
				      sizeof(WSHSAVEFORM)));
	waveSub.ncomps = (Int16) waveSubRead.ncompsL;
	for(chan=0; chan<ncomps; ++chan)
	{
	    waveSub.pyramidN[chan] = (Int16)waveSubRead.pyramidN[chan];
	    waveSub.min_layer[chan] = (Int16)waveSubRead.min_layer[chan];
	    waveSub.mindc[chan] = (Int16)((waveSubRead.mindcH[chan] << 8)|
					  waveSubRead.mindcL[chan]);
	    waveSub.quant[chan] = (Int16)((waveSubRead.quantH[chan] << 8)|
					  waveSubRead.quantL[chan]);
	    for(level=0; level< waveSub.pyramidN[chan]; ++level)
	    {
		waveSub.coefw[chan][level] = 
		    (Int16)((waveSubRead.coefwH[chan][level] << 8)|
			    waveSubRead.coefwL[chan][level]);
		waveSub.coefh[chan][level] = 
		    (Int16)((waveSubRead.coefhH[chan][level] << 8)|
			    waveSubRead.coefhL[chan][level]);
	    }
	}	
	ERR_PASS_CLEANUP(wp_waveDecoding(pIPWState, &waveSub, coef, 
				 ReadCallback, userData)); 

	ERR_PASS_CLEANUP(wp_waveDequatization(pIPWState, &waveSub, coef));

	for(chan=0; chan<ncomps; ++chan)
	{
	    coefBpl = waveSub.coefw[chan][0];
	    for(level=waveSub.pyramidN[chan]-1; level>=0; --level)
	    {
		coefw = waveSub.coefw[chan][level];
		coefh = waveSub.coefh[chan][level];
		if(level >= 0)
		    ERR_PASS_CLEANUP(i_wd_waveletInvertTrans (pIPWState,
							      coefw, coefh,
							      coefBpl,
							      r_lowpass,
							      r_highpass,
							      &userBuf,
							      &coef[chan]));
		else
		    ERR_PASS_CLEANUP(i_wd_waveletInvertTrans_LUT (pIPWState,
								  coefw,
								  coefh,
								  coefBpl,
								  &lowTab[0],
								  &highTab[0],
								  &userBuf,
								  &coef[chan]));
	    }
	    if(row == 0) clap[chan] = coef[chan];
	    else
	    {
		if( chan <1)
		    clap[chan] = coef[chan]+4*coefBpl;
		else 
		    clap[chan] = coef[chan]+2*coefBpl;
	    }
	}
	width = waveSub.coefw[0][0];
	height = waveSub.coefh[0][0];

	if(row >0 ) 
	{
	    height -=4;
	    ERR_PASS_CLEANUP(i_wd_buffer2image (pIPWState, ncomps, bpl,
						0, 0, width, 4,
						tempbuf, coef,
						&mincoef, &maxcoef));
	}
	ERR_PASS_CLEANUP(i_wd_buffer2image (pIPWState, ncomps, bpl,
					    0, row, width,
					    height, image, clap,
					    &mincoef, &maxcoef));
/*	fprintf(stderr," min=%d max=%d\n", mincoef, maxcoef); */
	if(row >0)
	{
	    for(chan=0; chan<ncomps; ++chan)
	    {
		ptm1 = image[chan]+ (row-4)*bpl;
		ptm2 = tempbuf[chan];
		for(line=0; line<4; ++line, ptm1+=bpl, ptm2+=bpl)
		{
		    for(col=0; col<bpl; ++col)
			ptm1[col] = (ptm1[col]+ptm2[col])>>1;
		}
	    }
	} 
	for(chan=0; chan<ncomps; ++chan)
	    IPW_FREE(coef[chan], cAllocatorData);	

	row += height;
    }

    IPW_FREE(lowTab[0], cAllocatorData);
    IPW_FREE(userBuf, cAllocatorData);
    for(chan=0; chan<ncomps; ++chan)
	IPW_FREE(tempbuf[chan], cAllocatorData);

    NORMAL_RETURN;

ERR_CLEANUP:

    if(lowTab[0] != NULL)
	IPW_FREE(lowTab[0], cAllocatorData);
    if(userBuf != NULL)
	IPW_FREE(userBuf, cAllocatorData);
    for(chan=0; chan<ncomps; ++chan)
    {
	if(tempbuf[chan] != NULL)
	    IPW_FREE(tempbuf[chan], cAllocatorData);
    }
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (WAVElet) Decompress a  file to 8 bit gray PIXR or a 3 channel x 
 *  8 bit RGB PIXR.
 *
 * LongDescription:::
 *    Decompress a  stored WAVElet format compressed image to
 *    an 8 bit gray pixr or a 3 channel x 8 bit RGB PIXR according 
 *    to compressed image information.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fdWAVE::         FILE ptr for the compressed file
 *
 * Output:::
 *  dstPixrPtr::     Ptr to ptr to 8 bit gray dest PIXR.
 *                   If *dstPixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_waveDecompress (
    IPWState           *pIPWState,
    FILE               *fdWAVE,
    PIXR              **dstPixrPtr)
{
    PROLOGUE;
    PIXR               *pixrBand;
    Int32              w, h, bpl,chan, ncomps;
    PIXR               *dstPixr;
    WAVEIMAGEHEAD      waveHead;
    WIHSAVEFORM        waveHeadRead;
    UInt8              *image[3];
    READINDATA         rData;

/* Verify arguments */
    ERR_GEN(fdWAVE == NULL, cErrNullPtr);
    ERR_GEN(dstPixrPtr == NULL, cErrNullPtr);

    readLen =0;
    rData.pbuf = NULL;
    rData.pbuf = (UInt8 *)IPW_CALLOC(sizeof(UInt8), BUFFLEN, 
					     cAllocatorData);
    ERR_GEN (rData.pbuf == NULL, cErrNomem);
    rData.fin = fdWAVE;
    rData.currIdx = BUFFLEN;

/* Read header */
    ERR_PASS(wp_ReadCallback(pIPWState, (void *)&rData, (UInt8*)&waveHeadRead, 
			     sizeof(WIHSAVEFORM)));
    waveHead.version = (Int16)waveHeadRead.versionL;
    waveHead.ncomps = (Int16)waveHeadRead.ncomps;
    waveHead.num_sub = (Int16)waveHeadRead.num_sub;
    waveHead.imageW = (Int16)((waveHeadRead.imageWH << 8) |
			      waveHeadRead.imageWL);
    waveHead.imageH = (Int16)((waveHeadRead.imageHH << 8) |
			      waveHeadRead.imageHL);
    ncomps = waveHead.ncomps;
    w = waveHead.imageW;
    h = waveHead.imageH;

/* Check whether is current version */
    ERR_GEN (waveHead.version != WAVE_VERSION, cErrBadVersion);

/*
  If dstPixrPtr is NULL, assume that the caller wants one created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
/* Use a local pixr variable to avoid all the indirection */
    dstPixr = *dstPixrPtr;
    if (dstPixr == NULL)
    {
	if(ncomps ==1)
	    ERR_PASS_CLEANUP( w_createPixr(pIPWState, w, h, cDepth8, 
					   Int32_TO_ScaledInt1616(300), 
					   Int32_TO_ScaledInt1616(300),
					   cFrameBits, &dstPixr));
	else
	    ERR_PASS_CLEANUP( w_createPixrExtended(pIPWState, w, h, cDepth8, 
						 Int32_TO_ScaledInt1616(300), 
						 Int32_TO_ScaledInt1616(300), 
                                   cFrameBits, 3, cColorRGB, NULL, &dstPixr));
    }
    else
    {
	if(ncomps ==1)
	    ERR_GEN_PIXR(dstPixr,cByteOrderBig,cDepth8, cChannels1, cFrame32);
	else
	    ERR_GEN_PIXR(dstPixr, cByteOrderBig,cDepth8,cChannels3, cFrame32);
    }

    *dstPixrPtr = dstPixr;
    ERR_PASS( w_pixrGetBpl(pIPWState, dstPixr, &bpl));
    for (chan=0; chan<ncomps; chan++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, dstPixr, chan, &pixrBand));
        ERR_GEN(pixrBand == NULL, cErrInvalidParm);
	ERR_PASS( w_pixrGetImage(pIPWState, pixrBand, &(image[chan])));
    }
   
    ERR_PASS_CLEANUP(wp_dwaveImageStrip( pIPWState, image, w, h,
				  bpl, ncomps, waveHead.num_sub,
				  wp_ReadCallback, (void *)&rData));

    IPW_FREE(rData.pbuf, cAllocatorData);
    *dstPixrPtr = dstPixr;
    NORMAL_RETURN;

ERR_CLEANUP:

    if(rData.pbuf != NULL)
	IPW_FREE(rData.pbuf, cAllocatorData);

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  (WAVElet) Decompress a  file to 8 bit gray PIXR or a 3 channel x 
 *  8 bit RGB PIXR.
 *
 * LongDescription:::
 *    Decompress a  stored WAVElet format compressed image to
 *    an 8 bit gray pixr or a 3 channel x 8 bit RGB PIXR according 
 *    to compressed image information. Callback function is passed
 *    in by user.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  ReadCallback::   Call back function for reading compressed data 
 *  userData::       Ptr to hook for passing in user data
 *
 * Output:::
 *  dstPixrPtr::     Ptr to ptr to 8 bit gray dest PIXR.
 *                   If *dstPixrPtr == NULL on input, PIXR is allocated and
 *                   filled in.  Else, passed PIXR is used and just filled in.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  Wavelet
 *
 * Keywords:::
 *  Wavelet, compression, file io, format conversion
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  xzhu::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_waveDecompress_option (
    IPWState           *pIPWState,
    ErrCode	(CDECL *ReadCallback) (IPWState *pIPWState,
				       void *userData, UInt8 *buf, 
				       UInt32 nbytes),
    void               *userData,
    PIXR              **dstPixrPtr)
{
    PROLOGUE;
    PIXR               *pixrBand;
    Int32               w, h, bpl,chan, ncomps;
    PIXR               *dstPixr;
    WAVEIMAGEHEAD       waveHead;
    WIHSAVEFORM         waveHeadRead;
    UInt8              *image[3];

/* Verify arguments */
    ERR_GEN(userData == NULL, cErrNullPtr);
    ERR_GEN(dstPixrPtr == NULL, cErrNullPtr);

/* Read header */
    readLen = sizeof(WAVEIMAGEHEAD);
    ERR_PASS(ReadCallback(pIPWState, (void *)userData, (UInt8*)&waveHeadRead, 
			     sizeof(WIHSAVEFORM)));
    waveHead.version = (Int16)waveHeadRead.versionL;
    waveHead.ncomps = (Int16)waveHeadRead.ncomps;
    waveHead.num_sub = (Int16)waveHeadRead.num_sub;
    waveHead.imageW = (Int16)((waveHeadRead.imageWH << 8) |
			      waveHeadRead.imageWL);
    waveHead.imageH = (Int16)((waveHeadRead.imageHH << 8) |
			      waveHeadRead.imageHL);
    ncomps = waveHead.ncomps;
    w = waveHead.imageW;
    h = waveHead.imageH;

/* Check whether is current version */
    ERR_GEN (waveHead.version != WAVE_VERSION, cErrBadVersion);

/*
  If dstPixrPtr is NULL, assume that the caller wants one created.
  Note that this requires the caller to initialize the destination
  pixr ptr to NULL.
*/
/* Use a local pixr variable to avoid all the indirection */
    dstPixr = *dstPixrPtr;
    if (dstPixr == NULL)
    {
	if(ncomps ==1)
	    ERR_PASS( w_createPixrNoInit(pIPWState, w, h, cDepth8, 
					   Int32_TO_ScaledInt1616(300), 
					   Int32_TO_ScaledInt1616(300),
					   cFrameBits, &dstPixr));
	else
	    ERR_PASS( w_createPixrExtendedNoInit(pIPWState, w, h, cDepth8, 
						 Int32_TO_ScaledInt1616(300), 
						 Int32_TO_ScaledInt1616(300), 
                                   cFrameBits, 3, cColorRGB, NULL, &dstPixr));
    }
    else
    {
	if(ncomps ==1)
	    ERR_GEN_PIXR(dstPixr,cByteOrderBig,cDepth8, cChannels1, cFrame32);
	else
	    ERR_GEN_PIXR(dstPixr, cByteOrderBig,cDepth8,cChannels3, cFrame32);
    }

    *dstPixrPtr = dstPixr;
    ERR_PASS( w_pixrGetBpl(pIPWState, dstPixr, &bpl));
    for (chan=0; chan<ncomps; chan++)
    {
	ERR_PASS( w_pixrGetChannel(pIPWState, dstPixr, chan, &pixrBand));
        ERR_GEN(pixrBand == NULL, cErrInvalidParm);
	ERR_PASS( w_pixrGetImage(pIPWState, pixrBand, &(image[chan])));
    }
   
    ERR_PASS(wp_dwaveImageStrip( pIPWState, image, w, h,
				  bpl, ncomps, waveHead.num_sub,
				  ReadCallback, (void *)userData));

    *dstPixrPtr = dstPixr;
    NORMAL_RETURN;

}

#undef _IW_DWAVE_C_

/*
$Log: iw_dwave.c,v $
Revision 1.1.1.1  1999/09/20 18:19:18  jlin
Imported sources

Revision 1.17  1998/01/15 01:49:46  ddavies
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Fix WATCOM warnings about unused procedure args.

Revision 1.16  1998/01/14 00:13:40  ddavies
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.15  1997/12/05 19:50:27  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Moved x_scan and y_scan from iw_cwave to iw_dwave.

Revision 1.14  1997/10/28 23:45:11  xzhu
xzhu on Tue Oct 28 15:43:51 PST 1997

iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

Fixed bug in decoding, which appears occasionally.

Revision 1.13  1997/10/01 23:33:32  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Added version info in and check it when decoding.

Revision 1.12  1997/09/19 19:09:30  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Changed wavelet coding method, it improved image quality. It can be progressive decoding.

Revision 1.11  1997/07/11 01:59:10  ddavies
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Change procedure pointers with CDECLs to make MSVC happy:
type CDECL (*procPtrName) -> type (CDECL *procPtrName).

Revision 1.10  1997/05/17 01:21:51  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: No print out for release version.

Revision 1.9  1997/05/10 00:05:43  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Fixed header byte order problem.

Revision 1.8  1997/04/28 16:27:09  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Added A=A to make compiler happy.

Revision 1.7  1997/04/25 22:07:21  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Changed coding algorithm and added new wrap functions.

Revision 1.6  1997/04/15 22:13:25  ddavies
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

Use non-initialized PIXR creation where possible.

Revision 1.5  1996/11/08 23:11:47  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Passed callback function to wp_waveCoding() and wp_wavedecoding().

Revision 1.4  1996/10/28 19:18:49  xzhu
xzhu on Mon Oct 28 11:14:46 PST 1996

iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

w_waveDecompress suports color image. But in description, only gray scale is mentioned. So we put descriptions about color image in.

Revision 1.3  1996/10/16 00:11:52  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Ipshared wavelet functions' name has been changed to fit ipshared style. So we did same changes on these functions' name.

Revision 1.2  1996/10/15 18:07:30  xzhu
iw_dwave.c -> /product/ipcore/ipwrap/src/RCS/iw_dwave.c,v

CommonCheck: Fixed coreCheck warnings and WATCOM complier warnings.

Revision 1.1  1996/10/14 19:12:48  xzhu
Initial revision

*/

