/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/* file - pcx.c
 *
 * $Header: /project/icons/cvsroot/ipcore/src/ipwrap/src/pcx.c,v 1.1.1.1 1999/09/20 18:19:18 jlin Exp $
 *
 * DESCRIPTION
 *   PCX file reader/writer (invoked via the imageio interface).
 *
 * $Log: pcx.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:18  jlin
 * Imported sources
 *
 * Revision 1.12  1995/11/13 20:06:12  ejaquith
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Changed wp_CONVERTFROMINTEL16() to wp_stupidPCXconvertFromIntel16()
 * invocation.  Included "io.prv" for prototype declaration.
 * Added needed IPWrap-related include files to include "io.prv".
 *
 * Revision 1.11  1995/10/16  22:57:46  dferrell
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.10  1995/10/13  02:03:32  ejaquith
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * The PCX spec. dictates that the only acceptable depth 4 image is to be
 * written with its data stored in 4 separate binary planes, with a sick
 * joke of line interlacing.  Apparently Mark Sydner and I are having
 * creative differences.  (off with his head!)
 *
 * Revision 1.9  1995/09/12  18:24:38  dferrell
 * dferrell on Tue Sep 12 11:24:14 PDT 1995
 *
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Added wp_ prefix to all instances of CONVERTFROMINTEL16.
 *
 * Revision 1.8  1995/09/01  02:23:16  ejaquith
 * ejaquith on Thu Aug 31 19:23:15 PDT 1995
 *
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.7  1995/08/31  21:15:46  ejaquith
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.6  1995/08/18  01:54:41  ejaquith
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Removed inclusion of "ipstate.pub" and "io.pub".
 * Moved routine CONVERTFROMINTEL16() from this file to pcx2.c
 *
 * Revision 1.5  1995/08/17  22:04:47  ejaquith
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.4  1995/07/27  00:03:00  ejaquith
 * ejaquith on Wed Jul 26 17:00:58 PDT 1995
 *
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Changed pcx_read_file() to pcx_read_line() because that's what it
 * does and that ALWAYS annoyed me!
 * Fixed one-byte-too-far-in-buffer read for writing.  Now I have to say,
 * this damn code is a real piece of work.  If I ever see that consultant
 * again, I will STRANGLE him.
 * Fixed unit'ed mem read in colormap when using 8bpp images.
 * The palette was never filled, but allocated, so now we fill it full of
 * zeros.
 *
 * Revision 1.3  1994/12/08  03:08:38  dferrell
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Added typecastes so that MPW could compile it.
 *
 * Revision 1.2  1994/12/06  22:33:06  dferrell
 * dferrell on Tue Dec 6 14:31:57 PST 1994
 *
 * pcx.c -> /product/ipcore/ipwrap/src/RCS/pcx.c,v
 *
 * Removed unnecessary #include of malloc.h.  stdlib.h has all the same stuff
 * in it and MPW doesn't have a malloc.h anyway.
 *
 * Revision 1.1  1994/10/21  17:18:20  ejaquith
 * Initial revision
 *
 * 
 *    Rev 1.36   18 Oct 1994 18:45:40   EHOPPE
 * 
 * Changed pack24 proc and callers to fix 24bit write.  Write all types
 * at version 5.  Cleaned up some unneeded logic.
 * 
 *    Rev 1.35   28 Sep 1994 16:39:18   EHOPPE
 * 
 * Added special case to fix binary PCX writing reverse contrast bug.  This
 * is should be reexamined: a special case should not be needed.
 * 
 *    Rev 1.34   27 Sep 1994 13:52:56   EHOPPE
 * Removed binary inversion kludge.
 * 
 *    Rev 1.33   26 Sep 1994 15:26:10   EHOPPE
 * Added use of resolution information.  Removed localized reverse contrast
 * kludges.  Fixed THE reverse contrast problem in imageio.  Still need to add
 * 'invert' flag for other clients.  Fixed gray images read as color (GIF, BMP
 * Added byte swapping code and removed nonportable language features.
 * 
 *    Rev 1.32   21 Sep 1994 14:19:36   EHOPPE
 * 
 * Read the 256 palette for 8-bit gray.
 * 
 *    Rev 1.31   20 Sep 1994 16:39:36   EHOPPE
 * 
 * Use the new GData resolution fields.
 * 
 *    Rev 1.30   19 Sep 1994 09:13:20   EHOPPE
 * Use bytes per line field from pcx header.  Cleaned up other 
 * strange code related to byte counts and image dimensions.
 */

/*
 * INCLUDES
 */
#include <stdio.h>
#include <stdlib.h>

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "imagerep.h"
#include "ptypes.h"
#include "gdefs.h"
#include "pcx.h"
#include "imageio.h"

/* prototypes */
#include "io.prv"
#include "gfio.h"
#include "gfiopro.h"
#include "epsutil.h"


/*
 * CONSTANTS
 */

/*
 * MACROS
 */

#if 0
static void
Break(int i)
{
    fprintf (stderr, "%s line %d: HEAPCHECK failed with code %d.\n",
        __FILE__, __LINE__, i);
}
#define HEAPCHECK() { int h = _heapchk(); if (h!=_HEAPOK&&h!=_HEAPEMPTY) { Break(h); } }
#else
#define HEAPCHECK()
#define printf(a)   /* ejaquith added this to wipe out prints */
#endif

/*
 * TYPEDEFS
 */

struct State {
   void*          file;
};

/*
 * ENUMS
 */

/*
 * GLOBAL VARIABLES
 */

/*--ejaquith-- added prototype after taking from pcx.h! */
static void pcx_planes( unsigned char *pixels, unsigned char *bitplanes,
			int bytesperline, int planes, int bitsperpixel );


static unsigned char pcx_buffer[8192];

/*  static int pcx_fmode; */
char *pcx_line_ptr;

/* this is the default EGA 16 color palette */
unsigned char default_palette[48] = {
                            0x00, 0x00, 0x0e, 0x00, 0x52, 0x07, 0x2c, 0x00,
                            0x0e, 0x00, 0x00, 0x00, 0xf8, 0x01, 0x2c, 0x00,
                            0x85, 0x0f, 0x42, 0x00, 0x21, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0x6a, 0x24, 0x9b, 0x49, 0xa1, 0x5e,
                            0x90, 0x5e, 0x18, 0x5e, 0x84, 0x14, 0xd9, 0x95,
                            0xa0, 0x14, 0x12, 0x00, 0x06, 0x00, 0x68, 0x1f
                           };

unsigned char bw_palette[48] = {
                            0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
                            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                            0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
                           };

unsigned char gray16_palette[48] =
{
    0x00, 0x00, 0x00, 0x11, 0x11, 0x11, 0x22, 0x22, 0x22, 0x33, 0x33, 0x33,
    0x44, 0x44, 0x44, 0x55, 0x55, 0x55, 0x66, 0x66, 0x66, 0x77, 0x77, 0x77,
    0x88, 0x88, 0x88, 0x99, 0x99, 0x99, 0xaa, 0xaa, 0xaa, 0xbb, 0xbb, 0xbb,
    0xcc, 0xcc, 0xcc, 0xdd, 0xdd, 0xdd, 0xee, 0xee, 0xee, 0xff, 0xff, 0xff,
};

unsigned char color16_palette[48] =
{
    0, 0, 0,
    128, 0, 0,
    0, 128, 0,
    128, 128, 0,
    0, 0, 128,
    128, 0, 128,
    0, 128, 128,
    192, 192, 192,
    128, 128, 128,
    255, 0, 0,
    0, 255, 0,
    255, 255, 0,
    0, 0, 255,
    255, 0, 255,
    0, 255, 255,
    255, 255, 255
};


unsigned char gray256_palette[768] = 
{
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x03,
	0x04, 0x04, 0x04, 0x05, 0x05, 0x05, 0x06, 0x06, 0x06, 0x07, 0x07, 0x07,
	0x08, 0x08, 0x08, 0x09, 0x09, 0x09, 0x0a, 0x0a, 0x0a, 0x0b, 0x0b, 0x0b,
	0x0c, 0x0c, 0x0c, 0x0d, 0x0d, 0x0d, 0x0e, 0x0e, 0x0e, 0x0f, 0x0f, 0x0f,
	0x10, 0x10, 0x10, 0x11, 0x11, 0x11, 0x12, 0x12, 0x12, 0x13, 0x13, 0x13,
	0x14, 0x14, 0x14, 0x15, 0x15, 0x15, 0x16, 0x16, 0x16, 0x17, 0x17, 0x17,
	0x18, 0x18, 0x18, 0x19, 0x19, 0x19, 0x1a, 0x1a, 0x1a, 0x1b, 0x1b, 0x1b,
	0x1c, 0x1c, 0x1c, 0x1d, 0x1d, 0x1d, 0x1e, 0x1e, 0x1e, 0x1f, 0x1f, 0x1f,
	0x20, 0x20, 0x20, 0x21, 0x21, 0x21, 0x22, 0x22, 0x22, 0x23, 0x23, 0x23,
	0x24, 0x24, 0x24, 0x25, 0x25, 0x25, 0x26, 0x26, 0x26, 0x27, 0x27, 0x27,
	0x28, 0x28, 0x28, 0x29, 0x29, 0x29, 0x2a, 0x2a, 0x2a, 0x2b, 0x2b, 0x2b,
	0x2c, 0x2c, 0x2c, 0x2d, 0x2d, 0x2d, 0x2e, 0x2e, 0x2e, 0x2f, 0x2f, 0x2f,
	0x30, 0x30, 0x30, 0x31, 0x31, 0x31, 0x32, 0x32, 0x32, 0x33, 0x33, 0x33,
	0x34, 0x34, 0x34, 0x35, 0x35, 0x35, 0x36, 0x36, 0x36, 0x37, 0x37, 0x37,
	0x38, 0x38, 0x38, 0x39, 0x39, 0x39, 0x3a, 0x3a, 0x3a, 0x3b, 0x3b, 0x3b,
	0x3c, 0x3c, 0x3c, 0x3d, 0x3d, 0x3d, 0x3e, 0x3e, 0x3e, 0x3f, 0x3f, 0x3f,
	0x40, 0x40, 0x40, 0x41, 0x41, 0x41, 0x42, 0x42, 0x42, 0x43, 0x43, 0x43,
	0x44, 0x44, 0x44, 0x45, 0x45, 0x45, 0x46, 0x46, 0x46, 0x47, 0x47, 0x47,
	0x48, 0x48, 0x48, 0x49, 0x49, 0x49, 0x4a, 0x4a, 0x4a, 0x4b, 0x4b, 0x4b,
	0x4c, 0x4c, 0x4c, 0x4d, 0x4d, 0x4d, 0x4e, 0x4e, 0x4e, 0x4f, 0x4f, 0x4f,
	0x50, 0x50, 0x50, 0x51, 0x51, 0x51, 0x52, 0x52, 0x52, 0x53, 0x53, 0x53,
	0x54, 0x54, 0x54, 0x55, 0x55, 0x55, 0x56, 0x56, 0x56, 0x57, 0x57, 0x57,
	0x58, 0x58, 0x58, 0x59, 0x59, 0x59, 0x5a, 0x5a, 0x5a, 0x5b, 0x5b, 0x5b,
	0x5c, 0x5c, 0x5c, 0x5d, 0x5d, 0x5d, 0x5e, 0x5e, 0x5e, 0x5f, 0x5f, 0x5f,
	0x60, 0x60, 0x60, 0x61, 0x61, 0x61, 0x62, 0x62, 0x62, 0x63, 0x63, 0x63,
	0x64, 0x64, 0x64, 0x65, 0x65, 0x65, 0x66, 0x66, 0x66, 0x67, 0x67, 0x67,
	0x68, 0x68, 0x68, 0x69, 0x69, 0x69, 0x6a, 0x6a, 0x6a, 0x6b, 0x6b, 0x6b,
	0x6c, 0x6c, 0x6c, 0x6d, 0x6d, 0x6d, 0x6e, 0x6e, 0x6e, 0x6f, 0x6f, 0x6f,
	0x70, 0x70, 0x70, 0x71, 0x71, 0x71, 0x72, 0x72, 0x72, 0x73, 0x73, 0x73,
	0x74, 0x74, 0x74, 0x75, 0x75, 0x75, 0x76, 0x76, 0x76, 0x77, 0x77, 0x77,
	0x78, 0x78, 0x78, 0x79, 0x79, 0x79, 0x7a, 0x7a, 0x7a, 0x7b, 0x7b, 0x7b,
	0x7c, 0x7c, 0x7c, 0x7d, 0x7d, 0x7d, 0x7e, 0x7e, 0x7e, 0x7f, 0x7f, 0x7f,
	0x80, 0x80, 0x80, 0x81, 0x81, 0x81, 0x82, 0x82, 0x82, 0x83, 0x83, 0x83,
	0x84, 0x84, 0x84, 0x85, 0x85, 0x85, 0x86, 0x86, 0x86, 0x87, 0x87, 0x87,
	0x88, 0x88, 0x88, 0x89, 0x89, 0x89, 0x8a, 0x8a, 0x8a, 0x8b, 0x8b, 0x8b,
	0x8c, 0x8c, 0x8c, 0x8d, 0x8d, 0x8d, 0x8e, 0x8e, 0x8e, 0x8f, 0x8f, 0x8f,
	0x90, 0x90, 0x90, 0x91, 0x91, 0x91, 0x92, 0x92, 0x92, 0x93, 0x93, 0x93,
	0x94, 0x94, 0x94, 0x95, 0x95, 0x95, 0x96, 0x96, 0x96, 0x97, 0x97, 0x97,
	0x98, 0x98, 0x98, 0x99, 0x99, 0x99, 0x9a, 0x9a, 0x9a, 0x9b, 0x9b, 0x9b,
	0x9c, 0x9c, 0x9c, 0x9d, 0x9d, 0x9d, 0x9e, 0x9e, 0x9e, 0x9f, 0x9f, 0x9f,
	0xa0, 0xa0, 0xa0, 0xa1, 0xa1, 0xa1, 0xa2, 0xa2, 0xa2, 0xa3, 0xa3, 0xa3,
	0xa4, 0xa4, 0xa4, 0xa5, 0xa5, 0xa5, 0xa6, 0xa6, 0xa6, 0xa7, 0xa7, 0xa7,
	0xa8, 0xa8, 0xa8, 0xa9, 0xa9, 0xa9, 0xaa, 0xaa, 0xaa, 0xab, 0xab, 0xab,
	0xac, 0xac, 0xac, 0xad, 0xad, 0xad, 0xae, 0xae, 0xae, 0xaf, 0xaf, 0xaf,
	0xb0, 0xb0, 0xb0, 0xb1, 0xb1, 0xb1, 0xb2, 0xb2, 0xb2, 0xb3, 0xb3, 0xb3,
	0xb4, 0xb4, 0xb4, 0xb5, 0xb5, 0xb5, 0xb6, 0xb6, 0xb6, 0xb7, 0xb7, 0xb7,
	0xb8, 0xb8, 0xb8, 0xb9, 0xb9, 0xb9, 0xba, 0xba, 0xba, 0xbb, 0xbb, 0xbb,
	0xbc, 0xbc, 0xbc, 0xbd, 0xbd, 0xbd, 0xbe, 0xbe, 0xbe, 0xbf, 0xbf, 0xbf,
	0xc0, 0xc0, 0xc0, 0xc1, 0xc1, 0xc1, 0xc2, 0xc2, 0xc2, 0xc3, 0xc3, 0xc3,
	0xc4, 0xc4, 0xc4, 0xc5, 0xc5, 0xc5, 0xc6, 0xc6, 0xc6, 0xc7, 0xc7, 0xc7,
	0xc8, 0xc8, 0xc8, 0xc9, 0xc9, 0xc9, 0xca, 0xca, 0xca, 0xcb, 0xcb, 0xcb,
	0xcc, 0xcc, 0xcc, 0xcd, 0xcd, 0xcd, 0xce, 0xce, 0xce, 0xcf, 0xcf, 0xcf,
	0xd0, 0xd0, 0xd0, 0xd1, 0xd1, 0xd1, 0xd2, 0xd2, 0xd2, 0xd3, 0xd3, 0xd3,
	0xd4, 0xd4, 0xd4, 0xd5, 0xd5, 0xd5, 0xd6, 0xd6, 0xd6, 0xd7, 0xd7, 0xd7,
	0xd8, 0xd8, 0xd8, 0xd9, 0xd9, 0xd9, 0xda, 0xda, 0xda, 0xdb, 0xdb, 0xdb,
	0xdc, 0xdc, 0xdc, 0xdd, 0xdd, 0xdd, 0xde, 0xde, 0xde, 0xdf, 0xdf, 0xdf,
	0xe0, 0xe0, 0xe0, 0xe1, 0xe1, 0xe1, 0xe2, 0xe2, 0xe2, 0xe3, 0xe3, 0xe3,
	0xe4, 0xe4, 0xe4, 0xe5, 0xe5, 0xe5, 0xe6, 0xe6, 0xe6, 0xe7, 0xe7, 0xe7,
	0xe8, 0xe8, 0xe8, 0xe9, 0xe9, 0xe9, 0xea, 0xea, 0xea, 0xeb, 0xeb, 0xeb,
	0xec, 0xec, 0xec, 0xed, 0xed, 0xed, 0xee, 0xee, 0xee, 0xef, 0xef, 0xef,
	0xf0, 0xf0, 0xf0, 0xf1, 0xf1, 0xf1, 0xf2, 0xf2, 0xf2, 0xf3, 0xf3, 0xf3,
	0xf4, 0xf4, 0xf4, 0xf5, 0xf5, 0xf5, 0xf6, 0xf6, 0xf6, 0xf7, 0xf7, 0xf7,
	0xf8, 0xf8, 0xf8, 0xf9, 0xf9, 0xf9, 0xfa, 0xfa, 0xfa, 0xfb, 0xfb, 0xfb,
	0xfc, 0xfc, 0xfc, 0xfd, 0xfd, 0xfd, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff
}; 


/*
 * STATIC VARIABLES
 */

/*
 * LOCAL FUNCTION PROTOTYPES
 */

int pcx_line( char *buf, void *fp, unsigned int length );
int FillPCXHead( LPPCXHEAD pPCX, GDATA *lpDIB );

/*
 * FUNCTION DEFINITIONS
 */

int IMXDLbegin_write( GDATA *pcx_param, void *ft )
{
    pcxhead         pcx;
    
    /*  fill in and write the pcx header info from the OptBlock */
    FillPCXHead( &pcx, pcx_param );
/*--ejaquith--added call here! */
wp_stupidPCXconvertFromIntel16(&pcx);
    if ( _lwrite( ft, (LPSTR)&pcx, sizeof(pcxhead)) != sizeof(pcxhead ) )
    {
        printf("error couldn't write PCX header %d\n");
        _lclose( ft );
        ft = NULL; 
        return( 0 );
    }  
    return( 1 );
}


/*                      */
/*   Helper functions   */
/*                      */


int FillPCXHead(LPPCXHEAD pPCX, GDATA *lpDIB)
{
    int i;
    unsigned char *palette_ptr; /*--ejaquith--added unsigned */
    int x;
    
    /*  first fill in the easy ones...   */

    for( x = 0; x < 58; x++ )
        pPCX->filler[x] = 0x0;
    pPCX->man = 0x0a;     /*  always!!! */
    pPCX->enc = 1;        /*  PCX encoding scheme */
    pPCX->win[0] = pPCX->win[1] = 0;
    pPCX->win[2] = (int)lpDIB->x_dim - 1;
    pPCX->win[3] = (int)lpDIB->y_dim - 1;
    pPCX->hrez = (short)lpDIB->xDPI;
    pPCX->vrez = (short)lpDIB->yDPI;
    pPCX->reserved = 0;
  

    /*  now fill in the tricky ones... */
    switch ( lpDIB->depth ) 
    {
        case 1:
            lpDIB->gray_or_color = 1;    /* b&w */
            pPCX->ver = 5;
            pPCX->bits = 1;
            pPCX->nplanes = 1;
            pPCX->palinfo = 0x01;  /* color or b&w */
            pPCX->byteline = (short)( (lpDIB->x_dim + 31) / 32 * 4 );
                /* PROBLEM: this should be passed in not "computed"
                 */
            break;

        case 4:
            pPCX->ver = 5;
	    /* --10oct95--ejaquith
	     * which makes this (bits per pixel) 1 when there are 4 planes.
	     */
            pPCX->bits = 1;
	    /* --10oct95--ejaquith
	     * I will kill you, Mark Sydner, if I ever see you again.
	     * The spec for PCX says 4 color images MUST be 4 planes
	     * of depth 1, you idiot!!
	     */
            pPCX->nplanes = 4;
            pPCX->palinfo = lpDIB->gray_or_color;  /* color or b&w */
	    /* --10oct95--ejaquith
	     * no you idiot.  can't you do anything right?
	     * it's four binary planes, each with this many bytes per line.
	     *  pPCX->byteline = (short)( (lpDIB->x_dim * 4) 
	     *   + 31) / 32 * 4 );
	     */
	     pPCX->byteline = (short)( (lpDIB->x_dim + 31) / 32 * 4 );
                /* PROBLEM: this should be passed in not "computed"
                 */
            break;

        case 8:
            pPCX->ver = 5;
            pPCX->bits =  8;
            pPCX->nplanes = 1;
            pPCX->palinfo = lpDIB->gray_or_color;  /* color or b&w */
            pPCX->byteline = (short)( ((lpDIB->x_dim + 3) / 4) * 4 );
            break;

        case 24:
            pPCX->ver = 5;
            pPCX->bits =  8;
            pPCX->nplanes = 3;
            pPCX->palinfo = 0;  /* no palette */
            pPCX->byteline = (short)( ((lpDIB->x_dim + 3) / 4) * 4 );
            break;           
    
    } /* End switch */
  
    lpDIB->bytesPerLine = pPCX->byteline;
        /* PROBLEM: the other fields are set by the client. 
         *   This one should be too for PCX. --EHoppe (see above)
         */
  
    /*  now go after the npalette info */
    switch( lpDIB->nplanes ) 
    {
        case 4:  /* this goes up front */
            /* 10oct95--ejaquith.  this is totally wrong; taken out
	     * pPCX->nplanes = 1;
	     */
            if( pPCX->palinfo == 1 )
                palette_ptr = &color16_palette[0];
            else
                palette_ptr = &gray16_palette[0];
                        
            for (i=0; i<48; i++) 
            {
                pPCX->colmap[i] = *palette_ptr++;
            }  
            break;
        

        case 3:
        case 1:
        default: 
            /* 10oct95--ejaquith.  this is totally wrong; taken out
	     * pPCX->nplanes = lpDIB->nplanes;
	     */
            switch (lpDIB->depth)
            {
                case 1:
                    palette_ptr = bw_palette;
                    break;
                    
                case 4:
                    if (lpDIB->gray_or_color == 2)
                        palette_ptr = gray16_palette;
                    else 
                        palette_ptr = color16_palette;    
                    break;
                
                default:
                    palette_ptr = NULL; /* e.g., depth == 24 => no palettes */   
                    break;
            }
            if (palette_ptr != NULL)
                for (i=0; i<48; i++) 
                    pPCX->colmap[i] = *palette_ptr++;
	    else
		/* Initialize it (for later writing out this blank color
		 * map) since we have no map for this.
		 * This fixes a Purify Unitialized Memory Read.
		 * --ejaquith
		 */
                for (i=0; i<48; i++) 
                    pPCX->colmap[i] = 0;
            break;
    } /* End switch */
    return 0;
}

void pcx_write_open( GDATA *pcx_param, void *ft )
{
    
    IMXDLbegin_write( pcx_param, ft );
}


int write_pcx_line( void *fptr, GDATA *phdr, char *in_bufr )
{
    int byte_cnt = phdr->bytesPerLine * phdr->nplanes;
    
    
    HEAPCHECK();

    if( phdr->depth >= 4 )
    {
        switch( phdr->depth )
        {
            case 24:
            {
                pack24( in_bufr, phdr->bytesPerLine );
                HEAPCHECK();
                if( write_line( in_bufr, fptr, (unsigned int)byte_cnt ) != IMAGEIO_NOERROR )
                {
                    printf("Error writing 24 bit PCX file\n");
                    return( IMAGEIO_ERROR_FILE );
                } /* End if */
                HEAPCHECK();
                break;
            }
            case 8:
                if( write_line( in_bufr, fptr, (unsigned int)byte_cnt ) != IMAGEIO_NOERROR )
                {
                    printf("Error writing 256 color PCX file\n");
                    return( IMAGEIO_ERROR_FILE );
                } /* End if */
                break;

            case 4:
                if( write_line( in_bufr, fptr, (unsigned int)byte_cnt ) != IMAGEIO_NOERROR )
                {
                    printf("Error writing 16 color PCX file\n");
                    return( IMAGEIO_ERROR_FILE );
                } /* End if */
        } /* End switch */
        
    } else
    {
#if 1
        pcx_line( in_bufr, fptr, byte_cnt );
#else
        if( write_line( in_bufr, fptr, (unsigned int)byte_cnt ) != IMAGEIO_NOERROR )
        {
            printf("Error writing binary PCX file\n");
            return( IMAGEIO_ERROR_FILE );
        } /* End if */
#endif
    } /* End else */       
    return( IMAGEIO_NOERROR );
}


void pcx_close( void *ft )
{
        _lclose( ft );
}

void pcx_write_palette( void *ft, GDATA *phdr )
{
    unsigned char fstring[2] = { 0x0, 0x0 }; 
    unsigned char *pal_ptr; /*--ejaquith--added unsigned */
    
    if( phdr->gray_or_color == 2 )
    {  /* save as a grayscal image */
         if( phdr->depth == 8 )
             pal_ptr = &gray256_palette[0];
         else
             pal_ptr = &gray16_palette[0];    
    } 
    else
    {
        if ( phdr->depth == 8 )    
            pal_ptr = phdr->palette_data;
        else if ( phdr->depth == 4 )
            pal_ptr = &color16_palette[0];
        else
            pal_ptr = bw_palette;    
    } /* End else */        
            
    switch( phdr->depth )
    {
        case 1:
        case 4: /* for (version >= 5) use palette at EOF */
            fstring[0] = (unsigned char)12;
            _lwrite( ft, (LPSTR)&fstring[0], 1 );
            _lwrite( ft, (LPSTR)pal_ptr, 16 * 3 );
            break;

        case 8:
            fstring[0] = (unsigned char)12;
            _lwrite( ft, (LPSTR)&fstring[0], 1 );
            _lwrite( ft, (LPSTR)pal_ptr, 256 * 3 );
            break;
            
        case 24:
            break;
    } /* End switch */    

}

int write_line( char *p, void *fp, unsigned int n )
{
    unsigned int i = 0;
    unsigned int j = 0;
    unsigned int t = 0;
    unsigned char fchar; 
    unsigned char *tmp_buf;
    unsigned char *save_ptr;
    int byte_cnt = 0;
        
    tmp_buf = (unsigned char *)malloc( n * 2 ); /*--ejaquith--added unsigned*/
    save_ptr = tmp_buf;
    
    do 
    {
        i = 0;
#if 0
        while( ( p[t+i] == p[t+i+1] ) && ((t+i) < n) && (i<63))
            ++i;
#endif
	/* --ejaquith 7/95
	 * Read a byte off the end of the buffer with the above
	 * STUPID while.  I @$%$# hate this contractor's code.
	 * He can't even copy from a book correctly...
	 */
        while((t+i+1 < n) && (i < 63) && (p[t+i] == p[t+i+1]))
	    /* Increase the repeated byte count */
            ++i;

        if( i > 0 )
        {
            fchar = i | 0xc0;
            *tmp_buf++ = fchar;
            byte_cnt++;

            fchar = p[t];
            *tmp_buf++ = fchar;
            byte_cnt++;

            t+=i;
            j+=2;            
        }
	else
        {
            if( ( ( p[t] ) & 0xc0 ) == 0xc0 )
            {
                fchar = 0xc1;
                *tmp_buf++ = fchar;
                byte_cnt++;
                ++j;
            } 
            fchar = p[t++];
            *tmp_buf++ = fchar;
            byte_cnt++;   
            ++j;                
        } /* End else */
    } while( t < n );
 
    GFIO_Write( fp, (char *)save_ptr, byte_cnt );/*--ejaquith--added (char *)*/
    free( save_ptr );
    return( IMAGEIO_NOERROR );
}
                 

int OpenFile( UINT fuMode, void *fu )
{
    if( fuMode != 0 && fu != NULL )
        return( IMAGEIO_NOERROR );
    else
        return( IMAGEIO_NOERROR );    
}

UINT _lwrite( void *hf, char *hpvBuffer, UINT cbBuffer )
{
    int x;
    
    x = GFIO_Write( hf, hpvBuffer, cbBuffer );
    return( x );
}

int _lclose( HFILE hf )
{

    if( hf != NULL )
        return( IMAGEIO_NOERROR );
    else
        return( IMAGEIO_NOERROR );        
}

void clr_mem( char *start, int length )
{
    int x;
    
    for( x = 0; x < length; x++ )
        *start++ = 0x0;
}   


/**********************************************************/
/* The PCX read functions are given below:                */
/**********************************************************/

/* opens and reads in the header of a pcx file */
int pcx_read_open( GDATA *hptr, void *ft )
{
    pcxhead pcx_ptr;
    unsigned char *src_ptr;
    unsigned char *dest_ptr;
    int x;
            
    init_rd_buffer();
    if( OpenFile( OF_READ, ft ) == IMAGEIO_ERROR_FILE )
        return( IMAGEIO_ERROR_FILE );           
    if( ! ( GFIO_Read( ft, (char *)&pcx_ptr, sizeof( pcxhead ) ) ) 
        == sizeof( pcxhead) )
            return( IMAGEIO_ERROR_FILE );
/*--ejaquith--added call here! */
wp_stupidPCXconvertFromIntel16(&pcx_ptr);
    hptr->bits_pixel = pcx_ptr.bits;
    hptr->nplanes = pcx_ptr.nplanes;
    hptr->bytesPerLine = pcx_ptr.byteline;

    if (pcx_ptr.hrez > 0)
        hptr->xDPI  = pcx_ptr.hrez;
    if (pcx_ptr.vrez > 0)
        hptr->yDPI  = pcx_ptr.vrez;
    hptr->x_dim = pcx_ptr.win[2] - pcx_ptr.win[0] + 1;
    hptr->y_dim = pcx_ptr.win[3] - pcx_ptr.win[1] + 1;
    
    if (hptr->nplanes == 3 && hptr->bits_pixel == 1)
       hptr->depth = hptr->bits_pixel * 4;
    else
       hptr->depth = hptr->bits_pixel * hptr->nplanes;

    /* default to color */
    if (pcx_ptr.palinfo == 2)
      hptr->gray_or_color = 2;
    else 
      hptr->gray_or_color = 1; 
    
    /*
     * Read or create the desired palette.
     */
    if (hptr->depth == 1)
    {
        /* For now, ignore the reverse contrast issue and
         *   use the default B&W palette
         */
        hptr->gray_or_color = 2; 
            /* PROBLEM: fileio expects this value.  Will this
             *   confuse other clients?
             */
        src_ptr = bw_palette;
        dest_ptr = hptr->palette_data;
        for( x = 0; x < 48; x++ )
            *dest_ptr++ = *src_ptr++;    
    }
    else if (hptr->gray_or_color == 2)
    {
        /* grayscale */
        if (hptr->depth <= 4)
        {
            /* read the default 16 gray scale palette */
            src_ptr = gray16_palette;
            dest_ptr = hptr->palette_data;
            for( x = 0; x < 48; x++ )
                *dest_ptr++ = *src_ptr++;    
        }
        else
        {
#if 1
            Int32    fsize = GFIO_FileSize(ft);
        
        
            /* use the 256 "color" palette at end of file */
            fsize -= 768;
            GFIO_Seek( ft, fsize );
	    /*--ejaquith--added (char *)*/
            pcx_read_palette( (char *)hptr->palette_data, ft );
            GFIO_Seek(ft, 128 ); /* point to beginning of data */

#else
            /* read the default 256 gray scale palette */
            src_ptr = gray256_palette;
            dest_ptr = hptr->palette_data;
            for( x = 0; x < 768; x++ )
                *dest_ptr++ = *src_ptr++;    
#endif
        }
        
    }
    else if (pcx_ptr.ver == 3 || pcx_ptr.ver == 0)
    {
        /* No palette found; read the default 16 color palette. */
        src_ptr = color16_palette;
        dest_ptr = hptr->palette_data;
        for( x = 0; x < 48; x++ )
            *dest_ptr++ = *src_ptr++;    
    }
    else if (hptr->depth <= 4)
    {
        /* read the header palette */
        src_ptr = &pcx_ptr.colmap[0];
        dest_ptr = hptr->palette_data;
        for( x = 0; x < 48; x++ )
            *dest_ptr++ = *src_ptr++;    
    }            
    else 
    {
        Int32    fsize = GFIO_FileSize(ft);
        
        
        /* else use the VGA palette at end of file */
        fsize -= 768;
        GFIO_Seek( ft, fsize );
	/*--ejaquith--added (char *)*/
        pcx_read_palette( (char *)hptr->palette_data, ft );
        GFIO_Seek(ft, 128 ); /* point to beginning of data */
    }
    
    return( IMAGEIO_NOERROR );                       
}/* eo pcx_read_open */


int pcx_read_line( GDATA *hptr, char *buf_ptr, void *fp )
{
    int x;
    int byte_cnt;
    int gbits;
    unsigned char *b;    
    unsigned char *p;
    unsigned char *save_ptr;
    int n = 0;
    int c, i;               
    unsigned char *plane_ptr;
        
    gbits = hptr->depth;    
    byte_cnt = hptr->bytesPerLine * hptr->nplanes;
      /*PROBLEM: This cannot be computed above imageio; an interface must be added --EHoppe */
    b = (unsigned char *)pcx_line_ptr; /*--ejaquith--added (unsigned char *)*/
    p = (unsigned char *)buf_ptr;      /*--ejaquith--added (unsigned char *)*/
    clr_mem( (char *)p, byte_cnt );	/*--ejaquith--added (char *)*/
    if( hptr->depth > 4 )
    {
	/*--ejaquith--added (char *)*/
        if( ! read_line( (char *)p, fp, byte_cnt ) )
        {
            printf("Error - can't read pcx file\n");
            return( IMAGEIO_ERROR_FILE );
        }    
        if ( hptr->depth == 24 )
	    /*--ejaquith--added (char *)*/
            unpack24( (char *)p, hptr->x_dim );
    } else
    if( hptr->depth == 4 )
    {
	/*--ejaquith--added (unsigned char *)*/
        plane_ptr = (unsigned char *)malloc( byte_cnt * 8);
            /* PROBLEM: Looks like pcx_planes somehow uses one byte 
             *   per pixel during computation.  This shouldn't be necessary.
             */
        save_ptr = plane_ptr;
	/*--ejaquith--added (char *)*/
        if( ! read_line( (char *)b, fp, byte_cnt ) ) {
            printf("Error - can't read pcx file\n");
	}
        pcx_planes( plane_ptr, b, byte_cnt, hptr->nplanes, hptr->depth );
        /* convert PCX pixels to perfectscan packed pixels */
        if( hptr->nplanes == 1 )
        {    
            for( x = 0; x < byte_cnt; x++ )
            {
                p[x] = ( (*plane_ptr << 4) & 0xf0 );
                p[x] += ( *plane_ptr++ & 0x0f );
            } /* End for */
        } 
        else
        {
            for( x = 0; x < byte_cnt; x++ )
            {
                p[x] = (*plane_ptr++ << 4) & 0xf0;
                p[x] += *plane_ptr++ & 0x0f;
            } /* End for */
        } /* End else */ 
        free( save_ptr );
    } else
    {
        do 
        {
            c = get_buffered_char( fp );
            if( ( c & 0xc0 ) == 0xc0 )
            {
                i = c & 0x3f;
                c = get_buffered_char( fp );
                while( i-- )
                p[n++] = ~c; /*--ejaquith--Added byte not */
            } else
                p[n++] = ~c; /*--ejaquith--Added byte not */
        } while( n < byte_cnt );                                  
    }
    
    return( IMAGEIO_NOERROR );
} 

/* reads an individual PCX scan line */                                       
int read_line( char *p, void *fp, int max )
{
    int n = 0;
    int c, i;
    char tmp_buf[2];

    clr_mem( p, max );
    do 
    {
        tmp_buf[0] = get_buffered_char( fp );
        c = tmp_buf[0] & 0xff;
        if( ( c & 0xc0 ) == 0xc0 )
        {
            i = c & 0x3f;
            tmp_buf[0] = get_buffered_char( fp );
            c = tmp_buf[0];
            while( i-- )
            {
                p[n++] = c;
            }    
        } 
        else
            p[n++] = c;            
    } while( n < max );
    return( n );
}                                                   

/* reads the 8 bit palette at the end of a pcx file */
int pcx_read_palette( char *p_addr, void *fptr )
{
    if( GFIO_Read( fptr, p_addr, 768 ) != GFIO_EOF )
    {
        return( IMAGEIO_NOERROR );
    } else
    {
        return( IMAGEIO_ERROR_FILE );    
    }    
}

int OpenPCX( GDATA *phdr, void *fptr, int mode )
{
    struct State* state;
      
    HEAPCHECK();

    /* allocate a state structure */
    state = (struct State*)malloc(sizeof(struct State));
    if (state==NULL) 
    return( IMAGEIO_ERROR_FILE );
    phdr->internal_data = state;

    /* initialize the structure that is passed back to the callback proc */
    state->file = fptr;

    if ( mode == GFIO_READ )
    {    
        if ( fptr != NULL )
        {
            phdr->ftype = GFIO_READ;
            pcx_read_open( phdr, fptr );

#if 1
	    /* 10oct95--ejaquith
	     * I swear I can't look at this any more.
	     * Mark S. can't program in C if his life depended on it.
	     * We ACTUALLY paid this guy.
	     */
            pcx_line_ptr = (char *)calloc( phdr->bytesPerLine, phdr->nplanes);
#else
            pcx_line_ptr = (char *)malloc( phdr->bytesPerLine *
					   phdr->nplanes);
            for( x = 0; x < (phdr->bytesPerLine * phdr->nplanes); x++ )
                pcx_line_ptr[x] = 0;            
#endif
            return( IMAGEIO_NOERROR );
        } 
        else
        {
            GFIO_DestroyFileToken( fptr );              
            return( IMAGEIO_ERROR_FILE );
        } /* End else */
    } 
    else if ( mode == GFIO_WRITE )
    {
       if ( fptr != NULL )
       {
           phdr->ftype = GFIO_WRITE;
           pcx_write_open( phdr, fptr );
           HEAPCHECK();
           return( IMAGEIO_NOERROR );
       } 
       else
       {
           return( IMAGEIO_ERROR_FILE );
       } /* End else */                         
    } 
    else
        return( IMAGEIO_ERROR_FILE );
}/* eo OpenPCX */         
 
int WritePCX( GDATA *phdr, char *buffer )
{
    struct State *state;
    void *fptr;
    
    state = phdr->internal_data;
    fptr = state->file;
    return( write_pcx_line( fptr, phdr, buffer ) );
}


int ClosePCX( GDATA *phdr )
{
    struct State *state;
    void *fptr;
    
    state = phdr->internal_data;
    fptr = state->file;

    if( phdr->ftype == GFIO_WRITE && phdr->depth != 24 && phdr->depth != 1 )
        pcx_write_palette( fptr, phdr ); 
    else if( phdr->ftype == GFIO_READ )
    {
        free( pcx_line_ptr );
    } /* End else */           
    pcx_close( fptr );
    free( state );
    return( IMAGEIO_NOERROR );
}      
 
int ReadPCX( GDATA *phdr, char *gbuffer )
{
    struct State *state;
    void *fptr;

    state = phdr->internal_data;
    fptr = state->file;
    return( pcx_read_line( phdr, gbuffer, fptr ) );
}

/* converts pcx 24 bit RRRRRRRGGGGGGGBBBBBBB into RGB */
void unpack24( char *bptr, int blen )
{
    char *tptr, *sptr;
    int x;
    char *redptr, *greenptr, *blueptr;
 
    
    tptr = (char *)malloc( blen * 4 );
    sptr = tptr;
     
    redptr = bptr;
    greenptr = bptr + ( 1 * blen );
    blueptr = bptr + ( 2 * blen );
        
    /* copy colors */
    for( x = 0; x < blen; x++ )
    {
        *tptr++ = *redptr++;
        *tptr++ = *greenptr++;
        *tptr++ = *blueptr++;
    } /* End for */
    
    tptr = sptr;
    for( x = 0; x < blen *3; x++ )
        *bptr++ = *tptr++;
    free( sptr );
}

/* converts RGB to pcx 24 bit RRRRRRRGGGGGGGBBBBBBB */
void 
pack24(char *bptr, UInt32 nBytesPerLine)
{
    unsigned char *tptr, *sptr, *zptr;
    int i;
    unsigned char *redptr, *greenptr, *blueptr;
 
    /*--ejaquith--added (unsigned char *)*/
    tptr = (unsigned char *)malloc( nBytesPerLine * 3 );
    sptr = tptr;
    zptr = (unsigned char *)bptr;    /*--ejaquith--added (unsigned char *)*/
     
    redptr = tptr;
    greenptr = tptr + ( 1 * nBytesPerLine );
    blueptr = tptr + ( 2 * nBytesPerLine );
        
    /* copy colors */
    for( i = 0; i < nBytesPerLine; i++ )
    {
        *redptr++ = *bptr++;
        *greenptr++ = *bptr++;
        *blueptr++ = *bptr++;
    } /* End for */

    bptr = (char *)zptr;    /*--ejaquith--added (char *)*/
    tptr = sptr;
    for( i = 0; i < nBytesPerLine * 3; i++ )
        *bptr++ = *tptr++;

    free( sptr );
}/* eo pack24 */

/* convert 4 plane 1 bit/pixel to 1 plane 4 bit/pixel */
static void pcx_planes( unsigned char *pixels, unsigned char *bitplanes,
                        int bytesperline, int planes, int bitsperpixel )
{
    int i, j;
    int npixels;
    unsigned char *p;
    unsigned int pixbit;
    unsigned char bits;
    unsigned int mask;
    
    if( planes == 4 )
    {
        npixels = (bytesperline * (bitsperpixel * 2) );
        p = pixels;
        while( --npixels >= 0 )
            *p++ = 0;
    
        for( i = 0; i < planes; i++ )
        {
            p = pixels;
            pixbit = ( 1 << i ); /* plane mask (selector) */
            
            /* NOTE: 'bytesperline' is the bytes per line per plane times 4 */
            for( j = 0; j < bytesperline/4; j++ )
            {
                bits = *bitplanes++;
                /*   OR every bit into a a dst byte.  Later, the
                 *   bytes will be packed into a 4bit/1plane image.
                 */
                for( mask = 0x80; mask != 0; mask >>= 1, p++ )
                {
                    if( bits & mask )
                        *p |= pixbit;
                }        
            } /* End for */
        } /* End for */
    } else
    {
        for( i = 0; i < bytesperline; i++ )
            *pixels++ = *bitplanes++;        
    } /* End else */
}
    
            
    
/*
 * PROBLEM: Shouldn't have to negate here as a special case.
 *      Why is binary PCX NOT inverted when it leaves PerfectScan?
 *      (all images inverted in imageio prior to calling WritePCX).
 *      --EHoppe
 */    
int pcx_line( char *buf, void *fp, unsigned int length )
{
    char *end;    /* ejaquith -- dropped the unsigned here */
    int c;
    int previous;
    int count;
    unsigned char *tmp_buf;
    unsigned char *save_ptr;
    int byte_cnt;
    
    
    byte_cnt = 0;
    tmp_buf = &pcx_buffer[0];    
    save_ptr = tmp_buf;
    end = buf + length;
    previous = ~(*buf++);
    count = 1;
    
    while( buf < end )
    {
        c = ~(*buf++);
        if( c == previous && count < 63 )
        {
            ++count;
            continue;
        }
        
        if( count > 1 || ( previous & 0xc0 ) == 0xc0 )
        {
            count |= 0xc0;
            *tmp_buf++ = count;
            byte_cnt++;
        }
        
        *tmp_buf++ = previous;
        byte_cnt++;
        previous = c;
        count = 1;
    }
    
    if( count > 1 || ( previous & 0xc0 ) == 0xc0 )
    {
        count |= 0xc0;
        *tmp_buf++ = count;
        byte_cnt++;
    }
    
    *tmp_buf++ = previous; 
    byte_cnt++;
    /*--ejaquith--added (char *)*/
    GFIO_Write( fp, (char *)save_ptr, byte_cnt );
    return( TRUE );
}
