/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 * pcx2.c:  routines for reading PCX formatted files into PIXRs and
 *          for writing PIXRs into PCX files.
 *
 *  Note:  all routines are private (thus, pointers and PIXRs are not checked
 *  at this level)
 *
 */

#include <stdio.h>

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "gdefs.h"
#include "frames.pub"
#include "imageio.h"
#include "imagerep.h"
#include "ptypes.h"

/* prototypes */
#include "pcx.h"
#include "props.pub"
#include "io.pub"
#include "io.prv"
#include "gfio.h"
#include "gfiopro.h"

IP_RCSINFO(RCSInfo, "$RCSfile: pcx2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:18 $");


#define BITS_PER_WORD   32
#define BITS_PER_BYTE   8




    /* WARNING WARNING WARNING!!
     * These 2 functions, wp_imageioAllocGData() and wp_imageioFreeGData() 
     * have been copied from perfectscan's imageio.c and had "Copy" appended.
     * We did NOT want to import just for these functions (to include
     * that file would mean including the whole rest of the imageio stuff!)
     * So upon new version updates of PCX reading/writing, get the "fresh"
     * functions from imageio.c again and copy them here if they change!!!
     * And yes, eventually this will probably change when we get the cleaned
     * up versions of code in and have decided on who-will-maintain-what.
     *    -ejaquith 10/20/94
     */

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make imageio gdata stuff.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppGData::        Ptr to ptr to imageio stuff.
 *                   ppGData is allocated and filled in. 
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_imageioAllocGDataCopy (
    IPWState           *pIPWState,
    GDATA             **ppGData)
{
    PROLOGUE;
    GDATA   *pGData = NULL;
    

    pGData = (GDATA *)IPW_CALLOC(1, sizeof(GDATA), cAllocatorData);
    ERR_GEN(pGData == NULL, cErrNomem);

    pGData->palette_data = NULL;
    pGData->internal_data = NULL;
    pGData->palette_data = NULL;

    /* Typical Windows screen driver resolution --EHoppe */
    pGData->xDPI = 96;
    pGData->yDPI = 96;
    pGData->palette_data = (UInt8 *)IPW_CALLOC(768, sizeof(UInt8),
						   cAllocatorData);

    ERR_GEN_CLEANUP(pGData->palette_data == NULL, cErrNomem);

    *ppGData = pGData;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pGData != NULL)
    {
	IPW_FREE(pGData->palette_data, cAllocatorData);
	IPW_FREE(pGData, cAllocatorData);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free imageio gdata stuff and set ptr to NULL.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  ppGData::        Ptr to ptr to imageio stuff.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_imageioFreeGDataCopy (
    IPWState           *pIPWState,
    GDATA             **ppGData)
{
    PROLOGUE;

    if (*ppGData != NULL)
    {
	IPW_FREE( (*ppGData)->palette_data, cAllocatorData );
	IPW_FREE( *ppGData, cAllocatorData );
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Byte flip routine for broken/stupid PCX code
 *
 * Input:::
 *  ph::	Ptr to PCX head
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
void CDECL
wp_stupidPCXconvertFromIntel16 (
    pcxhead *ph)
{
    UInt16     tmp;
    ErrCode    __error__; /* Gack, what a hack.  I am at a lack */
    IPWState  *pIPWState = NULL;
    

    /*  Note how we pass in NULL instead of std pIPWState -- this is a total
     *  hack! because we know getIntel16 doesn't use pIPWState (as of now)
     *  and we haven't got pIPWState propogating down through this PCX code.
     *  Temporary solution most likely -- one day I'll have to rewrite all
     *  this STUPID code because this code ...  well, you know how I feel.
     */

    /* NOTE!  These purposely do not have the standard error-handling macros
     * wrapping them because we have no IPWState.
     */

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->win[0], &tmp)) 
	return;
    ph->win[0]   = tmp;

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->win[1], &tmp))
        return;
    ph->win[1]   = tmp;

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->win[2], &tmp))
        return;
    ph->win[2]   = tmp;

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->win[3], &tmp))
        return;
    ph->win[3]   = tmp;

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->vrez, &tmp))
        return;
    ph->vrez     = tmp;

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->hrez, &tmp))
        return;
    ph->hrez     = tmp;

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->byteline, &tmp))
        return;
    ph->byteline = tmp;

    ERR_CHECK(w_getIntel16(NULL, (UInt8 *)&ph->palinfo, &tmp))
        return;
    ph->palinfo  = tmp;
}



#if 0  /* See comment at its invocation below */
/*
 * ColorMapPcxPr ()
 *
 *  	Translate the full 8/4 bit x 3 colormap to PIXR colormap
 *
 * Arguments:
 *  	hptr	-   ptr to generic graphics bitmap header
 *  	coltype	-   type of colormap (gray or color)
 *
 * Returns:
 *  	colmap 	-   PIXR colormap, or NULL on error
 */
static ColorMap CDECL *
wp_colorMapPcxPr(GDATA *hptr, ColorType coltype)
{
    ColorMap *colmap;
    ColorRecord *map_index;
/*
    unsigned char *palet_index;
*/
    UInt8 *palet_index;
    Int16 count;

    if ( hptr->depth != 8 && hptr->depth != 4)
    {
	/* We only have PCX colormaps for 4 and 8 bpp images */
	return (ColorMap *)NULL;
    }
    
    if ((colmap = (ColorMap *)IPW_MALLOC(sizeof(ColorMap))) == NULL)
    {
	return abortP("malloc fail for colormap", procName, NULL);
    }
	
    colmap->count = ( hptr->depth == 8 ? 256 : 16 );
    colmap->type = coltype;

    if ((colmap->map = (ColorRecord *)IPW_CALLOC(colmap->count,
					     sizeof(ColorRecord))) == NULL)
    {
	return abortP("malloc fail for colors", procName, NULL);
    }
/*    
    map_index = &(colmap->map[0]);
*/
    map_index = colmap->map;
/*
    palet_index = &(hptr->palette_data[0]);
*/
    palet_index = hptr->palette_data;
    for (count = 0; count < colmap->count; map_index++, count++) 
    {
	(*map_index)[0] = *(palet_index++); /*R*/
	(*map_index)[1] = *(palet_index++); /*G*/
	(*map_index)[2] = *(palet_index++); /*B*/
    }

    return(colmap);
}
#endif /* 0 */



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a PCX file into a PIXR.
 *
 * LongDescription:::
 *     Supported format is single plane, depth 1, 4, or 8.
 *     Gray-level colormaps may be inverted when read in.
 *     For now, only binary images have been tested at all, gray-level
 *     is up to you for testing.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream.
 *
 * Output:::
 *  pixr::           Ptr to ptr to returned PIXR with image that was read.
 *                   pixr is allocated and filled in.
 *                   *pixr must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_readPcx (
    IPWState           *pIPWState,
    FILE               *fp,
    PIXR              **pixr)
{
    PROLOGUE;
    UInt8   	*p;
    Int32    	 iHeight, bpl;
    Int16    	 row;
    UInt16   	 bytesperrow;
    char    	 tempfilename[] = "tempinfile\0"; 
    GDATA   	*hptr = NULL;
    void    	*ft = NULL;
    PIXR        *pr = NULL;

    ERR_GEN(pixr == NULL  ||  fp == NULL, cErrNullPtr);
    ERR_GEN(*pixr != NULL, cErrPtrNotNull);

    ERR_PASS( wp_imageioAllocGDataCopy(pIPWState, &hptr));
 
    ft = GFIO_CreateFileToken(tempfilename, fp);
    ERR_GEN_CLEANUP(ft == NULL, cErrIO_ReadError);

    /* make sure file is reset to start */
    ERR_GEN_CLEANUP(GFIO_Seek( ft, 0 ) != 0, cErrIO_SeekFailure);

    hptr->format = IMAGEIO_PCX;
    ERR_GEN_CLEANUP(OpenPCX( hptr, ft, GFIO_READ ) != IMAGEIO_NOERROR,
		    cErrIO_ReadError);

    ERR_GEN_CLEANUP(hptr->depth != 1 && hptr->depth != 4 && hptr->depth != 8,
		    cErrDepthNotSupported);

    iHeight = hptr->y_dim;

    /* Note the nplanes field will always be 1 for depth 1 and depth 8
     * images, and always be 4 for depth 4 images (these are 4 planes of 1 bit
     * per pixel).
     */
    bytesperrow = (UInt16)(hptr->bytesPerLine * hptr->nplanes);

    ERR_PASS_CLEANUP (w_createPixrNoInit (pIPWState, hptr->x_dim, iHeight,
					  hptr->depth,
					  Int32_TO_ScaledInt1616 (hptr->xDPI), 
					  Int32_TO_ScaledInt1616 (hptr->yDPI),
					  cFrameBits, &pr ));

#if 0
    /* Not used presently because things like destroyPixr() don't
     * destroy the colormap.  Thus, if we set it here now, it never
     * gets freed and is a memory leak.  (In laymen's terms: a liability).
     * Fine then, we won't set it and now noone can use it.  hmpf!
     */
    if (hptr->depth == 4 || hptr->depth == 8)
    {
	/* If the image is gray-level, do the best we can at trying 
	 * to get the actual colormap
	 */
	ColorMap *colmap;
	ColorType coltype;
	
	coltype = ( hptr->gray_or_color == 2 ? cColorGray : cColorRGB );
	colmap = wp_colorMapPcxPr( hptr, coltype );
	
	pixrSetType( pr, coltype );
	pixrSetColorMap( pr, colmap );
    }
#endif /* 0 */

    /* Bleah!  The underlying PCX reading is done ONLY in rows, that is ONE
     * row at a time.  Most likely that is not ideal, but we must conform.
     */
    ERR_PASS_CLEANUP( w_pixrGetImage(pIPWState,  pr , &p));
    ERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState,  pr , &bpl));
    for (row = 0; row < iHeight; row++, p += bpl) 
    {
	ERR_GEN_CLEANUP(ReadPCX( hptr, (char *)p ) != IMAGEIO_NOERROR,
			cErrIO_ReadError);
    }

    ERR_GEN_CLEANUP(ClosePCX( hptr ) != IMAGEIO_NOERROR,
		    cErrIO_ReadError);

    GFIO_DestroyFileToken( ft ); ft = NULL;
    ERR_PASS_CLEANUP( wp_imageioFreeGDataCopy( pIPWState, &hptr ));

    ERR_PASS_CLEANUP( w_flip(pIPWState, pr));

    *pixr = pr;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (ft != NULL)
	GFIO_DestroyFileToken( ft );
    wp_imageioFreeGDataCopy( pIPWState, &hptr );
    w_destroyPixr(pIPWState, &pr);

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a PIXR to file stream in PCX format.
 *
 * LongDescription:::
 *     Supported format is single plane.
 *     Gray-level images have a "standard" colormap written out,
 *     unfortunately ignoring any PIXR colormap information.
 *     For now, only binary images have been tested at all, gray-level
 *     is up to you for testing.
 *
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  fp::             Ptr to file stream to write to.
 *  pixr::           Ptr to PIXR, single channel, depth 1, 4, or 8.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  file io
 *
 * Keywords:::
 *  file io, compression
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_fwritePcx (
    IPWState           *pIPWState,
    FILE               *fp,
    PIXR               *pixr)
{
    PROLOGUE;
    UInt8        *pImage;
    Int32         iDepth, iWidth, iHeight, bpl, channels;
    Int16         row;
    ScaledInt1616 xdpi, ydpi;
    UInt16        bytesperrow;
    ColorType     type;
    char          tempfilename[] = "tempoutfile\0";
    GDATA        *hptr = NULL;
    UInt8        *buf = NULL;
    void         *ft = NULL;

    ERR_GEN(pixr == NULL, cErrInvalidParm);

    ft = GFIO_CreateFileToken(tempfilename, fp);
    ERR_GEN(ft == NULL, cErrIO_WriteError);

    ERR_PASS_CLEANUP( wp_imageioAllocGDataCopy(pIPWState, &hptr));

    /* Set up the structure we will use to write the PCX header out with */
    ERR_PASS_CLEANUP( w_pixrGetWidth(pIPWState,  pixr , &iWidth));
    ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState,  pixr , &iDepth));
    ERR_PASS_CLEANUP( w_pixrGetHeight(pIPWState,  pixr , &iHeight));
    ERR_PASS_CLEANUP( w_pixrGetChannelCount(pIPWState,  pixr , &channels));
    ERR_PASS_CLEANUP( w_pixrGetType(pIPWState, pixr, &type));
    ERR_PASS_CLEANUP( w_pixrGetResolutionDPI(pIPWState, pixr, &xdpi, &ydpi));

    hptr->xDPI        = ScaledInt1616_TO_Int32(xdpi);
    hptr->yDPI        = ScaledInt1616_TO_Int32(ydpi);
    hptr->x_dim       = iWidth;
    hptr->y_dim       = iHeight;

    ERR_GEN_CLEANUP(channels != 1, cErrPlanesNotSupported);

    /* grayscale or color: gray = 2, color = 1 */
    hptr->gray_or_color = 2;

    switch (iDepth)
    {
    case 1:
	hptr->depth         = 1;
	hptr->bits_pixel    = 1;
	hptr->nplanes       = 1;
	break;
    case 4:
	hptr->depth         = 4;
	hptr->bits_pixel    = 1;
	hptr->nplanes       = 4;
	break;
    case 8:
	hptr->depth         = 8;
	hptr->bits_pixel    = 8;
	hptr->nplanes       = 1;
	break;
    default:
	ERR_GEN_CLEANUP(1, cErrDepthNotSupported);
    }
    

    /* This field gets filled in in OpenPCX():
     *   hptr->bytesPerLine
     */
    ERR_GEN_CLEANUP(OpenPCX( hptr, ft, GFIO_WRITE ) != IMAGEIO_NOERROR,
		    cErrIO_WriteError);

    /* Each call to WritePCX() will write out this many bytes:
     *   hptr->bytesPerLine * hptr->nplanes
     */
    bytesperrow = (UInt16)(hptr->bytesPerLine * hptr->nplanes);

    ERR_PASS( w_flip(pIPWState, pixr));

    /* Bleah!  The underlying PCX writing is done ONLY in rows, that is ONE
     * row at a time.  Most likely that is not ideal, but we must conform.
     */
    ERR_PASS( w_pixrGetImage(pIPWState,  pixr , &pImage));
    ERR_PASS( w_pixrGetBpl(pIPWState,  pixr , &bpl));
    

    if (iDepth == 8  ||  iDepth == 1)
    {
	for (row = 0; row < iHeight; row++, pImage += bpl) 
	{
	    /* Since our PIXRs always have padding, we can get away with
	     * writing from the image ptr directly without fear we will
	     * read too far over when adding any extra padding at the end
	     * of the line.
	     */
	    ERR_GEN_CLEANUP(WritePCX( hptr, (char *)pImage )
			    != IMAGEIO_NOERROR,
			    cErrIO_WriteError);
	}
    }
    else
    {
	UInt8   *scanlineOfPlane1, *scanlineOfPlane2, *scanlineOfPlane3;
	UInt32  *pWord;
	Int32    byteNum;
	
	buf = (UInt8 *)IPW_CALLOC( bytesperrow, sizeof(UInt8), cAllocatorData);
	ERR_GEN_CLEANUP(buf == NULL, cErrNomem);

	/* For 4-bit (16 color) images, the buffer is to be set up as follows:
	 * the 0th (binary) plane, scan line 0, hptr->bytesPerLine #bytes.
	 * the 1st (binary) plane, scan line 0, hptr->bytesPerLine #bytes.
	 * the 2nd (binary) plane, scan line 0, hptr->bytesPerLine #bytes.
	 * the 3rd (binary) plane, scan line 0, hptr->bytesPerLine #bytes.
	 * the 0th (binary) plane, scan line 1, hptr->bytesPerLine #bytes.
	 * ...
	 * Since our PIXRs use a single plane of depth 4, in a sense we
	 * have to take our "interlaced" data and split it up into its
	 * "components", into 4 binary planes.
	 */
	scanlineOfPlane1 = buf + hptr->bytesPerLine;
	scanlineOfPlane2 = scanlineOfPlane1 + hptr->bytesPerLine;
	scanlineOfPlane3 = scanlineOfPlane2 + hptr->bytesPerLine;

	/* Thus, now our buffer has four pointers in it, so it is divided
	 * up as spec'd above, into four scanlines of binary planes.
	 */

	for (row = 0; row < iHeight; row++, pImage += bpl) 
	{
	    for (byteNum = 0, pWord = (UInt32 *)pImage;
		 byteNum < hptr->bytesPerLine; 
		 pWord++, byteNum++)
	    {
#if _ALPACA_IMAGE_FMT_ == cBigEndianFmt
 		/* Construct the 3rd plane from the MSB/leftmost bit of the
		 * first/next 8 pixels (which are stored in the first/next
		 * 4 bytes), that is from bits 7 and 3 from each byte.
		 */
		scanlineOfPlane3[byteNum] =
		    ((pWord[0] & 0x80000000) >> 24) |
		    ((pWord[0] & 0x08000000) >> 21) |
		    ((pWord[0] & 0x00800000) >> 18) |
		    ((pWord[0] & 0x00080000) >> 15) |
		    ((pWord[0] & 0x00008000) >> 12) |
		    ((pWord[0] & 0x00000800) >>  9) |
		    ((pWord[0] & 0x00000080) >>  6) |
		    ((pWord[0] & 0x00000008) >>  3);

		/* Construct the 2nd plane from the bit one to the right of
		 * the MSB/leftmost bit of the
		 * first/next 8 pixels (which are stored in the first/next
		 * 4 bytes), that is from bits 6 and 2 from each byte.
		 */
		scanlineOfPlane2[byteNum] =
		    ((pWord[0] & 0x40000000) >> 23) |
		    ((pWord[0] & 0x04000000) >> 20) |
		    ((pWord[0] & 0x00400000) >> 17) |
		    ((pWord[0] & 0x00040000) >> 14) |
		    ((pWord[0] & 0x00004000) >> 11) |
		    ((pWord[0] & 0x00000400) >>  8) |
		    ((pWord[0] & 0x00000040) >>  5) |
		    ((pWord[0] & 0x00000004) >>  2);

		/* Construct the other two planes in a similar manner */
		scanlineOfPlane1[byteNum] =
		    ((pWord[0] & 0x20000000) >> 22) |
		    ((pWord[0] & 0x02000000) >> 19) |
		    ((pWord[0] & 0x00200000) >> 16) |
		    ((pWord[0] & 0x00020000) >> 13) |
		    ((pWord[0] & 0x00002000) >> 10) |
		    ((pWord[0] & 0x00000200) >>  7) |
		    ((pWord[0] & 0x00000020) >>  4) |
		    ((pWord[0] & 0x00000002) >>  1);

		buf[byteNum] =
		    ((pWord[0] & 0x10000000) >> 21) |
		    ((pWord[0] & 0x01000000) >> 18) |
		    ((pWord[0] & 0x00100000) >> 15) |
		    ((pWord[0] & 0x00010000) >> 12) |
		    ((pWord[0] & 0x00001000) >>  9) |
		    ((pWord[0] & 0x00000100) >>  6) |
		    ((pWord[0] & 0x00000010) >>  3) |
		    ((pWord[0] & 0x00000001)      );
#else  /* _ALPACA_IMAGE_FMT_ */
		/* This is just like above, except that the bytes
		 * need to be byteswapped.  We could do so with a call
		 * to w_flip() before accessing image data, and then
		 * another call after we are all done.  But hey, we
		 * can just rearrange constants in a trivial way,
		 * so why not do it like so...?!
		 * (I mean, we ARE writing a PCX file, a PC based format,
		 * so most likely we want it to be as fast (if not faster)
		 * on a PC).
		 */

		/* The bit shifts were figured by following the comments
		 * on each line, where the numbers correspond to bit
		 * indices in the longword, bit 31 == MSB == leftmost bit
		 * and bit 0 == LSB == rightmost bit.
		 * Thus, to move longword bit 3 to bit 6, we left shift
		 * by 3.
		 */
                scanlineOfPlane3[byteNum] = (UInt8)
                    (((pWord[0] & 0x00000080)      )  /*  7==> 7 */  |
                     ((pWord[0] & 0x00000008) <<  3)  /*  3==> 6 */  |
                     ((pWord[0] & 0x00008000) >> 10)  /* 15==> 5 */  |
                     ((pWord[0] & 0x00000800) >>  7)  /* 11==> 4 */  |
                     ((pWord[0] & 0x00800000) >> 20)  /* 23==> 3 */  |
                     ((pWord[0] & 0x00080000) >> 17)  /* 19==> 2 */  |
                     ((pWord[0] & 0x80000000) >> 30)  /* 31==> 1 */  |
                     ((pWord[0] & 0x08000000) >> 27)) /* 27==> 0 */  ;
                  
                scanlineOfPlane2[byteNum] = (UInt8)              
                    (((pWord[0] & 0x00000040) <<  1)  /*  6==> 7 */  |
                     ((pWord[0] & 0x00000004) <<  4)  /*  2==> 6 */  |
                     ((pWord[0] & 0x00004000) >>  9)  /* 14==> 5 */  |
                     ((pWord[0] & 0x00000400) >>  6)  /* 10==> 4 */  |
                     ((pWord[0] & 0x00400000) >> 19)  /* 22==> 3 */  |
                     ((pWord[0] & 0x00040000) >> 16)  /* 18==> 2 */  |
                     ((pWord[0] & 0x40000000) >> 29)  /* 30==> 1 */  |
                     ((pWord[0] & 0x04000000) >> 26)) /* 26==> 0 */  ;
                  
                scanlineOfPlane1[byteNum] = (UInt8)              
                    (((pWord[0] & 0x00000020) <<  2)  /*  5==> 7 */  |
                     ((pWord[0] & 0x00000002) <<  5)  /*  1==> 6 */  |
                     ((pWord[0] & 0x00002000) >>  8)  /* 13==> 5 */  |
                     ((pWord[0] & 0x00000200) >>  5)  /*  9==> 4 */  |
                     ((pWord[0] & 0x00200000) >> 18)  /* 21==> 3 */  |
                     ((pWord[0] & 0x00020000) >> 15)  /* 17==> 2 */  |
                     ((pWord[0] & 0x20000000) >> 28)  /* 29==> 1 */  |
                     ((pWord[0] & 0x02000000) >> 25)) /* 25==> 0 */  ;
                  
                buf[byteNum] = (UInt8)  
                    (((pWord[0] & 0x00000010) <<  3)  /*  4==> 7 */  |
                     ((pWord[0] & 0x00000001) <<  6)  /*  0==> 6 */  |
                     ((pWord[0] & 0x00001000) >>  7)  /* 12==> 5 */  |
                     ((pWord[0] & 0x00000100) >>  4)  /*  8==> 4 */  |
                     ((pWord[0] & 0x00100000) >> 17)  /* 20==> 3 */  |
                     ((pWord[0] & 0x00010000) >> 14)  /* 16==> 2 */  |
                     ((pWord[0] & 0x10000000) >> 27)  /* 28==> 1 */  |
                     ((pWord[0] & 0x01000000) >> 24)) /* 24==> 0 */  ;

#endif  /* _ALPACA_IMAGE_FMT_ */

	    }
	    
	    ERR_GEN_CLEANUP(WritePCX( hptr, (char *)buf ) != IMAGEIO_NOERROR,
			    cErrIO_WriteError);
	}

	IPW_FREE(buf, cAllocatorData);
    }
    
    ERR_PASS_CLEANUP( w_flip(pIPWState, pixr));
    

    ERR_GEN_CLEANUP(ClosePCX( hptr ) != IMAGEIO_NOERROR,
		    cErrIO_WriteError);

    GFIO_DestroyFileToken( ft ); ft = NULL;
    ERR_PASS_CLEANUP( wp_imageioFreeGDataCopy( pIPWState, &hptr ));
    NORMAL_RETURN;

ERR_CLEANUP:
    if (ft != NULL)
	GFIO_DestroyFileToken( ft );
    wp_imageioFreeGDataCopy( pIPWState, &hptr );
    IPW_FREE(buf, cAllocatorData);
    ERR_RETURN;
}


/*

$Log: pcx2.c,v $
Revision 1.1.1.1  1999/09/20 18:19:18  jlin
Imported sources

Revision 1.32  1997/06/30 23:10:55  ddavies
pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v

Delete inclusion of Sandpiper's internal.h.

Revision 1.31  1997/04/15 22:18:48  ddavies
pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v

Use non-initialized PIXR creation where possible.

Revision 1.30  1996/09/26 22:16:34  dferrell
dferrell on Thu Sep 26 15:15:42 PDT 1996

pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v

Fixed bug in PCX file reading caused by missing conversion from Int32
to ScaledInt1616 when creating the new PIXR.

Revision 1.29  1996/01/04 20:25:29  ejaquith
pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v

CommonCheck: Removed all HTML tags that we had placed in our C code.
Removed double-quoted routine name after "BeginDocumentation" and
"EndDocumentation".

 * Revision 1.28  1995/12/13  04:11:19  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.27  1995/11/13  20:06:51  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Changed wp_CONVERTFROMINTEL16() to wp_stupidPCXconvertFromIntel16().
 * Updated usage of PROLOGUE macro.
 *
 * Revision 1.26  1995/11/10  02:03:58  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Fixed Documentation-related warnings.
 *
 * Revision 1.25  1995/10/30  18:04:25  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Bug fixes!
 * PCX writing of depth 1 and depth 4 and depth 8 images were incorrect
 * when viewed ON the PC with a viewer in that they were byteswapped.
 * Now call w_flip() to byteswap before (and after) writing out, and
 * after reading in.
 * Another problem was depth 8 images weren't writing out any colormap
 * so the net was that the image appeared all black on a PC viewer, even
 * though the bitmap was itself correct.
 * A third little annoyance was that the depth 4 and 8 images were
 * being considered color images, so had weird color mappings.  Now, we
 * write them out as gray images, so they look swell and dandy.
 * Long live PCX depth 1 4 and 8.
 *
 * Revision 1.24  1995/10/18  04:18:37  ddavies
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Add casts to make Visual C++ happy.
 *
 * Revision 1.23  1995/10/16  22:57:57  dferrell
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Fixed type conversion warnings from Microsoft Visual C++ compiler.
 *
 * Revision 1.22  1995/10/13  22:03:46  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.21  1995/10/13  02:10:09  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Yanked no-longer-needed wp_wordPaddedBytesPerRow() routine.
 * Sped up reading and writing of depth 1 and depth 8 files by removing
 * copying into an intermediate data buffer before passing to low level
 * PCX routines.
 * (Probably about 2x gain).
 * Fixed depth 4 writing which was attempting to write out the data in
 * a single plane, which is illegal.  Made it split the single plane into
 * four, one line at a time.  I decided to be oh-so-nice and didn't use
 * the w_flip() of the PIXR before and after, and just wrote separate
 * plane unpacking for the PC (I mean, this IS a PC-based format -- so it
 * should be faster on the PC than the SUN if possible).
 * Fixed all sorts of bits per pixel, planes, and depth funniness so all
 * the code gets along (and better yet, does the right thing).
 * Removed calculation of bytes per file line for writing, because the
 * low level routines calculate it for us.
 * Added tons of comments to sort out all the confusion and idiosyncrasies
 * of using the PCX code from imageio.
 *
 * Revision 1.20  1995/10/10  02:32:36  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Added inclusion of "stdio.h"
 * Bug Fixes.
 * PCX format doesn't allow for depth 2 files.
 *  Added file depth and channel checking to only allow valid PCX files to
 * be read/written.
 * We only allow for single plane images.
 * Note there is still a bug in either reading/writing PCX depth 4 files
 * on both the SUN and PC, and still a bug in reading/writing PCX depth 1
 * files on the PC...
 *
 * Revision 1.19  1995/09/29  00:35:45  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.18  1995/09/21  22:55:16  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/20  21:31:23  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * made wp_imageioAllocGDataCopy() use calloc calls instead of malloc calls
 * to set it up (purify (in nodebug only) whined about unitited mem read).
 *
 * Revision 1.16  1995/09/19  21:10:04  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Fixed transfer of DPI from PIXR struct to PCX struct by using
 * ScaledInt1616_TO_Int32() macros.
 *
 * Revision 1.15  1995/09/14  18:56:45  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 *  Fixed types to be used with w_pixrGetResolutionDPI().
 *
 * Revision 1.14  1995/09/12  16:55:15  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.13  1995/09/08  00:15:21  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Updated IPWrap routines to have "w_" prefix, and updated other IPWrap
 * routine invocations.  Fixed any parameters whose order have changed with
 * this latest round of updates.
 *
 * Revision 1.12  1995/09/07  01:33:58  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Ran script to add tag "_POCHANGE" in front of each routine's declaration
 * and any invocations to alert coders there has been a parameter order
 * change in the prototype files.
 * Script also did name changes for routines that were documented when the
 * names changed in the prototype files.  (Typically dropping the "Pixr"
 * or "Pr" suffix in a routine name).
 *
 * Revision 1.11  1995/09/06  20:22:03  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.10  1995/09/01  02:20:58  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.9  1995/08/31  21:15:49  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.8  1995/08/25  01:37:42  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.7  1995/08/18  01:54:07  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Moved routine CONVERTFROMINTEL16() from pcx.c to this file.
 *
 * Revision 1.6  1995/08/17  22:01:46  ejaquith
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.5  1995/07/26  23:59:49  ejaquith
 * ejaquith on Wed Jul 26 16:58:13 PDT 1995
 *
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Plugged some memory leaks.
 * Freed buffer for line writing.
 * I was allocating and filling a PIXR colormap from the PCX file
 * previously.  Since destroyPixr() et al don't free them, we got leak.s
 * Solution: I #if 0ed the function and its invocation out.
 *
 * Revision 1.4  1995/03/01  22:16:34  ejaquith
 * ejaquith on Wed Mar 1 14:14:04 PST 1995
 *
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Fixed memory leak where memory allocated for buffer in
 * readPcxPr() was not freed.
 *
 * Revision 1.3  1994/12/08  03:08:54  dferrell
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Added typecastes so that MPW could compile it.
 *
 * Revision 1.2  1994/10/21  16:58:10  ejaquith
 * ejaquith on Fri Oct 21 09:55:53 PDT 1994
 *
 * pcx2.c -> /product/ipcore/ipwrap/src/RCS/pcx2.c,v
 *
 * Added casts to get rid of acc warnings that gcc missed.
 * NOTE!  Also copied new versions of pcx.c and epsutil.c into /product
 * to get rid of acc warnings
 *
 * Revision 1.1  1994/10/21  01:20:47  ejaquith
 * Initial revision
 *

*/

