/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  props2.c:
 *
 *      Includes extraction ops for PIXR
 *		w_pixrGetBpl()
 *		w_pixrGetChannel ()
 *		w_pixrGetChannelCount ()
 *		w_pixrGetColorMap ()
 *		w_pixrGetDepth()
 *		w_pixrGetFrame()
 *		w_pixrGetHeight()
 *		w_pixrGetPrimaryChannel ()
 *		w_pixrGetResolutionDPI()
 *		w_pixrGetType ()
 *		w_pixrGetWidth()
 *
 *		w_pixrSetBpl()
 *		w_pixrSetChannel ()
 *		w_pixrSetChannelCount ()
 *		w_pixrSetColorMap ()
 *		w_pixrSetDepth()
 *		w_pixrSetFrame()
 *		w_pixrSetHeight()
 *		w_pixrSetResolutionDPI()
 *		w_pixrSetType ()
 *		w_pixrSetWidth()
 *
 *	 Ptr to image
 *		w_pixrGetImage()
 *		w_pixrGetImageAll()
 *		wp_pixrGetImageBuffer()
 *		w_getLine()	(moved here from readfile.c,
 *				 ought to be called w_pixrGetLine())
 *
 *		w_pixrSetImageAll()
 *		wp_pixrSetImageBuffer()
 *
 *	 _Calculate_ information about the image
 *		w_pixrBpl()
 *		w_pixrFramedWidth()
 *		w_pixrFramedHeight()
 *		w_pixrImageBytes()
 *		w_pixrImageProperties()
 *		w_pixrImageOffset()
 *
 *
 */




#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"

/* prototypes */
#include "props.pub"
#include "props.prv"
#include "shrpixr.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: props2.c,v $; $Revision: 1.2 $; $Date: 2001/09/28 23:38:22 $");

/*--------------------------------------------------------------------*
 *                        Extraction and Setting                      *
 *--------------------------------------------------------------------*/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the width (excluding frame and padding) of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pWidth::         Ptr to width of PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetWidth (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pWidth)
{
    PROLOGUE;
    Int32     channels, w;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pWidth == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        w = pixr->pixrData.plane.w;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetWidth(pIPWState, ch, &w));
    }

    *pWidth = w;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the width (excluding frame and padding) of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  width::          Width value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetWidth (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               width)
{
    PROLOGUE;
    Int32 i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_GEN(width < 0, cErrInvalidParm);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	pixr->pixrData.plane.w = width;
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetWidth(pIPWState, ch, width));
	}
    }
	
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the height (excluding frame and padding) of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pHeight::        Ptr to PIXR height
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetHeight (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pHeight)
{
    PROLOGUE;
    Int32     channels, h;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pHeight == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        h = pixr->pixrData.plane.h;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetHeight(pIPWState, ch, &h));
    }

    *pHeight = h;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the width (excluding frame and padding) of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  height::         height value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetHeight (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               height)
{
    PROLOGUE;
    Int32     i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_GEN(height < 0, cErrInvalidParm);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	pixr->pixrData.plane.h = height;
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetHeight(pIPWState, ch, height));
	}
    }
	
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the depth of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pDepth::         Ptr to PIXR depth
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetDepth (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pDepth)
{
    PROLOGUE;
    Int32     channels, d;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pDepth == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        d = pixr->pixrData.plane.d;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetDepth(pIPWState, ch, &d));
    }

    *pDepth = d;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the depth of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  depth::          depth value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetDepth (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               depth)
{
    PROLOGUE;
    Int32        i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_GEN(depth < 1, cErrInvalidParm);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	pixr->pixrData.plane.d = depth;
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetDepth(pIPWState, ch, depth));
	}
    }
	
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the frame of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pFrame::         Ptr to PIXR frame
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetFrame (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pFrame)
{
    PROLOGUE;
    Int32    channels, f;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pFrame == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        f = pixr->pixrData.plane.frame;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetFrame(pIPWState, ch, &f));
    }

    *pFrame = f;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the frame of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  frame::          frame value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetFrame (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               frame)
{
    PROLOGUE;
    Int32 i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_GEN(frame < 0, cErrInvalidParm);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	pixr->pixrData.plane.frame = frame;
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetFrame(pIPWState, ch, frame));
	}
    }
	
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the number of bytes per line for the PIXR's image(s)
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pBpl::           Ptr to PIXR's bytes per line
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetBpl (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pBpl)
{
    PROLOGUE;
    Int32        channels, bpl;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pBpl == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        bpl = pixr->pixrData.plane.bpl;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetBpl(pIPWState, ch, &bpl));
    }

    *pBpl = bpl;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the bytes per line of the PIXR's image(s) 
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  bpl::            bytes per line value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetBpl (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               bpl)
{
    PROLOGUE;
    Int32 i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	pixr->pixrData.plane.bpl = bpl;
    else
    {
	PIXR  *ch;

	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetBpl(pIPWState, ch, bpl));
	}
    }
	
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the X and Y d.p.i. resolution of the PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pXRes::          Ptr to PIXR's X dpi
 *  pYRes::          Ptr to PIXR's Y dpi
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetResolutionDPI (
    IPWState           *pIPWState,
    PIXR               *pixr,
    ScaledInt1616      *pXRes,
    ScaledInt1616      *pYRes)
{
    PROLOGUE;
    Int32  channels;
    ScaledInt1616  xRes, yRes;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pXRes == NULL  ||  pYRes == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
    {
	xRes = pixr->pixrData.plane.xRes;
	yRes = pixr->pixrData.plane.yRes;
    }
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetResolutionDPI(pIPWState, ch, &xRes, &yRes));
    }

    *pXRes = xRes;
    *pYRes = yRes;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the X and Y d.p.i. resolution of the PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  xRes::           X dpi value
 *  yRes::           Y dpi value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetResolutionDPI (
    IPWState           *pIPWState,
    PIXR               *pixr,
    ScaledInt1616       xRes,
    ScaledInt1616       yRes)
{
    PROLOGUE;
    Int32 i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
    {
	pixr->pixrData.plane.xRes = xRes;
	pixr->pixrData.plane.yRes = yRes;
    }
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetResolutionDPI(pIPWState, ch, xRes, yRes));
	}
    }
	
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the type of the PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pType::          Ptr to PIXR type
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetType (
    IPWState           *pIPWState,
    PIXR               *pixr,
    ColorType          *pType)
{
    PROLOGUE;
    Int32         channels;
    ColorType     type;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pType == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        type = pixr->pixrData.plane.type;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetType(pIPWState, ch, &type));
    }

    *pType = type;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the type of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  type::           type value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetType (
    IPWState           *pIPWState,
    PIXR               *pixr,
    ColorType           type)
{
    PROLOGUE;
    Int32     i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	pixr->pixrData.plane.type = type;
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetType(pIPWState, ch, type));
	}
    }
	
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the byte order PIXR image words
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pByteOrder::     Ptr to PIXR byteOrder.  Will return value of
 *		     cBigEndianByteOrder or cLittleEndianByteOrder.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetByteOrder (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt32             *pByteOrder)
{
    PROLOGUE;
    Int32         channels;
    UInt32	  byteOrder;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pByteOrder == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        byteOrder = pixr->pixrData.plane.byteOrder;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetByteOrder(pIPWState, ch, &byteOrder));
    }

    *pByteOrder = byteOrder;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the byteOrder of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  byteOrder::      byteOrder value
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  D. Davies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetByteOrder (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt32		byteOrder)
{
    PROLOGUE;
    Int32     i, channels;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN((byteOrder != cBigEndianByteOrder) &&
	    (byteOrder != cLittleEndianByteOrder), cErrInvalidParm);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	pixr->pixrData.plane.byteOrder = byteOrder;
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrSetByteOrder(pIPWState, ch, byteOrder));
	}
    }
	
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the colormap of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  ppMap::          Ptr to ptr to PIXR colormap
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetColorMap (
    IPWState           *pIPWState,
    PIXR               *pixr,
    ColorMap          **ppMap)
{
    PROLOGUE;
    Int32      channels;
    ColorMap  *pMap;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(ppMap == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        pMap = pixr->pixrData.plane.colorMap;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetColorMap(pIPWState, ch, &pMap));
    }

    *ppMap = pMap;
    NORMAL_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the colormap of a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  map::            Ptr to colormap
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetColorMap (
    IPWState           *pIPWState,
    PIXR               *pixr,
    ColorMap           *map)
{
    PROLOGUE;
    Int32 i, channels;
    ColorMap *prevCmap;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
    {
	ERR_PASS( w_pixrGetColorMap(pIPWState, pixr, &prevCmap));
	ERR_GEN(prevCmap != NULL, cErrPixr_colormap);
	pixr->pixrData.plane.colorMap = map;
    }
    else
    {
	PIXR  *ch;
	
	for (i = 0; i < channels; i++)
	{
	    ERR_PASS( w_pixrGetChannel(pIPWState, pixr, i, &ch));
	    ERR_PASS( w_pixrGetColorMap(pIPWState, ch, &prevCmap));
	    ERR_GEN(prevCmap != NULL, cErrPixr_colormap);
	    ERR_PASS( w_pixrSetColorMap(pIPWState, ch, map));
	}
    }
	
    NORMAL_RETURN;
}



/*--------------------------------------------------------------------*
 *                        Pointers to Image Data                      *
 *--------------------------------------------------------------------*/

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get ptr to upper left corner of image, including frame, of PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  ppImage::        Ptr to ptr to image
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetImageAll (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt8             **ppImage)
{
    PROLOGUE;
    Int32         channels;
    UInt8        *pImage;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(ppImage == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        pImage = pixr->pixrData.plane.image;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetImageAll(pIPWState, ch, &pImage));
    }

    *ppImage = pImage;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get ptr to buffer which contains the image.  This may return a different
 *  value than w_pixrGetImageAll due to alignment constraints.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  ppImageBuf::     Ptr to ptr to 1st byte of image buffer
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_pixrGetImageBuffer (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt8             **ppImageBuf)
{
    PROLOGUE;
    Int32         channels;
    UInt8        *pImageBuf;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(ppImageBuf == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        pImageBuf = pixr->pixrData.plane.imageBuffer;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( wp_pixrGetImageBuffer(pIPWState, ch, &pImageBuf));
    }

    *ppImageBuf = pImageBuf;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get ptr to specified scan line of a PIXR.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR.
 *  line::           Number of scan line to get ptr to.
 *
 * Output:::
 *  pLine::          Ptr to ptr to desired image scan line.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dbloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_getLine (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               line,
    UInt8             **pLine)
{
    PROLOGUE;
    Int32         h, Bpl;
    UInt8        *data;

    ERR_GEN(pixr == NULL, cErrNullPtr);
    ERR_GEN(pLine == NULL, cErrNullPtr);


    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));

    ERR_GEN(line < 0 || line >= h, cErrOutOfBounds);

    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &Bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, pixr, &data));
    
    ERR_GEN(data == NULL, cErrNoImageData);

    *pLine = data + line * Bpl;
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set ptr to upper left corner of image, including frame, of PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  image::          Ptr to image (including frame)
 *
 * Returns:::
 *
 * Notes:::
 *  This function really doesn't make any sense when applied to a multi-
 *  plane image.  An error is returned if called with a pointer to a
 *  multi-plane image.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetImageAll (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt8              *image)
{
    PROLOGUE;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannels1, cFrameAny);

	/* since there can be only one channel, just set it. */
    pixr->pixrData.plane.image = image;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set PIXR image buffer pointer
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  imageBuf::       Ptr to beginning of mem buffer that contains image
 *
 * Returns:::
 *
 * Notes:::
 *  This function really doesn't make any sense when applied to a multi-
 *  plane image.  An error is returned if called with a pointer to a
 *  multi-plane image.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_pixrSetImageBuffer (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt8              *imageBuf)
{
    PROLOGUE;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannels1, cFrameAny);

	/* since there can be only one channel, just set it. */
    pixr->pixrData.plane.imageBuffer = imageBuf;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get ptr to upper left corner of image, excluding frame, of PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  ppImage::        Ptr to ptr to image
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetImage (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt8             **ppImage)
{
    PROLOGUE;
    Int32         channels;
    UInt8        *pImage;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(ppImage == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
    {
        pImage = (UInt8 *)(pixr->pixrData.plane.image
			     + pixr->pixrData.plane.frame *
			     pixr->pixrData.plane.bpl
			     + pixr->pixrData.plane.frame *
			     pixr->pixrData.plane.d / 8);
    }
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrGetImage(pIPWState, ch, &pImage));
    }

    *ppImage = pImage;
    NORMAL_RETURN;
}




/*--------------------------------------------------------------------*
 *               widths and heights of various images                 *
 *--------------------------------------------------------------------*/


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Calculate the byte width of entire image
 *
 * LongDescription:::
 *  That is, calculate from the PIXR's width, frame, and depth how many
 *  bytes the PIXR will need per image line.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pBpl::           Ptr to bytes per line PIXR will/does require
 *
 * Returns:::
 *
 * Notes:::
 *  Doesn't set::
 *   This does not set the PIXR's bytes per line field
 *  PIXR depth, frame, width::
 *   Must all be set and valid prior to call!
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrBpl (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pBpl)
{
    PROLOGUE;
    Int32         channels, bpl;
    Int32	  bufSize;


    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pBpl == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
       ERR_PASS(w_pixrImageProperties(pIPWState,
				      pixr->pixrData.plane.w,
				      pixr->pixrData.plane.h,
				      pixr->pixrData.plane.d,
				      pixr->pixrData.plane.frame,
				      &bpl, &bufSize));
    else
    {
	PIXR  *ch = NULL;

	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrBpl(pIPWState, ch, &bpl));
    }

    *pBpl = bpl;
    NORMAL_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Calculate PIXR image width in pixels including frame, excluding padding
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pWidth::         Ptr to framed width of PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  PIXR frame, width::
 *   Must be set and valid prior to call!
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrFramedWidth (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pWidth)
{
    PROLOGUE;
    Int32         channels, w;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pWidth == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        w = pixr->pixrData.plane.w + 2 * pixr->pixrData.plane.frame;
    else
    {
	PIXR  *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrFramedWidth(pIPWState, ch, &w));
    }

    *pWidth = w;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Calculate PIXR image height in pixels including frame, excluding padding
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pHeight::        Ptr to framed height of PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  PIXR frame, height::
 *   Must be set and valid prior to call!
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrFramedHeight (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pHeight)
{
    PROLOGUE;
    Int32         channels, h;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pHeight == NULL, cErrNullPtr);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1) 
        h = pixr->pixrData.plane.h + 2 * pixr->pixrData.plane.frame;
    else
    {
	PIXR   *ch = NULL;
	
	ERR_PASS( w_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	ERR_PASS( w_pixrFramedHeight(pIPWState, ch, &h));
    }

    *pHeight = h;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Calculate the number of bytes in the entire PIXR image
 *
 * LongDescription:::
 *  That is, calculate from the PIXR's width, height, frame, and depth
 *  how many bytes the PIXR will need to hold the entire image.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pNBytes::        Ptr to number of bytes PIXR image will/does require
 *
 * Returns:::
 *
 * Notes:::
 *  Doesn't set::
 *   This does not set any PIXR field.
 *  PIXR width, height, depth, frame::
 *   Must all be set and valid prior to call!
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrImageBytes (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pNBytes)
{
    PROLOGUE;
    Int32    fh, bpl;

    ERR_GEN_PIXR(pixr, cByteOrderAny, cDepthAny, cChannelsAny, cFrameAny);
    ERR_GEN(pNBytes == NULL, cErrNullPtr);

    /* This gives the # of bytes in the "primary" plane */
    ERR_PASS( w_pixrBpl(pIPWState, pixr, &bpl));
    ERR_PASS( w_pixrFramedHeight(pIPWState, pixr, &fh));
    
    *pNBytes = fh * bpl;
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Calculate bytes/line and memory buffer size for PIXR
 *
 * LongDescription:::
 *  That is, calculate from the PIXR's width, height, frame, and depth
 *  the number of bytes in a scan line and the size of the memory buffer
 *  needed to hold the image.  This calculation takes into account
 *  padding for long accesses, alignment issues and extra padding to
 *  allow right angle rotation in place.
 *  This routine may be of use to clients who create their own blocks
 *  of image memory and attach PIXR structures to them.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  width::          width of image in pixels
 *  height::	     height of image in pixels
 *  depth::	     bits/pixel (we assume 1 sample/pixel)
 *  frame::	     number of pixels in frame (normally equal to cFrameBits)
 *
 * Output:::
 *  pBpl::           Ptr to image's bytes/line
 *  pBufSize::	     Ptr to size of memory buffer required to hold image
 *
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrImageProperties (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    Int32               depth,
    Int32               frame,
    Int32              *pBpl,
    Int32	       *pBufSize)
{
    PROLOGUE;

    ERR_GEN((pBpl == NULL) || (pBufSize == NULL), cErrNullPtr);

	/* This gives the # of bytes in the "primary" plane */
    ERR_PASS( i_calcImageProperties (pIPWState, width, height, depth, frame,
				     pBpl, pBufSize));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Calculate byte offset from the start of image buffer to location (0, 0)
 *
 * LongDescription:::
 *  Due to alignment constraints, especially with 64-bit processors,
 *  the upper left corner of a frame may not lie exactly on the first
 *  byte of an allocated memory buffer.  This routine calculates the
 *  byte offset from the start of a memory buffer to location (0, 0)
 *  of the image.  This is the upper left corner *inside* the frame.
 *  This routine may be of use to clients who create their own blocks
 *  of image memory and attach PIXR structures to them.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pImageBuf::	     Ptr to first byte of memory buffer holding image.
 *  bpl::            image's bytes/line (see w_pixrImageProperties)
 *  depth::	     bits/pixel (we assume 1 sample/pixel)
 *  frame::	     number of pixels in frame (normally equal to cFrameBits)
 *
 * Output:::
 *  pOffsetTo00::    Ptr to offset from pImageBuf to location (0,0).
 *
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrImageOffset (
    IPWState           *pIPWState,
    UInt8              *pImageBuf,
    Int32               bpl,
    Int32               depth,
    Int32               frame,
    Int32              *pOffsetTo00)
{
    PROLOGUE;

    ERR_GEN(pOffsetTo00 == NULL, cErrNullPtr);

	/* This gives offset */
    ERR_PASS( i_calcImageOffset (pIPWState,  pImageBuf, bpl, depth, frame,
				 pOffsetTo00));
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the number of channels in a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  pChannels::      Ptr to number of PIXR channels
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetChannelCount (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32              *pChannels)
{
    PROLOGUE;

    ERR_GEN(pixr == NULL, cErrPixr_null);
    ERR_GEN(pChannels == NULL, cErrNullPtr);

    *pChannels = pixr->channels;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of channels in a PIXR
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  channels::       number of channels value
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetChannelCount (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               channels)
{
    PROLOGUE;

    /* Keep anal compilers happy */
    pIPWState = pIPWState;

    ERR_GEN(pixr == NULL, cErrPixr_null);

    /* NOTE that if you set this on something that had more channels before,
       the sub-channels are then left danging. */
    pixr->channels = (UInt16) channels;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get ptr to one of the sub-channels (another PIXR) of a PIXR .
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  channel::        Which channel to get
 *
 * Output:::
 *  ppSubPixr::      Ptr to ptr to desired PIXR channel (a sub PIXR)
 *
 * Returns:::
 *
 * Notes:::
 *  no PIXR verification::
 *   The normal verification of the PIXR is not done here to avoid potential
 *   problems when destroying PIXRs.  However, the first call to a PIXR
 *   accessor routine on this returned channel will do that checking then.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetChannel (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               channel,
    PIXR              **ppSubPixr)
{
    PROLOGUE;
    Int32 channels;

    ERR_GEN(ppSubPixr == NULL, cErrNullPtr);
    ERR_GEN(pixr == NULL, cErrPixr_null);

    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    ERR_GEN(channel < 0 || channel >= channels, cErrInvalidParm);

    if (channels == 1)
	*ppSubPixr = pixr;
    else
	*ppSubPixr = pixr->pixrData.childPlanes[channel];

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set ptr to one of the sub-channels (another PIXR) of a PIXR .
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  channel::        Which channel to set
 *  subPixr::        Ptr to another PIXR to use as channel
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *  Dangling channels::
 *   If you set a channel that was already set, the previously set channel
 *   is left dangling.  One should not do this, or should get the channel
 *   first and deal with this old channel (by saving the ptr to it or by
 *   destroying it), and then set the channel to the new one.
 *  no PIXR verification::
 *   The normal verification of the PIXR is not done here to avoid potential
 *   problems when destroying PIXRs.  However, the first call to a PIXR
 *   accessor routine on this set channel will do that checking then,
 *   so client should have done any necessary PIXR checking prior to this
 *   routine call.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrSetChannel (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               channel,
    PIXR               *subPixr)
{
    PROLOGUE;
    Int32 channels;

    ERR_GEN(pixr == NULL, cErrPixr_null);
    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    ERR_GEN(channels == 1 || channel < 0 || channel >= channels,
	    cErrInvalidParm);

    pixr->pixrData.childPlanes[channel] = subPixr;

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ptr to the single "primary" channel of a PIXR.
 *
 * LongDescription:::
 *  Return the "primary" channel of the image.  If the image is RGB, return
 *  the green channel. If the image is Yxx, return the Y channel. If the image
 *  is CMYK, return the, uh, well, magenta channel.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  ppSubPixr::      Ptr to ptr to primary PIXR channel (a sub PIXR)
 *
 * Returns:::
 *
 * Notes:::
 *  Must be multi-channel::
 *   The PIXR must contain at least 2 channels
 *
 *  Internally called::
 *   This is a private routine used to avoid problem of recursive PIXR
 *   checking.  Thus, the PIXR must have already been verified at calling
 *   public PIXR routine.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrGetPrimaryChannel (
    IPWState           *pIPWState,
    PIXR               *pixr,
    PIXR              **ppSubPixr)
{
    PROLOGUE;
    PIXR       *ch;
    Int32       channel;

    ERR_GEN(pixr->channels == 1, cErrInternal);

    /* The types should be the same, so get the first one */
    ch = pixr->pixrData.childPlanes[cPseudoPlane];

    switch (ch->pixrData.plane.type)
    {
    case cColorRGB:	channel = cGreenPlane; break;
    case cColorCMYK:	channel = cMagentaPlane; break;
    case cColorCIELAB:
    case cColorCIELUV:	channel = cLuminancePlane; break;
    default:		channel = cPseudoPlane; break;
    }

    ERR_GEN(channel < 0 || (UInt32)channel >= (pixr->channels), cErrInternal);

    *ppSubPixr = pixr->pixrData.childPlanes[channel];
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ptr to the single "primary" channel of a PIXR.
 *
 * LongDescription:::
 *  Return the "primary" channel of the image. If the image is 1 channel, just
 *  return it.  If the image is RGB, return the green channel. If the image is
 *  Yxx, return the Y channel. If the image is CMYK, return the, uh, well,
 *  magenta channel.  
 *  <BR><BR>
 *  Typically this channel is used as the "exemplar" channel
 *  and its fields are used in most w_pixrGet**() calculations.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *
 * Output:::
 *  ppSubPixr::      Ptr to ptr to primary PIXR channel (a sub PIXR)
 *
 * Returns:::
 *
 * Notes:::
 *  no PIXR verification::
 *   The normal verification of the PIXR is not done here to avoid potential
 *   problems when destroying PIXRs.  However, the first call to a PIXR
 *   accessor routine on this returned channel will do that checking then,
 *   so client should have done any necessary PIXR checking prior to this
 *   routine call.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrGetPrimaryChannel (
    IPWState           *pIPWState,
    PIXR               *pixr,
    PIXR              **ppSubPixr)
{
    PROLOGUE;
    Int32	channels;
    PIXR       *ch = NULL;

    ERR_GEN(ppSubPixr == NULL, cErrNullPtr);
    ERR_PASS( w_pixrGetChannelCount(pIPWState, pixr, &channels));

    if (channels == 1)
	*ppSubPixr = pixr;
    else
    {
	/* Multi-channeled PIXR */
	ERR_PASS( wp_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	*ppSubPixr = ch;
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Validate a PIXR for byte order, depth, channel count, and frame.
 *
 * LongDescription:::
 *  If the routine returns without error, the PIXR has been validated.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR
 *  byteFlags::      Allowable value(s) for PIXR byte order.
 *                   May be {cByteOrderBig, cByteOrderLittle}
 *                   (or any OR-ed combination) or cByteOrderAny.
 *  depthFlags::     Allowable value(s) for PIXR depth.
 *                   May be {cDepth1, cDepth2, cDepth4, cDepth8, cDepth10,
 *                   cDepth12, cDepth16}
 *                   (or any OR-ed combination) or cDepthAny.
 *  channelFlags::   Allowable value(s) for PIXR number of channels.
 *                   May be {cChannels1, cChannels3}
 *                   (or any OR-ed combination) or cChannelsAny.
 *  frameFlags::     Allowable value(s) for PIXR frame (# pixels in frame).
 *                   May be {cFrame0, cFrame32}
 *                   (or any OR-ed combination) or cFrameAny.
 *
 * Returns:::
 *
 * Notes:::
 *  byte order::
 *    cByteOrderBig is Big-endian: MSB at lowest addr.
 *    cByteOrderLittle is Little-endian: LSB at lowest addr.
 *
 * FunctionCategory:::
 *  image management
 *
 * Keywords:::
 *  image management
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_pixrCheckFlags (
    IPWState           *pIPWState,
    PIXR               *pixr,
    UInt32              byteFlags,
    UInt32              depthFlags, 
    UInt32              channelFlags, 
    UInt32              frameFlags)
{
    PROLOGUE;
    PIXR    *ch;
    
    ERR_GEN(pixr == NULL, cErrPixr_null);

    switch (pixr->channels)
    {
    case  1: ERR_GEN((channelFlags & cChannels1) == 0, cErrPixr_channels);
	     break;
    case  3: ERR_GEN((channelFlags & cChannels3) == 0, cErrPixr_channels);
	     break;
    default: ERR_GEN(1, cErrPixr_channels); break;
    }

    if (pixr->channels != 1)
    {
	ERR_PASS( wp_pixrGetPrimaryChannel(pIPWState, pixr, &ch));
	pixr = ch;

	ERR_GEN(pixr == NULL, cErrPixr_null);
    }

    switch (pixr->pixrData.plane.d)
    {
    case  1: ERR_GEN((depthFlags & cDepth1 )== 0, cErrPixr_depth); break;
    case  2: ERR_GEN((depthFlags & cDepth2 )== 0, cErrPixr_depth); break;
    case  4: ERR_GEN((depthFlags & cDepth4 )== 0, cErrPixr_depth); break;
    case  8: ERR_GEN((depthFlags & cDepth8 )== 0, cErrPixr_depth); break;
    case 10: ERR_GEN((depthFlags & cDepth10)== 0, cErrPixr_depth); break;
    case 12: ERR_GEN((depthFlags & cDepth12)== 0, cErrPixr_depth); break;
    case 16: ERR_GEN((depthFlags & cDepth16)== 0, cErrPixr_depth); break;
    default: ERR_GEN(1, cErrPixr_depth); break;
    }

    switch (pixr->pixrData.plane.frame)
    {
    case  0: ERR_GEN((frameFlags & cFrame0) == 0, cErrPixr_frame); break;
    case 32: ERR_GEN((frameFlags & cFrame32)== 0, cErrPixr_frame); break;
    default: ERR_GEN(1, cErrPixr_frame); break;
    }

    switch (pixr->pixrData.plane.byteOrder)
    {
    case cBigEndianByteOrder: ERR_GEN((byteFlags & cByteOrderBig) == 0,
				      cErrPixr_byteOrder); break;
    case cLittleEndianByteOrder: ERR_GEN((byteFlags & cByteOrderLittle) == 0,
				         cErrPixr_byteOrder); break;
    default: ERR_GEN(1, cErrPixr_byteOrder); break;
    }

    NORMAL_RETURN;
}


/*

 $Log: props2.c,v $
 Revision 1.2  2001/09/28 23:38:22  dabercro
 w_getLine moved from readfile.c to props2.c (so that Digipaper doesn't need
 the library that readfile.c resides in)

 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.36  1998/01/14 00:14:21  ddavies
 props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.35  1997/11/01 21:02:29  aber
 aber on Sat Nov 1 13:02:05 PST 1997

 props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v

 Return an error if adding a colormap when one already exists.

 Revision 1.34  1996/09/13 23:26:39  ddavies
 ddavies on Fri Sep 13 16:19:39 PDT 1996

 props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v

 Add implementations of wp_pixrGetImageBuffer, wp_pixrSetImageBuffer,
 w_pixrImageProperties and w_pixrImageOffset.
 The get and set routines allow access to the new imageBuffer entry in a
 PIXR>  This may not point to the upper left corner of an image due to
 alignment constraints.  The Properties and Offset routines return the
 bytes/line, required memory buffer size and offset within a buffer for
 an image.  These can be used by clients to create legal images.
 Change w_pixrBpl to use w_pixrImageProperties.

 Revision 1.33  1996/05/18 00:39:14  ejaquith
 ejaquith on Fri May 17 17:37:37 PDT 1996

 props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v

 Fixed broken wp_pixrCheckFlags().
 While attempting to fix a compiler (gcc 2.7.2) warning on a bunch of
 lines doing "a & b == 0", the parens ended up in the wrong place,
 Net was PIXRs were always validating.  Now of form "(a & b) == 0".

 Revision 1.32  1996/02/10 01:08:27  ddavies
 ddavies on Fri Feb 9 17:08:08 PST 1996

 props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v

 Add parens around conditional expressions to make gcc 2.7.2 happy.

 Revision 1.31  1996/01/04 20:25:37  ejaquith
 props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v

 CommonCheck: Removed all HTML tags that we had placed in our C code.
 Removed double-quoted routine name after "BeginDocumentation" and
 "EndDocumentation".

 * Revision 1.30  1995/12/21  04:39:20  ddavies
 * ddavies on Wed Dec 20 20:37:55 PST 1995
 *
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Delete code that returns cErrInvalidParm if the bytes/line parameter
 * supplied to w_pixrSetBpl is negative.  We use negative bytes/line to
 * convince w_rasterOp to work from the bottom of the image to the top.
 * This is the only way to shift things down when there's overlap.
 *
 * Revision 1.29  1995/12/13  04:11:31  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.28  1995/12/01  19:32:47  ddavies
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Fix acc compiler warnings.
 *
 * Revision 1.27  1995/11/13  20:14:37  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.26  1995/10/28  01:35:47  ddavies
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Add implementations for w_pixrGetByteOrder and w_pixrSetByteOrder.  These
 * routines get and set the byteOrder flag of a PIXR respectively.
 * Also, change w_pixrSetImageAll so that it complains if handed a multi-plane
 * image instead of setting the image pointers on all planes to point to the
 * same image.
 * Also set the byteOrder flag to cBigEndianByteOrder whenever a PIXR is
 * created.  Users wishing to have little endian PIXRs can use
 * w_pixrSetByteOrder to flip it.
 * Also have wp_pixrCheckFlags check the byteOrder flag instead of just
 * saying it's ok.
 *
 * Revision 1.25  1995/10/13  22:03:55  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.24  1995/10/12  04:26:45  ddavies
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Add casts and constant types to make Visual C++ happy.
 *
 * Revision 1.23  1995/09/29  00:36:06  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Filled in standardized routine header comment blocks.
 *
 * Revision 1.22  1995/09/21  22:58:11  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.21  1995/09/19  03:20:17  ddavies
 * ddavies on Mon Sep 18 20:18:16 PDT 1995
 *
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Have all the pixrGet/SetChannel routines stop doing ERR_GEN_PIXR with
 * all the checks set to -Any.  Replace with explicit checks to see if the
 * pixr is NULL.  This is to avoid a circular problem when a color pixr
 * is being deleted.  Each call to pixrGetChannel invoked ERR_GEN_PIXR, which
 * invoked wp_pixrGetPrimaryChannel, which might have already been deleted,
 * which caused big trouble.
 *
 * Revision 1.20  1995/09/14  02:09:31  ddavies
 * ddavies on Wed Sep 13 19:09:14 PDT 1995
 *
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Changed all resolutions to ScaledInt1616 from Int16.
 *
 * Revision 1.19  1995/09/13  00:23:24  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Created wp_pixrGetPrimaryChannel().
 * Made pixrGetFlags() use PIXR fields directly.
 * Made PIXR accessor public functions use ERR_GEN_PIXR()
 * (and hence wp_pixrGetFlags())
 * Fixed recursive tricky stuff.
 *
 * Revision 1.18  1995/09/12  20:21:09  ejaquith
 * ejaquith on Tue Sep 12 13:20:11 PDT 1995
 *
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Made all routines check only if the pixr was NULL on entrance to the routine
 * instead of how I was previously using ERR_GEN_PIXR() with all flags
 * set to "Any".
 *
 * Revision 1.17  1995/09/12  16:55:30  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Made all public (w_* routines) check ALL pointers on entrance to routine.
 * Made all public routines passed PIXRs use ERR_GEN_PIXR().
 * Changed as necessary routines that were modifying output pointers before
 * NORMAL_RETURN.  (Now they only modify them the line before the successful
 * return).  Made some routines private (wp_* routines) as necessary.
 * Updated comments at top of file.
 *
 * Revision 1.16  1995/09/06  20:22:13  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.15  1995/09/01  02:21:09  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Put minimum number of include files back in.
 *
 * Revision 1.14  1995/08/31  21:15:57  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.13  1995/08/25  01:37:50  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Fixed warnings
 *
 * Revision 1.12  1995/08/22  02:21:35  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Nailed some PC warnings.
 *
 * Revision 1.11  1995/08/17  22:02:01  ejaquith
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Spring cleaning I.
 *
 * Revision 1.10  1994/09/30  15:52:41  lperry
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Add include of utils.pub
 *
 * Revision 1.9  1994/09/29  23:13:48  lperry
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Modify to use props.pub instead of pixrpros.h
 *
 * Revision 1.8  1994/06/07  02:39:35  macleod
 * macleod on Mon Jun 6 19:39:34 PDT 1994
 *
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Some of the routines which called themselves with a single-channel sub-pixr
 * when passed a multi-channel image were not checking the error return of the
 * recursive calls. This was a problem when a yieldproc indicated that an abort
 * was requested.
 *
 * Revision 1.7  1994/05/26  19:23:35  ddavies
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Add #ifndef PRODUCTION around the entire error check in pixrSetChannelCount.
 * Without this, the "if (!pixr)" has an empty statement when PRODUCTION is
 * enabled.
 *
 * Revision 1.6  1994/05/05  01:46:39  ddavies
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Correct pixrSetChannelCount's procName (used to be pixrGetChannelCount).
 * Make this procedure return nothing since it's now defined to return void.
 *
 * Revision 1.5  1994/04/12  01:15:03  macleod
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Added support for color PIXRs
 *
 * Revision 1.4  1994/04/05  00:41:47  macleod
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Added pixrGetResolution and pixrSetResolution, to deal with resolution
 * fields of PIXRs
 *
 * Revision 1.3  1993/12/03  01:17:35  ddavies
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:14:55  ddavies
 * props2.c -> /product/ipcore/ipwrap/src/RCS/props2.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

