/*<HTML><HEAD><TITLE>props6.c</TITLE></HEAD><BODY><PRE>*/
/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  props6.c:
 *
 *	     Pixr clearing (and setting)
 *			w_setFramePixels()	(sets pixel value)
 *			w_setFramePixelsGray() 
 *
 *	     Pixr bit inversion
 *			w_invert ()
 *
 *	     Pixr centering
 *			w_center()
 *
 *	     grabbing a subimage of a Pixr
 *			w_rectOf()
 *
 *	     Filling rectangles, etc
 *			w_paint()
 *
 *	     Clearing an image
 *			w_clear()
 *	     
 */


#include "ipwcore.prv"

#include "geomadt.h"
#include "imageref.pub"
#include "alplib.h"
#include "lmask.prv"
#include "frames.pub"
#include "defines.pub"
#include "pixr.h"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "rectops.pub"
#include "shrpixr.pub"
#include "shrrast.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: props6.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the frame pixels (to 0 or max value) surrounding the PIXR's image.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  pixr::           Ptr to PIXR.
 *  value::          The value can be nonzero(ON) or 0(OFF). 
 *                   (This routine can set pixels to either 0 or their maximum
 *                   value, not to any intermediate value.)
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_setFramePixels (
    IPWState           *pIPWState,
    PIXR               *pixr,
    Int32               value)
{
    PROLOGUE;
    Int32        w, h, op, f;
    Int32        iw, ih;


    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);
    
    ERR_PASS( w_pixrFramedWidth(pIPWState, pixr, &w));
    ERR_PASS( w_pixrFramedHeight(pIPWState, pixr, &h));
    ERR_PASS( w_pixrGetFrame(pIPWState, pixr, &f));

    /* return now if there is no frame */
    if ( f <= 0 )
	NORMAL_RETURN;

    iw = w - f - f;	/* width of proper image */
    ih = h - f - f;	/* height of proper image */

    op = (value != 0) ? PIX_SET | PIX_DONTCLIP : PIX_CLR | PIX_DONTCLIP;

    /* pixr_rop handles color images */

    /* top frame (full width) */
    ERR_PASS( w_rasterop (pIPWState,
			  0, 0,
			  NULL,
			  op, 
			  -f, -f, w, f,
			  pixr));

    /* left frame (starting below top frame, just image height) */
    ERR_PASS( w_rasterop (pIPWState,
			  0, 0,
			  NULL,
			  op, 
			  -f, 0, f, ih,
			  pixr));
    
    /* bottom frame (full width) */
    ERR_PASS( w_rasterop (pIPWState,
			  0, 0,
			  NULL, 
			  op, 
			  -f, ih, w, f, 
			  pixr));

    /* right frame (between top frame and bottom frame) */
    ERR_PASS( w_rasterop (pIPWState,
			  0, 0,
			  NULL,
			  op,
			  iw, 0, f, ih,
			  pixr));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the frame pixels (to any value) surrounding the PIXR's image.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  data::           ptr to PIXR image data
 *  w::              full framed width, in pixels
 *  h::              full framed height, in pixels
 *  bpl::            bytes per line of full image
 *  f::              frame width in pixels
 *  value::          new value to set frame pixels
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_setFramePixelsGray (
    IPWState           *pIPWState,
    UInt32             *data,
    Int32               w,
    Int32               h,
    Int32               bpl,
    Int32               f,
    Int32               value)
{
    PROLOGUE;
    Int32        i, j;
    Int32        wpl, height, width, fw;
    Int32        bigval, excess;
    UInt32      *pword, *pline;


    ERR_GEN(data == NULL, cErrNullPtr);

    bigval = value | (value << 8) | (value << 16) | (value << 24);
    wpl = bpl >> 2;
    height = h - 2 * f;  /* proper height */
    width = w - 2 * f;   /* proper width */
    fw = f >> 2;    /* number of words in the frame width */

	/* set top */
    pword = data;
    for (i = 0; i < f; i++)
	for (j = 0; j < wpl; j++)
	    *pword++ = bigval;

	/* set bottom */
    pword = data + (f + height) * wpl;
    for (i = 0; i < f; i++)
	for (j = 0; j < wpl; j++)
	    *pword++ = bigval;

	/* set left */
    pline = data + f * wpl;
    for (i = 0; i < height; i++)
    {
	pword = pline + i * wpl;
	for (j = 0; j < fw; j++)
	    *pword++ = bigval;
    }

	/* set right */
    /*    excess = width & 3; ...PSM: this is wrong, I think */
    excess = 4 - (width & 3);
    if (excess)
    {
	pline = data + f * wpl + fw;  /* start of proper image */
	for (i = 0; i < height; i++) {
	    for (j = 0; j < excess; j++)
		SET_IMAGE_BYTE(pline, width + j, (UInt8) value);
	    pline += wpl;
	}
    }

    pline = data + f * wpl + fw + ((width + 3) >> 2);
    for (i = 0; i < height; i++)
    {
	pword = pline + i * wpl;
	for (j = 0; j < fw; j++)
	    *pword++ = bigval;
    } 

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Bit invert PIXR's image(s)
 *
 * LongDescription:::
 *      Handles all bit depths and allows in-place inversion.  For 
 *	in-place inversion, call as w_invert(pIPWState, spr, &spr) <BR><BR>
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  spr::            Ptr to source PIXR
 *
 * Output:::
 *  dpr::            Ptr to destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *  Binary PIXRs::
 *   Routine leaves the frame's bits cleared.
 *
 *  Gray scale PIXRs::
 *   DOES differentiate between the frame's bits and other bits.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_invert (
    IPWState           *pIPWState,
    PIXR               *spr,
    PIXR              **dpr)
{
    PROLOGUE; 
    Int32    channels, channel;
    Int32    rightFrameColsToClear, bottomFrameRowsToClear;
    Bool     pPixrAllocatedHere = cFalse;

    ERR_GEN_PIXR(spr, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    /* generate the destination pixr if necessary.  if not, check
     * it for validity.
     */
    if (spr != *dpr)
        ERR_PASS(wp_clonePixrConditionalNoInit(pIPWState, spr, dpr,
					       &rightFrameColsToClear,
					       &bottomFrameRowsToClear,
					       &pPixrAllocatedHere));
 
    /* If it's multi-channel, do it on all the subchannels */
    ERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, spr, &channels));

    if (channels != 1)
    {
	PIXR  *dch, *sch;
	
	for (channel = 0; channel < channels; channel++)
	{
	    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, spr, channel,
			     &sch));
	    ERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, *dpr, channel,
			     &dch));
	    ERR_PASS_CLEANUP(w_invert(pIPWState, sch, &dch));
	}
    }
    else 
    {
	/* If src and dst are the same, do the inversion in place.
 	 * this is special cased 'cause the in place code should be faster
	 * than calling rasterop.
	 */
	if (*dpr != spr)
	{
	    Int32   dw, dh, dbpl, df, dd;
	    Int32   sw, sh, sbpl, sf, sd;
	    UInt8  *pdImage, *psImage;
	    
	    ERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, *dpr, &dd));
	    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, *dpr, &dw));
	    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, *dpr, &dh));
	    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, *dpr, &dbpl));
	    ERR_PASS_CLEANUP(w_pixrGetFrame(pIPWState, *dpr, &df));
	    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, *dpr, &pdImage));
	    
	    ERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, spr, &sd));
	    ERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, spr, &sw));
	    ERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, spr, &sh));
	    ERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, spr, &sbpl));
	    ERR_PASS_CLEANUP(w_pixrGetFrame(pIPWState, spr, &sf));
	    ERR_PASS_CLEANUP(w_pixrGetImageAll(pIPWState, spr, &psImage));

	    ERR_PASS_CLEANUP(i_rasterOpFull (pIPWState, pdImage,
					     dw, dh, df, df, 0, 0,
					     dw, dh, dd, dbpl,
					     PIX_NOT(PIX_SRC),
					     psImage, sw, sh, 0, 0, sf, sf,
					     sd, sbpl));
	    if ((rightFrameColsToClear | bottomFrameRowsToClear) != 0)
	        ERR_PASS_CLEANUP(
		    wp_clearRightBottomFrames(pIPWState, *dpr, 
				 	      rightFrameColsToClear, 
					      bottomFrameRowsToClear));
	    
	}
	else
	{
	    ERR_PASS_CLEANUP(wp_invertIP(pIPWState, *dpr));
	}
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (pPixrAllocatedHere == cTrue) w_destroyPixr(pIPWState, dpr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Bit invert PIXR's image(s) in place.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *
 * Output:::
 *  pixr::           Ptr to PIXR that gets bit inverted in place.
 *
 * Returns:::
 *
 * Notes:::
 *  Since this is a private procedure it does not check parameters
 *  for validity.  It also assumes that it WILL not get passed a
 *  multi-channel pixr.
 * 
 *  Frame bits::
 *   Routine leaves frame's bits as is.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_invertIP (
    IPWState           *pIPWState,
    PIXR               *pixr)
{
    PROLOGUE; 
    UInt32  *srcData, *pSrc;
    UInt8   *pImage;
    Int32    i, w, h, Bpl, depth;
    UInt32   wInBits, wInWords, extraBits;
    UInt32   lastWordMask;


    /* There's a magic thing about 2's complement arithmetic.  "Inverting"
     * a pixr - meaning subtracting each pixel from the maximum
     * pixel value - is equivalent to doing a simply 1's complement
     * inversion on each bit of each pixel.  That's easier.
     *
     * We do want to leave the borders untouched.  The variable
     * wInWords in the width of the gray pixr in units of words
     * *rounded down* to the nearest word boundary.  wLastWord is
     * the number of valid pixr bits in the last word.  We do
     * each line by flipping entire words, then flipping the leftmost
     * wLastWord bits in the rightmost word.
     *
     * For the sake of efficiency, we don't want the stupid compiler
     * shifting the array index every time it's used.  That means we
     * want a byte index.  However, we want it to flip entire words
     * each time.  So, we make the byte index point to an int by
     * rounding down to the next Int32 boundary.
     */

    /* # of image pixels per line */
    ERR_PASS( w_pixrGetDepth(pIPWState, pixr, &depth));
    ERR_PASS( w_pixrGetWidth(pIPWState, pixr, &w));
    wInBits      = w * depth;          /* Tot # of image bits */
    wInWords     = wInBits / 32;       /* # of full 32 bit words */
    extraBits    = wInBits & 31;       /* # of bits in last word, if any */

    /*
      If there are any image bits at all in the last 
      word, select a mask to use for inverting only
      the image bits. leftMask[n-1] contains a mask
      suitable for n leftmost bits.
    */
    lastWordMask = 0;
    if (extraBits != 0)
	lastWordMask = leftMask[extraBits - 1];

    ERR_PASS( w_pixrGetHeight(pIPWState, pixr, &h));
    ERR_PASS( w_pixrGetBpl(pIPWState, pixr, &Bpl));
    ERR_PASS( w_pixrGetImage(pIPWState, pixr, &pImage));
    srcData = (UInt32 *)pImage;

    for (i = 0; i < h; i++, srcData += Bpl/4)
    {
	/* initialize the word pointers.  Process from right to
	 * left in hopes of encouraging the compiler to generate
	 * slightly better code.  Also use a byte index so the compiler
	 * doesn't mess around with shifting it.  Finally, process
	 * long words so we reduce the number of writes by a factor
	 * of 4. */

	/*
	  If there are any image bits at all in the last 
	  word, do the inversion by xor'ing with lastWordMask.
	  This will prevent any alterations to the frame bits
	*/
	if (extraBits != 0)
	    srcData[wInWords] = srcData[wInWords] ^ lastWordMask;

	/* Test for an image of less than 32 bits */
	if (wInWords != 0)
	{
	    /* Do all the whole words */
	    for (pSrc=&srcData[wInWords-1]; pSrc>=srcData; pSrc--)
		*pSrc = ~ (*pSrc);
	}
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Center a PIXR within another PIXR.
 *
 * LongDescription:::
 *   Centers the source PIXR within the destination.  Destination
 *   pixels falling outside the new source are not affected.
 *   Works for gray or binary images, Assumes both source
 *   and dest have same depth.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  spr::            Ptr to source PIXR
 *
 * Output:::
 *  dpr::            Ptr to destination PIXR
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_center (
    IPWState           *pIPWState,
    PIXR               *spr,
    PIXR               *dpr)
{
    PROLOGUE;
    Int32              dw, dh, sw, sh, dd, sd;
    Int32              shiftX, shiftY;

    ERR_GEN_PIXR(spr, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);
    ERR_GEN_PIXR(dpr, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannels1, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, dpr, &dw));
    ERR_PASS( w_pixrGetWidth(pIPWState, spr, &sw));
    ERR_PASS( w_pixrGetHeight(pIPWState, dpr, &dh));
    ERR_PASS( w_pixrGetHeight(pIPWState, spr, &sh));
    ERR_PASS( w_pixrGetDepth(pIPWState, dpr, &dd));
    ERR_PASS( w_pixrGetDepth(pIPWState, spr, &sd));

    ERR_GEN(sd != dd, cErrInvalidParm);

    shiftX = ABS((dw - sw) / 2);
    shiftY = ABS((dh - sh) / 2);

	/* There are four cases here.  The source may be wider or not as
	 * wide as the dest; the source may be taller or not as tall as
	 * the dest.  Each case requires a slightly different movement
	 * as either slices of source are discarded or slices of dest are
	 * unaffected.
	 */
    if ((dw <= sw) && (dh <= sh))
	    /* the source must be trimmed on all sides */
	ERR_PASS( w_rasterop (pIPWState,
			      shiftX, shiftY,
			      spr,
			      PIX_SRC, 
			      0, 0, dw, dh,
			      dpr));

    else if (dw <= sw) /* so dh > sh */
	    /* the source must be trimmed on left and right, but bits on
	     * the top and bottom of the dest are unaffected */
	ERR_PASS( w_rasterop (pIPWState,
			      shiftX, 0,
			      spr,
			      PIX_SRC,
			      0, shiftY, dw, sh, 
			      dpr));

    else /* dw > sw */
    if ((dw > sw) && (dh <= sh))
	    /* the source must be trimmed on top and bottom, but bits on
	     * the left and right of the dest are unaffected */
	ERR_PASS( w_rasterop (pIPWState,
			      0, shiftY,
			      spr,
			      PIX_SRC,
			      shiftX, 0, sw, dh,
			      dpr));

    else /* dw > sw and dh > sh */
	    /* the source fits entirely inside the dest */
	ERR_PASS( w_rasterop (pIPWState, 
			      0, 0,
			      spr,
			      PIX_SRC, 
			      shiftX, shiftY, sw, sh, 
			      dpr));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a PIXR from a subrect of another PIXR.
 *
 * LongDescription:::
 *   Returns the rect part of the input PIXR.  The output
 *   PIXR is framed, regardless of whether or not the input PIXR was framed.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prs::            Ptr to source PIXR
 *  rect::           Ptr to rectangle.  If NULL, make PIXR identical
 *                   to source PIXR.  If rect is not valid, it clips the rect
 *                   to the PIXR; else, it uses the entire rect.
 *                   If entire rect is out of bounds, dest PIXR will have
 *                   0 width and height.
 *
 * Output:::
 *  prd::            Ptr to ptr to created dest PIXR.
 *                   prd is allocated and filled in.
 *                   *prd must point to NULL on input.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, cropping, multi-channel
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  bloomberg::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_rectOf (
    IPWState           *pIPWState,
    PIXR               *prs,
    RECTANGL           *rect,
    PIXR              **prd)
{
    PROLOGUE;
    Int32		 d, channels;
    ColorType		 type;
    ColorMap		*map = NULL;
    RECTANGL		*rectn = NULL;
    ScaledInt1616        xdpi, ydpi;
    Int8		 valid;
    PIXR		*pixr = NULL;

    ERR_GEN(prd == NULL, cErrNullPtr);    
    ERR_GEN(*prd != NULL, cErrPtrNotNull);    
    ERR_GEN_PIXR(prs, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    if (rect == NULL)
    { 
	/* return a copy of the input pixr */
	ERR_PASS( w_copy(pIPWState, prs, &pixr));
    }
    else 
    {
	ERR_PASS( w_isRectValid(pIPWState, prs, rect, &valid));
    
	if (!valid)
	    /*  rect exists, but extends beyond boundaries of pixr */
	    ERR_PASS( w_clipRectToPixr(pIPWState, prs, rect, &rectn));
	else
	    ERR_PASS( w_copyRect(pIPWState, rect, &rectn));

        /*  take the rect part of the Pixr */
	ERR_PASS_CLEANUP( w_pixrGetDepth(pIPWState, prs, &d));
	IP_YIELD;
    
	ERR_PASS_CLEANUP( w_pixrGetChannelCount (pIPWState, prs, &channels));
	ERR_PASS_CLEANUP( w_pixrGetType (pIPWState, prs, &type));
	ERR_PASS_CLEANUP( w_pixrGetColorMap (pIPWState, prs, &map));
	ERR_PASS_CLEANUP( w_pixrGetResolutionDPI (pIPWState, prs,
						  &xdpi, &ydpi));


	ERR_PASS_CLEANUP( w_createPixrExtendedNoInit (pIPWState,
						      rectn->w, rectn->h,
						      d, xdpi, ydpi, cFrameBits,
						      channels, type, map,
						      &pixr));
	IP_YIELD;

	if (rectn->w > 0  &&  rectn->h > 0) 
	{
	    ERR_PASS_CLEANUP( w_rasterop (pIPWState, 
					  rectn->x, rectn->y,
					  prs,
					  PIX_SRC, 
					  0, 0, rectn->w, rectn->h,
					  pixr));
	}
	/* else we made a rect and a PIXR of 0 width and height above! */
	
	ERR_PASS_CLEANUP( w_destroyRect(pIPWState, &rectn));
    }

    *prd = pixr;
    NORMAL_RETURN;

ERR_CLEANUP:
    w_destroyRect(pIPWState, &rectn);
    w_destroyPixr(pIPWState, &pixr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Paint a rectangular region of a PIXR.
 *
 * LongDescription:::
 *	Paint a color into a rectangular region of a PIXR. We'll use
 *	a fairly stupid method: fill a line and replicate it with
 *	raster operations if the image isn't binary. If the image is binary,
 *	we can just use one raster operation.
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prImage::        Ptr to PIXR into which to paint.
 *  rect::           Rectangular area to fill; if NULL, fill the whole PIXR.
 *  color::          Color value to fill with.
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_paint (
    IPWState           *pIPWState,
    PIXR               *prImage,
    RECTANGL           *rect,
    ColorRecord         color)
{
    PROLOGUE;
    Int32	 depth, op;
    Int32	 x, y, w, h;

    ERR_GEN_PIXR(prImage, cByteOrderBig, cDepth1 | cDepth2 | cDepth4 | cDepth8,
		 cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetDepth (pIPWState, prImage, &depth));

    if (rect == NULL)
    {
	x = 0; 
	y = 0;
	ERR_PASS( w_pixrGetWidth (pIPWState, prImage, &w));
	ERR_PASS( w_pixrGetHeight (pIPWState, prImage, &h));
    }
    else
    {
	x = rect->x;
	y = rect->y;
	w = rect->w;
	h = rect->h;
    }

    if (depth == 1)
    {
	/* We can just use a raster Op directly */
	op = (color[0]) ? PIX_SET : PIX_CLR;
	ERR_PASS( w_rasterop (pIPWState,
			      0, 0,
			      NULL,
			      op,
			      x, y, w, h,
			      prImage));
    }
    else
    {
	Int32	 channel, channels;
	Int32	 i, cVal, bpl, iw;
	UInt8	*img = NULL;
	PIXR    *ch;

	ERR_PASS( w_pixrGetChannelCount(pIPWState, prImage, &channels));
	ERR_PASS( w_pixrGetBpl (pIPWState, prImage, &bpl));

	for (channel = 0; channel < channels; channel++)
	{
	    ERR_PASS( w_pixrGetChannel (pIPWState, prImage, channel, &ch));
	    ERR_PASS( w_pixrGetImage (pIPWState, ch, &img));
	    img += y * bpl;
	    cVal = color[channel];
	    iw = x + w;

	    /* Very slow for the first line, especially 4 and 2 bit! */
	    if (depth == 8)
	    {
		for (i = x; i < iw; i++)
		{
		    SET_IMAGE_BYTE (img, i, (UInt8) cVal);
		}
	    }
	    else
	    {
		for (i = x; i < iw; i++)
		{
		    SET_IMAGE_BITS(img, i, cVal, depth);
		}
	    }
	}

        /* Copy line zero to all the other lines */

	ERR_PASS( w_rasterop (pIPWState, 
				  x, y,
				  prImage, 
				  PIX_SRC, 
				  x, y + 1, w, (h-1), 
				  prImage));
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the data portion of an image to 0.
 *
 * LongDescription:::
 *	Blindly clear an image.  This is surprisingly useful..
 *
 * Input:::
 *  pIPWState::      Ptr to IP struct for memory alloc/freeing routines
 *  prS::  	      Ptr to PIXR into which to paint.
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, multi-channel, color
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ddavies
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_clear (
    IPWState           *pIPWState,
    PIXR               *prS)
{
    PROLOGUE;
    Int32	 w, h;

    ERR_GEN_PIXR(prS, cByteOrderBig, cDepthAny, cChannelsAny, cFrame32);

    ERR_PASS( w_pixrGetWidth(pIPWState, prS, &w));
    ERR_PASS( w_pixrGetHeight(pIPWState, prS, &h));

	/* Just do it */
   ERR_PASS( w_rasterop (pIPWState,
			  0, 0,
			  NULL,
			  PIX_CLR, 
			  0, 0, w, h,
			  prS));

    NORMAL_RETURN;
}



/*

 $Log: props6.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.9  1998/01/14 00:14:32  ddavies
 props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v

 CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

 Revision 1.8  1997/11/13 17:36:21  xzhu
 xzhu on Thu Nov 13 09:35:10 PST 1997

 props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v

 Implemented Dan's suggestion, reduced more calling w_rasterop in w_paint.

 Revision 1.7  1997/11/13 00:58:44  xzhu
 xzhu on Wed Nov 12 16:57:23 PST 1997

 props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v

 Speeded up w_paint by reducing calling w_rasterop.

 Revision 1.6  1997/06/26 20:04:28  ddavies
 props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.5  1997/04/17 02:40:50  ddavies
 ddavies on Wed Apr 16 19:39:40 PDT 1997

 props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v

 Fix name of input arg for autodoc comment (prImage -> prS in w_clear).

 Revision 1.4  1997/04/15 22:27:47  ddavies
 props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v

 Remove procedure names from autodoc comments to be consistent.
 Use non-initialized PIXR creation where possible.
 Add new procedure, w_clear.  This proc sets the image portion of a bitmap
 to 0 using w_rasterOp.  It's a convenience.

 Revision 1.3  1995/12/13 04:11:48  ejaquith
 props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v

 CommonCheck: Added ";" after macro IP_RCSINFO()

 * Revision 1.2  1995/11/13  20:14:51  ejaquith
 * props6.c -> /product/ipcore/ipwrap/src/RCS/props6.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.1  1995/10/28  01:42:46  ddavies
 * Initial revision
 *

 */

