/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*
 *   rectops2.c:
 *
 *	Raster operations for rectangle lists: 
 *		w_pixrPutRectListBorder()
 *		w_pixrPutRectListWithNumber()
 *
 *	Basic "raster an integer" routine:
 *		w_pixrDrawNum()
 */

#include "ipwcore.prv"
#include "geomadt.h"
#include "pixr.h"
#include "frames.pub"
#include "defines.pub"

/* prototypes */
#include "props.pub"
#include "rast.pub"
#include "rectops.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: rectops2.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:19 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw0(
    IPWState *pIPWState,
    PIXR     *pixr,
    Int32     x,
    Int32     y)
{
    PROLOGUE;

    ERR_PASS( w_pixrPutLine( pIPWState, pixr, x,   y, x,   y+4, 1));
    ERR_PASS( w_pixrPutLine( pIPWState, pixr, x+2, y, x+2, y+4, 1));

    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+1, y, 1));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+1, y+4, 1));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw8(
    IPWState *pIPWState,
    PIXR     *pixr,
    Int32     x,
    Int32     y)
{
    PROLOGUE;
    ERR_PASS( wp_pixrDraw0( pIPWState, pixr, x, y));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+1, y+2, 1));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw6(
    IPWState *pIPWState,
    PIXR     *pixr,
    Int32     x,
    Int32     y)
{
    PROLOGUE;
    ERR_PASS( wp_pixrDraw8( pIPWState, pixr, x, y));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+2, y+1, 0));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw5(
    IPWState *pIPWState,
    PIXR     *pixr,
    Int32     x,
    Int32     y)
{
    PROLOGUE;
    ERR_PASS( wp_pixrDraw6( pIPWState, pixr, x, y));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x, y+3, 0));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw3(
    IPWState   *pIPWState,
    PIXR       *pixr,
    Int32       x,
    Int32       y)
{
    PROLOGUE;
    ERR_PASS( wp_pixrDraw8( pIPWState, pixr, x, y));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x, y+1, 0));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x, y+3, 0));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw2(
    IPWState    *pIPWState,
    PIXR        *pixr,
    Int32        x,
    Int32        y)
{
    PROLOGUE;
    ERR_PASS( wp_pixrDraw3( pIPWState, pixr, x, y));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x,   y+3, 1));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+2, y+3, 0));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw9(
    IPWState    *pIPWState,
    PIXR        *pixr,
    Int32        x,
    Int32        y)
{
    PROLOGUE;
    ERR_PASS( wp_pixrDraw8( pIPWState, pixr, x, y));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x,   y+3, 0));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x,   y+4, 0));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+1, y+4, 0));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw4(
    IPWState    *pIPWState,
    PIXR        *pixr,
    Int32        x,
    Int32        y)
{
    PROLOGUE;
    ERR_PASS( wp_pixrDraw9( pIPWState, pixr, x, y));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+1, y, 0));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw7(
    IPWState    *pIPWState,
    PIXR        *pixr,
    Int32        x,
    Int32        y)
{
    PROLOGUE;

    ERR_PASS( w_pixrPutLine( pIPWState, pixr, x+2, y, x+2, y+4, 1));
    ERR_PASS( w_pixrPutLine( pIPWState, pixr, x,   y, x+2, y,   1));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Draw this digit in binary PIXR -- see w_pixrDrawNum() for details
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  pixr::              Ptr to binary PIXR to draw in
 *  x::                 Ptr to upper left corner X position to draw digit at
 *  y::                 Ptr to upper left corner Y position to draw digit at
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_pixrDraw1(
    IPWState    *pIPWState,
    PIXR        *pixr,
    Int32        x,
    Int32        y)
{
    PROLOGUE;
    
    ERR_PASS( w_pixrPutLine( pIPWState, pixr, x+1, y, x+1, y+4, 1));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x,   y+4, 1));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x+2, y+4, 1));
    ERR_PASS( w_pixrPutPixel( pIPWState, pixr, x,   y+1, 1));

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Writes a number into a binary PIXR, using 3 x 5 pixels for each digit
 *
 * LongDescription:::
 *  A crude way of writing numerical information into an image.
 *  Not much real emphasis on speed.  Writes in "black" pixels.
 *
 * <PRE>
 *  Bitmaps for digits:
 *   xxx   x   xxx  xxx  x x  xxx  xxx  xxx  xxx  xxx
 *   x x  xx     x    x  x x  x    x      x  x x  x x
 *   x x   x   xxx  xxx  xxx  xxx  xxx    x  xxx  xxx
 *   x x   x   x      x    x    x  x x    x  x x    x
 *   xxx  xxx  xxx  xxx    x  xxx  xxx    x  xxx  xxx
 *
 * </PRE>
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  x::                 Upper left corner X location in PIXR to begin
 *                      rendering the number
 *  y::                 Upper left corner Y location in PIXR to begin
 *                      rendering the number
 *  num::               Nonnegative number to render into the PIXR.
 *
 * Output:::
 *  pixr::              Ptr to binary PIXR to write number into.
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, debug
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrDrawNum (
    IPWState    *pIPWState,
    Int32        x,
    Int32        y,
    Int32        num,
    PIXR        *pixr)
{
    PROLOGUE;
    Int32  numDigits, index;
    
    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepth1, cChannels1, cFrame32);
    ERR_GEN(num < 0, cErrInvalidParm);

    for (numDigits = 1, index = num; (index = index / 10) > 0; numDigits++)
    ;

    x += 4 * (numDigits-1);
    
    for ( ; numDigits > 0; num /= 10, x -=4, numDigits--)
    {
	switch (num % 10)
	{
	case 0: ERR_PASS( wp_pixrDraw0( pIPWState, pixr, x, y));break;
	case 1: ERR_PASS( wp_pixrDraw1( pIPWState, pixr, x, y));break;
	case 2: ERR_PASS( wp_pixrDraw2( pIPWState, pixr, x, y));break;
	case 3: ERR_PASS( wp_pixrDraw3( pIPWState, pixr, x, y));break;
	case 4: ERR_PASS( wp_pixrDraw4( pIPWState, pixr, x, y));break;
	case 5: ERR_PASS( wp_pixrDraw5( pIPWState, pixr, x, y));break;
	case 6: ERR_PASS( wp_pixrDraw6( pIPWState, pixr, x, y));break;
	case 7: ERR_PASS( wp_pixrDraw7( pIPWState, pixr, x, y));break;
	case 8: ERR_PASS( wp_pixrDraw8( pIPWState, pixr, x, y));break;
	case 9: ERR_PASS( wp_pixrDraw9( pIPWState, pixr, x, y));break;
	}
    }

    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Writes each rectangle in list onto a PIXR, with it's list order # in center
 *
 * LongDescription:::
 *  Similar to w_pixrPutRectListBorder().
 *  The rectangles are drawn so that their line segment thickness moves
 *  to the "interior" of the rect.  That is, if a rect is width x height
 *  of 50 x 60, and linethickness is 5, the rect drawn will be a 40 x 50
 *  white rect centered over a 50 x 60 black rect.<BR><BR>
 *
 *  Additionally, the first rect in the list is considered rect "0", so
 *  "0" is written into the center of the drawn rect, the second rect
 *  will get "1" written into its center, etc...<BR><BR>
 *
 *  It uses w_pixrDrawNum() to "write" these index numbers.
 *
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rectsList::         Ptr to list of rectangles
 *  inHimetrics::       cTrue if rectangls are in units of Himetrics.
 *                      cFalse if they are in units of pixels.
 *  linewidth::         Number of pixels to draw rectangle outline thickness
 *  rectsXRes::         Positive resolution along X axis rectangles are in
 *  rectsYRes::         Positive resolution along Y axis rectangles are in
 *  imageWidth::        Positive width value to create PIXR with
 *  imageHeight::       Positive height value to create PIXR with
 *  imageXRes::         Positive resolution along X axis PIXR is in, or is
 *                      to be created with
 *  imageYRes::         Positive resolution along Y axis PIXR is in, or is
 *                      to be created with
 *
 * Output:::
 *  ppPr::              If NULL, create a binary image of the above
 *                         desired specifications.
 *                      Else, ptr to binary PIXR on input.
 *                      On return, the rectangles have been written into it.
 *
 * Returns:::
 *
 * Notes:::
 *  scaling::
 *    If rectangles and PIXR are in different resolutions, will rescale
 *    rectangles accordingly before "drawing".
 *
 *  clipping::
 *    If rect is not valid, it clips the rect to the Pixr.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrPutRectListWithNumber (
    IPWState   *pIPWState,
    RECTANGL   *rectsList,
    Bool        inHimetrics,
    Int32       linewidth,
    Int32       rectsXRes,
    Int32       rectsYRes,
    Int32       imageWidth,
    Int32       imageHeight,
    Int32       imageXRes,
    Int32       imageYRes,
    PIXR      **ppPr)
{
    PROLOGUE;
    Int32      cnt;
    PIXR      *pr = NULL;
    RECTANGL  *pRect, *pRect2 = NULL;
    Float32    scaleX, scaleY;

    ERR_GEN(rectsList == NULL  ||  ppPr == NULL, cErrNullPtr);
    ERR_GEN(linewidth   <= 0  ||
	    rectsXRes   <= 0  ||
	    rectsYRes   <= 0  ||
	    imageWidth  <= 0  ||
	    imageHeight <= 0  ||
	    imageXRes   <= 0  ||
	    imageYRes   <= 0, cErrInvalidParm);
    if (*ppPr != NULL)
    {
	ERR_GEN_PIXR( *ppPr, cByteOrderBig, cDepth1, cChannels1, cFrame32 );
	pr = *ppPr;
    }
    else 
    {
	/* Make an image of the user's specified dimens */

	ERR_PASS (w_createPixr (pIPWState, 
				imageWidth, imageHeight, 1, 
				Int32_TO_ScaledInt1616(imageXRes),
				Int32_TO_ScaledInt1616(imageYRes),
				cFrameBits, &pr));
    }
    

    scaleX = ((Float32) imageXRes) / rectsXRes;
    scaleY = ((Float32) imageYRes) / rectsYRes;
    
    for (cnt = 0, pRect = rectsList; pRect != NULL; cnt++, pRect = pRect->next)
    {
	if (inHimetrics)
	    ERR_PASS_CLEANUP (w_rectHimetricsToPixels (pIPWState, 
						       pRect, 
						       imageXRes, imageYRes,
						       &pRect2));
	else if (scaleX != 1.  ||  scaleY != 1.)
	    ERR_PASS_CLEANUP (w_xformRect (pIPWState,
					   pRect, 0, 0, scaleX, scaleY,
					   &pRect2));
	else
	    pRect2 = pRect;
	
	ERR_PASS_CLEANUP (w_pixrPutRect (pIPWState, pr, pRect2, linewidth, 1));
	ERR_PASS_CLEANUP (w_pixrDrawNum (pIPWState, 
					 pRect2->x + MAX(pRect2->w,1)/2,
					 pRect2->y + MAX(pRect2->h,1)/2, cnt,
					 pr));
	
	if (pRect2 != pRect)
	    ERR_PASS_CLEANUP (w_destroyRect (pIPWState, &pRect2));
    }


    *ppPr = pr;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (pRect2 != pRect)
	w_destroyRect (pIPWState, &pRect2);
    if (*ppPr == NULL)
	w_destroyPixr (pIPWState, &pr);
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Writes rect outline for each rectangle in list onto a PIXR.
 *
 * LongDescription:::
 *  Here, the method is, for each rectangle,
 *  to write an outline of "linewidth" thickness black pixels
 *  so that a white block, dimensions rectangle width by rectangle height,
 *  is in the center of this outline.
 *
 * Input:::
 *  pIPWState::         IP struct for memory alloc/freeing routines
 *  rectlist::          Ptr to list of rectangles
 *  linewidth::         Number of pixels to draw rectangle outline thickness
 *
 * Output:::
 *  pixr::              Ptr to single-channel PIXR to write rectangles to.
 *
 * Returns:::
 *
 * Notes:::
 *  clipping::
 *    If rect is not valid, it clips the rect to the Pixr.
 *
 * FunctionCategory:::
 *  rasterOp
 *
 * Keywords:::
 *  rasterOp, geometry
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  ejaquith::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_pixrPutRectListBorder (
    IPWState   *pIPWState,
    RECTANGL   *rectlist,
    Int32       linewidth,
    PIXR       *pixr)
{
    PROLOGUE;
    RECTANGL *crect, trect;

    ERR_GEN_PIXR(pixr, cByteOrderBig, cDepthAny, cChannels1, cFrame32);
    ERR_GEN(rectlist == NULL, cErrNullPtr);

    for (crect = rectlist; crect != NULL; crect = crect->next)
    {
	trect.x = crect->x - linewidth/2;
	trect.y = crect->y - linewidth/2;
	trect.w = crect->w + (linewidth/2)*2;
	trect.h = crect->h + (linewidth/2)*2;
	ERR_PASS( w_pixrPutRect( pIPWState, pixr, &trect, linewidth, 1));
    }

    NORMAL_RETURN;
}



/*
 $Log: rectops2.c,v $
 Revision 1.1.1.1  1999/09/20 18:19:19  jlin
 Imported sources

 Revision 1.3  1997/06/26 20:04:33  ddavies
 rectops2.c -> /product/ipcore/ipwrap/src/RCS/rectops2.c,v

 CommonCheck: Add inclusion of new file rast.pub.

 Revision 1.2  1996/02/22 01:38:20  ejaquith
 rectops2.c -> /product/ipcore/ipwrap/src/RCS/rectops2.c,v

 Removed unused param

 * Revision 1.1  1996/01/27  22:33:50  ejaquith
 * Initial revision
 *
 */
