/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  strel1.c--includes subroutines
 *                   w_strel1Inits ()
 *                   w_strel1Term ()
 *
 *                  ...simple linear elements...
 *               strel2h, strel2v, strel3h, strel3v,
 *               strel4h, strel4v, strel5h, strel5v,
 *               strel6h, strel6v, strel7h, strel7v,
 *               strel8h, strel8v, strel9h, strel9v,
 *               strel10h, strel10v, strel11h, strel11v,
 *		 strel15h, strel15v, strel19h, strel19v,
 *               strel20h, strel20v, strel25h, strel25v,
 *               strel30h, strel30v, strel35h, strel35v,
 *               strel40h, strel40v, strel45h, strel45v,
 *               strel50h, strel50v;
 *
 *                  ...use for solid brick matches...
 *               strel1, strel2, strel3, strel4, strel5;
 *
 *                  ...use for 8-connected B.B. fills, etc...
 *               strel2dp, strel2dm;
 */



#include "ipwcore.prv"

#include "defines.pub"
#include "pixr.h"
#include "morphg.h"

#include "strel.pub"


#undef CHECKERR
#define CHECKERR(statement) \
if ((statement)) return abortI("No memory for strel", procName, cErrNomem);

IP_RCSINFO(RCSInfo, "$RCSfile: strel1.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initializes a basic set of commonly used structuring elements.
 *
 * LongDescription:::
 *  These STRELs are created and stored in the strel manager instance
 *  associated with the client state structure passed in and may
 *  be destroyed by calling w_strel1Term ():
 *
 *    strel2h, strel2v, strel3h, strel3v, strel4h, strel4v, strel5h, 
 *    strel5v, strel6h, strel6v, strel7h, strel7v, strel8h, strel8v, 
 *    strel9h, strel9v, strel10h, strel10v, strel11h, strel11v, strel15h, 
 *    strel15v, strel19h, strel19v, strel20h, strel20v, strel25h, strel25v, 
 *    strel30h, strel30v, strel35h, strel35v, strel40h, strel40v, strel45h, 
 *    strel45v, strel50h, strel50v, strel1, strel2, strel3, strel4, strel5,
 *    strel2dp, strel2dm
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel management, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strel1Inits (
    IPWState           *pIPWState)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pIPWState->strelInfo == NULL, cErrStrel_MgrNotInitted);

    if (pIPWState->strelInfo->strel1InitsDone)
	NORMAL_RETURN;

    /* horizontal strel of length 2: (1c 1) */
    ERR_PASS( w_initStrelByName (pIPWState, strel2h));

    /* vertical strel of length 2: v(1c 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel2v));

    /* horizontal strel of length 3: (1 1c 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel3h));

    /* vertical strel of length 3: v(1 1c 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel3v));

    /* horizontal strel of length 4: (1 1c 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel4h));

    /* vertical strel of length 4: v(1 1c 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel4v));

    /* horizontal strel of length 5: (1 1 1c 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel5h));

    /* vertical strel of length 5: v(1 1 1c 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel5v));

    /* horizontal strel of length 6: (1 1 1c 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel6h));

    /* vertical strel of length 6: v(1 1 1c 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel6v));

    /* horizontal strel of length 7: (1 1 1 1c 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel7h));

    /* vertical strel of length 7: v(1 1 1 1c 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel7v));

    /* horizontal strel of length 8: (1 1 1 1c 1 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel8h));

    /* vertical strel of length 8: v(1 1 1 1c 1 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel8v));

    /* horizontal strel of length 9: (1 1 1 1 1c 1 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel9h));

    /* vertical strel of length 9: v(1 1 1 1 1c 1 1 1 1) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel9v));

    /* horizontal strel of length 10: ((4:1) 1c (5:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel10h));

    /* vertical strel of length 10: v((4:1) 1c (5:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel10v));

    /* horizontal strel of length 11: ((5:1) 1c (5:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel11h));

    /* vertical strel of length 11: v((5:1) 1c (5:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel11v));

    /* horizontal strel of length 15: ((7:1) 1c (7:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel15h));

    /* vertical strel of length 15: v((7:1) 1c (7:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel15v));

    /* horizontal strel of length 19: ((9:1) 1c (9:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel19h));

    /* vertical strel of length 19: v((10:1) 1c (8:1)) */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel19v));

    /* horizontal strel of length 20: ((9:1) 1c (10:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel20h));

    /* vertical strel of length 20: v((9:1) 1c (10:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel20v));

    /* horizontal strel of length 25: ((12:1) 1c (12:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel25h));

    /* vertical strel of length 25: v((12:1) 1c (12:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel25v));

    /* horizontal strel of length 30: ((14:1) 1c (15:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel30h));

    /* vertical strel of length 30: v((14:1) 1c (15:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel30v));

    /* horizontal strel of length 35: ((17:1) 1c (17:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel35h));

    /* vertical strel of length 35: v((17:1) 1c (17:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel35v));

    /* horizontal strel of length 40: ((19:1) 1c (20:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel40h));

    /* vertical strel of length 40: v((19:1) 1c (20:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel40v));

    /* horizontal strel of length 45: ((22:1) 1c (22:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel45h));

    /* vertical strel of length 45: v((22:1) 1c (22:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel45v));

    /* horizontal strel of length 50: ((24:1) 1c (25:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel50h));

    /* vertical strel of length 50: v((24:1) 1c (25:1))  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel50v));

    /* pattern:  1c  */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel1));

    /* pattern:  1c  1 
		 1   1      */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel2));

    /* pattern:  1   1   1 
		 1   1c  1
		 1   1   1   */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel3));

    /* pattern:  1   1   1   1 
		 1   1c  1   1
		 1   1   1   1
		 1   1   1   1    */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel4));

    /* pattern:  1   1   1   1   1
		 1   1   1   1   1
		 1   1   1c  1   1
		 1   1   1   1   1
		 1   1   1   1   1     */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel5));

    /* pattern:  0c  1
		 1   0      */ 
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel2dp));

    /* pattern:  1c  0
		 0   1      */ 
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, strel2dm));


    pIPWState->strelInfo->strel1InitsDone = cTrue;
    NORMAL_RETURN;

ERR_CLEANUP:
    /* we purposefully ignore the return value here since
     * 1) it will mask the original error condition which brought us here, and
     * 2) strel1Term () will return an error since it can't de-init all the
     * strels (at least one of them has not been initted if we got to the
     * error condition).
     */
    w_strel1Term (pIPWState);
    pIPWState->strelInfo->strel1InitsDone = cFalse;
    ERR_RETURN;

}  /* strel1Inits () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroys the set structuring elements created by w_strel1Inits ()
 *
 * LongDescription:::
 *  If the strels are no longer used by any other instances of the strel
 *  manager, they are freed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel management, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strel1Term (
    IPWState           *pIPWState)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pIPWState->strelInfo == NULL, cErrStrel_MgrNotInitted);

    /* Note that we don't check to see if the strel1InitsDone
     * flag is set as a precondition as this function may be
     * called as a cleanup operation if an error occurred in
     * strel1Init () in which case some strels may have been
     * created, but the flag has not been set.
     */

    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel2h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel2v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel3h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel3v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel4h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel4v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel5h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel5v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel6h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel6v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel7h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel7v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel8h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel8v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel9h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel9v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel10h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel10v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel11h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel11v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel15h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel15v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel19h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel19v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel20h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel20v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel25h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel25v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel30h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel30v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel35h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel35v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel40h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel40v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel45h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel45v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel50h));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel50v));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel1));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel2));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel3));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel4));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel5));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel2dp));
    ERR_PASS_CLEANUP( w_releaseStrelByName(pIPWState, NULL, strel2dm));

    /* Mark this group of strels as uninitialized */
    pIPWState->strelInfo->strel1InitsDone = cFalse;
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;

}  /* strel1Term () */


/*
 *
 * $Log: strel1.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.21  1996/01/04 20:26:59  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.20  1995/12/13  04:13:21  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.19  1995/11/13  20:16:17  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.18  1995/10/13  22:05:25  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.17  1995/09/29  21:10:38  dferrell
 * dferrell on Fri Sep 29 14:10:18 PDT 1995
 *
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.16  1995/09/21  22:59:50  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.15  1995/09/07  22:08:39  jcraig
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * strel11Term => w_strel11Term
 *
 * Revision 1.14  1995/09/06  20:23:36  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.13  1995/09/01  18:49:48  dferrell
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.12  1995/08/31  21:39:21  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.11  1995/08/30  17:27:06  ejaquith
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Made routines accessing pIPWState first check that it is valid (not NULL).
 * If it's null, return error: cErrNullIPWState.
 *
 * Revision 1.10  1995/08/22  23:46:15  dferrell
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Added new parameter to all calls to releaseStrelByName ().
 *
 * Revision 1.9  1995/08/17  21:56:19  dferrell
 * dferrell on Thu Aug 17 14:56:13 PDT 1995
 *
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.8  1995/01/24  22:08:00  lperry
 * lperry on Tue Jan 24 14:05:50 PST 1995
 *
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Incorporate changes to use new strel manager interface.
 * Release function also changes.
 * "Global" strel variables (strel2h, strel5, etc)
 * are now members of an enumerated type.
 *
 * Revision 1.7  1994/09/30  17:08:22  lperry
 * lperry on Fri Sep 30 10:08:02 PDT 1994
 *
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Add include of utils.pub to get _abortX defs.
 *
 * Revision 1.6  1994/09/29  21:18:28  lperry
 * lperry on Thu Sep 29 14:18:03 PDT 1994
 *
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Change to use strel.pub instead of strlpros.h
 *
 * Revision 1.5  1994/08/10  22:03:57  lperry
 * lperry on Wed Aug 10 15:02:29 PDT 1994
 *
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Change to use the new strel manager functions to
 * allocate and track strels. There will only be a
 * single copy of each unique strel maintained, even
 * if multiple init calls are made.
 *
 * Revision 1.4  1993/12/03  01:19:10  ddavies
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.3  1993/11/06  03:17:19  ddavies
 * strel1.c -> /product/ipcore/ipwrap/src/RCS/strel1.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *
 */

