/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/*
 *  strel11.c--includes subroutine
 *                   w_strel11Inits ()
 *                   w_strel11Term ()
 *                   w_createVRuleStrel ()
 *                   w_createHRuleStrel ()
 *
 *                ... pepper...
 *                    s_pepper_1
 *                    s_pepper_2
 *                    s_pepper_3
 *                    s_pepper_4
 *
 *                ... salt ...
 *                    s_salt_1
 *                    s_salt_2
 *                    s_salt_3
 *
 *                ... replicative expansion ...
 *                    s_repl3
 *                    s_repl4
 *
 *		  ... detection of rules ...
 *		      s_rulew9h7    (75dpi)
 *		      s_ruleh9w7 
 */



#include "ipwcore.prv"

#include "defines.pub"
#include "pixr.h"
#include "morphg.h"

#include "strel.pub"


IP_RCSINFO(RCSInfo, "$RCSfile: strel11.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

#undef CHECKERR
#define CHECKERR(statement) \
if ((statement)) return abortI("No memory for strel", procName, cErrNomem);



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initializes an set of commonly used image cleanup structuring elements.
 *
 * LongDescription:::
 *  These STRELs are created and stored in the strel manager instance
 *  associated with the client state structure passed in and may
 *  be destroyed by calling w_strel11Term ():
 *
 *    s_pepper_1, s_pepper_2, s_pepper_3, s_pepper_4, s_salt_1, s_salt_2,
 *    s_salt_3, s_repl3, s_repl4, s_rulew9h7, s_ruleh9w7 
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel management, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strel11Inits (
    IPWState           *pIPWState)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pIPWState->strelInfo == NULL, cErrStrel_MgrNotInitted);

    if (pIPWState->strelInfo->strel11InitsDone)
	NORMAL_RETURN;


/**********************************************************************
 ******************          Pepper & Salt            *****************
 *********************************************************************/
    /* 1x1 pepper */
    ERR_PASS( w_initStrelByName (pIPWState, s_pepper_1));

    /* 2x2 pepper */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_pepper_2));

    /* 3x3 pepper */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_pepper_3));

    /* 4x4 pepper */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_pepper_4)); 

    /* 1x1 salt */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_salt_1)); 

    /* 2x2 salt */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_salt_2)); 

    /* 3x3 salt */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_salt_3)); 


/**********************************************************************
 *****************          Spread Replication        *****************
 *********************************************************************/
    /* Spread 3 */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_repl3)); 

    /* Spread 4 */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_repl4)); 


/*********************************************************************
 *****************          Rule Detection           *****************
 *********************************************************************/

    /*    s_rulew9h7:  matches horizontal rules of width at least 9
     *	         and height (thickness) up to 7.
     *		 Used for removal of rules at 75/inch
     *
     *    s_rulew9h7
     *    9 9 4 4 2
     *    2 0 0 0 2 0 0 0 2
     *    0 0 0 0 0 0 0 0 0
     *    0 0 0 0 0 0 0 0 0
     *    0 0 0 0 0 0 0 0 0
     *    1 0 1 0 1 0 1 0 1
     *    0 0 0 0 0 0 0 0 0
     *    0 0 0 0 0 0 0 0 0
     *    0 0 0 0 0 0 0 0 0
     *    2 0 0 0 2 0 0 0 2
     */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_rulew9h7)); 


    /*    s_ruleh9w7:  matches vertical rules of height at least 10
     *		 and width (thickness) up to 7.
     *		 Used for removal of rules at 75/inch
     *
     *    s_ruleh9w7
     *    9 9 4 4 2
     *    2 0 0 0 1 0 0 0 2
     *    0 0 0 0 0 0 0 0 0
     *    0 0 0 0 1 0 0 0 0
     *    0 0 0 0 0 0 0 0 0
     *    2 0 0 0 1 0 0 0 2
     *    0 0 0 0 0 0 0 0 0
     *    0 0 0 0 1 0 0 0 0
     *    0 0 0 0 0 0 0 0 0
     *    2 0 0 0 1 0 0 0 2
     */
    ERR_PASS_CLEANUP( w_initStrelByName (pIPWState, s_ruleh9w7)); 

    pIPWState->strelInfo->strel11InitsDone = cTrue;
    NORMAL_RETURN;


ERR_CLEANUP:
    /* we purposefully ignore the return value here since
     * 1) it will mask the original error condition which brought us here, and
     * 2) strel1Term () will return an error since it can't de-init all the
     * strels (at least one of them has not been initted if we got to the
     * error condition).
     */
    w_strel11Term (pIPWState);
    pIPWState->strelInfo->strel11InitsDone = cFalse;
    ERR_RETURN;

}  /* strel11Inits () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Destroys the set structuring elements created by w_strel11Inits ()
 *
 * LongDescription:::
 *  If the strels are no longer used by any other instances of the strel
 *  manager, they are freed.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *
 * Output:::
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel management, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  lperry
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_strel11Term (
    IPWState           *pIPWState)
{
    PROLOGUE;

    ERR_GEN (pIPWState == NULL, cErrNullIPWState);
    ERR_GEN (pIPWState->strelInfo == NULL, cErrStrel_MgrNotInitted);

    /* Note that we don't check to see if the strel11InitsDone
     * flag is set as a precondition as this function may be
     * called as a cleanup operation if an error occurred in
     * strel11Init () in which case some strels may have been
     * created, but the flag has not been set.
     */

    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_pepper_1));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_pepper_2));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_pepper_3));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_pepper_4));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_salt_1));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_salt_2));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_salt_3));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_repl3));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_repl4));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_rulew9h7));
    ERR_PASS_CLEANUP( w_releaseStrelByName (pIPWState, NULL, s_ruleh9w7));

    /* Mark this group of strels as uninitialized */
    pIPWState->strelInfo->strel11InitsDone = cFalse;
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;

}  /* strel11Term () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a vertical rule-detecting strel
 *
 * LongDescription:::
 *  The strel will be width+2 x height, with a column down the middle of 
 *  alternating hits and don't-cares, with misses along the edges (top, 
 *  middle, bottom).  The caller is responsible for keeping track of the 
 *  resulting strel--it is not stored in the strel manager.  Use 
 *  w_destroyStrel () to free it.  The result pointer should be set to 
 *  NULL before its address is passed in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  width::          Rule width to be removed
 *  height::         Height of strel
 *
 * Output:::
 *  strel::          The resulting strel
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createVRuleStrel (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    STREL             **strel)
{
    PROLOGUE;
    Int32 i;

    ERR_GEN (*strel != NULL, cErrInvalidParm);
    ERR_GEN (((width <= 0) || (height <= 0)), cErrInvalidParm);

    ERR_PASS( w_createStrel (pIPWState, width + 2, height, strel));

    (*strel)->cx = width / 2 + 1;
    (*strel)->cy = height / 2;
    for (i = 0; i < height; i++)
    {
	/* HIT (1) or DON'T CARE (0) depending on low bit of i */
	(*strel)->data[i][(*strel)->cx] = i & 1;
    }

    /* Scatter a few misses along the edge */
/* For now, take out the top and bottom misses */
/*    strel->data[0][0] = MISS; */
/*    strel->data[height - 1][0] = MISS; */

    (*strel)->data[(*strel)->cy][0] = MISS;

/*    strel->data[0][width-1] = MISS; */
/*    strel->data[height - 1][width-1] = MISS; */

    (*strel)->data[(*strel)->cy][width-1] = MISS;

    NORMAL_RETURN;

}  /* w_createVRuleStrel () */


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Make a horizontal rule-detecting strel
 *
 * LongDescription:::
 *  The strel will be width x height+2, with a row down the middle of 
 *  alternating hits and don't-cares, with misses along the edges.  The 
 *  caller is responsible for keeping track of the resulting strel--it is
 *  not stored in the strel manager.  Use w_destroyStrel () to free it.  
 *  The result pointer should be set to NULL before its address is passed 
 *  in.
 *
 * PrivateDescription:::
 *
 * Input:::
 *  pIPWState::      State pointer
 *  width::          Width of strel
 *  height::         Thickness of rules to be removed
 *
 * Output:::
 *  strel::          The resulting strel
 *
 * Returns:::
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  morphology
 *
 * Keywords:::
 *  strel, binary, morphology
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  macleod
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_createHRuleStrel (
    IPWState           *pIPWState,
    Int32               width,
    Int32               height,
    STREL             **strel)
{
    PROLOGUE;
    Int32 i;

    ERR_GEN (*strel != NULL, cErrInvalidParm);
    ERR_GEN (((width <= 0) || (height <= 0)), cErrInvalidParm);

    ERR_PASS( w_createStrel (pIPWState, width, height + 2, strel));

    (*strel)->cx = width / 2;
    (*strel)->cy = height / 2 + 1;
    for (i = 0; i < width; i++)
    {
	/* HIT (1) or DON'T CARE (0) depending on low bit of i */
	(*strel)->data[(*strel)->cy][i] = i & 1;
    }

    /* Scatter a few misses along the edge */
/* For now, take out the left and right misses */
/*    strel->data[0][0] = MISS; */
/*    strel->data[0][width-1] = MISS; */

    (*strel)->data[0][(*strel)->cx] = MISS;

/*    strel->data[height-1][0] = MISS; */
/*    strel->data[height-1][width-1] = MISS; */

    (*strel)->data[height-1][(*strel)->cx] = MISS;

    NORMAL_RETURN;

}  /* w_createHRuleStrel () */


/*
 *
 * $Log: strel11.c,v $
 * Revision 1.1.1.1  1999/09/20 18:19:20  jlin
 * Imported sources
 *
 * Revision 1.23  1996/01/04 20:27:02  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * CommonCheck: Removed all HTML tags that we had placed in our C code.
 * Removed double-quoted routine name after "BeginDocumentation" and
 * "EndDocumentation".
 *
 * Revision 1.22  1995/12/13  04:13:25  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * CommonCheck: Added ";" after macro IP_RCSINFO()
 *
 * Revision 1.21  1995/11/13  20:16:21  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * CommonCheck: Updated use of PROLOGUE macro to no longer take any arguments.
 * (Previously it took the string of the name of the routine).
 *
 * Revision 1.20  1995/10/13  22:05:28  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * CommonCheck: Some minor changes, like adding routine name after
 * BeginDocumentation tag, some keyword condensing.
 *
 * Revision 1.19  1995/09/29  21:10:53  dferrell
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Updated function comment blocks.
 *
 * Revision 1.18  1995/09/21  22:59:53  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * CommonCheck: Added standard function block comment at beginning of function.
 * Changed static routines to begin with "wp_" as needed.
 * Killed old routine declarations and special comment markers.
 *
 * Revision 1.17  1995/09/07  22:08:53  jcraig
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * strel12Term => w_strel12Term
 *
 * Revision 1.16  1995/09/06  20:23:39  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * CommonCheck: Used script to modify all IPWrap routines to begin with "w_".
 * Routines that were not of type "ErrCode CDECL" have not been changed.
 * IPWrap routines that use ERR_PASS() and ERR_PASS_CLEANUP() have been
 * modified to their new names.
 *
 * Revision 1.15  1995/09/01  18:49:53  dferrell
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Added necessary include files.
 *
 * Revision 1.14  1995/08/31  21:39:24  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * CommonCheck: Stripped all #includes.
 * Changed IPState to IPWState.
 *
 * Revision 1.13  1995/08/30  17:27:29  ejaquith
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Made routines accessing pIPWState first check that it is valid (not NULL).
 * If it's null, return error: cErrNullIPWState.
 *
 * Revision 1.12  1995/08/23  18:26:23  macleod
 * macleod on Wed Aug 23 11:25:24 PDT 1995
 *
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Checked the flag for "strel1InitsDone" instead of "strel11InitsDone",
 * causing strel11 inits to never happen if strel1 inits were done
 *
 * Revision 1.11  1995/08/22  23:46:27  dferrell
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Added new parameter to all calls to releaseStrelByName ().
 *
 * Revision 1.10  1995/08/17  21:56:25  dferrell
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * First pass of ipwrap transmogrification.
 *
 * Revision 1.9  1995/01/24  22:08:10  lperry
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Incorporate changes to use new strel manager interface.
 *
 * Revision 1.8  1994/09/30  17:08:34  lperry
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Add include of utils.pub to get _abortX defs.
 *
 * Revision 1.7  1994/09/29  21:18:42  lperry
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Change to use strel.pub instead of strlpros.h
 *
 * Revision 1.6  1994/08/10  22:04:31  lperry
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 *  Change to use the new strel manager functions to
 * allocate and track strels. There will only be a
 * single copy of each unique strel maintained, even
 * if multiple init calls are made.
 *
 * Revision 1.5  1994/01/05  22:07:25  macleod
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Added functions to compute larger rule-removing strels, which, stupidly,
 * I no longer use.
 *
 * Revision 1.4  1993/12/09  19:59:40  macleod
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Changed INT32 to Int32
 *
 * Revision 1.3  1993/12/03  01:19:14  ddavies
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Put RCS magic into macro that doesn't generate compiler warnings.
 *
 * Revision 1.2  1993/11/06  03:17:26  ddavies
 * strel11.c -> /product/ipcore/ipwrap/src/RCS/strel11.c,v
 *
 * Add RCS strings so file's name, revision and date appear in the object file.
 * Also automatically add a change log at the bottom of the file.
 *

 */

