/* $Header: /project/icons/cvsroot/ipcore/src/ipwrap/src/tif_appl.c,v 1.1.1.1 1999/09/20 18:19:20 jlin Exp $ */

/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994 Sam Leffler
 * Copyright (c) 1991, 1992, 1993, 1994 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
 * TIFF Library Macintosh-specific routines.
 *
 * These routines use only Toolbox and high-level File Manager traps.
 * They make no calls to the THINK C "unix" compatibility library.  Also,
 * malloc is not used directly but it is still referenced internally by
 * the ANSI library in rare cases.  Heap fragmentation by the malloc ring
 * buffer is therefore minimized.
 *
 * O_RDONLY and O_RDWR are treated identically here.  The tif_mode flag is
 * checked in TIFFWriteCheck().
 *
 * Create below fills in a blank creator signature and sets the file type
 * to 'TIFF'.  It is much better for the application to do this by Create'ing
 * the file first and TIFFOpen'ing it later.
 */

#include "tiffiop.h"
#include <Errors.h>
#include <Files.h>
#include <Memory.h>
#include "ipwcore.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif

#if 0
#include "memory.pub"   /* dferrell 1/96 */
#endif

/*  dferrell 1/26/96
#ifdef applec
*/
#ifdef macintosh
#define	CtoPstr	c2pstr
#endif

static tsize_t
_tiffReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return (FSRead((short) fd, (long*) &size, (char*) buf) == noErr ?
	    size : (tsize_t) -1);
}

static tsize_t
_tiffWriteProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return (FSWrite((short) fd, (long*) &size, (char*) buf) == noErr ?
	    size : (tsize_t) -1);
}

static toff_t
_tiffSeekProc(thandle_t fd, toff_t off, int whence)
{
	long fpos, size;

	if (GetEOF((short) fd, &size) != noErr)
		return EOF;
	(void) GetFPos((short) fd, &fpos);

	switch (whence) {
	case SEEK_CUR:
		if (off + fpos > size)
			SetEOF((short) fd, off + fpos);
		if (SetFPos((short) fd, fsFromMark, off) != noErr)
			return EOF;
		break;
	case SEEK_END:
		if (off > 0)
			SetEOF((short) fd, off + size);
		if (SetFPos((short) fd, fsFromStart, off + size) != noErr)
			return EOF;
		break;
	case SEEK_SET:
		if (off > size)
			SetEOF((short) fd, off);
		if (SetFPos((short) fd, fsFromStart, off) != noErr)
			return EOF;
		break;
	}

	return (toff_t)(GetFPos((short) fd, &fpos) == noErr ? fpos : EOF);
}

static int
_tiffMapProc(thandle_t fd, tdata_t* pbase, toff_t* psize)
{
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, tdata_t base, toff_t size)
{
}

static int
_tiffCloseProc(thandle_t fd)
{
	return (FSClose((short) fd));
}

static toff_t
_tiffSizeProc(TIFF* tif, thandle_t fd)
{
	long size;

	if (GetEOF((short) fd, &size) != noErr) {
		TIFFError(tif->tif_ipwstate, "_tiffSizeProc",
			  "%s: Cannot get file size");
		return (-1L);
	}
	return ((toff_t) size);
}

/*
 * Open a TIFF file descriptor for read/writing.
 */
TIFF*
TIFFFdOpen(void *pState, int fd, const char* name, const char* mode)
{
	TIFF* tif;

	tif = TIFFClientOpen(pState, name, mode, (thandle_t) fd,
	    _tiffReadProc, _tiffWriteProc, _tiffSeekProc, _tiffCloseProc,
	    _tiffSizeProc, _tiffMapProc, _tiffUnmapProc);
	if (tif)
		tif->tif_fd = fd;
	return (tif);
}

/*
 * Open a TIFF file for read/writing.
 */
TIFF*
TIFFOpen(void *pState, const char* name, const char* mode)
{
	static const char module[] = "TIFFOpen";
	Str255 pname;
	FInfo finfo;
	short fref;
	OSErr err;

	strcpy((char*) pname, name);
	CtoPstr((char*) pname);

	switch (_TIFFgetMode(pState, mode, module)) {
	default:
		return ((TIFF*) 0);
	case O_RDWR | O_CREAT | O_TRUNC:
		if (GetFInfo(pname, 0, &finfo) == noErr)
			FSDelete(pname, 0);
		/* fall through */
	case O_RDWR | O_CREAT:
		if ((err = GetFInfo(pname, 0, &finfo)) == fnfErr) {
			if (Create(pname, 0, '    ', 'TIFF') != noErr)
				goto badCreate;
			if (FSOpen(pname, 0, &fref) != noErr)
				goto badOpen;
		} else if (err == noErr) {
			if (FSOpen(pname, 0, &fref) != noErr)
				goto badOpen;
		} else
			goto badOpen;
		break;
	case O_RDONLY:
	case O_RDWR:
		if (FSOpen(pname, 0, &fref) != noErr)
			goto badOpen;
		break;
	}
	return (TIFFFdOpen(pState, (int) fref, name, mode));
badCreate:
	TIFFError(pState, module, "%s: Cannot create", name);
	return ((TIFF*) 0);
badOpen:
	TIFFError(pState, module, "%s: Cannot open", name);
	return ((TIFF*) 0);
}

void
_TIFFmemset(void* p, int v, size_t c)
{
	memset(p, v, c);
}

void
_TIFFmemcpy(void* d, const void* s, size_t c)
{
	memcpy(d, s, c);
}

int
_TIFFmemcmp(const void* p1, const void* p2, size_t c)
{
	return (memcmp(p1, p2, c));
}

void *
_TIFFmalloc(TIFF *tif, size_t s)
{
	/*return (NewPtr(s));*/
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    return (IPW_MALLOC(s, cAllocatorData));


 /*--ejaquith--oct94 malloc->MALLOC */ 
 /*--ejaquith--aug95 MALLOC->macMalloc */ 
 /*--dferrell--feb96 macMalloc->malloc for CW7 */
 /*--ddavies--jan98 malloc->IPW_MALLOC for consistency */
}

void
_TIFFfree(TIFF *tif, void *p)
{
	/* DisposePtr(p);*/
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    IPW_FREE(p, cAllocatorData);
}

void *
_TIFFrealloc(TIFF *tif, void* p, size_t s)
{
    /*Ptr n = p;

	SetPtrSize(p, s);
	if (MemError() && (n = NewPtr(s)) != NULL) {
		BlockMove(p, n, GetPtrSize(p));
		DisposePtr(p);
	}
	return (n);
     */
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    return (IPW_REALLOC(p, s, cAllocatorData));
}

static void
appleWarningHandler(void *pState, const char* module,
		    const char* fmt, va_list ap)
{
    char	errorString[512];
    char	outputString[512];
    IPWState   *pIPWState = (IPWState *)pState;

	/* DDavies, 11/18/96: Only print stuff is there's a printing routine */
    if ((pIPWState != NULL) && (pIPWState->errorProc != NULL))
    {
	    /* print using the IPWrap error handler.  That thing takes
	     * a string, so put the strings together.
	     * First, assemble "Warning, " and the module name.
	     */
	errorString[0] = '\0';		/* null strings so far */
	outputString[0] = '\0';
	if (module != NULL)
	{
	    strcat(&(outputString[0]), "Minor Problem in ");
	    strcat(&(outputString[0]), module);
		/* Add a colon and space after module */
	    strcat(&(outputString[0]), ": ");
	}
	else
	{
	    strcat(&(outputString[0]), "Minor Problem: ");
	}
    
	    /* create formatted error string */
	vsprintf(&(errorString[0]), fmt, ap);
    
	    /* Combine with warning/module string */
	strcat(&(outputString[0]), &(errorString[0]));
	    /* end with period and CR */
	strcat(&(outputString[0]), ".\n");

	    /* send the error string wherever the client wants */
	pIPWState->errorProc(pIPWState->errorInfo, &(outputString[0]));
    }
}
TIFFErrorHandler _TIFFwarningHandler = appleWarningHandler;

static void
appleErrorHandler(void *pState, const char* module, const char* fmt, va_list ap)
{
    char	errorString[512];
    char	outputString[512];
    IPWState   *pIPWState = (IPWState *)pState;

	/* DDavies, 11/18/96: Only print stuff is there's a printing routine */
    if ((pIPWState != NULL) && (pIPWState->errorProc != NULL))
    {
	    /* print using the IPWrap error handler.  That thing takes
	     * a string, so put the strings together.
	     * First, assemble "Error, " and the module name.
	     */
	errorString[0] = '\0';		/* null strings so far */
	outputString[0] = '\0';
	if (module != NULL)
	{
	    strcat(&(outputString[0]), " in ");
	    strcat(&(outputString[0]), module);
		/* Add a colon and space after module */
	    strcat(&(outputString[0]), "Disaster in ");
	}
	else
	{
	    strcat(&(outputString[0]), "Disaster: ");
	}

	    /* create formatted error string */
	vsprintf(&(errorString[0]), fmt, ap);
    
	    /* Combine with warning/module string */
	strcat(&(outputString[0]), &(errorString[0]));
	    /* end with period and CR */
	strcat(&(outputString[0]), ".\n");

	    /* send the error string wherever the client wants */
	pIPWState->errorProc(pIPWState->errorInfo, &(outputString[0]));
    }
}
TIFFErrorHandler _TIFFerrorHandler = appleErrorHandler;
