/* $Header: /project/icons/cvsroot/ipcore/src/ipwrap/src/tif_msdo.c,v 1.1.1.1 1999/09/20 18:19:21 jlin Exp $ */

/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994 Sam Leffler
 * Copyright (c) 1991, 1992, 1993, 1994 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
 * TIFF Library MSDOS-specific Routines.
 */
#if defined(__WATCOMC__) || defined(__BORLANDC__) || defined(_MSC_VER)
#include <io.h>		/* for open, close, etc. function prototypes */
#include <stdio.h>
#endif
#include "tiffiop.h"
#include "ipwcore.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif


static tsize_t 
_tiffReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return (read((int) fd, buf, size));
}

static tsize_t
_tiffWriteProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return (write((int) fd, buf, size));
}

static toff_t
_tiffSeekProc(thandle_t fd, toff_t off, int whence)
{
	return (lseek((int) fd, (off_t) off, whence));
}

static int
_tiffCloseProc(thandle_t fd)
{
	return (close((int) fd));
}

#include <sys/stat.h>

static toff_t
_tiffSizeProc(TIFF* tif, thandle_t fd)
{
	struct stat sb;
	tif = tif;
	return (fstat((int) fd, &sb) < 0 ? 0 : sb.st_size);
}

static int
_tiffMapProc(thandle_t fd, tdata_t* pbase, toff_t* psize)
{
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, tdata_t base, toff_t size)
{
}

/*
 * Open a TIFF file descriptor for read/writing.
 */
TIFF*
TIFFFdOpen(void *pState, int fd, const char* name, const char* mode)
{
	TIFF* tif;

	tif = TIFFClientOpen(pState, name, mode,
	    (void*) fd,
	    _tiffReadProc, _tiffWriteProc, _tiffSeekProc, _tiffCloseProc,
	    _tiffSizeProc, _tiffMapProc, _tiffUnmapProc);
	if (tif)
		tif->tif_fd = fd;
	return (tif);
}

/*
 * Open a TIFF file for read/writing.
 */
TIFF*
TIFFOpen(void *pState, const char* name, const char* mode)
{
	static const char module[] = "TIFFOpen";
	int m, fd;

	m = _TIFFgetMode(pState, mode, module);
	if (m == -1)
		return ((TIFF*)0);
	fd = open(name, m|O_BINARY, 0666);
	if (fd < 0) {
		TIFFError(pState, module, "%s: Cannot open", name);
		return ((TIFF*)0);
	}
	return (TIFFFdOpen(pState, fd, name, mode));
}

#ifdef __GNUC__
extern	char *malloc();
extern	char *realloc();
#else
#include <stdlib.h>
#endif

void *
_TIFFmalloc(TIFF *tif, size_t s)
{
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    return (IPW_MALLOC(s, cAllocatorData));
}

void
_TIFFfree(TIFF *tif, void* p)
{
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    IPW_FREE(p, cAllocatorData);
}

void *
_TIFFrealloc(TIFF *tif, void* p, size_t s)
{
    IPWState   *pIPWState = (IPWState *) (tif->tif_ipwstate);
    return (IPW_REALLOC(p, s, cAllocatorData));
}

void
_TIFFmemset(void* p, int v, size_t c)
{
	memset(p, v, c);
}

void
_TIFFmemcpy(void* d, const void* s, size_t c)
{
	memcpy(d, s, c);
}

int
_TIFFmemcmp(const void* p1, const void* p2, size_t c)
{
	return (memcmp(p1, p2, c));
}

static void
msdosWarningHandler(void *pState, const char* module,
		    const char* fmt, va_list ap)
{
    char	errorString[512];
    char	outputString[512];
    IPWState   *pIPWState = (IPWState *)pState;

	/* DDavies, 11/18/96: Only print stuff is there's a printing routine */
    if ((pIPWState != NULL) && (pIPWState->errorProc != NULL))
    {
	    /* print using the IPWrap error handler.  That thing takes
	     * a string, so put the strings together.
	     * First, assemble "Warning, " and the module name.
	     */
	errorString[0] = '\0';		/* null strings so far */
	outputString[0] = '\0';
	if (module != NULL)
	{
	    strcat(&(outputString[0]), "Minor Problem in ");
	    strcat(&(outputString[0]), module);
		/* Add a colon and space after module */
	    strcat(&(outputString[0]), ": ");
	}
	else
	{
	    strcat(&(outputString[0]), "Minor Problem: ");
	}

	    /* create formatted error string */
	vsprintf(&(errorString[0]), fmt, ap);
    
	    /* Combine with warning/module string */
	strcat(&(outputString[0]), &(errorString[0]));
	    /* end with period and CR */
	strcat(&(outputString[0]), ".\n");

	    /* send the error string wherever the client wants */
	pIPWState->errorProc(pIPWState->errorInfo, &(outputString[0]));
    }
}
TIFFErrorHandler _TIFFwarningHandler = msdosWarningHandler;

static void
msdosErrorHandler(void *pState, const char* module, const char* fmt, va_list ap)
{
    char	errorString[512];
    char	outputString[512];
    IPWState   *pIPWState = (IPWState *)pState;

	/* DDavies, 11/18/96: Only print stuff is there's a printing routine */
    if ((pIPWState != NULL) && (pIPWState->errorProc != NULL))
    {
	    /* print using the IPWrap error handler.  That thing takes
	     * a string, so put the strings together.
	     * First, assemble "Error, " and the module name.
	     */
	errorString[0] = '\0';		/* null strings so far */
	outputString[0] = '\0';
	if (module != NULL)
	{
	    strcat(&(outputString[0]), "Disaster in ");
	    strcat(&(outputString[0]), module);
		/* Add a colon and space after module */
	    strcat(&(outputString[0]), ": ");
	}
	else
	{
	    strcat(&(outputString[0]), "Disaster: ");
	}
	
	    /* create formatted error string */
	vsprintf(&(errorString[0]), fmt, ap);

	    /* Combine with warning/module string */
	strcat(&(outputString[0]), &(errorString[0]));
	    /* end with period and CR */
	strcat(&(outputString[0]), ".\n");

	    /* send the error string wherever the client wants */
	pIPWState->errorProc(pIPWState->errorInfo, &(outputString[0]));
    }
}
TIFFErrorHandler _TIFFerrorHandler = msdosErrorHandler;
