/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifconv.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifconv.c:	XIF Image conversion routines
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xiff.h"
#include "xifimage.h"
#include "xifconv.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifconv.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do internal image processing to convert from TIFF format to PIXR
 *
 * LongDescription:::
 *  w_XifConvertImage() does minimal image processing required to put
 *  image data into an expected PIXR state (binary: 0 = white, non-paletted:
 *  0 = black).
 *
 * Input:::
 *  xImg::			Valid image
 *  pPixr::			Image data structure
 *  dwFlags::			Processes (see xifconv.h XIF_CONV_xxx)
 *  rowOffset::			Offset of first row to process
 *  rowCount::			Number of rows to process
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifConvertImage(
	XImg *xImg,
	PIXR *pPixr,
	UInt32 dwFlags,
	UInt32 rowOffset,
	UInt32 rowCount
)
{
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState;
    PIXR *newPixr = NULL;
    PROLOGUE;

    /* Reference the image dimensions */
    pxImgInfo = &xImg->imgInfo;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    if (dwFlags & XIF_CONV_INVERT) {
	ERR_PASS(w_rasterop(pIPWState, 0, 0, NULL, PIX_NOT(PIX_DST),
	    0, rowOffset, pxImgInfo->dwWidth, rowCount, pPixr));
    }

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newPixr) {
	w_destroyPixr(pIPWState, &newPixr);
    }

    ERR_RETURN;
}


/*****************************************************************************
 * Begindocumentation:::
 *
 * OneLineDescription:::
 *  Convert the byte ordering of arbitrary imagery to PIXR order.
 *
 * LongDescription:::
 *  wp_PCtoPIXR() converts the byte ordering of arbitrary imagery to PIXR order.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  inImg::			Image to convert
 *  inBPL::			Bytes per image line in inImg
 *
 * Output:::
 *  outImg::			Image to receive the converted data
 *  outBPL::			Bytes per image line in outImg
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_PCtoPIXR(
	 XImg *xImg,
	 UInt8 **inImg,
	 UInt32 inBPL,
	 UInt8 **outImg,
	 UInt32 outBPL
)
{
    UInt32 row, rows;
    UInt32 quad, quads;
    UInt8 *in;
    UInt32 *out;
    PROLOGUE;

    rows = xImg->imgInfo.dwHeight;
    quads = (inBPL + 3) / 4;

    for (row = 0; row < rows; row++) {
	in = *inImg + inBPL * row;
        out = (UInt32 *)(*outImg + outBPL * row);
	for (quad = 0; quad < quads; quad++) {
	    *out++ = (in[0] << 24) | (in[1] << 16) | (in[2] << 8) | in[3];
	    in += 4;
	}
    }

    NORMAL_RETURN;
}

/*
$Log: xifconv.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.4  1997/06/26 22:24:34  aber
aber on Thu Jun 26 15:23:49 PDT 1997

xifconv.c -> /product/ipcore/ipwrap/src/RCS/xifconv.c,v

Add conversion routine for inverting "white-is-zero" images (XIF 2.0
reading only).

Revision 1.3  1997/05/30 00:50:56  aber
aber on Thu May 29 17:50:56 PDT 1997

xifconv.c -> /product/ipcore/ipwrap/src/RCS/xifconv.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.2  1997/05/20 22:29:00  aber
aber on Tue May 20 15:29:00 PDT 1997

xifconv.c -> /product/ipcore/ipwrap/src/RCS/xifconv.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.1  1997/05/17 00:46:33  aber
Initial revision

*/
