/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifread.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifread.c:	XIFF READ source
 *
 *	Handles read-only XIFF handling.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrcmprs.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifrtiff.h"
#include "xifrxiff.h"
#include "xifrpriv.h"
#include "xifrjpeg.h"
#include "xifrtok.h"
#include "xiff.pub"
#ifndef XIF_READONLY
#include "xifwxiff.h"
#endif /* XIF_READONLY */

IP_RCSINFO(RCSInfo, "$RCSfile: xifread.c,v $; $Revision: 1.2 $; $Date: 2000/04/12 23:22:54 $");



/* XIF_InitInstance()
 *
 * OneLineDescription:::
 *  Initialize a XIF instance, used to identify the XIF state.
 *
 * LongDescription:::
 *  XIF_InitInstance() initializes a XIF instance, used to store file I/O
 *  routines connecting the instance to a handle, for maintaining the
 *  malloc/free functions, and for retaining the default authoring description.
 *
 * Input:::
 *  dwClientInstID::	Client ID assigned to this instance
 *
 * Output:::
 *  pxInstHandle::		Pointer to the instance handle returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_InitInstance(
    UInt32 dwClientInstID,
    XIF_INSTHANDLE *pxInstHandle
)
{
    Instance *instData = NULL;
    XifMemoryTally *tally = NULL;
    PROLOGUE;

    /* Return nothing if failure */
    *pxInstHandle = (XIF_INSTHANDLE)instData;

    /* Here's the only calls to the system malloc */
    CHECK_MALLOC(instData = (Instance *)malloc(sizeof(*instData) +
				       sizeof(*tally)));
    tally = (XifMemoryTally *)((char *)instData + sizeof(*instData));
    
    memset(instData, 0, sizeof(*instData));
    memset(tally, 0, sizeof(*tally));
    
    /* Set defaults */
    instData->handleID = (UInt32)instData;	/* Consistency check */
    instData->nextDocID = 0;	/* Doc unique IDs */
    instData->xiffMagicNum = XIFF_MAGIC_NUM;
    
    /* Preserve the default ID for switching back and forth between the
     * client ID and the reserved default */
    instData->defaultMallocID = (void *)tally;
    
    /* Save user identifier */
    instData->clientInstID = dwClientInstID;
    
    /* Initialize the malloc functions to the defaults */
    XERR_PASS_CLEANUP(XIF_SetMallocFuncs((XIF_INSTHANDLE)instData,
			    instData->defaultMallocID, wp_XifMalloc,
			    wp_XifFree));
    
    /* Return what we did, obscured from the client */
    *pxInstHandle = (XIF_INSTHANDLE)instData;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (instData) {
	free(instData);
    }
    if (tally) {
	free(tally);
    }
    
    ERR_RETURN;
}


/* XIF_SetMallocFuncs()
 *
 * OneLineDescription:::
 *  Declare a malloc and free function to be used for memory calls.
 *
 * LongDescription:::
 *  XIF_SetMallocFuncs() declares a malloc and free function to be used in
 *  all subsequent memory calls. This allows the client to provide their own
 *  memory management. Without this call, default internal routines are used
 *  instead, which may use the system malloc()/free() pair.
 *
 * Input:::
 *  xInst::			XIF instance handle, returned from
 *				XIF_InitInstance()
 *  dwClientMallocID::		Identifier pass back to functions when called
 *  xMalloc::			Memory allocator function
 *  xFree::			Memory free'r function
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_SetMallocFuncs(
    XIF_INSTHANDLE xInst,
    void *dwClientMallocID,
    XIF_MALLOCFUNC xMalloc,
    XIF_FREEFUNC xFree
)
{
    Instance *instData = (Instance *)xInst;
    PROLOGUE;
    
    CheckHandleValidity(xInst);	/* Must be valid, even though not used */

    if (xMalloc && xFree) {	/* Must both be set to use clients */
	instData->malloc = xMalloc;
	instData->free = xFree;
	instData->mallocID = dwClientMallocID;
    } else if (!xMalloc || !xFree) {
	__error__ = XIF_BADPARAMETER;
	goto ERR_CLEANUP;
    } else {
	
	/* NULL functions indicate to use the defaults */
	instData->malloc = wp_XifMalloc;
	instData->free = wp_XifFree;
	instData->mallocID = instData->defaultMallocID;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/* XIF_GetInstance()
 *
 * OneLineDescription:::
 *  Get the "XIFF instance" ID associated with a XIFF document handle
 *
 * LongDescription:::
 *  XIF_EndInstance() returns the "XIFF instance" ID previously associated
 *  with a XIFF document handle.
 *
 * Input:::
 *  xDocHandle::		Active document handle
 *  pxInstHandle::		Pointer to the XIF instance handle to be
 *				returned, originally returned from
 *				XIF_InitInstance(), and to be passed in to
 *				XIF_EndInstance().
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetInstance(
    XIF_DOCHANDLE xDocHandle,
    XIF_INSTHANDLE *pxInstHandle
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    PROLOGUE;
    
    CheckHandleValidity(xDocHandle);

    *pxInstHandle = (XIF_INSTHANDLE)XIF_DOCINST(xDoc);
    NORMAL_RETURN;
}


/* XIF_EndInstance()
 *
 * OneLineDescription:::
 *  End a previously estalished XIF instance.
 *
 * LongDescription:::
 *  XIF_EndInstance() ends a previously estalished XIF instance. It is an
 *  error to end an instance with a current document open (a call to
 *  XIF_CloseDocument() is required first).
 *
 * Input:::
 *  xInst::			XIF instance handle, returned from
 *				XIF_InitInstance()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_EndInstance(
    XIF_INSTHANDLE xInst
)
{
    Instance *instData = (Instance *)xInst;
    PROLOGUE;
    
    CheckHandleValidity(xInst);
    
    /* End IPCore instance */
    if (instData->pIPWState && instData->myIPWState) {
	XERR_PASS_CLEANUP(w_IPWStateTerm((IPWState **)&instData->pIPWState));
    }
    
    /* Check known XIFF memory for memory leaks */
    if (ERR_IGNORE(wp_XifMemoryCheck((XifMemoryTally *)instData->defaultMallocID))) {
	/* Assume something wasn't called correctly */
	__error__ = XIF_SEQUENCE;
	/* Contine on with dealloc */
    }
    
    /* Due to potential realloc() calls, some operating systems allow this
     * pointer to propogate. Invalidate before we free it. */
    memset((void *)instData->defaultMallocID, 0, sizeof(XifMemoryTally));
    memset(instData, 0, sizeof(*instData));

    if (instData->defaultMallocID) {
	free((void *)instData->defaultMallocID);
    }
    free(instData);
    
    if (__error__ != XIF_NOERROR) {
	goto ERR_CLEANUP;    /* This is silly, but now we reference the label */
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/* XIF_IsXiff()
 *
 * OneLineDescription:::
 *  Check a file for a header identifying it as a currenly supported XIFF.
 *
 * LongDescription:::
 *  XIF_IsXiff() reads enough of a file to identify it as a currently
 *  support XIFF file. The client must only provide enough input parameters
 *  to allow reading and roaming the file to make such a decision. The client
 *  ID is a parameter passed back through the I/O functions, such that any of
 *  these supporting functions could use the ID to identify a file descriptor
 *  or file pointer, or even possibly a structure with this and a file state.
 *  The XIFF file's version and revision are returned if the file is a XIFF,
 *  and are zeroed otherwise. XIF_NOSUPPORT is returned if not supported.
 *  XIF_BADFILEFORMAT is returned if not a XIFF.
 *
 * Input:::
 *  xInst::			XIF instance handle, returned from
 *				XIF_InitInstance()
 *  dwClientID::		Client ID, passed back through all I/O funcs
 *  xReadFunc::			Client file read function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *
 * Output:::
 *  pdwVersion::		Pointer to the version
 *  pdwRevision::		Pointer to the revision
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_IsXiff(
	XIF_INSTHANDLE xInst,
	Int32 dwClientID,
	XIF_READFUNC xReadFunc,
	XIF_SEEKFUNC xSeekFunc,
	XIF_SIZEFUNC xSizeFunc,
	UInt8 *pVersion,
	UInt8 *pRevision
)
{
    XFile xFile;
    PROLOGUE;
    
    /* Return nothing until something found */
    *pVersion = 0UL;
    *pRevision = 0UL;
    
    /* Setup a file structure for temporary use */
    memset(&xFile, 0, sizeof(xFile));
    xFile.handleID = (UInt32)&xFile;
    xFile.instance = (Instance *)xInst;
    XIF_FILETOKEN(&xFile)->dwClientFileID = dwClientID;
    XIF_FILETOKEN(&xFile)->FileRead = xReadFunc;
    XIF_FILETOKEN(&xFile)->FileSeek = xSeekFunc;
    XIF_FILETOKEN(&xFile)->FileSize = xSizeFunc;
    
    /* Check to see if it's a XIFF file, using temporary xFile token */
    XERR_PASS_CLEANUP(wp_XifIsXiff(&xFile, pVersion, pRevision));
    
 ERR_CLEANUP:
    
    wp_XifFileUnLock(XIF_FILETOKEN(&xFile));
    NORMAL_RETURN;
}


/* XIF_OpenDocumentRead()
 *
 * OneLineDescription:::
 *  Open a document for read only
 *
 * LongDescription:::
 *  XIF_OpenDocumentRead() begins a session of reading a document. The
 *  client must already have opened the file at the operating system level,
 *  so that the read and seek methods in the file token are available.
 *  This function will perform any initialization required for reading
 *  the file, including reading enough header information to verify
 *  that the file is in a format supported by this version of
 *  the XIF library.
 *  <BR>
 *  A document handle is returned through the third parameter. This handle
 *  is passed to all subsequent function calls that use the document, and
 *  will encompass the I/O functions passed in here.
 *
 * Input:::
 *  xInstHandle::		XIF instance handle, returned from
 *				XIF_InitInstance()
 *  dwClientID::		Input client ID
 *  xReadFunc::			Client file read function
 *  xWriteFunc::		Client file write function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *
 * Output:::
 *  pxDocHandle::		Document handle used to refer to this
 *				page in other functions
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_OpenDocumentRead(
	XIF_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	XIF_READFUNC xReadFunc,
	XIF_WRITEFUNC xWriteFunc,
	XIF_SEEKFUNC xSeekFunc,
	XIF_SIZEFUNC xSizeFunc,
	XIF_DOCHANDLE *pxDocHandle
)
{
    Instance *instData = (Instance *)xInstHandle;
    XDoc *xDoc = NULL;
    XFile *xFile = NULL;
    UInt8 version = 0, revision = 0;
    PROLOGUE;

    /* Clear the results for now */
    *pxDocHandle = 0UL;

    CheckHandleValidity(xInstHandle);

    /* Check for proper calling sequence */


    /* Create a new document (handle) */
    CHECK_MALLOC(xDoc = (XDoc *)XIF_MALLOC(sizeof(XDoc)));
    memset(xDoc, 0, sizeof(*xDoc));
    xDoc->handleID = (UInt32)xDoc;	/* Self identity */
    xDoc->docID = ++instData->nextDocID;/* ID unique to this library instance
					 */

    /* Create a link back to the instance */
    CHECK_MALLOC(xFile = (XFile *)XIF_MALLOC(sizeof(XFile)));
    memset(xFile, 0, sizeof(*xFile));
    xFile->handleID = (UInt32)xFile;	/* Self identity */

    /* Setup the file handle */
    xFile->instance = instData;
    XIF_FILETOKEN(xFile)->dwClientFileID = dwClientID;
    XIF_FILETOKEN(xFile)->FileRead = xReadFunc;
    XIF_FILETOKEN(xFile)->FileWrite = xWriteFunc;
    XIF_FILETOKEN(xFile)->FileSeek = xSeekFunc;
    XIF_FILETOKEN(xFile)->FileSize = xSizeFunc;
    
    /* Retain the file handle */
    xDoc->xFile = xFile;
    
    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_XifFileLock(XIF_FILETOKEN(xFile)));
    
    /* Check the XIFF header */
    __error__ = wp_XifIsXiff(xFile, &version, &revision);
    switch ((int)__error__) {
    case XIF_NOERROR:
	break;
    case XIF_NOTTIFF:
	/* Not even a valid TIF */
	goto ERR_CLEANUP;
    case XIF_NOTXIFF:
	/* Not a valid XIF, but potentially readable. Continue on but prepare
	 * for later failure. */
	__error__ = XIF_NOERROR;
	break;
    case XIF_NOSUPPORT:
	/* Version not supported */
	goto ERR_CLEANUP;
    default:
	/* Other error (read/seek/etc.) */
	goto ERR_CLEANUP;
    }
    
    /* Initialize */
    xDoc->seqStatus |= XDOC_SEQ_READSTART;
    xDoc->readWrite = READ_ALLOWED;
    xDoc->docInfo.version = version;
    xDoc->docInfo.revision = revision;

    xDoc->readWrite = READ_ALLOWED;
    
    /* Return the result */
    *pxDocHandle = (UInt32)xDoc;
    
    XERR_PASS_CLEANUP(wp_XifFileUnLock(XIF_FILETOKEN(xFile)));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (xDoc) {
	XIF_FREE(xDoc);
    }
    if (xFile) {
	XIF_FREE(xFile);
    }
    
    if (xFile) {
	wp_XifFileUnLock(XIF_FILETOKEN(xFile));
	xFile = NULL;
    }
    NORMAL_RETURN;
}


/* XIF_GetDocInfo()
 *
 * OneLineDescription:::
 *  Get the basic content of the document.
 *
 * LongDescription:::
 *  XIF_GetDocInfo() returns the document specific content, including: XIFF
 *  version, XIFF revision, number of pages, authoring description, and the
 *  reference page.
 *
 * Input:::
 *  xDocHandle::		Input document handle
 *
 * Output:::
 *  pdwVersion::		Version of XIFF
 *  pdwRevision::		Revision of this version of XIFF
 *  pdwNPages::			Number of pages in file
 *  ppAuthorBlock::		Author string (NULL if none
 *  pwAuthorBlockBytes::	Size of author string
 *  pdwReferencePageOffset::	Reference page offset (-1 if none)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetDocInfo(
	XIF_DOCHANDLE xDocHandle,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt16 *pwNPages,
	const Int8 **ppAuthorBlock,
	UInt16 *pwAuthorBlockBytes,
	Int32 *pdwReferencePageOffset
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    XIF_DOCINFO *docInfo;
    PROLOGUE;
    
    CheckHandleValidity(xDocHandle);
    
    /* Check for proper calling sequence */
    
    /* Clear the inputs */
    *pVersion = *pRevision = 0;
    *pwNPages = 0;
    *pdwReferencePageOffset = -1;	/* None set */
    *pwAuthorBlockBytes = 0;
    *ppAuthorBlock = NULL;
    
    /* If it has already been read, don't read again */
    if ((xDoc->seqStatus & XDOC_SEQ_GETINFO) == 0) {
	
	/* Read the page extension tags */
	XERR_PASS_CLEANUP(wp_XifReadExtensions(xDoc));
	
	/* Read the extension data (pagetable, reference page, author, etc. */
	XERR_PASS_CLEANUP(wp_XifReadTagData(xDoc));

	/* If we have a page table, and the page count is > 0, and the
	 * number of entries per table is 0, something is wrong */
	XERR_GEN_CLEANUP(xDoc->pageTableHeader.pageCount > 0 &&
			 xDoc->pageTableHeader.entriesPerTable == 0,
			 XIF_BADFILEFORMAT);
	
	xDoc->seqStatus |= XDOC_SEQ_GETINFO | XDOC_SEQ_W_PAGETABLE;
    }
    docInfo = &xDoc->docInfo;

    /* Set the sequence */
    xDoc->seqStatus |= XDOC_SEQ_READSTART;

    /* Return the results */
    *pVersion = docInfo->version;
    *pRevision = docInfo->revision;
    *pwNPages = docInfo->wNPages;
    *pwAuthorBlockBytes = docInfo->wAuthorBlockBytes;
    *pdwReferencePageOffset = docInfo->dwReferencePageOffset;
    *ppAuthorBlock = docInfo->pAuthorBlock;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_GetPageInfo()
 *
 * OneLineDescription:::
 *  Get the basic content of the page
 *
 * LongDescription:::
 *  XIF_GetPageInfo() returns the page specific content, including:
 *  image layer counts, number of layers present, annotation count, and a
 *  handle used to reference this page for other functions.
 *
 * Input:::
 *  xDocHandle::		Document handle
 *  dwPageOffset::		Page offset within the document (-1 for the
 *				last page, including incompleted last page)
 *
 * Output:::
 *  ppdwImageCounts::		Number of images for each image layer (array,
 *				this may NULL, in which case no array is
 *				returned)
 *  pdwLayersPresent::		Number of image layers present (indicates size
 *				of array pdwImageLayers[], may have missing
 *				layers)
 *  pdwAnnotationCount::	Number of annotations on page
 *  pxPageHandle::		Page handle, used to identify this page
 *				in the page level routines
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetPageInfo(
	XIF_DOCHANDLE xDocHandle,
	Int32 dwPageOffset,
	UInt32 **ppdwImageCounts,
	UInt32 *pdwLayersPresent,
	UInt16 *pwAnnotationCount,
	XIF_PAGEHANDLE *pxPageHandle
)
{
    Instance *instData;
    XFile *xFile;
    XDoc *xDoc = (XDoc *)xDocHandle;
    XPage *xPage = NULL;
    XPage *searchPage = NULL;
    XImg *xImg = NULL;
    UInt8 flags;
    Int32 ifdOffset;
    XIF_PAGEINFO xPageInfo;
    UInt32 xOffset, yOffset, xRes, yRes, width, height;
    UInt32 maskImages, bestStripHeight;
    XIF_COMPRESSION compression;
    XIF_COLORTYPE colorType;
    UInt8 palette[256*3];
    XIF_IMGHANDLE primImgHandle;
    PROLOGUE;
    
    CheckHandleValidity(xDocHandle);

    /* Return nothing by default */
    if (ppdwImageCounts) {
	*ppdwImageCounts = NULL;
    }
    *pdwLayersPresent = 0;
    *pwAnnotationCount = 0;
    *pxPageHandle = 0;

    /* Check for proper calling sequence and parameters... */

    /* Must have the document open for read or write */
    XERR_GEN_CLEANUP(((xDoc->seqStatus &
	(XDOC_SEQ_READSTART | XDOC_SEQ_WRITESTART)) == 0),
	XIF_SEQUENCE);

    /* Must have gotten the document info (either from XIF_GetDocInfo() or
     * from XIF_OpenDocument... */
    XERR_GEN_CLEANUP((xDoc->seqStatus & XDOC_SEQ_GETINFO) == 0,
		     XIF_SEQUENCE);

    /* The page should be a valid existing page, or a page that was just
     * started in the write process (XIF_AddPageStart was called). */
    if (dwPageOffset < 0) {
	dwPageOffset = xDoc->docInfo.wNPages - 1;
    }
    XERR_GEN_CLEANUP(dwPageOffset >= xDoc->pageTableHeader.pageCount &&
	dwPageOffset >= xDoc->docInfo.wNPages, XIF_BADPARAMETER);
    
    /* Clear the outputs */
    if (ppdwImageCounts) {
	*ppdwImageCounts = NULL;
    }
    *pwAnnotationCount = 0;
    *pxPageHandle = 0UL;
    memset(&xPageInfo, 0, sizeof(xPageInfo));
    
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    
    /* Look for a previous read of this page */
    searchPage = xDoc->pages;
    if (searchPage != NULL) {
	do {
	    if (searchPage->pageInfo.dwPageOffset == dwPageOffset) {
		
		/* Found, this page already visited */
		xPage = searchPage;
		break;
	    }
	} while ((searchPage = searchPage->next) != NULL);
    }
    
    if (searchPage == NULL || searchPage->pageInfo.dwPageOffset != dwPageOffset) {
	
	/* Begin construction of the page data */
	CHECK_MALLOC(xPage = (XPage *)XIF_MALLOC(sizeof(XPage)));
	memset(xPage, 0, sizeof(XPage));
	xPage->handleID = (UInt32)xPage;	/* self identity */
	xPage->xDoc = xDoc;
	xPage->pageInfo.xPageHandle = (UInt32)xPage;
	xPage->pageInfo.dwPageOffset = dwPageOffset;
	
	/* Make sure we have the page table for the correct page */
	XERR_PASS_CLEANUP(wp_XifGetPageIFD(xDoc, dwPageOffset, &flags, &ifdOffset));
	
	/* Check for valid IFD offset */
	XERR_GEN_CLEANUP((ifdOffset <= XIFF_EXTS_OFFSET), XIF_BADFILEFORMAT);
	
	/* Begin construction of the primary image data */
	XERR_PASS_CLEANUP(wp_NewImg(xPage,
				    (XImg **)&xPage->pageInfo.xPrimImgHandle,
				    ifdOffset));

	/* Note: it is possible to have the primary image with a layer 1 or layer 3
	 * designation, however, no other layers and no other images within this layer
	 * would be allowed.
	 * This may come in a future revision of XIFF 3.0, but for now, we'll stick to
	 * this. */
	((XImg *)xPage->pageInfo.xPrimImgHandle)->imgInfo.dwImgLayer = 2;
	((XImg *)xPage->pageInfo.xPrimImgHandle)->imgInfo.dwImgOffset = 0;
	((XImg *)xPage->pageInfo.xPrimImgHandle)->imgInfo.xImgType = XIF_IMGTYPE_PRIMARY;
	
	/* Attach it to the appropriate spot in linked list */
	searchPage = xDoc->pages;
	if (searchPage == NULL) {
	    xDoc->pages = xPage;
	} else {
	    do {
		if (! searchPage->next ||
		    searchPage->next->pageInfo.dwPageOffset > dwPageOffset) {
		    
		    /* Found the right spot to insert this page (after) */
		    break;
		}
	    } while ((searchPage = searchPage->next) != NULL);

	    /* If at the beginning, insert as first */
	    if (searchPage == xDoc->pages &&
		searchPage->pageInfo.dwPageOffset > dwPageOffset) {
		xPage->next = searchPage;
		searchPage->prev = xPage;
		xDoc->pages = xPage;
	    } else {
	    
		/* Insert inbetween searchPage and searchPage->next */
		xPage->next = searchPage->next;
		xPage->prev = searchPage;
		if (searchPage->next) {
		    searchPage->next->prev = xPage;
		}
		searchPage->next = xPage;
	    }
	}
    }

    XERR_GEN_CLEANUP(xPage == NULL, XIF_INTERNAL_ERROR);

    /* We have the page (either already read or not), read if not */
    if ((xPage->seqStatus & XPAGE_SEQ_GETINFO) == 0) {

	/* If the primary image has not been added yet, return counts of zero */
	if ((xPage->pageInfo.dwLayersPresent == 0 &&
	    xDoc->pageTableHeader.pageCount != xDoc->docInfo.wNPages) ||
	    xPage->pageInfo.xPrimImgHandle == 0) {
	    *pxPageHandle = xPage->pageInfo.xPageHandle;
	    NORMAL_RETURN;
	}
	
	/* Set the sequence */
        xPage->seqStatus |= XPAGE_SEQ_READSTART;

	/* Parse through the page hierarchy, and tally construction stats */
	
	/* Go to the primary IFD and read the tags */
	xImg = (XImg *)xPage->pageInfo.xPrimImgHandle;
	XERR_PASS_CLEANUP(XIF_GetImageInfo((UInt32)xPage, 2,
	    0, &xOffset, &yOffset, &xRes, &yRes, &width, &height,
	    &bestStripHeight, &compression, &colorType,
	    palette, &maskImages, &primImgHandle));

	/* The annotation data is handled specially. Version 2.x stores the
	 * annotation data as an array of offsets, then stores the annotation
	 * characteristics in a header in front of the data at each of those
	 * offsets. During the TIFF field parsing, the annotation count was
	 * obtained directly from the IFD itself, if the version is 2.x.
	 * Otherwise, the count is contained within the annotation table header.
	 * Let's read it now, so we can pass that count back.
	 */
	XERR_PASS_CLEANUP(wp_XifGetAnnotTableHeader(xPage));
	xPage->pageInfo.wAnnotationCount = xPage->annotTableHeader.annotCount;

	xPage->seqStatus |= XPAGE_SEQ_GETINFO;
    } /* Info not already gotten */

    /* Copy out the page information
     */
    /* Make a copy of the real layer image counts, so that the client cannot
     * muck with the actual list */
    if (xPage->extImageCounts) {
	XIF_FREE(xPage->extImageCounts);
	xPage->extImageCounts = NULL;
    }

    /* Copy the image counts, for each layer, to the output */
    XERR_GEN_CLEANUP(xPage->pageInfo.dwLayersPresent == 0, XIF_INTERNAL_ERROR);
    CHECK_MALLOC(xPage->extImageCounts = 
	(UInt32 *)XIF_MALLOC(sizeof(UInt32) * xPage->pageInfo.dwLayersPresent));
    memcpy(xPage->extImageCounts, xPage->pageInfo.pdwImageCounts,
	sizeof(UInt32) * xPage->pageInfo.dwLayersPresent);
    if (ppdwImageCounts) {
	*ppdwImageCounts = xPage->extImageCounts;
    }
    *pdwLayersPresent = xPage->pageInfo.dwLayersPresent;
    *pwAnnotationCount = xPage->pageInfo.wAnnotationCount;
    *pxPageHandle = xPage->pageInfo.xPageHandle;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_GetImageInfo()
 *
 * OneLineDescription:::
 *  Get the image description, including MRC image counts; indicate support.
 *
 * LongDescription:::
 *  XIF_GetImageInfo() gets the description of an image: position relative
 *  to the Primary image, resolution, size, stored strip size, stored
 *  compression type, color type, palette (if any), and mask image count
 *  (if any).
 *  <BR>
 *  Returned is an image handle to be used to reference this image in later
 *  calls. A call to XIF_CloseDocument() renders this image handle invalid.
 *  <BR>
 *  This function also serves as the mechanism to indicate current support.
 *  If the characteristics of this image are unsupported, then an error is
 *  returned: XIF_NOSUPPORT.
 *
 * Input:::
 *  xPageHandle::		Page handle
 *  dwImgLayer::		Image layer (1, 2, ...)
 *  dwImgOffset::		Image offset for image within that layer (0,
 *				1, ...)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pxCompression::		Compression type for src/dst
 *  pxColorType::		Color type (see XIF_COLORTYPE
 *  pPalette::			Color palette if XIF_COLORTYPE is
 *				XIF_COLORTYPE_COLOR4 or XIF_COLORTYPE_COLOR8.
				Palette is RGB triples (chunky): RGBRGB...
 *  pdwMaskImages::		Number of mask images
 *  pxImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetImageInfo(
    XIF_PAGEHANDLE xPageHandle,
    UInt32 dwImgLayer,
    UInt32 dwImgOffset,
    UInt32 *pdwXOffset,
    UInt32 *pdwYOffset,
    UInt32 *pdwXResolution,
    UInt32 *pdwYResolution,
    UInt32 *pdwWidth,
    UInt32 *pdwHeight,
    UInt32 *pdwBestStripHeight,
    XIF_COMPRESSION *pxCompression,
    XIF_COLORTYPE *pxColorType,
    UInt8 *pPalette,
    UInt32 *pdwMaskImages,
    XIF_IMGHANDLE *pxImgHandle
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XImg *xPrimImg, *xImg = NULL;
    XPage *xPage = (XPage *)xPageHandle;
    XDoc *xDoc = NULL;
    XIF_IMGINFO *imgInfo;
    UInt32 layer;
    UInt32 *prevImageCounts = NULL;
    UInt32 prevLayersPresent;
    PROLOGUE;
    
    CheckHandleValidity(xPageHandle);
    
    /* Check for proper calling sequence */
    

    XERR_GEN_CLEANUP(dwImgLayer < 1 ||
	(xPage->pageInfo.dwLayersPresent > 0 &&
	 (dwImgLayer > xPage->pageInfo.dwLayersPresent ||
	 xPage->pageInfo.pdwImageCounts == NULL ||
	 xPage->pageInfo.pdwImageCounts[dwImgLayer - 1] <= dwImgOffset)),
	XIF_BADPARAMETER);
    
    xDoc = xPage->xDoc;
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    
    /* Clear the results */
    *pdwXOffset = *pdwYOffset = *pdwXResolution = *pdwYResolution = 0UL;
    *pdwWidth = *pdwHeight = *pdwBestStripHeight = *pdwMaskImages = 0UL;
    *pxCompression = XIF_COMP_UNKNOWN;
    *pxColorType = XIF_COLORTYPE_NONE;
    *pxImgHandle = 0UL;
    if (pPalette) {
	memset(pPalette, 0, 256*3);
    }
    
    /* Haven't gotten the image info for the parent image yet, get now */
    xPrimImg = (XImg *)xPage->pageInfo.xPrimImgHandle;
    XERR_GEN_CLEANUP(xPrimImg == NULL, XIF_SEQUENCE);
    XERR_GEN_CLEANUP(xPrimImg->imgInfo.dwImgLayer != 2, XIF_INTERNAL_ERROR);
    if ((xPrimImg->seqStatus & XIMG_SEQ_GETINFO) == 0) {
	
	/* This really gets the info for an individual image */
	XERR_PASS_CLEANUP(wp_XifGetImageDesc(xPrimImg));

	/* If no MRContent TIFF field exists, initialize the list with the
	 * Primary image. */
	if (xPage->pageInfo.dwLayersPresent == 0) {
	    XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage, 2, 1));
	}
    }
    
    if (dwImgLayer == 2 && dwImgOffset == 0) {
	    
	/* We're already referencing the primary image */
	xImg = xPrimImg;
    } else {
	if (xPage->xImgs == NULL) {

	    /* If we have previously used the MRContent TIFF field to
	     * explain the page content, let's clear it now, and let
	     * the true content guide the values. */
	    prevLayersPresent = xPage->pageInfo.dwLayersPresent;
	    prevImageCounts = xPage->pageInfo.pdwImageCounts;
	    xPage->pageInfo.pdwImageCounts = NULL;
	    xPage->pageInfo.dwLayersPresent = 0;

	    /* Initialize the list with the Primary image, which already
	     * exists */
	    XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage, 2, 1));

	    /* Now read the actual SubIFDs list */
	    XERR_PASS_CLEANUP(wp_XifTraverseSubImages(xPrimImg));

	    /* Compare to what we had. This is really a sanity check; we could
	     * simply return, and hope the client requests the information
	     * again after the first image is read, but that's too obscure. */
	    XERR_GEN_CLEANUP(xPage->pageInfo.dwLayersPresent != prevLayersPresent,
		XIF_BADFILEFORMAT);
	    for (layer = 0; layer < prevLayersPresent; layer++) {
		XERR_GEN_CLEANUP(xPage->pageInfo.pdwImageCounts[layer] !=
		    prevImageCounts[layer], XIF_BADFILEFORMAT);
	    }

	    XIF_FREE(prevImageCounts);
	    prevImageCounts = NULL;

	}

	/* Search for the image layer and order (linked list is maintained
         * in sequential order).*/
	xImg = xPage->xImgs;
	while (xImg->imgInfo.dwImgLayer < dwImgLayer ||
	    (xImg->imgInfo.dwImgLayer == dwImgLayer &&
	    xImg->imgInfo.dwImgOffset < dwImgOffset)) {

	    /* The image with this offset doesn't exist */
	    XERR_GEN_CLEANUP(((xImg = xImg->next) == NULL), XIF_INTERNAL_ERROR);
	}
    }

    /* Check to see that the requested image is actually found */
    XERR_GEN_CLEANUP((xImg == NULL), XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(xImg->imgInfo.dwImgLayer != dwImgLayer ||
	xImg->imgInfo.dwImgOffset != dwImgOffset, XIF_INTERNAL_ERROR);
    
    /* Now get the info for that image */
    XERR_PASS_CLEANUP(wp_XifGetImageDesc(xImg));
    imgInfo = &xImg->imgInfo;
    
    /* Let's make a last-minute check to see if the parameters
     * are consistent with the page */
    if (xImg != xPrimImg && xPrimImg != NULL) {
	XERR_GEN_CLEANUP(imgInfo->dwXOffset > xPrimImg->imgInfo.dwWidth,
	    XIF_BADFILEFORMAT);
	XERR_GEN_CLEANUP(imgInfo->dwYOffset > xPrimImg->imgInfo.dwHeight,
	    XIF_BADFILEFORMAT);

	/* Note: given the resolution differences, the last subimage pixel
	 * may overhang the edge of the page, meaning the next-to-last subimage
	 * pixel must not quite go up to the edge. That check made here: */
	XERR_GEN_CLEANUP(imgInfo->dwXOffset + (imgInfo->dwWidth - 1) *
	    xPrimImg->imgInfo.dwXResolution / imgInfo->dwXResolution >=
	    xPrimImg->imgInfo.dwWidth, XIF_BADFILEFORMAT);
	XERR_GEN_CLEANUP(imgInfo->dwYOffset + (imgInfo->dwHeight - 1) *
	    xPrimImg->imgInfo.dwYResolution / imgInfo->dwYResolution >=
	    xPrimImg->imgInfo.dwHeight, XIF_BADFILEFORMAT);
    }

    /* Check other consistencies */
    if (imgInfo->dwXResolution == 0) {
	imgInfo->dwXResolution = 300;
    }
    if (imgInfo->dwYResolution == 0) {
	imgInfo->dwYResolution = 300;
    }
    if (imgInfo->dwBestStripHeight == 0) {
	imgInfo->dwBestStripHeight = imgInfo->dwHeight;
    }

    /* File must be corrupt if not at least reasonable */
    XERR_GEN_CLEANUP(imgInfo->dwWidth == 0, XIF_BADFILEFORMAT);
    XERR_GEN_CLEANUP(imgInfo->dwHeight == 0, XIF_BADFILEFORMAT);
    XERR_GEN_CLEANUP(imgInfo->xCompression == XIF_COMP_UNKNOWN,
	XIF_BADFILEFORMAT);
    XERR_GEN_CLEANUP(imgInfo->dwMaskImages > 1, XIF_BADFILEFORMAT);

    /* Check parameters for what we expect to be currently supported
     * (Note: we can't catch everything, but this has come to bite us
     * already as color tokenization was unanticipated.)
     */
    switch (imgInfo->xColorType) {
    case XIF_COLORTYPE_BINARY:
	switch (imgInfo->xCompression) {
	case XIF_TOKENIZATION:
	case XIF_LLTOKENIZATION:
	case XIF_LOSSYDITHER:
	case XIF_G4:
	case XIF_NOC:
	    break;
	default:
	    XERR_PASS(XIF_NOSUPPORT);
	}
	break;
    case XIF_COLORTYPE_GRAY4:
	switch (imgInfo->xCompression) {
	case XIF_JPEG:
	case XIF_NOC:
	    break;
	default:
	    XERR_PASS(XIF_NOSUPPORT);
	}
	break;
    case XIF_COLORTYPE_GRAY8:
	switch (imgInfo->xCompression) {
	case XIF_JPEG:
	case XIF_WAVELET:
	case XIF_NOC:
	    break;
	default:
	    XERR_PASS(XIF_NOSUPPORT);
	}
	break;
    case XIF_COLORTYPE_COLOR4:
    case XIF_COLORTYPE_COLOR8:
	switch (imgInfo->xCompression) {
	case XIF_JPEG:
	case XIF_WAVELET:
	case XIF_NOC:
	    break;
	default:
	    XERR_PASS(XIF_NOSUPPORT);
	}
	break;
    case XIF_COLORTYPE_COLOR24:
	switch (imgInfo->xCompression) {
	case XIF_JPEG:
	case XIF_WAVELET:
	case XIF_NOC:
	case XIF_TOKENIZATION:
	case XIF_LLTOKENIZATION:
	    break;
	default:
	    XERR_PASS(XIF_NOSUPPORT);
	}
	break;
    default:
	XERR_PASS(XIF_NOSUPPORT);
    }
    
    /* Copy out the image information */
    *pdwXOffset = imgInfo->dwXOffset;
    *pdwYOffset = imgInfo->dwYOffset;
    *pdwXResolution = imgInfo->dwXResolution;
    *pdwYResolution = imgInfo->dwYResolution;
    *pdwWidth = imgInfo->dwWidth;
    *pdwHeight = imgInfo->dwHeight;
    *pdwBestStripHeight = imgInfo->dwBestStripHeight;
    *pxCompression = imgInfo->xCompression;
    *pxColorType = imgInfo->xColorType;
    if (imgInfo->pPalette && pPalette) {
	memcpy(pPalette, imgInfo->pPalette, 256*3);
    }
    *pdwMaskImages = imgInfo->dwMaskImages;
    *pxImgHandle = (XIF_IMGHANDLE)xImg;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (prevImageCounts) {
	XIF_FREE(prevImageCounts);
	prevImageCounts = NULL;
    }
    
    ERR_RETURN;
}


/* XIF_GetImageStats()
 *
 * OneLineDescription:::
 *  Get some image stats (compressed size, token dictionary ID/size)
 *
 * LongDescription:::
 *  XIF_GetImageStats() accumulates the total compressed image size, as
 *  stored in the file, and if appropriate, the associated token dictionary
 *  ID and size.
 *
 * Input:::
 *  xImgHandle::		Image handle
 *
 * Output:::
 *  pdwCmpSize::		Total size of the image, compressed
 *  pdwDictCount::		If tokenized, the number of token dictionaries
 *  ppdwDictIDs::		If tokenized, the dictionary ID list
 *  ppdwDictSizes::		If tokenized, the sizes of the token
 *				dictionaries
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetImageStats(
    XIF_IMGHANDLE xImgHandle,
    UInt32 *pdwCmpSize,
    UInt32 *pdwDictCount,
    UInt32 **ppdwDictIDs,
    Int32 **ppdwDictSizes
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XImg *xImg = (XImg *)xImgHandle;
    XDoc *xDoc = NULL;
    UInt8 *pCmpData = NULL;
    Int32 filePos, dictLocation;
    UInt32 cmpBytes, cmpTotal = 0;
    UInt16 dictNum, dictCount;
    UInt32 *dictIDs = NULL;
    Int32 *dictSizes = NULL;
    UInt8 *tundraData;
    UInt32 tundraBytes;
    UInt32 row, stripHeight;
    PROLOGUE;
    
    CheckHandleValidity(xImgHandle);
    
    /* Check for proper calling sequence */
    
    xDoc = xImg->xPage->xDoc;
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    
    /* Clear the results */
    *pdwCmpSize = *pdwDictCount = 0UL;
    *ppdwDictIDs = NULL;
    *ppdwDictSizes = NULL;

    if (xImg->imgInfo.xCompression == XIF_TOKENIZATION ||
	xImg->imgInfo.xCompression == XIF_LLTOKENIZATION) {

        /* Get the token dictionary information... */

	/* Read enough of the token stream to extract the dictionary ID */
	XERR_GEN_CLEANUP(xImg->fields.StripsPerImage < 1, XIF_INTERNAL_ERROR);
	filePos = xImg->fields.StripOffsets[0];
	cmpBytes = xImg->fields.StripByteCounts[0];

	/* We could try to avoid reading the entire data stream in, but we'd
	 * have to either interpret a portion of the data first (which goes
	 * against having the interpretation all in one place), or make some
	 * assumptions about the size of the header. I did the later and got
	 * bit. Since it's compressed, just allocate and read the whole thing.
	 */
	XERR_GEN_CLEANUP(cmpBytes == 0, XIF_BADFILEFORMAT);
	CHECK_MALLOC(pCmpData = XIF_MALLOC(cmpBytes));
	XIF_SEEK(xFile, filePos);
	XIF_READ(xFile, pCmpData, cmpBytes);

	/* Get the dictionary ID from the compressed data stream */
	XERR_PASS_CLEANUP(wp_XifInterpretTokenHeader(xFile, pCmpData, cmpBytes,
	    &dictCount, &dictIDs, &tundraData, &tundraBytes));
	*pdwDictCount = dictCount;

	/* Allocate an array for the dictionary sizes */
	CHECK_MALLOC(dictSizes = XIF_MALLOC(dictCount * sizeof(UInt32)));

	/* From the IDs, get the file location and byte count */
	for (dictNum = 0; dictNum < dictCount; dictNum++) {
	    XERR_PASS_CLEANUP(wp_XifGetSymlibOffset(xImg->xPage->xDoc,
		dictIDs[dictNum], &dictLocation, &dictSizes[dictNum]));
	}
    }

    /* Loop through each strip, and request the compressed data size */
    XERR_GEN_CLEANUP(xImg->imgInfo.dwBestStripHeight < 1, XIF_INTERNAL_ERROR);
    stripHeight = xImg->imgInfo.dwBestStripHeight;
    for (row = 0; row < xImg->imgInfo.dwHeight; row += stripHeight) {
	if (row + stripHeight > xImg->imgInfo.dwHeight) {
	    stripHeight = xImg->imgInfo.dwHeight - row;
	}
	XERR_PASS_CLEANUP(XIF_GetCompressedStripSize(xImgHandle, row,
						     &cmpBytes));
	cmpTotal += cmpBytes;
    }

    /* Preserve the arrays in structures that will persist until the image
     * is freed */
    if (xImg->dictList) {
	XIF_FREE(xImg->dictList);
    }
    xImg->dictList = dictIDs;
    dictIDs = NULL;
    if (xImg->dictSizes) {
	XIF_FREE(xImg->dictSizes);
    }
    xImg->dictSizes = dictSizes;
    dictSizes = NULL;
    if (pCmpData) {
	XIF_FREE(pCmpData);
	pCmpData = NULL;
    }

    /* Copy out the image information */
    *pdwCmpSize = cmpTotal;
    *ppdwDictIDs = xImg->dictList;
    *ppdwDictSizes = xImg->dictSizes;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (pCmpData) {
	XIF_FREE(pCmpData);
    }
    if (dictIDs) {
	XIF_FREE(dictIDs);
    }
    if (dictSizes) {
	XIF_FREE(dictSizes);
    }
    
    ERR_RETURN;
}


/* XIF_GetMaskImageInfo()
 *
 * OneLineDescription:::
 *  Get the image information for a transparency mask.
 *
 * LongDescription:::
 *  XIF_GetMaskImageInfo() returns the same information about an image as does
 *  XIF_GetImageInfo() (see above), but does so for a transparency mask
 *  attached to a non-binary image.
 *
 * Input:::
 *  xParentImgHandle::		Parent image handle
 *  dwImgOffset::		Image offset (should always be 0 for now)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pxCompression::		Compression type for src/dst
 *  pxImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetMaskImageInfo(
	XIF_IMGHANDLE xParentImgHandle,
	UInt32 dwImgOffset,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	XIF_COMPRESSION *pxCompression,
	XIF_IMGHANDLE *pxImgHandle
)
{
    Instance *instData;
    XFile *xFile;
    XImg *xParentImg = (XImg *)xParentImgHandle;
    XImg *xImg = NULL;
    XIF_IMGINFO *imgInfo;
    PROLOGUE;
    
    /* Clear the results */
    *pdwXOffset = *pdwYOffset = *pdwXResolution = *pdwYResolution = 0UL;
    *pdwWidth = *pdwHeight = *pdwBestStripHeight = 0UL;
    *pxCompression = XIF_COMP_UNKNOWN;
    *pxImgHandle = 0UL;
    
    CheckHandleValidity(xParentImgHandle);
    
    /* Check for proper calling sequence */


    instData = XIF_IMGINST(xParentImg);
    xFile = XIF_IMGFILE(xParentImg);
    
    /* Haven't gotten the image info for the parent image yet, get now */
    if ((xParentImg->seqStatus & XIMG_SEQ_GETINFO) == 0) {
	
	/* This really gets the info for an individual image */
	XERR_PASS_CLEANUP(wp_XifGetImageDesc(xParentImg));
    }
    
    /* Search for the first image of the type requested */
    XERR_GEN_CLEANUP((! xParentImg->imgInfo.dwMaskImages),
		     XIF_BADPARAMETER);
    if (xParentImg->xMaskImg == NULL) {
	XERR_PASS_CLEANUP(wp_XifTraverseSubImages(xParentImg));
    }
    xImg = xParentImg->xMaskImg;

    /* Check to see that the requested image is actually found */
    XERR_GEN_CLEANUP((xImg == NULL), XIF_BADPARAMETER);
    
    /* Search for the Nth image of that type */
    while (dwImgOffset > xImg->imgInfo.dwImgOffset) {

	/* The image with this offset doesn't exist */
	XERR_GEN_CLEANUP(((xImg = xImg->next) == NULL),
		 XIF_BADPARAMETER);
    }
    
    /* Now get the info for that image */
    XERR_PASS_CLEANUP(wp_XifGetImageDesc(xImg));
    
    /* Copy out the image information */
    imgInfo = &xImg->imgInfo;
    *pdwXOffset = imgInfo->dwXOffset;
    *pdwYOffset = imgInfo->dwYOffset;
    *pdwXResolution = imgInfo->dwXResolution;
    *pdwYResolution = imgInfo->dwYResolution;
    *pdwWidth = imgInfo->dwWidth;
    *pdwHeight = imgInfo->dwHeight;
    *pdwBestStripHeight = imgInfo->dwBestStripHeight;
    *pxCompression = imgInfo->xCompression;
    *pxImgHandle = (XIF_IMGHANDLE)xImg;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_ImageReadStart()
 *
 * OneLineDescription:::
 *  Specify the format of, and prepare to read image data.
 *
 * LongDescription:::
 *  XIF_ImageReadStart() prepares to read image data, and is used to specify
 *  the format of the data to be returned. These values are only stored for
 *  now, but used during later calls to XIF_ImageReadStrip().
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetImageInfo()
 *  dwFlags::			See XIF_IMGFLAGS_xxx
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ImageReadStart(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwFlags
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    PROLOGUE;
    
    CheckHandleValidity(xImgHandle);
    pxImgInfo = &xImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_READSTATUS) !=
		      XIMG_SEQ_READFINISH), XIF_SEQUENCE);

#ifndef XIF_READONLY
    /* If input file is in read/write mode, flush the page before we
     * can read it */
    if ((xImg->xPage->xDoc->readWrite & WRITE_ALLOWED) &&
	xImg->xPage->seqStatus & (XPAGE_SEQ_TOKENSTART)) {
	XERR_PASS_CLEANUP(XIF_WriteTokenImages((XIF_DOCHANDLE)xImg->xPage->xDoc));
    }
#endif /* XIF_READONLY */

    /* Store the parameters */
    xImg->imgFlags = dwFlags;
    
    /* Update the sequence status */
    xImg->seqStatus |= XIMG_SEQ_READSTART;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_GetCompressedStripSize()
 *
 * OneLineDescription:::
 *  Get the number of bytes of the compressed strip
 *
 * LongDescription:::
 *  XIF_GetCompressedStripSize() retrieves the size of the strip in compressed
 *  bytes.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetImageInfo()
 *  dwImageRowOffset::		The row offset in the image proper
 *
 * Output:::
 *  pdwBytes::			The number of bytes for the compressed strip
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetCompressedStripSize(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImageRowOffset,
	UInt32 *pdwBytes
)
{
    XImg *xImg = (XImg *)xImgHandle;
    UInt32 offset;
    PROLOGUE;

    /* Default is to return nothing */
    *pdwBytes = 0L;

    /* Check the image handle */
    CheckHandleValidity(xImgHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((xImg->seqStatus & XIMG_SEQ_GETINFO) == 0, XIF_SEQUENCE);
    XERR_GEN_CLEANUP(xImg->imgInfo.dwBestStripHeight == 0,
	XIF_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(dwImageRowOffset % xImg->imgInfo.dwBestStripHeight,
	XIF_BADPARAMETER);

    /* Get the size and offset of this strip */
    XERR_PASS_CLEANUP(wp_XifGetCompressedSize(xImg, dwImageRowOffset,
	&offset, pdwBytes));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_CompressedReadStrip()
 *
 * OneLineDescription:::
 *  Read an arbitrary strip in compressed form
 *
 * LongDescription:::
 *  XIF_CompressedReadStrip() returns a strip of compressed data from a
 *  stored strip.  The strip size must match that stored in the file, as
 *  indicated by the dwBestStripHeight parameter passed back from
 *  XIF_GetImageInfo().
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetImageInfo()
 *  dwImgRowOffset::		Source row offset of the image proper
 *  dwStripHeight::		Number of rows to fill
 *
 * Output:::
 *  ppCmpBuffer::		Pointer to pointer to preallocated
 *				compressed data buffer (may be updated)
 *  pdwCmpBufsize::		Pointer to the number of bytes in buffer
 *  pdwBytesFilled::		Number of bytes successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_CompressedReadStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppCmpBuffer,
	UInt32 *pdwCmpBufsize,
	UInt32 *pdwBytesFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesFilled = 0UL;
    
    CheckHandleValidity(xImgHandle);
    
    pxImgInfo = &xImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_READSTATUS) !=
		      XIMG_SEQ_READSTART), XIF_SEQUENCE);
    XERR_GEN_CLEANUP((dwImgRowOffset % pxImgInfo->dwBestStripHeight) != 0 &&
      (dwStripHeight == pxImgInfo->dwBestStripHeight ||
       dwImgRowOffset + dwStripHeight == pxImgInfo->dwHeight),
      XIF_BADPARAMETER);
    
    XERR_PASS_CLEANUP(wp_XifCompressedReadStrip(xImg, dwImgRowOffset,
	ppCmpBuffer, pdwCmpBufsize, pdwBytesFilled));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* wp_CopyFromTempImage()
 *
 * OneLineDescription:::
 *  Copy image data from the intermediate image buffer to the user's buffer
 *
 * LongDescription:::
 *  wp_CopyFromTempImage() copies data stored in the intermediate image buffer,
 *  to the user's image buffer. The number of rows filled is returned.
 *
 * Input:::
 *  xImg::			Data structure of valid image, contains
 *				intermediate image buffer
 *
 * Output:::
 *  ppImgChannels::		Destination image channel(s)
 *  dstBPL::			Destination image bytes per line (for one channel)
 *  pdwImgRowOffset::		Pointer to image proper row offset, updated for
 *				outer loop
 *  pdwDstImgRowOffset::	Pointer to destination row offset, updated for
 *				outer loop
 *  pdwRowsToRead::		Pointer to the number of rows in outer loop to
 *				read
 *  pdwTotalRowsFilled::	Pointer to the total number of rows filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_CopyFromTempImage(
    XImg *xImg,
    UInt8 **ppImgChannels,
    Int32 dstBPL,
    UInt32 *pdwImgRowOffset,
    UInt32 *pdwDstImgRowOffset,
    UInt32 *pdwRowsToRead,
    UInt32 *pdwTotalRowsFilled
)
{
    UInt8 *srcPtr, *dstPtr;
    Int32 byteCount;
    IPWState *pIPWState;
    Int32 channel, channels;
    PIXR *srcPixr, *pSubPixr;
    Int32 srcBPL;
    UInt32 rowsFilled = 0UL;
    UInt32 rowsAvailable;
    PROLOGUE;

    /* Default to nothing happened */
    *pdwTotalRowsFilled = 0UL;
    
    if (xImg->tmpPixr == NULL || *pdwImgRowOffset < xImg->tmpPixrRow ||
	*pdwImgRowOffset >= xImg->tmpPixrRow + xImg->tmpPixrRows) {
	NORMAL_RETURN;
    }

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    srcPixr = xImg->tmpPixr;
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, srcPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, srcPixr, &srcBPL));
    byteCount = (xImg->minBytesPerRow + 3) & ~3;
    if (byteCount > dstBPL) {
	byteCount = dstBPL;		/* This might be a problem */
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, srcPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &srcPtr));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, srcPixr, &srcPtr));
	}
	srcPtr += (*pdwImgRowOffset - xImg->tmpPixrRow) * srcBPL;
	dstPtr = ppImgChannels[channel] + *pdwDstImgRowOffset * dstBPL;
	rowsAvailable = xImg->tmpPixrRows - (*pdwImgRowOffset - xImg->tmpPixrRow);
	if (rowsAvailable > *pdwRowsToRead) {
	    rowsAvailable = *pdwRowsToRead;
	}

	/* Copy the user's image to the PIXR image */
	while (rowsAvailable--) {
	    memcpy(dstPtr, srcPtr, byteCount);
	    srcPtr += srcBPL;
	    dstPtr += dstBPL;
	    rowsFilled++;
	}
    } /* for each channel */
    rowsFilled /= channels;
    *pdwImgRowOffset += rowsFilled;
    *pdwDstImgRowOffset += rowsFilled;
    *pdwRowsToRead -= rowsFilled;
    (*pdwTotalRowsFilled) += rowsFilled;

    /* Convert image order according to user's specifications */

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_ImageReadStrip()
 *
 * OneLineDescription:::
 *  Read an arbitrary strip of image data.
 *
 * LongDescription:::
 *  XIF_ImageReadStrip() returns a strip of image data from anywhere within
 *  the image. Appropriate buffering may be necessary to read stored strips
 *  of data required to fill the requested rows. Each stored strip is
 *  decompressed entirely before the image data is buffered out.
 *  <BR>
 *  Image data is always top-to-bottom, and left-to-right. If the client
 *  does not choose to read the entire image in as a single strip, it is
 *  advisable to use a strip height that is a multiple of "dwBestStripHeight",
 *  such that not intermediate buffering is required for partial stored strips.
 *  <BR>
 *  This routine is really a wrapper on top of w_XifReadStrip().
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetImageInfo()
 *  dwImgRowOffset::		Source row offset of the image proper
 *  dwStripHeight::		Number of rows to fill
 *
 * Output:::
 *  ppImagePlanes::		Pointer to 1-3 image planes (channels)
 *  dwBytesPerRow::		Number of bytes per row of image data in each
 *				plane
 *  dwBufferRowOffset::		Destination row offset in the image planes
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ImageReadStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 *pdwRowsFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    UInt32 totalRowsFilled = 0UL;
    ColorType ctype;
    IPWState *pIPWState;
    Int32 tmpPixrHeight = 0;
    UInt32 rowsFilled = 0UL;
    Int32 readRows;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;
    
    CheckHandleValidity(xImgHandle);
    
    pxImgInfo = &xImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_READSTATUS) !=
		      XIMG_SEQ_READSTART), XIF_SEQUENCE);
    
    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    /* If we already have an intermediate PIXR, get its height */
    if (xImg->tmpPixr != NULL) {
	XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, xImg->tmpPixr, &tmpPixrHeight));
    } /* If no intermediate image buffer */

    /* Loop until the user's image buffer (region) is filled */
    while (dwStripHeight) {

	/* Check for overfill */
	if (dwImgRowOffset + dwStripHeight > pxImgInfo->dwHeight) {
	    dwStripHeight = pxImgInfo->dwHeight - dwImgRowOffset;
	}
	if (dwStripHeight <= 0) {
	    break;
	}

	/* Fill any left over data in the intermediate image */
	XERR_PASS_CLEANUP(wp_CopyFromTempImage(xImg, ppImagePlanes, dwBytesPerRow,
	    &dwImgRowOffset, &dwBufferRowOffset, &dwStripHeight, &rowsFilled));
	totalRowsFilled += rowsFilled;

	/* Break out of loop if we're finished */
	if (dwStripHeight == 0) {
	    break;
	}

	/* Compute the number of rows to read */
	readRows = pxImgInfo->dwBestStripHeight;
	if (readRows > (Int32)pxImgInfo->dwHeight - (Int32)dwImgRowOffset) {
	    readRows = pxImgInfo->dwHeight - dwImgRowOffset;
	}
		
	/* If wavelet decompressing, the entire PIXR must be a full TIFF
	 * strip (we can't tell the wavelet routines to decompress only a portion
	 * of the PIXR). */
	if (pxImgInfo->xCompression == XIF_WAVELET && xImg->tmpPixr != NULL) {
	    if (tmpPixrHeight != readRows) {
		XERR_GEN_CLEANUP(xImg->tmpPixrRow + xImg->tmpPixrRows > dwImgRowOffset,
		    XIF_INTERNAL_ERROR);
		XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &xImg->tmpPixr));
		xImg->tmpPixr = NULL;
	    }
	}

	/* If we don't have an intermediate PIXR to handle buffering,
         * create it now */
        if (xImg->tmpPixr == NULL) {
	    ScaledInt1616 xres, yres;
		
	    /* Prepare a temporary PIXR */
	    xres = Int32_TO_ScaledInt1616(pxImgInfo->dwXResolution);
	    yres = Int32_TO_ScaledInt1616(pxImgInfo->dwYResolution);
	    switch ((int)pxImgInfo->xColorType) {
	    case XIF_COLORTYPE_COLOR24:
		ctype = cColorRGB;
		break;
	    case XIF_COLORTYPE_BINARY:
	    case XIF_COLORTYPE_GRAY4:
	    case XIF_COLORTYPE_COLOR4:
	    case XIF_COLORTYPE_GRAY8:
	    case XIF_COLORTYPE_COLOR8:
		ctype = cColorGray;
		break;
	    default:
		goto ERR_CLEANUP;
	    }
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, pxImgInfo->dwWidth,
		readRows, XIF_BPPE(pxImgInfo->xColorType),
		xres, yres, 32, XIF_CHANNELS(pxImgInfo->xColorType),
		ctype, NULL, &xImg->tmpPixr));
	    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, xImg->tmpPixr, &tmpPixrHeight));
	} /* If no intermediate image buffer */

	/* Read a TIFF strip from the file if intermediate PIXR isn't filled with
	 * the next requested image line */
	if (xImg->tmpPixrRows == 0 || xImg->tmpPixrRow > dwImgRowOffset ||
	    xImg->tmpPixrRow + xImg->tmpPixrRows <= dwImgRowOffset) {
	    XERR_PASS_CLEANUP(w_XifReadStrip(pIPWState, xImgHandle,
		dwImgRowOffset, readRows, xImg->tmpPixr, 0, &rowsFilled));
		xImg->tmpPixrRow = dwImgRowOffset;
		xImg->tmpPixrRows = rowsFilled;
	}

	/* Fill any new data in the intermediate image */
	XERR_PASS_CLEANUP(wp_CopyFromTempImage(xImg, ppImagePlanes, dwBytesPerRow,
	    &dwImgRowOffset, &dwBufferRowOffset, &dwStripHeight, &rowsFilled));
	totalRowsFilled += rowsFilled;
    }

    /* Output the total number of rows in user's image filled */
    *pdwRowsFilled = totalRowsFilled;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_ImageReadFinish()
 *
 * OneLineDescription:::
 *  Do necessary cleanup for the image read sequence.
 *
 * LongDescription:::
 *  XIF_ImageReadFinish() notifies the XIF library that the client is
 *  finished reading an image. Every call to XIF_ImageReadStart must be
 *  eventually followed by a call to this function.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetImageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ImageReadFinish(
    XIF_IMGHANDLE xImgHandle
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    MISC_MEM *mmptr, *next_mm;
    Instance *instData;
    PROLOGUE;
    
    CheckHandleValidity(xImgHandle);
    
    pxImgInfo = &xImg->imgInfo;
    instData = XIF_IMGINST(xImg);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_READSTATUS) !=
		      XIMG_SEQ_READSTART), XIF_SEQUENCE);
    
    /* Free intermediate data */
    if (xImg->cmpBuffer) {
	XIF_FREE(xImg->cmpBuffer);
	xImg->cmpBuffer = NULL, xImg->cmpBufsize = xImg->cmpBytes = 0UL;
    }
    if (xImg->tmpPixr) {
	IPWState *pIPWState;

	/* Reference the IPWrap state */
	XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

	XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &xImg->tmpPixr));
    }
    if ((mmptr = xImg->tmpCmpData) != NULL) {
	while (mmptr) {
	    if (mmptr->mem) {
		XIF_FREE(mmptr->mem);
	    }
	    next_mm = mmptr->next;
	    XIF_FREE(mmptr);
	    mmptr = next_mm;
	}
	xImg->tmpCmpData = NULL;
    }
    
    /* Reset the read status */
    xImg->seqStatus &= ~XIMG_SEQ_READSTATUS;	/* Clear read status */
    xImg->seqStatus |= XIMG_SEQ_READFINISH;		/* Set to finished */
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_GetLayerInfo()
 *
 * OneLineDescription:::
 *  Get the image description for the image region of a binary "even" layer
 *
 * LongDescription:::
 *  XIF_GetLayerInfo() gets the description of a single image that could
 *  represent the entire layer of a binary even-numbered layer: position
 *  relative to the Primary image, resolution, and size.
 *  <BR>
 *  Returned is an image handle to be used to reference this image in later
 *  calls to XIF_ImageReadLayer(). A call to XIF_CloseDocument() renders this
 *  image handle invalid.
 *
 * Input:::
 *  xPageHandle::		Page handle
 *  dwImgLayer::		Image layer (1, 2, ...)
 *
 * Output:::
 *  pxCompression::		Pointer to the compression type of all images,
 *				or XIF_COMP_UNKNOWN if not just one
 *  pdwXOffset::		Pointer to the X location in page coordinates
 *  pdwYOffset::		Pointer to the Y location in page coordinates
 *  pdwXResolution::		Pointer to the X Resolution of image in dpm
 *  pdwYResolution::		Pointer to the Y Resolution of image in dpm
 *  pdwWidth::			Pointer to the width in image's X resolution
 *				units
 *  pdwHeight::			Pointer to the height in image's Y resolution
 *				units
 *  pxImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetLayerInfo(
    XIF_PAGEHANDLE xPageHandle,
    UInt32 dwImgLayer,
    XIF_COMPRESSION *pxCompression,
    UInt32 *pdwXOffset,
    UInt32 *pdwYOffset,
    UInt32 *pdwXResolution,
    UInt32 *pdwYResolution,
    UInt32 *pdwWidth,
    UInt32 *pdwHeight,
    XIF_IMGHANDLE *pxImgHandle
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XPage *xPage = (XPage *)xPageHandle;
    UInt32 imgNum;
    UInt32 xOffset, yOffset, xRes, yRes, width, height;
    UInt32 bestStripHeight, maskImages;
    XIF_COMPRESSION compression;
    XIF_COLORTYPE colorType;
    UInt8 palette[768];
    XIF_IMGHANDLE imgHandle;
    PROLOGUE;
    
    CheckHandleValidity(xPageHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xPage->seqStatus &
	(XPAGE_SEQ_READSTATUS | XPAGE_SEQ_WRITESTATUS)) == 0),
	XIF_SEQUENCE);
    
    XERR_GEN_CLEANUP(dwImgLayer < 1 ||
	(xPage->pageInfo.dwLayersPresent > 0 &&
	(dwImgLayer > xPage->pageInfo.dwLayersPresent ||
	xPage->pageInfo.pdwImageCounts == NULL ||
	xPage->pageInfo.pdwImageCounts[dwImgLayer - 1] <= 0)),
	XIF_BADPARAMETER);
    
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);
    
    /* Clear the results */
    *pdwXOffset = *pdwYOffset = *pdwXResolution = *pdwYResolution = 0UL;
    *pdwWidth = *pdwHeight = 0UL;
    *pxCompression = XIF_COMP_UNKNOWN;

    /* Loop through the images on that layer, looking for the total
     * dimensions */
    for (imgNum = 0; imgNum < xPage->pageInfo.pdwImageCounts[dwImgLayer-1];
    imgNum++) {

	/* Get the image information for this image */
	XERR_PASS_CLEANUP(XIF_GetImageInfo(xPageHandle, dwImgLayer,
	    imgNum, &xOffset, &yOffset, &xRes, &yRes, &width, &height,
	    &bestStripHeight, &compression, &colorType,
	    palette, &maskImages, &imgHandle));
    
	/* Copy out the image information if it's the first image */
	if (imgNum == 0) {
	    *pdwXOffset = xOffset;
	    *pdwYOffset = yOffset;
	    *pdwXResolution = xRes;
	    *pdwYResolution = yRes;
	    *pdwWidth = width;
	    *pdwHeight = height;
	    /* Use the first image of the layer as the reference */
	    *pxImgHandle = imgHandle;
	    *pxCompression = compression;
	} else {

	    /* Check the parameters that should be consistent */
	    XERR_GEN_CLEANUP(xRes != *pdwXResolution, XIF_BADFILEFORMAT);
	    XERR_GEN_CLEANUP(yRes != *pdwYResolution, XIF_BADFILEFORMAT);

	    /* If the dimensions grow the size of the layer image, update
	    * the dimensions */
	    if (xOffset < *pdwXOffset) {
		*pdwWidth += *pdwXOffset - xOffset;
		*pdwXOffset = xOffset;
	    }
	    if (yOffset < *pdwYOffset) {
		*pdwHeight += *pdwYOffset - yOffset;
		*pdwYOffset = yOffset;
	    }
	    if (xOffset + width > *pdwXOffset + *pdwWidth) {
		*pdwWidth = xOffset + width - *pdwXOffset;
	    }
	    if (yOffset + height > *pdwYOffset + *pdwHeight) {
		*pdwHeight = yOffset + height - *pdwYOffset;
	    }

	    /* Add in the compression type to the list (unique bits) */
	    *pxCompression |= compression;
	}
    }  /* For each image in the layer */
	
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_ImageReadLayer()
 *
 * OneLineDescription:::
 *  Get the image data used to represent one image for an even layer
 *
 * LongDescription:::
 *  XIF_ImageReadLayer() returns the entire image associated with all the
 *  images belonging to an even numbered layer, as described by the image
 *  return via XIF_GetLayerInfo().
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetLayerInfo()
 *  dwImageHeight::		Number of rows to fill
 *
 * Output:::
 *  pImageBuffer::		Pointer to single binary image plane
 *  dwBytesPerRow::		Number of bytes per row of image data
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function allows piecing together the various images separated by
 *  the tokenization and lossy dither algorithms: residual, tokenized,
 *  and lossy dithered halftones.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ImageReadLayer(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImageHeight,
	UInt8 *pImageBuffer,
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_COMPRESSION compression;
    UInt32 layerXOffset, layerYOffset, layerXRes, layerYRes;
    UInt32 layerWidth, layerHeight;
    IPWState *pIPWState;
    ScaledInt1616 xres, yres;
    PIXR *pPixr = NULL;
    UInt32 row, rowsFilled, bpl;
    Int32 srcBpl, dstBpl;
    UInt8 *src, *dst;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;
    
    CheckHandleValidity(xImgHandle);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->xPage->seqStatus & XPAGE_SEQ_READSTATUS) !=
		      XPAGE_SEQ_READSTART), XIF_SEQUENCE);
    /* Should be even layer */
    XERR_GEN_CLEANUP(xImg->imgInfo.dwImgLayer & 1, XIF_BADPARAMETER);

    /* Get the characteristics of the layer image */
    XERR_PASS_CLEANUP(XIF_GetLayerInfo((XIF_PAGEHANDLE)xImg->xPage,
	xImg->imgInfo.dwImgLayer, &compression, &layerXOffset, &layerYOffset,
	&layerXRes, &layerYRes, &layerWidth, &layerHeight, &xImgHandle));

    /* There should be some consistency */
    XERR_GEN_CLEANUP(xImgHandle != (XIF_IMGHANDLE)xImg, XIF_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(layerHeight != dwImageHeight, XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(dwBytesPerRow >= layerWidth * 32, XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(pImageBuffer == NULL, XIF_BADPARAMETER);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    /* Create a PIXR to hold the composite image */
    xres = Int32_TO_ScaledInt1616(layerXRes);
    yres = Int32_TO_ScaledInt1616(layerYRes);
    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, layerWidth,
    	layerHeight, 1, xres, yres, 32, 1, cColorGray, NULL, &pPixr));

    /* Populate the PIXR with the image data */
    XERR_PASS_CLEANUP(w_XifImageReadLayer(pIPWState, xImgHandle, pPixr,
	&rowsFilled));
    XERR_GEN_CLEANUP(rowsFilled != layerHeight, XIF_INTERNAL_ERROR);

    /* Copy the PIXR image data to the client image */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &srcBpl));
    dstBpl = dwBytesPerRow;
    bpl = MIN(srcBpl, dstBpl);
    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &src));
    dst = pImageBuffer;
    for (row = 0; row < layerHeight; row++) {
	memcpy(dst, src, bpl);
	dst += dstBpl;
	src += srcBpl;
    }
    
    /* Cleanup */
    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pPixr));
    
    /* Output the total number of rows in user's image filled */
    *pdwRowsFilled = layerHeight;

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pPixr));
    ERR_RETURN;
}


/* XIF_GetAnnotationInfo()
 *
 * OneLineDescription:::
 *  Get the characteristics of an annotation
 *
 * LongDescription:::
 *  XIF_GetAnnotationInfo() reads the data necessary to describe an
 *  annotation. This includes the annotation: type, position, flags, size,
 *  and number of bytes of raw data.
 *
 * Input:::
 *  xPageHandle::		Valid page handle, as returned from
 *				XIF_GetPageInfo()
 *  wAnnotOffset::		Absolute offset of this annotation on the
 *				page
 *
 * Output:::
 *  pwAnnotType::		Pointer to the annotation type
 *  pdwXOffset::		Pointer to the X offset
 *  pdwYOffset::		Pointer to the Y offset
 *  pdwWidth::			Pointer to the width
 *  pdwHeight::			Pointer to the height
 *  pdwFlags::			Pointer to the flags
 *  pdwAnnotLength::		Pointer to the number of data bytes
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetAnnotationInfo(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 wAnnotOffset,
	UInt16 *pwAnnotType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwFlags,
	UInt32 *pdwAnnotLength
)
{
    XPage *xPage = (XPage *)xPageHandle;
    AnnotTable *annotTable;
    AnnotTableEntry *annot;
    UInt32 entryOffset;
    PROLOGUE;
    
    CheckHandleValidity(xPageHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((xPage->seqStatus & XPAGE_SEQ_GETINFO) == 0,
		      XIF_SEQUENCE);

    /* Read the annotation headers, and extract the data for this particular
     * annotation. Note that XIFF 3.0 and higher will extract a set of
     * annotation descriptions all at once, when reading an annotation table.
     * For XIFF 2.0, the information isn't grouped in a table, and has to
     * be read individually */
    XERR_PASS(wp_XifGetAnnotTable(xPage, wAnnotOffset, &annotTable));

    /* Return the results */
    entryOffset = wAnnotOffset % xPage->annotTableHeader.entriesPerTable;
    annot = &annotTable->entries[entryOffset];
    *pwAnnotType = annot->type;
    *pdwXOffset = annot->xOffset;
    *pdwYOffset = annot->yOffset;
    *pdwWidth = annot->width;
    *pdwHeight = annot->height;
    *pdwFlags = annot->flags;
    *pdwAnnotLength = annot->bytes;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_GetAnnotationData()
 *
 * OneLineDescription:::
 *  Get the characteristics of an annotation
 *
 * LongDescription:::
 *  XIF_GetAnnotationData() reads the data associated with an annotation.
 *
 * Input:::
 *  xPageHandle::		Valid page handle, as returned from
 *				XIF_GetPageInfo()
 *  wAnnotOffset::		Absolute offset of this annotation on the
 *				page
 *
 * Output:::
 *  dwBytes::			Number of bytes to retrieve (shouldn't exceed
 *				the stored amount)
 *  pData::			Buffer to occupy annotation data
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_GetAnnotationData(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 wAnnotOffset,
	UInt32 dwBytes,
	UInt8 *pData
)
{
    XPage *xPage = (XPage *)xPageHandle;
    PROLOGUE;
    
    CheckHandleValidity(xPageHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((xPage->seqStatus & XPAGE_SEQ_GETINFO) == 0,
		      XIF_SEQUENCE);

    /* Read the annotation data */
    XERR_PASS(wp_XifGetAnnotData(xPage, wAnnotOffset, dwBytes, pData));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_CloseDocument()
 *
 * OneLineDescription:::
 *  Do the tasks necessary to complete communication to/from a document.
 *
 * LongDescription:::
 *  XIF_CloseDocument() will, if necessary, complete the processes associated
 *  with the storage of a partially written document, including completion of
 *  the Tokenization process (see XIF_WriteTokenDictionary() and
 *  XIF_WriteTokenImage()). All internal data associated with an open
 *  document is freed.
 *
 * Input:::
 *  xDocHandle::		Active document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_CloseDocument(
    XIF_DOCHANDLE xDocHandle
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    Instance *instData;
    IPWState *pIPWState;
    PROLOGUE;

    /* Get an pointer to the IPState, it's needed below */
    XERR_PASS (wp_XifIPWState (xDoc, &pIPWState));
    
    CheckHandleValidity(xDocHandle);
    instData = XIF_DOCINST(xDoc);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xDoc->seqStatus &
		       (XDOC_SEQ_READSTART | XDOC_SEQ_WRITESTART)) == 0),
		     XIF_SEQUENCE);

#ifndef XIF_READONLY

    /* Flush the token dictionary and resolve and write the images
     * that are in a tokenized state. */
    XERR_PASS_CLEANUP(XIF_WriteTokenImages(xDocHandle));

    /* For XIF 2.0 files, make sure we have stored a page table. In the
     * case of a 0-page document, a page table is still required.
     */
    if (xDoc->docInfo.version == 2 && xDoc->docInfo.wNPages == 0 &&
	(xDoc->seqStatus & XDOC_SEQ_WRITESTART)) {
	XERR_PASS_CLEANUP(wp_XifSetPageIFD(xDoc, 0UL, 0, 0UL, NULL));
    }

#endif /* ! XIF_READONLY */
    
    /* Fall through to do the rest of the cleanup */
    
 ERR_CLEANUP:
    
    /* Free up the shared data */
    if (xDoc->textCompInfo) {
	i_freeTextCompressionInfo (pIPWState,
				   xDoc->textCompInfo); /* Ignore return */
	xDoc->textCompInfo = NULL;
    }

    /* Clear the status, document state reset */
    xDoc->seqStatus = 0;

    if (xDoc) {
	wp_XifDocFree(&xDoc);
    }
    
    NORMAL_RETURN;
}


/* XIF_Decompress()
 *
 * OneLineDescription:::
 *  Decompress an image strip as given from XIF_ImageReadCompressed()
 *
 * LongDescription:::
 *  XIF_Decompress() decompresses an image strip returned from
 *  XIF_ImageReadCompressed().
 *
 * PrivateDescription:::
 *  This is a front-end to a structure oriented routine w_XifDecompress().
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetImageInfo()
 *  xCompression::		See XIF_COMPRESSION types
 *  xColorType::		Color type of image data
 *  dwFlags::			Destination image format (see XIF_IMGFLAGS_xxx)
 *  pSrc::			Source buffer (compressed)
 *  dwBytesInSrc::		Byte count of source buffer
 *
 * Output:::
 *  dwRowOffset::		Destination row offset
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwHeight::			Rows in dst image
 *  pDst[3]::			Pointer to destination image buffer channels
 *  dwBytesPerRow::		Bytes per row in image buffer
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_Decompress(
	XIF_IMGHANDLE xImgHandle,
	XIF_COMPRESSION xCompression,
	XIF_COLORTYPE xColorType,
	UInt32 dwFlags,
	UInt8 *pSrc,
	UInt32 dwBytesInSrc,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt8 *pDst[3],
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;
    ScaledInt1616 xres, yres;
    Int32 rowsLeft;
    UInt32 rowsFilled = 0UL;
    Int32 pixrBpl;
    UInt32 bytesPerRow;
    IPWState *pIPWState;
    PIXR *pPixr = NULL, *pSubPixr;
    ColorType ctype = cColorNone;
    UInt8 *pUserImage, *pPixrImage;
    Int32 channel, channels;
    PROLOGUE;

    *pdwRowsFilled = 0UL;
    CheckHandleValidity(xImgHandle);

    /* Get an pointer to the IPState, it's needed below */
    XERR_PASS (wp_XifIPWState (xImg->xPage->xDoc, &pIPWState));

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    /* Check to see if there's anything to do */
    if (dwBytesInSrc == 0) {
	NORMAL_RETURN;
    }

    /* Check buffers and sizes */
    XERR_GEN_CLEANUP(pSrc == NULL || pDst == NULL || pDst[0] == NULL ||
	dwBytesPerRow == 0, XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(dwRowOffset >= xImg->imgInfo.dwHeight, XIF_BADPARAMETER);
    
    /* Create a PIXR image to hold the image */
    xres = Int32_TO_ScaledInt1616(xImg->imgInfo.dwXResolution);  /* Any resolution is fine, I guess */
    yres = Int32_TO_ScaledInt1616(xImg->imgInfo.dwYResolution);  /* Any resolution is fine, I guess */
    switch (xColorType) {
    case XIF_COLORTYPE_COLOR24:
	ctype = cColorRGB;
	break;
    case XIF_COLORTYPE_BINARY:
    case XIF_COLORTYPE_GRAY4:
    case XIF_COLORTYPE_COLOR4:
    case XIF_COLORTYPE_GRAY8:
    case XIF_COLORTYPE_COLOR8:
	ctype = cColorGray;
	break;
    default:
	break;
    }
    channels = XIF_CHANNELS(xColorType);
    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, dwWidth, dwHeight,
	XIF_BPPE(xColorType), xres, yres, 32, channels, ctype, NULL, &pPixr));

    /* Call the underlying decompression routine */
    XERR_PASS_CLEANUP(w_XifDecompress(pIPWState, xImgHandle, xCompression,
	xColorType, dwFlags, pSrc, dwBytesInSrc, 0, dwWidth, dwHeight, pPixr,
	&rowsFilled));
    
    /* Copy the data to the user's buffer */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &pixrBpl));
    bytesPerRow = dwBytesPerRow;
    if (bytesPerRow > pixrBpl) {
	bytesPerRow = pixrBpl;
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &pPixrImage));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &pPixrImage));
	}

	rowsLeft = rowsFilled;
	pUserImage = pDst[channel] + dwRowOffset * dwBytesPerRow;
	while (rowsLeft > 0) {
	    memcpy(pUserImage, pPixrImage, bytesPerRow);
	    pUserImage += dwBytesPerRow;
	    pPixrImage += pixrBpl;
	    rowsLeft--;
	}
    }

    *pdwRowsFilled = rowsFilled;
    w_destroyPixr(pIPWState, &pPixr);

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    w_destroyPixr(pIPWState, &pPixr);
    
    ERR_RETURN;
}

/*
$Log: xifread.c,v $
Revision 1.2  2000/04/12 23:22:54  dabercro
Allow color JBIG2 as a valid compression.

Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.45  1998/10/28 20:36:04  aber
aber on Wed Oct 28 12:04:08 PST 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Enable compression of 8-bit gray using Wavelet compression.

Revision 1.44  1998/10/20 18:17:59  aber
aber on Tue Oct 20 11:17:04 PDT 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

When checking for compression consistency, allow lossy dither compression
for binary images (forgot to include it in last checkin).

Revision 1.43  1998/08/28 20:30:57  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Detect the inconsistency of the unsupported color tokenization, and
return a NOSUPPORT error for tokenized images that aren't binary. This
was practically impossible to anticipate, but should be gracefully
handled in all future releases.

Revision 1.42  1998/08/04 00:21:32  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.41  1998/07/30 22:39:42  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.40  1998/07/27 19:05:53  aber
aber on Mon Jul 27 12:04:31 PDT 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Temporary fix for stuff I'm about to check in.

Revision 1.39  1998/07/27 17:36:16  aber
aber on Mon Jul 27 10:35:43 PDT 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Use private memory tally structure.

Revision 1.38  1998/05/05 07:15:21  aber
aber on Tue May 5 00:15:21 PDT 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Additions necessary to fill in the never completed functions
XIF_Compress and XIF_Decompress.

Revision 1.37  1998/04/06 22:43:45  aber
aber on Mon Apr 6 15:43:44 PDT 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Bug fix: check resolution difference when checking the image size of
the image during read.
Also, check for unanticipated condition of the page resolution being
SMALLER than than the subimage resolution. Adjust subimage dimensions
accordingly.

Revision 1.36  1998/03/25 23:25:14  aber
aber on Wed Mar 25 15:25:13 PST 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Use a unique document ID for each document handle. This allows tagging
each document so that during "quick copy", the token dictionaries are
copy only once, and only once (in the case of two destination documents
somehow ending up with the same memory location for the structure data).

Revision 1.35  1998/02/09 20:21:23  aber
aber on Mon Feb 9 12:21:23 PST 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.34  1998/01/20 03:20:56  aber
aber on Mon Jan 19 19:18:52 PST 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Before reading a tokenized image, make sure it's flushed to disk.

Revision 1.33  1998/01/16 00:39:45  aber
aber on Thu Jan 15 16:38:01 PST 1998

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Allow page inserts.

Revision 1.32  1998/01/14 00:16:25  ddavies
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.31  1997/11/24 21:06:32  aber
aber on Mon Nov 24 13:04:42 PST 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Change calling sequence requirements for GetAnnotationData
to have previously gotten the page parameters.
(Was requiring to have been in read mode, but is a valid operation
just after writing).

Revision 1.30  1997/11/19 21:42:10  aber
aber on Wed Nov 19 13:41:32 PST 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Allow getting annotation info while writing.

Revision 1.29  1997/11/18 23:37:01  aber
aber on Tue Nov 18 15:36:16 PST 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Add new function: XIF_WriteTokenImages() -- which will complete
the tokenization process, write the token dictionary, and write
all token images which haven't been stored yet.
Add annotation table element: flags.
Add annotation header element: bytes per element.

Revision 1.28  1997/11/17 19:25:58  aber
aber on Mon Nov 17 11:25:58 PST 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Allow negative page offsets to automatically append the page to
the end of the document.

Revision 1.27  1997/11/14 02:08:57  aber
aber on Thu Nov 13 18:08:57 PST 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.26  1997/11/01 20:49:08  aber
aber on Sat Nov 1 12:48:29 PST 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Force buffering strips exactly equal to TIFF strip size -- allows
exact buffer for wavelet strip.

Revision 1.25  1997/10/24 21:38:54  aber
aber on Fri Oct 24 14:38:53 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.24  1997/10/23 18:53:44  aber
aber on Thu Oct 23 11:53:18 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

> should be >= when checking for a valid page offset.

Revision 1.23  1997/10/16 16:15:47  aber
aber on Thu Oct 16 09:15:47 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Remove carriage-returns (left over from PC).

Revision 1.22  1997/10/15 22:14:13  aber
aber on Wed Oct 15 15:12:07 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Store the page table in XIF 2 file, if no pages have been saved.

Revision 1.21  1997/10/14 17:01:44  aber
aber on Tue Oct 14 10:00:57 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Assumption made about the number of dictionaries referenced by
a single page was wrong. Assumption removed.

Revision 1.20  1997/10/01 23:35:04  aber
aber on Wed Oct 1 16:34:06 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Simple cleanup, doesn't affect anything relevant.

Revision 1.19  1997/08/29 17:27:59  aber
aber on Fri Aug 29 10:27:59 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.18  1997/08/13 22:48:18  aber
aber on Wed Aug 13 15:48:18 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Edit fix via Pedro (XResolution should be Yresolution).

Revision 1.17  1997/08/13 22:18:44  aber
aber on Wed Aug 13 15:18:32 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Allow page info query when writing.

Revision 1.16  1997/08/07 21:27:43  aber
aber on Thu Aug 7 14:27:43 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.15  1997/07/29 23:49:41  aber
aber on Tue Jul 29 16:47:16 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Add function XIF_GetInstance(), to retrieve the XIFF instance associated
with a document handle. Used to detach the instance when closing the
document.
Added function XIF_GetImageStats(), which retrieves the total compressed
size of the image, and the size of the token dictionary, if tokenized.
Allow requests for compressed data size without calling XIF_ImageReadStart().

Revision 1.14  1997/07/15 23:42:09  aber
aber on Tue Jul 15 16:41:19 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Some buffering fixes.

Revision 1.13  1997/07/02 01:03:59  aber
aber on Tue Jul 1 18:01:39 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Make memory functions match IPWrap. Expand the return values for
IsXiff to indicate: no support, not a TIFF, not a XIFF, and read/seek error.
Complete generic function XIF_ImageReadLayer(), which buffers a PIXR and
calls w_XifImageReadLayer().

Revision 1.12  1997/06/20 17:43:37  aber
aber on Fri Jun 20 10:43:12 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

More checking for corrupt XIF files.

Revision 1.11  1997/06/18 00:41:28  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Updated prototype for XIF_GetLayerInfo.

Revision 1.10  1997/06/17 01:29:15  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: XIFF 2.0 support. Annotation bug fix.

Revision 1.9  1997/06/12 00:49:42  aber
aber on Wed Jun 11 17:49:11 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

Fix row check for strip alignment.

Revision 1.8  1997/06/10 01:26:37  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.6  1997/05/30 22:11:31  aber
aber on Fri May 30 15:11:31 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Update documentation for web browsing. XIF_ functions are left
with no "BeginDocumentation" tag, since w_ and wp_ functions are expected.

Revision 1.5  1997/05/30 00:50:58  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:02  aber
xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:09  aber
aber on Mon May 19 17:54:09 PDT 1997

xifread.c -> /product/ipcore/ipwrap/src/RCS/xifread.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:34  aber
Initial revision

*/
