/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifrjpeg.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifrjpeg.c:	XIFF READ JPEG source
 *
 *	XIF read-only JPEG routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "jpeglib.h"
#include "iw_jpeg.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifrtiff.h"
#include "xifrjpeg.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifrjpeg.c,v $; $Revision: 1.3 $; $Date: 2000/06/12 22:06:05 $");

typedef struct ReadData {
    UInt8 *buffer;
    UInt8 *next;
    UInt32 bufBytes;
    UInt32 bytesLeft;
} ReadData;

static ErrCode CDECL
wp_JPEG_ReadCallback(
    IPWState *pIPWState,
    void *userData,
    UInt8 *ptr,
    UInt32 bytes,
    UInt32 *ncounts
)
{
    ReadData *readData = (ReadData *)userData;
    PROLOGUE;
    pIPWState = pIPWState;
    
    if (readData->bytesLeft < bytes) {
	bytes = readData->bytesLeft;
    }
    memcpy(ptr, readData->next, bytes);
    readData->next += bytes;
    readData->bytesLeft -= bytes;
    
    *ncounts = bytes;
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompresses a strip of JPEG data.
 *
 * LongDescription:::
 *  wp_XifDecompressStripJPEG() decompresses a strip of JPEG data. Temporary
 *  PIXR found in xImg is used to store the image data decompressed.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *  pPixr::			Pointer to image PIXR
 *  dwStripHeight::		Number of rows to decompress
 *  dwRowOffset::		Destination row offset
 *
 * Output:::
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifDecompressStripJPEG(
    XImg *xImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    PIXR *pPixr,
    UInt32 dwStripHeight,
    UInt32 dwRowOffset,
    UInt32 *pdwRowsFilled
)
{
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState;
    Int32 rowsToRead = (Int32)dwStripHeight;
    UInt32 totalRowsFilled = 0UL;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8 *image[3];
    Int32 w, h, bpl, ncomps;
    Int32 row, line;
    Int32 nlines;
    Int32 channel, channels;
    UInt32 memorg = 0UL;
    JSAMPIMAGE userBuf = NULL;
    ReadData readData;
    PROLOGUE;
    
    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));
    XERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pPixr, &bpl));

    pxImgInfo = &xImg->imgInfo;
    
    /* Initialize */
    readData.buffer = pCmpData;
    readData.next = readData.buffer;
    readData.bufBytes = dwCmpBytes;
    readData.bytesLeft = readData.bufBytes;
    switch (XIF_CHANNELS(xImg->imgInfo.xColorType)) {
    case 3:
	memorg = JMEMORG_BSQ_PIXR;
	break;
    case 1:
	switch (XIF_BPPE(xImg->imgInfo.xColorType)) {
	case 4:
	    memorg = JMEMORG_GRAY4_PIXR;
	    break;
	case 8:
	    memorg = JMEMORG_GRAY8_PIXR;
	    break;
	}
	break;
    }
    XERR_PASS_CLEANUP(w_jdSetProgressiveDisplay(pIPWState, cinfo,
						cFalse));
    XERR_PASS_CLEANUP(w_jdInit(pIPWState, cinfo, wp_JPEG_ReadCallback,
		  (void *)&readData, NULL, memorg, 1.0f,
		  &w, &h, &ncomps, 0));
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    XERR_GEN_CLEANUP(w > (Int32)((xImg->imgInfo.dwWidth + 31) & ~31),
	XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(h < (Int32)dwStripHeight, XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(ncomps != channels, XIF_BADPARAMETER);

    XERR_GEN_CLEANUP(h < rowsToRead, XIF_BADPARAMETER);
    
    /* Get the location of the image and the line pitch */
    {
	PIXR  *pSubPixr;
	for (channel = 0; channel < channels; channel++) {
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				&(image[channel])));
	    image[channel] += dwRowOffset * bpl;
	}
    }
    
    XERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState, cinfo, cFalse, 
		                     	       &userBuf));
    XERR_GEN_CLEANUP(userBuf == NULL, XIF_INTERNAL_ERROR);
    
    /* Decompress the image an MCU row at a time */
    XERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row < rowsToRead; ++row) {
	for (channel = 0; channel < channels; channel++) {
	    userBuf[channel][line] = image[channel];
	    image[channel] += bpl;
	}
	++line;
	if (line >= nlines || row == (h-1)) {    
	    /* Decompress an MCU row into userBuf */
	    XERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }
    
    XERR_PASS_CLEANUP(w_jdMCURowFree(pIPWState, cinfo, cFalse, userBuf));
    userBuf = NULL;
    
    XERR_PASS_CLEANUP( w_jdTerm(pIPWState, cinfo));

    totalRowsFilled = row * channels;
    *pdwRowsFilled = totalRowsFilled;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (userBuf != NULL)
	w_jdMCURowFree(pIPWState, cinfo, cFalse, userBuf);
    
    jpeg_destroy_decompress(pIPWState, &cinfo->j_dinfo);
    
    w_jdFreeErrorStr(pIPWState, cinfo);
    
    ERR_RETURN;
}

/*
$Log: xifrjpeg.c,v $
Revision 1.3  2000/06/12 22:06:05  jlin
change jpeg read write callback function

Revision 1.2  2000/02/23 20:18:38  dabercro
Use added parameter for options to JPEG decoding (enables specifying the color space, ITULAB).

Revision 1.13  1998/05/08 21:51:41  aber
aber on Fri May 8 14:50:35 PDT 1998

xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

To avoid purify warning, set the progressive display option explicitly;
although it is never being used, at least this avoids a complaint about
an UMR.

Revision 1.12  1998/01/14 00:16:29  ddavies
xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.11  1997/09/02 20:02:53  aber
xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.10  1997/07/14 23:21:28  aber
aber on Mon Jul 14 16:20:42 PDT 1997

xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

Update error checking. Avoid duplicate free when error occurs.

Revision 1.9  1997/07/10 21:52:04  aber
aber on Thu Jul 10 14:51:26 PDT 1997

xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

Use GEN_CLEANUP instead of PASS_CLEANUP, to indicate the
problem instead of just returning an error.

Revision 1.8  1997/06/26 22:25:12  aber
xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

Allow reading JPEG images with multiple tiles.

Revision 1.7  1997/06/10 17:59:10  aber
aber on Tue Jun 10 10:58:58 PDT 1997

xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

Remove commented out section of code.

Revision 1.6  1997/06/10 01:26:39  aber
xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:00  aber
xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:04  aber
xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:11  aber
xifrjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifrjpeg.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:35  aber
Initial revision

*/
