/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1997-1998, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifronly.c
 *
 *	XIFF read-only stubs (mainly for stubbing pieces of functions that
 *		are shared by read and write calls)
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifronly.c:	XIFF READ-ONLY source
 *
 *	Handles read-only stubs.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xiff.h"
#include "xiffile.h"
#include "xifimage.h"
#include "xifwxiff.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifronly.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");


/* Stub for XIF_WriteTokenImages()
 *****************************************************************************/
ErrCode CDECL
XIF_WriteTokenImages(
	XIF_DOCHANDLE xDocHandle
)
{
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    NORMAL_RETURN;
}


/* Stub for XIF_WriteTokenImage()
 *****************************************************************************/
ErrCode CDECL
XIF_WriteTokenImage(
	XIF_IMGHANDLE xImgHandle
)
{
    PROLOGUE;

    CheckHandleValidity(xImgHandle);

    NORMAL_RETURN;
}


/* Stub for XIF_WriteTokenDictionary()
 *****************************************************************************/
ErrCode CDECL
XIF_WriteTokenDictionary(
	XIF_DOCHANDLE xDocHandle
)
{
    PROLOGUE;

    CheckHandleValidity(xDocHandle);


    NORMAL_RETURN;
}


/* Stub for XIF_SetAuthor()
 *****************************************************************************/
ErrCode CDECL
XIF_SetAuthor(
	XIF_DOCHANDLE xDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
)
{
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    /* Avoid compiler complaints about unused variables */
    if (pAuthorBlock == NULL || wAuthorBytes == 0) {
      NORMAL_RETURN;
    }

    NORMAL_RETURN;
} 


/* Stub for XIF_SetReferencePage()
 *****************************************************************************/
ErrCode CDECL
XIF_SetReferencePage(
	XIF_DOCHANDLE xDocHandle,
	Int32 dwReferencePageOffset
)
{
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    /* Avoid compiler complaints about unused variables */
    dwReferencePageOffset = 0;

    NORMAL_RETURN;
} 


/* Stub for XIF_SetAnnotTableEntryCount()
 *****************************************************************************/
ErrCode CDECL
XIF_SetAnnotTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
)
{
    PROLOGUE;

    CheckHandleValidity(xDocHandle);
    
    /* Avoid compiler complaints about unused variables */
    wEntriesPerTable = 0;

    NORMAL_RETURN;
}


/* Stub for XIF_SetPageTableEntryCount()
 *****************************************************************************/
ErrCode CDECL
XIF_SetPageTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
)
{
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    /* Avoid compiler complaints about unused variables */
    wEntriesPerTable = 0;
    
    NORMAL_RETURN;
}


/* Stub for wp_XifSetPageIFD()
 *****************************************************************************/
ErrCode CDECL
wp_XifSetPageIFD(
	      XDoc *xDoc,
	      UInt32 pageOffset,
	      UInt8 flags,
	      UInt32 ifdOffset,
	      XPage *xPage
)
{
    PROLOGUE;
    
    /* Check parameters / mode */
    CheckHandleValidity(xDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xDoc->seqStatus & XDOC_SEQ_WRITESTART) == 0),
		     XIF_SEQUENCE);
    XERR_GEN_CLEANUP(ifdOffset == 0, XIF_BADPARAMETER);

    /* We can only add the next page, not allow a gap in the pages */
    XERR_GEN_CLEANUP((pageOffset > xDoc->pageTableHeader.pageCount + 1),
		     XIF_BADPARAMETER);
    XERR_GEN_CLEANUP((xDoc->readWrite & WRITE_ALLOWED) == 0,
		     XIF_INTERNAL_ERROR);

    /* Since this is a stub, it should never be called anyway. Check these
     * values, to keep the compiler from complaining */
    CheckHandleValidity(xPage);
    XERR_GEN_CLEANUP(flags != 0, XIF_BADPARAMETER);

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    ERR_RETURN;
}


/*
$Log: xifronly.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.2  1998/02/12 20:09:16  aber
aber on Thu Feb 12 12:08:46 PST 1998

xifronly.c -> /product/ipcore/ipwrap/src/RCS/xifronly.c,v

Check all parameters, to keep Watcom happy.

Revision 1.1  1998/02/11 23:30:38  aber
Initial revision

*/
