/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifrpixr.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifrpixr.c:	XIFF READ PIXR source
 *
 *	Handles read-only XIFF PIXR
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrcmprs.pub"
#include "color.pub"

#include "ipstate.pub"
#include "morphg.h"
#include "streldef.h"
#include "strel.pub"
#include "morph.pub"
#include "logicops.h"
#include "logicops.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"

#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xiff.pub"
#include "xifimage.h"
#include "xifrtiff.h"
#include "xifrxiff.h"
#include "xifrpriv.h"
#include "xifrjpeg.h"
#include "xifrwave.h"
#include "xifrtok.h"
#include "xifconv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifrpixr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");



#if 0
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize a XIF instance, used to reference files and memory.
 *
 * LongDescription:::
 *  w_XifInitInstance()
 *
 * Input:::
 *  pIPWState::			IPW state
 *
 * Output:::
 *  pxInst::			XIF instance returned (passed to
 *				w_XifEndInstance() when done)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifInitInstance(
	IPWState *pIPWState,
	UInt32 dwClientInstID,
	XIF_INSTHANDLE *pxInst
)
{
    Instance *instData = NULL;
    MemoryTally *tally = NULL;
    PROLOGUE;

    /* Here's the only calls to the system malloc */
    CHECK_MALLOC(instData = (Instance *)malloc(sizeof(*instData) +
				       sizeof(*tally)));
    tally = (MemoryTally *)((char *)instData + sizeof(*instData));

    memset(instData, 0, sizeof(*instData));
    memset(tally, 0, sizeof(*tally));

    /* Set defaults */
    instData->handleID = (UInt32)instData;	/* Consistency check */
    instData->xiffMagicNum = XIFF_MAGIC_NUM;

    /* Preserve the default ID for switching back and forth between the
     * client ID and the reserved default */
    instData->defaultMallocID = (UInt32)tally;

    /* Save user identifier */
    instData->clientInstID = dwClientInstID;

    /* Initialize the malloc functions to the defaults */
    if (xMalloc && xFree) {	/* Must both be set to use clients */
	instData->malloc = xMalloc;
	instData->free = xFree;
	instData->mallocID = dwClientMallocID;
    } else if (!xMalloc || !xFree) {
	__error__ = XIF_BADPARAMETER;
	goto ERR_CLEANUP;
    } else {

	/* NULL functions indicate to use the defaults */
	instData->malloc = XifMalloc;
	instData->free = XifFree;
	instData->mallocID = instData->defaultMallocID;
    }

    /* Return what we did, obscured from the client */
    *pxInst = (XIF_INSTHANDLE)instData;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (instData) {
	free(instData);
    }
    if (tally) {
	free(tally);
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  End a XIF instance.
 *
 * LongDescription:::
 *   w_XifEndInstance() ends a previously established XIF instance.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  xInst::			XIF instance
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifEndInstance(
	IPWState *pIPWState,
	XIF_INSTHANDLE xInst
)
{
    Instance *instData = (Instance *)xInst;
    PROLOGUE;

    CheckHandleValidity(xInst);

    /* End IPCore instance */
    if (instData->pIPWState) {
	w_IPWStateTerm((IPWState **)&instData->pIPWState);
    }

    /* Check known XIFF memory for memory leaks */
    if (XifMemoryCheck((MemoryTally *)instData->defaultMallocID)) {
	/* Assume something wasn't called correctly */
	__error__ = XIF_SEQUENCE;
	/* Contine on with dealloc */
    }

    /* Due to potential realloc() calls, some operating systems allow this
     * pointer to propogate. Invalidate before we free it. */
    memset((void *)instData->defaultMallocID, 0, sizeof(MemoryTally));
    memset(instData, 0, sizeof(*instData));

    if (instData->defaultMallocID) {
	free((void *)instData->defaultMallocID);
    }
    free(instData);

    if (__error__ != XIF_NOERROR) {
	goto ERR_CLEANUP;    /* This is silly, but now we reference the label */
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}
#endif


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Share an IPWrap state structure pointer with the XIF layer
 *
 * LongDescription:::
 *  w_XifShareIPWState() propogates an IPWrap state structure to the XIF layer.
 *  This can be necessary when using a mixture of XIF_ functions with the w_
 *  functions. Noteably, when two underlying memory allocators are used, one
 *  as passed to w_IPWStateInit() during the creation of the pIPWState passed
 *  to the w_ functions, and a different memory allocator passed in to
 *  XIF_SetMallocFuncs().
 *
 * Input:::
 *  instHandle::		Instance handle, returned from
 *				XIF_InitInstance()
 *  pIPWState::			Pointer to the IPWrap state to be shared
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifShareIPWState(
	XIF_INSTHANDLE instHandle,
	IPWState *pIPWState
)
{
    Instance *instance = (Instance *)instHandle;
    PROLOGUE;

    CheckHandleValidity(instHandle);
    XERR_GEN_CLEANUP(pIPWState == NULL, XIF_BADPARAMETER);

    instance->pIPWState = (UInt32)pIPWState;	/* A one-liner does all the work */

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy image data stored in PIXR buffer to output image buffer.
 *
 * LongDescription:::
 *  wp_CopyFromTempPixr() copies image data temporarily stored in the intermediate
 *  PIXR, to the output buffer. Image position and rows left maintained in an
 *  outer loop are updated.
 *  <BR>
 *  If no intermediate buffer is used, there's no reason to copy.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  dstPixr::			Destination PIXR
 *  pdwDstPixrRowOffset::	Pointer to the destination PIXR row offset
 *  pdwImgRowOffset::		Pointer to the image proper row offset
 *  pdwTotalRowsFilled::	Pointer to the number of rows filled in outer
 *				loop
 *  pdwRowsToRead::		Pointer to the number of rows to fill in
 *				outer loop
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_CopyFromTempPixr(
    XImg *xImg,
    PIXR *dstPixr,
    UInt32 *pdwDstPixrRowOffset,
    UInt32 *pdwImgRowOffset,
    UInt32 *pdwTotalRowsFilled,
    UInt32 *pdwRowsToRead
)
{
    UInt8 *srcPtr, *dstPtr;
    Int32 byteCount;
    IPWState *pIPWState;
    Int32 channel, channels;
    PIXR *srcPixr, *pSubPixr;
    Int32 srcBPL, dstBPL = 0L;
    UInt32 rowsFilled = 0UL;
    UInt32 rowsAvailable;
    PROLOGUE;

    if (xImg->tmpPixr == NULL || *pdwImgRowOffset < xImg->tmpPixrRow ||
	*pdwImgRowOffset >= xImg->tmpPixrRow + xImg->tmpPixrRows) {
	NORMAL_RETURN;
    }

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    srcPixr = xImg->tmpPixr;
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, dstPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, srcPixr, &srcBPL));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, dstPixr, &dstBPL));
    byteCount = (xImg->minBytesPerRow + 3) & ~3;
    if (byteCount > dstBPL) {
	byteCount = dstBPL;		/* This might be a problem */
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, srcPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &srcPtr));
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, dstPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &dstPtr));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, srcPixr, &srcPtr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, dstPixr, &dstPtr));
	}
	srcPtr += (*pdwImgRowOffset - xImg->tmpPixrRow) * srcBPL;
	dstPtr += *pdwDstPixrRowOffset * dstBPL;
	rowsAvailable = xImg->tmpPixrRows - (*pdwImgRowOffset - xImg->tmpPixrRow);
	if (rowsAvailable > *pdwRowsToRead) {
	    rowsAvailable = *pdwRowsToRead;
	}
	while (rowsAvailable--) {
	    memcpy(dstPtr, srcPtr, byteCount);
	    srcPtr += srcBPL;
	    dstPtr += dstBPL;
	    rowsFilled++;
	}
    } /* for each channel */
    rowsFilled /= channels;
    (*pdwTotalRowsFilled) += rowsFilled;
    *pdwImgRowOffset += rowsFilled;
    *pdwDstPixrRowOffset += rowsFilled;
    *pdwRowsToRead -= rowsFilled;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read an arbitrary stip image data, and decompress if necessary
 *
 * LongDescription:::
 *  w_XifReadStrip() reads the next strip of image data from an image that was
 *  requested by XIF_ImageReadStart. Image data is read from the top of the
 *  image to the bottom. If the client does not choose to read the entire
 *  image in as a single strip, it is advisable to use a strip height that is
 *  a multiple of the "dwBestStripHeight" (from the XIF_IMGINFO structure)
 *  for best performance.
 *  <BR>
 *  If the BottomToTop flag was selected, this pointer
 *  refer to the beginning of the last row in each buffer.
 *  <BR>
 *  In the case of Uncompressed data returned, the number of bytes filled
 *  should be equal to the bytes/row times either the strip height, or the
 *  number of rows successfully read.
 *  In the case of compressed data returned, the number of bytes filled will
 *  be the number of compressed bytes filled in. For compressed data, the
 *  dwBytesPerRow specifies the size of the buffer. If the buffer is not big
 *  enough to accomodate the compressed byte stream needed for the strip
 *  size, XIF_BUFFEREXCEEDED is returned.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  dwImgRowOffset::		Row offset of image proper
 *  dwStripHeight::		Number of lines to be retrieved
 *  pPixr::			Pointer to client's image PIXR to retrieve into
 *  dwPixrRowOffset::		PIXR row offset to place the strip
 *
 * Output:::
 *  pdwRowsFilled::		Pointer to the number of rows filled in
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifReadStrip(
	IPWState *pIPWState,
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	PIXR *pPixr,
	UInt32 dwPixrRowOffset,
	UInt32 *pdwRowsFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;	/* Image structure */
    IPWState *internalIPWState = NULL;	/* IPW state used by XIFF API */
    XIF_IMGINFO *pxImgInfo;		/* Image info structure */
    UInt32 clientRowsLeft;		/* Number of client rows to fill */
    Int32 bpl = 0L;			/* Bytes per line */
    UInt32 rowsFilled;			/* Rows filled in lower function */
    UInt32 totalRowsFilled = 0UL;	/* Total rows filled */
    ColorType ctype;			/* Color type of intermediate PIXR */
    UInt32 tiffRowsToRead;		/* Rows to read in TIFF strip */
    Int32 framedWidth, framedBytes;	/* Minimum framed PIXR size */
    Int32 pixrHeight, tmpPixrHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    CheckHandleValidity(xImgHandle);

    pxImgInfo = &xImg->imgInfo;

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_READSTATUS) !=
		     XIMG_SEQ_READSTART), XIF_SEQUENCE);

    /* If it's the first strip, populate the PIXR colormap */
    if (dwImgRowOffset == 0 &&
	(pxImgInfo->xColorType == XIF_COLORTYPE_COLOR4 ||
	 pxImgInfo->xColorType == XIF_COLORTYPE_COLOR8)) {
	ColorMap *cMap = NULL;

	XERR_PASS_CLEANUP(w_pixrGetColorMap(pIPWState, pPixr, &cMap));
	if (cMap == NULL) {

	    ColorRecord *cRec;
	    UInt8 *color, *srcColor;
	    UInt32 index, indices = 0UL;

	    indices = 1 << XIF_BPPE(pxImgInfo->xColorType);
	    XERR_PASS_CLEANUP(w_createColorMap(pIPWState, indices, &cMap));
	    XERR_PASS_CLEANUP(w_pixrSetColorMap(pIPWState, pPixr, cMap));
	    XERR_GEN_CLEANUP(cMap == NULL, XIF_INTERNAL_ERROR);
	    cRec = cMap->map;
	    srcColor = pxImgInfo->pPalette;
	    for (index = 0; index < indices; index++) {
		color = (UInt8 *)cRec++;
		color[cRedPlane] = *srcColor++;
		color[cGreenPlane] = *srcColor++;
		color[cBluePlane] = *srcColor++;
	    }
	}
    }

    /* Save the internally used IPWState, and use the passed in one.
     * Restore when finished. */
    if (xImg->xPage->xDoc->xFile->instance->pIPWState != 0) {
        internalIPWState = (IPWState *)xImg->xPage->xDoc->xFile->instance->pIPWState;
    }
    xImg->xPage->xDoc->xFile->instance->pIPWState = (UInt32)pIPWState;

    /* Copy over any left over rows from previous calls */
    clientRowsLeft = dwStripHeight;
    XERR_PASS_CLEANUP(wp_CopyFromTempPixr(xImg, pPixr, &dwPixrRowOffset,
					  &dwImgRowOffset, &totalRowsFilled,
					  &clientRowsLeft));

    /* Keep track of user PIXR status */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));

    /* Get the size of the PIXR passed in */
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    
    /* Check the size of the PIXR if framed properly */
    XERR_PASS_CLEANUP(w_pixrFramedWidth(pIPWState, pPixr, &framedWidth));
    framedBytes = (((framedWidth * XIF_BPPE(pxImgInfo->xColorType)) + 31) / 8) & ~3;

    /* Read strips from the file up to the last whole file strip that won't
     * overfill the client strip */
    while (clientRowsLeft > 0) {

	/* Compute the number of physical rows to read */
	tiffRowsToRead = pxImgInfo->dwBestStripHeight;
	if (dwImgRowOffset + tiffRowsToRead > pxImgInfo->dwHeight) {
	    tiffRowsToRead = pxImgInfo->dwHeight - dwImgRowOffset;
	}

	if (tiffRowsToRead == 0UL) {
	    break;
	}

	/* Determine if the data needs to be buffered:
	 * 1. If the client has requested something in the middle of a
	 *    TIFF strip.
	 * 2. The strip given to us by the client doesn't fill a full
	 *    TIFF strip.
	 * 3. I'm wavlet decompressing and the input PIXR is not a full TIFF
	 *    strip (either has a y-offset or not the TIFF strip height).
	 */
	if (((dwImgRowOffset % pxImgInfo->dwBestStripHeight) != 0) ||
	    (clientRowsLeft < tiffRowsToRead) ||
	    (framedBytes < bpl) ||
	    (pxImgInfo->xCompression == XIF_WAVELET &&
	    (pixrHeight != (Int32)tiffRowsToRead || dwPixrRowOffset != 0))) {

	    /* If wavelet decompressing, the entire PIXR must be a full TIFF
	     * strip (we can't tell the wavelet routines to decompress only
	     * a portion of the PIXR). */
	    if (pxImgInfo->xCompression == XIF_WAVELET &&
		xImg->tmpPixr != NULL) {
		XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, xImg->tmpPixr,
						  &tmpPixrHeight));
		if (tmpPixrHeight != (Int32)tiffRowsToRead) {

		    /* Check for an error condition where the calling
		     * function is using the same intermediate buffer. */
		    XERR_GEN_CLEANUP(pPixr == xImg->tmpPixr,
				     XIF_INTERNAL_ERROR);

		    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState,
						    &xImg->tmpPixr));
		    xImg->tmpPixr = NULL;
		}
	    }

	    /* Use intermediate buffer */
	    if (xImg->tmpPixr == NULL) {
		ScaledInt1616 xres, yres;

		/* Prepare a temporary PIXR */
		xres = Int32_TO_ScaledInt1616(pxImgInfo->dwXResolution);
		yres = Int32_TO_ScaledInt1616(pxImgInfo->dwYResolution);
		switch ((int)pxImgInfo->xColorType) {
		case XIF_COLORTYPE_COLOR24:
		    ctype = cColorRGB;
		    break;
		case XIF_COLORTYPE_BINARY:
		case XIF_COLORTYPE_GRAY4:
		case XIF_COLORTYPE_COLOR4:
		case XIF_COLORTYPE_GRAY8:
		case XIF_COLORTYPE_COLOR8:
		    ctype = cColorGray;
		    break;
		default:
		    ctype = cColorNone;
		}
		xImg->tmpPixrRow = xImg->tmpPixrRows = 0UL;
		XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState,
		    pxImgInfo->dwWidth,
		    tiffRowsToRead, XIF_BPPE(pxImgInfo->xColorType),
		    xres, yres, 32, XIF_CHANNELS(pxImgInfo->xColorType),
		    ctype, NULL, &xImg->tmpPixr));
	    }

	    /* Fill intermediate PIXR */
	    XERR_PASS_CLEANUP(wp_XifReadStrip(xImg, xImg->tmpPixr, 0UL,
		dwImgRowOffset, &tiffRowsToRead, &xImg->tmpPixrRows));
	    xImg->tmpPixrRow = dwImgRowOffset;

	} else {

	    /* Use client PIXR directly */
	    rowsFilled = 0;
	    XERR_PASS_CLEANUP(wp_XifReadStrip(xImg, pPixr, dwPixrRowOffset,
		dwImgRowOffset, &tiffRowsToRead, &rowsFilled));
	    clientRowsLeft -= rowsFilled;
	    dwPixrRowOffset += rowsFilled;
	    dwImgRowOffset += rowsFilled;
	    totalRowsFilled += rowsFilled;
	}

	/* Copy any lines needed to fill the client strip */
	XERR_PASS_CLEANUP(wp_CopyFromTempPixr(xImg, pPixr, &dwPixrRowOffset,
	    &dwImgRowOffset, &totalRowsFilled, &clientRowsLeft));
    }

    /* Copy any lines needed to fill the client strip */
    XERR_PASS_CLEANUP(wp_CopyFromTempPixr(xImg, pPixr, &dwPixrRowOffset,
	&dwImgRowOffset, &totalRowsFilled, &clientRowsLeft));

    *pdwRowsFilled = totalRowsFilled;

    if (internalIPWState != NULL) {
	xImg->xPage->xDoc->xFile->instance->pIPWState = (UInt32)internalIPWState;
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

    if (internalIPWState != NULL) {
	xImg->xPage->xDoc->xFile->instance->pIPWState = (UInt32)internalIPWState;
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clears a rectangular image region in a PIXR
 *
 * LongDescription:::
 *  wp_clearPixrRegion() clears (0's) a rectangular image region in a PIXR
 *  <BR>
 *  Note: does not actually clear an arbitrary region (yet). Needs modification
 *  to actually clear in PIXR format, just clears (long)s.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  pPixr::			PIXR to clear
 *  dstCol::			Destination column, must be 0 (for now)
 *  dstRow::			Destination row, must be 0 (for now)
 *  dstWidth::			Region width
 *  dstHeight::			Region height
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  Even though this was parameterized for a region, assumes the entire
 *  image is to be cleared.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_clearPixrRegion(
    IPWState *pIPWState,
    PIXR *pPixr,
    Int32 dstCol,
    Int32 dstRow,
    Int32 dstWidth,
    Int32 dstHeight
)
{
    UInt32 *imgPtr, *rowPtr;
    Int32 channels;
    Int32 row, rows, lastRow;
    Int32 col, cols, lastCol;
    Int32 bpl;
    PROLOGUE;

    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    if (channels > 1) {
	PIXR *pSubPixr;
	Int32 channel;

	for (channel = 0; channel < channels; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
		&pSubPixr));
	    XERR_PASS_CLEANUP(wp_clearPixrRegion(pIPWState, pSubPixr, dstCol,
		dstRow, dstWidth, dstHeight));
	}
	NORMAL_RETURN;
    }
    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, (UInt8 **)&imgPtr));
    XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, &cols));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &rows));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));

    /* Cleared region will include fractional longs on either side of
     * rectangle */
    dstWidth += (dstCol & 3) + 3;
    dstCol &= ~3;		/* Must be divisible by 4 */
    imgPtr += (dstRow * bpl + dstCol) / 4;
    if (dstCol + dstWidth > cols) {
	dstWidth = cols - dstCol;
    }
    if (dstRow + dstHeight > rows) {
	dstHeight = rows - dstRow;
    }
    lastCol = dstCol + dstWidth;
    lastRow = dstRow + dstHeight;
    for (row = dstRow; row < lastRow; row++) {
	rowPtr = imgPtr;
	for (col = 0; col < lastCol; col += 4) {
	    *rowPtr++ = 0UL;
	}
	imgPtr += bpl / 4;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress image strips as given from the XIFF read routines.
 *
 * LongDescription:::
 *  w_XifDecompress() decompresses image strips as given from the XIFF read
 *  routines. The image handle is used for extracting memory functions.
 *
 * Source buffer is decompressed into the destination image.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  xCompression::		See XIF_COMPRESSION types
 *  xColorType::		Type of color
 *  dwFlags::			XIF_IMGFLAGS_xxx bit combo
 *  pSrc::			Source buffer
 *  dwBytesInSrc::		Byte count of source buffer
 *  dwRowOffset::		Destination row offset
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwStripHeight::		Rows to decompress
 *
 * Output:::
 *  pPixr::			Destination PIXR
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifDecompress(
	IPWState *pIPWState,
	XIF_IMGHANDLE xImgHandle,
	XIF_COMPRESSION xCompression,
	XIF_COLORTYPE xColorType,
	UInt32 dwFlags,
	UInt8 *pSrc,
	UInt32 dwBytesInSrc,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwStripHeight,
	PIXR *pPixr,
	UInt32 *pdwRowsFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;
    Instance *instData;
    XFile *xFile;
    XIF_IMGINFO *pxImgInfo;
    Int32 channel, channels;
    UInt8 *imageChannels[3];
    PIXR *pSubPixr;
    UInt8 *pDst;
    UInt8 *iPtr, *cPtr;
    UInt32 rowsLeft, rowsFilled;
    UInt32 bytesLeft;
    Int32 srcBPL, dstBPL, bytesPerRow;
    Int32 dstHeight;
    UInt32 bytesUsed;
    PROLOGUE;

    *pdwRowsFilled = 0UL;
    CheckHandleValidity(xImgHandle);
    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);
    pxImgInfo = &xImg->imgInfo;

    /* Check buffers and sizes */
    XERR_GEN_CLEANUP((pSrc == NULL || pPixr == NULL), XIF_BADPARAMETER);
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &dstHeight));
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    XERR_GEN_CLEANUP(XIF_BPP(xColorType) / XIF_BPPE(xColorType) !=
	channels, XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(dwRowOffset >= dstHeight, XIF_BADPARAMETER);
    if (dwRowOffset + dwStripHeight > (UInt32)dstHeight) {
	dwStripHeight = dstHeight - dwRowOffset;
    }

    /* Check the input PIXR */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &dstBPL));
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				       &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				     &imageChannels[channel]));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
				     &imageChannels[channel]));
	}
    }
    pDst = imageChannels[0];

    /* Clear the image */
    XERR_PASS_CLEANUP(wp_clearPixrRegion(pIPWState, pPixr, 0, dwRowOffset,
	dwWidth, dwStripHeight));

    cPtr = pSrc;
    iPtr = pDst + (dwRowOffset * dstBPL) + 0;
    rowsLeft = dwStripHeight;
    bytesLeft = dwBytesInSrc;
    srcBPL = xImg->minBytesPerRow;
    bytesPerRow = srcBPL;
    if (bytesPerRow > dstBPL) {
	bytesPerRow = dstBPL;
    }
    switch ((int)xCompression) {
    case XIF_NOC:
        XERR_PASS(wp_XifUnpackImage(pIPWState, cPtr, srcBPL, rowsLeft, channels,
	    imageChannels, dstBPL, dwRowOffset, &rowsFilled));
	*pdwRowsFilled += rowsFilled;
	break;
    case XIF_G4:
	XERR_PASS_CLEANUP(wp_XifDecompressStripCCITT(xImg, cPtr, bytesLeft,
	    iPtr, dstBPL, 0, rowsLeft, &rowsFilled));
	*pdwRowsFilled += rowsFilled;
	break;
    case XIF_JPEG:
	XERR_PASS_CLEANUP(wp_XifDecompressStripJPEG(xImg, cPtr, bytesLeft,
	    pPixr, rowsLeft, dwRowOffset, &bytesUsed));
	*pdwRowsFilled += dwStripHeight;
	break;
    case XIF_TOKENIZATION:
    case XIF_LLTOKENIZATION:
	XERR_PASS_CLEANUP(wp_XifDetokenizeImage(xImg, cPtr, bytesLeft, iPtr,
	    dstBPL));
	*pdwRowsFilled += rowsLeft;
	break;
    case XIF_WAVELET:
	/* Must be the beginning of the image (underlying routines will not
	 * fill in from a row offset) */
	XERR_GEN_CLEANUP(dwRowOffset != 0, XIF_NOSUPPORT);
	XERR_PASS_CLEANUP(wp_XifDecompressStripWavelet(xImg, cPtr, bytesLeft,
	    pPixr, rowsLeft, &rowsFilled));
	*pdwRowsFilled += rowsFilled;
	break;
    case XIF_LOSSYDITHER:
	XERR_PASS_CLEANUP(wp_XifDecompressStripLossyDither(xImg, cPtr, bytesLeft,
	    iPtr, dstBPL, 0, rowsLeft, &rowsFilled));
	*pdwRowsFilled += dwStripHeight;
	break;
    case XIF_COMP_UNKNOWN:
    default:
	__error__ = XIF_NOSUPPORT;
	goto ERR_CLEANUP;
    }

    /* Convert the image if appropriate */
    dwFlags = 0;
    switch ((int)pxImgInfo->xColorType) {
    case XIF_COLORTYPE_BINARY:
	if (xImg->fields.PhotometricInterpretation ==
	    XTIFF_PHOTOMETRIC_BLACKISZERO) {
	    dwFlags = XIF_CONV_INVERT;
	}
	break;
    default:
	if (xImg->fields.PhotometricInterpretation ==
	    XTIFF_PHOTOMETRIC_WHITEISZERO) {
	    dwFlags = XIF_CONV_INVERT;
	}
    } /* switch color type */

    if (dwFlags) {
	XERR_PASS_CLEANUP(w_XifConvertImage(xImg, pPixr, dwFlags, dwRowOffset,
	    *pdwRowsFilled));
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy image data to a PIXR from a byte-packed image stream
 *
 * LongDescription:::
 *  wp_XifUnpackImage() copies image data to a PIXR from a byte-packed image
 *  buffer. This is used to retrieve the "uncompressed" form of data,
 *  independent of the platform. If the source image is multi-channel,
 *  then the source image is one channel after another.
 *  <BR>
 *  The source image buffer is left intact.
 *
 * Input:::
 *  pIPWState::			Ptr to IPCore state struct.
 *  src::			Input image stream
 *  srcBPL::			Bytes per line of packed data
 *  height::			Number of lines (each channel) to transfer
 *  channels::			Number of channels to transfer
 *
 * Output:::
 *  dstChannels::		Destination image buffer(s)
 *  dstBPL::			Destination image bytes per line
 *  dstRowOffset::		Destination row offset
 *  pRowsFilled::		Rows of image data sucessfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifUnpackImage(
	       IPWState *pIPWState,
	       UInt8 *src,
	       Int32 srcBPL,
	       Int32 height,
	       Int32 channels,
	       UInt8 **dstChannels,
	       Int32 dstBPL,
	       UInt32 dstRowOffset,
	       UInt32 *pRowsFilled
)
{
    GraySwapNearProc swapLineProc;
    UInt8 *dst;
    Int32 channel;
    Int32 row, lineBPL;
    UInt8 *line = NULL;
    PROLOGUE;

    /* Start off with nothing filled */
    *pRowsFilled = 0UL;

    XERR_GEN_CLEANUP(srcBPL > dstBPL, XIF_BADPARAMETER);

    ip_getByteSwapProc (pIPWState, cFalse, cPixrToPixr, NULL, NULL, &swapLineProc);

    /* Compute the intermediate line BPL */
    lineBPL = (srcBPL + 3) & ~3;

    /* Allocate an intermediate line buffer (for swapping) */
    XERR_GEN_CLEANUP(((line = (UInt8 *)malloc(lineBPL)) == NULL),
		      XIF_NOMEMORY);

    for (channel = 0; channel < channels; channel++) {
      dst = dstChannels[channel] + dstRowOffset * dstBPL;
      for (row = 0; row < height; row++) {

	/* Copy the byte packed image line to the intermediate line */
	memcpy(line, src, srcBPL);
	src += srcBPL;

	/* Swap the intermediate line to the destination image */
	(*swapLineProc)(pIPWState, (UInt32 *)line, (UInt32 *)dst, lineBPL / 4);
	dst += dstBPL;
      }
    }

    free(line);
    *pRowsFilled = height;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (line) {
      free(line);
    }

     ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image data used to represent one image for an even layer
 *
 * LongDescription:::
 *  w_XifReadLayer() returns the entire image associated with all the
 *  images belonging to an even numbered layer, as described by the image
 *  return via XIF_GetLayerInfo().
 *
 * Input:::
 *  pIPWState::			IPW state
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_GetLayerInfo()
 *
 * Output:::
 *  pPixr::			Image data structure
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function allows piecing together the various images separated by
 *  the tokenization and lossy dither algorithms: residual, tokenized,
 *  and lossy dithered halftones.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifImageReadLayer(
	IPWState *pIPWState,
	XIF_IMGHANDLE xImgHandle,
	PIXR *pPixr,
	UInt32 *pdwRowsFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    Int32 pixrHeight;
    UInt32 rowsFilled = 0UL;
    PIXR *newPixr = NULL, *openedPixr = NULL;
    XIF_COMPRESSION layerCompression;
    UInt32 layerXOffset, layerYOffset, layerXRes, layerYRes;
    UInt32 layerWidth, layerHeight;
    UInt32 imgXOffset, imgYOffset, imgXRes, imgYRes;
    UInt32 imgWidth, imgHeight, imgBestStripHeight, imgMasks;
    XIF_COMPRESSION imgComp;
    XIF_COLORTYPE imgColorType;
    XIF_IMGHANDLE imgHandle;
    UInt32 compNum;
    XIF_COMPRESSION compType[5] = {XIF_G4, XIF_NOC, XIF_TOKENIZATION,
	XIF_LLTOKENIZATION, XIF_LOSSYDITHER};
    UInt32 imgNum, layer;
    ScaledInt1616 xres, yres;
    STREL *pStrel3 = NULL;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    CheckHandleValidity(xImgHandle);

    pxImgInfo = &xImg->imgInfo;

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((xImg->xPage->seqStatus & XPAGE_SEQ_GETINFO) == 0,
	XIF_SEQUENCE);
    layer = xImg->imgInfo.dwImgLayer;
    XERR_GEN_CLEANUP(layer & 1, XIF_BADPARAMETER);  /* Should be even layer */

    /* If there is only one image, return it instead, since we don't
     * have to mess around with going through the binary pieces */
    if (xImg->xPage->pageInfo.pdwImageCounts[layer-1] == 1) {

	/* Get the characteristics of the image */
	XERR_PASS_CLEANUP(XIF_GetImageInfo((XIF_PAGEHANDLE)xImg->xPage,
	    layer, 0, &imgXOffset, &imgYOffset,
	    &imgXRes, &imgYRes, &imgWidth, &imgHeight, &imgBestStripHeight,
	    &imgComp, &imgColorType, NULL, &imgMasks, &imgHandle));

	/* Read it directly */
	XERR_PASS_CLEANUP(XIF_ImageReadStart(imgHandle, 0UL));
	XERR_PASS_CLEANUP(w_XifReadStrip(pIPWState, imgHandle,
	    0, imgHeight, pPixr, 0, pdwRowsFilled));
	XERR_PASS_CLEANUP(XIF_ImageReadFinish(imgHandle));

	/* Return */
	NORMAL_RETURN;
    }

    /* Get the characteristics of the layer image */
    XERR_PASS_CLEANUP(XIF_GetLayerInfo((XIF_PAGEHANDLE)xImg->xPage,
	layer, &layerCompression, &layerXOffset, &layerYOffset,
	&layerXRes, &layerYRes, &layerWidth, &layerHeight, &xImgHandle));

    /* Examine the PIXR */
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));

    /* There should be some consistency */
    XERR_GEN_CLEANUP(xImgHandle != (XIF_IMGHANDLE)xImg, XIF_INTERNAL_ERROR);
    XERR_GEN_CLEANUP((Int32)layerHeight != pixrHeight, XIF_BADPARAMETER);

    /* Loop through the compression methods */
    for (compNum = 0; compNum < sizeof(compType) / sizeof(*compType);
	 compNum++) {

	/* Loop through each image within the layer */
	for (imgNum = 0;
	     imgNum < xImg->xPage->pageInfo.pdwImageCounts[layer - 1];
	     imgNum++) {

	    /* Get the characteristics of the image */
	    XERR_PASS_CLEANUP(XIF_GetImageInfo((XIF_PAGEHANDLE)xImg->xPage,
		pxImgInfo->dwImgLayer, imgNum, &imgXOffset, &imgYOffset,
		&imgXRes, &imgYRes, &imgWidth, &imgHeight, &imgBestStripHeight,
		&imgComp, &imgColorType, NULL, &imgMasks, &imgHandle));

	    if (imgComp != compType[compNum]) {
		continue;
	    }

	    /* Allocate a PIXR for the entire image */
	    newPixr = NULL;
	    xres = Int32_TO_ScaledInt1616(imgXRes);
	    yres = Int32_TO_ScaledInt1616(imgYRes);
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, imgWidth,
		imgHeight, 1, xres, yres, 32, 1, cColorGray, NULL, &newPixr));

	    /* Clear the image */
	    XERR_PASS_CLEANUP(wp_clearPixrRegion(pIPWState, newPixr, 0, 0,
		imgWidth, imgHeight));

	    /* Get the entire image */
	    XERR_PASS_CLEANUP(XIF_ImageReadStart(imgHandle, 0UL));
	    XERR_PASS_CLEANUP(w_XifReadStrip(pIPWState, imgHandle,
		0, imgHeight, newPixr, 0, &rowsFilled));
	    XERR_PASS_CLEANUP(XIF_ImageReadFinish(imgHandle));

	    /* Add it in to the combined image */
	    switch ((int)imgComp) {
	    case XIF_NOC:
	    case XIF_G4:
	    case XIF_TOKENIZATION:
	    case XIF_LLTOKENIZATION:

		/* Use a boolean OR to recombine */
		XERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, newPixr,
		    PIX_SRC | PIX_DST, imgXOffset, imgYOffset, imgWidth,
		    imgHeight, pPixr));

		break;
	    case XIF_LOSSYDITHER:

		if (pStrel3 == NULL) {

		    /* Initialize the strels */
		    XERR_PASS_CLEANUP(w_strel1Inits(pIPWState));

		    /* Get the 3x3 strel */
		    XERR_PASS_CLEANUP(w_getStrel(pIPWState, strel3, &pStrel3));
		}

		/* Create a new PIXR the same size as lossy dithered image */
		XERR_PASS_CLEANUP(w_clonePixrNoInit(pIPWState, newPixr,
		    &openedPixr));

		/* Open the lossy dithered image with a 3x3 strel, send
		 * result to new image */
		XERR_PASS_CLEANUP(w_open(pIPWState, newPixr, pStrel3,
		    &openedPixr));

		/* Apply a boolean XOR to recombine opened image with output
		 * image */
		XERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, openedPixr,
		    PIX_SRC ^ PIX_DST, imgXOffset, imgYOffset, imgWidth,
		    imgHeight, pPixr));

		/* Subtract the opened image from the lossy dithered image */
		XERR_PASS_CLEANUP(w_logicOp(pIPWState, openedPixr, SUBTRACT,
		    newPixr));

		/* Apply a boolean OR to recombine */
		XERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, newPixr,
		    PIX_SRC | PIX_DST, imgXOffset, imgYOffset, imgWidth,
		    imgHeight, pPixr));
		break;
	    default:
		break;
	    } /* switch (compression type) */

	    /* Clean up */
	    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &newPixr));
	    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &openedPixr));

	}  /* For each image in layer */
    }  /* For each compression type in order */

    /* Output the total number of rows in user's image filled */
    *pdwRowsFilled = layerHeight;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newPixr) {
	w_destroyPixr(pIPWState, &newPixr);
    }
    if (openedPixr) {
	w_destroyPixr(pIPWState, &openedPixr);
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initiate a common IPWrap connection
 *
 * LongDescription:::
 *  wp_XifIPWState() initiates an IPWrap state, to be used throughout the
 *  XIFF API.  If the state has been initated previously, as attached through
 *  the document handle, it is simply returned.
 *
 * Input:::
 *  xDoc::			Valid open document structure
 *
 * Output:::
 *  ppIPWState::		Pointer to the resulting IPWrap state
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifIPWState(
	XDoc *xDoc,
	IPWState **ppIPWState
)
{
    PROLOGUE;

    /* Default to return nothing */
    *ppIPWState = NULL;

    CheckHandleValidity(xDoc);
    XERR_GEN_CLEANUP(ppIPWState == NULL, XIF_BADPARAMETER);

    *ppIPWState = (IPWState *)xDoc->xFile->instance->pIPWState;
    if (*ppIPWState == NULL) {

	/* Initiate an IPCore interface
	 */
	XERR_PASS_CLEANUP(w_IPWStateInit(wp_IPC_Malloc, wp_IPC_Calloc,
		wp_IPC_Realloc, wp_IPC_Free, (void *)xDoc->xFile->instance,
		wp_IPC_Malloc, wp_IPC_Calloc, wp_IPC_Realloc, wp_IPC_Free,
		(void *)xDoc->xFile->instance, NULL, stderr, 0, NULL, NULL,
		ppIPWState));
	xDoc->xFile->instance->pIPWState = (UInt32)*ppIPWState;
	xDoc->xFile->instance->myIPWState = 1;
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a CCITT G4 image (MMR)
 *
 * LongDescription:::
 *  wp_DecompressStripCCITT() decompresses a strip of CCITT G4 (MMR = T.6)
 *  data. The temporary PIXR found in xImg is used to store the image data
 *  decompressed.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *
 * Output:::
 *  pImgData::			Pointer to the first pixel of image data
 *  dwBytesPerRow::		Bytes per row of image data
 *  dwRowOffset::		Row offset in image data to place output
 *  dwStripHeight::		Number of rows to decompress
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifDecompressStripCCITT(
    XImg *xImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt32 *pdwRowsFilled
)
{
    Instance *instData;
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState;
    Int32 w;
    void *compInfo;
    UInt32 imgExh, cmpExh;
    Int32 stateBlockSize;
    UInt32 padType = cNoPad;
    UInt8 *iPtr, *pDst;
    UInt8 *cPtr;
    Int32 bytesLeft, bytesUsed, rowsLeft, pct;
    Bool contImg, contCmp;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));
    instData = XIF_IMGINST(xImg);

    pxImgInfo = &xImg->imgInfo;
    w = pxImgInfo->dwWidth;

    if (xImg->tmpCmpData == NULL) {	/* First strip */

	/* Prepare for pre-decode */
	XERR_PASS_CLEANUP(i_getSizeOfCCITTDecompressionInfo(
	    pIPWState,
	    w + 2*cFrameBits, &stateBlockSize));

	/* Allocate temporary compression data, attach to image */
	CHECK_MALLOC(xImg->tmpCmpData =
	    (MISC_MEM *)XIF_MALLOC(sizeof(MISC_MEM)));
	CHECK_MALLOC(xImg->tmpCmpData->mem =
	    (UInt8 *)XIF_MALLOC(stateBlockSize));
	xImg->tmpCmpData->next = NULL;
	compInfo = xImg->tmpCmpData->mem;

	XERR_PASS_CLEANUP(i_initializeCCITTDecompressionInfo(
	    pIPWState,
	    compInfo, w + 2*cFrameBits, cTrue,
	    padType, stateBlockSize));
    } else {
	compInfo = xImg->tmpCmpData->mem;
    }

    pDst = pImgData;
    iPtr = pDst - (cFrameBits * XIF_BPPE(pxImgInfo->xColorType)) / 8;
    cPtr = pCmpData;
    rowsLeft = dwStripHeight;
    bytesLeft = dwCmpBytes;
    contImg = cFalse;
    contCmp = cFalse;

    do {
	XERR_PASS_CLEANUP(i_decompressRowsMMR (pIPWState, compInfo,
	    iPtr, dwRowOffset,
	    rowsLeft, dwBytesPerRow, cFalse, cPtr, bytesLeft, contImg,
	    contCmp, cTrue, &imgExh, &cmpExh, &pct, &bytesUsed));

	bytesLeft -= bytesUsed;

	if (imgExh == cTrue && cmpExh == cFalse) {

	    /* Okay, we can live with it, but it's wrong */

	} else if (cmpExh == cTrue && imgExh == cFalse) {

	    /* Can't decompress all of it, bad! */
	    __error__ = XIF_BADFILEFORMAT;
	    goto ERR_CLEANUP;
	}
	contImg = cTrue;
	contCmp = cTrue;
    } while (imgExh == cFalse && cmpExh == cFalse && bytesLeft > 0);
    rowsLeft = 0;

    *pdwRowsFilled = dwStripHeight - rowsLeft;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a lossy dithered image
 *
 * LongDescription:::
 *  wp_DecompressStripLossyDither() decompresses a strip of lossy dither
 *  compressed data.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *
 * Output:::
 *  pImgData::			Pointer to the first pixel of image data
 *  dwBytesPerRow::		Bytes per row of image data
 *  dwRowOffset::		Row offset in image data to place output
 *  dwStripHeight::		Number of rows to decompress
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifDecompressStripLossyDither(
    XImg *xImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt32 *pdwRowsFilled
)
{
    Instance *instData;
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState;
    Int32 w;
    void *compInfo;
    UInt32 imgExh, cmpExh;
    UInt32 stateBlockSize;
    UInt8 *iPtr, *pDst;
    UInt8 *cPtr;
    Int32 bytesLeft, bytesUsed, rowsLeft, pct;
    Bool contImg, contCmp;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));
    instData = XIF_IMGINST(xImg);

    pxImgInfo = &xImg->imgInfo;
    w = pxImgInfo->dwWidth;

    if (xImg->tmpCmpData == NULL) {	/* First strip */

	/* Prepare for pre-decode */
	XERR_PASS_CLEANUP(i_getSizeOfLDDecompressionInfo(
	    pIPWState,
	    w + 2*cFrameBits, &stateBlockSize));

	/* Allocate temporary compression data, attach to image */
	CHECK_MALLOC(xImg->tmpCmpData =
	    (MISC_MEM *)XIF_MALLOC(sizeof(MISC_MEM)));
	CHECK_MALLOC(xImg->tmpCmpData->mem =
	    (UInt8 *)XIF_MALLOC(stateBlockSize));
	xImg->tmpCmpData->next = NULL;
	compInfo = xImg->tmpCmpData->mem;

	XERR_PASS_CLEANUP(i_initializeLDDecompressionInfo(
	    pIPWState,
	    compInfo, w + 2*cFrameBits, stateBlockSize));
    } else {
	compInfo = xImg->tmpCmpData->mem;
    }

    pDst = pImgData;
    iPtr = pDst - (cFrameBits * XIF_BPPE(pxImgInfo->xColorType)) / 8;
    cPtr = pCmpData;
    rowsLeft = dwStripHeight;
    bytesLeft = dwCmpBytes;
    contImg = cFalse;
    contCmp = cFalse;

    do {
	XERR_PASS_CLEANUP(i_decompressRowsLD (pIPWState, compInfo,
	    iPtr, dwRowOffset,
	    rowsLeft, dwBytesPerRow, cPtr, bytesLeft, contImg, contCmp,
	    cTrue, &imgExh, &cmpExh, &pct, &bytesUsed));

	bytesLeft -= bytesUsed;

	if (imgExh == cTrue && cmpExh == cFalse) {

	    /* Okay, we can live with it, but it's wrong */

	} else if (cmpExh == cTrue && imgExh == cFalse) {

	    /* Can't decompress all of it, bad! */
	    __error__ = XIF_BADFILEFORMAT;
	    goto ERR_CLEANUP;
	}
	contImg = cTrue;
	contCmp = cTrue;
    } while (imgExh == cFalse && cmpExh == cFalse && bytesLeft > 0);
    rowsLeft = 0;

    *pdwRowsFilled = dwStripHeight - rowsLeft;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*
$Log: xifrpixr.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.25  1998/07/27 20:54:11  aber
aber on Mon Jul 27 13:53:43 PDT 1998

xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Added a new function to specifically propogate the IPWState structure.

Revision 1.24  1998/05/05 07:15:24  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Additions necessary to fill in the never completed functions
XIF_Compress and XIF_Decompress.

Revision 1.23  1998/03/16 21:01:53  aber
aber on Mon Mar 16 13:00:48 PST 1998

xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

In read-layer function, allow uncompressed images to be added.

Revision 1.22  1998/02/09 20:21:25  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.21  1998/01/20 03:21:55  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Fix status check to work in read/write mode.

Revision 1.20  1998/01/14 00:16:30  ddavies
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.19  1997/11/01 20:51:18  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Fix color palette handling (don't add palette to PIXR if already there).
Add exact strip buffering for wavelet.

Revision 1.18  1997/07/15 23:42:27  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Buffering fixes.

Revision 1.17  1997/07/14 23:22:59  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Review image buffering during read. Rename variables to be
more intuitive. Allow non-recommended strip sizes.

Revision 1.16  1997/06/26 23:29:13  aber
aber on Thu Jun 26 16:28:26 PDT 1997

xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Make potential ansification fixes that the Mac might complain about.

Revision 1.15  1997/06/26 22:25:37  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Convert "white-is-zero" XIFF 2.0 image (invert).

Revision 1.14  1997/06/26 20:05:01  ddavies
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Add inclusion of new file rast.pub.

Revision 1.13  1997/06/23 22:59:03  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Complete function calls for paletted image support.

Revision 1.12  1997/06/20 17:44:38  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

When asked for a layer, return the single image (non-binary
primary image case).

Revision 1.11  1997/06/18 00:41:26  aber
aber on Tue Jun 17 17:41:26 PDT 1997

xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Updated prototype for XIF_GetLayerInfo.

Revision 1.10  1997/06/17 23:45:06  aber
aber on Tue Jun 17 16:44:44 PDT 1997

xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Update return value before returning.

Revision 1.9  1997/06/17 22:35:18  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Clean up if completed the image read -- was simply returning.

Revision 1.7  1997/06/12 00:50:21  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Enable lossless tokenization. If XIFF has unsupported compression type,
return no support, vs. bad file format.

Revision 1.6  1997/06/10 23:34:01  aber
aber on Tue Jun 10 16:33:31 PDT 1997

xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

Fix documentation (nag nag).
Clean up memory in proper loop, fixes memory leak.

Revision 1.5  1997/06/10 01:26:40  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.4  1997/05/30 00:51:02  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.3  1997/05/20 22:29:06  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.2  1997/05/20 00:54:13  aber
xifrpixr.c -> /product/ipcore/ipwrap/src/RCS/xifrpixr.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:37  aber
Initial revision

*/
