/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifrpriv.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifrpriv.c:	XIFF READ private routines
 *
 *	XIFF read-only private routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifrtiff.h"
#include "xifrxiff.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifrpriv.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");


/* XIFF internal memory allocation adds some info to help detect leaks
 * and page faults. I place a key word before and after the
 * memory during the allocation, and check that it is still
 * in tact during the free. I also place the size of memory
 * before, so that I know where the key is after the block,
 * and to keep track of memory usage at any point.
 */
#define MEMORY_KEY		0x75c38a19	/* Ya, I just made it up */
#define KEYED_MEMORY	(sizeof(long) * 3)

/* Memory and I/O */
void * CDECL
wp_XifMalloc(
    size_t bytes,
    Bool hasPtrs,
    void *clientMallocID
)
{
    XifMemoryTally *tally = (XifMemoryTally *)clientMallocID;
    void *memory;
    UInt32 *lmem;
	
    if (clientMallocID == NULL || bytes == 0) {
	hasPtrs = 0;	    /* reference once, to avoid compiler warnings */
	return NULL;
    }

    
    bytes = (bytes + 3UL) & ~3UL;	/* Round to next word */
    if ((memory = (void *)malloc(bytes + KEYED_MEMORY)) == NULL) {
	return memory;
    }
    tally->bytesAllocated += bytes;
    if (tally->bytesAllocated > tally->peak) {
	tally->peak = tally->bytesAllocated;
    }
    
    /* Place memory key and size around buffer */
    lmem = (UInt32 *)memory;
    lmem[0] = bytes;
    lmem[1] = (UInt32)MEMORY_KEY;
    lmem[((bytes + KEYED_MEMORY) >> 2) - 1] = (UInt32)MEMORY_KEY;
    
    return (void *)(lmem + 2);
}

void CDECL
wp_XifFree(
    void *buffer,
    size_t dwBytes,
    void *clientMallocID
)
{
    XifMemoryTally *tally = (XifMemoryTally *)clientMallocID;
    UInt32 *lmem;
    UInt32 bytes;
	
    if (clientMallocID == NULL) {
	dwBytes = 0;	    /* Reference once, to avoid compiler warnings */
	return;
    }
    
    if ((unsigned)buffer & 3) {
	fprintf(stderr, "Invalid memory free, buffer is not long aligned\n");
	goto memoryError;
    }
    
    lmem = ((UInt32 *)buffer) - 2;
    bytes = lmem[0];		/* If seg fault here, buffer doesn't contain 4
			 * bytes -- buffer is probably invalid */
    if (bytes < 4) {
	fprintf(stderr, "Memory corruption, memory before buffer invalid\n");
	goto memoryError;
    }
    if (lmem[1] != (UInt32)MEMORY_KEY) {
	fprintf(stderr, "Memory key corrupted, memory before buffer invalid\n");
	goto memoryError;
    }
    if (lmem[((bytes + KEYED_MEMORY) >> 2) - 1] != (UInt32)MEMORY_KEY) {
	fprintf(stderr, "Memory key corrupted, memory after buffer invalid\n");
	goto memoryError;
    }
    
    tally->bytesAllocated -= bytes;
    free((char *)lmem);
    return;		/* Normal return */
    
 memoryError:
    return;		/* Put break point here in debugger */
}

UInt32 CDECL
wp_XifMemoryCheck(
    XifMemoryTally *tally
)
{
    
#ifdef DEBUG
    fprintf(stderr, "%d bytes currently allocated, %d bytes used at peak.\n",
	    tally->bytesAllocated, tally->peak);
#endif /* DEBUG */
    return tally->bytesAllocated;
}


UInt32 CDECL
wp_XifAllocSize(
    void *buffer,
    void *clientID
)
{
    XifMemoryTally *tally = (XifMemoryTally *)clientID;
    UInt32 *lmem;
    UInt32 bytes;

    if (tally == NULL) {
	goto memoryError;
    }
    
    if ((unsigned)buffer & 3) {
	fprintf(stderr, "Invalid memory, buffer is not long aligned\n");
	goto memoryError;
    }
    
    lmem = ((UInt32 *)buffer) - 2;
    bytes = lmem[0];		/* If seg fault here, buffer doesn't contain 4
			 * bytes -- buffer is probably invalid */
    return bytes;		/* Normal return */
    
 memoryError:
    return 0UL;		/* Put break point here in debugger */
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Lock the file -- mark as unavailable by other threads.
 *
 * LongDescription:::
 *  wp_XifFileLock() locks the file for full ownership by the calling thread,
 *  disallowing another thread to request ownership until this thread has
 *  released the file using XIF_FileUnLock().
 *
 * Input:::
 *  fileToken::			File token referencing file to lock
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifFileLock(
    XIF_TOKEN *fileToken
)
{
    PROLOGUE;

    XERR_GEN_CLEANUP(fileToken == NULL, XIF_INTERNAL_ERROR);

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Unlock the file -- mark as free for the next thread to take control.
 *
 * LongDescription:::
 *  wp_XifFileUnLock() unlocks the file from the current thread, allowing
 *  another thread to request full ownership of the file.
 *
 * Input:::
 *  fileToken::		File token referencing file to unlock
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifFileUnLock(
    XIF_TOKEN *fileToken
)
{
    PROLOGUE;

    XERR_GEN_CLEANUP(fileToken == NULL, XIF_INTERNAL_ERROR);

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}



/***********************************************************
 *       IPCORE memory interface
 */

void * CDECL
wp_IPC_Malloc(
    size_t count,
    Bool hasPtrs,
    void *thing
)
{
    Instance *inst = (Instance *)thing;

    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
    return inst->malloc(count, hasPtrs, inst->mallocID);
}



void * CDECL
wp_IPC_Calloc(
    size_t nel,
    size_t size,
    Bool hasPtrs,
    void *thing
)
{
    Instance *inst = (Instance *)thing;
    void *buffer;
	     
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
      return NULL;
    }
    buffer = inst->malloc(nel * size, hasPtrs, inst->mallocID);
    memset(buffer, 0, nel * size);
 
    return buffer;
}



void * CDECL
wp_IPC_Realloc(
    void *ptr,
    size_t size,
    Bool hasPtrs,
    void *thing
)
{
    UInt32 oldSize;
    Instance *inst = (Instance *)thing;
    void *buffer;
	    
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
				    
    /* This isn't right in the case of client malloc!!!!!!!!!!!!!
     *
     * We need to know the size of the old buffer, to know how to 
     * reallocate it. In the case of the memory allocation I have,
     * I keep it stored with the buffer. The client may do anything
     * they desire, so I have no way of pinging for this information.
     */
    oldSize = wp_XifAllocSize(ptr, inst->mallocID);
    buffer = inst->malloc(size, hasPtrs, inst->mallocID);
    
    if (oldSize > size) {
	memcpy(buffer, ptr, size);
    } else {
	memcpy(buffer, ptr, oldSize);
    }
    inst->free(ptr, oldSize, inst->mallocID);
    
    return buffer;
}



void CDECL
wp_IPC_Free(
    void *ptr,
    size_t size,
    void *thing
)
{
    Instance *inst = (Instance *)thing;

    if (thing == NULL) {
	size = 0;	/* Reference once to avoid compiler warnings */
	return;
    }
    inst->free(ptr, 0, inst->mallocID);
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocate a new image handle, and assign it to its owner.
 *   
 *
 * LongDescription:::
 *  wp_NewImg() allocates a new image handle, and assign it to its owner.
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *  owner::			Pointer to pointer to image (points to new
 *				image's owner)
 *  ifdOffset::			IFD offset of the new image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_NewImg(
    XPage *xPage,
    XImg **owner,
    Int32 ifdOffset
)
{
    Instance *instData;
    XFile *xFile;
    XImg *xImg = NULL;
    PROLOGUE;
    
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);
    
    /* Begin construction of the primary image data */
    CHECK_MALLOC(xImg = (XImg *)XIF_MALLOC(sizeof(XImg)));
    memset(xImg, 0, sizeof(XImg));
    xImg->handleID = (UInt32)xImg;		/* self identity */
    xImg->xPage = xPage;
    xImg->imgInfo.xImgHandle = (UInt32)xImg;
    
    xImg->ifdOffset = ifdOffset;
    
    /* Attach it to the owner */
    *owner = xImg;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Increment the number of images within a layer
 *
 * LongDescription:::
 *  wp_XifIncrementImageCounts() makes the appropriate steps to increment the
 *  count of images within a layer, which may include reallocating the image
 *  counts array stored in the page info.
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *  layer::			Layer to increment the count
 *  inc::			Number of times to increment
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifIncrementImageCounts(
	XPage *xPage,
	UInt32 layer,
	Int32 inc
)
{
    Instance *instData = XIF_PAGEINST(xPage);
    UInt32 *newImageCounts = NULL;
    PROLOGUE;

    if (inc == 0) {
	NORMAL_RETURN;
    }

    /* If decrementing, see if this collapses the number of layers required */
    if (inc < 0) {
	XERR_GEN_CLEANUP(xPage->pageInfo.dwLayersPresent < layer,
			 XIF_INTERNAL_ERROR);
	if (layer == xPage->pageInfo.dwLayersPresent &&
	    xPage->pageInfo.pdwImageCounts[layer - 1] == (UInt32)-inc) {

	    CHECK_MALLOC(newImageCounts = (UInt32 *)XIF_MALLOC(sizeof(UInt32) *
							       layer - 1));
	    memset(newImageCounts, 0, sizeof(UInt32) * layer - 1);
	    memcpy(newImageCounts, xPage->pageInfo.pdwImageCounts,
		   sizeof(UInt32) * layer - 1);
	    XIF_FREE(xPage->pageInfo.pdwImageCounts);

	    xPage->pageInfo.pdwImageCounts = newImageCounts;
	    newImageCounts = NULL;
	    xPage->pageInfo.dwLayersPresent = layer - 1;
	} else {

	    /* Increment (decrement) the count for this layer */
	    xPage->pageInfo.pdwImageCounts[layer - 1] += inc;
	}

    } else {

	/* See if the array needs to be expanded */
	if (xPage->pageInfo.dwLayersPresent < layer) {

	    CHECK_MALLOC(newImageCounts = (UInt32 *)XIF_MALLOC(sizeof(UInt32) *
							       layer));
	    memset(newImageCounts, 0, sizeof(UInt32) * layer);
	    if (xPage->pageInfo.pdwImageCounts != NULL) {
		memcpy(newImageCounts, xPage->pageInfo.pdwImageCounts,
		       sizeof(UInt32) * xPage->pageInfo.dwLayersPresent);
		XIF_FREE(xPage->pageInfo.pdwImageCounts);
	    }
	    xPage->pageInfo.pdwImageCounts = newImageCounts;
	    newImageCounts = NULL;
	    xPage->pageInfo.dwLayersPresent = layer;
	}

	/* Increment the count for this layer */
	xPage->pageInfo.pdwImageCounts[layer - 1] += inc;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newImageCounts != NULL) {
	XIF_FREE(newImageCounts);
    }
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image description and fill in the ImgInfo structure.
 *
 * LongDescription:::
 *  wp_XifGetImageDesc() reads the IFD and fills in the ImgInfo structure with
 *  the basic image description as interpreted from the IFD.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGetImageDesc(
    XImg *xImg
)
{
    Instance *instData;
    XFile *xFile;
    XPage *xPage = NULL;
    XDoc *xDoc = NULL;
    XImg *childImg = NULL;
    UInt32 nextIFD;
    XImg **imgPtr;
    UInt32 imgNum;
    PROLOGUE;
    
    /* We may have previously gotten the image info... */
    if ((xImg->seqStatus & XIMG_SEQ_GETINFO) == 0) {
	
	xPage = xImg->xPage;
	xDoc = xPage->xDoc;
	instData = XIF_DOCINST(xDoc);
	xFile = XIF_DOCFILE(xDoc);
	
	/* Call the TIFF routine to read the tags */
	XERR_PASS_CLEANUP(wp_TifReadIFD(xImg));
	
	/* Mark as gotten now, otherwise we'll appear to need it again during a
	 * child image call */
	xImg->seqStatus |= XIMG_SEQ_GETINFO;
	xImg->imgInfo.xImgHandle = (UInt32)xImg;
	
	/* Retrieve information about the layout of the parent image */
	if (xImg->imgInfo.xImgType == XIF_IMGTYPE_PRIMARY) {

	    /* For version 3 and greater, the reset of the page information
	     * is located in the MixedRasterContent field */
	    if (xDoc->docInfo.version == 2) {

		/* Count the Primary image */
		XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage, 2, 1));
		
		/* Darn, we're gunna have to parse through the entire list
		 * of images to extract all the info we desire */
		/* Look for attached Subimages (overlay) or Mask images */
		if (xImg->fields.SubIFDs != NULL) {
		    nextIFD = xImg->fields.SubIFDs[0];
		} else {
		    nextIFD = 0;
		}
		imgPtr = &xPage->xImgs;
		imgNum = 0;
		while (nextIFD) {
		    
		    XERR_PASS_CLEANUP(wp_NewImg(xPage, imgPtr, nextIFD));
		    childImg = *imgPtr;
		    childImg->imgInfo.dwImgOffset = imgNum++;
		    XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage,
			childImg->imgInfo.dwImgLayer = 5, 1));
		    
		    XERR_PASS_CLEANUP(wp_TifGetNextIFD(childImg));
		    
		    nextIFD = childImg->fields.nextIFD;
		    imgPtr = &childImg->next;
		}
	    } /* Get page layout, version dependent */
	} else if (xImg->imgInfo.dwImgLayer > 0) {	/* Not a mask image */

	    /* Image type is NOT Primary or Mask */
	    
	    /* Look for attached MASK images */
	    XERR_PASS_CLEANUP(wp_XifTraverseMaskImages(xImg));
	} /* Image is the primary image */

	/* Check for a correct "best strip height" value */
	if (xImg->imgInfo.dwBestStripHeight > xImg->imgInfo.dwHeight) {
	    xImg->imgInfo.dwBestStripHeight = xImg->imgInfo.dwHeight;
	}
	
    } /* Info not already gotten */
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*
$Log: xifrpriv.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.13  1998/07/27 17:36:23  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

Use private memory tally structure.

Revision 1.12  1998/02/09 20:21:26  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.11  1997/12/18 00:11:13  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

Allow negative increments in image count, handle shrinking image counts
in MixedRasterContent field.

Revision 1.10  1997/08/29 17:28:02  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.9  1997/07/10 21:52:30  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

Use size_t as the memory allocation size (instead of UInt32).

Revision 1.8  1997/07/02 01:05:15  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

Convert memory functions to IPWrap functions.

Revision 1.7  1997/06/20 17:46:19  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

Test for sub-images before referencing the list, NULL-pointer reference
otherwise.

Revision 1.6  1997/06/10 01:26:42  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:03  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:07  aber
xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:56:25  aber
aber on Mon May 19 17:55:49 PDT 1997

xifrpriv.c -> /product/ipcore/ipwrap/src/RCS/xifrpriv.c,v

Fixed CDECL in prototype, and potentially unused variables.

Revision 1.1  1997/05/17 00:46:38  aber
Initial revision

*/
