/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifrtiff.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifrtiff.c:	READ TIFF source
 *
 *	TIFF read-only routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xiff.h"
#include "xifrtiff.h"
#include "xiftiff.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifrtiff.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

/* Data type sizes (in bytes) */
char FieldTypeSize[] = {0, 1, 1, 2, 4, 8, 1, 1, 2, 4, 8, 4, 8};


/* Short swap: BA -> AB */
void
wp_swaps(Int8 *x)
{
    Int8   tmp;
    Int8   *a = x;

    tmp = a[1];
    a[1]= a[0];
    a[0]= tmp;
}


/* Long swap: DCBA -> ABCD */
void
wp_swapl(Int8 *x)
{
    Int8  tmp;
    Int8   *a = (Int8 *)x;
    
    tmp = a[0];
    a[0]= a[3];
    a[3]= tmp;
    
    tmp = a[1];
    a[1]= a[2];
    a[2]= tmp;
}


/* Double long swap: HGFE DCBA -> ABCD EFGH */
void
wp_swapd(Int8 *x)
{
    Int8  tmp;
    Int8   *a = (Int8 *)x;
    
    tmp = a[0];
    a[0]= a[7];
    a[7]= tmp;
    
    tmp = a[1];
    a[1]= a[6];
    a[6]= tmp;
    
    tmp = a[2];
    a[2]= a[5];
    a[5]= tmp;
    
    tmp = a[3];
    a[3]= a[4];
    a[4]= tmp;
}


ErrCode CDECL
wp_TifConvcopy(
	 void *dst,
	 const void *src,
	 UInt16 type,   /* See "Data Types" in tifffields.h */
	 UInt32 count,
	 UInt8 swap
)
{
    UInt32 bytes = FieldTypeSize[type] * count;
    UInt32 index;
    UInt16 *sptr;
    UInt32 *lptr;
    double *dptr;
    PROLOGUE;
    
    /* Copy the data */
    memcpy(dst, src, bytes);

    if (swap) {
	switch (FieldTypeSize[type]) {
	case 2:
	    sptr = (UInt16 *)dst;
	    for (index = 0; index < count; index++) {
		SwapS(sptr);
		sptr++;
	    }
	    break;
	case 4:
	    lptr = (UInt32 *)dst;
	    for (index = 0; index < count; index++) {
		SwapL(lptr);
		lptr++;
	    }
	    break;
	case 8:
	    dptr = (double *)dst;
	    for (index = 0; index < count; index++) {
		SwapD(dptr);
		dptr++;
	    }
	    break;
	}
    }
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check to see if the file is a valid TIFF
 *
 * LongDescription:::
 *  wp_TifIsTiff() checks for the necessary header information to identify the
 *  file as a TIFF file.
 *
 * Input:::
 *  pxFile::			Pointer to file handle
 *
 * Output:::
 *  pdwFirstIfd::		Pointer to first IFD offset
 *
 * Returns:::
 *  XIF_NOTTIFF::		File is not a TIF
 *  XIF_NOERROR::		File is a supported TIF
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifIsTiff(
    XFile *pxFile,
    UInt32 *pdwFirstIfd
)
{
    TiffHeader header;
    PROLOGUE;

    /* Initialize */
    *pdwFirstIfd = 0;

    /* Check the TIFF header */
    XIF_SEEK(pxFile, 0UL);
    XIF_READ(pxFile, &header.byteOrder, sizeof(header.byteOrder));
    XIF_READ(pxFile, &header.version, sizeof(header.version));
    XIF_READ(pxFile, &header.firstIfd, sizeof(header.firstIfd));
    
    /* Check for a supported version */
    XERR_GEN_CLEANUP((header.byteOrder != MM && header.byteOrder != II),
		     XIF_NOTTIFF);
    XIF_FILETOKEN(pxFile)->bSwapBytes = header.byteOrder != NATIVEORDER;
    SwapShort(pxFile, &header.version);
    SwapLong(pxFile, &header.firstIfd);
    XERR_GEN_CLEANUP((header.version != 42), XIF_NOTTIFF);

    /* TIFF defines: a first page must exist. However, XIFF must provide
     * 0-page files. If we were to be clear about this, we would do
     * the following: */
    /* XERR_GEN_CLEANUP((header.firstIfd < 8), XIF_TIFFEMPTY); */

    *pdwFirstIfd = header.firstIfd;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the IFD information allocated
 *
 * LongDescription:::
 *  wp_TifFieldCleanup() frees any miscellaneous data allocated for an IFD.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  fieldList::			TIFF field structure to clean
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFieldCleanup(
    XImg *xImg,
    TiffFields *fieldList
)
{
    Instance *instData;
    PROLOGUE;
    
    CheckHandleValidity(xImg);
    
    instData = XIF_IMGINST(xImg);
    
    /* Check for memory blocks */
    if (fieldList->BitsPerSample) {
	XIF_FREE(fieldList->BitsPerSample);
    }
    if (fieldList->StripOffsets) {
	XIF_FREE(fieldList->StripOffsets);
    }
    if (fieldList->StripByteCounts) {
	XIF_FREE(fieldList->StripByteCounts);
    }
    if (fieldList->TileOffsets) {
	XIF_FREE(fieldList->TileOffsets);
    }
    if (fieldList->TileByteCounts) {
	XIF_FREE(fieldList->TileByteCounts);
    }
    if (fieldList->AnnotationOffsets) {
	XIF_FREE(fieldList->AnnotationOffsets);
    }
    if (fieldList->Colormap) {
	XIF_FREE(fieldList->Colormap);
    }
    if (fieldList->SubIFDs) {
	XIF_FREE(fieldList->SubIFDs);
    }
    if (fieldList->ILContent.imageCounts) {
	XIF_FREE(fieldList->ILContent.imageCounts);
    }
    
    NORMAL_RETURN;
}


#if 0
static long
FieldLong(
    FieldEntry *fieldEntry,
    int swap
)
{
    long lval = 0L;
    
    switch (fieldEntry->type) {
    case IFDBYTE:
    case IFDASCII:
    case IFDSHORT:
	break;
    case IFDLONG:
	if (fieldEntry->len == 1) {
	    lval = fieldEntry->offset;
	}
	break;
    case IFDRATIONAL:
    case IFDSBYTE:
    case IFDUNDEFINED:
    case IFDSSHORT:
    case IFDSLONG:
    case IFDSRATIONAL:
    case IFDFLOAT:
    case IFDDOUBLE:
	break;
    }
    
    if (swap) {
	SwapL(&lval);
    }
    return lval;
}
#endif


static int
wp_TifReadFieldData(
    XImg *xImg,
    FieldEntry *fieldEntry,
    void *data,
    UInt32 dataSize,
    int swap
)
{
    XFile *xFile = XIF_IMGFILE(xImg);
    short *sptr;
    long  *lptr;
    double *dptr;
    int bytes = FieldTypeSize[fieldEntry->type] * fieldEntry->len;
    int elementsLeft;
    PROLOGUE;
    
    if (bytes <= 4) {
	
	/* Data found in offset entry element */
	if (swap) {
	    SwapL(&fieldEntry->offset);	/* swap back */
	}
	memcpy(data, &fieldEntry->offset, bytes);
	if (swap) {
	    switch (bytes) {
	    case 1:
		break;
	    case 2:
		SwapS(data);
		break;
	    case 4:
		SwapL(data);
		break;
	    }
	}
	/* Data already swapped when placed in fieldEntry->offset */
	
    } else {	/* bytes > 4 */
	
	/* Data is still in the file, extract */
	XIF_SEEK(xFile, fieldEntry->offset);
	XIF_READ(xFile, data, bytes);
	
	if (swap) {
	    bytes = fieldEntry->len * FieldTypeSize[fieldEntry->type];
	    switch (fieldEntry->type) {
	    case IFDBYTE:
	    case IFDASCII:
	    case IFDSBYTE:
	    case IFDUNDEFINED:
		break;
		
	    case IFDSHORT:
	    case IFDSSHORT:
		elementsLeft = bytes / 2;
		sptr = (short *)data;
		while (elementsLeft--) {
		    SwapS(sptr++);
		}
		break;
		
	    case IFDLONG:
	    case IFDRATIONAL:
	    case IFDSLONG:
	    case IFDSRATIONAL:
	    case IFDFLOAT:
		elementsLeft = bytes / 4;
		lptr = (long *)data;
		while (elementsLeft--) {
		    SwapL(lptr++);
		}
		break;
		
	    case IFDDOUBLE:
		elementsLeft = bytes / 8;
		dptr = (double *)data;
		while (elementsLeft--) {
		    SwapD(dptr++);
		}
		break;
	    }
	}
    }
    
    /* Handle case where source and destination types don't match */
    switch (bytes) {
    case 1:
	switch ((int)dataSize) {
	case 2:
	    *(short *)data = *(char *)data;
	    break;
	case 4:
	    *(long *)data = *(char *)data;
	    break;
	}
	break;
    case 2:
	switch ((int)dataSize) {
	case 4:
	    *(long *)data = *(short *)data;
	    break;
	}
	break;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* Special macro to automatically pass in the size of the destination
 */
#define ReadFieldData(field) \
    wp_TifReadFieldData(xImg, fieldEntry, &(field), sizeof(field), swap);



static int CDECL
wp_TifParseField(
    XImg *xImg,
    FieldEntry *fieldEntry,
    TiffFields *fieldList,
    UInt32 fieldOffset,
    int swap
)
{
    Instance *instData = XIF_IMGINST(xImg);
    PROLOGUE;
    
    switch (fieldEntry->tagID) {
    case NEWSUBFILETYPE:
	ReadFieldData(fieldList->NewSubFileType);
	break;
    case SUBFILETYPE:
	break;
    case IMAGEWIDTH:
	ReadFieldData(fieldList->ImageWidth);
	break;
    case IMAGELENGTH:
	ReadFieldData(fieldList->ImageLength);
	break;
    case BITSPERSAMPLE:
	CHECK_MALLOC(fieldList->BitsPerSample =
		     XIF_MALLOC(fieldEntry->len * FieldTypeSize[fieldEntry->type]));
	wp_TifReadFieldData(xImg, fieldEntry, fieldList->BitsPerSample,
		       sizeof(*fieldList->BitsPerSample), swap);
	break;
    case COMPRESSION:
	ReadFieldData(fieldList->Compression);
	break;
    case PHOTOMETRICINTERPRETATION:
	ReadFieldData(fieldList->PhotometricInterpretation);
	break;
    case THRESHHOLDING:
    case CELLWIDTH:
    case CELLLENGTH:
    case FILLORDER:
    case DOCUMENTNAME:
    case IMAGEDESCRIPTION:
    case MAKE:
    case MODEL:
	break;
    case STRIPOFFSETS:
	if (fieldList->StripsPerImage == 0) {
	    fieldList->StripsPerImage = fieldEntry->len;
	} else {
	    XERR_GEN_CLEANUP(fieldList->StripsPerImage != fieldEntry->len,
		XIF_BADFILEFORMAT);	    
	}
	CHECK_MALLOC(fieldList->StripOffsets =
		     XIF_MALLOC(fieldList->StripsPerImage *
		     sizeof(*fieldList->StripOffsets)));
	wp_TifReadFieldData(xImg, fieldEntry, fieldList->StripOffsets,
		       sizeof(*fieldList->StripOffsets), swap);
	break;
    case ORIENTATION:
	break;
    case SAMPLESPERPIXEL:
	ReadFieldData(fieldList->SamplesPerPixel);
	break;
    case ROWSPERSTRIP:
	ReadFieldData(fieldList->RowsPerStrip);
	break;
    case STRIPBYTECOUNTS:
	if (fieldList->StripsPerImage == 0) {
	    fieldList->StripsPerImage = fieldEntry->len;
	} else {
	    XERR_GEN_CLEANUP(fieldList->StripsPerImage != fieldEntry->len,
		XIF_BADFILEFORMAT);	    
	}
	CHECK_MALLOC(fieldList->StripByteCounts =
		     XIF_MALLOC(fieldList->StripsPerImage *
		     sizeof(*fieldList->StripByteCounts)));
	wp_TifReadFieldData(xImg, fieldEntry, fieldList->StripByteCounts,
		       sizeof(*fieldList->StripByteCounts), swap);
	break;
    case MINSAMPLEVALUE:
    case MAXSAMPLEVALUE:
	break;
    case XRESOLUTION:
	ReadFieldData(fieldList->XResolution);
	break;
    case YRESOLUTION:
	ReadFieldData(fieldList->YResolution);
	break;
    case PLANARCONFIGURATION:
	ReadFieldData(fieldList->PlanarConfiguration);
	break;
    case PAGENAME:
	break;
    case XPOSITION:
	ReadFieldData(fieldList->XPosition);
	break;
    case YPOSITION:
	ReadFieldData(fieldList->YPosition);
	break;
    case FREEOFFSETS:
    case FREEBYTECOUNTS:
    case GRAYRESPONSEUNIT:
    case GRAYRESPONSECURVE:
    case T4OPTIONS:
    case T6OPTIONS:
    case RESOLUTIONUNIT:
    case PAGENUMBER:
    case COLORRESPONSEUNIT:
    case TRANSFERFUNCTION:
    case SOFTWARE:
    case DATETIME:
    case ARTIST:
    case HOSTCOMPUTER:
    case PREDICTOR:
    case WHITEPOINT:
    case PRIMARYCHROMATICITIES:
	break;
    case COLORMAP:
	XERR_GEN_CLEANUP((FieldTypeSize[fieldEntry->type] != 2), XIF_BADFILEFORMAT);
	CHECK_MALLOC(fieldList->Colormap =
		     XIF_MALLOC(fieldEntry->len * sizeof(*fieldList->Colormap)));
	memset(fieldList->Colormap, 0, fieldEntry->len * sizeof(*fieldList->Colormap));
	wp_TifReadFieldData(xImg, fieldEntry, fieldList->Colormap,
		       sizeof(*fieldList->Colormap), swap);
	break;
    case HALFTONEHINTS:
	break;
    case TILEWIDTH:
	ReadFieldData(fieldList->TileWidth);
	break;
    case TILELENGTH:
	ReadFieldData(fieldList->TileLength);
	break;
    case TILEOFFSETS:
	if (fieldList->TileCount == 0) {
	    fieldList->StripsPerImage = fieldList->TileCount = fieldEntry->len;
	} else if (fieldList->TileCount != fieldEntry->len) {
	    __error__ = XIF_BADFILEFORMAT;
	    goto ERR_CLEANUP;
	}
	CHECK_MALLOC(fieldList->TileOffsets =
		     XIF_MALLOC(fieldEntry->len * FieldTypeSize[fieldEntry->type]));
	wp_TifReadFieldData(xImg, fieldEntry, fieldList->TileOffsets,
		       sizeof(*fieldList->TileOffsets), swap);
	break;
    case TILEBYTECOUNTS:
	if (fieldList->TileCount == 0) {
	    fieldList->StripsPerImage = fieldList->TileCount = fieldEntry->len;
	} else if (fieldList->TileCount != fieldEntry->len) {
	    __error__ = XIF_BADFILEFORMAT;
	    goto ERR_CLEANUP;
	}
	CHECK_MALLOC(fieldList->TileByteCounts =
		     XIF_MALLOC(fieldEntry->len * FieldTypeSize[fieldEntry->type]));
	wp_TifReadFieldData(xImg, fieldEntry, fieldList->TileByteCounts,
		       sizeof(*fieldList->TileByteCounts), swap);
	break;
    case SUBIFDS:
	fieldList->SubIFDCount = fieldEntry->len;
	CHECK_MALLOC(fieldList->SubIFDs =
		     XIF_MALLOC(fieldEntry->len * FieldTypeSize[fieldEntry->type]));
	wp_TifReadFieldData(xImg, fieldEntry, fieldList->SubIFDs,
		       sizeof(*fieldList->SubIFDs), swap);
	if (fieldEntry->len <= 1) {
	    xImg->subIFDsTagOffset = fieldOffset + 8;
	} else {
	    xImg->subIFDsTagOffset = fieldEntry->offset;
	}
	break;
    case INKSET:
    case INKNAMES:
    case NUMBEROFINKS:
    case DOTRANGE:
    case TARGETPRINTER:
    case EXTRASAMPLES:
    case SAMPLEFORMAT:
    case SMINSAMPLEVALUE:
    case SMAXSAMPLEVALUE:
    case TRANSFERRANGE:
    case JPEGPROC:
    case JPEGINTERCHANGEFORMAT:
    case JPEGINTERCHANGEFORMATLENGTH:
    case JPEGRESTARTINTERVAL:
    case JPEGLOSSLESSPREDICTORS:
    case JPEGPOINTTRANSFORMS:
    case JPEGQTABLES:
    case JPEGDCTABLES:
    case JPEGACTABLES:
    case YCBCRCOEFFICIENTS:
    case YCBCRSUBSAMPLING:
    case YCBCRPOSITIONING:
    case REFERENCEBLACKWHITE:
    case COPYRIGHT:
	break;
    case ANNOTATIONOFFSETS:
	fieldList->AnnotationCount = (UInt16)fieldEntry->len;
	if (fieldList->AnnotationCount > 0) {
	    CHECK_MALLOC(fieldList->AnnotationOffsets =
		XIF_MALLOC(fieldList->AnnotationCount *
		sizeof(*fieldList->AnnotationOffsets)));
	    wp_TifReadFieldData(xImg, fieldEntry, fieldList->AnnotationOffsets,
		sizeof(*fieldList->AnnotationOffsets), swap);
	    xImg->annotationOffsetsCountOffset = fieldOffset + 4;
	    if (fieldEntry->len <= 1) {
		xImg->annotationOffsetsDataOffset = fieldOffset + 8;
	    } else {
		xImg->annotationOffsetsDataOffset = fieldEntry->offset;
	    }
	} /* Annotations attached */
	break;
    case MASKIFDS:
	ReadFieldData(fieldList->MaskIFDs);
	xImg->maskIFDsTagOffset = fieldOffset + 8;
	break;
    case IMAGELAYER:
	ReadFieldData(fieldList->ImageLayer);

	/* Remember the field data location */
	xImg->imageLayerLocation = fieldEntry->offset;
	break;
    case IMAGELAYERCONTENT:
	fieldList->ILContent.layersPresent = fieldEntry->len;
	if (fieldList->ILContent.imageCounts) {
	    XIF_FREE(fieldList->ILContent.imageCounts);
	}
	if (fieldList->ILContent.layersPresent > 0) {
	    CHECK_MALLOC(fieldList->ILContent.imageCounts = (UInt32 *)
		XIF_MALLOC(fieldList->ILContent.layersPresent * sizeof(UInt32)));
	    ReadFieldData(fieldList->ILContent.imageCounts[0]);
	}
	xImg->ILContentTagOffset = fieldOffset + 8;
	if (fieldEntry->len <= 1) {
	    xImg->ILContentDataOffset = fieldOffset + 8;
	} else {
	    xImg->ILContentDataOffset = fieldEntry->offset;
	}
	break;
    case RESERVEDTAG:
	break;
    } /* which field ID */
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


static int CDECL
wp_TifFieldList2Info(
    XImg *xImg,
    XIF_IMGINFO *imgInfo
)
{
    Int32 baseXres;
    Int32 baseYres;
    TiffFields *fieldList = &xImg->fields;
    XPage *xPage;
    PROLOGUE;
    
    imgInfo->dwXResolution = Rational2Long(fieldList->XResolution);
    imgInfo->dwYResolution = Rational2Long(fieldList->YResolution);
    baseXres = ((XImg *)xImg->xPage->pageInfo.xPrimImgHandle)->
	imgInfo.dwXResolution;
    baseYres = ((XImg *)xImg->xPage->pageInfo.xPrimImgHandle)->
	imgInfo.dwYResolution;
    imgInfo->dwXOffset = (UInt32)(Rational2Float(fieldList->XPosition) *
			      (float)baseXres + 0.5);
    imgInfo->dwYOffset = (UInt32)(Rational2Float(fieldList->YPosition) *
			      (float)baseYres + 0.5);
    imgInfo->dwWidth = fieldList->ImageWidth;
    imgInfo->dwHeight = fieldList->ImageLength;
    if (imgInfo->xImgType == XIF_IMGTYPE_PRIMARY) {
	Int32 layer;

	xPage = xImg->xPage;
	for (layer = fieldList->ILContent.layersPresent; layer > 0; layer--) {
	    XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage, layer,
		fieldList->ILContent.imageCounts[layer-1]));
	}
	if (xPage->xDoc->docInfo.version == 2) {
	    xPage->pageInfo.wAnnotationCount = fieldList->AnnotationCount;
	}
    }
    imgInfo->dwMaskImages = 0;
    if (fieldList->RowsPerStrip) {
	imgInfo->dwBestStripHeight = fieldList->RowsPerStrip;
    } else if (fieldList->TileCount) {
	if (fieldList->TileLength) {
	    imgInfo->dwBestStripHeight = fieldList->TileLength;
	} else {
	    imgInfo->dwBestStripHeight = 0;	/* ??? */
	}
    } else {
	imgInfo->dwBestStripHeight = fieldList->ImageLength;
    }
    
    switch (fieldList->Compression) {
    case XTIFF_NOCOMPRESSBYTE:
	imgInfo->xCompression = XIF_NOC;
	break;
    case XTIFF_CCITT3:
    case XTIFF_FAXCCITT3:
    case XTIFF_FAXCCITT4:
    case XTIFF_FAXCCITT3_MR:
	imgInfo->xCompression = XIF_G4;
	break;
    case XTIFF_JPEG:
	imgInfo->xCompression = XIF_JPEG;
	break;
    case XTIFF_XEROX_TOKENIZATION:
	imgInfo->xCompression = XIF_TOKENIZATION;
	break;
    case XTIFF_XEROX_LLTOKENIZATION:
	imgInfo->xCompression = XIF_LLTOKENIZATION;
	break;
    case XTIFF_XEROX_WAVELET:
	imgInfo->xCompression = XIF_WAVELET;
	break;
    case XTIFF_XEROX_LOSSYDITHER:
	imgInfo->xCompression = XIF_LOSSYDITHER;
	break;
    case XTIFF_LZW:
    case XTIFF_NOCOMPRESSWORD:
    case XTIFF_PACKBITS:
    case XTIFF_XEROX_HVQ:
    default:
	imgInfo->xCompression = XIF_COMP_UNKNOWN;
    }
    switch (fieldList->SamplesPerPixel) {
    case 1:
	switch (fieldList->BitsPerSample? fieldList->BitsPerSample[0]: 1) {
	case 1:
	    imgInfo->xColorType = XIF_COLORTYPE_BINARY;
	    break;
	case 4:
	    if (fieldList->Colormap) {
		imgInfo->xColorType = XIF_COLORTYPE_COLOR4;
	    } else {
		imgInfo->xColorType = XIF_COLORTYPE_GRAY4;
	    }
	    break;
	case 8:
	    if (fieldList->Colormap) {
		imgInfo->xColorType = XIF_COLORTYPE_COLOR8;
	    } else {
		imgInfo->xColorType = XIF_COLORTYPE_GRAY8;
	    }
	    break;
	default:
	    imgInfo->xColorType = XIF_COLORTYPE_NONE;
	}
	break;
    case 3:
	switch (fieldList->BitsPerSample? fieldList->BitsPerSample[0]: 1) {
	case 8:
	    imgInfo->xColorType = XIF_COLORTYPE_COLOR24;
	    break;
	default:
	    imgInfo->xColorType = XIF_COLORTYPE_NONE;
	}
    } /* Samples/pixel */
    
    /* Convert colormap */
    if (imgInfo->xColorType == XIF_COLORTYPE_COLOR4 ||
	imgInfo->xColorType == XIF_COLORTYPE_COLOR8) {
	{
	    UInt32 component;
	    UInt32 components = 1 << fieldList->BitsPerSample[0];
	    UInt16 *red, *green, *blue;
	    UInt8 *dstCPtr = imgInfo->pPalette;
	    
	    /* Missing colormap */
	    XERR_GEN_CLEANUP((fieldList->Colormap == NULL), XIF_BADFILEFORMAT);
	    XERR_GEN_CLEANUP((components !=
		      (1UL << XIF_BPPE(imgInfo->xColorType))),
		     XIF_BADFILEFORMAT);

	    red = fieldList->Colormap;
	    green = red + components;
	    blue = green + components;
	    for (component = 0; component < components; component++) {
		*dstCPtr++ = (UInt8)((*red++) >> 8);
		*dstCPtr++ = (UInt8)((*green++) >> 8);
		*dstCPtr++ = (UInt8)((*blue++) >> 8);
	    }
	} /* Convert colormap */
    }

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the nextIFD offset
 *
 * LongDescription:::
 *  wp_TifGetNextIFD() reads the nextIFD stored as part of the IFD for the
 *  given image.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifGetNextIFD(
    XImg *xImg
)
{
    XFile *xFile;
    UInt16 fieldCount;
    int swap = 0;
    PROLOGUE;
    
    CheckHandleValidity(xImg);
    xFile = XIF_IMGFILE(xImg);
    swap = XIF_FILETOKEN(xFile)->bSwapBytes;
    
    /* Go to the beginning of the IFD */
    XIF_SEEK(xFile, xImg->ifdOffset);
    
    /* Read the field count */
    XIF_READ(xFile, &fieldCount, sizeof(fieldCount));
    SwapShort(xFile, &fieldCount);
    xImg->fields.fieldCount = fieldCount;
    
    /* Go past the fields */
    XIF_SEEK(xFile, xImg->ifdOffset + sizeof(fieldCount) + fieldCount * XTIFF_FIELDSIZE);
    
    /* Read the next IFD */
    XIF_READ(xFile, &xImg->fields.nextIFD, sizeof(xImg->fields.nextIFD));
    SwapLong(xFile, &xImg->fields.nextIFD);
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the TIFF fields
 *
 * LongDescription:::
 *  wp_TifReadFields() reads the IFD fields, and copies over the pertinent
 *  information regarding the image.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifReadFields(
    XImg *xImg
)
{
    XFile *xFile;
    UInt16 fieldCount;
    UInt16 fieldNum;
    FieldEntry fieldEntry;
    XIF_IMGINFO *pxImgInfo = NULL;
    UInt32 fieldSize;
    int swap = 0;
    Int32 fileOffset;
    PROLOGUE;
    
    CheckHandleValidity(xImg);
    xFile = XIF_IMGFILE(xImg);
    pxImgInfo = &xImg->imgInfo;
    swap = XIF_FILETOKEN(xFile)->bSwapBytes;
    
    /* Use default values */
    memset(&xImg->fields, 0, sizeof(xImg->fields));	/* 0 for everything */
    xImg->fields.SamplesPerPixel = 1;
    xImg->fields.XResolution.numerator =
	xImg->fields.YResolution.numerator = 300;
    xImg->fields.XResolution.denominator =
	xImg->fields.YResolution.denominator = 1;
    
    
    /* Go to the beginning of the IFD */
    XIF_SEEK(xFile, xImg->ifdOffset);
    
    /* Read the field count */
    XIF_READ(xFile, &fieldCount, sizeof(fieldCount));
    SwapShort(xFile, &fieldCount);
    
    fieldSize = sizeof(fieldEntry.tagID) + sizeof(fieldEntry.type) +
	sizeof(fieldEntry.len) + sizeof(fieldEntry.offset);
    
#ifdef DEBUG
    printf("\tField\tID\tType\tLength\tOffset\n");
#endif
    
    /* Loop through the fields */
    for (fieldNum=0; fieldNum < fieldCount; fieldNum++) {
	
	/* Go to the beginning of the field */
	fileOffset = xImg->ifdOffset + sizeof(fieldCount) + fieldNum * fieldSize;
	XIF_SEEK(xFile, fileOffset);
	
	/* Read a field */
	XIF_READ(xFile, &fieldEntry.tagID, sizeof(fieldEntry.tagID));
	SwapShort(xFile, &fieldEntry.tagID);
	XIF_READ(xFile, &fieldEntry.type, sizeof(fieldEntry.type));
	SwapShort(xFile, &fieldEntry.type);
	XIF_READ(xFile, &fieldEntry.len, sizeof(fieldEntry.len));
	SwapLong(xFile, &fieldEntry.len);
	XIF_READ(xFile, &fieldEntry.offset, sizeof(fieldEntry.offset));
	SwapLong(xFile, &fieldEntry.offset);
	
#ifdef DEBUG
	printf("\t%2d:\t%2d\t%2d%8d%8d\n", fieldNum, fieldEntry.fieldID,
	       fieldEntry.type, fieldEntry.len, fieldEntry.offset);
#endif
	
	XERR_PASS_CLEANUP(wp_TifParseField(xImg, &fieldEntry, &xImg->fields, fileOffset, swap));
    }
    
    XERR_PASS_CLEANUP(wp_TifFieldList2Info(xImg, pxImgInfo));
    
    /*
      xImg->minBytesPerRow = ((xImg->imgInfo.dwWidth *
      XIF_BPPE(xImg->imgInfo.dwColorType) + 31) / 8) & ~3;
      */
    xImg->minBytesPerRow = (xImg->imgInfo.dwWidth *
		    XIF_BPPE(xImg->imgInfo.xColorType) + 7) / 8;
    
    /* Go past the fields */
    XIF_SEEK(xFile, xImg->ifdOffset + sizeof(fieldCount) + fieldCount * fieldSize);
    
    /* Read the next IFD */
    XIF_READ(xFile, &xImg->fields.nextIFD, sizeof(xImg->fields.nextIFD));
    SwapLong(xFile, &xImg->fields.nextIFD);

    xImg->fields.dataTransferred = 1 /* cTrue */;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the entire IFD
 *
 * LongDescription:::
 *  wp_TifReadIFD() reads the entire IFD and stores the pertinent data.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifReadIFD(
    XImg *xImg
)
{
    PROLOGUE;
    
    CheckHandleValidity(xImg);

    if (xImg->fields.dataTransferred == 0) {
    
	XERR_PASS_CLEANUP(wp_TifReadFields(xImg));
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

/*
$Log: xifrtiff.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.19  1998/01/20 03:22:32  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Rename MixedRasterContent field to ImageLayerContent, per Lloyd's suggestion.

Revision 1.18  1997/12/18 00:12:39  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Remember the ImageLayer TIFF field data location, in case the image
ordering has to be changed (as in the case of a deleted image).

Revision 1.17  1997/10/15 21:35:44  aber
aber on Wed Oct 15 14:30:51 PDT 1997

xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Modify return parameters from IsTIFF, allow no first IFD.

Revision 1.16  1997/09/02 20:02:52  aber
aber on Tue Sep 2 13:02:52 PDT 1997

xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.15  1997/08/29 17:28:03  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.14  1997/07/10 21:59:11  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Use the field count offset to update the count also.

Revision 1.13  1997/07/02 17:03:37  aber
aber on Wed Jul 2 10:02:25 PDT 1997

xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Have type conversion routine return a value (can't use void, since
a call to this function MUST have a return value checked, otherwise
coreCheckC complains that the return value isn't used).

Revision 1.12  1997/07/02 01:05:32  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Use private names for swap functions.

Revision 1.11  1997/06/26 22:26:12  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Change return values to indicate "not XIFF".

Revision 1.10  1997/06/23 22:59:05  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

CommonCheck: Complete function calls for paletted image support.

Revision 1.9  1997/06/20 17:47:15  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Round the X/Y position vs. truncation.

Revision 1.8  1997/06/17 22:34:05  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Annotation count type cast.

Revision 1.7  1997/06/12 00:50:53  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

Revision 1.6  1997/06/10 01:26:45  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:05  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:09  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:14  aber
xifrtiff.c -> /product/ipcore/ipwrap/src/RCS/xifrtiff.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:39  aber
Initial revision

*/
