/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifrtok.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifrtok.c:	XIFF READ TOKENIZATION source
 *
 *	XIFF read-only routines for Tokenization.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "shrcmprs.pub"
#include "iaerror.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifrtiff.h"
#include "xifrtok.h"
#include "xifrpriv.h"
#include "xifrxiff.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifrtok.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

static Instance *globalInstData;
static void * CDECL
wp_XifTextMalloc(
	   IPWState *pIPWState,
	   UInt32 bytes,
	   UInt32 fixupPtrs
)
{
    Instance *instData = globalInstData;
    void *buf = NULL;
    PROLOGUE;
    pIPWState = pIPWState;

    CHECK_MALLOC(buf = XIF_MALLOC(bytes));
    memset(buf, 0, bytes);

    if (fixupPtrs) {
      return (void *)buf;
    }

ERR_CLEANUP:

    return (void *)buf;
}


static void CDECL
wp_XifTextFree(
	 IPWState *pIPWState,
	 void *buf,
	 UInt32 len
)
{
    Instance *instData = globalInstData;
    pIPWState = pIPWState;

    if (len == 0) {
      return;
    }

    XIF_FREE(buf);

    return;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts the compressed Tundra data into a binary image
 *
 * LongDescription:::
 *  wp_XifDetokenizeImage() converts a JBIG-2 Tundra position/ID block to
 *  positions and IDs, used by the IPShared routines, and "stamps" them
 *  back onto a blank image. The data is assumed to be found in the
 *  cmpBuffer/cmpBytes structure elements of the xImg pointer passed in.
 *  The pointer to the image (pImgData) must point to the entire binary image.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *  pImgData::			Pointer to image data
 *  dwBytesPerRow::		Bytes per image row of image data
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifDetokenizeImage(
    XImg *xImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow
)
{
    Instance *instData;
    XFile *xFile;
    XDoc *xDoc;
    XIF_IMGINFO *pxImgInfo;
    void *textInfo;
    UInt8 *tundraData;
    UInt32 tundraBytes;
    UInt32 w, h;
    UInt16 dictNum, dictCount;
    UInt32 *dictIDs = NULL;
    Bool reload_dictionaries;
    IPWState *pIPWState;
    PROLOGUE;

    CheckHandleValidity(xImg);
    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);
    xDoc = xImg->xPage->xDoc;
    pxImgInfo = &xImg->imgInfo;

    /* Extract the shared IPCore state */
    XERR_PASS_CLEANUP(wp_XifIPWState(xDoc, &pIPWState));

    /* Get the header information */
    XERR_PASS_CLEANUP(wp_XifInterpretTokenHeader(xFile, pCmpData,
	dwCmpBytes, &dictCount, &dictIDs, &tundraData, &tundraBytes));

    if ((xImg->dictCount = dictCount) == 0) {
	/* No token dictionary (blank image) */
	NORMAL_RETURN;
    }

    if (xImg->imgInfo.xColorType != XIF_COLORTYPE_BINARY) {
	/* We could be nasty and return a real error. We'll be nice and
	 * consider that the image is probably an up-and-coming color
	 * tokenized foreground plane, and just return (the image has
	 * already been set to zero (black), so the image will look
	 * correct when rendered with this image.
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * Fix this later. */
	NORMAL_RETURN;
    }

    /* If the loaded dictionary list is different than the current
     * required list, and in the same order, reload all dictionaries */
    textInfo = xDoc->textCompInfo;
    reload_dictionaries = cFalse;
    if (textInfo == NULL) {
	reload_dictionaries = cTrue;
    } else if (xDoc->curDictCount > 0 &&
	xDoc->curDictCount != dictCount) {
	reload_dictionaries = cTrue;
    } else {

	/* Check to see that the lists match */
	for (dictNum = 0; dictNum < dictCount; dictNum++) {
	    if (dictIDs[dictNum] != xDoc->curDictIDs[dictNum]) {
		reload_dictionaries = cTrue;
		break;
	    }
	}
    }

    /* Free everything, abandons the current token dictionaries */
    if (textInfo && reload_dictionaries) {
	i_freeTextCompressionInfo (pIPWState, textInfo); /* Ignore return */
	textInfo = xDoc->textCompInfo = NULL;
	xDoc->curDictCount = 0;
    }

    /* Initiate a new tokenization process */
    if (textInfo == NULL) {
	/* Work around a client ID problem (for now) */
	globalInstData = XIF_IMGINST(xImg);

	XERR_PASS_CLEANUP(i_allocateTextCompressionInfo (pIPWState,
	    &textInfo, wp_XifTextMalloc, wp_XifTextFree));
	xDoc->textCompInfo = (UInt8 *)textInfo;
    }

    /* If appropriate, read the dictionary used for this image */
    if (reload_dictionaries) {

	/* Free the old "current" list */
	if (xDoc->curDictIDs) {
	    XIF_FREE(xDoc->curDictIDs);
	}

	/* Allocate a new list */
	CHECK_MALLOC(xDoc->curDictIDs = XIF_MALLOC(dictCount * sizeof(UInt32)));

	/* Load each of them */
	for (dictNum = 0; dictNum < dictCount; dictNum++) {
	    XERR_PASS_CLEANUP(wp_XifReadTokenDictionary(xDoc, dictIDs[dictNum]));
	    xDoc->curDictIDs[dictNum] = dictIDs[dictNum];
	    xDoc->curDictCount++;
	}
    }

    /* Decode the position blocks */
    pImgData -= 32 * dwBytesPerRow + 4;
    w = pxImgInfo->dwWidth + 64;
    h = pxImgInfo->dwHeight + 64;
    XERR_PASS_CLEANUP(i_readTundraPositionBlock (pIPWState, textInfo,
	tundraData, tundraBytes, pImgData, w, h, dwBytesPerRow));

ERR_CLEANUP:

    if (dictIDs) {
	XIF_FREE(dictIDs);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Interpret the token dictionary header
 *
 * LongDescription:::
 *  wp_XifInterpretTokenHeader() reads the non-Tundra header placed at the
 *  beginning of the token image data stream. The dictionary count, dictionary
 *  ID, pointer to the Tundra data portion, and Tundra data byte count are
 *  returned.
 *
 * Input:::
 *  xFile::			Active file handle
 *  pCmpData::			Tokenized data stream
 *  dwCmpBytes::		Number of bytes in the tokenized data stream
 *
 * Output:::
 *  ppdwDictIDList::		Dictionary identifier list, TO BE FREED BY
 *				CLIENT
 *  pdwDictCount::		Number of dictionary associated with Tundra
 *				data
 *  ppTundraData::		Pointer to the Tundra data section (byte count
 *				implied by the delta from pCmpData)
 *  pdwTundraBytes::		Number of bytes for Tundra data (note, with
 *				position block extensions, the byte count may
 *				not add up to the token stream byte count)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifInterpretTokenHeader(
    XFile *xFile,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt16 *pdwDictCount,
    UInt32 **ppdwDictIDList,
    UInt8 **ppTundraData,
    UInt32 *pdwTundraBytes
)
{
    Instance *instData;
    UInt8 *ptr, positionVersion;
    UInt16 dictNum, dictCount;
    UInt32 tundraBytes;
    UInt32 dictID, *newIDs = NULL;
    PROLOGUE;

    /* Return nothing initally */
    *pdwDictCount = 0;
    *ppdwDictIDList = NULL;
    *ppTundraData = pCmpData;

    XERR_GEN(dwCmpBytes < 6, XIF_BADPARAMETER);
    instData = xFile->instance;

    /* Read the pre-Tundra header, also stored in xifwpub.c */
    ptr = pCmpData;
    memcpy(&positionVersion, ptr + cXifTCPosPositionVersion,
	   sizeof(positionVersion));
    memcpy(&tundraBytes, ptr + cXifTCPosPositionSize, sizeof(tundraBytes));
    SwapLong(xFile, &tundraBytes);
    *pdwTundraBytes = tundraBytes;
    memcpy(&dictCount, ptr + cXifTCPosPositionDictCount, sizeof(dictCount));
    SwapShort(xFile, &dictCount);
    *pdwDictCount = dictCount;
    if (*pdwDictCount == 0) {	/* No token dictionary (blank image) */
	*ppTundraData = ptr + cXifTCPosPositionDictIDs;
    } else {

	/* Allocate a list of new dictionary IDs */
	CHECK_MALLOC(newIDs = XIF_MALLOC(dictCount * sizeof(UInt32)));

	/* Accumulate the IDs in a list */
	for (dictNum = 0; dictNum < dictCount; dictNum++) {
	    memcpy (&dictID, ptr + cXifTCPosPositionDictIDs + dictNum * sizeof(dictID),
		sizeof(UInt32));
	    SwapLong(xFile, &dictID);
	    newIDs[dictNum] = dictID;
	}
	
	/* Return the results */
	*ppdwDictIDList = newIDs;
	newIDs = NULL;
	*ppTundraData = ptr + cXifTCPosPositionDictIDs + dictCount * sizeof(dictID);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if (newIDs) {
	XIF_FREE(newIDs);
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a token dictionary
 *
 * LongDescription:::
 *  wp_XifReadTokenDictionary() reads the Tundra format for the token
 *  dictionary, and loads it as the single internal dictionary. Client is
 *  responsible for adding this ID to the list of "current" IDs.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  dictID::			Dictionary identifier
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifReadTokenDictionary(
	XDoc *xDoc,
	UInt32 dictID
)
{
    Instance *instData = NULL;
    XFile *xFile;
    void *textInfo;
    Int32 dictBytes;
    UInt8 *dictBlock = NULL;
    Int32 offset;
    UInt32 dataOffset;
    UInt8 dictVersion;
    UInt32 dictSize;
    IPWState *pIPWState;
    PROLOGUE;

    CheckHandleValidity(xDoc);

    /* Extract the shared IPCore state */
    XERR_PASS_CLEANUP(wp_XifIPWState(xDoc, &pIPWState));

    instData = XIF_DOCINST(xDoc);
    xFile= XIF_DOCFILE(xDoc);

    /* Get the offset and size of the symlib with this ID */
    XERR_PASS_CLEANUP(wp_XifGetSymlibOffset(xDoc, dictID, &offset,
					    &dictBytes));

    /* Allocate a buffer for the Tundra dictionary */
    CHECK_MALLOC(dictBlock = XIF_MALLOC(dictBytes));

    /* Read the dictionary */
    XIF_SEEK(xFile, offset);
    XIF_READ(xFile, dictBlock, dictBytes);

    /* Load it */
    textInfo = xDoc->textCompInfo;
    XERR_GEN_CLEANUP((textInfo == NULL), XIF_INTERNAL_ERROR);

    /* Extract the header */
    memcpy(&dictVersion, dictBlock + cXifTCPosDictVersion,
	   sizeof(dictVersion));
    dataOffset = sizeof(dictVersion);
    memcpy(&dictSize, dictBlock + cXifTCPosDictSize, sizeof(dictSize));
    SwapLong(xFile, &dictSize);
    dataOffset = cXifTCPosDictSize + sizeof(dictSize);
    XERR_GEN_CLEANUP(dictSize > dictBytes - dataOffset,
		     XIF_BADFILEFORMAT);

    /* Check the version for support */
    XERR_GEN_CLEANUP(dictVersion > cXifTCDictVersion, XIF_NOSUPPORT);

    XERR_PASS_CLEANUP(i_readTundraDictionaryBlock (pIPWState, textInfo,
	dictBlock + dataOffset, dictSize, dictID, cTrue));

    /* Fall through to clean up temporary memory */

 ERR_CLEANUP:

    if (dictBlock) {
	XIF_FREE(dictBlock);
	dictBlock = NULL;
    }
    NORMAL_RETURN;
}

/*
$Log: xifrtok.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.16  1998/08/28 20:30:55  aber
aber on Fri Aug 28 13:30:55 PDT 1998

xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: Detect the inconsistency of the unsupported color tokenization, and
return a NOSUPPORT error for tokenized images that aren't binary. This
was practically impossible to anticipate, but should be gracefully
handled in all future releases.

Revision 1.15  1998/01/16 00:40:23  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

Local text data allocation given privat names (wp_).

Revision 1.14  1998/01/14 00:16:33  ddavies
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.13  1997/08/27 21:50:35  aber
aber on Wed Aug 27 14:49:45 PDT 1997

xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

Update image structure with dictionary count, used to determine when
to reload token dictionary.

Revision 1.12  1997/08/16 00:15:29  aber
aber on Fri Aug 15 17:15:09 PDT 1997

xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

Forgot to swap bytes in endian conversion case.

Revision 1.11  1997/08/15 01:54:17  aber
aber on Thu Aug 14 18:54:17 PDT 1997

xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: Modify token dictionary header to include the Tunrda dictionary
byte count, allows for extensible dictionary data.

Revision 1.10  1997/08/15 00:09:44  aber
aber on Thu Aug 14 17:07:19 PDT 1997

xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

Clean up pre-Tundra header code, use predefined offsets, and
add the pre-Tundra version numbers for both the position blocks,
and the dictionaries. This will obsolete previous versions of
tokenized XIFF files, but will enable later modifications to
the tokenization stream, without breaking things.

Revision 1.9  1997/08/07 21:27:45  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.8  1997/07/29 23:51:29  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

Separate into a new function the code to parse the pre-Tundra token image
stream, allowing XIF_GetImageStats() to retrieve the dictionary ID.
/

Revision 1.7  1997/06/26 22:27:27  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

If no token dictionary associated with token position stream, know
immediately that the image is blank and return.

Revision 1.6  1997/06/10 01:26:47  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:07  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:11  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:16  aber
xifrtok.c -> /product/ipcore/ipwrap/src/RCS/xifrtok.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:40  aber
Initial revision

*/
