/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifrwave.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifrwave.c:	XIFF READ Wavelet source
 *
 *	XIFF read-only Wavelet routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "iw_wave.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifrtiff.h"
#include "xifrwave.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifrwave.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

typedef struct ReadData {
    UInt8 *buffer;
    UInt8 *next;
    UInt32 bufBytes;
    UInt32 bytesLeft;
} ReadData;

static Int32 CDECL
wp_Wavelet_ReadCallback(
    IPWState *pIPWState,
    void *userData,
    UInt8 *ptr,
    UInt32 bytes
)
{
    ReadData *readData = (ReadData *)userData;
    pIPWState = pIPWState;
    
    if (readData->bytesLeft < bytes) {
	bytes = readData->bytesLeft;
    }
    memcpy(ptr, readData->next, bytes);
    readData->next += bytes;
    readData->bytesLeft -= bytes;
    
    return cSuccessful;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompresses a strip of Wavelet data.
 *
 * LongDescription:::
 *  wp_XifDecompressStripWavelet() decompresses a strip of Wavelet data.
 *  Temporary PIXR found in xImg is used to store the image data decompressed.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *  pPixr::			Pointer to image PIXR
 *  dwStripHeight::		Number of rows to decompress
 *
 * Output:::
 *  pRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifDecompressStripWavelet(
    XImg *xImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    PIXR *pPixr,
    UInt32 dwStripHeight,
    UInt32 *pRowsFilled
)
{
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState;
    ReadData readData;
    Int32 pixrHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pRowsFilled = 0UL;

    if (dwStripHeight == 0) {
        NORMAL_RETURN;
    } else {
      /* Should be the PIXR height */
      /* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */
    }

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    pxImgInfo = &xImg->imgInfo;
    
    /* Initialize */
    readData.buffer = pCmpData;
    readData.next = readData.buffer;
    readData.bufBytes = dwCmpBytes;
    readData.bytesLeft = readData.bufBytes;

    /* Decompress the entire image */
    XERR_PASS_CLEANUP(w_waveDecompress_option(pIPWState,
					      wp_Wavelet_ReadCallback,
					      (void *)&readData, &pPixr));

    /* Since w_waveDecompress_option doesn't return the number of rows
     * successfully filled, we have to assume that the requested amount was
     * the successful number */
    *pRowsFilled = dwStripHeight;

    /* Limit this to the actual number of rows in the PIXR */
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    if (pixrHeight < dwStripHeight) {
	*pRowsFilled = pixrHeight;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    ERR_RETURN;
}

/*
$Log: xifrwave.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.9  1998/05/05 07:15:26  aber
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

CommonCheck: Additions necessary to fill in the never completed functions
XIF_Compress and XIF_Decompress.

Revision 1.8  1998/01/14 00:16:34  ddavies
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.7  1997/11/01 20:53:16  aber
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

Comment added to reflect that wavelet routine expect to fill the entire
PIXR.

Revision 1.6  1997/09/02 20:02:56  aber
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.5  1997/06/10 01:26:49  aber
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.4  1997/05/30 00:51:08  aber
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.3  1997/05/20 22:29:13  aber
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.2  1997/05/20 00:54:18  aber
xifrwave.c -> /product/ipcore/ipwrap/src/RCS/xifrwave.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:41  aber
Initial revision

*/
