/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifrxiff.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifrxiff.c:	READ XIFF source
 *
 *	XIFF read-only routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif


#include "xifimage.h"
#include "xifrtiff.h"
#include "xifrxiff.h"
#include "xifrpriv.h"

/* The idea is to remove this one day: */
#include "xiff.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: xifrxiff.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the internal data associated with an image
 *
 * LongDescription:::
 *  wp_XifImgFree() frees the data associated with an image handle, and
 *  optionally the subsequent images also. Since the images belonging to
 *  a page are stored internally in a linked list, any image freed could
 *  either have it's link removed, or all links down the chain.
 *
 * Input:::
 *  pxImg::			Data structure of valid image
 *  subsequentImagesToo::	Boolean for: free connected images also
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifImgFree(
    XImg **pxImg,
    Int32 subsequentImagesToo
)
{
    Instance *instData = NULL;
    XImg *xImg, *nextImg = NULL;
    MISC_MEM *mmptr, *next_mm;
    PROLOGUE;

    CheckHandleValidity(*pxImg);

    xImg = *pxImg;
    instData = XIF_IMGINST(xImg);

    while (xImg) {
	nextImg = xImg->next;

	if (xImg->xMaskImg) {
	    XERR_PASS_CLEANUP(wp_XifImgFree(&xImg->xMaskImg, 1));
	}

	XERR_PASS_CLEANUP(wp_TifFieldCleanup(xImg, &xImg->fields));
	mmptr = xImg->tmpCmpData;
	while (mmptr) {
	    if (mmptr->mem) {
		XIF_FREE(mmptr->mem);
	    }
	    next_mm = mmptr->next;
	    XIF_FREE(mmptr);
	    mmptr = next_mm;
	}
	if (xImg->cmpBuffer) {
	    XIF_FREE(xImg->cmpBuffer);
	    xImg->cmpBuffer = NULL, xImg->cmpBytes = 0UL;
	}
	if (xImg->tmpPixr) {
	    IPWState *pIPWState;
	    
	    /* Reference the IPWrap state */
	    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

	    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &xImg->tmpPixr));
	}
	if (xImg->dictList) {
	    XIF_FREE(xImg->dictList);
	}
	if (xImg->dictSizes) {
	    XIF_FREE(xImg->dictSizes);
	}

	XIF_FREE(xImg);

	if (! subsequentImagesToo) {
	    break;
	}

	xImg = nextImg;
    }

    *pxImg = nextImg;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the internal data associated with a page
 *
 * LongDescription:::
 *   wp_XifPageFree() frees the data associated with an page handle, and
 *  optionally the subsequent pages also. Since the pages belonging to
 *  a document are stored internally in a linked list, any page freed could
 *  either have it's link removed, or all links down the chain. All images
 *  tied to the page(s) are also freed.
 *
 * Input:::
 *  pxPage::			Pointer to data structure of valid page
 *  subsequentPagesToo::	Boolean for: free other pages along chain
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifPageFree(
    XPage **pxPage,
    Int32 subsequentPagesToo
)
{
    Instance *instData = NULL;
    XPage *xPage, *nextPage = NULL;
    AnnotTable *at, *nextAt;
    PROLOGUE;

    CheckHandleValidity(*pxPage);

    xPage = *pxPage;
    instData = XIF_PAGEINST(xPage);

    while (xPage) {
	nextPage = xPage->next;

	/* Free the images associated with this page */
	if (xPage->pageInfo.xPrimImgHandle) {
	    XERR_PASS_CLEANUP(wp_XifImgFree((XImg **)
					 &xPage->pageInfo.xPrimImgHandle, 1));
	}
	if (xPage->pageInfo.pdwImageCounts) {
	  XIF_FREE(xPage->pageInfo.pdwImageCounts);
	}
	if (xPage->extImageCounts) {
	  XIF_FREE(xPage->extImageCounts);
	}
	if (xPage->xImgs) {
	    XERR_PASS_CLEANUP(wp_XifImgFree(&xPage->xImgs, 1));
	}
	at = xPage->firstAnnotTable;
	while (at) {
	    nextAt = at->next;
	    if (at->entries) {
		XIF_FREE(at->entries);
	    }
	    XIF_FREE(at);
	    at = nextAt;
	}

	/* Free the page itself */
	XIF_FREE(xPage);

	if (! subsequentPagesToo) {
	    break;
	}

	xPage = nextPage;
    }

    *pxPage = nextPage;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free all data associated with a document handle
 *
 * LongDescription:::
 *  wp_XifDocFree() free all data associated with a document handle, including
 *  all page and image data.
 *
 * Input:::
 *  pxDoc::			Pointer to data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifDocFree(
    XDoc **pxDoc
)
{
    Instance *instData = NULL;
    XDoc *xDoc;	
    PageTable *pt, *nextPt;
    SymlibTable *st, *nextSt;
    CopiedDicts *pd, *nextPd;
    PROLOGUE;

    CheckHandleValidity(*pxDoc);

    xDoc = *pxDoc;
    instData = XIF_DOCINST(xDoc);

    if (xDoc->pages) {
	XERR_PASS_CLEANUP(wp_XifPageFree(&xDoc->pages, 1));
    }

    if (xDoc->xFile) {
	XIF_FREE(xDoc->xFile);
    }
    if (xDoc->docInfo.pAuthorBlock) {
	XIF_FREE(xDoc->docInfo.pAuthorBlock);
    }
    pt = xDoc->firstPageTable;
    while (pt) {
	nextPt = pt->next;
	if (pt->entries) {
	    XIF_FREE(pt->entries);
	}
	XIF_FREE(pt);
	pt = nextPt;
    }

    st = xDoc->firstSymlibTable;
    while (st) {
	nextSt = st->next;
	if (st->entries) {
	    XIF_FREE(st->entries);
	}
	XIF_FREE(st);
	st = nextSt;
    }

    pd = xDoc->prevTDCopied;
    while (pd) {
	nextPd = pd->next;
	XIF_FREE(pd);
	pd = nextPd;
    }

    if (xDoc->curDictIDs) {
	XIF_FREE(xDoc->curDictIDs);
    }

    XIF_FREE(xDoc);

    *pxDoc = NULL;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check to see if the file is a valid XIFF, and get the version number
 *
 * LongDescription:::
 *  wp_XifIsXiff() checks for the necessary header information to identify the
 *  file as a XIFF file. The version and revision are returned if it has
 *  the XIFF features.
 *
 * Input:::
 *  xFile::			File handle structure
 *
 * Output:::
 *  pVersion::			Version number
 *  pRevision::			Revision number
 *
 * Returns:::
 *  XIF_NOTTIFF::		File is not even a TIF
 *  XIF_NOTXIFF::		File is not ag XIF
 *  XIF_NOSUPPORT::		File appears to be a unsupported version of XIF
 *  XIF_NOERROR::		File is a supported XIF
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifIsXiff(
	XFile *xFile,
	UInt8 *pVersion,
	UInt8 *pRevision
)
{
    Int8 xiffKey[10];
    UInt32 firstIFD;
    PROLOGUE;

    /* Return nothing until something found */
    *pVersion = 0UL;
    *pRevision = 0UL;

    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_XifFileLock(XIF_FILETOKEN(xFile)));

    /* Check the TIFF header */
    XERR_PASS_CLEANUP(wp_TifIsTiff(xFile, &firstIFD));

    /* Now check the XIFF header */
    XIF_SEEK(xFile, 8UL);
    XIF_READ(xFile, xiffKey, sizeof(xiffKey));
    XIF_READ(xFile, pVersion, sizeof(*pVersion));
    XIF_READ(xFile, pRevision, sizeof(*pRevision));

    /* Assume not a XIF */
    __error__ = XIF_NOTXIFF;

    /* Check for a supported version */
    if (! strncmp((char *)xiffKey, XIFF_KEY_2_0, sizeof(xiffKey))) {
	if (*pVersion == 2 && *pRevision == 0) {	/* 2.0 only */
	    __error__ = XIF_NOERROR;
	} else {
	    __error__ = XIF_NOTXIFF;
	}
    } else if (! strncmp((char *)xiffKey, XIFF_KEY_3_0, sizeof(xiffKey))) {
	if (*pVersion == 3) {		/* Assume support for all 3.x */
	    __error__ = XIF_NOERROR;
	} else {
	    __error__ = XIF_NOSUPPORT;
	}
    }

    NORMAL_RETURN;

 ERR_CLEANUP:


    wp_XifFileUnLock(XIF_FILETOKEN(xFile));
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the XIFF fields (DocTags) from a single field list
 *
 * LongDescription:::
 *  wp_XifReadExtensionList() reads the XIFF extension fields in one set of
 *  DocTags, and recursively call the routine again for each NEXTEXT field
 *  found. Requires the file offset to be that of the document field count.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  recursionLevel::		Number of lists read so far
 *  fileOffset::		Current file offset
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifReadExtensionList(
    XDoc *xDoc,
    UInt32 recursionLevel,
    Int32 fileOffset
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt16 fieldCount;
    UInt32 nextExtOffset = 0UL;
    DocField docField;
    PROLOGUE;

    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((recursionLevel > 100), XIF_BADPARAMETER);

    /* Read the document field count */
    XIF_READ(xFile, &fieldCount, sizeof(fieldCount));
    SwapShort(xFile, &fieldCount);
    fileOffset += sizeof(fieldCount);

    /* Read the fields and fill in the document information from them */
    while (fieldCount--) {

	/* Make sure we're on a 4-byte boundary */

	XIF_READ(xFile, &docField.tagID, sizeof(docField.tagID));
	XIF_READ(xFile, &docField.offset, sizeof(docField.offset));
	SwapShort(xFile, &docField.tagID);
	SwapLong(xFile, &docField.offset);

	switch (docField.tagID) {
	case DOCTAG_NULL:
	    /* Only a placeholder for later docfields */
	    xDoc->tagOffset[docField.tagID] = fileOffset;
	    break;

	case DOCTAG_PAGETABLE:

	    /* Page table header */
	    xDoc->tagOffset[docField.tagID] = fileOffset;
	    xDoc->pageTableHeaderOffset = docField.offset;
	    break;

	case DOCTAG_AUTHOR:

	    /* Author block of data */
	    xDoc->tagOffset[docField.tagID] = fileOffset;
	    xDoc->authorOffset = docField.offset;
	    break;

	case DOCTAG_REFERENCEPAGE:

	    /* Reference page offset */
	    xDoc->tagOffset[docField.tagID] = fileOffset;
	    xDoc->refPageOffset = docField.offset;
	    break;

	case DOCTAG_NEXTEXT:

	    /* Next list of extensions */
	    xDoc->tagOffset[docField.tagID] = fileOffset;
	    nextExtOffset = docField.offset;
	    break;

	case DOCTAG_SYMLIBTABLE:

	    /* First symbol library */
	    xDoc->tagOffset[docField.tagID] = fileOffset;
	    xDoc->symlibTableHeaderOffset = docField.offset;
	    break;

#ifdef _DEBUG
	default:
	    /* Unrecognized field, skip */
	    fprintf(stderr, "Unrecognized docfield: %d\n", docField.tagID);
#endif
	}

	/* Update the offset to the beginning of the next field */
	fileOffset += sizeof(docField.tagID) + sizeof(docField.offset);
    }

    /* If more docfields, seek there and read them too */
    if (nextExtOffset) {
	XIF_SEEK(xFile, nextExtOffset);

	/* Recursively read the extension list */
	XERR_PASS_CLEANUP(wp_XifReadExtensionList(xDoc,
	    recursionLevel + 1, nextExtOffset));
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the data associated with the XIFF fields
 *
 * LongDescription:::
 *  wp_XifReadTagData() reads the data associated with the XIFF fields. The
 *  location is that derived from the previously read fields.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifReadTagData(
    XDoc *xDoc
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt16 sval;
    PROLOGUE;

    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);

    /* Check for proper calling sequence */


    /* Read the page table header */
    if (xDoc->pageTableHeaderOffset != 0UL) {
	XIF_SEEK(xFile, xDoc->pageTableHeaderOffset);
	XIF_READ(xFile, &xDoc->pageTableHeader.pageCount,
	    sizeof(xDoc->pageTableHeader.pageCount));
	SwapShort(xFile, &xDoc->pageTableHeader.pageCount);
	XIF_READ(xFile, &xDoc->pageTableHeader.entriesPerTable,
	    sizeof(xDoc->pageTableHeader.entriesPerTable));
	SwapShort(xFile, &xDoc->pageTableHeader.entriesPerTable);
	XIF_READ(xFile, &xDoc->pageTableHeader.lastOrdinal,
	    sizeof(xDoc->pageTableHeader.lastOrdinal));
	SwapShort(xFile, &xDoc->pageTableHeader.lastOrdinal);
	XIF_READ(xFile, &xDoc->pageTableHeader.firstTableOffset,
	    sizeof(xDoc->pageTableHeader.firstTableOffset));
	SwapLong(xFile, &xDoc->pageTableHeader.firstTableOffset);

	/* Read the page table, handled separately, due to multiple table entries */
	if (xDoc->pageTableHeader.pageCount > 0) {
	    XERR_PASS_CLEANUP(wp_XifReadPageTable(xDoc, 0UL, NULL));
	}

	/* Fill in the page count */
	xDoc->docInfo.wNPages = xDoc->pageTableHeader.pageCount;
    } else {
	xDoc->docInfo.wNPages = 0;
    }

    /* Read the author block */
    if (xDoc->authorOffset) {
	XIF_SEEK(xFile, xDoc->authorOffset);
	XIF_READ(xFile, &sval, sizeof(sval));
	SwapShort(xFile, &sval);
	xDoc->docInfo.wAuthorBlockBytes = sval;
	if (xDoc->docInfo.wAuthorBlockBytes > 0) {
	    CHECK_MALLOC(xDoc->docInfo.pAuthorBlock = (Int8 *)
			 XIF_MALLOC(xDoc->docInfo.wAuthorBlockBytes));
	    XIF_READ(xFile, xDoc->docInfo.pAuthorBlock,
		     (Int32)xDoc->docInfo.wAuthorBlockBytes);
	} else {
	    xDoc->docInfo.pAuthorBlock = NULL;
	}
    }

    /* Read the reference page offset */
    if (xDoc->refPageOffset) {
	XIF_SEEK(xFile, xDoc->refPageOffset);
	XIF_READ(xFile, &xDoc->docInfo.dwReferencePageOffset,
		 sizeof(xDoc->docInfo.dwReferencePageOffset));
	SwapLong(xFile, &xDoc->docInfo.dwReferencePageOffset);

	/* The stored value is actually the ordinal, convert
	 * to offset */
	xDoc->docInfo.dwReferencePageOffset--;
    }

    /* Read the symlib table header */
    if (xDoc->symlibTableHeaderOffset != 0UL) {
	XIF_SEEK(xFile, xDoc->symlibTableHeaderOffset);
	XIF_READ(xFile, &xDoc->symlibTableHeader.symlibCount,
	    sizeof(xDoc->symlibTableHeader.symlibCount));
	SwapShort(xFile, &xDoc->symlibTableHeader.symlibCount);
	XIF_READ(xFile, &xDoc->symlibTableHeader.entriesPerTable,
	    sizeof(xDoc->symlibTableHeader.entriesPerTable));
	SwapShort(xFile, &xDoc->symlibTableHeader.entriesPerTable);
	XIF_READ(xFile, &xDoc->symlibTableHeader.lastOrdinal,
	    sizeof(xDoc->symlibTableHeader.lastOrdinal));
	SwapShort(xFile, &xDoc->symlibTableHeader.lastOrdinal);
	XIF_READ(xFile, &xDoc->symlibTableHeader.firstTableOffset,
	    sizeof(xDoc->symlibTableHeader.firstTableOffset));
	SwapLong(xFile, &xDoc->symlibTableHeader.firstTableOffset);

	/* Read the symbol library header, handled separately */
	XERR_PASS_CLEANUP(wp_XifReadSymlibTable(xDoc, 0UL, NULL));
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read all XIFF fields
 *
 * LongDescription:::
 *  wp_XifReadExtensions() reads all XIFF fields stored in the file. These
 *  fields may be store in one or more list.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifReadExtensions(
    XDoc *xDoc
)
{
    Instance *instData = NULL;
    XFile *xFile;
    PROLOGUE;

    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);

    /* Check for proper calling sequence */

    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_XifFileLock(XIF_DOCFILETOKEN(xDoc)));

    /* Seek to base extension location */
    XIF_SEEK(xFile, XIFF_EXTS_OFFSET);

    /* Recursively read the extension list */
    XERR_PASS_CLEANUP(wp_XifReadExtensionList(xDoc, 0, XIFF_EXTS_OFFSET));

    NORMAL_RETURN;

 ERR_CLEANUP:

    wp_XifFileUnLock(XIF_DOCFILETOKEN(xDoc));
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a single page table
 *
 * LongDescription:::
 *  wp_XifReadPageTable() reads a single page table, and inserts it into a
 *  linked list of page tables.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  tableOffset::		Offset of page table to read
 *
 * Output:::
 *  ppPageTable::		If NULL, the table is read and nothing
 *				returned. Otherwise the pointer will point to
 *				the resulting table.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifReadPageTable(
    XDoc *xDoc,
    UInt32 tableOffset,
    PageTable **ppPageTable
)
{
    Instance *instData = NULL;
    XFile *xFile;
    PageTable *pt, *newpt = NULL;		/* Linked page table */
    PageTableEntry *ep;
    UInt32 mallocbytes;
    UInt32 offset;
    UInt16 entriesLeft;
    PROLOGUE;

    /* Default to return nothing */
    if (ppPageTable) {
	*ppPageTable = NULL;
    }

    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    pt = xDoc->firstPageTable;
    offset = xDoc->pageTableHeader.firstTableOffset;

    /* Check for proper calling sequence */


    /* If initial page table is not in place, we cannot link
     * another */
    XERR_GEN_CLEANUP((tableOffset > 0 && pt == NULL), XIF_SEQUENCE);

    /* If we're installing the first, don't look for cached tables */
    if (pt) {

	/* Since the page tables are always read in order, and cached
	 * when read, go through the cached list until we come to the
	 * last one cached */
	while (pt->ordinal <= tableOffset) {
	    /* There is no such page table */
	    XERR_GEN_CLEANUP(((offset = pt->nextTableOffset) == 0),
		     XIF_BADPARAMETER);
	    if (pt->next == NULL) {
		/* Table exists, but hasn't been read yet */
		break;
	    }
	    pt = pt->next;
	}

	/* If we're at the right page table, and it's already read, bail now */
	if (pt->ordinal == tableOffset + 1) {
	    goto found;
	}
    }

    /* Read as many page tables as necessary to retrieve the one requested */
    do {
	/* Seek to the next page table offset.
	 * If there aren't the number of page tables that the page
	 * table header claims, the file is corrupt */
	XERR_GEN_CLEANUP((! offset), XIF_BADFILEFORMAT);
	XIF_SEEK(xFile, offset);

	CHECK_MALLOC(newpt = (PageTable *)XIF_MALLOC(sizeof(PageTable)));
	memset(newpt, 0, sizeof(PageTable));
	mallocbytes = sizeof(PageTableEntry) * 
	    xDoc->pageTableHeader.entriesPerTable;
	CHECK_MALLOC(newpt->entries =
		     (PageTableEntry *)XIF_MALLOC(mallocbytes));
	memset(newpt->entries, 0, mallocbytes);

	if (pt) {
	    /* Attach to end of linked list */
	    pt->next = newpt;
	    pt = newpt;
	} else {
	    /* Begin list here */
	    pt = xDoc->firstPageTable = newpt;
	}
	newpt = NULL;	/* Avoid freeing if later failure */

	/* Read the table */
	XIF_READ(xFile, &pt->ordinal, sizeof(pt->ordinal));
	SwapShort(xFile, &pt->ordinal);
	ep = pt->entries;
	entriesLeft = xDoc->pageTableHeader.entriesPerTable;
	while (entriesLeft--) {
	    XIF_READ(xFile, &ep->flags, sizeof(ep->flags));
	    XIF_READ(xFile, &ep->IFDoffset, sizeof(ep->IFDoffset));
	    SwapLong(xFile, &ep->IFDoffset);
	    ep++;
	}

	/* Read the next page table offset */
	XIF_READ(xFile, &pt->nextTableOffset, sizeof(pt->nextTableOffset));
	SwapLong(xFile, &pt->nextTableOffset);
	offset = pt->nextTableOffset;

    } while (pt->ordinal != tableOffset + 1);

 found:

    /* Return the one found */
    if (ppPageTable) {
	*ppPageTable = pt;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (newpt) {
	XIF_FREE(newpt);
	if (newpt->entries) {
	    XIF_FREE(newpt->entries);
	}
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the page IFD offset
 *
 * LongDescription:::
 *  wp_XifGetPageIFD() returns the page IFD offset, reading the appropriate
 *  page table, if necessary.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  pageOffset::		Offset of the desired page (0, 1, ...)
 *
 * Output:::
 *  pFlags::			Pointer to the flags stored for the page
 *				(see xxxxxxxxxxxxxxxxxx)
 *  pIfdOffset::		Pointer to the IFD offset
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGetPageIFD(
    XDoc *xDoc,
    UInt32 pageOffset,
    UInt8 *pFlags,
    Int32 *pIfdOffset
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt32 tableOffset;
    UInt32 entryOffset;
    PageTable *pageTable;
    PROLOGUE;

    /* Default to no return value */
    *pFlags = 0;
    *pIfdOffset = 0UL;

    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((xDoc->seqStatus & XDOC_SEQ_GETINFO) == 0,
	XIF_SEQUENCE);

    /* Read the appropriate page table */
    XERR_GEN_CLEANUP(xDoc->pageTableHeader.entriesPerTable == 0,
	XIF_INTERNAL_ERROR);
    tableOffset = pageOffset / xDoc->pageTableHeader.entriesPerTable;
    entryOffset = pageOffset % xDoc->pageTableHeader.entriesPerTable;
    XERR_PASS_CLEANUP(wp_XifReadPageTable(xDoc, tableOffset, &pageTable));

    /* Return the IFD offset for the requested page */
    *pFlags = pageTable->entries[entryOffset].flags;
    *pIfdOffset = pageTable->entries[entryOffset].IFDoffset;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a single symbol libaray table
 *   
 *
 * LongDescription:::
 *  wp_XifReadSymlibTable() reads a single symbol library table, and inserts
 *  it into a linked list of symlib tables.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  tableOffset::		Symlib table offset
 *
 * Output:::
 *  ppSymlibTable::		If NULL, the table is read and nothing
 *				returned. Otherwise the pointer will point to
 *				the resulting table.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifReadSymlibTable(
    XDoc *xDoc,
    UInt32 tableOffset,
    SymlibTable **ppSymlibTable
)
{
    Instance *instData = NULL;
    XFile *xFile;
    SymlibTable *st, *newst = NULL;		/* Linked symlib table */
    SymlibTableEntry *ep;
    UInt32 mallocbytes;
    Int32 offset;
    UInt16 entriesLeft;
    PROLOGUE;

    /* Default to return nothing */
    if (ppSymlibTable) {
	*ppSymlibTable = NULL;
    }

    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    st = xDoc->firstSymlibTable;
    offset = xDoc->symlibTableHeader.firstTableOffset;

    /* Check for proper calling sequence */


    /* If initial symlib table is not in place, we cannot link
     * another */
    XERR_GEN_CLEANUP((tableOffset > 0 && st == NULL), XIF_SEQUENCE);

    /* If we're installing the first, don't look for cached tables */
    if (st) {

	/* Since the symlib tables are always read in order, and cached
	 * when read, go through the cached list until we come to the
	 * last one cached */
	while (st->ordinal <= tableOffset) {
	    /* There is no such symlib table */
	    XERR_GEN_CLEANUP(((offset = st->nextTableOffset) == 0),
		     XIF_BADFILEFORMAT);
	    if (st->next == NULL) {
		/* Table exists, but hasn't been read yet */
		break;
	    }
	    st = st->next;
	}

	/* If we're at the right symlib table, and it's already read, bail now */
	if (st->ordinal == tableOffset + 1) {
	    goto found;
	}
    }

    /* Read as many symlib tables as necessary to retrieve the one requested */
    do {
	/* Seek to the next symlib table offset.
	 * If there aren't the number of symlib tables that the symlib
	 * table header claims, the file is corrupt */
	XERR_GEN_CLEANUP((! offset), XIF_BADFILEFORMAT);
	XIF_SEEK(xFile, offset);

	CHECK_MALLOC(newst = (SymlibTable *)XIF_MALLOC(sizeof(SymlibTable)));
	memset(newst, 0, sizeof(SymlibTable));
	mallocbytes = sizeof(SymlibTableEntry) * 
	    xDoc->symlibTableHeader.entriesPerTable;
	CHECK_MALLOC(newst->entries =
		     (SymlibTableEntry *)XIF_MALLOC(mallocbytes));
	memset(newst->entries, 0, mallocbytes);

	if (st) {
	    /* Attach to end of linked list */
	    st->next = newst;
	    st = newst;
	} else {
	    /* Begin list here */
	    st = xDoc->firstSymlibTable = newst;
	}
	newst = NULL;	/* Avoid freeing if later failure */

	/* Read the table */
	XIF_READ(xFile, &st->ordinal, sizeof(st->ordinal));
	SwapShort(xFile, &st->ordinal);
	ep = st->entries;
	entriesLeft = xDoc->symlibTableHeader.entriesPerTable;
	while (entriesLeft--) {
	    XIF_READ(xFile, &ep->bytes, sizeof(ep->bytes));
	    SwapLong(xFile, &ep->bytes);
	    XIF_READ(xFile, &ep->offset, sizeof(ep->offset));
	    SwapLong(xFile, &ep->offset);
	    ep++;
	}

	/* Read the next symlib table offset */
	XIF_READ(xFile, &st->nextTableOffset, sizeof(st->nextTableOffset));
	SwapLong(xFile, &st->nextTableOffset);
	offset = st->nextTableOffset;

    } while (st->ordinal != tableOffset + 1);

 found:

    /* Return the one found */
    if (ppSymlibTable) {
	*ppSymlibTable = st;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (newst) {
	XIF_FREE(newst);
	if (newst->entries) {
	    XIF_FREE(newst->entries);
	}
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the size and offset of a token dictionary
 *
 * LongDescription:::
 *  wp_XifGetSymlibOffset() returns the size and file offset of a token
 *  dictionary found within the symbol library.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  dictID::			Unique ID of the token dictionary
 *
 * Output:::
 *  pOffset::			Pointer to the data offset
 *  pBytes::			Pointer to the number of token dictionary bytes
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGetSymlibOffset(
    XDoc *xDoc,
    UInt32 dictID,
    Int32 *pOffset,
    Int32 *pBytes
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt32 tableOffset;
    UInt32 entryOffset;
    Int32 symlibOffset;
    SymlibTable *symlibTable;
    PROLOGUE;

    /* Default to no return value */
    *pOffset = 0UL;
    *pBytes = 0UL;

    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);

    /* Check for proper calling sequence */


    /* Read the appropriate symlib table */
    symlibOffset = dictID;
    XERR_GEN_CLEANUP(symlibOffset < 0, XIF_BADPARAMETER);
    if (xDoc->symlibTableHeader.entriesPerTable == 0) {
	XERR_PASS_CLEANUP(wp_XifReadSymlibTable(xDoc, 0L, NULL));
    }

    tableOffset = symlibOffset / xDoc->symlibTableHeader.entriesPerTable;
    entryOffset = symlibOffset % xDoc->symlibTableHeader.entriesPerTable;
    XERR_PASS_CLEANUP(wp_XifReadSymlibTable(xDoc, tableOffset, &symlibTable));

    /* Return the offset and size for the requested symlib */
    *pOffset = symlibTable->entries[entryOffset].offset;
    *pBytes = symlibTable->entries[entryOffset].bytes;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read one strip of image data, leave compressed
 *
 * LongDescription:::
 *  wp_XifCompressedReadStrip() reads a single strip of image data from a TIFF
 *  (XIFF), returning the compressed data.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  dwImgRowOffset::		Row offset in image proper, to read
 *
 * Output:::
 *  ppCmpBuffer::		Pointer to pointer to compressed buffer
 *				to store data (may be updated to accomodate
 *				the compressed strip)
 *  pdwCmpBufsize::		Pointer to size of compressed buffer (may be
 *				updated)
 *  pdwBytesFilled::		Pointer to the number of bytes filled in
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifCompressedReadStrip(
    XImg *xImg,
    UInt32 dwImgRowOffset,
    UInt8 **ppCmpBuffer,
    UInt32 *pdwCmpBufsize,
    UInt32 *pdwBytesFilled
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt32 stripOffset;
    UInt32 stripBytes;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesFilled = 0UL;

    CheckHandleValidity(xImg);
    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* Check the strip offset */
    XERR_GEN_CLEANUP((! xImg->imgInfo.dwBestStripHeight), XIF_BADFILEFORMAT);

    /* Get the offset and size of the compressed data */
    XERR_PASS_CLEANUP(wp_XifGetCompressedSize(xImg, dwImgRowOffset,
	&stripOffset, &stripBytes));

    /* Make sure we have allocated the right number of bytes */
    if (*pdwCmpBufsize < (UInt32)stripBytes || *ppCmpBuffer == NULL) {
	if (*pdwCmpBufsize > 0 && *ppCmpBuffer != NULL) {
	    XIF_FREE(*ppCmpBuffer);
	}
	CHECK_MALLOC(*ppCmpBuffer = XIF_MALLOC(stripBytes));
	*pdwCmpBufsize = stripBytes;
    }

    /* Read the "compressed" data */
    XIF_SEEK(xFile, stripOffset);
    XIF_READ(xFile, (char *)*ppCmpBuffer, stripBytes);

    /* Return the number of bytes stored */
    *pdwBytesFilled = stripBytes;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the compressed strip data offset and byte count
 *
 * LongDescription:::
 *  wp_XifGetCompressedSize() returns the appropriate file offset and byte
 *  count for the strip data, as previously retrieve from the TIFF fields.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  dwImgRowOffset::		Row offset in image proper, to read
 *
 * Output:::
 *  pdwOffset::			Pointer to the file offset
 *  pdwBytes::			Pointer to the byte count
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGetCompressedSize(
    XImg *xImg,
    UInt32 dwImgRowOffset,
    UInt32 *pdwOffset,
    UInt32 *pdwBytes
)
{
    UInt32 stripNum;
    Int32 stripOffset;
    UInt32 stripBytes;
    UInt32 stripRows;
    PROLOGUE;

    /* Default to return nothing */
    *pdwOffset = 0L;
    *pdwBytes = 0UL;

    CheckHandleValidity(xImg);

    /* Check the strip offset */
    XERR_GEN_CLEANUP((! xImg->imgInfo.dwBestStripHeight), XIF_BADFILEFORMAT);

    stripRows = xImg->imgInfo.dwBestStripHeight;
    stripNum = dwImgRowOffset / stripRows;
    while (stripNum > xImg->fields.StripsPerImage) {
	if (xImg->fields.StripsPerImage == 0) {
	    if (xImg->imgInfo.dwHeight == 0 || stripRows == 0) {
		NORMAL_RETURN;
	    }
	    xImg->fields.StripsPerImage = (xImg->imgInfo.dwHeight + stripRows - 1)
		/ stripRows;
	    XERR_GEN_CLEANUP((xImg->fields.StripsPerImage == 0),
			      XIF_BADFILEFORMAT);
	} else {
	    __error__ = XIF_BADPARAMETER;
	    goto ERR_CLEANUP;
	}
    }

    /* Get the offset and size of the compressed data */
    if (xImg->fields.StripOffsets == NULL || xImg->fields.StripByteCounts == NULL) {
	XERR_GEN_CLEANUP((xImg->fields.TileOffsets == NULL ||
	    xImg->fields.TileByteCounts == NULL), XIF_BADFILEFORMAT);

	stripOffset = xImg->fields.TileOffsets[stripNum];
	stripBytes = xImg->fields.TileByteCounts[stripNum];
    } else {
	stripOffset = xImg->fields.StripOffsets[stripNum];
	stripBytes = xImg->fields.StripByteCounts[stripNum];
    }

    /* Return the desired results */
    *pdwOffset = stripOffset;
    *pdwBytes = stripBytes;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read and decompress one strip of image data
 *
 * LongDescription:::
 *  wp_XifReadStrip() reads and decompresses a single strip of image data
 *  from a TIFF (XIFF).
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pPixr::			Pointer to image PIXR
 *  dwPixrRowOffset::		Row offset in PIXR to store image strip
 *  dwImgRowOffset::		Row offset in image proper, to read
 *
 * Output:::
 *  pdwRowsToRead::		Rows of image to read, in outer loop
 *  pdwRowsFilled::		Pointer to the number of rows filled in
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifReadStrip(
    XImg *xImg,
    PIXR *pPixr,
    UInt32 dwPixrRowOffset,
    UInt32 dwImgRowOffset,
    UInt32 *pdwRowsToRead,
    UInt32 *pdwRowsFilled
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt32 stripOffset;
    UInt32 stripBytes;
    UInt32 rowsFilled;
    IPWState *pIPWState;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    CheckHandleValidity(xImg);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* Check the strip size & offset:
     *  1. There must be a TIFF "strip" size set.
     *  2. The requested strip size must align with the TIFF strip size.
     */
    XERR_GEN_CLEANUP(xImg->imgInfo.dwBestStripHeight == 0, XIF_BADFILEFORMAT);
    XERR_GEN_CLEANUP(dwImgRowOffset % xImg->imgInfo.dwBestStripHeight != 0,
	XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(*pdwRowsToRead != xImg->imgInfo.dwBestStripHeight &&
	dwImgRowOffset + *pdwRowsToRead != xImg->imgInfo.dwHeight,
	XIF_BADPARAMETER);

    /* Get the strip location and size */
    XERR_PASS_CLEANUP(wp_XifGetCompressedSize(xImg, dwImgRowOffset,
	&stripOffset, &stripBytes));

    /* If there's no data, then there's no need to decompress anything */
    if (stripBytes == 0) {
	rowsFilled = *pdwRowsToRead;
	*pdwRowsToRead -= rowsFilled;
        *pdwRowsFilled = rowsFilled;

	/* It's assumed that the image data was previously cleared; otherwise it
	 * would be appropriate to clear it here */

	NORMAL_RETURN;
    }

    /* Allocate a buffer for the compressed data */
    if (xImg->cmpBytes < (UInt32)stripBytes || xImg->cmpBuffer == NULL) {
	if (xImg->cmpBytes > 0 || xImg->cmpBuffer != NULL) {
	    XIF_FREE(xImg->cmpBuffer);
	}
	XERR_GEN_CLEANUP(stripBytes == 0, XIF_BADFILEFORMAT);
	CHECK_MALLOC(xImg->cmpBuffer = XIF_MALLOC(stripBytes));
	xImg->cmpBytes = stripBytes;
    }

    /* Read the "compressed" data */
    XIF_SEEK(xFile, stripOffset);
    XIF_READ(xFile, (char *)xImg->cmpBuffer, stripBytes);

    /* Decompress into output image */
    XERR_PASS_CLEANUP(w_XifDecompress(pIPWState, (UInt32)xImg,
	xImg->imgInfo.xCompression, xImg->imgInfo.xColorType,
	xImg->imgFlags, xImg->cmpBuffer, stripBytes,
	dwPixrRowOffset, xImg->imgInfo.dwWidth, *pdwRowsToRead,
	pPixr, &rowsFilled));
    XERR_GEN_CLEANUP((rowsFilled < 1), XIF_BADFILEFORMAT);
    XERR_GEN_CLEANUP(rowsFilled > *pdwRowsToRead, XIF_INTERNAL_ERROR);
    *pdwRowsToRead -= rowsFilled;
    *pdwRowsFilled = rowsFilled;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read through the SubIFDs list(s) and store the IFD information
 *
 * LongDescription:::
 *  wp_XifTraverseSubImages() reads through each of the SubIFDs chains and
 *  stores the information contained within the IFDs.
 *
 * Input:::
 *  parentImg::			Image handle of parent image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifTraverseSubImages(
    XImg *parentImg
)
{
    XPage *xPage = parentImg->xPage;
    XImg *childImg = NULL;	/* Image found on subimage path (unlinked) */
    XImg *curImg;		/* Currently linked image */
    XImg *siblingImg;		/* Sibling already linked */
    XImg *nextImg;		/* Next sibling in links */
    UInt32 nextIFD;
    UInt32 maskOffset = 0;
    Int32 imgOffset = 0L;
    UInt32 layer = 0UL;
    XImg **listPtr = NULL;
    UInt32 subPath;
    XIF_IMGTYPE xImgType;
    PROLOGUE;

    CheckHandleValidity(parentImg);

    /* Search through the images starting with the first subimage. There may
     * exist more than one starting node. */
    for (subPath = 0; subPath < parentImg->fields.SubIFDCount; subPath++) {
	nextIFD = parentImg->fields.SubIFDs[subPath];

	/* Loop while there's one attached */
	while (nextIFD) {

	    /* Add the next in the sequence */
	    XERR_PASS_CLEANUP(wp_NewImg(xPage, &childImg, nextIFD));

	    /* Get the scoop on this image (esp. ImageLayer) */
	    XERR_PASS_CLEANUP(wp_TifReadFields(childImg));

	    /* Check to see if it was a mask image */
	    xImgType = XIF_IMGTYPE_UNKNOWN;
	    if (xPage->xDoc->docInfo.version > 2) {
		if (childImg->fields.NewSubFileType & XTIFF_SUBFILE_MASK) {
		    xImgType = XIF_IMGTYPE_MASK;
		    layer = 0;	    /* Not part of the MRC */
		    imgOffset = maskOffset++;
		    listPtr = &childImg->xMaskImg;
		}
	    }

	    if (xImgType == XIF_IMGTYPE_UNKNOWN) {

		/* Determine which type of image it is */
		switch ((int)childImg->fields.ImageLayer[0]) {
		case 0:
		    if (xPage->xDoc->docInfo.version == 2) {
			XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage,
			    layer = 5, 1));
			xImgType = XIF_IMGTYPE_LAYER;
			imgOffset = (Int32)xPage->pageInfo.pdwImageCounts[5-1]
			    - 1;
			listPtr = &xPage->xImgs;
		    } else if ((xPage->xDoc->docInfo.version > 2) &&
			(childImg->fields.PhotometricInterpretation == 4) &&
			((childImg->fields.NewSubFileType & XTIFF_SUBFILE_MASK) != 0)) {
			xImgType = XIF_IMGTYPE_MASK;
			listPtr = &parentImg->xMaskImg;
			imgOffset = maskOffset++;
		    } else {
			__error__ = XIF_BADFILEFORMAT;
			goto ERR_CLEANUP;
		    }
		    break;
		default:
		    XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage,
			layer = childImg->fields.ImageLayer[0], 1));
		    xImgType = XIF_IMGTYPE_LAYER;
		    imgOffset = (Int32)childImg->fields.ImageLayer[1] - 1;
		    if (imgOffset < 0) {
			imgOffset = xPage->pageInfo.pdwImageCounts[layer - 1];
		    }
		    listPtr = &xPage->xImgs;
		    break;
		}
	    }

	    /* Retain it's identity */
	    childImg->imgInfo.xImgType = xImgType;
	    childImg->imgInfo.dwImgLayer = layer;
	    childImg->imgInfo.dwImgOffset = (UInt32)imgOffset;

	    /* Reference the appropriate linked list */
	    siblingImg = *listPtr;

	    /* If this is the first, just insert it */
	    if (siblingImg == NULL) {
		*listPtr = curImg = childImg;
		childImg = NULL;
	    } else if (siblingImg->imgInfo.dwImgLayer > layer ||
		(siblingImg->imgInfo.dwImgLayer == layer &&
		siblingImg->imgInfo.dwImgOffset > (UInt32)imgOffset)) {

		/* This one should really be first. Insert as such. */
		childImg->next = siblingImg;
		siblingImg->prev = childImg;
		*listPtr = curImg = childImg;
		childImg = NULL;
	    } else {

		/* Search for the proper point to insert *after* */
		while (((nextImg = siblingImg->next) != NULL) &&
		    (nextImg->imgInfo.dwImgLayer < layer ||
		    (nextImg->imgInfo.dwImgLayer == layer &&
		    nextImg->imgInfo.dwImgOffset < (UInt32)imgOffset))) {
		    siblingImg = nextImg;
		}

		/* Link this image to the one that was next */
		childImg->next = nextImg;
		if (nextImg) {
		    nextImg->prev = childImg;
		}

		/* Link prev image to this one */
		siblingImg->next = curImg = childImg;
		childImg->prev = siblingImg;
		childImg = NULL;
	    }

	    /* Keep the loop going if there's another IFD attached */
	    nextIFD = curImg->fields.nextIFD;

	    if (xPage->xDoc->docInfo.version == 2) {
		/* Well, it can't be a mask, since they're attached via
		 * the MaskIFDs field. Since it's not a mask, look for one. */
		XERR_PASS_CLEANUP(wp_XifTraverseMaskImages(curImg));
	    }
	} /* For each subimage in sub path */
    } /* For each sub path */

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (childImg) {
	Instance *instData = XIF_IMGINST(parentImg);
	XIF_FREE(childImg);
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the MaskIFDs chain and store the IFD information
 *
 * LongDescription:::
 *  wp_XifTraverseMaskImages() reads through the MaskIFDs chain and
 *  stores the information contained within the IFDs as transparency mask
 *  information.
 *
 * Input:::
 *  parentImg::			Image handle of parent image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifTraverseMaskImages(
    XImg *parentImg
)
{
    XPage *xPage = parentImg->xPage;
    XDoc *xDoc;
    XImg *childImg = NULL;
    UInt32 nextIFD = 0UL;
    UInt32 overlayOffset;
    UInt32 imgOffset = 0UL;
    XImg **listPtr;
    PROLOGUE;

    CheckHandleValidity(parentImg);

    /* Search through the images starting with the first subimage.
     */
    xDoc = xPage->xDoc;
    if (xDoc->docInfo.version > 2) {
	if (parentImg->fields.SubIFDCount > 0) {
	    nextIFD = parentImg->fields.SubIFDs[0];
	}
    } else {
	nextIFD = parentImg->fields.MaskIFDs;
    }

    overlayOffset = 0;

    /* Loop while there's one attached */
    while (nextIFD) {

	/* Add the next in the sequence */
	XERR_PASS_CLEANUP(wp_NewImg(xPage, &childImg, nextIFD));

	/* Get the scoop on this image */
	XERR_PASS_CLEANUP(wp_TifReadFields(childImg));

	listPtr = &parentImg->xMaskImg;

	/* Retain it's identity */
	childImg->imgInfo.xImgType = XIF_IMGTYPE_MASK;
	childImg->imgInfo.dwImgOffset = imgOffset++;

	/* Mark parent as having a mask */
	parentImg->seqStatus |= XIMG_SEQ_MASKADDED;

	/* Place it in the list of the appropriate type */
	while (*listPtr && (*listPtr)->imgInfo.dwImgOffset <= imgOffset) {
	    listPtr = &((*listPtr)->next);
	}
	if (*listPtr) {
	    childImg->next = *listPtr;  /* Relink the next one here */
	}
	*listPtr = childImg;	    /* Insert this one */
	nextIFD = childImg->fields.nextIFD;
	childImg = NULL;
    } /* For each subimage in sub path */

    /* Mark mask count in parent */
    parentImg->imgInfo.dwMaskImages = imgOffset;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (childImg) {
	Instance *instData = XIF_IMGINST(parentImg);
	XIF_FREE(childImg);
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read or retrieve the annotation table header
 *
 * LongDescription:::
 *  wp_XifGetAnnotTableHeader() reads the annotation table header. If it has
 *  previously been gotten, nothing is done. The annotation table header is
 *  stored in the xPage structure.
 *
 * Input:::
 *  xPage::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGetAnnotTableHeader(
    XPage *xPage
)
{
    XFile *xFile;
    XImg *primImg = NULL;
    Int32 fileOffset;
    UInt16 sval; UInt32 lval;
    PROLOGUE;

    CheckHandleValidity(xPage);

    xFile = XIF_PAGEFILE(xPage);


    /* Refer to primary image for annotation field */
    primImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

    if (xPage->annotTableHeader.entriesPerTable > 0) {
	NORMAL_RETURN;
    }

    if (xPage->xDoc->docInfo.version == 2) {

    /* Since no real "table" exists, call each entry in the array
	* of offsets a "table entry" */
	xPage->annotTableHeader.entriesPerTable = 1;

	/* Make sure we don't "think" there are annotations when there aren't */
	if (primImg->fields.AnnotationOffsets == NULL) {
	    xPage->annotTableHeader.annotCount = 0;
	} else {
	    xPage->annotTableHeader.annotCount = (UInt16)primImg->fields.AnnotationCount;
	}
    } else if (xPage->xDoc->docInfo.version > 2) {

	/* Check to see if any exist */
	if (primImg->fields.AnnotationOffsets == NULL) {
	    xPage->annotTableHeader.annotCount = 0;
	} else {

	    /* The count should always be 1 in a 3.x file */
	    XERR_GEN_CLEANUP(primImg->fields.AnnotationCount != 1, XIF_BADFILEFORMAT);

	    /* Read the header */
	    xPage->annotTableHeaderOffset = fileOffset =
		primImg->fields.AnnotationOffsets[0];

	    /* If the offset is zero, there is no annotation table header */
	    if (fileOffset == 0) {
		xPage->annotTableHeader.annotCount = 0;
		xPage->annotTableHeader.entriesPerTable = 20;
		xPage->annotTableHeader.lastOrdinal = 0;
		xPage->annotTableHeader.firstTableOffset = 0;
		NORMAL_RETURN;
	    }
	    XERR_GEN_CLEANUP(fileOffset == 0, XIF_BADFILEFORMAT);
	    XIF_SEEK(xFile, fileOffset);

	    XIF_READ(xFile, &sval, sizeof(sval));
	    SwapShort(xFile, &sval);
	    xPage->annotTableHeader.annotCount = sval;
	    XIF_READ(xFile, &sval, sizeof(sval));
	    SwapShort(xFile, &sval);
	    xPage->annotTableHeader.entriesPerTable = sval;
	    XIF_READ(xFile, &sval, sizeof(sval));
	    SwapShort(xFile, &sval);
	    xPage->annotTableHeader.bytesPerEntry = sval;
#if 1 	/* To be taken out when those reading non-updated XIFF 3.0 files
	 * have updated their files */
	    {
		AnnotTableEntry *ep = NULL;

		if (xPage->annotTableHeader.bytesPerEntry !=
		    sizeof(ep->id) + sizeof(ep->type) +
		    sizeof(ep->xOffset) + sizeof(ep->yOffset) +
		    sizeof(ep->width) + sizeof(ep->height) +
		    sizeof(ep->flags) +
		    sizeof(ep->bytes) + sizeof(ep->offset)) {
		    xPage->annotTableHeader.lastOrdinal =
			xPage->annotTableHeader.bytesPerEntry;

		    /* Must be missing the new flags element */
		    xPage->annotTableHeader.bytesPerEntry = 28;
		} else {
		    XIF_READ(xFile, &sval, sizeof(sval));
		    SwapShort(xFile, &sval);
		    xPage->annotTableHeader.lastOrdinal = sval;
		}
	    }
#else
	    XIF_READ(xFile, &sval, sizeof(sval));
	    SwapShort(xFile, &sval);
	    xPage->annotTableHeader.lastOrdinal = sval;
#endif
	    XIF_READ(xFile, &lval, sizeof(lval));
	    SwapLong(xFile, &lval);
	    xPage->annotTableHeader.firstTableOffset = lval;
	} /* AnnotationOffset tag exists */
    } /* version 2 else > 2 */

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a single annotation table
 *   
 *
 * LongDescription:::
 *  wp_XifGetAnnotTable() reads a single annotation table, and inserts it into
 *  a linked list of annotation tables. If the data has previously been read
 *  or added, the data is found and returned.*
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *  annotOffset::		Annotation offset seeking for
 *
 * Output:::
 *  ppAnnotTable::		If NULL, the table is read and nothing
 *				returned. Otherwise the pointer will point to
 *				the resulting table.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGetAnnotTable(
    XPage *xPage,
    UInt16 annotOffset,
    AnnotTable **ppAnnotTable
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XImg *primImg = NULL;
    AnnotTable *at, *newAt = NULL;		/* Linked annotation table */
    AnnotTableEntry *ep;
    UInt32 mallocbytes;
    Int32 fileOffset;
    UInt16 entriesLeft;
    UInt16 ordinal = 0;		    /* Table ordinal */
    UInt32 tableOffset;		    /* Offset of all tables that annotation is
				     * found */
    UInt32 entryOffset;		    /* Entry within table of annotation */
    UInt32 curAnnotOffset;	    /* Annotation offset within loop */
    UInt8 *annotEntryData = NULL;   /* Annotation table intermediate buffer */
    UInt8 *entryStart, *entry;	    /* annotEntryData pointers */
    UInt32 bytesLeft;	 	    /* Bytes left in entry data */
    UInt32 twoHeaderSize = 0;	    /* If v 2.0, the bytes in the header
				     * portion */
    PROLOGUE;

    /* Default to return nothing */
    if (ppAnnotTable) {
	*ppAnnotTable = NULL;
    }

    CheckHandleValidity(xPage);
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);
    at = xPage->firstAnnotTable;

    /* Check for proper calling sequence */


    /* Make sure we've read the annotation table header (if any) */
    if (xPage->annotTableHeader.entriesPerTable == 0UL) {
	XERR_GEN_CLEANUP(xPage->annotTableHeader.annotCount > 0,
			 XIF_INTERNAL_ERROR);

	XERR_PASS_CLEANUP(wp_XifGetAnnotTableHeader(xPage));
    }

    /* Check parameters */
    XERR_GEN_CLEANUP(annotOffset >= xPage->annotTableHeader.annotCount,
		     XIF_BADPARAMETER);

    /* Refer to primary image for annotation field */
    primImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

    if (xPage->xDoc->readWrite & READ_ALLOWED) {

        /* For XIFF 2.0, the annotation information is at the same place as
	 * the data itself. Get the file offset of the header.
	 * For XIFF 3.0, the table offset is contained in the table header. */
	if (xPage->xDoc->docInfo.version == 2 &&
	    xPage->annotTableHeader.firstTableOffset == 0) {
	    XERR_GEN_CLEANUP(primImg->fields.AnnotationOffsets == NULL,
		XIF_INTERNAL_ERROR);
	    fileOffset = primImg->fields.AnnotationOffsets[annotOffset];
	} else {
	    fileOffset = xPage->annotTableHeader.firstTableOffset;
	}
    } else {
	fileOffset = 0;
    }

    /* Compute the table and entry */
    tableOffset = annotOffset / xPage->annotTableHeader.entriesPerTable;
    entryOffset = annotOffset % xPage->annotTableHeader.entriesPerTable;

    /* If we're installing the first, don't look for cached tables */
    if (at) {

	/* Since the annotation tables are always read in order, and cached
	 * when read, go through the cached list until we come to the
	 * last one cached */
	while (at->ordinal <= tableOffset) {

	    /* Should be in sequence */
	    XERR_GEN_CLEANUP(++ordinal != at->ordinal, XIF_INTERNAL_ERROR);

	    /* There is no such annotation table */
	    if (xPage->xDoc->docInfo.version >= 3 &&
		(xPage->xDoc->readWrite & READ_ALLOWED)) {
		XERR_GEN_CLEANUP(((fileOffset = at->nextTableOffset) == 0),
		    XIF_BADPARAMETER);
	    }
	    if (at->next == NULL) {

		/* Table may exist, but hasn't been read yet */
		break;
	    }
	    at = at->next;
	}

	/* If we're at the right annotation table, and it's already read,
	 * bail now */
	if (at->ordinal == tableOffset + 1) {
	    goto found;
	}
    }

    /* We shouldn't get here if we're not reading */
    XERR_GEN_CLEANUP((xPage->xDoc->readWrite & READ_ALLOWED) == 0,
	XIF_INTERNAL_ERROR);

    /* Prepare to read the entire table (3.0) or entry (2.0) into a single
     * buffer in a single read */
    if (xPage->xDoc->docInfo.version == 2) {
	mallocbytes = twoHeaderSize = sizeof(ep->bytes) + sizeof(ep->type) +
	    sizeof(ep->xOffset) + sizeof(ep->yOffset) +
	    sizeof(ep->width) + sizeof(ep->height);
    } else {
	mallocbytes = sizeof(short) + xPage->annotTableHeader.entriesPerTable *
	    xPage->annotTableHeader.bytesPerEntry;
    }
    CHECK_MALLOC(annotEntryData = (UInt8 *)XIF_MALLOC(mallocbytes));

    /* Read as many annotation tables as necessary to retrieve the one
     * requested */
    do {
	/* Seek to the next annotation table offset.
	 * If there aren't the number of annotation tables that the annotation
	 * table header claims, the file is corrupt */
	XERR_GEN_CLEANUP(fileOffset < 8, XIF_BADFILEFORMAT);
	XIF_SEEK(xFile, fileOffset);

	CHECK_MALLOC(newAt = (AnnotTable *)XIF_MALLOC(sizeof(AnnotTable)));
	memset(newAt, 0, sizeof(AnnotTable));
	mallocbytes = sizeof(AnnotTableEntry) * 
	    xPage->annotTableHeader.entriesPerTable;
	CHECK_MALLOC(newAt->entries =
		     (AnnotTableEntry *)XIF_MALLOC(mallocbytes));
	memset(newAt->entries, 0, mallocbytes);

	if (at) {
	    /* Attach to end of linked list */
	    at->next = newAt;
	    at = newAt;
	} else {
	    /* Begin list here */
	    at = xPage->firstAnnotTable = newAt;
	}
	newAt = NULL;	/* Avoid freeing if later failure */

	/* Read the table */
	ep = at->entries;
	if (xPage->xDoc->docInfo.version == 2) {
	    at->ordinal = annotOffset + 1;
	    ep->id = annotOffset;
	    ep->offset = primImg->fields.AnnotationOffsets[annotOffset] +
		twoHeaderSize;

	    /* Read the entire header portion of the data (expect that the
	     * byte count should at least include the header portion). */
	    XIF_READ(xFile, annotEntryData, (Int32)twoHeaderSize);
	    entry = annotEntryData;

	    memcpy(&ep->bytes, entry, sizeof(long));
	    SwapLong(xFile, &ep->bytes);
	    entry += sizeof(long);

	    XERR_GEN_CLEANUP(ep->bytes < sizeof(ep->type) + sizeof(ep->xOffset)
		+ sizeof(ep->yOffset) + sizeof(ep->width) + sizeof(ep->height),
		XIF_BADFILEFORMAT);
	    ep->bytes -= sizeof(ep->type) + sizeof(ep->xOffset) +
		sizeof(ep->yOffset) + sizeof(ep->width) + sizeof(ep->height);

	    memcpy(&ep->type, entry, sizeof(short));
	    SwapShort(xFile, &ep->type);
	    entry += sizeof(short);
	    memcpy(&ep->xOffset, entry, sizeof(long));
	    SwapLong(xFile, &ep->xOffset);
	    entry += sizeof(long);
	    memcpy(&ep->yOffset, entry, sizeof(long));
	    SwapLong(xFile, &ep->yOffset);
	    entry += sizeof(long);
	    memcpy(&ep->width, entry, sizeof(long));
	    SwapLong(xFile, &ep->height);
	    entry += sizeof(long);
	    memcpy(&ep->height, entry, sizeof(long));
	    SwapLong(xFile, &ep->height);
	    entry += sizeof(long);
	    if (annotOffset + 1 < primImg->fields.AnnotationCount) {
		at->nextTableOffset =
		  primImg->fields.AnnotationOffsets[annotOffset+1];
	    }
	} else if (xPage->xDoc->docInfo.version > 2) {

	    /* Read the ID and whole list of table entries */
	    XIF_READ(xFile, annotEntryData, (Int32)(sizeof(short) +
		     xPage->annotTableHeader.entriesPerTable *
		     xPage->annotTableHeader.bytesPerEntry));

	    memcpy(&at->ordinal, annotEntryData, sizeof(short));
	    SwapShort(xFile, &at->ordinal);
	    entryStart = annotEntryData + sizeof(short);

	    entriesLeft = xPage->annotTableHeader.entriesPerTable;
	    curAnnotOffset = tableOffset *
	      xPage->annotTableHeader.entriesPerTable;

	    /* Now parse them out: pick out the number of bytes that
	     * the annotation table header says are in this version,
	     * and stuff them as they fit into the structure known
	     * at the time this implementation was made.  In future
	     * versions of XIFF, there will be more annotation data
	     * than we know what to do with, so we'll just take what
	     * we know about. And in older versions of XIFF, we will
	     * have less data than our structure, so we won't fill in
	     * the whole structure. */
	    while (entriesLeft--) {
		entry = entryStart;
		bytesLeft = xPage->annotTableHeader.bytesPerEntry;
		if (bytesLeft >= sizeof(short)) {
		    memcpy(&ep->id, entry, sizeof(short));
		    SwapShort(xFile, &ep->id);
		    entry += sizeof(short), bytesLeft -= sizeof(short);
		    XERR_GEN_CLEANUP(ep->id != curAnnotOffset &&
			curAnnotOffset < xPage->annotTableHeader.annotCount,
			XIF_BADFILEFORMAT);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(short)) {
		    memcpy(&ep->type, entry, sizeof(short));
		    SwapShort(xFile, &ep->type);
		    entry += sizeof(short), bytesLeft -= sizeof(short);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->xOffset, entry, sizeof(long));
		    SwapLong(xFile, &ep->xOffset);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->yOffset, entry, sizeof(long));
		    SwapLong(xFile, &ep->yOffset);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->width, entry, sizeof(long));
		    SwapLong(xFile, &ep->width);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->height, entry, sizeof(long));
		    SwapLong(xFile, &ep->height);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->bytes, entry, sizeof(long));
		    SwapLong(xFile, &ep->bytes);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->offset, entry, sizeof(long));
		    SwapLong(xFile, &ep->offset);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->flags, entry, sizeof(long));
		    SwapLong(xFile, &ep->flags);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}

		/* Always add new entry fields to the end of the list
		 * for backwards compatibility */

	    nextEntry:
		entryStart += xPage->annotTableHeader.bytesPerEntry;
		ep++;
		curAnnotOffset++;
	    } /* While entries left */

	    /* Read the next annotation table offset */
	    XIF_READ(xFile, &at->nextTableOffset, sizeof(at->nextTableOffset));
	    SwapLong(xFile, &at->nextTableOffset);
	    fileOffset = at->nextTableOffset;
	} /* version > 2 */

	/* Expect to find the tables in sequence */
	ordinal++;

    } while (at->ordinal != tableOffset + 1);

    /* Go to the entry offset */
    ep = &at->entries[entryOffset];

    /* Expect not to find an annotation with an offset of zero */
    XERR_GEN_CLEANUP(ep->offset == 0, XIF_BADFILEFORMAT);

    /* Read the header portion of the pertinent entry if version 2 and it
     * hasn't already been read */
    if (xPage->xDoc->docInfo.version == 2 && ep->bytes == 0) {

	/* Seek to the data (already there) */

	/* Read the header */
	XIF_READ(xFile, &ep->bytes, sizeof(ep->bytes));
	SwapLong(xFile, &ep->bytes);

	/* Remove the header from the byte count and offset */
	ep->bytes -= sizeof(UInt32) * 5 + sizeof(UInt16);
	ep->offset += sizeof(UInt32) * 5 + sizeof(UInt16);

	/* Continue reading */
	XIF_READ(xFile, &ep->type, sizeof(ep->type));
	SwapShort(xFile, &ep->type);
	XIF_READ(xFile, &ep->xOffset, sizeof(ep->xOffset));
	SwapLong(xFile, &ep->xOffset);
	XIF_READ(xFile, &ep->yOffset, sizeof(ep->yOffset));
	SwapLong(xFile, &ep->yOffset);
	XIF_READ(xFile, &ep->width, sizeof(ep->width));
	SwapLong(xFile, &ep->width);
	XIF_READ(xFile, &ep->height, sizeof(ep->height));
	SwapLong(xFile, &ep->height);
    } /* version */

 found:

    if (annotEntryData) {
	XIF_FREE(annotEntryData);
	annotEntryData = NULL;
    }

    /* Return the one found */
    if (ppAnnotTable) {
	*ppAnnotTable = at;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (annotEntryData) {
	XIF_FREE(annotEntryData);
	annotEntryData = NULL;
    }

    if (newAt) {
	XIF_FREE(newAt);
	if (newAt->entries) {
	    XIF_FREE(newAt->entries);
	}
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the data associated with an annotation
 *
 * LongDescription:::
 *  wp_XifGetAnnotData() reads the data associated with an annotation. The data
 *  is disjoint from the annotation table, which contains a reference to the
 *  data.
 *
 * Input:::
 *  xPage::			Data structure of valid document
 *  annotOffset::		Annotation offset
 *  bytes::			Number of data bytes (excludes the XIFF 2.0
 *				annotation description)
 *
 * Output:::
 *  pData::			Pointer to a preallocated data buffer with
 *				the same number of bytes
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGetAnnotData(
    XPage *xPage,
    UInt16 annotOffset,
    UInt32 bytes,
    UInt8 *pData
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt16 tableOffset;
    UInt16 entryOffset;
    AnnotTable *annotTable;
    AnnotTableEntry *ep;
    PROLOGUE;

    CheckHandleValidity(xPage);
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);

    /* Check for proper calling sequence */


    /* Read the appropriate annot table */
    tableOffset = annotOffset / xPage->annotTableHeader.entriesPerTable;
    entryOffset = annotOffset % xPage->annotTableHeader.entriesPerTable;
    XERR_PASS_CLEANUP(wp_XifGetAnnotTable(xPage,
	(UInt16)(tableOffset * xPage->annotTableHeader.entriesPerTable),
	&annotTable));

    /* Check to make sure the annotation information has already been requested */
    ep = &annotTable->entries[entryOffset];
    XERR_GEN_CLEANUP(ep->offset == 0, XIF_SEQUENCE);
    XERR_GEN_CLEANUP(ep->bytes != bytes, XIF_BADPARAMETER);

    /* Go to the file offset of the data */
    XIF_SEEK(xFile, ep->offset);

    /* Read the data */
    XIF_READ(xFile, pData, (Int32)bytes);

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}

/*
$Log: xifrxiff.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.35  1998/09/25 22:14:09  aber
aber on Fri Sep 25 15:13:04 PDT 1998

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

If reading an image that was supposedly compressed to zero bytes,
handle it gracefully. It can't really have zero bytes, since no compression
algorithm we have will compress to nothing.

Revision 1.34  1998/08/04 00:21:34  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.33  1998/07/30 22:39:40  aber
aber on Thu Jul 30 15:39:40 PDT 1998

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.32  1998/04/14 01:34:57  aber
aber on Mon Apr 13 18:33:22 PDT 1998

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Bug fix when read-writing a XIFF 2.0 file and storing multiple annotations
in the file.

Revision 1.31  1998/02/09 20:21:27  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.30  1998/01/22 00:23:52  aber
aber on Wed Jan 21 16:23:52 PST 1998

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Allow a 0-byte author block.

Revision 1.29  1998/01/16 00:41:20  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Retain annotation table header file location, for rewrite.

Revision 1.28  1997/11/19 22:30:44  aber
aber on Wed Nov 19 14:30:43 PST 1997

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Used right error code.

Revision 1.27  1997/11/19 21:42:57  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Allow deleting all annotations, but retaining the annotation
table header.

Revision 1.26  1997/11/19 17:54:31  aber
aber on Wed Nov 19 09:53:58 PST 1997

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Avoid compiler complaint, it's smart, but not TOO smart -- the
complaint is unwarranted.

Revision 1.25  1997/11/18 23:43:39  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Add bytesPerEntry field in annotation header. For now there's some
code to accept the absense of this entry, but once Pagis ships we'll
take it back out.
Streamline the reading of the annotation table -- read into a single
array, then remove from the array as needed. Allows for using the
bytesPerEntry header field, and only extracts the bytes that are
supposed to be in the file version. This should make it both backwards
compatible and forward extendible.
Fixed a byte swapping problem when reading annotation type (XIF 2.0
only), that doesn't really appear since byte swapping hasn't been
required yet.

Revision 1.24  1997/11/14 02:09:00  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.23  1997/11/01 20:57:58  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Add strip buffering for wavelet compression.

Revision 1.22  1997/10/24 21:38:56  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.21  1997/10/16 16:15:50  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Remove carriage-returns (left over from PC).

Revision 1.20  1997/10/15 22:15:41  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Allow reading the page table header without reading the page table itself.

Revision 1.19  1997/08/29 17:36:57  aber
aber on Fri Aug 29 10:36:53 PDT 1997

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Fix documentation.

Revision 1.18  1997/08/29 17:28:05  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.17  1997/08/15 23:48:01  aber
aber on Fri Aug 15 16:48:00 PDT 1997

xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.16  1997/08/15 00:10:05  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Slight code cleanup.

Revision 1.15  1997/08/07 21:27:47  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.14  1997/07/24 00:38:28  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

When reading a mask image, mark parent as having one.
Fix logic to keep track of subimages in order.

Revision 1.13  1997/07/14 23:23:30  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Enforce TIFF strip size during lower level image reads.

Revision 1.12  1997/07/10 22:01:26  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Support the ID within XIFF 3.0, fix multiple-table reads.

Revision 1.11  1997/07/02 01:06:03  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Fix return values for no support vs. not a XIFF.

Revision 1.10  1997/06/26 22:28:01  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

Return a new error parameter if the file is not a XIFF.

Revision 1.9  1997/06/17 01:29:28  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: XIFF 2.0 support. Annotation bug fix.

Revision 1.8  1997/06/10 01:26:50  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:10  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:14  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:20  aber
xifrxiff.c -> /product/ipcore/ipwrap/src/RCS/xifrxiff.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:42  aber
Initial revision

*/
