/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwjpeg.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifwjpeg.c:	XIFF WRITE JPEG source
 *
 *	Compression (write) for JPEG.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "jpeglib.h"
#include "iw_jpeg.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifrtiff.h"
#include "xifwjpeg.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwjpeg.c,v $; $Revision: 1.2 $; $Date: 2000/06/12 22:06:05 $");

typedef struct WriteData {
    Instance *instData;
    UInt8 **ppBuffer;
    UInt8 *next;
    UInt32 *pBufBytes;
    UInt32 bytesLeft;
} WriteData;

static ErrCode CDECL
wp_JPEG_WriteCallback(
    IPWState *pIPWState,
    void *userData,
    UInt8 *ptr,
    UInt32 bytes,
    UInt32 *ncount
)
{
    Instance *instData;
    WriteData *writeData = (WriteData *)userData;
    UInt32 newbufsize;
    UInt8 *newbuf = NULL;
    PROLOGUE;

    pIPWState = pIPWState;
    
    if (writeData->bytesLeft < bytes) {
	instData = writeData->instData;

	/* Must reallocate the buffer to accomodate the new amount,
	 * increase by 20% more than what we know we need now */
	newbufsize = (*writeData->pBufBytes - writeData->bytesLeft + bytes)
	    * 120 / 100;
	CHECK_MALLOC(newbuf = XIF_MALLOC(newbufsize));

	/* Copy the data to the new buffer */
	memcpy(newbuf, *writeData->ppBuffer,
	    writeData->next - *writeData->ppBuffer);
	writeData->next = newbuf +
	    (*writeData->pBufBytes - writeData->bytesLeft);
	writeData->bytesLeft = newbufsize -
	    (*writeData->pBufBytes - writeData->bytesLeft);

	/* Update the external buffer */
	XIF_FREE(*writeData->ppBuffer);
	*writeData->ppBuffer = newbuf;
	*writeData->pBufBytes = newbufsize;

	/* Dereference the now stored buffer */
	newbuf = NULL;
    }
    memcpy(writeData->next, ptr, bytes);
    writeData->next += bytes;
    writeData->bytesLeft -= bytes;
    *ncount = bytes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (newbuf) {
	instData = writeData->instData;
	XIF_FREE(newbuf);
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  JPEG compress a strip of image data
 *
 * LongDescription:::
 *  wp_XifCompressStripJPEG() JPEG compresses a strip of image data.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pPixr::			Pointer to image PIXR
 *  dwRowOffset::		Row offset of image strip to compress
 *  dwStripHeight::		Number of rows to compress
 *
 * Output:::
 *  ppCmpData::			Pointer to pointer to the compression buffer,
 *				may be reallocated if not enough space
 *  pdwCmpBytes::		Number of bytes in compression buffer, may
 *				be modified if not enough space
 *  pdwBytesFilled::		Number of bytes filled, in compression buffer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifCompressStripJPEG(
    XImg *xImg,
    PIXR *pPixr,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt8 **ppCmpData,
    UInt32 *pdwCmpBytes,
    UInt32 *pdwBytesFilled
)
{
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState = NULL;
    UInt32 rowsToWrite = dwStripHeight;
    CompressInfo      cinfo_struct;
    Compress_info_ptr cinfo=&cinfo_struct;
    UInt8 *image[3];
    Int32 h, bpl;
    Int32 row, line;
    Int32 nlines;
    Int32 channel, channels;
    UInt32 memorg = 0UL;
    JSAMPIMAGE userBuf = NULL;
    WriteData writeData;
    PROLOGUE;
    
    /* Default to return nothing */
    *pdwBytesFilled = 0;

    /* Can't offset into PIXR for now */
    pxImgInfo = &xImg->imgInfo;
    XERR_GEN_CLEANUP(dwRowOffset + dwStripHeight > pxImgInfo->dwHeight,
	XIF_BADPARAMETER);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &h));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));

    /* Initialize */
    writeData.instData = XIF_IMGINST(xImg);
    writeData.ppBuffer = ppCmpData;
    writeData.next = *writeData.ppBuffer;
    writeData.pBufBytes = pdwCmpBytes;
    writeData.bytesLeft = *writeData.pBufBytes;
    switch (XIF_CHANNELS(xImg->imgInfo.xColorType)) {
    case 3:
	memorg = JMEMORG_BSQ_PIXR;
	break;
    case 1:
	switch (XIF_BPPE(xImg->imgInfo.xColorType)) {
	case 4:
	    memorg = JMEMORG_GRAY4_PIXR;
	    break;
	case 8:
	    memorg = JMEMORG_GRAY8_PIXR;
	    break;
	}
	break;
    }
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    XERR_PASS_CLEANUP(w_jcInit(pIPWState, cinfo, xImg->imgInfo.dwWidth,
	dwStripHeight, channels, xImg->JPEGquality, JC_DEFAULT,
	xImg->imgInfo.dwXResolution, xImg->imgInfo.dwYResolution,
	wp_JPEG_WriteCallback, (void *)&writeData, NULL, memorg));
    
    /* Get the location of the image and the line pitch */
    {
	PIXR  *pSubPixr;
	for (channel = 0; channel < channels; channel++) {
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				&(image[channel])));
	    image[channel] += dwRowOffset * bpl;
	}
    }
    if (h < rowsToWrite) {
	rowsToWrite = h;
    }
    
    XERR_PASS_CLEANUP(w_jcAllocMCURowBuffer(pIPWState, cinfo, cFalse, 
		                      	       &userBuf));
    XERR_PASS_CLEANUP(userBuf == NULL);
    
    /* Decompress the image an MCU row at a time */
    XERR_PASS_CLEANUP(w_jcGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row < rowsToWrite; ++row) {
	for (channel = 0; channel < channels; channel++) {
	    userBuf[channel][line] = image[channel];
	    image[channel] += bpl;
	}
	++line;
	if (line >= nlines || row == (dwStripHeight-1)) {    
	    /* Decompress an MCU row into userBuf */
	    XERR_PASS_CLEANUP( w_jcMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }
    
    XERR_PASS_CLEANUP(w_jcMCURowFree(pIPWState, cinfo, cFalse, userBuf));
    userBuf = NULL;
    
    XERR_PASS_CLEANUP( w_jcTerm(pIPWState, cinfo));

    *pdwBytesFilled = *writeData.pBufBytes - writeData.bytesLeft;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (userBuf != NULL)
	w_jcMCURowFree(pIPWState, cinfo, cFalse, userBuf);
    
    jpeg_destroy_compress(pIPWState, &cinfo->j_cinfo);
    
    w_jcFreeErrorStr(pIPWState, cinfo);
    
    ERR_RETURN;
}

/*
$Log: xifwjpeg.c,v $
Revision 1.2  2000/06/12 22:06:05  jlin
change jpeg read write callback function

Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.14  1998/07/23 21:06:58  xzhu
xzhu on Thu Jul 23 14:06:41 PDT 1998

xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

Follow the change of w_jcInit.

Revision 1.13  1998/01/14 00:16:36  ddavies
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.12  1997/11/20 00:16:53  aber
aber on Wed Nov 19 16:16:22 PST 1997

xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

Add new parameters to set JPEG stream X/Y resolution.

Revision 1.11  1997/09/02 20:02:55  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.10  1997/06/26 22:28:38  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

Allow multiple Wavelet tiles per image, untested.

Revision 1.9  1997/06/17 01:29:30  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

CommonCheck: XIFF 2.0 support. Annotation bug fix.

Revision 1.8  1997/06/10 21:16:29  xzhu
xzhu on Tue Jun 10 14:15:08 PDT 1997

xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

Followed the w_jcInit prototype changes.

Revision 1.7  1997/06/10 17:59:22  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

Remove commented out section of code.

Revision 1.6  1997/06/10 01:26:52  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:12  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:16  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:22  aber
xifwjpeg.c -> /product/ipcore/ipwrap/src/RCS/xifwjpeg.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:44  aber
Initial revision

*/
