/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifpixrw.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifpixrw.c:	XIFF IPWrap WRITE source
 *
 *	XIFF write-only IPWrap level routines. Calls the "Generic XIFF"
 *  routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifwtiff.h"
#include "xifwxiff.h"
#include "xifwpriv.h"
#include "xifrpriv.h"
#include "xifwjpeg.h"
#include "xifwwave.h"
#include "xifwtok.h"
#include "xiff.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwpixr.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

	
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress and write a strip of image data
 *
 * LongDescription:::
 *  w_XifWriteStrip() compresses and writes a strip of image data. If necessary,
 *  the image data is broken up into TIFF strips, and stored as separate strips.
 *  The last strip may be stored until the next block of data is added to
 *  complete a TIFF strip.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  pPixr::			Pointer to the image PIXR
 *  dwPixrRowOffset::		Row offset in PIXR of strip to write
 *  dwStripHeight::		Number of rows in PIXR to write
 *  dwImgRowOffset::		Row offset in image to write strip
 *
 * Output:::
 *  pdwRowsFilled::		Number of rows written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifWriteStrip(
	IPWState *pIPWState,
	XIF_IMGHANDLE xImgHandle,
	PIXR *pPixr,
	UInt32 dwPixrRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pdwRowsFilled
)
{
    Instance *instData;
    IPWState *internalIPWState = NULL;
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    UInt32 mallocBytes;
    UInt32 stripOffset;
    UInt32 channel, channels;
    UInt32 clientRowsLeft;
    UInt32 tiffRows;
    UInt32 rowsFilled;			    /* Rows filled in intermediate buffer */
    Int32 srcBPL, dstBPL;
    UInt8 *srcPtr, *dstPtr;		    /* Image data pointers */
    UInt32 totalRowsStored = 0UL;	    /* Number of rows stored (buffered/written) */
    UInt32 bytesWritten;
    Int32 pixrHeight, tmpPixrHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    CheckHandleValidity(xImgHandle);

    instData = XIF_IMGINST(xImg);    
    pxImgInfo = &xImg->imgInfo;
    channels = XIF_BPP(pxImgInfo->xColorType) /
      XIF_BPPE(pxImgInfo->xColorType);
    
    /* Check for proper calling sequence */
    if ((xImg->seqStatus & XIMG_SEQ_WRITESTATUS) != XIMG_SEQ_WRITESTART) {
	__error__ = XIF_SEQUENCE;
    }
    XERR_GEN_CLEANUP(dwImgRowOffset >= pxImgInfo->dwHeight,
	XIF_BADPARAMETER);

    /* Be kind, and ignore any extra that was given */
    if (dwImgRowOffset + dwStripHeight > pxImgInfo->dwHeight) {
	dwStripHeight = pxImgInfo->dwHeight - dwImgRowOffset;
    }
    clientRowsLeft = dwStripHeight;

    /* Save the internally used IPWState, and use the passed in one.
     * Restore when finished. */
    if (xImg->xPage->xDoc->xFile->instance->pIPWState != 0) {
        internalIPWState =
	    (IPWState *)xImg->xPage->xDoc->xFile->instance->pIPWState;
    }
    xImg->xPage->xDoc->xFile->instance->pIPWState = (UInt32)pIPWState;

    /* Create a compressed buffer, if we haven't already */
    if ((xImg->cmpBuffer == NULL) &&
	(pxImgInfo->xCompression != XIF_COLORFORTOKENS)) {

	/* Expect reasonable compression, unless storing uncompressed */
	if (pxImgInfo->xCompression == XIF_NOC) {
	    mallocBytes = xImg->minBytesPerRow * pxImgInfo->dwBestStripHeight *
		XIF_CHANNELS(pxImgInfo->xColorType);
	} else {
	    mallocBytes = (xImg->minBytesPerRow * pxImgInfo->dwBestStripHeight
		* 50) / 100;
	}
	if (mallocBytes < 1000) {
	    mallocBytes = 1000;
	}
	CHECK_MALLOC(xImg->cmpBuffer = XIF_MALLOC(mallocBytes));
	xImg->cmpBufsize = mallocBytes;
    }

    /* Get the size of the input PIXR */
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    
    /* Continue filling intermediate buffer (if buffering) */
    /*
    CopyToTempImage(xImg, &pBuffer, &clientRowsLeft, dwBytesPerRow);
    */
    
    /* Loop over input data until it's exhausted */
    while (clientRowsLeft > 0) {

	/* Compute the TIFF strip size */
	stripOffset = dwImgRowOffset / pxImgInfo->dwBestStripHeight;
	if (stripOffset * pxImgInfo->dwBestStripHeight +
	    pxImgInfo->dwBestStripHeight > pxImgInfo->dwHeight) {
	    tiffRows = pxImgInfo->dwHeight -
		stripOffset * pxImgInfo->dwBestStripHeight;
	} else {
	    tiffRows = pxImgInfo->dwBestStripHeight;
	}

	/* Determine if the data needs to be buffered:
	 * 1. If the number of rows doesn't fill a TIFF strip.
	 * 2. I have previously buffered data in an unwritten strip.
	 * 3. I'm wavlet compressing and the input PIXR is not a full TIFF
	 *    strip (either has a y-offset or not the same height).
	 */
	if (((xImg->tmpPixrRows > 0 && xImg->tmpPixrRows < tiffRows)
	    || clientRowsLeft < tiffRows) ||
	    (pxImgInfo->xCompression == XIF_WAVELET &&
	     (pixrHeight != tiffRows || dwPixrRowOffset != 0))) {

	    /* If wavelet compressing, the entire PIXR must be a full TIFF
	     * strip (we can't tell the wavelet routines to compress only a
	     * portion of the PIXR). */
	    if (pxImgInfo->xCompression == XIF_WAVELET &&
		xImg->tmpPixr != NULL) {
		XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, xImg->tmpPixr,
						  &tmpPixrHeight));
		if (tmpPixrHeight != tiffRows) {
		    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &xImg->tmpPixr));
		    xImg->tmpPixr = NULL;
		}
	    }

	    /* If no intermediate buffer in place, create it */
	    if (xImg->tmpPixr == NULL) {

		/* If this is the first buffered row, it should fall on
		 * a TIFF strip boundary */
		XERR_GEN_CLEANUP(dwImgRowOffset % pxImgInfo->dwBestStripHeight,
		    XIF_SEQUENCE);

		/* Prepare to allocate the data */
		{
		    ScaledInt1616 xres, yres;
		    ColorType ctype = cColorNone;
		
		    /* Prepare a temporary PIXR */
		    xres = Int32_TO_ScaledInt1616(pxImgInfo->dwXResolution);
		    yres = Int32_TO_ScaledInt1616(pxImgInfo->dwYResolution);
		    switch ((int)pxImgInfo->xColorType) {
		    case XIF_COLORTYPE_COLOR24:
			ctype = cColorRGB;
			break;
		    case XIF_COLORTYPE_BINARY:
		    case XIF_COLORTYPE_GRAY4:
		    case XIF_COLORTYPE_COLOR4:
		    case XIF_COLORTYPE_GRAY8:
		    case XIF_COLORTYPE_COLOR8:
			ctype = cColorGray;
			break;
		    default:
			goto ERR_CLEANUP;
		    }   
		    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, pxImgInfo->dwWidth,
			tiffRows,
			XIF_BPPE(pxImgInfo->xColorType),
			xres, yres, 32,
			XIF_CHANNELS(pxImgInfo->xColorType),
			ctype, NULL, &xImg->tmpPixr));
		}
		xImg->tmpPixrRow = dwImgRowOffset;
		xImg->tmpPixrRows = 0UL;
	    } /* If creating an intermediate PIXR */

	    /* Make sure there are no gaps in the image data (adjacent calls
	     * didn't specify adjacent image data) */
	    XERR_GEN_CLEANUP(xImg->tmpPixrRow + xImg->tmpPixrRows !=
		dwImgRowOffset, XIF_SEQUENCE);

	    /* Copy the rows relevant to this TIFF strip */
	    rowsFilled = 0;
	    for (channel = 0; channel < channels; channel++) {
		PIXR *pSubPixr;
		UInt32 rowsAvailable, byteCount;

		XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &srcBPL));
		XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, xImg->tmpPixr,
					       &dstBPL));
		if (channels > 1) {
		    pSubPixr = NULL;
		    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr,
						       channel,
			&pSubPixr));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
						     &srcPtr));
		    pSubPixr = NULL;
		    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState,
						       xImg->tmpPixr, channel,
			&pSubPixr));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
						     &dstPtr));
		} else {
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
						     &srcPtr));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, xImg->tmpPixr,
						     &dstPtr));
		}
		srcPtr += dwPixrRowOffset * srcBPL;
		dstPtr += (dwImgRowOffset - xImg->tmpPixrRow) * dstBPL;
		rowsAvailable = tiffRows - (dwImgRowOffset - xImg->tmpPixrRow);
		if (rowsAvailable > clientRowsLeft) {
		    rowsAvailable = clientRowsLeft;
		}
		byteCount = MIN(srcBPL, dstBPL);

		/* Copy the PIXR image to the user's image */
		while (rowsAvailable--) {
		    memcpy(dstPtr, srcPtr, byteCount);
		    srcPtr += srcBPL;
		    dstPtr += dstBPL;
		    rowsFilled++;
		}
	    } /* for each channel */
	    rowsFilled /= channels;
	    xImg->tmpPixrRows += rowsFilled;
	    dwPixrRowOffset += rowsFilled;
	    dwImgRowOffset += rowsFilled;
	    clientRowsLeft -= rowsFilled;
	    totalRowsStored += rowsFilled;

	    if (pxImgInfo->xCompression != XIF_COLORFORTOKENS) {

		/* Compress intermediate buffer if filled */
		if (xImg->tmpPixrRows == tiffRows) {
		    XERR_PASS_CLEANUP(w_XifCompress(pIPWState, (UInt32)xImg,
			pxImgInfo->xCompression, pxImgInfo->xColorType,
			xImg->imgFlags, xImg->tmpPixr, 0,
			pxImgInfo->dwWidth, xImg->tmpPixrRows,
			&xImg->cmpBuffer, &xImg->cmpBufsize, &xImg->cmpBytes));
		}
	    }
	} else { /* Not using an intermediate buffer: */

	    /* Compress from client image directly */
	    XERR_PASS_CLEANUP(w_XifCompress(pIPWState, (UInt32)xImg,
		pxImgInfo->xCompression, pxImgInfo->xColorType,
		xImg->imgFlags, pPixr, dwPixrRowOffset, pxImgInfo->dwWidth,
		tiffRows, &xImg->cmpBuffer, &xImg->cmpBufsize,
		&xImg->cmpBytes));
	    dwPixrRowOffset += tiffRows;
	    dwImgRowOffset += tiffRows;
	    clientRowsLeft -= tiffRows;
	    totalRowsStored += tiffRows;
	} /* Not using an intermediate buffer */

	/* Write the strip */
	if (xImg->cmpBytes) {
	    /* Enable writing compressed data */
	    xImg->imgFlags |= XIF_IMGFLAGS_Compressed;

	    /* Write the compressed data */
	    XERR_PASS_CLEANUP(XIF_CompressedWriteStrip(xImgHandle,
		xImg->tmpPixrRow, tiffRows, xImg->cmpBuffer,
		xImg->cmpBytes, &bytesWritten));
	    	
	    /* Disable writing compressed data */
	    xImg->imgFlags &= ~XIF_IMGFLAGS_Compressed;

	    /* Reset for next TIFF strip */
	    xImg->tmpPixrRow = dwImgRowOffset;
	    xImg->tmpPixrRows = 0;
	    xImg->cmpBytes = 0UL;
	}
    } /* While rows to write */

    /* Restore the state */
    if (internalIPWState != NULL) {
	xImg->xPage->xDoc->xFile->instance->pIPWState =
	    (UInt32)internalIPWState;
    }

    /* Return the number of rows buffered/written */
    *pdwRowsFilled += totalRowsStored;
    NORMAL_RETURN;
    
 ERR_CLEANUP:

    if (internalIPWState != NULL) {
	xImg->xPage->xDoc->xFile->instance->pIPWState =
	    (UInt32)internalIPWState;
    }

    ERR_RETURN;
}

	
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress a strip of image data
 *
 * LongDescription:::
 *  w_XifCompress() compresses a single image strip. Used as a standalone
 *  compressor, allows an application to retrieve the image data from file
 *  in raw compressed form, and decode individual strips on the fly as needed.
 *  <BR>
 *  Source image is compressed into the destination buffer. If the destination
 *  buffer is filled and would overflow, XIF_BUFFEREXCEEDED is returned.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  xCompression::		See XIF_COMPRESSION types
 *  xColorType::		Type of color
 *  pPixr::			Pointer to the source PIXR
 *  dwFlags::			XIF_IMGFLAGS_xxx combo
 *  dwRowOffset::		Row offset in image of data to compress
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwHeight::			Rows in dst image
 *
 * Output:::
 *  ppDst::			Pointer to the destination compressed data buffer,
 *				to be modified if compression requires more
 *  pdwBytesInDst::		Pointer to the byte count of destination buffer,
 *				to be modified if compression requires more
 *  pdwBytesFilled::		Actual bytes in destination filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_XifCompress(
	IPWState *pIPWState,
	XIF_IMGHANDLE xImgHandle,
	XIF_COMPRESSION xCompression,
	XIF_COLORTYPE xColorType,
	UInt32 dwFlags,
	PIXR *pPixr,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt8 **ppDst,
	UInt32 *pdwBytesInDst,
	UInt32 *pdwBytesFilled
)
{
    Instance *instData;
    XImg *xImg = (XImg *)xImgHandle;
    Int32 channel, channels;
    UInt32 rowsLeft;
    UInt32 bytesLeft;
    Int32 pixrWidth;
    Int32 pixrHeight;
    Int32 srcBPL, dstBPL, bytesPerRow;
    UInt32 bytesUsed = 0UL;
    Int32 bytesFilled;
    Int32 colorType;
    Int32 colorDepth;
    PIXR *pSubPixr;
    UInt8 *imageChannels[3];
    UInt8 *src, *dst;
    UInt8 continued = cFalse;
    UInt8 stage;
    UInt8 *newbuf;
    UInt32 newbufSize;
    PROLOGUE;

    *pdwBytesFilled = 0UL;

    CheckHandleValidity(xImgHandle);
    instData = XIF_IMGINST(xImg);

    /* Check buffers and sizes */
#if 0
//    XERR_GEN_CLEANUP((pPixr == NULL || ppDst == NULL || *ppDst == NULL),
//		     XIF_BADPARAMETER);
#endif
    XERR_GEN_CLEANUP(pPixr == NULL || ppDst == NULL,
		     XIF_BADPARAMETER);

    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetType(pIPWState, pPixr, &colorType));
    XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, &pixrWidth));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    XERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, pPixr, &colorDepth));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &srcBPL));
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				       &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				     &imageChannels[channel]));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
				     &imageChannels[channel]));
	}
    }

    /* If there's no memory preallocated, initialize to something */
    if (*pdwBytesInDst == 0 || *ppDst == NULL) {
	if ((int)xCompression == XIF_NOC) {
	    newbufSize = dwWidth * dwHeight * XIF_BPP(xColorType) / 8;
	} else {
	    newbufSize = dwWidth * dwHeight * XIF_BPP(xColorType) / 8 / 100;    /* 100:1 compression */
	}
	CHECK_MALLOC(newbuf = XIF_MALLOC(newbufSize));

	/* Replace old buffer */
	if (*ppDst) {
	    XIF_FREE(*ppDst);
	}
	*ppDst = newbuf;
	*pdwBytesInDst = newbufSize;
    }
    
    rowsLeft = dwHeight;
    bytesLeft = *pdwBytesInDst;
    dstBPL = xImg->minBytesPerRow;
    bytesPerRow = dstBPL;
    if (bytesPerRow > srcBPL) {
	bytesPerRow = srcBPL;
    }
    dst = *ppDst;

    /* Convert the image format */
    if (dwFlags) {
#if 1
      dwFlags = 0;
#else
      XifConvImage(xImg, dwFlags, etc.);
#endif
    }

    switch ((int)xCompression) {
    case XIF_NOC:

	XERR_GEN_CLEANUP((bytesPerRow * dwHeight > bytesLeft),
		 XIF_BUFFEREXCEEDED);
	XERR_PASS(wp_XifPackImage(pIPWState, imageChannels, srcBPL, dwRowOffset,
	    dwHeight, channels, dst, dstBPL, (UInt32 *)&bytesFilled));
	bytesUsed += bytesFilled;
	break;
    case XIF_G4:
	{
	    void *compInfo;
	    UInt32 imgExh, cmpExh;
	    Int32 pct = 0;
	    
	    if (xImg->tmpCmpData == NULL) {	/* First strip */
		Int32 stateBlockSize;
		UInt32 padType = cNoPad;
		
		/* Prepare for pre-decode */
		XERR_PASS_CLEANUP(i_getSizeOfCCITTCompressionInfo(
		    pIPWState, dwWidth + 2*cFrameBits, &stateBlockSize));
		
		/* Allocate temporary compression data, attach to image */
		CHECK_MALLOC(xImg->tmpCmpData =
		     (MISC_MEM *)XIF_MALLOC(sizeof(MISC_MEM) +
				    stateBlockSize));
		xImg->tmpCmpData->next = NULL;
		
		/* Data follows immediately */
		xImg->tmpCmpData->mem = compInfo = (void *)(xImg->tmpCmpData + 1);
		
		XERR_PASS_CLEANUP(i_initializeCCITTCompressionInfo(
		    pIPWState, compInfo, dwWidth + 2*cFrameBits, cTrue,
		    (xImg->imgInfo.dwXResolution > 150 ? 4 : 2), padType,
		    stateBlockSize));
	    } else {
		compInfo = xImg->tmpCmpData->mem;
	    }

	    src = imageChannels[0] - (cFrameBits * XIF_BPPE(xColorType)) / 8;
	    continued = cFalse;
	    bytesUsed = 0UL;
	    stage = 1;
	    do {
		bytesFilled = 0;
		switch (stage) {
		case 1:
		    /* Compress using MMR, don't flush yet */
		    XERR_PASS_CLEANUP(i_compressRowsMMR (pIPWState,
		        compInfo, src,
			dwWidth + 2 * cFrameBits, dwRowOffset, rowsLeft,
			srcBPL, dst, bytesLeft, continued, continued,
			&imgExh, &cmpExh, &pct, &bytesFilled));

		    if (imgExh == 0) {
			XERR_GEN_CLEANUP((cmpExh == 0), XIF_INTERNAL_ERROR);
		    } else {
			stage++;
		    }

		    break;
		case 2:
		    /* Add the two EOL markers required by TIFF G4 */
		    XERR_PASS_CLEANUP(i_writeCCITTEOL (pIPWState,
			compInfo, cNoPad,
			cPlainEOL, dst, bytesLeft, &bytesFilled));
		    stage++;
		    break;
		case 3:
		    /* Flush the rest of the data */
		    XERR_PASS_CLEANUP(i_flushCCITT (pIPWState,
			compInfo, dst, bytesLeft,
			&bytesFilled));
		    stage++;
		    break;
		}
		bytesUsed += bytesFilled;
		bytesLeft -= bytesFilled;
		dst += bytesFilled;

		/* Increase the compressed buffer size, if needed */
		if ((imgExh == 0 && cmpExh == 1) || bytesLeft < 10) {

		    XERR_GEN_CLEANUP(imgExh && cmpExh == 0 && bytesLeft >= 10,
				     XIF_INTERNAL_ERROR);

		    /* Filled the compressed buffer, reallocate the compressed
		     * data buffer */

		    /* Anticipate the compression due to the percentage, plus a
		     * fudge */
		    if (pct > 0 && pct <= 100) {
			newbufSize = *pdwBytesInDst * (100 + 10) / pct;
		    } else {
			newbufSize = *pdwBytesInDst * (200 - pct + 10) / 100;
		    }
		    CHECK_MALLOC(newbuf = XIF_MALLOC(newbufSize));
		    memcpy(newbuf, *ppDst, bytesUsed);

		    /* Replace old buffer */
		    XIF_FREE(*ppDst);
		    *ppDst = newbuf;
		    *pdwBytesInDst = newbufSize;
		    dst = newbuf + bytesUsed;
		    bytesLeft = newbufSize - bytesUsed;
		    continued = cTrue;
		}
	    } while (imgExh == 0 || stage < 4);

	    /* Can't compress given the buffer size, reallocate and go back
	     * in */
	    XERR_GEN_CLEANUP((cmpExh && imgExh == 0), XIF_BUFFEREXCEEDED);
	}	/* G4 write */
	break;
    case XIF_JPEG:
	XERR_PASS_CLEANUP(wp_XifCompressStripJPEG(xImg, pPixr, dwRowOffset,
	    rowsLeft, ppDst, pdwBytesInDst, (UInt32 *)&bytesFilled));

	bytesUsed += bytesFilled;
	break;
    case XIF_LLTOKENIZATION:
    case XIF_TOKENIZATION:
	break;
    case XIF_WAVELET:
	XERR_PASS_CLEANUP(wp_XifCompressStripWavelet(xImg, pPixr, dwRowOffset,
	    rowsLeft, ppDst, pdwBytesInDst, (UInt32 *)&bytesFilled));

	bytesUsed += bytesFilled;
	break;
    case XIF_LOSSYDITHER:
	XERR_PASS_CLEANUP(wp_XifCompressStripLossyDither(xImg,
	    imageChannels[0], srcBPL, dwRowOffset, dwHeight, ppDst,
	    pdwBytesInDst, (UInt32 *)&bytesFilled));

	bytesUsed += bytesFilled;
	break;
    default:
	break;
    }

    /* Return the compressed buffer size */
    *pdwBytesFilled = bytesUsed;

    /* Set the appropriate mode */
#if 0
//    xImg->imgFlags |= XIF_IMGFLAGS_Compressed;
#endif


    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy image data in PIXR order to a byte-packed image stream
 *
 * LongDescription:::
 *  wp_XifPackImage() copies image data from a PIXR to a byte-packed image
 *  buffer. This is used to store the "uncompressed" form of data, independent
 *  of platform. If the source image is multi-channel, then the destination
 *  image is one channel after another.
 *  <BR>
 *  The source image buffers are left intact.
 *
 * Input:::
 *  pIPWState::			Pointer to shared IPCore state
 *  srcChannels::		Input image channels (1-3)
 *  srcBPL::			Bytes per line of data (each channel)
 *  srcRowOffset::		Source row offset
 *  height::			Number of lines to transfer
 *  channels::			Number of channels to transfer
 *
 * Output:::
 *  dst::			Destination image buffer
 *  dstBPL::			Destination image bytes per line
 *  pBytesFilled::		Bytes filled in destination image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifPackImage(
	     IPWState *pIPWState,
	     UInt8 **srcChannels,
	     Int32 srcBPL,
	     UInt32 srcRowOffset,
	     Int32 height,
	     Int32 channels,
	     UInt8 *dst,
	     Int32 dstBPL,
	     UInt32 *pBytesFilled
)
{
    GraySwapNearProc swapLineProc;
    UInt8 *src;
    Int32 channel;
    Int32 row, lineBPL;
    UInt8 *line = NULL;
    PROLOGUE;

    XERR_GEN_CLEANUP(srcBPL < dstBPL, XIF_BADPARAMETER);

    ip_getByteSwapProc (pIPWState, cFalse, cPixrToPixr,
			NULL, NULL, &swapLineProc);

    /* Compute the intermediate line BPL */
    lineBPL = (dstBPL + 3) & ~3;

    /* Allocate an intermediate line buffer (for swapping) */
    XERR_GEN_CLEANUP(((line = (UInt8 *)malloc(lineBPL)) == NULL),
		      XIF_NOMEMORY);

    for (channel = 0; channel < channels; channel++) {
      src = srcChannels[channel] + srcRowOffset * srcBPL;
      for (row = 0; row < height; row++) {

	/* Swap the source line to the intermediate one */
	(*swapLineProc)(pIPWState, (UInt32 *)src, (UInt32 *)line, lineBPL / 4);
	src += srcBPL;

	/* Copy the intermediate line to the destination, byte packing */
	memcpy(dst, line, dstBPL);
	dst += dstBPL;
      }
    }

    free(line);

    *pBytesFilled = height * dstBPL * channels;
    NORMAL_RETURN;

 ERR_CLEANUP:
    if (line) {
      free(line);
    }

     ERR_RETURN;
}

/*
$Log: xifwpixr.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.24  1998/08/04 00:21:35  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.23  1998/05/08 05:59:35  aber
aber on Thu May 7 22:58:20 PDT 1998

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Limit how small the estimated compression buffer will be to be at least
1000 bytes. Found a case where a 1-line strip estimates to 0 bytes, which
returns NULL from malloc, which signals an error condition.

Revision 1.22  1998/05/05 10:00:37  aber
aber on Tue May 5 02:59:24 PDT 1998

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

One more G4 cleanup. Added special case for output compression size
when leaving uncompressed.

Revision 1.21  1998/05/05 08:45:09  aber
aber on Tue May 5 01:44:35 PDT 1998

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Logic fix: avoid generating an error when just adding a small
cushion for G4 compression.

Revision 1.20  1998/05/05 07:52:53  aber
aber on Tue May 5 00:52:01 PDT 1998

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Changed logic in G4 compress to add a cushion when adding the
EOL marker. Found a case where it could fail.

Revision 1.19  1998/05/05 07:38:53  aber
aber on Tue May 5 00:37:28 PDT 1998

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Modified compression expansion equation when G4 compressed buffer
estimate was off, improves efficiency on a test case I looked at.

Revision 1.18  1998/05/05 07:15:28  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: Additions necessary to fill in the never completed functions
XIF_Compress and XIF_Decompress.

Revision 1.17  1998/04/06 22:44:50  aber
aber on Mon Apr 6 15:44:02 PDT 1998

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

To make things work potentially a little smoother, allow the client to
pass in more data than specified in the image, but ignore the extra.

Revision 1.16  1998/03/16 21:03:24  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Fix to uncompressed image write (recompute byte count when storing
3-channel imagery). Normally RGB images are compressed, but we should
still handle uncompressed mode.

Revision 1.15  1998/01/14 00:16:38  ddavies
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.14  1997/11/01 20:59:22  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Add strip buffering for wavelet compression.

Revision 1.13  1997/07/29 23:52:05  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Fix bug where G4 data < 10 bytes, but failure to read.

Revision 1.12  1997/07/15 23:43:20  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Redo the buffering logic (finish it).

Revision 1.11  1997/06/24 17:44:21  aber
aber on Tue Jun 24 10:43:49 PDT 1997

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Less-than should be greater-than.

Revision 1.10  1997/06/23 22:59:01  aber
aber on Mon Jun 23 15:59:00 PDT 1997

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: Complete function calls for paletted image support.

Revision 1.9  1997/06/23 16:54:58  aber
aber on Mon Jun 23 09:54:32 PDT 1997

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Reference previously "unused variable".

Revision 1.8  1997/06/20 17:47:32  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Added parameter check.

Revision 1.7  1997/06/17 22:25:27  aber
aber on Tue Jun 17 15:24:31 PDT 1997

xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

Use "compression buffer continued" = cTrue for CCITT compression. Bug fix.

Revision 1.5  1997/06/10 01:26:54  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.4  1997/05/30 00:51:13  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.3  1997/05/20 22:29:17  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.2  1997/05/20 00:54:24  aber
xifwpixr.c -> /product/ipcore/ipwrap/src/RCS/xifwpixr.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:45  aber
Initial revision

*/
