/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwpriv.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifwpriv.c:	XIFF WRITE private routines
 *
 *	XIFF private miscellaneous routines
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifwtiff.h"
#include "xifwxiff.h"
#include "xifwpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwpriv.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Align a file to a "long" byte count
 *
 * LongDescription:::
 *  wp_FileLongAlign() adjusts the size of the file such that the total byte
 *  count is divisible by 4. Empty bytes are used to pad the end of the file
 *  to adjust the file size.
 *
 * Input:::
 *  xFile::			Active file token
 *
 * Output:::
 *  offset::			Offset of the resulting file size
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_FileLongAlign(
	      XFile *xFile,
	      Int32 *offset
)
{
    UInt32 zero = 0UL;
    PROLOGUE;

    XIF_FILESIZE(xFile, *offset);
    XIF_SEEK(xFile, *offset);
    if (*offset & 3) {
	XIF_WRITE(xFile, &zero, 4 - (*offset & 3));
	*offset = (*offset + 3) & ~3;
    }

    NORMAL_RETURN;
ERR_CLEANUP:

    ERR_RETURN;
}

/*
$Log: xifwpriv.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.6  1997/06/12 00:52:13  aber
xifwpriv.c -> /product/ipcore/ipwrap/src/RCS/xifwpriv.c,v

Actually seek to the end of the file when long aligning, instead
of assuming that the file size routine seeks there.

Revision 1.5  1997/05/30 00:51:15  aber
xifwpriv.c -> /product/ipcore/ipwrap/src/RCS/xifwpriv.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:19  aber
xifwpriv.c -> /product/ipcore/ipwrap/src/RCS/xifwpriv.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:27  aber
xifwpriv.c -> /product/ipcore/ipwrap/src/RCS/xifwpriv.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:46  aber
Initial revision

*/
