/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwpub.c
 *
 *	XIFF write public "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifwtiff.h"
#include "xifwxiff.h"
#include "xifwpriv.h"
#include "xifrxiff.h"
#include "xifwpub.h"
#include "xifrpriv.h"
#include "xifrtok.h"
#include "xifwtok.h"
#include "xifconv.h"
#include "xiff.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwpub.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy token dictionaries associated with a position block to a new document
 *
 * LongDescription:::
 *  wp_XifCopyTokenDictionary() copies the token dictionary/ies, associated
 *  with a position block, from one XIF document to another; and modifies
 *  the token dictionary IDs found in the Tundra position/ID stream.  The
 *  token dictionary is left in Tundra format during the copy, although the
 *  dictionary IDs associated with the Tundra dictionaries may change during
 *  the transfer.  This routine is used for image/page coping, and allows
 *  much faster copies.
 *  <BR>
 *  Since token dictionaries span more than one page/image, there is some
 *  internal states that are maintained to keep track of which token
 *  dictionaries have already been copied.  These are keyed off the ID of
 *  the document handle to which the dictionary is copied to, and
 *  maintained in the document handle of the originating dictionary.
 *  <BR>
 *  If the dictionary has previously been copied, the copy is not repeated.
 *  The new dictionary ID is returned, and replaces the ID stored in the
 *  Tundra position/ID header of the input stream.
 *
 * Input:::
 *  srcImg::			Source image handle, with Tundra
 *				position/ID block
 *  imgData::			Tundra position/ID block (header
 *				containing old dictionary ID will be
 *				updated)
 *  dataBytes::			Number of bytes in imgData
 *
 * Output:::
 *  dstDoc::			Destination document
 *  ppDictIDList::		Pointer to the new dictionay ID list (memory
 *				is attached to the image, and should not
 *				be freed by the client)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifCopyTokenDictionary(
	     XImg *srcImg,
	     UInt8 *imgData,
	     UInt32 dataBytes,
	     XDoc *dstDoc,
	     UInt32 **ppDictIDList
)
{
    Instance *instData;
    XFile *srcFile, *dstFile;
    XDoc *srcDoc;
    XIF_IMGINFO *inImgInfo;
    UInt8 *dictData = NULL, *ptr;
    Int32 dictBytes;
    CopiedDicts *pd;		/* Previously copied dictionaries */
    UInt32 srcDictOffset, newOffset;
    Int32 srcOffset, dstOffset;	/* File offsets */
    UInt16 dictNum, dictCount;
    UInt32 tempLong;
    UInt32 *newIDs = NULL;
    PROLOGUE;

    /* Reference the image information */
    instData = XIF_IMGINST(srcImg);
    inImgInfo = &srcImg->imgInfo;
    XERR_GEN_CLEANUP(inImgInfo->xCompression != XIF_TOKENIZATION &&
	inImgInfo->xCompression != XIF_LLTOKENIZATION, XIF_INTERNAL_ERROR);

    /* At least the header should exist */
    XERR_GEN_CLEANUP(dataBytes < 6, XIF_BADPARAMETER);

    /* Reference the source document, and document files */
    srcDoc = srcImg->xPage->xDoc;
    srcFile = XIF_IMGFILE(srcImg);
    dstFile = XIF_DOCFILE(dstDoc);

    /* Get the token dictionary ID from the Tundra data
     */

    /* Read the pre-Tundra header, also stored in xifrtok.c */
    ptr = imgData;
    memcpy(&dictCount, ptr + cXifTCPosPositionDictCount, sizeof(dictCount));
    SwapShort(srcFile, &dictCount);

    /* Allocate a list of new dictionary IDs */
    if (dictCount) {
	CHECK_MALLOC(newIDs = XIF_MALLOC(dictCount * sizeof(UInt32)));
    }

    /* Point now to the array of dictionary IDs */
    for (dictNum = 0; dictNum < dictCount; dictNum++) {
	memcpy (&srcDictOffset, ptr + cXifTCPosPositionDictIDs +
		 dictNum * sizeof(srcDictOffset), sizeof(UInt32));
        SwapLong(srcFile, &srcDictOffset);

	/* Check the list of previously copied token dictionary IDs */
	pd = srcDoc->prevTDCopied;
	while (pd && (pd->dstDocHandle != (XIF_DOCHANDLE)dstDoc ||
		      pd->dstDocID != dstDoc->docID ||
		      pd->dictOffset != srcDictOffset)) {
	    pd = pd->next;
	}

	/* If we found it, use it */
	if (pd) {
	    newOffset = pd->newOffset;
	} else {
	    newOffset = dstDoc->symlibTableHeader.symlibCount;
	}

	/*
	 * Copy the token dictionary if needed
	 */
	if (! pd) {		/* Not found in list of previous copies */
	    
	    /* Get the size and offset of the token dictionary */
	    XERR_PASS_CLEANUP(wp_XifGetSymlibOffset(srcDoc, srcDictOffset,
		&srcOffset, &dictBytes));

	    /* Check for a valid dictionary file position */
	    XERR_GEN_CLEANUP(srcOffset == 0, XIF_BADFILEFORMAT);

	    /* Allocate a dictionary buffer */
	    CHECK_MALLOC(dictData = (UInt8 *)XIF_MALLOC(dictBytes));

	    /* Read the dictionary */
	    XIF_SEEK(srcFile, srcOffset);
	    XIF_READ(srcFile, dictData, dictBytes);

	    /* At this point we need to reswap the endian dependent
	     * items */
	    if (XIF_FILETOKEN(srcFile)->bSwapBytes !=
		XIF_FILETOKEN(dstFile)->bSwapBytes) {
	      UInt32 dictSize;

	      memcpy(&dictSize, dictData + cXifTCPosDictSize,
		     sizeof(dictSize));
	      SwapL(&dictSize);
	      memcpy(dictData + cXifTCPosDictSize, &dictSize,
		     sizeof(dictSize));
	    }

	    /* Store the dictionary */
	    XERR_PASS_CLEANUP(wp_FileLongAlign(dstFile, &dstOffset));
	    XIF_WRITE(dstFile, dictData, dictBytes);

	    /* Deallocate compressed data buffer */
	    XIF_FREE(dictData);
	    dictData = NULL;

	    /* Update the symbol library table */
	    XERR_PASS_CLEANUP(wp_XifSetSymlibOffset(dstDoc,
		dstDoc->symlibTableHeader.symlibCount, dictBytes, dstOffset));

	    /* Remember that this dictionary was transferred */
	    CHECK_MALLOC(pd = (CopiedDicts *)XIF_MALLOC(sizeof(CopiedDicts)));
	    memset(pd, 0, sizeof(*pd));
	    pd->dstDocHandle = (XIF_DOCHANDLE)dstDoc;
	    pd->dstDocID = dstDoc->docID;
	    pd->dictOffset = srcDictOffset;
	    pd->newOffset = newOffset;

	    if (srcDoc->prevTDCopied) {
		pd->next = srcDoc->prevTDCopied;
	    }
	    srcDoc->prevTDCopied = pd;
	} /* Not previously copied */

	/* Replace the ID in the input position block Tundra stream */
	if (srcDictOffset != newOffset ||
	    XIF_FILETOKEN(srcFile)->bSwapBytes !=
	    XIF_FILETOKEN(dstFile)->bSwapBytes) {
	    tempLong = newOffset;
	    SwapLong(dstFile, &tempLong);
	    memcpy(ptr + cXifTCPosPositionDictIDs + dictNum *
		   sizeof(tempLong), &tempLong, sizeof(tempLong));
	}

	/* Add it to the list of new IDs */
	newIDs[dictNum] = newOffset;
    } /* For each dictionary referenced */

    /* Attach the list of dictionary IDs to the document, for
     * memory management (freed when document is closed). */
    if (srcDoc->curDictIDs) {
	XIF_FREE(srcDoc->curDictIDs);
    }
    srcDoc->curDictIDs = newIDs;
	
    /* Return the new ID */
    *ppDictIDList = newIDs;
    newIDs = NULL;
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (dictData) {
	XIF_FREE(dictData);
	dictData = NULL;
    }
    if (newIDs) {
	XIF_FREE(newIDs);
	newIDs = NULL;
    }
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy an individual image from one XIF to another
 *
 * LongDescription:::
 *  wp_XifCopyImage() copies an individual image from one XIF to another,
 *  leaving the image data compressed during the copy.  This allows a much
 *  faster copy, since the data is not decompressed and recompressed.  It also
 *  means that the use of a lossy algorithm (like JPEG) will not produce added
 *  lossy affects by recompressing with additional or different loss.
 *
 * Input:::
 *  srcImg::			Source image handle
 *
 * Output:::
 *  parentImg::			If source image is a transparency mask, then
 *				this is the parent image to attach the output
 *				image to, and disregarded otherwise
 *  dstPage::			Page to attach copy of image to
 *  pOutImgInfo::		Resulting output image information. Image
 *				characteristics will be adopted from the srcImg
 *				except for the following structure elements of
 *				pOutImgInfo: layer, resolution, position,
 *				palette.
 *				Resulting xImgHandle will be returned via
 *				this structure also.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifCopyImage(
	     XImg *srcImg,
	     XImg *parentImg,
	     XPage *dstPage,
	     XIF_IMGINFO *pOutImgInfo
)
{
    Instance *instData;
    UInt32 row, rowsPerStrip;
    XImg *outImg = NULL;
    XIF_IMGINFO *inImgInfo;
    XIF_COMPRESSION compressionType;
    UInt32 quality = 0UL;
    UInt8 *cmpBuffer = NULL;
    UInt32 cmpBufsize = 0UL;
    UInt32 cmpBytes = 0UL;
    UInt32 bytesFilled;
    UInt32 *dictIDs;
    PROLOGUE;

    /* Reference the image information */
    inImgInfo = &srcImg->imgInfo;
    instData = XIF_IMGINST(srcImg);

    /* Begin the image read and write processes */
    XERR_PASS_CLEANUP(XIF_ImageReadStart(inImgInfo->xImgHandle, 0UL));
    if (inImgInfo->xImgType == XIF_IMGTYPE_MASK) {
	XERR_GEN_CLEANUP(parentImg == NULL, XIF_BADPARAMETER);
	XERR_PASS_CLEANUP(XIF_AddMaskImage((XIF_IMGHANDLE)parentImg,
	   inImgInfo->dwBestStripHeight, (XIF_IMGHANDLE *)&outImg));
    } else {
	XERR_PASS_CLEANUP(XIF_AddImage((XIF_PAGEHANDLE)dstPage,
	    pOutImgInfo->dwImgLayer, inImgInfo->xColorType,
	    pOutImgInfo->dwXOffset, pOutImgInfo->dwYOffset,
	    pOutImgInfo->dwXResolution, pOutImgInfo->dwYResolution,
	    inImgInfo->dwWidth, inImgInfo->dwHeight,
	    inImgInfo->dwBestStripHeight, pOutImgInfo->pPalette,
	    (XIF_IMGHANDLE *)&outImg));
    }
    compressionType = inImgInfo->xCompression;
    XERR_PASS_CLEANUP(XIF_ImageWriteStart((XIF_IMGHANDLE)outImg,
	XIF_IMGFLAGS_Compressed, &compressionType, &quality));

    /* Loop through the image strips, copying them compressed one by one */
    rowsPerStrip = inImgInfo->dwBestStripHeight;
    for (row = 0; row < inImgInfo->dwHeight; row += rowsPerStrip) {
	if (row + rowsPerStrip > inImgInfo->dwHeight) {
	    rowsPerStrip = inImgInfo->dwHeight - row;
	}
	XERR_PASS_CLEANUP(XIF_GetCompressedStripSize(inImgInfo->xImgHandle,
	    row, &cmpBytes));
	if (cmpBufsize < cmpBytes) {
	    if (cmpBufsize && cmpBuffer) {
		XIF_FREE(cmpBuffer);
	    }
	    CHECK_MALLOC(cmpBuffer = (UInt8 *)XIF_MALLOC(cmpBytes));
	    cmpBufsize = cmpBytes;
	}

	/* Read the compressed strip */
	XERR_PASS_CLEANUP(XIF_CompressedReadStrip(inImgInfo->xImgHandle,
	    row, rowsPerStrip, &cmpBuffer, &cmpBufsize, &bytesFilled));
	cmpBytes = bytesFilled;

        /* If it's a tokenized image, copy the token dictionary first */
	if (row == 0) {
	    switch ((int)inImgInfo->xCompression) {
	    case XIF_TOKENIZATION:
	    case XIF_LLTOKENIZATION:
		XERR_PASS_CLEANUP(wp_XifCopyTokenDictionary(srcImg,
		    cmpBuffer, cmpBytes, dstPage->xDoc, &dictIDs));
		break;
	    default:
		break;
	    }
	}

	/* Write the compressed strip */
	XERR_PASS_CLEANUP(XIF_CompressedWriteStrip((XIF_IMGHANDLE)outImg,
	    row, rowsPerStrip, cmpBuffer, cmpBytes, &bytesFilled));
	XERR_GEN_CLEANUP(bytesFilled < cmpBytes, XIF_INTERNAL_ERROR);
    }

    /* Deallocate compressed data buffer */
    XIF_FREE(cmpBuffer);
    cmpBuffer = NULL, cmpBufsize = 0UL;

    /* Finish up the read and write processes */
    XERR_PASS_CLEANUP(XIF_ImageReadFinish(inImgInfo->xImgHandle));
    XERR_PASS_CLEANUP(XIF_ImageWriteFinish((XIF_IMGHANDLE)outImg));

    /* Return the resulting image info */
    memcpy(pOutImgInfo, &outImg->imgInfo, sizeof(*pOutImgInfo));
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (cmpBuffer) {
	XIF_FREE(cmpBuffer);
	cmpBuffer = NULL, cmpBufsize = 0UL;
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation::
 *
 * OneLineDescription:::
 *  Copies the entire contents of a page from one XIFF to another
 *
 * LongDescription:::
 *  wp_XifCopyPage() copies a specified page from one XIFF document to
 *  another. The page can be an arbitrary page in the source document,
 *  but must be the next available page in the destination document
 *  (no page insertions).  The image data is left compressed during the
 *  copy, enabling fast copies.  Version differences are accounted for
 *  by decompressing and recompressing (using the default compression
 *  method).
 *  <BR>
 *  This function can be used to facilitate page insertion, deletion and
 *  reordering; but you must place it in a loop with the appropriate page
 *  offsets to sequentially specify the new destination page order.
 *
 * Input:::
 *  xSrcDocHandle::		Source document handle, returned from
 *				XIF_OpenDocumentRead() or
 *				XIF_OpenDocumentWrite()
 *  xDstDocHandle::		Destination document handle, returned from
 *				XIF_CreateDocument() or
 *				XIF_OpenDocumentWrite()
 *  dwSrcPageOffset::		Source document offset of page to copy
 *  dwDstPageOffset::		Destination document offset of page to add
 *				(this should be the end of the document)
 *  copyAnnotations::		Boolean for copy the page annotations
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifCopyPage(
	XDoc *xSrcDoc,
	Int32 dwSrcPageOffset,
	XDoc *xDstDoc,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
)
{
    Instance *instData;
    UInt32 imgNum;
    UInt16 annotNum;
    XIF_DOCHANDLE inXDoc = 0, outXDoc = 0;
    XIF_PAGEHANDLE inXPage, outXPage;
    XIF_IMGHANDLE primImg, outXImg;
    XImg *xImg;
    XIF_PAGEINFO inPageInfo;
    XIF_IMGINFO inImgInfo, inLayerInfo, outImgInfo;
    XIF_IMGINFO inMaskImgInfo, outMaskImgInfo;
    XIF_ANNOTINFO annot;
    UInt32 layer;
    UInt8 *annotData = NULL;
    UInt16 annotID;
    IPWState *pIPWState = NULL;
    PIXR *pPixr = NULL;
    UInt8 *imageChannels[3];
    Int32 bpl;
    UInt32 rowsFilled, rowsWritten;
    ScaledInt1616 xres, yres;
    ColorType ctype = cColorNone;
    Int32 channel, channels;
    PIXR *pSubPixr;
    Bool recompress;
    XIF_COMPRESSION compression;
    UInt32 quality;
    PROLOGUE;

    CheckHandleValidity(xSrcDoc);
    CheckHandleValidity(xDstDoc);

    /* Check for proper calling sequence */
    XERR_GEN((xSrcDoc->seqStatus & XDOC_SEQ_GETINFO) == 0,
	XIF_SEQUENCE);
    
    /* Reference the handles/data */
    inXDoc = (XIF_DOCHANDLE)xSrcDoc;
    outXDoc = (XIF_DOCHANDLE)xDstDoc;
    instData = XIF_DOCINST(xSrcDoc);

    /* Make sure that any unwritten pages for the input document are flushed,
     * before we get the image counts for the layers */
    if ((xSrcDoc->readWrite & WRITE_ALLOWED) &&
	xSrcDoc->seqStatus & (XDOC_SEQ_TOKENSTART)) {
	XERR_PASS_CLEANUP(XIF_WriteTokenImages((XIF_DOCHANDLE)xSrcDoc));
    }

    /* Make sure that any unwritten pages for the output document are flushed,
     * before we potentially start a new token dictionary */
    if ((xDstDoc->readWrite & WRITE_ALLOWED) &&
	xDstDoc->seqStatus & (XDOC_SEQ_TOKENSTART)) {
	XERR_PASS_CLEANUP(XIF_WriteTokenImages((XIF_DOCHANDLE)xDstDoc));
    }

    /* Get the page info */
    inPageInfo.dwPageOffset = dwSrcPageOffset;
    XERR_PASS_CLEANUP(XIF_GetPageInfo(inXDoc, dwSrcPageOffset,
	&inPageInfo.pdwImageCounts, &inPageInfo.dwLayersPresent,
	&inPageInfo.wAnnotationCount, &inXPage));

    /* Begin addition of a new page */
    XERR_PASS_CLEANUP(XIF_AddPageStart(outXDoc,
	dwDstPageOffset, &outXPage));
	    
    /* Start with some basic assumptions about all images */
    memset(&inImgInfo, 0, sizeof(inImgInfo));
    memset(&inMaskImgInfo, 0, sizeof(inMaskImgInfo));
    inMaskImgInfo.xImgType = XIF_IMGTYPE_MASK;
    inMaskImgInfo.xColorType = XIF_COLORTYPE_BINARY;
	    
    /* Get the Primary image info */
    inImgInfo.xImgType = XIF_IMGTYPE_PRIMARY;
    XERR_PASS_CLEANUP(XIF_GetImageInfo(inXPage,
	inImgInfo.dwImgLayer = 2,
	inImgInfo.dwImgOffset = 0UL,
	&inImgInfo.dwXOffset, &inImgInfo.dwYOffset,
	&inImgInfo.dwXResolution, &inImgInfo.dwYResolution,
	&inImgInfo.dwWidth, &inImgInfo.dwHeight,
	&inImgInfo.dwBestStripHeight, &inImgInfo.xCompression,
	&inImgInfo.xColorType, (UInt8 *)&inImgInfo.pPalette[0],
	&inImgInfo.dwMaskImages, &inImgInfo.xImgHandle));
    
    /* Borrow the image characteristics for the output image */
    memcpy(&outImgInfo, &inImgInfo, sizeof(outImgInfo));
    primImg = inImgInfo.xImgHandle;
    xImg = (XImg *)primImg;

    /* Copy the primary image first (always) */
    recompress = cFalse;
    layer = 2;
    if (((XPage *)outXPage)->xDoc->docInfo.version < 3) {

	/* For XIF 2 files, if there is more than one layer 2 image,
	 * or if XIF 2 cannot support the compression type, decompress
	 * and store off in default compression type */
	if (inPageInfo.pdwImageCounts[layer - 1] > 1) {
	    recompress = cTrue;
	} else {
	    switch (inImgInfo.xCompression) {
	    case XIF_NOC:
	    case XIF_G4:
	    case XIF_JPEG:
		break;
	    default:
		recompress = cTrue;
	    }
	}

	/* If the strip height will change, reform the strips */
	if (inImgInfo.dwBestStripHeight != outImgInfo.dwBestStripHeight) {
	    recompress = cTrue;
	}

	/* If PhotometricInterpretation is not right, force fixing it */
	switch (inImgInfo.xColorType) {
	case XIF_COLORTYPE_BINARY:
	    if (xImg->fields.PhotometricInterpretation !=
		XTIFF_PHOTOMETRIC_WHITEISZERO) {
		recompress = cTrue;
	    }
	    break;
	case XIF_COLORTYPE_GRAY4:
	case XIF_COLORTYPE_GRAY8:
	    if (xImg->fields.PhotometricInterpretation !=
		XTIFF_PHOTOMETRIC_BLACKISZERO) {
		recompress = cTrue;
	    }
	    break;
	case XIF_COLORTYPE_COLOR4:
	case XIF_COLORTYPE_COLOR8:
	    if (xImg->fields.PhotometricInterpretation !=
		XTIFF_PHOTOMETRIC_PALETTE) {
		recompress = cTrue;
	    }
	    break;
	case XIF_COLORTYPE_COLOR24:
	    if (xImg->fields.PhotometricInterpretation !=
		XTIFF_PHOTOMETRIC_WHITEISZERO) {
		recompress = cTrue;
	    }
	    break;
	}
    }
	
    /* Copy the image */
    if (recompress) {

	/* This image is always the entire page, based on the primary IFD.
	 * Read the entire layer. */
	XERR_PASS_CLEANUP(XIF_GetLayerInfo(inXPage,
	    inLayerInfo.dwImgLayer = layer, &inLayerInfo.xCompression,
	    &inLayerInfo.dwXOffset, &inLayerInfo.dwYOffset,
	    &inLayerInfo.dwXResolution, &inLayerInfo.dwYResolution,
	    &inLayerInfo.dwWidth, &inLayerInfo.dwHeight,
	    &inLayerInfo.xImgHandle));

	inLayerInfo.dwImgOffset = 0;
	inLayerInfo.xColorType = inImgInfo.xColorType;
	inLayerInfo.dwMaskImages = 0UL;

	/* Will be default, set elsewhere */
	inLayerInfo.dwBestStripHeight = 0;

	/* Borrow the image characteristics for the output image */
	memcpy(&outImgInfo, &inLayerInfo, sizeof(outImgInfo));
	outImgInfo.xImgType = XIF_IMGTYPE_PRIMARY;

	/* Access the common IPWState */
	XERR_PASS_CLEANUP(wp_XifIPWState(((XPage *)outXPage)->xDoc,
					 &pIPWState));

	/* Create a PIXR image to hold the image */
	xres = Int32_TO_ScaledInt1616(inLayerInfo.dwXResolution);
	yres = Int32_TO_ScaledInt1616(inLayerInfo.dwYResolution);
	switch (inLayerInfo.xColorType) {
	case XIF_COLORTYPE_COLOR24:
	    ctype = cColorRGB;
	    break;
	case XIF_COLORTYPE_BINARY:
	case XIF_COLORTYPE_GRAY4:
	case XIF_COLORTYPE_COLOR4:
	case XIF_COLORTYPE_GRAY8:
	case XIF_COLORTYPE_COLOR8:
	    ctype = cColorGray;
	    break;
	default:
	    break;
	}
	channels = XIF_CHANNELS(inLayerInfo.xColorType);
	XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, outImgInfo.dwWidth,
	    outImgInfo.dwHeight, XIF_BPPE(outImgInfo.xColorType),
	    xres, yres, 32, channels, ctype, NULL, &pPixr));
	
	/* Get some characteristics of the PIXR, retain the image data
	 * location for each channel(s) */
	XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));
	for (channel = 0; channel < channels; channel++) {
	    if (channels > 1) {
		pSubPixr = NULL;
		XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
		    &pSubPixr));
		XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
		    &imageChannels[channel]));
	    } else {
		XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
		    &imageChannels[channel]));
	    }
	}
	
	/* Read the entire layer */
	rowsFilled = 0;
	XERR_PASS_CLEANUP(w_XifImageReadLayer(pIPWState,
	    inLayerInfo.xImgHandle, pPixr, &rowsFilled));

	XERR_PASS_CLEANUP(XIF_AddImage(outXPage, outImgInfo.dwImgLayer,
	    outImgInfo.xColorType, outImgInfo.dwXOffset,
	    outImgInfo.dwYOffset, outImgInfo.dwXResolution,
	    outImgInfo.dwYResolution, outImgInfo.dwWidth,
	    outImgInfo.dwHeight, outImgInfo.dwBestStripHeight,
	    outImgInfo.pPalette, &outImgInfo.xImgHandle));
	outXImg = outImgInfo.xImgHandle;

	compression = XIF_COMP_UNKNOWN;	    /* Let it pick up the default */
	quality = 0;			    /* Let it pick up the default */

	XERR_PASS_CLEANUP(XIF_ImageWriteStart(outXImg, 0, &compression,
					      &quality));

	/* Write the entire image (as one strip). Let this function parcel it
	 * into strips. */
	XERR_PASS_CLEANUP(w_XifWriteStrip(pIPWState, outXImg, pPixr, 0,
    		rowsFilled, 0, &rowsWritten));

	/* Finish the write */
	XERR_PASS_CLEANUP(XIF_ImageWriteFinish(outXImg));

	XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pPixr));
    } else {	/* Not recompressing */
	XERR_PASS_CLEANUP(wp_XifCopyImage((XImg *)inImgInfo.xImgHandle,
	    0UL, (XPage *)outXPage, &outImgInfo));
    } /* recompress */

    /* Loop through each layer present */
    for (layer = 1; layer <= inPageInfo.dwLayersPresent; layer++) {

	/* For XIF 2 output, skip FG, BG and all selector layers */
	if (((XPage *)outXPage)->xDoc->docInfo.version < 3) {
	    if (((layer & 1) == 0) || layer < 5) {
		continue;
	    }
	}

	/* Skip the primary image (already stored above) */
	if (layer == 2) {
	    imgNum = 1;
	} else {
	    imgNum = 0;
	}
	for ( ; imgNum < inPageInfo.pdwImageCounts[layer - 1]; imgNum++) {
	    
	    inImgInfo.xImgType = XIF_IMGTYPE_LAYER;
	    XERR_PASS_CLEANUP(XIF_GetImageInfo(inXPage,
		inImgInfo.dwImgLayer = layer,
		inImgInfo.dwImgOffset = imgNum,
		&inImgInfo.dwXOffset, &inImgInfo.dwYOffset,
		&inImgInfo.dwXResolution, &inImgInfo.dwYResolution,
		&inImgInfo.dwWidth, &inImgInfo.dwHeight,
		&inImgInfo.dwBestStripHeight, &inImgInfo.xCompression,
		&inImgInfo.xColorType, &inImgInfo.pPalette[0],
		&inImgInfo.dwMaskImages, &inImgInfo.xImgHandle));

#if 0
	    printImgInfo(2, &inImgInfo);
#endif

	    recompress = cFalse;
	    if (((XPage *)outXPage)->xDoc->docInfo.version < 3) {

		/* For XIF 2 files, if the file is stored in an unsupported
		 * compression type, decompress and store off in default
		 * compression type. */
		switch (inImgInfo.xCompression) {
		case XIF_NOC:
		case XIF_G4:
		case XIF_JPEG:
		    break;
		default:
		    recompress = cTrue;
		}

		/* If PhotometricInterpretation is not right, force fixing it
		 */
		xImg = (XImg *)inImgInfo.xImgHandle;
		switch (inImgInfo.xColorType) {
		case XIF_COLORTYPE_BINARY:
		    if (xImg->fields.PhotometricInterpretation !=
			XTIFF_PHOTOMETRIC_WHITEISZERO) {
			recompress = cTrue;
		    }
		    break;
		case XIF_COLORTYPE_GRAY4:
		case XIF_COLORTYPE_GRAY8:
		    if (xImg->fields.PhotometricInterpretation !=
			XTIFF_PHOTOMETRIC_BLACKISZERO) {
			recompress = cTrue;
		    }
		    break;
		case XIF_COLORTYPE_COLOR4:
		case XIF_COLORTYPE_COLOR8:
		    if (xImg->fields.PhotometricInterpretation !=
			XTIFF_PHOTOMETRIC_PALETTE) {
			recompress = cTrue;
		    }
		    break;
		case XIF_COLORTYPE_COLOR24:
		    break;
		}
	    }
	    
	    /* Borrow the image characteristics for the output image */
	    memcpy(&outImgInfo, &inImgInfo, sizeof(outImgInfo));

	    /* Test the validity of the parameters */
	    XERR_PASS_CLEANUP(XIF_CheckImageInfo(outXPage,
		&outImgInfo.dwImgLayer, &outImgInfo.dwImgOffset,
		&outImgInfo.xColorType,
		&outImgInfo.dwXOffset, &outImgInfo.dwYOffset,
		&outImgInfo.dwXResolution, &outImgInfo.dwYResolution,
		&outImgInfo.dwWidth, &outImgInfo.dwHeight,
		&outImgInfo.dwBestStripHeight, outImgInfo.pPalette,
		&outImgInfo.xCompression));

	    /* If the strip height will change, reform the strips */
	    if (inImgInfo.dwBestStripHeight != outImgInfo.dwBestStripHeight) {
		recompress = cTrue;
	    }

	    /* Copy the image */
	    if (recompress) {

		/* Will be default, set elsewhere */
		inImgInfo.dwBestStripHeight = 0;

		/* Access the common IPWState */
		XERR_PASS_CLEANUP(wp_XifIPWState(((XPage *)outXPage)->xDoc,
						 &pIPWState));

		/* Create a PIXR image to hold the image */
		xres = Int32_TO_ScaledInt1616(inImgInfo.dwXResolution);
		yres = Int32_TO_ScaledInt1616(inImgInfo.dwYResolution);
		switch (inImgInfo.xColorType) {
		case XIF_COLORTYPE_COLOR24:
		    ctype = cColorRGB;
		    break;
		case XIF_COLORTYPE_BINARY:
		case XIF_COLORTYPE_GRAY4:
		case XIF_COLORTYPE_COLOR4:
		case XIF_COLORTYPE_GRAY8:
		case XIF_COLORTYPE_COLOR8:
		    ctype = cColorGray;
		    break;
		default:
		    break;
		}
		channels = XIF_CHANNELS(inImgInfo.xColorType);
		XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState,
		    outImgInfo.dwWidth, outImgInfo.dwHeight,
		    XIF_BPPE(outImgInfo.xColorType), xres, yres,
		    32, channels, ctype, NULL, &pPixr));
	
		/* Get some characteristics of the PIXR, retain the image
		 * data location for each channel(s) */
		XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));
		for (channel = 0; channel < channels; channel++) {
		    if (channels > 1) {
			pSubPixr = NULL;
			XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr,
			    channel, &pSubPixr));
			XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
			    &imageChannels[channel]));
		    } else {
			XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
			    &imageChannels[channel]));
		    }
		}
	
		/* Read the entire image */
		XERR_PASS_CLEANUP(XIF_ImageReadStart(inImgInfo.xImgHandle, 0));

		rowsFilled = 0;
		XERR_PASS_CLEANUP(w_XifReadStrip(pIPWState,
		    inImgInfo.xImgHandle, 0, inImgInfo.dwHeight, pPixr,
		    0, &rowsFilled));
		
		XERR_PASS_CLEANUP(XIF_ImageReadFinish(inImgInfo.xImgHandle));

		/* Add the image to the output document */
		XERR_PASS_CLEANUP(XIF_AddImage(outXPage, outImgInfo.dwImgLayer,
		    outImgInfo.xColorType, outImgInfo.dwXOffset,
		    outImgInfo.dwYOffset, outImgInfo.dwXResolution,
		    outImgInfo.dwYResolution, outImgInfo.dwWidth,
		    outImgInfo.dwHeight, outImgInfo.dwBestStripHeight,
		    outImgInfo.pPalette, &outImgInfo.xImgHandle));
		outXImg = outImgInfo.xImgHandle;

		/* Let it pick up the defaults: */
		compression = XIF_COMP_UNKNOWN;
		quality = 0;

		XERR_PASS_CLEANUP(XIF_ImageWriteStart(outXImg, 0, &compression,
						      &quality));

		/* Write the entire image (as one strip). Let this function
		 * parcel it into strips. */
		XERR_PASS_CLEANUP(w_XifWriteStrip(pIPWState, outXImg, pPixr, 0,
		    rowsFilled, 0, &rowsWritten));

		/* Finish the write */
		XERR_PASS_CLEANUP(XIF_ImageWriteFinish(outXImg));

		XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pPixr));
	    } else {
		XERR_PASS_CLEANUP(wp_XifCopyImage((XImg *)inImgInfo.xImgHandle,
		    NULL, (XPage *)outXPage, &outImgInfo));
	    } /* recompress */
	    
	    if (inImgInfo.dwMaskImages) {
		XERR_PASS_CLEANUP(XIF_GetMaskImageInfo(inImgInfo.xImgHandle,
		    inMaskImgInfo.dwImgOffset = 0UL,
		    &inMaskImgInfo.dwXOffset, &inMaskImgInfo.dwYOffset,
		    &inMaskImgInfo.dwXResolution,
		    &inMaskImgInfo.dwYResolution,
		    &inMaskImgInfo.dwWidth, &inMaskImgInfo.dwHeight,
		    &inMaskImgInfo.dwBestStripHeight,
		    &inMaskImgInfo.xCompression,
		    &inMaskImgInfo.xImgHandle));

		/* Copy the image */
		XERR_PASS_CLEANUP(wp_XifCopyImage(
		    (XImg *)inMaskImgInfo.xImgHandle,
		    (XImg *)outImgInfo.xImgHandle, (XPage *)outXPage,
		    &outMaskImgInfo));
	    }
	} /* For each image */
    } /* For each layer */

    if (copyAnnotations) {
	/* Loop through the annotations */
	for (annotNum = 0; annotNum < inPageInfo.wAnnotationCount;
	     annotNum++) {
	    XERR_PASS_CLEANUP(XIF_GetAnnotationInfo(inXPage, annotNum,
		&annot.wAnnotType, &annot.dwXOffset, &annot.dwYOffset,
		&annot.dwWidth, &annot.dwHeight, &annot.dwFlags,
		&annot.dwAnnotLength));
	
	    CHECK_MALLOC(annotData = (UInt8 *)XIF_MALLOC(annot.dwAnnotLength));
	    XERR_PASS_CLEANUP(XIF_GetAnnotationData(inXPage, annotNum,
		annot.dwAnnotLength, annotData));
	    XERR_PASS_CLEANUP(XIF_AddAnnotation(outXPage, annot.wAnnotType,
		annot.dwXOffset, annot.dwYOffset, annot.dwWidth,
		annot.dwHeight, annot.dwFlags, annotData, annot.dwAnnotLength,
		&annotID));
	    XIF_FREE(annotData), annotData = NULL;
	} /* For each annotation */
    } /* If copying annotations */
    
    XERR_PASS_CLEANUP(XIF_AddPageFinish(outXPage));
    
    NORMAL_RETURN;
	
ERR_CLEANUP:
    if (annotData) {
	XIF_FREE(annotData);
	annotData = NULL;
    }
    if (pPixr) {
	w_destroyPixr(pIPWState, &pPixr);
    }

    ERR_RETURN;
}

/*
$Log: xifwpub.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.28  1998/07/30 22:39:44  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.27  1998/03/25 23:25:15  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Use a unique document ID for each document handle. This allows tagging
each document so that during "quick copy", the token dictionaries are
copy only once, and only once (in the case of two destination documents
somehow ending up with the same memory location for the structure data).

Revision 1.26  1998/02/09 20:20:16  aber
aber on Mon Feb 9 12:17:48 PST 1998

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Flush token dictionary for both source and destination when preparing
to copy a page, instead of just the source. This was causing the destination
token dictionary to get trampled on, in the case where the destination
file had both new pages and copied "old" pages.

Revision 1.25  1998/01/20 03:22:40  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Before reading a tokenized image, make sure it's flushed to disk.

Revision 1.24  1998/01/16 00:41:38  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Allow page inserts.

Revision 1.23  1997/11/24 21:08:11  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Check strip height before deciding whether to recompress during
a page copy in compressed mode.

Revision 1.22  1997/11/20 02:54:05  aber
aber on Wed Nov 19 18:53:36 PST 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

If copying from and to images with different strip sizes, recompress.

Revision 1.21  1997/11/18 23:44:55  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Add new parameter for disabling annotation copy during page copy.
Add annotation flags argument to annotation calls.

Revision 1.20  1997/11/17 19:26:01  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Allow negative page offsets to automatically append the page to
the end of the document.

Revision 1.19  1997/11/14 23:04:42  aber
aber on Fri Nov 14 15:03:43 PST 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Fix bug in copy page, the layer wasn't being set correctly when
not recompressing some images.

Revision 1.18  1997/11/11 22:44:27  aber
aber on Tue Nov 11 14:43:41 PST 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Allow new image parameters to change when using XIF_CopyImage() in lower
level function wp_XifCopyImage().

Revision 1.17  1997/11/06 22:49:53  aber
aber on Thu Nov 6 14:49:48 PST 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Cosmetic changes.

Revision 1.16  1997/10/27 18:13:20  aber
aber on Mon Oct 27 10:12:18 PST 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Change handling of error check, to avoid condition compiler "warns" about.

Revision 1.15  1997/10/24 21:38:59  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.14  1997/10/16 19:04:22  aber
aber on Thu Oct 16 12:01:55 PDT 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Improve the functionality of the CopyPage function (quick copy). Now
the images are checked for consistency across version change copies.
If the output version has different support, the image is decompressed
and recompressed automatically. Also, the multiple images in layer 2
are recombined when the output is XIF 2.0.

Revision 1.13  1997/10/02 00:51:36  aber
aber on Wed Oct 1 17:50:33 PDT 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

If no token dictionaries referenced by the tokenization stream
(blank image), don't allocate a 0-byte array for the dictionary
references when copying in compressed mode.

Revision 1.12  1997/08/29 17:28:07  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.11  1997/08/15 23:48:02  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.10  1997/08/15 01:54:20  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Modify token dictionary header to include the Tunrda dictionary
byte count, allows for extensible dictionary data.

Revision 1.9  1997/08/15 00:10:46  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Use predefined pre-Tundra header offsets for dictionary copies.

Revision 1.8  1997/08/07 21:27:49  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.7  1997/07/10 22:01:44  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Use 16-bit value for the annotation ID.

Revision 1.6  1997/06/17 01:29:36  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: XIFF 2.0 support. Annotation bug fix.

Revision 1.5  1997/06/10 01:26:58  aber
xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.3  1997/05/30 22:36:52  aber
aber on Fri May 30 15:36:20 PDT 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Use ERR_PASS vs. ERR_PASS_CLEANUP to avoid possible use of uninitialized
variable.

Revision 1.2  1997/05/30 17:14:31  aber
aber on Fri May 30 10:14:13 PDT 1997

xifwpub.c -> /product/ipcore/ipwrap/src/RCS/xifwpub.c,v

Check variable ranges for validity, avoids Watcom complaint.

Revision 1.1  1997/05/30 00:51:16  aber
Initial revision

*/
