/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwrite.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifwrite.c:	XIFF WRITE source
 *
 *	XIFF write-only routines
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifwtiff.h"
#include "xifwxiff.h"
#include "xifwpriv.h"
#include "xifrxiff.h"
#include "xifwpub.h"
#include "xifrpriv.h"
#include "xifwtok.h"
#include "xifconv.h"
#include "xiff.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwrite.c,v $; $Revision: 1.2 $; $Date: 2000/04/07 18:56:15 $");


/* XIF_CreateDocument()
 *
 * OneLineDescription:::
 *  Create a new XIFF document
 *
 * LongDescription:::
 *  XIF_CreateDocument() creates a new XIFF document, and prepares to write
 *  data to it. The client must have already created the file at the operating
 *  system level, so that the read, write, and seek methods are available.
 *  <BR>
 *  A handle representing the document is passed back through a parameter.
 *  This handle is passed in to all subsequent XIF_  function calls, and will
 *  encompass the XIF_INSTHANDLE passed in here.
 *
 * Input:::
 *  xInstHandle::		XIF instance handle, returned from
 *				XIF_InitInstance()
 *  dwClientID::		Client ID, passed back through all I/O funcs
 *  version::			Version of XIFF document to create
 *  xReadFunc::			Client file read function
 *  xWriteFunc::		Client file write function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *
 * Output:::
 *  pxDocHandle::		Pointer to document handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_CreateDocument(
	XIF_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	UInt8 version,
	XIF_READFUNC xReadFunc,
	XIF_WRITEFUNC xWriteFunc,
	XIF_SEEKFUNC xSeekFunc,
	XIF_SIZEFUNC xSizeFunc,
	XIF_DOCHANDLE *pxDocHandle
)
{
    Instance *instData = (Instance *)xInstHandle;
    XDoc *xDoc = NULL;
    XFile *xFile = NULL;
    UInt8 revision;
    PROLOGUE;
    
    /* Clear the results for now */
    *pxDocHandle = 0UL;
    
    CheckHandleValidity(xInstHandle);
    
    /* Check for proper calling sequence */

    
    /* Create a new document (handle) */
    CHECK_MALLOC(xDoc = (XDoc *)XIF_MALLOC(sizeof(XDoc)));
    memset(xDoc, 0, sizeof(*xDoc));
    xDoc->handleID = (UInt32)xDoc;	/* Self identity */
    xDoc->docID = ++instData->nextDocID;/* ID unique to this library instance
					 */
    
    /* Create a link back to the instance */
    CHECK_MALLOC(xFile = (XFile *)XIF_MALLOC(sizeof(XFile)));
    memset(xFile, 0, sizeof(*xFile));
    xFile->handleID = (UInt32)xFile;	/* Self identity */
    
    /* Setup the file handle */
    xFile->instance = instData;
    XIF_FILETOKEN(xFile)->dwClientFileID = dwClientID;
    XIF_FILETOKEN(xFile)->FileRead = xReadFunc;
    XIF_FILETOKEN(xFile)->FileWrite = xWriteFunc;
    XIF_FILETOKEN(xFile)->FileSeek = xSeekFunc;
    XIF_FILETOKEN(xFile)->FileSize = xSizeFunc;
    
    /* Retain the file handle */
    xDoc->xFile = xFile;
    
    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_XifFileLock(XIF_FILETOKEN(xFile)));
    
    /* Create the XIFF header */
    switch (version) {
    case 2:
	revision = 0;
	break;
    default:
	version = CURRENT_XIFF_VERSION;
	revision = CURRENT_XIFF_REVISION;
	break;
    }
    XERR_PASS_CLEANUP(wp_XifWriteHeader(xDoc, &version, &revision));

    /* Initialize */
    xDoc->seqStatus |= XDOC_SEQ_WRITESTART;
    xDoc->readWrite = WRITE_ALLOWED;
    xDoc->docInfo.version = version;
    xDoc->docInfo.revision = revision;
    
    /* We could store a page table now with the following call:
     *     XERR_PASS_CLEANUP(wp_XifSetPageIFD(xDoc, 0UL, 0, 0UL, NULL));
     * but instead the page table is added when the first page is written,
     * or when the document is closed (in the case of XIF 3 only, with 0 pages).
     * The reason it's not done here, is that XIF 3+ files will not have
     * a page table attached when no pages exist (more compact), and the
     * number of page table entries per table is allowed to be set up until
     * the first page is added (see XIF_SetPageTableEntryCount()).
     */
    
    /* Return the result */
    *pxDocHandle = (UInt32)xDoc;
    
    XERR_PASS_CLEANUP(wp_XifFileUnLock(XIF_FILETOKEN(xFile)));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (xDoc) {
	XIF_FREE(xDoc);
    }
    if (xFile) {
	XIF_FREE(xFile);
    }
    wp_XifFileUnLock(XIF_FILETOKEN(xFile));

     ERR_RETURN;
}


/* XIF_OpenDocumentReadWrite()
 *
 * OneLineDescription:::
 *  Open a document for read and write
 *
 * LongDescription:::
 *  XIF_OpenDocumentReadWrite() begins a session of reading and writing
 *  an existing document. The client must already have opened the file at
 *  the operating system level, so that the read and seek methods in the
 *  file token are available.
 *  This function will perform any initialization required for reading
 *  the file, including reading enough header information to verify
 *  that the file is in a format supported by this version of
 *  the XIF library.
 *  <BR>
 *  A document handle is returned through the third parameter. This handle
 *  is passed to all subsequent function calls that use the document, and
 *  will encompass the I/O functions passed in here.
 *
 * Input:::
 *  xInstHandle::		XIF instance handle, returned from
 *				XIF_InitInstance()
 *  dwClientID::		Input client ID
 *  xReadFunc::			Client file read function
 *  xWriteFunc::		Client file write function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *
 * Output:::
 *  pxDocHandle::		Document handle used to refer to this
 *				page in other functions
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_OpenDocumentReadWrite(
	XIF_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	UInt8 version,
	XIF_READFUNC xReadFunc,
	XIF_WRITEFUNC xWriteFunc,
	XIF_SEEKFUNC xSeekFunc,
	XIF_SIZEFUNC xSizeFunc,
	XIF_DOCHANDLE *pxDocHandle
)
{
    Instance *instData = (Instance *)xInstHandle;
    XDoc *xDoc;
    XFile xFile;
    UInt8 revision;
    UInt16 nPages;
    Int32 refPage;
    UInt16 authorBytes;
    const Int8 *authorBlock;
    Int32 fileSize = 0;
    PROLOGUE;

    /* Clear results for now */
    *pxDocHandle = 0UL;

    CheckHandleValidity(xInstHandle);

    /* Create a link back to the instance */
    memset(&xFile, 0, sizeof(xFile));
    xFile.handleID = (UInt32)&xFile;	/* Self identity */

    /* Setup the file handle */
    xFile.instance = instData;
    XIF_FILETOKEN(&xFile)->dwClientFileID = dwClientID;
    XIF_FILETOKEN(&xFile)->FileRead = xReadFunc;
    XIF_FILETOKEN(&xFile)->FileWrite = xWriteFunc;
    XIF_FILETOKEN(&xFile)->FileSeek = xSeekFunc;
    XIF_FILETOKEN(&xFile)->FileSize = xSizeFunc;
    
    /* Check the file size, if it's zero, create the document */
    XIF_FILESIZE(&xFile, fileSize);
    if (fileSize == 0) {
	XERR_PASS_CLEANUP(XIF_CreateDocument(xInstHandle, dwClientID,
	    version, xReadFunc, xWriteFunc, xSeekFunc, xSizeFunc, pxDocHandle));
    } else {

	/* Open the document read-only */
	XERR_PASS_CLEANUP(XIF_OpenDocumentRead(xInstHandle, dwClientID,
	    xReadFunc, xWriteFunc, xSeekFunc, xSizeFunc, pxDocHandle));
    }

    /* Reference the document */
    xDoc = (XDoc *)*pxDocHandle;

    /* Change the status from read to read/write */
    xDoc->readWrite |= READ_ALLOWED | WRITE_ALLOWED;
    xDoc->seqStatus |= XDOC_SEQ_READSTART | XDOC_SEQ_WRITESTART;

    /* Get the document information */
    XERR_PASS_CLEANUP(XIF_GetDocInfo(*pxDocHandle, &version, &revision,
	&nPages, &authorBlock, &authorBytes, &refPage));

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}



/* XIF_CopyImage()
 *
 * OneLineDescription:::
 *  Copy an individual image from one XIF to another (leave compressed)
 *
 * LongDescription:::
 *  XIF_CopyImage() copies an individual image from one XIF to another,
 *  leaving the image data compressed during the copy.  This allows a much
 *  faster copy, since the data is not decompressed and recompressed.  It also
 *  means that the use of a lossy algorithm (like JPEG) will not produce added
 *  lossy affects by recompressing with additional or different loss.
 *  <BR>
 *  Note: this function will NOT keep track of version inconsistencies (like
 *  the fact that a XIFF 2.0 does not support tokenization).
 *
 * Input:::
 *  xSrcImgHandle::		Source image handle, returned from
 *				XIF_GetImageInfo()
 *  xParentImgHandle::		Image handle for parent image (if source image
 *				is a transparency mask)
 *  xDstPageHandle::		Destination page handle, returned from
 *				XIF_AddPageStart() or XIF_GetPageInfo()
 *  dwImgLayer::		Image layer in MRC
 *  dwXOffset::			X pixel offset in Primage image coordinates
 *  dwYOffset::			Y pixel offset in Primage image coordinates
 *  dwBestStripHeight::		Height of strips stored
 *
 * Output:::
 *  pxDstImgHandle::		Pointer to resulting image handle, can be used
 *				to later reference this image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_CopyImage(
	XIF_IMGHANDLE xSrcImgHandle,
	XIF_IMGHANDLE xParentImgHandle,
	XIF_PAGEHANDLE xDstPageHandle,
	UInt32 dwImgLayer,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwBestStripHeight,
	XIF_IMGHANDLE *pxDstImgHandle
)
{
    XImg *xSrcImg, *xParentImg = NULL;
    XPage *xDstPage;
    XDoc *xSrcDoc, *xDstDoc;
    XIF_IMGINFO *pxSrcImgInfo, xDstImgInfo;
    UInt32 *pDstPageImageCounts = NULL, dstPageLayersPresent = 0;
    UInt16 wAnnotationCount;
    XIF_PAGEHANDLE xOutPage;
    PROLOGUE;

    CheckHandleValidity(xSrcImgHandle);
    CheckHandleValidity(xDstPageHandle);

    xSrcImg = (XImg *)xSrcImgHandle;
    xDstPage = (XPage *)xDstPageHandle;
    pxSrcImgInfo = &xSrcImg->imgInfo;

    /* If image is a transparency mask, check the parent handle */
    if (pxSrcImgInfo->xImgType == XIF_IMGTYPE_MASK) {
	CheckHandleValidity(xParentImgHandle);
	xParentImg = (XImg *)xParentImgHandle;
    }

    /* Access the documents */
    xSrcDoc = xSrcImg->xPage->xDoc;
    xDstDoc = xDstPage->xDoc;

    /* Get (make sure they're fresh) the page info */
    if (xDstDoc->pageTableHeader.pageCount > xDstPage->pageInfo.dwPageOffset) {
	XERR_PASS_CLEANUP(XIF_GetPageInfo((XIF_DOCHANDLE)xDstDoc,
	    xDstPage->pageInfo.dwPageOffset,
	    &pDstPageImageCounts, &dstPageLayersPresent,
	    &wAnnotationCount, &xOutPage));
    }

    /* Make sure the compression types jive */
    if (xDstDoc->docInfo.version < 3) {
	switch (pxSrcImgInfo->xCompression) {
	case XIF_TOKENIZATION:
	case XIF_LLTOKENIZATION:
	case XIF_LOSSYDITHER:
	    XERR_GEN_CLEANUP(1, XIF_NOSUPPORT);
	}
    }

    /* Create the resulting image info structure */
    memcpy(&xDstImgInfo, pxSrcImgInfo, sizeof(xDstImgInfo));

    /* These items will be different */
    xDstImgInfo.dwImgLayer = dwImgLayer;
    xDstImgInfo.dwXOffset = dwXOffset;
    xDstImgInfo.dwYOffset = dwYOffset;
    xDstImgInfo.dwBestStripHeight = dwBestStripHeight;
    xDstImgInfo.dwMaskImages = 0;
    xDstImgInfo.dwImgOffset = 0;
    if (pxSrcImgInfo->xImgType != XIF_IMGTYPE_MASK) {
	switch (dwImgLayer) {
	case 2:
	    if (dstPageLayersPresent == 0) {
		xDstImgInfo.xImgType = XIF_IMGTYPE_PRIMARY;
	    } else {
		xDstImgInfo.xImgType = XIF_IMGTYPE_LAYER;
	    }
	    break;
	default:
	    xDstImgInfo.xImgType = XIF_IMGTYPE_LAYER;
	}
	if (dstPageLayersPresent > 0 && dstPageLayersPresent >= dwImgLayer) {
	    xDstImgInfo.dwImgOffset = pDstPageImageCounts[dwImgLayer - 1];
	}
    }

    /* Call lower level routine to do the copy */
    XERR_PASS_CLEANUP(wp_XifCopyImage(xSrcImg, xParentImg, xDstPage,
	&xDstImgInfo));

    /* Return the new image handle */
    *pxDstImgHandle = (XIF_IMGHANDLE)xDstImgInfo.xImgHandle;

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/* XIF_CopyPage()
 *
 * OneLineDescription:::
 *  Copies the entire contents of a page from one XIFF to another
 *
 * LongDescription:::
 *  XIF_CopyPage() copies a specified page from one XIFF document to
 *  another. The page can be an arbitrary page in the source document,
 *  but must be the next available page in the destination document
 *  (no page insertions).  The image data is left compressed during the
 *  copy, enabling fast copies.  Version differences are NOT fixed in
 *  this copy. If the compression stream of the source image data is of
 *  an outdated format, the stream is not updated.
 *  <BR>
 *  This function can be used to facilitate page insertion, deletion and
 *  reordering; but you must place it in a loop with the appropriate page
 *  offsets to sequentially specify the new destination page order.
 *
 * Input:::
 *  xSrcDocHandle::		Source document handle, returned from
 *				XIF_OpenDocumentRead() or
 *				XIF_OpenDocumentWrite()
 *  xDstDocHandle::		Destination document handle, returned from
 *				XIF_CreateDocument() or
 *				XIF_OpenDocumentWrite()
 *  dwSrcPageOffset::		Source document offset of page to copy
 *  dwDstPageOffset::		Destination document offset of page to add
 *				(this should be the end of the document).
 *				-1 indicates the next available page.
 *  copyAnnotations::		Boolean for copy the page annotations
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_CopyPage(
	XIF_DOCHANDLE xSrcDocHandle,
	XIF_DOCHANDLE xDstDocHandle,
	Int32 dwSrcPageOffset,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
)
{
    XDoc *xSrcDoc, *xDstDoc;
    PROLOGUE;

    CheckHandleValidity(xSrcDocHandle);
    CheckHandleValidity(xDstDocHandle);

    xSrcDoc = (XDoc *)xSrcDocHandle;
    xDstDoc = (XDoc *)xDstDocHandle;

    /* Allow negative number to indicate the next available page */
    if (dwDstPageOffset < 0) {
	dwDstPageOffset = xDstDoc->pageTableHeader.pageCount;
    }

    XERR_PASS_CLEANUP(wp_XifCopyPage(xSrcDoc, dwSrcPageOffset,
	xDstDoc, dwDstPageOffset, copyAnnotations));

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/* XIF_AddPageStart()
 *
 * OneLineDescription:::
 *  Begin the process of adding a new page
 *
 * LongDescription:::
 *  XIF_AddPageStart() adds a new page to a multipage document. Page offsets
 *  start at zero, but setting dwPageNumber < 0 causes the new page to be
 *  added at the end of the document. The operation will fail if a page
 *  is added beyond the next available page at the end of the document, i.e.
 *  {0, 1, 2, 7} is not a valid page sequence, and the attempt to add page
 *  offset 7 would fail.
 *  <BR>
 *  Note that when adding pages A, B, and C, each with offset = 0, will result
 *  in the pages having a reverse page order -- the final page offsets will be
 *  2, 1, and 0.
 *  <BR>
 *  The page handle returned from this routine is used in later calls to
 *  attach images and annotations to this page.
 *  <BR>
 *  XIF_AddPageStart() may be followed by adding a Primary image using
 *  XIF_AddImage(). 
 *
 * Input:::
 *  xDocHandle::		Document handle
 *  dwPageOffset::		Offset of page to add, should be up to the
 *				existing page count, or negative, to indicate
 *				the end of the document
 *  pxPageHandle::		Page handle returned
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_AddPageStart(
	XIF_DOCHANDLE xDocHandle,
	Int32 dwPageOffset,
	XIF_PAGEHANDLE *pxPageHandle
)
{
    Instance *instData;
    XFile *xFile;
    XDoc *xDoc = (XDoc *)xDocHandle;
    XPage *xPage = NULL;
    XPage *searchPage = NULL, *prevPage;
    UInt8 flags = 0;
    Int32 pageOffset;
    Bool inserting = cFalse;
    PageTable *pt;
    Int32 tableOffset;
    PROLOGUE;
    
    CheckHandleValidity(xDocHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN((xDoc->seqStatus & XDOC_SEQ_WRITESTART) == 0,
	XIF_SEQUENCE);

    /* Allow negative number to indicate the next available page */
    if (dwPageOffset < 0) {
	dwPageOffset = xDoc->pageTableHeader.pageCount;
    }

    /* Pages must be inserted, or appended */
    XERR_GEN(dwPageOffset > xDoc->pageTableHeader.pageCount,
	XIF_BADPARAMETER);
    
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);

    if ((xDoc->readWrite & READ_ALLOWED) &&
	xDoc->pageTableHeader.entriesPerTable > 0) {

	/* Search *PAST* the last page table that is stored in memory */
	pt = xDoc->firstPageTable;
	tableOffset = 0;
	if (pt != NULL) {
	    do {
		/* Pages tables should ALWAYS be sequential */
		XERR_GEN_CLEANUP(tableOffset + 1 != pt->ordinal,
		    XIF_INTERNAL_ERROR);
		
		tableOffset++;
	    } while ((pt = pt->next) != NULL);
	}

	/* If we haven't yet read the last page table entry, we must
	 * read it now, since we'll either add to the last table, add
	 * a new table, or shift the tables (in the case of an insert)
	 */
	if (tableOffset <= xDoc->pageTableHeader.pageCount /
	    xDoc->pageTableHeader.entriesPerTable) {
	    XERR_PASS_CLEANUP(wp_XifReadPageTable(xDoc, 0UL, NULL));
	}

#if 1	/* It's unclear that this is needed, and it will cause a lot of
	 * time to be spent reading through all the IFDs */
	/* If we know we're inserting, make sure we have the page info
	 * for all the pages, since we'll be shifting their ordinals */
	if (dwPageOffset < xDoc->pageTableHeader.pageCount) {
	    Int32 pageNum;
	    UInt32 imageLayers;
	    UInt16 annotations;
	    XIF_PAGEHANDLE pageHandle;

	    for (pageNum = 0; pageNum < xDoc->pageTableHeader.pageCount; pageNum++) {
		XERR_PASS_CLEANUP(XIF_GetPageInfo((XIF_DOCHANDLE)xDoc,
		    (UInt32)pageNum, NULL, &imageLayers, &annotations,
		    &pageHandle));
	    }
	}
#endif
	inserting = cTrue;
    }
    
    /* Look for a previous read/create of this page */
    searchPage = xDoc->pages;
    pageOffset = 0;
    if (searchPage != NULL) {
	do {
	    /* Found, this page already exists.  We'll be inserting this page */
	    if (searchPage->pageInfo.dwPageOffset == dwPageOffset) {
		inserting = cTrue;
		break;
	    }

	    pageOffset++;
	} while ((searchPage = searchPage->next) != NULL);
    }
    
    if (searchPage == NULL ||
	searchPage->pageInfo.dwPageOffset != dwPageOffset || inserting) {
	
	/* Begin construction of the page data */
	CHECK_MALLOC(xPage = (XPage *)XIF_MALLOC(sizeof(XPage)));
	memset(xPage, 0, sizeof(XPage));
	xPage->handleID = (UInt32)xPage;	/* self identity */
	xPage->xDoc = xDoc;
	xPage->pageInfo.xPageHandle = (UInt32)xPage;
	xPage->pageInfo.dwPageOffset = dwPageOffset;

	/* This can get tricky here: write a made-up offset for the last page.
	 * If this causes a new page table to be written, then the following
	 * call to FileLongAlign() won't return a bogus value once the call
	 * to XifSetPageIFD() that follows is called. */
	if ((xDoc->pageTableHeader.entriesPerTable == 0UL) ||
	    ((xDoc->docInfo.wNPages %
	    xDoc->pageTableHeader.entriesPerTable == 0UL) &&
	    (xDoc->docInfo.wNPages / xDoc->pageTableHeader.entriesPerTable >=
	    xDoc->pageTableHeader.lastOrdinal))) {
	    XERR_PASS_CLEANUP(wp_XifSetPageIFD(xDoc, xDoc->docInfo.wNPages,
		flags, 0UL, xPage));
	}

	/* Attach it to the appropriate spot in linked list */
	prevPage = NULL;
	searchPage = xDoc->pages;
	if (searchPage == NULL) {
	    xDoc->pages = xPage;
	} else {
	    do {
		if (searchPage->pageInfo.dwPageOffset >= dwPageOffset) {

		    /* Only = when inserting */
		    XERR_GEN_CLEANUP(inserting != cTrue &&
			searchPage->pageInfo.dwPageOffset == dwPageOffset,
			XIF_INTERNAL_ERROR);
	    
		    /* Found the right spot to insert this page before */
		    break;
		}
		prevPage = searchPage;
	    } while ((searchPage = searchPage->next) != NULL);
	    
	    /* Insert before searchPage */
	    xPage->prev = prevPage;
	    if (prevPage) {	/* There's a previous page */
		prevPage->next = xPage;
	    } else {
		xDoc->pages = xPage;
	    }
	    xPage->next = searchPage;
	    if (searchPage) {	/* There's a page which will follow */
		searchPage->prev = xPage;

		/* Update the page offset of all those that follow */
		while (searchPage) {
		    searchPage->pageInfo.dwPageOffset++;
		    searchPage = searchPage->next;
		}
	    }
	}

	/* Update the page count */
	xDoc->docInfo.wNPages++;
    }

    /* If the page was inserted, rewrite the affected page table entries */
    if (inserting) {
	XERR_PASS_CLEANUP(wp_XifRewritePageTableEntries(xDoc, xPage));
    }

    /* Return the page handle */
    *pxPageHandle = (XIF_PAGEHANDLE)xPage;

    /* Update status */
    xPage->seqStatus |= XPAGE_SEQ_WRITESTART;

    xPage = NULL;
	
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (xPage) {
	XIF_FREE(xPage);
	xPage = NULL;
    }
    
    ERR_RETURN;
}


/* XIF_AddPageFinish()
 *
 * OneLineDescription:::
 *  Complete the construction of a page, tokenization still pending
 *
 * LongDescription:::
 *  XIF_AddPageFinish() completes the general construction of a page, although
 *  the completion of the tokenization process later may actually store the
 *  data for the Primary IFD, and may also add other subimages for Layer 2.
 *
 * Input:::
 *  xPageHandle::		Page handle of page to complete
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_AddPageFinish(
	XIF_PAGEHANDLE xPageHandle
)
{
    XPage *xPage = (XPage *)xPageHandle;
    PROLOGUE;

    CheckHandleValidity(xPageHandle);
    
    /* Check for proper calling sequence */

    /* We must have started to write this page */
    XERR_GEN_CLEANUP((xPage->seqStatus & XPAGE_SEQ_WRITESTART) == 0,
	XIF_SEQUENCE);

    /* Mark as completely written */
    if ((xPage->seqStatus & XPAGE_SEQ_TOKENSTART) == 0) {
	xPage->seqStatus &= ~XPAGE_SEQ_WRITESTATUS; /* Clear write status */
	xPage->seqStatus |= XPAGE_SEQ_WRITEFINISH;  /* Set to finished */
	xPage->seqStatus |= XPAGE_SEQ_READSTART;	/* Mark as readable */
    } else {
	/* Mark as finshed, but pending tokenization completion */
	xPage->seqStatus |= XPAGE_SEQ_WRITEPENDING;  /* Pending tokenization */
    }	

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/* XIF_DeletePage()
 *
 * OneLineDescription:::
 *  Delete an entire page from the document (including all images on the page)
 *
 * LongDescription:::
 *  XIF_DeletePage() deletes (removes all references to) an entire page in
 *  the file.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  page are removed.  This includes: page count in the page table header,
 *  removing the primary IFD from the page table (and shifting subsequent
 *  page references to accomodate the removal), and relinking the previous
 *  page to the subsequent page via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the page is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' a
 *  page deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see XIF_CopyPage() and
 *  XIF_CopyImage()), allowing the disowned file fragments to remain behind.
 *  <BR>
 *  Note also that the page offsets of all subsequent pages will automatically
 *  be shifted down one (previous page offsets will now be wrong, only for
 *  pages that followed this one).
 *
 * Input:::
 *  xPageHandle::		Page handle of the page to be removed, as
 *				returned from XIF_GetPageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_DeletePage(
	XIF_PAGEHANDLE xPageHandle
)
{
    XPage *xPage = (XPage *)xPageHandle;
    XDoc *xDoc;
    UInt8 flags;
    Int32 ifdOffset;
    XPage *nextPage;	/* Page left in deleted pages place */
    PROLOGUE;

    CheckHandleValidity(xPageHandle);
    xDoc = xPage->xDoc;

    /* Make sure we've read all the page tables (since the internal structures
     * are actually stored back out in the revised page table). */
    XERR_PASS_CLEANUP(wp_XifGetPageIFD(xDoc, xDoc->pageTableHeader.pageCount - 1,
	&flags, &ifdOffset));

#if 1
    /* It's unclear that this is needed, and it will cause a lot of
     * time to be spent reading through all the IFDs */
    /* If we know we're inserting, make sure we have the page info
     * for all the pages, since we'll be shifting their ordinals */
    if (xPage->pageInfo.dwPageOffset < xDoc->pageTableHeader.pageCount) {
	Int32 pageNum;
	UInt32 imageLayers;
	UInt16 annotations;
	XIF_PAGEHANDLE pageHandle;

	for (pageNum = 0; pageNum < xDoc->pageTableHeader.pageCount; pageNum++) {
	    XERR_PASS_CLEANUP(XIF_GetPageInfo((XIF_DOCHANDLE)xDoc,
		(UInt32)pageNum, NULL, &imageLayers, &annotations,
		&pageHandle));
	}
    }
#endif

    /* Actually remove the page from the linked list */
    if (xPage->prev) {
	xPage->prev->next = nextPage = xPage->next;
    } else {
	xDoc->pages = nextPage = xPage->next;
    }
    if (xPage->next) {
	xPage->next->prev = xPage->prev;
    }
    xPage->prev = xPage->next = NULL;

    /* Remove the page from existance */
    XERR_PASS_CLEANUP(wp_XifPageFree(&xPage, cFalse));

    /* Revise the page offsets of all succeeding pages */
    xPage = nextPage;
    while (xPage != NULL) {
	xPage->pageInfo.dwPageOffset--;
	xPage = xPage->next;
    }

    /* Shift the page table entries by one */
    XERR_PASS_CLEANUP(wp_XifRewritePageTableEntries(xDoc, nextPage == NULL?
	xDoc->pages: nextPage));

    /* Update the page count */
    xDoc->docInfo.wNPages--;
    xDoc->pageTableHeader.pageCount--;
    XERR_PASS_CLEANUP(wp_XifWritePageTableHeader(xDoc));
    XERR_GEN_CLEANUP(xDoc->docInfo.wNPages != xDoc->pageTableHeader.pageCount,
    	XIF_INTERNAL_ERROR);

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/* XIF_CheckImageInfo()
 *
 * OneLineDescription:::
 *  Check and adjust (if necessary) the image info
 *
 * LongDescription:::
 *  XIF_CheckImageInfo() verifies the consistency of the image info based
 *  on the XIF version, and compression type. Changes are made to the
 *  parameters based on what XIF_AddImage(), XIF_AddMaskImage(), and/or
 *  XIF_ImageWriteStart() would do, or what XIF_TokenizeImage() would do.
 *
 * Input:::
 *  xPageHandle::		Page handle
 *  pdwImgLayer::		Image layer in MRC (1, 2, ...)
 *  pdwImgOffset::		Image rank within the layer (0, 1, 2, ...)
 *  pxColorType::		Color type
 *  pdwXOffset::		X pixel offset in Primage image coordinates
 *  pdwYOffset::		Y pixel offset in Primage image coordinates
 *  pdwXResolution::		Resolution in the X direction
 *  pdwYResolution::		Resolution in the Y direction
 *  pdwWidth::			Width
 *  pdwHeight::			Height
 *  pdwBestStripHeight::	Height of strips stored
 *  pPalette::			Palette (if XIF_COLORTYPE_COLOR8 or COLOR4)
 *  pxCompression::		Type of compression
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_CheckImageInfo(
	XIF_PAGEHANDLE xPageHandle,
	UInt32 *pdwImgLayer,
	UInt32 *pdwImgOffset,
	XIF_COLORTYPE *pxColorType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	UInt8 *pPalette,
	XIF_COMPRESSION *pxCompression
)
{
    XPage *xPage = (XPage *)xPageHandle;
    XImg *xParentImg;
    PROLOGUE;

    CheckHandleValidity(xPageHandle);

    if (*pdwWidth == 0 || *pdwHeight == 0) {
	goto ERR_CLEANUP;
    }

    /* Either we need a primary image, or we're adding it now */
    if (*pdwImgLayer != 2 && xPage->pageInfo.xPrimImgHandle == 0 &&
	(*pdwImgLayer < 1 || *pdwImgLayer > 3)) {
	*pdwImgLayer = 2;	/* Could be 1, 2 or 3 */
	*pdwImgOffset = 0;	/* Make first in layer */
    }
    xParentImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

    /* Make some page consistency checks */
    if (xParentImg != NULL) {
	if (*pdwXResolution == 0) {
	    *pdwXResolution = xParentImg->imgInfo.dwXResolution;
	}
	if (*pdwYResolution == 0) {
	    *pdwYResolution = xParentImg->imgInfo.dwYResolution;
	}
	if (*pdwXOffset > xParentImg->imgInfo.dwWidth) {
	    *pdwXOffset = 0;
	}
	if (*pdwYOffset > xParentImg->imgInfo.dwHeight) {
	    *pdwYOffset = 0;
	}

	/* Note: given the resolution differences, the last subimage pixel
	 * may overhang the edge of the page, but not have a pixel completely
	 * outside of the page */
	if (*pdwXOffset + (*pdwWidth - 1) * xParentImg->imgInfo.dwXResolution
	    / *pdwXResolution >= xParentImg->imgInfo.dwWidth) {
	    if (xParentImg->imgInfo.dwXResolution >= *pdwXResolution) {
		*pdwWidth = (xParentImg->imgInfo.dwWidth - *pdwXOffset +
		    xParentImg->imgInfo.dwXResolution / *pdwXResolution - 1) *
		    *pdwXResolution / xParentImg->imgInfo.dwXResolution;
	    } else {
		*pdwWidth = ((xParentImg->imgInfo.dwWidth - *pdwXOffset) *
		    *pdwXResolution + xParentImg->imgInfo.dwXResolution - 1) /
		    xParentImg->imgInfo.dwXResolution;
	    }
	}
	if (*pdwYOffset + (*pdwHeight - 1) * xParentImg->imgInfo.dwYResolution
	    / *pdwYResolution >= xParentImg->imgInfo.dwHeight) {
	    if (xParentImg->imgInfo.dwYResolution >= *pdwYResolution) {
		*pdwHeight = (xParentImg->imgInfo.dwHeight - *pdwYOffset +
		    xParentImg->imgInfo.dwYResolution / *pdwYResolution - 1) *
		    *pdwYResolution / xParentImg->imgInfo.dwYResolution;
	    } else {
		*pdwHeight = ((xParentImg->imgInfo.dwHeight - *pdwYOffset) *
		    *pdwYResolution + xParentImg->imgInfo.dwYResolution - 1) /
		    xParentImg->imgInfo.dwYResolution;
	    }
	}
    } else {
	if (*pdwXOffset != 0) {
	    *pdwXOffset = 0;
	}
	if (*pdwYOffset != 0) {
	    *pdwYOffset = 0;
	}
    }

    if (*pdwBestStripHeight == 0) {

	/* General rule (in TIFF) is that it shouldn't be more than 8K
	 * uncompressed, but we'll use 64K for non-binary */
	switch (*pxColorType) {
	case XIF_COLORTYPE_COLOR24:
	    *pdwBestStripHeight = 65536 /
		(*pdwWidth * XIF_BPP(*pxColorType) / 8);

	    /* Round up to the next multiple of 16 */
	    *pdwBestStripHeight = (*pdwBestStripHeight + 15) & ~15;
	    break;
	case XIF_COLORTYPE_GRAY4:
	case XIF_COLORTYPE_COLOR4:
	case XIF_COLORTYPE_GRAY8:
	case XIF_COLORTYPE_COLOR8:
	    *pdwBestStripHeight = 65536 /
		(*pdwWidth * XIF_BPP(*pxColorType) / 8);

	    /* Round up to the next multiple of 8 */
	    *pdwBestStripHeight = (*pdwBestStripHeight + 7) & ~7;
	    break;
	default:
	    *pdwBestStripHeight = 8192 /
		(*pdwWidth * XIF_BPP(*pxColorType) / 8);
	}

	/* If truncated to nothing, use the full image */
	if (*pdwBestStripHeight == 0) {
	    *pdwBestStripHeight = *pdwHeight;
	}
    }

    /* Make sure the compression type is supported */
    if (xPage->xDoc->docInfo.version < 3) {
	switch (*pxCompression) {
	case XIF_WAVELET:
	    *pxCompression = XIF_JPEG;
	    break;
	case XIF_TOKENIZATION:
	case XIF_LLTOKENIZATION:
	case XIF_LOSSYDITHER:
	    *pxCompression = XIF_G4;
	    break;
	}
    }

    /* Make sure the compression matches the color type, and that a palette
     * exists if required. */
    switch ((int)*pxColorType) {
    case XIF_COLORTYPE_BINARY:
	switch ((int)*pxCompression) {
	case XIF_NOC:
	case XIF_G4:
	    break;  /* These are valid */
	case XIF_TOKENIZATION:
	case XIF_LLTOKENIZATION:
	case XIF_LOSSYDITHER:
	    if (xPage->xDoc->docInfo.version < 3) {
		*pxCompression = XIF_G4;   /* Only this Supported */
	    } else {
		*pdwBestStripHeight = *pdwHeight;
	    }
	    break;  /* Otherwise valid */
	case XIF_COLORFORTOKENS:
	    /* CHANGE ME TO 4 *//* CHANGE ME TO 4 *//* CHANGE ME TO 4 */
	    /* CHANGE ME TO 4 *//* CHANGE ME TO 4 *//* CHANGE ME TO 4 */
	    /* CHANGE ME TO 4 *//* CHANGE ME TO 4 *//* CHANGE ME TO 4 */
	    /* CHANGE ME TO 4 *//* CHANGE ME TO 4 *//* CHANGE ME TO 4 */
	    /* CHANGE ME TO 4 *//* CHANGE ME TO 4 *//* CHANGE ME TO 4 */
	    /* CHANGE ME TO 4 *//* CHANGE ME TO 4 *//* CHANGE ME TO 4 */
	    if (xPage->xDoc->docInfo.version < 3 /* CHANGE ME TO 4 */) {
		*pxCompression = XIF_JPEG;   /* Only this Supported */
	    }
	    *pdwBestStripHeight = *pdwHeight;
	    break;
	default:
	    *pxCompression = XIF_G4; /* The default */
	}
	break;
    case XIF_COLORTYPE_GRAY4:
    case XIF_COLORTYPE_GRAY8:
    case XIF_COLORTYPE_COLOR24:
	switch ((int)*pxCompression) {
	case XIF_NOC:
	case XIF_JPEG:
	    break;  /* This is valid */
	case XIF_TOKENIZATION:
	case XIF_LLTOKENIZATION:
	case XIF_WAVELET:
	    if (xPage->xDoc->docInfo.version < 3) {
		*pxCompression = XIF_JPEG;   /* Only this Supported */
	    }
	    break;  /* Otherwise valid */
	default:
	    *pxCompression = XIF_JPEG; /* The default */
	}
	break;
    case XIF_COLORTYPE_COLOR4:
    case XIF_COLORTYPE_COLOR8:
	switch ((int)*pxCompression) {
	case XIF_NOC:
	    break;  /* This are valid */
	default:
	    *pxCompression = XIF_NOC; /* The default */
	}
	XERR_GEN_CLEANUP(pPalette == NULL, XIF_BADPARAMETER);
	break;
    default:
	break;
    }

    /* XIF 2.0 file must always use a single image when using JPEG compression.
     */
    if (xPage->xDoc->docInfo.version < 3 && *pxCompression == XIF_JPEG) {
	switch (*pxColorType) {
	case XIF_COLORTYPE_COLOR24:
	case XIF_COLORTYPE_GRAY4:
	case XIF_COLORTYPE_GRAY8:
	    *pdwBestStripHeight = *pdwHeight;
	}
    }

    /* Always limit it to the image height */
    if (*pdwBestStripHeight > *pdwHeight) {
	*pdwBestStripHeight = *pdwHeight;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/* XIF_AddImage()
 *
 * OneLineDescription:::
 *  Add an image to a page
 *
 * LongDescription:::
 *  XIF_AddImage() adds a new image to a page. Image should not be a
 *  transparancy mask, XIF_AddMaskImage() is used for handling this.
 *  <BR>
 *  Note that every empty page must start with a Primary image which defines
 *  the boundaries of the page, both in width and height.  XIF_BADPARAMETER
 *  will be returned if the first image added is not a primary image (image
 *  layer == 2).
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: XIF_ImageWriteStart(), XIF_ImageWriteStrip(), and
 *  XIF_ImageWriteFinish(); or to the two functions:
 *  XIF_ImageTokenizeImage(), and XIF_WriteTokenImage(); or the function:
 *  XIF_ImageTokenizeImages().
 *
 * Input:::
 *  xPageHandle::		Page to attach to
 *  dwImgLayer::		Image layer in MRC
 *  xColorType::		Color type
 *  dwXOffset::			X pixel offset in Primage image coordinates
 *  dwYOffset::			Y pixel offset in Primage image coordinates
 *  dwXResolution::		Resolution in the X direction
 *  dwYResolution::		Resolution in the Y direction
 *  dwWidth::			Width
 *  dwHeight::			Height
 *  dwBestStripHeight::		Height of strips stored
 *  pPalette::			Palette (if XIF_COLORTYPE_COLOR8 or COLOR4)
 *
 * Output:::
 *  pImgHandle::		Pointer to image handle, used to later
 *				reference this image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_AddImage(
	XIF_PAGEHANDLE xPageHandle,
	UInt32 dwImgLayer,
	XIF_COLORTYPE xColorType,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwXResolution,
	UInt32 dwYResolution,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwBestStripHeight,
	UInt8 *pPalette,
	XIF_IMGHANDLE *pImgHandle
)
{
    Instance *instData;
    XFile *xFile;
    XPage *xPage = (XPage *)xPageHandle;	/* Page */
    XImg *xParentImg = NULL;		/* Parent image */
    Int32 ifdOffset;
    UInt8 flags = 0;
    XImg *curImg;			/* Current image */
    XImg *newImg = NULL;		/* Unlinked image */
    XImg *prevImg, *nextImg;		/* Prev/next image in ordered link list */
    XIF_COMPRESSION xCompression;
    UInt32 dwImgOffset;
    PROLOGUE;

    /* Default return value */
    *pImgHandle = 0UL;

    CheckHandleValidity(xPageHandle);
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);

    /* Either we need a primary image, or we're adding it now */
    XERR_GEN(dwImgLayer != 2 && xPage->pageInfo.xPrimImgHandle == 0,
		     XIF_SEQUENCE);
    xParentImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

    /* Check for proper calling sequence:
     * 1 - not writing this page
     * 2 - primary image not written for this page, and this isn't
     *     the primary image
     */
    XERR_GEN((((xPage->seqStatus & XPAGE_SEQ_WRITESTART) == 0) ||
	(((xPage->seqStatus & XPAGE_SEQ_WRITEPRIMARY) == 0)
	&& (dwImgLayer != 2))), XIF_SEQUENCE);

    /* Check for version support: only layer 2 offset 0 valid in even layers */
    XERR_GEN(xPage->xDoc->docInfo.version < 3 &&
	(dwImgLayer % 2) == 0 &&		    /* Even numbered layer */
	(dwImgLayer != 2 || xParentImg != NULL),    /* Not primary image */
	XIF_NOSUPPORT);

    /* Make some page consistency checks */
    xCompression = XIF_COMP_UNKNOWN;
    dwImgOffset = 0UL;
    XERR_PASS_CLEANUP(XIF_CheckImageInfo(xPageHandle, &dwImgLayer,
	&dwImgOffset, &xColorType, &dwXOffset, &dwYOffset,
	&dwXResolution, &dwYResolution,	&dwWidth, &dwHeight,
	&dwBestStripHeight, pPalette, &xCompression));

    /* Align long (for all TIFF IFDs) */
    XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &ifdOffset));

    /* Allocate a new image */
    XERR_PASS_CLEANUP(wp_NewImg(xPage, &newImg, ifdOffset));
    newImg->imgInfo.xImgType = XIF_IMGTYPE_LAYER;
    newImg->imgInfo.dwImgLayer = dwImgLayer;
    newImg->imgInfo.dwImgOffset = 0UL;

    /* Retain only the transferable information */
    newImg->imgInfo.dwXOffset = dwXOffset;
    newImg->imgInfo.dwYOffset = dwYOffset;
    newImg->imgInfo.dwXResolution = dwXResolution;
    newImg->imgInfo.dwYResolution = dwYResolution;
    newImg->imgInfo.dwWidth = dwWidth;
    newImg->imgInfo.dwHeight = dwHeight;
    newImg->imgInfo.dwBestStripHeight = dwBestStripHeight;
    newImg->imgInfo.xCompression = XIF_COMP_UNKNOWN;
    newImg->imgInfo.xColorType = xColorType;
    switch ((int)xColorType) {
    case XIF_COLORTYPE_COLOR4:
    case XIF_COLORTYPE_COLOR8:
	XERR_GEN_CLEANUP((pPalette == NULL),
		 XIF_BADPARAMETER);
	{
	    UInt32 paletteSize;

	    paletteSize = (1 << XIF_BPPE(xColorType)) * 3;
	    memcpy(newImg->imgInfo.pPalette, pPalette, paletteSize);
	}
    default:
	break;
    }

    /* Write out the IFD */
    XERR_PASS_CLEANUP(wp_XifWriteMockIFD(newImg));

    /* If this is the primary image, link it to the page */
    if (xParentImg == NULL) {

	/* Assign this one as the primary image */
	xPage->pageInfo.xPrimImgHandle = (UInt32)(curImg = newImg);
	newImg = NULL;
	curImg->imgInfo.xImgType = XIF_IMGTYPE_PRIMARY;

	/* In the case of an inserted page, shift the page table entries to
	 * accomodate the insertion of this page. */
	if (xPage->next) {
	    XERR_PASS_CLEANUP(wp_XifRewritePageTableEntries(xPage->xDoc, xPage));
	}

	/* Make sure we have the page table for the correct page */
	XERR_PASS_CLEANUP(wp_XifGeneratePageFlags(xPage, &flags));
	XERR_PASS_CLEANUP(wp_XifSetPageIFD(xPage->xDoc, xPage->pageInfo.dwPageOffset,
					flags, ifdOffset, xPage));

	/* Start off with a image count of 1 for this layer (usually layer 2) */
	XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage, dwImgLayer, 1));

	/* Update the sequence status */
	xPage->seqStatus |= XPAGE_SEQ_WRITEPRIMARY | XPAGE_SEQ_GETINFO;
	
    } else {	/* Otherwise link it to the parent image */
	XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage, dwImgLayer, 1));
	newImg->imgInfo.dwImgOffset =
	    xPage->pageInfo.pdwImageCounts[dwImgLayer - 1] - 1;
	if (xPage->xImgs == NULL) {
	    xPage->xImgs = curImg = newImg;
	    newImg = NULL;
	} else {

	    /* Search for the proper point to insert *before* */
	    prevImg = nextImg = xPage->xImgs;
	    while (nextImg &&
		(nextImg->imgInfo.dwImgLayer < dwImgLayer ||
		(nextImg->imgInfo.dwImgLayer == dwImgLayer &&
		 nextImg->imgInfo.dwImgOffset < newImg->imgInfo.dwImgOffset)
		)) {
		prevImg = nextImg;
		nextImg = nextImg->next;
	    }

	    /* It is an error to insert (for now) */
	    XERR_GEN_CLEANUP(nextImg && nextImg->imgInfo.dwImgLayer == dwImgLayer &&
		nextImg->imgInfo.dwImgOffset == newImg->imgInfo.dwImgOffset,
		XIF_INTERNAL_ERROR);

	    /* Link this image to the one that is now the next one */
	    if (nextImg) {
		newImg->next = nextImg;
		nextImg->prev = newImg;
	    }

	    /* Link prev image to this one */
	    if (prevImg == nextImg) {

		/* Link as new first image */
		xPage->xImgs = newImg;
	    } else {
		prevImg->next = newImg;
		newImg->prev = prevImg;

	    }
	    curImg = newImg;
	    newImg = NULL;
	}
	XERR_PASS_CLEANUP(wp_TifUpdateILContent(xParentImg));

	/* Attach this image to its parent */
	XERR_PASS_CLEANUP(wp_TifAttachChildIFD(xParentImg, curImg, ifdOffset));
    }

    /* Return the image added */
    *pImgHandle = (UInt32)curImg;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newImg) {
	XIF_FREE(newImg);
    }

    ERR_RETURN;
}


/* XIF_DeleteImage()
 *
 * OneLineDescription:::
 *  Delete an image from a page
 *
 * LongDescription:::
 *  XIF_DeleteImagee() deletes (removes all references to) an single image on
 *  a page.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  image are removed.  This includes: MRC layer image counts and relinking
 *  the previous image to the subsequent image via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the image is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' an
 *  image deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see XIF_CopyPage() and
 *  XIF_CopyImage()), allowing the disowned file fragments to remain behind.
 *
 * Input:::
 *  xImgHandle::		Image handle of the image to be removed, as
 *				returned from XIF_GetImageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_DeleteImage(
	XIF_IMGHANDLE xImgHandle
)
{
    XImg *xImg = (XImg *)xImgHandle;	/* Image */
    PROLOGUE;

    CheckHandleValidity(xImgHandle);

    /* Check for proper calling sequence */
    XERR_GEN((xImg->xPage->xDoc->readWrite & WRITE_ALLOWED) == 0,
	     XIF_SEQUENCE);

    /* Let the TIFF routine do the dirty work */
    XERR_PASS_CLEANUP(wp_TifDeleteImage(xImg));

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/* XIF_AddMaskImage()
 *
 * OneLineDescription:::
 *  Add a transparency mask to a non-binary image
 *
 * LongDescription:::
 *  XIF_AddMaskImage() is like XIF_AddImage(), except that instead of adding
 *  an image to a page, we are adding a transparency mask image to a parent
 *  non-binary image.  The image layer is inherited from the parent image.
 *  This image must have the same X/Y location, width, height and resolutions
 *  as the parent image.  This image MUST also be binary, with 0 indicating
 *  the transparent  pixels, and 1 indicating the opaque pixels of the parent.
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: XIF_ImageWriteStart(), XIF_ImageWriteStrip(), and
 *  XIF_ImageWriteFinish().
 *
 * Input:::
 *  xParentImgHandle::		Parent image to attach to
 *  dwBestStripHeight::		For best performance, write strips with
 *				height a multiple of this
 *
 * Output:::
 *  pImgHandle::		Returned image handle for image to be
 *				added
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_AddMaskImage(
	XIF_IMGHANDLE xParentImgHandle,
	UInt32 dwBestStripHeight,
	XIF_IMGHANDLE *pImgHandle
)
{
    XFile *xFile;
    XPage *xPage;
    XImg *xParentImg = (XImg *)xParentImgHandle;/* Parent image */
    Int32 ifdOffset;
    XImg *curImg;			/* Current mask image */
    XImg *newImg = NULL;		/* Unlinked mask image */
    PROLOGUE;

    /* Default return value */
    *pImgHandle = 0UL;

    CheckHandleValidity(xParentImgHandle);

    xPage = xParentImg->xPage;

    /* Mask image cannot be applied to an even layered image, with the
     * special exception of the primary image (this exception does not
     * normally conform to XIFF 3.0, but this library will allow it). */
    XERR_GEN_CLEANUP(((xParentImg->imgInfo.dwImgLayer & 1) != 1) &&
		     (xParentImg->imgInfo.xImgType != XIF_IMGTYPE_PRIMARY),
		     XIF_BADPARAMETER);
    /* Image must not already have a mask */
    XERR_GEN_CLEANUP((xParentImg->seqStatus & XIMG_SEQ_MASKADDED) != 0,
		     XIF_BADPARAMETER);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((((xPage->seqStatus & XPAGE_SEQ_WRITESTART) == 0) ||
		      (xPage->seqStatus & XPAGE_SEQ_WRITEPENDING) ||
	((xPage->seqStatus & XPAGE_SEQ_WRITEPRIMARY) == 0)),
		     XIF_SEQUENCE);

    if (dwBestStripHeight == 0) {

	/* General rule is that it shouldn't be more than 8K uncompressed */
	dwBestStripHeight = (8192 / (xParentImg->imgInfo.dwWidth *
	    XIF_BPP(XIF_COLORTYPE_BINARY) / 8));
    }

    xFile = XIF_PAGEFILE(xPage);

    /* Align long (for all TIFF IFDs) */
    XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &ifdOffset));

    /* Allocate a new image */
    XERR_PASS_CLEANUP(wp_NewImg(xPage, &newImg, ifdOffset));
    newImg->imgInfo.xImgType = XIF_IMGTYPE_MASK;
    newImg->imgInfo.dwImgLayer = 0;	/* Not in a layer */
    newImg->imgInfo.dwImgOffset = 0UL;

    /* Retain only the transferable information */
    newImg->imgInfo.dwXOffset = xParentImg->imgInfo.dwXOffset;
    newImg->imgInfo.dwYOffset = xParentImg->imgInfo.dwYOffset;
    newImg->imgInfo.dwXResolution = xParentImg->imgInfo.dwXResolution;
    newImg->imgInfo.dwYResolution = xParentImg->imgInfo.dwYResolution;
    newImg->imgInfo.dwWidth = xParentImg->imgInfo.dwWidth;
    newImg->imgInfo.dwHeight = xParentImg->imgInfo.dwHeight;
    newImg->imgInfo.dwBestStripHeight = dwBestStripHeight;
    newImg->imgInfo.xCompression = XIF_COMP_UNKNOWN;
    newImg->imgInfo.xColorType = XIF_COLORTYPE_BINARY;

    /* Write out the IFD */
    XERR_PASS_CLEANUP(wp_XifWriteMockIFD(newImg));

    /* Link it to the parent image */
    if (xParentImg->xMaskImg == NULL) {
	xParentImg->xMaskImg = curImg = newImg;
	newImg = NULL;
    } else {
	/* Only one transparency mask is allowed */
	__error__ = XIF_BADPARAMETER;
	goto ERR_CLEANUP;
    }

    /* Attach this image to its parent */
    XERR_PASS_CLEANUP(wp_TifAttachChildIFD(xParentImg, curImg, ifdOffset));

    /* For now, only one attached mask is allowed */
    xParentImg->imgInfo.dwMaskImages++;
    XERR_GEN_CLEANUP(xParentImg->imgInfo.dwMaskImages != 1, XIF_INTERNAL_ERROR);

    /* Mark as having a mask */
    xParentImg->seqStatus |= XIMG_SEQ_MASKADDED;

    /* Return the image added */
    *pImgHandle = (UInt32)curImg;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newImg) {
	Instance *instData;

	if (xParentImg) {
	    instData = XIF_IMGINST(xParentImg);
	    XIF_FREE(newImg);
	}
    }

     ERR_RETURN;
}


/* XIF_ImageWriteStart()
 *
 * OneLineDescription:::
 *  Prepare to write image data to the XIFF image
 *
 * LongDescription:::
 *  XIF_ImageWriteStart() initiates the write operation for an image. The
 *  image handle is that returned by XIF_AddImage(). This call will be
 *  followed by one or more calls to XIF_ImageWriteStrip(), and a closing
 *  call to XIF_ImageWriteFinish.
 *  <BR>
 *  pCompression points to the desired compression method. If the compression
 *  type is invalid or unsupported, the actual compression to be used is
 *  returned through this pointer. Note: if the image is the color plane for
 *  later use in color tokenization, the data is not actually written, but is
 *  stored for later extraction and assignment to symbols.
 *  <BR>
 *  If JPEG compression is used, pJPEGQuality points to the desired JPEG
 *  quality (1-100). If the JPEG quality changes from what is desired to what
 *  will be used, the actual value is passed back through the pointer. An
 *  unreasonable quality factor will yield a different output than passed in.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  dwFlags::		
 *  pxCompression::		Pointer to the requested compression type.
 *				XIF_COMP_UNKNOWN yields default.
 *				The compression type used is passed back
 *				through this pointer
 *  pdwJPEGQuality::		Pointer to the requested JPEG quality (0-100)
 *				  0 - default used and passed back
 *				 75 - good quality, best compression
 *				 90 - better quality, good compression
 *				100 - best quality, okay compression
 *				The JPEG quality used is passed back
 *				through this pointer
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ImageWriteStart(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwFlags, 
	XIF_COMPRESSION *pxCompression,
	UInt32 *pdwJPEGQuality
)	  
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    PROLOGUE;
    
    CheckHandleValidity(xImgHandle);
    pxImgInfo = &xImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_WRITESTATUS) !=
	XIMG_SEQ_WRITEFINISH), XIF_SEQUENCE);
    
    /* Store the flags parameters passed */
    xImg->imgFlags = dwFlags;
    pxImgInfo->xCompression = *pxCompression;
    xImg->JPEGquality = *pdwJPEGQuality;

    /* Check to see if compression options are valid */
    /* Make some page consistency checks */
    XERR_PASS_CLEANUP(XIF_CheckImageInfo((XIF_PAGEHANDLE)xImg->xPage,
	&pxImgInfo->dwImgLayer, &pxImgInfo->dwImgOffset,
	&pxImgInfo->xColorType,
	&pxImgInfo->dwXOffset, &pxImgInfo->dwYOffset,
	&pxImgInfo->dwXResolution, &pxImgInfo->dwYResolution,
	&pxImgInfo->dwWidth, &pxImgInfo->dwHeight,
	&pxImgInfo->dwBestStripHeight, pxImgInfo->pPalette,
	&pxImgInfo->xCompression));
 
    switch ((int)pxImgInfo->xCompression) {
    case XIF_JPEG:
    case XIF_WAVELET:
	if (xImg->JPEGquality == 0) {
	    xImg->JPEGquality = 75;	/* The default */
	} else if (xImg->JPEGquality > 100) {
	    xImg->JPEGquality = 100;	/* The max */
	} else if (xImg->JPEGquality < 1) {
	    xImg->JPEGquality = 1;	/* My hard-coded min */
	}
	break;
    default:
	xImg->JPEGquality = 0;
    }

    /* Now write the real IFD */
    XERR_PASS_CLEANUP(wp_TifWriteIFD(xImg));

    /* Return the compression to be used */
    *pxCompression = pxImgInfo->xCompression;
    *pdwJPEGQuality = xImg->JPEGquality;

    /* Update the sequence status */
    xImg->seqStatus |= XIMG_SEQ_WRITESTART | XIMG_SEQ_GETINFO;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/* wp_CopyToTempImage()
 *
 * OneLineDescription:::
 *  Copy user image buffer to intermediate TIFF buffer (PIXR)
 *
 * LongDescription:::
 *  wp_CopyToTempImage() copies fragments of the user buffer necessary to buffer
 *  up a TIFF strip, such that a whole TIFF strip can be accumulated before the
 *  write. The temporary PIXR xImg->tmpPixr is used as this intermediate buffer.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  ppImgChannels::		Pointer to array of 1 to 3 image channels to
 *				copy from
 *  srcBPL::			Source image bytes per line (for one channel)
 *  pdwBufferRowOffset::	Pointer to source image row offset, updated for
 *				outer loop
 *  pdwImgRowOffset::		Pointer to image proper row offset, updated for
 *				outer loop
 *  pdwRowsToWrite::		Pointer to the number of rows in outer loop to
 *				write
 *  pdwTotalRowsFilled::	Pointer to the total number of rows buffered out
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_CopyToTempImage(
    XImg *xImg,
    UInt8 **ppImgChannels,
    Int32 srcBPL,
    UInt32 *pdwBufferRowOffset,
    UInt32 *pdwImgRowOffset,
    UInt32 *pdwRowsToWrite,
    UInt32 *pdwTotalRowsFilled
)
{
    UInt8 *srcPtr, *dstPtr;
    UInt32 byteCount;
    IPWState *pIPWState;
    Int32 channel, channels;
    PIXR *dstPixr, *pSubPixr;
    Int32 dstBPL = 0L;
    UInt32 rowsFilled = 0UL;
    UInt32 rowsAvailable;
    Int32 pixrRow, pixrHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pdwTotalRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    /* Examine image features of the destination PIXR */
    dstPixr = xImg->tmpPixr;
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, dstPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, dstPixr, &pixrHeight));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, dstPixr, &dstBPL));
    byteCount = (xImg->minBytesPerRow + 3) & ~3;
    XERR_GEN_CLEANUP(byteCount > dstBPL, XIF_BADPARAMETER);

    /* Check to see if nothing needs to be done */
    if (xImg->tmpPixr == NULL || *pdwImgRowOffset < xImg->tmpPixrRow ||
	*pdwImgRowOffset >= xImg->tmpPixrRow + pixrHeight) {
	NORMAL_RETURN;
    }

    /* Make sure we haven't been asked to skip lines */
    pixrRow = *pdwImgRowOffset - xImg->tmpPixrRow;
    XERR_GEN_CLEANUP(pixrRow != xImg->tmpPixrRows, XIF_SEQUENCE);
    
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, dstPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &dstPtr));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, dstPixr, &dstPtr));
	}
	srcPtr = ppImgChannels[channel] + *pdwBufferRowOffset * srcBPL;
	dstPtr += pixrRow * dstBPL;
	rowsAvailable = pixrHeight - pixrRow;
	if (rowsAvailable > *pdwRowsToWrite) {
	    rowsAvailable = *pdwRowsToWrite;
	}

	/* Copy the PIXR image to the user's image */
	while (rowsAvailable--) {
	    memcpy(dstPtr, srcPtr, byteCount);
	    srcPtr += srcBPL;
	    dstPtr += dstBPL;
	    rowsFilled++;
	}
    } /* for each channel */
    rowsFilled /= channels;
    xImg->tmpPixrRows += rowsFilled;	    /* Update intermediate PIXR row count */
    *pdwBufferRowOffset += rowsFilled;	    /* Update external source buffer row */
    *pdwImgRowOffset += rowsFilled;	    /* Update external image proper row */
    *pdwRowsToWrite -= rowsFilled;	    /* Update external "rows left" */
    (*pdwTotalRowsFilled) = rowsFilled;	    /* Return the count of rows filled */
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_ImageWriteStrip()
 *
 * OneLineDescription:::
 *  Compress and write a strip of image data
 *
 * LongDescription:::
 *  XIF_ImageWriteStrip() compresses and writes a strip of image data. If
 *  necessary, the image data is broken up into TIFF strips, and stored as
 *  separate strips.  The last strip may be stored until the next block of
 *  data is added to complete a TIFF strip.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  ppImagePlanes::		Pointer to image buffer(s)
 *  dwBytesPerRow::		Bytes per image row
 *  dwBufferRowOffset::		Row offset in image of the strip to write
 *  dwStripHeight::		Number of rows in PIXR to write
 *  dwImgRowOffset::		Row offset in image proper to write strip
 *
 * Output:::
 *  pRowsFilled::		Number of rows written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data to this
 *  buffer as needed, and writing the intermediate buffer when filled.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ImageWriteStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pRowsFilled
)
{
    Instance *instData;
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState;
    UInt32 channels;
    Int32 tmpPixrHeight = 0;
    ColorType ctype;
    UInt32 rowsFilled = 0UL;
    UInt32 totalRowsFilled = 0UL;
    Int32 writeRows;
    PROLOGUE;

    /* Default to return nothing */
    *pRowsFilled = 0UL;
    
    CheckHandleValidity(xImgHandle);

    instData = XIF_IMGINST(xImg);    
    pxImgInfo = &xImg->imgInfo;
    channels = XIF_BPP(pxImgInfo->xColorType) / XIF_BPPE(pxImgInfo->xColorType);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_WRITESTATUS) !=
		     XIMG_SEQ_WRITESTART), XIF_SEQUENCE);

    /* Check for proper compression mode */
    XERR_GEN_CLEANUP(((xImg->imgFlags & XIF_IMGFLAGS_Compression) !=
	XIF_IMGFLAGS_Uncompressed), XIF_SEQUENCE);

    /* Reference the IPWrap state */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    /* Get the temporary PIXR height, if it exists */
    if (xImg->tmpPixr != NULL) {
	XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, xImg->tmpPixr,
					  &tmpPixrHeight));
    }

    /* Loop until the user's image buffer (region) is emptied */
    while (dwStripHeight) {

	/* Check for overfill */
	if (dwImgRowOffset + dwStripHeight > pxImgInfo->dwHeight) {
	    dwStripHeight = pxImgInfo->dwHeight - dwImgRowOffset;
	}
	if (dwStripHeight <= 0) {
	    break;
	}

	/* Compute the number of strips to be written (to match TIFF strip) */
	writeRows = pxImgInfo->dwBestStripHeight;
	if (writeRows > pxImgInfo->dwHeight - xImg->tmpPixrRow) {
	    writeRows = pxImgInfo->dwHeight - xImg->tmpPixrRow;
	}

	/* If wavelet compressing, the entire PIXR must be a full TIFF
	 * strip (we can't tell the wavelet routines to compress only a portion
	 * of the PIXR). */
	if (pxImgInfo->xCompression == XIF_WAVELET && xImg->tmpPixr != NULL) {
	    if (tmpPixrHeight != writeRows) {
		XERR_GEN_CLEANUP(xImg->tmpPixrRows > 0, XIF_INTERNAL_ERROR);
		XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &xImg->tmpPixr));
		xImg->tmpPixr = NULL;
	    }
	}

	/* If we don't have an intermediate PIXR to handle buffering,
	 * create it now */
	if (xImg->tmpPixr == NULL) {
	    ScaledInt1616 xres, yres;
		
	    /* Prepare a temporary PIXR */
	    xres = Int32_TO_ScaledInt1616(pxImgInfo->dwXResolution);
	    yres = Int32_TO_ScaledInt1616(pxImgInfo->dwYResolution);
	    switch ((int)pxImgInfo->xColorType) {
	    case XIF_COLORTYPE_COLOR24:
		ctype = cColorRGB;
		break;
	    case XIF_COLORTYPE_BINARY:
	    case XIF_COLORTYPE_GRAY4:
	    case XIF_COLORTYPE_COLOR4:
	    case XIF_COLORTYPE_GRAY8:
	    case XIF_COLORTYPE_COLOR8:
		ctype = cColorGray;
		break;
	    default:
		goto ERR_CLEANUP;
	    }
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, pxImgInfo->dwWidth,
		writeRows, XIF_BPPE(pxImgInfo->xColorType),
		xres, yres, 32, XIF_CHANNELS(pxImgInfo->xColorType),
		ctype, NULL, &xImg->tmpPixr));
	    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, xImg->tmpPixr, &tmpPixrHeight));
	} /* If no intermediate image buffer */

	/* Fill up the intermediate image if we haven't already, update:
	 *	buffer row offset	    dwBufferRowOffset
	 *	image row offset	    dwImgRowOffset
	 *	rows left to buffer out	    dwStripHeight
	 * We update:
	 *	number of rows in buffer    xImg->tmpPixrRows
	 */
	XERR_PASS_CLEANUP(wp_CopyToTempImage(xImg, ppImagePlanes, dwBytesPerRow,
	    &dwBufferRowOffset, &dwImgRowOffset, &dwStripHeight, &rowsFilled));
	totalRowsFilled += rowsFilled;

	/* If the intermediate buffer is full, write out the strip */
	if (xImg->tmpPixrRows == pxImgInfo->dwBestStripHeight ||
	    xImg->tmpPixrRow + xImg->tmpPixrRows == pxImgInfo->dwHeight) {

	    /* Convert image order from user's specifications to PIXR */

	    /*********** for now we assume the user has passed in PIXR image
	     *********** order
	     ***********/

	    /* Write out the PIXR (should now be one full TIFF strip) */
	    XERR_PASS_CLEANUP(w_XifWriteStrip(pIPWState, xImgHandle,
		xImg->tmpPixr, 0, writeRows, xImg->tmpPixrRow,
		&rowsFilled));
	    xImg->tmpPixrRow = dwImgRowOffset;
	    xImg->tmpPixrRows = 0;
	}
    }

    /* Return the number of rows written out */
    *pRowsFilled = totalRowsFilled;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/* XIF_CompressedWriteStrip()
 *
 * OneLineDescription:::
 *  Write a strip of compressed image data
 *
 * LongDescription:::
 *  XIF_CompressedWriteStrip() writes a strip of already compressed image
 *  data. The strip *must* be the strip height as given in the parameter
 *  dwBestStripHeight during the call to XIF_AddImage(), except for the
 *  very last strip which may fill up to the end of the image.  The strip
 *  *must* also lie on a strip boundary.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  dwImgRowOffset::		Row offset in image proper to write strip
 *  dwStripHeight::		Number of rows in PIXR to write
 *  pCmpBuffer::		Pointer to the compressed image buffer
 *  dwCmpBytes::		Bytes of compressed data found at pCompressed
 *
 * Output:::
 *  pdwBytesStored::		Number of bytes written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_CompressedWriteStrip(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 *pCmpBuffer,
	UInt32 dwCmpBytes,
	UInt32 *pdwBytesStored
)
{
    Instance *instData;
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    UInt8 *tmpCmp = NULL;
    UInt32 tmpCmpBytes = 0UL, tmpCmpBufsize = 0UL;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesStored = 0UL;

    CheckHandleValidity(xImgHandle);

    instData = XIF_IMGINST(xImg);    
    pxImgInfo = &xImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_WRITESTATUS) !=
	XIMG_SEQ_WRITESTART), XIF_SEQUENCE);

    /* Check for proper compression mode */
#if 0
// Must change this later
//    XERR_GEN_CLEANUP(((xImg->imgFlags & XIF_IMGFLAGS_Compression) !=
//	XIF_IMGFLAGS_Compressed), XIF_SEQUENCE);
#endif

    /* We *must* have the data in a compressed strip already */
    XERR_GEN_CLEANUP((dwStripHeight != pxImgInfo->dwBestStripHeight) &&
	(dwStripHeight + dwImgRowOffset != pxImgInfo->dwHeight),
	XIF_BADPARAMETER);

    /* The strip must lie on a strip boundary */
    XERR_GEN_CLEANUP(dwImgRowOffset % pxImgInfo->dwBestStripHeight != 0,
	XIF_BADPARAMETER);
    
    /* Internal routines use the compressed buffer attached to the
     * image handle. Repoint previously used values to the ones passed
     * in to this function, then restore later. Ya, ugly, but I haven't
     * had time to consider a better alternative. */
    tmpCmp = xImg->cmpBuffer;
    tmpCmpBytes = xImg->cmpBytes;
    tmpCmpBufsize = xImg->cmpBufsize;

    /* Repoint internal buffer to this one */
    xImg->cmpBuffer = pCmpBuffer;
    xImg->cmpBytes = dwCmpBytes;
    if (xImg->cmpBufsize < xImg->cmpBytes) {
	xImg->cmpBufsize = xImg->cmpBytes;
    }
    XERR_PASS_CLEANUP(wp_TifWriteStrip(xImg));

    /* Restore the internal state */
    xImg->cmpBuffer = tmpCmp;
    xImg->cmpBytes = tmpCmpBytes;
    xImg->cmpBufsize = tmpCmpBufsize;

    /* Update the status (used as a check for number of rows written,
     * see XIF_ImageWriteFinish()) */
    xImg->tmpPixrRow += dwStripHeight;

    /* Return the number of bytes written */
    *pdwBytesStored = dwCmpBytes;

    /* Reset the mode */
#if 0
//    xImg->imgFlags &= ~XIF_IMGFLAGS_Compressed;
#endif

	
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (tmpCmp) {
	xImg->cmpBuffer = tmpCmp;
	xImg->cmpBytes = tmpCmpBytes;
	xImg->cmpBufsize = tmpCmpBufsize;
    }
    
    ERR_RETURN;
}

	
/* XIF_ImageWriteFinish()
 *
 * OneLineDescription:::
 *  Complete an image write
 *
 * LongDescription:::
 *  XIF_ImageWriteFinish() cleans up the write operation state and performs
 *  consistecy checks.  Fails if correct number of rows not written.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ImageWriteFinish(
	XIF_IMGHANDLE xImgHandle
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XIF_IMGINFO *pxImgInfo;
    MISC_MEM *mmptr, *next_mm;
    Instance *instData;
    PROLOGUE;
    
    CheckHandleValidity(xImgHandle);
    
    pxImgInfo = &xImg->imgInfo;
    instData = XIF_IMGINST(xImg);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_WRITESTATUS) !=
		      XIMG_SEQ_WRITESTART), XIF_SEQUENCE);
    
    /* Free intermediate data */
    if (xImg->cmpBuffer) {
	XIF_FREE(xImg->cmpBuffer);
	xImg->cmpBuffer = NULL, xImg->cmpBufsize = xImg->cmpBytes = 0UL;
    }
    if (xImg->tmpPixr) {
	IPWState *pIPWState;

	/* Reference the IPWrap state */
	XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

	XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &xImg->tmpPixr));
    }
    if ((mmptr = xImg->tmpCmpData) != NULL) {
	while (mmptr) {
	    next_mm = mmptr->next;
	    XIF_FREE(mmptr);
	    mmptr = next_mm;
	}
	xImg->tmpCmpData = NULL;
    }
    
    /* Reset the read status */
    xImg->seqStatus &= ~XIMG_SEQ_WRITESTATUS;	/* Clear write status */
    xImg->seqStatus |= XIMG_SEQ_WRITEFINISH;	/* Set to finished */

    /* Now check to see if the right number of rows was written */
    XERR_GEN(xImg->tmpPixrRow != xImg->imgInfo.dwHeight, XIF_SEQUENCE);

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/* XIF_TokenizeImage()
 *
 * OneLineDescription:::
 *  Store a binary image internally using tokenization
 *
 * LongDescription:::
 *  XIF_TokenizeImage() takes as input an entire binary image and performs
 *  the necessary tasks to add the page to the existing (or created) token
 *  dictionary. All image data is left undisturbed, and the tokenized version
 *  of the entire page can be reconstructed from tokens found in the dictionary,
 *  when completed, albeit with the lossy artifacts of the algorithm.
 *  <BR>
 *  Note, the image is not stored off at this time, since the token dictionary
 *  will not be complete until the last page is tokenized (see
 *  XIF_WriteTokenImage() and XIF_WriteTokenDictionary(), or
 *  XIF_WriteTokenImages()).
 *  <BR>
 *  Also, if a sequence of calls to XIF_TokenizeImage() is interrupted by a
 *  call to either XIF_WriteTokenImage(), XIF_WriteTokenDictionary() or
 *  XIF_WriteTokenImages(), then the next call to XIF_TokenizeImage()
 *  following those calls will cause the previous token dictionary to be
 *  flushed (if not), and a new one to be initialized.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  pImgData::			One channel only
 *  dwBytesPerRow::		Bytes per row
 *  dwFlags::			Tokenization flags (see TOKENIZE_xxx in xiff.h)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_TokenizeImage(
	XIF_IMGHANDLE xImgHandle,
	UInt8 *pImgData,
	UInt32 dwBytesPerRow,
	UInt32 dwFlags
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XImg *xImg = (XImg *)xImgHandle;
    UInt8 *imgptr = NULL;
    PROLOGUE;

    CheckHandleValidity(xImgHandle);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xImg->seqStatus & XIMG_SEQ_WRITESTATUS) !=
	XIMG_SEQ_WRITEFINISH), XIF_SEQUENCE);
    XERR_GEN_CLEANUP(xImg->imgInfo.xImgType != XIF_IMGTYPE_PRIMARY,
		     XIF_BADPARAMETER);
#if 0	/* // To be added shortly */
    XERR_GEN_CLEANUP((dwFlags & TOKENIZE_COLORTOO) &&
		     (xImg->xPage->seqStatus & XPAGE_SEQ_TKCOLORADDED) == 0,
		     XIF_SEQUENCE);
    XERR_GEN_CLEANUP((dwFlags & TOKENIZE_COLORTOO) &&
		     xImg->xPage->fgImg == NULL, XIF_INTERNAL_ERROR);
#endif

    if (xImg->xPage->xDoc->docInfo.version < 3 ||
	(dwFlags & TOKENIZE_ON) == TOKENIZE_OFF) {

	/* Tokenization not supported for this version (or not enabled),
	 * let's be nice and store it G4 */
	{
	    UInt8 *imagePlanes[3];
	    UInt32 rowsFilled;
	    XIF_COMPRESSION compression = XIF_G4;
	    UInt32 quality = 0UL;

	    /* Begin writing */
	    XERR_PASS_CLEANUP(XIF_ImageWriteStart(xImgHandle, 0UL, &compression,
		&quality));

	    /* Write the whole strip */
	    imagePlanes[0] = pImgData;
	    XERR_PASS_CLEANUP(XIF_ImageWriteStrip(xImgHandle, imagePlanes,
		dwBytesPerRow, 0, xImg->imgInfo.dwHeight, 0, &rowsFilled));

	    /* Finish writing */
	    XERR_PASS_CLEANUP(XIF_ImageWriteFinish(xImgHandle));
	}

	NORMAL_RETURN;
    }

    /* Transfer the color tokenization mode to the page */
    if (dwFlags & TOKENIZE_COLORTOO) {
	xImg->xPage->seqStatus |= XPAGE_SEQ_TKCOLORADDED;
#if 0	/* // To be added shortly */
	xImg->xPage->xDoc->dictFlags |= TOKEN_COLORADDED;
#endif
	xImg->xPage->xDoc->seqStatus |= XDOC_SEQ_MULTIDICTS;
    }

    if (dwFlags & TOKENIZE_LOSSLESS) {
	xImg->imgInfo.xCompression = XIF_LLTOKENIZATION;
    } else {
	xImg->imgInfo.xCompression = XIF_TOKENIZATION;
    }

    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* Check the current tokenization status */
    if (xImg->xPage->xDoc->seqStatus & XDOC_SEQ_TOKENSTART) {

	/* If this is the case, we must save the previous tokenized
	 * images before proceeding. This is a bad state to be in,
	 * and should have already been taken care of.
	 */
	if (xImg->xPage->xDoc->seqStatus & XDOC_SEQ_TOKENDICTSAVED) {
	    XERR_PASS_CLEANUP(XIF_WriteTokenImages((XIF_DOCHANDLE)
						   xImg->xPage->xDoc));
	}
    }

    /* Tokenize the image */
    XERR_PASS_CLEANUP(wp_XifTokenizeImage(xImg, pImgData, dwBytesPerRow,
	dwFlags));

    /* Update the sequence status */
    xImg->seqStatus |= XIMG_SEQ_GETINFO | XIMG_SEQ_WRITESTART;

    /* This should always be TRUE */
    if (xImg->imgInfo.xImgType == XIF_IMGTYPE_PRIMARY) {
	xImg->xPage->seqStatus |= XPAGE_SEQ_WRITEPRIMARY;
    }	

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (imgptr) {
	XIF_FREE(imgptr);
    }
    NORMAL_RETURN;
}


/* XIF_WriteTokenImages()
 *
 * OneLineDescription:::
 *  Store all previously tokenized images, and token dictionary
 *
 * LongDescription:::
 *  XIF_WriteTokenImages() is a replacement to having to loop through
 *  XIF_WriteTokenImage(), since the list of images that have been tokenized
 *  but not resolved to a finished token dictionary are maintained via the
 *  document handle. This function will also call XIF_WriteTokenDictionary().
 *  See the description for both of these functions for further explanation.
 *
 * Input:::
 *  xDocHandle::		Valid document handle, returned from
 *				XIF_CreateDocument() or XIF_OpenDocumentWrite()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_WriteTokenImages(
	XIF_DOCHANDLE xDocHandle
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    IPWState *pIPWState;
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    /* Get shared IPCore state struct */
    XERR_PASS_CLEANUP(wp_XifIPWState(xDoc, &pIPWState));

    /* Store the symbol IDs & positions, and the residual image, for each
     * tokenized (but unresolved) image. */
    if (xDoc->seqStatus & (XDOC_SEQ_TOKENSTART)) {
	XPage *xPage;
	UInt32 page, pages;
	XImg *xImg;

	/* Write the token dictionary, if it hasn't already been written */
	XERR_PASS_CLEANUP(XIF_WriteTokenDictionary(xDocHandle));

	/* For each page that has been tokenized, store off the symbol IDs
	 * & posisions, and the residual image */
	pages = xDoc->pageTableHeader.pageCount;
	xPage = xDoc->pages;
	for (page = 0; page < pages; page++) {
	    if (xPage == NULL) {
		break;
	    }
	    if (xPage->seqStatus & (XPAGE_SEQ_TOKENSTART)) {
		xImg = (XImg *)xPage->pageInfo.xPrimImgHandle;
		XERR_PASS_CLEANUP(XIF_WriteTokenImage((UInt32)xImg));
    	    }
	    xPage = xPage->next;
	}

	/* Clear the tokenization state */
	xDoc->seqStatus &=
	    ~(XDOC_SEQ_TOKENDICTSAVED | XDOC_SEQ_TOKENSTART |
	    XDOC_SEQ_MULTIDICTS | XDOC_SEQ_SHAREDDICT);
    }

    /* Fall through to do the rest of the cleanup */
 ERR_CLEANUP:
    
    /* Free up the shared data */
    if (xDoc->textCompInfo) {
	i_freeTextCompressionInfo (pIPWState,
				   xDoc->textCompInfo); /* Ignore return */
	xDoc->textCompInfo = NULL;
    }

    ERR_RETURN;
}


/* XIF_WriteTokenImage()
 *
 * OneLineDescription:::
 *  Store a previously tokenized image
 *
 * LongDescription:::
 *  XIF_WriteTokenImage() is the partner to XIF_TokenizeImage(), which
 *  previously converted an image to tokens, and preserved it internally.
 *  This call resolves the image into token IDs and positions, and a residual
 *  image, and stores the data for both.  The token IDs and positions are
 *  stored in the token dictionary, which must be resolved (cannot be added
 *  to anymore) and is consolidated (like symbols mapped together) and stored.
 *  Then symbols from the image that repeat are converted to a compressed
 *  tokenization stream and stored. Finally, the symbols that do not repeat
 *  are retained as a separate image and stored as run-lengths (MMR), the
 *  residual image -- a SubIFD.
 *  <BR>
 *  This function should only be called after all of the binary images which
 *  are to share the same token dictionary have been tokenized. Implied is the
 *  need to store the token dictionary, which may have either taken place via
 *  a call to XIF_WriteTokenDictionary(), or is done automatically during the
 *  first call to XIF_WriteTokenImage().
 *  <BR>
 *  Internally, the token IDs are resolved (if not already from a previous call
 *  to XIF_WriteTokenImage() or XIF_WriteTokenDictionary(), and the image
 *  fragments with a count across the pages of 1 (this page only) are removed
 *  from the token dictionary, and placed back on a blank image of the original
 *  size, and stored as a residual image. The remaining tokens (with count 2 or
 *  greater), are then stored with their locations, using the Tundra format, as
 *  proposed for JBIG-2, and indices are assigned Huffman codes, or Lempel-Ziv
 *  compressed.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_WriteTokenImage(
	XIF_IMGHANDLE xImgHandle
)
{
    XImg *xImg = (XImg *)xImgHandle;
    XPage *xPage;
    PROLOGUE;

    CheckHandleValidity(xImgHandle);
    xPage = xImg->xPage;

    /* Write the token dictionary, if it hasn't already been written */
    XERR_PASS_CLEANUP(XIF_WriteTokenDictionary((UInt32)xPage->xDoc));

    /* If color tokenization is used, store off the unique symbols in
     * a separate dictionary */
    if (xPage->seqStatus & XPAGE_SEQ_TKCOLORADDED) {
	XERR_PASS_CLEANUP(wp_XifWriteTokenDictionary(xPage->xDoc, xPage));
    }

    /* Store the symbol IDs & positions, and the residual image */
    XERR_PASS_CLEANUP(wp_XifWriteTokenImage(xImg));

    /* Clear the tokenization state */
    xImg->seqStatus &= ~XIMG_SEQ_TOKENSTART;
    xPage->seqStatus &= ~XPAGE_SEQ_TOKENSTART;

    /* If appropriate mark as completely written */
    if (xPage->seqStatus & XPAGE_SEQ_WRITEPENDING) {
	xPage->seqStatus &= ~XPAGE_SEQ_WRITESTATUS;
	xPage->seqStatus |= XPAGE_SEQ_WRITEFINISH;
	xPage->seqStatus |= XPAGE_SEQ_READSTART;	/* Mark as readable */
    }

    /* Check to see if it was the last page to be stored for the given
     * dictionary */
    if (xPage->xDoc->seqStatus & (XDOC_SEQ_TOKENSTART)) {
	XPage *pxPage;
	UInt32 page, pages;
	Bool pagesLeft = 0;

	/* Check each page */
	pages = xPage->xDoc->pageTableHeader.pageCount;
	pxPage = xPage->xDoc->pages;
	for (page = 0; page < pages; page++) {
	    if (pxPage == NULL) {
		break;
	    }
	    if (pxPage->seqStatus & (XPAGE_SEQ_TOKENSTART)) {
		pagesLeft++;
		break;
    	    }
	    pxPage = pxPage->next;
	}

	/* If all the tokenized images have been stored, reallow tokenization
	 */
	if (pagesLeft == 0) {
	    /* Clear the tokenization state */
	    xPage->xDoc->seqStatus &=
		~(XDOC_SEQ_TOKENDICTSAVED | XDOC_SEQ_TOKENSTART |
		XDOC_SEQ_MULTIDICTS | XDOC_SEQ_SHAREDDICT);
	}
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/* XIF_WriteTokenDictionary()
 *
 * OneLineDescription:::
 *  Store the dictionary of accumulated tokens
 *
 * LongDescription:::
 *  XIF_WriteTokenDictionary() resolves a minimal set of token dictionary
 *  symbols by remapping similar symbols into a fairly distinct set, and
 *  stores the resulting token dictionary using the Tundra format, as
 *  proposed for JBIG-2.
 *  <BR>
 *  The token dictionary will consist of all repeated symbols found during
 *  the calls to XIF_TokenizeImage() for all images passed in either since
 *  the last call to XIF_WriteTokenImage(), XIF_WriteTokenDictionary() or
 *  XIF_WriteTokenImages().
 *  
 * Input:::
 *  xDocHandle::		Document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_WriteTokenDictionary(
	XIF_DOCHANDLE xDocHandle
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    if ((xDoc->seqStatus & (XDOC_SEQ_TOKENSTART)) == 0) {
	NORMAL_RETURN;	/* No tokenization ongoing */
    }
    if (xDoc->seqStatus & (XDOC_SEQ_TOKENDICTSAVED)) {
	NORMAL_RETURN;	/* Already saved */
    }

    /* If the document has an open token dictionary, create a consolidation
     * mapping table */
    XERR_PASS_CLEANUP(wp_XifConsolidateTokens(xDoc));

    /* Write the consolidated dictionary */
    XERR_PASS_CLEANUP(wp_XifWriteTokenDictionary(xDoc, NULL));

    /* Mark as stored (you cannot tokenize after this) */
    xDoc->seqStatus |= XDOC_SEQ_TOKENDICTSAVED;

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/* XIF_SetAuthor()
 *
 * OneLineDescription:::
 *  Set (write or rewrite) the authoring block
 *
 * LongDescription:::
 *  XIF_SetAuthor() sets the authoring information block. If the authoring block
 *  already exists in the file, the block is rewritten to the new block. The data
 *  will overwrite the previous block if the old block was as big or bigger than
 *  the new block, otherwise it is appended to the file.
 *
 * Input:::
 *  xDocHandle::		Document handle
 *  pAuthorBlock::		Pointer to the authoring block of data
 *  wAuthorBytes::		Number of bytes in authoring block
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_SetAuthor(
	XIF_DOCHANDLE xDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    XERR_PASS(wp_XifSetAuthor(xDoc, pAuthorBlock, wAuthorBytes));

    NORMAL_RETURN;
} 


/* XIF_SetReferencePage()
 *
 * OneLineDescription:::
 *  Set (write or rewrite) the reference page offset
 *
 * LongDescription:::
 *  XIF_SetReferencePage() sets the reference page offset. If the reference
 *  page is already stored in the file, it is overwritten.
 *
 * Input:::
 *  xDocHandle::		Document handle
 *  dwReferencePageOffset::	Reference page offset
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_SetReferencePage(
	XIF_DOCHANDLE xDocHandle,
	Int32 dwReferencePageOffset
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    XERR_PASS(wp_XifSetReferencePage(xDoc, dwReferencePageOffset));

    NORMAL_RETURN;
} 


/* XIF_SetAnnotTableEntryCount()
 *
 * OneLineDescription:::
 *  Set the number of entries each annotation table will contain.
 *
 * LongDescription:::
 *  XIF_SetAnnotTableEntryCount() sets the number of annotation table entries
 *  to be used, before the annotation table header is actually written. This
 *  routine can be called at any reasonable time, but the count passed in will
 *  only apply to new annotation table headers created, and only when an
 *  annotation is actually saved.
 *
 * Input:::
 *  xDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_SetAnnotTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    PROLOGUE;

    CheckHandleValidity(xDocHandle);
    
    /* If version 2, there is no annotation table */
    if (xDoc->docInfo.version < 3) {
	NORMAL_RETURN;
    }

    /* For a minimum value */
    if (wEntriesPerTable < 3) {
	wEntriesPerTable = 3;
    }      

    /* Will be used in any new annotation tables written later */
    xDoc->annotsPerTable = wEntriesPerTable;

    NORMAL_RETURN;
}


/* XIF_SetPageTableEntryCount()
 *
 * OneLineDescription:::
 *  Set the number of entries each page table will contain.
 *
 * LongDescription:::
 *  XIF_SetPageTableEntryCount() sets the number of page table entries
 *  to be used, before the page table header is actually written. It is
 *  best to call this routine just after the call to XIF_CreateDocument(),
 *  although you should be able to call it any time before the first page
 *  is actually started.
 *
 * Input:::
 *  xDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_SetPageTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    PROLOGUE;

    CheckHandleValidity(xDocHandle);
    
    /* For a minimum value */
    if (wEntriesPerTable < 10) {
	wEntriesPerTable = 10;
    }      

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(xDoc->seqStatus & XDOC_SEQ_W_PAGETABLE, XIF_SEQUENCE);

    /* If the file exists, we cannot go back and change the value...
     * unless it was never set in the first place */
    XERR_GEN_CLEANUP((xDoc->readWrite & READ_ALLOWED) &&
		     xDoc->pageTableHeader.entriesPerTable != 0, XIF_NOSUPPORT);

    xDoc->pageTableHeader.entriesPerTable = wEntriesPerTable;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_SetSymlibTableEntryCount()
 *
 * OneLineDescription:::
 *  Set the number of entries each symbol library table will contain.
 *
 * LongDescription:::
 *  XIF_SetSymlibTableEntryCount() sets the number of symbol library table entries
 *  to be used, before the symbol library table header is actually written. It is
 *  best to call this routine just after the call to XIF_CreateDocument(),
 *  although you should be able to call it any time before the first token
 *  dictionary is actually written (up to the call to XIF_WriteTokenImage(),
 *  XIF_WriteTokenDictionary(), XIF_WriteTokenImages(), or
 *  XIF_CloseDocument()).
 *
 * Input:::
 *  xDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_SetSymlibTableEntryCount(
	XIF_DOCHANDLE xDocHandle,
	UInt16 wEntriesPerTable
)
{
    XDoc *xDoc = (XDoc *)xDocHandle;
    PROLOGUE;

    CheckHandleValidity(xDocHandle);

    /* For a minimum value */
    if (wEntriesPerTable < 2) {
	wEntriesPerTable = 2;
    }      
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(xDoc->seqStatus & XDOC_SEQ_W_SYMLIBTABLE, XIF_SEQUENCE);

    /* If the file exists, we cannot go back and change the value...
     * unless it was never set in the first place */
    XERR_GEN_CLEANUP((xDoc->readWrite & READ_ALLOWED) &&
		     xDoc->symlibTableHeader.entriesPerTable != 0, XIF_NOSUPPORT);

    xDoc->symlibTableHeader.entriesPerTable = wEntriesPerTable;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* XIF_AddAnnotation()
 *
 * OneLineDescription:::
 *  Add an annotation to a page
 *
 * LongDescription:::
 *  XIF_AddAnnotation() adds an annotation to a page
 *
 * Input:::
 *  xPageHandle::		Active page handle
 *  wAnnotType::		Type of annotation
 *  dwXOffset::			Offset in X direction in page coordinates
 *  dwYOffset::			Offset in Y direction in page coordinates
 *  dwWidth::			Width in page coordinates
 *  dwHeight::			Height in page coordinates
 *  pBuf::			Annotation data
 *  dwAnnotBytes::		Length of annotation data, in bytes
 *
 * Output:::
 *  pdwAnnotID::		Annotation ID assigned to this annotation
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_AddAnnotation(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 wAnnotType,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwFlags,
	UInt8 *pBuf,
	UInt32 dwAnnotBytes,
	UInt16 *pdwAnnotID
)
{
    XPage *xPage = (XPage *)xPageHandle;
    AnnotTable *annotTable;
    PROLOGUE;

    CheckHandleValidity(xPageHandle);

    /* Make sure that we've read the appropriate table that the annotation
     * will be attached to */
    if ((xPage->xDoc->readWrite & READ_ALLOWED) &&
	xPage->annotTableHeader.annotCount > 0) {
	XERR_PASS(wp_XifGetAnnotTable(xPage,
	    (UInt16)(xPage->annotTableHeader.annotCount - 1), &annotTable));
    }

    /* Now add the annotation */
    XERR_PASS(wp_XifAddAnnotation(xPage, wAnnotType, dwXOffset, dwYOffset,
	dwWidth, dwHeight, dwFlags, pBuf, dwAnnotBytes, pdwAnnotID));

    NORMAL_RETURN;
} 


/* XIF_DeleteAnnotation()
 *
 * OneLineDescription:::
 *  Deletes an annotation from a page
 *
 * LongDescription:::
 *  XIF_DeleteAnnotation() deletes an annotation from a page. Note that the
 *  remaining subsequent annotations will automatically shift down one
 *  position.
 *
 * Input:::
 *  xPageHandle::		Page handle, returned from XIF_GetPageInfo()
 *  dwAnnotOffset::		Offset (or ID) of annotation to delete (0, 1, ...)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_DeleteAnnotation(
	XIF_PAGEHANDLE xPageHandle,
	UInt16 dwAnnotOffset
)
{
    XPage *xPage = (XPage *)xPageHandle;
    UInt16 oldCount = 0, newCount = 0;
    PROLOGUE;

    /* Check page handle */
    CheckHandleValidity(xPageHandle);

    /* Get the old annotation count */
    oldCount = xPage->annotTableHeader.annotCount;

    /* Make sure there's one to delete, and that the requested one is there */
    XERR_GEN_CLEANUP(dwAnnotOffset >= oldCount, XIF_BADPARAMETER);

    /* Delete the annotation (just remove the reference, data is left in
     * the file) */
    XERR_PASS_CLEANUP(wp_XifDeleteAnnotation(xPage, dwAnnotOffset, &newCount));

    XERR_GEN_CLEANUP(newCount != oldCount - 1, XIF_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(newCount != xPage->annotTableHeader.annotCount,
	XIF_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(newCount != xPage->pageInfo.wAnnotationCount,
	XIF_INTERNAL_ERROR);

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


/* XIF_Compress()
 *
 * OneLineDescription:::
 *  Compress a strip of image data
 *
 * LongDescription:::
 *  XIF_Compress() compresses a single image strip. Used as a standalone
 *  compressor, allows an application to retrieve the image data from file
 *  in raw compressed form, and decode individual strips on the fly as needed.
 *  <BR>
 *  Source image is compressed into the destination buffer. If the destination
 *  buffer is filled and would overflow, XIF_BUFFEREXCEEDED is returned.
 *
 * Input:::
 *  xImgHandle::		Valid image handle, as returned from
 *				XIF_AddImage()
 *  xCompression::		See XIF_COMPRESSION types
 *  pdwQuality::		Pointer to desired quality (actual quality
 *				returned), see XIF_ImageWriteStart() for more
 *  xColorType::		Type of color
 *  dwFlags::			XIF_IMGFLAGS_xxx combo
 *  pSrc[3]::			Source buffer channels
 *  dwBytesPerRow::		Bytes per row in image buffer
 *  dwRowOffset::		Row offset in image of data to compress
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwHeight::			Rows in dst image
 *
 * Output:::
 *  ppDst::			Pointer to the destination compressed data buffer
 *  pdwBytesInDst::		Pointer to the byte count of destination buffer
 *  pdwBytesFilled::		Actual bytes in destination filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_Compress(
	XIF_IMGHANDLE xImgHandle,
	XIF_COMPRESSION xCompression,
	UInt32 *pdwQuality,
	XIF_COLORTYPE xColorType,
	UInt32 dwFlags,
	UInt8 *pSrc[3],
	UInt32 dwBytesPerRow,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt8 **ppDst,
	UInt32 *pdwBytesInDst,
	UInt32 *pdwBytesFilled
)
{
    XImg *xImg = (XImg *)xImgHandle;
    ScaledInt1616 xres, yres;
    Int32 rowsLeft;
    Int32 pixrBpl;
    UInt32 bytesPerRow;
    IPWState *pIPWState;
    PIXR *pPixr = NULL, *pSubPixr;
    ColorType ctype = cColorNone;
    UInt8 *pUserImage, *pPixrImage;
    Int32 channel, channels;
    PROLOGUE;

    *pdwBytesFilled = 0UL;
    CheckHandleValidity(xImgHandle);

    /* Borrow the quality passed in */
    xImg->JPEGquality = *pdwQuality;

    switch ((int)xCompression) {
    case XIF_JPEG:
    case XIF_WAVELET:
	if (xImg->JPEGquality == 0) {
	    xImg->JPEGquality = 75;	/* The default */
	} else if (xImg->JPEGquality > 100) {
	    xImg->JPEGquality = 100;	/* The max */
	} else if (xImg->JPEGquality < 1) {
	    xImg->JPEGquality = 1;	/* My hard-coded min */
	}
	break;
    default:
	xImg->JPEGquality = 0;
    }

    /* Get an pointer to the IPState, it's needed below */
    XERR_PASS (wp_XifIPWState (xImg->xPage->xDoc, &pIPWState));

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    /* Check buffers and sizes */
    XERR_GEN_CLEANUP(pSrc == NULL || pSrc[0] == NULL || ppDst == NULL ||
	dwBytesPerRow == 0, XIF_BADPARAMETER);
    XERR_GEN_CLEANUP(dwRowOffset >= xImg->imgInfo.dwHeight, XIF_BADPARAMETER);
    
    /* Create a PIXR image to hold the image */
    xres = Int32_TO_ScaledInt1616(xImg->imgInfo.dwXResolution);  /* Any resolution is fine, I guess */
    yres = Int32_TO_ScaledInt1616(xImg->imgInfo.dwYResolution);  /* Any resolution is fine, I guess */
    switch (xColorType) {
    case XIF_COLORTYPE_COLOR24:
	ctype = cColorRGB;
	break;
    case XIF_COLORTYPE_BINARY:
    case XIF_COLORTYPE_GRAY4:
    case XIF_COLORTYPE_COLOR4:
    case XIF_COLORTYPE_GRAY8:
    case XIF_COLORTYPE_COLOR8:
	ctype = cColorGray;
	break;
    default:
	break;
    }
    channels = XIF_CHANNELS(xColorType);
    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, dwWidth, dwHeight,
	XIF_BPPE(xColorType), xres, yres, 32, channels, ctype, NULL, &pPixr));

    /* Copy the data to the user's buffer */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &pixrBpl));
    bytesPerRow = dwBytesPerRow;
    if (bytesPerRow > pixrBpl) {
	bytesPerRow = pixrBpl;
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &pPixrImage));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &pPixrImage));
	}

	rowsLeft = dwHeight;
	pUserImage = pSrc[channel] + dwRowOffset * dwBytesPerRow;
	while (rowsLeft > 0) {
	    memcpy(pPixrImage, pUserImage, bytesPerRow);
	    pUserImage += dwBytesPerRow;
	    pPixrImage += pixrBpl;
	    rowsLeft--;
	}
    }

    /* Call the underlying compression routine */
    XERR_PASS_CLEANUP(w_XifCompress(pIPWState, xImgHandle, xCompression,
	xColorType, dwFlags, pPixr, 0, dwWidth, dwHeight, ppDst, pdwBytesInDst,
	pdwBytesFilled));
    
    w_destroyPixr(pIPWState, &pPixr);

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    w_destroyPixr(pIPWState, &pPixr);
    
    ERR_RETURN;
}

/*
$Log: xifwrite.c,v $
Revision 1.2  2000/04/07 18:56:15  jlin
Add fix for color xif

 * Revision 1.1.1.1  1999/09/20  18:19:22  jlin
 * Imported sources
 *
Revision 1.58  1998/08/04 00:21:37  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.57  1998/07/30 22:39:46  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.56  1998/05/11 18:25:00  aber
aber on Mon May 11 11:24:09 PDT 1998

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Do not propogate the source layer number when copying images compressed.
Allow the user specified image layer to be assigned.

Revision 1.55  1998/05/05 17:22:40  aber
aber on Tue May 5 10:22:11 PDT 1998

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Added a parameter to XIF_Compress to allow changing the JPEG/Wavelet
quality.

Revision 1.54  1998/05/05 07:15:29  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Additions necessary to fill in the never completed functions
XIF_Compress and XIF_Decompress.

Revision 1.53  1998/04/06 22:43:47  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Bug fix: check resolution difference when checking the image size of
the image during read.
Also, check for unanticipated condition of the page resolution being
SMALLER than than the subimage resolution. Adjust subimage dimensions
accordingly.

Revision 1.52  1998/04/02 00:09:52  aber
aber on Wed Apr 1 16:09:51 PST 1998

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Allow attaching a transparency mask to the primary image, special request.

Revision 1.51  1998/03/25 23:25:17  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Use a unique document ID for each document handle. This allows tagging
each document so that during "quick copy", the token dictionaries are
copy only once, and only once (in the case of two destination documents
somehow ending up with the same memory location for the structure data).

Revision 1.50  1998/02/24 01:41:35  aber
aber on Mon Feb 23 17:40:25 PST 1998

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Fix bounding problem, where the FG/BG image could be larger than the
selector image (accounting for the resolution differences). Adjust
the FG/BG image size, rather than allow the X/Y position to go negative.

Revision 1.49  1998/01/20 03:23:55  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Expand read/write mode to do a create, if the file doesn't already exist.
Fix read/write problems. Proto update for MixedRasterContent renaming.

Revision 1.48  1998/01/16 00:43:12  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Allow page inserts.
To be done later: the page table is regenerated from the linked-list
of pages. This should really be regenerated from the page table itself.

Revision 1.47  1998/01/14 00:16:40  ddavies
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.46  1997/12/18 00:19:24  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Keep track of when a page is completed (pending resolved tokenization),
and is completely written (able to read back).
Add new function XIF_DeletePage() (unwritten) and XIF_DeleteImage().

Revision 1.45  1997/11/27 00:59:56  aber
aber on Wed Nov 26 16:59:39 PST 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Fix PC complaints.

Revision 1.44  1997/11/26 23:26:58  aber
aber on Wed Nov 26 15:26:07 PST 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Remove stipulation that a mask image must mask a non-binary image.
Force tokenization of primary image ONLY.

Revision 1.43  1997/11/25 01:20:37  aber
aber on Mon Nov 24 17:20:08 PST 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Reset the tokenization state after the last tokenized image is
stored; this allows tokenization to start all over again with
a fresh token dictionary.

Revision 1.42  1997/11/24 21:12:05  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Add new function XIF_CheckImageInfo() that is used to localize the
requirements of the supported compression types, and the strip height
that matches the XIF version and compression type.
This allows the check to be made before XIF_CopyImage() is called,
so that a strip size change will force recompression, instead of
writing a bad strip height out (say when going from a XIF 3.0 to 2.0,
the strip height of a JPEG image MUST be the image height).

Revision 1.41  1997/11/20 02:54:29  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Resolve (once again) that XIF 2.0 files cannot have multi-tiled JPEG images.

Revision 1.40  1997/11/19 20:23:53  aber
aber on Wed Nov 19 12:22:54 PST 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Bug fix: when adding an image to a linked list, order by both image
layer, then order within the layer.

Revision 1.39  1997/11/18 23:46:34  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Add new parameter to XIF_CopyPage() to allow (disable) annotation
copy.
Add new function XIF_WriteTokenImages() to loop through the pages that
have been tokenized but not written yet. Automatically store the token
dictionary (if not stored yet).
Add flags parameter to annotation functions.

Revision 1.38  1997/11/17 19:26:02  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Allow negative page offsets to automatically append the page to
the end of the document.

Revision 1.37  1997/11/14 02:09:03  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.36  1997/11/11 18:12:43  aber
aber on Tue Nov 11 10:12:05 PST 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

In XIF_CopyImage(), assign transparency mask image offset to be 0, even
though it's not used.

Revision 1.35  1997/11/07 01:59:49  aber
aber on Thu Nov 6 17:58:50 PST 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Keep track of rows buffered, not rows written, when buffering data out.

Revision 1.34  1997/11/07 01:13:09  aber
aber on Thu Nov 6 17:12:46 PST 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Fix logic for new function: XIF_CopyImage().

Revision 1.33  1997/11/06 22:52:17  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

New function added: XIF_CopyImage(). This allows copying an individual
image compressed. Formerly only the entire page could be copied as a whole,
while the individual images were copied compressed. This allows an editor
(like Pagis) to copy selected images compressed, then copy over the remaining
edited images over with edits.
This is the "rough draft" of this function, it may need work (I need a
test environment).

Revision 1.32  1997/11/01 20:59:59  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Remove stipulation that wavelet compression requires the full image.
Add strip buffering for wavelet compression.

Revision 1.31  1997/10/24 21:39:00  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.30  1997/10/16 19:04:50  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Allow GetImageInfo any time after AddImage.

Revision 1.29  1997/10/15 21:36:52  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Modify parameters to XifSetPageIFD. Update page count in document during
page add (wasn't being updated at all).

Revision 1.28  1997/10/01 23:41:29  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Recommend strip height a multiple of 8 when 4/8-bit gray/color.
Always check that the strip is not larger than the image.
Don't force JPEG compression when NOC (uncompressed) requested.
Do recommend a full-height strip when using JPEG on a XIFF 2.0 file,
but allow 3.0 files to vary.

Revision 1.27  1997/09/19 00:17:34  aber
aber on Thu Sep 18 17:15:37 PDT 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

When writing, allow XIF_GetPageInfo() call.
Set minimum entries per table for page (10), symlib (2), and annotation
(3).

Revision 1.26  1997/08/29 17:27:33  aber
aber on Fri Aug 29 10:27:01 PDT 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Add 3 new functions to preset the size of the tables: annot table,
page table, and symlib table.
Adjust parameter types to match what's actually stored in the file.

Revision 1.25  1997/08/27 00:17:10  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Update strip sizes to be 64k (instead of whole image, or 8k TIFF
recommendation), and round up to the next multiple of 16. Don't
let it exceed the image height.

Revision 1.24  1997/08/15 00:11:12  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Slight code cleanup.

Revision 1.23  1997/08/13 22:48:20  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Edit fix via Pedro (XResolution should be Yresolution).

Revision 1.22  1997/07/30 01:08:57  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

If strip height not specified by client, use TIFF 8K computation. If JPEG
or Wavelet, force strip to be the entire image (really bad compression if
we use the 8K rule). Mark tokenized images as having image characteristics
set.

Revision 1.21  1997/07/24 00:37:29  aber
aber on Wed Jul 23 17:33:36 PDT 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Add check for existance of a mask when adding one (one shouldn't
already exist).

Revision 1.20  1997/07/22 22:43:04  aber
aber on Tue Jul 22 15:42:10 PDT 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Append the page to the document when adding page zero,
carries over from XFILE.

Revision 1.19  1997/07/15 23:44:13  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Change some parameters to isolate XIF_ImageWriteStrip from w_XifImageWriteStrip.

Revision 1.18  1997/07/14 23:23:59  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

If no recommended strip size, use the TIFF recommendation of 8K.

Revision 1.17  1997/07/10 22:02:30  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Use 16-bit ID for annotations. Support annotation deletion.

Revision 1.16  1997/07/02 01:06:48  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Add sequence checks for tokenization.

Revision 1.15  1997/06/26 22:32:10  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

If the XIF 2.0 file is to store a JPEG image, it must be all one tile.
Override the tile height to include the entire image (to be buffered).

Revision 1.14  1997/06/23 22:59:39  aber
aber on Mon Jun 23 15:59:15 PDT 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Update function call to match prototype update for palette support.

Revision 1.13  1997/06/23 22:48:08  aber
aber on Mon Jun 23 15:48:08 PDT 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Sync up image pointers so that lossy dither works right.

Revision 1.12  1997/06/20 17:49:13  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Added comments. Rethink the control parameters to strip buffering.

Revision 1.11  1997/06/17 22:36:28  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Check new flag bit for tokenization on. Define renames.

Revision 1.9  1997/06/10 01:27:00  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.7  1997/05/30 22:12:44  aber
aber on Fri May 30 15:12:32 PDT 1997

xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

Variable rename to match documentation.

Revision 1.6  1997/05/30 22:11:34  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Update documentation for web browsing. XIF_ functions are left
with no "BeginDocumentation" tag, since w_ and wp_ functions are expected.

Revision 1.5  1997/05/30 00:51:18  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:21  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:29  aber
xifwrite.c -> /product/ipcore/ipwrap/src/RCS/xifwrite.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:47  aber
Initial revision

*/
