/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwtiff.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifwtiff.c:	WRITE TIFF source
 *
 *	TIFF write-only routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xiff.h"
#include "xifwpriv.h"
#include "xifwtiff.h"
#include "xifwxiff.h"
#include "xiftiff.h"
#include "xifrtiff.h"
#include "xifrxiff.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwtiff.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:22 $");


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write the TIFF header.
 *
 * LongDescription:::
 *  wp_TifWriteHeader() writes the TIFF header.
 *
 * Input:::
 *  pxFile::			Pointer to XIFF file handle
 *  dwFirstIfd::		Offset of first IFD
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifWriteHeader(
    XFile *pxFile,
    UInt32 dwFirstIfd
)
{
    TiffHeader header;
    UInt8 cr[1] = {10};
    Int32 fileSize;
    PROLOGUE;

    /* We're gunna have some fun first. Write out a carriage-return, and
     * see if only one byte was written */
    XIF_FILESIZE(pxFile, fileSize);
    XERR_GEN_CLEANUP(fileSize != 0L, XIF_BADFILEFORMAT);    /* File not empty */
    XIF_SEEK(pxFile, 0UL);
    XIF_WRITE(pxFile, cr, 1);
    XIF_FILESIZE(pxFile, fileSize);
    XERR_GEN_CLEANUP(fileSize == 0L, XIF_WRITE_ERR);
    XERR_GEN_CLEANUP(fileSize == 2L, XIF_BADFILEMODE);

    /* Create the header */
    header.byteOrder = NATIVEORDER;
    header.version = 42;
    header.firstIfd = dwFirstIfd;
    
    /* Write the TIFF header */
    XIF_SEEK(pxFile, 0UL);
    XIF_WRITE(pxFile, &header.byteOrder, sizeof(header.byteOrder));
    XIF_WRITE(pxFile, &header.version, sizeof(header.version));
    XIF_WRITE(pxFile, &header.firstIfd, sizeof(header.firstIfd));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write the TIFF fields
 *
 * LongDescription:::
 *  wp_TifWriteFields() write the TIFF fields necessary to describe the XIFF file.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *  pFieldCount::			Pointer to the number of fields written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifWriteFields(
    XImg *xImg,
    UInt16 *pFieldCount
)
{
    Instance *instData;
    XFile *xFile;
    XIF_IMGINFO *pxImgInfo = NULL;
    UInt8 paletted = 0;
    UInt8 bpp;
    UInt8 bppe;
    UInt8 channels = 1;
    XIF_COMPRESSION compression;
    UInt16 sval;
#ifdef USE_T6OPTIONS
    UInt16 svals[2];
#endif
    UInt32 lval;
    UInt32 lvals[4];
    Int32 dataOffset;
    TiffFields *fieldList;
    Rational rval;
    UInt32 version;

    PROLOGUE;
    
    CheckHandleValidity(xImg);
    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);
    pxImgInfo = &xImg->imgInfo;
    fieldList = &xImg->fields;
    version = xImg->xPage->xDoc->docInfo.version;

    /* Use default values */
    fieldList->SamplesPerPixel = 1;
    xImg->minBytesPerRow = (xImg->imgInfo.dwWidth *
		    XIF_BPPE(xImg->imgInfo.xColorType) + 7) / 8;
    
    /* Start with no fields */
    *pFieldCount = 0;

    XERR_GEN_CLEANUP((pxImgInfo->dwBestStripHeight == 0),
		     XIF_INTERNAL_ERROR);

    /* Determine mode */
    compression = pxImgInfo->xCompression;
    bpp = XIF_BPP(pxImgInfo->xColorType);
    bppe = XIF_BPPE(pxImgInfo->xColorType);
    channels = bpp / bppe;
    if (pxImgInfo->xColorType == XIF_COLORTYPE_COLOR4 ||
	pxImgInfo->xColorType == XIF_COLORTYPE_COLOR8) {
	paletted = 1;
    }
    
#ifdef DEBUG
    printf("\tField\tID\tType\tLength\tOffset\n");
#endif

    /* Write the NewSubFileType field, if appropriate */
    fieldList->NewSubFileType = 0;
    if (pxImgInfo->xImgType == XIF_IMGTYPE_PRIMARY) {
	fieldList->NewSubFileType |= XTIFF_SUBFILE_MULTIPAGE;
    }
    if (pxImgInfo->xImgType != XIF_IMGTYPE_MASK) {
	if (version >= 3) {
	    fieldList->NewSubFileType |= XTIFF_SUBFILE_MRC;
	}
    } else {
	fieldList->NewSubFileType |= XTIFF_SUBFILE_MASK;
    }
    lval = (UInt32)fieldList->NewSubFileType;
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, NEWSUBFILETYPE, IFDLONG,
	1, &dataOffset, &lval, pFieldCount));

    /* Write the ImageWidth field, if appropriate */
    lval = fieldList->ImageWidth = pxImgInfo->dwWidth;
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, IMAGEWIDTH, IFDLONG,
	1, &dataOffset, &lval, pFieldCount));

    /* Write the ImageLength field, always */
    lval = fieldList->ImageLength = pxImgInfo->dwHeight;
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, IMAGELENGTH, IFDLONG,
	1, &dataOffset, &lval, pFieldCount));

    /* Write the BitsPerSample field, always */
    CHECK_MALLOC(fieldList->BitsPerSample =
	XIF_MALLOC(channels * FieldTypeSize[IFDSHORT]));
    for (sval = 0; sval < channels; sval++) {
	fieldList->BitsPerSample[sval] = bppe;
    }
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, BITSPERSAMPLE, IFDSHORT,
	channels, &dataOffset, fieldList->BitsPerSample, pFieldCount));

    /* Write the Compression field, always */
    switch ((int)pxImgInfo->xCompression) {
    case XIF_NOC:
	fieldList->Compression = XTIFF_NOCOMPRESSBYTE;
	break;
    case XIF_G4:
	fieldList->Compression = XTIFF_FAXCCITT4;
	break;
    case XIF_JPEG:
	fieldList->Compression = XTIFF_JPEG;
	break;
    case XIF_TOKENIZATION:
	fieldList->Compression = XTIFF_XEROX_TOKENIZATION;
	break;
    case XIF_LLTOKENIZATION:
	fieldList->Compression = XTIFF_XEROX_LLTOKENIZATION;
	break;
    case XIF_WAVELET:
	fieldList->Compression = XTIFF_XEROX_WAVELET;
	break;
    case XIF_LOSSYDITHER:
	fieldList->Compression = XTIFF_XEROX_LOSSYDITHER;
	break;
    default:
	__error__ = XIF_BADPARAMETER;
	goto ERR_CLEANUP;
    }
    sval = fieldList->Compression;
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, COMPRESSION, IFDSHORT,
	1, &dataOffset, &sval, pFieldCount));

    /* Write the Photometric field, always */
    if (pxImgInfo->xCompression == XIF_JPEG) {
	fieldList->PhotometricInterpretation = XTIFF_PHOTOMETRIC_YCBCR;
    } else if (paletted) {
	fieldList->PhotometricInterpretation = XTIFF_PHOTOMETRIC_PALETTE;
    } else if (pxImgInfo->xImgType == XIF_IMGTYPE_MASK) {
	fieldList->PhotometricInterpretation = XTIFF_PHOTOMETRIC_TRANSPARENCY;
    } else if (pxImgInfo->xColorType == XIF_COLORTYPE_COLOR24) {
	fieldList->PhotometricInterpretation = XTIFF_PHOTOMETRIC_RGB;
    } else if (bppe == 1) {
	fieldList->PhotometricInterpretation = XTIFF_PHOTOMETRIC_WHITEISZERO;
    } else {
	fieldList->PhotometricInterpretation = XTIFF_PHOTOMETRIC_BLACKISZERO;
    }
    sval = fieldList->PhotometricInterpretation;
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, PHOTOMETRICINTERPRETATION, IFDSHORT,
	1, &dataOffset, &sval, pFieldCount));

    /* Write the StripOffset field if it's not JPEG */
    if (pxImgInfo->xCompression != XIF_JPEG) {

	/* Compute the number of strips */
	fieldList->StripsPerImage = (pxImgInfo->dwHeight +
	    pxImgInfo->dwBestStripHeight - 1) / pxImgInfo->dwBestStripHeight;
	CHECK_MALLOC(fieldList->StripOffsets =
	    XIF_MALLOC(fieldList->StripsPerImage * FieldTypeSize[IFDLONG]));
	memset(fieldList->StripOffsets, 0, FieldTypeSize[IFDLONG] *
	    fieldList->StripsPerImage);
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, STRIPOFFSETS, IFDLONG,
	    fieldList->StripsPerImage, &dataOffset, fieldList->StripOffsets, pFieldCount));

	/* Store the location of the strip offsets */
	xImg->offsetsLocation = dataOffset;
    }

    /* Write the SamplesPerPixel field if it is RGB */
    if (channels > 1) {
	sval = fieldList->SamplesPerPixel = channels;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, SAMPLESPERPIXEL, IFDSHORT,
	    1, &dataOffset, &sval, pFieldCount));
    }

    /* Write the RowsPerStrip field if it's not JPEG */
    if (pxImgInfo->xCompression != XIF_JPEG) {
	sval = (UInt16)(fieldList->RowsPerStrip = pxImgInfo->dwBestStripHeight);
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, ROWSPERSTRIP, IFDSHORT,
	    1, &dataOffset, &sval, pFieldCount));
    }

    /* Write the StripByteCounts field if it's not JPEG */
    if (pxImgInfo->xCompression != XIF_JPEG) {
	CHECK_MALLOC(fieldList->StripByteCounts =
	    XIF_MALLOC(fieldList->StripsPerImage * FieldTypeSize[IFDLONG]));
	memset(fieldList->StripByteCounts, 0, FieldTypeSize[IFDLONG] *
	    fieldList->StripsPerImage);
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, STRIPBYTECOUNTS, IFDLONG,
	    fieldList->StripsPerImage, &dataOffset, fieldList->StripByteCounts,
	    pFieldCount));

	/* Store the location of the strip byte count */
	xImg->byteCountsLocation = dataOffset;
    }

    /* Write the XResolution field, always */
    rval.numerator = fieldList->XResolution.numerator = pxImgInfo->dwXResolution;
    rval.denominator = fieldList->XResolution.denominator = 1;
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, XRESOLUTION, IFDRATIONAL,
	1, &dataOffset, &rval, pFieldCount));

    /* Write the YResolution field, always */
    rval.numerator = fieldList->YResolution.numerator = pxImgInfo->dwYResolution;
    rval.denominator = fieldList->YResolution.denominator = 1;
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, YRESOLUTION, IFDRATIONAL,
	1, &dataOffset, &rval, pFieldCount));

    /* Write the PlanarConfiguration if it's JPEG */
    if (pxImgInfo->xCompression == XIF_JPEG) {
	sval = (UInt16)(fieldList->PlanarConfiguration = XTIFF_PLANAR_CHUNKY);
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, PLANARCONFIGURATION, IFDSHORT,
	    1, &dataOffset, &sval, pFieldCount));
    }

    /* Write the XPosition field, if it's not the primary IFD */
    if (pxImgInfo->xImgType != XIF_IMGTYPE_PRIMARY) {
	rval.numerator = fieldList->XPosition.numerator = pxImgInfo->dwXOffset;
        rval.denominator = fieldList->XPosition.denominator =
	    ((XImg *)(xImg->xPage->pageInfo.xPrimImgHandle))->imgInfo.dwXResolution;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, XPOSITION, IFDRATIONAL,
	    1, &dataOffset, &rval, pFieldCount));
    }

    /* Write the YPosition field, if it's not the primary IFD */
    if (pxImgInfo->xImgType != XIF_IMGTYPE_PRIMARY) {
	rval.numerator = fieldList->YPosition.numerator = pxImgInfo->dwYOffset;
        rval.denominator = fieldList->YPosition.denominator =
	    ((XImg *)(xImg->xPage->pageInfo.xPrimImgHandle))->imgInfo.dwYResolution;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, YPOSITION, IFDRATIONAL,
	    1, &dataOffset, &rval, pFieldCount));
    }

#ifdef USE_T6OPTIONS
    /* Write the T6Options field, if it's G4 compressed */
    if (pxImgInfo->xCompression == XIF_G4) {
	lval = 0;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, T6OPTIONS, IFDLONG,
	    1, &dataOffset, &lval, pFieldCount));
    }

    /* Write the PageNumber field if it's the primary IFD */
    if (pxImgInfo->xImgType == XIF_IMGTYPE_PRIMARY) {
	svals[0] = xImg->xPage->pageInfo.dwPageOffset + 1;
	svals[1] = 0;	/* Don't know the page count yet */
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, PAGENUMBER, IFDSHORT,
	    2, &dataOffset, svals, pFieldCount));
    }
#endif

    /* Write the ResolutionUnit field, always */
    sval = (UInt16)(fieldList->ResolutionUnit = INCHES);
    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, RESOLUTIONUNIT, IFDSHORT,
	1, &dataOffset, &sval, pFieldCount));

    /* Write the Software field on primary images, XIFF 2.0 only */
    if (version == 2 && pxImgInfo->xImgType == XIF_IMGTYPE_PRIMARY) {
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, SOFTWARE, IFDASCII,
	    sizeof(XIFF_SW_2_0), &dataOffset, XIFF_SW_2_0, pFieldCount));
    }

    /* Write the Colormap field and data, if it's paletted */
    if (paletted) {
	UInt32 index;
	UInt32 indices = 1 << bppe;
	UInt16 *red, *green, *blue;
	UInt8 *cmap;

	CHECK_MALLOC(fieldList->Colormap =
	    XIF_MALLOC(indices * 3 *sizeof(*fieldList->Colormap)));
	memset(fieldList->Colormap, 0, indices * sizeof(*fieldList->Colormap));

	/* Convert from the internal colormap, to a TIFF format */
	red = fieldList->Colormap;
	green = red + indices;
	blue = green + indices;
	cmap = pxImgInfo->pPalette;
	for (index = 0; index < indices; index++) {
	    *red++ = ((UInt16)*cmap << 8) | *cmap;
	    cmap++;
	    *green++ = ((UInt16)*cmap << 8) | *cmap;
	    cmap++;
	    *blue++ = ((UInt16)*cmap << 8) | *cmap;
	    cmap++;
	}

	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, COLORMAP, IFDSHORT,
	    indices * 3, &dataOffset, fieldList->Colormap, pFieldCount));
    }

    /* Write the TileWidth field if it's JPEG */
    if (pxImgInfo->xCompression == XIF_JPEG) {
	lval = fieldList->TileWidth = pxImgInfo->dwWidth;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, TILEWIDTH, IFDLONG,
	    1, &dataOffset, &lval, pFieldCount));
    }

    /* Write the TileLength field if it's JPEG */
    if (pxImgInfo->xCompression == XIF_JPEG) {
	lval = fieldList->TileLength = pxImgInfo->dwBestStripHeight;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, TILELENGTH, IFDLONG,
	    1, &dataOffset, &lval, pFieldCount));
    }

    /* Write the TileOffsets field if it's JPEG */
    if (pxImgInfo->xCompression == XIF_JPEG) {

	fieldList->TileCount = (pxImgInfo->dwHeight +
	    pxImgInfo->dwBestStripHeight - 1) / pxImgInfo->dwBestStripHeight;
	CHECK_MALLOC(fieldList->TileOffsets =
	    XIF_MALLOC(fieldList->TileCount * FieldTypeSize[IFDLONG]));
	memset(fieldList->TileOffsets, 0, FieldTypeSize[IFDLONG] *
	    fieldList->TileCount);
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, TILEOFFSETS, IFDLONG,
	    fieldList->TileCount, &dataOffset, fieldList->TileOffsets,
	    pFieldCount));

	/* Store the location of the tile offsets */
	xImg->offsetsLocation = dataOffset;
    }

    /* Write the TileByteCounts field if it's JPEG */
    if (pxImgInfo->xCompression == XIF_JPEG) {

	CHECK_MALLOC(fieldList->TileByteCounts =
	    XIF_MALLOC(fieldList->TileCount * FieldTypeSize[IFDLONG]));
	memset(fieldList->TileByteCounts, 0, FieldTypeSize[IFDLONG] *
	    fieldList->TileCount);
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, TILEBYTECOUNTS, IFDLONG,
	    fieldList->TileCount, &dataOffset, fieldList->TileByteCounts,
	    pFieldCount));

	/* Store the location of the tile offsets */
	xImg->byteCountsLocation = dataOffset;
    }

    /* Write the SubIFDs field, if it could have child images:
     *	- primary image and binary
     *  - odd numbered layer
     *  - special exception: primary image may now have a transparency mask
     */
    if ((pxImgInfo->xImgType == XIF_IMGTYPE_PRIMARY) ||
	(pxImgInfo->dwImgLayer & 1)) {

	/* Allocate the list of SubIFDs if not already done (may have
	 * been done when attaching child images to a tokenized parent
	 * that hasn't been actually written */
	if (fieldList->SubIFDCount == 0) {
	    /* Extend up to 1 without having to rewrite when extending */
	    fieldList->SubIFDCount = 1;

	    CHECK_MALLOC(fieldList->SubIFDs =
		XIF_MALLOC(fieldList->SubIFDCount * FieldTypeSize[IFDLONG]));
	    memset(fieldList->SubIFDs, 0, fieldList->SubIFDCount *
		FieldTypeSize[IFDLONG]);
	}
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, SUBIFDS, IFDLONG,
	    fieldList->SubIFDCount, &dataOffset, fieldList->SubIFDs,
	    pFieldCount));

	/* Preserve the field location */
	xImg->subIFDsTagOffset = dataOffset;
    }

    /* If it's compressed using JPEG, write the JPEG interchange stats */
    if (pxImgInfo->xCompression == XIF_JPEG) {

	/* Write the JPEG Process field */
	fieldList->JPEGProc = 1;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, JPEGPROC, IFDSHORT,
	    1, &dataOffset, &fieldList->JPEGProc, pFieldCount));

	/* These two fields are the location of the JFIF stream, including the
	 * APP0 marker. They are the same as the tile data, which hasn't been
	 * set yet, but will be during the strip write. */
	
	/* Write the JPEG Interchange Format field */
	XERR_GEN_CLEANUP(fieldList->TileOffsets == NULL, XIF_INTERNAL_ERROR);
	fieldList->JPEGInterchangeFormat = fieldList->TileOffsets[0];
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, JPEGINTERCHANGEFORMAT,
	    IFDLONG, 1, &dataOffset, &fieldList->JPEGInterchangeFormat, pFieldCount));

	/* Remember the field data location */
	xImg->interchangeLocation = dataOffset;

	/* Write the JPEG Interchange Format Length field */
	XERR_GEN_CLEANUP(fieldList->TileByteCounts == NULL, XIF_INTERNAL_ERROR);
	fieldList->JPEGInterchangeFormatLength = fieldList->TileByteCounts[0];
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, JPEGINTERCHANGEFORMATLENGTH,
	    IFDLONG, 1, &dataOffset, &fieldList->JPEGInterchangeFormatLength,
	    pFieldCount));

	/* Remember the field data location */
	xImg->interchangeLengthLocation = dataOffset;
    }

    /* Write the AnnotationOffsets field, if it's the primary image */
    if (pxImgInfo->xImgType == XIF_IMGTYPE_PRIMARY) {

	if (version == 2) {

	    /* If the annotation table list is more current than the retained
	     * TIFF field, update the TIFF field */
	    if (fieldList->AnnotationCount <
		xImg->xPage->annotTableHeader.annotCount) {
		AnnotTable *at;
		UInt32 entryOffset = 0UL;

		if (fieldList->AnnotationOffsets) {
		    XIF_FREE(fieldList->AnnotationOffsets);
		}
		fieldList->AnnotationCount = xImg->xPage->annotTableHeader.annotCount;
		CHECK_MALLOC(fieldList->AnnotationOffsets =
		    XIF_MALLOC(fieldList->AnnotationCount *
		    FieldTypeSize[IFDLONG]));
		memset(fieldList->AnnotationOffsets, 0,
		    fieldList->AnnotationCount * FieldTypeSize[IFDLONG]);
		at = xImg->xPage->firstAnnotTable;
		while (at && entryOffset < fieldList->AnnotationCount) {
		    fieldList->AnnotationOffsets[entryOffset++] = at->entries[0].offset;
		    at = at->next;
		}
	    }

	    /* Preserve the count location */
	    xImg->annotationOffsetsCountOffset = xImg->ifdOffset + sizeof(*pFieldCount)
		+ *pFieldCount * XTIFF_FIELDSIZE + 4;

	    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, ANNOTATIONOFFSETS, IFDLONG,
		fieldList->AnnotationCount, &dataOffset,
		fieldList->AnnotationOffsets, pFieldCount));

	    /* Preserve the data location */
	    xImg->annotationOffsetsDataOffset = dataOffset;  /* *pFieldCount; */
	} else {
	    if (fieldList->AnnotationOffsets == NULL) {
		CHECK_MALLOC(fieldList->AnnotationOffsets =
		    XIF_MALLOC(1 * FieldTypeSize[IFDLONG]));  /* 1 empty entry */
		fieldList->AnnotationOffsets[0] = xImg->xPage->annotTableHeaderOffset;
	    }
	    XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, ANNOTATIONOFFSETS, IFDLONG,
		1, &dataOffset, fieldList->AnnotationOffsets, pFieldCount));

	    /* Preserve the field location */
	    xImg->annotationOffsetsDataOffset = dataOffset;  /* *pFieldCount; */
	} /* version 2 or not */
    }

    /* Write the MaskIFDs field, if it could have a transparency mask
     * (XIFF 2.0 only)
     */
    if (version < 3 && pxImgInfo->xImgType != XIF_IMGTYPE_PRIMARY && bpp != 1) {

	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, MASKIFDS, IFDLONG,
	    1, &dataOffset, &fieldList->MaskIFDs, pFieldCount));

	/* Preserve the field location */
	xImg->maskIFDsTagOffset = dataOffset;
    }

    /* Write the ImageLayer field if it's an image within the normal model */
    if (pxImgInfo->xImgType != XIF_IMGTYPE_MASK && version >= 3) {
	lvals[0] = fieldList->ImageLayer[0] = pxImgInfo->dwImgLayer;
	lvals[1] = fieldList->ImageLayer[1] = pxImgInfo->dwImgOffset + 1;
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, IMAGELAYER, IFDLONG,
	    2, &dataOffset, &lvals, pFieldCount));

	/* Remember the field data location */
	xImg->imageLayerLocation = dataOffset;
    }

    /* Write the ImageLayerContent field, if it's the primary image */
    if (pxImgInfo->xImgType == XIF_IMGTYPE_PRIMARY && version > 2) {
	XERR_PASS_CLEANUP(wp_TifWriteFieldEntry(xImg, IMAGELAYERCONTENT,
	    IFDLONG, 0, &dataOffset, NULL, pFieldCount));

	/* Will be the tag offset because < 4 bytes was written . Something
	 * key to note here, is that the TIFF count is what tells the maximum
	 * MRC layer, and will be located 4 bytes before this offset. */
	xImg->ILContentTagOffset = dataOffset;

	/* Now actually write the data */
	XERR_PASS_CLEANUP(wp_TifUpdateILContent(xImg));

	/* Leave the file pointer positioned at the end of the field */
	XIF_SEEK(xFile, xImg->ifdOffset + sizeof(*pFieldCount) +
		 *pFieldCount * XTIFF_FIELDSIZE);
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#define TEMP_DATA_BUFFER_SIZE	1024
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an individual TIFF field
 *
 * LongDescription:::
 *  wp_TifWriteFieldEntry() writes a single field entry, including the data
 *  associated with it. The field entry must have the fields fieldID, type,
 *  and len filled in, and then the associated data pointer passed in
 *  through the *fieldData parameter. This function will fill in the
 *  appropriate value for the offset field. This function will also do
 *  the necessary byte swapping, as appropriate for the endian of the
 *  file, and the type of data.
 *  <BR>
 *  Some of the field 'offset' values are retained for later update. This
 *  is a somewhat obscure thing to do, but it avoids rereading the IFD to
 *  relocate the data, which can't be done anyway if the file was opened
 *  write only (no read).
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  tagID::			TIFF field tag ID
 *  type::			TIFF field data type
 *  len::			TIFF field data length
 *  pFieldData::		Data to store
 *
 * Output:::
 *  pOffset::			Offset of where the data was stored
 *  pFieldCount::		Pointer to accumulating field count
 *
 * Returns:::
 *  error status, sets the appropriate
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifWriteFieldEntry(
    XImg *xImg,
    UInt16 tagID,
    UInt16 type,
    UInt32 len,
    Int32 *pOffset,
    const void *pFieldData,
    UInt16 *pFieldCount
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt16 sval;
    UInt32 lval;
    UInt8 tempBuffer[TEMP_DATA_BUFFER_SIZE];
    UInt8 *tempBuf = NULL;
    UInt8 swap;
    UInt32 bytes;
    PROLOGUE;
    
    CheckHandleValidity(xImg);

    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);
    swap = (UInt8)XIF_FILETOKEN(xFile)->bSwapBytes;

    /* If we haven't made room for the field, we can't fit it in */
    XERR_GEN_CLEANUP((xImg->reservedFieldCount <= *pFieldCount),
		     XIF_INTERNAL_ERROR);

    /* Check to see if the data will fit into the offset value */
    bytes = FieldTypeSize[type] * len;
    if (bytes <= 4) {
	XERR_PASS_CLEANUP(wp_TifConvcopy(pOffset, pFieldData, type, len, swap));

    } else {	/* Data won't stuff into the offset value */

	/* Place at the end of the file */
	XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, pOffset));

	if (swap) {

	    /* If data buffer is small, use local stack */
	    if (bytes <= TEMP_DATA_BUFFER_SIZE) {
		tempBuf = tempBuffer;
	    } else {	/* Otherwise use the client memory */
		CHECK_MALLOC(tempBuf = XIF_MALLOC(bytes));
	    }

	    /* Copy and swap the data */
	    XERR_PASS_CLEANUP(wp_TifConvcopy(&pOffset, tempBuf, type, len,
		swap));

	    /* Write the field data out */
	    XIF_WRITE(xFile, tempBuf, bytes);
	} else {
	    /* Write the field data out */
	    XIF_WRITE(xFile, pFieldData, bytes);
	}

    }

    /* Go to the beginning of the field */
    XIF_SEEK(xFile, xImg->ifdOffset + sizeof(*pFieldCount) + *pFieldCount *
	     XTIFF_FIELDSIZE);
	
    /* Write the field. Keep in mind, if we're writing to an existing file, we
     * may need to swap if the endian mismatches */
    sval = tagID;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = type;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    lval = len;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));
    lval = *pOffset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));
	
    /* Retain the data location (used for some fields) */
    if (bytes <= 4) {
    	*pOffset = xImg->ifdOffset + sizeof(*pFieldCount) + *pFieldCount *
	    XTIFF_FIELDSIZE + sizeof(sval) * 2 + sizeof(lval);
    }

#ifdef DEBUG
    printf("\t%2d:\t%2d\t%2d%8d%8d\n", *pFieldCount, fieldID,
	type, len, pOffset);
#endif

    /* Update the field count */
    (*pFieldCount)++;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    if (tempBuf && (tempBuf != tempBuffer)) {
	XIF_FREE(tempBuf);
    }
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Update the "nextIFD" part of either the previous IFD or TIFF header.
 *
 * LongDescription:::
 *  wp_TifWriteNextIFD() writes the "nextIFD" to the IFD of the previous IFD,
 *  or the TIFF header itself. If the prevImg is passed in, it will be used
 *  to identify the location of the nextIFD location to set. Otherwise, the
 *  nextIFD of the IFD located at prevIFDOffset is set. If both are 0 (or
 *  NULL), then the nextIFD in the TIFF header is set.
 *
 * Input:::
 *  xDoc::			Parent document (in case where deleting last
 *				page of a document, in which case xImg will be
 *				NULL)
 *  xImg::			Data structure of valid image
 *  prevImg::			Previous IFD in chain
 *  prevIFDOffset::		If no prevImg, this IFD is modified
 *  newIFDOffset::		IFD offset of the IFD being added
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifWriteNextIFD(
		XDoc *xDoc,
		XImg *xImg,
		XImg *prevImg,
		UInt32 prevIFDOffset,
		UInt32 newIFDOffset
)
{
    XFile *xFile;
    UInt16 fieldCount;
    UInt32 lval;
    Int32 seekOffset = 0L;
    PROLOGUE;
    
    CheckHandleValidity(xDoc);

    xFile = XIF_DOCFILE(xDoc);

    if (prevImg != NULL && prevIFDOffset == 0UL) {
	prevIFDOffset = prevImg->ifdOffset;
    }

    if (prevIFDOffset == 0UL) {
	TiffHeader tiffHeader;	/* Use only for sizeof macros */

	/* Not used, avoids compiler warnings */
	tiffHeader.firstIfd = newIFDOffset;
	seekOffset = sizeof(tiffHeader.byteOrder) + sizeof(tiffHeader.version);

	/* Make sure we're not referencing something that doesn't exist.
	 * If we have a previous image, but we think we're pointing to the
	 * TIFF header, something is really wrong */
	XERR_GEN_CLEANUP(prevImg != NULL, XIF_BADPARAMETER);
    } else {

	CheckHandleValidity(xImg);

	if (xImg->imgInfo.xImgType == XIF_IMGTYPE_PRIMARY &&
	       xImg->xPage->prev) {

	    /* Use the previous primary IFD from memory */

	    XPage *prevPage;

	    /* Look for the previous page */
	    prevPage = xImg->xPage->prev;
	    prevImg = (XImg *)prevPage->pageInfo.xPrimImgHandle;

	    /* If the image hasn't been stored yet, or it already
	     * points to this IFD offset, no need to continue */
	    if (prevImg == NULL || (prevImg->fields.dataTransferred &&
		prevImg->fields.nextIFD == newIFDOffset)) {
		NORMAL_RETURN;
	    }
	    
	    /* Store the nextIFD on the previous image (may be written later)
	     */
	    prevImg->fields.nextIFD = newIFDOffset;

	    /* If we haven't written the data yet, save it for later */
	    if (prevImg->fields.dataTransferred == 0) {
		NORMAL_RETURN;
	    }
	}
    }

    if (prevImg) {

	/* Make sure the IFD offset is saved internally */
	prevImg->fields.nextIFD = newIFDOffset;
    }

    /* If we have found the previous image, see whether the IFD has been
     * written */
    if (prevImg && prevImg->fields.fieldCount) {

	/* Go to the nextIFD location, past the fields */
	seekOffset = prevIFDOffset + 2 +
	    prevImg->fields.fieldCount * XTIFF_FIELDSIZE;
    } else if (prevImg) {

	/* The nextIFD value (now stored in the fields structure) will be
	 * written to the file later (this happens when the image has been
	 * tokenized, but not written until the dictionary is resolved).
	 */
	NORMAL_RETURN;

    } else if (seekOffset == 0 && xImg &&
	(xImg->xPage->xDoc->readWrite & READ_ALLOWED)) {

	/* Go to the beginning of the IFD */
	XIF_SEEK(xFile, prevIFDOffset);

	/* Read the field count */
	XIF_READ(xFile, &fieldCount, sizeof(fieldCount));
	SwapShort(xFile, &fieldCount);

	/* Go to the nextIFD location, past the fields */
	seekOffset = prevIFDOffset + 2 + fieldCount * XTIFF_FIELDSIZE;
    }

    /* Make sure we've found what we're looking for */
    XERR_GEN_CLEANUP(seekOffset == 0, XIF_INTERNAL_ERROR);

    /* Seek to the nextIFD location */
    XIF_SEEK(xFile, seekOffset);

    /* Write the nextIFD */
    lval = newIFDOffset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an entire IFD
 *
 * LongDescription:::
 *  wp_TifWriteIFD() writes the various components of a XIFF to complete a
 *  separate IFD (to be attached outside of this function).
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifWriteIFD(
    XImg *xImg
)
{
    XFile *xFile;
    UInt16 fieldCount = 0;
    XIF_IMGINFO *pxImgInfo = NULL;
    UInt16 sval;
    UInt32 lval;
    PROLOGUE;
    
    CheckHandleValidity(xImg);
    xFile = XIF_IMGFILE(xImg);
    pxImgInfo = &xImg->imgInfo;

    /* Write an IFD placeholder */
    if (xImg->reservedFieldCount == 0) {
	XERR_PASS_CLEANUP(wp_XifWriteMockIFD(xImg));
    }

    /* Use default values */
    xImg->fields.SamplesPerPixel = 1;

    /* Populate the field list */

    /* Go to the beginning of the IFD */
    XERR_GEN_CLEANUP((xImg->ifdOffset == 0UL),
		     XIF_INTERNAL_ERROR);
    
    /* Write the field count */
    XIF_SEEK(xFile, xImg->ifdOffset);
    sval = fieldCount;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));

    /* Signal data as being written so that all writes take place */
    xImg->fields.dataTransferred = 1 /* cTrue */;

    /* Write the fields */
    XERR_PASS_CLEANUP(wp_TifWriteFields(xImg, &fieldCount));

    /* If we don't have the IFD (as in the case of a page insertion), get it
     * from the page table */
    if (xImg->fields.nextIFD == 0
	&& xImg->imgInfo.xImgType == XIF_IMGTYPE_PRIMARY &&
	xImg->xPage->xDoc->docInfo.wNPages >
	xImg->xPage->pageInfo.dwPageOffset + 1) {
	UInt8 nextPageFlags;
	Int32 nextPageIFDOffset;

	XERR_PASS_CLEANUP(wp_XifGetPageIFD(xImg->xPage->xDoc,
	    xImg->xPage->pageInfo.dwPageOffset + 1,
	    &nextPageFlags, &nextPageIFDOffset));
	xImg->fields.nextIFD = nextPageIFDOffset;
    }

	    /* Write the next IFD (must be positioned just after the fields) */
    lval = xImg->fields.nextIFD;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));

    /* Now go back and write the real field count */
    XIF_SEEK(xFile, xImg->ifdOffset);
    sval = fieldCount;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    xImg->fields.fieldCount = fieldCount;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Attach a child image to a parent, along the SubIFDs list
 *
 * LongDescription:::
 *  wp_TifAttachChildIFD() writes the "nextIFD" to the IFD of the previous IFD
 *  in the SubIFDs list, or the SubIFDs field itself (if the first child).
 *
 * Input:::
 *  parentImg::			Parent image to attach to
 *  xImg::			Data structure of valid image
 *  childIFDOffset::		Offset of child IFD
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifAttachChildIFD(
		XImg *parentImg,
		XImg *xImg,
		UInt32 childIFDOffset
)
{
    Instance *instData;
    XFile *xFile;
    UInt32 lval;
    UInt32 offset;
    PROLOGUE;
    
    CheckHandleValidity(xImg);

    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* The image must not be the primary image if it has a parent. All other
     * images are children of another image. Mask images cannot be a child
     * of a primary image. */
    XERR_GEN_CLEANUP((xImg->imgInfo.xImgType == XIF_IMGTYPE_PRIMARY ||
	(parentImg->imgInfo.xImgType != XIF_IMGTYPE_PRIMARY &&
	xImg->imgInfo.xImgType != XIF_IMGTYPE_MASK)),
		     XIF_BADPARAMETER);

    /* The IFD of the parent images must be known */
    XERR_GEN_CLEANUP((parentImg->ifdOffset == 0UL),
		     XIF_INTERNAL_ERROR);

    /* If we have another subIFD to attach it to, go ahead and do it
     */
    if (parentImg->lastSubImg != NULL){

	/* Shouldn't have another transparency mask for version 2.0 */
	XERR_GEN_CLEANUP(xImg->xPage->xDoc->docInfo.version < 3 &&
	    xImg->imgInfo.xImgType == XIF_IMGTYPE_MASK, XIF_INTERNAL_ERROR);

	/* Attach to the last SubIFD in the list via the nextIFD */
	XERR_PASS_CLEANUP(wp_TifWriteNextIFD(xImg->xPage->xDoc, xImg,
	    parentImg->lastSubImg, 0UL, childIFDOffset));

    } else if (parentImg->fields.dataTransferred &&
	((xImg->xPage->xDoc->docInfo.version >= 3 &&
	parentImg->fields.SubIFDCount > 0UL) ||
	(xImg->xPage->xDoc->docInfo.version < 3))) {

	/* We have an active parent image, attach to the SubIFDs/MaskIFDs
	 * field of the parent */
	if (xImg->imgInfo.xImgType == XIF_IMGTYPE_MASK &&
	    xImg->xPage->xDoc->docInfo.version < 3) {
	    offset = parentImg->maskIFDsTagOffset;
	    parentImg->fields.MaskIFDs = lval = childIFDOffset;
	} else {
	    offset = parentImg->subIFDsTagOffset;
	    parentImg->fields.SubIFDs[0] = lval = childIFDOffset;
	}
	XIF_SEEK(xFile, offset);
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
    } else {

	/* The only time we should wait to write later is when we're
	 * tokenizing an image and waiting for other pages before actually
	 * writing out the data. In XIFF 2.0, this can't happen */
	XERR_GEN_CLEANUP(xImg->xPage->xDoc->docInfo.version < 3 &&
	    xImg->imgInfo.xImgType == XIF_IMGTYPE_MASK, XIF_INTERNAL_ERROR);

	XERR_GEN_CLEANUP((parentImg->fields.SubIFDCount > 0UL),
		 XIF_INTERNAL_ERROR);

	/* Keep it around as the first one for now (write later) */
	parentImg->fields.SubIFDCount = 1;
	if (xImg->imgInfo.xImgType == XIF_IMGTYPE_MASK &&
	    xImg->xPage->xDoc->docInfo.version < 3) {

	    parentImg->fields.MaskIFDs = childIFDOffset;

	} else {
	    CHECK_MALLOC(parentImg->fields.SubIFDs =
		XIF_MALLOC(parentImg->fields.SubIFDCount *
		FieldTypeSize[IFDLONG]));
	    memset(parentImg->fields.SubIFDs, 0,
		parentImg->fields.SubIFDCount * FieldTypeSize[IFDLONG]);

	    parentImg->fields.SubIFDs[0] = childIFDOffset;
	}
    }

    /* Retain the position of the last in the string */
    parentImg->lastSubImg = xImg;


    /* Update the parent image */
    /* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */


    /* Update the page statistics */
    /* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */


    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}



/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a strip of data and update the offset/byteCount field entries
 *
 * LongDescription:::
 *  wp_TifWriteStrip() write a strip of (usually compressed) data, and updates
 *  the StripOffsets and ByteCounts fields.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifWriteStrip(
	      XImg *xImg
)
{
    XFile *xFile;
    Int32 offset;
    UInt32 stripNum;
    UInt32 seekOffset;
    UInt32 lval;
    PROLOGUE;

    CheckHandleValidity(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* Align the file to a long boundary */
    XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &offset));

    /* Write the data */
    XIF_WRITE(xFile, xImg->cmpBuffer, xImg->cmpBytes);

    /* Write the strip offset and byte count to the fields */
    stripNum = xImg->tmpPixrRow / xImg->imgInfo.dwBestStripHeight;
    seekOffset = xImg->offsetsLocation + stripNum * sizeof(long);
    XERR_GEN_CLEANUP(seekOffset == 0, XIF_INTERNAL_ERROR);
    XIF_SEEK(xFile, seekOffset);
    lval = offset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));
    seekOffset = xImg->byteCountsLocation + stripNum * sizeof(long);
    XERR_GEN_CLEANUP(seekOffset == 0, XIF_INTERNAL_ERROR);
    XIF_SEEK(xFile, seekOffset);
    lval = xImg->cmpBytes;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));

    /* Update the field values */
    if (xImg->fields.StripOffsets == NULL ||
	xImg->fields.StripByteCounts == NULL) {
	XERR_GEN_CLEANUP((xImg->fields.TileOffsets == NULL ||
	    xImg->fields.TileByteCounts == NULL), XIF_INTERNAL_ERROR);

	xImg->fields.TileOffsets[stripNum] = offset;
	xImg->fields.TileByteCounts[stripNum] = xImg->cmpBytes;
    } else {
	xImg->fields.StripOffsets[stripNum] = offset;
	xImg->fields.StripByteCounts[stripNum] = xImg->cmpBytes;
    }

    /* Also update the JPEG Interchange values */
    if (stripNum == 0 && xImg->imgInfo.xCompression == XIF_JPEG) {
	seekOffset = xImg->interchangeLocation;
	XERR_GEN_CLEANUP(seekOffset == 0, XIF_INTERNAL_ERROR);
        XIF_SEEK(xFile, seekOffset);
        lval = offset;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	seekOffset = xImg->interchangeLengthLocation;
	XERR_GEN_CLEANUP(seekOffset == 0, XIF_INTERNAL_ERROR);
	XIF_SEEK(xFile, seekOffset);
	lval = xImg->cmpBytes;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Update the ImageLayerContent TIFF field
 *
 * LongDescription:::
 *  wp_TifUpdateILContent() update the contents of the ImageLayerContent TIFF
 *  field.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifUpdateILContent(
		   XImg *xImg
)
{
    Instance *instData;
    XFile *xFile;
    XPage *xPage = xImg->xPage;
    UInt32 lval, *lvals = NULL;
    UInt32 layer;
    TiffFields *fieldList;
    PROLOGUE;

    fieldList = &xImg->fields;

    if (xImg->fields.dataTransferred == 0 ||
	xImg->xPage->xDoc->docInfo.version < 3) {

	/* The field hasn't been written yet, no need to update now */
	NORMAL_RETURN;
    }

    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* If the number of stored image counts has increased, we need to
     * rewrite the data to a new location */
    if (xPage->pageInfo.dwLayersPresent > fieldList->ILContent.layersPresent) {
	XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &xImg->ILContentDataOffset));
	XIF_SEEK(xFile, xImg->ILContentTagOffset - sizeof(UInt32));
	lval = xPage->pageInfo.dwLayersPresent;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	lval = xImg->ILContentDataOffset;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	fieldList->ILContent.layersPresent = xPage->pageInfo.dwLayersPresent;
    }
    CHECK_MALLOC(lvals = (UInt32 *)XIF_MALLOC(
	fieldList->ILContent.layersPresent * sizeof(UInt32)));
    for (layer = 0; layer < fieldList->ILContent.layersPresent; layer++) {
	lvals[layer] = xPage->pageInfo.pdwImageCounts[layer];
    }

    if ((UInt8)XIF_FILETOKEN(xFile)->bSwapBytes) {
	XERR_PASS_CLEANUP(wp_TifConvcopy(lvals, lvals, IFDLONG,
	    fieldList->ILContent.layersPresent,
	    (UInt8)XIF_FILETOKEN(xFile)->bSwapBytes));
    }

    XIF_SEEK(xFile, xImg->ILContentDataOffset);
    XIF_WRITE(xFile, lvals, sizeof(lvals) * fieldList->ILContent.layersPresent);
    XIF_FREE(lvals);
    lvals = NULL;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (lvals) {
	XIF_FREE(lvals);
	lvals = NULL;
    }
    
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an image from a page
 *
 * LongDescription:::
 *  wp_TifDeleteImage() deletes (removes all references to) an single image on
 *  a page.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  image are removed.  This includes: MRC layer image counts and relinking
 *  the previous image to the subsequent image via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the image is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' an
 *  image deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see XIF_CopyPage() and
 *  XIF_CopyImage()), allowing the disowned file fragments to remain behind.
 *
 * Input:::
 *  xImg::			Image to be removed
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifDeleteImage(
	XImg *xImg
)
{
    Instance *instData;
    XFile *xFile;
    XImg *xParentImg = NULL;		/* Parent image */
    XImg *childImg;
    XPage *xPage = xImg->xPage;
    XDoc *xDoc = xPage->xDoc;
    XIF_IMGINFO *pxImgInfo;
    UInt32 lval;
    PROLOGUE;

    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);
    pxImgInfo = &xImg->imgInfo;

    /* Deleting the primary image from a page is invalid, since the page
     * has the primary image as it's foundation (without it, there is no
     * page). What is probably meant here is a call to XIF_DeletePage().
     */
    XERR_GEN(pxImgInfo->xImgType == XIF_IMGTYPE_PRIMARY, XIF_BADPARAMETER);
    XERR_GEN(xImg == xPage->xImgs, XIF_BADPARAMETER);

    /* Either this is the primary image, or there should be one. Primary
     * image must be called layer 1, 2 or 3. */
    XERR_GEN(xPage->pageInfo.xPrimImgHandle == 0 &&
	     (pxImgInfo->dwImgLayer != 1 || pxImgInfo->dwImgLayer != 2 ||
	      pxImgInfo->dwImgLayer != 3) && pxImgInfo->dwImgOffset != 0,
	     XIF_SEQUENCE);
    xParentImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

    /* Check for proper calling sequence */
    XERR_GEN((xDoc->readWrite & WRITE_ALLOWED) == 0, XIF_SEQUENCE);

    /* Unlink it from the image(s) that follows */
    if (xImg->next) {

	/* Traverse the following images, and decrement the image offset
	 * of any that reside on the same level, and have a subsequent offset
	 */
	childImg = xImg;  /* Start here, but immediately go to the next */
	while ((childImg = childImg->next) != NULL) {
	    if (childImg->imgInfo.dwImgLayer == pxImgInfo->dwImgLayer &&
		childImg->imgInfo.dwImgOffset > pxImgInfo->dwImgOffset) {
		childImg->imgInfo.dwImgOffset--;

		/* Rewrite the ImageLayer field item: image offset */
		XIF_SEEK(xFile, childImg->imageLayerLocation + sizeof(long));
		lval = childImg->imgInfo.dwImgOffset + 1;	/* 1-based */
		SwapLong(xFile, &lval);
		XIF_WRITE(xFile, &lval, sizeof(long));
	    }
	}
    }

    /* Point the previous image to the next one, via the NextIFD IFD element */
    if (xImg->prev) {

	/* Rewrite the NextIFD for the elder sibling (or parent) */
	XERR_PASS_CLEANUP(wp_TifWriteNextIFD(xDoc, xImg->prev, xImg->prev->prev,
					     xImg->prev->ifdOffset,
					     xImg->fields.nextIFD));
	
    } else {

	/* There must only be one if there's no backwards link, which
	 * is in error, since we checked above for deleting the primary
	 * image. */
	xPage->xImgs = NULL;
	XERR_GEN_CLEANUP(xPage->xImgs == xImg, XIF_INTERNAL_ERROR);
    }

    /* Decrement the ImageLayerContent field value for the layer */
    XERR_PASS_CLEANUP(wp_XifIncrementImageCounts(xPage, pxImgInfo->dwImgLayer,
						 -1));

    XERR_PASS_CLEANUP(wp_TifUpdateILContent(xParentImg));

    /* Unlink it from the page list */
    if (xImg->prev) {
	xImg->prev->next = xImg->next;
	xImg->next->prev = xImg->prev;
    } else {

	/* There must only be one if there's no backwards link, which
	 * is in error, since we checked above for deleting the primary
	 * image. */
	xPage->xImgs = NULL;
	XERR_GEN_CLEANUP(xPage->xImgs == xImg, XIF_INTERNAL_ERROR);
    }

    /* Free the image structure */
    XERR_PASS_CLEANUP(wp_XifImgFree(&xImg, 0));

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*
$Log: xifwtiff.c,v $
Revision 1.1.1.1  1999/09/20 18:19:22  jlin
Imported sources

Revision 1.21  1998/07/30 22:42:53  aber
aber on Thu Jul 30 15:42:24 PDT 1998

xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Allow color tokenized images, call them RGB.

Revision 1.20  1998/04/02 00:09:55  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

CommonCheck: Allow attaching a transparency mask to the primary image, special request.

Revision 1.19  1998/01/21 22:56:10  aber
aber on Wed Jan 21 14:54:57 PST 1998

xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Change logic to allow a binary image to have a mask, in an odd-numbered
layer only. Note that this formally breaks MRC.

Revision 1.18  1998/01/20 03:43:53  aber
aber on Mon Jan 19 19:43:22 PST 1998

xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Prototype update to allow deleting last page of a document.

Revision 1.17  1998/01/20 03:24:17  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Rename MixedRasterContent field to ImageLayerContent, per Lloyd's suggestion.

Revision 1.16  1998/01/16 00:43:57  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Allow page inserts.

Revision 1.15  1997/12/18 00:24:50  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Keep track of the ImageLayer field data location.
Make sure the nextIFD value is kept internally.
Keep track of the strip sizes and locations (required when in read/write)

Revision 1.14  1997/11/24 21:14:07  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Return the file pointer to the end of the field list, so that when
writing the NextIFD, it written to the right place. This only happens
when writing the MixedRasterContent field, when the field is updated
after the subimages have been attached (tokenized with subimages).

Revision 1.13  1997/08/29 17:28:08  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.12  1997/07/10 22:02:56  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Keep track of the annotation count within the TIFF field.

Revision 1.11  1997/07/02 17:04:58  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Make the right return value check.

Revision 1.10  1997/07/02 01:07:24  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Use private function names and check return values for type
conversion.

Revision 1.9  1997/06/26 22:33:09  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Return a new parameter if the file is not a TIF.
Allow multiple tiles (JPEG/Wavelet).

Revision 1.8  1997/06/17 22:37:15  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

Fixed bug (editing error) in last checkin.

Revision 1.6  1997/06/10 01:27:02  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:20  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:23  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:32  aber
xifwtiff.c -> /product/ipcore/ipwrap/src/RCS/xifwtiff.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:49  aber
Initial revision

*/
