/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwtok.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifwtok.c:	XIFF WRITE TOKENIZATION source
 *
 *	XIFF write-only routines for Tokenization.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "shrcmprs.pub"
#include "iaerror.pub"
#include "frames.pub"
#include "halftone.pub"

#include "ipstate.pub"
#include "morphg.h"
#include "streldef.h"
#include "strel.pub"
#include "morph.pub"
#include "logicops.h"
#include "logicops.pub"

#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifwtiff.h"
#include "xifrtok.h"
#include "xifwtok.h"
#include "xifwpriv.h"
#include "xifwxiff.h"
#include "xifrpriv.h"
#include "xiff.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwtok.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");

static Instance *globalInstData;
static void * CDECL
wp_XifTextMalloc(
	   IPWState *pIPWState,
	   UInt32 bytes,
	   UInt32 fixupPtrs
)
{
    Instance *instData = globalInstData;
    void *buf = NULL;
    PROLOGUE;

    pIPWState = pIPWState;
    CHECK_MALLOC(buf = XIF_MALLOC(bytes));
    memset(buf, 0, bytes);

    if (fixupPtrs) {
      return (void *)buf;
    }

ERR_CLEANUP:

    return (void *)buf;
}


static void CDECL
wp_XifTextFree(
	 IPWState *pIPWState,
	 void *buf,
	 UInt32 len
)
{
    Instance *instData = globalInstData;

    pIPWState = pIPWState;
    if (len == 0) {
      return;
    }

    XIF_FREE(buf);

    return;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a binary image to token positions and IDs, to be stored later.
 *
 * LongDescription:::
 *  wp_XifTokenizeImage() converts a binary image from raster form to token
 *  positions and IDs. This is the first of several processes in the
 *  tokenization list. Later, when all pages have been tokenized, the
 *  symbols in the dictionary are consolidated and all position information
 *  is remapped to the cosolidated dictionary. Then all data is converted
 *  to JBIG-2 "Tundra" format, which is more compressed.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pImgData::			Pointer to first pixel of image data
 *  dwBytesPerRow::		Bytes per row of image data
 *  dwFlags::			Tokenization flags (see TOKEN_... in xiff.h)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifTokenizeImage(
    XImg *xImg,
    UInt8 *pImgData,
    Int32 dwBytesPerRow,
    UInt32 dwFlags
)
{
    Instance *instData = NULL;
    XPage *xPage;
    XDoc *xDoc;
    IPWState *pIPWState = NULL;
    XIF_IMGINFO *imgInfo;
    UInt8 *positionBuffer = NULL;
    UInt32 positionBufferSize;
    UInt32 positionBytes;
    void *textInfo;
    ScaledInt1616 xres, yres;
    Int32 h, bpl, dpi;
    PIXR *imgPixr = NULL;
    UInt8 *srcptr, *dstptr;
    UInt32 row;
    PixrRef *stipplePixrs;
    Int32 numStippleImgs = 0, stippleNum;
    Int32 stippleWidth, stippleHeight;
    UInt32 stippleXOffset = 0, stippleYOffset = 0;
    UInt8 *imgData, *cmpBuffer = NULL, *imgPlanes[3];
    UInt32 cmpBufsize, cmpBytes;
    XIF_IMGHANDLE stippleImgHandle;
    XImg *stippleImg;
    UInt32 bytesPerRow;
    XIF_COMPRESSION compression;
    UInt32 quality;
    UInt32 bytesWritten;
    Int32 pctIncrease;
    UInt8 *pFramedImgData;
    Int32 framedWidth, framedHeight;
#if 0
    // Fix when color actually added
    UInt8 *fgImgPlanes[3];
    Int32 fgImgBpl;
#endif
    TCData callbackData;
    PROLOGUE;

    CheckHandleValidity(xImg);
    instData = XIF_IMGINST(xImg);
    imgInfo = &xImg->imgInfo;
    xPage = xImg->xPage;
    xDoc = xPage->xDoc;

    /* If in a color tokenization state, make sure we pass the color
     * information downward */

#if 0
    // Fix when color actually added

    if (xPage->seqStatus & XPAGE_SEQ_TKCOLORADDED) {
	XImg *fgImg = xPage->fgImg;
	Int32 channel, channels;
	PIXR *pSubPixr;

	XERR_GEN_CLEANUP(fgImg == NULL, XIF_INTERNAL_ERROR);
	XERR_GEN_CLEANUP(fgImg->tmpPixr == NULL, XIF_INTERNAL_ERROR);
	XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, fgImg->tmpPixr,
						&channels));
	XERR_GEN_CLEANUP(channels != 3, XIF_INTERNAL_ERROR);
	for (channel = 0; channel < channels; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, fgImg->tmpPixr,
					       channel, &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				     &fgImgPlanes[channel]));
	}
	XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, fgImg->tmpPixr, &fgImgBpl));
    }
    // Fix when color actually added
#endif

    /* Reference the IPWrap state */
    XERR_PASS_CLEANUP(wp_XifIPWState(xDoc, &pIPWState));

    /* If scanned in binary, halftones may exist which will cause the
     * tokenization process to spend a lot of effort on tiny pixel clumps.
     * Attempt to remove the clumps */
    if (dwFlags & TOKENIZE_BINARYSCAN) {

#if 1
 	XERR_GEN_CLEANUP(imgInfo->dwXResolution <= 0 ||
			 imgInfo->dwYResolution <= 0, XIF_BADFILEFORMAT);
#else
	/* Make sure the resolution isn't totally bogus, since it's
	 * important here */
	if (imgInfo->dwXResolution < 50 || imgInfo->dwXResolution > 2000) {
	    if (imgInfo->dwYResolution < 50 || imgInfo->dwYResolution > 2000) {
		imgInfo->dwXResolution = imgInfo->dwYResolution = 300;
	    } else {
		imgInfo->dwXResolution = imgInfo->dwYResolution;
	    }
	}
	if (imgInfo->dwYResolution < 50 || imgInfo->dwYResolution > 2000) {
	    imgInfo->dwYResolution = imgInfo->dwXResolution;
	}
#endif


	/* Copy the image data to a PIXR that can be edited */
	xres = Int32_TO_ScaledInt1616(imgInfo->dwXResolution);
	yres = Int32_TO_ScaledInt1616(imgInfo->dwYResolution);
	dpi = MIN(imgInfo->dwXResolution, imgInfo->dwYResolution);
 	XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, imgInfo->dwWidth,
	    imgInfo->dwHeight, XIF_BPPE(imgInfo->xColorType),
	    xres, yres, 32, 1, cColorGray, NULL, &imgPixr));

	h = imgInfo->dwHeight;
	srcptr = pImgData;
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, imgPixr, &dstptr));
	XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, imgPixr, &bpl));
	bytesPerRow = MIN(dwBytesPerRow, bpl);
	for (row = 0; row < h; row++) {
	    memcpy(dstptr, srcptr, bytesPerRow);
	    srcptr += dwBytesPerRow;
	    dstptr += bpl;
	}

	/* Initialize strels */
	XERR_PASS_CLEANUP(w_strel1Inits(pIPWState));
	XERR_PASS_CLEANUP(w_strel11Inits(pIPWState));

	/* Remove the stipple from the image */
	if (dpi >= 300) {
	    XERR_PASS_CLEANUP(w_segmentBinaryPagePixrs(pIPWState, imgPixr,
		4, 8, 1, 99, 15, &stipplePixrs, &numStippleImgs));
	} else if (dpi >= 200) {
	    XERR_PASS_CLEANUP(w_segmentBinaryPagePixrs(pIPWState, imgPixr,
		2, 4, 1, 66, 10, &stipplePixrs, &numStippleImgs));
	} else {
	    XERR_PASS_CLEANUP(w_segmentBinaryPagePixrs(pIPWState, imgPixr,
		1, 2, 1, 33, 5, &stipplePixrs, &numStippleImgs));
	}

	/* If there's no stipple image(s), continue as normal */
	if (stipplePixrs == NULL) {
	    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &imgPixr));
	} else {

	    /* A stipple was found, loop through them and remove each.
	     * Use lossy dither algorithm to compress */
	    for (stippleNum = 0; stippleNum < numStippleImgs; stippleNum++) {

		/* Get the size */
		XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState,
		    stipplePixrs[stippleNum].pPixr, &stippleWidth));
		XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState,
		    stipplePixrs[stippleNum].pPixr, &stippleHeight));
		XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState,
		    stipplePixrs[stippleNum].pPixr, &imgData));
		stippleXOffset = stipplePixrs[stippleNum].xOffset;
		stippleYOffset = stipplePixrs[stippleNum].yOffset;

		/* Store the image */
		XERR_PASS_CLEANUP(XIF_AddImage((XIF_PAGEHANDLE)xImg->xPage,
		    xImg->imgInfo.dwImgLayer, XIF_COLORTYPE_BINARY, stippleXOffset,
		    stippleYOffset, xImg->imgInfo.dwXResolution,
		    xImg->imgInfo.dwYResolution, stippleWidth, stippleHeight,
		    stippleHeight, NULL, &stippleImgHandle));
		stippleImg = (XImg *)stippleImgHandle;

		/* Initiate the write process */
		compression = XIF_LOSSYDITHER;
		quality = 0;
		XERR_PASS_CLEANUP(XIF_ImageWriteStart(stippleImgHandle,
		    XIF_IMGFLAGS_Compressed, &compression, &quality));
		XERR_GEN_CLEANUP(compression != XIF_LOSSYDITHER, XIF_NOSUPPORT);
		    
		/* Create an image for the lossy dithered data */
		cmpBufsize = (stippleWidth * stippleHeight / 4) / 8;
		CHECK_MALLOC(cmpBuffer = XIF_MALLOC(cmpBufsize));

		/* Lossy dither the image */
		imgPlanes[0] = imgData;
		XERR_PASS_CLEANUP(w_XifCompress(pIPWState, stippleImgHandle,
		    XIF_LOSSYDITHER, XIF_COLORTYPE_BINARY, 0,
		    stipplePixrs[stippleNum].pPixr, 0, stippleWidth,
		    stippleHeight, &cmpBuffer, &cmpBufsize, &cmpBytes));

		XERR_PASS_CLEANUP(XIF_CompressedWriteStrip(stippleImgHandle,
		    0, stippleHeight, cmpBuffer, cmpBytes, &bytesWritten));

		XERR_PASS_CLEANUP(XIF_ImageWriteFinish(stippleImgHandle));

		/* Remove it from the original */
		XERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0,
		    stipplePixrs[stippleNum].pPixr, PIX_SRC ^ PIX_DST,
		    stipplePixrs[stippleNum].xOffset,
		    stipplePixrs[stippleNum].yOffset, stippleWidth,
		    stippleHeight, imgPixr));

		/* Delete PIXR */
		XERR_PASS_CLEANUP(w_destroyPixr(pIPWState,
		    &stipplePixrs[stippleNum].pPixr));

		/* Free the compressed data buffer */
		XIF_FREE(cmpBuffer), cmpBuffer = NULL;

	    } /* For each stipple image */

	    /* Have free the same way we allocated */
	    IPW_FREE(stipplePixrs, cAllocatorData);
	    stipplePixrs = NULL, numStippleImgs = 0;

	    /* Now repoint the input image to the image with the stipples
	     * removed */
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, imgPixr, &pImgData));
	    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, imgPixr, &dwBytesPerRow));
	}
    }   /* If a binary scan */

    /* Setup */
    textInfo = xDoc->textCompInfo;
    if (textInfo == NULL) {
	/* ***********  Work around a client ID problem (for now)*********** */
	globalInstData = XIF_IMGINST(xImg);
	XERR_PASS_CLEANUP(i_allocateTextCompressionInfo (pIPWState,
	    &textInfo, wp_XifTextMalloc, wp_XifTextFree));
	xDoc->textCompInfo = (UInt8 *)textInfo;
    }

    /* If we have no dictionary, or the one we have is inconsistent with the
     * one we need, start with a new dictionary */
    if (((xDoc->dictFlags & TOKENDICT_WRITABLE) == 0) ||
	(((xDoc->dictFlags & TOKENDICT_LOSSLESS) == TOKENDICT_LOSSLESS) ^
	((dwFlags & TOKENIZE_LOSSLESS) == TOKENIZE_LOSSLESS))) {

	/* If we have a token dictionary already, flush it */
	if (((xDoc->seqStatus & (XDOC_SEQ_TOKENSTART)) == 1) &&
	    ((xDoc->dictFlags & TOKENDICT_WRITABLE) == 0)) {
	    XERR_PASS_CLEANUP(XIF_WriteTokenDictionary((XIF_DOCHANDLE)xDoc));
	}

	/* Start a new list of "current" dictionary IDs */
	if (xDoc->curDictIDs) {
	    XIF_FREE(xDoc->curDictIDs);
	}
	xDoc->curDictCount = 1;
	CHECK_MALLOC(xDoc->curDictIDs =
	    XIF_MALLOC(xDoc->curDictCount * sizeof(UInt32)));

	/* Set the dictionary number (must start at 0 for IPShared routines) */
	xDoc->curDictIDs[0] = xDoc->symlibTableHeader.symlibCount;
	XERR_PASS_CLEANUP(i_setDictionaryNumber (pIPWState,
	    textInfo, 0));

	/* Fix the dictionary ID now, since compressed dictionaries could
	 * later be copied before this dictionary is resolved */
	XERR_PASS_CLEANUP(wp_XifSetSymlibOffset(xDoc, xDoc->curDictIDs[0],
	    0, 0));

	xDoc->dictFlags = TOKENDICT_WRITABLE;
	if (dwFlags & TOKENIZE_LOSSLESS) {
	    xDoc->dictFlags |= TOKENDICT_LOSSLESS;
	}
    }

    /* If we're gunna use a previously created dictionary, load now */
#ifdef MERGING_TO_ANOTHER_TOKEN_DICTIONARY
    /* This won't happen, but here's the call we'd use */
    XERR_PASS_CLEANUP(i_loadDictionaryBlock (pIPWState));
#endif

    if (xPage->seqStatus & XPAGE_SEQ_TKCOLORADDED) {
	callbackData.img = xPage->fgImg;

	/* Add a function to the text compression handle, to make use of the
	 * color information */
	XERR_PASS_CLEANUP(i_initializeTextColorCallback(pIPWState,
	    textInfo, wp_XifAssignTokenColor, (void *)&callbackData));
    }

    /* Assign a page ID */
    xPage->tokenPageID = ++xDoc->nextTokenPageID;

    /* Point to frameless data */
    pFramedImgData = pImgData - ((dwBytesPerRow << cLogFrameBits) +
				 cFrameBytes);
    framedWidth = imgInfo->dwWidth + 64;
    framedHeight = imgInfo->dwHeight + 64;

    /*  Estimate the size of the position information */
    XERR_PASS_CLEANUP(i_estimateSizeOfPositionInfo (pIPWState,
	pFramedImgData, framedWidth, framedHeight,
	XIF_BPP(imgInfo->xColorType), dwBytesPerRow,
	&positionBufferSize));

    if (dwFlags & TOKENIZE_LOSSLESS) { /* Double the estimate */
	positionBufferSize *= 2;
    }

    pctIncrease = 120;	    /* Initial buffer size percentage increase */

    do {

	/* Allocate that buffer (okay, be overly thorough and add a little) */
	positionBufferSize = (positionBufferSize * pctIncrease) / 100;
	CHECK_MALLOC(positionBuffer = XIF_MALLOC(positionBufferSize));

	/* Convert the image to position information. Keep the error around
	 * in case we need a larger compression buffer. */
	if (dwFlags & TOKENIZE_LOSSLESS) {
	    __error__ = i_convertImageToSymbolsExactly (pIPWState,
		pFramedImgData, framedWidth, framedHeight, dwBytesPerRow,
		textInfo, xPage->tokenPageID, positionBuffer,
		positionBufferSize, &positionBytes);
	} else {
	    __error__ = i_convertImageToSymbols (pIPWState,
		pFramedImgData, framedWidth, framedHeight, dwBytesPerRow,
		textInfo, xPage->tokenPageID, positionBuffer,
		positionBufferSize, &positionBytes);
	}

	if (__error__ == ia_bufferExceeded) {
	    XIF_FREE(positionBuffer);
	    positionBuffer = NULL;
	    pctIncrease = MIN(pctIncrease + 50, 300);	/* At most triple it */
	} else if (__error__ != ia_successful) {
	    goto ERR_CLEANUP;
	}
    } while (__error__ == ia_bufferExceeded);

    /* Attach the data to the image, create a new memory buffer with only
     * the bytes necessary */
    CHECK_MALLOC(xImg->tmpCmpData = XIF_MALLOC(sizeof(MISC_MEM)));
    xImg->tmpCmpData->next = NULL;
    CHECK_MALLOC(xImg->tmpCmpData->mem = XIF_MALLOC(positionBytes));
    xImg->tmpCmpData->bytes = positionBytes;
    memcpy(xImg->tmpCmpData->mem, positionBuffer, positionBytes);
    XIF_FREE(positionBuffer);
    positionBuffer = NULL;

    /* Reflect tokenization environment (for later info queries) */
    xImg->imgInfo.dwBestStripHeight = xImg->imgInfo.dwHeight;

    /* Indicate tokenization in progress */
    xImg->seqStatus |= XIMG_SEQ_TOKENSTART;
    xImg->xPage->seqStatus |= XPAGE_SEQ_TOKENSTART;
    xDoc->seqStatus |= XDOC_SEQ_TOKENSTART;

    /* Fall through to free up any temporary memory */
    
 ERR_CLEANUP:

    if (cmpBuffer) {
	XIF_FREE(cmpBuffer);
    }

    if (positionBuffer) {
	XIF_FREE(positionBuffer);
	positionBuffer = NULL;
    }

    if (numStippleImgs) {
	for (stippleNum = 0; stippleNum < numStippleImgs; stippleNum++) {
	    if (stipplePixrs[stippleNum].pPixr != NULL) {
		w_destroyPixr(pIPWState, &stipplePixrs[stippleNum].pPixr);
	    }
	}
    }

    if (imgPixr) {
	w_destroyPixr(pIPWState, &imgPixr);
    }

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Consolidate "like" tokens.
 *
 * LongDescription:::
 *  wp_XifConsolidateTokens() finds tokens in the current token dictionary that
 *  "match" each other, and map to a common symbol. The mapping table is
 *  retained for remapping the previous IDs.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifConsolidateTokens(
	XDoc *xDoc
)
{
    Instance *instData = NULL;
    XPage *xPage;
    XImg *xImg;
    Int32 numOldSymbols, numNewSymbols;
    void *compInfo;
    UInt32 *remapTable = NULL;
    UInt32 retval;
    Int32 tableSize;
    UInt8 *newMem = NULL;
    UInt32 remappedSize, newMemsize;
    IPWState *pIPWState = NULL;
    PROLOGUE;

    CheckHandleValidity(xDoc);

    /* Extract the IPCore state structure */
    XERR_PASS(wp_XifIPWState(xDoc, &pIPWState));

    instData = XIF_DOCINST(xDoc);

    /* Extract the text compression info structure from the document */
    compInfo = xDoc->textCompInfo;
    XERR_GEN_CLEANUP((compInfo == NULL),
		     XIF_INTERNAL_ERROR);

    /* Get a count of the number of symbols before consolidation */
    XERR_PASS_CLEANUP(i_getNumberOfSymbols (pIPWState,
	compInfo, cSymbolsAll, 0, &numOldSymbols));

    /* Remap only if lossy tokenization */
    if ((xDoc->dictFlags & TOKENDICT_LOSSLESS) == 0 && numOldSymbols > 0) {

	/* Get the size of the remap table */
	XERR_PASS_CLEANUP(i_getSizeOfSymbolRemapArray (pIPWState,
	    compInfo, &tableSize));

	/* Create a table to do the remapping */
	CHECK_MALLOC(remapTable = (UInt32 *)XIF_MALLOC(tableSize));

	/* Consolidate symbols, create a remapping */
	XERR_PASS_CLEANUP(i_consolidateShapeInfo (pIPWState,
	    compInfo, remapTable));

	/* Get a count of the number of symbols *after* consolidation */
	XERR_PASS_CLEANUP(i_getNumberOfSymbols (pIPWState,
	    compInfo, cSymbolsAll, 0, &numNewSymbols));

#ifdef DEBUG
	printf("Symbol consolidation reduced %d symbols to %d\n", numOldSymbols,
	    numNewSymbols);
#endif

	/* Apply remapping */
        xPage = xDoc->pages;
	while (xPage) {
	    xImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

	    /* If there's failure here, there's a good chance that a previous
	     * failure occured, and we're cleaning up the document */
	    XERR_GEN_CLEANUP(xImg == NULL, XIF_INTERNAL_ERROR);

	    if (xImg->seqStatus & XIMG_SEQ_TOKENSTART) {

		do {
		    retval = i_remapSymbolIDs (pIPWState, compInfo, remapTable,
			xImg->tmpCmpData->mem, xImg->tmpCmpData->bytes,
			xImg->tmpCmpData->mem, xImg->tmpCmpData->bytes,
			&remappedSize);
		    if (retval == ia_bufferExceeded) {

			/* Is this a 1:1 mapping?  Try a larger buffer. */
			newMemsize = xImg->tmpCmpData->bytes * 12 / 10;
			CHECK_MALLOC(newMem = (UInt8 *)XIF_MALLOC(newMemsize));
			memcpy(newMem, xImg->tmpCmpData->mem,
			    xImg->tmpCmpData->bytes);
			XIF_FREE(xImg->tmpCmpData->mem);
			xImg->tmpCmpData->mem = (void *)newMem;
			xImg->tmpCmpData->bytes = newMemsize;
			newMem = NULL;
		    } else {
			xImg->tmpCmpData->bytes = remappedSize;
		    }
		} while (retval == ia_bufferExceeded);

		if (retval != ia_successful) {
		    __error__ = retval;
		    goto ERR_CLEANUP;
		}

	    }
	    xPage = xPage->next;
	} /* For each page that nedds to be remapped */
    } /* If not a lossless dictionary */

    /* Fall through to free up any temporary memory */

 ERR_CLEANUP:

    if (remapTable) {
	XIF_FREE(remapTable);
	remapTable = NULL;
    }
    if (newMem) {
	XIF_FREE(newMem);
	newMem = NULL;
    }
    NORMAL_RETURN;
} 


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert the token dictionary to "Tundra" format, and store in the symlib.
 *
 * LongDescription:::
 *  wp_XifWriteTokenDictionary() converts the current token dictionary to
 *  "Tundra" format and attach it to the list of token dictionaries in the
 *  symbol library.
 *  <BR>
 *  If the multi-page dictionary hasn't been written yet, then it is resolved
 *  and stored first. If a specific page is passed in, then the residual (non-
 *  matching symbols) are stored in a separate dictionary.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  xPage::			Data structure of individual page to store
 *				dictionary of residual
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteTokenDictionary(
	XDoc *xDoc,
	XPage *xPage
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XPage *tmpPage;
    XImg *xImg;
    Bool *dictArray = NULL;
    UInt32 dicts;
    UInt32 headerSize;		/* Number of bytes in the header */
    UInt32 dictBytes;		/* Number of bytes allocated for dict */
    UInt8 *dictBlock = NULL;
    Int32 offset;
    Int32 numSymbols;		/* Number of multi-page symbols */
    Int32 numPageSymbols = 0;	/* Number of unique symbols for this page */
    void *compInfo;
    UInt8 dictVersion;
    UInt32 dictSize, tempLong;	/* Number of bytes used for dictionary */
    UInt32 retval;
    IPWState *pIPWState = NULL;
    UInt32 pageRequest = 0;
    PROLOGUE;

    CheckHandleValidity(xDoc);

    /* Extract the IPCore state structure */
    XERR_PASS(wp_XifIPWState(xDoc, &pIPWState));

    instData = XIF_DOCINST(xDoc);
    xFile= XIF_DOCFILE(xDoc);

    /* Check sequence */
    if ((xDoc->dictFlags & TOKENDICT_WRITABLE) == 0) {
	if (! xPage || (xPage->seqStatus & XPAGE_SEQ_TOKENDICTSAVED) != 0) {
	    NORMAL_RETURN;
	}
    } else if (xPage) {
	/* Make sure we've written the multi-page dictionary first */
	XERR_PASS_CLEANUP(wp_XifWriteTokenDictionary(xDoc, NULL));
    }

    /* Extract the text compression info structure from the document */
    compInfo = xDoc->textCompInfo;
    XERR_GEN_CLEANUP(compInfo == NULL, XIF_INTERNAL_ERROR);

    /* Determine the set of dictionary symbols that will be written */
    if (xPage && xPage->tokenPageID) {
	pageRequest = xPage->tokenPageID;

	/* Get a count of the number of page symbols in the dictionary */
	XERR_PASS_CLEANUP(i_getNumberOfSymbols (pIPWState, compInfo,
	    cSymbolsOnceOnPage, pageRequest, &numPageSymbols));

	/* If no symbols exist, there's no dictionary to write */
	if (numPageSymbols == 0UL) {
	    xPage->seqStatus |= XPAGE_SEQ_TOKENDICTSAVED;

	    /* Continue on and make an estimate of the position block */
	    xImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

	    if (xImg->seqStatus & XIMG_SEQ_TOKENSTART) {
		XERR_PASS_CLEANUP(i_estimateTundraPositionBlockSize (pIPWState,
		    xDoc->textCompInfo, xImg->tmpCmpData->mem,
		    xImg->tmpCmpData->bytes, pageRequest, &xImg->posSize));
	    }
	    NORMAL_RETURN;
	}

    } else {
	pageRequest = 0;

	/* Get a count of the number of multi-page symbols in the dictionary */
	XERR_PASS_CLEANUP(i_getNumberOfSymbols (pIPWState, compInfo,
	    cSymbolsMoreThanOnce, 0, &numSymbols));

	/* If no symbols exist, there's nothing to write */
	if (numSymbols == 0UL) {
	    xDoc->dictFlags &= ~TOKENDICT_WRITABLE;
	    xDoc->seqStatus &= ~XDOC_SEQ_SHAREDDICT;

	    NORMAL_RETURN;
	}

	/* Add to the dictionary the list of single-count symbols from
	 * this page only (prune any single-count symbols from other pages) */
	if (xDoc->seqStatus & XDOC_SEQ_MULTIDICTS) {
	    XERR_PASS_CLEANUP(i_assignDictionaryNumbersForSingleUse(
		pIPWState, compInfo));
	}

    }

    /* Create an array of internal dictionaries to request the size of */
    dicts = xDoc->nextTokenPageID + 1;
    CHECK_MALLOC(dictArray = (Bool *)XIF_MALLOC(dicts * sizeof(Bool)));

    /* Clear the dictionary list */
    memset(dictArray, 0, dicts * sizeof(Bool));

    if (xPage) {
	XERR_GEN_CLEANUP(xPage->tokenPageID > xDoc->nextTokenPageID,
	    XIF_INTERNAL_ERROR);
	dictArray[xPage->tokenPageID] = cTrue;
    } else {
	dictArray[0] = cTrue;	/* Use the shared dictionary */
    }

    /* Estimate the dictionary size when converted to Tundra format */
    XERR_PASS_CLEANUP(i_estimateTundraDictionaryBlockSize (pIPWState,
	xDoc->textCompInfo, dictArray, dicts, pageRequest, &dictBytes));

    /* Compute the estimates for the position blocks, only if the
     * token dictionary has the complete symbol list (multi-count
     * dictionay or individual multi-count and single-count page
     * symbols) */
    if (pageRequest || (xDoc->seqStatus & XDOC_SEQ_MULTIDICTS) == 0) {
	tmpPage = xDoc->pages;
	while (tmpPage) {
	    xImg = (XImg *)tmpPage->pageInfo.xPrimImgHandle;

	    if ((xImg->seqStatus & XIMG_SEQ_TOKENSTART) &&
		(pageRequest == tmpPage->tokenPageID ||
	    pageRequest == 0)) {
		XERR_PASS_CLEANUP(i_estimateTundraPositionBlockSize (pIPWState,
		    xDoc->textCompInfo, xImg->tmpCmpData->mem,
		    xImg->tmpCmpData->bytes, tmpPage->tokenPageID, &xImg->posSize));
	    }
	    tmpPage = tmpPage->next;
	}
    }

    headerSize = cXifTCPosDictSize + sizeof(dictSize);

    /* Loop until the buffer is big enough */
    do {
	dictBytes = (dictBytes + 10) * 11 / 10;

	if (dictBlock) {
	    XIF_FREE(dictBlock);
	    dictBlock = NULL;
	}

	/* Allocate a buffer for the dictionary and header */
	CHECK_MALLOC(dictBlock = (UInt8 *)
		     XIF_MALLOC(dictBytes + headerSize));

	/* Attach a header */
	dictVersion = cXifTCDictVersion;
	memcpy(dictBlock + cXifTCPosDictVersion, &dictVersion,
	    sizeof(dictVersion));

	/* Write the symbols (compressed) to the block */
	retval = i_writeTundraDictionaryData (pIPWState,
	    xDoc->textCompInfo, dictArray, dicts, dictBlock + headerSize,
	    dictBytes, pageRequest, &dictSize);
    } while (retval == ia_bufferExceeded);

    if (retval != ia_successful) {
	__error__ = retval;
	goto ERR_CLEANUP;
    }

    /* Store the size in the header */
    tempLong = dictSize;
    SwapLong(xFile, &tempLong);
    memcpy(dictBlock + cXifTCPosDictSize, &tempLong, sizeof(tempLong));

    /* Store the data in the file */
    XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &offset));
    XIF_WRITE(xFile, dictBlock, headerSize + dictSize);

/* Note: for multiple dictionaries, we would set the size and offset of
 * each here. */

    /* Update the symbol library table */
    /* Mark this dictionary as unwritable (already written) */
    if (pageRequest) {
	xPage->tokenDictID = xDoc->symlibTableHeader.symlibCount;
	XERR_PASS_CLEANUP(wp_XifSetSymlibOffset(xDoc, xPage->tokenDictID,
	    headerSize + dictSize, offset));
	xPage->seqStatus |= XPAGE_SEQ_TOKENDICTSAVED;
    } else {
	XERR_PASS_CLEANUP(wp_XifSetSymlibOffset(xDoc, xDoc->curDictIDs[0],
	    headerSize + dictSize, offset));
	xDoc->dictFlags &= ~TOKENDICT_WRITABLE;
	xDoc->seqStatus |= XDOC_SEQ_SHAREDDICT;
    }

    /* Fall through to free up any temporary memory */

 ERR_CLEANUP:

    if (dictArray) {
	XIF_FREE(dictArray);
	dictArray = NULL;
    }
    if (dictBlock) {
	XIF_FREE(dictBlock);
	dictBlock = NULL;
    }
    NORMAL_RETURN;
} 


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Remap the symbol IDs, convert position/IDs to Tundra, and store.
 *
 * LongDescription:::
 *  wp_XifWriteTokenImage() remaps the symbol IDs and compresses the IDs
 *  and positions using the JBIG-2 "Tundra" format, then stores the compressed
 *  Tundra data. The residual is converted to MMR and stored as a separate
 *  image as a layer 2 image.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  The format of the pre-Tundra header is:
 *	1 byte	version number
 *	4 bytes	Tundra position block byte count
 *	2 bytes	Number of dictionaries associated with the position block
 *	4 bytes	First dictionary ID
 *	...	Next dictionary ID
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteTokenImage(
	XImg *xImg
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XDoc *xDoc;
    UInt8 *posBlock = NULL, *residBlock = NULL;
    UInt32 posSize, residSize;
    UInt32 cmpBytes = 0UL;
    XIF_COMPRESSION compression;
    UInt32 quality;
    UInt32 residImgHandle;
    XImg *rImg;
    UInt32 retval;
    UInt32 xpos, ypos;
    UInt32 width, height;
    UInt8 *ptr;
    UInt8 *tundraData;
    UInt32 tundraBytes;
    UInt32 bytesWritten = 0UL;
    UInt8 posBlockVersion;
    UInt16 dictCount = 0, tempShort;
    UInt32 dict, dictID, tempLong;
    IPWState *pIPWState = NULL;
    UInt32 pageRequest = 0;
    UInt32 binaryCompBytes = 0;
    PROLOGUE;

    CheckHandleValidity(xImg);
    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);
    xDoc = xImg->xPage->xDoc;

    /* Extract the IPCore state structure */
    XERR_PASS(wp_XifIPWState(xDoc, &pIPWState));

    /* Estimate the residual, in case we need to decide whether to store the
     * the primary image */
    if (xImg->xPage->seqStatus & XPAGE_SEQ_TKCOLORADDED) {
	residSize = 0;
    } else {

	/* If this pointer is NULL, there's a good chance that this image is
	 * not the primary image */
	XERR_GEN_CLEANUP(xImg->tmpCmpData == NULL, XIF_INTERNAL_ERROR);

	/* Convert the positions to Tundra format */
	XERR_PASS_CLEANUP(i_estimateTundraResidualBlockSize (pIPWState,
	    xDoc->textCompInfo,	xImg->tmpCmpData->mem,
	    xImg->tmpCmpData->bytes, &residSize));
    }

    if (xImg->posSize > 0 &&
	(xImg->xPage->xDoc->seqStatus & XDOC_SEQ_SHAREDDICT) != 0) {
	dictCount = (UInt16)1;	/* Shared dictionary */
    } else {
	dictCount = 0;
    }

    if (xImg->xPage->seqStatus & XPAGE_SEQ_TKCOLORADDED) {

	/* If we stored a second dictionary, include it also */
	if ((xImg->xPage->tokenDictID > 0) && xImg->posSize > 0) {
	    dictCount++;
	    pageRequest = xImg->xPage->tokenPageID;
	}
    }

    /* Convert the POSITIONS to Tundra format */
    posSize = xImg->posSize;

    /* Loop until we have the right size */
    do {
	posSize = (posSize * 11) / 10 + cXifTCPosPositionDictIDs +
	    2 * sizeof(dictID);	    /* Add a fudge factor */

	/* Allocate a buffer for the Tundra position data */
	CHECK_MALLOC(posBlock = (UInt8 *)XIF_MALLOC(posSize));
	memset(posBlock, 0, posSize);

	/* Create the pre-Tundra header */
	ptr = (UInt8 *)posBlock;
	    
	/* Set to the current Tundra dictionary version */
#define cTundraDictionaryVersion    0
	posBlockVersion = cTundraDictionaryVersion;
	memcpy(ptr + cXifTCPosPositionVersion, &posBlockVersion,
	    sizeof(posBlockVersion));
	    
	tempShort = dictCount;
	SwapShort(xFile, &tempShort);
	memcpy(ptr + cXifTCPosPositionDictCount, &tempShort,
	    sizeof(tempShort));
	    
	/* Add the dictionary ID(s) to the list */
	dict = 0;
	if (xDoc->curDictIDs && xDoc->curDictCount > 0 &&
	    (xDoc->seqStatus & XDOC_SEQ_SHAREDDICT) != 0 &&
	    xImg->posSize > 0) {
	    tempLong = dictID = xDoc->curDictIDs[0];
	    SwapLong(xFile, &tempLong);
	    memcpy(ptr + cXifTCPosPositionDictIDs +
		dict * sizeof(tempLong), &tempLong,
		sizeof(tempLong));
	    dict++;
	}
	if (xImg->xPage->tokenDictID > 0 && xImg->posSize > 0) {
	    tempLong = dictID = xImg->xPage->tokenDictID;
	    SwapLong(xFile, &tempLong);
	    memcpy(ptr + cXifTCPosPositionDictIDs +
		dict * sizeof(tempLong), &tempLong,
		sizeof(tempLong));
	    dict++;
	}
	XERR_GEN_CLEANUP(dict != dictCount, XIF_INTERNAL_ERROR);
	
	tundraData = ptr + cXifTCPosPositionDictIDs +
	    dictCount * sizeof(dictID);
	tundraBytes = posSize - (tundraData - posBlock);
	
	/* Write the Tundra position data to the block */
	if (xImg->posSize > 0UL) {
	    retval = i_writeTundraPositionBlock (pIPWState, xDoc->textCompInfo,
		xImg->tmpCmpData->mem, xImg->tmpCmpData->bytes, tundraData,
		tundraBytes, pageRequest, &cmpBytes);
	    if (retval == ia_bufferExceeded) {
		XIF_FREE(posBlock);
		posBlock = NULL;
	    }
	} else {
	    retval = ia_successful;
	}
    } while (retval == ia_bufferExceeded);
    
    if (retval != ia_successful) {
	__error__ = retval;
	goto ERR_CLEANUP;
    }

    /* Add the Tundra data size to the pre-Tundra header */
    tempLong = cmpBytes;
    SwapLong(xFile, &tempLong);
    memcpy(ptr + cXifTCPosPositionSize, &tempLong, sizeof(tempLong));

    /* Attach to the image (for writing a compressed strip) */
    xImg->cmpBuffer = posBlock;
    if (posSize) {
	xImg->cmpBytes = cmpBytes + (tundraData - posBlock);
    }
    xImg->cmpBufsize = posSize;
    posBlock = NULL;

    /* Write it to the file */
    XERR_PASS_CLEANUP(wp_TifWriteIFD(xImg));
    if (xImg->cmpBytes) {
	XERR_PASS_CLEANUP(wp_TifWriteStrip(xImg));
    }

    /* Complete the write for this image (do cleanup and memory freeing) */
    if (xImg->cmpBuffer) {
	XIF_FREE(xImg->cmpBuffer);
	xImg->cmpBuffer = NULL, xImg->cmpBufsize = xImg->cmpBytes = 0UL;
    }

    /* Now work on the RESIDUAL... */

    if (residSize) {

	/* Loop until we have the right size */
	do {
	    residSize = (residSize * 11) / 10 + 8;    /* Add a fudge factor */

	    /* Allocate a buffer for the Tundra residual data */
	    CHECK_MALLOC(residBlock = (UInt8 *)XIF_MALLOC(residSize));

	    /* Write the Tundra position data to the block */
	    retval = i_writeTundraResidualBlock (pIPWState, xDoc->textCompInfo,
		xImg->tmpCmpData->mem, xImg->tmpCmpData->bytes, residBlock,
		residSize, &xpos, &ypos, &width, &height, &cmpBytes);
	    if (retval == ia_bufferExceeded) {
		XIF_FREE(residBlock);
		residBlock = NULL;
	    }
	} while (retval == ia_bufferExceeded);

	if (retval != ia_successful) {
	    __error__ = retval;
	    goto ERR_CLEANUP;
	}

	/* Create a residual image attached to this parent */
	XERR_PASS_CLEANUP(XIF_AddImage((UInt32)xImg->xPage, 2,
	    XIF_COLORTYPE_BINARY, xpos, ypos, xImg->imgInfo.dwXResolution,
	    xImg->imgInfo.dwYResolution, width, height, height, NULL,
	    &residImgHandle));
	rImg = (XImg *)residImgHandle;

	/* Write it to the file, precompressed */
	compression = XIF_G4;
	quality = 0UL;
	XERR_PASS_CLEANUP(XIF_ImageWriteStart(residImgHandle,
	    XIF_IMGFLAGS_Compressed, &compression, &quality));
	XERR_PASS_CLEANUP(XIF_CompressedWriteStrip(residImgHandle, 0, height,
	    residBlock, cmpBytes, &bytesWritten));
	XERR_PASS_CLEANUP(XIF_ImageWriteFinish(residImgHandle));
    } /* If there's residual data */

    /* Attach the FG image using the same compressed image stream, and
     * removed the color section from the size of the stream just added */
    if (binaryCompBytes) {
	XERR_PASS_CLEANUP(XIF_ShareColorTokenImage((XIF_IMGHANDLE)xImg,
	    binaryCompBytes, XIF_COLORTYPE_COLOR24));
    }


    /* Fall through to free up any temporary buffers */

 ERR_CLEANUP:

    if (posBlock) {
	XIF_FREE(posBlock);
	posBlock = NULL;
    }
    if (residBlock) {
	XIF_FREE(residBlock);
	residBlock = NULL;
    }
    NORMAL_RETURN;
} 


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reuse the color tokenized image data for both the mask and color layers.
 *
 * LongDescription:::
 *  XIF_ShareColorTokenImage() copies the token image IFD for an image in an
 *  even layer (mask layer, already stored) to a matching image in the layer
 *  just above (foreground layer) image, and shares the compressed image data
 *  (both size and location); then resets the byte count of the compressed
 *  image in the even layer image to the binary subset, described by the byte
 *  count passed in (second parameter).
 *  <BR>
 *  Note that the simple case is layers 2 and 3, where the color tokenized
 *  stream is first stored in layer 2 as a binary image, but with the color
 *  section included. Then the correct size of the binary portion is passed
 *  in to XIF_SharedColorTokenImage(), which then clones the IFD to make a
 *  layer 3 IFD (modifying the color channel count and photometric), points
 *  to the exact same data in it's entirety, then modifies the layer 2 byte
 *  count.
 *  <BR>
 *  Ya, a bit hokey, but it lets you keep the previous way of doing things,
 *  with a minimal number of extra calls.
 *
 * Input:::
 *  xImgHandle::		Valid image handle
 *  dwBinaryImageBytes::	Byte count of compressed image data (already
 *				stored in IFD of even layer) of the binary
 *				subsection. The current byte count will
 *				be relayed to the IFD of the next layer,
 *				and this value will be written over the
 *				current value.
 *  xColorType::		Color type of color tokens store in odd
 *				layer IFD
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
XIF_ShareColorTokenImage(
	XIF_IMGHANDLE xImgHandle,
	UInt32 dwBinaryImageBytes,
	XIF_COLORTYPE xColorType
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XImg *xImg = (XImg *)xImgHandle;	/* Token image (even layer) */
    UInt32 colorTokenBytesOffset = 0;	/* File offset of even layer image
				         * byte count */
    UInt32 colorTokenBytes = 0;	/* Original color token byte count */
    XIF_IMGHANDLE fgImgHandle;	/* FG image handle */
    XImg *fgImg = NULL;		/* FG image being added */
    UInt32 lval;		/* Temporary long value, used for byte
				 * swapping */
    PROLOGUE;

    CheckHandleValidity(xImg);
    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* Make sure it's token compressed */
    XERR_GEN_CLEANUP(xImg->imgInfo.xCompression != XIF_TOKENIZATION &&
	xImg->imgInfo.xCompression != XIF_LLTOKENIZATION, XIF_BADPARAMETER);

    /* Get the previously stored token compression stream size */
    XERR_GEN_CLEANUP((xImg->fields.StripOffsets == NULL ||
	xImg->fields.StripByteCounts == NULL), XIF_SEQUENCE);
    colorTokenBytesOffset = xImg->fields.StripOffsets[0];
    colorTokenBytes = xImg->fields.StripByteCounts[0];
    XERR_GEN_CLEANUP(colorTokenBytesOffset == 0, XIF_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(colorTokenBytes == 0, XIF_INTERNAL_ERROR);

    /* Create a new image for the color tokenized foreground */
    XERR_PASS_CLEANUP(XIF_AddImage((XIF_PAGEHANDLE)xImg->xPage,
	xImg->imgInfo.dwImgLayer + 1, xColorType, 0, 0,
	xImg->imgInfo.dwXResolution, xImg->imgInfo.dwYResolution,
	xImg->imgInfo.dwWidth, xImg->imgInfo.dwHeight,
	 xImg->imgInfo.dwHeight, NULL, &fgImgHandle));
    fgImg = (XImg *)fgImgHandle;

    /* Mark it with the same flavor of token compressed */
    fgImg->imgInfo.xCompression = xImg->imgInfo.xCompression;

    /* Write the new IFD */
    XERR_PASS_CLEANUP(wp_TifWriteIFD(fgImg));

    /* Point the strip data to the existing token stream */
    XERR_GEN_CLEANUP((fgImg->offsetsLocation == 0 ||
	fgImg->byteCountsLocation == 0), XIF_SEQUENCE);
    XIF_SEEK(xFile, fgImg->offsetsLocation);
    lval = colorTokenBytesOffset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));
    XIF_SEEK(xFile, fgImg->byteCountsLocation);
    lval = colorTokenBytes;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));

    /* Now go back to the original image, and revise the byte count */
    XIF_SEEK(xFile, xImg->byteCountsLocation);
    lval = dwBinaryImageBytes;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));

    NORMAL_RETURN;

 ERR_CLEANUP:
    ERR_RETURN;
} 


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Lossy dither compress a PIXR binary image
 *
 * LongDescription:::
 *  wp_XifCompressStripLossyDither() compresses an image using the lossy
 *  dither algorithm.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pImgData::			Image data buffer (unframed)
 *  dwBytesPerRow::		Bytes per scanline
 *  dwRowOffset::		Row offset of image data to compress
 *  dwStripHeight::		Number of rows to compress
 *
 * Output:::
 *  ppCmpBuffer::		Pointer to the compression buffer (may be
 *				reallocated)
 *  pdwCmpBufsize::		Number of bytes in compression buffer (may be
 *				revised)
 *  pdwBytesFilled::		Number of bytes filled, in compression buffer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifCompressStripLossyDither(
    XImg *xImg,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt8 **ppCmpBuffer,
    UInt32 *pdwCmpBufsize,
    UInt32 *pdwBytesFilled
)
{
    Instance *instData;
    XIF_IMGINFO *imgInfo;
    IPWState *pIPWState;
    UInt32 imgExh, cmpExh;
    UInt32 compinfoSize;
    void *compInfo;
    UInt8 *newbuf = NULL;
    UInt32 newbufSize;
    UInt8 *src, *dst;
    Int32 bytesUsed, bytesLeft, bytesFilled, pct;
    UInt32 rowsLeft;
    Bool imgContinued;
    UInt32 stage = 1;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesFilled = 0;

    /* Reference the overall image */
    instData = XIF_IMGINST(xImg);
    imgInfo = &xImg->imgInfo;
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));
    XERR_GEN_CLEANUP(ppCmpBuffer == NULL, XIF_BADPARAMETER);


    if (xImg->tmpCmpData == NULL) {	/* First strip */

	/* Get the size of the necessary compression info */
	XERR_PASS_CLEANUP(i_getSizeOfLDCompressionInfo (pIPWState, 
	    imgInfo->dwWidth + 2 * cFrameBits, &compinfoSize));
		
	/* Allocate temporary compression data, attach to image */
	CHECK_MALLOC(xImg->tmpCmpData =
	    (MISC_MEM *)XIF_MALLOC(sizeof(MISC_MEM) + compinfoSize));
	xImg->tmpCmpData->next = NULL;
		
	/* Data follows immediately */
	xImg->tmpCmpData->mem = compInfo = (void *)(xImg->tmpCmpData + 1);

	/* Initialize the lossy-dither state */
	XERR_PASS_CLEANUP(i_initializeLDCompressionInfo (pIPWState, compInfo,
	    imgInfo->dwWidth + 2 * cFrameBits, compinfoSize));
    } else {
	compInfo = xImg->tmpCmpData->mem;
    }
    
    /* If we don't have a compressed data buffer, create one.
     * If it's not big enough from the previous loop, increase
     * the size. */
    if (*ppCmpBuffer == NULL) {

	/* Estimate the size of the buffer, use a quarter of the image size */
	*pdwCmpBufsize = (imgInfo->dwWidth * imgInfo->dwHeight / 4) / 8;

	CHECK_MALLOC(*ppCmpBuffer = XIF_MALLOC(*pdwCmpBufsize));
    }

    /* Prepare to loop through the data until all image data is compressed */
    src = pImgData - ((dwBytesPerRow << cLogFrameBits) + cFrameBytes);
    rowsLeft = dwStripHeight;
    dst = *ppCmpBuffer;
    bytesLeft = *pdwCmpBufsize;
    if (dwRowOffset == 0) {
	imgContinued = cFalse;
    } else {
	imgContinued = cTrue;
    }
    bytesUsed = 0UL;
    do {

	bytesFilled = 0;
	switch ((int)stage) {
	case 1:

	    /* Compress using Lossy Dither */
	    XERR_PASS_CLEANUP(i_compressRowsLD (pIPWState, compInfo, src,
		imgInfo->dwWidth + 2 * cFrameBits, dwRowOffset + cFrameBits,
		rowsLeft, dwBytesPerRow, dst, bytesLeft, imgContinued,
		imgContinued, &imgExh, &cmpExh, &pct, &bytesFilled));

	    /* This is an illegal state */
	    XERR_GEN_CLEANUP((imgExh == cFalse) && (cmpExh == cFalse),
		XIF_INTERNAL_ERROR);

	    break;
	case 2:
	case 3:	/* Place 2 EOL markers */

	    /* Place an end-of-file marker */
	    XERR_PASS_CLEANUP(i_writeLDEOL (pIPWState, compInfo, dst, bytesLeft,
		&bytesFilled));
	    break;
	case 4:

	    /* Flush the stream */
	    XERR_PASS_CLEANUP(i_flushLD (pIPWState, compInfo, dst, bytesLeft,
		&bytesFilled));
	}

	/* Update byte counts/pointers */
	bytesUsed += bytesFilled;
        bytesLeft -= bytesFilled;
	dst += bytesFilled;

	/* Increase the compressed buffer size, if needed */
	if (imgExh == cFalse || bytesLeft < 10) {
	    
	    /* Filled the compressed buffer, reallocate the compressed
	     * data buffer */
	    
	    /* Anticipate the compression due to the percentage, plus a
	     * fudge */
	    newbufSize = *pdwCmpBufsize * (200 - pct + 20) / 100;
	    CHECK_MALLOC(newbuf = XIF_MALLOC(newbufSize));
	    memcpy(newbuf, *ppCmpBuffer, bytesUsed);
	    
	    /* Replace old buffer */
	    XIF_FREE(*ppCmpBuffer);
	    *ppCmpBuffer = newbuf;
	    *pdwCmpBufsize = newbufSize;
	    newbuf = NULL;
	    dst = *ppCmpBuffer + bytesUsed;
	    bytesLeft = newbufSize - bytesUsed;
	    imgContinued = cTrue;
	}
	if (imgExh == cTrue) {
	    stage++;
	}
    } while (imgExh == cFalse || stage < 5);
    
    /* Can't compress given the buffer size, we should have continued in the
     * loop */
    XERR_GEN_CLEANUP((cmpExh == cTrue && imgExh == cFalse), XIF_INTERNAL_ERROR);

    *pdwBytesFilled = bytesUsed;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newbuf) {
	XIF_FREE(newbuf);
	newbuf = NULL;
    }
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Assign the color to a connected component, found during tokenization
 *
 * LongDescription:::
 *  wp_XifAssignTokenColor() determines the color associated with a binary
 *  connected component, found during tokenization. If the color is fairly
 *  constant, then the color will be placed in a list, and the color ID will
 *  be attached to the token symbol. If the color is not fairly consistent,
 *  then a miniature image is kept with the token dictionary, and the color
 *  ID for the image is attached to the token symbol.
 *
 * Input:::
 *  pIPWState::			Pointer to IPWrap state structure
 *  textInfo::			Text info structure used in symbol compression
 *  callbackData::		Callback function used to assign color to
 *				symbol
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifAssignTokenColor(
	IPWState *pIPWState,
	void *textInfo,
	void *callbackData
)
{
    TCData *tcData = (TCData *)callbackData;
    PROLOGUE;

    pIPWState = pIPWState;
    textInfo = textInfo;
    tcData = tcData;

    NORMAL_RETURN;
#if 0    
 ERR_CLEANUP:
    ERR_RETURN;
#endif
}


/*
$Log: xifwtok.c,v $
Revision 1.1.1.1  1999/09/20 18:19:21  jlin
Imported sources

Revision 1.35  1998/08/06 16:43:20  aber
aber on Thu Aug 6 09:43:05 PDT 1998

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Use unused variables, to avoid Watcom complaints.

Revision 1.34  1998/08/05 23:31:11  aber
aber on Wed Aug 5 16:30:58 PDT 1998

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Fixed coreCheckC complaints.

Revision 1.33  1998/08/04 00:21:39  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.32  1998/08/01 19:53:43  aber
aber on Sat Aug 1 12:53:26 PDT 1998

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Correct the name of the parameter in the comment.

Revision 1.31  1998/07/30 22:41:55  aber
aber on Thu Jul 30 15:41:55 PDT 1998

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: New function XIF_ShareColorTokenImage(), used to add a FG image IFD, while
reusing the existing token data, but then overwriting the byte count of
the binary image to reflect the non-color subset of the token stream.

Revision 1.30  1998/05/21 06:14:55  aber
aber on Wed May 20 23:08:24 PDT 1998

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Remove temporary fix intended to be "nice". If the resolution is set too
low (like 0), I automatically bumped the resolution to what we normally
see (300). However, I have one case where I've forced the text plane
to be very low (10 dpi), and this "nice" fix causes it to crash.
So for now, we'll force the file to have the correct resolution.

Revision 1.29  1998/04/03 00:00:55  aber
aber on Thu Apr 2 15:58:49 PST 1998

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Fix a bug when tokenizing a page (first dictionary, but not yet saved),
then copying a second token dictionary from another file (now the first),
then saving the thought-to-be first token dictionary; which is now the
second.

Revision 1.28  1998/02/19 21:28:14  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Rename local text allocation routines to use wp_ names.

Revision 1.27  1998/01/20 18:06:11  aber
aber on Tue Jan 20 10:05:57 PST 1998

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Compiler signed/unsigned complaint fix.

Revision 1.26  1998/01/20 03:25:15  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Reset the current dictionary status, when writing and then reading back
a page that has been tokenized.

Revision 1.25  1998/01/14 00:16:44  ddavies
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.24  1997/11/26 23:28:06  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Added a check to make sure that the tokenized image (supposedly the
primary image) does indeed have the tokenized data attached to the
image structure.

Revision 1.23  1997/10/01 23:44:22  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Loop with more memory until the dictionary compresses. Don't add
the dictionary offset to the header when the tokenization stream
is empty.

Revision 1.22  1997/08/31 17:09:31  aber
aber on Sun Aug 31 10:07:50 PDT 1997

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Handle case where residual buffer doesn't increase (Andrew found bug).

Revision 1.21  1997/08/27 23:40:05  aber
aber on Wed Aug 27 16:39:41 PDT 1997

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Fix the token dictionary ID, start at 0, not -1.

Revision 1.20  1997/08/27 00:17:40  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Avoid repeated file seeks.

Revision 1.19  1997/08/15 23:48:03  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.18  1997/08/15 01:54:19  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Modify token dictionary header to include the Tunrda dictionary
byte count, allows for extensible dictionary data.

Revision 1.17  1997/08/15 00:11:41  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Use predefined offsets for pre-Tundra header.

Revision 1.16  1997/08/07 21:27:50  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.15  1997/07/02 01:09:04  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Use the same unframed parameters for both stipple and non-stipple modes.

Revision 1.14  1997/06/26 23:29:22  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Make potential ansification fixes that the Mac might complain about.

Revision 1.13  1997/06/26 22:34:10  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Check for no symbols, don't output the token dictionary if such,
but store the minimum position information with a dictionary count of 0.

Revision 1.12  1997/06/23 22:48:10  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Sync up image pointers so that lossy dither works right.

Revision 1.11  1997/06/17 22:35:53  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Define renames.

Revision 1.9  1997/06/12 00:52:42  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Enable lossless tokenization.

Revision 1.7  1997/06/10 23:35:00  aber
aber on Tue Jun 10 16:34:20 PDT 1997

xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

Clean up temporary memory, fixes memory leak. Fix documentation.

Revision 1.6  1997/06/10 01:27:04  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:21  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:25  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:35  aber
xifwtok.c -> /product/ipcore/ipwrap/src/RCS/xifwtok.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:50  aber
Initial revision

*/
