/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwwave.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	xifwwave.c:	XIFF WRITE WAVE source
 *
 *	Compression (write) for WAVE.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "iw_wave.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xifimage.h"
#include "xifrtiff.h"
#include "xifwwave.h"
#include "xifrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwwave.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:21 $");

typedef struct WriteData {
    Instance *instData;
    UInt8 **ppBuffer;
    UInt8 *next;
    UInt32 *pBufBytes;
    UInt32 bytesLeft;
} WriteData;

static Int32 CDECL
wp_Wavelet_WriteCallback(
    IPWState *pIPWState,
    void *userData,
    UInt8 *ptr,
    UInt32 bytes
)
{
    Instance *instData;
    WriteData *writeData = (WriteData *)userData;
    UInt32 newbufsize;
    UInt8 *newbuf = NULL;
    PROLOGUE;

    pIPWState = pIPWState;
    if (writeData->bytesLeft < bytes) {
	instData = writeData->instData;

	/* Must reallocate the buffer to accomodate the new amount,
	 * increase by 20% more than what we know we need now */
	newbufsize = (*writeData->pBufBytes - writeData->bytesLeft + bytes)
	    * 120 / 100;
	CHECK_MALLOC(newbuf = XIF_MALLOC(newbufsize));

	/* Copy the data to the new buffer */
	memcpy(newbuf, *writeData->ppBuffer,
	    writeData->next - *writeData->ppBuffer);
	writeData->next = newbuf +
	    (*writeData->pBufBytes - writeData->bytesLeft);
	writeData->bytesLeft = newbufsize -
	    (*writeData->pBufBytes - writeData->bytesLeft);

	/* Update the external buffer */
	XIF_FREE(*writeData->ppBuffer);
	*writeData->ppBuffer = newbuf;
	*writeData->pBufBytes = newbufsize;

	/* Dereference the now stored buffer */
	newbuf = NULL;
    }
    memcpy(writeData->next, ptr, bytes);
    writeData->next += bytes;
    writeData->bytesLeft -= bytes;
    return 0;

ERR_CLEANUP:
    if (newbuf) {
	instData = writeData->instData;
	XIF_FREE(newbuf);
    }

    return cErrNomem;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Wavelet compress a strip of image data
 *
 * LongDescription:::
 *  wp_XifCompressStripWAVE() WAVE compresses a strip of image data.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *  pPixr::			Pointer to image PIXR
 *  dwRowOffset::		Row offset of image strip to compress
 *				(MUST BE 0)
 *  dwStripHeight::		Number of rows to compress
 *				(MUST BE THE PIXR HEIGHT)
 *
 * Output:::
 *  ppCmpData::			Pointer to pointer to the compression buffer,
 *				may be reallocated if not enough space
 *  pdwCmpBytes::		Number of bytes in compression buffer, may
 *				be modified if not enough space
 *  pdwBytesFilled::		Number of bytes filled, in compression buffer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifCompressStripWavelet(
    XImg *xImg,
    PIXR *pPixr,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt8 **ppCmpData,
    UInt32 *pdwCmpBytes,
    UInt32 *pdwBytesFilled
)
{
    XIF_IMGINFO *pxImgInfo;
    IPWState *pIPWState;
    WriteData writeData;
    Int32 ratio;
    Int32 pixrHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesFilled = 0;

    if (dwStripHeight == 0) {
        NORMAL_RETURN;
    }

    /* Can't offset into PIXR for now, nor can we compress a portion of
     * the PIXR */
    pxImgInfo = &xImg->imgInfo;
    XERR_GEN_CLEANUP(dwRowOffset != 0, XIF_BADPARAMETER);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_XifIPWState(xImg->xPage->xDoc, &pIPWState));

    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    XERR_GEN_CLEANUP(dwStripHeight != pixrHeight, XIF_BADPARAMETER);

    /* Initialize */
    writeData.instData = XIF_IMGINST(xImg);
    writeData.ppBuffer = ppCmpData;
    writeData.next = *writeData.ppBuffer;
    writeData.pBufBytes = pdwCmpBytes;
    writeData.bytesLeft = *writeData.pBufBytes;

    /* Determine a crude mapping from JPEG quality to compression ratio */
    {
      Int32 quality, index;
      Int32 qualities[6] = {0, 10, 50, 75, 90, 100};
      Int32 ratios[6] = {100, 64, 15, 10, 5, 4};

      quality = xImg->JPEGquality;
      if (quality < 0) quality = 0;
      if (quality > 100) quality = 100;
      index = 0;
      while (quality > qualities[index]) {
	index++;
      }

      /* Use exact quality conversion */
      if (quality == qualities[index]) {
	ratio = ratios[index];
      } else {	/* Linearly interpolate between values */
	ratio = ratios[index-1] + (quality - qualities[index-1]) *
	  (ratios[index] - ratios[index-1]) /
	  (qualities[index] - qualities[index-1]);
      }
    }

    /* Compress the PIXR */
    XERR_PASS_CLEANUP(w_waveCompress_option(pIPWState, pPixr,
	wp_Wavelet_WriteCallback, (void *)&writeData, 0L, ratio));

    *pdwBytesFilled = *writeData.pBufBytes - writeData.bytesLeft;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    ERR_RETURN;
}

/*
$Log: xifwwave.c,v $
Revision 1.1.1.1  1999/09/20 18:19:21  jlin
Imported sources

Revision 1.11  1998/01/14 00:16:46  ddavies
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.10  1997/11/01 21:00:53  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

Expect strips to equal the entire PIXR (PIXR is now a TIFF strip).

Revision 1.9  1997/09/02 20:02:57  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.8  1997/07/30 01:09:35  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

Bug fix: wavlet write routine expects error condition returned, not byte count.

Revision 1.7  1997/06/26 22:35:21  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

Enforce provision that Wavelet image must be the entire image.
Calls don't allow offsetting into a PIXR to compress just a portion.

Revision 1.6  1997/06/17 01:29:44  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

CommonCheck: XIFF 2.0 support. Annotation bug fix.

Revision 1.5  1997/06/10 01:27:05  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.4  1997/05/30 00:51:23  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.3  1997/05/20 22:29:27  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.2  1997/05/20 00:54:37  aber
xifwwave.c -> /product/ipcore/ipwrap/src/RCS/xifwwave.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:51  aber
Initial revision

*/
