/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1997, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * xifwxiff.c
 *
 *	XIFF write of XIFF specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#ifdef _MSC_VER
#pragma pack(push, PRE_XIFF, 1)
#endif
#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_XIFF)
#endif

#include "xiff.h"
#include "xifimage.h"
#include "xifwtiff.h"
#include "xifwxiff.h"
#include "xifwpriv.h"
#include "xifrxiff.h"
#include "xifrpriv.h"

/* The idea is to remove this one day: */
#include "xiff.pub"

IP_RCSINFO(RCSInfo, "$RCSfile: xifwxiff.c,v $; $Revision: 1.1.1.1 $; $Date: 1999/09/20 18:19:20 $");

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an individual XIFF extension (at current file location)
 *   
 *
 * LongDescription:::
 *   wp_XifWriteExtension() writes an individual XIFF extension (DocField) at
 *   the current file location.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  docField::			XIFF fields to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteExtension(
    XDoc *xDoc,
    DocField *docField
)
{
    XFile *xFile = XIF_DOCFILE(xDoc);
    PROLOGUE;

    XIF_WRITE(xFile, &docField->tagID, sizeof(docField->tagID));
    XIF_WRITE(xFile, &docField->offset, sizeof(docField->offset));

    NORMAL_RETURN;
ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write the XIFF extension fields at the current file location.
 *   
 *
 * LongDescription:::
 *   wp_XifWriteExtensionList() writes the XIFF extension fields (DocFields) at the
 *   current file location.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteExtensionList(
    XDoc *xDoc
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt16 tagCount = 0;
    DocField docField;
    UInt32 docFieldBytes;
    Int32 fileOffset;
    PROLOGUE;
    
    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    docFieldBytes = sizeof(docField.tagID) + sizeof(docField.offset);

    /* Write the document tag count (come back later and fill it in) */
    XIF_SEEK(xFile, XIFF_EXTS_OFFSET);
    XIF_WRITE(xFile, &tagCount, sizeof(tagCount));
    fileOffset = XIFF_EXTS_OFFSET + sizeof(tagCount);

    /* Write a pagetable extension */
    docField.tagID = DOCTAG_PAGETABLE;
    docField.offset = xDoc->pageTableHeaderOffset;
    xDoc->tagOffset[docField.tagID] = fileOffset;
    XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));
    fileOffset += docFieldBytes;
    tagCount++;
    
    /* Write an author extension */
    docField.tagID = DOCTAG_AUTHOR;
    docField.offset = xDoc->authorOffset;
    xDoc->tagOffset[docField.tagID] = fileOffset;
    XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));
    fileOffset += docFieldBytes;
    tagCount++;
    
    /* Write a reference page extension */
    docField.tagID = DOCTAG_REFERENCEPAGE;
    docField.offset = xDoc->refPageOffset;
    xDoc->tagOffset[docField.tagID] = fileOffset;
    XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));
    fileOffset += docFieldBytes;
    tagCount++;
    
    /* Write a symbol library table extension */
    if (xDoc->docInfo.version >= 3) {
	docField.tagID = DOCTAG_SYMLIBTABLE;
	docField.offset = xDoc->symlibTableHeaderOffset;
	xDoc->tagOffset[docField.tagID] = fileOffset;
	XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));
	fileOffset += docFieldBytes;
	tagCount++;
    }
    
    /* Write a next extensions extension */
    docField.tagID = DOCTAG_NEXTEXT;
    docField.offset = 0UL;
    xDoc->tagOffset[docField.tagID] = fileOffset;
    XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));
    fileOffset += docFieldBytes;
    tagCount++;

    /* Go back and write the actual tag count */
    XIF_SEEK(xFile, XIFF_EXTS_OFFSET);
    XIF_WRITE(xFile, &tagCount, sizeof(tagCount));
    
    /* END OF EXTENSIONS, now append the data to the end of the extension list */

    XERR_PASS_CLEANUP(wp_XifWritePageTableHeader(xDoc));

    /* Mark information as set */
    xDoc->seqStatus |= XDOC_SEQ_GETINFO;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    ERR_RETURN;    
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write XIFF extension fields data, derived from offsets
 *
 * LongDescription:::
 *  wp_XifWriteFieldData() writes the data associated with a XIFF fields
 *  (DocFields).
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  pxDocInfo::			Pointer to document info, internal data copied
 *				here to expose to client
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress(incomplete)
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteFieldData(
    XDoc *xDoc,
    XIF_DOCINFO *pxDocInfo
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt16 sval; UInt32 lval;
    PROLOGUE;
    
    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    
    /* Check for proper calling sequence */
    
    
    /* Write the page table header */
    XIF_SEEK(xFile, xDoc->pageTableHeaderOffset);
    sval = xDoc->pageTableHeader.pageCount;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xDoc->pageTableHeader.entriesPerTable;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xDoc->pageTableHeader.lastOrdinal;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    lval = xDoc->pageTableHeader.firstTableOffset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));

    /* Fill in the page count */
    xDoc->docInfo.wNPages = xDoc->pageTableHeader.pageCount;
    pxDocInfo->wNPages = xDoc->pageTableHeader.pageCount;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write the XIFF fields or DocFields and data.
 *
 * LongDescription:::
 *  wp_XifWriteExtensions() write extension "DocFields" and data associated with
 *  the DocFields.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteExtensions(
    XDoc *xDoc
)
{
    Instance *instData = NULL;
    XFile *xFile;
    PROLOGUE;
    
    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    
    /* Check for proper calling sequence */


    /* Seek to base extension location */
    XIF_SEEK(xFile, XIFF_EXTS_OFFSET);
    
    /* Write the extension list */
    XERR_PASS_CLEANUP(wp_XifWriteExtensionList(xDoc));

    /* Fill in the extension data */
    XERR_PASS_CLEANUP(wp_XifWriteFieldData(xDoc, &xDoc->docInfo));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write XIFF header and necessary components to initiate a XIFF file.
 *
 * LongDescription:::
 *  wp_XifWriteHeader() writes the TIFF and XIFF headers, and the necessary
 *  components to initiate a XIFF file.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  pVersion::			Pointer to the XIFF version
 *  pRevision::			Pointer to the XIFF revision
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteHeader(
    XDoc *xDoc,
    UInt8 *pVersion,
    UInt8 *pRevision
)
{
    Instance *instData = NULL;
    XFile *xFile;
    PROLOGUE;
    
    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    
    /* Check for proper calling sequence */


    /* Begin a write sequence */
    if ((xDoc->seqStatus & XDOC_SEQ_WRITESTART) == 0) {
	xDoc->seqStatus |= XDOC_SEQ_WRITESTART;
    }
    
    /* Write a TIFF header first */
    XERR_PASS_CLEANUP(wp_TifWriteHeader(xFile, 0UL));
    
    /* Seek to XIFF key location */
    XIF_SEEK(xFile, XIFF_KEY_OFFSET);

    /* Write the key and version/revision */
    if (*pVersion < 3) {
	XIF_WRITE(xFile, XIFF_KEY_2_0, XIFF_KEY_LENGTH);
    } else {
	XIF_WRITE(xFile, XIFF_KEY_3_0, XIFF_KEY_LENGTH);
    }
    XIF_WRITE(xFile, pVersion, sizeof(*pVersion));
    XIF_WRITE(xFile, pRevision, sizeof(*pRevision));
    xDoc->docInfo.version = *pVersion;
    xDoc->docInfo.revision = *pRevision;
    
    /* Write the extension list */
    XERR_PASS_CLEANUP(wp_XifWriteExtensionList(xDoc));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write or rewrite the author block
 *
 * LongDescription:::
 *  wp_XifSetAuthor() writes the authoring block of data and attaches it to the
 *  XIF. If the authoring block previously existed, it is overwritten (if it
 *  fits), or appended to the file.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  pAuthorBlock::		Pointer to the authoring block of data
 *  wAuthorBytes::		Number of bytes in authoring block
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifSetAuthor(
	XDoc *xDoc,
    	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
)
{
    Instance *instData = NULL;
    XFile *xFile = XIF_DOCFILE(xDoc);
    Int32 fileOffset = 0UL;
    DocField docField;
    UInt16 sval;
    PROLOGUE;

    instData = XIF_DOCINST(xDoc);

    /* Make sure we're allowed to write */
    XERR_GEN_CLEANUP((xDoc->readWrite & WRITE_ALLOWED) == 0, XIF_BADFILEMODE);

    /* If the file is read/write, check to see if we can write over the previous
     * authoring block */
    if (xDoc->readWrite & READ_ALLOWED) {

	/* Make sure we've read the current authoring block first */
	XERR_GEN_CLEANUP((xDoc->seqStatus & XDOC_SEQ_GETINFO) == 0,
	    XIF_INTERNAL_ERROR);

	if (xDoc->docInfo.wAuthorBlockBytes >= wAuthorBytes) {
	    fileOffset = xDoc->authorOffset;

	    /* Seek to the authoring block location */
	    XIF_SEEK(xFile, fileOffset);
	}
    }
    
    /* If we haven't found a spot, either it didn't fit, or there wasn't one */
    if (fileOffset == 0L) {

	/* Place it at the end of the file */
	XERR_PASS(wp_FileLongAlign(xFile, &fileOffset));
	xDoc->authorOffset = fileOffset;
    }

    /* Write the authoring block out */
    sval = wAuthorBytes;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    if (wAuthorBytes > 0) {
	XERR_GEN_CLEANUP(pAuthorBlock == NULL, XIF_BADPARAMETER);
	XIF_WRITE(xFile, pAuthorBlock, wAuthorBytes);
	CHECK_MALLOC(xDoc->docInfo.pAuthorBlock = XIF_MALLOC(wAuthorBytes));
	memcpy(xDoc->docInfo.pAuthorBlock, pAuthorBlock, wAuthorBytes);
    } else {
	xDoc->docInfo.pAuthorBlock = NULL;
    }
    xDoc->docInfo.wAuthorBlockBytes = wAuthorBytes;

    /* Update the XIFF authoring block fields
     */
    docField.tagID = DOCTAG_AUTHOR;
    docField.offset = xDoc->authorOffset;
    fileOffset = xDoc->tagOffset[docField.tagID];

    /* Write the authoring block fields at the correct location */
    XIF_SEEK(xFile, fileOffset);
    XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write or rewrite the reference page offset
 *
 * LongDescription:::
 *  wp_XifSetReferencePage() writes the reference page offset to the XIF. If
 *  the reference page was previously written, it is overwritten; otherwise
 *  it is appended to the file.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  dwReferencePageOffset::	Reference page offset
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifSetReferencePage(
	XDoc *xDoc,
	Int32 dwReferencePageOffset
)
{
    XFile *xFile = XIF_DOCFILE(xDoc);
    Int32 fileOffset = 0UL;
    DocField docField;
    Int32 lval;
    PROLOGUE;

    /* Make sure we're allowed to write */
    XERR_GEN_CLEANUP((xDoc->readWrite & WRITE_ALLOWED) == 0, XIF_BADFILEMODE);

    /* If the file is read/write, check to see if we can write over the
     * previous authoring block */
    if (xDoc->readWrite & READ_ALLOWED) {

	/* Make sure we've read the current authoring block first */
	XERR_GEN_CLEANUP((xDoc->seqStatus & XDOC_SEQ_GETINFO) == 0,
	    XIF_INTERNAL_ERROR);

	if ((fileOffset = xDoc->refPageOffset) > 0) {

	    /* Seek to the existing reference page data location */
	    XIF_SEEK(xFile, fileOffset);
	}
    }
    
    /* If we haven't found a spot, either it didn't fit, or there wasn't one */
    if (fileOffset == 0L) {

	/* Place it at the end of the file */
	XERR_PASS(wp_FileLongAlign(xFile, &fileOffset));
	xDoc->refPageOffset = fileOffset;
    }

    /* Write the authoring block out */
    xDoc->docInfo.dwReferencePageOffset = dwReferencePageOffset;

    /* Stored value is page number, not offset */
    lval = xDoc->docInfo.dwReferencePageOffset + 1;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));

    /* Update the XIFF reference page fields
     */
    docField.tagID = DOCTAG_REFERENCEPAGE;
    docField.offset = fileOffset;
    fileOffset = xDoc->tagOffset[docField.tagID];

    /* Write the reference page fields at the correct location */
    XIF_SEEK(xFile, fileOffset);
    XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a page table entry.
 *
 * LongDescription:::
 *   wp_XifWritePageTableEntry() writes an individual page table entry.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  fileOffset::		Where to write the table entry
 *  pageTableEntry::		Table entry to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWritePageTableEntry(
    XDoc *xDoc,
    Int32 fileOffset,
    PageTableEntry *pageTableEntry
)
{
    XFile *xFile = XIF_DOCFILE(xDoc);
    PageTableEntry tempEntry;
    PROLOGUE;

    /* Seek to the entry location */
    XIF_SEEK(xFile, fileOffset);

    /* Copy to a temporary entry (for swapping) */
    memcpy(&tempEntry, pageTableEntry, sizeof(tempEntry));

    /* Swap if necessary */
    SwapLong(xFile, &tempEntry.IFDoffset);

    /* Write this puppy out */
    XIF_WRITE(xFile, &tempEntry.flags, sizeof(tempEntry.flags));
    XIF_WRITE(xFile, &tempEntry.IFDoffset, sizeof(tempEntry.IFDoffset));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rewrites the page table entries, used to accomodate a page insert/delete
 *
 * LongDescription:::
 *  wp_XifRewritetPageTableEntries() starts with a given page, and rewrites
 *  the page table entries. The TIFF primary IFD nextIFD entries are also
 *  checked for consistency, and adjusted if necessary (as is the case AFTER
 *  the insertion or deletion of a page from the document).
 *  <BR>
 *  It is assumed that after the insertion or deletion of a page (or pages),
 *  that the internal linked list of pages is now correct, and that the page
 *  table needs to be adjusted to match the insertion or removal of a link or
 *  links.
 *  <BR>
 *  All shifted data is taken from that of the internal data structures --
 *  the entire page table MUST be read (or written) prior to this function.
 *
 * Input:::
 *  xDoc::			Document handle, required when last page was
 *				just deleted.
 *  xPage::			Page where shift will begin (inclusive). This
 *				is only a hint, the very first page could be
 *				used, but only the affected pages will result
 *				in change.
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifRewritePageTableEntries(
    XDoc *xDoc,
    XPage *xPage
)
{
    Int32 pages;
#if 0
    Int32 startOffset;
    Int32 shiftsLeft;			/* Shifts left to apply */
    Int32 srcEntryOffset, dstEntryOffset;
    XPage *dstPage;
#endif
    Int32 pageOffset;
    UInt8 flags;
    Int32 ifdOffset;
    XImg *xImg;
    PROLOGUE;

    CheckHandleValidity(xDoc);

    /* Allow the case where the page handle is NULL (no pages = last page
     * was just deleted) */
    if (xPage) {
	CheckHandleValidity(xPage);
	pages = xDoc->docInfo.wNPages;
	pageOffset = xPage->pageInfo.dwPageOffset;  /* This must be correct */
    } else {
	pageOffset = 0;

	/* No pages remain: special case, reset the TIFF header to have no first
	 * page */
	XERR_PASS_CLEANUP(wp_XifSetPageIFD(xDoc, pageOffset, 0, 0, NULL));
    }

    /* Compute the last entry positions to be copied (source and destination)
     */
    while (xPage) {

	/* Make sure things are consistent */
	XERR_GEN_CLEANUP(pageOffset != xPage->pageInfo.dwPageOffset,
	    XIF_INTERNAL_ERROR);

	/* Get the page table information */
	XERR_PASS_CLEANUP(wp_XifGetPageIFD(xDoc, pageOffset, &flags,
	    &ifdOffset));

	/* Update the IFD pointers info if not correct */
	xImg = (XImg *)xPage->pageInfo.xPrimImgHandle;
	if (xImg && ifdOffset != xImg->ifdOffset) {

	    /* Write the source information to the new destination position */
	    XERR_PASS_CLEANUP(wp_XifGeneratePageFlags(xPage, &flags));
	    XERR_PASS_CLEANUP(wp_XifSetPageIFD(xDoc, pageOffset, flags,
		xImg->ifdOffset, xPage));
	}
      
	/* Go to the next page */
	xPage = xPage->next;
	pageOffset++;
    }

#if 0


    if (shifts > 0) {	/* Insertion case */

	srcEntryOffset = pages - 1 - shifts;	/* New last page position */
	dstEntryOffset = pages - 1;		/* Last page */
	shiftsLeft = srcEntryOffset - startOffset + 1;

	/* Get the new destination page */
	pageOffset = 0;
	dstPage = xDoc->pages;
	while (pageOffset++ < dstEntryOffset) {
	    dstPage = dstPage->next;
	    XERR_GEN_CLEANUP(dstPage == NULL, XIF_INTERNAL_ERROR);
	}

	/*
         * Regenerate the destination entries, one at a time, and place them in
	 * their new (shifted) positions.
	 */
	while (shiftsLeft > 0) {
	    /* Get the source information */
	    XERR_PASS_CLEANUP(wp_XifGetPageIFD(xDoc, srcEntryOffset, &flags,
		&ifdOffset));

		/* Update the internal structure so that the nextIFD doesn't change,
		 * which would trigger a rewrite of the nextIFD field during the call
		 * to wp_XifSetPageIFD() */

	    /* Write the source information to the new destination position */
	    XERR_PASS_CLEANUP(wp_XifSetPageIFD(xDoc, dstEntryOffset, flags,
		ifdOffset, dstPage));
      
	    /* Go to the next page (previous) */
	    srcEntryOffset--, dstEntryOffset--, shiftsLeft--;
	    dstPage = dstPage->prev;
	    XERR_GEN_CLEANUP(dstPage == NULL, XIF_INTERNAL_ERROR);
	}
    } else {	/* Deletion case */

	srcEntryOffset = startOffset - shifts;		/* New "next" page */
	dstEntryOffset = startOffset;			/* First "clobbered" page */
	shiftsLeft = pages - srcEntryOffset;
	dstPage = xPage;

	/*
	 * Regenerate the destination entries, one at a time, and place them in
	 * their new (shifted) positions.
	 */
	while (shiftsLeft > 0) {
	    /* Get the source information */
	    XERR_PASS_CLEANUP(wp_XifGetPageIFD(xDoc, srcEntryOffset, &flags,
		&ifdOffset));

	    /* Update the internal structure so that the nextIFD doesn't change,
	     * which would trigger a rewrite of the nextIFD field during the call
	     * to wp_XifSetPageIFD() */
	    
	    /* Write the source information to the new destination position */
	    XERR_PASS_CLEANUP(wp_XifSetPageIFD(xDoc, dstEntryOffset, flags,
		ifdOffset, dstPage));
      
	    /* Go to the next page */
	    srcEntryOffset++, dstEntryOffset++, shiftsLeft--;
	    dstPage = dstPage->next;
	    XERR_GEN_CLEANUP(dstPage == NULL, XIF_INTERNAL_ERROR);
	}
    }

    /* If there's nothing that needs shifting, we don't need to continue */
    if (startOffset >= pages || shiftsLeft <= 0) {
	NORMAL_RETURN;
    }

#endif


    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add an annotation to a page
 *   
 * LongDescription:::
 *  wp_XifAddAnnotation() adds annotation data and an updated annotation
 *  offsets list of a XIFF 2.0 file, or the annotation table entry (and table)
 *  of a XIFF 3.0 and later file.
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *  type::			Type of annotation
 *  xOffset::			X offset in page coordinates
 *  yOffset::			Y offset in page coordinates
 *  width::			Width in page coordinates
 *  height::			Height in page coordinates
 *  flags::			Annotation flags (see ANNOT_xxx in xifpags.h)
 *  pData::			Annotation data
 *  bytes::			Number of bytes in annotation data
 *
 * Output:::
 *  pID::			ID assigned to this annotation
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifAddAnnotation(
	      XPage *xPage,
	      UInt16 type,
	      UInt32 xOffset,
	      UInt32 yOffset,
	      UInt32 width,
	      UInt32 height,
	      UInt32 flags,
	      UInt8 *pData,
	      UInt32 bytes,
	      UInt16 *pID
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XImg *primImg;
    UInt16 annotationOffset;
    UInt32 tableOffset;
    UInt32 entryOffset;
    AnnotTable *prevTbl, *tblPtr;
    AnnotTable *newTblPtr = NULL;
    AnnotTableEntry *ep;
    Int32 tblOffset, prevTblOffset;
    UInt32 mallocbytes;
    Int32 fileOffset;
    UInt16 sval; UInt32 lval;
    Int32 annotationFileOffset = 0UL;
    PROLOGUE;

    /* Check parameters / mode */
    CheckHandleValidity(xPage);
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xPage->seqStatus &
	(XPAGE_SEQ_READSTART | XPAGE_SEQ_WRITESTART)) == 0),
	XIF_SEQUENCE);
    XERR_GEN_CLEANUP((xPage->xDoc->readWrite & WRITE_ALLOWED) == 0,
	XIF_INTERNAL_ERROR);

    /* For now, we must have a primary image attached */
    XERR_GEN_CLEANUP(xPage->pageInfo.xPrimImgHandle == 0,
		     XIF_SEQUENCE);

    /* Write the annotation table header if none is attached */
    if (xPage->annotTableHeader.entriesPerTable == 0UL) {
	XERR_GEN_CLEANUP(xPage->annotTableHeader.annotCount > 0,
			 XIF_INTERNAL_ERROR);
	if (xPage->xDoc->docInfo.version < 3) {
	    /* Tables only internal */
	    xPage->xDoc->annotsPerTable =
	      xPage->annotTableHeader.entriesPerTable = 1;
	} else {
	    if (xPage->xDoc->annotsPerTable == 0) {
	      xPage->xDoc->annotsPerTable  = 4;
	    }
	    xPage->annotTableHeader.entriesPerTable =
		xPage->xDoc->annotsPerTable;
	}

	/* Figure out the byte count of a table entry */
	xPage->annotTableHeader.bytesPerEntry =
	    sizeof(ep->id) + sizeof(ep->type) +
	    sizeof(ep->xOffset) + sizeof(ep->yOffset) +
	    sizeof(ep->width) + sizeof(ep->height) +
	    sizeof(ep->flags) + sizeof(ep->bytes) + sizeof(ep->offset);

	XERR_PASS_CLEANUP(wp_XifWriteAnnotTableHeader(xPage));
    }

    /* We can only add the next annotation, not allow a gap in the list */
    annotationOffset = xPage->annotTableHeader.annotCount;
    
    tableOffset = annotationOffset / xPage->annotTableHeader.entriesPerTable;
    entryOffset = annotationOffset % xPage->annotTableHeader.entriesPerTable;

    /* Make sure we've read all the annotation tables if the destination is
     * a read-write file */
    if ((xPage->xDoc->readWrite & READ_ALLOWED) &&
	xPage->annotTableHeader.annotCount > 0) {
	UInt32 tmpAnnotationOffset;
	UInt16 tmpTableOffset;
	AnnotTable *pAnnotTable = NULL;

	tmpAnnotationOffset = xPage->annotTableHeader.annotCount - 1;
	tmpTableOffset = (UInt16)(tmpAnnotationOffset /
	    xPage->annotTableHeader.entriesPerTable);

	XERR_PASS_CLEANUP(wp_XifGetAnnotTable(xPage, tmpTableOffset,
					      &pAnnotTable));
    }

    /* Search for the table itself */
    prevTbl = tblPtr = xPage->firstAnnotTable;
    prevTblOffset = 0UL;
    tblOffset = xPage->annotTableHeader.firstTableOffset;
    while (tblPtr) {
	if (tblPtr->ordinal == tableOffset + 1) {
	    break;
	}
	prevTblOffset = tblOffset;
	tblOffset = tblPtr->nextTableOffset;	
	prevTbl = tblPtr;
	tblPtr = tblPtr->next;
    }

    /* Create a new table if the one desired doesn't exist yet */
    if (tblPtr == NULL) {

	CHECK_MALLOC(newTblPtr = (AnnotTable *)XIF_MALLOC(sizeof(AnnotTable)));
	memset(newTblPtr, 0, sizeof(AnnotTable));
	mallocbytes = sizeof(AnnotTableEntry) * 
	    xPage->annotTableHeader.entriesPerTable;
	CHECK_MALLOC(newTblPtr->entries =
		     (AnnotTableEntry *)XIF_MALLOC(mallocbytes));
	memset(newTblPtr->entries, 0, mallocbytes);
	newTblPtr->ordinal = (UInt16)tableOffset + 1;

	/* Add to the linked list */
	if (prevTbl) {

	    /* Attach to end of linked list */
	    prevTbl->next = newTblPtr;

	} else {

	    /* Begin list here */
	    xPage->firstAnnotTable = newTblPtr;
	    xPage->annotTableHeader.firstTableOffset = tblOffset;
	}

	/* Write the new annotation table */
	if (xPage->xDoc->docInfo.version > 2) {
	    XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &tblOffset));
	    sval = newTblPtr->ordinal;
	    SwapShort(xFile, &sval);
	    XIF_WRITE(xFile, &sval, sizeof(sval));

	    /* Regardless of structure alignment issues, the structure entry
	     * data is at least the size of the file entries, and is zero'd
	     * out. Just write out the number of 0-bytes to fill the file
	     * entries */
	    XIF_WRITE(xFile, newTblPtr->entries,
		xPage->annotTableHeader.entriesPerTable *
		xPage->annotTableHeader.bytesPerEntry);
	    lval = newTblPtr->nextTableOffset;
	    SwapLong(xFile, &lval);
	    XIF_WRITE(xFile, &lval, sizeof(lval));

	    if (prevTbl) {

		/* Look for the annotation table 'nextTableOffset' location
		 * of the table were attaching this one to */
		fileOffset = prevTblOffset +
		    sizeof(xPage->firstAnnotTable->ordinal) +
		    xPage->annotTableHeader.entriesPerTable *
		    xPage->annotTableHeader.bytesPerEntry;

		/* Link the new table to the previous one, in the file */
		XIF_SEEK(xFile, fileOffset);
		lval = tblOffset;
		SwapLong(xFile, &lval);
		XIF_WRITE(xFile, &lval, sizeof(lval));
		prevTbl->nextTableOffset = tblOffset;
	    } else {
		xPage->annotTableHeader.firstTableOffset = tblOffset;
	    }
	} /* if version > 2 */
	tblPtr = newTblPtr;
	newTblPtr = NULL;	/* Avoid freeing if later failure */

	/* Update the header */
	xPage->annotTableHeader.lastOrdinal++;
    } /* Adding a new table */

    /* Prepare to write the data at the end of the file */
    XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &annotationFileOffset));

    /* For XIFF 2.0 files, store the characteristics before the data */
    if (xPage->xDoc->docInfo.version < 3) {

	/* Reference the entry */
	ep = &tblPtr->entries[entryOffset];
	ep->id = (UInt16)annotationOffset;
	ep->offset = annotationFileOffset;

	ep->bytes = bytes;
	lval = bytes + sizeof(ep->type) + sizeof(ep->xOffset) +
	    sizeof(ep->yOffset) + sizeof(ep->width) + sizeof(ep->height);
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	sval = ep->type = type;
	SwapShort(xFile, &lval);
	XIF_WRITE(xFile, &sval, sizeof(sval));
	lval = ep->xOffset = xOffset;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	lval = ep->yOffset = yOffset;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	lval = ep->width = width;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	lval = ep->height = height;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
    } /* version < 3 */

    /* Write the annotation data */
    XIF_WRITE(xFile, pData, bytes);

    /* Write the new annotation offsets list */
    if (xPage->xDoc->docInfo.version < 3) {

	if (xPage->annotTableHeader.annotCount > 0) {

	    /* Place annotation offsets list at the end of the file */
	    XERR_PASS_CLEANUP(wp_FileLongAlign(xFile,
		&xPage->annotTableHeaderOffset));
	    xPage->annotTableHeader.firstTableOffset =
		xPage->annotTableHeaderOffset;

	    /* Loop through each table */
	    tblPtr = xPage->firstAnnotTable;
	    while (tblPtr) {

		/* Loop through each table entry */
		ep = tblPtr->entries;
		for (entryOffset = 0;
		entryOffset < xPage->annotTableHeader.entriesPerTable;
		entryOffset++) {
		    lval = ep->offset;
		    SwapLong(xFile, &lval);
		    XIF_WRITE(xFile, &lval, sizeof(lval));

		    if (ep->id == annotationOffset) {
			break;
		    }
		    ep++;
		}

		if (tblPtr->ordinal == tableOffset + 1) {
		    break;
		}
		tblPtr = tblPtr->next;
	    } /* for each table */
	}

	/* Update the TIFF field */
	primImg = (XImg *)xPage->pageInfo.xPrimImgHandle;
	XERR_GEN_CLEANUP(primImg == NULL, XIF_INTERNAL_ERROR);
	XERR_GEN_CLEANUP(primImg->annotationOffsetsCountOffset == 0UL,
	    XIF_INTERNAL_ERROR);
	XERR_GEN_CLEANUP(primImg->annotationOffsetsDataOffset !=
	    primImg->annotationOffsetsCountOffset + 4, XIF_INTERNAL_ERROR);
	XIF_SEEK(xFile, primImg->annotationOffsetsCountOffset);

	/* (xPage->annotTableHeader.annotCount to be incremented shortly) */
	lval = xPage->annotTableHeader.annotCount + 1;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));	/* Write the new count */
	if (xPage->annotTableHeader.annotCount == 0) {
	    /* Write the annotation offset of the sole annotation */
	    lval = annotationFileOffset;
	} else {
	    lval = xPage->annotTableHeaderOffset;
	}
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));	/* Write the new data offset */

    } else { /* version >= 3 */

	/* Write the new entry in the annotation table */
	XERR_GEN_CLEANUP(tblOffset == 0, XIF_INTERNAL_ERROR);
	fileOffset = tblOffset + sizeof(xPage->firstAnnotTable->ordinal) +
	    entryOffset * xPage->annotTableHeader.bytesPerEntry;
	ep = &tblPtr->entries[entryOffset];
	ep->id = annotationOffset;
	ep->type = (UInt16)type;
	ep->xOffset = xOffset;
	ep->yOffset = yOffset;
	ep->width = width;
	ep->height = height;
	ep->flags = flags;
	ep->bytes = bytes;
	ep->offset = annotationFileOffset;
	XERR_PASS_CLEANUP(wp_XifWriteAnnotTableEntry(xPage, fileOffset, ep));
    }

    /* Increment the total count */
    xPage->annotTableHeader.annotCount++;    
    xPage->pageInfo.wAnnotationCount++;    
    XERR_PASS_CLEANUP(wp_XifWriteAnnotTableHeader(xPage));

    /* Return the annotation ID (offset) */
    *pID = annotationOffset;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newTblPtr) {
	if (newTblPtr->entries) {
	    XIF_FREE(newTblPtr->entries);
	}
	XIF_FREE(newTblPtr);
    }
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write/rewrite the annotation table header
 *   
 *
 * LongDescription:::
 *  wp_XifWriteAnnotTableHeader() writes or rewrites the annotation table header
 *  to take account of an update in the information (usually the number of
 *  annotations).
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteAnnotTableHeader(
    XPage *xPage
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt16 sval; UInt32 lval;
    PROLOGUE;
    
    CheckHandleValidity(xPage);
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);

    /* Check for proper calling sequence */


    /* If version 2, the list of annotation data offsets has already been
     * written */
    if (xPage->xDoc->docInfo.version < 3) {
	NORMAL_RETURN;
    }

    /* If this is the first time it is written, place at the end of file */
    if (xPage->annotTableHeaderOffset == 0UL) {
	XImg *primImg;

	/* Make sure we have the TIFF fields written */
	primImg = (XImg *)xPage->pageInfo.xPrimImgHandle;
	XERR_GEN_CLEANUP(primImg == NULL, XIF_SEQUENCE);

	/* Place the annotation table header at the end of the file */
	XERR_PASS_CLEANUP(wp_FileLongAlign(xFile,
	    &xPage->annotTableHeaderOffset));

	/* Update the TIFF fields */
	if (primImg->fields.AnnotationOffsets == NULL) {
	    CHECK_MALLOC(primImg->fields.AnnotationOffsets =
		XIF_MALLOC(1 * sizeof(*primImg->fields.AnnotationOffsets)));
	    primImg->fields.AnnotationCount = 1;
	}
	primImg->fields.AnnotationOffsets[0] = xPage->annotTableHeaderOffset;
	if (primImg->annotationOffsetsDataOffset != 0) {
	    Int32 lval;

	    /* Actually attach it */
	    XIF_SEEK(xFile, primImg->annotationOffsetsDataOffset);
	    lval = xPage->annotTableHeaderOffset;
	    SwapLong(xFile, &lval);
	    XIF_WRITE(xFile, &lval, sizeof(lval));
	}
    }

    /* If we have written the primary image, attach the annotations */

    /* Seek to the header */
    XIF_SEEK(xFile, xPage->annotTableHeaderOffset);

    /* Write out the header */
    sval = xPage->annotTableHeader.annotCount;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xPage->annotTableHeader.entriesPerTable;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xPage->annotTableHeader.bytesPerEntry;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xPage->annotTableHeader.lastOrdinal;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    lval = xPage->annotTableHeader.firstTableOffset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an annotation table entry.
 *   
 * LongDescription:::
 *  wp_XifWriteAnnotTableEntry() stores a single annotation entry.
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *  fileOffset::		Where to write the table entry
 *  annotTableEntry::		Table entry to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteAnnotTableEntry(
    XPage *xPage,
    Int32 fileOffset,
    AnnotTableEntry *annotTableEntry
)
{
    XFile *xFile = XIF_PAGEFILE(xPage);
    AnnotTableEntry tempEntry;
    Int32 bytesLeft;
    PROLOGUE;

    /* Seek to the entry location */
    XIF_SEEK(xFile, fileOffset);

    /* Copy to a temporary entry (for swapping) */
    memcpy(&tempEntry, annotTableEntry, sizeof(tempEntry));

    /* Swap if necessary */
    SwapShort(xFile, &tempEntry.id);
    SwapShort(xFile, &tempEntry.type);
    SwapLong(xFile, &tempEntry.xOffset);
    SwapLong(xFile, &tempEntry.yOffset);
    SwapLong(xFile, &tempEntry.width);
    SwapLong(xFile, &tempEntry.height);
    SwapLong(xFile, &tempEntry.flags);
    SwapLong(xFile, &tempEntry.bytes);
    SwapLong(xFile, &tempEntry.offset);

    /* Keep track of how many bytes we're allowed to write */
    bytesLeft = xPage->annotTableHeader.bytesPerEntry;

    /* Write this puppy out */
    if (bytesLeft >= sizeof(tempEntry.id)) {
	XIF_WRITE(xFile, &tempEntry.id, sizeof(tempEntry.id));
	bytesLeft -= sizeof(tempEntry.id);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.type)) {
	XIF_WRITE(xFile, &tempEntry.type, sizeof(tempEntry.type));
	bytesLeft -= sizeof(tempEntry.type);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.xOffset)) {
	XIF_WRITE(xFile, &tempEntry.xOffset, sizeof(tempEntry.xOffset));
	bytesLeft -= sizeof(tempEntry.xOffset);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.yOffset)) {
	XIF_WRITE(xFile, &tempEntry.yOffset, sizeof(tempEntry.yOffset));
	bytesLeft -= sizeof(tempEntry.yOffset);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.width)) {
	XIF_WRITE(xFile, &tempEntry.width, sizeof(tempEntry.width));
	bytesLeft -= sizeof(tempEntry.width);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.height)) {
	XIF_WRITE(xFile, &tempEntry.height, sizeof(tempEntry.height));
	bytesLeft -= sizeof(tempEntry.height);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.bytes)) {
	XIF_WRITE(xFile, &tempEntry.bytes, sizeof(tempEntry.bytes));
	bytesLeft -= sizeof(tempEntry.bytes);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.offset)) {
	XIF_WRITE(xFile, &tempEntry.offset, sizeof(tempEntry.offset));
	bytesLeft -= sizeof(tempEntry.offset);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.flags)) {
	XIF_WRITE(xFile, &tempEntry.flags, sizeof(tempEntry.flags));
	bytesLeft -= sizeof(tempEntry.flags);
    } else {
	goto EntryDone;
    }

 EntryDone:
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an annotation from a page
 *   
 * LongDescription:::
 *  wp_XifDeleteAnnotation() deletes an annotation from a page. For 2.0 files,
 *  the annotation offsets list is rewritten, with the deleted annotation
 *  removed from the list. For 3.0 and later versions of XIFF, the table
 *  entries are updated to reflect one less annotation; if appropriate, the
 *  entries are shifted by one, including any shifts from later tables. For
 *  this reason, it might be most efficient to delete the latest entry
 *  first.
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *  offset::			Annotation offset
 *
 * Output:::
 *  *pdwAnnotationCount::	New annotation count
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  IDs returned to any of the annotations on this page will be invalid
 *  after an annotation is deleted. This is due to the fact that the ID
 *  is the offset of the annotation in an absolute list (0, 1, 2, ...),
 *  which may change depending on the position of the deleted annotation.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifDeleteAnnotation(
	      XPage *xPage,
	      UInt16 offset,
	      UInt16 *pdwAnnotationCount
)
{
    Instance *instData = NULL;
    XFile *xFile;
    XImg *xPrimImg;
    UInt16 entryOffset, tableOffset;
    AnnotTable *prevTbl = NULL, *al = NULL, *annotTable;
    AnnotTableEntry *ep, *srcEp, *dstEp, *prevEp, *storeEp = NULL;
    Int32 fileOffset;
    UInt16 annotOffset, annotCount;
    UInt8 *entryData = NULL, *bp;
    UInt32 *lp;
    UInt16 shiftEntryCount;
    UInt16 storeEntryCount;
    UInt32 storeEntryOffset = 0UL;
    UInt16 sval; UInt32 lval;
    UInt32 prevTableFileOffset = 0UL;	/* File offset of previous table */
    UInt32 thisTableFileOffset = 0UL;	/* File offset of this table */
    UInt32 tempDataSize;
    PROLOGUE;

    /* Check parameters / mode */
    CheckHandleValidity(xPage);
    instData = XIF_PAGEINST(xPage);
    xFile = XIF_PAGEFILE(xPage);

    annotCount = xPage->annotTableHeader.annotCount;
    XERR_GEN_CLEANUP(offset >= annotCount, XIF_BADPARAMETER);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xPage->seqStatus & XPAGE_SEQ_GETINFO) == 0),
		     XIF_SEQUENCE);
    XERR_GEN_CLEANUP((xPage->xDoc->readWrite & WRITE_ALLOWED) == 0,
		     XIF_INTERNAL_ERROR);

    /* Make sure we have an accurate accouting of all the annotations */
    XERR_PASS_CLEANUP(wp_XifGetAnnotTable(xPage, (UInt16)(annotCount - 1),
	&annotTable));

    /* For XIFF 2.0 files, simple remove the annotation data offset
     * from the list, and rewrite it in place */
    if (xPage->xDoc->docInfo.version < 3) {
	xPrimImg = (XImg *)xPage->pageInfo.xPrimImgHandle;

	/* There should only be one entry per table, as generated for XIFF 2.0
	 */
	XERR_GEN_CLEANUP(xPage->annotTableHeader.entriesPerTable != 1,
	    XIF_INTERNAL_ERROR);
#if 0
	XERR_GEN_CLEANUP(xPrimImg->fields.AnnotationCount != 
	    xPage->annotTableHeader.annotCount, XIF_INTERNAL_ERROR);
	XERR_GEN_CLEANUP(xPrimImg->fields.AnnotationOffsets == NULL,
	    XIF_INTERNAL_ERROR);
#endif

	if (offset < xPage->annotTableHeader.annotCount - 1) {

	    /* The shift is necessary. Regenerate the annotation offsets
	     * list from the offsets stored internally (linked list), but
	     * only generate the portion of the list that gets shifted.
	     * Swap as necessary. */

	    /* Allocate an array that will contain an entire table of entries
	     */
	    tempDataSize = xPage->annotTableHeader.annotCount - offset - 1;
	    CHECK_MALLOC(entryData = (UInt8 *)XIF_MALLOC(tempDataSize
		* sizeof(long))); 
	    lp = (UInt32 *)entryData;
	    annotOffset = 0;

	    /* Loop through each table */
	    al = xPage->firstAnnotTable;
	    prevTbl = NULL;
	    while (al) {

		ep = al->entries;   /* Only 1 per table for XIFF 2.0 */
		if (annotOffset == offset) {

		    /* Remove the item from the links. We only have to worry
		     * about the table for XIFF 2.0, instead of worrying about
		     * entries shifting from one table to the next. Just
		     * remove the entire table. */
		    if (! prevTbl) {
			xPage->firstAnnotTable = al->next;
		    } else {
			prevTbl->next = al->next;
		    }
		    XIF_FREE(al->entries);
		    XIF_FREE(al);
		    /* Point to previous one so we loop to next one */
		    al = prevTbl;
		} else if (annotOffset > offset) {
		    lval = ep->offset;
		    SwapLong(xFile, &lval);
		    *lp++ = lval;	    /* Store in temporary array */
		    ep->id = annotOffset - 1;   /* Update the ID */
		    al->ordinal = annotOffset;	/* Update the table ordinal */
		}
		annotOffset++;	/* Annotation offset before the removal */

		prevTbl = al;
		al = al->next;
	    } /* for each table */

	    /* Store the new portion of the offsets list */
	    fileOffset = xPage->annotTableHeaderOffset +
		offset * sizeof(long);
	    XIF_SEEK(xFile, fileOffset);
	    XIF_WRITE(xFile, entryData,
		(xPage->annotTableHeader.annotCount - offset - 1) *
		sizeof(long));

	    /* Free the temporary buffer */
	    XIF_FREE(entryData);
	    entryData = NULL;

	} /* if offsets list needs rearranging */

	/* Update the TIFF field (annotation offsets count) */
	XERR_GEN_CLEANUP(xPrimImg == NULL, XIF_INTERNAL_ERROR);
	XERR_GEN_CLEANUP(xPrimImg->annotationOffsetsCountOffset == 0UL,
	    XIF_INTERNAL_ERROR);
	XIF_SEEK(xFile, xPrimImg->annotationOffsetsCountOffset);
	lval = --xPage->annotTableHeader.annotCount;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));	    /* Write the new count */

	/* Note the location of the data remains the same. The array
	 * resides in the same spot, just has an unused "hole" after it. */

    } else { /* if version >= 3 */
	
	/* For XIFF 3.0 and later, the table entries must be modified
	 */

	/* If it's already the last item, no update is necessary, go straight
	 * to updating the count of annotations in the header; otherwise do
	 * the shifting. */
	if (offset < annotCount - 1) {

	    /* If the items do need shifting, we have to make sure we can
	     * handle shifting data from later tables to earlier tables.
	     */

	    /* Allocate an array that will contain an entire table of entries
	     */
	    CHECK_MALLOC(entryData = (UInt8 *)XIF_MALLOC(
		xPage->annotTableHeader.entriesPerTable *
		xPage->annotTableHeader.bytesPerEntry)); 
	    bp = entryData;
	    annotOffset = 0;

	    /* Loop through the structures stored in memory, reconstructing
	     * the pieces of the tables that would have moved in the file.
	     * Store the pieces whenever the "rest of" a table is completed.
	     */

	    /* Find the first affected table */
	    tableOffset = offset / xPage->annotTableHeader.entriesPerTable;
	    al = xPage->firstAnnotTable;
	    prevTableFileOffset = 0;
	    thisTableFileOffset = xPage->annotTableHeader.firstTableOffset;
	    annotOffset = 0;
	    while (al) {
		if (al->ordinal == tableOffset + 1) {
		    break;
		}
		prevTableFileOffset = thisTableFileOffset;
		thisTableFileOffset = al->nextTableOffset;
		al = al->next;
		annotOffset += xPage->annotTableHeader.entriesPerTable;
	    }
	    annotOffset = tableOffset * xPage->annotTableHeader.entriesPerTable;

	    /* Compute the number of items to shift within this table */
	    shiftEntryCount = xPage->annotTableHeader.entriesPerTable - (offset
		- annotOffset) - 1;

	    /* If this is the last table, only shift the rest of them (not the
	     * entire rest within the table */
	    if (tableOffset == annotCount /
		xPage->annotTableHeader.entriesPerTable) {
		shiftEntryCount = annotCount - (offset - annotOffset) - 1;
	    }
	    entryOffset = offset - annotOffset;
	    dstEp = al->entries + entryOffset;
	    srcEp = dstEp + 1;

	   /* Loop through the remaining tables, shifting the entries
	    * in both the internal structures, and the tables in file */
	    prevTbl = NULL;
	    prevEp = NULL;
	    storeEntryCount = 0;
	    while (al || prevTbl) {

		/* Finish up the copy: move first item of this table to
		 * the tail of the previous one, or simply flush the previous
		 * one
		 */
		if (prevTbl != NULL) {

		    /* Copy the first from this table to the tail of the
		     * previous table */
		    if (al) {
			srcEp = al->entries;
			memcpy(dstEp, srcEp, sizeof(AnnotTableEntry));
			storeEntryCount++;
		    }

		    /* Loop through the entries that shifted, and copy them to
		     * a buffer to be stored in the file */
		    while (storeEntryCount--) {
			sval = --storeEp->id;    /* Shift the IDs down one */
			SwapShort(xFile, &sval);
			memcpy(bp, &sval, sizeof(sval)), bp += sizeof(sval);
			sval = storeEp->type;
			SwapShort(xFile, &sval);
			memcpy(bp, &sval, sizeof(sval)), bp += sizeof(sval);
			lval = storeEp->xOffset;
			SwapLong(xFile, &lval);
			memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
			lval = storeEp->yOffset;
			SwapLong(xFile, &lval);
			memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
			lval = storeEp->width;
			SwapLong(xFile, &lval);
			memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
			lval = storeEp->height;
			SwapLong(xFile, &lval);
			memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
			lval = storeEp->flags;
			SwapLong(xFile, &lval);
			memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
			lval = storeEp->bytes;
			SwapLong(xFile, &lval);
			memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
			lval = storeEp->offset;
			SwapLong(xFile, &lval);
			memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);

			storeEp++;
		    }

		    /* Store the data */
		    XERR_GEN_CLEANUP(prevTableFileOffset == 0,
				     XIF_INTERNAL_ERROR);
		    XERR_GEN_CLEANUP((bp - entryData) /
			xPage->annotTableHeader.bytesPerEntry >
			xPage->annotTableHeader.entriesPerTable -
			(storeEntryOffset % xPage->annotTableHeader.entriesPerTable),
			XIF_INTERNAL_ERROR);
		    XIF_SEEK(xFile, prevTableFileOffset + sizeof(short) +
			(storeEntryOffset % xPage->annotTableHeader.entriesPerTable)
			* xPage->annotTableHeader.bytesPerEntry);
		    XIF_WRITE(xFile, entryData, bp - entryData);

		    /* If no more tables follow, break out of loop */
		    if (al == NULL) {
			break;
		    }
		    /* Prepare for next table */
		    shiftEntryCount =
			xPage->annotTableHeader.entriesPerTable - 1;
		    dstEp = al->entries;
		    srcEp = dstEp + 1;
		    bp = entryData;
		}

		/* Make sure we're gunna store no more than we have */
		if (shiftEntryCount > annotCount - annotOffset) {
		    if (annotOffset > annotCount) {
			shiftEntryCount = 0;
		    } else {
			shiftEntryCount = annotCount - annotOffset;
		    }
		}

		/* Shift the entries */
		memcpy(dstEp, srcEp, shiftEntryCount * sizeof(AnnotTableEntry));

		/* Remember the location of the first item that shifted */
		storeEp = dstEp;
		storeEntryCount = shiftEntryCount;
		storeEntryOffset = annotOffset + entryOffset;

		/* Rember where to store the first entry of the next table */
		dstEp += shiftEntryCount;

		/* Remember this as previous table, then go to next */
		prevTableFileOffset = thisTableFileOffset;
		thisTableFileOffset = al->nextTableOffset;
		prevTbl = al;
		al = al->next;
		annotOffset += xPage->annotTableHeader.entriesPerTable;
		entryOffset = 0;
		if (al) {
		    srcEp = al->entries;
		}
	    } /* for each table */

	    /* Free the temporary buffer */
	    XIF_FREE(entryData);
	    entryData = NULL;
	}

	/* Update the total */
	xPage->annotTableHeader.annotCount--;
	xPage->pageInfo.wAnnotationCount--;
	XERR_PASS_CLEANUP(wp_XifWriteAnnotTableHeader(xPage));
    } /* if version >= 3 */

    /* Return the new count */
    *pdwAnnotationCount = xPage->annotTableHeader.annotCount;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (entryData) {
	XIF_FREE(entryData);
	entryData = NULL;
    }
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Update the nextIFD and page table entry necessary to attach a new IFD.
 *
 * LongDescription:::
 *  wp_XifSetPageIFD() makes the updates necessary to place a new IFD in the
 *  primary chains of IFDs for a new page. This includes updating the page
 *  table entry, and adding a new page table if necessary; and also, if
 *  necessary, relinking the "nextIFD" fields of the previous IFD, or updating
 *  the "nextIFD" in the TIFF header.
 *  <BR>
 *  Note that "if necessary" means that the update occurs only if the data
 *  has changed.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  pageOffset::		Offset of this page
 *  flags::			Page table flags
 *  ifdOffset::			Offset of primary IFD
 *  xPage::			Page being added (only used if page is
 *				attached to a new location, which includes
 *				actually being added)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *   If just attempting to add a page table during the initial write, the last
 *   four parameters can all be zero.
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifSetPageIFD(
	      XDoc *xDoc,
	      UInt32 pageOffset,
	      UInt8 flags,
	      UInt32 ifdOffset,
	      XPage *xPage
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt32 tableOffset;
    UInt32 entryOffset;
    PageTable *prevPt, *pt;
    PageTable *newPt = NULL;
    PageTableEntry *tableEntry;
    Int32 ptOffset, prevPtOffset;
    UInt32 mallocbytes;
    Int32 fileOffset;
    UInt16 sval; UInt32 lval;
    UInt8 rewriteHeader = 0;
    UInt8 prevPageFlags;
    Int32 prevPageIFDOffset;
    XImg *xImg = NULL;
    PROLOGUE;
    
    /* Check parameters / mode */
    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xDoc->seqStatus & XDOC_SEQ_WRITESTART) == 0),
		     XIF_SEQUENCE);

    /* We can only add the next page, not allow a gap in the pages */
    XERR_GEN_CLEANUP((pageOffset > xDoc->pageTableHeader.pageCount + 1),
		     XIF_BADPARAMETER);
    XERR_GEN_CLEANUP((xDoc->readWrite & WRITE_ALLOWED) == 0,
		     XIF_INTERNAL_ERROR);

    /* Make sure we've read all the page tables if the destination is
     * a read-write file */
    if ((xDoc->readWrite & READ_ALLOWED) &&
	xDoc->pageTableHeader.pageCount > 0) {
	UInt8 flags;
	Int32 offset;

	XERR_PASS_CLEANUP(wp_XifGetPageIFD(xDoc,
	    xDoc->pageTableHeader.pageCount - 1, &flags, &offset));
    }

    /* Update the appropriate page table */
    if (xDoc->pageTableHeader.entriesPerTable == 0UL) {
	xDoc->pageTableHeader.entriesPerTable = 32;
    }
    XERR_GEN_CLEANUP(xDoc->pageTableHeader.entriesPerTable == 0,
	XIF_INTERNAL_ERROR);
    tableOffset = pageOffset / xDoc->pageTableHeader.entriesPerTable;
    entryOffset = pageOffset % xDoc->pageTableHeader.entriesPerTable;

    /* Search for the table itself */
    prevPt = pt = xDoc->firstPageTable;
    prevPtOffset = 0UL;
    ptOffset = xDoc->pageTableHeader.firstTableOffset;
    while (pt) {
	if (pt->ordinal == tableOffset + 1) {
	    break;
	}
	prevPtOffset = ptOffset;
	ptOffset = pt->nextTableOffset;	
	prevPt = pt;
	pt = pt->next;
    }

    /* Create a new table if the one desired doesn't exist yet */
    if (pt == NULL) {

	CHECK_MALLOC(newPt = (PageTable *)XIF_MALLOC(sizeof(PageTable)));
	memset(newPt, 0, sizeof(PageTable));
	mallocbytes = sizeof(PageTableEntry) * 
	    xDoc->pageTableHeader.entriesPerTable;
	CHECK_MALLOC(newPt->entries =
		     (PageTableEntry *)XIF_MALLOC(mallocbytes));
	memset(newPt->entries, 0, mallocbytes);
	newPt->ordinal = (UInt16)tableOffset + 1;

	/* Write the new page table */
	XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &ptOffset));
	sval = newPt->ordinal;
	SwapShort(xFile, &sval);
	XIF_WRITE(xFile, &sval, sizeof(sval));
	XIF_WRITE(xFile, newPt->entries, xDoc->pageTableHeader.entriesPerTable
	    * (sizeof(tableEntry->flags) + sizeof(tableEntry->IFDoffset)));
	lval = newPt->nextTableOffset;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));
	
	/* Update the status to reflect the first page table written */
	xDoc->seqStatus |= XDOC_SEQ_W_PAGETABLE;

	if (prevPt) {
	    /* Attach to end of linked list */
	    prevPt->next = newPt;

	    /* Look for the page table 'nextTableOffset' location of the table
	     * were attaching this one to */
	    fileOffset = prevPtOffset + sizeof(xDoc->firstPageTable->ordinal) +
		xDoc->pageTableHeader.entriesPerTable *
		(sizeof(tableEntry->flags) + sizeof(tableEntry->IFDoffset));

	    /* Link the new table to the previous one, in the file */
	    XIF_SEEK(xFile, fileOffset);
	    lval = ptOffset;
	    SwapLong(xFile, &lval);
	    XIF_WRITE(xFile, &lval, sizeof(lval));
	    prevPt->nextTableOffset = ptOffset;
	} else {
	    /* Begin list here */
	    xDoc->firstPageTable = newPt;
	    xDoc->pageTableHeader.firstTableOffset = ptOffset;
	}
	pt = newPt;
	newPt = NULL;	/* Avoid freeing if later failure */

	/* Update the header */
	xDoc->pageTableHeader.lastOrdinal++;
	rewriteHeader = 1;
    } /* Adding a new table */

    /* Point to the right entry */
    tableEntry = &pt->entries[entryOffset];

    if (pageOffset == xDoc->pageTableHeader.pageCount && ifdOffset) {
	xDoc->pageTableHeader.pageCount = (UInt16)(pageOffset + 1);
	rewriteHeader = 1;
    }

    if (rewriteHeader) {
	XERR_PASS_CLEANUP(wp_XifWritePageTableHeader(xDoc));
    }

    /* Write the new IFD offset to the page table entry, and if appropriate,
     * to the previous IFD (or TIFF header). */
    if (ifdOffset != pt->entries[entryOffset].IFDoffset) {

	/* Write to the previous IFD */
	if (pageOffset > 0) {

	    /* Get the IFD of the previous page */
	    XERR_PASS_CLEANUP(wp_XifGetPageIFD(xDoc, pageOffset - 1,
		&prevPageFlags, &prevPageIFDOffset));
	} else {
	    prevPageIFDOffset = 0UL;
	}

	if (xPage != NULL) {

	    /* Now we must have the page handle, so we can attach the nextIFD
	     * field of the previous page */
	    CheckHandleValidity(xPage);
	    xImg = (XImg *)xPage->pageInfo.xPrimImgHandle;
	    XERR_PASS_CLEANUP(wp_TifWriteNextIFD(xDoc, xImg, NULL, prevPageIFDOffset,
		    ifdOffset));

	    /* Now write the page table entry */
	    XERR_GEN_CLEANUP(ptOffset == 0, XIF_INTERNAL_ERROR);
	    fileOffset = ptOffset + sizeof(xDoc->firstPageTable->ordinal) +
		entryOffset * (sizeof(tableEntry->flags) +
		sizeof(tableEntry->IFDoffset));
	    tableEntry->flags = flags;
	    tableEntry->IFDoffset = ifdOffset;
	    XERR_PASS_CLEANUP(wp_XifWritePageTableEntry(xDoc, fileOffset,
		tableEntry));
	} else {
	    /* Special case: last page was just deleted, remove the first IFD
	     * offset from the TIFF header */
	    XERR_PASS_CLEANUP(wp_TifWriteNextIFD(xDoc, xImg, NULL, prevPageIFDOffset,
		    ifdOffset));
	}
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newPt) {
	if (newPt->entries) {
	    XIF_FREE(newPt->entries);
	}
	XIF_FREE(newPt);
    }
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generate the compatible page table flags for a page
 *
 * LongDescription:::
 *  wp_XifGeneratePageFlags() generates the page table flags associated with
 *  a page.  Currently, the flags are not used, but maybe some day?
 *
 * Input:::
 *  xPage::			Data structure of valid page
 *
 * Output:::
 *  pFlags::			Pointer to 8-bit flags
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifGeneratePageFlags(
		XPage *xPage,
		UInt8 *pFlags
)
{
    PROLOGUE;
    
    CheckHandleValidity(xPage);

    /* Currently, the flags aren't used (yet) */
    *pFlags = 0;
    
    NORMAL_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write or rewrite the page table header
 *
 * LongDescription:::
 *  wp_XifWritePageTableHeader() writes or rewrites the page table header,
 *  usually to accomodate the addition of a page to update the page count.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWritePageTableHeader(
		XDoc *xDoc
)
{
    XFile *xFile;
    UInt16 sval; UInt32 lval;
    PROLOGUE;
    
    CheckHandleValidity(xDoc);
    
    /* Check for proper calling sequence */

    xFile = XIF_DOCFILE(xDoc);

    /* If this is the first time it is written, place at the end of file */
    if (xDoc->pageTableHeaderOffset == 0UL) {
	DocField docField;

	XERR_PASS_CLEANUP(wp_FileLongAlign(xFile,
	    &xDoc->pageTableHeaderOffset));
	docField.tagID = DOCTAG_PAGETABLE;
	docField.offset = xDoc->pageTableHeaderOffset;
	XIF_SEEK(xFile, xDoc->tagOffset[DOCTAG_PAGETABLE]);
	XERR_GEN_CLEANUP(xDoc->tagOffset[DOCTAG_PAGETABLE] == 0,
	    XIF_INTERNAL_ERROR);
	XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));
    }
    XIF_SEEK(xFile, xDoc->pageTableHeaderOffset);
    sval = xDoc->pageTableHeader.pageCount;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xDoc->pageTableHeader.entriesPerTable;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xDoc->pageTableHeader.lastOrdinal;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    lval = xDoc->pageTableHeader.firstTableOffset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the symbol library file offset for an individual dictionary.
 *   
 * LongDescription:::
 *  wp_XifSetSymlibOffset() sets the file offset for an individual token
 *  dictionary in the list of dictionaries (symbol library).
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  symlibOffset::		Offset of new symlib
 *  bytes::			Number of bytes stored
 *  offset::			Offset in the file of the dictionary
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifSetSymlibOffset(
	      XDoc *xDoc,
	      UInt32 symlibOffset,
	      UInt32 bytes,
	      UInt32 offset
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt32 tableOffset;
    UInt32 entryOffset;
    SymlibTable *prevSl, *sl;
    SymlibTable *newSl = NULL;
    SymlibTableEntry *tableEntry;
    Int32 slOffset, prevSlOffset;
    UInt32 mallocbytes;
    Int32 fileOffset;
    UInt16 sval; UInt32 lval;
    UInt8 rewriteHeader = 0;
    PROLOGUE;
    
    /* Check parameters / mode */
    CheckHandleValidity(xDoc);
    instData = XIF_DOCINST(xDoc);
    xFile = XIF_DOCFILE(xDoc);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((xDoc->seqStatus & XDOC_SEQ_WRITESTART) == 0),
		     XIF_SEQUENCE);

    /* We can only add the next symlib, not allow a gap in the list */
    XERR_GEN_CLEANUP((symlibOffset > xDoc->symlibTableHeader.symlibCount),
		     XIF_BADPARAMETER);
    XERR_GEN_CLEANUP((xDoc->readWrite & WRITE_ALLOWED) == 0,
		     XIF_INTERNAL_ERROR);
    
    /* If we haven't yet read the last symlib table, make sure we read it */
    if ((xDoc->readWrite & READ_ALLOWED) &&
	xDoc->symlibTableHeader.symlibCount > 0) {
	Int32 offset, bytes;

	XERR_PASS_CLEANUP(wp_XifGetSymlibOffset(xDoc,
	    xDoc->symlibTableHeader.symlibCount - 1, &offset, &bytes));
    }

    /* Update the appropriate symlib table */
    if (xDoc->symlibTableHeader.entriesPerTable == 0UL) {
	xDoc->symlibTableHeader.entriesPerTable = 5;
    }
    tableOffset = symlibOffset / xDoc->symlibTableHeader.entriesPerTable;
    entryOffset = symlibOffset % xDoc->symlibTableHeader.entriesPerTable;

    /* Search for the table itself */
    prevSl = sl = xDoc->firstSymlibTable;
    prevSlOffset = 0UL;
    slOffset = xDoc->symlibTableHeader.firstTableOffset;
    while (sl) {
	if (sl->ordinal == tableOffset + 1) {
	    break;
	}
	prevSlOffset = slOffset;
	slOffset = sl->nextTableOffset;	
	prevSl = sl;
	sl = sl->next;
    }

    /* Create a new table if the one desired doesn't exist yet */
    if (sl == NULL) {

	CHECK_MALLOC(newSl = (SymlibTable *)XIF_MALLOC(sizeof(SymlibTable)));
	memset(newSl, 0, sizeof(SymlibTable));
	mallocbytes = sizeof(SymlibTableEntry) * 
	    xDoc->symlibTableHeader.entriesPerTable;
	CHECK_MALLOC(newSl->entries =
		     (SymlibTableEntry *)XIF_MALLOC(mallocbytes));
	memset(newSl->entries, 0, mallocbytes);
	newSl->ordinal = (UInt16)tableOffset + 1;

	/* Write the new symlib table */
	XERR_PASS_CLEANUP(wp_FileLongAlign(xFile, &slOffset));
	sval = newSl->ordinal;
	SwapShort(xFile, &sval);
	XIF_WRITE(xFile, &sval, sizeof(sval));
	XIF_WRITE(xFile, newSl->entries,
	    xDoc->symlibTableHeader.entriesPerTable *
	    (sizeof(tableEntry->bytes) + sizeof(tableEntry->offset)));
	lval = newSl->nextTableOffset;
	SwapLong(xFile, &lval);
	XIF_WRITE(xFile, &lval, sizeof(lval));

	/* Update the status to reflect the page table header written */
	xDoc->seqStatus |= XDOC_SEQ_W_SYMLIBTABLE;

	if (prevSl) {
	    /* Attach to end of linked list */
	    prevSl->next = newSl;

	    /* Look for the symlib table 'nextTableOffset' location of the
	     * table were attaching this one to */
	    fileOffset = prevSlOffset + sizeof(xDoc->firstSymlibTable->ordinal)
		+ xDoc->symlibTableHeader.entriesPerTable *
		(sizeof(tableEntry->bytes) + sizeof(tableEntry->offset));

	    /* Link the new table to the previous one, in the file */
	    XIF_SEEK(xFile, fileOffset);
	    lval = slOffset;
	    SwapLong(xFile, &lval);
	    XIF_WRITE(xFile, &lval, sizeof(lval));
	    prevSl->nextTableOffset = slOffset;
	} else {
	    /* Begin list here */
	    xDoc->firstSymlibTable = newSl;
	    xDoc->symlibTableHeader.firstTableOffset = slOffset;
	}
	sl = newSl;
	newSl = NULL;	/* Avoid freeing if later failure */

	/* Update the header */
	xDoc->symlibTableHeader.lastOrdinal++;
	rewriteHeader = 1;
    } /* Adding a new table */

    /* Point to the entry */
    tableEntry = &sl->entries[entryOffset];

    if (symlibOffset == xDoc->symlibTableHeader.symlibCount) {
	xDoc->symlibTableHeader.symlibCount = (UInt16)(symlibOffset + 1);
	rewriteHeader = 1;
    }

    if (rewriteHeader) {
	XERR_PASS_CLEANUP(wp_XifWriteSymlibTableHeader(xDoc));
    }

    /* Write the new entry in the symlib table */
    if (symlibOffset || offset) {
	XERR_GEN_CLEANUP(slOffset == 0, XIF_INTERNAL_ERROR);
	fileOffset = slOffset + sizeof(xDoc->firstSymlibTable->ordinal) +
	    entryOffset * (sizeof(tableEntry->bytes) +
			   sizeof(tableEntry->offset));
	tableEntry->bytes = bytes;
	tableEntry->offset = offset;
	XERR_PASS_CLEANUP(wp_XifWriteSymlibTableEntry(xDoc, fileOffset,
	    tableEntry));
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newSl) {
	if (newSl->entries) {
	    XIF_FREE(newSl->entries);
	}
	XIF_FREE(newSl);
    }
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a symlib table entry.
 *   
 * LongDescription:::
 *  wp_XifWriteSymlibTableEntry() stores a single symbol library entry for a
 *  token dictionary.
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *  fileOffset::		Where to write the table entry
 *  symlibTableEntry::		Table entry to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteSymlibTableEntry(
    XDoc *xDoc,
    Int32 fileOffset,
    SymlibTableEntry *symlibTableEntry
)
{
    XFile *xFile = XIF_DOCFILE(xDoc);
    SymlibTableEntry tempEntry;
    PROLOGUE;

    /* Seek to the entry location */
    XIF_SEEK(xFile, fileOffset);

    /* Copy to a temporary entry (for swapping) */
    memcpy(&tempEntry, symlibTableEntry, sizeof(tempEntry));

    /* Swap if necessary */
    SwapLong(xFile, &tempEntry.bytes);
    SwapLong(xFile, &tempEntry.offset);

    /* Write this puppy out */
    XIF_WRITE(xFile, &tempEntry.bytes, sizeof(tempEntry.bytes));
    XIF_WRITE(xFile, &tempEntry.offset, sizeof(tempEntry.offset));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write/rewrite the symlib table header
 *   
 * LongDescription:::
 *  wp_XifWriteSymlibTableHeader() writes or rewrites the symbol library header
 *  to take account of an update in the information (usually the token
 *  dictionary count).
 *
 * Input:::
 *  xDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteSymlibTableHeader(
    XDoc *xDoc
)
{
    XFile *xFile;
    UInt16 sval; UInt32 lval;
    PROLOGUE;
    
    CheckHandleValidity(xDoc);
    
    /* Check for proper calling sequence */

    xFile = XIF_DOCFILE(xDoc);

    /* If this is the first time it is written, place at the end of file */
    if (xDoc->symlibTableHeaderOffset == 0UL) {
	DocField docField;

	XERR_PASS_CLEANUP(wp_FileLongAlign(xFile,
	    &xDoc->symlibTableHeaderOffset));
	docField.tagID = DOCTAG_SYMLIBTABLE;
	docField.offset = xDoc->symlibTableHeaderOffset;
	XIF_SEEK(xFile, xDoc->tagOffset[DOCTAG_SYMLIBTABLE]);
	XERR_GEN_CLEANUP(xDoc->tagOffset[DOCTAG_SYMLIBTABLE] == 0,
	    XIF_INTERNAL_ERROR);
	XERR_PASS_CLEANUP(wp_XifWriteExtension(xDoc, &docField));
    }
    XIF_SEEK(xFile, xDoc->symlibTableHeaderOffset);
    sval = xDoc->symlibTableHeader.symlibCount;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xDoc->symlibTableHeader.entriesPerTable;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    sval = xDoc->symlibTableHeader.lastOrdinal;
    SwapShort(xFile, &sval);
    XIF_WRITE(xFile, &sval, sizeof(sval));
    lval = xDoc->symlibTableHeader.firstTableOffset;
    SwapLong(xFile, &lval);
    XIF_WRITE(xFile, &lval, sizeof(lval));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a temporary IFD for filling in later
 *
 * LongDescription:::
 *  wp_XifWriteMockIFD() makes space for a temporary IFD, such that data may be
 *  placed after the IFD. This allows a later "rewrite" of the IFD with the
 *  actual file offsets of the data. This can seem somewhat silly, but is
 *  necessary if the IFD is to preceed the IFD data, which may be a
 *  requirement for internet fax.
 *
 * Input:::
 *  xImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  XIFF
 *
 * Keywords:::
 *  file io, MRC, XIFF
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  work-in-progress
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_XifWriteMockIFD(
		XImg *xImg
)
{
    Instance *instData = NULL;
    XFile *xFile;
    UInt8 *zeros = NULL;
    UInt32 mallocSize;
    PROLOGUE;
    
    CheckHandleValidity(xImg);
    instData = XIF_IMGINST(xImg);
    xFile = XIF_IMGFILE(xImg);

    /* Check for proper calling sequence */

    /* Use a temporary buffer with zeros as a place-holder for the IFD */
    if (xImg->reservedFieldCount == 0) {
	xImg->reservedFieldCount = MAX_EXPECTED_TAG_COUNT;
    }

    mallocSize = 2 + xImg->reservedFieldCount * 12 /* tag size */ + 4;
    CHECK_MALLOC(zeros = XIF_MALLOC(mallocSize));
    memset(zeros, 0, mallocSize);
    XIF_WRITE(xFile, zeros, mallocSize);
    
    /* Fall through to free up any temporary memory */
    
 ERR_CLEANUP:
    
    if (zeros) {
	XIF_FREE(zeros);
    }
    
    NORMAL_RETURN;
}

/*
$Log: xifwxiff.c,v $
Revision 1.1.1.1  1999/09/20 18:19:20  jlin
Imported sources

Revision 1.36  1998/05/07 20:01:41  aber
aber on Thu May 7 13:00:28 PDT 1998

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Just to be real sure, I'm gunna do the same thing for the annotation
tables also; although it looks like it doesn't impact Pagis.

Revision 1.35  1998/05/07 19:36:01  aber
aber on Thu May 7 12:35:34 PDT 1998

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

I found the same problem with the page tables.

Revision 1.34  1998/05/07 17:39:27  aber
aber on Thu May 7 10:38:29 PDT 1998

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

If checking to read all symlib tables on a read-write file, make sure
there even is any. This causes a crash if there isn't any.

Revision 1.33  1998/05/06 23:10:48  aber
aber on Wed May 6 16:09:58 PDT 1998

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Make sure we've read the N-th symbol library table, before we append
the N+1th symbol dictionary.

Revision 1.32  1998/04/30 21:08:59  aber
aber on Thu Apr 30 14:07:19 PDT 1998

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Fixed write to AnnotationOffsets field when storing a single annotation
to a XIF 2.0 file. This hasn't causes a problem before (and won't in the
future) because Pagis has always written 2 annotations. However, this
is the right thing to do.

Revision 1.31  1998/03/28 01:25:22  aber
aber on Fri Mar 27 17:24:49 PST 1998

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

I checked in the wrong file last time. This one errors if adding an
annotation to a page with no primary image.

Revision 1.30  1998/01/22 00:23:55  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Allow a 0-byte author block.

Revision 1.29  1998/01/20 03:44:03  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Prototype update to allow deleting last page of a document.

Revision 1.28  1998/01/20 03:27:44  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Allow deleting the last page of a single-page document.
Update the symbol library location and size during write, so that we
can read the data without having to read the table header.

Revision 1.27  1998/01/16 00:45:00  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Allow page inserts.
New private function used to generate the page table flage (in one
common place). Currently no page table flags are used.

Revision 1.26  1997/12/18 00:25:28  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Remove unreferenced function.

Revision 1.25  1997/11/19 22:30:47  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Used right error code.

Revision 1.24  1997/11/19 21:43:45  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Allow deleting all annotations, but retaining the annotation
table header.

Revision 1.23  1997/11/19 20:25:46  aber
aber on Wed Nov 19 12:25:21 PST 1997

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Allow annotation deletion with either a call to XIF_AddPageStart(),
or XIF_GetPageInfo().

Revision 1.22  1997/11/18 23:48:27  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Add annotation flags entry to annotation functions.
Add bytesPerEntry field to annotation header.
Store only the annotation table entries that match the existing
annotation table header byte count (ignore the rest).

Revision 1.21  1997/11/14 02:09:07  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.20  1997/10/15 21:37:53  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Update SetPageIFD function to handle adding a page table without
actually adding a page.

Revision 1.19  1997/09/19 00:19:15  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Update page info for annotation count when adding or deleting one.

Revision 1.18  1997/08/29 17:28:09  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.17  1997/08/27 21:55:33  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

When attaching another page table or symlib table, keep track of it,
otherwise we start clobbering data elsewhere in the file.

Revision 1.16  1997/08/27 00:18:21  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Correct the symlib count check.

Revision 1.15  1997/08/16 00:16:33  aber
aber on Fri Aug 15 17:16:23 PDT 1997

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Update documentation to match ID removal.

Revision 1.14  1997/08/15 23:48:04  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.13  1997/08/15 00:11:56  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Slight code cleanup.

Revision 1.12  1997/07/14 23:24:20  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Complete annotation deletion.

Revision 1.11  1997/07/10 22:10:43  aber
aber on Thu Jul 10 15:09:25 PDT 1997

xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Remove unused variable. Initial 2 variables to NULL, since the compiler
complains they might be used before set, although it isn't really smart
enough to realize that they can't be used the first time through the
loop. Isn't it just only so clever!

Revision 1.10  1997/07/10 22:03:56  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Initial annotation deletion code. Use 16-bit ID.

Revision 1.9  1997/06/17 22:33:45  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

Fix XIFF 3.0 annotation writes, was clobbering the XIFF extension fields.

Revision 1.7  1997/06/10 01:27:07  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
XIF_ImageReadLayer().

Revision 1.6  1997/05/30 00:51:25  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.5  1997/05/20 22:29:30  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.4  1997/05/20 00:54:39  aber
xifwxiff.c -> /product/ipcore/ipwrap/src/RCS/xifwxiff.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:52  aber
Initial revision

*/
